-- name=common
--------------------------------------------------------------------------------

local _f = {
  LogError   = function () end,
  LogWarning = function () end,
  LogInfo    = function () end
};

local _this = {
  time  = {
    shift = 0
  },
  music = {},
  sound = {}  
};

public.log      = {};
public.fps      = {};
public.mod      = {};
public.delegate = {};
public.time     = {};
public.music    = {};
public.sound    = {};
public.math     = {};
public.table    = {};
public.string   = {};
public.event    = {};

--------------------------------------------------------------------------------
function public.__InitModule ()
  
  private.delegate_events = {};
  
  private.ExtendLuaApi();
  private.LoadStrings();

  SetCursor( CURSOR_DEFAULT );
  
  ObjCreate( "ng_level", "obj" );
  ObjSet(    "ng_level", { pos_z = 13 } );
  ObjAttach( "ng_level", "ng_application" );
  
  InterfaceName = "ng_interface";
  ObjCreate( InterfaceName, "obj" );
  ObjSet(    InterfaceName, { pos_z = 121 } );
  ObjAttach( InterfaceName, "ng_application" );
  
  private.InitAudio();

  ne.msg.Subscribe(
    "Event_notEngine_ScriptHub_Script_Error",
    function ( name, params )

      private.error_info = params;

      public.MessageSend( "Event_Common_ScriptError", params );

      DbgTrace( "[ ERROR ] LUA "..public.table.ToString( params ) );

      --public.ERROR = private.error_info;

    end
  );

--  ne.msg.Subscribe(
--    "Command_EditorShowLogMesage",
--    function ( name, params )
--
--      DbgTrace( "+++ [ "..os.time().." ] "..tostring( params.message ) );
--
--    end
--  );

  if ( debug ) then
    _f.LogError   = private.LogError;
    _f.LogWarning = private.LogWarning;
    _f.LogInfo    = private.LogInfo;
  end;

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function public.Init ()
  
  --
  
end;
--------------------------------------------------------------------------------
function public.Trace ( text, force )

  local all = false;

  if ( all or force ) then

    DbgTrace( text );

  end;

end;
--------------------------------------------------------------------------------
function public.Save ( to_server )

  ApplicationSave( to_server );

end;
--------------------------------------------------------------------------------
function public.Return ( with_save )

  wiz.bfg.ccs.HideButton();

  RestartApplication( with_save );

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.mod.Load ( arg )

  if ( type( arg ) == "string" ) then
  
    local module_name = arg;
    private.ModLoad( module_name );
  
  else
  
    local module_list = arg;
    for _, module_name in pairs( module_list ) do
      private.ModLoad( module_name );
    end;
  
  end;

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.delegate.Set ( module_name, delegates, skip_unsupported )

  if ( not _G[ module_name ].__GetDelegateScheme ) then

    DbgTrace( "[ ERROR ] "..module_name.." doesn't support delegates." );

    return false;

  end;

  local scheme  = _G[ module_name ].__GetDelegateScheme();
  local success = false;

  for function_name, value in pairs( delegates ) do

    if ( scheme[ function_name ] ) then

      if ( type( value ) == "function" ) then

        scheme[ function_name ] = value;
          
        success = true;

      else

        DbgTrace( "[ ERROR ] "..function_name.." is not a function." );

      end;

    elseif ( not skip_unsupported ) then

      DbgTrace( "[ ERROR ] "..module_name.." doesn't support "..function_name );

    end;

  end;

  return success;
    
end;
--------------------------------------------------------------------------------
function public.delegate.Clear ( module_name )

  if ( not _G[ module_name ].__GetDelegateScheme ) then

    return;

  end;

  local scheme = _G[ module_name ].__GetDelegateScheme();

  for function_name, _ in pairs( scheme ) do

    scheme[ function_name ] = function () end;

  end;

end;
--------------------------------------------------------------------------------
function public.event.Set ( module_name, delegate, event )
  
  if ( not private.delegate_events[ module_name ] ) then
    private.delegate_events[ module_name ] = {};
  end;
  
  if ( not private.delegate_events[ module_name ][ delegate ] ) then
    private.delegate_events[ module_name ][ delegate ] = {};
  end;

  private.delegate_events[ module_name ][ delegate ][ event.name ] = event.handler;
  
  local delegates = {};
  
  delegates[ delegate ] = function ( params )
    
    local events = private.delegate_events[ module_name ][ delegate ];
    
    for _, value in pairs( events ) do
      value( params );
    end;
    
  end;
  
  if ( not public.delegate.Set( module_name, delegates ) ) then
    private.delegate_events[ module_name ][ delegate ] = nil;
  end;
  
end;
--------------------------------------------------------------------------------
function public.event.Delete ( module_name, delegate, event )
  
  if ( not private.delegate_events[ module_name ] ) then
    return;
  end;
  
  if ( not private.delegate_events[ module_name ][ delegate ] ) then
    return;
  end;
  
  private.delegate_events[ module_name ][ delegate ][ event.name ] = nil;
  
end;
--------------------------------------------------------------------------------
function public.event.Clear ( module_name, delegate )
  
--  if ( not _G[ module_name ].__GetDelegateScheme ) then
--
--    return;
--
--  end;
--
--  local scheme = _G[ module_name ].__GetDelegateScheme();
--  
--  scheme[ delegate ] = function () end;
--    
--  private.delegate_events[ module_name ][ delegate ] = nil;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.MessageSend ( msg, param )

  return ne.msg.Send( msg, param );

end;
--------------------------------------------------------------------------------
function public.TimeFromSeconds ( seconds )

  seconds = seconds or 0;
  seconds = math.floor( seconds + 0.5 );
  
  local day_sec    = 60 * 60 * 24;
  local hour_sec   = 60 * 60;
  local minute_sec = 60; 
  
  local days = math.floor( seconds / day_sec );
  seconds = seconds - ( days * day_sec );
  
  local hours = math.floor( seconds / hour_sec );
  seconds = seconds - ( hours * hour_sec );
  
  local minutes = math.floor( seconds / minute_sec );
  seconds = seconds - ( minutes * minute_sec );
  
  return {
    d = days,
    h = hours,
    m = minutes,
    s = seconds
  };

end;
--------------------------------------------------------------------------------
function public.IsEntityInList( list, entity, field, force_boolean )

  -- AL> Возвращает порядковый номер entity, если он есть в list.
  -- Возможно сравнение по entity.field == list[N].field, если указан field

  if ( list and #list > 0 ) then

    for l = 1, #list do

      if ( field and ( list[ l ][ field ] == entity[ field ] ) ) or
         ( list[ l ] == entity                                 ) then

        if ( force_boolean ) then

          return true;

        end;

        return l;

      end;

    end;

  end;

  return false;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.log.E ( message, message_compat )

  _f.LogError( message, message_compat );

end;
--------------------------------------------------------------------------------
function public.log.W ( message, message_compat )

  _f.LogWarning( message, message_compat );

end;
--------------------------------------------------------------------------------
function public.log.I ( message, message_compat )

  _f.LogInfo( message, message_compat );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.fps.Get ()
  
  local fps = math.floor( ne.GetFPS() );
  
  if ( private.limited_fps ) then
    fps = math.min( private.limited_fps, fps );
  end;
  
  return fps;
  
end;
--------------------------------------------------------------------------------
function public.fps.SetLimit ( value )
  
  if ( type( value ) == "number" ) then
    private.limited_fps = value;
  else
    Log.w( "SetLimit accepts numbers only" );
  end;
  
end;
--------------------------------------------------------------------------------
function public.fps.UnsetLimit ()
  
  private.limited_fps = nil;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.time.GetCurrent ()

  local  seconds = os.time();
  local  current = seconds + _this.time.shift;
  return current;

end;
--------------------------------------------------------------------------------
function public.time.SetShift ( seconds )
  
  _this.time.shift = _this.time.shift + seconds;

end;
--------------------------------------------------------------------------------
function public.time.ResetShift ()
  
  _this.time.shift = 0;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.sound.Play ( name, loop, fade, alt_name )

  loop = loop or 0;
  fade = fade or 0;

  local is_absolute = string.find( name, "/" );
  local path        = tern( is_absolute )(
    name,
    "assets/content/audio/sound/aud_"..name
  );

  local is_played = wiz.snd.Play( path, "sfx", loop, "", fade );
  
  if ( not is_played ) and
     ( alt_name      ) then

    is_absolute = string.find( alt_name, "/" );
    path        = tern( is_absolute )(
      alt_name,
      "assets/content/audio/sound/aud_"..alt_name
    );
     
    is_played = wiz.snd.Play( path, "sfx", loop, "", fade );
     
  end;
  
  return is_played;

end;
--------------------------------------------------------------------------------
function public.sound.Stop ( name, fade )

  fade = fade or 0;

  local is_absolute = string.find( name, "/" );
  local path        = tern( is_absolute )(
    name,
    "assets/content/audio/sound/aud_"..name
  );

  ne.snd.Stop( path, fade );

end;
--------------------------------------------------------------------------------
function public.sound.IsEnabled ()

  return _this.sound.is_enabled;

end;
--------------------------------------------------------------------------------
function public.sound.Toggle ()

  local is_enabled       = public.sound.IsEnabled();
  local volume           = tern( is_enabled )( 0.0, 0.7 );
  _this.sound.is_enabled = not is_enabled;

  SetSoundVolume( "sfx", volume );

  return _this.sound.is_enabled;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.music.Play ( name, loop, fade )

  loop = loop or 1;
  fade = fade or 2;

  local is_absolute = string.find( name, "/" );
  local path        = tern( is_absolute )(
    name,
    "assets/content/audio/music/aud_"..name.."_mus"
  );
  
  ne.snd.Play( path, "soundtrack", loop, "", fade );

end;
--------------------------------------------------------------------------------
function public.music.Stop ( name, fade )
  
  fade = fade or 0.5;

  local is_absolute = string.find( name, "/" );
  local path        = tern( is_absolute )(
    name,
    "assets/content/audio/music/aud_"..name.."_mus"
  );

  ne.snd.Stop( path, fade );

end;
--------------------------------------------------------------------------------
function public.music.IsEnabled ()

  return _this.music.is_enabled;

end;
--------------------------------------------------------------------------------
function public.music.Toggle ()

  local is_enabled       = public.music.IsEnabled();
  local volume           = tern( is_enabled )( 0.0, 0.5 );
  _this.music.is_enabled = not is_enabled;

  SetSoundVolume( "soundtrack", volume );

  return _this.music.is_enabled;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.table.ToString ( table_in, table_depth, symbol_lf, symbol_ct )

  if not ( table_in and "table" == type( table_in ) ) then

    return "";

  end;

  symbol_lf = symbol_lf or "\n";
  symbol_ct = symbol_ct or "\t";

  table_depth = table_depth or 0;

  local string_depth = "";

  for d = 1, table_depth do

    string_depth = string_depth..symbol_ct;

  end;
      
  local string_out = "";

  if ( 0 == table_depth ) then

    string_out = symbol_lf;

  end;

  for key, value in pairs( table_in ) do

    if ( "string" == type( value ) ) then

      value = "'"..value.."'";

    end;

    string_out = string_out..string_depth..tostring( key ).." = "..tostring( value )..symbol_lf;

    if ( "table" == type( value ) ) then

      string_out = string_out..public.table.ToString( value, 1 + table_depth, symbol_lf, symbol_ct );

    end;

  end;

  return string_out;

end;
--------------------------------------------------------------------------------
function public.table.ToIndexed ( accosiated )

  -- JK: you will lose all keys, they will be replaced by indexes
  
  local indexed = {};

  for key, value in pairs( associated ) do
  
    table.insert( indexed, value );

  end;

  return indexed;

end;
--------------------------------------------------------------------------------
function public.table.ToAssociated ( indexed, field )
  
  if ( type( indexed ) ~= "table" ) then
    public.Trace( "[ ERROR ][ COMMON ] ToAssociated -> Not table");
    return;
  end;
  
  if ( #indexed == 0 ) then
    public.Trace( "[ WARNING ][ COMMON ] ToAssociated -> Already associated or empty");
    return;
  end;
  
  field = field or "name";
  local associated = {};
  
  for i = 1, #indexed do
    local key = indexed[ i ][ field ];
    associated[ key ] = indexed[ i ];
  end;
  
  return associated;
  
end;
--------------------------------------------------------------------------------
-- Returns deep copy of data (recursive function)
-- Able to fix value types as set in params:
--
--  params = { key = "count", type = "number" } will convert strings to numbers
--
--------------------------------------------------------------------------------
function public.table.DeepCopy ( data, params )

  if ( type( data ) ~= "table" ) then

    return data;

  end;

  local copy   = {};
  local backup = ( getmetatable( data ) or {} ).__watch_backup;

  for key, value in pairs( backup or data ) do
        
    local copied_item = public.table.DeepCopy( value, params );
    
    -- JK: fixing values as set in params
    if ( params                 ) and
       ( copied_item ~= "table" ) and
       ( params.key  == key     ) then
       
      if     ( params.type == "number" ) then
        
        copied_item = tonumber( copied_item );
        
      elseif ( params.type == "string" ) then
        
        copied_item = tostring( copied_item );
        
      end;
      
    end;
    
    copy[ key ] = copied_item;
    
  end;

  return copy;

end;
--------------------------------------------------------------------------------
function public.table.Reverse ( table )

  local s, e = 1, #table;

  while ( s < e ) do
    table[ s ], table[ e ] = table[ e ], table[ s ];
    s = s + 1;
    e = e - 1;
  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Timer_Create ( placeholder, timer_name, step, func_exec )
  
  local properties = {
    time    = step,
    playing = true,
    endtrig = func_exec
  };
  
  ui.obj.Create(
    timer_name,
    "timer",
    placeholder,
    properties
  );

end;
--------------------------------------------------------------------------------
function public.Timer_Delete ( timer_name )
  
  ui.obj.Delete( timer_name );
  
end;
--------------------------------------------------------------------------------
function public.Timer_Start ( timer_name, step, func_exec, multiset )

  local properties =
  {
    time    = step, 
    playing = true, 
    endtrig = func_exec
  };

  ui.obj.Set( timer_name, properties, multiset );

end;
--------------------------------------------------------------------------------
function public.Timer_Stop ( timer_name, multiset )

  local properties =
  {
    time    = 0,
    playing = false,
    endtrig = function () end
  };

  ui.obj.Set( timer_name, properties, multiset );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Run callback funcs sequence
--   callback_list : list, contain pointers to callback funcs, that accept ONE
--                   argument - pointer to callback func
-- Usage
--   common.RunCallbackList( { model.RequestData, view.Show, interface.Show } )
--------------------------------------------------------------------------------
function public.RunCallbackList ( callback_list )
  
  local callback_count = #callback_list;
  
  if     ( callback_count == 0 ) then
      
    public.Trace( "[ ERROR ] Call empty callback list." );
    return;
      
  elseif ( callback_count == 1 ) then
      
    local last_callback = table.remove( callback_list, 1 );
    last_callback();
      
  else
      
    local first_callback = table.remove( callback_list, 1 );
    
    local function run_other_callbacks ()
      public.RunCallbackList( callback_list );
    end;
    
    first_callback( run_other_callbacks );
      
  end;

end;

--------------------------------------------------------------------------------
-- Process array data without EventLoop locking
--   array           : processing data array
--   element_handler : array element handler, accept that as argument
--   func_after      : after processing callback
--   step            : time step
-- NOTE: async function, control will be passed to next instruction after call
--------------------------------------------------------------------------------
function public.ProcessArrayByTimer (
    array,
    element_handler,
    func_after,
    at_once,
    step
  )
  
  if ( private.array_processing ) then
    DbgTrace( "[ ERROR ] common.ProcessArrayByTimer is busy now!" );
    return;
  end;

  if ( type( array ) ~= "table" ) then
    if ( func_after ) then func_after(); end;
    return;
  end;

  at_once = at_once or 1;
  step    = step or 0.01;
  
  local timer_name = "tmr_common_array_processor";
  local total      = #array;
 
  local function func_iteration ( current )

    public.Timer_Start(
      timer_name,
      step,
      function ()
    
        public.Timer_Stop( timer_name );

        for i = 1, at_once do

          if ( array[ current ] ) then

            local progress = public.math.Round( current / total, 2 );

            element_handler( array[ current ], progress, current );

            current = current + 1;

          else

            break;

          end;

        end;

        if ( array[ current ] ) then
            
          func_iteration( current );
          
        else
          
          ObjDetach( timer_name );
          private.array_processing = false;
          
          if ( func_after ) then func_after(); end;

        end;
      
      end
    );
    
  end;

  if ( total > 0 ) then

    ObjAttach( timer_name, "ng_application" );
    private.array_processing = true;
    func_iteration( 1 );

  else

    if ( func_after ) then func_after(); end;

  end;
  
end;

--------------------------------------------------------------------------------
-- Defer callback call to specified time
--   action_name : name of defer action, must be unique
--   defer_time  : defer time, seconds
--   callback    : deferred function callback
-- NOTE: async function, control will be passed to next instruction after call
--------------------------------------------------------------------------------
function public.Defer ( action_name, defer_time, callback )
  
  if ( type( callback ) ~= "function" ) then return; end;
  
  local timer_name = "tmr_common_defer_" .. action_name;
  
  if ( not ne.obj.Get( timer_name ) ) then
    ne.obj.Create( timer_name, "timer" );
  end;
  
  ne.obj.Attach( timer_name, "ng_application" );
  
  local function func_deferred ()
    callback();
    ne.obj.Delete( timer_name );
  end;
  
  ne.obj.Set( timer_name,
  { 
    time    = defer_time,
    playing = true,
    endtrig = func_deferred
  } );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
-- Return value rounded to nearest number
--  value      : rounding numeric value
--  digit_count: count of decimal digits after comma (optional, default = 0)
--------------------------------------------------------------------------------
function public.math.Round( value, digit_count )
  
  digit_count = digit_count or 0;
  
  local precision = 10^digit_count;
  
  return math.floor( ( value * precision ) + 0.5 ) / precision;
  
end;
--------------------------------------------------------------------------------
function public.math.NumberToDigits( number )
  
  local result = {};
  
  while ( number > 0 ) do
  
    table.insert( result, 1, math.fmod(number, 10) );
    number = math.modf(number / 10);
  
  end;

  return result;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
-- Return array of substrings splitted by separator
--------------------------------------------------------------------------------
function public.string.Split ( string, separator )
  
  local split = {};
  local start = 0;

  if ( #string == 1 ) then return { string }; end;
   
  while true do
    
    local pos = string.find( string, separator, start, true );
    
    if ( pos ~= nil ) then
    
      table.insert( split, string.sub( string, start, pos - 1 ) ); 
      
      start = pos + 1;
      
    else
      
      table.insert( split, string.sub( string, start ) );
      
      break;
      
    end;
    
  end
  
  return split;
  
end;

function public.string.FormatNameStage ( string )
  
  local format_string;
  
  if ( string:find( "_daily" ) ) then
    
    format_string = StringGet( "str_stage_name_dailyquest" );
    
  elseif ( string:find( "_minievent" ) ) then
    
    format_string = StringGet( "str_stage_name_minievent" );
    
  elseif ( string:find( "_bonus" ) ) then
    
    format_string = StringGet( "str_stage_name_"..string );
    
  else
    local number = ui.obj.GetNamePostfix( string );
    local format_number = math.fmod( number, 110 );
    if ( format_number == 0 ) then format_number = 110; end;
    format_string = StringGet( "str_stage_name_saga", format_number );
    
  end;
  
  return format_string;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Trace function and parent function info where DbgCallpoint called 
--------------------------------------------------------------------------------
function public.DbgCallpoint ( message, stack_shift )
  
  if ( debug ) then
  
    stack_shift = stack_shift or 0;
    
    local info    = debug.getinfo( 2 + stack_shift, "Sln" ) or {};
    local info_up = debug.getinfo( 3 + stack_shift, "Sln" ) or {};
    
    local trace = string.format(
      "%s:%s:%s -> %s:%s:%s",
      info_up.source, info_up.currentline, info_up.name,
      info.source,    info.currentline,    info.name
    );

    if ( message ) then
      trace = trace .. string.format( ": %s", message );
    end;
    
    DbgTrace( trace );
    
  else
    
    DbgTrace( "[ WARNING ] common.DbgCallpoint: debug module not loaded" );
    
    if ( message ) then
      DbgTrace( "common.DbgCallpoint: " .. tostring( message )  );
    end;
    
  end;
  
end;

--------------------------------------------------------------------------------
-- Return watching copy of table
--------------------------------------------------------------------------------
function public.DbgWatchTable( tab, name, field, writeonly )

  name = name or "table";

  if ( type(tab) == "table" ) then
  
    -- Backup table
    local backup = {};
    for k, v in pairs( tab ) do
      backup[ k ] = v;
      tab[ k ] = nil;
    end;
    
    local mt = {};
    mt.__watch_backup = backup;
    mt.__index = function ( t, k )
    
      local value = backup[ k ];
    
      if ( field == nil or field == k and not writeonly ) then
      
        local msg = ("INFO: READ %s.%s = %s"):format(name, k, value);
        public.DbgCallpoint( msg, 1 );
      
      end;
    
      return value;
    
    end;
    
    mt.__newindex = function ( t, k, v )
    
      if ( field == nil or field == k ) then
      
        local msg = ("WARNING: WRITE %s.%s = %s"):format(name, k, v);
        public.DbgCallpoint( msg, 1 );
      
      end;
      
      backup[ k ] = v;
    
    end;
    
    mt.__len = function ()
    
      return #backup;
    
    end;
    
    setmetatable( tab, mt );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.LogError ( message, message_compat )

  local tag = private.DetectTag();
  message   = tostring( message or message_compat );
  DbgTrace( "Error/"..tag..": "..message );

end;
--------------------------------------------------------------------------------
function private.LogWarning ( message, message_compat )

  local tag = private.DetectTag();
  message   = tostring( message or message_compat );
  DbgTrace( "Warning/"..tag..": "..message );

end;
--------------------------------------------------------------------------------
function private.LogInfo ( message, message_compat )

  local tag = private.DetectTag();
  message   = tostring( message or message_compat );
  DbgTrace( "Info/"..tag..": "..message );

end;
--------------------------------------------------------------------------------
function private.DetectTag ()

  local info_up = debug.getinfo( 4, "Sln" ) or {};
  return info_up.source;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.ModLoad ( module_name )

  ne.mod.LoadLazily( "assets/content/scripts/mod_"..module_name );

end;
--------------------------------------------------------------------------------
function private.InitAudio ()

  local func_InitChannel = function ( channel, volume ) 
    
    local current_volume = GetSoundVolume( channel );
    
    if ( current_volume > 0 ) then
      
      current_volume = volume or current_volume;
      
    end;
    
    SetSoundVolume( channel, current_volume );

  end; 

  func_InitChannel( "sfx" );
  func_InitChannel( "soundtrack", 0.5 );
  
  _this.sound.is_enabled = ( GetSoundVolume( "sfx" ) > 0 );
  _this.music.is_enabled = ( GetSoundVolume( "soundtrack" ) > 0 );

end;
--------------------------------------------------------------------------------
function private.LoadStrings ()
  
  ne.strings.Load( "assets/content/strings/strings_common" );
  
end;
--------------------------------------------------------------------------------
function private.ExtendLuaApi ()

  -- Extend default Lua API
  private.Extend( { "ne.dbg.Trace", "DbgTrace" }, 
    function ( func_trace, ... )
    
      local args = { ... };
      for _, arg in pairs( args ) do
        local argtype = type( arg );
        if     ( argtype == "table"  ) then arg = public.table.ToString( arg );
        elseif ( argtype ~= "string" ) then arg = tostring( arg );
        end;
        func_trace( arg );
      end;
    
    end
  );
  
  private.Extend( { "ne.obj.Create", "ObjCreate" },
    function ( func_create, name, obj_type, parent, params )
  
        func_create( name, obj_type );
        if ( params ) then ObjSet( name, params );    end;
        if ( parent ) then ObjAttach( name, parent ); end;
        
    end
  );
  
  private.Extend( { "ne.obj.anim.Play", "AnimPlay" },
    function ( func_animplay, object, animfunc, func_end )
  
      func_end = func_end or "";
      func_animplay( object, animfunc, func_end );
        
    end
  );

  -- Platform api
  local platform = GetPlatform();
  
  IsPC      = function () return ( platform == "pc"       ); end;
  IsIOS     = function () return ( platform == "ios"      ); end;
  IsAndroid = function () return ( platform == "android"  ); end;
  IsMobile  = function () return ( IsIOS() or IsAndroid() ); end;
  IsGoogle  = function () return ( IsAndroid() and wiz.Android.GetPlatform() == "google" ); end;
  IsAmazon  = function () return ( IsAndroid() and wiz.Android.GetPlatform() == "amazon" ); end;
  
    -- Ternary method: local result = tern( a == b )( "same", "not same" );
  tern = function ( condition )
  
    return function ( a, b )
      if ( condition ) then return a; end;
      return b;
    end;
    
  end;

  -- Log method
  log = {
    E = public.log.E,
    W = public.log.W,
    I = public.log.I
  };

  Log = {
    e = public.log.E,
    w = public.log.W,
    i = public.log.I
  };

end;
--------------------------------------------------------------------------------
function private.Extend ( names, func_replace )
  
  for _, name in pairs( names ) do

    local parent = _G;
    local tokens = public.string.Split( name, '.' );
    
    for i, token in pairs( tokens ) do
      if ( i == #tokens ) then
      
        local func_orig = parent[ token ];
        parent[ "__Old"..token ] = func_orig;
      
        parent[ token ] = function ( ... )
          return func_replace( func_orig, ... );
        end;
        
      else
      
        parent = parent[ token ];
        
      end;
    end;
  
  end;

end;
--------------------------------------------------------------------------------