-- name=int_dialog_carrier_main
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Build             = function () end,
  Close             = function () end,
  OnClose           = function () end,
  OnPostClose       = function () end,
  GetTab1Data       = function () end,
  OnShow            = function () end,
  Upgrade           = function () end,
  Send              = function () end,
  Speedup           = function () end,
  SpeedupGetFrame   = function () end,
  SpeedupGetTimeSec = function () end,
  SpeedupGetTimeStr = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()
  
  private.Init();
  private.LoadStrings();
  private.Adapt();
  private.CreateButtons();

end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function public.Show ( tab_number, tab1_variant )

  if ( private.active ) then return; end;

  private.active = true;

  private.ShowTab( tab_number, tab1_variant );

  ui.window.Open( "carrier_main", _f.OnShow );

end;
--------------------------------------------------------------------------------
function public.Hide ()

  if ( not private.active ) then return; end;

  private.active = false;

  ui.window.Close(
    "carrier_main",
    function ()

      private.Clean();
      _f.OnClose();
      _f.OnPostClose();

      _f.OnPostClose = function () end;

    end
  );

end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function public.Speedup_OnAccept ( func_end )

  common.Timer_Stop( private.this.timer );

  ui.obj.Set(
    private.this.send_progress,
    {
      frame = 120
    }
  );

  ui.anim.Play( private.this.skip_wrapper, "btn_hide", func_end );

end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function private.Init ()
  
  private.app_o = ui.screen.Orientation();
  
  private.this = {
    dialog = "obj_int_dialog_carrier_main",
    window = "anm_int_dialog_carrier_main_window",
    button = {
      close   = "btn_int_dialog_carrier_main_close",
      build   = "btn_int_dialog_carrier_main_build",
      upgrade = "btn_int_dialog_carrier_main_add",
      send    = "btn_int_dialog_carrier_main_send",
      skip    = "btn_int_dialog_carrier_main_speedup"
    },
    timer         = "tmr_int_dialog_carrier_main",    
    carrier       = "anm_int_dialog_carrier_main_carrier",
    upgrade_bar   = "anm_int_dialog_carrier_main_upgrade_scale",
    selector      = "anm_int_dialog_carrier_main_resources",
    resource1     = "anm_int_dialog_carrier_main_resource_1",
    resource2     = "anm_int_dialog_carrier_main_resource_2",
    resource3     = "anm_int_dialog_carrier_main_resource_3",
    send_wrapper  = "anm_int_dialog_carrier_main_btn_send",
    skip_wrapper  = "anm_int_dialog_carrier_main_btn_speedup",
    send_progress = "anm_int_dialog_carrier_main_send_progressbar"
  };
  
end;
--------------------------------------------------------------------------------
function private.LoadStrings ()
  
  local str_prefix = "str_int_dialog_carrier_main_";
  
  private.strings = {
    tab_1_1_title = ne.strings.Get( str_prefix.."title_build" ),
    tab_1_2_title = ne.strings.Get( str_prefix.."title_send" ),
    tab_1_3_title = ne.strings.Get( str_prefix.."title_speedup" ),
    tab_1_4_title = ne.strings.Get( str_prefix.."title_blocked" ),
    
    tab_1_1_desc = ne.strings.Get( str_prefix.."description_build" ),
    tab_1_2_desc = ne.strings.Get( str_prefix.."description_send" ),
    tab_1_3_desc = ne.strings.Get( str_prefix.."description_speedup" ),
    tab_1_4_desc = ne.strings.Get( str_prefix.."description_blocked" ),
    suspended    = ne.strings.Get( str_prefix.."description_suspended" ),

    max_upgrade = ne.strings.Get( str_prefix.."description_max_upgrade" ),

    select_resources = ne.strings.Get( str_prefix.."select_resources" ),
    no_resource      = ne.strings.Get( str_prefix.."no_resource" ),
    capacity         = ne.strings.Get( str_prefix.."capacity" ),

    eta  = ne.strings.Get( str_prefix.."eta" ),
    from = ne.strings.Get( str_prefix.."from" ),
    to   = ne.strings.Get( str_prefix.."to" ),

    have = ne.strings.Get( "str_int_dialog_carrier_send_you_have" )
  };
  
end;
--------------------------------------------------------------------------------
function private.Adapt ()

  local multiset = {};
  
  ui.obj.ToCenter( private.this.dialog, 0, 0, multiset );
  
  private.window_area = {
    portrait  = { -311, -420, 625, 898 },
    landscape = { -249, -330, 501, 712 }
  };
  
  ui.window.SetArea(
    private.this.dialog,
    private.window_area[ private.app_o ],
    multiset
  );
  
  ui.obj.Set(
    private.this.window,
    {
      tt_select = private.strings.select_resources
    },
    multiset
  );

  ui.obj.Set(
    private.this.selector,
    {
      tt_warning = private.strings.no_resource
    },
    multiset
  );
  
  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
function private.CreateButtons ()
  
  ui.button.Create(
    private.this.button.close,
    {
      design = "close"
    },
    function() _f.Close(); end
  );

  ui.button.Create(
    private.this.button.build,
    {
      title  = "str_int_dialog_carrier_main_button_build",
      design = "medium",
      color  = "green",
      price  = true
    },
    function() _f.Build(); end
  );
  
  ui.button.Create(
    private.this.button.upgrade,
    {
      design = "add"
    },
    function() _f.Upgrade(); end
  );
  
  ui.button.Create(
    private.this.button.send,
    {
      title  = "str_int_dialog_carrier_main_button_send",
      design = "medium",
      color  = "green"
    },
    function() 

      if ( _f.Send( private.selected_resource ) ) then

        _f.Close();

      end;

    end
  );

  ui.button.Create(
    private.this.button.skip,
    {
      title  = "str_int_dialog_carrier_main_button_skip",
      design = "medium",
      color  = "green",
      price  = true
    },
    function() _f.Speedup(); end
  );
  
  for r = 1, 3 do
    
    ui.control.EventsSet(
      private.this[ "resource"..r ],
      {
        func_on_click = function ()
          
          private.Send_SelectResource( r );
          
        end
      }
    );
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.Clean ()

  private.current_tab_number   = nil;
  private.current_tab1_variant = nil;
  private.current_tab_name     = nil;
  private.selected_resource    = nil;
  private.send_enabled         = nil;
  
end;
--------------------------------------------------------------------------------
function private.ShowTab ( tab_number, tab1_variant )

  if ( private.current_tab_number   == tab_number   ) and
     ( private.current_tab1_variant == tab1_variant ) then
    
    return;
    
  end;

  private.current_tab_number = tab_number;
  
  if ( tab1_variant ) then
  
    private.current_tab1_variant = tab1_variant;
    
  else
  
    tab1_variant = private.current_tab1_variant or 1;
    
  end;
  
  local tab_name = "tab_"..tab_number.."_"..tab1_variant;

  local multiset = {};
  
  local has_carrier = {
    tab_1_1 = true,
    tab_1_2 = false,
    tab_1_3 = true,
    tab_1_4 = true
  };
    
  if ( has_carrier[ tab_name ] ) then
    
    ui.obj.Enable( private.this.carrier, multiset );
    
  else
    
    ui.obj.Disable( private.this.carrier, multiset );
    
  end;

  private.current_tab_name = tab_name;
  
  local tab_data = _f[ "GetTab"..tab_number.."Data" ]();
  
  local func_SetupTab = {
    tab_1_1 = private.Build_SetupTab,
    tab_1_2 = private.Send_SetupTab,
    tab_1_3 = private.Speedup_SetupTab,
    tab_1_4 = private.Blocked_SetupTab,
  };
  
  func_SetupTab[ tab_name ]( tab_data, multiset );

  local description = private.strings[ tab_name.."_desc" ] or "";

  if ( tab_name == "tab_1_2" ) then

    if ( tab_data.carrier_state == 6 ) then

      description = private.strings.max_upgrade;

    end;

  end;

  if ( tab_name == "tab_1_4" ) then
    
    if ( tab_data.is_suspended ) then

      description = private.strings.suspended;

    end;

  end;

  ui.obj.Set(
    private.this.window,
    {
      animfunc = tab_name.."_"..private.app_o,
      tt_title = private.strings[ tab_name.."_title" ],
      tt_desc  = description
    },
    multiset
  );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function private.Build_SetupTab ( tab_data, multiset )
  
  ui.obj.Set(
    private.this.carrier,
    {
      animfunc = "state_1",
      playing  = true
    },
    multiset
  );

  ui.button.SetPrice(
    private.this.button.build,
    tab_data.price,
    true,
    multiset
  );
  
end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function private.Blocked_SetupTab ( tab_data, multiset )

  local carrier_state = tab_data.carrier_state or 1;

  ui.obj.Set(
    private.this.carrier,
    {
      animfunc = "state_"..carrier_state,
      playing  = true
    },
    multiset
  );
  
end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function private.Send_SetupTab ( tab_data, multiset )
  
  private.resources     = tab_data.resources;
  local resources_count = #private.resources;
  
  ui.obj.Set(
    private.this.window,
    {
      tt_capacity = private.strings.capacity.." "..tostring( tab_data.capacity )
    },
    multiset
  );
  
  ui.obj.Set(
    private.this.selector,
    {
      animfunc = "fireflies_"..resources_count
    },
    multiset
  );
  
  if ( resources_count > 0 ) then
  
    for r = 1, resources_count do
    
      ui.obj.Set(
        "spr_int_dialog_carrier_main_resource_"..r,
        {
          animfunc = "sleep_"..private.resources[ r ],
          playing  = true
        },
        multiset
      );
      
    end;

  end;
  
  ui.obj.Set(
    private.this.upgrade_bar,
    {
      animfunc = "scale_upgrade_"..tab_data.carrier_state,
      frame    = 60,
      playing  = false
    },
    multiset
  );

  private.Send_ResetResource( multiset ) ;

  ui.obj.Disable( private.this.send_wrapper, false, multiset );
    
end;
--------------------------------------------------------------------------------
function private.Send_SelectResource ( resource_index )
  
  local multiset      = {};
  local resource_name = private.resources[ resource_index ];
  local condition     = ( resource_name ~= private.selected_resource );
  
  private.selected_resource = tern( condition )( resource_name, nil );
  
  --DbgTrace( private.selected_resource.." is selected." );
  
  for r = 1, #private.resources do
  
    local anm_name = private.this[ "resource"..r ];
    
    if ( r == resource_index       ) and
       ( private.selected_resource ) then
      
      ui.anim.Play(
        anm_name,
        "firefly_action",
        function ()
      
           ui.anim.Play( anm_name, "firefly_select" );
          
        end
      );
        
    else
      
      ui.obj.Set(
        anm_name,
        {
          animfunc = "firefly_stat"
        },
        multiset
      );
      
    end;
  
  end;
  
  private.Send_SwitchSendButton( multiset );
  
  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
function private.Send_ResetResource ( multiset )
  
  local multiset = {};
  
  for r = 1, 3 do    
  
    local anm_name = private.this[ "resource"..r ];
    
    ui.obj.Set(
      anm_name,
      {
        animfunc = "firefly_stat"
      },
      multiset
    );    
  
  end;
    
  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
function private.Send_SwitchSendButton ( multiset )
  
  if ( private.selected_resource ) then
  
    if ( private.send_enabled ) then
      
      return;
      
    end;
    
    ui.obj.Enable( private.this.send_wrapper, true, multiset );
    
    ui.anim.Play(
      private.this.send_wrapper,
      "btn_show",
      function ()
       
        ui.input.On( private.this.send_wrapper );
        
      end
    );
  
    private.send_enabled = true;
    
  else
    
    if ( not private.send_enabled ) then
      
      return;
      
    end;
    
    ui.input.Off( private.this.send_wrapper, multiset );
    
    ui.anim.Play(
      private.this.send_wrapper,
      "btn_hide",
      function ()
        
        ui.obj.Disable( private.this.send_wrapper );
        
      end
    );
    
    private.send_enabled = nil;
    
  end;
  
end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function private.Speedup_SetupTab ( tab_data, multiset )

  local prefix           = "str_int_dialog_worldmap_";
  local origin_name      = ne.strings.Get( prefix..tab_data.origin );
  local destination_name = ne.strings.Get( prefix..tab_data.destination );

  ui.obj.Set(
    private.this.window,
    {
      tt_res_count  = tab_data.resource_count,
      tt_eta        = private.strings.eta,
      tt_from       = private.strings.from,
      tt_to         = private.strings.to,
      tt_eta_value  = tab_data.time,
      tt_from_value = origin_name,
      tt_to_value   = destination_name
    },
    multiset
  );

  ui.obj.Set(
    private.this.carrier,
    {
      animfunc = "state_"..tab_data.carrier_state,
      playing  = true
    },
    multiset
  );

  ui.obj.Set(
    "spr_int_dialog_carrier_main_send_resource",
    {
      res = ui.obj.Res( "resource", "icon_"..tab_data.resource_name )
    },
    multiset
  );

  ui.obj.Set(
    private.this.send_progress,
    {
      frame = tab_data.frame
    },
    multiset
  );

  ui.obj.Set(
    private.this.skip_wrapper,
    {
      animfunc = "btn_show",
      frame    = 24,
      playing  = false
    },
    multiset
  );

  ui.button.SetPrice(
    private.this.button.skip,
    tab_data.skip_price,
    multiset
  );
  
  private.Speedup_OnTimer();

end;
--------------------------------------------------------------------------------
function private.Speedup_OnTimer ()

  local time = _f.SpeedupGetTimeSec();

  if ( time < 1 ) then

    common.Timer_Stop( private.this.timer );

    return;

  end;

  local multiset = {};

  ui.obj.Set(
    private.this.window,
    {
      tt_eta_value = _f.SpeedupGetTimeStr()
    },
    multiset
  );

  ui.obj.Set(
    private.this.send_progress,
    {
      frame = _f.SpeedupGetFrame()
    },
    multiset
  );
  
  ui.timer.Show(
    "int_dialog_carrier_main",
    time,
    private.Speedup_OnTimer,
    {
      anm_timer = private.this.window,
      text_tag  = "tt_eta_value"
    }
  );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------