-- name=int_dialog_carrier_send
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Select  = function () end,
  Close   = function () end,
  OnOpen  = function () end,
  OnClose = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()
  
  private.app_o = ui.screen.Orientation();
  
  private.this = {
    dialog = "obj_int_dialog_carrier_send",
    window = "anm_int_dialog_carrier_send_window",
    button = {
      close = "btn_int_dialog_carrier_send_close"
    },
    scroller     = "scc_int_dialog_carrier_send_slots",
    hub_scroller = "scc_int_dialog_carrier_send"
  };
  
  private.window_area = {
    portrait  = { -311, -320, 625, 700 },
    landscape = { -311, -320, 625, 700 }
  };

  private.cell_size = {
    portrait  = 186,
    landscape = 185
  };
  
  private.strings = {
    title = ne.strings.Get( "str_int_dialog_carrier_send_title" ),
    send  = ne.strings.Get( "str_int_dialog_carrier_send_send" ),
    have  = ne.strings.Get( "str_int_dialog_carrier_send_you_have" ),
    eta   = ne.strings.Get( "str_int_dialog_carrier_send_eta" ),
    description = ne.strings.Get( "str_int_dialog_carrier_send_description" )
  };
  
  private.Adapt();
  private.CreateButtons();

end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function public.Init ( delivery_time )

  private.delivery_time = delivery_time;

end;
--------------------------------------------------------------------------------
function public.Show ( section_data, delivery_time )

  if ( private.active ) then return; end;

  private.active = true;
  private.delivery_time = delivery_time;
  
  private.SetDefault( section_data );

  ui.window.Open( "carrier_send", _f.OnOpen );

end;
--------------------------------------------------------------------------------
function public.Hide ()

  if ( not private.active ) then return; end;

  private.active = false;

  ui.window.Close(
    "carrier_send",
    function ()

      private.Clean();
      _f.OnClose();

    end
  );
  
end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function private.Adapt ()

  local multiset = {};
  
  ui.obj.ToCenter( private.this.dialog, 0, 0, multiset );
  
  ui.window.SetArea(
    private.this.dialog,
    private.window_area[ private.app_o ],
    multiset
  );
  
  ui.obj.Set(
    private.this.window,
    {
      animfunc = private.app_o,
      tt_title = private.strings.title,
      tt_description = private.strings.description
    },
    multiset
  );
  
  local scc_size = {
    portrait = {
      x = -280,
      y = -400,
      w =  570,
      h =  746
    },
    landscape = {
      x = -280,
      y = -320,
      w =  570,
      h =  556
    }
  };

  local x, y, w, h = scc_size[ private.app_o ].x,
                     scc_size[ private.app_o ].y,
                     scc_size[ private.app_o ].w,
                     scc_size[ private.app_o ].h;

  ui.obj.Set(
    private.this.hub_scroller,
    {
      inputrect_init = true,
      inputrect_x    = x,
      inputrect_y    = y,
      inputrect_w    = w,
      inputrect_h    = h,
      croprect_init  = true,
      croprect_x     = x,
      croprect_y     = y,
      croprect_w     = w,
      croprect_h     = h,
      cellsize       = private.cell_size[ private.app_o ]
    },
    multiset
  );

  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
function private.CreateButtons ()
  
  ui.button.Create(
    private.this.button.close,
    {
      design = "close"
    },
    function() _f.Close(); end
  );
  
end;
--------------------------------------------------------------------------------
function private.SetDefault ( section_data )
  
  local multiset = {};

  ui.obj.Clone(
    "scc_int_dialog_carrier_send",
    private.this.window,
    "_slots"
  );
  
  for s = 1, math.max( 4, #section_data ) do
    
    local slot_data = section_data[ s ];

    if ( slot_data ) then

      local section_name = "section_"..slot_data.section_number;
      local btn_name     = "btn_int_dialog_carrier_send_slot_send_"..s;
      local anim_func    = tern( slot_data.unlock )( "island", "island_lock" );
      
      ui.obj.Clone(
        "anm_int_dialog_carrier_send_slot",
        private.this.scroller,
        "_"..s,
        1
      );
      
      ui.button.Create(
        btn_name,
        {
          title  = private.strings.send,
          design = "medium",
          color  = "green"
        },
        function ()

          if ( _f.Select( slot_data.section_number ) ) then

            _f.Close();

          end;

        end
      );
      local str  = "str_int_dialog_worldmap_";
      local prew = StringGet( str.."section_"..( slot_data.section_number - 1 ) );
      local text = StringGet( str.."description_locked", prew );
      
      ui.obj.Set(
        "anm_int_dialog_carrier_send_slot_"..s,
        {
          animfunc       = anim_func,
          island         = slot_data.section_string,
          tt_res_1_count = slot_data.resource1,
          tt_res_2_count = slot_data.resource2,
          tt_res_3_count = slot_data.resource3,
          tt_you_have    = private.strings.have,
          tt_eta         = private.strings.eta.." "..private.delivery_time,
          lock_text      = text,
          rt_back        = "assets/content/worlds/"..section_name,
          rt_red = slot_data.color.red,
          rt_green = slot_data.color.green,
          rt_blue = slot_data.color.blue
        },
        multiset
      );

    else

      local obj_name = "obj_int_dialog_carrier_send_slot_"..s

      ui.obj.Create( obj_name );
      ui.obj.Attach( obj_name, private.this.scroller );

    end;
    
  end;
  
  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
function private.Clean ()
    
  ui.buttons.Delete(
    "int_dialog_carrier_send",
    {
      private.this.button.close
    }
  );

  ui.obj.Delete( private.this.scroller );
  
end;
--------------------------------------------------------------------------------