-- name=int_dialog_dailyquest
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  RewardTap       = function () end,
  RewardNotify    = function () end,
  ChallengeTap    = function () end,
  CooldownGetTime = function () end,
  CooldownSkip    = function () end,
  Close           = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  private.this = {
    window = "anm_int_dialog_dailyquest_window"
  };

  private.cooldown        = {};
  private.challenge_count = 4;
  
  private.window_area = {
    portrait  = { -320, -420, 640, 900 },
    landscape = { -250, -320, 500, 700 },
    none      = { 0,    0,    0,   0   }
  };
  
  private.int_o = ui.screen.Orientation();

  private.strings =
  {
    caption     = ne.strings.Get( "str_int_dialog_dailyquest_caption" ),
    description = {
      challenge  = ne.strings.Get( "str_int_dialog_dailyquest_challenge_description" ),
      dailyquest = ne.strings.Get( "str_int_dialog_dailyquest_dailyquest_description" ),
      countdown  = ne.strings.Get( "str_int_dialog_dailyquest_countdown_description" )
    },
    skip        = ne.strings.Get( "str_int_dialog_dailyquest_skip" ),
    hint        = ne.strings.Get( "str_int_dialog_dailyquest_hint" ),
    woodchest   = ne.strings.Get( "str_dialog_items_woodchest" ),
    silverchest = ne.strings.Get( "str_dialog_items_silverchest" ),
    goldchest   = ne.strings.Get( "str_dialog_items_goldchest" ),
    epicchest   = ne.strings.Get( "str_dialog_items_epicchest" )
  };

  private.Adapt();
  private.CreateButtons();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show ( states, dailyquest )

  private.challenge_states = states;
  private.reward_id        = dailyquest.reward_id;
  private.is_ready         = dailyquest.ready;

  if ( private.active ) then return; end;

  private.active = true;

  private.SetDefault( states, dailyquest );
  
  ui.window.Open(
    "dailyquest",
    nil,
    nil,
    "daily_quest"
  );
  
end;
--------------------------------------------------------------------------------
function public.Hide ( func_OnHide )
  
  if ( private.active ) then
  
    private.active = false;
    
    ui.window.Close( "dailyquest", function ()

      if ( func_OnHide ) then
        
        func_OnHide();

      end;

    end );

  end;

end;
--------------------------------------------------------------------------------
function public.IsActive ()
  
  return private.active;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Challenge_Set (
    id,
    new_dailyquest,
    old_dailyquest
  )

  private.Challenge_Set( 
    id, 
    new_dailyquest[ "challenge"..id ], 
    old_dailyquest[ "challenge"..id ].state,
    function ()

      private.Timer_TryShow( "challenge" );

    end
  );

end;
--------------------------------------------------------------------------------
function public.Challenge_Update ( id, multiset )
 
  local state = ( not private.Reward_IsPending() );

  private.Challenge_Update( id, state, true, multiset );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Reward_Set ( reward_id, reward )

  private.reward_id = reward_id;

  private.Reward_Set( reward );
  
end;
--------------------------------------------------------------------------------
function public.Reward_Enable ()

  private.Reward_Enable();
  
end;
--------------------------------------------------------------------------------
function public.Reward_Disable ( hide_reward )

  private.Reward_Disable( hide_reward );
  
end;
--------------------------------------------------------------------------------
function public.Reward_ShowPopup ()

  private.Reward_ShowPopup();
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Timer_TryShow ( timer_type, skip_price )

  private.Timer_TryShow( timer_type, skip_price );

end;
--------------------------------------------------------------------------------
function public.Timer_Hide ()

  private.Timer_Hide();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Adapt ()

  local multiset = {};

  ui.obj.ToCenter( "obj_int_dialog_dailyquest", nil, nil, multiset );  

  for c = 1, private.challenge_count do

    ui.obj.Set(
      "anm_int_dialog_dailyquest_challenge_"..c,
      {
        inputrect_init = true,
        inputrect_x    = -100,
        inputrect_y    = -100,
        inputrect_w    = 200,
        inputrect_h    = 200
      },
      multiset
    );

    ui.control.TapSet( "anm_int_dialog_dailyquest_challenge_"..c,
    {    
      func_true = function ()

        _f.ChallengeTap( c );

         common.sound.Play( "simply_click" );

      end
    } );

    ui.control.TapSet( "anm_int_dialog_dailyquest_challenge_point"..c,
    {    
      func_true = function ()

        _f.ChallengeTap( c );

         common.sound.Play( "simply_click" );

      end
    } );

  end;

  ui.obj.Set(
    private.this.window,
    {
      animfunc = private.int_o,
      title    = private.strings.caption,
      rt_bg    = "assets/content/interface/int_dialog_dailyquest/daily_quest_background"
    },
    multiset
  );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.CreateButtons ()

  ui.button.Create(
    "btn_int_dialog_dailyquest_close",
    {
      design = "close",
      sound  = "simply_click",
    },
    function() _f.Close(); end
  );
  
  ui.button.Create(
    "btn_int_dialog_dailyquest_skip",
    {
      design = "price",
      sound  = "buy"
    },
    function() _f.CooldownSkip( "dailyquest" ); end
  );  

  ui.control.EventsSet(
    "obj_int_dialog_dailyquest_cloud",
    {    
      func_on_click = function () _f.RewardTap(); end
    }
  );

end;
--------------------------------------------------------------------------------
function private.SetDefault ( states, dailyquest )

  local multiset = {};

  private.Cooldown_Update( "challenge" );
  private.Cooldown_Update( "countdown" );

  for id = 1, private.challenge_count do

    local challenge = dailyquest[ "challenge"..id ];
    
    private.Challenge_Set( id, challenge );
    
  end;

  private.Reward_Set( dailyquest.reward );

  if ( not private.cooldown.challenge ) then

    if ( private.Reward_IsPending () ) then

      private.Reward_Enable();

    else

      private.Reward_Disable();

    end;

  end;

  local timer_type = "challenge";
  local skipprice;

  if ( private.cooldown.countdown ) then

    timer_type = "countdown";
    
  elseif ( not dailyquest.ready           ) and
         ( not private.cooldown.challenge ) then

    timer_type = "dailyquest";
    skipprice = dailyquest.skipprice;

  end;

  private.Timer_TryShow( timer_type, skipprice );

  if ( private.Reward_IsPending() ) then
    _f.RewardTap();
  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Timer_TryShow ( timer_type, skip_price )
  
  local multiset = {};
    
  if ( private.Timer_TryStart( timer_type ) ) then

    local animfunc_appear = "appear";
    local animfunc_static = "static";
    
    ui.button.Disable( "btn_int_dialog_dailyquest_skip" );

    if ( timer_type == "dailyquest" ) then

      animfunc_appear = "appear_skip";
      animfunc_static = "static_skip";

      ui.button.Enable( "btn_int_dialog_dailyquest_skip" );
      ui.button.SetText( "btn_int_dialog_dailyquest_skip", skip_price );
      
    end;

    ui.input.Off( "obj_int_dialog_dailyquest_cloud", multiset );
    
    ui.obj.Set(
      "anm_int_dialog_dailyquest_cooldown",
      { 
        active      = true,
        input       = true, 
        visible     = true, 
        animfunc    = animfunc_appear,
        hint        = private.strings.hint,
        tt_skip     = private.strings.skip,
        description = private.strings.description[ timer_type ],
        chest_icon  = private.reward_res,
        playing     = true,
        frame       = 0,
        endtrig     = function ()
          if ( private.Cooldown_IsActive( timer_type ) ) then
            ui.anim.Play(
              "anm_int_dialog_dailyquest_cooldown",
              animfunc_static
            );
          else
            private.Timer_Hide();
          end;
        end
      },
      multiset
    );
    
    ui.window.SetArea(
      "obj_int_dialog_dailyquest",
      private.window_area.none,
      multiset
    );

    ui.obj.Multiset( multiset );

    return true;
    
  end;

  ui.input.On( "obj_int_dialog_dailyquest_cloud", multiset );

  ui.obj.Disable(
    "anm_int_dialog_dailyquest_cooldown",
    false,
    multiset
  );
    
  ui.window.SetArea(
    "obj_int_dialog_dailyquest",
    private.window_area[ private.int_o ],
    multiset
  );

  ui.obj.Multiset( multiset );

  return false;

end;
--------------------------------------------------------------------------------
function private.Timer_TryStart ( timer_type )
  
  private.Cooldown_Update( timer_type );
  
  if ( private.Cooldown_IsActive( timer_type ) ) and
     ( not private.Reward_IsPending() ) then  

    private.active_timer = timer_type;
    private.Timer_Update( timer_type );

    return true;
    
  end;

  private.active_timer = nil;
  common.Timer_Stop( "tmr_int_dialog_dailyquest_timer" );

  return false;
  
end;
--------------------------------------------------------------------------------
function private.Timer_Update ( timer_type )

  private.Cooldown_Update( timer_type );
  
  if ( private.Cooldown_IsActive( timer_type ) ) then

    local multiset = {};

    common.Timer_Start(
      "tmr_int_dialog_dailyquest_timer",
      1,
      function ()

        private.Timer_Update( timer_type );

      end,
      multiset
    );

    if ( private.active_timer ) then

      ui.obj.Set(
        "anm_int_dialog_dailyquest_cooldown",
        {
          time = ui.text.FormatTime( private.cooldown[ timer_type ] )
        },
        multiset
      );
      
    end;
      
    ui.obj.Multiset( multiset );

  else

    private.Timer_Hide();

  end;

end;
--------------------------------------------------------------------------------
function private.Timer_Hide ()
  
  ui.window.SetArea(
    "obj_int_dialog_dailyquest",
    private.window_area[ private.int_o ]
  );

  local multiset = {};

  common.Timer_Stop( "tmr_int_dialog_dailyquest_timer", multiset );

  if ( private.active_timer ) then

    ui.anim.Stop( "anm_int_dialog_dailyquest_cooldown" );

    local animfunc = "disappear";

    if ( private.active_timer == "dailyquest" ) then

      animfunc = "disappear_skip";
      
      ui.button.Disable( "btn_int_dialog_dailyquest_skip" );

    end;

    ui.input.On( "obj_int_dialog_dailyquest_cloud", multiset );
  
    ui.obj.Set(
      "anm_int_dialog_dailyquest_cooldown",
      { 
        animfunc = animfunc,
        playing  = true,
        frame    = 0,
        endtrig  = function ()
          ui.obj.Disable( "anm_int_dialog_dailyquest_cooldown" );
        end
      },
      multiset
     );
    
  end;

  for number = 1, private.challenge_count do

    private.Challenge_Update(
      number,
      ( private.reward_id == false ),
      true,
      multiset
    );

  end;

  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Challenge_Set (
    id,
    challenge,
    state_old,
    func_OnAnimEnd
  )
  
  local anm_slot     = "anm_int_dialog_dailyquest_challenge_"..id;
  local new_animfunc = private.GetAnimfuncByState( challenge.state );
  local old_animfunc = private.GetAnimfuncByState( state_old );
  local anm_playing  = true;

  local func_set_state = function ( anm_slot, animfunc, playing )   

    AnimStop( anm_slot );

    local anm_params = 
    {
      animfunc = animfunc,
      playing  = playing,
      frame    = 0,
      endtrig  = function() end
    };

    if ( func_OnAnimEnd ) then
        
      func_OnAnimEnd();
        
    end;

    ui.obj.Set( anm_slot, anm_params );

  end;

  if ( challenge.state == private.challenge_states[ 1 ] ) then

    anm_playing = false;

    private.Challenge_ShowEnemy(
      id,
      challenge.enemy.name,
      challenge.enemy.content
    );

  else 

    private.Challenge_HideEnemy( id );

  end;

  private.Challenge_Update(
    id,
    ( not private.Reward_IsPending() )
  );
  
  if ( old_animfunc ) then
    
    if ( new_animfunc == "state_3" ) then

       common.sound.Play( "daily_quest_new" );

    end;
    
    ui.anim.Play(
      anm_slot,
      old_animfunc,
      function ()

        func_set_state( anm_slot, new_animfunc, anm_playing );

      end
    );

  else

    func_set_state( anm_slot, new_animfunc, anm_playing );

  end;

end;
--------------------------------------------------------------------------------
function private.Challenge_Update ( id, state, animated, multiset )

  if ( private.Cooldown_IsActive( "challenge" ) ) or
     ( private.Cooldown_IsActive( "countdown" ) ) or
     ( not private.is_ready                     ) then

    state = false;

  end;

  local color    = tern( state )( 1.0, 0.5 );
  local obj_name = "anm_int_dialog_dailyquest_challenge_"..id;

  if ( animated ) then

    local start = ui.obj.Get( obj_name );

    if ( start.color_r ~= color ) then

      ui.input.Set( obj_name, state, multiset );

      ObjAnimate( obj_name, "color_rgb", 0, 0, "", 
      {
        0.0, 0, start.color_r, start.color_g, start.color_b,
        0.3, 0, color,         color,         color,
      } );

    end;

  else

    ui.obj.Set(
      obj_name,
      { 
        input   = state,
        color_r = color,
        color_g = color,
        color_b = color
      },
      multiset
    );

  end;

end;
--------------------------------------------------------------------------------
function private.Challenge_ShowEnemy ( num, enemy_name, enemy_content )

  --FF Костыль для гигантских монстров
  local scale = 1.0;

  if ( enemy_name == "octocrab" ) then

    scale = 0.65;

  elseif ( enemy_name == "bonedragon" ) then

    scale = 0.8;

  end;

  local multiset = {};

  ui.obj.Set(
    "anm_int_dialog_dailyquest_challenge_"..num.."_enemy",
    {
      active   = true,
      visible  = true,
      res      = "assets/content/"..enemy_content.."/"..enemy_name.."/anm_map_"..enemy_name,
      animfunc = "static",
      playing  = true,
      scale_x  = scale,
      scale_y  = scale
    },
    multiset
  );

  private.Point_SetActive( num, multiset );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.Challenge_HideEnemy ( num, enemy_name, enemy_content )

  ui.obj.Set(
    "anm_int_dialog_dailyquest_challenge_"..num.."_enemy",
    {
      active  = false,
      visible = false,
      playing = false
    }
  );

  private.Point_SetComplete( num );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Point_SetActive ( num, multiset )

  local input = true;

  if ( private.Cooldown_IsActive( "challenge" ) ) or
     ( private.Cooldown_IsActive( "countdown" ) ) or
     ( not private.is_ready                     ) or
     ( private.Reward_IsPending()               ) then

    input = false;

  end;

  ui.obj.Set(
    "anm_int_dialog_dailyquest_challenge_point"..num.."_bg",
    {
      animfunc = "bg_active"
    },
    multiset 
  );

  ui.input.Set(
    "anm_int_dialog_dailyquest_challenge_point"..num,
    input,
    multiset
  );

end;
--------------------------------------------------------------------------------
function private.Point_SetComplete ( num )

  local multiset = {};

  ui.obj.Set(
    "anm_int_dialog_dailyquest_challenge_point"..num.."_bg",
    {
      animfunc = "bg_disabled"
    },
    multiset 
  );

  ui.input.Off( "anm_int_dialog_dailyquest_challenge_point"..num, multiset );
  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Reward_IsPending ()

  return ( type( private.reward_id ) == "number" );
  
end;
--------------------------------------------------------------------------------
function private.Reward_Enable ()

  ui.obj.Enable( "spr_int_dialog_dailyquest_challenge_reward", true );

  ui.anim.Play(
    "anm_int_dialog_dailyquest_cloud",
    "cloud_reward"
  );
  
end;
--------------------------------------------------------------------------------
function private.Reward_Disable ( hide_reward )

  ui.anim.Play(
    "anm_int_dialog_dailyquest_cloud",
    "cloud"
  );

  if ( hide_reward ) then

    ui.obj.Disable( "spr_int_dialog_dailyquest_challenge_reward", true );

  else

    ui.obj.Enable( "spr_int_dialog_dailyquest_challenge_reward", true );

  end;
  
end;
--------------------------------------------------------------------------------
function private.Reward_ShowPopup ()
  
  ui.popup.Cloud_Show( 
    "monolog", 
    "anm_int_dialog_dailyquest_cloud", 
    "str_int_dialog_dailyquest_reward_description",
    nil,
    { -20, -50 }
  );
  
end
--------------------------------------------------------------------------------
function private.Reward_Set ( reward )

  local multiset = {};

  if ( reward ) then

    ui.obj.Set(
      "anm_int_dialog_dailyquest_cloud",
      {
        chest = private.strings[ reward.name ]
      },
      multiset
    );

  else

    ui.obj.Set(
      "anm_int_dialog_dailyquest_cloud",
      {
        chest = ""
      },
      multiset
    );

  end;

  if ( not reward ) then

    ui.obj.Disable(
      "spr_int_dialog_dailyquest_challenge_reward",
      false,
      multiset
    );

    ui.obj.Multiset( multiset );

    return;

  end;

  local scale = 1.0;

  if ( reward.type == "booster" ) then

    scale = 0.7;  

  end;

  private.reward_res = ui.obj.Res( reward.type, reward.name );

  ui.obj.Set(
    "spr_int_dialog_dailyquest_challenge_reward",
    {
      res     = private.reward_res,
      scale_x = scale,
      scale_y = scale,
    },
    multiset
  );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.GetAnimfuncByState ( state )

  if ( state == private.challenge_states[ 1 ] ) then

    return "state_1";

  elseif ( state == private.challenge_states[ 2 ] ) then

    return "state_2";

  elseif ( state == private.challenge_states[ 3 ] ) then

    return "state_3";

  end;

  return nil;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Cooldown_IsActive ( timer_type )
  
  private.Cooldown_Update( timer_type );
  
  return ( private.cooldown[ timer_type ]     ) and
         ( private.cooldown[ timer_type ] > 0 );
  
end;
--------------------------------------------------------------------------------
function private.Cooldown_Update ( timer_type )

  private.cooldown[ timer_type ] = _f.CooldownGetTime( timer_type );

end;
--------------------------------------------------------------------------------