-- name=int_dialog_hearts
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  ShowAskDialog        = function () end,
  ShowSendDialog       = function () end,
  Accept               = function () end,
  Close                = function () end,
  OnClose              = function () end,
  ShowTopPanel         = function () end,
  GetAttemptsVideo     = function () end,
  ShowVideo            = function () end,
  IsImmortality        = function () end,
  IsFacebook           = function () end,
  IsInvite             = function () end,
  GetHeartsData        = function () end,
  IsEventExist         = function () end,
  IsEventTimeRemaining = function () end,
  IsVideoAvailable     = function () end,
  VideoCanWatch        = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  private.int_w       = ui.screen.Width();
  private.int_h       = ui.screen.Height();
  private.orientation = ui.screen.Orientation();

  private.this = {
    dialog = "obj_int_dialog_hearts",
    window = "anm_int_dialog_hearts_window",
    button = {
      close    = "btn_int_dialog_hearts_close",
      accept   = "btn_int_dialog_hearts_accept",
      ask      = "btn_int_dialog_hearts_ask",
      send     = "btn_int_dialog_hearts_send",
      invite   = "btn_int_dialog_hearts_invite",
      video    = "btn_int_dialog_hearts_video",
    },
    timer = {
      main   = "tmr_int_dialog_hearts_main",
      social = "tmr_int_dialog_hearts_social",
      count  = "tmr_int_dialog_hearts_count"
    }
  };

  private.strings = {
    title       = ne.strings.Get( "str_int_dialog_hearts_caption" ),
    title_full  = ne.strings.Get( "str_int_dialog_hearts_caption_full" ),
    full        = ne.strings.Get( "str_int_dialog_hearts_full" ),
    immortality = ne.strings.Get( "str_int_dialog_hearts_immortality" )
  };
  
  private.window_area = {
    static_landscape               = { -313, -310, 626, 520 },
    static_full_landscape          = { -313, -250, 626, 410 },
    static_portrait                = { -313, -360, 626, 520 },
    static_full_portrait           = { -313, -360, 626, 410 },
    static_fb_portrait             = { -313, -390, 626, 610 },
    static_fb_full_portrait        = { -313, -390, 626, 610 },
    static_fb_invite_portrait      = { -313, -390, 626, 610 },
    static_fb_invite_full_portrait = { -313, -360, 626, 500 }
  }; 
  
  private.Adapt();
  private.CreateButtons();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Adapt ()
  
  local multiset = {};

  ui.obj.ToCenter( private.this.dialog, 0, 0, multiset );

  ui.obj.Set(
    private.this.window,
    {
      title = private.strings.title
    },
    multiset
  );

  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
function private.CreateButtons ()
  
  ui.button.Create(
    private.this.button.close,
    {
      design = "close",
      sound  = "simply_click",
    },
    function() _f.Close(); end
  );

  ui.button.Create(
    private.this.button.accept,
    {
      title  = "str_int_dialog_hearts_accept",
      design = "heart",
      price  = true
    },
    function() _f.Accept(); end
  );

  ui.button.Create(
    private.this.button.ask,
    {
      title  = "str_int_dialog_hearts_ask",
      design = "facebook_large",
      sound  = "simply_click"
    },
    function() _f.ShowAskDialog(); end
  );
  
  ui.button.Create(
    private.this.button.send,
    {
      title  = "str_int_dialog_hearts_send",
      design = "medium",
      color  = "green"
    },
    function() _f.ShowSendDialog(); end
  );
  
  if ( IsMobile() ) then
    ui.button.Create(
      private.this.button.video,
      {
        title  = "str_int_dialog_hearts_video",
        design = "video"
      },
      function() _f.ShowVideo(); end
    );
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Show ( skin_current )
  
  private.active = true;

  local is_immortality = _f.IsImmortality();

  private.SetSkin( skin_current, is_immortality );
  private.SetDefault( is_immortality );

  ui.window.Open( "hearts",
    function ()
      _f.ShowTopPanel();
      if ( is_immortality ) then private.ShowHeartIncreaseFX(); end;
    end
  );

end;
--------------------------------------------------------------------------------
function public.Hide ()

  if ( not private.active ) then return; end;

  private.active = false;

  ui.window.Close(
    "hearts",
    function ()
      ui.skin.Detach( private.hero );
      _f.OnClose();
    end
  );

end;
--------------------------------------------------------------------------------
function public.Update ( reason )
  
  if ( not private.active ) then return; end;
  
  private.SetDefault();

  -- JK: always play fx on update
  private.ShowHeartIncreaseFX();
  
  if ( ( reason ) and ( reason == "video" ) ) then
    local count = _f.GetHeartsData();
    ui.obj.Set( private.this.window, { hearts_count = tostring( math.max( count - 1, 0 ) ) } );
    ui.obj.Set( private.this.timer.count,
    {
      playing = true,
      time = 0.2,
      endtrig = function ()
        
        ui.obj.Set( private.this.window, { hearts_count = tostring( count ) } );
        ui.obj.Set( private.this.timer.count, { endtrig = "", playing = false } );
        
      end;
    } );
  end;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Timer_Main_Show ()

  local multiset = {};
  local is_immortality = _f.IsImmortality();
  local time_string  = tern( is_immortality )(
    private.strings.immortality,
    private.strings.full
  );

  local seconds  = _f.IsEventTimeRemaining();
  local window_title = private.strings.title;
  
  if ( seconds <= 0 ) then

    common.Timer_Stop( private.this.timer.main, multiset );

    window_title = private.strings.title_full;

  else

    common.Timer_Start(
      private.this.timer.main,
      1,
      private.Timer_Main_Show,
      multiset
    );

    time_string = ui.text.FormatTime( seconds );
    
  end;

  ui.obj.Set(
    private.this.window,
    {
      time  = time_string,
      title = window_title
    },
    multiset
  );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.SetDefault ( is_immortality )

  local multiset = {};
  local count, max_count, price = _f.GetHeartsData();
  local facebook = _f.IsFacebook();
  local is_invite = _f.IsInvite();
  local hearts_count;

  if ( is_immortality ) then
    hearts_count = ui.text.InfinitySign();
  else
    hearts_count = ui.text.FormatNumber( count );
  end;

  ui.button.SetPrice( private.this.button.accept, price, multiset );

  ui.button.SetTextTag(
    private.this.button.accept,
    "count",
    "+",--"+"..ui.text.FormatNumber( max_count - count ),
    multiset
  );

  local full = tern( count == 0 )( "", "full_" );
  local fb = tern( facebook )( "fb_", "" );
  local invite = tern( is_invite and facebook )( "invite_", "" );
  local animfunc = "static_"..fb..invite..full..private.orientation;
  
  ui.window.SetArea(
    private.this.dialog,
    private.window_area[ animfunc ],
    multiset
  );
  
  ui.obj.Set(
    private.this.window,
    {
      hearts_count = hearts_count,
      animfunc = animfunc,
      playing  = true
    },
    multiset
  );
  
  local attempts, max_attempts = _f.GetAttemptsVideo();
  local is_video_available = _f.IsVideoAvailable();
  
  if ( attempts >= max_attempts ) or ( is_immortality ) or 
    ( count >= max_count ) or ( not is_video_available ) then
    _f.VideoCanWatch( false );
    ui.obj.Set( private.this.button.video, { input = false, visible = false }, multiset );
  else
    _f.VideoCanWatch( true );
    ui.obj.Set( private.this.button.video, { input = true, visible = true }, multiset );
  end;
  
  ui.obj.Multiset( multiset );  
  
  private.Timer_Main_Show();
  
end;
--------------------------------------------------------------------------------
function private.SetSkin ( skin_current, is_immortality )
  
  local skin_type = tern( is_immortality )( "fight", "ride" );
  local skin_func = tern( is_immortality )( "win", "dialog" );

  private.hero = ui.skin.Attach( private.this.window, skin_current, skin_type );

  ui.obj.Set( private.hero,
  {
    animfunc = skin_func,
    playing  = true,
    anim_tag = "hero" 
  } );
  
end;
--------------------------------------------------------------------------------
function private.ShowHeartIncreaseFX ()

  ui.anim.Play( "anm_int_dialog_hearts_change_count", "heart_inc" );

end;
--------------------------------------------------------------------------------
function private.ShowImmortalityFX ()

  ui.anim.Play( "anm_int_dialog_hearts_change_count", "heart_immortality" );

end;
--------------------------------------------------------------------------------