-- name=int_dialog_hearts_send
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Accept = function () end,
  Close  = function () end,
  OnClose  = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  private.this = {
    dialog = "obj_int_dialog_hearts_send",
    window = "anm_int_dialog_hearts_send_window",
    button = {
      close  = "btn_int_dialog_hearts_send_close",
      accept = "btn_int_dialog_hearts_send_accept"     
    },
    photo = "spr_int_dialog_hearts_send_photo"
  };

  private.app_o = ui.screen.Orientation();

  private.strings = {
    title         = ne.strings.Get( "str_int_dialog_hearts_send_caption" ),
    desc_allow    = ne.strings.Get( "str_int_dialog_hearts_send_description_allowed" ),
    desc_disallow = ne.strings.Get( "str_int_dialog_hearts_send_description_disallowed" )
  };

  private.window_area = {
    default = { -270, -220, 540, 460 }
  };

  private.Adapt();
  private.CreateButtons();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Show (
    is_limit_reached,
    friend_name,
    friend_res,
    friend_url,
    pos_obj_friend
  )

  if ( private.active ) then return; end;

  private.active = true;
  private.pos_obj_friend = pos_obj_friend;
  
  local animfunc = tern( is_limit_reached )(
    private.app_o.."_nobutton",
    private.app_o.."_button"
  );

  local description = tern( is_limit_reached )(
    private.strings.desc_disallow,
    private.strings.desc_allow
  );

  local multiset = {};
  local visible = ( friend_res or friend_url ) and true or false;

  ui.obj.Set(
    private.this.photo,
    {
      url           = friend_url,
      res           = friend_res,
      default_res   = friend_res,
      visible       = visible,
      trg_resloaded = function ()

        ui.obj.SpriteAdjust( private.this.photo, 160 );

      end
    },
    multiset
  );

  ui.obj.Set( "anm_int_dialog_hearts_send_heart", { visible = true } );
  
  ui.obj.Set(
    private.this.window,
    {
      animfunc = animfunc,
      tt_desc  = description,
      tt_name  = friend_name
    },
    multiset
  );

  ui.obj.Multiset( multiset );
  ui.obj.SpriteAdjust( private.this.photo, 160 );
  ui.window.Open( "hearts_send" );

end;
--------------------------------------------------------------------------------
function public.Hide ()

  if ( not private.active ) then return; end;

  private.active = false;

  ui.window.Close( "hearts_send", function () _f.OnClose(); end );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Adapt ()
  
  local multiset = {};

  ui.obj.ToCenter( private.this.dialog, 0, 0, multiset );

  ui.window.SetArea(
    private.this.dialog,
    private.window_area.default,
    multiset
  );

  ui.obj.Set(
    private.this.window,
    {
      title = private.strings.title,
    },
    multiset
  );

  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
function private.CreateButtons ()
  
  ui.button.Create(
    private.this.button.close,
    {
      design = "close",
      sound  = "simply_click",
    },
    function() _f.Close(); end
  );

  ui.button.Create(
    private.this.button.accept,
    {
      title  = "str_int_dialog_hearts_send_accept",
      color  = "green",
      design = "medium"
    },
    function()
  
      private.Accept();
      _f.Close();
      
    end
  );

end;
--------------------------------------------------------------------------------
function private.Accept ()
  
  if ( ui.obj.Exists( "anm_int_dialog_hearts_send_heart_to_friend" ) ) then
    ObjStopAnimate( "anm_int_dialog_hearts_send_heart_to_friend", "pos_xy" );
    ObjStopAnimate( "anm_int_dialog_hearts_send_heart_to_friend", "scale_xy" );
    ui.obj.Delete( "anm_int_dialog_hearts_send_heart_to_friend" );
  end;
  
  ObjClone( "anm_int_dialog_hearts_send_heart", "_to_friend", 1, "ng_interface" );
  
  local control_hub = ObjGet( "int_control_hub" );
  local pos_z = tern( control_hub )( control_hub.pos_z + 1, 701 );
  local pos_start = ui.obj.Position( "anm_int_dialog_hearts_send_heart", "ng_interface" );

  ui.obj.Set( "anm_int_dialog_hearts_send_heart_to_friend",{
    pos_x = pos_start[1],
    pos_y = pos_start[2],
    pos_z = pos_z
  } );
  ui.obj.Set( "anm_int_dialog_hearts_send_heart", { visible = false } );
  
  ObjAnimate( "anm_int_dialog_hearts_send_heart_to_friend", "pos_xy", 0, 0, 
    function ()
      
      ui.obj.Delete( "anm_int_dialog_hearts_send_heart_to_friend" );
      
    end, 
    {
      0.0, 2, pos_start[1], pos_start[2],
      0.3, 2, pos_start[1], pos_start[2],
      0.8, 2, private.pos_obj_friend[1], private.pos_obj_friend[2]
    } 
  );
  
  ObjAnimate( "anm_int_dialog_hearts_send_heart_to_friend", "scale_xy", 0, 0, "", 
    {
      0.0, 0, 1.0, 1.0,
      0.2, 0, 1.0, 1.0,
      0.3, 0, 1.3, 1.3,
      0.8, 2, 0.1, 0.1
    } 
  );
  
  _f.Accept();
  
end;
--------------------------------------------------------------------------------
