-- name=int_dialog_info
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Accept = function () end,
  Reject = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  ui.hub.Create( "int_dialog_info" );

  private.Adapt();
  private.CreateButtons();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Adapt ()

  local app_w = ui.screen.Width();
  local app_h = ui.screen.Height();

  private.this = "obj_int_dialog_info";

  private.window_area = {
    default  = { -272, -300, 544, 600 }
  };

  local multiset = {};

  ui.window.SetArea( private.this, private.window_area.default, multiset );
  ui.obj.ToCenter( private.this, 0, 0, multiset );

  ui.obj.Set(
    "spr_int_dialog_info_bg_black",
    {
      scale_x = app_w / 64,
      scale_y = app_h / 64,
    },
    multiset
  );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.CreateButtons ()

  ui.button.Create(
    "btn_int_dialog_info_close",
    {
      design = "close"
    },
    function() private.FuncClose(); end
  );
  
  ui.button.Create(
    "btn_int_dialog_info_accept",
    {
      color  = "green",
      design = "medium",
      timer  = "down",
      info   = true
    },
    function () private.FuncOk(); end
  );
  
  ui.button.Create(
    "btn_int_dialog_info_accept_alt",
    {
      color  = "green",
      design = "medium",
      timer  = "down",
      info   = true
    },
    function () _f.Reject(); end
  );    
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show ( message_type, params )

  ui.hub.AttachContent( "int_dialog_info" );
  
  private.active = true;
  
  private.FuncClose = public.Hide;
  private.FuncOk = function ()
    _f.Accept(); 
    private.FuncClose();
  end;

  local button_text = "str_dialog_info_button_"..message_type;
  local message     = "str_dialog_info_message_"..message_type;
  local animfunc    = "portrait_btn";
  
  ObjMultiSet(
    {
      {
        "anm_int_dialog_info",
        {
          attention = message,
          title     = "str_dialog_info_caption",
          tt_player_id = ""
        }
      },
      {
        "btn_int_dialog_info_accept",
        {
          active = true,
          input  = true,
          visible = true,
          anim_tag = "btn_accept"
        }
      },
      {
        "btn_int_dialog_info_accept_alt",
        {
          active = false,
          input = false,
          visible = false
        }
      }
    }
  );

   common.sound.Play( "app_info" );

  ui.button.SetText( "btn_int_dialog_info_accept", button_text );
  
  local multiset = {};

  ui.button.SetTime( "btn_int_dialog_info_accept",     false, multiset );
  ui.button.SetTime( "btn_int_dialog_info_accept_alt", false, multiset );
  ui.button.SetInfo( "btn_int_dialog_info_accept",     false, multiset );
  ui.button.SetInfo( "btn_int_dialog_info_accept_alt", false, multiset );
  
  if ( message_type == "connectionlost" ) then

    ui.obj.Disable( "btn_int_dialog_info_close", false, multiset );

  elseif ( message_type == "newserver" ) then

    button_text = "str_dialog_info_message_"..message_type.."_"..( params.state or "old" );
    
    ui.obj.Disable( "btn_int_dialog_info_close", false, multiset );

    private.FuncOk = _f.Reject;

  elseif ( message_type == "savedelete" ) then

    ui.obj.Enable( "btn_int_dialog_info_close", false, multiset );

    private.FuncOk = _f.Accept;

  elseif ( message_type == "contentobsolete" ) then

    ui.obj.Disable( "btn_int_dialog_info_close", false, multiset );

    private.FuncOk = _f.Accept;

  elseif ( message_type == "playhaven_notify" ) then

    ui.obj.Enable( "btn_int_dialog_info_close", false, multiset );

    private.FuncOk = function () _f.Accept(); private.FuncClose(); end;
      
  elseif ( message_type == "exitconfirm" ) then

    ui.obj.Enable( "btn_int_dialog_info_close", false, multiset );

    private.FuncOk    = _f.Accept;
    private.FuncClose = _f.Reject;
      
  elseif ( message_type == "languageconfirm" ) then

    ui.obj.Enable( "btn_int_dialog_info_close", false, multiset );

    private.FuncOk = _f.Accept;
      
  elseif ( message_type == "widescreen_map" ) then

    ui.obj.Enable( "btn_int_dialog_info_close", false, multiset );

    private.FuncOk = _f.Accept;
      
  elseif ( message_type == "widescreen_stage" ) then

    ui.obj.Enable( "btn_int_dialog_info_close", false, multiset );

    private.FuncOk = _f.Accept;
 
  elseif ( message_type == "prioritysession" ) then
      
    ui.obj.Disable( "btn_int_dialog_info_close", false, multiset ); 

  elseif ( message_type == "conflictstorage" ) then
    
    animfunc = "portrait_two_btn";
    
    ui.obj.SetAnimTag( "btn_int_dialog_info_accept", "at_btn_accept_1" );

    ui.obj.SetTextTag(
      "anm_int_dialog_info",
      "title",
      message.."_title",
      multiset
    );     

    ui.obj.Disable( "btn_int_dialog_info_close", false, multiset );
    ui.obj.Enable( "btn_int_dialog_info_accept_alt", false, multiset );

    private.FuncOk = _f.Accept;
    _f.Reject = _f.Reject;
    
    ui.button.SetTime( "btn_int_dialog_info_accept", params.btn1[ 1 ], multiset );
    ui.button.SetTime( "btn_int_dialog_info_accept_alt", params.btn2[ 1 ], multiset );
    ui.button.SetText( "btn_int_dialog_info_accept", button_text.."1", multiset );
    ui.button.SetText( "btn_int_dialog_info_accept_alt", button_text.."2", multiset );
    
  elseif ( message_type == "conflictstorage_local"  ) or 
         ( message_type == "conflictstorage_server" ) then
    
    ui.obj.SetTextTag(
      "anm_int_dialog_info",
      "title",
      message.."_title",
      multiset
    );

    ui.obj.Enable( "btn_int_dialog_info_close", false, multiset );

    private.FuncOk = function ()
      _f.Accept(); 
      public.Hide();
    end;    
    private.FuncClose = _f.Reject;
      
  elseif ( message_type == "dailyquest_feature_disabled" ) then
    
      private.FuncClose = function () 
        public.Hide( _f.Reject );
      end;
      private.FuncOk = function ()
        public.Hide( _f.Accept );
      end;
    
  else

    ui.obj.Enable( "btn_int_dialog_info_close", false, multiset );

    private.FuncOk = function () private.FuncClose(); end;

  end;

  ui.obj.Multiset( multiset );  
  ui.anim.Play( "anm_int_dialog_info", animfunc );
  
end;
--------------------------------------------------------------------------------
function public.Hide ( func_end )

  ObjDelete( "tmr_dialog_info" );
  ObjDelete( "tmr_dialog_info_send" );
    
  ui.hub.DetachContent( "int_dialog_info" );
  
  private.active = false;

  common.sound.Play( "ok_info" );

  if ( func_end ) then

    func_end();

  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.IsActive ()

  return private.active;

end;
--------------------------------------------------------------------------------
function public.GetCloseFunc ()

  local btn_close = ui.obj.Get( "btn_int_dialog_info_close" ) or {};

  if ( btn_close.input ) then
  
    return private.FuncClose;
  
  end;
  
  return nil;

end;
--------------------------------------------------------------------------------
function public.GetOkFunc ()

  local btn_altok = ui.obj.Get( "btn_int_dialog_info_accept_alt" ) or {};
  
  if ( btn_altok.input ) then
    return nil;
  end;
  
  if ( private.FuncOk ) then
    return private.FuncOk;
  end;
  
  return nil;

end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function public.ShowLikeSysMsg ( params )

  ui.hub.AttachContent( "int_dialog_info" );
  
  private.active = true;
  
  local title     = params.title;
  local text      = params.body;
  local btn_left  = params.buttons[ 1 ];
  local btn_right = params.buttons[ 2 ];
  
  local animfunc = "portrait";
  local at_left  = "btn_accept";
  
  local set = {};
  
  local obj =
  {
    btn_close = "btn_int_dialog_info_close",
    btn_left  = "btn_int_dialog_info_accept",
    btn_right = "btn_int_dialog_info_accept_alt",
    dialog    = "anm_int_dialog_info"
  };
  
  -- Setup buttons
  ui.button.Disable( obj.btn_close, set );
  ui.button.Disable( obj.btn_right, set );
  
  if ( btn_left ) then
  
    ui.button.Enable( obj.btn_left, set );
    ui.button.SetText( obj.btn_left, btn_left.title, set );
    ui.button.SetTime( obj.btn_left, btn_left.params[ 1 ] or false, set );
    ui.button.SetInfo( obj.btn_left, btn_left.params[ 2 ] or false, set );
    
    private.FuncOk = function () public.Hide( btn_left.action ); end;
  
  end;
  
  if ( btn_right ) then
  
    ui.button.Enable( obj.btn_right, set );
    ui.button.SetText( obj.btn_right, btn_right.title, set );
    ui.button.SetTime( obj.btn_right, btn_right.params[ 1 ] or false, set );
    ui.button.SetInfo( obj.btn_right, btn_right.params[ 2 ] or false, set );
    
    _f.Reject = function () public.Hide( btn_right.action ); end;
    
    at_left  = "at_btn_accept_1";
  
  end;
  
  if ( btn_left and btn_right ) then
    animfunc = "portrait_two_btn";
  elseif ( btn_left or btn_right ) then
    animfunc = "portrait_btn";
  end;
  
  local id_player = "";
  
  if ( string.find( params.dialog_type, "maintenance" ) ) then
    local id  = params.id;
    local str = "str_int_dialog_hero_info_";
    
    id_player = tern( id == "0" )(
      function() return ne.strings.Get( str.."id_offline" ); end,
      function() return ne.strings.Get( str.."id", id ); end
    )();
  end;
  
  ui.obj.Set( obj.btn_left, { anim_tag = at_left }, set );
  
  -- Setup dialog
  ui.obj.Set(
    obj.dialog,
    {
      title     = title,
      attention = text,
      animfunc  = animfunc,
      playing   = true,
      tt_player_id = ""
    },
    set
  );
  
  ui.obj.Multiset( set );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.CheckConnection ()

  local timer = "tmr_dialog_info_send";

  ObjCreate( timer, "timer" );
  ObjAttach( timer, "obj_int_dialog_info" );

  ObjSet( timer,
  {
    time = 1.0,
    playing = true,
    endtrig = function()

      ObjDelete( timer );

      _f.Accept();
      private.CheckConnection();
      
    end;
  } );

end;
--------------------------------------------------------------------------------
function private.ShowMessage ( message_type )

  ObjSet( "anm_int_dialog_info", { attention = "str_dialog_info_message_wait" } );
  ObjSet( "btn_int_dialog_info_accept", { active = false, input = false, visible = false } );
  
  local timer = "tmr_dialog_info";

  ObjCreate( timer, "timer" );
  ObjAttach( timer, "int_dialog_info" );

  ObjSet( timer,
  {
    time = 10.0,
    playing = true,
    endtrig = function()

      ObjDelete( timer );
      ObjDelete( "tmr_dialog_info_send" );
      ObjSet( "anm_int_dialog_info", { attention = "str_dialog_info_message_"..message_type } );
      ObjSet( "btn_int_dialog_info_accept", { active = true, input = true, visible = true } );

    end;
  } );

end;
--------------------------------------------------------------------------------