-- name=int_dialog_items_chest
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  GetRewards   = function () end,
  CountRewards = function () end,
  OnShow       = function () end,
  OnHide       = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  private.window  = "anm_int_dialog_items_chest_window";
  private.chest   = "anm_int_dialog_items_chest";
  private.counter = "anm_int_dialog_items_chest_count";

  private.Adapt();
  private.CreateButtons();
    
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show ( chest_name, chest_count )

  chest_count = chest_count or 1;

  private.type = chest_name;
  
  private.SetInput( "none" );

  local multiset = {};

  ui.obj.SetTextTag( 
    private.chest,
    "reward",
    ne.strings.Get( "str_dialog_items_"..chest_name ),
    multiset
  );

  if ( chest_count > 1 ) then

    ui.obj.Enable( private.counter, true, multiset );

    ui.obj.SetTextTag( 
      private.counter,
      "tt_counter",
      "#"..chest_count,
      multiset
    );

  else

    ui.obj.Disable( private.counter, true, multiset );

  end;
  
  ui.obj.Multiset( multiset );
  
  ui.anim.Play(
    private.window,
    "open_medium",
    function ()
   
      ui.anim.Play( private.window, "static_medium" );
    
      common.sound.Play( "items_chest_appear" );
    
      ui.anim.Play(
        private.chest,
        private.type.."_appear",
        function ()
    
          private.SetInput( "chest" );
        
           common.sound.Play( "items_chest_waiting", 1 );

          _f.OnShow();
        
          private.Tap_Chest();
      
        end
      );
    
    end
  );
  
  private.active = true;
  
end;
--------------------------------------------------------------------------------
function public.Hide ()

  if ( not private.active ) then return; end;
  
  private.active = false;
  
  _f.OnHide();
 
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Adapt ()

  local w, h = ui.screen.Width(), ui.screen.Height();

  private.this = "obj_int_dialog_items_chest";

  local multiset = {};
  
  ui.obj.ToCenter( private.this, nil, nil, multiset );

  local biggest = ui.screen.MaxDimension();

  private.window_area = {
    default  = { -0.5 * biggest, -0.5 * biggest, biggest, biggest }
  };
  
  ui.window.SetArea(
    private.this,
    private.window_area.default,
    multiset
  );
  
  ui.obj.Set(
    "obj_int_dialog_items_chest_input",
    {
      inputrect_init = true,
      --inputrect_x    = 0,
      --inputrect_y    = 0,
      inputrect_w    = w,
      inputrect_h    = h
    },
    multiset
  );

  ui.obj.Set( 
    private.chest,
    {
      click  = ""--ne.strings.Get( "str_dialog_items_chest_tap" )
    },
    multiset
  );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.CreateButtons ()

  ui.button.Create(
    "btn_int_dialog_items_chest_claim", 
    {
      title  = "str_dialog_items_chest_claim",
      color  = "green",
      design = "medium"
    },
    function() private.Tap_Button(); end
  );
  
  local multiset = {};

  --ui.control.EventsSet(
  --  "obj_int_dialog_items_chest_input",
  --  {
  --    func_on_click = function() private.Tap_Chest(); end
  --  },
  --  multiset
  --);

  for i = 1, 5 do

    ui.control.EventsSet(
      "anm_int_dialog_items_chest_slot"..i,
      {
        func_on_click = function() private.Tap_Reward( i ); end
      },
      multiset
    );

  end;

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.SetRewards ()
  
  local rewards  = _f.GetRewards();
  local multiset = {};

  private.reward_count = #rewards;

  for r = 1, 5 do
    
    local reward    = rewards[ r ];
    local slot_name = "anm_int_dialog_items_chest_slot"..r;
    
    if ( reward ) then

      local tt_count = "";
      
      if ( reward.type ~= "heart" ) then
      
        tt_count = tern( reward.count > 1 )(
          ui.text.FormatNumber( reward.count ),
          tt_count
        );
        
      else
        
        tt_count = ui.text.FormatTime( reward.count );
        
      end;
      
      ui.obj.Set(
        slot_name,
        {
          active    = true,
          visible   = true,
          input     = true,
          animfunc  = "reward_"..reward.type,
          rt_reward = ui.obj.Res( reward.type, reward.name ),
          tt_count  = tt_count
        },
        multiset
      );
      
    else
      
      ui.obj.Disable( slot_name, false, multiset );
      
    end;
    
  end;
  
  ui.obj.Set(
    "anm_int_dialog_items_chest_slots",
    {
      animfunc = "rewards"..private.reward_count
    },
    multiset
  );
  
  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
function private.SetInput ( target )
  
  local multiset = {};
  
  local window = "anm_int_dialog_items_chest_window";
  local input  = "obj_int_dialog_items_chest_input";
  local button = "btn_int_dialog_items_chest_claim";
  
  if     ( target == "none" ) then
    
    ui.input.Off( window, multiset );
    ui.input.Off( input, multiset );
      
  elseif ( target == "chest" ) then
  
    ui.input.On( window, multiset );
    ui.input.On( input, multiset );
    ui.input.Off( button, multiset );
    
  elseif ( target == "button" ) then
  
    ui.input.On( window, multiset );
    ui.input.Off( input, multiset );
    ui.input.On( button, multiset );

  end;
  
  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Tap_Chest ()

  private.SetInput( "none" );
  
  common.sound.Stop( "items_chest_waiting", 1 );
  common.sound.Play( "items_chest_click" );

  private.SetRewards();
  private.animation_close_started = false;
  
  common.sound.Play( "open_chest_derty" );

  ui.anim.Play(
    private.chest,
    private.type.."_open",
    function () private.SetInput( "button" ); end,
    private.type.."_static_open"
  );
    
end;
--------------------------------------------------------------------------------
function private.Tap_Button ()

  if ( private.animation_close_started ) then return; end;

  private.animation_close_started = true;

  private.SetInput( "none" );
  
  _f.CountRewards();
  
  ui.anim.Play(
    private.chest,
    private.type.."_disappear",
    function ()
    
      ui.anim.Play(
        private.window,
        "close_medium",
        function () _f.OnHide(); end
      );
  
    end
  );
  
end;
--------------------------------------------------------------------------------
function private.Tap_Reward ( index )
  
  local slot_name = "anm_int_dialog_items_chest_slot"..index

  ui.input.Off( slot_name );

  ui.anim.Receive_Gold(
    slot_name,
    function ()

      ui.obj.Disable( slot_name );

      private.reward_count = private.reward_count - 1;

      if ( private.reward_count < 1 ) then

        private.Tap_Button();

      end;

    end
  );  

end;
--------------------------------------------------------------------------------