-- name=int_dialog_map_build_construct
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Accept           = function () end,
  Reject           = function () end,
  Find             = function () end,
  ResolveDeficient = function () end,
  ShowTopPanel     = function () end,
  HideTopPanel     = function () end,
  LockTopPanel     = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  private.app_h = ui.screen.Height();
  private.app_o = ui.screen.Orientation();
  
  private.strings = {
    res_desc    = ne.strings.Get( "str_dialog_map_build_construct_resources" ),
    reward      = ne.strings.Get( "str_dialog_map_build_construct_reward" ),
    for_str     = ne.strings.Get( "str_dialog_map_build_construct_for" ),
    title       = ne.strings.Get( "str_dialog_map_build_construct_caption" ),
    description = ne.strings.Get( "str_dialog_map_build_construct_description" )
  };
  
  private.this = {
    dialog = "obj_int_dialog_map_build_construct",
    window = "anm_int_dialog_map_build_construct_window",
    button = {
      close  = "btn_int_dialog_map_build_construct_close",
      buy    = "btn_int_dialog_map_build_construct_buy",
      find   = "btn_int_dialog_map_build_construct_find",
      accept = "btn_int_dialog_map_build_construct_accept"
    },
    item = "spr_int_dialog_map_build_construct_item",
    resources = "anm_int_dialog_map_build_construct_resources"
  };

  private.Adapt();
  private.CreateButtons();
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show ( 
    build_name,
    build_content,
    price_data,
    deficient_data,
    fragments_data,
    rewards_data
  )

  private.active = true;
  
  private.price_count = #price_data;
  
  private.SetDefaults(
    build_name,
    price_data,
    deficient_data,
    fragments_data,
    rewards_data
  );
  
  ui.window.Open(
    "map_build_construct",
    function ()

      ui.anim.Play(
        "anm_int_dialog_map_build_construct_caption",
        "appear",
        function ()

          ui.anim.Play(
            "anm_int_dialog_map_build_construct_caption",
            "static"
          );

        end
      );

      _f.ShowTopPanel();

    end
  );

end;
--------------------------------------------------------------------------------
function public.Hide ( func_execute )

  if ( not private.active ) then 

    return; 

  end;

  private.active = false;

  _f.HideTopPanel();
  
  ui.window.Close(
    "map_build_construct",
    function ()
      ui.window.Hide( "int_dialog_map_build_construct", func_execute );
    end
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Adapt ()

  ui.obj.ToCenter( private.this.dialog );
  
  private.window_area = 
  {
    static_1_normal              = { -266, -280, 536, 545 },
    static_2_normal              = { -268, -260, 536, 402 },
    static_3_normal              = { -268, -260, 536, 582 },
    static_3_portrait_normal     = { -270, -310, 540, 680 },
    static_3_landscape_normal    = { -225, -230, 450, 500 },
    static_4_portrait_normal     = { -266, -340, 536, 730 },
    static_4_landscape_normal    = { -214, -250, 430, 580 },
    static_1_deficient           = { -266, -280, 536, 608 },
    static_2_deficient           = { -268, -250, 536, 450 },
    static_3_deficient           = { -268, -280, 536, 610 },
    static_3_portrait_deficient  = { -270, -310, 540, 710 },
    static_3_landscape_deficient = { -225, -280, 450, 560 },
    static_4_portrait_deficient  = { -268, -350, 538, 757 },
    static_4_landscape_deficient = { -215, -280, 430, 615 }
  };

  -- AL> Random phase for header's fireflies animations
  local func_play;
  func_play = function ( anm_name )
    ObjSet( anm_name, { 
      playing = true, 
      endtrig = function () 
        func_play( anm_name ); 
      end } );
  end;
  
  for c, color in ipairs( { "yellow", "blue", "magneta" } ) do
    ObjSet( "anm_int_dialog_map_build_construct_"..color, { 
      frame = math.random( 1, 1000 ), 
      endtrig = function ()
        func_play( "anm_int_dialog_map_build_construct_"..color );
      end } );
  end;

end;
--------------------------------------------------------------------------------
function private.CreateButtons ()
  ui.button.Create(
    private.this.button.accept,
    {
      title  = "str_dialog_map_build_construct_accept",
      color  = "green",
      design = "large",
      sound  = "simply_click"
    },
    function()
      
      private.ApplyBuild();
      
    end
  );
  
  ui.button.Create(
    private.this.button.close,
    {
      design = "close",
      sound  = "simply_click"
    },
    function() _f.Reject(); end
  );
  
  ui.button.Create(
    private.this.button.find,
    {
      title  = "str_dialog_map_build_construct_find",
      color  = "green",
      design = "medium",
      sound  = "simply_click"
    },
    function() _f.FindFragment(); end
  );
  
  ui.button.Create(
    private.this.button.buy,
    {
      title  = "str_dialog_map_build_construct_buy",
      design = "buy_resource",
      tt_for = private.strings.for_str,
      sound  = "simply_click"
    },
    function()
      
      if ( _f.ResolveDeficient() ) then
        
        private.SetNormal();
        private.ApplyBuild();
        
      end;
      
    end
  );

end;
--------------------------------------------------------------------------------
function private.SetDefaults (
    build_name,
    price_data,
    deficient_data,
    fragments_data,
    reward_data
  )
  
  local multiset = {};
  
  local price_count = private.price_count;
  local postfix     = tern( not deficient_data )( "normal", "deficient" );
  local is_fragment = ( fragments_data and ( #fragments_data > 0 ) );
  local is_reward   = ( reward_data and ( #reward_data > 0 ) );
  
  local win_anim, item_count;
  
  win_anim = tern( is_fragment               )( "static_1", "static_2" );
  win_anim = tern( is_reward                 )( "static_3_"..private.app_o, win_anim );
  win_anim = tern( is_fragment and is_reward )( "static_4_"..private.app_o, win_anim );
  
  if ( is_fragment ) then

    local fragment        = fragments_data[ 1 ].name;
    local collected_count = 0;

    for f = 1, #fragments_data do
      
      collected_count = tern( fragments_data[ f ].collected )(
        collected_count + 1,
        collected_count
      );
      
    end;

    item_count = tostring( collected_count.." / "..#fragments_data );

    local func_SetFindButton = tern( collected_count < #fragments_data )(
      function () ui.button.Enable(  private.this.button.find, multiset ); end,
      function () ui.button.Disable( private.this.button.find, multiset ); end
    );
    
    func_SetFindButton();
    
    ui.obj.Set(
      private.this.item,
      { 
        res = ui.obj.Res( "fragment", fragment ),
        alp = tern( collected_count < #fragments_data )( 0.5, 1 )
      },
      multiset
    );

    _f.FindFragment = function () _f.Find( fragment ); end;

  end;
  
  if ( is_reward ) then
    
    ui.obj.Set(
      "spr_int_dialog_map_build_construct_reward",
      { 
        res = ui.obj.Res( reward_data[ 1 ].type, reward_data[ 1 ].name )
      },
      multiset
    );

  end;
  
  for r = 1, price_count do
    
    local condition = false;
    
    if ( deficient_data ) then
    
      condition = common.IsEntityInList(
        deficient_data.list,
        price_data[ r ].resource
      );
      
    end;
    
    local res_animfunc = tern( condition )(
      "res_"..r.."_deficient",
      "res_"..r.."_normal"
    );
    
    ui.obj.Set(
      "anm_int_dialog_map_build_construct_resource"..r,
      {
        animfunc = res_animfunc,
        tt_count = tostring( price_data[ r ].count )
      },
      multiset
    );
    
  end;
  
  ui.obj.Set(
    "anm_int_dialog_map_build_construct_resources",
    {
      animfunc   = "res_area_"..postfix.."_"..price_count,
      frame      = 0,
      playing    = true
    },
    multiset
  );
  
  local win_animfunc = win_anim.."_"..postfix;
  
  ui.window.SetArea( private.this.dialog, private.window_area[ win_animfunc ] );
  
  ui.obj.Set(
    private.this.window,
    {
      animfunc        = win_animfunc,
      frame           = 0,
      playing         = true,
      item_count      = item_count,
      title           = private.strings.title,
      tt_description  = private.strings.description,
      res_description = private.strings.res_desc,
      reward          = private.strings.reward
    },
    multiset
  );
  
  ui.obj.Set(
    "anm_int_dialog_map_build_construct_caption",
    {
      animfunc = "appear",
      playing  = false,
      frame    = 0
    },
    multiset
  );
  
  if ( deficient_data ) then
  
    ui.button.Enable( private.this.button.buy, multiset );
    
    ui.button.SetTextTag(
      private.this.button.buy,
      "tt_count",
      tostring( deficient_data.gold ),
      multiset
    );
    
  else
    
    ui.button.Disable( private.this.button.buy, multiset );
    
  end;
  
  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
function private.ApplyBuild ()
  
  if ( _f.Accept ) then
  
    _f.LockTopPanel();
    _f.Accept();
  
  end;
  
end;
--------------------------------------------------------------------------------
function private.SetNormal ()
  
  local multiset = {};
  
  for r = 1, private.price_count do
    
    ui.obj.Set(
      "anm_int_dialog_map_build_construct_resource"..r,
      {
        animfunc = "res_"..r.."_normal"
      },
      multiset
    );
    
  end;
  
  ui.obj.Set(
    private.this.resources,
    {
      animfunc   = "res_area_normal_"..private.price_count
    },
    multiset
  );
  
  local win_anim = ui.obj.Get( private.this.window ).animfunc;
  
  win_anim = string.gsub( win_anim, "deficient", "normal" );
  
  ui.obj.Set(
    private.this.window,
    {
      animfunc = win_anim
    },
    multiset
  );
  
  ui.button.Disable( private.this.button.buy, multiset );
  
  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------