-- name=int_dialog_map_quest_reward
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Accept = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  private.Adapt();
  private.CreateButtons();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show (
    character_data,
    reward_data
  )

  private.nogold = true;
  private.active = true;

  private.SetDefault( character_data, reward_data );

  local animfunc_postfix = tern( private.nogold )(
    "nogold_"..private.orientation, 
    private.orientation
  );
  
  ui.anim.Play(
    "anm_int_dialog_map_quest_reward_window",
    "appear_"..animfunc_postfix,
    nil,
    "static_"..animfunc_postfix
  );
  
  ui.window.Open(
    "map_quest_reward",
    nil,
    nil,
    "quest_complete"
  );

end;
--------------------------------------------------------------------------------
function public.Hide ( func_preclose, func_close )

  if ( not private.active ) then return; end;

  private.active = false;
  
  local animfunc_postfix = tern( private.nogold )(
    "nogold_"..private.orientation, 
    private.orientation
  );  
  
  ui.anim.Play(
    "anm_int_dialog_map_quest_reward_window",
    "disappear_"..animfunc_postfix,
    function ()
      
      if ( func_preclose ) then

        -- JK: this should be the first to prevent sprite duplication
        func_preclose();

        ui.obj.Disable( "spr_int_dialog_map_quest_reward_item_icon", true );
        ui.obj.Disable( "anm_int_dialog_map_quest_reward_item_counter", true );

      end;
      
      ui.window.Close( "map_quest_reward", func_close );
      
    end 
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Adapt ()

  local pos_y = {
    portrait  = -20,
    landscape = -40
  };

  private.orientation = ui.screen.Orientation();

  local multiset = {};

  ui.obj.ToCenter( "obj_int_dialog_map_quest_reward", 0, 0, multiset );

  ui.obj.SetTextTag(
    "anm_int_dialog_map_quest_reward_window", 
    "title",
    ne.strings.Get( "str_int_dialog_map_reward_caption" ),
    multiset
  );
  
  ui.obj.Set(
    "obj_int_dialog_map_quest_reward_window",
    {
      pos_y = pos_y[ private.orientation ]
    },
    multiset
  );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.CreateButtons ()
  
  ui.button.Create(
    "btn_int_dialog_map_quest_reward_accept",
    {
      title  = "str_int_dialog_map_quest_reward_accept",
      design = "medium",
      color  = "green",
      sound  = "bottom_button"
    },
    function() _f.Accept(); end
  );

  ui.control.TapSet( "spr_int_dialog_map_quest_reward_item_icon",
  {    
    func_true = function ()

      _f.Accept();

       common.sound.Play( "bottom_button" );

    end
  } );
  
end;
--------------------------------------------------------------------------------
function private.SetDefault ( character_data, reward_data )

  local objects = {};

  local character_name, character_content = character_data.name, character_data.content;
  local character_animfunc = character_data.animfunc;
  local content_dir = "assets/content/"..character_content;
      
  table.insert( objects, { "anm_int_dialog_map_quest_reward_character",
  { 
    res      = content_dir.."/anm_character_"..character_name,
    animfunc = character_animfunc,
    playing  = true 
  } } );
  
  table.insert( objects, { "anm_int_dialog_map_quest_reward_window",
  {
    tt_message = ne.strings.Get(
      "str_int_popup_cloud_characters_completed_"..character_name
    )
  } } );

  for r = 1, #reward_data do

    if ( reward_data[ r ].type == "gold" ) then

      table.insert( objects, { "anm_int_dialog_map_quest_reward_window",
      {
        gold_count = reward_data[ r ].count
      } } );
      
      private.nogold = false;

    else

      local rew_type = reward_data[ r ].type;
      local rew_name = reward_data[ r ].name;

      local icon     = "spr_int_dialog_map_quest_reward_item_icon";
      local params   = 
      {
        active  = true,
        visible = true,
        res     = ui.obj.Res( rew_type, rew_name )
      };
      
      if ( rew_name == "immortality" ) then
        icon     = "anm_int_dialog_map_quest_reward_item_counter";
        params   = 
        {
          active   = true,
          visible  = true,
          animfunc = "reward",
          rt_icon  = ui.obj.Res( rew_type, rew_name ),
          tt_count = ui.text.FormatTime( reward_data[ r ].count )
        };
      end;

      table.insert( objects, { icon, params } );

    end;

  end;
  
  ui.obj.Multiset( objects );

end;
--------------------------------------------------------------------------------