-- name=int_dialog_map_reward
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Accept = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  private.app_o = ui.screen.Orientation();

  ui.button.Create(
    "btn_int_dialog_map_reward_accept",
    {
      title  = "str_int_dialog_map_reward_accept",
      design = "medium",
      color  = "green",
      sound  = "simply_click"
    },
    function () _f.Accept(); end
  );

  local multiset = {};

  ui.obj.SetTextTag(
    "anm_int_dialog_map_reward_window", 
    "title",
    ne.strings.Get( "str_int_dialog_map_reward_caption" ),
    multiset
  );

  ui.obj.ToCenter( "obj_int_dialog_map_reward", 0, 0, multiset );
  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show ( reward_data )

  private.active = true;

  private.SetDefault( reward_data );
  
  ui.window.Open(
    "map_reward",
    function ()

      ui.anim.Play(
        "anm_int_dialog_map_reward_window",
        "appear_"..private.app_o,
        nil,
        "static_"..private.app_o
      );

      common.sound.Play( "level_win_on_map" );
    
    end
  );

end;
--------------------------------------------------------------------------------
function public.Hide ( func_preclose, func_close )

  if ( not private.active ) then return; end;

  private.active = false;

  if ( func_preclose ) then

    -- JK: this should be first to prevent sprite duplication
    func_preclose();
    
    if ( private.reward.item ) then
      
      ui.obj.Disable( "spr_int_dialog_map_reward_item_icon", true );
      ui.obj.Disable( "anm_int_dialog_map_reward_item_counter", true );
      
    end;

  end;

  ui.anim.Play(
    "anm_int_dialog_map_reward_window",
    "disappear_"..private.app_o,
    function ()

      ui.window.Close( "map_reward", func_close, "ok_info" );

    end
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.SetDefault ( reward_data )

  private.reward = 
  {
    gold = false,
    res  = false,
    item = false
  }

  local multiset = {};

  local wnd_obj = "anm_int_dialog_map_reward_content";
  local res_obj = "anm_int_dialog_map_reward_resources"

  local wnd_params = {};
  local res_params = {};

  local res_count = 0;

  for id, reward in pairs( reward_data ) do

    local name  = reward.name;
    local count = reward.count;

    if ( name == "gold" ) then
      
      private.reward.gold   = true;
      wnd_params.gold_count = ui.text.FormatNumber( count );
      
    elseif ( string.find( name, "resource" ) ) then

      private.reward.res = true;
      res_count = res_count + 1;
      res_params[ "res"..res_count.."_count" ] = ui.text.FormatNumber( count );

      local icon   = "spr_int_dialog_map_reward_resource"..res_count.."_icon";
      local params =
      { 
        res = "assets/content/items/"..name
      };

      table.insert( multiset, { icon, params } );

    else

      private.reward.item = true;

      if ( reward.type == "fragment" ) then
        name = name:gsub( "_%d+$", "" );
      end;

      local icon     = "spr_int_dialog_map_reward_item_icon";
      local params   = 
      {
        active  = true,
        visible = true,
        res     = ui.obj.Res( reward.type, name )
      };
      
      if ( name == "immortality" ) then
        icon     = "anm_int_dialog_map_reward_item_counter";
        params   = 
        {
          active   = true,
          visible  = true,
          animfunc = "reward",
          rt_icon  = ui.obj.Res( reward.type, name ),
          tt_count = ui.text.FormatTime( reward.count )
        };
      end;

      table.insert( multiset, { icon, params } );

    end;

  end;

  local wnd_anim = "static";

  if ( private.reward.gold ) then wnd_anim = wnd_anim.."_gold"; end;
  if ( private.reward.res  ) then wnd_anim = wnd_anim.."_res";  end;
  if ( private.reward.item ) then wnd_anim = wnd_anim.."_item"; end;

  wnd_params.animfunc = wnd_anim;
  wnd_params.playing  = true;

  table.insert( multiset, { wnd_obj, wnd_params } );

  res_params.animfunc = "resources"..res_count;
  res_params.playing  = true;

  table.insert( multiset, { res_obj, res_params } );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------