-- name=int_dialog_map_spell
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Accept       = function () end,
  Close        = function () end,
  ShowTopPanel = function () end,
  HideTopPanel = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects()
  
  private.Adapt();
  private.CreateButtons();
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Adapt ()  

  private.this = "obj_int_dialog_map_spell";

  private.window_area = {
    default  = { -270, -330, 538, 640 }
  };

  ui.obj.ToCenter( private.this );  
  ui.window.SetArea( private.this, private.window_area.default );
    
  ObjSet( "anm_int_dialog_map_spell_window",
  {
    title              = ne.strings.Get( "str_int_dialog_map_spell_caption" ),
    input_description  = ne.strings.Get( "str_int_dialog_map_spell_input" ),
    output_description = ne.strings.Get( "str_int_dialog_map_spell_output" )
 } );

end;
--------------------------------------------------------------------------------
function private.CreateButtons ()
  
  ui.button.Create(
    "btn_int_dialog_map_spell_accept",
    {
      title  = "str_int_dialog_map_spell_accept",
      color  = "green",
      design = "large",
      clear  = "window_outer_tap",
      sound  = "ok_info",
    },
    function()
      
      private.TapAccept();
      common.sound.Play( "use_book" );
      
    end
  );
  
  ui.button.Create(
    "btn_int_dialog_map_spell_close",
    {
      design = "close",
      sound  = "ok_info"
    },
    function()
      
      _f.Close();
      
    end
  );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show ( spell_data, zone_data )

  if ( private.active ) then return; end;

  private.active = true;

  private.SetDefault( spell_data, zone_data );

  ui.window.Open( "map_spell", function ()

    _f.ShowTopPanel();

  end );

end;
--------------------------------------------------------------------------------
function public.Hide ( func_execute )

  if ( not private.active ) then return; end;

  private.active = false;

  _f.HideTopPanel();

  ui.window.Close( "map_spell",
  function ()
    ui.window.Hide( "int_dialog_map_spell", func_execute );
  end );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.SetDefault ( spell_data, zone_data )

  private.input_count = #spell_data.input;

  local text_colors = {
    resource1 = { 0.882353, 0.372549, 0.000000 },
    resource2 = { 0.078431, 0.450980, 0.627451 },
    resource3 = { 0.647059, 0.196078, 0.803922 }
  };

  local objects = {};
  local count   = { "", "" };

  for i = 1, private.input_count do
    
    table.insert( objects, { "spr_int_dialog_map_spell_input_slot"..i.."_icon",
    {
      res = ui.obj.Res( "resource", "icon_"..spell_data.input[ i ].name, "" )
    } } );

    count[ i ] = tostring( ui.text.FormatNumber( spell_data.input[ i ].count ) );

  end;
    
  table.insert( objects, { "anm_int_dialog_map_spell_input",
  {
    animfunc      = "book_input_"..private.input_count,
    input_count_1 = count[ 1 ],
    input_count_2 = count[ 2 ] } } );

  table.insert( objects, { "spr_int_dialog_map_spell_output_slot_icon",
  {
    animfunc = "sleep_"..spell_data.output[ 1 ].name,
    playing  = true } } );
    
  table.insert( objects, { "txt_int_dialog_map_spell_output_slot_count",
  {
    text        = "+"..ui.text.FormatNumber( spell_data.output[ 1 ].count ),
    fontcolor_r = text_colors[ spell_data.output[ 1 ].name ][ 1 ],
    fontcolor_g = text_colors[ spell_data.output[ 1 ].name ][ 2 ],
    fontcolor_b = text_colors[ spell_data.output[ 1 ].name ][ 3 ] } } );

  local bg = "assets/content/%s/bg_%s_fight";
  
  table.insert( objects, { "anm_int_dialog_map_spell_window",
  {
    bg       = bg:format( zone_data.content, zone_data.name ),
    animfunc = "book_static",
    playing  = true } } );
    
  table.insert( objects, { "btn_int_dialog_map_spell_accept",
  {
    input = true } } );
    
  table.insert( objects, { "btn_int_dialog_map_spell_close",
  {
    input = true,
    alp   = 1.0 } } );

  ObjMultiSet( objects );

end;
--------------------------------------------------------------------------------
function private.TapAccept ()

  local spell_data = _f.Accept();

  if ( spell_data ) then
    
    local objects = {};
    
    table.insert( objects, { "btn_int_dialog_map_spell_accept",
    {
      input = false } } );
      
    table.insert( objects, { "btn_int_dialog_map_spell_close",
    {
      input = false } } );

    table.insert( objects, { "spr_int_dialog_map_spell_output_slot_icon",
      {
        animfunc = "static_"..spell_data.output[ 1 ].name,
        playing  = true } } );

    ObjMultiSet( objects );

    ui.anim.FadeOut( 0.3, "btn_int_dialog_map_spell_close" );

    ui.anim.Play(
      "anm_int_dialog_map_spell_window",
      "book_activate",
      function ()

        _f.Close( spell_data );

      end
    );

  end;

end;
--------------------------------------------------------------------------------