-- name=int_dialog_minibank
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Accept = function () end,
  Reject = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  private.this = "obj_int_dialog_minibank";

  private.window_area = {
    default = { -270, -300, 540, 580 }
  };
  
  private.Func_Buy_1 = function () end;
  private.Func_Buy_2 = function () end;
  
  ui.button.Create(
    "btn_int_dialog_minibank_close",
    {
      design = "close"
    },
    function()
      
      public.Hide();
      _f.Reject();
      
    end
  );

  ui.button.Create(
    "btn_int_dialog_minibank_buy_1",
    {
      color  = "green",
      design = "medium"
    },
    function() private.Func_Buy_1(); end
  );
  
  ui.button.Create(
    "btn_int_dialog_minibank_buy_2",
    {
      color  = "green",
      design = "medium"
    },
    function() private.Func_Buy_2(); end
  );

  local multiset = {};

  ui.obj.ToCenter( "obj_int_dialog_minibank", 0, 0, multiset );
  ui.window.SetArea( private.this, private.window_area.default, multiset );

  ui.obj.Set(
    "anm_int_dialog_minibank_window",
    {
      title           = StringGet( "str_dialog_minibank_caption" ),
      defecient_text  = StringGet( "str_dialog_minibank_text" )
    },
    multiset
  );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show ( deficient )

  if ( private.active ) then return; end;

  private.active = true;
  
  private.SetLots( deficient );

  local window = "anm_int_dialog_minibank_window";

  ui.window.Open( "minibank" );
  ui.anim.Play( window, "open", nil, "static" );

end;
--------------------------------------------------------------------------------
function public.Hide ()

  if ( not private.active ) then return; end;

  private.active = false;

  ui.window.Close( "minibank", function ()

    ui.window.Hide( "int_dialog_minibank" );

  end );

end;

--------------------------------------------------------------------------------
function public.UpdateLots ( gold_list )

  if ( private.active ) then
  
    local price_1, price_2 = "", "";
    for _, lot in pairs( gold_list ) do
    
      if ( lot.name_ios == private.lot_1_id ) then price_1 = lot.price; 
      end;
      if ( lot.name_ios == private.lot_2_id ) then price_2 = lot.price; 
      end;
    
    end;
    
    local str_buy = ne.strings.Get( "str_dialog_shop_gold_buy" );
    local txt_price_1 = tern( price_1 == "" )( str_buy, price_1 );
    local txt_price_2 = tern( price_2 == "" )( str_buy, price_2 );
    
    ui.button.SetText( "btn_int_dialog_minibank_buy_1", txt_price_1 );
    ui.button.SetText( "btn_int_dialog_minibank_buy_2", txt_price_2 );
  
  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.SetLots ( deficient, objects )
  
  --DbgTrace( deficient );

  local window  = "anm_int_dialog_minibank_window";
  local btn_1   = "btn_int_dialog_minibank_buy_1";
  local btn_2   = "btn_int_dialog_minibank_buy_2";
  local icon_1  = "anm_int_dialog_minibank_lot_1";
  local icon_2  = "anm_int_dialog_minibank_lot_2";
  local heart_1 = "anm_int_dialog_minibank_heart_1";
  local heart_2 = "anm_int_dialog_minibank_heart_2";

  local lots = deficient.lots;

  local objects = {};
  
  if ( lots and #lots > 0 ) then
    
    local lot_1 = lots[ 1 ];
    local lot_2 = lots[ 2 ];
    
    private.lot_1_id = lot_1.name_ios;
    private.lot_2_id = lot_2.name_ios;
    
    local header_1 = "str_dialog_minibank_gold_1";
    local header_2 = "str_dialog_minibank_gold_2";
    
    if ( lot_1.currency_count > lot_2.currency_count ) then
      
      header_2 = "str_dialog_minibank_gold_less";
      
    end;
    
    table.insert( objects, 
    { window,
      {
        header_1        = StringGet( header_1 ),
        header_2        = StringGet( header_2 ),
        count_1         = "+"..tostring( lot_1.currency_count ),
        count_2         = "+"..tostring( lot_2.currency_count ),
        defecient_count = tostring( deficient.count ),
      }
    } );
    
    table.insert( objects, { icon_1, {
      animfunc = "icon_pack_gold_"..lot_1.number,
      playing = true
    } } );
    
    table.insert( objects, { icon_2, {
      animfunc = "icon_pack_gold_"..lot_2.number,
      playing = true
    } } );

    local heart_1_animfunc = "heart_off";
    local heart_2_animfunc = "heart_off";
    local heart_1_count    = "";
    local heart_2_count    = "";

    if ( lot_1.immortality     ) and
       ( lot_1.immortality > 0 ) then

      heart_1_animfunc = "heart_on";
      heart_1_count    = ui.text.FormatTime( lot_1.immortality );

    end;

    if ( lot_2.immortality     ) and
       ( lot_2.immortality > 0 ) then

      heart_2_animfunc = "heart_on";
      heart_2_count    = ui.text.FormatTime( lot_2.immortality );

    end;

    table.insert( objects, { heart_1, {
      animfunc = heart_1_animfunc,
      tt_count = heart_1_count
    } } );
    
    table.insert( objects, { heart_2, {
      animfunc = heart_2_animfunc,
      tt_count = heart_2_count
    } } );
    
    private.Func_Buy_1 = function () 
      _f.Accept( lot_1.name );
    end;
  
    private.Func_Buy_2 = function () 
      _f.Accept( lot_2.name );
    end;
    
    ObjMultiSet( objects );
    
    local lot_1_price = tern( lot_1.price == "" )( ne.strings.Get( "str_dialog_shop_gold_buy" ), lot_1.price );
    local lot_2_price = tern( lot_2.price == "" )( ne.strings.Get( "str_dialog_shop_gold_buy" ), lot_2.price );

    ui.button.SetText( btn_1, lot_1_price );
    ui.button.SetText( btn_2, lot_2_price );    

  end;

end;
--------------------------------------------------------------------------------