-- name=int_dialog_minievent
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  OnShow             = function () end,
  ShowStartDialog    = function () end,
  ShowRewardDialog   = function () end,
  ShowPreviewDialog  = function () end,
  Close              = function () end,
  GetSeconds         = function () end,
  ShowTopPanel       = function () end,
  UpdateTopPanel     = function () end,
  HideTopPanel       = function () end,
  IsTopPanelVisible  = function () end,
  SkipCooldown       = function () end,
  StopCooldown       = function () end,
  GetCooldownSeconds = function () end,
  ShowChestReward    = function () end,
  GetNextReward      = function () end,
  GetDialogData      = function () end,
  SetDialogComplete  = function () end,
  SwitchScene        = function () end,
  GetBookData        = function () end,
  SetBookComplete    = function () end,
  CheckShare         = function () end,
  CheckRewardStage   = function () end,
  GetCurrentSkin     = function () end,
  GetCrownDataChanged = function () end
};

local _this = {};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  _this = {
    is_active = false,
    int      = "int_dialog_minievent",
    hub      = "hub_int_dialog_minievent",
    dialog   = "obj_int_dialog_minievent",
    window   = "anm_int_dialog_minievent_window",
    bg       = "anm_int_dialog_minievent_window_bg",
    decor    = "anm_int_dialog_minievent_window_decor",
    header   = "anm_int_dialog_minievent_window_header",
    footer   = "anm_int_dialog_minievent_window_footer",
    finished = "anm_int_dialog_minievent_window_finished",
    suspended = "anm_int_dialog_minievent_window_suspended",
    button = {
      close   = "btn_int_dialog_minievent_close",
      play    = "btn_int_dialog_minievent_play",
      skip    = "btn_int_dialog_minievent_skip",
      finish  = "btn_int_dialog_minievent_finished_ok",
      suspend = "btn_int_dialog_minievent_suspended_ok",
      reward = {
        point4 = "btn_int_dialog_minievent_reward_point4_get",
        point7 = "btn_int_dialog_minievent_reward_point7_get",
        final  = "btn_int_dialog_minievent_reward_final_get"
      }
    },
    timer  = { 
      duration = "tmr_int_dialog_minievent_duration",
      cooldown = "tmr_int_dialog_minievent_cooldown",
      animation = "tmr_int_dialog_minievent_animation"
    },
    reward = {
      anim   = "anm_int_dialog_minievent_reward",
      sprite = "spr_int_dialog_minievent_reward"
    },
    marker = {
      anim = "anm_int_dialog_minievent_avatar",
      icon = "anm_int_dialog_minievent_avatar_icon"
    },
    character = "anm_int_dialog_minievent_character",
    activity = "anm_int_dialog_minievent_activity",
    progressbar = "anm_int_dialog_minievent_target_progressbar",
    point = "anm_int_dialog_minievent_point",
    character_dialog = {
      body       = "anm_int_dialog_minievent_window_character_dialog",
      character  = "anm_int_dialog_minievent_window_character_dialog_character",
      item       = "anm_int_dialog_minievent_window_character_dialog_item"
    },
    book = {
      body   = "anm_int_dialog_minievent_window_book",
      button = "btn_int_dialog_minievent_window_book_play"
    },
    point_count = 0,
    stage_last  = {},
    anm_progress_hero = "anm_int_dialog_minievent_window_progress_hero"
  };

  private.app_o = ui.screen.Orientation();
  private.app_h = ui.screen.Height();
  private.app_w = ui.screen.Width();
  
  --private.size = { 
  --  [ "portrait"  ] = { w = 600, h = 700 }, 
  --  [ "landscape" ] = { w = 600, h = 700 } 
  --};
  
  --[[ui.obj.Set( "obj_int_dialog_minievent_dragger",
  {
    inputrect_init =   true,
    inputrect_x    = - private.size[ private.app_o ].w / 2,
    inputrect_y    = - private.size[ private.app_o ].h / 2,
    inputrect_w    =   private.size[ private.app_o ].w,
    inputrect_h    =   private.size[ private.app_o ].h,

    drag            = true,
    realdrag        = true,
    inputrect_init  = true,
    
    event_startdrag = private.Map_ScrollStart,
    event_dragdrop  = private.Map_ScrollEnd,
    event_drag      = private.Map_Scroll,
  } );]]

  private.Adapt();
  private.LoadStrings();
  private.CreateButtons();
  private.CreatePoints();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Show ( minievent_data, skin_data )

  --DbgTrace( minievent_data );

  if ( _this.is_active ) then return; end;

  _this.is_active    = true;
  _this.is_suspended = minievent_data.is_suspended;
  _this.rewards_showed = false;
  
  _this.is_reward      = minievent_data.is_reward;
  _this.minievent_name = minievent_data.name;
  _this.minievent_type = minievent_data.type;
  _this.scene_number   = minievent_data.scene or 1;
  _this.stage_count    = #minievent_data.stages;

  private.player_skin = ui.skin.Attach(
    _this.anm_progress_hero,
    _f.GetCurrentSkin()
  );
  
  ui.obj.Set(
    private.player_skin,
    {
      animfunc = "dialog",
      playing  = true,
      anim_tag = "at_hero"
    }
  );
  
  if ( minievent_data.animate ) then

    _this.is_animating = true;

  else

    private.StageLast_Set( minievent_data.id, #minievent_data.stages_completed + 1 );    
    
    if ( minievent_data.type ~= "chests" ) then
    
      private.AttachPoints( _this.stage_count );
    
    end;
  
    private.Defaults_Set( minievent_data, skin_data );

  end;

  private.Header_Footer_Reset();
  private.Window_Set( minievent_data );
  private.Timer_SetDuration();
  
  ui.window.Open(
    "minievent",
    function ()

      if ( not minievent_data.is_reward ) then

        if ( minievent_data.is_cooldown ) then _f.ShowTopPanel(); end;

        private.Header_Footer_On();

      end;

      if ( minievent_data.is_finished ) then

        if ( minievent_data.is_reward ) then

          private.Animate( minievent_data, skin_data );

        else

          private.Window_ShowFinished();

        end;

      else

        private.Book_TryShow(
          minievent_data,
          function ()

            private.Character_TryShowDialog(
              minievent_data,
              "start",
              function () private.Animate( minievent_data, skin_data ); end
            );

          end
        );

      end;

      _f.OnShow();

    end
  );

end;
--------------------------------------------------------------------------------
function public.Hide ( func_OnHide, is_instantly )

  if ( not _this.is_active ) then return; end;

  _this.is_active    = false;
  _this.is_suspended = false;

  ui.input.Off( _this.dialog );
  
  local func_OnPanelHide = function ()

    if ( not is_instantly ) then

      private.Header_Footer_Off();

      ui.window.Close(
        "minievent",
        function ()

          ui.skin.Detach( private.player_skin );
          private.Header_Footer_Reset();
          ui.input.On( _this.dialog );
          if ( func_OnHide ) then func_OnHide(); end;

        end
      );

    elseif ( func_OnHide ) then
      ui.skin.Detach( private.player_skin );
      ui.input.On( _this.dialog );
      if ( func_OnHide ) then func_OnHide(); end;

    end;

  end;

  _f.HideTopPanel( func_OnPanelHide, true );

end;
--------------------------------------------------------------------------------
function public.Update ( minievent_data, func_update )

  if ( _this.is_reward ) then return; end;

  if ( _this.is_animating ) then

    private.func_Update = func_update;

    return;

  end;

  if ( _this.rewards_showed ) then
    return;
  end;
  
  if ( not minievent_data         ) or
     ( minievent_data.is_finished ) then

    private.Window_ShowFinished();
    
    return;
    
  end;

  if ( minievent_data.is_suspended ) then

    private.Window_ShowSuspended();
    return;

  elseif ( _this.is_suspended ) then

    private.Window_HideSuspended();

  end;

  _this.point_count    = #minievent_data.stages;
  _this.minievent_name = minievent_data.name;

  if ( minievent_data.type ~= "chests" ) then
  
    private.AttachPoints( _this.point_count );

  end;

  private.Defaults_Set( minievent_data );
  private.Window_Set( minievent_data );
  private.Timer_SetDuration();

  if ( not _this.is_header_footer_on ) then

    private.Header_Footer_On();

  end;

  ui.input.On( _this.window );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Adapt ()
  
  local multiset = {};

  ui.obj.ToCenter( _this.dialog, 0, 0, multiset );

  ui.obj.Set(
    _this.window,
    {
      animfunc = private.app_o,
      pos_y    = tern( private.app_o == "landscape" )( 0, - private.app_h / 2 )
    },
    multiset
  );

  if ( private.app_o == "portrait" ) then

    ui.obj.Disable( _this.decor, false, multiset );

  end;

  ui.obj.Set(
    _this.header,
    {
      pos_y = - private.app_h / 2
    },
    multiset
  );

  ui.obj.Set(
    _this.footer,
    {
      pos_y = private.app_h / 2
    },
    multiset
  );

  ui.obj.SetTextTag(
    _this.finished,
    "tt_desc",
    "str_int_dialog_minievent_finished",
    multiset
  );

  ui.obj.SetTextTag(
    _this.suspended,
    "tt_desc",
    "str_int_dialog_minievent_suspended",
    multiset
  );

  ui.obj.SetTextTag(
    _this.reward.anim.."_final",
    "tt_reward_title",
    "str_int_dialog_minievent_reward_title",
    multiset
  );

  ui.obj.Set(
    _this.character_dialog.body,
    {
      inputrect_init = true,
      inputrect_x    = private.app_w / -2,
      inputrect_y    = private.app_h / -2,
      inputrect_w    = private.app_w,
      inputrect_h    = private.app_h
    },
    multiset
  );

  ui.obj.Set(
    _this.book.body,
    {
      inputrect_init = true,
      inputrect_x    = private.app_w / -2,
      inputrect_y    = private.app_h / -2,
      inputrect_w    = private.app_w,
      inputrect_h    = private.app_h
    },
    multiset
  );

  ui.control.EventsSet(
    _this.character_dialog.body,
    {
      func_on_click = function () private.Character_NextDialog(); end
    },
    multiset
  );

  ui.control.EventsSet(
    _this.character,
    {
      func_on_click = function () private.Character_OnTap(); end
    },
    multiset
  );

  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
function private.LoadStrings ()

  _this.strings = {
    dialog_skip = ne.strings.Get( "str_int_dialog_minievent_dialog_skip" )
  };

end;
--------------------------------------------------------------------------------
function private.CreateButtons ()
  
  ui.button.Create(
    _this.button.close,
    {
      design = "close"
    },
    function() _f.Close(); end
  );

  ui.button.Create(
    _this.button.finish,
    {
      title  = "str_int_dialog_minievent_finished_ok",
      design = "medium",
      color  = "green"
    },
    function() _f.Close(); end
  );

  ui.button.Create(
    _this.button.suspend,
    {
      title  = "str_int_dialog_minievent_suspended_ok",
      design = "medium",
      color  = "green"
    },
    function() _f.Close(); end
  );
  
  ui.button.Create(
    _this.button.play,
    {
      title  = "str_int_dialog_minievent_play",
      design = "medium",
      color  = "green"
    },
    function () _f.ShowStartDialog(); end
  );
  
  ui.button.Create(
    _this.button.skip,
    {
      title  = "str_int_dialog_minievent_skip",
      design = "wheel",
      timer  = "down"
    },
    function () private.Cooldown_Skip(); end
  );

  ui.button.Create(
    _this.button.reward.final,
    {
      title  = "str_int_dialog_minievent_get_reward",
      design = "medium",
      color  = "green"
    },
    function() private.Reward_Get( "final" ); end
  );
  
end;
--------------------------------------------------------------------------------
function private.CreatePoints ( points_count )
  
  local end_count   = points_count or 10;
  local begin_count = tern( points_count )( end_count - 9, 1 );
  
  local multiset = {};

  for i = begin_count, end_count do
    
    ui.obj.Clone(
      "anm_int_dialog_minievent_point",
      _this.hub, 
      "_"..i
    );
    
    ui.obj.Set(
      "anm_int_dialog_minievent_point_"..i,
      {
        anim_tag = "at_event_stage"..i
      },
      multiset
    );
    
  end;
  
  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.AttachPoints ( point_count )

  if ( point_count > 10 ) then
    
    private.CreatePoints( point_count );
    
  end;
  
  for i = 1, point_count do
    
    ui.obj.Attach(
      "anm_int_dialog_minievent_point_"..i,
      "anm_int_dialog_minievent_window_bg"
    );
    
  end;

  _this.point_count = point_count;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Animate ( minievent_data, skin_data )

  --DbgTrace( "Animate" );
  --DbgTrace( minievent_data );

  if ( minievent_data.animate ) then
    
    ui.input.Off( _this.window );
    
    if     ( minievent_data.animate == "win" ) then

      private.Animate_OnWin( minievent_data, skin_data );
    
    elseif ( minievent_data.animate == "fail" ) then
      
      private.Animate_OnFail( minievent_data );
      
    else
      
      private.Animate_OnEnd( minievent_data );
      
    end;

  else

    mediator.Notify( "Interface_Dialog_Minievent_Ready" );
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.Animate_OnWin ( minievent_data, skin_data )

  --DbgTrace( "Animate_OnWin" );

  local callbacks = {
    function ( func_next )

      -- JK: in story, move the activity forward
      if ( minievent_data.type == "story" ) and
         ( minievent_data.activity        ) then

        local stage_index = private.StageLast_Get( minievent_data.id );

        ui.anim.Play(
          _this.activity,
          "move"..stage_index,
          function ()

            if ( _this.stage_count == stage_index ) then

              if ( minievent_data.activity.hide_on_end ) then

                ui.obj.Disable( _this.activity );

              end;

              if ( minievent_data.character ) then

                local pose_index = minievent_data.character.end_pose;
                ui.anim.Play( _this.character, "static"..pose_index );

              end;

            end;
            
            if ( minievent_data.nature == "location" ) then
              _f.CheckRewardStage( func_next );
            else
              func_next();
            end;

          end,
          tern( _this.stage_count < stage_index )(
            "static"..( stage_index + 1 ),
            nil
          )
        );

      else

        func_next();

      end;

    end
  };

  table.insert(
    callbacks,
    function ( func_next )

      private.Character_TryShowDialog( minievent_data, "win", func_next );

    end
  );

  table.insert(
    callbacks,
    function ( func_next )

      local reward_group = minievent_data.reward_group;
      local is_reward    = minievent_data.is_reward;

      if ( is_reward ) then

        local funcs = {
          point4 = private.Animate_OnWin_ProgressReward,
          point7 = private.Animate_OnWin_ProgressReward,
          chests = private.Animate_OnWin_ChestReward,
          final  = private.Animate_OnWin_FinalReward
        };

        local func_Reward = funcs[ reward_group ];

        if ( func_Reward ) then func_Reward( minievent_data ); end;

      else

        private.Animate_OnWin_NoReward( minievent_data );
        
      end;

    end
  );

  common.RunCallbackList( callbacks );

end;
--------------------------------------------------------------------------------
function private.Animate_OnWin_FinalReward ( minievent_data )

  if ( minievent_data.type == "story" ) then

    private.Animate_OnWin_FinalReward_Chest( minievent_data );

  else

    private.Animate_OnWin_FinalReward_Default( minievent_data );

  end;

end;
--------------------------------------------------------------------------------
function private.Animate_OnWin_FinalReward_Default ( minievent_data )

  --DbgTrace( "Animate_OnWin_FinalReward_Default" );

  local callbacks = {
    function ( func_on_end_anim )

      local stage_last = private.StageLast_Get( minievent_data.id );

      private.Point_Complete( stage_last, func_on_end_anim );

    end
  };

  table.insert(
    callbacks,
    function ( func_on_end_anim )

      private.Reward_Appear( minievent_data.reward_group, func_on_end_anim );

    end
  );

  table.insert(
    callbacks,
    function ( func_on_end_anim )

      if ( not minievent_data.is_finished ) and
         ( minievent_data.is_suspended    ) then

        private.Window_ShowSuspended();

      end;

      func_on_end_anim();

    end
  );

  table.insert(
    callbacks,
    function ()

      private.Animate_OnEnd( minievent_data );

    end
  );

  common.RunCallbackList( callbacks );

end;
--------------------------------------------------------------------------------
function private.Animate_OnWin_FinalReward_Chest ( minievent_data )

  local callbacks = {
    function ( func_on_end_anim )

      local stage_last = private.StageLast_Get( minievent_data.id );

      private.Point_Complete( stage_last, func_on_end_anim );

    end
  };

  table.insert(
    callbacks,
    function ( func_on_end_anim )

      local chests = minievent_data.reward.final;

      private.Reward_Chests( chests, false, true, func_on_end_anim );

    end
  );

  table.insert(
    callbacks,
    function ( func_on_end_anim )

      _f.SwitchScene();

    end
  );

--  table.insert(
--    callbacks,
--    function ( func_on_end_anim )
--
--      if ( not minievent_data.is_finished ) and
--         ( minievent_data.is_suspended    ) then
--
--        private.Window_ShowSuspended();
--
--      end;
--
--      func_on_end_anim();
--
--    end
--  );
--
--  table.insert(
--    callbacks,
--    function ( func_on_end_anim )
--
--      private.Animate_OnEnd( minievent_data, func_on_end_anim );
--
--    end
--  );
--
--  table.insert(
--    callbacks,
--    function ( func_on_end_anim )
--
--      if ( minievent_data.is_finished ) then
--
--        private.Window_ShowFinished();
--
--      end;
--
--    end
--  );

  common.RunCallbackList( callbacks );

end;
--------------------------------------------------------------------------------
function private.Animate_OnWin_ProgressReward ( minievent_data )

  --DbgTrace( "Animate_OnWin_ProgressReward" );

  local stage_last = private.StageLast_Get( minievent_data.id );

  local callbacks = {
    function ( func_on_end_anim )

      local stage_last = private.StageLast_Get( minievent_data.id );

      private.Point_Complete( stage_last, func_on_end_anim );

    end
  };

  table.insert(
    callbacks,
    function ( func_on_end_anim )

      private.Reward_Appear( minievent_data.reward_group, func_on_end_anim );

    end
  );
  
  if ( stage_last < _this.stage_count ) then

    table.insert(
      callbacks,
      function ( func_on_end_anim )

        private.Point_Active( stage_last + 1, func_on_end_anim );

      end
    );

    table.insert(
      callbacks,
      function ( func_on_end_anim )

        private.Avatar_Hide( func_on_end_anim );

      end
    );

    table.insert(
      callbacks,
      function ( func_on_end_anim )

        private.Avatar_Show( stage_last + 1, func_on_end_anim );

      end
    );

  end;

  table.insert(
    callbacks,
    function ( func_on_end_anim )

      if ( not minievent_data.is_finished ) and
         ( minievent_data.is_suspended    ) then

        private.Window_ShowSuspended();

      end;

      func_on_end_anim();

    end
  );

  table.insert(
    callbacks,
    function ( func_on_end_anim )

      private.Animate_OnEnd( minievent_data, func_on_end_anim );

    end
  );

  table.insert(
    callbacks,
    function ( func_on_end_anim )

      if ( minievent_data.is_finished ) then

        private.Window_ShowFinished();

      end;

    end
  );

  common.RunCallbackList( callbacks );

end;
--------------------------------------------------------------------------------
function private.Animate_OnWin_TargetReward ( minievent_data, func_end )

  local callbacks = {
    function ( func_next )

      private.Reward_Target( func_next );

    end
  };

  table.insert( callbacks, func_end );
  common.RunCallbackList( callbacks );
  
end;
--------------------------------------------------------------------------------
function private.Animate_OnWin_ChestReward ( minievent_data )

  --DbgTrace( "Animate_OnWin_ChestReward" );

  local is_scene_done = ( minievent_data.type == "story" );

  local callbacks = {};

  if ( minievent_data.nature == "stages" ) then

    table.insert(
      callbacks,
      function ( func_on_end_anim )

        local stage_last = private.StageLast_Get( minievent_data.id );

        private.Point_Complete( stage_last, func_on_end_anim );

      end
    );

  end;

  table.insert(
    callbacks,
    function ( func_on_end_anim )

      local chests      = minievent_data.reward.chests;
      local is_cooldown = minievent_data.is_cooldown;

      private.Reward_Chests(
        chests,
        is_cooldown,
        is_scene_done,
        func_on_end_anim
      );

    end
  );

  if ( minievent_data.is_target_reward ) then

    table.insert(
      callbacks,
      function ( func_next )

        private.Animate_OnWin_TargetReward( minievent_data, func_next );

      end
    );

  end;

  if ( not is_scene_done ) then

    table.insert(
      callbacks,
      function ( func_on_end_anim )

        if ( not minievent_data.is_finished ) and
           ( minievent_data.is_suspended    ) then

          private.Window_ShowSuspended();

        end;

        func_on_end_anim();

      end
    );

    table.insert(
      callbacks,
      function ( func_on_end_anim )

        private.Animate_OnEnd( minievent_data, func_on_end_anim );

      end
    );

    table.insert(
      callbacks,
      function ( func_on_end_anim )

        _f.CheckShare( func_on_end_anim );

      end
    );
    
    table.insert(
      callbacks,
      function ( func_on_end_anim )

        if ( minievent_data.is_finished ) then

          private.Window_ShowFinished();

        end;

      end
    );

  else

    table.insert(
      callbacks,
      function ( func_on_end_anim )

        _f.CheckShare( func_on_end_anim );

      end
    );
    
    table.insert(
      callbacks,
      function ( func_on_end_anim )

        _f.Close();

      end
    );

  end;

  common.RunCallbackList( callbacks );

end;
--------------------------------------------------------------------------------
function private.Animate_OnWin_NoReward ( minievent_data )

  --DbgTrace( "Animate_OnWin_NoReward" );

  local stage_last = private.StageLast_Get( minievent_data.id );

  local callbacks = {};

  if ( minievent_data.nature == "stages" ) then

    table.insert(
      callbacks,
      function ( func_on_end_anim )

        local stage_last = private.StageLast_Get( minievent_data.id );

        private.Point_Complete( stage_last, func_on_end_anim );

      end
    );

    if ( stage_last < _this.stage_count ) then

      table.insert(
        callbacks,
        function ( func_on_end_anim )

          private.Point_Active( stage_last + 1, func_on_end_anim );

        end
      );

      table.insert(
        callbacks,
        function ( func_on_end_anim )

          private.Avatar_Hide( func_on_end_anim );

        end
      );

      table.insert(
        callbacks,
        function ( func_on_end_anim )

          private.Avatar_Show( stage_last + 1, func_on_end_anim );

        end
      );

    end;

  end;

  table.insert(
    callbacks,
    function ( func_on_end_anim )

      if ( not minievent_data.is_finished ) and
         ( minievent_data.is_suspended    ) then

        private.Window_ShowSuspended();

      end;

      func_on_end_anim();

    end
  );

  table.insert(
    callbacks,
    function ( func_on_end_anim )

      private.Animate_OnEnd( minievent_data, func_on_end_anim );

    end
  );

  table.insert(
    callbacks,
    function ( func_on_end_anim )

      if ( minievent_data.is_finished ) then

        private.Window_ShowFinished();

      end;

    end
  );

  common.RunCallbackList( callbacks );

end;
--------------------------------------------------------------------------------
function private.Animate_OnFail ( minievent_data )

  --DbgTrace( "Animate_OnFail" );

  local callbacks = {};

  --- Not discarded ------------------------------------------------------------
  if ( not minievent_data.discard ) then

    table.insert(
      callbacks,
      function ( func_on_end_anim )

        if ( not minievent_data.is_finished ) and
           ( minievent_data.is_suspended    ) then

          private.Window_ShowSuspended();

        end;

        func_on_end_anim();

      end
    );

    table.insert(
      callbacks,
      function ()

        private.Animate_OnEnd( minievent_data );

      end
    );
    
    common.RunCallbackList( callbacks );
    
    return;

  end;

  --- Discarded ----------------------------------------------------------------
  callbacks = {
    function ( func_on_end_anim )

      local stage_last = private.StageLast_Get( minievent_data.id );

      private.Point_InactiveCascade(
        stage_last,
        minievent_data.id,
        func_on_end_anim
      );

    end,
    function ( func_on_end_anim )

      private.Point_Active( 1, func_on_end_anim );

    end,
    function ( func_on_end_anim )

      private.Avatar_Hide( func_on_end_anim );

    end,
    function ( func_on_end_anim )

      private.Avatar_Show( 1, func_on_end_anim );

    end,
    function ()

      private.Animate_OnEnd( minievent_data );

    end
  };
  
  if ( private.StageLast_Get( minievent_data.id ) == 1 ) then

    callbacks = {
      function ()

        private.Animate_OnEnd( minievent_data );

      end
    };

  end;

  table.insert(
    callbacks,
    #callbacks,
    function ( func_on_end_anim )

      if ( not minievent_data.is_finished ) and
         ( minievent_data.is_suspended    ) then

        private.Window_ShowSuspended();

      end;

      func_on_end_anim();

    end
  );
  
  common.RunCallbackList( callbacks );

end;
--------------------------------------------------------------------------------
function private.Animate_OnEnd ( minievent_data, func_on_end_anim )

  --DbgTrace( "Animate_OnEnd" );

  _this.is_animating = nil;
  _this.is_reward    = nil;

  local reward_group = minievent_data.reward_group;
  local is_reward    = minievent_data.is_reward;

  if ( is_reward               ) and
     ( reward_group == "final" ) then

    private.Point_ClearEvents();
    ui.control.EventsClear( _this.reward.anim.."_final" );

  else

    private.StageLast_Set( minievent_data.id, #minievent_data.stages_completed + 1 );

    if ( not private.Callback_TryUpdate() ) then
      
      if ( is_reward               ) and
         ( reward_group ~= "final" ) then
           
        if ( minievent_data.type == "tourney" ) then
          minievent_data.reward_visible[ reward_group ] = nil;
        end;
      
      end;

      if ( minievent_data.type == "chests" ) then
        minievent_data.reward = _f.GetNextReward();
      end;

      private.Defaults_Set( minievent_data );

    end;

  end;

  ui.input.On( _this.window );
  mediator.Notify( "Interface_Dialog_Minievent_Ready" );

  if ( func_on_end_anim ) then func_on_end_anim(); end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Point_Complete ( point_index, func_on_end_anim )

  local is_reward_point = private.Point_IsReward( point_index );

  local animfunc = tern( is_reward_point )(
    "stage_reward_to_complete",
    "stage_to_complete"
  );

  local func_after = function ()
    ui.anim.Play(
      _this.point.."_"..point_index,
      animfunc,
      func_on_end_anim
    );
  end;
  
  _f.CheckRewardStage( func_after, point_index );
  
end;
--------------------------------------------------------------------------------
function private.Point_Active ( point_index, func_on_end_anim )
  
  local is_reward_point = private.Point_IsReward( point_index );

  local animfunc_start = tern( is_reward_point )(
    "stage_inactive_reward",
    "stage_to_active"
  );

  local animfunc_finish = tern( is_reward_point )(
    "stage_inactive_reward",
    "stage_active"
  );

  local func_end = function ()
    
    ui.obj.Set(
      _this.point.."_"..point_index,
      {
        animfunc = animfunc_finish,
        playing  = true
      }
    );
    
    func_on_end_anim();
    
  end;

  ui.anim.Play(
    _this.point.."_"..point_index,
    animfunc_start,
    func_end
  );
  
end;
--------------------------------------------------------------------------------
function private.Point_SetComplete ( minievent_data, multiset )
  
  for i = 1, #minievent_data.stages_completed do
    
    local stage = minievent_data.stages_completed[ i ];
    
    ui.control.EventsClear( _this.point.."_"..i );

    ui.obj.Set(
      _this.point.."_"..i,
      {
        anim_tag  = "at_event_stage"..i,
        animfunc  = "stage_complete",
        tt_number = i,
        playing   = true,
        endtrig   = function() end
      },
      multiset
    );

  end;
  
end;
--------------------------------------------------------------------------------
function private.Point_ClearEvents ()
  
  for i = 1, _this.point_count do
    
    ui.control.EventsClear( _this.point.."_"..i );

  end;
  
end;
--------------------------------------------------------------------------------
function private.Point_SetActive ( minievent_data, multiset )
  
  local stages_completed = minievent_data.stages_completed;
  local stages_total     = minievent_data.stages;
  
  for i = #stages_completed + 1, #stages_total do
    
    if ( i == #stages_completed + 1 ) then
      
      ObjAttach( _this.marker.anim, _this.point.."_"..i );

      local is_reward_point = private.Point_IsReward( i );

      local animfunc = tern( is_reward_point )(
        "stage_inactive_reward",
        "stage_active"
      );
     
      ui.obj.Set(
        _this.point.."_"..i,
        {
          anim_tag  = "at_event_stage"..i,
          animfunc  = animfunc,
          tt_number = i,
          playing   = true,
          endtrig   = ""
        },
        multiset
      );
      
      ui.control.EventsSet( _this.point.."_"..i, 
      {
        func_on_click = function () _f.ShowStartDialog(); end
      } );
      
    else

      local is_reward_point = private.Point_IsReward( i );

      local animfunc = tern( is_reward_point )(
        "stage_inactive_reward",
        "stage_inactive"
      );

      ui.obj.Set(
        _this.point.."_"..i,
        {
          animfunc  = animfunc,
          tt_number = i,
          playing   = true,
          endtrig   = ""
        },
        multiset
      );
      
      ui.control.EventsClear( "anm_int_dialog_minievent_point_"..i );
      
    end;

  end;
  
end;
--------------------------------------------------------------------------------
function private.Point_InactiveCascade ( point_index, event_id, func_on_end_anim )

  local animfunc;

  if ( private.Point_IsReward( point_index ) ) then

    animfunc = "stage_to_inactive_reward";

  else

    local stage_last = private.StageLast_Get( event_id );

    animfunc = tern( point_index == stage_last )(
      "stage_to_inactive1", 
      "stage_to_inactive2"
    );

  end;
  
  if ( point_index == 0 ) then

    func_on_end_anim();
    return;

  end;

  if ( point_index > 0 ) then 

    ui.anim.Play(
      _this.point.."_"..point_index,
      animfunc,
      function ()

        private.Point_InactiveCascade( point_index - 1, event_id, func_on_end_anim );

      end
    );

  end;                                                                                                        
  
end;
--------------------------------------------------------------------------------
function private.Point_IsReward ( point_index )

  -- JK: Tower -----------------------------------------------------------------
  if ( _this.minievent_type == "tower" ) then

    return ( point_index == _this.stage_count );

  end;
  
  -- JK: Tourney ---------------------------------------------------------------
  if ( _this.minievent_type == "tourney" ) then

    return (
      ( point_index == 4                 ) or
      ( point_index == 7                 ) or
      ( point_index == _this.stage_count )
    );

  end;
  
  -- JK: Chests ----------------------------------------------------------------
  if ( _this.minievent_type == "chests" ) then

    return false;

  end;
  
  -- JK: Story -----------------------------------------------------------------
  if ( _this.minievent_type == "story" ) then

    return ( point_index == _this.stage_count );

  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
function public.Point_Reward_ShowFireflies ( emitter, reward, func_end )
  
  local obj_fly   = "obj_map_flying_resources";
  local anm_fly   = "anm_map_flying_resources";
  local anm_res   = "anm_map_flying_resources_icon";
  local obj_anims = "obj_int_dialog_minievent_anims";
  local tmr_fly   = "tmr_int_dialog_minievent_resource";

  local emitter_obj = "anm_int_dialog_minievent_point_"..emitter;
  
  local emitter_pos  = private.Point_PositionToFireflies( emitter_obj );
  local count        = 0;
  local flight_times = {};
  local max_count    = math.min( reward.count, 15 );
  local tme          = { 0.45, 0.47, 0.51, 0.54, 0.58, 0.60, 0.64, 0.66 };
  
  local pattern = "_%s_%s_%s";
  
  for j = 1, max_count do

    count = count + 1;
    
    local suffix = pattern:format( emitter, reward.name, count );

    local obj_name = obj_fly..suffix;
    local anm_name = anm_fly..suffix;
    local res_name = anm_res..suffix;
    local tmr_name = tmr_fly..suffix;

    local index = math.random( 1, 8 );
    local pos_x = math.random( - 150, 150 );
    local pos_y = math.random( - 80, 80 );

    ui.obj.Clone( obj_fly, obj_anims, suffix, true );
    
    local multiset = {};
    
    ui.obj.Set(
      obj_name,
      {
        pos_x   = emitter_pos[ 1 ],
        pos_y   = emitter_pos[ 2 ],
        pos_z   = ( 10 + index ),
        scale_x = 0,
        scale_y = 0
      },
      multiset
    );
    
    ui.obj.Set(
      res_name, 
      {
        animfunc = reward.name,
        playing  = true
      },
      multiset
    );

    local delay = math.random( 100, 500 ) / 1000;

    ui.obj.Create( tmr_name, "timer"  );
    ui.obj.Attach( tmr_name, obj_name );
    
    local res_type = tern( reward.name:find( "gold" ) )( "gold", "resource" );
    
    ui.obj.Set(
      tmr_name, 
      {
        time    = delay,
        playing = true,
        endtrig = function() ui.anim.Play( anm_name, res_type..index ); end
      },
      multiset
    );
    
    ui.obj.Multiset( multiset );

    local fly_time = tme[ index ] + delay + 0.5;

    table.insert( flight_times, fly_time );

    ObjAnimate( 
      obj_name, "pos_xy", 0, 0,
      function () 
        
        local name_without_prefix = ui.obj.GetNameWithoutPrefix( obj_name );
        
        private.Point_Reward_Get( name_without_prefix, reward );
        
      end, 
      {
        0.0,            1, emitter_pos[ 1 ],         emitter_pos[ 2 ],
        delay,          1, emitter_pos[ 1 ],         emitter_pos[ 2 ],
        fly_time - 0.5, 1, emitter_pos[ 1 ] + pos_x, emitter_pos[ 2 ] + pos_y,
        fly_time,       1, emitter_pos[ 1 ] + pos_x, emitter_pos[ 2 ] + pos_y
      }
    );

    ObjAnimate( obj_name, "scale_xy", 0, 0, "", 
      {
        0.00, 0, 0.0, 0.0,
        0.10, 0, 1.2, 1.2,
        0.20, 0, 0.8, 0.8,
        0.25, 0, 1.1, 1.1,
        0.3,  0, 1.0, 1.0      
      }
    );

  end;
  
  table.sort( flight_times );
  
  local max_flight_time = flight_times[ #flight_times ] + 0.5;
  local tmr_reward      = "tmr_int_dialog_minievent_"..reward.name;
  
  ui.obj.Attach( tmr_reward, obj_anims );
  
  local func_OnTimerEnd = function()

    ObjDetach( tmr_reward );
    if ( func_end ) then func_end(); end;

  end;

  ui.obj.Set(
    tmr_reward,
    {
      time    = max_flight_time,
      playing = true,
      endtrig = func_OnTimerEnd
    }
  );
  
  common.sound.Play( "assets/content/map/sounds/aud_fireflies" );
  
end;
--------------------------------------------------------------------------------
function private.Point_Reward_Get ( emitter, reward_data )

  local resource_name = reward_data.name;

  local obj_name = "obj_"..emitter;
  local pfx_name = "pfx_"..emitter;
  local anm_name = "anm_"..emitter;

  ObjStopAnimate( obj_name, "pos_x" );

  local start_pos = ObjGet( obj_name );
  --ES>
  if ( not v_interface.Resource_IsPanelActive( resource_name ) ) then

    if ( string.find( resource_name, "gold" ) ) then

      v_interface.Gold_ShowPanel();

    else

      v_interface.Resource_ShowPanel( resource_name );

    end;

  end;
  
  local stop_pos  = ui.obj.Position( 
    "obj_int_panel_interface_top_"..resource_name,
    "obj_int_dialog_minievent_anims"
  );
  -->ES> блок выше переделать как будет панель для красных светляков
  if ( string.find( resource_name, "resource" ) ) then

    stop_pos[ 2 ] = stop_pos[ 2 ] + 140;

  end;

  local func_end = function ()

    ObjDelete( obj_name );

  end;

  ObjSet( pfx_name, { visible = true, active = true } );

  PartSysStart( pfx_name );

  ObjAnimate( obj_name, "pos_xy", 0, 0, func_end, 
  {
    0.0, 0, start_pos.pos_x, start_pos.pos_y,
    0.5, 0, stop_pos[ 1 ], stop_pos[ 2 ]
  } );

  ObjAnimate( obj_name, "alp", 0, 0, "", 
  {
    0.00, 1, 1,
    0.45, 1, 1,
    0.50, 1, 0,
  } );

end;
--------------------------------------------------------------------------------
function private.Point_PositionToFireflies ( emitter_obj )
  
  local emitter_pos = ui.obj.Position( emitter_obj, InterfaceName );
  
  emitter_pos[ 1 ] = emitter_pos[ 1 ] - 0.5 * private.app_w;
  emitter_pos[ 2 ] = emitter_pos[ 2 ] - 0.5 * private.app_h;
  
  return emitter_pos;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Avatar_Hide ( func_on_end_anim )

  ui.anim.Play(
    _this.marker.anim,
    "avatar_disappear",
    func_on_end_anim
  );  
  
end;
--------------------------------------------------------------------------------
function private.Avatar_Show ( point_index, func_on_end_anim )

  local point_anim = _this.point;
  
  ui.obj.Attach(
    _this.marker.anim,
    point_anim.."_"..point_index
  );
  
  ui.anim.Play(
    _this.marker.anim,
    "avatar_appear",
    func_on_end_anim,
    "avatar_static"
  );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Reward_Set ( minievent_data, multiset )

  local reward_groups = {
    "point4",
    "point7",
    "chests",
    "final"
  };

  local reward_placeholders = {
    "point4",
    "point7",
    "chest1",
    "chest2",
    "chest3",
    "final"
  };

  for _, placeholder in pairs( reward_placeholders ) do

    local group = tern( placeholder:find( "chest" ) )( "chests", placeholder );

    if ( not minievent_data.reward[ group ] ) then

      ui.obj.Disable( _this.reward.anim.."_"..placeholder, false, multiset );

    else

      ui.obj.Enable( _this.reward.anim.."_"..placeholder, false, multiset );

    end;

  end;

  for g, data in pairs( minievent_data.reward ) do

    --- JK: Final reward -------------------------------------------------------
    if ( g == "final"               ) and
       ( minievent_data.reward[ g ] ) then

      private.Reward_SetFinal( minievent_data, g, data, multiset, true );
       
    --- JK: Point reward  ------------------------------------------------------
    elseif ( g:find( "point" )          ) and
           ( minievent_data.reward[ g ] ) then

      private.Reward_SetPoint( minievent_data, g, multiset );
    
    --- JK: chest rewards ------------------------------------------------------
    elseif ( g == "chests"              ) and
           ( minievent_data.reward[ g ] ) then
    
      private.Reward_SetChests( minievent_data, g, multiset );
      
    end;
  
  end;

  --- JK: target reward is final-like ------------------------------------------
  if ( minievent_data.targets ) then

    local g      = "final";
    local target = minievent_data.targets[ 1 ];
    local reward = target.reward;

    if ( not target.is_completed         ) or
       ( minievent_data.is_target_reward ) then

      local goal_count      = minievent_data.targets[ 1 ].goal.count;
      local target_progress = minievent_data.targets[ 1 ].progress;
      local stages_count    = #minievent_data.stages;

      local is_reachable = ( stages_count >= ( goal_count - target_progress ) );

      if ( is_reachable ) then
       
        private.Reward_SetFinal( minievent_data, g, { reward }, multiset, false );

      end;

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Reward_SetFinal ( minievent_data, g, data, multiset, is_clickable )
  
  ui.obj.Disable( _this.reward.anim.."_"..g.."_heart", nil, multiset );
  ui.obj.Disable( _this.reward.anim.."_"..g.."_slot1", nil, multiset );
  ui.obj.Disable( _this.reward.anim.."_"..g.."_slot2", nil, multiset );
  ui.obj.Disable( _this.reward.anim.."_"..g.."_slot3", nil, multiset );

  if ( data[ 1 ].type == "chest" ) then

    is_clickable = false;

    local anim_tag_postfix = minievent_data.chest_variant or "";

    private.Reward_SetFinal_Chest( g, data, anim_tag_postfix, multiset );

  else

    private.Reward_SetFinal_Default( g, data, multiset );

  end;
  
  if ( is_clickable ) then
  
    -- JK: Set final reward tap event
    ui.control.EventsSet(
      _this.reward.anim.."_"..g,
      {
        func_on_click = function ()

          if ( not _this.is_reward ) then

            _f.ShowPreviewDialog( g );

          end;

        end
      },
      multiset
    );
    
  else
    
    ui.control.EventsClear( _this.reward.anim.."_"..g );
  
  end;
  
end;
--------------------------------------------------------------------------------
function private.Reward_SetFinal_Default ( g, data, multiset )

  local slot_number         = 1;
  private.top_panel_updates = nil;

  for _, reward in pairs( data ) do

    local slot_name = _this.reward.anim.."_"..g.."_slot"..slot_number;

    if ( reward.type == "heart" ) then

      ui.obj.Set(
        _this.reward.anim.."_"..g.."_heart",
        {
          active  = true,
          visible = true,
          tt_time = ui.text.FormatTime( reward.count ),
          tt_sign = ui.text.InfinitySign(),
        },
        multiset
      );

    else
      
      if ( reward.type == "gold"     ) or
         ( reward.type == "resource" ) then
        private.top_panel_updates = private.top_panel_updates or {};
        table.insert( private.top_panel_updates, reward.name );
      end;
      
      ui.obj.Set(
        slot_name,
        {
          active     = true,
          visible    = true,
          animfunc   = "reward_"..reward.type,
          rt_reward  = ui.obj.Res( reward.type, reward.name ),
          tt_count   = tern( reward.count > 1 )(
            ui.text.FormatNumber( reward.count ),
            ""
          )
        },
        multiset
      );

      slot_number = slot_number + 1;

    end;

  end;
  
  ui.obj.Enable( _this.reward.anim.."_final", false, multiset );
  ui.anim.Play( _this.reward.anim.."_final", "reward" );

end;
--------------------------------------------------------------------------------
function private.Reward_SetFinal_Chest ( g, data, anim_tag_postfix, multiset )

  ui.obj.Disable( _this.reward.anim.."_final", false, multiset );
  
  local chest_name = data[ 1 ].name;
  local anim_tag   = "at_chest1"..anim_tag_postfix;
  local anm_chest  = "anm_int_dialog_minievent_reward_chest1";
  local anm_icon   = "anm_int_dialog_minievent_reward_chest1_icon";

  ui.obj.Set(
    anm_chest,
    {
      anim_tag = anim_tag,
      animfunc = "chest_static",
      frame    = 0
    },
    multiset
  );

  ui.obj.Set(
    anm_icon,
    {
      animfunc     = chest_name.."_static_closed",
      reward       = "",
      click        = "",
      txt_reward_1 = "",
      txt_reward_2 = "",
      txt_reward_3 = ""
    },
    multiset
  );

  ui.control.EventsSet(
    anm_chest,
    {
      func_on_click = function ()

        ui.input.Off( _this.dialog );
        ui.anim.Play( anm_icon, chest_name.."_static_closed" );

        local func_end = function ()

          ui.obj.Set( anm_icon, { frame = 0 } );
          ui.anim.Stop( anm_icon );
          ui.input.On( _this.dialog );

        end;

        _f.ShowPreviewDialog( g, 1, func_end );

      end
    },
    multiset
  );

  ui.obj.Enable( anm_chest, false, multiset );

end;
--------------------------------------------------------------------------------
function private.Reward_SetPoint ( minievent_data, g, multiset )
  
  if ( minievent_data.reward_visible[ g ] ) then

    ui.obj.Enable( _this.reward.anim.."_"..g, false, multiset );

    ui.control.EventsSet(
      _this.reward.anim.."_"..g,
      {
        func_on_click = function () _f.ShowPreviewDialog( g ); end
      },
      multiset
    );

  else

    ui.obj.Disable( _this.reward.anim.."_"..g, false, multiset );
    ui.control.EventsClear( _this.reward.anim.."_"..g );

  end;
  
end;
--------------------------------------------------------------------------------
function private.Reward_SetChests ( minievent_data, g, multiset )
  
  local chests           = minievent_data.reward[ g ];
  local anim_tag_postfix = minievent_data.chest_variant or "";

  for id, chest in pairs( chests ) do

    local chest_name = chest.name;
    local anim_tag   = "at_chest"..id..anim_tag_postfix;
    local anm_chest  = _this.reward.anim.."_chest"..id;
    local anm_icon   = _this.reward.anim.."_chest"..id.."_icon";

    ui.obj.Set(
      anm_chest,
      {
        anim_tag = anim_tag,
        animfunc = "chest_static",
        frame    = 0
      },
      multiset
    );
    
    ui.obj.Set(
      anm_icon,
      {
        animfunc     = chest_name.."_static_closed",
        reward       = "",
        click        = "",
        txt_reward_1 = "",
        txt_reward_2 = "",
        txt_reward_3 = ""
      },
      multiset
    );

    ui.control.EventsSet(
      anm_chest,
      {
        func_on_click = function ()

          ui.input.Off( _this.dialog );

          ui.anim.Play(
            anm_icon,
            chest_name.."_static_closed"

          );

          local func_end = function ()

            ui.obj.Set(
              anm_icon,
              {
                frame = 0
              }
            );

            ui.anim.Stop( anm_icon );
            ui.input.On( _this.dialog );

          end;

          _f.ShowPreviewDialog( g, id, func_end );

        end
      },
      multiset
    );

  end;
  
end;
--------------------------------------------------------------------------------
function private.Reward_Appear ( reward_group, func_on_end_anim )

  if ( reward_group ~= "final" ) then

    _f.ShowRewardDialog(
      reward_group,
      nil,
      function ()

        private.Header_Footer_On();

        _this.is_reward = nil;

        if ( func_on_end_anim ) then func_on_end_anim(); end;

      end
    );

    private.Reward_Disappear( reward_group );

    return;

  end;

  local animfunc = {
    final = {
      appear = "appear_reward",
      static = "static_reward"
    }
  };

  if ( private.top_panel_updates ) then
    _f.ShowTopPanel( private.top_panel_updates );
  end;
  
  local multiset = {};

  ui.input.Off( _this.button.close, multiset );

  ui.obj.SetAnimTag(
    _this.reward.anim.."_"..reward_group,
    "at_getreward",
    multiset
  );

  ui.obj.Multiset( multiset );

  ui.obj.Attach(
    _this.reward.anim.."_"..reward_group,
    _this.window
  );
  
  common.sound.Play( "tourney_general_reward" );
  
  ui.anim.Play(
    _this.reward.anim.."_"..reward_group,
    animfunc[ reward_group ].appear,
    func_on_end_anim,
    animfunc[ reward_group ].static
  );
  
end;
--------------------------------------------------------------------------------
function private.Reward_Get ( reward_group )
  
  ui.screen.Lock( true );
  
  common.sound.Play( "tourney_reward_animation" );
  
  local func_on_end = function ()

    if ( reward_group == "final" ) then

      local func_after_close = function ()

        local multiset = {};

        ui.input.On( _this.button.close, multiset );

        ui.obj.SetAnimTag(
          _this.reward.anim.."_"..reward_group,
          "at_reward",
          multiset
        );

        ui.obj.Multiset( multiset );

        ui.obj.Attach(
          _this.reward.anim.."_"..reward_group,
          _this.bg
        );

        ui.anim.Play( _this.reward.anim.."_"..reward_group, "reward" );

      end;
      
      _f.CheckShare( function () _f.Close( func_after_close ); end );

    end;

    ui.screen.Lock( false );
    _this.is_reward = nil;

  end;

  if ( private.top_panel_updates ) then
    ui.anim.Play(
      _this.reward.anim.."_dummy",
      "get_reward_dummy",
      function () _f.UpdateTopPanel( private.top_panel_updates ); end
    );
  end;

  ui.anim.Play(
    _this.reward.anim.."_"..reward_group,
    "get_reward",
    function ()
      
      if ( private.top_panel_updates ) then
        _f.HideTopPanel( func_on_end );
      else
        func_on_end();
      end;

    end
  );

end;
--------------------------------------------------------------------------------
function private.Reward_Disappear ( reward_group )

  ui.obj.Disable( _this.reward.anim.."_"..reward_group );
  ui.control.EventsClear( _this.reward.anim.."_"..reward_group );

end;
--------------------------------------------------------------------------------
function private.Reward_Chests (
    chests,
    is_cooldown,
    is_scene_done,
    func_on_end_anim
  )

  --DbgTrace( "Reward_Chests" );

  local callbacks = {};

  for id, chest in pairs( chests ) do

    if ( chest.count > 0 ) then

      if ( chest.factor     ) and 
         ( chest.factor > 1 ) then

        local counter = _this.reward.anim.."_chest"..id.."_count";

        ui.obj.Set(
          counter,
          {
            active     = true,
            visible    = true,
            alp        = 0,
            tt_counter = chest.factor
          }
        );

        ui.anim.FadeIn( 0.3, counter );

      end;

      ui.anim.Play(
        _this.reward.anim.."_chest"..id.."_icon",
        chest.name.."_static_closed"
      );

    end;

  end;

  for id, chest in pairs( chests ) do

    if ( chest.count > 0 ) then

      table.insert(
        callbacks,
        function( func_next )

          ui.anim.Play(
            _this.reward.anim.."_chest"..id,
            "chest_get",
            function ()

              _f.ShowChestReward( id, func_next );

              local anm_icon = _this.reward.anim.."_chest"..id.."_icon";

              ui.obj.Set(
                anm_icon,
                {
                  frame = 0
                }
              );

              ui.anim.Stop( anm_icon );

            end
          );

        end
      );

    end;

  end;

  table.insert(
    callbacks,
    function ( func_next )

      _this.is_reward = nil;

      if ( not is_scene_done ) then

        private.Header_Footer_On( func_next );

      end;

      for id, chest in pairs( chests ) do

        if ( chest.count > 0 ) then

          if ( chest.factor     ) and 
             ( chest.factor > 1 ) then

            ui.obj.Set(
              _this.reward.anim.."_chest"..id.."_count",
              {
                active     = false,
                visible    = false,
                alp        = 0,
                tt_counter = ""
              }
            );

          end;

          if ( not is_scene_done ) then

            ui.anim.Play(
              _this.reward.anim.."_chest"..id,
              "chest_appear",
              nil,
              "chest_static"
            );

          end;

        end;

      end;

      if ( is_scene_done ) then

        func_next();

      end;

    end
  );

  if ( is_cooldown ) then

    table.insert(
      callbacks,
      function ( func_next )

        ui.anim.FadeIn( 0.6, _this.button.skip, nil, func_next );
        _f.ShowTopPanel();

      end
    );

  end;

  table.insert(
    callbacks,
    func_on_end_anim
  );

  common.RunCallbackList( callbacks );

end;
--------------------------------------------------------------------------------
function private.Reward_Target ( func_next )

  private.Header_Footer_Off();

  local reward_group = "final";

  local btn_event = ui.button.EventGet( _this.button.reward.final );

  ui.button.EventSet(
    _this.button.reward.final,
    function () private.Reward_Target_Get( func_next, btn_event ); end
  );

  local animfunc = {
    final = {
      appear = "appear_reward",
      static = "static_reward"
    }
  };

  if ( private.top_panel_updates ) then
    _f.ShowTopPanel( private.top_panel_updates );
  end;
  
  local multiset = {};

  ui.input.Off(
    {
      _this.button.close,
      _this.button.play,
      _this.button.skip,
      _this.reward.anim.."_chest1",
      _this.reward.anim.."_chest2",
      _this.reward.anim.."_chest3"
    },
    multiset
  );
  
  _this.rewards_showed = true;
  
  ui.obj.SetAnimTag(
    _this.reward.anim.."_"..reward_group,
    "at_getreward",
    multiset
  );

  ui.obj.Multiset( multiset );

  ui.obj.Attach(
    _this.reward.anim.."_"..reward_group,
    _this.window
  );
  
  common.sound.Play( "tourney_general_reward" );
  
  ui.anim.Play(
    _this.reward.anim.."_"..reward_group,
    animfunc[ reward_group ].appear,
    function() ui.input.On( _this.window ); end,
    animfunc[ reward_group ].static
  );

  ui.anim.FadeOut(
    0.6,
    _this.progressbar,
    function () ui.obj.Disable( _this.progressbar ); end
  );

  ui.anim.FadeOut(
    0.6,
    _this.button.skip
  );

end;
--------------------------------------------------------------------------------
function private.Reward_Target_Get ( func_next, btn_event )

  ui.screen.Lock( true );

  local anm_reward = _this.reward.anim.."_final";

  ui.anim.Play(
    anm_reward,
    "get_reward",
    function ()
      
      ui.button.EventSet( _this.button.reward.final, btn_event );

      ui.anim.FadeIn(
        0.6,
        _this.button.skip
      );

      local multiset = {};
      
      ui.input.On(
        {
          _this.button.close,
          _this.button.play,
          _this.button.skip,
          _this.reward.anim.."_chest1",
          _this.reward.anim.."_chest2",
          _this.reward.anim.."_chest3"
        },
        multiset
      );

      ui.obj.SetAnimTag( anm_reward, "at_reward", multiset );
      ui.obj.Multiset( multiset );
      
      ui.obj.Set(
        anm_reward,
        {
          active   = false,
          visible  = false,
          input    = false,
          alp      = 1,
          animfunc = "",
          frame    = 0,
          playing  = false
        }
      );
      
      private.Header_Footer_On(
        function ()
          
          ui.obj.Attach( anm_reward, _this.bg );
          
          ui.obj.Set(
            anm_reward,
            {
              active   = false,
              visible  = false,
              input    = false,
              alp      = 1,
              animfunc = "reward",
              frame    = 0,
              playing  = false
            }
          );
          
        end
      );

      if ( func_next ) then func_next() end;

      ui.screen.Lock( false );

    end
  );

end;
--------------------------------------------------------------------------------
function private.Header_Footer_Off ( func_after )

  if ( not _this.is_header_footer_on ) then

    if ( func_after ) then func_after(); end;
    return;

  end;

  _this.is_header_footer_on = nil;

  ui.anim.Play( _this.header, "header_disappear", func_after );
  ui.anim.Play( _this.footer, "footer_disappear" );

end;
--------------------------------------------------------------------------------
function private.Header_Footer_On ( func_after )
  
  if ( _this.is_suspended ) or
     ( _this.is_finished  ) then

    if ( func_after ) then func_after(); end;
    return;
    
  end;
  
  _this.is_header_footer_on = true;
  _this.is_header_animating = true;
  
  ui.anim.Play(
    _this.header,
    "header_appear",
    function()

      _this.is_header_animating = false;

      if ( func_after ) then func_after(); end;

    end
  );

  ui.anim.Play( _this.footer, "footer_appear" );

end;
--------------------------------------------------------------------------------
function private.Header_Footer_Reset ()

  _this.is_header_footer_on = nil;
  
  local multiset = {};

  ui.anim.Stop( _this.header, multiset );
  ui.anim.Stop( _this.footer, multiset );

  ui.obj.Set(
    _this.header,
    {
      animfunc = "header_appear",
      playing  = false,
      frame    = 0
    },
    multiset
  );

  ui.obj.Set(
    _this.footer,
    {
      animfunc = "footer_appear",
      playing  = false,
      frame    = 0
    },
    multiset
  );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Callback_TryUpdate ()
  
  if ( private.func_Update ) then

    private.func_Update();
    private.func_Update = nil;

    return true;

  end;

  return false;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.StageLast_Set ( event_id, stage_number )

  _this.stage_last[ "id_"..event_id ] = stage_number;

end;
--------------------------------------------------------------------------------
function private.StageLast_Get ( event_id )

  return _this.stage_last[ "id_"..event_id ] or 1;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Cooldown_Skip ()

  if ( _this.cooldown_processing ) then return; end;

  _this.cooldown_processing = true;

  if ( _f.SkipCooldown() ) then

    private.Cooldown_OnFinished();

  else

    _this.cooldown_processing = nil;

  end;

end;
--------------------------------------------------------------------------------
function private.Cooldown_SetTimer ()

  local seconds = _f.GetCooldownSeconds();
        seconds = math.max( 0, seconds );

  ui.button.SetTime(
    _this.button.skip,
    ui.text.FormatTime( seconds )
  );

  if ( seconds > 0 ) then

    common.Timer_Start( _this.timer.cooldown, 1.0, private.Cooldown_SetTimer );

    return;

  end;

  common.Timer_Stop( _this.timer.cooldown );

  if ( not _this.cooldown_processing ) then

    _this.cooldown_processing = true;

    private.Cooldown_OnFinished();

  end;
  
end;
--------------------------------------------------------------------------------
function private.Cooldown_OnFinished ()

  ui.input.Off( _this.dialog );

  local multiset = {};

  ui.button.Disable( _this.button.skip, multiset );
  ui.button.Enable(  _this.button.play, multiset );

  _f.HideTopPanel();

  ui.obj.Set(
    _this.header,
    {
      tt_title = "str_int_dialog_minievent_".._this.minievent_name.."_title"
    },
    multiset
  );

  ui.obj.Set(
    _this.footer,
    {
      tt_desc = "str_int_dialog_minievent_".._this.minievent_name.."_description"
    },
    multiset
  );

  ui.obj.Multiset( multiset );
  
  ui.input.On( _this.dialog );
  _this.cooldown_processing = nil;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Defaults_Set ( minievent_data, skin_data )
  
  local multiset = {};
  
  ui.input.On( _this.window, multiset );
  ui.obj.Disable( _this.finished, false, multiset );

  if ( not _this.is_suspended ) then

    ui.obj.Disable( _this.suspended, false, multiset );

  end;

  private.Reward_Set( minievent_data, multiset );
  private.Point_SetComplete( minievent_data, multiset );
  private.Point_SetActive( minievent_data, multiset );
  
  if ( skin_data ) then
    private.Marker_Set( skin_data, multiset );
  end;

  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Window_Set ( minievent_data )
  
  local minievent_name = minievent_data.name;
  local background     = minievent_data.background.."_background_"..private.app_o;

  --- 1st pass to get minievent full name
  local minievent_fullname = minievent_name:match( "^(.*)_%d+$" ) or
                             minievent_name;

  --- 2nd pass to get minievent full name
  minievent_fullname = minievent_fullname:match( "^(%w+)$" ) or
                       minievent_fullname:match( "^(%w+_%w+)$" );

  -- JK: path to minievent resources -------------------------------------------
  local res_path   = "assets/content/minievent_"..minievent_fullname.."/";
  local scene_path = res_path;
  
  if ( minievent_data.scene ) then

    scene_path = res_path.."scene_"..minievent_data.scene.."/";

  end;

  local multiset = {};

  if ( minievent_data.nature ~= "location" ) then
  
    ui.buttons.Disable( { _this.button.play, _this.button.skip } );
    
  elseif ( minievent_data.is_cooldown ) then
    
    ui.button.Disable( _this.button.play, multiset );
    ui.button.Enable( _this.button.skip, multiset );
    
    ui.button.SetTextTag(
      _this.button.skip,
      "price",
      minievent_data.cooldown.price,
      multiset
    );
    
    private.Cooldown_SetTimer();

    if ( minievent_data.is_reward ) then

      ui.obj.Set( _this.button.skip, { alp = 0 }, multiset );

    end;
    
  else
    
    ui.button.Disable( _this.button.skip, multiset );

    if ( not minievent_data.is_target_reward ) then
      ui.button.Enable( _this.button.play, multiset );
    else
      ui.button.Disable( _this.button.play, multiset );
    end;
    
  end;

  --- JK: target progressbar ---------------------------------------------------
  if ( minievent_data.targets ) then

    local target = minievent_data.targets[ 1 ];

    if ( not target.is_completed         ) or
       ( minievent_data.is_target_reward ) then

      local total    = target.goal.count;
      local progress = math.min( target.progress, total );
      local ratio    = progress / total;
      local frame    = common.math.Round( 120 * ratio );

      ui.obj.Set(
        _this.progressbar,
        {
          active       = true,
          visible      = true,
          frame        = frame,
          tt_collected = "str_int_dialog_minievent_target_collected",
          tt_counter   = progress.." / "..total
        },
        multiset
      );

    else

      ui.obj.Disable( _this.progressbar, false, multiset );

    end;
    
  else

    ui.obj.Disable( _this.progressbar, false, multiset );
    
  end;
  
  --- JK: character ------------------------------------------------------------
  if ( minievent_data.character ) then

    local character_name = minievent_data.character.name;
    local pose_index     = minievent_data.character.start_pose;

    ui.obj.Set(
      _this.character,
      {
        active   = true,
        visible  = true,
        input    = true,
        res      = res_path.."characters/anm_"..character_name,
        animfunc = "static"..pose_index,
        playing  = true,
        frame    = 0
      },
      multiset
    );
    
  else
    
    ui.obj.Disable( _this.character, false, multiset );
    
  end;

  -- JK: activity ----------------------------------------------------------------
  if ( minievent_data.activity ) then

    local stage_index = private.StageLast_Get( minievent_data.id );

    ui.obj.Set(
      _this.activity,
      {
        active   = true,
        visible  = true,
        input    = true,
        res      = res_path.."activities/anm_"..minievent_data.activity.name,
        animfunc = "static"..stage_index,
        playing  = true,
        frame    = 0
      },
      multiset
    );

  else

    ui.obj.Disable( _this.activity, false, multiset );

  end;

  -- JK: dialogs ---------------------------------------------------------------
  if ( minievent_data.dialogs ) then

    if ( _this.character_dialog_is_active ) then

      ui.obj.Enable( _this.character_dialog.body, false, multiset );

    end;

  else
    
    ui.obj.Disable( _this.character_dialog.body, false, multiset );

  end;

  -- JK: book ------------------------------------------------------------------
  if ( minievent_data.type == "story" ) then

    if ( _this.book_is_active ) then

      ui.obj.Enable( _this.book.body, false, multiset );

    end;

  else
    
    ui.obj.Disable( _this.book.body, false, multiset );

  end;

  -- JK: window props ----------------------------------------------------------  
  local window_props = {
    tt_title     = "str_int_dialog_minievent_"..minievent_name.."_title",
    tt_desc      = "str_int_dialog_minievent_"..minievent_name.."_description",
    tt_event_end = "str_int_dialog_minievent_end"
  };

  local scrollback_id = minievent_data.type.."_"..minievent_data.nature;

  if ( minievent_data.nature == "stages" ) then

    scrollback_id = scrollback_id.."_".._this.stage_count;

  end;

  if ( minievent_data.nature == "location" ) then
    
    if ( minievent_data.is_reward ) then
      
      if ( minievent_data.type ~= "chests" ) then
      
        scrollback_id = scrollback_id.."_final";
      
      end;
    
    end;

  end;

  local bg_props = {
    rt_mini_event_bg = scene_path..background,
    animfunc         = "scrollback_"..private.app_o.."_"..scrollback_id
  };
  
  if ( private.app_o == "landscape" ) then

    window_props.rt_back = "assets/content/interface/int_dialog_minievent/back";
    
    ui.obj.Set(
      _this.decor,
      {
        res      = scene_path.."anm_landscape_"..minievent_fullname.."_bg",
        animfunc = "landscape_bg_"..minievent_fullname,
      },
      multiset
    );

  end;

  ui.obj.Set( _this.window, window_props, multiset );
  ui.obj.Set( _this.bg, bg_props, multiset );

  local title = tern( minievent_data.is_cooldown )(
    "",
    "str_int_dialog_minievent_"..minievent_name.."_title"
  );
  
  ui.obj.Set(
    _this.header,
    {
      tt_event     = "str_int_dialog_minievent_"..minievent_data.type,
      tt_title     = title,
      tt_event_end = "str_int_dialog_minievent_end",
      rt_icon      = ui.obj.Res( "item", "icon_minievent_"..minievent_data.type )
    },
    multiset
  );

  local description = tern( minievent_data.is_cooldown )(
    "str_int_dialog_minievent_"..minievent_name.."_cooldown",
    "str_int_dialog_minievent_"..minievent_name.."_description"
  );
  
  if ( minievent_data.type == "progress" ) then
    description = "str_int_dialog_minievent_"..minievent_name.."_description_1";
  end;
  
  ui.obj.SetTextTag( _this.footer, "tt_desc", description, multiset );
  ui.obj.Multiset( multiset );
  
  private.Progress_SetState( minievent_data );
  
end;
--------------------------------------------------------------------------------
function private.Window_ShowFinished ()

  local multiset = {};

  ui.input.Off( _this.window, multiset );
  ui.obj.Enable( _this.finished, false, multiset );

  ui.obj.Multiset( multiset );

  private.Header_Footer_Off();

end;
--------------------------------------------------------------------------------
function private.Window_ShowSuspended ()

  _this.is_suspended = true;

  private.Point_ClearEvents();

  local multiset = {};

  ui.input.Off( _this.window, multiset );
  ui.obj.Enable( _this.suspended, false, multiset );

  ui.obj.Multiset( multiset );

  private.Header_Footer_Off();

end;
--------------------------------------------------------------------------------
function private.Window_HideSuspended ()

  if ( not _this.is_suspended ) then return; end;

  local multiset = {};

  ui.obj.Disable( _this.suspended, false, multiset );
  ui.input.On( _this.window, multiset );

  ui.obj.Multiset( multiset );

  _this.is_suspended = nil;

  private.Header_Footer_On();
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Progress_SetState ( minievent_data )
  
  if ( minievent_data.type ~= "progress" ) then
    ui.obj.Disable( _this.anm_progress_hero );
    return;
  end;
  
  ui.obj.Enable( _this.anm_progress_hero );
  ui.obj.Set( _this.anm_progress_hero, {
    res = "assets/content/minievent_progress_crowns/activities/anm_hero"
  } );
  
  local new_progress, old_progress = _f.GetCrownDataChanged();
  
  if ( new_progress == old_progress ) then
    ui.anim.Play( _this.anm_progress_hero, "static"..new_progress );
  else
    local func_end = function ()
      ui.input.On( _this.button.close );
      ui.input.On( _this.button.play );
      ui.anim.Play( _this.anm_progress_hero, "static"..new_progress );
    end;
    
    common.Timer_Start( _this.timer.animation, 0.6, function ()
      local anim_func = tern( new_progress > old_progress )( "move"..new_progress, "reset"..old_progress );
      common.Timer_Stop( _this.timer.animation );
      ui.input.Off( _this.button.close );
      ui.input.Off( _this.button.play );
      ui.anim.Play( _this.anm_progress_hero, anim_func, func_end );
    end );
  end;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Marker_Set ( skin_data, multiset )   
  
  ui.obj.Set(
    _this.marker.icon,
    {
      res      = ui.obj.Res( "skin", skin_data.name ),
      animfunc = "marker",
      playing  = true
    },
    multiset
  );
  
  ui.obj.Set(
    _this.marker.anim,
    {
      animfunc = "avatar_static",
      playing  = true,
      endtrig  = function() end
    },
    multiset
  ); 
  
end;
--------------------------------------------------------------------------------
function private.Timer_SetDuration ()

  local seconds = math.max( 0, _f.GetSeconds() );

  ui.obj.SetTextTag(
    _this.header,
    "tt_event_timer",
    ui.text.FormatTime( seconds )
  );

  if ( seconds > 0 ) then

    common.Timer_Start( _this.timer.duration, 1.0, private.Timer_SetDuration );

    _this.is_finished = false;

    return;

  end;
  
  _this.is_finished = true;

  if ( not _this.is_reward ) then
    
    _f.StopCooldown();
    private.Window_ShowFinished();
    
  end;
  
  common.Timer_Stop( _this.timer.duration );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Map_ScrollStart ()

  local dragger   = "obj_int_dialog_minievent_dragger";
  local container = "anm_int_dialog_minievent_window";
  
  local shift_pos = GetObjPosByObj( dragger, container );
  private.drag_begin_pos = shift_pos[ 2 ];
  
end;
--------------------------------------------------------------------------------
function private.Map_ScrollEnd ()

  local dragger = "obj_int_dialog_minievent_dragger";
  ObjSet( dragger, { pos_x = 0, pos_y = 0 } );
  
end;
--------------------------------------------------------------------------------
function private.Map_Scroll ()
  
  local dragger    = "obj_int_dialog_minievent_dragger";
  local container  = "anm_int_dialog_minievent_window";
  local map        = "anm_int_dialog_minievent_window_bg";

  local current_pos = GetObjPosByObj( dragger, container );
  local delta = current_pos[ 2 ] - private.drag_begin_pos;
  private.drag_begin_pos = current_pos[ 2 ];
  
  local posy = ObjGet( map ).pos_y;
  local newpos = posy + delta;
  
  local top_margin = - 0.5 * private.app_h;
  local bot_margin =   0.5 * private.app_h;
  
  if ( newpos <= top_margin ) then
    
    ObjSet( map, { pos_y = top_margin } );
    
  elseif ( newpos >= bot_margin ) then
    
    ObjSet( map, { pos_y = bot_margin } );
    
  else
    
    ObjSet( map, { pos_y = newpos } );
    
  end;  
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Character_OnTap ()

  ui.popup.Cloud_Show(
    "monolog",
    "obj_int_dialog_minievent_character_indicator",
    "str_minievent_".._this.minievent_name..
    "_scene_".._this.scene_number..
    "_character_speech"
  );

  --ui.anim.Play( _this.character, "static2", nil, "static1" );

end;
--------------------------------------------------------------------------------
function private.Character_TryShowDialog ( minievent_data, when, func_next )

  --DbgTrace( "CharacterDialog_TryShow" );

  if ( _this.character_dialog_is_active ) then return; end;

  if ( not minievent_data.dialogs ) then

    if ( func_next ) then func_next(); end; return;

  end;

  local stage_index = private.StageLast_Get( minievent_data.id );

  if ( minievent_data.dialogs[ "stage_"..stage_index ]         ) and
     ( minievent_data.dialogs[ "stage_"..stage_index ] == when ) then

    local dialog_data = _f.GetDialogData( stage_index, when );

    --DbgTrace( "dialog data:" );
    --DbgTrace( dialog_data );

    if ( dialog_data ) then

      local respath = "assets/content/minievent_"..minievent_data.name.."/";

      _this.character_dialog.data             = dialog_data;
      _this.character_dialog.data.step        = 1;
      _this.character_dialog.data.func_next   = func_next;
      _this.character_dialog.data.respath     = respath;
      _this.character_dialog.data.res         = respath.."anm_int_dialog_minievent_story_character";
      _this.character_dialog.data.str_prefix  = "str_int_dialog_minievent_"..dialog_data.name.."_";
      _this.character_dialog.data.item_prefix = "str_int_dialog_minievent_"..minievent_data.name.."_item_";

      private.Character_ShowDialog();

    else

      if ( func_next ) then func_next(); end;

    end;

  else

    if ( func_next ) then func_next(); end;

  end;

end;
--------------------------------------------------------------------------------
function private.Character_ShowDialog ()

  _this.character_dialog_is_active = true;

  local dialog_data = _this.character_dialog.data;
  local step        = dialog_data.step;
  local func_next   = dialog_data.func_next;
  local respath     = dialog_data.respath;
  local res         = dialog_data.res;
  local str         = dialog_data.str_prefix;
  local item_prefix = dialog_data.item_prefix;

  local multiset = {};

  ui.obj.Set(
    _this.character_dialog.body,
    {
      active         = true,
      visible        = true,
      input          = true,
      res            = res,
      tt_description = ne.strings.Get( str..step ),
      tt_skip        = _this.strings.dialog_skip
    },
    multiset
  );

  ui.obj.Set(
    _this.character_dialog.character,
    {
      res      = res,
      animfunc = "character"..dialog_data.list[ step ].character_view
    },
    multiset
  );

  private.Character_SetDialogItems(
    _this.character_dialog.item,
    dialog_data.list[ step ].items,
    respath,
    res,
    item_prefix,
    multiset
  );

  ui.obj.Multiset( multiset );
  
  ui.anim.Play(
    _this.character_dialog.body,
    "appear_"..private.app_o,
    nil,
    "static_"..private.app_o
  );

  _this.character_dialog.data.step = step + 1;

end;
--------------------------------------------------------------------------------
function private.Character_NextDialog ()

  if ( not _this.character_dialog.data ) then return; end;

  local dialog_data = _this.character_dialog.data;
  local step        = dialog_data.step;
  local func_next   = dialog_data.func_next;
  local respath     = dialog_data.respath;
  local res         = dialog_data.res;
  local str         = dialog_data.str_prefix;
  local item_prefix = dialog_data.item_prefix;

  if ( not dialog_data.list[ step ] ) then

    _f.SetDialogComplete( dialog_data.name );

    private.Character_HideDialog( func_next );

    _this.character_dialog.data      = nil;
    _this.character_dialog_is_active = nil;

    return;

  end;

  ui.obj.Clone(
    _this.character_dialog.character,
    _this.hub,
    "_1"  
  );

  ui.obj.Clone(
    _this.character_dialog.character,
    _this.hub,
    "_2"  
  );

  ui.obj.Clone(
    _this.character_dialog.item,
    _this.hub,
    "_1"  
  );

  ui.obj.Clone(
    _this.character_dialog.item,
    _this.hub,
    "_2"  
  );

  local step     = dialog_data.step;
  local animfunc = "character"..dialog_data.list[ step ].character_view;

  local multiset = {};

  ui.obj.Set(
    _this.character_dialog.body,
    {
      tt_description = ne.strings.Get( str..step )
    },
    multiset
  );

  ui.obj.Set( 
    _this.character_dialog.character.."_1",
    {
      anim_tag = "at_character1"
    },
    multiset
  );

  ui.obj.Set( 
    _this.character_dialog.character.."_2",
    {
      anim_tag = "at_character2",
      animfunc = animfunc
    },
    multiset
  );

  ui.obj.Set( 
    _this.character_dialog.item.."_1",
    {
      anim_tag = "at_item1"
    },
    multiset
  );

  ui.obj.Set( 
    _this.character_dialog.item.."_2",
    {
      anim_tag = "at_item2"
    },
    multiset
  );

  private.Character_SetDialogItems(
    _this.character_dialog.item.."_1",
    dialog_data.list[ step - 1 ].items,
    respath,
    res,
    item_prefix,
    multiset
  );

  private.Character_SetDialogItems(
    _this.character_dialog.item.."_2",
    dialog_data.list[ step ].items,
    respath,
    res,
    item_prefix,
    multiset
  );

  ui.obj.Multiset( multiset );

  ui.obj.Attach(
    _this.character_dialog.character.."_1", _this.character_dialog.body
  );

  ui.obj.Attach(
    _this.character_dialog.character.."_2", _this.character_dialog.body
  );

  ui.obj.Attach(
    _this.character_dialog.item.."_1", _this.character_dialog.body
  );

  ui.obj.Attach(
    _this.character_dialog.item.."_2", _this.character_dialog.body
  );

  ui.anim.Play(
    _this.character_dialog.body,
    "transition_"..private.app_o,
    function ()

      local multiset = {};

      ui.obj.Set( 
        _this.character_dialog.character,
        {
          animfunc = animfunc
        },
        multiset
      );

      private.Character_SetDialogItems(
        _this.character_dialog.item,
        dialog_data.list[ step ].items,
        respath,
        res,
        item_prefix,
        multiset
      );

      ui.obj.Multiset( multiset );

      ui.obj.Delete( _this.character_dialog.character.."_1" );
      ui.obj.Delete( _this.character_dialog.character.."_2" );
      ui.obj.Delete( _this.character_dialog.item.."_1" );
      ui.obj.Delete( _this.character_dialog.item.."_2" );

    end,
    "static_"..private.app_o
  );

  _this.character_dialog.data.step = step + 1;

end;
--------------------------------------------------------------------------------
function private.Character_SetDialogItems (
    obj_name,
    items,
    respath,
    res,
    item_prefix,
    multiset
  )

  if ( items ) then

    local rt_item = { "", "", "" };
    local tt_item = { "", "", "" };

    local item_count = #items;

    for i, item in pairs( items ) do

      if ( item.name:find( "chest" ) ) then

        rt_item[ i ] = ui.obj.Res( "item", item.name );

      else

        rt_item[ i ] = respath.."items/item_"..item.name;

      end;

      tt_item[ i ] = item_prefix..item.substr;

    end;

    ui.obj.Set(
      obj_name,
      {
        active    = true,
        visible   = true,
        res       = res,
        animfunc  = "item"..item_count,
        rt_item1  = rt_item[ 1 ],
        rt_item2  = rt_item[ 2 ],
        rt_item3  = rt_item[ 3 ],
        tt_item1  = tt_item[ 1 ],
        tt_item2  = tt_item[ 2 ],
        tt_item3  = tt_item[ 3 ]
      },
      multiset
    );

  else

    ui.obj.Disable( obj_name, false, multiset );

  end;

end;
--------------------------------------------------------------------------------
function private.Character_HideDialog ( func_next )

  ui.input.Off( _this.character_dialog.body );

  ui.anim.Play(
    _this.character_dialog.body,
    "disappear_"..private.app_o,
    function ()

      ui.obj.Disable( _this.character_dialog.body );

      if ( func_next ) then func_next(); end;

    end
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Book_TryShow ( minievent_data, func_next )

  if ( _this.book_is_active ) then return; end;

  if ( minievent_data.type ~= "story" ) then

    if ( func_next ) then func_next(); end; return;

  end;

  local book_data = _f.GetBookData();

  if ( book_data.already_shown ) then

    if ( func_next ) then func_next(); end;
    return;

  end;
  
  _this.book_is_active = true;

  local respath = "assets/content/minievent_"..minievent_data.name.."/";

  ui.button.Create(
    _this.book.button,
    {
      title  = "str_int_dialog_minievent_play",
      design = "medium",
      color  = "green"
    },
    function() private.Book_Hide( func_next ); end
  );
  
  ui.obj.Set(
    _this.book.body,
    {
      active         = true,
      visible        = true,
      input          = true,
      res            = respath.."anm_int_dialog_minievent_story_book",
      tt_description = "str_int_dialog_minievent_"..minievent_data.name.."_book"
    }
  );

  ui.anim.Play(
    _this.book.body,
    "appear_"..private.app_o,
    nil,
    "static_"..private.app_o
  );

end;
--------------------------------------------------------------------------------
function private.Book_Hide ( func_next )

  ui.input.Off( _this.book.body );

  ui.anim.Play(
    _this.book.body,
    "disappear_"..private.app_o,
    function ()

      if ( func_next ) then func_next(); end;

      _f.SetBookComplete();
      ui.obj.Disable( _this.book.body );
      ui.button.Delete( _this.book.button );

      _this.book_is_active = nil;

    end
  );  

end;
--------------------------------------------------------------------------------