-- name=int_dialog_minievent_notification
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Close  = function () end,
  Play   = function () end,
  Toggle = function () end
};

local _this = {};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  _this = {
    int      = "int_dialog_minievent_notification",
    dialog   = "obj_int_dialog_minievent_notification",
    window   = "anm_int_dialog_minievent_notification_window",
    button = {
      close = "btn_int_dialog_minievent_notification_close",
      play  = "btn_int_dialog_minievent_notification_window_play"
    },
    switcher = {
      anm      = "anm_int_dialog_minievent_notification_window_check",
      btn      = "btn_int_dialog_minievent_notification_window_check",
      on       = "check",
      off      = "uncheck",
      on_text  = "str_int_dialog_minievent_notification_switcher_on",
      off_text = "str_int_dialog_minievent_notification_switcher_off"
    },
    dcr = {
      obj = "anm_int_dialog_minievent_notification_decoration",
      res = "anm_decoration_top"
    },
    checkbox_state = true
  };

  private.app_o = ui.screen.Orientation();
  private.app_h = ui.screen.Height();
  private.app_w = ui.screen.Width();
  private.window_area = {
    landscape = { -220, -240, 440, 450 },
    portrait = { -250, -250, 500, 500 };
  };

  private.Adapt();
  private.CreateButtons();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Show ( minievent_data )
  
  private.SetDefault( minievent_data );
  
  ui.window.Open( "minievent_notification", function ()
    
    ui.anim.Play(
      _this.window,
      "appear_"..private.app_o,
      function ()
        if ( minievent_data.type ~= "gift" ) then
          private.Decoration_Enable( minievent_data.name );
        end;
      end,
      "static_"..private.app_o
    );
    
  end );
  
end;
--------------------------------------------------------------------------------
function public.Hide ( func_OnHide )
  
  ui.anim.Play(
    _this.window,
    "disappear_"..private.app_o,
    function ()
      
      ui.window.Close( "minievent_notification", func_OnHide );
      
    end
  );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Adapt ()
  
  local multiset = {};
  
  ui.obj.ToCenter( _this.dialog, 0, 0, multiset );
  ui.window.SetArea( _this.dialog, private.window_area[ private.app_o ], multiset );
  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
function private.CreateButtons ()
  
  ui.button.Create(
    _this.button.close,
    {
      design = "close"
    },
    function() _f.Close(); end
  );

  ui.button.Create(
    _this.button.play,
    {
      title  = "str_int_dialog_minievent_notification_play",
      design = "medium",
      color  = "green"
    },
    function() _f.Play(); end
  );
  
  ui.control.EventsSet( 
    _this.switcher.btn, 
    { 
      func_on_click = private.SwitcherClick 
    } 
  );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.SetDefault ( minievent_data )
  
  local minievent_name     = minievent_data.name;
  local minievent_fullname = minievent_data.type;
  
  if ( minievent_data.type ~= "chests" ) then

    minievent_fullname = minievent_name:match( "^(.*)_%d+$" )  or minievent_name;
    minievent_fullname = minievent_fullname:match( "^(%w+)$" ) or minievent_fullname:match( "^(%w+_%w+)$" );

  end;

  local res_back;
  local res_path = "assets/content/minievent_"..minievent_fullname;

  if ( minievent_data.type ~= "story" ) then
    if ( minievent_data.type == "gift" ) then
      local zone = minievent_data.zone;
      res_back = "assets/content/"..zone.content.."/bg_"..zone.name.."_ride";
    elseif ( minievent_data.type == "section" ) then
      local zone = minievent_data.zone;
      res_back = res_path.."/bg_"..zone.name.."_ride";
    else
      res_back = res_path.."/bg_"..minievent_fullname.."_ride";
    end;
  else
    res_back = res_path.."/"..minievent_data.level_background;
  end;

  local res_icon = "assets/content/items/icon_minievent_"..minievent_data.type;
  local multiset = {};
  
  _this.checkbox_state = not minievent_data.is_context;

  private.UpdateCheckbox( multiset );
  
  ui.obj.Set( 
    _this.window, 
    {
      tt_title = "str_int_dialog_minievent_"..minievent_name.."_title",
      tt_desc = "str_int_dialog_minievent_"..minievent_name.."_description",
      rt_icon = res_icon,
      rt_back = res_back
    },
    multiset
  );
  
  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
function private.SwitcherClick ()
  
  _this.checkbox_state = _f.Toggle();

  local multiset = {};
  
  private.UpdateCheckbox( multiset );
  
  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
function private.UpdateCheckbox ( multiset )
  
  ui.obj.SetTextTag(
    _this.window,
    "tt_switcher",
    _this.switcher.on_text,
    multiset
  );

  ui.obj.Set(
    _this.switcher.anm,
    {
      animfunc = tern( _this.checkbox_state )( _this.switcher.off, _this.switcher.on ),
      playing  = true
    },
    multiset
  );
  
end;
--------------------------------------------------------------------------------
function private.Decoration_Enable ( minievent_name )

  local decoration = _this.dcr.obj;
  local resource   = _this.dcr.res;
  
  ui.obj.Set( decoration, {
    pos_y = - ( 0.5 * private.app_h + 50 ),
    playing = true,
    res = "assets/content/minievent_"..minievent_name.."/"..resource,
    animfunc = private.app_o.."_static"
  } );
  
  ui.obj.Enable( decoration, true );

end;
--------------------------------------------------------------------------------