-- name=int_dialog_minievent_reward
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  OnShow = function () end,
  Accept = function () end,
  Close  = function () end,
  OnHide = function () end
};
local _this = {
  is_active = false
};
local _strings = {};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  local module_name = "_int_dialog_minievent_reward_";

  local str = "str"..module_name;
  local btn = "btn"..module_name;
  local anm = "anm"..module_name;
  
  _this = {
    int    = "int_dialog_minievent_reward",
    app_o  = ui.screen.Orientation(),
    dialog = "obj_int_dialog_minievent_reward",
    header = anm.."window_header",
    body   = anm.."window_body",
    button = {
      close  = btn.."close",
      accept = btn.."accept"
    },
    slot   = "anm"..module_name.."slot",
    sprite = {
      heart = anm.."heart"
    }
  };

  _this.strings = {
    title_preview = ne.strings.Get( str.."title_preview" ),
    title_reward  = ne.strings.Get( str.."title_reward" ),
    description   = ne.strings.Get( str.."description" )
  };

  _this.window_area = { -186, -159, 373, 385 };

  private.Adapt();
  private.CreateButtons();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show (
    reward_group,
    reward_data,
    is_reward,
    is_heart,
    stage_number
  )

  if ( _this.is_active ) then return; end;

  _this.is_active    = true;
  _this.is_reward    = is_reward;
  _this.reward_count = #reward_data;
  _this.func_prefix  = tern( is_heart )( "heart_", "" );

  private.SetDefault( reward_data, is_reward, stage_number );

  local header_appear = tern( is_reward )(
    "appear_".._this.app_o,
    "appear_preview_".._this.app_o
  );

  local header_static = tern( is_reward )(
    "static_".._this.app_o,
    "static_preview_".._this.app_o
  );

  local body_static  = tern( is_reward )(
    _this.func_prefix.."static".._this.reward_count,
    _this.func_prefix.."preview".._this.reward_count
  );

  ui.anim.Play(
    _this.header,
    header_appear,
    _f.OnShow,
    header_static
  );     

  ui.anim.Play( _this.body, body_static );
  
  local open_sound = tern( is_reward )( "tourney_reward_window", "level_click" );

  ui.window.Open(
    "minievent_reward",
    _f.OnShow,
    nil,
    open_sound
  );

end;
--------------------------------------------------------------------------------
function public.Hide ()

  if ( not _this.is_active ) then return; end;

  _this.is_active = false;

  local header_close = tern( _this.is_reward )(
    "disappear_".._this.app_o,
    "disappear_preview_".._this.app_o
  );

  ui.anim.Play(
    _this.header,
    header_close
  );

  ui.window.Close(
    "minievent_reward",
    function () _f.OnHide(); end
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Adapt ()

  local multiset = {};

  ui.obj.ToCenter( _this.dialog, 0, 0, multiset );
  ui.window.SetArea( _this.dialog, _this.window_area, multiset );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.CreateButtons ()

  ui.button.Create(
    _this.button.accept,
    {
      title  = "str_int_dialog_minievent_reward_accept",
      design = "medium",
      color  = "green"
    },
    function() private.AcceptAction(); end
  );

  ui.button.Create(
    _this.button.close,
    {
      design = "close"
    },
    function() _f.Close(); end
  );

end;
--------------------------------------------------------------------------------
function private.SetDefault ( reward_data, is_reward, stage_number )

  local multiset = {};

  if ( is_reward ) then

    ui.input.On( _this.button.accept, multiset );

  end;

  local title = tern( is_reward )(
    _this.strings.title_reward,
    _this.strings.title_preview
  );

  local desc = _this.strings.description;

  if ( not is_reward ) and
     ( stage_number  ) then

    local module_name = "_int_dialog_minievent_reward_";
    local str         = "str"..module_name;
    desc = ne.strings.Get( str.."description_with_number", stage_number );

  end;

  if ( is_reward ) then
    desc = "str_int_dialog_minievent_reward_description_claim";
  end;

  ui.obj.SetTextTag( _this.header, "tt_title", title, multiset );
  ui.obj.SetTextTag( _this.body, "tt_reward_desc", desc, multiset );
  
  for i = 1, 3 do

    ui.obj.Disable( _this.slot..i, true, multiset );

  end;

  local slot_id = 1;

  for _, reward in pairs( reward_data ) do

    if ( reward.type ~= "heart" ) then
      
      reward.count = reward.count or 1;

      local tt_count = tern( reward.count > 1 )( tostring( reward.count ), "" );

      ui.obj.Set(
        _this.slot..slot_id,
        {
          active    = true,
          visible   = true,
          playing   = true,
          animfunc  = "reward_"..reward.type,
          rt_reward = ui.obj.Res( reward.type, reward.name ),
          tt_count  = tt_count
        },
        multiset
      );

      slot_id = slot_id + 1;

    else

      ui.obj.Set(
        _this.sprite.heart,
        {
          active  = true,
          visible = true,
          tt_time = ui.text.FormatTime( reward.count ),
          tt_sign = ui.text.InfinitySign()
        },
        multiset
      );

    end;

  end;

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.AcceptAction ()

  if ( _f.Accept() ) then
    
    common.sound.Play( "tourney_reward_animation" );

    ui.input.Off( _this.button.accept );

    ui.anim.Play(
      _this.body,
      _this.func_prefix.."get_reward".._this.reward_count,
      _f.Close
    );

  end;

end;
--------------------------------------------------------------------------------