-- name=int_dialog_minievent_start
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Accept      = function () end,
  Reject      = function () end,
  BuyBooster  = function () end,
  DataBooster = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  private.app_w = ui.screen.Width();
  private.app_h = ui.screen.Height();
  
  private.orientation = ui.screen.Orientation();
  
  private.this = {
    dialog = "obj_int_dialog_minievent_start",
    window = "anm_int_dialog_minievent_start_window",
    button = {
      close = "btn_int_dialog_minievent_start_close",
      play  = "btn_int_dialog_minievent_start_play"
    }
  };

  private.anim_time = 0.3;

  private.strings = {
    title     = ne.strings.Get( "str_stage_name_minievent" ),
    play      = ne.strings.Get( "str_int_dialog_minievent_start_play" ),
    moves     = ne.strings.Get( "str_int_dialog_minievent_start_moves" ),
    targets   = ne.strings.Get( "str_int_dialog_minievent_start_targets" ),
    boosters  = ne.strings.Get( "str_int_dialog_minievent_start_boosters" ),
    resources = ne.strings.Get( "str_int_dialog_minievent_start_resources" )
  };

  ui.button.Create(
    private.this.button.close,
    {
      design = "close"
    },
    function()
      
      private.Reject();
      
    end
  );
  
  ui.button.Create(
    private.this.button.play,
    {
      title  = private.strings.play,
      color  = "green",
      design = "medium",
      sound  = "button_click"
    },
    function()
      
      private.Accept( private.boosters_selected );
      
    end
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Show (
    stage_name,
    stage_data,
    func_accept,
    func_reject,
    func_booster_buy
  )

  private.boosters          = {};
  private.boosters_selected = {};
  private.stage_name        = stage_name;  
  private.Accept            = func_accept;
  private.Reject            = func_reject;
  private.BoosterBuy        = func_booster_buy;

  private.Default_Set( stage_name, stage_data );

  private.active = true;
  
  ui.window.Open(
    "minievent_start",
    function()
 
    ui.window.SetAreaByButtons(
      private.this.dialog,
      private.this.button.play
    );
  
    end,
    nil,
    "level_click"
  );

end;
--------------------------------------------------------------------------------
function public.Hide ( func_on_close_end )

  if ( not private.active ) then

    return;

  end;

  private.active = false;
  
  ui.window.Close( "minievent_start", func_on_close_end );

end;
--------------------------------------------------------------------------------
function public.Update ( booster_data )

  if ( not private.active ) then

    return;

  end;

  local booster_name = booster_data.name;
  local booster_count = booster_data.count;
  local index = private.boosters[ booster_name ];

  if ( index ) then

    private.Booster_TapSet( booster_name, index, booster_count );

    local obj = "anm_int_dialog_minievent_start_boosters_"..index.."_icon";

    local params = { count = ui.text.FormatNumber( booster_count ) };

    if ( booster_count > 0 ) then

      params.animfunc = "booster";
      ObjSet( obj, params );

    else

      params.animfunc = "booster_add";
      ObjSet( obj, params );

    end;

    if ( ( booster_data.source or "" ) ~= "tutorial" ) and ( booster_data.delta > 0 ) then
    
      ( ui.control.TapGet( "btn_int_dialog_minievent_start_boosters_"..index ).func_true or function () end )();

    end;

  end;

end;
--------------------------------------------------------------------------------
function public.UpdateSaleBoosters ()
  
  if ( not private.active ) then

    return;

  end;
  
  for key, value in pairs( private.boosters ) do
    
    local booster_data = _f.DataBooster( key );
    
    ObjSet( "anm_int_dialog_minievent_start_boosters_"..value.."_icon", {
      rt_sale = tern( booster_data.sale )( "assets/content/items/icon_sale","" ),
      tt_sale = tern( booster_data.sale )( "str_int_sale_label","" )
    } );
    
  end;
  
end;
--------------------------------------------------------------------------------
function public.IsActive ()

  return private.active;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Default_Set ( stage_name, stage_data )
  
  local multiset_table = {};
  
  -- Detect stage type.
  local stage_type = stage_data.type;

  -- Moves.
  local moves_data = {};
  moves_data.moves = stage_data.moves;
  private.Moves_Set( multiset_table, moves_data );  
  
  -- Targets.
  local targets_data = {};
  targets_data.targets = stage_data.targets;
  targets_data.stage_type = stage_type;
  targets_data.section = stage_data.section or "section_1";
  private.Targets_Set( multiset_table, targets_data );
  
  -- Boosters.
  private.Boosters_Set( multiset_table, stage_data.boosters );
  
  -- Anim.
  local anim_data = stage_data;
  anim_data.stage_type = stage_type;
  anim_data.stage_state = stage_data.state;
  private.Animation_Set( multiset_table, anim_data );
  
  -- Resources
  private.Resources_Set( multiset_table, stage_data );
  
  private.Window_Set( multiset_table, stage_data );
  
  ui.obj.Multiset( multiset_table );

  local animfunc = "static_"..private.orientation;
  
  if ( #stage_data.resources > 0 ) then
    animfunc = "static_resources_"..private.orientation
  end;
  
  ui.anim.Play( private.this.window, animfunc );

end;
--------------------------------------------------------------------------------
function private.Moves_Set ( multiset_table, moves_data )

  if ( moves_data.moves ) then
    
    table.insert( multiset_table, 
    { 
      "txt_int_dialog_minievent_start_moves_count", { text = ui.text.FormatNumber( moves_data.moves ) }
    } );
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.Targets_Set ( multiset_table, targets_data )
  
  local targets_pos_y = { { 0 }, { - 25, 25 }, { - 50, 0, 50 } };
  
  local pos_x = 40; 
  
  for i = 1, 3 do

    table.insert( multiset_table, 
    { 
      "obj_int_dialog_minievent_start_targets_target_"..i, { active = false, visible = false }
    } );
    ui.obj.Set(
      "spr_int_dialog_minievent_start_targets_target_"..i.."_icon",
      {
        res = ""
      },
      multiset_table
    );

  end;
  
  if ( targets_data.targets and #targets_data.targets > 0 ) then
    
    local resource = {
      DropItem       = "items/icon_%s",
      DropItem_key   = targets_data.section.."/icon_%s_"..( targets_data.section:match( "minievent_(.*)" ) or targets_data.section:match( "%d+" ) ),
      DestroyGem     = "items/icon_%s",
      DestroyCell    = "items/icon_%s"
    };
    
    for i = 1, #targets_data.targets do
      
      local t_type = targets_data.targets[ i ].type;
      local t_name = targets_data.targets[ i ].name;

      table.insert( multiset_table, 
      { 
        "obj_int_dialog_minievent_start_targets_target_"..i,
        {
          active  = true,
          visible = true,
          pos_y   = targets_pos_y[ #targets_data.targets ][ i ]
        }
      } );

      if ( targets_data.targets[ i ].type == "DropItem"    ) or
         ( targets_data.targets[ i ].type == "DestroyGem"  ) or
         ( targets_data.targets[ i ].type == "DestroyCell" ) then
         
        pos_x = 40; 
         
        local resource_path = tern( resource[ t_type.."_"..string.lower( t_name ) ] )(
          resource[ t_type.."_"..string.lower( t_name ) ],
          resource[ t_type ]
        )
        
        local res_part = string.format(
          resource_path,
          string.lower( targets_data.targets[ i ].name )
        );

        ui.obj.Set(
          "spr_int_dialog_minievent_start_targets_target_"..i.."_icon",
          {
            active  = true,
            visible = true,
            res     = "assets/content/"..res_part
          },
          multiset_table
        );
          
      else

        pos_x = 0; 
          
      end;

      table.insert( multiset_table, 
      { 
        "txt_int_dialog_minievent_start_targets_target_"..i,
        {
          pos_x = pos_x,
          text  = ui.text.FormatNumber( targets_data.targets[ i ].count, true )
        } 
      } );

    end;
  
  else

    if ( targets_data.stage_type == "fight" ) then
      
      pos_x = 0;
      
      table.insert( multiset_table, 
      { 
        "obj_int_dialog_minievent_start_targets_target_1", { active = true, visible = true, pos_y = - 20 }
      } );

      table.insert( multiset_table, 
      { 
        "txt_int_dialog_minievent_start_targets_target_1", {
          pos_x  = pos_x,
          text   = "str_dialog_minievent_start_killenemy", 
          param0 = StringGet( "str_enemy_name_"..targets_data.enemy_name )
        }
      } );

    end;

  end;  
  
end;
--------------------------------------------------------------------------------
function private.Boosters_Set ( multiset_table, boosters )

  local boosters_tags = { { 3 }, { 2, 4 }, { 1, 3, 5 } };
  
  for i = 1, 3 do

    table.insert( multiset_table, 
    { 
      "btn_int_dialog_minievent_start_boosters_"..i, 
      {
        anim_tag = "",
        active   = false,
        visible  = false,
        input    = false 
      }
    } );
    
    if ( boosters[ i ] ) then

      local booster = boosters[ i ];

      private.boosters[ booster.name ] = i;

      table.insert( multiset_table, 
      { 
        "btn_int_dialog_minievent_start_boosters_"..i,
        { 
          anim_tag = "booster_"..boosters_tags[ #boosters ][ i ],
          active   = true,
          visible  = true,
          input    = true 
        }
      } );

      local params =
      {
        booster_icon = ui.obj.Res( "booster", booster.name ),
        count        = ui.text.FormatNumber( booster.count ),
        playing      = true,
        rt_sale      = tern( booster.sale )( "assets/content/items/icon_sale","" ),
        tt_sale      = tern( booster.sale )( "str_int_sale_label","" )
      };

      if ( booster.count > 0 ) then

        params.animfunc = "booster";

      else

        params.animfunc = "booster_add";

      end;

      table.insert( multiset_table, 
      { "anm_int_dialog_minievent_start_boosters_"..i.."_icon", params } );

      private.Booster_TapSet( booster.name, i, booster.count );

    end;

  end;
  
end;
--------------------------------------------------------------------------------
function private.Animation_Set ( multiset_table, anim_data )
 
  local light_colors = { r = 0.2352, g = 0.3058, b = 0.3058 };

  local light_color_r = anim_data.color.r + light_colors.r;
  if ( light_color_r > 1 ) then light_color_r = 1; end;

  local light_color_g = anim_data.color.g + light_colors.g;
  if ( light_color_g > 1 ) then light_color_g = 1; end;

  local light_color_b = anim_data.color.b + light_colors.b;
  if ( light_color_b > 1 ) then light_color_b = 1; end;
  
  table.insert( multiset_table, 
  { 
    private.this.window, 
    {
      color_l_r    = light_color_r,
      color_l_g    = light_color_g,
      color_l_b    = light_color_b,
      color_d_r    = anim_data.color.r,
      color_d_g    = anim_data.color.g,
      color_d_b    = anim_data.color.b
    }
  } );
  
end;
--------------------------------------------------------------------------------
function private.Resources_Set ( multiset_table, resources_data )
  
  ui.obj.Set( "obj_int_dialog_minievent_start_resources", 
  { 
    active = false, visible = false
  }, multiset_table );

  for i = 1, 3 do

    ui.obj.Set( "spr_int_dialog_minievent_start_resources_"..i, 
    { 
      active = false, visible = false
    }, multiset_table );

  end;  

  if ( 0 < #resources_data.resources ) then

    ui.obj.Set( "obj_int_dialog_minievent_start_resources", 
    { 
      active = true, visible = true
    }, multiset_table );

    for i = 1, math.min( #resources_data.resources, 3 ) do

      ui.obj.Set( "spr_int_dialog_minievent_start_resources_"..i,
      {
        active   = true,
        visible  = true,
        res = "assets/content/items/icon_"..string.lower( resources_data.resources[ i ] )
      }, multiset_table );
      
    end;

  end; 
  
end;
--------------------------------------------------------------------------------
function private.Window_Set ( multiset, stage_data )

  local shift_y = 0; --tern( IsPC() )( 0, -100 );

  ui.obj.Set( 
    private.this.dialog, 
    { 
      pos_x = 0.5 * private.app_w,
      pos_y = 0.5 * private.app_h + shift_y
    },
    multiset
  );
  
  local zone = stage_data.zone.background;
  
  ui.obj.Set( 
    private.this.window, 
    {
      tt_title         = private.strings.title,
      tt_targets       = private.strings.targets,
      tt_resources     = private.strings.resources,
      tt_moves         = private.strings.moves,
      tt_booster_title = tern( #stage_data.boosters > 0 )( private.strings.boosters, "" ),
      bg_image         = "assets/content/"..zone.content.."/bg_"..zone.name.."_"..stage_data.type
    },
    multiset
  );
  
end;
--------------------------------------------------------------------------------
function private.Booster_TapSet ( name, index, count )

  ui.control.TapClear( "btn_int_dialog_minievent_start_boosters_"..index );

  ui.control.TapSet( "btn_int_dialog_minievent_start_boosters_"..index,
  {
    func_true = function ()

      if ( not private.popup_active ) then

        if ( count > 0 ) then

          private.Booster_Select( name, index );
        
        else

          private.BoosterBuy( name );
      
        end;
          
        common.sound.Play( "select_boost" );
          
      end;

      private.popup_active = false;
           
      ui.popup.Flat_Hide( true );

    end,

    func_menter = function ()

      --

    end,

    func_mleave = function ()
      
      if ( private.popup_active ) then
        
        ui.popup.Flat_Hide( true );
        
      end;
      
    end,

    func_hold_down = function ()
   
      private.popup_active = true;
    
      ui.popup.Flat_Show( 
        "str_item_"..name.."_caption",
        "str_item_"..name.."_description",
        "btn_int_dialog_minievent_start_boosters_"..index,
        { 0, -160 }
      );

    end,

    func_hold_up = function ()
    
      ui.popup.Flat_Hide( true );
  
    end;
  } );

end;
--------------------------------------------------------------------------------
function private.Booster_Select ( booster_name, index )

  local anm_icon = "anm_int_dialog_minievent_start_boosters_"..index.."_icon";

  if ( private.boosters_selected[ booster_name ] ) then

    ui.anim.Play( anm_icon, "booster" );
    private.boosters_selected[ booster_name ] = nil;

  else

    ui.anim.Play( anm_icon, "booster_selected" );
    private.boosters_selected[ booster_name ] = true;

  end;

end;
--------------------------------------------------------------------------------