-- name=int_dialog_player_code_enter
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  EnterCode = function () end,
  Close     = function () end,
  OnHide    = function () end
};

local _this = {
  button = {
    close = "btn_int_dialog_player_code_enter_close",
    ok    = "btn_int_dialog_player_code_enter_ok",
    clear = "btn_int_dialog_player_code_enter_clear"
  },
  textedit = "ted_int_dialog_player_code_enter"
};
  
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  private.app_w = ui.screen.Width();
  private.app_h = ui.screen.Height();
  private.app_o = ui.screen.Orientation();

  private.CreateButtons();
  private.Adapt();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show ()
  
  if ( _this.is_active ) then return; end;

  _this.is_active  = true;
  _this.is_success = false;

  private.SetDefault();
  
  ui.anim.Play(
    private.anim,
    "appear_code_"..private.app_o,
    function ()
    
      ui.input.On( _this.textedit );
      
    end,
    "static_code_"..private.app_o
  );
  
  ui.window.Open( "player_code_enter" );
  
end;
--------------------------------------------------------------------------------
function public.Hide ( func_end )

  if ( not _this.is_active ) then return; end;
  
  _this.is_active = false;
 
  ui.obj.Set( _this.textedit, 
    {
      input       = false,
      active      = false
    }
  );

  --ui.anim.Play( private.anim, "disappear_code_"..private.app_o );
  
  ui.window.Close(
    "player_code_enter",
    function ()
      _f.OnHide();
      if ( func_end ) then func_end(); end;
    end
  );
  
end;
--------------------------------------------------------------------------------
function public.IsActive ()
  
  return _this.is_active;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.OnCodeAccepted ()
  
  _this.is_success = true;
  
  local frame = ui.obj.Get( private.anim ).frame;
  
  ui.obj.Set(
    private.anim,
    {
      animfunc   = "success_code_"..private.app_o,
      playing    = true,
      frame      = frame,
      tt_success = ne.strings.Get( "str_int_dialog_player_code_enter_success" )
    }
  );
  
  ui.input.On({
    _this.button.close,
    _this.button.ok,
    _this.button.clear
  });
  
end;
--------------------------------------------------------------------------------
function public.OnCodeRejected( reason )
  
  local animfunc      = "error_code_"..private.app_o;
  local error_message = "str_int_dialog_player_code_enter_error";

  if ( reason == "code_used" ) then

    error_message = error_message .. "_used";
    animfunc      = "success_code_"..private.app_o;

  end;
  
  local frame = ui.obj.Get( private.anim ).frame;

  ui.obj.Set(
    private.anim,
    {
      animfunc = animfunc,
      playing  = true,
      frame    = frame,
      tt_error   = ne.strings.Get( error_message ),
      tt_success = ne.strings.Get( error_message )
    }
  );

  if ( reason == "code_used" ) then

    ui.input.On({
      _this.button.close,
      _this.button.ok,
      _this.button.clear
    });

  else

    ui.anim.Play(
      private.anim,
      animfunc,
      function ()

        local frame = ui.obj.Get( private.anim ).frame;
        
        ui.obj.Set(
          private.anim,
          {
            animfunc = "static_code_"..private.app_o,
            playing  = true,
            frame    = frame
          }
        );
        
        ui.obj.Enable( _this.textedit );
        
        ui.input.On({
          _this.button.close,
          _this.button.ok,
          _this.button.clear
        });

      end
    );

  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Adapt ()
  
  private.this_obj = "obj_int_dialog_player_code_enter";
  private.anim = "anm_int_dialog_player_code_enter_window";
  
  private.window_area = {
    portrait  = { -270, -280, 540, 420 },
    landscape = { -225, -100, 450, 200 }
  };

  local multiset = {};

  ui.obj.ToCenter( private.this_obj, 0, 0, multiset );
  
  ui.window.SetArea(
    private.this_obj,
    private.window_area[ private.app_o ],
    multiset
  );
  
  ui.obj.Set(
    private.anim,
    {
      animfunc   = "static_code_"..private.app_o,
      tt_error   = ne.strings.Get( "str_int_dialog_player_code_enter_error" ),
      tt_success = ne.strings.Get( "str_int_dialog_player_code_enter_success" )
    },
    multiset
  );

  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
function private.CreateButtons ()
  
  ui.button.Create(
    _this.button.close,
    {
      design = "close"
    },
    function() _f.Close(); end
  );

  ui.button.Create(
    _this.button.clear,
    {
      design = "backspace"
    },
    function() private.ClearInput(); end
  );
  
  ui.button.Create(
    _this.button.ok,
    {
      title  = "str_int_dialog_player_code_enter_ok",
      design = "medium",
      color  = "green"
    },
    function() private.Accept(); end
  );
  
end;
--------------------------------------------------------------------------------
function private.SetDefault ()

  local multiset = {};

  ui.obj.Set(
    private.anim,
    {
      tt_codetitle = ne.strings.Get( "str_int_dialog_player_code_enter_title" )
    },
    multiset
  );
  local updateTexteditParams;
  updateTexteditParams = function ( msg, params )

    ne.obj.Set( _this.textedit, {
      --native_h           = 50,
      --native_w           = private.app_w,
      --native_y = private.app_h - (50 + params.height),
      native_h           = 0,
      native_w           = 0,
      native_y       = 0
    } );
    ne.msg.Unsubscribe("Event_Keyboard_WillShow", updateTexteditParams);

  end;
  ne.msg.Subscribe("Event_Keyboard_WillShow", updateTexteditParams);

  ui.obj.Set( _this.textedit,
    {
      text               = ne.strings.Get( "str_int_dialog_player_code_enter_code" ),
      input              = false,
      visible            = true,
      active             = true,
      allowed_characters = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz",
      maxchars           = 11,
      clearonfirstinput  = true,
      show_native        = true,
      cursorpos          = 0,
      fontcolor_r        = 0.7568,
      fontcolor_g        = 0.7568,
      fontcolor_b        = 0.7568,
      cursorcolor_r      = 0.0039,
      cursorcolor_g      = 0.3765,
      cursorcolor_b      = 0.4941,
      trg_enter          = private.Accept,
      trg_endediting     = private.Accept,
      trg_onfirstinput   = function ()
      
        ui.obj.Set(
          _this.textedit,
          {
            fontcolor_r = 0,
            fontcolor_g = 0.4274,
            fontcolor_b = 0.5176
          }
        );
       
      end,
      transformtoupper = true
    },
    multiset 
  );
  
  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.ClearInput ()

  local props = ne.obj.Get( _this.textedit );

  if ( props.text        ~= ""     ) and
     ( props.fontcolor_r ~= 0.7568 ) then

    ui.obj.Set(
      _this.textedit,
      {
        text      = "",
        cursorpos = 0
      } 
    );

  end;

end;
--------------------------------------------------------------------------------
function private.Accept ()
  if ( not _this.is_active ) then return; end;

  ui.input.Off({
    _this.button.close,
    _this.button.ok,
    _this.button.clear
  });

  ui.obj.Disable( _this.textedit );

  _f.EnterCode( ne.obj.Get( _this.textedit ).text );
  
end;
--------------------------------------------------------------------------------
