-- name=int_dialog_player_main
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  GetLotInfo                      = function () end,
  PurchaseSkin                    = function () end, 
  GetInviteSkin                   = function () end,
  IsInviteSkinAvailable           = function () end,    
  ChangeSkin                      = function () end,  
  ToggleTopPanel                  = function () end,
  UpdatePanel                     = function () end,
  IsNeedLogin                     = function () end,
  Login                           = function () end,
  GetAchievementList              = function () end,
  GetAchievementReward            = function () end,
  GetAchievementRewardData        = function () end,
  GetAchievementNotificationCount = function () end,
  OpenGoogleAchievements          = function () end,
  Close                           = function () end,
  Show                            = function () end,
  PotionsNotification             = function () end,
  PotionUse                       = function () end,
  PotionBuy                       = function () end,
  PotionGet                       = function () end,
  PotionGetRemainingTime          = function () end,
  PotionIsActive                  = function () end,
  IsSkinTutorialStarted           = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  private.app_w = ui.screen.Width();
  private.app_h = ui.screen.Height();
  private.app_o = ui.screen.Orientation();

  private.this = {
    window = "anm_int_dialog_player_main_window",
    btn_purchase = "btn_int_dialog_player_main_skin_purchase",
    btn_buy = "btn_int_dialog_player_main_skin_buy",
    btn_get_invite = "btn_int_dialog_player_main_skin_get_invite",
    btn_info = "btn_int_dialog_player_main_skin_btninfo"
  };

  local str_prefix = "str_int_dialog_hero";

  private.strings = {
    title             = {
      skins        = ne.strings.Get( "str_int_dialog_hero_tab_skins" ),
      achievements = ne.strings.Get( "str_int_dialog_hero_tab_achievements" ),
      info         = ne.strings.Get( "str_int_dialog_hero_tab_info" ),
      bonus        = ne.strings.Get( "str_int_dialog_hero_tab_potions" )
    },
    tab_skins         = ne.strings.Get( str_prefix.."_tab_skins" ),
    tab_achievements  = ne.strings.Get( str_prefix.."_tab_achievements" ),
    tab_info          = ne.strings.Get( str_prefix.."_tab_info" ),
    tab_bonus         = ne.strings.Get( str_prefix.."_tab_potions" ),
    collect_skinparts = ne.strings.Get( str_prefix.."_collect_skinparts" ),
    already_wearing   = ne.strings.Get( str_prefix.."_already_wearing" ),
    bonus_title       = ne.strings.Get( str_prefix.."_skins_bonus_ribbon" ),
    skin_unavailable  = ne.strings.Get( str_prefix.."_skin_unavailable" )
  };

  private.skins = 
  {
    list = {}, -- skins list
    map  =     -- skins info by name
    {
      data = {},
      id   = {}
    },
    notified = nil, -- notified skin list
    markers  = nil, -- notification markers
    focus =
    {
      skin = nil, -- skin name
      obj  = nil  -- attached object
    },
    scc =
    {
      width      = { portrait = 600, landscape = 600 },
      scales     = { focus = 1.2, near = 0.9, far = 0.75 },
      scale_time = 0.3
    },
    anim_file   = "anm_int_dialog_hero_tab_skin",
    tab_created = false,
    func_create = private.SkinsTab_Create,
    func_update = private.SkinsTab_Update
  }

  private.achievements =
  {
    list = {},
    id   = {},
    scc  =
    {
      size =
      { 
        portrait  = { w = 600, h = 700 }, 
        landscape = { w = 600, h = 700 } 
      }
    },
    anim_file   = "anm_int_dialog_hero_tab_achievements",
    tab_created = false,
    func_create = private.AchievementsTab_Create,
    func_update = private.AchievementsTab_Update
  }

  private.info =
  {
    statistic   = {},
    skin     = nil, -- Skin & attached object
    skin_obj = nil,
    anim_file   = "anm_int_dialog_hero_tab_info",
    tab_created = false,
    func_create = private.InfoTab_Create,
    func_update = private.InfoTab_Update
  }
  
  private.potions =
  {
    list  = {},
    id    = {},
    timer_object = "tmr_int_dialog_player_main_potions_timer",
    scc  =
    {
      size =
      { 
        portrait  = { w = 600, h = 700 }, 
        landscape = { w = 600, h = 700 } 
      }
    },
    anim_file   = "anm_int_dialog_hero_tab_bonus",
    tab_created = false,
    func_create = private.PotionTab_Create,
    func_update = private.PotionTab_Update
  }
  
  private.window_area = {
    [ "portrait"  ] = { -320, -420, 640, 900 },
    [ "landscape" ] = { -250, -320, 500, 700 }
  };

  ui.window.SetArea( "obj_int_dialog_player_main", 
                    private.window_area[ private.app_o ] );
  
  local multiset = {};

  -- Setup static text tags
  local str_prefix = "str_int_dialog_hero";

  -- Centering dialog
  table.insert( multiset, { "obj_int_dialog_player_main", {
    pos_x       = private.app_w / 2,
    pos_y       = private.app_h / 2
  } } );

  -- Skin scroller width
  local skins_scc_w = private.skins.scc.width[ private.app_o ];

  table.insert( multiset, { "scc_int_dialog_player_main_skin_scroller",
  {
    inputrect_x   = - skins_scc_w / 2,
    inputrect_w   =   skins_scc_w,
    croprect_x    = - skins_scc_w / 2,
    croprect_w    =   skins_scc_w,

    begindragtrig = private.SkinsTab_DragStart,
    enddragtrig   = private.SkinsTab_DragEnd,
    scrollendtrig   = private.SkinsTab_DragEnd
  } } );

  -- Achievments scroller
  local ach_scc = private.achievements.scc.size[ private.app_o ];

  table.insert( multiset, { "scc_int_dialog_player_main_achievements_scroller",
  {
    inputrect_init = true,
    inputrect_x   = - ach_scc.w / 2,
    inputrect_y   = - ach_scc.h / 2,
    inputrect_w   =   ach_scc.w,
    inputrect_h   =   ach_scc.h,

    croprect_init = true,
    croprect_x    = - ach_scc.w / 2,
    croprect_y    = - ach_scc.h / 2,
    croprect_w    =   ach_scc.w,
    croprect_h    =   ach_scc.h
  } } );
  
  -- Potions scroller
  local potions_scc = private.potions.scc.size[ private.app_o ];

  table.insert( multiset, { "scc_int_dialog_player_main_potions_scroller",
  {
    inputrect_init = true,
    inputrect_x   = - potions_scc.w / 2,
    inputrect_y   = - potions_scc.h / 2,
    inputrect_w   =   potions_scc.w,
    inputrect_h   =   potions_scc.h,

    croprect_init = true,
    croprect_x    = - potions_scc.w / 2,
    croprect_y    = - potions_scc.h / 2,
    croprect_w    =   potions_scc.w,
    croprect_h    =   potions_scc.h
  } } );
  
  ObjMultiSet( multiset );
  
  ui.button.Create(
    "btn_int_dialog_player_main_close",
    {
      design = "close"
    },
    function () private.Hide( _f.UpdatePanel ); end 
  );
  
  ui.button.Create(
    private.this.btn_info,
    {
      design = "info"
    },
    function () _f.GetInviteSkin( private.focused_skin ); end
  );  
  
  ui.button.Create(
    "btn_int_dialog_player_main_skin_select",
    {
      title  = str_prefix.."_skin_wear",
      design = "medium",
      color  = "green",
      sound  = "wear"
    },
    function () _f.ChangeSkin( private.focused_skin ); end 
  );
  
  ui.button.Create(
  private.this.btn_buy,
    {
      title  = "0",
      design = "price"
    },
    function () _f.PurchaseSkin( private.focused_skin ); end 
  );

  ui.button.Create(
    private.this.btn_purchase,
    {
      title  = "",
      design = "medium",
      color  = "green",
    },
    function () _f.PurchaseSkin( private.focused_skin ); end
  );
  
  ui.button.Create(
    private.this.btn_get_invite,
    {
      title  = str_prefix.."_skin_get_invite",
      design = "medium",
      color  = "blue",
    },
    function () _f.GetInviteSkin( private.focused_skin ); end
  );
  
  ui.button.Create(
    "btn_int_dialog_player_main_info_login",
    {
      title  = str_prefix.."_info_login",
      design = "facebook"
    },
    function () private.InfoTab_Login() end 
  );
  
  if ( IsAndroid() and ( wiz.Android.GetPlatform() == "google" ) ) then
    ui.button.Create(
      "btn_int_dialog_player_main_googleachievements",
      {
        design = "googleach"
      },
      function () _f.OpenGoogleAchievements(); end
    );
  end;  

  ui.control.EventsSet( "btn_int_dialog_player_main_tab_skins", 
  { 
    func_on_click = function () private.ShowTab( "skins" ); end 
  } );

  ui.control.EventsSet( "btn_int_dialog_player_main_tab_achievements", 
  { 
    func_on_click = function () private.ShowTab( "achievements" ); end 
  } );

  ui.control.EventsSet( "btn_int_dialog_player_main_tab_info", 
  { 
    func_on_click = function () private.ShowTab( "info" ); end 
  } );
    
  ui.control.EventsSet( "btn_int_dialog_player_main_tab_potions", 
  { 
    func_on_click = function () private.ShowTab( "potions" ); end 
  } );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.HasSelectButton ()

  local animfunc = ui.obj.Get( private.this.window ).animfunc;

  return ( animfunc:find( "select" ) );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.AchievementsTab_DisableScrolling ()

  local scroller    = "scc_int_dialog_player_main_achievements_scroller";
  private.scrollpos = ui.obj.Get( scroller ).scrollpos;
  
  ui.obj.Set(
    scroller,
    {
      dragenabled = false
    }
  );
  
end;
--------------------------------------------------------------------------------
function public.AchievementsTab_EnableScrolling ()

  ui.obj.Set(
    "scc_int_dialog_player_main_achievements_scroller",
    {
      dragenabled = true,
      scrollpos   = private.scrollpos
    }
  );
  
  private.scrollpos = nil;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Show ( 
    tab_name,
    skins,
    potions,
    achievements,
    statistic,
    focused_skin,
    skins_notified,
    skinparts_markers
  )

  private.skins.list        = skins;
  private.achievements.list = achievements;
  private.potions.list      = potions;
  private.info.statistic    = statistic;

  private.skins.notified = skins_notified;
  private.skins.markers  = skinparts_markers;

  private.Tab_UpdateMarkers( "potions", _f.PotionsNotification() );
  
  private.SkinsTab_MakeIndex();
  
  private.weared_skin  = private.GetWearedSkin( skins );
  private.focused_skin = focused_skin or private.weared_skin;

  tab_name = private.forced_tab or tab_name;
  private.forced_tab = nil;
  
  private.ShowTab( tab_name );
  private.active = true;

  private.Tab_UpdateMarkers( "skins", #skins_notified );
  private.Tab_UpdateMarkers( "achievements", _f.GetAchievementNotificationCount() );

  _f.Show();
  
  ui.window.Open( "player_main", function ()

    -- Achievements notify
    if ( private.notify ) then
      
      mediator.Notify( private.notify.message, private.notify.data );
      private.notify = nil;
      
    else

      mediator.Notify( "Interface_DialogHero_Opened", tab_name );

    end;

  end );
  
end;
--------------------------------------------------------------------------------
function public.Hide( func_execute )
  private.Hide ( func_execute );
end;
--------------------------------------------------------------------------------
function private.Hide ( func_execute )

  private.active = false;
  
  ui.window.Close( "player_main", function ()

    _f.Close();
    
    ui.window.Hide( "int_dialog_player_main", func_execute );

  end );
  
  if ( private.tab_current == "skins" ) then
    v_log.Playhaven_LogPlacement( "outfits_close" );
  end;

end;
--------------------------------------------------------------------------------
function private.ShowTab ( tab )

  if ( tab == "skins" or tab == "achievements" or tab == "info" or tab == "potions" ) then
    
    _f.ToggleTopPanel( tab );
    
    local tab_data = private[ tab ];
    
    ObjSet( private.this.window, 
    { 
      res = "assets/content/interface/int_dialog_player_main/"..tab_data.anim_file 
    } );
    ObjSet( "anm_int_dialog_player_main_tabs", 
    { 
      res = "assets/content/interface/int_dialog_player_main/"..tab_data.anim_file,
      animfunc = "tabs_"..private.app_o,
      playing  = true
    } );
    
    local window_params = {
      tab_skins        = private.strings.tab_skins,
      tab_achievements = private.strings.tab_achievements,
      tab_info         = private.strings.tab_info,
      tab_bonus        = private.strings.tab_bonus,
      tt_message       = private.strings.collect_skinparts
    };

    ObjSet( private.this.window, window_params );
    
    if ( private.tab_current == tab ) then
    
      tab_data.func_update();

      return;

    end;
    
    if ( private.active ) then common.sound.Play( "tab_switch" ); end;

    private.ToggleTabsInput( tab );

    if not ( tab_data.tab_created ) then

      tab_data.func_create();
      tab_data.tab_created = true;

    end;
    
    if ( tab == "skins" ) then
      v_log.Playhaven_LogPlacement( "outfits_open" );
    end;

    tab_data.func_update();

  else

    DbgTrace( "[ INTERFACE ][ DIALOG HERO ] Unknown tab: " .. tostring( tab ) );

  end;

end;
--------------------------------------------------------------------------------
function private.ToggleTabsInput ( tab )

  local tab_prewious  = private.tab_current;
  private.tab_current = tab;

  if ( tab_prewious ~= tab ) then

    local btn = "btn_int_dialog_player_main_tab_";

    if ( tab_prewious ) then
      ObjSet( btn..tab_prewious, { input = true  } );
    end;

    ObjSet( btn..tab,  { input = false } );

  end;

end;
--------------------------------------------------------------------------------
function public.IsActive ()
  
  return private.active;
  
end;
--------------------------------------------------------------------------------
function public.SetForcedTab ( force_tab )
  
  private.forced_tab = force_tab;
  
end;
--------------------------------------------------------------------------------
function public.GetCurrentTab ()
  
  return private.tab_current;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.GetWearedSkin ( skins )

  for s = 1, #skins do
    if ( skins[ s ].weared ) then return skins[ s ].name; end;
  end;

  return false;

end;
--------------------------------------------------------------------------------
function private.Tab_UpdateMarkers ( tab, count )

  count = count or 0;

  local objects = {};

  local obj_ind = "obj_int_dialog_player_main_tab_"..tab.."_notification";
  local txt_ind = "anm_int_dialog_player_main_tab_"..tab.."_notification_bg";
  
  if ( count > 0 ) then

    table.insert( objects, { obj_ind, { active = true, visible = true } } );
    table.insert( objects, { txt_ind, { count  = tostring( count )    } } );

  else

    table.insert( objects, { obj_ind, { active = false, visible = false } } );

  end;

  ObjMultiSet( objects );
  
end;
--------------------------------------------------------------------------------
function private.Skin_UpdateMarker ( skin )

  local obj = "obj_int_dialog_skin_slot_notifications_"..skin;
  local cnt = "anm_int_dialog_skin_slot_notifications_bg_"..skin;

  local text;

  if ( common.IsEntityInList( private.skins.notified, skin ) ) then
    text = "!";
  end;

  if ( private.skins.markers[ skin ] ) then
    text = private.skins.markers[ skin ];
  end;

  local objects = {};

  if ( text ) then

    table.insert( objects, { obj, { active = true, visible = true } } );
    table.insert( objects, { cnt, { count  = tostring( text )     } } );

  else

    table.insert( objects, { obj, { active = false, visible = false } } );

  end;

  ObjMultiSet( objects );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.SkinsTab_MakeIndex ()

  local list = private.skins.list;
  local map  = private.skins.map;

  for id = 1, #list do

    local name = list[ id ].name;

    map.data[ name ] = list[ id ];
    map.id[ name ]   = id;

  end;

end;
--------------------------------------------------------------------------------
function private.SkinsTab_Create ()

  local skin_list = private.skins.list;
  local scc       = private.skins.scc;

  local scroller   = "scc_int_dialog_player_main_skin_scroller";
  local item_dummy = "anm_int_dialog_player_main_skin_slot";
  local spacer     = "obj_int_dialog_player_main_skin_spacer";

  ObjClone( spacer, "_pre_1", true, scroller );
  ObjClone( spacer, "_pre_2", true, scroller );

  for s = 1, #skin_list do

    local skin = skin_list[ s ];
    local name = skin.name;

    ObjClone( item_dummy, "_"..name, true, scroller );

    ui.obj.Disable( "spr_int_dialog_skin_slot_hero_icon_"..name, true );

    ui.control.EventsSet( item_dummy.."_"..name,
    { 
      func_on_click = function () private.SkinsTab_SkinTap( name ) end
    } );

  end;

  ObjClone( spacer, "_post_1", true, scroller );
  ObjClone( spacer, "_post_2", true, scroller );

  local ranged_list = private.SkinsTab_RangedList();

  local function attach_list_item ( skin )
    
    private.SkinsTab_UpdateScroll( skin );

  end;

  common.ProcessArrayByTimer( ranged_list, attach_list_item );

  private.SkinsTab_AttachFocusedSkin();

end;
--------------------------------------------------------------------------------
function private.SkinsTab_RangedList ()
  
  -- Async skins list attach (start from focused skin)
  local skin_list = private.skins.list;
  local id = private.skins.map.id[ private.focused_skin ];

  local ranged_list = {};
  table.insert( ranged_list, skin_list[ id ] );

  for i = 1, #skin_list do

    if ( skin_list[ id - i ] ) then
      table.insert( ranged_list, skin_list[ id - i ] );
    end;

    if  ( skin_list[ id + i ] ) then
      table.insert( ranged_list, skin_list[ id + i ] );
    end;

  end;
  
  return ranged_list;
  
end;
--------------------------------------------------------------------------------
function private.SkinsTab_Update ()

  private.SkinsTab_UpdateContainer();
  private.SkinsTab_Focus();

end;
--------------------------------------------------------------------------------
function private.SkinsTab_UpdateScroll ( skin )

  local skin_tag = "obj_int_dialog_skin_slot_hero_"..skin.name;
  local skin_obj;
  
  if ( skin.animated ) then
    
    skin_obj = ui.skin.Attach( skin_tag, skin.name );

    ObjSet( skin_obj, 
    {
      animfunc = "dialog",
      playing  = true
    } );
    
  else
    
    skin_obj = "spr_int_dialog_skin_slot_hero_icon_"..skin.name;
    ui.obj.Enable( skin_obj, true );
    ui.obj.Set( skin_obj, 
    {
      res = "assets/content/"..skin.content.."/"..skin.name.."_skin"
    } );
    ui.obj.Attach( skin_obj, skin_tag );
    
  end;

end;
--------------------------------------------------------------------------------
function private.SkinsTab_UpdateContainer ()

  local scroller   = "scc_int_dialog_player_main_skin_scroller";
  local item_dummy = "anm_int_dialog_player_main_skin_slot";
  local updated = false;
  
  local skin_list = private.skins.list;
  
  local function attach_list_item ( skin )

    private.SkinsTab_UpdateScroll( skin );

  end;

  for s = 1, #skin_list do

    local skin     = skin_list[ s ];
    local name     = skin.name;
    local active   = skin.wearable;
    local selected = skin.weared;

    local animfunc = "skin_inactive";
    
    if     ( selected ) then animfunc = "skin_selected";
    elseif ( active   ) then animfunc = "skin_active";
    end;

    local item = item_dummy.."_"..name;
    
    if ( not ui.obj.Exists( item ) ) then
      
      ObjClone( item_dummy, "_"..name, true, scroller );

      ui.control.EventsSet( item,
      { 
        func_on_click = function () private.SkinsTab_SkinTap( name ) end
      } );
      
      updated = true;
        
      attach_list_item( skin );
        
    end;
      
    ui.anim.Play( item, animfunc );

    private.Skin_UpdateMarker( name );
    
  end;
  
  if ( updated ) then
  
    local ranged_list = private.SkinsTab_RangedList();
    
    for s = 1, #ranged_list do
      
      local name = ranged_list[ s ].name;
      
      ObjDetach( "anm_int_dialog_player_main_skin_slot_"..name );
      ObjAttach( "anm_int_dialog_player_main_skin_slot_"..name, scroller );
      
    end;
  
  end;
  
  local spacer = "obj_int_dialog_player_main_skin_spacer";
  
  ObjDetach( spacer.."_post_1" );
  ObjDetach( spacer.."_post_2" );
  
  ObjAttach( spacer.."_post_1", scroller );
  ObjAttach( spacer.."_post_2", scroller ); 

end;
--------------------------------------------------------------------------------
function private.SkinsTab_Focus ()

  local skin_map  = private.skins.map;
  local focused   = private.skins.focus;

  local skin = private.focused_skin;

  private.SkinsTab_UpdateInfo();

  ui.skin.Detach( focused.obj );
  private.SkinsTab_AttachFocusedSkin();

  local skin_data  = skin_map.data[ skin ];
  local selectable = skin_data.wearable and not skin_data.weared;
  
  local af = "tab_1_"..private.app_o;
  if ( selectable ) then af = af .. "_select"; end;
  if ( skin_data.price and not skin_data.wearable ) then 
    if ( skin_data.price.lot ) then
      af = af .. "_purchase"; 
    elseif ( skin_data.price.gold ) then
      af = af .. "_buy"; 
    end;
  end;
  
  local message = private.strings.collect_skinparts;  
  
  if ( skin_data.invite and not skin_data.wearable ) then
    
    if ( _f.IsInviteSkinAvailable( skin_data.name ) ) then
      af = af .. "_get_invite";
    else
      message = private.strings.skin_unavailable.."_"..skin_data.name;
    end;
  
  end;
  
  local id    = skin_map.id[ private.focused_skin ];
  local total = #private.skins.list;

  if ( skin_data.weared ) then

    message = private.strings.already_wearing;

  end;

  ObjSet( private.this.window,
  {
    animfunc      = af,
    skins_counter = id .. "/" .. total,
    title         = private.strings.title.skins,
    tt_message    = message
  } );

  local scrollpos = id - 1;
  ObjSet( "scc_int_dialog_player_main_skin_scroller", { scrollpos = scrollpos } );

  private.SkinsTab_ScalesReset();
  private.SkinsTab_ScalesSet();

  if ( skin_data.animated ) then

    ui.anim.Play( focused.obj, "skinview_fun1", nil, "dialog" );

  end;

end;
--------------------------------------------------------------------------------
function private.SkinsTab_AttachFocusedSkin ()
  
  local parent    = "obj_int_dialog_player_main_skin_hero";
  local picture   = "spr_obj_int_dialog_player_main_skin_hero_icon";
  local skin_name = private.focused_skin;
  local skin_data = private.skins.map.data[ skin_name ];
 
  local focused = private.skins.focus;
  focused.skin  = skin_name;
  
  if ( skin_data.animated ) then
    focused.obj = ui.skin.Attach( parent, focused.skin );
    ui.obj.Disable( picture, true );
  else
    ui.obj.Enable( picture, true );
    ui.obj.Set( picture, {
      res = "assets/content/"..skin_data.content.."/"..skin_name.."_skin"
    } );
  end;

  local grayscale = false;

  if not ( private.skins.map.data[ skin_name ].wearable ) then
    grayscale = true;
  end;

  ObjSet( parent, { grayscale = grayscale } );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.SkinsTab_UpdateInfo ()

  local skin = private.focused_skin;

  if ( private.skins.markers[ skin ] ) then

    local to_step = private.skins.map.data[ skin ].collected_parts;
    local steps   = private.skins.map.data[ skin ].parts;

    local delta     = private.skins.markers[ skin ];
    local from_step = to_step - delta;

    private.SkinsTab_SetInfo();
    private.SkinsTab_AnimateInfo( from_step, to_step, steps );

  else

    private.SkinsTab_SetInfo();

  end;

end;
--------------------------------------------------------------------------------
function private.SkinsTab_SetInfo ()

  local info       = "anm_int_dialog_player_main_skin_info";
  local str_prefix = "str_skins_";

  local skin      = private.focused_skin;
  local skin_data = private.skins.map.data[ skin ];

  local animfunc = tern( skin_data.wearable )(
    "skin_info_complete",
    "skin_info_progress"
  );

  local multiset      = {};
  local progress_text = "";  

  if ( skin_data.parts           ) and
     ( skin_data.collected_parts ) and
     ( skin_data.parts > 0       ) then

    local parts = skin_data.collected_parts;
    local total = skin_data.parts;

    progress_text = parts.."/"..total;

    ui.obj.Set(
      "spr_int_dialog_player_main_skin_info_icon",
      { 
        res = ui.obj.Res( "skinpart", skin.."_item" )
      },
      multiset
    );

    ui.obj.Set(
      "anm_int_dialog_player_main_skin_info_progress", 
      { 
        frame = parts / total * 120
      },
      multiset
    );

  else

    animfunc = "skin_info_complete";

  end;

  ui.obj.Set(
    info,
    {
      skin_name        = ne.strings.Get( str_prefix..skin ),
      skin_description = ne.strings.Get( str_prefix..skin.."_description" ),
      bonus_title      = private.strings.bonus_title,
      bonus_text       = ne.strings.Get( str_prefix..skin.."_bonus" ),
      brief_text       = ne.strings.Get( str_prefix..skin.."_brief" ),
      progress_text    = progress_text,
      animfunc         = animfunc,
      playing          = true
    },
    multiset
  );

  ui.button.Disable( private.this.btn_get_invite, multiset );
  ui.button.Disable( private.this.btn_info, multiset );

  if ( not skin_data.wearable ) then
    
    if ( skin_data.price ) then
      
      if ( skin_data.price.gold ) then
        
        ui.button.SetText( private.this.btn_buy, skin_data.price.gold );
        
      elseif ( skin_data.price.lot ) then

        local price_string = _f.GetLotInfo( skin_data.price.lot ).price;
        price_string = tern( price_string == "" )( "str_int_dialog_hero_skin_buy", price_string );

        ui.button.SetText( private.this.btn_purchase, price_string );
        
      end;
      
    end;
    
    if ( skin_data.invite and _f.IsInviteSkinAvailable( skin_data.name ) ) then
      ui.button.Enable( private.this.btn_get_invite, multiset );
    end;
    
  else
    
    if ( skin_data.invite and _f.IsInviteSkinAvailable( skin_data.name ) ) then
      ui.button.Enable( private.this.btn_info, multiset );
    end;
    
  end;
  
  ui.obj.Multiset( multiset );

  local id = common.IsEntityInList( private.skins.notified, skin );
  if ( id ) then

    table.remove( private.skins.notified, id );

  end;

  private.Tab_UpdateMarkers( "skins", #private.skins.notified );
  private.Skin_UpdateMarker( skin );

end;
--------------------------------------------------------------------------------
function private.SkinsTab_AnimateInfo ( from_step, to_step, steps )

  local inf = "anm_int_dialog_player_main_skin_info";
  local bar = "anm_int_dialog_player_main_skin_info_progress";

  AnimStop( inf );
  ui.anim.Play( inf, "skin_info_progress" );

  local anim_list = {};

  local delta    = to_step - from_step;
  local anm_time = 0.6 / delta;

  local function func_anim_step ( from, to, func_after )

    local pb_from = from / steps * 120;
    local pb_to   = to   / steps * 120;

    ObjSet( inf, { progress_text = from .. "/" .. steps } );

    ObjStopAnimate( bar, "frame" );
    ObjAnimate( bar, "frame", 0, 0, 
    function ()
      ObjSet( inf, { progress_text = to .. "/" .. steps } );
      if ( func_after ) then func_after(); end;
    end, 
    {
      0.0,      0, pb_from,
      anm_time, 0, pb_to
    } );

  end;

  for i = 1, delta do

    local from     =   from_step + i - 1;
    local to       =   from_step + i;

    local function func_step ( func_after )
      func_anim_step( from, to, func_after )
    end;

    table.insert( anim_list, func_step );

  end;
  
  common.RunCallbackList( anim_list );
  
  local af = "skin_info_action";
  if ( to_step == steps ) then   
  
    af = "skin_info_action_final"

  end;
  
  local function func_update_marker ()
    
    private.skins.markers[ private.focused_skin ] = nil;
    private.Skin_UpdateMarker( private.focused_skin );
    
    ui.anim.Play( inf, af, private.SkinsTab_SetInfo );    
    
  end;
  
  if ( _f.IsSkinTutorialStarted() ) then
    
    local func_end = function()

      func_update_marker();
      
    end;
    
    mediator.SubscribeOnce( "View_Tutorial_SkinProgressComplete", func_end ); 

  else

    func_update_marker();
    
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.SkinsTab_DragStart ()

  private.SkinsTab_ScalesReset();

end;
--------------------------------------------------------------------------------
function private.SkinsTab_DragEnd ()

  common.sound.Play( "button_click" );

  local id = ObjGet( "scc_int_dialog_player_main_skin_scroller" ).scrollpos + 1;

  local skin_name = private.skins.list[ id ].name;

  if ( private.focused_skin ~= skin_name ) then

    private.focused_skin = skin_name;
    private.SkinsTab_Focus();

  else

    private.SkinsTab_ScalesSet();

  end;

end;
--------------------------------------------------------------------------------
function private.SkinsTab_SkinTap ( skin_name )
  
  if ( private.focused_skin == skin_name ) then

    return;

  end;

  private.focused_skin = skin_name;

  private.SkinsTab_Focus();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.SkinsTab_ScalesReset ( func_after )

  local function func_after_safe ()

    if ( func_after ) then func_after() end;

  end;

  local scale_new = private.skins.scc.scales.far;
  local tme       = private.skins.scc.scale_time;

  local list = private.skins.list;

  for id, skin in pairs( list ) do

    local obj = "obj_int_dialog_skin_slot_hero_"..skin.name;

    ObjStopAnimate( obj, "scale_xy" );

    local scale = ObjGet( obj ).scale_x;

    ObjAnimate( obj, "scale_xy",  0, 0, func_after_safe,
    {
      0.0, 0, scale,     scale,
      tme, 0, scale_new, scale_new
    } );

  end;

end;
--------------------------------------------------------------------------------
function private.SkinsTab_ScalesSet ()

  local skin_list = private.skins.list;
  local skin_map  = private.skins.map;
  local scc       = private.skins.scc;
  local tme       = scc.scale_time;

  local scrollpos = ObjGet( "scc_int_dialog_player_main_skin_scroller" ).scrollpos;

  local scales =
  {
    scc.scales.far, 
    scc.scales.near,
    scc.scales.focus,
    scc.scales.near,
    scc.scales.far
  };

  for i = 1, #scales do

    local skin_id = scrollpos + i - 2;

    if ( skin_id > 0 and skin_id <= #skin_list ) then

      local name = skin_list[ skin_id ].name;
      local obj  = "obj_int_dialog_skin_slot_hero_"..name;

      ObjStopAnimate( obj, "scale_xy" );

      local scale = ObjGet( obj ).scale_x;

      ObjAnimate( obj, "scale_xy",  0, 0, "",
      {
        0.0, 0, scale,       scale,
        tme, 0, scales[ i ], scales[ i ]
      } );

    end;

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Skins_Update ( skins )

  private.skins.list  = skins;
  private.weared_skin = private.GetWearedSkin( skins );

  private.SkinsTab_MakeIndex();
  private.SkinsTab_Update();

end;
--------------------------------------------------------------------------------
function public.Skins_Focus ( skin_name, func_end )

  if ( func_end ) then

    func_end();

  end;

  if ( private.focused_skin == skin_name ) then

    return;

  end;

  private.focused_skin = skin_name;

  private.ShowTab( "skins" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.AchievementsTab_Update ()

  private.achievements.list = _f.GetAchievementList();

  for _, achievement in pairs( private.achievements.list ) do
    private.AchievementsTab_UpdateItem( achievement );
  end;

end;
--------------------------------------------------------------------------------
function public.AchievementsTab_UpdateItem ( achievement )

  if ( private[ "achievements" ].tab_created ) then

    private.AchievementsTab_UpdateItem( achievement );

  end;

end;
--------------------------------------------------------------------------------
function private.AchievementsTab_Create ()

  local scc    = "scc_int_dialog_player_main_achievements_scroller";
  local slot   = "anm_int_dialog_player_main_achievements_slot";
  local reward = "anm_int_dialog_player_main_achievements_slot_reward";
  local button = "btn_int_dialog_player_main_achievements_slot_button";
  local icon   = "anm_int_dialog_player_main_achievements_slot_icon";

  local str_ach = "str_achievement_";
  local str_dlg = "str_int_dialog_hero_achievement_";

  local multiset = {};

  private.achievements.list = _f.GetAchievementList();

  for i, achievement in pairs( private.achievements.list ) do

    local name   = achievement.name;
    local suffix = "_"..name;

    private.achievements.id[ name ] = i;

    ObjClone( slot, suffix, true, scc );

    local item        =   slot..suffix;
    local item_reward = reward..suffix;
    local item_button = button..suffix;
    local item_icon   =   icon..suffix;
    
    ui.button.Create( item_button,
      {
        title  = str_dlg.."get_reward",
        design = "medium",
        color  = "green"
      },
      function () end 
    );

    table.insert( multiset, 
    { item, 
      { achievement_name = StringGet( str_ach..name ) }
    } );

    table.insert( multiset, 
    { item_reward, 
      { 
        reward_label = StringGet( str_dlg.."reward" ),
        tt_complete  = StringGet( str_dlg.."complete" )
      } 
    } );

  end;

  ObjMultiSet( multiset );

end;
--------------------------------------------------------------------------------
function private.AchievementsTab_Update ()

  for _, achievement in pairs( private.achievements.list ) do
    private.AchievementsTab_UpdateItem( achievement );
  end;

  ui.obj.Set(
    private.this.window,
    {
      animfunc = "tab_3_"..private.app_o,
      title    = private.strings.title.achievements
    }
  );

end;
--------------------------------------------------------------------------------
function private.AchievementsTab_UpdateItem ( achievement )

  local scc    = "scc_int_dialog_player_main_achievements_scroller";
  local slot   = "anm_int_dialog_player_main_achievements_slot";
  local reward = "anm_int_dialog_player_main_achievements_slot_reward";
  local icon   = "anm_int_dialog_player_main_achievements_slot_icon";
  local button = "btn_int_dialog_player_main_achievements_slot_button";

  local str_ach = "str_achievement_";
  local str_dlg = "str_int_dialog_hero_achievement_";

  local name   = achievement.name;
  local suffix = "_"..name;

  local item        =   slot..suffix;
  local item_reward = reward..suffix;
  local item_icon   =   icon..suffix;
  local item_button = button..suffix;

  local level     = achievement.level;
  local level_max = 3; 

  local reward_level;
  for i, reward in pairs( achievement.reward ) do

    if ( reward.get and not reward.done ) then
      reward_level = i;
      break;
    end;

  end;

  if ( reward_level ) then
    level = reward_level;
  end;
  
  local item_animfunc, description_id, progress, progress_text;

  local icon_animfunc   = "icon_state"  ..level;
  local reward_animfunc = "reward_state"..level;

  if ( achievement.count == 0 and level == 1 ) then

    icon_animfunc = "icon_state0";

  end;

  if ( reward_level ) then
   
    ui.button.Enable( item_button );

    level          = reward_level;
    item_animfunc  = "achievement_slot_button";
    description_id = str_ach..name.."_description_"..level;

    ui.button.EventSet( item_button,
      function ()
        
        ui.button.Disable( item_button );

        local count       = achievement.levels[ level ].reward_count;
        local reward_type = achievement.levels[ level ].reward_type;

        _f.GetAchievementReward( name, reward_type, count, level );
        achievement.reward = _f.GetAchievementRewardData( name );
        private.AchievementsTab_AnimateItem( achievement, 
                                             item, item_reward, item_icon,
                                             level );

        private.Tab_UpdateMarkers( "achievements", _f.GetAchievementNotificationCount() );
        
        common.sound.Play( "achievement_reward_click" );

      end );

    private.notify =
    { 
      message = "Interface_Hero_AchievementUpdated", 
      data    = name 
    };

  elseif ( level > level_max ) then

    item_animfunc  = "achievement_slot_complete";
    description_id = "str_achievement_complete_description";

    icon_animfunc   = "icon_state"  ..level;
    reward_animfunc = "reward_state"..level;

  else

    item_animfunc  = "achievement_slot_progress";
    description_id = str_ach..name.."_description_"..level;

    local step  = achievement.count;
    local steps = achievement.levels[ level ].value;

    progress = step / steps * 120;
    progress_text = ui.text.FormatNumber( step ).."/"..ui.text.FormatNumber( steps );

  end;

  local levels       = achievement.levels;
  local reward_count = levels[ math.min( level, level_max ) ].reward_count;
  local value        = levels[ math.min( level, level_max ) ].value;

  local multiset = {};

  table.insert( multiset, 
  { item, 
    { 
      animfunc         = item_animfunc,
      frame            = progress,
      progress_text    = progress_text,
      achievement_desc = StringGet( description_id, value )
    }
  } );

  table.insert( multiset, 
  { item_icon, 
    { 
      animfunc = icon_animfunc,
      playing  = true
    }
  } );

  table.insert( multiset, 
  { item_reward, 
    { 
      animfunc     = reward_animfunc,
      reward_count = ui.text.FormatNumber( reward_count )
    }
  } );

  -- Focus at achievement with reward
  if ( reward_level ) then

    local list = private.achievements.list;
    local id   = private.achievements.id[ name ] or 0;
    local pos  = math.max( id - 3, 0 );
    local max_pos = #list - 4;
    
    pos = math.min( max_pos, pos );
    
    table.insert( multiset, { scc, { scrollpos = pos } } );

  end;

  ObjMultiSet( multiset );

end;
--------------------------------------------------------------------------------
function private.AchievementsTab_AnimateItem ( achievement, 
                                               item, item_reward, item_icon,
                                               level )

  local suffix = level.."_"..( level + 1 );

  local icon_func   = "icon_state"..suffix
  local reward_func = "reward_state"..suffix;

  local icon_func_next = "icon_state"..( level + 1 );
  
  ui.anim.Play( item_icon,   icon_func,
    function ()

        ui.anim.Play( item_icon, icon_func_next );
        ui.anim.Play( item, "achievement_slot_action",
        function ()

          ui.anim.Play( item_reward, reward_func,
            function ()
              private.AchievementsTab_UpdateItem( achievement );
            end );

        end );

    end );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.InfoTab_Create ()

  private.InfoTab_AttachWearedSkin();

end;
--------------------------------------------------------------------------------
function private.InfoTab_Update ()

  local stat = private.info.statistic;

  local ach     = stat.achievements.completed_count;
  local ach_tot = stat.achievements.total_count;
  local quests  = stat.quests;
  local stg     = stat.stages.completed_count;
  local stg_tot = stat.stages.total_count - stat.stages.bonus_total;

  local id  = stat.id;
  local str = "str_int_dialog_hero_info_";
  
  local id_label = tern( id == "0" )(
    function() return ne.strings.Get( str.."id_offline" ); end,
    function() return ne.strings.Get( str.."id", id ); end
  )();

  ui.obj.Set(
    "anm_int_dialog_player_main_info_data",
    {
      achievements_label   = StringGet( str.."achievements" ),
      achievements_counter = StringGet( str.."achievements_value", ach, ach_tot ),
      quests_label         = StringGet( str.."quests" ),
      quests_counter       = StringGet( str.."quests_value", quests ),
      levels_label         = StringGet( str.."levels" ),
      levels_counter       = StringGet( str.."levels_value", stg, stg_tot ),
      player_id            = id_label;
    }
  );

  if ( private.info.skin ~= private.weared_skin ) then

    ui.skin.Detach( private.info.skin_obj );
    private.InfoTab_AttachWearedSkin();

  end;

  local animfunc = "tab_4_"..private.app_o

  if ( _f.IsNeedLogin() ) then

    animfunc = animfunc .. "_fb";

  end;

  ui.obj.Set(
    private.this.window,
    {
      animfunc = animfunc,
      title    = private.strings.title.info
    }
  );

  ui.anim.Play( private.info.skin_obj, "skinview_fun1",
    function ()
      ui.anim.Play( private.info.skin_obj, "dialog" );
    end );

end;
--------------------------------------------------------------------------------
function private.InfoTab_AttachWearedSkin ()

  local skin_name = private.weared_skin;
  local parent = "obj_int_dialog_player_main_info_skin";

  private.info.skin     = skin_name;
  private.info.skin_obj = ui.skin.Attach( parent, skin_name );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.InfoTab_Login ()

  _f.Login();
  mediator.SubscribeOnce( "Model_Social_LoginComplete", 
                          private.InfoTab_Update );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.PotionTab_Create ()
  
  local scc    = "scc_int_dialog_player_main_potions_scroller";
  local slot   = "anm_int_dialog_player_main_potions_slot";
  local button_buy = "btn_int_dialog_player_main_potions_slot_button_buy";
  local button_use = "btn_int_dialog_player_main_potions_slot_button_use";
  local str_bonus_name = "str_";
  local str_bonus_desc = "str_description_";
  local str_dlg = "str_int_dialog_hero_potions_";
  local multiset = {};

  for i, potion in pairs( private.potions.list ) do

    local name = potion.name;
    local suffix = "_"..name;

    private.potions.id[ name ] = i;

    ObjClone( slot, suffix, true, scc );

    local item = slot..suffix;

    ui.button.Create( button_use..suffix,
      {
        title  = str_dlg.."use",
        design = "medium",
        color  = "green"
      },
      function () 
        _f.PotionUse( name );
      end 
    );
    
    ui.button.Create( button_buy..suffix,
      {
        title  = potion.price.count,
        design = "price_"..potion.price.resource,
      },
      function () 
        _f.PotionBuy( name );
      end 
    );

    ui.obj.Set( item, 
    { 
      rt_icon = ui.obj.Res( "potion", name ),
      tt_bonus_name = str_bonus_name..name,
      tt_bonus_desc = str_bonus_desc..name,
    },
    multiset );

  end;

  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
function private.PotionTab_Update ()
  
  local animfunc = "tab_2_"..private.app_o
  
  private.Tab_UpdateMarkers( "potions", _f.PotionsNotification( true ) );
  
  ui.obj.Set(
    private.this.window,
    {
      animfunc = animfunc,
      title    = private.strings.title.bonus
    }
  );
  
  for _, potion in pairs( private.potions.list ) do
    private.PotionTab_Update_Item( potion );
  end;
  
  private.CheckTimer();
  
end;
--------------------------------------------------------------------------------
function private.PotionTab_Update_Item ( potion )
  
  local slot   = "anm_int_dialog_player_main_potions_slot";
  local name   = potion.name;
  local suffix = "_"..name;
  local item   = slot..suffix;
  local animfunc = "bonus_slot_button";
  local button_use = "btn_int_dialog_player_main_potions_slot_button_use"..suffix;
  local button_buy = "btn_int_dialog_player_main_potions_slot_button_buy"..suffix;
  
  if ( potion.count == 0 ) then
    ui.button.Enable( button_buy );
    ui.button.Disable( button_use );
  else
    ui.button.Enable( button_use );
    ui.button.Disable( button_buy );
  end;
  
  if ( _f.PotionIsActive( name ) ) then
    animfunc = "bonus_slot_active";
  end;
    
  ObjSet( item, 
  { 
    animfunc = animfunc,
    playing = true,
    tt_bonus_count = "x"..tostring( potion.count ),
    tt_timer = "",
    tt_bonus_not = "str_int_dialog_hero_potions_not",
    tt_timer_title = "str_int_dialog_hero_potions_timer_title"
  } );
  
end;
--------------------------------------------------------------------------------
function public.PotionsUpdate ( potions, potion, action )
  
  if ( ( not public.IsActive() ) or ( not private.potions.tab_created ) ) then
    return;
  end;
  
  private.potions.list = potions;
  
  if ( action == "update" ) then
    _f.PotionsNotification( true );
  end;
  
  private.PotionTab_Update_Item( potion );
  
  private.CheckTimer();
  
end;
--------------------------------------------------------------------------------
function private.CheckTimer ()
  
  local yet = false;
  local multiset = {};
  
  for key, value in pairs( private.potions.list ) do
    if ( _f.PotionIsActive( value.name ) ) then
      local time = _f.PotionGetRemainingTime( value.name );
      local slot_name = "anm_int_dialog_player_main_potions_slot_"..value.name;
      
      yet = true;
      ui.obj.Set( slot_name, { tt_timer = ui.text.FormatTime( time ) }, multiset );
    end;
  end;
  
  if ( yet and ( private.tab_current == "potions" ) ) then
    
    ui.obj.Set( private.potions.timer_object, 
    {  
      playing = true,
      time = 1.0,
      endtrig = private.CheckTimer
    }, 
    multiset );
    
  end;
  
  if ( #multiset ~= 0 ) then
    ui.obj.Multiset( multiset );
  end;
  
end;
--------------------------------------------------------------------------------