-- name=int_dialog_social_mail
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  ReadMessage            = function () end, 
  SendHeart              = function () end, 
  IsAttachmentObtainable = function () end,
  Hide                   = function () end,
  ShowTopPanel           = function () end,
  EnterPromocode         = function () end,
  ShowAskDialog          = function () end,
  ShowSendDialog         = function () end,
  GetMessages            = function () end,
  OnHide                 = function () end
};

local _this = {};
_this.animating = {};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  _this.app_o = ui.screen.Orientation();
  _this.path  = "assets/content/interface/social/";
  _this.types = 
  {
    custom = "red",   -- custom letter
    gift   = "green", -- letter with attached heart
    help   = "blue",  -- letter with heart request
    thanks = "blue"   -- letter with 'thanks' message
  };
  _this.messages_visible = 3;

  _this.dialog = "obj_int_dialog_social_mail";
  _this.button = {
    close = "btn_int_dialog_social_mail_close"
  };
  _this.scroller   = "scc_int_dialog_social_mail_slots";
  _this.mail_photo = "spr_int_dialog_social_mail_letter_";

  private.CreateButtons();
  private.Adapt();

  private.strings =
  {
    letter_title_help   = ne.strings.Get( "str_int_dialog_social_mail_title_help" ),
    letter_title_gift   = ne.strings.Get( "str_int_dialog_social_mail_title_gift" ),
    letter_title_thanks = ne.strings.Get( "str_int_dialog_social_mail_title_thanks" ),
    letter_title_custom = ne.strings.Get( "str_int_dialog_social_mail_title_custom" ),
    notavailable        = ne.strings.Get( "str_int_dialog_social_mail_heart_unavailable" ),
    removed             = ne.strings.Get( "str_int_dialog_social_mail_letter_removed" ),
    sent                = ne.strings.Get( "str_int_dialog_social_mail_letter_sent" ),
    received            = ne.strings.Get( "str_int_dialog_social_mail_letter_received" ),
    thank               = ne.strings.Get( "str_int_dialog_social_mail_thanks" )
  };

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Adapt ()
  
  local window_area = {
    portrait  = { -279, -388, 540, 788 },
    landscape = { -223, -304, 433, 631 }
  };

  local multiset = {};

  ui.obj.ToCenter( _this.dialog, 0, 0, multiset );
  ui.window.SetArea( _this.dialog, window_area[ _this.app_o ], multiset );

  ui.obj.Set(
    "scc_int_dialog_social_mail",
    { 
      enddragtrig = function () private.Scroller_OnScroll(); end
    }
  );

  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
function private.CreateButtons ()
  
  ui.button.Create(
    _this.button.close,
    {
      design = "close"
    },
    function() public.Hide(); end
  );
  
  ui.button.Create(
    "btn_int_dialog_social_mail_enter_promocode",
    {
      title  = "str_int_dialog_social_mail_enter_promocode",
      design = "large",
      color  = "blue"
    },
    function() _f.EnterPromocode(); end
  );
  
  ui.button.Create(
    "btn_int_dialog_social_mail_ask",
    {
      title  = "str_int_dialog_social_mail_ask_hearts",
      design = "facebook",
      sound  = "simply_click"
    },
    function() _f.ShowAskDialog(); end
  );

  ui.button.Create(
    "btn_int_dialog_social_mail_send",
    {
      title  = "str_int_dialog_social_mail_send_hearts",
      design = "medium",
      color  = "green"
    },
    function() _f.ShowSendDialog(); end
  );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show ( mail, mail_total, is_login, is_promocode, is_invite )

  if ( _this.is_active ) then return; end;
  
  _this.is_active    = true;
  _this.mail_load    = #mail;
  _this.mail_total   = mail_total;
  _this.is_full_load = false;

  --DbgTrace( "[ INFO ][ int_dialog_social_mail ] Total mail: ".._this.mail_total );
  --DbgTrace( "[ INFO ][ int_dialog_social_mail ] Loaded so far: ".._this.mail_load );
  
  private.SetDefault( mail, is_login, is_promocode, is_invite );

  ui.window.Open(
    "social_mail",
    _f.ShowTopPanel,
    nil,
    "daily_quest"
  );

end;
--------------------------------------------------------------------------------
function public.Hide ( func_end )

  if ( not _this.is_active ) then return; end;
  _this.is_active = false;
  
   _f.Hide();
  
  if ( ui.popup.Flat_AttachedTo() ) then
    ui.popup.Flat_Hide();
  end;

  ui.window.Close(
    "social_mail",
    function()
      private.Cleanup();
      _f.OnHide();
      if ( func_end ) then func_end(); end;
    end
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.SetDefault ( mail, is_login, is_promocode, is_invite )

  ui.obj.Clone(
    "scc_int_dialog_social_mail",
    "anm_int_dialog_social_mail_window",
    "_slots"
  );

  local multiset = {};
  local is_mail_empty = ( _this.mail_load == 0 );
    
  ui.obj.Set(
    "anm_int_dialog_social_empty",
    {
      visible = is_mail_empty,
      active  = is_mail_empty
    },
    multiset
  );
  
  local animfunc = tern( is_promocode )(
    _this.app_o.."_promocode",
    _this.app_o
  );
  
  if ( not is_login ) then
    animfunc = _this.app_o.."_logout";
    if ( is_promocode ) then
      animfunc = _this.app_o.."_logout_promocode";
    end;
  else
    if ( is_invite ) then
      animfunc = tern( is_promocode )(
        _this.app_o.."_invite_promocode",
        _this.app_o.."_invite"
      );
    end;
  end;
  
  ui.obj.Set(
    "anm_int_dialog_social_mail_window",
    {
      animfunc = animfunc,
      tt_desc  = tern( is_mail_empty )( "str_int_dialog_social_mail_empty", "" ),
      title    = ne.strings.Get( "str_int_dialog_social_mail_caption" ),
      playing  = true
    },
    multiset
  );
  
  for i, message in pairs( mail ) do

    private.Message_Setup( i, message, multiset );

  end;
    
  tern( #multiset > 0 )(
    function() ObjMultiSet( multiset ); end, 
    function() end 
  )();

  for i = 1, _this.mail_load do
    ui.obj.SpriteAdjust( _this.mail_photo..i, 100 );
  end;

  --JK: dummies if all mail loaded at start
  if ( _this.mail_total == _this.mail_load ) then

    for i = ( _this.mail_load + 1 ),
            ( _this.mail_total + _this.messages_visible - 1 ) do

      ui.obj.Clone( 
        "obj_int_dialog_social_mail_slot",
        "scc_int_dialog_social_mail_slots",
        "_"..i
      );

    end;

  end;
  
end;
--------------------------------------------------------------------------------
function private.Cleanup ()

  ui.obj.Delete( "scc_int_dialog_social_mail_slots" );

  ui.buttons.Delete(
    "int_dialog_social_mail",
    {
      _this.button.close,
      "btn_int_dialog_social_mail_enter_promocode",
      "btn_int_dialog_social_mail_ask",
      "btn_int_dialog_social_mail_send",
      "btn_int_dialog_social_mail_invite"
    }
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Scroller_OnScroll ()

  if ( _this.mail_load <= _this.messages_visible ) then return; end;
  if ( _this.mail_load > _this.mail_total        ) then return; end;

  local scrollpos   = ui.obj.Get( _this.scroller ).scrollpos;
  local next_loaded = ( _this.mail_load - scrollpos );

  if ( next_loaded <= ( 2 * _this.messages_visible ) ) then

    local multiset = {};
    local messages = _f.GetMessages( _this.mail_load + 1, _this.messages_visible );

    if ( #messages > 0 ) then

      for i, message in pairs( messages ) do

        private.Message_Setup( _this.mail_load + i, message, multiset );

      end;

      ui.obj.Multiset( multiset );

      for i, message in pairs( messages ) do
        local id = _this.mail_load + i;
        ui.obj.SpriteAdjust( _this.mail_photo..id, 100 );
      end;

      _this.mail_load = _this.mail_load + #messages;

      if ( _this.mail_load >= _this.mail_total ) then
        _this.is_full_load = true;
        --DbgTrace( "[ INFO ][ int_dialog_social_mail ] All mail loaded" );
      end;

    else

      _this.is_full_load = true;
      --DbgTrace( "[ INFO ][ int_dialog_social_mail ] All mail loaded" );

    end;

  end;

  --DbgTrace( "[ INFO ][ int_dialog_social_mail ] Loaded so far: ".._this.mail_load );

  if ( _this.is_full_load ) then

    local max_load = ( _this.mail_load + _this.messages_visible - 1 );

    --JK: dummies
    for i = ( _this.mail_load + 1 ), max_load do

      ui.obj.Clone( 
        "obj_int_dialog_social_mail_slot",
        "scc_int_dialog_social_mail_slots",
        "_"..i
      );

      _this.mail_load = _this.mail_load + 1;

    end;

  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Message_Setup ( i, message, multiset )

  local message_type = "custom";
  
  for key, value in pairs( _this.types ) do
    if ( message.text.string:find( key ) ) then
      message_type = key;
      break;
    end;
  end;

  local obj_slot = "obj_int_dialog_social_mail_slot_"..i;

  if ( not ui.obj.Exists( obj_slot ) ) then

    ui.obj.Clone(
      "obj_int_dialog_social_mail_slot",
      "scc_int_dialog_social_mail_slots",
      "_"..i,
      1
    );

  end;
  
  -- Set message and title
  local title, msgtext = "dummy", "dummy";
  local msgtype, str   = message_type, message.text.string;
  local params         = message.text.params;
  
  if ( msgtype ~= "custom" or str == "" ) then

    title = private.strings[ "letter_title_"..message_type ];
    
    local attach = ne.strings.Get(
      "str_int_dialog_social_mail_attach_heart", 
      message.attach.name
    );

    local name_friend = message.from.name;
    
    if ( message.from.userId == 3 ) then
      name_friend = ne.strings.Get( "str_social_bot_name_3" );
    end;
    
    msgtext = ne.strings.Get(
      "str_int_dialog_social_mail_message_"..message_type,
      name_friend,
      attach
    );
  
  else

    title   = ne.strings.Get( str.."_title", table.unpack( params ) );
    msgtext = ne.strings.Get( str, table.unpack( params ) );
  
  end;
    
  local expdate = tern( message.expDate.month == 0 and 
                        message.expDate.day   == 0 and
                        message.expDate.year  == 0 )
                  (
                    "", 
                    ne.strings.Get(
                      "str_int_dialog_social_mail_expdate",
                      message.expDate.month,
                      message.expDate.day,
                      message.expDate.year
                  ) );

  ui.obj.Set(
    "anm_int_dialog_social_mail_slot_"..i,
    {
      animfunc = "mail_static"
    },
    multiset
  );
  
  local reward_count = tern( message.attach and message.attach.count )(
    message.attach.count,
    0
  );
  
  reward_count = tern( reward_count > 1 )( reward_count, "" );

  if ( message.attach                       ) and
     ( message.attach.name == "immortality" ) then

    reward_count = "";

  end;

  ui.obj.Set(
    "anm_int_dialog_social_mail_letter_"..i,
    {
      animfunc     = _this.types[ message_type ],
      letter_title = title,
      deleted      = expdate,
      message      = msgtext,
      tt_count     = reward_count
    },
    multiset
  );

  if ( message.from.avatar == '' ) then message.from.avatar = nil; end;

  local icon_res         = _this.path.."default_avatar";
  local icon_default_res = _this.path.."default_avatar";
  
  if ( message.from.userId == 3 ) then

    icon_res         = _this.path.."bot_avatar_3";
    icon_default_res = _this.path.."bot_avatar_3";
    
  end;
  
  ui.obj.Set(
    _this.mail_photo..i,
    {
      url           = message.from.avatar,
      res           = icon_res,
      default_res   = icon_default_res,
      trg_resloaded = function ()
        ui.obj.SpriteAdjust( _this.mail_photo..i, 100 );
      end
    },
    multiset 
  );

  local btn_action_name = "btn_int_dialog_social_mail_letter_action_"..i;
  
  -- Default button is "DELETE".
  ui.button.Create(
    btn_action_name,
    {
      color  = "green",
      design = "medium"
    },
    function () private.Message_MarkRemoved( i, message.messageId ); end
  );
  
  local btn_delete_name = "btn_int_dialog_social_mail_letter_delete_"..i;
  
  -- Default button is "DELETE".
  ui.button.Create(
    btn_delete_name,
    {
      design = "delete"
    },
    function () private.Message_MarkRemoved( i, message.messageId ); end
  );

  ui.button.SetText(
    btn_action_name,
    "str_int_dialog_social_mail_btn_delete",
    multiset
  );
  
  ui.obj.Disable(
    "spr_int_dialog_social_mail_letter_attach_"..i, false, multiset
  );

  if ( message.attach ) then

    local func_get_heart = function ()

      local is_obtainable, reason = _f.IsAttachmentObtainable( message.attach );

      if ( is_obtainable ) then

        private.Message_MarkReceived( i, message.messageId );

      else

        ui.input.Off( btn_action_name );

        ui.obj.SetTextTag(
          "anm_int_dialog_social_mail_slot_"..i,
          "notavailable",
          private.strings.notavailable
        );

        ui.anim.Play(
          "anm_int_dialog_social_mail_slot_"..i,
          "mail_notavailable",
          function () ui.input.On( btn_action_name ); end,
          "mail_static"
        );

      end;

    end;
    
    local func_get_other = function ( button_name )

      local is_obtainable, reason = _f.IsAttachmentObtainable( message.attach );
    
      if ( is_obtainable ) then
      
        private.Message_MarkReceived( i, message.messageId );
      
      elseif ( reason == "wait" ) then
        
        ui.popup.Flat_Show(
          "str_dialog_social_mail_locked_title",
          "str_dialog_social_mail_locked_message",
          btn_action_name,
          { 0, -130 }
        );
        
      end;
      
    end;

    ui.button.SetText(
      btn_action_name,
      "str_int_dialog_social_mail_btn_claim",
      multiset
    );

    ui.button.EventSet( btn_action_name,
      function ()    
        tern( message_type == "gift" )( func_get_heart, func_get_other )();    
      end
    );
  
    local reward_name = message.attach.name;
    local reward_type = message.attach.type;
    local scales =
    {
      booster   = 0.6,
      chest     = 0.7,
      heart     = 1.2,
      container = 0.7,
    };
    local reward_scale = scales[ reward_type ] or 1.0;
    
    if ( reward_type == "gold" ) then    
      reward_name = "gold_1";
    elseif ( reward_type == "container" ) then    
      reward_name = "goldchest";      
    elseif ( reward_type == "chest" and reward_name:find( "event" ) ) then    
      -- TODO: Fix it in 1.12
      reward_name = "woodchest";    
    end;
    
    local res = ui.obj.Res( reward_type, reward_name );

    if ( message_type == "help"   ) or
       ( message_type == "thanks" ) then
      res = "assets/content/interface/social/icon_"..message_type;
    end;
    
    ui.obj.Set(
      "spr_int_dialog_social_mail_letter_attach_"..i,
      {
        active  = true,
        visible = true,
        input   = false,
        res     = res,
        scale_x = reward_scale,
        scale_y = reward_scale
      },
      multiset
    );
    
  end;
  
  -- Gift message (heart). Button "CLAIM".
  if ( message_type == "gift" ) then
    --
  end;

  -- Help message. Button "SEND".
  if ( message_type == "help" ) then
    
    ui.button.SetText(
      btn_action_name,
      "str_int_dialog_social_mail_btn_help",
      multiset
    );

    ui.button.EventSet(
      btn_action_name,
      function ()
        private.Message_MarkSent( i, message.messageId );
        _f.SendHeart( message.from.userId );
      end
    );

  end;

  -- Thanks message. Button "DELETE".
  if ( message_type == "thanks" ) then

    ui.button.SetText(
      btn_action_name,
      "str_int_dialog_social_mail_btn_delete",
      multiset
    );

    ui.button.EventSet(
      btn_action_name,
      function ()
        private.Message_MarkRemoved( i, message.messageId );
      end
    );
        
  end;
  
  -- Custom message (event, admin, etc).
  if ( message_type == "custom" ) then
    
    ui.obj.Set(
      _this.mail_photo..i,
      {
        url           = message.from.avatar,
        res           = _this.path.."admin_avatar",
        default_res   = _this.path.."admin_avatar",
        trg_resloaded = function ()
          ui.obj.SpriteAdjust( _this.mail_photo..i, 100 );
        end
      },
      multiset
    );

  end;

end;
--------------------------------------------------------------------------------
function private.Message_MarkRemoved ( index, message_id )

  private.Message_MarkAsRead( index, message_id, "removed" );

end;
--------------------------------------------------------------------------------
function private.Message_MarkReceived ( index, message_id )

  private.Message_MarkAsRead( index, message_id, "received" );

end;
--------------------------------------------------------------------------------
function private.Message_MarkSent ( index, message_id )

  private.Message_MarkAsRead( index, message_id, "sent" );

end;
--------------------------------------------------------------------------------
function private.Message_MarkAsRead ( index, message_id, operation_type )

  local animfunc = {
    removed  = "mail_delete",
    received = "mail_received",
    sent     = "mail_sent"
  };

  ui.input.Off({
    "anm_int_dialog_social_mail_slot_"..index,
    _this.button.close
  });

  _this.animating[ "id_"..message_id ] = true;

  ui.obj.SetTextTag(
    "anm_int_dialog_social_mail_slot_"..index,
    operation_type,
    private.strings[ operation_type ]
  );

  ui.anim.Play(
    "anm_int_dialog_social_mail_slot_"..index,
    animfunc[ operation_type ],
    function ()
      _f.ReadMessage( message_id );
      _this.animating[ "id_"..message_id ] = nil;
      if ( next( _this.animating ) ) then return; end;
      ui.input.On( _this.button.close );
    end
  );

end;
--------------------------------------------------------------------------------