-- name=int_dialog_stage_continue
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  OnShow           = function () end,
  OnHide           = function () end,
  BuyBooster       = function () end,
  Reject           = function () end,
  ShowVideo        = function () end,
  GetAttemptsVideo = function () end,
  IsVideoAvailable = function () end,
  VideoCanWatch    = function () end,
  GetTargets       = function () end,
  GetSection       = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.__InitObjects ()
  
  private.app_w = ui.screen.Width();
  private.app_h = ui.screen.Height();
  private.app_o = ui.screen.Orientation();

  private.this = {
    dialog = "obj_int_dialog_stage_continue",
    window = {
      title = "str_int_dialog_stage_continue_caption",
      anim  = "anm_int_dialog_stage_continue_window",
      desc  = "str_int_dialog_stage_continue_description",
      label = "str_int_dialog_stage_continue_label",
      count = "str_int_dialog_stage_continue_count"
    },
    button = {
      accept = "btn_int_dialog_stage_continue_accept",
      close  = "btn_int_dialog_stage_continue_close",
      video  = "btn_int_dialog_stage_continue_video"
    },
    targets = {
      panel  = "anm_int_dialog_stage_continue_targets",
      goal_1 = "anm_int_dialog_stage_continue_targets_goal_1",
      goal_2 = "anm_int_dialog_stage_continue_targets_goal_2",
      goal_3 = "anm_int_dialog_stage_continue_targets_goal_3",
      goals  = "str_int_dialog_stage_continue_goals",
      score  = "str_int_dialog_stage_continue_score"
    }
  };

  private.window_area = {
    default  = { -0.5 * 420, -0.6 * 380, 420, 380 }
  };
  
  ui.window.SetArea( private.this.dialog, private.window_area.default );

  private.Adapt();
  private.CreateButtons();
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Adapt ()

  ui.obj.ToCenter( private.this.dialog, 0, 0 );

end;
--------------------------------------------------------------------------------
function private.CreateButtons ()

  ui.button.Create(
    private.this.button.accept,
    {
      title  = "str_int_dialog_stage_continue_use",
      color  = "green",
      design = "medium",
      price  = true
    },
    function() _f.BuyBooster(); end
  );

  ui.button.Create(
    private.this.button.close,
    {
      design = "close"
    },
    function() _f.Reject(); end
  );

  if ( IsMobile() ) then
    
    ui.button.Create(
      private.this.button.video,
      {
        title  = "str_int_dialog_stage_continue_video",
        design = "video"
      },
      function() _f.ShowVideo(); end
    );
    
  end;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Show ( continue_data )
  
  private.is_active = true;
  
  private.Window_Set( continue_data );

  ui.window.Open(
    "stage_continue",
    _f.OnShow,
    { outer_disable = true },
    "click_energy"
  );

end;
--------------------------------------------------------------------------------
function public.Hide ( func_end )

  private.is_active = false;
  
  ui.window.Close(
    "stage_continue",
    function ()

      if ( func_end ) then func_end(); end;

      _f.OnHide();
      
    end
  );

end;
--------------------------------------------------------------------------------
function public.Update ()
  
  if ( not private.is_active ) then return; end;
  
  private.SetButtonAttemptsVideo();
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Window_Set ( continue_data )
  
  local continue_name, continue_count, continue_price = 
        continue_data.name, continue_data.count, continue_data.price;

  local title = private.this.window.title.."_"..continue_name;
  local label = private.this.window.label.."_"..continue_name;
  local desc  = private.this.window.desc.."_"..continue_name;
  desc = tern( continue_count > 0 )( desc.."_exist", desc );
  
  local animfunc = private.app_o.."_"..
                   string.match( continue_name, "extra(%w+)" ).."_"..
                   tern( continue_count > 3 )( 3, continue_count );
                   
  local count = ne.strings.Get( private.this.window.count, continue_count );
  
  local anm_price    = "anm_price_int_dialog_stage_continue_accept";
  local multiset     = {};
  local section_name = _f.GetSection();
  local targets      = _f.GetTargets();
  local target_count = 0;
  
  for target_type, target_data in pairs( targets ) do
    for index, target in pairs( target_data ) do
      
      target_count = target_count + 1;
      
      if ( target_type == "GainScore" ) then
        ui.obj.Set(
          private.this.targets.panel,
          {
            tt_score_count = ui.text.FormatNumber( target.progress, true )
          },
          multiset
        );
      end;
      
      local count;
      if ( target_type ~= "GainScore" ) then
        count = ui.text.FormatNumber( target.progress, true ).."/"..ui.text.FormatNumber( target.count, true );
      else
        count = ui.text.FormatNumber( target.count, true );
      end;
      
      ui.obj.Set(
        private.this.targets[ "goal_"..target_count ],
        {
          animfunc = tern( target.done )( "complete_goal", "static_goal" ),
          tt_count = count,
          rt_icon  = ui.obj.Res( "goal", target, nil, section_name )
        },
        multiset
      );
    
    end;
  end;
  
  ui.obj.Set(
    private.this.targets.panel,
    {
      animfunc = "targets_"..target_count,
      tt_goals = ne.strings.Get( private.this.targets.goals ),
      tt_score = ne.strings.Get( private.this.targets.score )
    },
    multiset
  );
  
  ui.obj.Set( anm_price, { tt_count = continue_price }, multiset );

  if ( continue_count > 0 ) then
    
    ui.obj.Disable( anm_price, false, multiset );
  
  else
    
    animfunc = private.app_o.."_"..
               string.match( continue_name, "extra(%w+)" )..
               "_3";
    count    = ne.strings.Get( private.this.window.count, 3 );
    
    ui.obj.Enable( anm_price, true, multiset );
  
  end;
  
  ui.obj.Set( private.this.window.anim,
  {
    title       = title,
    description = desc,
    animfunc    = animfunc,
    label       = label, 
    count       = count,
    playing     = true
  },
  multiset );

  private.SetButtonAttemptsVideo();
  
  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
function private.SetButtonAttemptsVideo ()
  
  local show_button = true;
  local attempts, max_attempts = _f.GetAttemptsVideo();
  local is_video_available = _f.IsVideoAvailable();
  
  if ( attempts >= max_attempts ) or ( not is_video_available ) then
    show_button = false;
  end;
  
  _f.VideoCanWatch( show_button );
  
  ui.obj.Set( private.this.button.video, 
  { 
    visible = show_button, 
    input   = show_button 
  } );
  
end;
--------------------------------------------------------------------------------