-- name=int_dialog_stage_exit
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Accept = function () end,
  Reject = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  private.this = {
    int = "int_dialog_stage_exit",
    obj = "obj_int_dialog_stage_exit",
    anm = {
      window = "anm_int_dialog_stage_exit_window"
    },
    button = {
      exit  = "btn_int_dialog_stage_exit_exit",
      close = "btn_int_dialog_stage_exit_close"
    }
  };

  private.strings = {
    message_default     = ne.strings.Get( "str_dialog_stage_exit_message_default" ),
    message_immortality = ne.strings.Get( "str_dialog_stage_exit_message_immortality" ),
    message_nomoves     = ne.strings.Get( "str_dialog_stage_exit_message_no_moves" )
  };

  private.window_area = {
    default = { -272, -220, 544, 440 },
    heart   = { -272, -220, 544, 540 }
  };

  local multiset = {};

  ui.obj.ToCenter( private.this.obj, 0, 0, multiset );

  ui.window.SetArea(
    private.this.obj,
    private.window_area.default,
    multiset
  );

  ui.obj.Multiset( multiset );

  ui.button.Create(
    private.this.button.exit,
    {
      title  = "str_dialog_stage_exit_exit",
      design = "medium",
      clear  = "window_outer_tap",
      color  = "blue",
      sound  = "ok_info"
    },
    function() _f.Accept(); end
  );

  ui.button.Create(
    private.this.button.close,
    {
      design = "close"
    },
    function() _f.Reject(); end
  );  

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show (
    stage_data,
    state_window_failed,
    heart_count,
    hearts_max,
    skin_current,
    is_immortality
  )

  if ( private.active ) then return; end;
  
  private.active      = true;
  private.state_window_failed  = state_window_failed;
  private.heart_count = heart_count;
  private.immortality = is_immortality;

  private.hero = ui.skin.Attach( private.this.anm.window, skin_current );

  ui.obj.Set(
    private.hero,
    {
      animfunc = "dialog",
      playing  = true,
      anim_tag = "hero"
    }
  );

  private.SetDefault( stage_data, heart_count, hearts_max );

  ui.window.Open(
    "stage_exit",
    nil,
    nil,
    "app_info"
  );

end;
--------------------------------------------------------------------------------
function public.Hide ( func_execute, no_animation )

  if ( not private.active ) then return; end;

  private.active = false;
  
  if ( private.state_window_failed ) and ( not no_animation ) then

    local func_end = function () private.Close( func_execute ); end;
  
    if ( not private.immortality ) then

      ui.anim.Play( private.hero, "fail_dialog" );

    end;
   
    ui.anim.Play(
      "anm_int_dialog_stage_exit_heart",
      "heart",
      func_end
    );

    ui.anim.Play(
      "anm_int_dialog_stage_exit_heart_dummy",
      "heart_dummy",
      function()

        if ( not private.immortality ) then

          ui.obj.Set(
            "anm_int_dialog_stage_exit_heart",
            {
              heart_count = tostring( private.heart_count - 1 )
            }
          );

        end;

      end
    );

    ui.input.Off({
      private.this.button.close,
      private.this.button.exit
    });

    ui.anim.FadeOut( 0.3, private.this.button.close );

    common.sound.Play( "heart_break" );

  else

    private.Close( func_execute );

  end;

end;
--------------------------------------------------------------------------------
function public.IsActive ()

  return private.active;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.SetDefault ( stage_data, heart_count, hearts_max )

  local multiset       = {};
  local animfunc       = "quit_nomoves_portrait";
  local message = tern( private.immortality )(
    private.strings.message_immortality,
    private.strings.message_default
  );

  if ( private.immortality ) then

    heart_count = ui.text.InfinitySign();

  else

    heart_count = tostring( math.min( heart_count, hearts_max ) );

  end;

  ui.obj.Set(
    private.this.button.close,
    {
      input = true,
      alp   = 1.0
    },
    multiset
  );
  
  ui.input.On( private.this.button.exit, multiset );

  if ( private.state_window_failed ) then

    animfunc = "quit_portrait";

    ui.window.SetArea( private.this.obj, private.window_area.heart );

    ui.obj.Set(
      "anm_int_dialog_stage_exit_heart",
      {
        heart_count = heart_count,
        active      = true,
        visible     = true
      },
      multiset
    );
    
    ui.obj.Disable( "anm_int_dialog_stage_exit_character", true, multiset );

  else

    ui.obj.Disable( "anm_int_dialog_stage_exit_heart", true, multiset );
    ui.obj.Enable( "anm_int_dialog_stage_exit_character", true, multiset );
    
    message = private.strings.message_nomoves;

  end;
  
  local zone = stage_data.zone.background;
  local background_res =
  "assets/content/"..zone.content.."/bg_"..zone.name.."_"..stage_data.type;
  
  if ( stage_data.type == "fight" ) then
    background_res = "assets/content/zones_1/bg_arena_fight";
  end;
  
  ui.obj.Set(
    "anm_int_dialog_stage_exit_window",
    {
      animfunc  = animfunc,
      title     = common.string.FormatNameStage( stage_data.name ),
      attention = message,
      bg        = background_res
    },
    multiset
  );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.Close ( func_execute )

  ui.window.Close( "stage_exit", function ()

    ui.skin.Detach( private.hero );

    if ( func_execute ) then

      func_execute();

    end;

  end );

end;
--------------------------------------------------------------------------------