-- name=int_dialog_stage_fail
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Close              = function () end,
  Retry              = function () end,
  Restore            = function () end,
  ShowVideo          = function () end,
  IsImmortality      = function () end,
  GetHeartsData      = function () end,
  GetAttemptsVideo   = function () end,
  IsEventExist       = function () end,
  EventTimeRemaining = function () end,
  IsVideoAvailable   = function () end,
  VideoCanWatch      = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  private.this   = {
    obj = "obj_int_dialog_stage_fail",
    anm = "anm_int_dialog_stage_fail_window",
    explosion = "anm_int_dialog_stage_fail_explosion",
    buttons = {
      video = "btn_int_dialog_stage_fail_video"
    },
    timer = "tmr_int_dialog_stage_fail_count"
  };
  
  private.app_o = ui.screen.Orientation();

  private.window_area = {
    default = { -272, -255, 544, 510 }
  };
  
  ui.window.SetArea( private.this.obj, private.window_area.default );
  
  private.Adapt();
  private.CreateButtons();
  private.CreateGoals();
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show (
    stage_data,
    targets,
    heart_count,
    heart_max,
    skin_current,
    is_retry_allowed
  )

  private.active = true;

  private.hero = ui.skin.Attach( private.this.anm, skin_current );

  local score    = 0;
  local multiset = {};
  local is_immortality = _f.IsImmortality();

  local hero_animfunc = tern( is_immortality )( "dialog", "fail_dialog" );

  ui.obj.Set(
    private.hero,
    {
      animfunc = hero_animfunc,
      playing  = true,
      anim_tag = "hero"
    },
    multiset
  );

  for g = 1, 3 do
    
    ui.obj.Disable( "anm_int_dialog_stage_fail_goal_"..g, false, multiset );
    
  end;

  ui.obj.Set( 
    private.this.explosion, 
    { 
      visible = tern( stage_data.explosion_field )( true, false ) 
    },
    multiset
  );
  
  if ( stage_data.type == "ride"  ) or
     ( stage_data.type == "bonus" ) then

    score = private.SetRide( targets, stage_data, multiset );

  elseif( stage_data.type == "fight" ) then
    
    score = private.SetFight( targets, stage_data, multiset );
    
  end;
  
  local zone = stage_data.zone.background;
  local background_res =
  "assets/content/"..zone.content.."/bg_"..zone.name.."_"..stage_data.type;
  
  local button_text = tern( is_retry_allowed )(
    "str_dialog_stage_fail_retry",
    "str_dialog_stage_fail_ok"
  );
  
  ui.button.SetText( "btn_int_dialog_stage_fail_retry", button_text, multiset );
  
  ui.obj.Set(
    private.this.anm,
    {
      title       = common.string.FormatNameStage( stage_data.name ),
      animfunc    = private.app_o,
      frame       = 0,
      playing     = false,
      bg          = background_res,
      goals       = "str_dialog_stage_fail_targets",
      description = "str_dialog_stage_fail_fail",
      score       = StringGet( "str_dialog_stage_fail_current_score", score )
    },
    multiset
  );

  ui.obj.Multiset( multiset );

  private.SetButtonAttemptsVideo();
  private.CheckTimer( heart_count, heart_max );
  
  ui.window.Open(
    "stage_fail",
    function()
      private.AnimHearts( heart_count, -1 );
    end
  );

end;
--------------------------------------------------------------------------------
function public.Hide ( func_end )

  if ( not private.active ) then

    return;

  end;

  private.active = false;
  
  ui.window.Close(
    "stage_fail",
    function()

      ui.skin.Detach( private.hero );

      if ( func_end ) then

        func_end();

      end;

    end
 );

end;
--------------------------------------------------------------------------------
function public.Update ( reason )
  
  if ( not private.active ) then return; end;
  
  local heart_count, heart_max = _f.GetHeartsData();
  
  private.SetButtonAttemptsVideo();
  private.CheckTimer( heart_count, heart_max );
  
  if ( reason == "video_hearts_add" ) or ( reason == "hearts_add" ) then
    heart_count = heart_count - 1;
    private.AnimHearts( heart_count, 1 );
  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.CreateGoals ()
  
  for g = 1, 3 do
    
    ObjClone( 
      "anm_int_dialog_stage_fail_goal", 
      "_"..g, 
      1, 
      private.this.anm 
    );
    
    ObjSet( "anm_int_dialog_stage_fail_goal_"..g, {
      anim_tag = "goal_"..g
    } );
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.CreateButtons ()
  
  ui.button.Create(
    "btn_int_dialog_stage_fail_close",
    {
      design = "close",
      sound  = "click_wood"
    },
    function() _f.Close(); end
  );
  
  ui.button.Create(
    "btn_int_dialog_stage_fail_retry",
    {
      title  = "str_dialog_stage_fail_retry",
      color  = "green",
      design = "medium"
    },
    function() 
      local count = _f.GetHeartsData();
      
      if ( count > 0 ) then
        _f.Retry();
      else
        _f.Restore();
      end;
    end
  );
  
  if ( IsMobile() ) then
    ui.button.Create(
      private.this.buttons.video,
      {
        title  = "str_int_dialog_hearts_video",
        design = "video"
      },
      function() _f.ShowVideo(); end
    );
  end;
  
end;
--------------------------------------------------------------------------------
function private.Adapt ()

  private.app_w       = ui.screen.Width();
  private.app_h       = ui.screen.Height();
  private.orientation = ui.screen.Orientation();
  
  ui.obj.ToCenter( private.this.obj );

end;
--------------------------------------------------------------------------------
function private.SetRide ( targets, stage_data, objects )

  local score;
  
  stage_data.section = stage_data.section or "section_1";
  
  local resource = {
    default    = "items/icon_%s",
    DropItem   = stage_data.section.."/icon_%s_%s",
    DestroyGem = "items/icon_%s"
  };
  
  local chests_resources = {
    woodchest = "items/icon_%s",
    silverchest = "items/icon_%s",
    goldchest = "items/icon_%s"
  };
  
  for t_type, t_table in pairs( targets ) do

    for t, target in pairs( t_table ) do

      local icon_postfix = "";
      local done_prefix  = "un";
      local text;
      
      if not ( target.type == "GainScore" ) then
      
        icon_postfix = "_icon";
        
        table.insert( objects, { "obj_int_dialog_stage_fail_goal_icon_"..t, { 
          visible = true
        } } );
        
        text = ui.text.FormatNumber( target.count, true );
        
        local res_part = string.format( resource[ target.type ] or resource.default, 
                                        string.lower( target.name ), 
                                        stage_data.section:match( "%d+" ) );
        
        if ( target.type == "DropItem" and chests_resources[ target.name ] ) then
          res_part = string.format( chests_resources[ target.name ], target.name );
        end;
                                        
        table.insert( objects, { "spr_int_dialog_stage_fail_goal_icon_"..t, {
          res     = "assets/content/"..res_part,
          visible = true
        } } );
        
      else
        
        score = ui.text.FormatNumber( target.progress or 0, true );
        text  = ui.text.FormatNumber( target.count or 0, true ); 
        
        table.insert( objects, { "obj_int_dialog_stage_fail_goal_icon_"..t, { 
          visible  = false
        } } );
        
      end;
      
      if ( target.done ) then
        
        done_prefix = "";
        
      end;
      
      table.insert( objects, { "anm_int_dialog_stage_fail_goal_"..t, {
        active   = true,
        visible  = true,
        animfunc = done_prefix.."performed"..icon_postfix,
        text     = text
      } } );

    end;

  end;
  
  return score;
  
end;
--------------------------------------------------------------------------------
function private.SetFight ( targets, stage_data, objects )
    
  table.insert( objects, { "anm_int_dialog_stage_fail_goal_2", {
    active   = true,
    visible  = true,
    animfunc = "unperformed",
    text     = ne.strings.Get(
        "str_dialog_stage_fail_battle",
        ne.strings.Get( "str_enemy_name_gender_"..stage_data.enemy.name )
      )
  } } );
  
  table.insert( objects, { "obj_int_dialog_stage_fail_goal_icon_2", {
    visible = false
  } } );
  
end;
--------------------------------------------------------------------------------
function private.SetButtonAttemptsVideo ()
  
  local is_immortality = _f.IsImmortality();
  local attempts, max_attempts = _f.GetAttemptsVideo();
  local count, max_count = _f.GetHeartsData();
  local show_button = true;
  local is_video_available = _f.IsVideoAvailable();
  
  if ( attempts >= max_attempts ) or ( is_immortality ) or 
    ( count >= max_count ) or ( not is_video_available ) then
    show_button = false;
  end;
  
  _f.VideoCanWatch( show_button );
  
  ui.obj.Set( private.this.buttons.video, 
  { 
    visible = show_button, 
    input = show_button 
  } );
  
end;
--------------------------------------------------------------------------------
function private.AnimHearts ( heart_count, delta )
  
  ui.anim.Play( "anm_int_dialog_stage_fail_heart", "heart" );
  
  if( heart_count == 0 ) then  common.sound.Play( "need_live" ); end;

  ui.anim.Play(
    "anm_int_dialog_stage_fail_heart_dummy",
    "heart_dummy",
    function()

      local is_immortality = _f.IsImmortality();
      
      if ( not is_immortality ) then

        ui.obj.SetTextTag(
          "anm_int_dialog_stage_fail_heart",
          "heart_count",
          ( heart_count + delta )
        );

      else
        
        ui.obj.SetTextTag(
          "anm_int_dialog_stage_fail_heart",
          "heart_count",
          ui.text.InfinitySign()
        );
        
      end;

    end
  );
  
end;
--------------------------------------------------------------------------------
function private.CheckTimer ( heart_count, heart_max )
  
  local is_immortality = _f.IsImmortality();
  
  if ( not is_immortality ) then

    ui.obj.SetTextTag(
      "anm_int_dialog_stage_fail_heart",
      "heart_count",
      math.min( heart_count, heart_max )
    );

  else

    ui.obj.SetTextTag(
      "anm_int_dialog_stage_fail_heart",
      "heart_count",
      ui.text.InfinitySign()
    );

  end;
  
  private.ShowTimer();
  
end;
--------------------------------------------------------------------------------
function private.ShowTimer ()
  
  local multiset = {};
  local time_string  = "";
  local seconds = _f.EventTimeRemaining();
  local is_exist = _f.IsEventExist();
  
  if ( is_exist ) or ( seconds > 0 ) then
    common.Timer_Start(
      private.this.timer,
      1,
      private.ShowTimer,
      multiset
    );
    time_string = ui.text.FormatTime( seconds );
  else
    common.Timer_Stop( private.this.timer, multiset );
  end;

  ui.obj.Set( private.this.anm, { tt_timer = time_string }, multiset );

  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------