-- name=int_dialog_stage_fortunewheel
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  ShowVideo          = function () end,
  GetAttemptsVideo   = function () end,
  IsRecover          = function () end,
  IsVideoAvailable   = function () end,
  GetCooldownSeconds = function () end,
  VideoCanWatch      = function () end,
  VideoRefused       = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()
  
  private.app_w = ui.screen.Width();
  private.app_h = ui.screen.Height();
  private.app_o = ui.screen.Orientation();

  private.hint_time_popup = 3;
  private.wheel_process_anim = false;
  
  private.this = {
    dialog = "obj_int_dialog_stage_fortunewheel",
    anm = {
      wheel = "anm_int_dialog_stage_fortunewheel_window",
      body  = "anm_int_dialog_stage_fortunewheel_body",
      teeth = "anm_int_dialog_stage_fortunewheel_teeth",
      win   = "anm_int_dialog_stage_fortunewheel_win"
    },
    button = {
      wheel = {
        hub   = "anm_int_dialog_stage_fortunewheel_buttons",
        open  = "btn_int_dialog_stage_fortunewheel_open",
        close = "btn_int_dialog_stage_fortunewheel_close",
        run   = "btn_int_dialog_stage_fortunewheel_run",
        skip  = "btn_int_dialog_stage_fortunewheel_skip",
        tap   = "btn_int_dialog_stage_fortunewheel_tap",
        video = "btn_int_dialog_stage_fortunewheel_video"
      }
    },
    timer = {
      hint = "tmr_int_dialog_stage_fortunewheel_waiting",
      skip = "tmr_int_dialog_stage_fortunewheel_skip"
    }
  };

  private.Adapt();
  private.CreateButtons();
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Adapt ()

  ui.obj.ToCenter( private.this.dialog, 0, 0 );

end;
--------------------------------------------------------------------------------
function private.CreateButtons ()
  
  ui.control.EventsSet(
    private.this.button.wheel.open,
    {
      func_on_click = function ()

        private.Wheel_Open();

      end
    }
  );  
  
  ui.control.EventsSet(
    private.this.button.wheel.tap,
    {
      func_on_click = function ()

        if ( not private.wheel_process_anim ) then
          private.Wheel_Launch();
        end;

      end
    }
  );  
  
  ui.button.Create(
    private.this.button.wheel.run,
    {
      design = "medium",
      color  = "green",
      title  = "str_int_dialog_stage_fortunewheel_run",
    },
    function()
      
      if ( not private.wheel_process_anim ) then
        private.Wheel_Launch();
      end;
      
    end
  );   
  
  ui.button.Create(
    private.this.button.wheel.close,
    {
      design = "close"
    },
    function ( automatically )
       
      if ( not automatically ) then
        _f.VideoRefused();
      end;
      private.Wheel_Close();
      
    end
  );  
  
  ui.button.Create(
    private.this.button.wheel.skip,
    {
      design = "wheel",
      title  = "str_int_dialog_stage_fortunewheel_skip",
      timer  = "down"
    },
    function()
      
      _f.VideoRefused();
      private.Wheel_ButtonsOff();
      
      if ( not private.FuncResetCooldown() ) then

        private.Wheel_ButtonsWait();

      end;

      private.Wheel_Launch();
      
    end
  );  

  if ( IsMobile() ) then
    ui.button.Create(
      private.this.button.wheel.video,
      {
        title  = "str_int_dialog_stage_fortunewheel_video",
        design = "video"
      },
      function() _f.ShowVideo(); end
    );
  end;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Show ( 
    wheel_data,
    func_maximize,
    func_minimize,
    func_turn,
    func_reset_cooldown,
    func_get_reward,
    func_get_flight_target,
    popup_funcs,
    func_hero_fun,
    func_show_panel_stage_top,
    func_hide_panel_stage_top,
    func_attach_res_panel,
    func_detach_res_panel
  )
  
  if ( private.is_active ) then return; end;
  
  private.is_active  = true;
  private.wheel_show = false;
  
  private.Window_Maximize     = func_maximize;
  private.Window_Minimize     = func_minimize;
  
  private.FuncTurn            = func_turn;
  private.FuncResetCooldown   = func_reset_cooldown;
  private.FuncItemGet         = func_get_reward;
  
  private.FuncPopupShow       = popup_funcs.show;
  private.FuncPopupHide       = popup_funcs.hide;  
  private.FuncHeroFun         = func_hero_fun;   
  private.FuncGetFlightTarget = func_get_flight_target;
  
  private.Show_Stage_TopPanel = func_show_panel_stage_top;
  private.Hide_Stage_TopPanel = func_hide_panel_stage_top;   
  
  private.TopPanel_Show       = func_attach_res_panel;
  private.TopPanel_Hide       = func_detach_res_panel;

  private.FuncNotifyReady     = wheel_data.notify_ready;

  private.wheel_visual_state = "hidden";  
  private.wheel_logic_state  = wheel_data.state;
  private.prizes             = wheel_data.prizes;
  private.spin_cost          = wheel_data.spin_cost;
  private.continue_name      = wheel_data.continue_name;
  
  private.Wheel_Create( private.prizes );
  private.Wheel_ButtonsOff();  

  local multiset = {};

  ui.anim.Play( private.this.anm.wheel, "wheel_close_static_"..private.app_o );
  ui.anim.Play( private.this.anm.body,  "static_close" );

  private.Teeth_CloseInstantly( multiset );

  ui.input.On( private.this.button.wheel.open, multiset );
 
  ui.button.SetTextTag(
    private.this.button.wheel.skip,
    "price",
    wheel_data.spin_cost,
    multiset
  );

  ui.obj.Multiset( multiset );

  private.SetStateButtonVideoAttempts();
  
  if ( private.wheel_logic_state == "recover" ) then
   
    private.Timer_Create();
  
  end;

  ui.window.Open( "stage_fortunewheel", function ()
    wheel_data.notify_opened();
  end );

end;
--------------------------------------------------------------------------------
function public.Hide ( func_end )

  if ( not private.is_active ) then return; end;
  
  private.is_active = false;
  
  private.Timer_Delete();
  private.Hint_Hide();

  ui.window.Close( "stage_fortunewheel", func_end );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Update ( reason )
  
  if ( not private.is_active ) then return; end;
  
  if ( reason == "video_attempts" ) then
    private.SetStateButtonVideoAttempts();
  end;
  
end;
--------------------------------------------------------------------------------
function public.Cooldown_OnStarted ()

  private.wheel_logic_state = "recover";

  private.Timer_Create();

  private.SetStateButtonVideoAttempts();
  
end;
--------------------------------------------------------------------------------
function public.Cooldown_OnFinished ()

  private.wheel_logic_state = "ready";

  if ( private.wheel_visual_state == "inactive" ) then
    private.Wheel_Activate();
  end;
  
  private.SetStateButtonVideoAttempts();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Hint_Show ( time, func_exec )

  local func_exec = func_exec or function () 
    private.FuncPopupShow( "dragme", { 60, - 110 } ); 
  end;

  common.Timer_Start(
    private.this.timer.hint,
    time or private.hint_time_popup,
    function ()
      
      if ( private.popup_enabled ) then
  
        func_exec();
        
      end;

    end );

end;
--------------------------------------------------------------------------------
function private.Hint_Hide ()
  
  private.FuncPopupHide();
  common.Timer_Stop( private.this.timer.hint );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Wheel_Create ( content )

  private.current_pos = math.floor( #content / 2 );

  local multiset = {};
  
  for i = 1, 3 do

    local props = private.Teeth_IconProperties( 
      private.Teeth_RandomRewards( content )
    );

    ui.obj.Set(
      "spr_int_dialog_stage_fortunewheel_teeth_2_"..i,
      { 
        res = ui.obj.Res( "booster", private.continue_name );
      },
      multiset
    );    

    ui.obj.Set(
      "spr_int_dialog_stage_fortunewheel_teeth_main_"..i,
      { 
        res     = props.res,
        scale_x = props.scale,
        scale_y = props.scale
      },
      multiset
    );
        
  end;
  
  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.Wheel_Launch ()

  if ( private.wheel_logic_state == "ready" ) then

    private.Hint_Hide();
    private.Wheel_ButtonsOff();

    local finish_pos = private.FuncTurn();
    private.Wheel_Spin( finish_pos );

    common.sound.Play( "wheel_rotate" );

  end;

end;
--------------------------------------------------------------------------------
function private.Wheel_Spin ( finish_pos )
  
  private.popup_enabled = false;

  local multiset    = {};
  local content     = private.prizes;
  local reward_name = content[ finish_pos ].name;

  ui.input.Off( private.this.button.wheel.close );
  
  if ( reward_name == "empty" ) then

    local random = private.Teeth_RandomRewards( content );

    for i = 1, 3 do

      local props = private.Teeth_IconProperties( random[ i ] );

      ui.obj.Set(
        "spr_int_dialog_stage_fortunewheel_teeth_main_"..i,
        {
          res     = props.res,
          scale_x = props.scale,
          scale_y = props.scale
        },
        multiset
      );

      ui.obj.Set(
        "spr_int_dialog_stage_fortunewheel_win_"..i,
        {
          res     = props.res,
          scale_x = props.scale,
          scale_y = props.scale
        },
        multiset
      );
      
    end;

  else

    local props = private.Teeth_IconProperties( reward_name );

    for i = 1, 3 do

      ui.obj.Set(
        "spr_int_dialog_stage_fortunewheel_teeth_main_"..i,
        {
          res     = props.res,
          scale_x = props.scale,
          scale_y = props.scale
        },
        multiset
      );

      ui.obj.Set(
        "spr_int_dialog_stage_fortunewheel_win_"..i,
        {
          res     = props.res,
          scale_x = props.scale,
          scale_y = props.scale
        },
        multiset
      );

    end;

    ui.obj.Set(
      "spr_int_dialog_stage_fortunewheel_win_icon",
      {
        res = props.res
      },
      multiset
    );

  end;
  
  ui.obj.Multiset( multiset );
      
  common.sound.Play( "wheel_working" );

  ui.anim.Play( private.this.anm.body, "spin" );
  ui.anim.Play(
    private.this.anm.teeth,
    "tooth_spin",
    function () private.Wheel_OnSpinEnd( finish_pos ); end
  );
  
  private.Hint_Show(
    1,
    function ()
      common.sound.Play( "wheel_stop" );
      private.Hint_Hide();
    end
  );

end;
--------------------------------------------------------------------------------
function private.Wheel_OnSpinEnd ( finish_pos )

  private.current_pos = finish_pos;

  ui.anim.Play( private.this.anm.body, "static_open" );

  local reward_name  = private.prizes[ finish_pos ].name;
  local reward_count = private.prizes[ finish_pos ].count;

  ui.control.EventsClear( private.this.anm.win );

  if ( reward_name == "empty" ) then

    ui.input.On( private.this.button.wheel.close );
    private.Wheel_Deactivate();

  else

    -- Win!
    private.FuncHeroFun();

    local func_reward_get;
    local multiset = {};
    
    ui.obj.Disable(
      "txt_int_dialog_stage_fortunewheel_win_icon_count",
      true,
      multiset
    );
    
    ui.obj.Disable(
      "txt_int_dialog_stage_fortunewheel_win_count",
      true,
      multiset
    );

    if ( reward_name == "gold"  ) or
       ( reward_name == "heart" ) then
      
      local target_name = private.FuncGetFlightTarget( reward_name );

      local count_pos = {
        gold  = { 40, 45 },
        heart = { -5, 0 }
      };

      ui.obj.Set(
        "txt_int_dialog_stage_fortunewheel_win_count",
        { 
          active  = true, 
          visible = true, 
          text    = reward_count,
          pos_x   = count_pos[ reward_name ][1],
          pos_y   = count_pos[ reward_name ][2]
        },
        multiset
      );

      func_reward_get = function ()

        ui.screen.Lock( true );
        
        local target_coords = ui.obj.Position( target_name, private.this.anm.win );

        local func_anim_end = function ()

          ui.obj.Set(
            "spr_int_dialog_stage_fortunewheel_win_icon",
            {
              pos_x = 0,
              pos_y = 0,
              alp   = 1.0
            }
          );
          
          ObjStopAnimate( private.this.anm.win, "alp" );
          ObjStopAnimate( "spr_int_dialog_stage_fortunewheel_win_icon",  "alp" );
          ObjDetach( "spr_int_dialog_stage_fortunewheel_win_icon" );

          private.FuncItemGet(
            private.prizes[ finish_pos ],
            function ()
            
              ui.button.EventGet( private.this.button.wheel.close )( true );
              ui.screen.Lock( false );
            
            end
          );

        end;
        
        local multiset = {};

        ui.obj.Set(
          private.this.anm.win,
          { 
            visible = false, 
            input   = false, 
            endtrig = "", 
            playing = false 
          },
          multiset
        );

        ui.obj.Attach(
          "spr_int_dialog_stage_fortunewheel_win_icon",
          private.this.anm.wheel
        );

        ui.obj.Set(
          "txt_int_dialog_stage_fortunewheel_win_icon_count",
          {
            active  = true,
            visible = true,
            text    = reward_count
          },
          multiset
        );
        
        ui.obj.Multiset( multiset );

        ObjAnimate(
          "spr_int_dialog_stage_fortunewheel_win_icon", 
          "pos_xy", 0, 0,
          func_anim_end,
          { 
            0.0, 3, 0, 0, 
            0.6, 3, target_coords[1], target_coords[2]
          }
        );

        ObjAnimate(
          "spr_int_dialog_stage_fortunewheel_win_icon",
          "alp", 0, 0, "",
          {
            0.0, 0, 1.0,
            0.3, 0, 1.0, 
            0.6, 3, 0.0
          }
        );

         common.sound.Play( "rotate_win" );

      end;

    elseif ( reward_name == "extramoves" ) or 
           ( reward_name == "extralives" ) then
           
      if ( reward_name == "extramoves" ) then
        
        private.Show_Stage_TopPanel();
        
      end;

      func_reward_get = function ()

        local target_name   = private.FuncGetFlightTarget( reward_name );
        local target_coords = ui.obj.Position(
          target_name,
          private.this.anm.win
        );

        local func_anim_end = function ()

          ui.obj.Set(
            "spr_int_dialog_stage_fortunewheel_win_icon",
            {
              pos_x = 0,
              pos_y = 0,
              alp   = 1.0
            }
          );
          
          ObjStopAnimate( private.this.anm.win, "alp" );
          ObjStopAnimate( "spr_int_dialog_stage_fortunewheel_win_icon",  "alp" );
          ObjDetach( "spr_int_dialog_stage_fortunewheel_win_icon" );

          private.FuncItemGet( private.prizes[ finish_pos ] );
          --private.FuncReject( true );

        end;

        ui.obj.Set(
          private.this.anm.win,
          { 
            visible = false, 
            input   = false, 
            endtrig = "", 
            playing = false 
          }
        );

        ui.obj.Attach(
          "spr_int_dialog_stage_fortunewheel_win_icon",
          private.this.anm.wheel
        );

        ObjAnimate(
          "spr_int_dialog_stage_fortunewheel_win_icon",
          "pos_xy", 0, 0, func_anim_end,
          { 
            0.0, 3, 0, 0, 
            0.6, 3, target_coords[1], target_coords[2]
          }
        );

        ObjAnimate(
          "spr_int_dialog_stage_fortunewheel_win_icon",
          "alp", 0, 0, "",
          {
            0.0, 0, 1.0,
            0.3, 0, 1.0, 
            0.6, 3, 0.0
          }
        );

         common.sound.Play( "rotate_win" );

      end;

    else

      func_reward_get = function ()

        ObjStopAnimate( private.this.anm.win, "alp" );
        ObjStopAnimate( "spr_int_dialog_stage_fortunewheel_win_icon",  "alp" );

        ui.obj.Set(
          private.this.anm.win,
          {
            input   = false,
            endtrig = ""
          }
        );
        
        ObjAnimate(
          private.this.anm.win,
          "alp", 0, 0,
          function ()
            
            ui.obj.Hide( private.this.anm.win );
            
          end,
          {
            0.0, 3, 1.0, 
            0.3, 3, 0.0
          }
        );

        if ( reward_name ~= "empty" ) then

          private.FuncItemGet( private.prizes[ finish_pos ] );

        end;
        
        ui.button.EventGet( private.this.button.wheel.close )( true );

         common.sound.Play( "rotate_win" );

      end;

    end;

    private.Teeth_CloseInstantly( multiset );

    ui.obj.Set(
      private.this.anm.win,
      { 
        alp      = 1.0, 
        input    = true, 
        visible  = true,
        active   = true,
        animfunc = "win",
        playing  = true,
        endtrig  = func_reward_get,
        tt_award = "str_int_dialog_stage_fortunewheel_reward"
      },
      multiset
    );

    ui.obj.Multiset( multiset );
    
    ui.control.EventsSet(
      private.this.anm.win,
      {
        func_on_click = func_reward_get
      }
    );

  end;

end;
--------------------------------------------------------------------------------
function private.Wheel_Open ()
  
  private.Window_Maximize();
    
  private.popup_enabled      = true;
  private.wheel_process_anim = true;
  
  ui.input.Off({
    private.this.button.wheel.open,
    private.this.button.wheel.close
  });

  local function func_on_open ()

    private.wheel_process_anim = false;
    private.wheel_show = true;
    
    if ( private.wheel_visual_state == "ready" ) then
    
      private.Hint_Show();

    end;

    private.FuncNotifyReady();
    
    private.SetStateButtonVideoAttempts();
    
    ui.input.On( private.this.button.wheel.close );
    
  end;

  if ( private.wheel_logic_state == "ready" ) then

    private.wheel_visual_state = "ready";

     common.sound.Play( "wheel_open_mouth" );

    private.TopPanel_Hide();

    private.Wheel_ButtonsReady();

    ui.anim.Play( private.this.anm.body,  "open", nil, "static_open" );
    ui.anim.Play( private.this.anm.teeth, "tooth_open", nil, "tooth_static_open" );

  else

    private.wheel_visual_state = "inactive";

    private.TopPanel_Show();

    private.Wheel_ButtonsWait();

    ui.anim.Play( private.this.anm.body, "static_close" );

  end;
  
  ui.anim.Play( private.this.anm.wheel,
    "wheel_appear_"..private.app_o,
    func_on_open,
    "wheel_open_static_"..private.app_o
  );

end;
--------------------------------------------------------------------------------
function private.Wheel_Close ()

  ui.input.Off( private.this.button.wheel.close );
  
  private.popup_enabled = false;
  private.wheel_process_anim = true;
  private.wheel_show = false;
  
  private.wheel_visual_state = "hidden";

  ui.input.Off( private.this.button.wheel.run );
  ui.input.Off( private.this.button.wheel.tap );

  ui.obj.Disable( private.this.anm.win );

  if ( ui.anim.Func( private.this.anm.teeth ) ~= "tooth_close" ) then

    ui.anim.Play( private.this.anm.teeth, "tooth_close" );

  end;

  if ( ui.anim.Func( private.this.anm.body ) ~= "static_close" ) then

    ui.anim.Play( private.this.anm.body, "close", nil, "static_close" );

  end;
  
  local function func_on_close ()
    
    private.wheel_process_anim = false;
    
    private.Window_Minimize();
    
    ui.input.On({
      private.this.button.wheel.close,
      private.this.button.wheel.open
    });

    private.Teeth_CloseInstantly();
    
    -- Simulate window closed.
    mediator.Notify( "Interface_Window_WindowClosedEnd", "int_dialog_stage_fortunewheel" );
    
  end;  

  ui.anim.Play(
    private.this.anm.wheel,
    "wheel_disappear_"..private.app_o,
    func_on_close,
    "wheel_close_static_"..private.app_o
  );
  
  private.SetStateButtonVideoAttempts();
  private.Hint_Hide();

end;
--------------------------------------------------------------------------------
function private.Wheel_Activate ()

  private.wheel_visual_state = "ready";

  common.sound.Play( "wheel_open_mouth" );

  ui.anim.Play( private.this.anm.body, "open", nil, "static_open" );
  ui.anim.Play( private.this.anm.teeth, "tooth_open" );

  private.Wheel_ButtonsReady();

  private.Hint_Show();

  private.TopPanel_Hide();

end;
--------------------------------------------------------------------------------
function private.Wheel_Deactivate ()

  private.wheel_visual_state = "inactive";

  ui.anim.Play( private.this.anm.body, "close", nil, "static_close" );
  ui.anim.Play( private.this.anm.teeth, "tooth_close" );
    
  private.Wheel_ButtonsWait();

  private.FuncPopupShow( "tryagain", { 50, - 110 } );

  private.TopPanel_Show();

end;
--------------------------------------------------------------------------------
function private.Wheel_ButtonsReady ()

  ui.input.On( { 
    private.this.button.wheel.run,
    private.this.button.wheel.close,
    private.this.button.wheel.tap
  } );
  ui.input.Off( private.this.button.wheel.skip );

  ui.anim.Play( private.this.button.wheel.hub, "buttons_ready" );

end;
--------------------------------------------------------------------------------
function private.Wheel_ButtonsWait ()

  ui.input.On( private.this.button.wheel.skip );
  ui.input.Off( private.this.button.wheel.run );
  ui.input.Off( private.this.button.wheel.tap );

  ui.anim.Play( private.this.button.wheel.hub, "buttons_wait" );

end;
--------------------------------------------------------------------------------
function private.Wheel_ButtonsOff ()

  ui.input.Off( {
    private.this.button.wheel.skip,
    private.this.button.wheel.run,
    private.this.button.wheel.tap
  } );

  ui.anim.Play( private.this.button.wheel.hub, "buttons_off" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Timer_Create ()

  if ( not ui.obj.Exists( private.this.timer.skip ) ) then

    ui.obj.Create( private.this.timer.skip, "timer" );

  end;

  ui.obj.Attach(
    private.this.timer.skip,
    private.this.dialog
  );

  private.Timer_Update();

end;
--------------------------------------------------------------------------------
function private.Timer_Update ()

  local seconds = _f.GetCooldownSeconds();

  if ( seconds > 0 ) then

    ui.button.SetTime(
      private.this.button.wheel.skip,
      ui.text.FormatTime( seconds )
    );

    common.Timer_Start(
      private.this.timer.skip,
      1,
      private.Timer_Update
    );

  else

    private.Timer_Delete();

    if ( private.wheel_visual_state == "inactive" ) then

      private.Wheel_Activate();

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Timer_Delete ()

  common.Timer_Delete( private.this.timer.skip );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Teeth_IconProperties ( item_name )

  local properties = {};

  if ( item_name == "extramoves" ) then

    if ( private.continue_name == "extramoves" ) then

      properties.res = ui.obj.Res( "booster", "extramoves" );
      
    elseif ( private.continue_name == "extralives" ) then

      properties.res = ui.obj.Res( "booster", "extralives" );

    end;

    properties.scale = 0.6;

  elseif ( item_name == "heart" ) then 
    
    properties.res   = ui.obj.Res( "item", "heart" );
    properties.scale = 1.0;

  elseif ( item_name == "gold" ) then 

    properties.res   = ui.obj.Res( "item", "gold_1" );
    properties.scale = 1.0;

  end;

  return properties;

end;
--------------------------------------------------------------------------------
function private.Teeth_RandomRewards ( content )
  
  local reward_names = {};
  reward_names[ 1 ] = content[ math.random( 1, #content - 1 ) ].name;
  reward_names[ 2 ] = content[ math.random( 1, #content - 1 ) ].name;
  reward_names[ 3 ] = content[ math.random( 1, #content - 1 ) ].name;

  while 
  ( reward_names[ 3 ] == reward_names[ 2 ] ) or 
  ( reward_names[ 3 ] == reward_names[ 1 ] ) do

    reward_names[ 3 ] = content[ math.random( 1, #content - 1 ) ].name;

  end;
  
  return reward_names;
    
end;
--------------------------------------------------------------------------------
function private.Teeth_CloseInstantly ( multiset )

  ui.obj.Set(
    private.this.anm.teeth,
    {
      animfunc = "tooth_close",
      frame    = 60,
      playing  = false,
      endtrig  = function () end
    },
    multiset
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.SetStateButtonVideoAttempts ()
  
  if ( not private.is_active ) then return; end;
  
  local attempts, max_attempts = _f.GetAttemptsVideo();
  local is_recover = _f.IsRecover();
  local is_video_available = _f.IsVideoAvailable();
  local obj_data = { visible = false, input = false };
  
  if ( is_recover ) and ( attempts < max_attempts ) and 
    ( private.wheel_show ) and ( is_video_available ) then
    obj_data = { visible = true, input = true };
  end;

  _f.VideoCanWatch( obj_data.visible );
  ui.obj.Set( private.this.button.wheel.video, obj_data );
  
end;
--------------------------------------------------------------------------------