-- name=int_dialog_stage_friends
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Close = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  private.app_w = ui.screen.Width();
  private.app_h = ui.screen.Height();
  
  ui.hub.Create( "int_dialog_stage_friends" );

  private.Adapt();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Prepare ( friends_data )

  private.friends = friends_data;

  -- friends
  private.candidates       = {};
  private.multiset_free    = {};
  private.interim_index    = 0;
  private.object_index     = 0;
  private.object_top_index = 0;

  private.Friends_Sort();

end;
--------------------------------------------------------------------------------
function public.Hide ()

  private.friends = {};

  if not ( private.active ) then

    return;

  end;

  private.active  = false;

  local func_end = function ()

    if ( private.object_index ) then

      for i = 1, private.object_index do

        ObjDelete( "obj_int_dialog_stage_friends_friendmarker_"..i );

      end;

    end;

    if ( private.object_top_index ) then

      for i = 1, private.object_top_index do

        ObjDelete( "obj_int_dialog_stage_friends_friend_"..i );

      end;

    end;
    
    ui.obj.Disable( "obj_int_dialog_stage_friends" );

    ui.hub.DetachContent( "int_dialog_stage_friends" );

  end;

  if ( private.animated ) then

    private.Anim_Hide( func_end );

  else

    func_end();
  
  end;

end;
--------------------------------------------------------------------------------
function public.Update ( score, is_win )

  if ( score == 0           ) or
     ( #private.friends < 1 ) then

    return;

  end;

  -- JK: Disable friends big panel in center of match3 field
  is_win = false;

  private.Friends_Update( score, is_win );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Show ()

  if ( private.active ) then

    return;

  end;

  private.active = true; 

  ui.hub.AttachContent( "int_dialog_stage_friends" );

  ui.obj.Set(
    "obj_int_dialog_stage_friends",
    {
      active  = true,
      visible = true,
      input   = false
    }
  );
  
end;
--------------------------------------------------------------------------------
function private.Anim_Show ()

  if ( private.animated ) then return; end;

  ui.anim.Play(
    "anm_int_dialog_stage_friends",
    "open",
    nil,
    "static"
  );

  common.sound.Play( "app_info" );

  private.animated = true;

end;
--------------------------------------------------------------------------------
function private.Anim_Hide ( func_end )

  if not ( private.animated ) then return; end;

  ui.anim.Play( "anm_int_dialog_stage_friends", "close", func_end );

  private.animated = false;

end;
--------------------------------------------------------------------------------
function private.Friends_Sort ()

  if ( #private.friends == 0 ) then

    return;

  end;

  for i = ( #private.friends - 1 ), 1, - 1 do 

    for j = 1, i do

      if ( private.friends[ j ].score > private.friends[ j + 1 ].score ) then

        private.friends[ j ], private.friends[ j + 1 ] = private.friends[ j + 1 ], private.friends[ j ];

      end;

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Friends_Update ( score, is_win )

  local index = 0;

  for i = 1, #private.friends do

    if ( score >= private.friends[ i ].score ) then

      table.insert( private.candidates, private.friends[ i ] );

      index = i;

    end;

  end;

  if ( index == 0 ) then

    return;

  end;

  for i = 1, index do

    table.remove( private.friends, 1 );

  end;

  private.Show();

  if ( is_win ) then

    private.Anim_Show();

  end;

  private.Friends_ParamsSet();
  private.Friends_Animate( is_win );

end;
--------------------------------------------------------------------------------
function private.Friends_ParamsSet ()

  local multiset = {};

  for i = 1, #private.candidates do

    local distance = ( private.candidates[ i ].score - private.candidates[ 1 ].score ) / 5;

    local object_index = 0;

    if ( #private.multiset_free > 0 ) then

      object_index = table.remove( private.multiset_free, 1 );

    else

      object_index = private.Friends_ObjectIndex();

      ui.obj.Clone(
        "obj_int_dialog_stage_friends_friendmarker",
        "obj_int_dialog_stage_friends",
        "_"..object_index,
        1
      );

    end;
    
    local friend_id = private.candidates[ i ].userId;
    local friend_url = private.candidates[ i ].url;
    local friend_score = private.candidates[ i ].score;
    
    local res = private.candidates[ i ].res or
                "assets/content/interface/int_dialog_stage_friends/default_0";

    local start_y = -0.5 * private.app_h + ( ui.screen.HeightPercent() * 30 );

    ui.obj.Clone(
      "spr_int_dialog_stage_friends_friend",
      "anm_int_dialog_stage_friends_friendmarker_"..object_index,
      "_"..friend_id,
      1
    );

    table.insert( multiset, { "obj_int_dialog_stage_friends_friendmarker_"..object_index,
    {    
      pos_x   = 50 + 0.5 * private.app_w + distance, 
      pos_y   = start_y,
      active  = true,
      visible = true
    } } );

    table.insert( multiset, { "anm_int_dialog_stage_friends_friendmarker_"..object_index,
    {    
      animfunc = "static",
      playing = true
    } } );

    table.insert( multiset, { "spr_int_dialog_stage_friends_friend_"..friend_id,
    {    
      url = friend_url,
      trg_resloaded = function ()
        ui.obj.SpriteAdjust( "spr_int_dialog_stage_friends_friend_"..friend_id, 
                         64 ); 
      end,
      res         = res,
      default_res = res,
      anim_tag = "icon"
    } } );

    table.insert( multiset, { "txt_int_dialog_stage_friends_friendmarker_score_"..object_index,
    {    
      text = ui.text.FormatNumber( friend_score, true );
    } } );

    ui.obj.SpriteAdjust(
      "spr_int_dialog_stage_friends_friend_"..friend_id,
      64
    );

    private.interim_index = private.interim_index + 1;

    private.candidates[ i ].object_index = object_index;
    private.candidates[ i ].interim_index = private.interim_index;
    private.candidates[ i ].distance = distance;

  end;

  ui.obj.Multiset( multiset );

  for i = 1, #private.candidates do

    ui.obj.SpriteAdjust(
      "spr_int_dialog_stage_friends_friend_"..private.candidates[ i ].userId,
      64
    );

  end;

end;
--------------------------------------------------------------------------------
function private.Friends_Animate ( is_win )

  local func_Anim = function ( candidate )

    local object_index  = candidate.object_index;
    local interim_index = candidate.interim_index;
    local distance      = candidate.distance;

    local object_name = "obj_int_dialog_stage_friends_friendmarker_"..object_index;

    local obj_params = ObjGet( object_name );

    ObjAnimate( object_name, "pos_xy", 0, 0, function ( params )

       common.sound.Play( "friend_drop" );

      if ( is_win ) then

        private.Friends_Fly( object_index, candidate.userId );

      else

        ui.anim.Play( "anm_int_dialog_stage_friends_friendmarker_"..object_index, "action" );

        ObjAnimate( params.sender, "pos_xy", 0, 0, function ()

          table.insert( private.multiset_free, object_index );

          ObjDelete( "spr_int_dialog_stage_friends_friend_"..candidate.userId );

        end, 
        { 
          0.0, 0, 0, obj_params.pos_y - 25,
          0.6 + 0.2, 0, - 0.5 * private.app_w - 50, obj_params.pos_y 
        } );

      end;

    end, 
    { 
      0.0, 0, 50 + 0.5 * private.app_w + distance, obj_params.pos_y,
      0.0 + 0.2 * candidate.interim_index, 0, 50 + 0.5 * private.app_w + distance, obj_params.pos_y,
      0.6 + 0.2 * candidate.interim_index, 0, 0, obj_params.pos_y - 25
    } );

  end;

  while ( #private.candidates > 0 ) and
        ( private.active          ) do

    local candidate = table.remove( private.candidates, 1 );

    func_Anim( candidate );

  end;

end;
--------------------------------------------------------------------------------
function private.Friends_Fly ( object_index, friend_id )

  local object_name = "obj_int_dialog_stage_friends_friendmarker_"..object_index;
  local obj_params  = ui.obj.Get( object_name );
  
  local func_end = function ( index )

    private.Friends_Show( index, friend_id );

    ui.obj.Disable( object_name, true );

    table.insert( private.multiset_free, object_index );

  end;

  local object_top_index = private.Friends_ObjectIndexTop();

  ui.obj.Clone(
    "obj_int_dialog_stage_friends_friend",
    "obj_int_dialog_stage_friends_friends",
    "_"..object_top_index,
    1
  );

  ui.anim.Play(
    "anm_int_dialog_stage_friends_friendmarker_"..object_index,
    "action"
  );
  
  ObjAnimate(
    object_name,
    "pos_x", 0, 0,
    function () func_end( object_top_index ) end, 
    { 
      0.0, 0, obj_params.pos_x,
      1.5, 0, private.Friends_Positions( object_top_index )
    }
  );

end;
--------------------------------------------------------------------------------
function private.Friends_Show ( index, friend_id )
  
  local obj_name = "obj_int_dialog_stage_friends_friend_hub_"..index;

  ObjAttach( "spr_int_dialog_stage_friends_friend_"..friend_id, obj_name );

  ui.obj.SpriteAdjust( "spr_int_dialog_stage_friends_friend_"..friend_id, 160 );

  local multiset = {};

  ui.obj.Set(
    "obj_int_dialog_stage_friends_friend_"..index, 
    { 
      active  = true, 
      visible = true, 
      pos_x   = private.Friends_Positions( index )
    },
    multiset
  );

  ui.obj.Set(
    "spr_int_dialog_stage_friends_friend_"..friend_id, 
    { 
      anim_tag = ""
    },
    multiset
  );

  ui.obj.Multiset( multiset );

  local anim_name = "anm_int_dialog_stage_friends_friend_"..index;

  ui.anim.Play( anim_name, "appear_in_panel" );

   common.sound.Play( "friend_appear" );

  if ( index > 1 ) then

    private.Friends_MoveRight( index );

  end;

end;
--------------------------------------------------------------------------------
function private.Friends_MoveRight ( index )

  local positions = 
  {
    { 0 },
    { - ( 0.125 * private.app_w + 30 ), 0.125 * private.app_w + 30 },
    { - ( 0.25 * private.app_w + 30 ), 0, 0.25 * private.app_w + 30 }
  }; 
  
  local count = 1;

  for i = index - 1, ( index - math.min( index, 3 ) ) + 1, - 1  do

    count = count + 1;

    local new_pos = positions[ math.min( index, 3 ) ][ count ];

    if ( new_pos ) then

      local obj_params = ObjGet( "obj_int_dialog_stage_friends_friend_"..i );

      ObjAnimate( "obj_int_dialog_stage_friends_friend_"..i, "pos_x", 0, 0, "", 
      { 
        0.0, 0, obj_params.pos_x, 
        0.3, 0, new_pos 
      } );

    end;

  end;

  local obj_params = ObjGet( "obj_int_dialog_stage_friends_friend_"..( index - math.min( index, 3 ) ) );

  if ( obj_params ) then

    ObjAnimate( "obj_int_dialog_stage_friends_friend_"..( index - math.min( index, 3 ) ), "pos_x", 0, 0, function ( params)

      ObjDelete( params.sender );

    end, 
    { 
      0.0, 0, obj_params.pos_x, 
      0.3, 0, private.app_w 
    } );

  end;

end;
--------------------------------------------------------------------------------
function private.Friends_ObjectIndex ()
  
  private.object_index = private.object_index + 1;

  return private.object_index;

end;
--------------------------------------------------------------------------------
function private.Friends_ObjectIndexTop ()

  private.object_top_index = private.object_top_index + 1;

  return private.object_top_index;

end;
--------------------------------------------------------------------------------
function private.Friends_Positions ( index )

  local positions = 
  {
    { 0 },
    { - ( 0.125 * private.app_w + 30 ), 0.125 * private.app_w + 30 },
    { - ( 0.25 * private.app_w + 30 ), 0, 0.25 * private.app_w + 30 }
  };

  return positions[ math.min( index, 3 ) ][ 1 ];

end;
--------------------------------------------------------------------------------
function private.Adapt ()

  ui.obj.ToCenter( "obj_int_dialog_stage_friends" )

end;
--------------------------------------------------------------------------------