-- name=int_dialog_stage_start
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Accept      = function () end,
  Reject      = function () end,
  BuyBooster  = function () end,
  OnShow      = function () end,
  DataBooster = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  private.app_w = ui.screen.Width();
  private.app_h = ui.screen.Height();
  
  private.this = {
    dialog = "obj_int_dialog_stage_start",
    window = "anm_int_dialog_stage_start_window",
    button = {
      close = "btn_int_dialog_stage_start_close",
      play  = "btn_int_dialog_stage_start_play"
    }
  };
  
  private.show_event_marker = false;
  private.event_marker      = "anm_int_dialog_stage_start_event_marker";

  private.anim_time = 0.3;

  private.strings = {
    moves   = ne.strings.Get( "str_dialog_stage_start_moves" ),
    targets = ne.strings.Get( "str_dialog_stage_start_targets" ),
    reward  = ne.strings.Get( "str_dialog_stage_start_rewards" )
  };

  ui.button.Create(
    private.this.button.close,
    {
      design = "close"
    },
    function() _f.Reject(); end
  );
  
  ui.button.Create(
    private.this.button.play,
    {
      title  = "str_dialog_stage_start_play",
      color  = "green",
      design = "medium",
      sound  = "button_click"
    },
    function()
      
      _f.Accept( private.boosters_selected );
      
    end
  );
  
  ui.obj.Set(
    private.event_marker,
    {
      pos_y    = - private.app_h / 2,
      tt_event = ne.strings.Get( "str_dialog_stage_start_event_marker" )
    }
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show (
    stage_name,
    stage_data,
    show_event_marker
  )

  private.boosters          = {};
  private.boosters_selected = {};
  private.stage_name        = stage_name;
  private.show_event_marker = show_event_marker;

  private.Default_Set( stage_name, stage_data );
  
  private.active = true;
  
  ui.obj.Set(
    private.event_marker,
    {
      visible = private.show_event_marker
    }
  );

  if ( private.show_event_marker ) then
  
    ui.anim.Play(
      private.event_marker,
      "event_marker_appear",
      nil,
      "event_marker_static"
    );
  
  end;

  ui.window.Open( "stage_start", function()
 
    ui.window.SetAreaByButtons(
      private.this.dialog,
      private.this.button.play
    );

    _f.OnShow();
  
  end );
    
end;
--------------------------------------------------------------------------------
function public.Hide ( func_execute, skip_map_enable, is_instantly )

  if ( not private.active ) then return; end;

  private.active = false;

  local function func_HideWindow ()
    ui.window.Hide( "int_dialog_stage_start", func_execute, skip_map_enable );
  end;

  if ( is_instantly ) then
    func_HideWindow();
    return;
  end;
  
  if ( private.show_event_marker ) then
    ui.anim.Play( private.event_marker, "event_marker_disappear" );
  end;
  
  ui.window.Close( "stage_start", func_HideWindow );

end;
--------------------------------------------------------------------------------
function public.Update ( booster_data )

  if ( not private.active ) then

    return;

  end;

  local booster_name = booster_data.name;
  local booster_count = booster_data.count;
  local index = private.boosters[ booster_name ];

  if ( index ) then

    private.Booster_TapSet( booster_name, index, booster_count );

    local obj = "anm_int_dialog_stage_start_boosters_"..index.."_icon";

    local params = { count = ui.text.FormatNumber( booster_count ) };

    if ( booster_count > 0 ) then

      params.animfunc = "booster";
      ObjSet( obj, params );

    else

      params.animfunc = "booster_add";
      ObjSet( obj, params );

    end;
    
    if ( ( booster_data.source or "" ) ~= "tutorial" ) and ( booster_data.delta > 0 ) then
    
      ( ui.control.TapGet( "btn_int_dialog_stage_start_boosters_"..index ).func_true or function () end )();

    end;

  end;

end;
--------------------------------------------------------------------------------
function public.OnUpdateBoosters ()
  
  if ( not private.active ) then

    return;

  end;
  
  for key, value in pairs( private.boosters ) do
    
    local booster_data = _f.DataBooster( key );
    
    ObjSet( "anm_int_dialog_stage_start_boosters_"..value.."_icon", {
      rt_sale = tern( booster_data.sale )( "assets/content/items/icon_sale","" ),
      tt_sale = tern( booster_data.sale )( "str_int_sale_label","" )
    } );
    
  end;
  
end;
--------------------------------------------------------------------------------
function public.GetStageName ()

  return private.stage_name;

end;
--------------------------------------------------------------------------------
function public.GetObjectPlayName ()
  
  return private.this.button.play;
  
end;
--------------------------------------------------------------------------------
function public.IsActive ()
  
  return private.active;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Default_Set ( stage_name, stage_data )

  private.section = stage_data.section;
  
  local multiset_table = {};
  
  -- Detect stage type.
  local stage_type = stage_data.type;
  
  table.insert( multiset_table, 
  { 
    "anm_int_dialog_stage_start_window", 
    {
      title = common.string.FormatNameStage( stage_name )
    }
  } );
    
  -- Moves.
  local moves_data = {};
  moves_data.moves = stage_data.moves;
  private.Moves_Set( multiset_table, moves_data );  
  
  -- Targets.
  local targets_data = {};
  targets_data.targets = stage_data.targets;
  targets_data.stage_type = stage_type;
  if ( stage_type == "fight" ) then
    
    targets_data.enemy_name = stage_data.enemy.name;  
    
  end;
  private.Targets_Set( multiset_table, targets_data );
  
  -- Rewards.
  local rewards_data = {};
  rewards_data.reward = stage_data.reward;
  rewards_data.stage_state = stage_data.state;
  rewards_data.noreward = stage_data.noreward;
  local quest_reward = private.Rewards_Set( multiset_table, rewards_data, stage_data.reward_chest );
  
  -- Boosters.
  private.Boosters_Set( multiset_table, stage_data.boosters );
  
  -- Resources.
  local resources_data = {};
  resources_data.resources = stage_data.resources;
  private.Resources_Set( multiset_table, resources_data );

  -- Enemy.
  local enemy_data = {};
  enemy_data = stage_data.enemy or enemy_data;
  enemy_data.stage_type = stage_type;
  private.Enemy_Set( enemy_data, multiset_table );
 
  -- Anim.
  local anim_data = stage_data;
  anim_data.stage_type = stage_type;
  anim_data.stage_state = stage_data.state;
  private.Animation_Set( multiset_table, anim_data );
  
  private.Window_Set( multiset_table );
  
  ui.obj.Multiset( multiset_table );
 
  local get_animfunc = function ()
  
    local animfunc;

    if ( stage_type == "fight" ) then
      
      animfunc = stage_type.."_static";
      
    else

      if ( stage_data.stateless ) then
        
        if ( not stage_data.boosters ) or ( #stage_data.boosters == 0 ) then
          
          animfunc = stage_type.."_stateless_noboosters_"..stage_data.state.."_static";
        
        else
          
          animfunc = stage_type.."_stateless_"..stage_data.state.."_static";

        end;
        
      else
        
        animfunc = stage_type.."_"..stage_data.state.."_static";
        
      end;
      
      if ( stage_data.noreward ) then
        
        if ( stage_data.stateless ) then
          
          if ( not stage_data.boosters ) or ( #stage_data.boosters == 0 ) then
            
            animfunc = stage_type.."_stateless_noboosters_2_static";
            
          else
            
            animfunc = stage_type.."_stateless_2_static";            
            
          end;

        else
      
          animfunc = stage_type.."_4_static";
          
          if ( quest_reward ) then
            
            animfunc = stage_type.."_"..stage_data.state.."_static";
            
          end;

        end;
        
      end;
      
    end;
    
    if ( stage_data.reward_chest and ( stage_data.state == 1 ) ) then
      
      if ( stage_data.stateless ) then
      
        animfunc = stage_type.."_stateless_5_static";
        
      else
        
        animfunc = stage_type.."_5_static";
        
      end;
      
    end;

    return animfunc;
  
  end;

  ui.anim.Play( "anm_int_dialog_stage_start_window", get_animfunc() );

end;
--------------------------------------------------------------------------------
function private.Moves_Set ( multiset_table, moves_data )

  if ( moves_data.moves ) then
    
    table.insert( multiset_table, 
    { 
      "txt_int_dialog_stage_start_moves_count", { text = ui.text.FormatNumber( moves_data.moves ) }
    } );
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.Targets_Set ( multiset_table, targets_data )
  
  local targets_pos_y = { { 0 }, { - 25, 25 }, { - 50, 0, 50 } };
  
  for i = 1, 3 do

    table.insert( multiset_table, 
    { 
      "obj_int_dialog_stage_start_targets_target_"..i, { active = false, visible = false }
    } );    

  end;
  
  if ( targets_data.targets and #targets_data.targets > 0 ) then
    
    local resource = {
      default     = "items/icon_%s",
      DropItem    = private.section.."/icon_%s_%s",
      DestroyGem  = "items/icon_%s",
      DestroyCell = "items/icon_%s"
    };
    
    for i = 1, #targets_data.targets do

      table.insert( multiset_table, 
      { 
        "obj_int_dialog_stage_start_targets_target_"..i,
        {
          active  = true,
          visible = true,
          pos_y   = targets_pos_y[ #targets_data.targets ][ i ]
        }
      } );

      if not ( targets_data.targets[ i ].type == "GainScore" ) then
        
        local res_part = string.format(
          resource[ targets_data.targets[ i ].type ] or resource.default,
          string.lower( targets_data.targets[ i ].name ),
          private.section:match( "%d+" )
        );

        ui.obj.Set(
          "spr_int_dialog_stage_start_targets_target_"..i.."_icon",
          {
            active  = true,
            visible = true,
            res     = "assets/content/"..res_part
          },
          multiset_table
        );

      end;

      table.insert( multiset_table, 
      { 
        "txt_int_dialog_stage_start_targets_target_"..i,
        {
          text = ui.text.FormatNumber( targets_data.targets[ i ].count, true )
        } 
      } );

    end;
  
  else

    if ( targets_data.stage_type == "fight" ) then

      table.insert( multiset_table, 
      { 
        "obj_int_dialog_stage_start_targets_target_1", { active = true, visible = true, pos_y = - 20 }
      } );

      table.insert( multiset_table, 
      { 
        "txt_int_dialog_stage_start_targets_target_1", { 
          text = "str_dialog_stage_start_killenemy", 
          param0 = StringGet( "str_enemy_name_gender_"..targets_data.enemy_name )
        }
      } );

    end;

  end;  
  
end;
--------------------------------------------------------------------------------
function private.Rewards_Set ( multiset_table, rewards_data, reward_chest )
  
  local quest_reward = nil;
  
  for i = rewards_data.stage_state, #rewards_data.reward do
    
    for j = 1, #rewards_data.reward[ i ] do
      
      if ( rewards_data.reward[ i ][ j ].type == "quest" ) then
        
        quest_reward = { i, j };
        
      end;
      
    end;
    
  end;
  
  for i = 1, 4 do
    
    ui.obj.Set(
      "spr_int_dialog_stage_start_reward_"..i,
      {
        active  = false,
        visible = false
      },
      multiset_table
    );
    
  end;  
  
  if ( rewards_data.noreward ) then

    if ( quest_reward ) then

      local reward_item = rewards_data.reward[ quest_reward[ 1 ] ][ quest_reward[ 2 ] ];

      table.insert( multiset_table, 
      { 
        "spr_int_dialog_stage_start_reward_3",
        {
          active  = true,
          visible = true,
          res     = ui.obj.Res( reward_item.type, reward_item.name )
        }
      } );

    end;

  else

    if ( rewards_data.reward and #rewards_data.reward > 0 ) then

      for i = 1, #rewards_data.reward do
     
        for j = 1, #rewards_data.reward[ i ] do

          if ( reward_chest and ( i == 1 ) ) then

            ui.obj.Set(
              "spr_int_dialog_stage_start_reward_4",
              {
                active  = true,
                visible = true,
              },
              multiset_table
            );

          end;
          
          local reward = rewards_data.reward[ i ][ j ];

          if ( reward.name == "gold" ) then

            ui.obj.Set(
              "spr_int_dialog_stage_start_reward_"..i,
              {
                active  = true,
                visible = true,
                res     = ui.obj.Res( "gold", "gold_"..i )
              },
              multiset_table
            );

          else

            table.insert( multiset_table, 
            { 
              "spr_int_dialog_stage_start_reward_"..i,
              {
                active  = true,
                visible = true,
                res     = ui.obj.Res( reward.type, reward.name )
              }
            } );
            
            break;

          end;
          
        end;
      
      end;  

    end;

  end;
  
  return quest_reward;
  
end;
--------------------------------------------------------------------------------
function private.Boosters_Set ( multiset_table, boosters )
  
  local boosters_tags = { { 3 }, { 2, 4 }, { 1, 3, 5 } };

  if ( #boosters > 0 ) then

    table.insert( multiset_table, 
    { 
      "anm_int_dialog_stage_start_window", 
      { booster_title = StringGet( "str_dialog_stage_start_boosters_title" ) }
    } );

  else

    table.insert( multiset_table, 
    { 
      "anm_int_dialog_stage_start_window", { booster_title = "" }
    } );

  end;
  
  for i = 1, 3 do

    table.insert( multiset_table, 
    { 
      "btn_int_dialog_stage_start_boosters_"..i, 
      {
        anim_tag = "",
        active   = false,
        visible  = false,
        input    = false 
      }
    } );
    
    if ( boosters[ i ] ) then

      local booster = boosters[ i ];

      private.boosters[ booster.name ] = i;

      table.insert( multiset_table, 
      { 
        "btn_int_dialog_stage_start_boosters_"..i,
        { 
          anim_tag = "booster_"..boosters_tags[ #boosters ][ i ],
          active   = true,
          visible  = true,
          input    = true 
        }
      } );

      local params =
      {
        booster_icon = ui.obj.Res( "booster", booster.name ),
        count        = ui.text.FormatNumber( booster.count ),
        playing      = true,
        rt_sale      = tern( booster.sale )( "assets/content/items/icon_sale","" ),
        tt_sale      = tern( booster.sale )( "str_int_sale_label","" )
      };

      if ( booster.count > 0 ) then

        params.animfunc = "booster";

      else

        params.animfunc = "booster_add";

      end;

      table.insert( multiset_table, 
      { "anm_int_dialog_stage_start_boosters_"..i.."_icon", params } );

      private.Booster_TapSet( booster.name, i, booster.count );

    end;

  end;
  
end;
--------------------------------------------------------------------------------
function private.Resources_Set ( multiset_table, resources_data )
 
  table.insert( multiset_table, 
  { 
    "obj_int_dialog_stage_start_resources", { active = false, visible = false }
  } );

  for i = 1, 3 do

    table.insert( multiset_table, 
    { 
      "spr_int_dialog_stage_start_resources_"..i, { active = false, visible = false }
    } );

  end;  

  if ( 0 < #resources_data.resources ) then

    table.insert( multiset_table, 
    { 
      "obj_int_dialog_stage_start_resources", { active = true, visible = true }
    } );
    table.sort( resources_data.resources );
    for i = 1, math.min( #resources_data.resources, 3 ) do
      
      table.insert( multiset_table, 
      { 
        "spr_int_dialog_stage_start_resources_"..i,
          {
            active   = true,
            visible  = true,
            res      = ui.obj.Res( "resource", 
                         "icon_"..string.lower( resources_data.resources[ i ] ), 
                         "" )
          }
      } );

    end;

  end; 
  
end;
--------------------------------------------------------------------------------
function private.Enemy_Set ( enemy_data, multiset )
    
  if ( enemy_data.stage_type == "fight" ) then
    
    local path = "assets/content/"..enemy_data.content.."/"..enemy_data.name;
    local anm  = "anm_"..enemy_data.name;

    ui.obj.Set(
      "anm_int_dialog_stage_start_enemy",
      {
        active   = true,
        visible  = true,
        res      = path.."/"..anm,
        animfunc = "static",
        playing  = true
      },
      multiset
    );  

  else
    
    ui.obj.Disable( "anm_int_dialog_stage_start_enemy", true, multiset );
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.Animation_Set ( multiset_table, anim_data )
  
  local bg_mainname = anim_data.zone.background.name;
  local zone = anim_data.zone.background.content;

  if ( anim_data.stage_type == "fight" ) then
    bg_mainname = "arena";
    zone = "zones_1";
  end;

  local bg_fullname = "bg_"..bg_mainname.."_"..anim_data.stage_type;
 
  local light_colors = { r = 0.2352, g = 0.3058, b = 0.3058 };

  local light_color_r = anim_data.color.r + light_colors.r;
  if ( light_color_r > 1 ) then light_color_r = 1;end;

  local light_color_g = anim_data.color.g + light_colors.g;
  if ( light_color_g > 1 ) then light_color_g = 1; end;

  local light_color_b = anim_data.color.b + light_colors.b;
  if ( light_color_b > 1 ) then light_color_b = 1; end;
  
  local resources = "";
  
  if ( anim_data.resources ) and ( 0 < #anim_data.resources ) then
    
    resources = StringGet( "str_dialog_stage_start_resources" );
    
  end;
  
  local rewards      = anim_data.reward[ anim_data.stage_state ];
  local reward_count = "";
  if ( #rewards == 1 ) and ( rewards[ 1 ].name == "gold" ) then
    reward_count = tostring( rewards[ 1 ].count );
  end;
  if ( #rewards == 1 ) and ( rewards[ 1 ].name == "immortality" ) then
    reward_count = ui.text.FormatTime( rewards[ 1 ].count );
  end;
  
  table.insert( multiset_table, 
  { 
    "anm_int_dialog_stage_start_window", 
    {
      color_l_r    = light_color_r,
      color_l_g    = light_color_g,
      color_l_b    = light_color_b,
      color_d_r    = anim_data.color.r,
      color_d_g    = anim_data.color.g,
      color_d_b    = anim_data.color.b,
      reward_count = reward_count,
      moves        = private.strings.moves,
      targets      = private.strings.targets,
      reward       = private.strings.reward,
      resources    = resources,
      one          = "1.",
      two          = "2.",
      three        = "3.",
      bg_image     = "assets/content/"..zone.."/"..bg_fullname
    }
  } );
  
end;
--------------------------------------------------------------------------------
function private.Window_Set ( multiset )

  local shift_y = tern( IsPC() )( 0, -100 );

  ui.obj.Set( 
    "obj_int_dialog_stage_start", 
    { 
      pos_x = 0.5 * private.app_w,
      pos_y = 0.5 * private.app_h + shift_y
    },
    multiset
  );   
  
end;
--------------------------------------------------------------------------------
function private.Booster_TapSet ( name, index, count )

  ui.control.TapClear( "btn_int_dialog_stage_start_boosters_"..index );

  ui.control.TapSet( "btn_int_dialog_stage_start_boosters_"..index,
  {
    func_true = function ()

      if ( not private.popup_active ) then

        if ( count > 0 ) then

          private.Booster_Select( name, index );
        
        else

          _f.BuyBooster( name );
      
        end;  
          
         common.sound.Play( "select_boost" );
          
      end;

      private.popup_active = false;
           
      ui.popup.Flat_Hide( true ); 

    end,

    func_menter = function ()

      --

    end,

    func_mleave = function ()
      
      if ( private.popup_active ) then
        
        ui.popup.Flat_Hide( true ); 
        
      end;
      
    end,

    func_hold_down = function ()
   
      private.popup_active = true;
    
      ui.popup.Flat_Show( 
        "str_item_"..name.."_caption",
        "str_item_"..name.."_description",
        "btn_int_dialog_stage_start_boosters_"..index,
        { 0, -160 }
      );

    end,

    func_hold_up = function ()
    
      ui.popup.Flat_Hide( true ); 
  
    end;
  } );

end;
--------------------------------------------------------------------------------
function private.Booster_Select ( booster_name, index )

  local obj = "anm_int_dialog_stage_start_boosters_"..index.."_icon";

  if ( private.boosters_selected[ booster_name ] ) then

    ObjSet( obj, { animfunc = "booster", playing = true } );
    private.boosters_selected[ booster_name ] = nil;

  else

    ObjSet( obj, { animfunc = "booster_selected", playing = true } );
    private.boosters_selected[ booster_name ] = true;

  end;

end;
--------------------------------------------------------------------------------