-- name=int_dialog_stage_win
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Accept = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()
  
  ui.button.Create(
    "btn_int_dialog_stage_win_next",
    {
      title  = "str_dialog_stage_win_next",
      color  = "green",
      design = "medium",
      sound  = "simply_click",
    },
    function() _f.Accept(); end
  );
  
  local multiset = {};

  ui.obj.SetTextTag(
    "anm_int_dialog_stage_win_window", 
    "title",
    ne.strings.Get( "str_dialog_stage_win_caption" ),
    multiset
  );

  ui.obj.ToCenter( "obj_int_dialog_stage_win", 0, 0, multiset );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show (
    stage_name,
    stage_type,
    stage_data,
    skin_current,
    social_panel
  )

  private.SetDefault();
  
  private.hero_anm_name = ui.skin.Attach(
    "obj_int_dialog_stage_win_hero",
    skin_current,
    stage_type
  );
  
  local subtitle = tern( stage_type == "ride" )( 
    "str_dialog_stage_win_score", 
    "str_dialog_stage_win_fight"
  );

  local params_table = {
    score_count = tern( stage_type == "ride" )( 
      function () 
        return ui.text.FormatNumber( stage_data.score, true );
      end,
      function () 
        return StringGet( "str_dialog_stage_win_moves", 
                          ui.text.FormatNumber( stage_data.moves, true ) );
      end 
      )()
  };
    
  params_table.animfunc = "static_victory";
  params_table.tt_gold_count = "";
  params_table.rt_gold = "";
  params_table.tt_resource_count = "";
  params_table.rt_resource = "";
  params_table.tt_heart_count = "";
  params_table.rt_heart = "";

  if ( stage_data.rewards and #stage_data.rewards > 0 ) then
    
    params_table.animfunc = "static_victory_reward_"..math.min( #stage_data.rewards, 2 );
    
    for i = 1, #stage_data.rewards do
      
      local reward = stage_data.rewards[ i ];
      local icon_path = "assets/content/items/"..reward.name;
 
      if ( reward.type == "gold" ) then
        icon_path = "assets/content/items/gold_"..reward.stage_state;
      end;

      if ( reward.name == "immortality" ) then
        reward.count = ui.text.FormatTime( reward.count );
      end;

      params_table[ "tt_"..reward.type.."_count" ] = reward.count;
      params_table[ "rt_"..reward.type ] = icon_path;
  
    end;
    
  end;
  
  if ( stage_data.chests ) then
    
    local indexes = {
      woodchest   = 1,
      silverchest = 2,
      goldchest   = 3
    };
    
    local overall_count = 0;
    
    params_table.score_count = "";
    subtitle = "str_dialog_stage_win_chests";

    for chest, count in pairs( stage_data.chests ) do
      
      overall_count = overall_count + count;

      params_table[ "tt_count_"..indexes[ chest ] ] = count;
      params_table[ "rt_icon_"..indexes[ chest ]  ] = "assets/content/items/icon_"..chest;
      
    end;
    
    if ( overall_count == 0 ) then
      ui.button.SetText( "btn_int_dialog_stage_win_next", 
                         "str_dialog_stage_win_tryagain" );
      
      ui.obj.SetTextTag( "anm_int_dialog_stage_win_window", 
        "title",
        ne.strings.Get( "str_dialog_stage_win_caption_tryagain" )
      );                    
                         
      subtitle = "str_dialog_stage_win_chests_zero";
    end;
    
  end;
  
  params_table.score_title = subtitle;
  
  ObjMultiSet( {
    { "anm_int_dialog_stage_win_window_content", params_table }
  } );

  private.SetPanelForm( social_panel );
  
  ui.window.Open(
    "stage_win",
    nil,
    nil,
    "win_stage"
  );
  
  local func_OnTimer = function( func_end )
    
    common.Timer_Start(
      "tmr_int_dialog_stage_win_snd",
      4,
      function ()
      
        common.sound.Play( "win_stage_"..math.random( 1, 2 ) );
        
        func_end( func_end );
    
      end
    );
  
  end;
  
  ui.anim.Play(
    "anm_int_dialog_stage_win_window",
    "appear_portrait",
    function ()
    
      common.sound.Play( "win_stage_"..math.random( 1, 2 ) );
      
      ui.anim.Play(
        "anm_int_dialog_stage_win_window",
        "static_portrait_start",
        function ()

          common.sound.Play( "win_stage_"..math.random( 1, 2 ) );
          
          func_OnTimer( func_OnTimer );
        
        end,
        "static_portrait"
      );
    
    end
  );

end;
--------------------------------------------------------------------------------
function public.Hide ( func_close )
  
  ui.anim.Play( "anm_int_dialog_stage_win_window", "disappear_portrait" );
  
  ui.window.Close( "stage_win", function ()
    
    ui.skin.Detach( private.hero_anm_name );
    private.hero_anm_name = nil;
    
    common.Timer_Stop( "tmr_int_dialog_stage_win_snd" );

    ui.window.Hide( "int_dialog_stage_win", function () 
      
      if ( func_close ) then
        func_close();
      end;
      
    end);
    
  end);

end;
--------------------------------------------------------------------------------
function public.ShowHero ( stage_type )
  
  if ( "ride" == stage_type ) then

    ui.obj.Set( private.hero_anm_name, {
      res      = "assets/content/hero/anm_hero_ride",
      animfunc = "win_cycle",
      playing  = true
    } );

  else

    ui.obj.Set( private.hero_anm_name, {
      res      = "assets/content/hero/anm_hero_fight",
      animfunc = "win",
      playing  = true
    } );

  end;

  --ui.anim.Play( private.hero_anm_name, "hero_appear" );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.SetDefault ()
  
  local multiset_table = {};

  for i = 1, 3 do

    multiset_table[ "tt_count_"..i ] = "";
    multiset_table[ "rt_icon_"..i  ] = "";
    
  end;
  
  ui.button.SetText( "btn_int_dialog_stage_win_next", 
                     "str_dialog_stage_win_next" );

  ui.obj.SetTextTag(
    "anm_int_dialog_stage_win_window", 
    "title",
    ne.strings.Get( "str_dialog_stage_win_caption" )
  );
  
  ObjMultiSet( {
    { "anm_int_dialog_stage_win_window_content", multiset_table }
  } );

end;
--------------------------------------------------------------------------------
function private.SetPanelForm ( social_panel )

  local screen_w = GetAppWidth();
  local screen_h = GetAppHeight();

  if ( social_panel ) then

    ui.obj.Set(
      "obj_int_dialog_stage_win",
      {
        pos_x = screen_w/2,
        pos_y = screen_h / 2 - 60
      }
    );

  else

    ui.obj.Set(
      "obj_int_dialog_stage_win",
      {
        pos_x = screen_w/2,
        pos_y = screen_h / 2
      }
    );

  end;

end;
--------------------------------------------------------------------------------