-- name=int_dialog_starterpack
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Accept = function () end,
  Reject = function () end,
  OnClose = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  private.app_o = ui.screen.Orientation();

  private.area = {
    portrait  = { -260, -475, 520, 950 },
    landscape = { -207, -380, 416, 760 }
  };

  private.this = {
    dialog = "obj_int_dialog_starterpack",
    button = {
      close = "btn_int_dialog_starterpack_close",
      buy   = "btn_int_dialog_starterpack_buy"
    }
  };
  
  local multiset = {};

  ui.window.SetArea(
    private.this.dialog,
    private.area[ private.app_o ],
    multiset
  );
  
  ui.obj.ToCenter( private.this.dialog, 0, 0, multiset );
  
  ui.button.Create(
    private.this.button.close,
    {
      design = "close"
    },
    function () _f.Reject(); end
  );
  
  ui.button.Create(
    private.this.button.buy,
    {
      design = "buy",
      timer  = ""
    },
    function () _f.Accept(); end
  );

  -- JK: increase buy button area
  ui.obj.Set(
    private.this.button.buy,
    {
      inputrect_x = -200,
      inputrect_y = -450,
      inputrect_w =  400,
      inputrect_h =  500
    },
    multiset
  );

  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show ( starterpack )
  
  private.active = true;
  
  private.SetDefaults( starterpack );
 
  private.Timer();
  
  ui.anim.Play( "anm_int_dialog_starterpack_window", "static_"..private.app_o );
  
  ui.window.Open(
    "starterpack",
    nil,
    nil,
    "starter_pack"
  );
  
  common.music.Stop( "assets/content/audio/music/aud_map_mus" );
  
  common.Timer_Create(
    "int_dialog_starterpack",
    "tmr_int_dialog_starterpack_music",
    25.0,
    function ()
      
      common.Timer_Delete( "tmr_int_dialog_starterpack_music" );
      
      common.music.Play(
        "assets/content/audio/music/aud_map_mus",
        nil,
        nil,
        true
      );
    
    end
  );

end;
--------------------------------------------------------------------------------
function public.Hide ( func_after )
  
  private.active = false;
  
  if ( ui.obj.Get( "tmr_int_dialog_starterpack_music" ) ) then
    common.Timer_Delete( "tmr_int_dialog_starterpack_music" );
    common.sound.Stop( "starter_pack" );  
    common.music.Play( "assets/content/audio/music/aud_map_mus", nil, nil, true );
  end;
  
  local close = function ()
    if (func_after) then
      func_after(_f.OnClose);
    else
      _f.OnClose();
    end;
  end;
  
  ui.window.Close( "starterpack", close );

end;
--------------------------------------------------------------------------------
function public.UpdatePrice ( price )

  if ( private.active and price and price ~= "" ) then
  
    ui.button.SetText( private.this.button.buy, price );
  
  end;

end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function private.Timer ()

  if ( not private.active ) then

    return;

  end;

  local multiset = {};
  local seconds  = TimeGetEventRemainingTime( "mte_starterpack_cooldown" );
  
  local time;

  if ( seconds <= 0 ) then

    common.Timer_Stop(
     "tmr_int_dialog_starterpack",
      multiset
    );

  else

    common.Timer_Start(
      "tmr_int_dialog_starterpack",
      1,
      function()

        private.Timer();

      end,
      multiset
    );
    
    time = ui.text.FormatTime( seconds );
    
  end;
  
  ui.button.SetTime(
    private.this.button.buy,
    time,
    multiset
  );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.SetDefaults ( starterpack )
  
  local objects = {};  

  local multiset_params = {};  
  
  local boosters = { hammer = 1, magicgift = 2 };

  multiset_params.tt_title   = "str_int_dialog_starterpack_title";
  multiset_params.tt_percent = "str_int_dialog_starterpack_percent";
  multiset_params.tt_header  = "str_int_dialog_starterpack_header";
  
  for i = 1, #starterpack.currency_count do
    
    local starterpack_content = starterpack.currency_count[ i ];
    
    if ( starterpack_content.type == "gold" ) then
      multiset_params.tt_gold = ui.text.FormatNumber( starterpack_content.count );
    else
      multiset_params[ "tt_booster"..boosters[ starterpack_content.name ] ] = 
      ne.strings.Get( "str_int_dialog_starterpack_countx", ui.text.FormatNumber( starterpack_content.count ) ); 
    end;
    
  end;
  
  local param0, param1, param2 = multiset_params.tt_gold, multiset_params.tt_booster1, multiset_params.tt_booster2;
  
  multiset_params.tt_description = ne.strings.Get( "str_int_dialog_starterpack_desc", param0, param1, param2 );
  
  ui.obj.Set( "anm_int_dialog_starterpack_window", multiset_params, objects );

  local price = tern( starterpack.price == "" )( ne.strings.Get( "str_dialog_shop_gold_buy" ),
                                                starterpack.price );
  
  ui.button.SetText( private.this.button.buy, price, objects );
  
  ObjMultiSet( objects );
  
end;
--------------------------------------------------------------------------------