-- name=int_dialog_tutorial
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Close = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  private.this = {
    int = "int_dialog_tutorial"
  };

  private.screen_w = ui.screen.Width();
  private.screen_h = ui.screen.Height();

  private.time_show = 0.25;
  private.time_hide = 0.25;

  private.arrows_shown = {};

  private.Default_SetDialog();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show (
    text_id,
    positions,
    hole_params,
    func_next_end,
    text_param,
    interface_lock
  )

  private.NextEnd = func_next_end;

  ObjAttach( private.this.int, "hub_tutorial" );
  ObjStopAnimate( private.this.int, "color_rgb" );

  local objects = {};
  
  table.insert( objects, { private.this.int,
  {
    color_r = 1,
    color_g = 1,
    color_b = 1 } } );

  table.insert( objects, { "obj_int_dialog_tutorial",
  {
    input = true,
    alp   = 0 } } );
    
  table.insert( objects, { "obj_int_dialog_tutorial_block",
  {
    input = true } } );

  ObjMultiSet( objects );

  private.Default_SetContent( positions, hole_params );

  private.active = true;
  mediator.Notify( "Interface_Dialog_Tutorial_Open" );

  local func_show_end = function()
  
    if ( not interface_lock ) then

      ui.screen.Lock( false );

    end;  

    ObjSet( "obj_int_dialog_tutorial_main",
    {
      active  = true,
      input   = true,
      visible = true } );

    private.Dialog_Show(
    { 
      text_id           = text_id,
      text_param        = text_param,
      dlg_start         = positions.dlg_start or - 1700,
      dlg_end           = positions.dlg_end or private.screen_w / 2,
      character         = true,
      character_start   = positions.character_start or nil,
      character_end     = positions.character_end or private.screen_w / 2,
      character_scale   = positions.character_scale or 1,
      character_pos_y   = positions.character_pos_y or -96
    } );

  end;

  local tme = private.time_show;

  ObjAnimate( "obj_int_dialog_tutorial", "alp", 0, 0, func_show_end,
  {
    0.0, 1, 0.0,
    tme, 1, 1.0
  } );

end;
--------------------------------------------------------------------------------
function public.Hide ( func_end, fast_hide, params )

  if ( not private.active ) then return; end;

  private.active = false;
  mediator.Notify( "Interface_Dialog_Tutorial_Hide" );

  local func_hide_end = function()

    local objects = {};

    table.insert( objects, { "obj_int_dialog_tutorial_main",
    {
      input   = false,
      visible = false } } );
      
    table.insert( objects, { "obj_int_dialog_tutorial_dark",
    {
      input   = false,
      visible = false } } );
      
    table.insert( objects, { "btn_int_dialog_tutorial_black",
    {
      input   = false,
      visible = false } } );
      
    table.insert( objects, { "obj_int_dialog_tutorial",
    {
      input = false,
      alp   = 0 } } );

    ObjMultiSet( objects );

    --ObjAttach( private.this.int, "ng_interface_anim" );

    func_end();

    if ( params ) and ( params.lock ) then

      ui.screen.Lock( true ); 

    end;

  end;

  AnimStop( "anm_int_dialog_tutorial_hand" );
  AnimStop( "anm_int_dialog_tutorial_sticker" );
  
  ObjSet( "obj_int_dialog_tutorial_block", { input = true } );
  ObjSet( "obj_int_dialog_tutorial_lock",  { input = false } );

  if ( fast_hide ) then

    func_hide_end();

  else

    local tme = private.time_hide;
  
    ObjAnimate( "obj_int_dialog_tutorial", "alp", 0, 0, func_hide_end,
    {
      0.0, 1, 1.0,
      tme, 1, 0.0
    } );

  end;

end;
--------------------------------------------------------------------------------
function public.IsActive ()

  return private.active;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Dialog_Show ( params )

  local text_id         = params.text_id;
  local text_param      = params.text_param;
  local character       = params.character;
  local character_start = params.character_start;
  local character_scale = params.character_scale;
  local character_pos_y = params.character_pos_y;
  local character_end   = params.character_end;
  local dlg_start       = params.dlg_start;
  local dlg_end         = params.dlg_end;

  local func_end = function()

    ObjSet( "obj_int_dialog_tutorial_block", { input = false } );

  end;

  if ( character ) then

    local scale = 0.8;
    local pos_x = 135;
    
    ObjSet( "obj_int_dialog_tutorial_character",
    {
      active  = true,
      visible = true,
      scale_x = scale * character_scale,
      pos_x   = pos_x * character_scale,
      pos_y   = character_pos_y } );
      
    ui.anim.Play( "anm_int_dialog_tutorial_character", "static" );

    if ( character_start ) then

      ObjAnimate( "obj_int_dialog_tutorial_character", "pos_x", 0, 0, "",
      {
        0.0,               1, character_start,
        0.1,               1, character_start,
        private.time_show, 1, character_end
      } );

    end;
  
  end;

  ObjSet( "txt_int_dialog_tutorial_dialog",
  {
    text     = "str_tutorial_"..text_id,
    param0   = text_param,
    fontsize = 30
  } );
  
  ui.text.FontsizeByHeight( "txt_int_dialog_tutorial_dialog", 120, 30 );

  ObjAnimate( "obj_int_dialog_tutorial_dialog", "pos_x", 0, 0, func_end,
  {
    0.0,               1, dlg_start,
    0.1,               1, dlg_start,
    private.time_show, 1, dlg_end
  } );

   common.sound.Play( "tutorial_start" );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Flashlight_Set ( index, params )

  local dark_names = { "left", "right", "top", "bottom" };

  local flashlight_object = "";

  local objects = {};

  if ( params.black == nil ) then
  
    params.black = true;
    
  end;
  
  params.lock = params.lock or false;
  
  table.insert( objects, 
  { "obj_int_dialog_tutorial_dark",
    {
      input = params.lock
    } 
  } );

  for d = 1, #dark_names do

    table.insert( objects,
      { "spr_int_dialog_tutorial_dark_"..dark_names[ d ],
        {
          input   = params.lock,
          visible = params.black
        }
      } );

  end;

  if ( params.input ) then

    local lock_object = "obj_int_dialog_tutorial_lock";

    table.insert( objects, 
    { lock_object,
      { 
        input = true,
        event_mup = function ()

          ObjSet( lock_object, { input = false } );
          
          private.Flashlight_Hide();
          private.NextEnd();

        end
      }
    } );

  end;

  local pos_x = params.pos_x;
  local pos_y = params.pos_y;

  params.angle = params.angle or 0;

  if ( params.focus ) then

    params.offset_x = params.offset_x or 0;
    params.offset_y = params.offset_y or 0;

    local focus_pos = ui.obj.Position( params.focus );

    pos_x = math.floor( focus_pos[ 1 ] ) + ( params.offset_x );
    pos_y = math.floor( focus_pos[ 2 ] ) + ( params.offset_y );

  end;

  local grid_w;
  local grid_h;
    
  if ( params.grid ) then

    flashlight_object = "grd_int_dialog_tutorial_hole";

    table.insert( objects, { flashlight_object,
      {
        visible = true,
        input   = true,
        pos_x   = pos_x,
        pos_y   = pos_y,
        ang     = params.angle,
        scale_x = params.scale_x or params.scale or 1,
        scale_y = params.scale_y or params.scale or 1,
        res     = "assets/content/interface/int_dialog_tutorial/grd_round_hole_"..params.res_id
      } 
    } );

    local corner_w = params.grid.corner_width;
    local corner_h = params.grid.corner_height;
    
    grid_w = params.grid.width + 2 * corner_w;
    grid_h = 2 * corner_h;

    local grid_params = 
    {
       0, { pos_x = - ( ( grid_w / 2 ) ),            pos_y = - ( ( grid_h / 2 ) ) },
       1, { pos_x = - ( ( grid_w / 2 ) - corner_w ), pos_y = - ( ( grid_h / 2 ) ) },
       2, { pos_x =   ( ( grid_w / 2 ) - corner_w ), pos_y = - ( ( grid_h / 2 ) ) },
       3, { pos_x =   ( ( grid_w / 2 ) ),            pos_y = - ( ( grid_h / 2 ) ) },

       4, { pos_x = - ( ( grid_w / 2 ) ),            pos_y =   ( ( grid_h / 2 ) ) },
       5, { pos_x = - ( ( grid_w / 2 ) - corner_w ), pos_y =   ( ( grid_h / 2 ) ) },
       6, { pos_x =   ( ( grid_w / 2 ) - corner_w ), pos_y =   ( ( grid_h / 2 ) ) },
       7, { pos_x =   ( ( grid_w / 2 ) ),            pos_y =   ( ( grid_h / 2 ) ) }
    };

    GridSet( flashlight_object, 0, grid_params );

    table.insert( objects, { "spr_int_dialog_tutorial_hole", { visible = false, input = false } } );

  else

    flashlight_object = "spr_int_dialog_tutorial_hole";

    table.insert( objects, { flashlight_object,
      {
        visible = params.black,
        input   = true,
        pos_x   = pos_x,
        pos_y   = pos_y,
        ang     = params.angle,
        scale_x = params.scale_x or params.scale or 1,
        scale_y = params.scale_y or params.scale or 1,
        res     = "assets/content/interface/int_dialog_tutorial/hole_"..params.res_id
      } 
    } );

    table.insert( objects, { "grd_int_dialog_tutorial_hole", { visible = false, input = false } } );

  end;

  if ( params.hand ) then

    table.insert( objects, { "obj_int_dialog_tutorial_hand",
      {
        visible = true, 
        pos_x   = pos_x + ( params.hand.offset_x or 0 ), 
        pos_y   = pos_y + ( params.hand.offset_y or 0 ),
        ang     = params.hand.angle or 0,
        scale_x = params.hand.scale_x or 1,
        scale_y = params.hand.scale_y or 1
      } 
    } );

    local animfunc = "hand_move";

    if ( params.hand.type ) then

      animfunc = "hand_"..params.hand.type;

    end;

    ui.anim.Play( "anm_int_dialog_tutorial_hand", animfunc );

  end;

  if ( params.arrow ) then

    local arrow_pos_x = params.pos_x;
    local arrow_pos_y = params.pos_y;

    if ( params.focus ) then

      params.arrow.offset_x = params.arrow.offset_x or 0;
      params.arrow.offset_y = params.arrow.offset_y or 0;

      local focus_pos = ui.obj.Position( params.focus );

      arrow_pos_x = math.floor( focus_pos[ 1 ] ) + ( params.arrow.offset_x );
      arrow_pos_y = math.floor( focus_pos[ 2 ] ) + ( params.arrow.offset_y );

    elseif ( params.arrow.offset_x or params.arrow.offset_y ) then

      params.arrow.offset_x = params.arrow.offset_x or 0;
      params.arrow.offset_y = params.arrow.offset_y or 0;

      arrow_pos_x = math.floor( arrow_pos_x ) + ( params.arrow.offset_x );
      arrow_pos_y = math.floor( arrow_pos_y ) + ( params.arrow.offset_y ) ;

    end;

    table.insert( objects, { "obj_int_dialog_tutorial_arrow",
      {
        alp     = 0,
        visible = true,
        input   = false,
        pos_x   = arrow_pos_x,
        pos_y   = arrow_pos_y,
        ang     = params.arrow.angle or 0,
        scale_x = 1,
        scale_y = 1
      } 
    } );

    table.insert( objects, { "grd_int_dialog_tutorial_arrow", { active = false, visible = false, pos_y = 0 } } );
    table.insert( objects, { "anm_int_dialog_tutorial_arrow", { active = true,  visible = true } } );

    local grid = false;

    if ( params.length ) then

      table.insert( objects, { "anm_int_dialog_tutorial_arrow", { active = false, visible = false } } );
      table.insert( objects, { "grd_int_dialog_tutorial_arrow", { active = true,  visible = true } } );
      
      grid = true;

      local length = params.length or 38; 

      local grid_params = 
      {
         0, { pos_y = -39 },
         1, { pos_y = -39 },
         2, { pos_y = length - 49 },
         3, { pos_y = length - 49 },

         4, { pos_y = length },
         5, { pos_y = length }
      };

      GridSet( "grd_int_dialog_tutorial_arrow", 0, grid_params );

    end;

    ObjMultiSet( objects );

    private.Arrow_Show( grid );

  end;
  
  ObjMultiSet( objects );

  private.Flashlight_SetParts(
    flashlight_object,
    pos_x,
    pos_y,
    params.angle,
    grid_w,
    grid_h
  );

  private.Flashlight_Show();

end;
--***********************************************************************************************************
--***********************************************************************************************************
function private.Flashlight_SetParts ( obj_name, spr_pos_x, spr_pos_y, spr_angle, draw_width, draw_height )

  local objects = {};
  local dark_obj = ObjGet( obj_name );

  table.insert( objects, { "obj_int_dialog_tutorial_dark", { input = true, visible = true, pos_x = spr_pos_x, pos_y = spr_pos_y } } );

  local spr_pos_x       = math.floor( spr_pos_x );
  local spr_pos_y       = math.floor( spr_pos_y );
  local spr_draw_width  = ( draw_width  or dark_obj.draw_width  ) * dark_obj.scale_x ;
  local spr_draw_height = ( draw_height or dark_obj.draw_height ) * dark_obj.scale_y;

  for i = 1, 3 do

    if ( spr_pos_x % 4 ~= 0 ) then

      spr_pos_x = spr_pos_x + 1;

    else

      break;

    end;

  end;

  for i = 1, 3 do

    if ( spr_pos_y % 4 ~= 0 ) then

      spr_pos_y = spr_pos_y + 1;

    else

      break;

    end;

  end;

  local spr_draw_repos  = math.abs( math.floor( spr_angle, 0.5 * math.pi ) );

  if ( spr_draw_repos == 1 or ( 0.5 * math.pi == math.abs( spr_angle ) ) ) then

    spr_draw_width, spr_draw_height = spr_draw_height, spr_draw_width;

  end;

  local app_width  = private.screen_w;
  local app_height = private.screen_h;

  table.insert( objects, { "spr_int_dialog_tutorial_dark_left", 
    {
      pos_x   = - spr_draw_width  / 2, 
      pos_y   =   spr_draw_height / 2, 
    }
  } );

  table.insert( objects, { "spr_int_dialog_tutorial_dark_top", 
    {
      pos_x   = - spr_draw_width  / 2, 
      pos_y   = - spr_draw_height / 2, 
    }
  } );

  table.insert( objects, { "spr_int_dialog_tutorial_dark_right", 
    {
      pos_x   =   spr_draw_width  / 2, 
      pos_y   = - spr_draw_height / 2, 
    }
  } );

  table.insert( objects, { "spr_int_dialog_tutorial_dark_bottom", 
    {
      pos_x   = spr_draw_width  / 2, 
      pos_y   = spr_draw_height / 2, 
    }
  } );

  ObjMultiSet( objects );

end;
--------------------------------------------------------------------------------
function private.Flashlight_Show ()

  

end;
--------------------------------------------------------------------------------
function private.Flashlight_Hide ()

  

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Arrow_Set ( index, params )

  local objects = {};

  local lock_object = "tpr_int_dialog_tutorial_lock";

  if ( params.lock ) then

    lock_object = "obj_int_dialog_tutorial_lock";

  end;

  table.insert( objects, { "tpr_int_dialog_tutorial_lock",
  {
    input   = false,
    visible = false } } );
    
  table.insert( objects, { "obj_int_dialog_tutorial_lock",
  {
    input   = false,
    visible = false } } );

  if not ( params.wait ) then

    table.insert( objects, { lock_object,
    {
      input     = true,
      visible   = true,
      event_mup = function ()

        ObjSet( lock_object,
        {
          input   = false,
          visible = false } );

        private.Arrow_Hide();
        private.NextEnd();

    end } } );

  end;

  local pos_x = params.pos_x;
  local pos_y = params.pos_y;

  if ( params.focus ) then

    params.offset_x = params.offset_x or 0;
    params.offset_y = params.offset_y or 0;

    local focus_pos = ui.obj.Position( params.focus );

    pos_x = math.floor( focus_pos[ 1 ] ) + ( params.offset_x );
    pos_y = math.floor( focus_pos[ 2 ] ) + ( params.offset_y );

  end;

  table.insert( objects, { "obj_int_dialog_tutorial_arrow_"..index,
    {
      alp     = 0,
      visible = true,
      input   = false,
      pos_x   = pos_x,
      pos_y   = pos_y,
      ang     = params.angle or 0,
      scale_x = params.scale or 1,
      scale_y = params.scale or 1
    } 
  } );

  table.insert( objects, { "grd_int_dialog_tutorial_arrow_"..index,
  {
    active  = false,
    visible = false,
    pos_y   = 0 } } );
    
  table.insert( objects, { "anm_int_dialog_tutorial_arrow_"..index,
  {
    active  = true,
    visible = true } } );

  local grid = false;

  if ( params.length ) then

    table.insert( objects, { "anm_int_dialog_tutorial_arrow_"..index,
    {
      active  = false,
      visible = false } } );
      
    table.insert( objects, { "grd_int_dialog_tutorial_arrow_"..index,
    {
      active  = true,
      visible = true } } );
    
    grid = true;

    local length = params.length or 38; 

    local grid_params = 
    {
       0, { pos_y = -39 },
       1, { pos_y = -39 },
       2, { pos_y = length - 49 },
       3, { pos_y = length - 49 },

       4, { pos_y = length },
       5, { pos_y = length }
    };

    GridSet( "grd_int_dialog_tutorial_arrow_"..index, 0, grid_params );

  end;

  ObjMultiSet( objects );

  private.Arrow_Show( index, grid );

end;
--------------------------------------------------------------------------------
function private.Arrow_Show ( index, grid )

  table.insert( private.arrows_shown, index );

  local arrow_object = "obj_int_dialog_tutorial_arrow_"..index;

  ObjAnimate( arrow_object, "alp", 0, 0, "", 
  { 
    0.0, 0, 0.0, 
    0.3, 0, 1.0 
  } );

  if ( grid ) then

    local arrow_sprite = "grd_int_dialog_tutorial_arrow_"..index;

    local pos_y = 0;

    ObjAnimate( arrow_sprite, "pos_y", 1, 0, "", 
    { 
      0.0, 0, pos_y, 
      0.5, 0, pos_y + 10, 
      1.0, 0, pos_y
    } );

  end;

end;
--------------------------------------------------------------------------------
function private.Arrow_Hide ()

  for a = 1, #private.arrows_shown do

    local index = private.arrows_shown[ a ];

    local arrow_object = "obj_int_dialog_tutorial_arrow_"..index;
    local arrow_sprite = "grd_int_dialog_tutorial_arrow_"..index;

    ObjStopAnimate( arrow_sprite, "pos_y" );

    ObjAnimate( arrow_object, "alp", 0, 0, "", 
    { 
      0.0, 0, 1.0, 
      0.3, 0, 0.0 
    } );

  end;

  private.arrows_shown = {};

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Sticker_Set ( index, params )

  local objects = {};
  
  local sticker_object = "obj_int_dialog_tutorial_sticker";
  local sticker_anim   = "anm_int_dialog_tutorial_sticker";
  
  local lock_object = "tpr_int_dialog_tutorial_lock";
  
  if ( params.lock ) then

    lock_object = "obj_int_dialog_tutorial_lock";

  end;

  table.insert( objects, { "tpr_int_dialog_tutorial_lock",
  {
    input   = false,
    visible = false } } );
    
  table.insert( objects, { "obj_int_dialog_tutorial_lock",
  {
    input   = false,
    visible = false } } );

  local start = params.start;
  
  local finish = params.finish;
  
  table.insert( objects, { sticker_object, {
    active  = true,
    visible = true,
    input   = false,
    pos_x   = start[1] or 0,
    pos_y   = start[2] or 0,
    ang     = params.angle or 0,
    scale_x = params.scale or 1,
    scale_y = params.scale or 1
  } } );
  
  local func_animate;
  func_animate = function ()
  
    ObjSet( sticker_object, {
      pos_x   = start[1] or 0,
      pos_y   = start[2] or 0
    } );
  
    ui.anim.Play( sticker_anim, "start", function ()
    
      local func_on_drag_end = function ()
      
        ui.anim.Play( sticker_anim, "stick", func_animate );
        
      end;
    
      ui.anim.Play( sticker_anim, "drag" );
      
      ObjAnimate( sticker_object, "pos_x", 0, 0, func_on_drag_end, 
      { 
        0.0, 1, start[1],
        1.5, 1, finish[1]
      } );
      
      ObjAnimate( sticker_object, "pos_y", 0, 0, "", 
      { 
        0.0, 2, start[2],
        1.5, 2, finish[2]
      } );
      
    end );
    
  end;
  
  func_animate();
  
  ObjMultiSet( objects );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Default_SetContent ( positions, hole_params )

  local objects = {};

  local set = {
    visible = false,
    ang     = 0
  };
  
  ui.obj.Set( "obj_int_dialog_tutorial_hand",    set, objects );
  
  for a = 1, 4 do
  
    ui.obj.Set( "obj_int_dialog_tutorial_arrow_"..a, set, objects );
  
  end;
  
  ui.obj.Hide( "spr_int_dialog_tutorial_hole",  objects );
  ui.obj.Hide( "btn_int_dialog_tutorial_black", objects );
  ui.obj.Hide( "grd_int_dialog_tutorial_hole",  objects );
  
  ui.obj.Disable( "obj_int_dialog_tutorial_character", true, objects );
  ui.obj.Disable( "obj_int_dialog_tutorial_sticker",   true, objects );

  ui.obj.Set(
    "obj_int_dialog_tutorial_dark",
    {
      inputrect_init = false,
      inputrect_w    = 0,
      inputrect_h    = 0
    },
    objects
  );

  if ( positions ) then

    ui.obj.Set(
      "obj_int_dialog_tutorial_dialog",
      { 
        pos_y   = positions.dlg_pos_y or 252,
        scale_x = positions.dlg_scale_x or positions.dlg_scale or 1.0,
        scale_y = positions.dlg_scale_y or positions.dlg_scale or 1.0
      },
      objects
    );
    
  end;

  local dark_names = { "left", "right", "top", "bottom" };

  for d = 1, #dark_names do
 
    ui.input.Off( "spr_int_dialog_tutorial_dark_"..dark_names[ d ], objects );

  end;

  ui.obj.Multiset( objects );

  if ( hole_params ) then

    local preset_functions = 
    {
      flashlight = private.Flashlight_Set,
      arrow      = private.Arrow_Set,
      sticker    = private.Sticker_Set
    };

    if ( hole_params.type ) then

      preset_functions[ hole_params.type ]( 1, hole_params );

    else

      for h = 1, #hole_params do

        preset_functions[ hole_params[ h ].type ]( h, hole_params[ h ] );

      end;
      
      ObjSet( "obj_int_dialog_tutorial_lock",
      {
        visible = hole_params.black,
        input   = hole_params.lock
      } );      
      
      ui.control.TapClear( "obj_int_dialog_tutorial_lock" );

      ui.control.TapSet( "obj_int_dialog_tutorial_lock",
      {
        func_true = function ()

          ui.input.Off( "obj_int_dialog_tutorial_lock" );

          private.Arrow_Hide();
          private.NextEnd();

        end
      } );      

    end;
   
  else

    ui.control.TapClear( "btn_int_dialog_tutorial_black" );

    ui.control.TapSet( "btn_int_dialog_tutorial_black",
    {
      func_true = function ()

        ui.input.Off( "btn_int_dialog_tutorial_black" );

        private.NextEnd();

      end
    } );

    ObjSet( "btn_int_dialog_tutorial_black",
    {
      visible = true,
      input   = true
    } );
     
  end;

end;
--------------------------------------------------------------------------------
function private.Default_SetDialog ()

  local screen_w = private.screen_w;
  local screen_h = private.screen_h;

  local objects = {};

  table.insert( objects, { "obj_int_dialog_tutorial_block", { inputrect_w = screen_w, inputrect_h = screen_h } } );
  table.insert( objects, { "tpr_int_dialog_tutorial_lock",  { inputrect_w = screen_w, inputrect_h = screen_h } } );
  table.insert( objects, { "obj_int_dialog_tutorial_lock",  { inputrect_w = screen_w, inputrect_h = screen_h } } );
  table.insert( objects, { "btn_int_dialog_tutorial_black", { inputrect_w = screen_w, inputrect_h = screen_h } } );
  table.insert( objects, { "spr_int_dialog_tutorial_black", { 
    pos_x   = 0.5 * screen_w, -- 1/2
    pos_y   = 0.5 * screen_h,
    scale_x = 0.015625 * screen_w, -- 1/64
    scale_y = 0.015625 * screen_h
  } } );

  ObjMultiSet( objects );  

  local grd_back_params =
  {
     0, { pos_x = -   0.5 * screen_w },
     1, { pos_x = - ( 0.5 * screen_w - 2 ) },
     2, { pos_x =   ( 0.5 * screen_w - 2 ) }, 
     3, { pos_x =     0.5 * screen_w }, 

     4, { pos_x = -   0.5 * screen_w }, 
     5, { pos_x = - ( 0.5 * screen_w - 2 ) }, 
     6, { pos_x =   ( 0.5 * screen_w - 2 ) }, 
     7, { pos_x =     0.5 * screen_w },
 
     8, { pos_x = -   0.5 * screen_w },
     9, { pos_x = - ( 0.5 * screen_w - 2 ) },
    10, { pos_x =   ( 0.5 * screen_w - 2 ) }, 
    11, { pos_x =     0.5 * screen_w }, 

    12, { pos_x = -   0.5 * screen_w }, 
    13, { pos_x = - ( 0.5 * screen_w - 2 ) }, 
    14, { pos_x =   ( 0.5 * screen_w - 2 ) }, 
    15, { pos_x =     0.5 * screen_w }, 
  }

  GridSet( "grd_int_dialog_tutorial_dialog_back", 0, grd_back_params );

end;
--------------------------------------------------------------------------------
function private.AnimTwoHand ( number )

  ObjSet( "obj_int_dialog_tutorial_hand", private.hand_params[ number ] );

  local animfunc = private.hand_params[ number ].animfunc;

  if ( number == 1 ) then

    number = 2;

  else

    number = 1;

  end;

  local func_end = function ()

    private.AnimTwoHand ( number );

  end;

  AnimStop( "anm_int_dialog_tutorial_hand" );
  ui.anim.Play( "anm_int_dialog_tutorial_hand", animfunc, func_end );

end;
--------------------------------------------------------------------------------