-- name=int_dialog_worldmap
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  IsContentReady  = function () end,
  GoTo            = function () end,
  OnClose         = function () end,
  Share           = function () end,
  GetSharingWorld = function () end
};

local _this = {};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function  public.__InitObjects ()

  private.orientation = ui.screen.Orientation();
  private.is_active = false;
  private.scc_size =
  {
    portrait           = { w = 600, h = 800 }, 
    landscape          = { w = 600, h = 800 },
    portrait_featured  = { w = 600, h = 600 }, 
    landscape_featured = { w = 600, h = 600 }
  };

  private.window_area =
  {
    portrait  = { -320, -420, 640, 900 },
    landscape = { -250, -320, 500, 700 }
  };

  local multiset = {};

  ui.obj.ToCenter( "obj_int_dialog_worldmap", 0, 0, multiset );

  ui.window.SetArea(
    "obj_int_dialog_worldmap",
    private.window_area[ private.orientation ],
    multiset
  );

  ui.obj.Set( 
    "anm_int_dialog_worldmap_slot_share",
    { 
      rt_back = "assets/content/worlds/share",
      tt_description = ne.strings.Get( "str_int_dialog_worldmap_share_description" ),
      rt_red = 0.196,
      rt_green = 0.812,
      rt_blue = 0.776
    },
    multiset
  );
  
  ui.button.Create(
    "btn_int_dialog_worldmap_close",
    {
      design = "close"
    },
    function () 
      public.Hide();
      v_log.Playhaven_LogPlacement( "worlds_close" );
    end
  );

  ui.button.Create(
    "btn_int_dialog_worldmap_slot_share_click",
    {
      title  = "str_int_dialog_worldmap_share",
      design = "facebook",
      sound  = "simply_click"
    },
    function () 

      _f.Share();
      
    end
  );
  
  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show ( section, sections_list, is_featured, skin, unlock, new_worlds )

  if ( private.is_active ) then return; end;
  
  private.is_active = true;
  
  _this.is_featured = is_featured;

  private.new_worlds = new_worlds;
  
  local layout = private.orientation;
        layout = tern( is_featured )( layout.."_featured", layout );

  local size = private.scc_size[ layout ];

  local multiset = {};

  ne.obj.Detach( "anm_int_dialog_worldmap_slot_share" );
  
  private.SetupScroller( size, multiset );

  ui.obj.Set(
    "anm_int_dialog_worldmap_window",
    {
      animfunc = layout,
      title    = ne.strings.Get( "str_int_dialog_worldmap_title" )
    },
    multiset
  );

  ui.obj.Multiset( multiset );

  local indexed_list = private.IndexList( sections_list );

  private.CreateListItems( sections_list );
  private.UpdateListItems( sections_list, section, skin );
  
  local world = _f.GetSharingWorld();
  local state_panel_fb = false;
  
  if ( world ) then
    state_panel_fb = private.InsertSlotShareToIndex( indexed_list[ world ].id, sections_list );
  end;
  
  private.Focus( indexed_list[ section ].id, #sections_list, is_featured, state_panel_fb );
  
  if ( unlock ) then

    local id   = indexed_list[ section ].id + 1;
    local name = sections_list[ id ].name;
    local item = "anm_int_dialog_worldmap_slot_"..name;

    ObjSet( item, { animfunc = "island_lock" } );
    private.Focus( indexed_list[ section ].id, #sections_list, is_featured, true );
    
    ui.window.Open( "worldmap", 
    function()

      common.sound.Play( "new_land" );
      ui.anim.Play( item, "island_unlock" );

    end );

  else

    ui.window.Open( "worldmap" );

  end;
  
end;
--------------------------------------------------------------------------------
function public.Hide ( func_on_hide )

  if ( not private.is_active ) then return; end;
  
  private.is_active = false;
  
  ui.window.Close(
    "worldmap",
    function ()
      
      local objects =
        ObjGetRelations( "scc_int_dialog_worldmap_sections" ).childs;

      for i = 1, #objects do
  
        ne.obj.Detach( objects[ i ] );
        
      end;
      
      ui.window.Hide( "int_dialog_worldmap", func_on_hide );

      if ( _this.is_featured ) then

        mediator.Notify( "Interface_Dialog_Worldmap_ClosedWithFeatured" );

      end;

      _f.OnClose()

    end
  );

end;
--------------------------------------------------------------------------------
function public.ShareComplete ( section )

  ne.obj.Detach( "anm_int_dialog_worldmap_slot_share" );
  
  local childs = ne.obj.GetRelations( "scc_int_dialog_worldmap_sections" ).childs;
  
  if ( childs[ #childs ]:find( section ) ) then
  
    local scc_pos = ObjGet( "scc_int_dialog_worldmap_sections" ).scrollpos;
    
    ObjSet( "scc_int_dialog_worldmap_sections", { scrollpos = scc_pos - 1 } );
    
  end;
  
end;
--------------------------------------------------------------------------------
function public.IsActive ()
  
  return private.is_active;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Featured_Show ()

  local layout = private.orientation.."_featured";
  local size   = private.scc_size[ layout ];

  ui.anim.Play(
    "anm_int_dialog_worldmap_window",
    private.orientation.."_featured_appear",
    function ()

      private.SetupScroller( size );

    end,
    private.orientation.."_featured"
  );

end;
--------------------------------------------------------------------------------
function public.Featured_Hide ()

  local layout = private.orientation;
  local size   = private.scc_size[ layout ];

  ui.anim.Play(
    "anm_int_dialog_worldmap_window",
    private.orientation.."_featured_disappear",
    function ()

      mediator.Notify( "Interface_Dialog_Worldmap_FeaturedHidden" );

    end,
    private.orientation
  );

  private.SetupScroller( size );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Focus ( id, count_section, is_featured, unlock )
  
  local scc = "scc_int_dialog_worldmap_sections";
  local scc_pos = id - 2;
  
  local visible_count = tern( is_featured )( 3, 4 );
  
  local min_pos = 0;
  local max_pos = count_section - visible_count;
  
  if ( unlock ) and ( _f.GetSharingWorld() ) then
    scc_pos = scc_pos + 1;
    max_pos = max_pos + 1;
  end;
  
  if     ( scc_pos < min_pos ) then scc_pos = min_pos;
  elseif ( scc_pos > max_pos ) then scc_pos = max_pos;
  end;
  
  ObjSet( scc, { scrollpos = scc_pos } );
  
end;
--------------------------------------------------------------------------------
function private.SetupScroller ( size, multiset )

  ui.obj.Set(
    "scc_int_dialog_worldmap_sections",
    {
      inputrect_init = true,
      inputrect_x   = - size.w / 2,
      inputrect_y   = - size.h / 2,
      inputrect_w   =   size.w,
      inputrect_h   =   size.h,

      croprect_init = true,
      croprect_x    = - size.w / 2,
      croprect_y    = - size.h / 2,
      croprect_w    =   size.w,
      croprect_h    =   size.h
    },
    multiset
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.CreateListItems ( aList )

  local scc    = "scc_int_dialog_worldmap_sections";
  local slot   = "anm_int_dialog_worldmap_slot";
  local button = "btn_int_dialog_worldmap_slot_play";

  local multiset = {};

  for id, section in pairs( aList ) do
    
    if ( not ui.obj.Get( slot.."_"..section.name ) ) then
      
      local item_button = button.."_"..section.name;
      
      ui.obj.Clone( slot, scc, "_"..section.name, 1 );
      
      ui.obj.Set(
        slot.."_"..section.name,
        {
          rt_back  = "assets/content/worlds/"..section.name,
          rt_red   = section.color.red,
          rt_green = section.color.green,
          rt_blue  = section.color.blue--,
          --rt_boss  = "assets/content/worlds/"..section.name.."_boss"
        },
        multiset
      );

      ui.button.Create(
        item_button,
        {
          title  = "str_int_dialog_worldmap_play",
          design = "medium",
          color  = "green",
          sound  = "simply_click"
        },
        function () 

          if ( _f.IsContentReady( section.name ) ) then

            _f.GoTo( tonumber( section.name:match( "%d+" ) ), public.Hide );

          end;
          
        end
      );
      
    end;

    if ( next( multiset ) ) then

      ui.obj.Multiset( multiset );

    end;
    
    ui.obj.Attach( slot.."_"..section.name, scc );

  end;

end;
--------------------------------------------------------------------------------
function private.UpdateListItems ( aList, sCurrentSection, tSkin )

  local slot   = "anm_int_dialog_worldmap_slot";
  local avatar = "anm_int_dialog_worldmap_marker"
  local str    = "str_int_dialog_worldmap_";

  local multiset = {};

  for id, section in pairs( aList ) do

    local item = slot .. "_" .. section.name;
    local obj_indicatior = "anm_int_dialog_worldmap_slot_notification_"..section.name;

    local params =
    { 
      animfunc = nil,
      frame    = 0, 
      playing  = true,
      island   = StringGet( str..section.name )
    };

    if ( section.name == sCurrentSection ) then
    
      params.animfunc = "island_current";
      params.stages   = private.GetSectionStages( section );
      
      local anm_avatar = avatar .. "_" .. section.name;
      local res = "assets/content/%s/anm_skin_%s"
      local ava_params =
      {
        res = res:format( tSkin.content, tSkin.name ),
        animfunc = "marker",
        playing  = true
      };
      table.insert( multiset, { anm_avatar, ava_params } );
    
    elseif ( private.IsSectionCompleted( id, aList ) ) then

      params.animfunc = "island_complete";
      params.stages   = private.GetSectionStages( section );

    elseif ( private.IsSectionUnlocked( section ) ) then

      params.animfunc = "island_unlock";
      params.frame    = 54;
      playing         = false; -- Static unlock state (last frame)
      
      params.stages   = private.GetSectionStages( section );

      local indicator = true;
      if ( private.IsSectionUnlockedAndStarted( section ) ) then
        indicator = false;
      end;
      table.insert( multiset, { obj_indicatior, { visible = indicator } } );
      
    else
      local prew = StringGet( str.."section_"..( tonumber( section.name:match( "%d+" ) ) - 1 ) );
      local text = StringGet( str.."description_locked", prew );

      params.animfunc  = "island_lock";
      params.lock_text = text;

      local indicator = false;
      if ( private.IsSectionLockedAndNew( section.name ) ) then
        indicator = true;
      end;
      table.insert( multiset, { obj_indicatior, { visible = indicator } } );
      
    end;

    table.insert( multiset, { item, params } );

  end;

  ObjMultiSet( multiset );

end;
--------------------------------------------------------------------------------
function private.InsertSlotShareToIndex ( id, sections_list )
  
  if ( not _f.GetSharingWorld() ) then return false; end;
  
  local start_index = math.min( id + 1, #sections_list );
  
  for i = id + 1, #sections_list do
    ne.obj.Detach( "anm_int_dialog_worldmap_slot_section_"..i );
  end;
  
  ne.obj.Attach( "anm_int_dialog_worldmap_slot_share", "scc_int_dialog_worldmap_sections" );
  
  for i = id + 1, #sections_list do
    ne.obj.Attach( "anm_int_dialog_worldmap_slot_section_"..i, "scc_int_dialog_worldmap_sections" );
  end;
  
  return true;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.IndexList ( aList )

  local map = {};

  for id, section in pairs( aList ) do

    map[ section.name ] = section;
    map[ section.name ].id = id;

  end;

  return map;
  
end;

--------------------------------------------------------------------------------
function private.IsSectionUnlocked ( section )

  return ( section and section.enabled );

end;
--------------------------------------------------------------------------------
function private.IsSectionUnlockedAndStarted ( section )

  local intro = false;
  
  if ( section.intro ) then
    if ( type( section.intro ) ~= "table" ) then
      intro = true;
    end;
  end;
  
  return ( section and section.enabled and intro );

end;
--------------------------------------------------------------------------------
function private.IsSectionCompleted ( id, aList )

  local nxt = id + 1;

  local section_next = aList[ nxt ];
  return ( section_next and section_next.enabled );

end;
--------------------------------------------------------------------------------
function private.IsSectionLockedAndNew ( name )
  
  local worlds = private.new_worlds;
  
  for i = 1, #worlds do
    if ( worlds[ i ] == name ) then
      return true;
    end;
  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
function private.GetSectionStages ( section )

  local pattern = "%s/%s";
  
  return pattern:format( section.stages_done, section.stages_count );

end;
--------------------------------------------------------------------------------