-- name=int_loading
--------------------------------------------------------------------------------
function public.__InitObjects () 

  private.app_w = ui.screen.Width();
  private.app_h = ui.screen.Height();

  local bg_height = 1136;
  local bg_scale  = math.max( 1, ( private.app_h / bg_height ) );

  local multiset = {};

  ui.obj.ToCenter( "obj_int_loading", 0, 0, multiset );
  
  ui.obj.Set(
    "obj_int_loading",
    {
      event_mup      = function () end,
      input          = true,
      inputrect_init = true,
      inputrect_x    = - private.app_w / 2,
      inputrect_y    = - private.app_h / 2,
      inputrect_w    =   private.app_w,
      inputrect_h    =   private.app_h
    },
    multiset
  );

  ui.obj.Set(
    "obj_int_loading_progressbar",
    { 
      pos_y = 0.5 * private.app_h - 50
    },
    multiset
  );

  ui.obj.Set(
    "obj_int_loading_hint",
    { 
      pos_y = 0.5 * private.app_h - 230
    },
    multiset
  );

  ui.obj.Set(
    "spr_int_loading_bg",
    {
      scale_x = bg_scale,
      scale_y = bg_scale
    },
    multiset
  );

  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.Show ( need_fade, func_continue, func_notify )

  if ( private.on_screen ) then

    need_fade = false;
    
  else
    
    private.Generate_Hint();

  end;

  func_continue = func_continue or function () end;

  private.on_screen = true;
  public.SetProgressBar( 0, 0, 0 );

  ObjAttach( "obj_int_loading", "ng_application" );

  ObjSet( "obj_int_loading", { active = true, visible = true, pos_z = 1001 } );

  ShowLoadingWheel( true );

  local func_next = function ()

    ObjSet( "obj_int_loading", { alp = 1.0 } );

    func_notify();
    func_continue();

  end;

  if ( not need_fade ) then

    func_next();

    return;

  end;

  ObjAnimate( "obj_int_loading", "alp", 0, 0, func_next,
  {
    0.0, 0, 0.0,
    0.3, 0, 1.0
  } );

end;
--------------------------------------------------------------------------------
function public.Hide ( need_fade, func_continue )

  func_continue = func_continue or function () end;
  
  if not ( private.on_screen ) then

    func_continue();
    return;

  end;

  local func_next = function ()

    ShowLoadingWheel( false );

    ObjSet( "obj_int_loading", { active = false, visible = false, alp = 0.0 } );
    ObjDetach( "obj_int_loading" );

    private.on_screen = false;

    func_continue();

  end;

  if ( not need_fade ) then

    func_next();

    return;

  end;

  ObjAnimate( "obj_int_loading", "alp", 0, 0, func_next,
  {
    0.0, 0, 1.0,
    0.3, 0, 0.0
  } );

end;
--------------------------------------------------------------------------------
-- Set progress bar to specified value in range from 'from_abs' to 'to_abs'
-- For example: public.SetProgressBar( 0.5, 0.8, 0.9 ) will set bar to 0.85
--------------------------------------------------------------------------------
function public.SetProgressBar ( relative, from_abs, to_abs )
  
  if ( not private.on_screen ) then return; end;
  
  local crop_current = ObjGet( "grd_int_loading_progress" ).croprect_w;
  
  local crop_new = 352 * ( from_abs + ( to_abs - from_abs ) * relative ) + 0.1;
  
  ObjAnimate( "grd_int_loading_progress", "croprect_w", 0, 0, "", 
  {
    0.000, 0, crop_current,
    0.017, 0, crop_new
  } );
  
end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function private.Generate_Hint ()
  
  local random_index = math.random( 1, 17 );
  local help_show = tern( random_index > 7 )( false, true );
  
  ObjMultiSet( {
    { "txt_int_loading_hint",    { visible = ( not help_show ), 
                               text = "str_loader_hint_"..random_index } },
    { "anm_int_loading_hint",    { visible = help_show,
                               animfunc = "help_"..random_index,
                               help_text = "str_loader_hint_"..random_index } }
  } );
  
end;
--------------------------------------------------------------------------------