-- name=int_osd_effects
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.this = {
    int     = "int_osd_effects",
    timer   = "tmr_int_osd_effects",
    effects = "obj_int_osd_effects"
  };

  private.count   = 0;
  private.queue   = {};
  private.process = false;

  mediator.Subscribe(
    "View_Map_MineGathered",
    private.MineGathered
  );

  mediator.Subscribe(
    "Interface_PanelStageProgress_HealthChanged", 
    private.HealthChanged
  );

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  ui.hub.Create( private.this.int );
  ui.hub.AttachContent( private.this.int );

end;

--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Clear ()
  
  common.Timer_Stop( private.this.timer );

  private.process = false;
  private.queue   = {};

  ui.obj.Delete( private.this.effects );
  ui.obj.Create( private.this.effects, nil, private.this.int );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.MineGathered ( params )

  if ( params.delta ~= 0 ) then

    local show_params =
    {
      text        = params.delta,
      posx        = params.posx, 
      posy        = params.posy,
      fontcolor_r = params.fontcolor_r, 
      fontcolor_g = params.fontcolor_g, 
      fontcolor_b = params.fontcolor_b
    };

    private.Show( show_params );  

  end;

end;
--------------------------------------------------------------------------------
function private.HealthChanged ( params )

  if ( params.delta ~= 0 ) then

    local show_params =
    {
      time        = params.time,
      offset      = params.offset,
      direction   = params.direction,
      text        = params.delta,
      posx        = params.posx, 
      posy        = params.posy,
      res         = "items/heart_small",
      fontcolor_r = params.fontcolor_r, 
      fontcolor_g = params.fontcolor_g, 
      fontcolor_b = params.fontcolor_b
    };

    private.Show( show_params );  

  end;

end;
--------------------------------------------------------------------------------
function private.Show ( params )

  local tPos = { iPosX = 512, iPosY = 384 };

  local default_fontcolor = { 1, 0.8, 0 };

  table.insert( private.queue, 
  { 
    time        = params.time or 3.0,
    offset      = params.offset or 150,
    direction   = params.direction or nil,
    text        = params.text, 
    posx        = params.posx or tPos.iPosX, 
    posy        = params.posy or tPos.iPosY,
    fontcolor_r = params.fontcolor_r or default_fontcolor[ 1 ], 
    fontcolor_g = params.fontcolor_g or default_fontcolor[ 2 ], 
    fontcolor_b = params.fontcolor_b or default_fontcolor[ 3 ], 
    res         = params.res or nil,
    animfunc    = params.animfunc or nil 
  } );

  if not ( private.process ) then

    private.ShowNext();

  end;

end;
--------------------------------------------------------------------------------
function private.ShowNext ()

  if ( #private.queue == 0 ) then

    private.process = false;

    return;

  else

    private.process = true;

  end;

  local params = private.queue[ 1 ];

  table.remove( private.queue, 1 );

  private.count = private.count + 1;

  local obj_name  = "int_osd_effects_default";

  ui.obj.Clone( obj_name, private.this.effects, "_"..private.count, 1 );

  obj_name = "int_osd_effects_default_"..private.count;

  ObjMultiSet( {
    { obj_name, { pos_x = params.posx, pos_y = params.posy } },
    { "txt_int_osd_effects_default_"..private.count, 
      { 
        text = params.text, 
        fontcolor_r = params.fontcolor_r, 
        fontcolor_g = params.fontcolor_g, 
        fontcolor_b = params.fontcolor_b
      } }
  } );

  if ( params.res ) then

    ObjSet( "spr_int_osd_effects_default_"..private.count, { active = true, visible = true, res = "assets/content/"..params.res } );

  elseif ( params.animfunc ) then

    ObjSet( "spr_int_osd_effects_default_"..private.count, { active = true, visible = true, animfunc = params.animfunc, playing = true } );

  else

    ObjSet( "spr_int_osd_effects_default_"..private.count, { active = false, visible = false } );
    ObjSet( "txt_int_osd_effects_default_"..private.count, { pos_x = - 50 } );

  end;

  local func = function ()

    ui.obj.Delete( obj_name );

  end;
  
  local tme = params.time;

  local factors = { up = - 1, down = 1 };

  if ( params.direction ) then

    local direction = factors[ params.direction ];

    ObjAnimate( obj_name, "pos_y", 0, 0, func,
    {
      ( 0.0 * tme ), 3, params.posy,
      ( 1.0 * tme ), 3, params.posy + direction * params.offset
    } );

  else

    local direction = - 1;

    if ( type( params.text ) == "number" ) then
      
      if ( params.text < 0 ) then

        direction = 1;

      else

        ObjSet( "txt_int_osd_effects_default_"..private.count, { text = "+"..params.text } );

      end;

    end;

    ObjAnimate( obj_name, "pos_y", 0, 0, func,
    {
      ( 0.0 * tme ), 3, params.posy,
      ( 1.0 * tme ), 3, params.posy + direction * params.offset
    } );

  end;

  ObjAnimate( obj_name, "alp", 0, 0, "",
  {
    ( 0.0 * tme ), 3, 0.0,
    ( 0.2 * tme ), 3, 1.0,
    ( 0.5 * tme ), 3, 1.0,
    ( 1.0 * tme ), 3, 0.0
  } );

  ObjAnimate( obj_name, "scale_xy", 0, 0, "",
  {
    ( 0.0 * tme ), 3, 0.6, 0.6,
    ( 0.2 * tme ), 3, 1.0, 1.0,
    ( 0.4 * tme ), 3, 1.0, 1.0,
    ( 1.0 * tme ), 3, 0.8, 0.8
  } );

  common.Timer_Start( private.this.timer, 0.35, private.ShowNext );

end;
--------------------------------------------------------------------------------