-- name=int_panel_interface_bottom
--------------------------------------------------------------------------------
local TAG = "int_panel_interface_bottom";
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  OptionsClickHandler  = function () end,
  PlayerClickHandler   = function () end,
  ShopClickHandler     = function () end,
  StickersClickHandler = function () end,
  WorldsClickHandler   = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.app_w = ui.screen.Width();
  private.app_h = ui.screen.Height() - ui.screen.BottomSafeArea();
  private.app_o = ui.screen.Orientation();

  private.buttons =
  {
    options =
    {
      order         = 1,
      enabled       = false,
      btn           = "btn_int_panel_interface_bottom_settings",
      anm           = "anm_int_panel_interface_bottom_settings",
      label         = ne.strings.Get( "str_int_panel_map_settings" ),
      animfunc      =
      {
        normal       = "btn_globalmap_option"
      },
      handler = function ()
        _f.OptionsClickHandler();
        end,
      offset   = - 0.5 * private.app_w + 5
    },
    player =
    {
      order         = 2,
      enabled       = false,
      notifications = {
        default = 0,
        custom = {
          anm  = "anm_int_panel_interface_bottom_player_notification",
          func = function ( res, count )
            local state = tern( count > 0 )( true, false );
            local set_table = { visible = state, input = state };
            return set_table;
          end
        }
      },
      btn           = "btn_int_panel_interface_bottom_player",
      anm           = "anm_int_panel_interface_bottom_player",
      label         = ne.strings.Get( "str_int_panel_map_hero" ),
      animfunc      =
      {
        normal       = "btn_player_normal_static",
        notify       = "btn_player_notification_static"
      },
      handler = function ()
        _f.PlayerClickHandler();
      end,
    },
    shop =
    {
      order         = 3,
      enabled       = false,
      notifications = {
        default = 0,
        custom = {
          anm  = "anm_int_panel_interface_bottom_shop",
          func = function ( res, count )
            local text = tern( count > 0 )( "str_int_sale_label", "" );
            local set_table = { rt_sale = res, tt_sale = text };
            return set_table;
          end
        }
      },
      btn           = "btn_int_panel_interface_bottom_shop",
      anm           = "anm_int_panel_interface_bottom_shop",
      label         = ne.strings.Get( "str_int_panel_map_shop" ),
      animfunc      =
      {
        normal       = "btn_shop_normal_static",
        notify       = "btn_shop_notification_static"
      },
      handler = function ()
        _f.ShopClickHandler();
      end
    },
    stickers =
    {
      order         = 4,
      enabled       = false,
      notifications = { default = 0, custom = false },
      btn           = "btn_int_panel_interface_bottom_stickers",
      anm           = "anm_int_panel_interface_bottom_stickers",
      label         = ne.strings.Get( "str_int_panel_map_fireflies" ),
      animfunc      =
      {
        normal       = "btn_stickers_normal_static",
        notify       = "btn_stickers_notification_static"
      },
      handler = function ()
        _f.StickersClickHandler();
      end
    },
    worlds =
    {
      order         = 5,
      enabled       = false,
      notifications = {
        default = 0,
        custom = {
          anm  = "anm_int_panel_interface_bottom_worldmap",
          func = function ( res, count, animfunc )

            local set_table = {
              animfunc = animfunc,
              rt_event = res,
              count    = tern( count > 0 )( "!", "" )
            };

            return set_table;

          end
        }
      },
      btn           = "btn_int_panel_interface_bottom_worldmap",
      anm           = "anm_int_panel_interface_bottom_worldmap",
      label         = ne.strings.Get( "str_int_panel_map_worldmap" ),
      animfunc      =
      {
        normal       = "btn_world_normal_static",
        notify       = "btn_world_notification_static"
      },
      handler = function ()
        _f.WorldsClickHandler();
      end
    }
  };

  private.this = {
    int = "int_panel_interface_bottom",
    obj = "obj_int_panel_interface_bottom",
    anm = "anm_int_panel_interface_bottom",
    dcr = {
      obj = "anm_int_panel_interface_bottom_decoration",
      res = "anm_decoration_bottom"
    }
  };

  private.animfunc = {
    btn1 = {
      appear    = private.app_o.."_1btn_appear",
      disappear = private.app_o.."_1btn_disappear",
      static    = private.app_o.."_1btn"
    },
    btn4 = {
      appear    = private.app_o.."_4btn_appear",
      disappear = private.app_o.."_4btn_disappear",
      static    = private.app_o.."_4btn"
    },
    btn5 = {
      appear    = private.app_o.."_5btn_appear",
      disappear = private.app_o.."_5btn_disappear",
      static    = private.app_o.."_5btn"
    }
  };

  private.hidden_frame = 10;
  private.active       = false;

  private.Layout_Set( "btn1" );

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  ui.hub.Create( private.this.int );

  local multiset = {};

  ui.obj.Set(
    private.this.obj,
    {
      pos_x = private.app_w /2,
      pos_y = private.app_h
    },
    multiset
  );

  ui.obj.Set(
    private.this.anm,
    {
      animfunc = private.animfunc[ private.layout ].disappear,
      frame    = 30
    },
    multiset
  );

  ui.obj.Multiset( multiset );

  ui.hub.AttachContent( private.this.int );

  for key, value in pairs( private.buttons ) do
    ui.control.EventsSet(
      value.btn,
      {
        func_on_click = function ()
          value.handler();
        end
      }
    );
  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show ( func_end )

  if ( private.active ) then
    if ( func_end ) then func_end(); end;
    return;
  end;

  private.active = true;
  private.off    = false;

  ui.anim.Play(
    private.this.anm,
    private.animfunc[ private.layout ].appear,
    function ()
      ui.input.On( private.this.obj );
      if ( func_end ) then func_end(); end;
    end,
    private.animfunc[ private.layout ].static
  );
  
  private.Decoration_Update();

end;
--------------------------------------------------------------------------------
function public.Hide ( func_end )

  if ( not private.active ) then

    return;

  end;

  private.active = false;

  ui.input.Off( private.this.obj );

  ui.anim.Play(
    private.this.anm,
    private.animfunc[ private.layout ].disappear,
    function ()

      if ( func_end ) then func_end(); end;
      private.off = true;

    end
  );

end;
--------------------------------------------------------------------------------
function public.Off ()

  if ( private.off ) then return; end;

  ui.obj.Set(
    private.this.anm,
    {
      animfunc = private.animfunc[ private.layout ].disappear,
      playing  = false,
      frame    = private.hidden_frame,
      endtrig  = function () end
    }
  );

  ui.input.Off( private.this.obj );

  private.active = false;
  private.off    = true;

end;
--------------------------------------------------------------------------------
function public.Input_On ()

  ui.input.On( private.this.obj );

end;
--------------------------------------------------------------------------------
function public.Input_Off ()

  ui.input.Off( private.this.obj );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Options_Enable ()

  private.Button_Enable( "options" );

end;
--------------------------------------------------------------------------------
function public.Options_Disable ()

  private.Button_Disable( "options" );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Player_Enable ()

  private.Button_Enable( "player" );

end;
--------------------------------------------------------------------------------
function public.Player_Disable ()

  private.Button_Disable( "player" );

end;
--------------------------------------------------------------------------------
function public.Player_SetNotification ( count )

  private.Button_SetCountNotification( "player", count );

end;
--------------------------------------------------------------------------------
function public.Player_SetCustomNotification ( params )

  private.Button_SetCustomNotification( "player", params );

end;
--------------------------------------------------------------------------------
function public.Player_GetButtonObjName ()

  return private.Button_GetObjectName( "player" );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Shop_Enable ()

  private.Button_Enable( "shop" );

end;
--------------------------------------------------------------------------------
function public.Shop_Disable ()

  private.Button_Disable( "shop" );

end;
--------------------------------------------------------------------------------
function public.Shop_SetNotification ( count )

  private.Button_SetCountNotification( "shop", count );

end;
--------------------------------------------------------------------------------
function public.Shop_SetCustomNotification ( params )

  private.Button_SetCustomNotification( "shop", params );

end;
--------------------------------------------------------------------------------
function public.Shop_GetButtonObjName ()

  return private.Button_GetObjectName( "shop" );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Stickers_Enable ()

  private.Button_Enable( "stickers" );

end;
--------------------------------------------------------------------------------
function public.Stickers_Disable ()

  private.Button_Disable( "stickers" );

end;
--------------------------------------------------------------------------------
function public.Stickers_SetNotification ( count )

  private.Button_SetCountNotification( "stickers", count );

end;
--------------------------------------------------------------------------------
function public.Stickers_GetButtonObjName ()

  return private.Button_GetObjectName( "stickers" );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Worlds_Enable ()

  private.Button_Enable( "worlds" );

end;
--------------------------------------------------------------------------------
function public.Worlds_Disable ()

  private.Button_Disable( "worlds" );

end;
--------------------------------------------------------------------------------
function public.Worlds_SetNotification ( count )

  private.Button_SetSimpleNotification( "worlds", ( count > 0 ) );

end;
--------------------------------------------------------------------------------
function public.Worlds_SetCustomNotification ( params )

  private.Button_SetCustomNotification( "worlds", params );

end;
--------------------------------------------------------------------------------
function public.Worlds_Warn ( count, icon_path )

  private.Button_SetCountNotification( "worlds", count );
  private.Button_SetOverlayAnimIcon( "worlds", icon_path );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Decoration_Enable ( entity_name )

  private.Decoration_Enable( entity_name );

end;
--------------------------------------------------------------------------------
function public.Decoration_Disable ()

  private.Decoration_Disable();

end;
--------------------------------------------------------------------------------
function private.Decoration_Enable ( entity_name )

  local decoration = private.this.dcr.obj;
  local resource   = private.this.dcr.res;
  
  private.this.dcr.enbl = true;
  
  ui.obj.Set( decoration, {
    playing = true,
    res = "assets/content/"..entity_name.."/"..resource,
    animfunc = private.animfunc[ private.layout ].static
  } );
  
  ui.obj.Enable( decoration, true );

end;
--------------------------------------------------------------------------------
function private.Decoration_Update ()
  
  local decoration = private.this.dcr.obj;

  local animfunc_cur = ui.obj.Get( decoration ).animfunc;
  local animfunc_new = private.animfunc[ private.layout ].static;
  
  if ( private.this.dcr.enbl and 
       animfunc_cur ~= animfunc_new ) then

    ui.obj.Set( decoration, {
      animfunc = animfunc_new
    } );
    
  end;

end;
--------------------------------------------------------------------------------
function private.Decoration_Disable ()

  local decoration = private.this.dcr.obj;
  
  private.this.dcr.enbl = nil;
  
  ui.obj.Set( decoration, {
    playing = false,
    animfunc = ""
  } );

  ui.obj.Disable( decoration );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Button_Enable ( button_name, func_action )

  local button = private.buttons[ button_name ];

  ui.obj.Set(
    button.btn,
    {
      pos_x = button.offset or 0
    }
  );
  
  local animfunc = ( button.notifications and 
                     button.notifications.default > 0 ) and 
                     button.animfunc.notify or button.animfunc.normal;

  ui.obj.Set(
    button.anm,
    {
      btn_name = button.label,
      animfunc = animfunc,
      playing  = true
    }
  );

  if ( button.enabled ) then
    return;
  end;

  button.enabled = true;

  ui.obj.Enable( button.btn );

  private.ReorderButtons();

end;
--------------------------------------------------------------------------------
function private.Button_Disable ( button_name )

  local button = private.buttons[ button_name ];

  if not ( button.enabled ) then

    return;

  end;

  button.enabled = false;

  ui.obj.Set(
    button.btn,
    {
      anim_tag = ""
    }
  );

  ui.obj.Disable( button.btn );

  private.ReorderButtons();

end;
--------------------------------------------------------------------------------
function private.Button_SetSimpleNotification ( button_name, state )

  local button = private.buttons[ button_name ];
  button.notifications.default = tern( state )( 1, 0 );

  if ( state ) then

    ui.obj.Set(
      button.anm,
      {
        animfunc = button.animfunc.notify,
        count    = "!",
        playing  = true
      }
    );

  else

    ui.anim.Play( button.anm, button.animfunc.normal );

  end;

end;
--------------------------------------------------------------------------------
function private.Button_SetCountNotification ( button_name, count )

  local button = private.buttons[ button_name ];
  button.notifications.default = count;

  if ( count > 0 ) then

    ui.obj.Set(
      button.anm,
      {
        animfunc = button.animfunc.notify,
        count    = tern( button_name ~= "worlds" )( tostring( count ), "!" ),
        playing  = true
      }
    );

  else

    ui.anim.Play( button.anm, button.animfunc.normal );

  end;

end;
--------------------------------------------------------------------------------
function private.Button_SetCustomNotification ( button_name, params )

  local button = private.buttons[ button_name ];

  if ( button.notifications and
       button.notifications.custom ) then

    local multiset = {};

    for key, value in pairs( params ) do

      local has_count = ( value.count and value.count > 0 );

      local animfunc = tern( has_count )(
        button.animfunc.notify,
        button.animfunc.normal
      );

      local res = tern( has_count )( value.icon_path, "" );
      local set = button.notifications.custom.func(
        res,
        value.count,
        animfunc
      );

      ui.obj.Set( button.notifications.custom.anm, set, multiset );

    end;

    ui.obj.Multiset( multiset );

  end;

end;
--------------------------------------------------------------------------------
function private.Button_GetObjectName ( button_name )

  return private.buttons[ button_name ].btn;

end;
--------------------------------------------------------------------------------
function private.Button_SetOverlayAnimIcon ( button_name, icon_path )

  icon_path = icon_path or "";
    
  local base_name = "int_panel_interface_bottom_"..button_name.."_overlay";
  local icon      = ui.obj.Get( "anm_"..base_name );

  if ( icon_path == "" ) then
    if ( icon ) then ui.obj.Delete( icon.name ); end;
    return;
  end;

  local button = private.buttons[ button_name ];

  if ( not icon ) then
    local properties = {
      res      = icon_path,
      anim_tag = "at_overlay_icon",
      animfunc = "static",
      playing  = true
    };
    ui.obj.CreateAnimation( base_name, button.anm, properties );
    return;
  end;

  if ( icon.res ~= icon_path ) then
    ui.obj.Set( icon.name, { res = icon_path } );
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.ReorderButtons ()

  local enabled_list = {};
  for id, button in pairs( private.buttons ) do

    if ( button.enabled ) then table.insert( enabled_list, id ); end;

  end;

  if ( #enabled_list > 0 ) then

    table.sort( enabled_list, private.Order );

    local multiset = {};

    for i, button in pairs( enabled_list ) do

      ui.obj.Set(
        private.buttons[ button ].btn,
        {
          anim_tag = "at_btn_"..i
        },
        multiset
      );

    end;

    ui.obj.Multiset( multiset );

    private.Layout_Set( "btn"..#enabled_list );

  end;

end;
--------------------------------------------------------------------------------
function private.Order ( btn_one, btn_two )

  return private.buttons[ btn_one ].order < private.buttons[ btn_two ].order;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Layout_Set ( layout )

  if ( private.animfunc[ layout ] ) then
    private.layout = layout;
  else
    -- Log.e( TAG, "layout not found: "..layout );
  end;

end;
--------------------------------------------------------------------------------
