-- name=int_panel_interface_left
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  HeartsEventTimeRemaining = function () end,
  HeartsIsImmortality      = function () end,
  HeartsGetCount           = function () end,
  HeartsDialogMain         = function () end,
  Arena_OnTap              = function () end,
  Minievent_OnTap          = function () end,
  Countdown_GetSeconds     = function () end,
  Countdown_GetLabel       = function () end,
  Countdown_GetHint        = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  local base_name = "int_panel_interface_left";

  private.this = {
    int  = base_name,
    obj  = "obj_"..base_name,
    anm  = "anm_"..base_name,
    grid = "anm_"..base_name.."_grid",
    dcr = {
      obj = "anm_"..base_name.."_decoration",
      res = "anm_decoration_left"
    }
  };

  private.animfunc = {
    appear         = "panel_appear",
    static         = "panel_static",
    disappear      = "panel_disappear",
    grid_normal    = "static_buttons",
    grid_highlight = "static_button_highlight"
  };

  private.hidden_frame = 12;
  
  private.buttons =
  {
    hearts =
    {
      order    = 1,
      enabled  = false,
      btn      = "btn_"..base_name.."_hearts",
      anm      = "anm_"..base_name.."_hearts",
      label    = ne.strings.Get( "str_int_panel_map_hearts_label" ),
      animfunc =
      {
        normal = "btn_hearts_static"
      },
      timer         =
      {
        enabled      = false,
        name         = base_name.."_hearts",
        func_seconds = nil
      }
    },
    dailyquest = 
    {
      order         = 2,
      enabled       = false,
      notifications = 0,
      btn           = "btn_"..base_name.."_dailyquest",
      anm           = "anm_"..base_name.."_dailyquest",
      label         = ne.strings.Get( "str_int_panel_map_dailyquest_label" ),
      animfunc      =
      {
        normal = "btn_daily_quest_normal_static",
        notify = "btn_daily_quest_notification_static"
      },
      timer         =
      {
        enabled      = false,
        name         = base_name.."_dailyquest",
        func_seconds = nil
      }
    },
    minievent = 
    {
      order         = 3,
      enabled       = false,
      notifications = 0,
      btn           = "btn_"..base_name.."_minievent",
      anm           = "anm_"..base_name.."_minievent",
      icon          = "anm_"..base_name.."_minievent_icon",
      label         = ne.strings.Get( "str_"..base_name.."_minievent" ),
      animfunc      =
      {
        normal = "minievent_panel_static"
      },
      timer         =
      {
        enabled      = false,
        name         = base_name.."_minievent",
        func_seconds = nil
      }
    },
    countdown =
    {
      order    = 5,
      enabled  = false,
      btn      = "btn_"..base_name.."_countdown",
      anm      = "anm_"..base_name.."_countdown",
      label    = "",
      animfunc =
      {
        normal = "countdown_static",
        notify = "countdown_static"
      },
      timer         =
      {
        enabled      = false,
        name         = base_name.."_countdown",
        func_seconds = _f.Countdown_GetSeconds
      }
    }
  };

  private.dailyquest_notification = false;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  ui.hub.Create( private.this.int );
  
  ui.obj.Multiset(
    {
      { 
        private.this.obj,
        {
          input = false,
          pos_x = private.app_w,
          pos_y = ui.screen.TopSafeArea()
        }
      },
      {
        private.this.grid,
        {
          animfunc = private.animfunc.grid_normal,
          playing  = true
        }
      }
    }
  );
  
  for _, button in pairs( private.buttons ) do
  
    if not ( button.enabled ) then
      
      ui.obj.Set(
        button.btn, 
        { 
          anim_tag = "",
          active   = false,
          input    = false,
          visible  = false
        }
      );
      
    end;
  
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Show ( highlight_button )
  
  if ( private.active ) then return; end;
  
  private.active = true;
  private.off    = false;  

  ui.hub.AttachContent( private.this.int );
  
  if ( highlight_button ) then
  
    private.highlighted_button = highlight_button;
    private.Button_EnableHighlight( private.highlighted_button );
  
  end;
  
  ui.anim.Play(
    private.this.anm,
    private.animfunc.appear,
    function ()
      
      private.TimersShow();
      ui.input.On( private.this.obj );

    end,
    private.animfunc.static
  );

  private.Decoration_Update();

end;
--------------------------------------------------------------------------------
function public.Hide ( func_end )

  if ( not private.active ) then return; end;

  private.active = false;

  private.Minievent_CloseTray();
  
  ui.input.Off( private.this.obj );
  
  private.TimersHide( true );

  ui.anim.Play(
    private.this.anm,
    private.animfunc.disappear,
    function ()

      private.off = true;
      if ( func_end ) then func_end(); end;

      ui.hub.DetachContent( private.this.int );
      
      if ( private.highlighted_button ) then
      
        private.ReorderButtons();
        private.highlighted_button = nil;
        
      end;

    end
  );

end;
--------------------------------------------------------------------------------
function public.Off ()

  if ( private.off ) then return; end;

  private.active = false;

  private.Minievent_ForceCloseTray();

  ui.input.Off( private.this.obj );
  
  private.TimersHide( false );

  ui.obj.Set(
    private.this.anm,
    {
      animfunc = private.animfunc.disappear,
      frame    = private.hidden_frame,
      endtrig  = function () end
    }
  );

  ui.hub.DetachContent( private.this.int );

  private.off = true;
  
  if ( private.highlighted_button ) then
  
    private.ReorderButtons();
    private.highlighted_button = nil;
    
  end;
  
end;
--------------------------------------------------------------------------------
function public.Input_On ()
  
  ui.input.On( private.this.obj );
  
end;
--------------------------------------------------------------------------------
function public.Input_Off ()
  
  ui.input.Off( private.this.obj );
  
end;
--------------------------------------------------------------------------------
function public.IsActive ()

  return private.active;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Hearts_Enable ()

  private.Button_Enable( "hearts", _f.HeartsDialogMain );

end;
--------------------------------------------------------------------------------
function public.Hearts_Disable ()

  private.Button_Disable( "hearts" );

end;
--------------------------------------------------------------------------------
function public.Hearts_SetCount ()

  local count = _f.HeartsGetCount();
  local is_immortality = _f.HeartsIsImmortality();
  
  if ( is_immortality ) then 
    count = ui.text.InfinitySign();
  else 
    count = ui.text.FormatNumber( count );
  end;

  ui.obj.Set( private.buttons[ "hearts" ].anm, { count = count } );

end;
--------------------------------------------------------------------------------
function public.Hearts_EnableTimer ()

  private.Button_EnableTimer( "hearts", _f.HeartsEventTimeRemaining );

end;
--------------------------------------------------------------------------------
function public.Hearts_DisableTimer ()

  private.Button_DisableTimer( "hearts" );

end;
--------------------------------------------------------------------------------
function public.Hearts_GetButtonObjName ()

  return private.Button_GetObjectName( "hearts" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Dailyquest_Enable ()

  private.Button_Enable( "dailyquest", _f.Arena_OnTap );
 
end;
--------------------------------------------------------------------------------
function public.Dailyquest_Disable ()

  private.Button_Disable( "dailyquest" );

end;
--------------------------------------------------------------------------------
function public.Dailyquest_EnableTimer ( func_seconds )

  private.Button_EnableTimer( "dailyquest", func_seconds );

end;
--------------------------------------------------------------------------------
function public.Dailyquest_SetNotifications( count )

  private.Button_SetNotifications( "dailyquest", count );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Minievent_Enable ()

  private.Button_Enable( "minievent", _f.Minievent_OnTap );

end;
--------------------------------------------------------------------------------
function public.Minievent_Disable ()

  private.Minievent_ForceCloseTray();
  private.Button_Disable( "minievent" );

end;
--------------------------------------------------------------------------------
function public.Minievent_EnableSlot (
    slot_index,
    icon_name,
    str_title,
    func_tap,
    func_seconds
  )

  local slot_obj = private.buttons.minievent.anm.."_"..slot_index;
  local icon_obj = private.buttons.minievent.icon.."_"..slot_index;
  local tmr_obj  = "int_panel_interface_left_minievent_"..slot_index;

  local multiset = {};

  ui.obj.Set(
    slot_obj,
    {
      active   = true,
      visible  = true,
      animfunc = "btn_minievent_static",
      btn_name = private.buttons.minievent.label --ne.strings.Get( str_title )
    },
    multiset
  );

  ui.obj.Set(
    icon_obj,
    {
      animfunc = "event_icon_"..icon_name
    },
    multiset
  );

  ui.control.EventsSet(
    slot_obj,
    {
      func_on_click = func_tap
    },
    multiset
  );

  local custom_obj = {
    anm_timer = slot_obj,
    text_tag  = "tt_timer"
  };

  ui.obj.SetTextTag(
    custom_obj.anm_timer,
    custom_obj.text_tag,
    ui.text.FormatTime( func_seconds() ),
    multiset
  );

  ui.obj.Multiset( multiset ); 

  local function func_tick ()

    ui.timer.Show(
      tmr_obj,
      func_seconds(),
      func_tick,
      custom_obj
    );
  
  end;

  func_tick(); 

end;
--------------------------------------------------------------------------------
function public.Minievent_DisableSlot ( slot_index )

  local slot_obj  = private.buttons.minievent.anm.."_"..slot_index;
  local is_active = ui.obj.Get( slot_obj ).active;

  if ( is_active ) then

    ui.control.EventsClear( slot_obj );
    ui.input.Off( slot_obj );

    ui.anim.FadeOut(
      0.3, slot_obj, function ()

        ui.obj.Set(
          slot_obj,
          {
            active  = false,
            visible = false,
            input   = false,
            alp     = 1
          }
        );

      end
    );

  end;

end;
--------------------------------------------------------------------------------
function public.Minievent_EnableTimer ( slot_index, func_seconds )

  local custom_obj = {
    anm_timer = "anm_int_panel_interface_left_minievent"..slot_index,
    text_tag  = "tt_timer"
  };

  ui.obj.SetTextTag(
    custom_obj.anm_timer,
    custom_obj.text_tag,
    ui.text.FormatTime( func_seconds() )
  );

  private.Button_EnableTimer( "minievent", func_seconds, custom_obj );

end;
--------------------------------------------------------------------------------
function public.Minievent_ToggleTray ()

  if ( not private.minievent_tray_expanded ) then

    private.minievent_tray_expanded        = true;
    private.buttons.minieventevents_backup =
      ui.control.EventsGet( private.buttons.minievent.btn );

    ui.control.EventsClear( private.buttons.minievent.btn );

    ui.anim.Play(
      private.buttons.minievent.anm,
      "minievent_panel_appear",
      function ()

        ui.input.On({
          private.buttons.minievent.anm.."_1",
          private.buttons.minievent.anm.."_2"
        });

      end
    );

    common.Defer( "minievent_tray_timeout", 3, private.Minievent_CloseTray );

  end;

end;
--------------------------------------------------------------------------------
function private.Minievent_CloseTray ()

  if ( private.minievent_tray_expanded ) then

    ui.input.Off({
      private.buttons.minievent.anm.."_1",
      private.buttons.minievent.anm.."_2"
    });

    ui.anim.Play(
      private.buttons.minievent.anm,
      "minievent_panel_disappear",
      function ()

        private.minievent_tray_expanded = nil;

        ui.control.EventsSet(
          private.buttons.minievent.btn,
          private.buttons.minieventevents_backup
        );

      end,
      "minievent_panel_static"
    );

  end;

end;
--------------------------------------------------------------------------------
function private.Minievent_ForceCloseTray ()

  if ( private.minievent_tray_expanded ) then

    ui.input.Off({
      private.buttons.minievent.anm.."_1",
      private.buttons.minievent.anm.."_2"
    });

    private.minievent_tray_expanded = nil;

    ui.control.EventsSet(
      private.buttons.minievent.btn,
      private.buttons.minieventevents_backup
    );

    ui.anim.Play( private.buttons.minievent.anm, "minievent_panel_static" );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Countdown_Enable ()

  private.buttons[ "countdown" ].label = _f.Countdown_GetLabel() or "";

  local custom_obj = {
    anm_timer = "anm_int_panel_interface_left_countdown",
    text_tag  = "tt_timer"
  };

  ui.obj.SetTextTag(
    custom_obj.anm_timer,
    custom_obj.text_tag,
    ui.text.FormatTime( _f.Countdown_GetSeconds() )
  );

  local function func_tap ()

    local hint = _f.Countdown_GetHint();

    if ( hint ) then

      ui.popup.Flat_Show(
        hint.caption,
        hint.text,
        private.buttons[ "countdown" ].btn,
        { 0, -130 },
        true,
        true
      );

    else

      DbgTrace( "[ int_panel_interface ] Countdown_Enable: no hint data" );

    end;

  end;

  private.Button_Enable( "countdown", func_tap );
  private.Button_EnableTimer(
    "countdown",
    _f.Countdown_GetSeconds,
    custom_obj
  );

end;
--------------------------------------------------------------------------------
function public.Countdown_Disable ()

  private.buttons[ "countdown" ].label = "";

  private.Button_DisableTimer( "countdown" );
  private.Button_Disable( "countdown" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Decoration_Enable ( entity_name )

  private.Decoration_Enable( entity_name );

end;
--------------------------------------------------------------------------------
function public.Decoration_Disable ()

  private.Decoration_Disable();

end;
--------------------------------------------------------------------------------
function private.Decoration_Enable ( entity_name )

  local decoration = private.this.dcr.obj;
  local resource   = private.this.dcr.res;
  
  private.this.dcr.enbl = true;
  
  ui.obj.Set( decoration, {
    playing = true,
    res = "assets/content/"..entity_name.."/"..resource,
    animfunc = private.animfunc.static
  } );
  
  ui.obj.Enable( decoration, true );

end;
--------------------------------------------------------------------------------
function private.Decoration_Update ()
  
  local decoration = private.this.dcr.obj;

  local animfunc_cur = ui.obj.Get( decoration ).animfunc;
  local animfunc_new = private.animfunc.static;
  
  if ( private.this.dcr.enbl and 
       animfunc_cur ~= animfunc_new ) then

    ui.obj.Set( decoration, {
      animfunc = animfunc_new
    } );
    
  end;

end;
--------------------------------------------------------------------------------
function private.Decoration_Disable ()

  local decoration = private.this.dcr.obj;
  
  private.this.dcr.enbl = nil;
  
  ui.obj.Set( decoration, {
    playing = false,
    animfunc = ""
  } );

  ui.obj.Disable( decoration );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Button_Enable ( button_name, func_action )

  local button = private.buttons[ button_name ];

  --DbgTrace( button );
  
  ui.obj.Set(
    button.anm,
    {
      btn_name = button.label,
      animfunc = button.animfunc.normal,
      playing  = true
    }
  );

  if ( button.enabled ) then return; end;
  
  button.enabled = true;
  
  if ( func_action ) then
    
    ui.control.EventsSet(
      button.btn,
      {
        func_on_click = function ( ... )

          --mediator.Notify( "Interface_PanelButton_Clicked" );
          func_action( ... );

        end
      }
    );
    
  end;

  ui.obj.Enable( button.btn );
  
  if ( button.notifications and
       button.notifications > 0 ) then
    
    private.Button_SetNotifications( button_name, button.notifications );
       
  end;
  
  private.ReorderButtons();

end;
--------------------------------------------------------------------------------
function private.Button_Disable ( button_name )

  local button = private.buttons[ button_name ];

  if not ( button.enabled ) then return; end;
  
  button.enabled = false;
  
  ui.control.EventsClear( button.btn );
  
  ui.obj.Set(
    button.btn,
    { 
      anim_tag = "",
      active   = false,
      input    = false,
      visible  = false
    }
  );
  
  
  private.ReorderButtons();

end;
--------------------------------------------------------------------------------
function private.Button_UpdateIcon ( button_name, icon_name, icon_index )

  icon_index = icon_index or 1;

  local anm_icon = private.buttons[ button_name ].icon;

  -- < FIX IT.
  --icon_name = icon_name:match( "(%a+)[_%a+]?" );
  --if ( icon_name == "tower" ) then
  --  icon_name = "treasury";
  --end;
  -- >

  if ( not anm_icon ) then

    DbgTrace(
      "[ ERROR ] [ int_panel_interface_left ] Failed to set icon <"..icon_name..
      "> for button <"..button_name.."> - no icon object found."
    );

    return;

  end;

  ui.anim.Play( anm_icon, "event_icon_"..icon_name );
  
end;
--------------------------------------------------------------------------------
function private.Button_GetObjectName ( button_name )

  return private.buttons[ button_name ].btn;

end;
--------------------------------------------------------------------------------
function private.Button_SetNotifications ( button_name, count )

  local button = private.buttons[ button_name ];
  
  button.notifications = count;

  if ( count > 0 ) then

    ui.obj.Set( button.anm,
    { 
      animfunc = button.animfunc.notify,
      count    = tern( button_name ~= "invite" )( tostring( count ), "!" ),
      playing  = true
    } );

  else
  
    ui.anim.Play( button.anm, button.animfunc.normal );

  end;

end;
--------------------------------------------------------------------------------
function private.Button_EnableHighlight ( button_name )

  ui.obj.Set(
    private.this.grid,
    {
      animfunc = private.animfunc.grid_highlight,
      playing  = true
    }
  );
  
  ui.obj.Set(
    private.buttons[ button_name ].btn,
    {
      anim_tag = "at_btn_highlighted"
    }
  );

end;
--------------------------------------------------------------------------------
function private.Button_EnableTimer (
    button_name,
    func_seconds,
    custom_obj
  )
  
  local timer = private.buttons[ button_name ].timer;
  
  timer.enabled      = true;
  timer.func_seconds = func_seconds;
  timer.custom_obj   = custom_obj;
  
  if ( private.active ) then
    private.Button_ShowTimer( button_name );
  end;

end;
--------------------------------------------------------------------------------
function private.Button_DisableTimer ( button_name )
  
  if ( private.buttons[ button_name ].timer.enabled ) then
  
    private.buttons[ button_name ].timer.enabled = false;
    private.Button_HideTimer( button_name, false );
  
  end;

end;
--------------------------------------------------------------------------------
function private.Button_ShowTimer ( button_name )

  if not ( private.active ) then return; end;

  local timer = private.buttons[ button_name ].timer;

  local function func_tick ()
    --DbgTrace( "[ int_panel_interface_left ] "..timer.name.." tick" );
    ui.timer.Show(
      timer.name,
      timer.func_seconds(),
      func_tick,
      timer.custom_obj
    );
  end;

  func_tick();

end;
--------------------------------------------------------------------------------
function private.Button_HideTimer ( button_name, animated )

  ui.timer.Hide(
    private.buttons[ button_name ].timer.name,
    animated,
    private.buttons[ button_name ].timer.custom_obj
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.TimersShow ()

  for button_name, button in pairs( private.buttons ) do
  
    if ( button.enabled and button.timer and button.timer.enabled ) then
    
      private.Button_ShowTimer( button_name );
    
    end;
  
  end;

end;
--------------------------------------------------------------------------------
function private.TimersHide ( animated )

  for button_name, button in pairs( private.buttons ) do
  
    if ( button.enabled and button.timer and button.timer.enabled ) then
    
      private.Button_HideTimer( button_name, animated );
    
    end;
  
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.ReorderButtons ()

  local enabled_list = {};
  for id, button in pairs( private.buttons ) do
  
    if ( button.enabled ) then table.insert( enabled_list, id ); end;
  
  end;
  
  if ( #enabled_list > 0 ) then
  
    table.sort( enabled_list, private.Order );
  
    local multiset = {}
    for i, button in pairs( enabled_list ) do
    
      ui.obj.Set(
        private.buttons[ button ].btn, 
        { 
          anim_tag = "at_btn_"..i
        },
        multiset
      );
    
    end;
    
    ui.obj.Set(
      private.this.grid,
      {
        animfunc = private.animfunc.grid_normal,
        frame    = 0,
        playing  = true
      },
      multiset
    );
    
    ui.obj.Multiset( multiset );
  
  end;

end;
--------------------------------------------------------------------------------
function private.Order ( btn_one, btn_two )

  return private.buttons[ btn_one ].order < private.buttons[ btn_two ].order;

end;
--------------------------------------------------------------------------------
