-- name=int_panel_interface_right
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Sale_OnTap   = function () end,
  Sale_GetTime = function () end,
  Google_OnTap = function () end,
  Mail_OnTap   = function () end,
  Social_OnTap = function () end,
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.app_w = ui.screen.Width();

  private.this = {
    int  = "int_panel_interface_right",
    obj  = "obj_int_panel_interface_right",
    anm  = "anm_int_panel_interface_right",
    grid = "anm_int_panel_interface_right_grid",
    dcr = {
      obj = "anm_int_panel_interface_right_decoration",
      res = "anm_decoration_right"
    }
  };

  private.active  = false;
  private.visible = false;

  private.animfunc = {
    appear    = "panel_appear",
    static    = "panel_static",
    disappear = "panel_disappear"
  };

  private.hidden_frame = 12;
  
  local base_name = "int_panel_interface_right";

  private.buttons =
  {
    mail = 
    {
      order         = 1,
      enabled       = false,
      notifications = 0,
      btn           = "btn_"..base_name.."_mail",
      anm           = "anm_"..base_name.."_mail",
      label         = ne.strings.Get( "str_int_panel_map_mail_label" ),
      animfunc      =
      {
        normal = tern( IsPC() )( "btn_mail_static_landscape", "btn_mail_static" ),
        notify = tern( IsPC() )( "btn_mail_notification_static_landscape", "btn_mail_notification_static" )
      }
    },
    google =
    {
      order         = 3,
      enabled       = false,
      btn           = "btn_int_panel_interface_right_google",
      anm           = "anm_int_panel_interface_right_google",
      label         = ne.strings.Get( "str_int_panel_map_google_label" ),
      animfunc      =
      {
        normal       = "btn_google_normal_static",
        normal_click = "btn_google_normal_click"
      }
    },
    social =
    {
      order    = 2,
      enabled  = false,
      btn      = "btn_"..base_name.."_social",
      anm      = "anm_"..base_name.."_social",
      label    = ne.strings.Get( "str_int_panel_map_social_label" ),
      animfunc =
      {
        -- DG: always show notification sign
        normal = "btn_social_notification_static"
      }
    },
    sale = 
    {
      order         = 4,
      enabled       = false,
      btn           = "btn_int_panel_interface_right_sale",
      anm           = "anm_int_panel_interface_right_sale",
      label         = ne.strings.Get( "str_int_panel_map_sale_label" ),
      animfunc      =
      {
        normal       = "btn_sale_normal_static",
        normal_click = "btn_sale_normal_click"
      },
      timer         =
      {
        enabled      = false,
        name         = "int_panel_interface_right_sale",
        func_seconds = nil
      }
    },
    ccs_dummy =
    {
      order         = 999,
      enabled       = false,
      btn           = "dmy_int_panel_interface_right_ccs"
    }
  };
  
end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  ui.hub.Create( private.this.int );
  
  ui.obj.Multiset(
    {
      { 
        private.this.obj,
        {
          input = false,
          pos_x = private.app_w,
          pos_y = ui.screen.TopSafeArea()
        }
      },
      {
        private.this.grid,
        {
          animfunc = "static_buttons",
          playing  = true
        }
      }
    }
  );
  
  for _, button in pairs( private.buttons ) do
  
    if not ( button.enabled ) then
      
      ui.obj.Set(
        button.btn, 
        { 
          anim_tag = "",
          active  = false,
          input   = false,
          visible = false
        }
      );
      
    end;
  
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Show ()

  if ( private.active ) then return; end;
  
  private.active = true;
  private.off    = false;  

  ui.hub.AttachContent( private.this.int );

  ui.anim.Play(
    private.this.anm,
    private.animfunc.appear,
    function ()
      
      private.TimersShow();
      
      private.visible = true;
      mediator.Notify( "Interface_RightPanel_Visible", private.buttons );
      
      ui.input.On( private.this.obj );

    end,
    private.animfunc.static
  );

  private.Decoration_Update();

end;
--------------------------------------------------------------------------------
function public.Hide ( func_end )

  if ( not private.active ) then return; end;
  private.active = false;
  
  ui.input.Off( private.this.obj );
  
  private.TimersHide( true );

  ui.anim.Play(
    private.this.anm,
    private.animfunc.disappear,
    function ()

      private.off     = true;
      if ( func_end ) then func_end(); end;

      ui.hub.DetachContent( private.this.int );

    end
  );
  
  private.visible = false;
  mediator.Notify( "Interface_RightPanel_Invisible", private.buttons );
  
end;
--------------------------------------------------------------------------------
function public.Off ()

  if ( private.off ) then return; end;

  private.active = false;

  ui.input.Off( private.this.obj );
  
  private.TimersHide( false );

  ui.obj.Set(
    private.this.anm,
    {
      animfunc = private.animfunc.disappear,
      frame    = private.hidden_frame,
      endtrig  = function () end
    }
  );

  ui.hub.DetachContent( private.this.int );

  private.off = true;
  
  private.visible = false;
  mediator.Notify( "Interface_RightPanel_Invisible" );

end;
--------------------------------------------------------------------------------
function public.Input_On ()
  
  ui.input.On( private.this.obj );
  
end;
--------------------------------------------------------------------------------
function public.Input_Off ()
  
  ui.input.Off( private.this.obj );
  
end;
--------------------------------------------------------------------------------
function public.IsActive ()

  return private.active;

end;
--------------------------------------------------------------------------------
function public.IsVisible ()

  return private.visible;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Sale_Enable ()

  private.Button_Enable( "sale", _f.Sale_OnTap );
  private.EnableButtonTimer( "sale", _f.Sale_GetTime );

end;
--------------------------------------------------------------------------------
function public.Sale_Warn ( count )

  private.Button_SetNotifications( "sale", count );

end;
--------------------------------------------------------------------------------
function public.Sale_Disable ()
  
  private.Button_Disable( "sale" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Google_Enable ()

  private.Button_Enable( "google", _f.Google_OnTap );

end;
--------------------------------------------------------------------------------
function public.Google_Disable ()

  private.Button_Disable( "google" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.CCS_Enable ()

  private.Button_Enable( "ccs_dummy" );

end;
--------------------------------------------------------------------------------
function public.CCS_Disable ()

  private.Button_Disable( "ccs_dummy" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Mail_Enable ( )

  private.Button_Enable( "mail", _f.Mail_OnTap );

end;
--------------------------------------------------------------------------------
function public.Mail_Disable ()

  private.Button_Disable( "mail" );

end;
--------------------------------------------------------------------------------
function public.Mail_SetNotifications( count )

  private.Button_SetNotifications( "mail", count );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Social_Enable ( button_type )

  private.Button_Enable( button_type, _f.Social_OnTap );

end;
--------------------------------------------------------------------------------
function public.Social_Disable ( button_type )

  private.Button_Disable( button_type );

end;
--------------------------------------------------------------------------------
function public.Social_SetNotifications( button_type, count )

  private.Button_SetNotifications( button_type, count );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Decoration_Enable ( entity_name )

  private.Decoration_Enable( entity_name );

end;
--------------------------------------------------------------------------------
function public.Decoration_Disable ()

  private.Decoration_Disable();

end;
--------------------------------------------------------------------------------
function private.Decoration_Enable ( entity_name )

  local decoration = private.this.dcr.obj;
  local resource   = private.this.dcr.res;
  
  private.this.dcr.enbl = true;
  
  ui.obj.Set( decoration, {
    playing = true,
    res = "assets/content/"..entity_name.."/"..resource,
    animfunc = private.animfunc.static
  } );
  
  ui.obj.Enable( decoration, true );

end;
--------------------------------------------------------------------------------
function private.Decoration_Update ()
  
  local decoration = private.this.dcr.obj;

  local animfunc_cur = ui.obj.Get( decoration ).animfunc;
  local animfunc_new = private.animfunc.static;
  
  if ( private.this.dcr.enbl and 
       animfunc_cur ~= animfunc_new ) then

    ui.obj.Set( decoration, {
      animfunc = animfunc_new
    } );
    
  end;

end;
--------------------------------------------------------------------------------
function private.Decoration_Disable ()

  local decoration = private.this.dcr.obj;
  
  private.this.dcr.enbl = nil;
  
  ui.obj.Set( decoration, {
    playing = false,
    animfunc = ""
  } );

  ui.obj.Disable( decoration );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Button_Enable ( button_name, func_action )

  local button = private.buttons[ button_name ];
  
  if ( button.anm ) then

    ui.obj.Set(
      button.anm,
      {
        btn_name = button.label,
        animfunc = button.animfunc.normal,
        playing  = true
      }
    );
    
  end;

  if ( button.enabled ) then

    return;
    
  end;
  
  button.enabled = true;
  
  if ( func_action ) then
    
    ui.control.EventsSet(
      button.btn,
      {
        func_on_click = function ( ... )

          --mediator.Notify( "Interface_PanelButton_Clicked" );
          func_action( ... );

        end
      }
    );
    
  end;

  ui.obj.Enable( button.btn );
 
  if ( button.notifications and
       button.notifications > 0 ) then
    
    private.Button_SetNotifications( button_name, button.notifications );
       
  end;
  
  private.ReorderButtons();
  
  mediator.Notify(
    "Interface_RightPanel_Updated",
    {
      name    = button_name,
      action  = "enabled",
      active  = private.active,
      visible = private.visible
    }
  );

end;
--------------------------------------------------------------------------------
function private.Button_Disable ( button_name )

  local button = private.buttons[ button_name ];

  if not ( button.enabled ) then

    return;
    
  end;
  
  button.enabled = false;
  
  ui.control.EventsClear( button.btn );
  
  ui.obj.Set(
    button.btn,
    { 
      anim_tag = "",
      active  = false,
      input   = false,
      visible = false
    }
  );
  
  if ( button.timer and button.timer.enabled ) then
    private.DisableButtonTimer( button_name );
  end;
  
  private.ReorderButtons();
  
  mediator.Notify(
    "Interface_RightPanel_Updated",
    {
      name    = button_name,
      action  = "disabled",
      active  = private.active,
      visible = private.visible
    }
  );

end;
--------------------------------------------------------------------------------
function private.Button_GetObjectName ( button_name )

  return private.buttons[ button_name ].btn;

end;
--------------------------------------------------------------------------------
function private.Button_SetNotifications ( button_name, count )

  local button = private.buttons[ button_name ];
  
  button.notifications = count;

  if ( count > 0 ) then

    ui.obj.Set( button.anm,
    { 
      animfunc = button.animfunc.notify,
      count    = tostring( count ),
      playing  = true
    } );

  else
  
    ui.anim.Play( button.anm, button.animfunc.normal );

  end;

end;
--------------------------------------------------------------------------------
function private.Button_SetOverlayIcon ( button_name, icon_path )

  icon_path = icon_path or "";

  local button = private.buttons[ button_name ];
  
  ui.obj.Set(
    button.anm,
    { 
      rt_icon = icon_path
    }
  );

end;
--------------------------------------------------------------------------------
function private.Button_SetOverlayAnimIcon ( button_name, icon_path )

  icon_path = icon_path or "";
    
  local base_name = "int_panel_interface_right_"..button_name.."_overlay";
  local icon      = ui.obj.Get( "anm_"..base_name );

  if ( icon_path == "" ) then
    if ( icon ) then ui.obj.Delete( icon.name ); end;
    return;
  end;

  local button = private.buttons[ button_name ];

  if ( not icon ) then
    local properties = {
      res      = icon_path,
      anim_tag = "at_overlay_icon",
      animfunc = "static",
      playing  = true
    };
    ui.obj.CreateAnimation( base_name, button.anm, properties );
    return;
  end;

  if ( icon.res ~= icon_path ) then
    ui.obj.Set( icon.name, { res = icon_path } );
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.EnableButtonTimer ( button_name, func_seconds )
  
  local timer = private.buttons[ button_name ].timer;
  
  timer.enabled = true;
  timer.func_seconds = func_seconds;
  
  if ( private.active ) then
  
    private.ShowButtonTimer( button_name );
  
  end;

end;
--------------------------------------------------------------------------------
function private.DisableButtonTimer ( button_name )
  
  if ( private.buttons[ button_name ].timer.enabled ) then
  
    private.buttons[ button_name ].timer.enabled = false;
    private.HideButtonTimer( button_name, false );
  
  end;

end;
--------------------------------------------------------------------------------
function private.ShowButtonTimer ( button_name )

  if not ( private.active ) then return; end;

  local timer = private.buttons[ button_name ].timer;
  local function func_tick ()
  
    ui.timer.Show( timer.name, timer.func_seconds(), func_tick );
  
  end;
  func_tick();

end;
--------------------------------------------------------------------------------
function private.HideButtonTimer ( button_name, animated )

  ui.timer.Hide( private.buttons[ button_name ].timer.name, animated );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.TimersShow ()

  for button_name, button in pairs( private.buttons ) do
  
    if ( button.enabled and button.timer and button.timer.enabled ) then
    
      private.ShowButtonTimer( button_name );
    
    end;
  
  end;

end;
--------------------------------------------------------------------------------
function private.TimersHide ( animated )

  for button_name, button in pairs( private.buttons ) do
  
    if ( button.enabled and button.timer and button.timer.enabled ) then
    
      private.HideButtonTimer( button_name, animated );
    
    end;
  
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.ReorderButtons ()

  local enabled_list = {};
  for id, button in pairs( private.buttons ) do
  
    if ( button.enabled ) then table.insert( enabled_list, id ); end;
  
  end;
  
  if ( #enabled_list > 0 ) then
  
    table.sort( enabled_list, private.Order );
  
    local multiset = {};

    for i, button in pairs( enabled_list ) do
    
      ui.obj.Set(
        private.buttons[ button ].btn,
        { 
          anim_tag = "at_btn_"..i
        },
        multiset
      );
    
    end;

    ui.obj.Multiset( multiset );


  
  end;

end;
--------------------------------------------------------------------------------
function private.Order ( btn_one, btn_two )

  return private.buttons[ btn_one ].order < private.buttons[ btn_two ].order;

end;
--------------------------------------------------------------------------------
