-- name=int_panel_minievent_character
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Complete = function () end,
  OnShow   = function () end,
  OnHide   = function () end
};

local PANEL = "int_panel_minievent_character";

local _this = {
  anm  = "anm_int_panel_minievent_character",
  def  = "def_int_panel_minievent_character",
  body = "anm_int_panel_minievent_character_body",
  item = "anm_int_panel_minievent_character_item"
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  private.app_o = ui.screen.Orientation();
  private.app_h = ui.screen.Height();
  private.app_w = ui.screen.Width();

  ui.hub.Create( PANEL );

  local multiset = {};
  
  ui.obj.ToCenter( PANEL, 0, 0, multiset );
  
  ui.obj.Set(
    PANEL,
    {
      inputrect_init = true,
      inputrect_x    = private.app_w / -2,
      inputrect_y    = private.app_h / -2,
      inputrect_w    = private.app_w,
      inputrect_h    = private.app_h
    },
    multiset
  );

  ui.control.EventsSet(
    PANEL,
    {
      func_on_click = private.Next
    },
    multiset
  );

  ui.obj.Multiset( multiset );

  _this.strings = {
    dialog_skip = ne.strings.Get( "str_int_dialog_minievent_dialog_skip" )
  };

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Show ( dialog_data )

  if ( _this.is_active ) then return; end;

  _this.is_active = true;

  local respath = "assets/content/minievent_"..dialog_data.minievent.."/";

  _this.data             = dialog_data;
  _this.data.step        = 1;
  _this.data.respath     = respath;
  _this.data.res         = respath.."anm_int_dialog_minievent_section_character";
  _this.data.str_prefix  = "str_int_dialog_minievent_"..dialog_data.name.."_";
  _this.data.item_prefix = "str_int_dialog_minievent_"..dialog_data.minievent.."_item_";

  ui.hub.AttachContent( PANEL );

  private.Start();

end;
--------------------------------------------------------------------------------
function private.Start ()  

  local step     = _this.data.step;
  local multiset = {};

  ui.obj.Set(
    _this.anm,
    {
      active         = true,
      visible        = true,
      input          = true,
      res            = _this.data.res,
      tt_description = ne.strings.Get( _this.data.str_prefix..step ),
      tt_skip        = _this.strings.dialog_skip
    },
    multiset
  );

  ui.obj.Set(
    _this.body,
    {
      res      = _this.data.res,
      animfunc = "character".._this.data.list[ step ].character_view
    },
    multiset
  );

  private.SetItem(
    _this.item,
    _this.data.list[ step ].items,
    _this.data.respath,
    _this.data.res,
    _this.data.item_prefix,
    multiset
  );

  ui.obj.Multiset( multiset );
  
  common.sound.Play( "character_panel_appear" );

  ui.anim.Play(
    _this.anm,
    "appear_"..private.app_o,
    function ()
      ui.input.On( PANEL );
      _f.OnShow();
    end,
    "static_"..private.app_o
  );

  _this.data.step = step + 1;

end;
--------------------------------------------------------------------------------
function private.Next ()

  if ( not _this.data ) then return; end;

  common.sound.Play( "click" );

  local step = _this.data.step;

  if ( not _this.data.list[ step ] ) then

    _f.Complete( _this.data.name );

    private.Hide();

    _this.data      = nil;
    _this.is_active = nil;

    return;

  end;

  ui.obj.Clone( _this.body, _this.def, "_1" );
  ui.obj.Clone( _this.body, _this.def, "_2" );
  ui.obj.Clone( _this.item, _this.def, "_1" );
  ui.obj.Clone( _this.item, _this.def, "_2" );

  local animfunc = "character".._this.data.list[ step ].character_view;

  local multiset = {};

  ui.obj.SetTextTag( 
    _this.anm,
    "tt_description",
    ne.strings.Get( _this.data.str_prefix..step ),
    multiset
  );

  ui.obj.SetAnimTag( _this.body.."_1", "at_character1", multiset );

  ui.obj.Set(
    _this.body.."_2",
    {
      anim_tag = "at_character2",
      animfunc = animfunc
    },
    multiset
  );

  ui.obj.SetAnimTag( _this.item.."_1", "at_item1", multiset );
  ui.obj.SetAnimTag( _this.item.."_2", "at_item2", multiset );

  private.SetItem(
    _this.item.."_1",
    _this.data.list[ step - 1 ].items,
    _this.data.respath,
    _this.data.res,
    _this.data.item_prefix,
    multiset
  );

  private.SetItem(
    _this.item.."_2",
    _this.data.list[ step ].items,
    _this.data.respath,
    _this.datares,
    _this.data.item_prefix,
    multiset
  );

  ui.obj.Multiset( multiset );

  ui.obj.Attach( _this.body.."_1", _this.anm );
  ui.obj.Attach( _this.body.."_2", _this.anm );
  ui.obj.Attach( _this.item.."_1", _this.anm );
  ui.obj.Attach( _this.item.."_2", _this.anm );

  ui.anim.Play(
    _this.anm,
    "transition_"..private.app_o,
    function ()

      local multiset = {};

      ui.obj.Set( 
        _this.body,
        {
          animfunc = animfunc
        },
        multiset
      );

      private.SetItem(
        _this.item,
        _this.data.list[ step ].items,
        _this.data.respath,
        _this.data.res,
        _this.data.item_prefix,
        multiset
      );

      ui.obj.Multiset( multiset );

      ui.obj.Delete( _this.body.."_1" );
      ui.obj.Delete( _this.body.."_2" );
      ui.obj.Delete( _this.item.."_1" );
      ui.obj.Delete( _this.item.."_2" );

    end,
    "static_"..private.app_o
  );

  _this.data.step = step + 1;

end;
--------------------------------------------------------------------------------
function private.SetItem (
    obj_name,
    items,
    respath,
    res,
    item_prefix,
    multiset
  )

  if ( items ) then

    local rt_item = { "", "", "" };
    local tt_item = { "", "", "" };

    local item_count = #items;

    for i, item in pairs( items ) do

      if ( item.name:find( "chest" ) ) then
        rt_item[ i ] = ui.obj.Res( "item", item.name );
      else
        rt_item[ i ] = respath.."items/item_"..item.name;
      end;

      tt_item[ i ] = item_prefix..item.substr;

    end;

    ui.obj.Set(
      obj_name,
      {
        active    = true,
        visible   = true,
        res       = res,
        animfunc  = "item"..item_count,
        rt_item1  = rt_item[ 1 ],
        rt_item2  = rt_item[ 2 ],
        rt_item3  = rt_item[ 3 ],
        tt_item1  = tt_item[ 1 ],
        tt_item2  = tt_item[ 2 ],
        tt_item3  = tt_item[ 3 ]
      },
      multiset
    );

  else

    ui.obj.Disable( obj_name, false, multiset );

  end;

end;
--------------------------------------------------------------------------------
function private.Hide ()

  ui.input.Off( PANEL );

  ui.anim.Play(
    _this.anm,
    "disappear_"..private.app_o,
    function ()
      ui.hub.DetachContent( PANEL );
      _this.is_active = nil;
      _f.OnHide();
    end
  );

end;
--------------------------------------------------------------------------------