-- name=int_panel_minievent_worldmap
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Play             = function () end,
  GetRemainingTime = function () end
};

local _this = {
  int    = "int_panel_minievent_worldmap",
  anm    = "anm_int_panel_minievent_worldmap",
  button = {
    play = "btn_int_panel_minievent_worldmap_play"
  },
  timer  = {
    obj = "tmr_int_panel_minievent_worldmap",
    tag = "tt_event_timer"
  },
  dwn = "obj_int_panel_minievent_worldmap_downloading"
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  ui.button.Create(
    _this.button.play,
    {
      title  = "str_int_dialog_minievent_play",
      design = "medium",
      color  = "green"
    },
    function () _f.Play(); end
  );

  _this.timer.name = ui.obj.GetNameWithoutPrefix( _this.timer.obj );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Setup ( view_type, minievent_data )

  local multiset = {};

  if ( view_type == "started" ) then

    --DbgTrace( minievent_data );

    local minievent_name = minievent_data.name;
    local res_path       = "assets/content/minievent_"..minievent_name.."/";

    ui.obj.Set(
      _this.anm,
      { 
        res          = res_path.."anm_int_panel_minievent_worldmap",
        animfunc     = "static",
        tt_title     = "str_int_dialog_minievent_"..minievent_name.."_title",
        tt_event_end = "str_int_dialog_minievent_end",
        rt_back      = res_path.."panel_background",
        rt_red       = minievent_data.color.red,
        rt_green     = minievent_data.color.green,
        rt_blue      = minievent_data.color.blue
      },
      multiset
    );

    _this.attachable = _this.anm;

  end;

  if ( view_type == "downloading" ) then    

    local str_minievent_name =
      ne.strings.Get(
        "str_int_dialog_minievent_"..minievent_data.name.."_title"
      );

    local str_downloading =
      ne.strings.Get(
        "str_int_panel_minievent_worldmap_downloading",
        str_minievent_name
      );

    ui.obj.SetTextTag(
      "anm_int_panel_minievent_worldmap_downloading",
      "tt_wait",
      str_downloading,
      multiset
    );

    ui.obj.Set(
      "spr_int_panel_minievent_worldmap_downloading_bg",
      {
        res = ui.obj.Res( "item", "icon_minievent_"..minievent_data.type )
      },
      multiset
    );

    _this.attachable = _this.dwn;

  end;

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function public.Attach ( to )

  if ( _this.attached ) then public.Detach(); end;

  ui.obj.Attach( _this.attachable, to );

  _this.attached = _this.attachable;

end;
--------------------------------------------------------------------------------
function public.Detach ()

  if ( _this.attached ) then

    ui.obj.Attach( _this.attached, _this.int );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Timer_Start ()

  --DbgTrace( "Timer_Start" );

  local custom_obj = {
    anm_timer = _this.anm,
    text_tag  = _this.timer.tag
  };

  ui.obj.SetTextTag(
    custom_obj.anm_timer,
    custom_obj.text_tag,
    ui.text.FormatTime( _f.GetRemainingTime() )
  );

  local function func_tick ()

    ui.timer.Show(
      _this.timer.name,
      _f.GetRemainingTime(),
      func_tick,
      custom_obj
    );
  
  end;

  func_tick();

end;
--------------------------------------------------------------------------------
function public.Timer_Stop ()

  --DbgTrace( "Timer_Stop" );

  local custom_obj = {
    anm_timer = _this.anm,
    text_tag  = _this.timer.tag
  };

  ui.timer.Hide(
    _this.timer.name,
    false,
    custom_obj
  );

end;
--------------------------------------------------------------------------------