-- name=int_panel_social_score
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  SendHeart = function () end,
  Share     = function () end
};
--------------------------------------------------------------------------------
local _this = {
  friends          = {},
  slots_attached   = 0,
  view_type        = "start",
  is_logged        = false,
  friend_set       = {},
  --preload          = {},
  --preloaded        = {},
  player_photo_url = nil,
  player_photo_res = "assets/content/interface/social/default_avatar"
};

private._this = _this;
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  private.app_h = ui.screen.Height();
  private.app_w = ui.screen.Width();
  private.app_o = ui.screen.Orientation();
  
  private.this = {
    obj    = "obj_int_panel_social_score",
    button = {
      top   = "btn_int_panel_social_score_top",
      me    = "btn_int_panel_social_score_me",
      share = "btn_int_panel_social_score_share"
    }
  };
  
  private.slot_animfunc = {
    no_heart   = "static_"..private.app_o.."_no_heart",
    with_heart = "static_"..private.app_o.."_with_heart",
    me_top     = "static_"..private.app_o.."_me_top",
    me         = "static_"..private.app_o.."_me",
    opponent   = "static_"..private.app_o.."_opponent"
  };
  
  private.scroller = "scc_int_panel_social_score_slots";

  private.transition = {
    portrait  = "pos_y",
    landscape = "pos_x"
  };
  
  private.hide_pos = {
    portrait  = private.app_h + 180,
    landscape = private.app_w + 300
  };
  
  private.show_pos = {
    portrait  = private.app_h,
    landscape = private.app_w
  };
  
  private.z_invite = 100;
  private.z_dummy  = 200;

  private.size_click_area = {
    portrait = {
      input_rects = { 200, 130 },
      input_shift = { -75, 20 }
    },
    landscape = {
      input_rects = { 190, 130 },
      input_shift = { -60, 35 }
    }
  };

  private.Adapt();
  private.CreateButtons();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Adapt ()
  
  ui.hub.Create( "int_panel_social_score" );

  local multiset = {};
  
  private.Scroller_Create( multiset );
  
  local position = {
    portrait  = { private.app_w / 2, private.app_h },
    landscape = { private.app_w,     private.app_h / 2 }
  };
  
  ui.obj.Set(
    "obj_int_panel_social_score",
    {
      pos_x = position[ private.app_o ][ 1 ],
      pos_y = position[ private.app_o ][ 2 ]
    },
    multiset
  );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.CreateButtons ()
  
  ui.button.Create(
    private.this.button.top,
    {
      title  = "str_int_panel_social_score_top",
      design = "green_top"
    },
    function() private.Scroller_ScrollTo( "top" ); end
  );
  
  ui.button.Create(
    private.this.button.me,
    {
      title  = "str_int_panel_social_score_me",
      design = "blue_me"
    },
    function() private.Scroller_ScrollTo( "me" ); end
  );
  
  ui.button.Create(
  private.this.button.share,
    {
      title  = "str_int_panel_social_score_share",
      design = "fb_invite",
    },
    function() _f.Share(); end
  );
  
end;
--------------------------------------------------------------------------------
function private.Clean ()
  
  for i = 1, _this.slots_attached do

    local user_id  = _this.friends[ i ].userId;
    local anm_slot = "anm_int_panel_social_score_slot_"..user_id;

    if ( ui.obj.Exists( anm_slot ) ) then

      ui.obj.Set( anm_slot, { pos_z = 99999 } );
      ui.obj.Detach( anm_slot );

    end;

  end;

  _this.friends        = {};
  _this.slots_attached = 0;

  ui.hub.DetachContent( "int_panel_social_score" );
  
end;
--------------------------------------------------------------------------------
function private.GetMyPos ()

  for f = 1, #_this.friends do
    
    if ( tonumber( _this.friends[ f ].userId ) == private.my_id ) then
      
      return f;
      
    end;
    
  end;

end;
--------------------------------------------------------------------------------
function private.Win_SetupRating ( player, data, place, animfunc, multiset )
  
  local anm_slot  = "anm_int_panel_social_score_slot_"..player;
  local spr_photo = "spr_int_panel_social_score_slot_photo_"..player;
 
  if ( not ui.obj.Exists( anm_slot ) ) then
 
    ui.obj.Clone(
      "anm_int_panel_social_score_slot",
      "obj_int_panel_social_score_"..player,
      "_"..player,
      1
    );

  end;
  
  ui.obj.Set(
    anm_slot,
    {
      animfunc = animfunc,
      tt_name  = data.name,
      score    = ui.text.FormatNumber( data.score, true ),
      place    = tostring( place ),
    },
    multiset
  );
  
  data.res = data.res or
    "assets/content/interface/int_dialog_stage_friends/default_0";
  
  private.SetPhoto( spr_photo, data.url, data.res, data.res );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show (
    point, -- Point can be "start" or "win"
    friends,
    my_id,
    is_logged
  )

  if ( private.active ) then
    
    return;
    
  end;
  
  private.active = true;
  
  ui.obj.Show( "int_panel_social_score" );

  _this.is_logged = is_logged;
  _this.view_type = tern( point:find( "win" ) )( "win", "start" );

  _this.friends   = friends;
  private.my_id   = tonumber( my_id );
  private.my_pos  = private.GetMyPos();

  -- JK: update player photo if changed
  local player_data = _this.friends[ private.my_pos ];

  --Log.i( "Current photo URL: "..tostring( _this.player_photo_url ) );
  --Log.i( "Current photo res: "..tostring( _this.player_photo_res ) );
  --Log.i( "New photo URL: "..tostring( player_data.url ) );
  --Log.i( "New photo res: "..tostring( player_data.res ) );

  if ( type( player_data.url ) == "string" ) and
     ( player_data.url ~= ""               ) then

    if ( _this.player_photo_url ~= ""              ) and
       ( _this.player_photo_url ~= player_data.url ) then

      private.UpdatePhoto( player_data );

    end;

  else

    if ( _this.player_photo_res ~= ""              ) and
       ( _this.player_photo_res ~= player_data.res ) then

      private.UpdatePhoto( player_data );

    end;

  end;
  
  private.SetDefault( point, friends );

  ui.obj.Set( "anm_int_panel_social_score_win", { visible = false } );
  
  if ( point ~= "start" ) then
    
    if ( private.my_pos == 1 ) then
      ui.obj.Set( "anm_int_panel_social_score_win", 
      { 
        visible = true, 
        rt_icon = "assets/content/items/icon_top_1", 
        playing = true 
      } );
    end;

  end;
  
  ui.hub.AttachContent( "int_panel_social_score" );

  local func_OnShow;

  local func_SetupView = function ()

    if ( _this.view_type:find( "win" ) ) then
      
      private.Scroller_ScrollTo( "me" );
      
      local multiset = {};

      local player_data = friends[ private.my_pos ];
      
      private.Win_SetupRating(
        "player",
        player_data,
        private.my_id,
        private.slot_animfunc.me,
        multiset
      );
      
      local opponent_pos = private.my_pos + 1;
      
      player_data = friends[ opponent_pos ];

      if ( player_data ) then
      
        private.Win_SetupRating(
          "opponent",
          player_data,
          opponent_pos,
          private.slot_animfunc.opponent,
          multiset
        );
        
      end;

      ui.obj.Multiset( multiset );

      ui.obj.SpriteAdjust(
        "spr_int_panel_social_score_slot_photo_player",
        74
      );
      
      ui.obj.SpriteAdjust(
        "spr_int_panel_social_score_slot_photo_opponent",
        74
      );

    else

      private.Scroller_ScrollTo( "top", true );

      func_OnShow = function () private.Scroller_OnScroll(); end;
      
    end;
      
  end;

  func_SetupView();

  public.VisualShow( func_OnShow );

end;
--------------------------------------------------------------------------------
function public.Hide ()

  if ( not private.active ) then
    
    return;
    
  end;
  
  public.VisualHide( private.Clean );
  
  private.active = false;

end;
--------------------------------------------------------------------------------
function public.IsActive ()

  return private.active;

end;
--------------------------------------------------------------------------------
function public.Prepare ( friends )

--  if ( #friends > 0 ) then
--
--    _this.preload = friends;
--
--    ui.obj.Create(
--      "dmy_int_panel_social_score_preload",
--      "dsprite",
--      InterfaceName,
--      { alp = 0 }
--    );
--
--    private.TryPreload();
--
--    return;
--
--  end;
  
end;
--------------------------------------------------------------------------------
function public.ShareComplete ()
  
  ui.obj.Set(
    "anm_int_panel_social_score",
    {
      animfunc = "win_unlogged_"..private.app_o
    }
  );
  
end;
--------------------------------------------------------------------------------
--[[function private.TryPreload ()

  local preloaded_count = #_this.preloaded;

  if ( preloaded_count >= #_this.preload ) then

    ui.obj.Delete( "dmy_int_panel_social_score_preload" );

    return;

  end;

  local friend = _this.preload[ preloaded_count + 1 ];

  if ( friend.url == "" ) then friend.url = nil; end;

  friend.res = friend.res or
    "assets/content/interface/int_dialog_stage_friends/default_0";

  DbgTrace( "Preloading: "..tostring( friend.url ) );

  local function on_preloaded ()

    table.insert( _this.preloaded, friend.userId );
    private.TryPreload();

  end;

  if ( friend.url ) then

    private.SetPhoto( spr_photo, friend.url, friend.res, friend.res, function () on_preloaded(); end );
    
  else

    on_preloaded();

  end;

end;]]
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.SetDefault ( point, friends )
  
  local multiset      = {};
  local cellsize      = ui.obj.Get( private.scroller ).cellsize;
  _this.cells_visible = math.floor( private.app_w / cellsize ) - 1;

  private.Scroller_AttachSlots( ( _this.cells_visible + 1 ) );
  private.Scroller_TryCreateDummies( friends );
  private.Scroller_Adapt( _this.view_type );
  
  if ( private.app_o == "portrait" ) then
    if ( _this.is_logged ) then
      ui.button.Enable( private.this.button.share, multiset );
    else
      ui.button.Disable( private.this.button.share, multiset );
      _this.view_type = _this.view_type.."_unlogged";
    end;
  else
    ui.button.Disable( private.this.button.share, multiset );
  end;
  
  ui.obj.Set(
    "anm_int_panel_social_score",
    {
      animfunc = _this.view_type.."_"..private.app_o
    },
    multiset
  );
  
  ui.obj.Set(
    "obj_int_panel_social_score",
    {
      [ private.transition[ private.app_o ] ] = private.hide_pos[ private.app_o ],
      alp = 0
    },
    multiset
  );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.SetFriend ( place, friend, multiset )

  local id        = friend.userId;
  local anm_slot  = "anm_int_panel_social_score_slot_"..id;
  local spr_photo = "spr_int_panel_social_score_slot_photo_"..id;

  if ( friend.url == "" ) then friend.url = nil; end;
  if ( friend.res == "" ) then friend.res = nil; end;
  
  local url = friend.url;
  local res = friend.res;
  
  if ( id == private.my_id ) then
    _this.player_photo_url = url or _this.player_photo_url;
    _this.player_photo_res = res or _this.player_photo_res;
    url = _this.player_photo_url;
    res = _this.player_photo_res;
  end;

  if ( not friend.is_bot              ) and
     ( _this.view_type:find( "start" ) ) then

    local btn_name = "btn_int_panel_social_score_slot_heart_"..id;

    ui.button.Create(
      btn_name,
      {
        design      = "heart_send",
        input_rects = private.size_click_area[ private.app_o ].input_rects,
        input_shift = private.size_click_area[ private.app_o ].input_shift
      },
      function()
        
        _f.SendHeart(
          friend.userId,
          friend.name,
          res,
          url,
          ui.obj.Position( btn_name, "ng_interface" )
        );

      end
    );

  end;

  private.SetPhoto( spr_photo, url, res );
  
  _this.friend_set[ "slot_"..id ] = true;

end;
--------------------------------------------------------------------------------
function private.UpdatePhoto ( friend )

  local id  = friend.userId;
  local url = friend.url;
  local res = friend.res

  if ( id == private.my_id ) then
    _this.player_photo_url = url or _this.player_photo_url;
    _this.player_photo_res = res or _this.player_photo_res;
    url = _this.player_photo_url;
    res = _this.player_photo_res;
  end;

  local spr_photo = "spr_int_panel_social_score_slot_photo_"..id;
  
  private.SetPhoto( spr_photo, url, res );

end;
--------------------------------------------------------------------------------
function private.SetPhoto ( spr_photo, url, res, default_res )
  
  if ( ui.obj.Exists( spr_photo ) ) then
    
    local data = {
      url = url,
      res = res,
      trg_resloaded = function () ui.obj.SpriteAdjust( spr_photo, 74 ); end
    };
    
    if ( default_res ) then
      data.default_res = default_res;
    end;
    
    ui.obj.Set( spr_photo, data );  
    ui.obj.SpriteAdjust( spr_photo, 74 );

  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Scroller_Create ( multiset )

  ui.obj.Clone(
    "scc_int_panel_social_score_default_slots",
    "anm_int_panel_social_score",
    "_rename"
  );
  
  local scroller_layout = {
    portrait = {
      cellsize =  148,
      rect_x   = -272,
      rect_y   = -85,
      rect_w   =  484,
      rect_h   =  170
    },
    landscape = {
      cellsize =  102,
      rect_x   = -150,
      rect_y   = -210,
      rect_w   =  300,
      rect_h   =  630
    }
  };
  
  ui.obj.Set(
    "scc_int_panel_social_score_default_slots_rename",
    {
      name        = private.scroller,
      scrollpos   = 0,
      dragenabled = true,
      horizontal  = ( private.app_o == "portrait" ),
      cellsize    = scroller_layout[ private.app_o ].cellsize,
      inputrect_x = scroller_layout[ private.app_o ].rect_x,
      inputrect_y = scroller_layout[ private.app_o ].rect_y,
      inputrect_w = scroller_layout[ private.app_o ].rect_w,
      inputrect_h = scroller_layout[ private.app_o ].rect_h,
      croprect_x  = scroller_layout[ private.app_o ].rect_x,
      croprect_y  = scroller_layout[ private.app_o ].rect_y,
      croprect_w  = scroller_layout[ private.app_o ].rect_w,
      croprect_h  = scroller_layout[ private.app_o ].rect_h,
      enddragtrig = function ()

        local scrollpos = ui.obj.Get( private.scroller ).scrollpos;

        private.Scroller_AttachSlots( scrollpos + ( _this.cells_visible * 2 ) );
        private.Scroller_OnScroll();

      end
    },
    multiset
  );

end;
--------------------------------------------------------------------------------
function private.Scroller_Adapt ( point )
  
  local scroller_layout = {
    start = {
      portrait = {
        pos_x    = - 215,
        rect_x   =   0,
        rect_w   =   492
      },
      landscape = {
        rect_x   =   0,
        rect_w   =   300
      }    
    },
    win = {
      portrait = {
        pos_x    = - 215,
        rect_x   =   0,
        rect_w   =   436
      },
      landscape = {
        rect_x   =   0,
        rect_w   =   300
      }    
    }
  };
  
  ui.obj.Set(
    private.scroller,
    {
      scrollpos   = 0,
      pos_x       = scroller_layout[ point ][ private.app_o ].pos_x,
      inputrect_x = scroller_layout[ point ][ private.app_o ].rect_x,
      inputrect_w = scroller_layout[ point ][ private.app_o ].rect_w,
      croprect_x  = scroller_layout[ point ][ private.app_o ].rect_x,
      croprect_w  = scroller_layout[ point ][ private.app_o ].rect_w
    }
  );
  
end;  
--------------------------------------------------------------------------------
function private.Scroller_ScrollTo ( target, ignore_onscroll )
  
  local target_pos = 0;

  if ( target == "top" ) then

    --

  elseif ( target == "me" ) then

    local viewable = math.floor( _this.cells_visible / 2 ) + 1;

    target_pos = math.max( 0, private.my_pos - viewable );
    
  end;

  if ( ui.obj.Get( private.scroller ).scrollpos ~= target_pos ) then

    ui.obj.Set( private.scroller, { scrollpos = target_pos } );

  end;

  if ( target == "me" ) then

    private.Scroller_AttachSlots( target_pos + _this.cells_visible * 2 );

  end;

  if ( not ignore_onscroll ) then

    private.Scroller_OnScroll();

  end;

end;
--------------------------------------------------------------------------------
function private.Scroller_OnScroll ()

  local multiset  = {};
  local affected  = {};
  local scrollpos = ui.obj.Get( private.scroller ).scrollpos + 1;

  for i = scrollpos, ( scrollpos + _this.cells_visible ) do

    if ( #_this.friends < 1 ) then
      
      return;

    end;

    local friend = _this.friends[ i ];

    if ( friend                                        ) and
       ( not _this.friend_set[ "slot_"..friend.userId ] ) then

      private.SetFriend( i, friend, multiset );

      table.insert( affected, friend.userId );

    end;

  end;

  if ( #multiset > 0 ) then

    ui.obj.Multiset( multiset );

    for _, user_id in pairs( affected ) do

      local sprite_name = "spr_int_panel_social_score_slot_photo_"..user_id;

      ui.obj.SpriteAdjust( sprite_name, 74 );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Scroller_TryCreateDummies ( friends )
 
  if ( #friends < _this.cells_visible + 1 ) then
    
    for i = 1, ( _this.cells_visible - #friends ) do
      
      private.Scroller_CreateDummy( i );
      
    end;
    
  end;

end;
--------------------------------------------------------------------------------
function private.Scroller_CreateDummy ( i )

  local obj_dummy = "obj_int_panel_social_score_slot_dummy_"..i;
  
  if ( not ui.obj.Exists( obj_dummy ) ) then
    
    ui.obj.Create( obj_dummy, "obj" );
    ui.obj.Set( obj_dummy, { pos_z = private.z_dummy } );
    
  end;
  
  ui.obj.Attach( obj_dummy, private.scroller );

end;
--------------------------------------------------------------------------------
function private.Scroller_AttachSlots ( to_number )

  to_number = math.min( to_number, #_this.friends );

  if ( _this.slots_attached >= to_number ) then

    return;

  end;

  local multiset = {}; 

  for i = _this.slots_attached + 1, to_number do

    local friend   = _this.friends[ i ];
    local is_me    = ( friend.userId == private.my_id );
    local anm_slot = "anm_int_panel_social_score_slot_"..friend.userId;

    local animfunc, tt_name;

    animfunc = tern( friend.is_bot or ( _this.view_type:find( "win" ) ) )(
      private.slot_animfunc.no_heart,
      private.slot_animfunc.with_heart
    );

    if ( is_me ) then

      animfunc = private.slot_animfunc.me_top;

    end;

    if ( not ui.obj.Exists( anm_slot ) ) then
      
      ui.obj.Clone(
        "anm_int_panel_social_score_slot",
        private.scroller,
        "_"..friend.userId,
        1
      );

      tt_name = tern( is_me )( "str_int_panel_social_score_me", friend.name );

    end;

    ui.obj.Set(
      anm_slot,
      {
        animfunc = animfunc,
        tt_name  = tt_name,
        score    = ui.text.FormatNumber( friend.score, true ),
        place    = tostring( i ),
        pos_z    = i,
        playing = true
      },
      multiset
    );

    ui.obj.Attach( anm_slot, private.scroller );

  end;

  ui.obj.Multiset( multiset );

  _this.slots_attached = to_number;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.VisualShow ( func_end )
  
  if ( private.active ) then
    
    ui.input.Off( private.this.obj );
    
    local transition = private.transition[ private.app_o ];
    local start_pos  = ui.obj.Get( "obj_int_panel_social_score" )[ transition ];
    local stop_pos   = private.show_pos[ private.app_o ];
    
    ObjAnimate(
      "obj_int_panel_social_score",
      transition, 0, 0,
      function ()
        
        if ( func_end ) then func_end(); end;
        ui.input.On( private.this.obj );
        
      end, 
      {
        0.0, 0, start_pos,
        0.3, 0, stop_pos
      }
    );

    ui.anim.FadeIn( 0.3, "obj_int_panel_social_score" );
     
  end;
  
end;
--------------------------------------------------------------------------------
function public.VisualHide ( func_end )
  
  if ( private.active ) then
    
    ui.input.Off( private.this.obj );
    
    local transition = private.transition[ private.app_o ];
    local start_pos  = ui.obj.Get( "obj_int_panel_social_score" )[ transition ];
    local stop_pos   = private.hide_pos[ private.app_o ];
    
    ObjAnimate(
      "obj_int_panel_social_score",
      transition, 0, 0, func_end, 
      {
        0.0, 0, start_pos,
        0.3, 0, stop_pos
      }
    );

    ui.anim.FadeOut( 0.3, "obj_int_panel_social_score" );
    
  end;
  
end;
--------------------------------------------------------------------------------
