-- name=int_panel_stage_boosters
--------------------------------------------------------------------------------

local _this = {
  int = "int_panel_stage_boosters",
  obj = "obj_int_panel_stage_boosters",
  pos = {
    show = { x = 0, y = 0 },
    hide = { x = 0, y = 0 }
  },
  anim_time   = 0.3,
  scale       = 1.0,
  is_vertical = false,
  is_active   = false,
  is_hideable = true,
  choice_booster = nil;
};

--------------------------------------------------------------------------------
function public.__InitObjects ()

  ui.hub.Create( _this.int );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.SetScale ( s )
  
  _this.scale = s;
    
end;
--------------------------------------------------------------------------------
function public.SetPosVisible ( x, y )
  
  _this.pos.show.x = x;
  _this.pos.show.y = y;
  
end;
--------------------------------------------------------------------------------
function public.SetPosHidden ( x, y )
  
  _this.pos.hide.x = x;
  _this.pos.hide.y = y;
  
end;
--------------------------------------------------------------------------------
function public.SetVertical ( is_vertical )
  
  _this.is_vertical = is_vertical;
  
end;
--------------------------------------------------------------------------------
function public.SetHideable ( is_hideable )
  
  _this.is_hideable = is_hideable;
  
end;
--------------------------------------------------------------------------------
function public.IsHideable ()
  
  return _this.is_hideable;
  
end;
--------------------------------------------------------------------------------
function public.IsActive ()
  
  return _this.is_active;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show ( boosters_data, func_tap )
  
  if ( _this.is_active ) then return end;
  
  _this.is_active = true;
  _this.choice_booster = nil;
  
  private.booster_index  = {};
  private.emergency_mode = false;

  -- JK: throw away locked boosters
  local unlocked_boosters = {};

  for _, booster in pairs( boosters_data ) do

    if ( booster.unlocked ) then

      table.insert( unlocked_boosters, booster );

    end;

  end;

  private.booster_count = #unlocked_boosters;

  -- JK: if booster count is zero, panel will not show
  if ( private.booster_count < 1 ) then

    _this.is_active = false;
    return;

  end;

  private.Tap = func_tap;

  ui.hub.AttachContent( _this.int );
  ui.obj.Enable( _this.int );
  
  private.SetDefault( unlocked_boosters );
  
  ui.anim.Play(
    "anm_int_panel_stage_boosters",
    "booster"..private.booster_count.."_static"
  );
  
  ui.obj.Enable( "anm_int_panel_stage_boosters" );

  local tme = _this.anim_time;

  ObjAnimate( _this.obj, "pos_xy", 0, 0, "", 
  {
    0.0, 2, _this.pos.hide.x, _this.pos.hide.y,
    tme, 2, _this.pos.show.x, _this.pos.show.y
  } );

end;
--------------------------------------------------------------------------------
function public.Hide ()

  if ( not _this.is_active ) then return; end;

  _this.is_active        = false;
  private.booster_count = nil;

  ui.input.Off( _this.int );

  local func_end = function ()

    ui.obj.Disable( _this.int, true );
    ui.hub.DetachContent( _this.int );

  end;

  local tme = _this.anim_time;

  ObjAnimate( _this.obj, "pos_xy", 0, 0, func_end, 
  {
    0.0, 2, _this.pos.show.x, _this.pos.show.y,
    tme, 2, _this.pos.hide.x, _this.pos.hide.y
  } );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Update ( boosters_list )

  if ( not _this.is_active ) then

    return;

  end; 

  for index, booster in pairs( boosters_list ) do

    if ( private.booster_index[ booster.name ] ) then

      local index = private.booster_index[ booster.name ];

      ui.obj.Set(
        "anm_int_panel_stage_booster_"..index.."_icon",
        {
          count    = ui.text.FormatNumber( booster.count ),
          playing  = true,
          animfunc = tern( booster.count > 0 )( "booster", "booster_add" )
        }
      );

    end;

  end;

end;
--------------------------------------------------------------------------------
function public.UpdateInfo ( func_data_booster )
  
  if ( not _this.is_active ) then return; end;
  
  for key, value in pairs( private.booster_index ) do
    
    local booster_data = func_data_booster( key );
    local booster_res_name = 
      tern( booster_data.phase )( booster_data.phase, booster_data.name );
    
    ObjSet( "anm_int_panel_stage_booster_"..value.."_icon", {
      rt_sale = tern( booster_data.sale )( "assets/content/items/icon_sale","" ),
      tt_sale = tern( booster_data.sale )( "str_int_sale_label","" ),
      booster_icon = ui.obj.Res( "booster", booster_res_name )
    } );
    
  end;
  
end;
--------------------------------------------------------------------------------
function public.SetInput ( flag )
  
  ui.input.Set( _this.obj, flag );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.OnBoosterTap ( booster_name )
  
  common.sound.Play( "select_boost" );

  if ( private.Tap( booster_name ) ) then

    private.StartAnim( booster_name );

  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.EmergencyMode_TryStart ()
  
  private.EmergencyMode_TryStart();
  
end;
--------------------------------------------------------------------------------
function public.EmergencyMode_Stop ()
  
  private.EmergencyMode_Stop();
  
end;
--------------------------------------------------------------------------------
function private.EmergencyMode_TryStart ()
  
  if      ( private.booster_count     )
  and     ( private.booster_count > 0 )
  and not ( private.emergency_mode    ) then
  
    private.emergency_mode = true;
    
    private.EmergencyMode_Cycle();
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.EmergencyMode_Cycle ()

  local function func_bounce ( booster_id, func_after_bounce )

    local sprite = "anm_int_panel_stage_booster_"..booster_id.."_icon";
    ui.anim.Bounce( sprite, func_after_bounce );

  end;

  local callbacks = {};

  for i = 1, private.booster_count do
    table.insert( callbacks, function ( after ) func_bounce( i, after ) end );
  end;

  table.insert( callbacks, private.EmergencyMode_Stop );

  common.RunCallbackList( callbacks );
  
end;
--------------------------------------------------------------------------------
function private.EmergencyMode_Stop ()
  
  private.emergency_mode = false;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.StartAnim ( booster_name )

  private.StartAnim( booster_name );

end;
--------------------------------------------------------------------------------
function public.StopAnim ( func_end )

  if ( not _this.choice_booster ) then 
    if ( func_end ) then func_end(); end;
    return;
  end;

  local anm_name = "booster1_desc_disappear";
  local booster_name = _this.choice_booster;
  
  if ( private.booster_count ~= 1 ) then 
    
    anm_name = "booster"..private.booster_count..private.booster_index[ booster_name ].."_desc_disappear";
    
  end;

  ObjSet( "anm_int_panel_stage_boosters", { 
    description = "str_item_"..booster_name.."_description"
  } );
  
  ui.anim.Play( "anm_int_panel_stage_boosters", anm_name, function ()
    _this.choice_booster = nil;
    if ( func_end ) then func_end(); end;
  end );
  
  private.StopAnim( booster_name );
    
end;
--------------------------------------------------------------------------------
function private.StartAnim ( booster_name )
  
  if ( _this.choice_booster ) then
    return;
  end;
  
  local anm_name = "booster1_desc_appear";
    
  if ( private.booster_count ~= 1 ) then 

    anm_name = "booster"..private.booster_count..private.booster_index[ booster_name ].."_desc_appear";
  
  end;
  
  ObjSet( "anm_int_panel_stage_boosters", { 
    description = "str_item_"..booster_name.."_description",
    animfunc = anm_name,
    playing = true;
  } );
  
  _this.choice_booster = booster_name;
  
  local index = private.booster_index[ booster_name ];

  ObjAttach( "anm_int_panel_stage_booster_selection", "obj_int_panel_stage_booster_"..index );
  ObjSet( "anm_int_panel_stage_booster_selection", { active = true, visible = true, scale_x = 1, scale_y = 1 } );
  
  ui.anim.Play( "anm_int_panel_stage_booster_selection", "select" );
    
  local spr_params = ObjGet( "anm_int_panel_stage_booster_"..index.."_icon" );

  ObjAnimate( "anm_int_panel_stage_booster_"..index.."_icon", "scale_xy", 1, 0, "", 
  {
    0.0, 0, spr_params.scale_x, spr_params.scale_y,
    0.5, 0, spr_params.scale_x + 0.1, spr_params.scale_y + 0.1,
    1.0, 0, spr_params.scale_x, spr_params.scale_y
  } );

end;
--------------------------------------------------------------------------------
function private.StopAnim ( booster_name )

  AnimStop( "anm_int_panel_stage_booster_selection" );
  ObjSet( "anm_int_panel_stage_booster_selection", { active = false, visible = false } );

  local index = private.booster_index[ booster_name ];

  for i = 1, 3 do

    if not ( i == index ) then

      ObjStopAnimate( "anm_int_panel_stage_booster_"..i.."_icon", "scale_xy" );
      ObjSet( "anm_int_panel_stage_booster_"..i.."_icon", { scale_x = 1, scale_y = 1 } );

    end;

  end;

  ObjStopAnimate( "anm_int_panel_stage_booster_"..index.."_icon", "scale_xy" );
  ObjSet( "anm_int_panel_stage_booster_"..index.."_icon", { scale_x = 1, scale_y = 1 } );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.SetDefault ( boosters_data )

  local multiset = {};
  
  ui.obj.Set(
    _this.obj,
    {
      active  = true,
      visible = true,
      pos_x   = _this.pos.hide.x,
      pos_y   = _this.pos.hide.y
    },
    multiset
  );
  
  for i = 1, 3 do

    ui.obj.Disable( "obj_int_panel_stage_booster_"..i, false, multiset );

  end;

  ui.obj.Multiset( multiset );

  multiset = {};

  -- Set booster contents
  for index, booster in pairs( boosters_data ) do

    local booster_res_name = tern( booster.phase )( booster.phase, booster.name );
    
    private.booster_index[ booster.name ] = index;

    private.StopAnim( booster.name );
      
    ui.obj.Enable( "obj_int_panel_stage_booster_"..index, false, multiset );

    ui.obj.Set(
      "anm_int_panel_stage_booster_"..index.."_icon",
      { 
        count        = ui.text.FormatNumber( booster.count ),
        booster_icon = ui.obj.Res( "booster", booster_res_name ),
        scale_x      = 1,
        scale_y      = 1,
        animfunc     = tern( booster.count > 0 )( "booster", "booster_add" ),
        playing      = true,
        rt_sale      = tern( booster.sale )( "assets/content/items/icon_sale", "" ),
        tt_sale      = tern( booster.sale )( "str_int_sale_label", "" )
      },
      multiset
    );

    ui.control.EventsSet(
      "obj_int_panel_stage_booster_"..index,
      {
        func_on_click = function ()

          private.OnBoosterTap( booster.name );

          common.sound.Play( "app_info" );

        end
      },
      multiset
    );

  end;

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Popup_Show ()
  
  if ( private.booster_count and private.booster_count > 0 ) then
    
    ui.popup.Flat_Show( 
      "str_int_panel_boosters_use_booster_caption", 
      "str_int_panel_boosters_use_booster",
      "obj_int_panel_stage_booster_2",
      { 0, - 160 }
    );
    
  end;
  
end;
--------------------------------------------------------------------------------