-- name=int_panel_stage_progress_fight
--------------------------------------------------------------------------------

local _this = {};

--------------------------------------------------------------------------------
function public.__InitModule ()

  _this = {
    int         = "int_panel_stage_progress_fight",
    obj         = "obj_int_panel_stage_progress_fight",
    enemy_touch = "obj_int_panel_stage_progress_fight_touch_enemy",
    hero_touch  = "obj_int_panel_stage_progress_fight_touch_hero",
    healthbar   = {
      hero  = "anm_int_panel_stage_progress_fight_healthbars_hero_healthbar",
      enemy = "anm_int_panel_stage_progress_fight_healthbars_enemy_healthbar"
    },
    health = {
      hero  = "anm_int_panel_stage_progress_fight_healthbars_hero",
      enemy = "anm_int_panel_stage_progress_fight_healthbars_enemy"
    },
    attack_time = {
      hero  = "anm_int_panel_stage_progress_fight_animations_hero_attack_time",
      enemy = "anm_int_panel_stage_progress_fight_animations_enemy_attack_time"
    },
    ability = {
      hero  = "anm_int_panel_stage_progress_fight_healthbars_hero_ability",
      enemy = "anm_int_panel_stage_progress_fight_healthbars_enemy_ability"
    },
    icon = {
      hero  = "anm_int_panel_stage_progress_fight_healthbars_hero_icon",
      enemy = "spr_int_panel_stage_progress_fight_healthbars_enemy_icon"
    }
  };
  
  private.active = false;
  
  private.screen_w = ui.screen.Width();
  private.screen_h = ui.screen.Height();

  private.attack_queue = {};
  
  private.hero_audio  = "assets/content/hero/aud_hero_";
  private.healthbar_w = 200;

  private.anim_time = 0.3;
  private.fg_scale  = 1.0;
  
  -- Hero and enemy distance
  private.dst_near = 350;
  private.dst_far  = 350;
  private.dst_off  = 350;
  
  private.count_frames_in_bar = 120;
  
end;
--------------------------------------------------------------------------------
function public.__InitObjects ()
 
  ui.hub.CreateInStage( _this.int );

  private.Adapt();

  mediator.Subscribe(
    "Model_Stage_HeroHealthRecovered",
    private.Hero_Respawn
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Show ( stage_data, hero_data, enemy_data )
  
  if ( private.active ) then

    private.Reset();

  end;

  private.attack_queue   = {};
  private.combo_hit      = nil;
  private.attack_pending = nil;
  
  local multiset = {};
  
  ui.input.On({
    _this.enemy_touch,
    _this.hero_touch
  });
  
  ui.obj.Disable( _this.obj, true, multiset );
  
  ui.hub.AttachContent( _this.int );

  private.SetDefaults( stage_data, hero_data, enemy_data, multiset );
  
  ui.obj.Multiset( {
    { _this.int,
      {
        input   = true,
        active  = true,
        visible = true,
        pos_x   = 0.5 * private.screen_w
      }
    },
    { "anm_int_panel_stage_progress_fight_animations_enemy",
      {
        pos_x = 0.5 * ( private.dst_far - private.dst_off )
      }
    },
    { "anm_int_panel_stage_progress_fight_animations_enemy_hero",
      {
        pos_x = -private.dst_far + private.dst_off
      }
    },
    { _this.health.hero,
      {
        pos_x = private.healthbar_hero_pos_start
      }
    },
    { _this.health.enemy,
      {
        pos_x = private.healthbar_enemy_pos_start
      }
    },
  } );
  
  public.ShowDots( 0, 0 );

  ObjSet( _this.int, { pos_y = private.pos_y } );

  ObjAnimate(
  _this.health.hero,
  "pos_x", 0, 0, "", 
  {
    0.0,               2, private.healthbar_hero_pos_start,
    private.anim_time, 2, private.healthbar_hero_pos_stop
  } );

  ui.anim.FadeIn(
    private.anim_time,
    _this.health.hero,
    true
  );

  ObjAnimate(
  _this.health.enemy,
  "pos_x", 0, 0, "", 
  {
    0.0,               2, private.healthbar_enemy_pos_start,
    private.anim_time, 2, private.healthbar_enemy_pos_stop
  } );

  ui.anim.FadeIn(
    private.anim_time,
    _this.health.enemy,
    true
  );
  
  private.active = true;

end;
--------------------------------------------------------------------------------
function public.Hide ()

  private.active = false;

  ui.obj.Disable( _this.int );
  
  ui.anim.Stop( _this.attack_time.hero );
  ui.anim.Stop( _this.attack_time.enemy );
  ui.anim.Stop( "anm_int_panel_stage_progress_fight_animations_enemy" );
  ui.anim.Stop( private.player_skin );
  
  ObjStopAnimate( _this.healthbar.hero,  "frame" );
  ObjStopAnimate( _this.healthbar.enemy, "frame" );
  
  ui.input.On( "obj_int_panel_stage_progress_fight_touch" );
  
  ui.hub.DetachContent( _this.int );
  private.Reset();

end;
--------------------------------------------------------------------------------
function public.IsWinAnimEnded ()

  return private.WinAnimEnd;

end;
--------------------------------------------------------------------------------
function private.Adapt ()

  local percent_w = private.screen_w / 100;

  private.healthbar_hero_pos_start  = - ( private.screen_w / 2 ) - private.healthbar_w;
  private.healthbar_hero_pos_stop   = - ( 50 * percent_w );
  private.healthbar_enemy_pos_start =     private.screen_w;
  private.healthbar_enemy_pos_stop  =   ( 50 * percent_w );
  
  local objects = {};

  table.insert( objects,
  { _this.health.enemy,
    {
      pos_x = private.healthbar_enemy_pos_start } } );
      
  table.insert( objects,
  { _this.health.hero,
    {
      pos_x = private.healthbar_hero_pos_start } } );

  local parts_count = private.screen_w / 128;

  for i = 1, parts_count do

    ObjClone(
      "spr_int_panel_stage_progress_fight_fg_decoration",
      "_"..i, 1,
      "obj_int_panel_stage_progress_fight_fg"
    );

    table.insert( objects,
    { "spr_int_panel_stage_progress_fight_fg_decoration_"..i,
      {
        pos_x = 128 * ( i - 0.5 )
      }
    } );

  end;

  ui.obj.Multiset( objects );
  
  -- Resize decoartion grid ( horizontal layout )
  local grd_size = {
    left   = private.screen_w / 2,
    right  = private.screen_w / 2,
    bottom = 0,
    top    = 40
  }
  
  local grd_border = {
    left = 0,
    right = 0,
    bottom = 0,
    top = 40
  }
  
  ui.grid.Resize(
    "grd_int_panel_stage_progress_fight_fg",
    grd_size,
    grd_border
  );

  ui.control.EventsSet(
    _this.hero_touch,
    { func_on_click = function () private.Hero_OnTap(); end }
  );

  ui.control.EventsSet(
    _this.enemy_touch,
    { func_on_click = function () private.Enemy_OnTap(); end }
  );

end;
--------------------------------------------------------------------------------
function public.SetOffset ( offset )
  
  private.pos_y = offset;
  
  local bg_w, bg_h = 768, 340; -- BG texture size
  
  local bg_scale_h = offset / bg_h;
  local bg_scale_w = private.screen_w / bg_w;
  
  local bg_scale = common.math.Round( math.max( bg_scale_h, bg_scale_w ), 2 );
  
  local objects = {};
  
  table.insert( objects,
  { _this.obj,
    {
      croprect_h =  offset, 
      croprect_y = -offset,
      croprect_w =  private.screen_w,
      croprect_x = -private.screen_w / 2 } } );
      
  table.insert( objects,
  { "spr_int_panel_stage_progress_fight_bg",
    {
      drawoff_y = -0.5 * bg_h,
      scale_x   =  bg_scale, 
      scale_y   =  bg_scale
    }
  } );
   
  table.insert( objects,
  { "obj_int_panel_stage_progress_fight_healthbars",
    {
      pos_y = -offset + 5
    }
  } );
  
  table.insert( objects, 
  { "obj_int_panel_stage_progress_fight_fg",
    {
      pos_x = -private.screen_w / 2
    }
  } );
  
  table.insert( objects, 
  { "anm_int_panel_stage_progress_fight_fg_decoration",
    {
      pos_x     = private.screen_w / 2,
      scale_x   = bg_scale,
      scale_y   = bg_scale
    }
  } );
  
  table.insert( objects, 
  { _this.hero_touch,
    {
      pos_x        = -private.screen_w / 2,
      pos_y        = -offset,
      inputrect_w  =  private.screen_w / 2,
      inputrect_h  =  offset
    }
  } );
  
  table.insert( objects, 
  { _this.enemy_touch,
    {
      pos_x        =  0,
      pos_y        = -offset,
      inputrect_w  =  private.screen_w / 2,
      inputrect_h  =  offset
    }
  } );

  ui.obj.Multiset( objects );
  
end;
--------------------------------------------------------------------------------
function public.SetSmoothBorder ( is_smooth_border )
  
  local is_waves_visible = true;
  local is_grad_visible  = false;
  
  if ( is_smooth_border ) then
    
    is_waves_visible, is_grad_visible = is_grad_visible, is_waves_visible;
    
  end;
  
  ui.obj.Multiset( {
    { "obj_int_panel_stage_progress_fight_fg", 
      { visible = is_waves_visible } },
    { "grd_int_panel_stage_progress_fight_fg", 
      { visible = is_grad_visible  } },
  } );
  
end;
--------------------------------------------------------------------------------
function public.SetDistances ( near, far )
  
  private.dst_near = near;
  private.dst_far  = far;
  
end;
--------------------------------------------------------------------------------
function public.SetScale ( scale )
  
  private.fg_scale = scale;
  
end;
--------------------------------------------------------------------------------
function private.Reset ()

  private.WinAnimEnd = false;

  ui.skin.Detach( private.player_skin );

end;
--------------------------------------------------------------------------------
function private.SetColor ( color )

  local objects     = {};
  local parts_count = private.screen_w / 128;

  for i = 1, parts_count do

    table.insert( objects,
    { "spr_int_panel_stage_progress_fight_fg_decoration_"..i,
      {
        color_r = color.r,
        color_g = color.g,
        color_b = color.b } } );

  end;
  
  table.insert( objects,
  { "grd_int_panel_stage_progress_fight_fg",
    {
      color_r = color.r,
      color_g = color.g,
      color_b = color.b } } );
  
  ui.obj.Multiset( objects );

end;

--------------------------------------------------------------------------------
function private.SetDefaults ( stage_data, hero_data, enemy_data, multiset )

  private.WinAnimEnd    = false;
  private.enemy_ability = false;

  private.SetColor( stage_data.color );

  private.enemy_name = enemy_data.name;

  -- player name and skin
  private.player_skin = ui.skin.Attach(
    "anm_int_panel_stage_progress_fight_animations_enemy",
    hero_data.skin_name,
    "fight"
  );

  -- health buffers
  private.enemy_health = enemy_data.health;
  private.hero_health  = hero_data.health;

  -- detect zone name
  local bg_zone_name    = "arena";
  local bg_zone_content = "zones_1";
  local obj_zone_name    = "arena";
  local obj_zone_content = "zones_1";  

  -- enemy resources
  local enemy_content = enemy_data.content;
  local enemy_res     = "assets/content/"..enemy_content.."/"..enemy_data.name.."/anm_"..enemy_data.name;
  private.enemy_audio = "assets/content/"..enemy_content.."/"..enemy_data.name.."/aud_";

  local bg_name = "bg_"..bg_zone_name.."_fight";
  local decoration_res = "assets/content/"..obj_zone_content.."/anm_"..obj_zone_name.."_fight_objects";
  
  table.insert( multiset,
  { "obj_int_panel_stage_progress_fight_animations",
    {
      scale_x = private.fg_scale,
      scale_y = private.fg_scale,
      pos_y   = -50 * private.fg_scale
    }
  } );
  
  private.static_suffix = tern( ui.screen.IsLandscape() )(
    "_static_landscape",
    "_static"
  );

  table.insert( multiset,
  { _this.health.enemy,
    {
      animfunc = "enemy_healthbar"..private.static_suffix
    }
  } );
  
  table.insert( multiset,
  { _this.health.hero,
    {
      animfunc = "hero_healthbar"..private.static_suffix
    }
  } );
  
  table.insert( multiset,
  { _this.healthbar.enemy,
    {
      animfunc = "full",
      count    = tostring( math.floor( private.enemy_health ) )
    }
  } );
  
  table.insert( multiset,
  { _this.healthbar.hero,
    {
      animfunc = "full",
      count    = tostring( math.floor( private.hero_health ) )
    }
  } );
  
  ui.obj.Set( 
    "anm_int_panel_stage_progress_fight_animations_enemy",
    {
      res = enemy_res
    },
    multiset
  );
  
  ui.obj.Set(
    _this.attack_time.enemy,
    {
      res = enemy_res
    },
    multiset
  );
  
  ui.obj.Set(
    "spr_int_panel_stage_progress_fight_bg",
    {
      res = "assets/content/"..bg_zone_content.."/"..bg_name
    },
    multiset
  );

  ui.obj.Set(
    "anm_int_panel_stage_progress_fight_fg_decoration",
    {
      res      = decoration_res,
      animfunc = "decoration_fg",
      playing  = true
    },
    multiset
  );

  ui.obj.Enable( _this.obj, true, multiset );
  
  ui.obj.SetTextTag(
    _this.health.hero,
    "tt_name",
    hero_data.player_name,
    multiset
  );
  
  ui.obj.SetTextTag(
  _this.health.enemy,
    "tt_name",
    "str_enemy_name_"..enemy_data.name,
    multiset
  );
  
  ui.obj.Hide( _this.ability.hero,  multiset );
  ui.obj.Hide( _this.ability.enemy, multiset );
  
  ui.obj.Set( _this.icon.hero,
    {
      visible  = true,
      res      = ui.obj.Res( "skin", hero_data.skin_name ),
      animfunc = "marker"
    },
    multiset
  );
  
  ui.obj.Set( _this.icon.enemy,
    {
      visible = true,
      res     = ui.obj.Res( "enemy_icon", enemy_data.name )
    },
    multiset
  );
  
  ui.obj.Multiset( multiset );

  -- play static at the beginning
  private.Hero_Static();
  private.Enemy_Static();

  private.Enemy_Abilities_Hide();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.GameContinue ()
  
  ui.input.On({
    _this.enemy_touch,
    _this.hero_touch
  });
  
end;
--------------------------------------------------------------------------------
function public.Update ( action_data, func_TryContinue )

  if ( not private.active ) then return; end;

  if ( not action_data.delta  ) or
     ( action_data.delta == 0 ) then

    return;

  end;

  private.Attack_InsertToQueue( action_data, func_TryContinue );
  private.Attack_TryProcessFromQueue();

end;
--------------------------------------------------------------------------------
function public.Update_Ability ()

  private.Hero_SetAbility( nil );
  private.Enemy_SetAbility( nil );

end;
--------------------------------------------------------------------------------
function public.Update_ComboHit ( power )
  
  -- Ignore combo hit on already started attack
  if ( private.attack_pending ) then return; end;
  
  -- JK: prepare for a combo hit only if enemy is still alive
  if ( private.enemy_health > 0 ) then
  
    if ( private.combo_hit         ) and
       ( private.combo_hit > power ) then
    
      return;
  
    end;
    
    private.combo_hit = power;
    
    ui.anim.Play(
      private.player_skin,
      "start_attack_combo"..private.combo_hit,
      private.Hero_Prepare
    );
    
    common.sound.Play(
      private.hero_audio.."start_attack_combo"..private.combo_hit
    );
    
  end;

end;
--------------------------------------------------------------------------------
function public.Update_AttackPrepare ()
  
  private.Hero_Prepare();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Attack_IsQueueEmpty ()
  
  return ( #private.attack_queue < 1 );

end;
--------------------------------------------------------------------------------
function private.Attack_InsertToQueue ( action_data, func_TryContinue )
  
  table.insert(
    private.attack_queue,
    {
      action_data      = action_data,
      func_TryContinue = func_TryContinue
    }
  );
  
end;
--------------------------------------------------------------------------------
function private.Attack_TryProcessFromQueue ()
  
  if ( #private.attack_queue < 1 ) or
     ( private.attack_pending    ) then
    
    return;
    
  end;
  
  private.attack_pending = true;
  
  local action_data      = private.attack_queue[ 1 ].action_data;
  local func_TryContinue = private.attack_queue[ 1 ].func_TryContinue;
  
  local func_prefinish = function ()
    
    table.remove( private.attack_queue, 1 );
    
  end;
  
  local func_finish = function ()
    
    private.attack_pending = nil;
    private.Attack_TryProcessFromQueue();
    
  end;
  
  if ( action_data.type == "EnemyAttackActivated" ) then

    local func_OnAnimationEnd = function()

      func_prefinish();
      
      if ( private.hero_health < 1 ) then

        private.Enemy_Win( func_TryContinue );

      else

        func_TryContinue();

      end;
      
      func_finish();

    end;
    
    private.Enemy_Attack(
      action_data,
      private.Enemy_AttackStyle(),
      func_OnAnimationEnd
    );

  end;

  if ( action_data.type == "HeroAttackActivated" ) then

    local func_OnAnimationEnd = function()

      func_prefinish();
      
      if ( private.enemy_health < 1 ) then

        private.Hero_Win( func_TryContinue );

      else

        func_TryContinue();

      end;
      
      func_finish();

    end;
    
    private.Hero_Attack(
      action_data,
      func_OnAnimationEnd
    );    
    
  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.ShowDots ( left, right )
  
  ui.obj.Multiset( {
    { _this.ability.hero,  { animfunc = "state_"..left  } },
    { _this.ability.enemy, { animfunc = "state_"..right } }
  } );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function  public.Hero_Fun_Wheel ()
  
  common.sound.Play( private.hero_audio.."win_bonus" );
  
  private.Hero_OnTap();
  
end;
--------------------------------------------------------------------------------
function private.Hero_Prepare ()
  
  local animfunc = "prepare_attack1";
  
  if ( private.combo_hit ) then
  
    animfunc = "prepare_attack_combo"..private.combo_hit;
    
  end;
  
  ui.anim.Play(
    private.player_skin,
    animfunc,
    private.Hero_Prepare
  );
  
end;
--------------------------------------------------------------------------------
function private.Hero_Attack ( action_data, func_end )

  local attack_id = private.Hero_AttackStyle( action_data );
  
  local func_OnDefenceEnd = function()

    private.Enemy_Static();

    if ( action_data.ability ) then

      private.Enemy_SetAbility( action_data.ability );
      
      if ( action_data.ability == "stun" or action_data.ability == "superhit" ) then
        
        if ( action_data.ability == "superhit" ) then
          
          private.Hero_SetAbility( nil );
          
        end;
        
      end;

    end;

    func_end();

  end;

  local func_OnAttackEnd = function()

    --DbgTrace( "Enemy defends" );
    
    ui.anim.Play(
      "anm_int_panel_stage_progress_fight_animations_enemy",
      "defense",
      func_OnDefenceEnd
    );
    
    private.Enemy_SetHealth( action_data );

  end;

  if ( action_data.ability ) then

    private.Hero_SetAbility( action_data.ability );

  end;

  common.sound.Play( private.hero_audio.."attack"..attack_id );

  --DbgTrace( "Hero attacks" );
  
  ui.anim.Play(
    private.player_skin,
    "attack"..attack_id,
    private.Hero_Static
  );
  
  ui.anim.Play(
    _this.attack_time.hero,
    "attack_time"..attack_id,
    func_OnAttackEnd
  );  

end;
--------------------------------------------------------------------------------
function private.Hero_AttackStyle ( action_data )

  --Log.i( "AttackStyle_Hero: "..action_data.delta );
  
  if ( private.combo_hit ) then
    
    local combo_id    = private.combo_hit;
    private.combo_hit = nil;
    
    return "_combo"..combo_id;
    
  end;

  if ( action_data.ability                                    ) and
     ( private.Hero_AvailableAbilities( action_data.ability ) ) then

    return 4;

  end;

  if ( math.abs( action_data.delta ) > 0.5 * private.enemy_health ) then

    return 3;

  end;

  if ( math.abs( action_data.delta ) > 0.25 * private.enemy_health ) then

    return 2;

  end;

  return 1;

end;
--------------------------------------------------------------------------------
function private.Hero_Static()

  ui.anim.Play(
    private.player_skin,
    "static",
    private.Hero_Static
  );

end;
--------------------------------------------------------------------------------
function private.Hero_Win ( func_end )

  ui.input.Off({
    _this.enemy_touch,
    _this.hero_touch
  });
  
  ui.anim.Play(
    _this.attack_time.hero,
    "end_time",
    func_end
  );
  
  ui.anim.Play( private.player_skin, "win" );
  
  ui.anim.Play(
    "anm_int_panel_stage_progress_fight_animations_enemy",
    "fail",
    function ()

      mediator.Notify(
        "Interface_PanelStageProgress_WinAnimEnd",
        "fight"
      );

      private.WinAnimEnd = true;

    end
  );

  common.sound.Play( private.enemy_audio..private.enemy_name.."_die" );

end;
--------------------------------------------------------------------------------
function private.Hero_Respawn ( action_data )

  private.Hero_Static();
  private.Hero_SetHealth( action_data );
  private.Enemy_Static();

  common.sound.Play( private.hero_audio.."respawn" );

end;
--------------------------------------------------------------------------------
function private.Hero_SetHealth ( action_data )

  local before = private.hero_health;
  private.hero_health = action_data.health_current;
  
  local total   = action_data.health_total;
  local current = action_data.health_current;
  
  local before_frame = common.math.Round( before / total * private.count_frames_in_bar );
  local frame        = common.math.Round( current / total * private.count_frames_in_bar );
  
  ui.obj.Set(
    _this.healthbar.hero,
    { 
      count    = tostring( math.floor( current ) ),
      animfunc = "fill"
    }
  );
  
  local function func_notify ()
  
    local hero_pos = ui.obj.Position( _this.health.hero );
    local notify_data = 
    {
      time      = 1.5,
      offset    = 50,
      direction = "up", 
      posx      = ( hero_pos[ 1 ] + private.healthbar_w / 2 ), 
      posy      = hero_pos[ 2 ] + 100, 
      delta     = action_data.delta 
    };
  
    mediator.Notify(
      "Interface_PanelStageProgress_HealthChanged",
      notify_data
    );
  
  end;
  
  ObjAnimate( _this.healthbar.hero,
              "frame", 0, 0, func_notify, 
              {
                0.0, 0, before_frame,
                0.3, 0, frame
              } );

end;
--------------------------------------------------------------------------------
function private.Hero_SetAbility ( ability )

  local anm_char = "anm_int_panel_stage_progress_fight_animations_enemy_hero_ability";
  local anm_abil = _this.ability.hero;
  local anm_icon = "anm_int_panel_stage_progress_fight_healthbars_hero_ability_icon";
  local avatar   = _this.icon.hero;

  if ( not ability ) then

    if ( not private.hero_ability ) then

      return;

    end;

    private.hero_ability = false;

    ui.obj.Multiset( {
      { anm_char, { visible = false } },
      { anm_abil, { visible = false } },
      { anm_icon, { visible = false } },
      { avatar,   { visible = true  } }
    } );

    ObjAttach( anm_char, "hub_int_panel_stage_progress_fight_default" );

    if not ( ui.screen.IsLandscape() ) then
      
      local func_end = function ()

        ui.anim.Play(
          _this.health.hero,
          "hero_healthbar_static"
        );

      end;
      
      ui.anim.Play(
        _this.health.hero,
        "hero_ability_hide",
        func_end
      );
      
    end;

    return;

  end;

  if ( private.hero_ability ) then return; end;

  if ( private.Hero_AvailableAbilities( ability ) ) then

    private.hero_ability = true;

    ObjAttach( anm_char, private.player_skin );

    ui.obj.Multiset( {
      { anm_char,
        {
          visible  = true,
          animfunc = ability,
          playing  = true,
          anim_tag = ability } },
          
      { anm_abil,
        { 
          visible = true } },
          
      { anm_icon,
        {
          visible  = true,
          animfunc = "icon_"..ability,
          playing  = true } },
          
      { avatar,   { visible = false  } }
    } );
    
    if not ( ui.screen.IsLandscape() ) then
      
      local func_end = function ()

        ui.anim.Play(
          _this.health.hero,
          "hero_ability_static"
        );

      end;

      ui.anim.Play(
        _this.health.hero,
        "hero_ability_show",
        func_end
      );
        
    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Hero_AvailableAbilities ( ability )

  local abilities = {
    damage   = true,
    superhit = true
  };

  return abilities[ ability ];

end;
--------------------------------------------------------------------------------
function private.Hero_OnTap ()

  common.sound.Play( "hero_touch" );
  
  ui.input.Off( _this.hero_touch );

  local func_OnAnimationEnd = function ()

    private.Hero_Static();

    ui.input.On( _this.hero_touch );

  end;

  ui.anim.Play( private.player_skin, "fun1", func_OnAnimationEnd );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Enemy_Attack ( action_data, attack_index, func_end )

  local func_OnAnimationEnd = function()

    --DbgTrace( "Enemy attack ended" );
    
    private.Enemy_Static();

    func_end();

  end;

  local func_OnDefenseEnd = function()

    private.Hero_Static();

    if ( private.hero_health < 1 ) then

      func_end();

    end;

  end;

  local func_OnAttackEnd = function()

    --DbgTrace( "Hero defends" );
    
    ui.anim.Play(
      private.player_skin,
      "defense",
      func_OnDefenseEnd
    );
    
    private.Hero_SetHealth( action_data );

  end;

  --DbgTrace( "Enemy attacks" );
  
  common.sound.Play(
    private.enemy_audio..private.enemy_name.."_attack"..attack_index
  );

  ui.anim.Play(
    "anm_int_panel_stage_progress_fight_animations_enemy",
    "attack"..attack_index,
    func_OnAnimationEnd
  );
  
  ui.anim.Play(
    _this.attack_time.enemy,
    "attack_time"..attack_index,
    func_OnAttackEnd
  );

end;
--------------------------------------------------------------------------------
function private.Enemy_AttackStyle ()

  return math.random( 1, 3 );

end;
--------------------------------------------------------------------------------
function private.Enemy_Static ()

  ui.anim.Play(
    "anm_int_panel_stage_progress_fight_animations_enemy",
    "static",
    private.Enemy_Static
  );

end;
--------------------------------------------------------------------------------
function private.Enemy_Win ( func_end )

  ui.input.Off({
    _this.enemy_touch,
    _this.hero_touch
  });
  
  ui.anim.Play(
    _this.attack_time.hero,
    "end_time",
    func_end
  );
  
  ui.anim.Play(
    "anm_int_panel_stage_progress_fight_animations_enemy",
    "win"
  );
  
  ui.anim.Play(
    private.player_skin,
    "fail"
  );
  
  common.sound.Play( private.enemy_audio..private.enemy_name.."_win" );

end;
--------------------------------------------------------------------------------
function private.Enemy_SetHealth ( action_data )

  local before = private.enemy_health;
  private.enemy_health = action_data.health_current;
  
  local total   = action_data.health_total;
  local current = action_data.health_current;
  
  local before_frame = common.math.Round( before / total * private.count_frames_in_bar );
  local frame        = common.math.Round( current / total * private.count_frames_in_bar );
  
  ui.obj.Set(
    _this.healthbar.enemy,
    { 
      count    = tostring( math.floor( current ) ),
      animfunc = "fill"
    }
  );
  
  local function func_notify ()
  
    local enemy_pos   = ui.obj.Position( _this.health.enemy );
    local notify_data = 
    { 
      time      = 1.5,
      offset    = 50,
      direction = "up", 
      posx      = ( enemy_pos[ 1 ] - private.healthbar_w / 2 ),
      posy      = enemy_pos[ 2 ] + 100, 
      delta     = ( action_data.delta - ( action_data.ability_delta or 0 ) )
    };

    local notify_data_ability = nil;  

    if ( action_data.ability_delta ) then

      notify_data_ability = 
      { 
        time        = 1.5,
        offset      = 50,
        direction   = "up",
        posx        = ( enemy_pos[ 1 ] - private.healthbar_w / 2 ), 
        posy        = enemy_pos[ 2 ] + 100, 
        delta       = action_data.ability_delta, 
        fontcolor_r = 0.5, 
        fontcolor_g = 1, 
        fontcolor_b = 0.5 
      };
    
    end;
    
    mediator.Notify(
      "Interface_PanelStageProgress_HealthChanged",
      notify_data
    );

    if ( notify_data_ability ) then

      mediator.Notify(
        "Interface_PanelStageProgress_HealthChanged",
        notify_data_ability
      );

    end;
  
  end;

  ObjAnimate( _this.healthbar.enemy,
              "frame", 0, 0, func_notify, 
              {
                0.0, 0, before_frame,
                0.3, 0, frame
              } );
  
end;
--------------------------------------------------------------------------------
function private.Enemy_SetAbility ( ability )

  local anm_char = "anm_int_panel_stage_progress_fight_animations_enemy_ability_"..( ability or private.enemy_ability or "" );
  local anm_abil = _this.ability.enemy;
  local anm_icon = "anm_int_panel_stage_progress_fight_healthbars_enemy_ability_icon";
  local avatar   = _this.icon.enemy;
  
  if ( not ability ) then

    if ( not private.enemy_ability ) then return; end;

    private.enemy_ability = false;

    ObjAnimate( anm_char, "alp", 0, 0, function ()

      ui.input.On( _this.enemy_touch );
      ui.obj.Set(
        anm_char,
        {
          visible = false,
          alp     = 1.0
        }
      );

    end, 
    { 
      0.0, 0, 1.0, 
      0.3, 0, 0.0 
    } );

    ui.obj.Multiset( {
      { anm_abil, { visible = false } },
      { anm_icon, { visible = false } },
      { avatar,   { visible = true  } }
    } );

    if not ( private.enemy_health < 1 ) then

      private.Enemy_Static();

    end;
    
    if ( not ui.screen.IsLandscape() ) then
      
      ui.anim.Play(
        _this.health.enemy,
        "enemy_ability_hide",
        nil,
        "enemy_healthbar_static"
      );
      
    end;

    return;

  end;
  
  local ability_data = private.Enemy_AvailableAbilities( ability );

  if ( ability_data and ability_data.available ) then
    
    if ( ability_data.animated ) then
      ui.anim.Play( "anm_int_panel_stage_progress_fight_animations_enemy",
        ability
      );
    end;

    if ( ability == "stun" ) then
      ui.input.Off( _this.enemy_touch );
    end;
    
    if ( private.enemy_ability ) then return; end;

    private.enemy_ability = ability;

    ui.obj.Multiset( {
      { anm_char,
        {
          visible  = true,
          animfunc = ability,
          playing  = true } },
          
      { anm_abil,
        { visible  = true } },
          
      { anm_icon,
        {
          visible  = true,
          animfunc = "icon_"..ability,
          playing  = true } },
          
      { avatar,   { visible = false  } }
    } );
    
    if ( not ui.screen.IsLandscape() ) then

      ui.anim.Play(
        _this.health.enemy,
        "enemy_ability_show",
        nil,
        "enemy_ability_static"
      );
      
    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Enemy_AvailableAbilities ( ability )

  local abilities = {
    stun = { available = true, animated = true  },
    dot  = { available = true, animated = false }
  };

  return abilities[ ability ];

end;
--------------------------------------------------------------------------------
function private.Enemy_Abilities_Hide ()

  local abilities = {
    stun = true,
    dot  = true
  };
  
  local multiset = {};

  for key, value in pairs( abilities ) do

    ui.obj.Hide(
      "anm_int_panel_stage_progress_fight_animations_enemy_ability_"..key,
      multiset
    );

  end;

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.Enemy_OnTap ()

  common.sound.Play(
    private.enemy_audio..private.enemy_name.."_tap",
    nil,
    nil,
    "enemy_click"
  );
  
  ui.input.Off( _this.enemy_touch );

  ui.anim.Play(
    "anm_int_panel_stage_progress_fight_animations_enemy",
    "defense",
    function ()
      
      ui.input.On( _this.enemy_touch );
      
    end,
    "static"
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Move_Near ( time )
  
  ui.input.Off( "obj_int_panel_stage_progress_fight_touch" );
  
  local near, far = private.dst_near, private.dst_far;
  local offset    = private.dst_off; -- internal animation objects offset
  
  ObjAnimate( "anm_int_panel_stage_progress_fight_animations_enemy", 
              "pos_x", 0, 0, "", 
              { 
                0.0,  0, 0.5 * ( far  - offset ), 
                time, 0, 0.5 * ( near - offset )
              } );
  
  ObjAnimate( "anm_int_panel_stage_progress_fight_animations_enemy_hero", 
              "pos_x", 0, 0, "", 
              { 
                0.0,  0, -far  + offset, 
                time, 0, -near + offset
              } );
              
end;
--------------------------------------------------------------------------------
function public.Move_Far ( time )
  
  local near, far = private.dst_near, private.dst_far;
  local offset    = private.dst_off; -- internal animation objects offset
  
  local function func_after_move ()
  
    ui.input.On( "obj_int_panel_stage_progress_fight_touch" );
  
  end;
  
  ObjAnimate( "anm_int_panel_stage_progress_fight_animations_enemy",
              "pos_x", 0, 0, "", 
              { 
                0.0,  0, 0.5 * ( near - offset ),
                time, 0, 0.5 * ( far  - offset )
              } );
  
  ObjAnimate( "anm_int_panel_stage_progress_fight_animations_enemy_hero",
              "pos_x", 0, 0, func_after_move,
              { 
                0.0,  0, -near + offset,
                time, 0, -far  + offset
              } );
  
end;
--------------------------------------------------------------------------------