-- name=int_panel_stage_progress_ride
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.active     = false;
  private.screen_w   = GetAppWidth();
  private.screen_h   = GetAppHeight();
  private.fg_scale   = common.math.Round( private.screen_h / 1024 );
  private.hero_audio = "assets/content/hero/aud_hero_";
  private.anim_time  = 0.3;
  
end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  ui.hub.CreateInStage( "int_panel_stage_progress_ride" );

  private.CreateDecorations( 36, 16 );

  ui.control.TapSet( "obj_int_panel_stage_progress_ride_animations",
  {
    func_true = function ()

      if ( ObjGet( private.player_skin ).animfunc == "static" ) then

        private.Hero_Fun_Click();

      end;

    end
  } );

end;
--------------------------------------------------------------------------------
function public.Show ( stage_data, skin_current )

  if ( private.active ) then

    private.Reset();

  end;

  ui.hub.AttachContent( "int_panel_stage_progress_ride" );

  private.SetDefaults( stage_data, skin_current );

  private.active = true;
 
end;
--------------------------------------------------------------------------------
function public.Hide ()

  private.active = false;

  ObjSet( "int_panel_stage_progress_ride", { active = false, visible = false, input = false } );
  ui.hub.DetachContent( "int_panel_stage_progress_ride" );
  private.Reset();

end;
--------------------------------------------------------------------------------
function public.IsWinAnimEnded ()

  return private.WinAnimEnd;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.CreateDecorations ( near_cnt, far_cnt )

  -- Setup disks
  private.disk      = {};
  private.disk[ 1 ] = {};
  private.disk[ 2 ] = {};

  private.disk[ 1 ].obj_cnt = near_cnt;
  private.disk[ 1 ].obj_max = 10;

  private.disk[ 2 ].obj_cnt = far_cnt;  
  private.disk[ 2 ].obj_max = 10;

  for i = 1, private.disk[ 1 ].obj_cnt do
    
    ObjCreate( "anm_int_panel_stage_progress_ride_disk1_obj"..i, "anim" );
    ObjAttach( "anm_int_panel_stage_progress_ride_disk1_obj"..i, "obj_int_panel_stage_progress_ride_disk1" );

  end;

  for i = 1, private.disk[ 2 ].obj_cnt do

    ObjCreate( "anm_int_panel_stage_progress_ride_disk2_obj"..i, "anim" );
    ObjAttach( "anm_int_panel_stage_progress_ride_disk2_obj"..i, "obj_int_panel_stage_progress_ride_disk2" );

  end;

  -- friends
  private.friends_buffer = 4;
--
--  for i = 1, private.friends_buffer do
--
--    ObjClone( "obj_int_panel_stage_progress_ride_friend", "_"..i, true, "obj_int_panel_stage_progress_ride_miles" );
--
--  end;

  -- FIXME: object generation at init, non adaptive
  -- Make decoration border ( vertical layout )
  local parts_count = private.screen_w / 128;

  local multiset_table = {};
  
  table.insert( multiset_table, 
    { 
      "obj_int_panel_stage_progress_ride_decoration",  
      { pos_x = - 0.5 * private.screen_w }
    } );

  for i = 0, parts_count do

    ObjClone( "spr_int_panel_stage_progress_ride_decoration", "_"..i, 1, "obj_int_panel_stage_progress_ride_decoration" );

    table.insert( multiset_table, 
    {
    "spr_int_panel_stage_progress_ride_decoration_"..i, { pos_x = 128 * ( i - 0.5 ) }
    } );

  end;

  ObjMultiSet( multiset_table );
  
  -- Resize decoartion grid ( horizontal layout )
  local grd_size = {
    left   = private.screen_w / 2,
    right  = private.screen_w / 2,
    bottom = 0,
    top    = 40
  }
  
  local grd_border = {
    left = 0,
    right = 0,
    bottom = 0,
    top = 40
  }
  
  ui.grid.Resize(
    "grd_int_panel_stage_progress_ride_decoration",
    grd_size,
    grd_border
  );
  
end;
--------------------------------------------------------------------------------
function public.SetSmoothBorder ( is_smooth_border )
  
  local is_waves_visible = true;
  local is_grad_visible  = false;
  
  if ( is_smooth_border ) then
    
    is_waves_visible, is_grad_visible = is_grad_visible, is_waves_visible;
    
  end;
  
  ObjMultiSet( {
    { "obj_int_panel_stage_progress_ride_decoration", 
      { visible = is_waves_visible } },
    { "grd_int_panel_stage_progress_ride_decoration", 
      { visible = is_grad_visible  } },
  } );
  
end;
--------------------------------------------------------------------------------
function public.SetBgOffset ( offset )
  
  local bg_w, bg_h = 768, 340; -- BG texture size
  
  local bg_scale_h = offset / bg_h;
  local bg_scale_w = private.screen_w / bg_w;
  
  local bg_scale = common.math.Round( math.max( bg_scale_h, bg_scale_w ), 2 );

  ObjMultiSet( {
    { "obj_int_panel_stage_progress_ride",
      { 
        croprect_h =  offset,
        croprect_y = -offset,
        croprect_w =  private.screen_w,
        croprect_x = -private.screen_w / 2 
      }
    },
    { "spr_int_panel_stage_progress_ride_bg",
      { 
        drawoff_y = -0.5 * bg_h,
        scale_x   = bg_scale, 
        scale_y   = bg_scale
      }
    }
  } );
  
end;
--------------------------------------------------------------------------------
function public.SetFgPos ( x, y )
  
  private.pos_y  = y;
  x = x - ( private.screen_w / 2 );

  ObjMultiSet( {
    { "obj_int_panel_stage_progress_ride_fg",          { pos_x = x } },
    { "anm_int_panel_stage_progress_ride_fg_win",      { pos_y = y } },
    { "anm_int_panel_stage_progress_ride_environment",
      { 
        active  = false,
        visible = false
      }
    }
  } );
  
end;
--------------------------------------------------------------------------------
function public.SetScale ( scale )
  
  private.fg_scale = scale;
  
end;
--------------------------------------------------------------------------------
function private.SetDefaults ( stage_data, skin_current )

  private.WinAnimEnd = false;

  private.Color_Set( stage_data.color );

  -- player skin
  private.player_skin = ui.skin.Attach(
      "obj_int_panel_stage_progress_ride_animations",
      skin_current
    );

  local objects = {};

  -- detect cycles
  private.miles_cycles = 1;

--  if ( stage_data.state > 1 ) then
--
--    private.miles_cycles = 2;
--
--  end;

  -- detect cycle distance 
  private.cycle_distance = 0;
  
  local targets = stage_data.targets;

  private.total_score = 0;

  for t = 1, #targets do
    
    if ( targets[ t ].type == "GainScore" ) then

      private.cycle_distance = targets[ t ].count / private.miles_cycles;

      private.total_score = targets[ t ].count;

    end;

  end;

  -- detect zone name
  local bg_zone_name    = stage_data.zone.background.name;
  local bg_zone_content = stage_data.zone.background.content;
  local obj_zone_name    = stage_data.zone.objects.name;
  local obj_zone_content = stage_data.zone.objects.content;
  
  local obj_res_path = "assets/content/"..obj_zone_content..
                       "/anm_"..obj_zone_name.."_ride_objects";
  local bg_res_path = "assets/content/"..bg_zone_content..
                      "/bg_"..bg_zone_name.."_ride";
  
  local spec_effects = {
    bay    = { name = "wave", pos_z = -2, pos_y = 50 },
    beach  = { name = "wave", pos_z = -2, pos_y = 50 },
    forest = { name = "light_sun", pos_z = 2, pos_y = 50 }
  };
  -- waves for bay
  if ( spec_effects[ bg_zone_name ] ) then
    
    local effect_data = spec_effects[ bg_zone_name ];

    table.insert( objects, { "anm_int_panel_stage_progress_ride_environment",
    {
      active   = true,
      visible  = true,
      res      = obj_res_path,
      animfunc = effect_data.name,
      playing  = true,
      pos_z    = effect_data.pos_x,
      pos_y    = effect_data.pos_y
    } } );

  else

    table.insert( objects, { "anm_int_panel_stage_progress_ride_environment", { active = false, visible = false } } );

  end;

  -- milestones
  private.score_queue       = {};
  private.milestones_cycles = 0;
  private.milestones_map    = {};
  private.milestones_locked = false;
  private.score_buffer      = 0;
  private.cycle_buffer      = 0;

  -- popup, stage & targets status
  private.popup_hit       = false;
  private.target_complete = false;
  private.stage_complete  = false;
  private.stage_failed    = false;
  private.bonus_run       = false;

  -- disk objects
  private.disk[ 1 ].objects = {}; 
  private.disk[ 2 ].objects = {};

  -- speedup
  private.hero_speedup = 0;

  -- static cycles
  private.static_cycles = 0;
  
  -- speeds
  private.world_speed_pace = 7;
  private.world_speed_slow = 9;
  private.world_speed_stop = 12;

  private.World_Stop();
  private.Miles_Stop();

  ObjStopAnimate( "obj_int_panel_stage_progress_ride_fg", "scale_xy" );
  ObjStopAnimate( "obj_int_panel_stage_progress_ride_fg", "ang" );
  ObjStopAnimate( "obj_int_panel_stage_progress_ride_fg", "pos_y" );

  AnimStop( "anm_int_panel_stage_progress_ride_bg_win" );
  AnimStop( "anm_int_panel_stage_progress_ride_fg_win" );

  table.insert( objects, { "anm_int_panel_stage_progress_ride_bg_win", 
    { active = false, visible = false, animfunc = "" } } );
  table.insert( objects, { "anm_int_panel_stage_progress_ride_fg_win", 
    { active = false, visible = false, animfunc = "" } } );
  
  for i = 1, private.disk[ 1 ].obj_cnt do

    table.insert( objects, { "anm_int_panel_stage_progress_ride_disk1_obj"..i, 
      { res = obj_res_path } } );

  end;

  for i = 1, private.disk[ 2 ].obj_cnt do

    table.insert( objects, { "anm_int_panel_stage_progress_ride_disk2_obj"..i, 
      { res = obj_res_path } } );

  end;

  table.insert( objects, { "anm_int_panel_stage_progress_ride_disk1_ground", 
    { res = obj_res_path, animfunc = "base_ground" } } );
  table.insert( objects, { "spr_int_panel_stage_progress_ride_bg",           
    { res = bg_res_path } } );

  table.insert( objects, { "obj_int_panel_stage_progress_ride_fg",
  { ang     = 0,
    scale_x = private.fg_scale,
    scale_y = private.fg_scale,
    pos_y   = -150 * private.fg_scale
  } } );

  table.insert( objects, { "obj_int_panel_stage_progress_ride_disk1",      { ang = 0 } } );
  table.insert( objects, { "obj_int_panel_stage_progress_ride_disk2",      { ang = 0 } } );
  table.insert( objects, { "obj_int_panel_stage_progress_ride_miles",      { ang = 0 } } );
  table.insert( objects, { "obj_int_panel_stage_progress_ride_milestone1", { active = false, visible = false, alp = 1.0 } } );
  table.insert( objects, { "obj_int_panel_stage_progress_ride_milestone2", { active = true, visible = true  } } );
  table.insert( objects, { "obj_int_panel_stage_progress_ride_milestone3", { active = true, visible = true  } } );
  table.insert( objects, { "obj_int_panel_stage_progress_ride_milestone4", { active = true, visible = true  } } );

  ObjMultiSet( objects );

  private.Miles_Set();
  private.World_Create( 1 );
  private.World_Create( 2 );

  -- friends reset
  --private._Friends_Hide();
  private.friend_drop = {};

  ObjSet( "int_panel_stage_progress_ride", 
  { 
    input   = true,
    active  = true,
    visible = true, 
    pos_x   = private.screen_w / 2,
    pos_y   = private.pos_y
  } );

  private.Hero_Static();

end;
--------------------------------------------------------------------------------
function private.Reset ()

  private.WinAnimEnd = false;

  ui.skin.Detach( private.player_skin );

  ObjStopAnimate( "obj_int_panel_stage_progress_ride_fg", "scale_xy" );
  ObjStopAnimate( "obj_int_panel_stage_progress_ride_fg", "ang" );
  ObjStopAnimate( "obj_int_panel_stage_progress_ride_fg", "pos_y" );

  AnimStop( "anm_int_panel_stage_progress_ride_bg_win" );
  AnimStop( "anm_int_panel_stage_progress_ride_fg_win" );

  ObjMultiSet( {
    { "anm_int_panel_stage_progress_ride_bg_win", { active = false, visible = false, animfunc = "" } },
    { "anm_int_panel_stage_progress_ride_fg_win", { active = false, visible = false, animfunc = "" } }
  } );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Hero_Start ()

  private.static_cycles = 0;

  ui.anim.Play( private.player_skin, "start", private.Hero_Dynamic );
  private.World_Rotate( private.world_speed_pace );

end;
--------------------------------------------------------------------------------
function private.Hero_Dynamic ()

  if not private.user_active then

    private.Hero_Slow();   

  else

    if ( private.popup_hit ) then

      private.Hero_PopupHit();

    elseif ( private.target_complete ) then

      private.Hero_Target();

    elseif ( private.stage_complete ) or ( private.stage_failed ) then

      private.Hero_Stop();

--    elseif ( private.bonus_run ) then
--
--      private.Hero_Run( 3 );

    else

      ui.anim.Play( private.player_skin, "dynamic1", private.Hero_Dynamic );
      private.World_Rotate( private.world_speed_pace );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Hero_Slow ()

  local func_OnAnimationEnd = function()

    if ( private.stage_complete ) or ( private.stage_failed ) then

      private.Hero_Stop();

      return;

    end;

    if not private.user_active then

      private.Hero_Stop();

    else

      private.Hero_Dynamic();  

    end;

  end;

  ui.anim.Play( private.player_skin, "slow", func_OnAnimationEnd );
  private.World_Rotate( private.world_speed_slow );

end;
--------------------------------------------------------------------------------
function private.Hero_Stop ()

  if ( private.stage_complete ) then

    local func_end = function()
        
      --private.Hero_Win();

    end;

    private.World_Win();
    
    ui.anim.Play( private.player_skin, "stop", func_end );
    private.World_Rotate( private.world_speed_stop );

     common.sound.Play( private.hero_audio.."win_zoom" );

  elseif ( private.stage_failed ) then

    local func_end = function()

      private.World_Stop();
      private.Hero_Fail();

    end;    

    ui.anim.Play( private.player_skin, "stop", func_end );
    private.World_Rotate( private.world_speed_stop );

  else

    local func_end = function()
        
      private.World_Stop();
      private.Hero_Static();      

    end;
    
    ui.anim.Play( private.player_skin, "stop", func_end );
    private.World_Rotate( private.world_speed_stop );

  end;

end;
--------------------------------------------------------------------------------
function private.Hero_Static ()

  private.static_cycles = private.static_cycles + 1;

  if ( private.static_cycles > 1 ) then

    private.Hero_Fun_Cycle();

    return;

  end;

  ui.anim.Play( private.player_skin, "static", private.Hero_Static );

end;
--------------------------------------------------------------------------------
function private.Hero_Target ()

   ui.anim.Play( private.player_skin, "target", private.Hero_Dynamic );
   private.World_Rotate( private.world_speed_pace );

   private.target_complete = false;

end;
--------------------------------------------------------------------------------
function private.Hero_Win ( func_end )

  private.score_queue = {};

  private.World_Stop();
  private.Miles_Stop();

  ObjMultiSet( {
    { "anm_int_panel_stage_progress_ride_bg_win", { active = true, visible = true } },
    { "anm_int_panel_stage_progress_ride_fg_win", { active = true, visible = true } }
  } );

--  local func_OnPauseEnd = function ()
--
--     common.sound.Play( "level_completed" );
--
--  end;
--
--  ObjAnimate( "obj_int_panel_stage_progress_ride_decoration", "alp", 0, 0, func_OnPauseEnd, { 0,0,1, 2,0,1 } );

  --[[ui.anim.Play( private.player_skin, "win", function ()

    mediator.Notify( "Interface_PanelStageProgress_WinAnimEnd", "ride" );

    private.WinAnimEnd = true;

  end );]]

  ui.anim.Play( private.player_skin, "run4" );

  ObjAnimate( private.player_skin, "pos_x", 0, 0, function ()
  
    ui.anim.Play( private.player_skin, "static" );

    mediator.Notify( "Interface_PanelStageProgress_WinAnimEnd", "ride" );

    private.WinAnimEnd = true;
    
    if ( func_end ) then
      
      func_end();
      
    end;

  end, 
  { 
    0.0, 0, ( ObjGet( private.player_skin ).pos_x or 0 ), 
    0.6, 0, 0.5 * private.screen_w + 500
  } );

  ui.anim.Play(
    "anm_int_panel_stage_progress_ride_bg_win",
    "fx_win_bg_start"
  );

  ui.anim.Play(
    "anm_int_panel_stage_progress_ride_fg_win",
    "fx_win_fg_start",
    nil,
    "fx_win_fg_static"
  );

  private.stage_complete = false;

end;
--------------------------------------------------------------------------------
function private.Hero_Fail ()

  ui.anim.Play( private.player_skin, "fail" );

end;
--------------------------------------------------------------------------------
function private.Hero_Respawn ()
  
  private.Hero_Static();
  private.World_Stop();
  private.stage_failed = false;

end;
--------------------------------------------------------------------------------
function private.Hero_Fun_Click ()

  private.static_cycles = 0;
  
  ui.anim.Play( private.player_skin, "fun1", private.Hero_Static );

end;
--------------------------------------------------------------------------------
function public.Hero_Fun_Wheel ()

   common.sound.Play( private.hero_audio.."win_bonus" );

  ui.anim.Play( private.player_skin, "fun1", private.Hero_Static );

end;
--------------------------------------------------------------------------------
function private.Hero_Fun_Cycle ()

  private.static_cycles = 0;

  local index = math.random( 1, 4 );
  
  ui.anim.Play( private.player_skin, "fun"..index, private.Hero_Static );

end;
--------------------------------------------------------------------------------
function private.Hero_PopupHit ()
  
  ui.anim.Play( private.player_skin, "dynamic2", private.Hero_Dynamic );
  private.World_Rotate( private.world_speed_pace );

  private.popup_hit = false;

end;
--------------------------------------------------------------------------------
function private.Hero_Run ( speed )

  private.hero_speedup = private.hero_speedup + speed;

  if ( private.hero_speedup > 4 ) then

    private.hero_speedup = 4;

  end;

  if ( private.hero_speedup > 1 ) then

    private.World_Run( private.hero_speedup );

  end;

  ui.anim.Play( private.player_skin, "run"..private.hero_speedup, function ()

    private.Hero_Retard( private.hero_speedup );

  end );

  private.World_Rotate( private.world_speed_pace - private.hero_speedup, true );

  if ( private.hero_speedup == 3 ) then

     common.sound.Play( private.hero_audio.."run_fast" );

  end;

  if ( private.hero_speedup == 4 ) then

     common.sound.Play( private.hero_audio.."fly_fast" );

  end;

end;
--------------------------------------------------------------------------------
function private.Hero_Retard ( speed )

  private.hero_speedup = 0;

  if ( private.bonus_run ) then

    ui.anim.Play( private.player_skin, "run2", function () private.Hero_Retard( 2 ); end );

    private.World_Retard();
    private.World_Rotate( private.world_speed_pace - 2, true );

  elseif ( speed > 1 ) then

    ui.anim.Play( private.player_skin, "run"..( speed - 1 ), function () private.Hero_Retard( speed - 1 ); end );

    private.World_Retard();
    private.World_Rotate( private.world_speed_pace - ( speed - 1 ), true );

  else

    ui.anim.Play( private.player_skin, "retard", private.Hero_Dynamic );

    private.World_Retard();
    private.World_Rotate( private.world_speed_pace, nil, 2 );

  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.World_Create ( disk_index )

  local disk  = ObjGet( "obj_int_panel_stage_progress_ride_disk"..disk_index );
  disk.radius = disk.inputrect_w / 2;

  local sector = math.floor( 2 * math.pi * disk.radius / private.disk[ disk_index ].obj_cnt );
  
  for i = 1, private.disk[ disk_index ].obj_cnt do

    local ang   = - ( math.random( sector * ( i - 1 ), sector * i ) / disk.radius );
    local pos_x = math.sin( ang ) * disk.radius;
    local pos_y = math.cos( ang ) * disk.radius;

    private.disk[ disk_index ].objects[ i ] = ang;
    
    local obj_index = math.random( 1, private.disk[ disk_index ].obj_max );

    ObjSet( "anm_int_panel_stage_progress_ride_disk"..disk_index.."_obj"..i,
    {    
      pos_x    = pos_x,
      pos_y    = pos_y,
      ang      = ang + math.pi,
      animfunc = "disk"..disk_index.."_obj"..obj_index,
      playing  = true
    } );

  end;

end;
--------------------------------------------------------------------------------
function private.World_Rotate ( time, turbo, mode )

  private.World_Stop();

  local disk_2 = ObjGet( "obj_int_panel_stage_progress_ride_disk2" );
  local disk_1 = ObjGet( "obj_int_panel_stage_progress_ride_disk1" );

  local acceleration = 1;
  local anm_mode     = mode or 0;

  if ( turbo ) then

    acceleration = 2 + ( 1 - time / 10 );

  end;

  disk_2.step = acceleration * math.pi / time / 4;
  disk_1.step = acceleration * math.pi / time;

  ObjAnimate( "obj_int_panel_stage_progress_ride_disk2", "ang", 0, 0, "", 
    {
      0,    0,        disk_2.ang, 
      time, anm_mode, disk_2.ang + disk_2.step
    }
  );

  ObjAnimate( "obj_int_panel_stage_progress_ride_disk1", "ang", 0, 0, "", 
    {
      0,    0,        disk_1.ang, 
      time, anm_mode, disk_1.ang + disk_1.step
    }
  );

end;
--------------------------------------------------------------------------------
function private.World_Stop ()

  ObjStopAnimate( "obj_int_panel_stage_progress_ride_disk1", "ang" );
  ObjStopAnimate( "obj_int_panel_stage_progress_ride_disk2", "ang" );

end;
--------------------------------------------------------------------------------
function private.World_Win ()
  
  ObjAnimate( "obj_int_panel_stage_progress_ride_fg", "scale_xy", 0, 0, "",
  {
    0,    0, 1.00 * private.fg_scale, 1.00 * private.fg_scale,
    0.75, 0, 1.20 * private.fg_scale, 1.20 * private.fg_scale,
    1.5,  0, 0.75 * private.fg_scale, 0.75 * private.fg_scale 
  } );

  ObjAnimate( "obj_int_panel_stage_progress_ride_fg", "ang", 0, 0, "",
  {
    0,    0, 0.0,
    0.75, 0, 0.1,
    1.5,  0, 0.0
  } );

  ObjAnimate( "obj_int_panel_stage_progress_ride_fg", "pos_y", 0, 0, "",
  {
    0,    0, 20 * private.fg_scale,
    0.75, 0, 40 * private.fg_scale,
    1.5,  0, 80 * private.fg_scale 
  } );

end;
--------------------------------------------------------------------------------
function private.World_Run ( speedup )

  local obj = ObjGet( "obj_int_panel_stage_progress_ride_fg" );

  local scale_x = ( 1.1 + 0.01 * speedup ) * private.fg_scale;
  local scale_y = ( 1.1 + 0.01 * speedup ) * private.fg_scale;

  ObjAnimate( "obj_int_panel_stage_progress_ride_fg", "scale_xy", 0, 0, "",
  {
    0.0, 0, obj.scale_x, obj.scale_y,
    0.5, 0, scale_x,     scale_y
  } );

  ObjAnimate( "obj_int_panel_stage_progress_ride_fg", "ang", 0, 0, "",
  {
    0.0, 0, obj.ang,
    0.5, 0, obj.ang + ( 0.015 * speedup )
  } );

end;
--------------------------------------------------------------------------------
function private.World_Retard ()

  local obj = ObjGet( "obj_int_panel_stage_progress_ride_fg" );

  local scale_x = 1.0 * private.fg_scale;
  local scale_y = 1.0 * private.fg_scale;

  ObjAnimate( "obj_int_panel_stage_progress_ride_fg", "scale_xy", 0, 0, "",
  {
    0.0, 0, obj.scale_x, obj.scale_y,
    0.6, 0, scale_x,     scale_y
  } );

  ObjAnimate( "obj_int_panel_stage_progress_ride_fg", "ang", 0, 0, "",
  {
    0.0, 0, obj.ang,
    0.6, 0, 0
  } );
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Miles_Set ()

  local disk  = ObjGet( "obj_int_panel_stage_progress_ride_miles" );
  disk.radius = disk.inputrect_w / 2;
  disk.length = 2 * math.pi * disk.radius;

  local ratio = private.total_score / disk.length;

  private.Miles_Map_Create( 0 );

  local objects = {};
  
  for i = 1, 4 do

    local ang = - ( private.milestones_map[ i ].score / ratio ) / disk.radius ;

    local pos_x = math.sin( ang ) * disk.radius;
    local pos_y = math.cos( ang ) * disk.radius;

    --DbgTrace(i.." "..ang.." "..pos_x.." "..pos_y);

    table.insert( objects, { "obj_int_panel_stage_progress_ride_milestone"..i, { pos_x = pos_x, pos_y = pos_y, ang = ang + math.pi } } );
    
  end;

  ObjMultiSet( objects );

  private.Miles_Map_Create( 1 );
  private.Miles_Map_Create( 2 );
  private.Miles_Map_Create( 3 );

  private.Miles_Update( 1, 0 );

end;
--------------------------------------------------------------------------------
function private.Miles_Map_Create ( cycle )

  local segment = private.cycle_distance / 4;

  for i = 1, 4 do

    local checkpoint = {};

    checkpoint.score = math.ceil( segment * ( i + cycle * 4 ) - segment );
    checkpoint.cycle = cycle;
    checkpoint.index = i; 

    table.insert( private.milestones_map, checkpoint );

  end;

end;
--------------------------------------------------------------------------------
function private.Miles_Map_Queue ( score )

  local map_queue = {};

  for i = 1, #private.milestones_map do

    if ( private.milestones_map[ i ].score > private.score_buffer ) and ( private.milestones_map[ i ].score <= score ) then

      table.insert( map_queue, private.milestones_map[ i ] );

    end;

  end;

  return map_queue;

end;
--------------------------------------------------------------------------------
function private.Miles_Map_Trim ()
  
  table.remove( private.milestones_map, 1, 4 );

end;
--------------------------------------------------------------------------------
function private.Miles_Match ( score )

  --DbgTrace("Score "..score);

  for i = 1, #private.milestones_map do

    --DbgTrace("MScore "..private.milestones_map[ i ].score);

    if ( private.milestones_map[ i ].score > 0 ) and ( private.milestones_map[ i ].score > private.score_buffer ) and ( private.milestones_map[ i ].score <= score ) then

      return private.milestones_map[ i ];

    end;

  end;

  return false;

end;
--------------------------------------------------------------------------------
function private.Miles_Rotate ()

  if ( #private.score_queue > 0 ) then
    
    private.milestones_locked = true;

    --local friend_ahead = private._Friends_Match( private.score_queue[ 1 ] );
    local marker_ahead = private.Miles_Match( private.score_queue[ 1 ] );

    --[[if ( friend_ahead ) then

      DbgTrace( "Soon we meet friend: "..private.friend_drop[ 1 ].name );

      table.insert( private.score_queue, 1, private.friend_drop[ 1 ].score );

    end;]]

    if ( marker_ahead ) then

      --DbgTrace( "Soon we meet marker: "..marker_ahead.index );

      table.insert( private.score_queue, 1, marker_ahead.score );

    end;

    local step = 2 * math.pi / private.cycle_distance;
    local time = 1 / #private.score_queue; 

    local score = private.score_queue[ 1 ];
    local cycle = math.floor( score / private.cycle_distance );

    if ( cycle > 2 ) and not ( cycle == private.cycle_buffer ) then

      private.Miles_Map_Create( cycle + 1 );
      private.Miles_Map_Trim();

    end;
    
    table.remove( private.score_queue, 1 );    
    
    local milestones = ObjGet( "obj_int_panel_stage_progress_ride_miles" );
  
    ObjAnimate( "obj_int_panel_stage_progress_ride_miles", "ang", 0, 0, function ()

      private.Miles_Rotate();

      --[[if ( friend_ahead ) then

        private._Friends_Drop();

      end;]]

      if ( marker_ahead ) then

        private.Miles_OnProgress( marker_ahead.index, marker_ahead.cycle, marker_ahead.score );

      end;

    end, 
    {
      0,    0, milestones.ang, 
      time, 0, score * step;
    } );  

    private.score_buffer = score;
    private.cycle_buffer = cycle;

  else

    private.milestones_locked = false;

  end;

end;
--------------------------------------------------------------------------------
function private.Miles_Update ( marker_number, cycle )

  if ( marker_number == 1 ) then

    local texts = {};

    for i = 1, 3 do

      local distance = math.ceil( private.cycle_distance / 4 * i + private.cycle_distance * cycle );

      --private.milestones_map[ i + 1 ] = distance;

      texts[ i ] = ui.text.FormatNumber( distance, true );

    end;

    ObjMultiSet( {
      { "txt_int_panel_stage_progress_ride_milestone2_distance", { text = texts[ 1 ] } },
      { "txt_int_panel_stage_progress_ride_milestone3_distance", { text = texts[ 2 ] } },
      { "txt_int_panel_stage_progress_ride_milestone4_distance", { text = texts[ 3 ] } }
    } );

  end;

  if ( marker_number == 2 ) then

    local distance = private.cycle_distance * ( cycle + 1 );

    --private.milestones_map[ 1 ] = distance;

    ObjSet( "txt_int_panel_stage_progress_ride_milestone1_distance", { text = ui.text.FormatNumber( distance, true ) } );

  end;

end;
--------------------------------------------------------------------------------
function private.Miles_OnProgress ( marker_number, cycle, score )

  --DbgTrace("Marker: "..marker_number..", cycle: "..cycle );

  -- SCORE GOAL ----------------------------------------------------------------
  if ( cycle < private.miles_cycles ) then

    private.Miles_Passed( marker_number );

    if ( marker_number == 1 ) then

      private.Miles_Update( marker_number, cycle );

    end;

    if ( marker_number == 2 ) then

      private.Miles_Update( marker_number, cycle );

      -- JK: show first marker
      if ( cycle == 0 ) then

        ObjSet( "obj_int_panel_stage_progress_ride_milestone1", { active = true, visible = true } );

      end;

    end;

    if ( marker_number == 3 ) then

      --
     
    end;

    if ( marker_number == 4 ) then 

      if ( cycle == private.miles_cycles - 1 ) then

        --private._Friends_Set( score, marker_number );

      end;

    end;
  
  -- BEYOND SCORE GOAL ---------------------------------------------------------
  elseif ( cycle >= private.miles_cycles ) then

    if ( marker_number == 1 ) then

      -- JK: hide markers 2,3,4 on score goal
      if ( cycle == private.miles_cycles ) then

        private.Miles_Passed( marker_number );
        private.Miles_Update( marker_number, cycle );

        ObjMultiSet( {
          { "obj_int_panel_stage_progress_ride_milestone2", { active = false, visible = false } },
          { "obj_int_panel_stage_progress_ride_milestone3", { active = false, visible = false } },
          { "obj_int_panel_stage_progress_ride_milestone4", { active = false, visible = false } }
        } );
        
      end;

    end;

    if ( marker_number == 2 ) then

      -- JK: hide first marker if beyond score goal
      if ( cycle == private.miles_cycles ) then

        private.Miles_Update( marker_number, cycle );

        ObjSet( "obj_int_panel_stage_progress_ride_milestone1", { active = false, visible = false } );

      end;

    end;

    if ( marker_number == 3 ) then

      --

    end;

    if ( marker_number == 4 ) then



    end;

    --if ( cycle > private.miles_cycles ) then

      --private._Friends_Set( score, marker_number );

    --end;

  end;

end;
--------------------------------------------------------------------------------
function private.Miles_Stop ()

  ObjStopAnimate( "obj_int_panel_stage_progress_ride_miles", "ang" );

end;
--------------------------------------------------------------------------------
function private.Miles_Passed ( index, func_end )

  ui.anim.Spin( "obj_int_panel_stage_progress_ride_milestone"..index, func_end );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private._Friends_Index ( marker_number )

  local index = { 2, 3, 4, 1 };

  return index[ marker_number ];

end;
--------------------------------------------------------------------------------
function private._Friends_Set ( score, marker_number )

  if ( #private.friends == 0 ) then

    --private._Friends_Hide();

    return false;

  end;

  local disk  = ObjGet( "obj_int_panel_stage_progress_ride_miles" );
  disk.radius = disk.inputrect_w / 2;
  disk.length = 2 * math.pi * disk.radius;

  local ratio = private.total_score / disk.length;

  local segment    = private.cycle_distance / 4;
  local from_score = score + segment;
  local till_score = score + 2 * segment;

  local candidates = {};

  --DbgTrace("Pick friends from "..from_score.." to "..till_score );

  for i = 1, #private.friends do

    if ( private.friends[ i ].score > from_score ) and ( private.friends[ i ].score <= till_score ) then

      table.insert( candidates, private.friends[ i ] );

    end;    

  end;

  local index = private._Friends_Index( marker_number );

  if ( #candidates == 0 ) then

    ObjSet( "obj_int_panel_stage_progress_ride_friend_"..index, { active = false, visible = false } );

    return false;

  end;

  for i = 1, #candidates do

    table.remove( private.friends, 1 );

  end;

  local chosen_one = candidates[ math.random( 1, #candidates ) ];

  --DbgTrace( "Chosen one: "..chosen_one.name..", score: "..chosen_one.score );  

  local ang   = - ( chosen_one.score / ratio ) / disk.radius ;
  local pos_y = math.cos( ang ) * disk.radius;
  local pos_x = math.sin( ang ) * disk.radius;

  local objects = {}; 

  table.insert( objects, { "obj_int_panel_stage_progress_ride_friend_"..index,
  {    
    pos_x   = pos_x,
    pos_y   = pos_y,
    ang     = ang + math.pi,
    active  = true,
    visible = true
  } } );

  table.insert( objects, { "txt_int_panel_stage_progress_ride_friend_score_"..index,
  {    
    text = chosen_one.score;
  } } );

  table.insert( objects, { "anm_int_panel_stage_progress_ride_friend_marker_"..index,
  {    
    animfunc = "static",
    playing = true
  } } );

  table.insert( objects, { "spr_int_panel_stage_progress_ride_friend_avatar_"..index,
  {    
    url = chosen_one.url;
  } } );

  ui.obj.SpriteAdjust( "spr_int_panel_stage_progress_ride_friend_avatar_"..index, 64 );

  ObjMultiSet( objects );

  chosen_one.index = index;

  table.insert( private.friend_drop, chosen_one );

  return true;

end;
--------------------------------------------------------------------------------
function private._Friends_Hide ()

  local multiset = {};

  for i = 1, private.friends_buffer do

    local anm_friend = "anm_int_panel_stage_progress_ride_friend_"..i;
    local obj_friend = "obj_int_panel_stage_progress_ride_friend_"..i;

    ui.obj.Attach( anm_friend, obj_friend );
    ui.obj.Disable( obj_friend, true, multiset );

  end;

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private._Friends_Match ( score )

  return ( private.friend_drop[ 1 ] )
  and    ( private.friend_drop[ 1 ].score )
  and    ( private.friend_drop[ 1 ].score > private.total_score )
  and    ( private.friend_drop[ 1 ].score > private.score_buffer )
  and    ( private.friend_drop[ 1 ].score <= score );  

end;
--------------------------------------------------------------------------------
function private._Friends_Drop ()

  --FF
  if ( #private.friend_drop < 1 ) then return; end;

  --DbgTrace( "Friend drop: "..private.friend_drop[ 1 ].name );

  local index = private.friend_drop[ 1 ].index;

  ui.anim.Play(
    "anm_int_panel_stage_progress_ride_friend_marker_"..index,
    "action"
  );
  
  table.remove( private.friend_drop, 1 );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Update ( action_data )

  if not private.active then

    return;

  end;

  -- start world animation as stage launched and first score achieved
  if ( action_data.score_current > 0 ) and not private.user_active then
    
    private.Hero_Start();

  end;

  table.insert( private.score_queue, action_data.score_current );

  if not private.milestones_locked then
  
    private.Miles_Rotate();

  end;
  
  private.UserActivity_Detect();

end;
--------------------------------------------------------------------------------
function public.Update_PopupHit ()

  private.popup_hit = true;

end;
--------------------------------------------------------------------------------
function public.Update_TargetComplete ()

  --private.target_complete = true;

end;
--------------------------------------------------------------------------------
function public.Update_StageComplete ( func_end )

  private.stage_complete = true;
  private.bonus_run      = false;

  private.Hero_Win( func_end );

end;
--------------------------------------------------------------------------------
function public.Update_StageFailed ()

  private.stage_failed = true;

end;
--------------------------------------------------------------------------------
function public.Update_HeroRespawn ()

  private.Hero_Respawn();

end;
--------------------------------------------------------------------------------
function public.Update_ComboSpeed ( speed )

  private.Hero_Run( speed );

end;
--------------------------------------------------------------------------------
function public.Update_BonusRun ()

  private.bonus_run = true;  
  --private.Hero_Run( 4 );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.UserActivity_Detect ()

  private.user_active = true;

  local func_OnTimerEnd = function()

    private.user_active = false;

  end;

  ObjSet( "tmr_int_panel_stage_progress_ride_user_activity_timeout",
  {
    playing = true,
    time    = 2.0,
    endtrig = func_OnTimerEnd
  } );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Color_Set ( color )

  local parts_count = private.screen_w / 128;

  local multiset_table = {};

  for i = 1, parts_count do

    table.insert( multiset_table, 
    {
      "spr_int_panel_stage_progress_ride_decoration_"..i, { color_r = color.r, color_g = color.g, color_b = color.b }
    } );

  end;
  
  table.insert( multiset_table, 
  {
    "grd_int_panel_stage_progress_ride_decoration", { color_r = color.r, color_g = color.g, color_b = color.b }
  } );

  ObjMultiSet( multiset_table );

end;
--------------------------------------------------------------------------------