-- name=int_panel_stage_settings
--------------------------------------------------------------------------------
function public.__InitModule ()

  ui.hub.Create( "int_panel_stage_settings" );

  private.obj_panel = "obj_int_panel_stage_settings";
  private.anm_panel = "anm_int_panel_stage_settings";

  private.obj_splash = "obj_int_panel_stage_settings_pause";
  private.spr_splash = "spr_int_panel_stage_settings_pause";

  private.af_show = "anim_in";
  private.af_hide = "anim_out";

  private.opened = false;

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  ObjSet( private.obj_panel,
  {
    pos_x = 0,
    pos_y = ui.screen.Height() - ui.screen.BottomSafeArea()
  } );

  ui.obj.ToCenter( private.obj_splash );
  ObjSet( private.obj_splash,
  {
    visible        = false,
    input          = false,

    inputrect_init = true,
    inputrect_w    = ui.screen.Width(),
    inputrect_h    = ui.screen.Height(),
    inputrect_x    = -0.5 * ui.screen.Width(),
    inputrect_y    = -0.5 * ui.screen.Height(),
    event_mup      = function () 

      if ( ObjGet( private.anm_panel ).playing ) then return; end;
      private.MenuClose();

    end
  } );

  local button = "btn_int_panel_stage_settings_";
  local anim   = "anm_int_panel_stage_settings_";

  ui.control.EventsSet( button.."panel",
  {
    func_on_click = function ()

      if ( ObjGet( private.anm_panel ).playing ) then return; end;

      if ( private.opened ) then private.MenuClose();
      else                       private.MenuOpen();
      end;
      
      if ( IsMobile() ) then
      
        private.ScaleUp( anim.."panel" );
      
      end;

    end,
    func_on_mouse_enter = function () private.ScaleDown( anim.."panel" ) end,
    func_on_mouse_leave = function () private.ScaleUp( anim.."panel" ) end
  } );

  ui.control.EventsSet( button.."options",
  {
    func_on_click = function ()

      private.Handler_ShowOptions( private.MenuClose );
      -- private.MenuClose();
      
      if ( IsMobile() ) then
      
        private.ScaleUp( anim.."panel" );
      
      end;

    end,
    func_on_mouse_enter = function () private.ScaleDown( anim.."options" ) end,
    func_on_mouse_leave = function () private.ScaleUp( anim.."options" ) end
  } );

  ui.control.EventsSet( button.."exit",
  {
    func_on_click = function ()

      private.Handler_ExitStage( private.MenuClose, private.MenuClose );
      -- private.MenuClose();
      
      if ( IsMobile() ) then
      
        private.ScaleUp( anim.."panel" );
      
      end;

    end,
    func_on_mouse_enter = function () private.ScaleDown( anim.."exit" ) end,
    func_on_mouse_leave = function () private.ScaleUp( anim.."exit" ) end
  } );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.SetHandlers ( func_options, func_exit_stage, 
                              func_show_others, func_hide_others )

  private.Handler_ShowOptions = func_options;
  private.Handler_ExitStage   = func_exit_stage;

  private.Handler_ShowOthers = func_show_others;
  private.Handler_HideOthers = func_hide_others;

end;
--------------------------------------------------------------------------------
function public.Show ()

  ObjSet( "int_panel_stage_settings", { input = true } );

  private.SetDefault();

  ui.hub.AttachContent( "int_panel_stage_settings" );
  
end
--------------------------------------------------------------------------------
function public.Hide ()

  ObjSet( "int_panel_stage_settings", { input = false } );
  ui.hub.DetachContent( "int_panel_stage_settings" );
  
end;
--------------------------------------------------------------------------------
function public.SetInput( flag )

  ObjSet( "int_panel_stage_settings", { input = flag } );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.SetDefault ()

  private.opened = false;

  ObjSet( private.obj_splash, { visible = false, input = false } );

  ObjSet( private.anm_panel, 
  {
    animfunc = private.af_show,
    frame    = 0,
    playing  = false
  } );

  ObjSet( "btn_int_panel_stage_settings_options", { input = false } );
  ObjSet( "btn_int_panel_stage_settings_exit",    { input = false } );

end;
--------------------------------------------------------------------------------
function private.MenuOpen ()

  if ( private.opened ) then return; end;
  private.opened = true;

  ObjSet( "btn_int_panel_stage_settings_options", { input = true } );
  ObjSet( "btn_int_panel_stage_settings_exit",    { input = true } );

  private.Handler_HideOthers();
  ui.anim.Play( private.anm_panel, private.af_show );

  ObjSet( private.obj_splash, { visible = true, input = true } );

  ObjStopAnimate( private.spr_splash, "alp" );
  local alp = ObjGet( private.spr_splash ).alp;

  ObjAnimate( private.spr_splash, "alp", 0, 0, "", 
  {
    0.0, 0, alp,
    0.3, 0, 0.35,
  } );

end;
--------------------------------------------------------------------------------
function private.MenuClose ()

  if not ( private.opened ) then return; end;
  private.opened = false;

  ObjSet( "btn_int_panel_stage_settings_options", { input = false } );
  ObjSet( "btn_int_panel_stage_settings_exit",    { input = false } );

  ui.anim.Play( private.anm_panel, private.af_hide, private.Handler_ShowOthers );

  ObjStopAnimate( private.spr_splash, "alp" );
  ObjAnimate( private.spr_splash, "alp", 0, 0,
  function ()
    ObjSet( private.obj_splash, { visible = false, input = false } );
  end, 
  {
    0.0, 0, 0.35,
    0.3, 0, 0.00,
  } );

end;
--------------------------------------------------------------------------------
function public.IsOpened ()

  return private.opened;

end;
--------------------------------------------------------------------------------
function public.MenuOpen ()

  private.MenuOpen();

end;
--------------------------------------------------------------------------------
function public.MenuClose ()

  private.MenuClose();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.ScaleDown ( object )

  ObjStopAnimate( object, "scale_xy" );
  local scl = ObjGet( object ).scale_x;

  ObjAnimate( object, "scale_xy", 0, 0, "", 
  {
    0.0, 0, scl, scl,
    0.3, 0, 0.8, 0.8,
  } );

end;
--------------------------------------------------------------------------------
function private.ScaleUp ( object )

  ObjStopAnimate( object, "scale_xy" );
  local scl = ObjGet( object ).scale_x;

  ObjAnimate( object, "scale_xy", 0, 0, "", 
  {
    0.0, 0, scl, scl,
    0.3, 0, 1.0, 1.0,
  } );

end;
--------------------------------------------------------------------------------
