-- name=int_panel_stage_top
--------------------------------------------------------------------------------
function public.__InitModule ()
  
  private.app_h = ui.screen.Height() - ui.screen.BottomSafeArea();
  private.app_w = ui.screen.Width();
  private.app_o = ui.screen.Orientation();
  
  private.this = {
    obj   = "obj_int_panel_stage_top",
    hub   = "hub_int_panel_stage_top_level",
    panel = "anm_int_panel_stage_top",
    level = "anm_int_panel_stage_top_level",
    state = "anm_int_panel_stage_top_level_state",
    moves = "anm_int_panel_stage_top_moves",
    goals = "anm_int_panel_stage_top_goals"
  };
  
  private.strings = {
    moves = ne.strings.Get( "str_int_panel_stage_moves" ),
    goals = ne.strings.Get( "str_int_panel_stage_targets" )
  };

  private.goal_queue         = {};
  private.receiver_animating = {};
  private.active = false;
  
end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  ui.hub.Create( "int_panel_stage_top" );
  private.Adapt();
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Show ()

  if ( private.active ) then return; end;
  
  private.active  = true;
  private.visible = true;
  
  ui.hub.AttachContent( "int_panel_stage_top" );
  
  ui.anim.Play(
    private.this.panel,
    "appear_"..private.app_o,
    function ()
  
      ui.input.On( private.this.obj );
      
      ui.anim.Play(
        private.this.panel,
        "static_"..private.app_o
      );
      
    end
  );

end;
--------------------------------------------------------------------------------
function public.Hide ()

  private.Level_Hide();
  
  if ( not private.active ) then return; end;
  
  private.active  = false;
  private.visible = false;
  
  mediator.Unsubscribe( "View_Stage_StageWon", private.Goals_OnStageWon );
  
  ui.input.Off( private.this.obj );
  
  ui.anim.Play(
    private.this.panel,
    "disappear_"..private.app_o,
    function ()
  
      private.Level_Clean();
      private.Goals_Clean();
      
      ui.hub.DetachContent( "int_panel_stage_top" );
      
    end
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.SetVisible ()
  
  if ( not private.active ) then return; end;
  
  if ( private.visible ) then return; end;
  
  private.visible = true;
  
  ui.anim.Play(
    private.this.panel,
    "appear_"..private.app_o,
    function () ui.input.On( private.this.obj ); end
  );

end;
--------------------------------------------------------------------------------
function public.SetInvisible ()
  
  if ( not private.active ) then return; end;
  
  if ( not private.visible ) then return; end;
  
  private.visible = false;
  
  ui.input.Off( private.this.obj );
  
  ui.anim.Play(
    private.this.panel,
    "disappear_"..private.app_o
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.IsActive ()
  
  return private.active;
  
end;
--------------------------------------------------------------------------------
function public.IsVisible ()
  
  return private.visible;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Level_Setup (
    stage_number,
    stage_state,
    stage_type,
    is_stateless,
    is_minievent
  )
  
  local multiset       = {};
  local stage_animfunc = stage_type.."_stage";
  
  ui.obj.Enable( private.this.level, false, multiset );
  
  if ( is_stateless ) then
    
    if ( stage_type ~= "fight" ) then
    
      stage_animfunc = "stateless";
    
    end;

    if ( is_minievent ) then

      stage_animfunc = "minievent";

    end;
  
    ui.obj.Disable( private.this.state, false, multiset );
    
  else
    
    ui.obj.Enable( private.this.state, false, multiset );
    
    ui.obj.Set(
      private.this.state,
      {
        animfunc = "state_"..stage_state
      },
      multiset
    );
  
  end;
  
  ui.obj.Set(
    private.this.level,
    {
      animfunc  = stage_animfunc,
      playing   = true,
      tt_number = tostring( stage_number )
    },
    multiset
  );
    
  if ( stage_type == "fight"       ) or
     ( private.app_o == "portrait" ) then
     
    ui.obj.Enable( private.this.hub, false, multiset );
     
    ui.obj.Attach(
      private.this.level,
      private.this.hub
    );
     
  else
   
    ui.obj.Disable( private.this.hub, false, multiset );
    
    ui.obj.Attach(
      private.this.level,
      private.this.panel
    );
     
  end;
  
  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function public.Level_Disable ()
  
  ui.obj.Disable( private.this.level );
  
end;
--------------------------------------------------------------------------------
function public.Level_Show ()
  
  private.level_standalone = true;
  
  ui.obj.Set(
    private.this.hub,
    {
      pos_z = 2
    }
  );
  
  ui.obj.Attach( private.this.hub, ui.hub.Name() );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Moves_Update ( moves, delta, is_win )

  local animated = false;
  local mode     = "static";
  
  if ( moves < 5 and not is_win ) then
  
    mode = "action";
  
  elseif ( moves >= 5 and delta > 1 ) then
  
    animated = true;
    mode = "increase";
  
  end;
  
  local animfunc = ui.obj.Get( private.this.moves ).animfunc;
  
  if ( not string.find( animfunc, mode ) ) then
  
    ui.anim.Play(
      private.this.moves,
      mode.."_"..private.app_o.."_moves"
    );
  
  end
  
  if ( not animated ) then
  
    ui.obj.Set(
      private.this.moves,
      { tt_moves_count = tostring( moves ) }
    );
  
  else
  
    local from  = moves - delta;
    local queue = {};
    
    for count = from, moves do
    
      table.insert(
        queue,
        function ( func_after )
        
          ui.obj.Set(
            private.this.moves,
            { tt_moves_count = tostring( count ) }
          );
        
          ui.anim.Play(
            private.this.moves,
            mode.."_"..private.app_o.."_moves",
            function ()
            
              if ( func_after ) then func_after(); end;
            
            end
          );
        
        end
      );
    
    end
    
    common.RunCallbackList( queue );
  
  end
  
end;

--------------------------------------------------------------------------------
function public.Moves_GetPosition ()

  return ui.obj.Position( private.this.moves );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Goals_GetReceiver ( goal_index )
  
  return "anm_int_panel_stage_top_goals_goal"..goal_index;

end;
--------------------------------------------------------------------------------
function public.Goals_Setup ( goals, package )

  local multiset = {};  

  private.goal_queue = {};

  private.goals = {
    number   = 0,
    data     = {},
    complete = 0
  };
  
  for goal_type, goal_data in pairs( goals ) do
    
    for goal_index, goal_value in pairs( goal_data ) do

      private.goals.number = math.max( private.goals.number, goal_index );
      private.goals.data[ goal_index ] = private.goals.data[ goal_index ] or {};
      private.goals.data[ goal_index ].done = goal_value.done or false;
      
      ui.obj.Set(
        "anm_int_panel_stage_top_goals_goal"..goal_index,
        {
          animfunc = "static_goal",
          tt_count = private.Goals_Progress( goal_value ),
          rt_icon  = ui.obj.Res( "goal", goal_value, nil, package )
        },
        multiset
      );
      
      if ( private.app_o == "portrait" ) then
      
        ui.obj.Attach(
          "anm_int_panel_stage_top_goals_goal"..goal_index,
          "anm_int_panel_stage_top_goals"
        );
      
      else
        
        ui.obj.Attach(
          "anm_int_panel_stage_top_goals_goal"..goal_index,
          "anm_int_panel_stage_top"
        );
        
      end;
      
    end;
    
  end;
  
  if ( private.app_o == "portrait" ) then
    
    ui.obj.Set(
      "anm_int_panel_stage_top_goals",
      {
        animfunc = "static_"..private.app_o.."_goal_"..private.goals.number
      },
      multiset
    );
    
    ui.obj.Attach(
      "anm_int_panel_stage_top_goals",
      "anm_int_panel_stage_top"
    );
    
  end;
  
  ui.obj.Multiset( multiset );

  mediator.SubscribeOnce(
    "View_Stage_StageWon",
    private.Goals_OnStageWon
  );
  
end;
--------------------------------------------------------------------------------
function public.Goals_AddToQueue ( goal_data )
  
  if ( not private.goal_queue[ goal_data.index ] ) then

    private.goal_queue[ goal_data.index ] = {};

  end;
  
  local queue_size = #private.goal_queue[ goal_data.index ];

  if ( queue_size > 0 ) then

    local prev_data = private.goal_queue[ goal_data.index ][ queue_size ];

    if ( prev_data.progress < goal_data.progress ) then

      table.insert( private.goal_queue[ goal_data.index ], goal_data );

    end;

  else

    table.insert( private.goal_queue[ goal_data.index ], goal_data );

  end;

end;
--------------------------------------------------------------------------------
function public.Goals_Update ( goal_data, goal_index, func_end )

  if ( not private.goals ) then return; end;
  
  local goal_number = ( goal_data or {} ).index or goal_index;

  if ( not private.goals.data[ goal_number ] )  then return; end;
  if ( private.goals.data[ goal_number ].done ) then return; end;
  
  goal_data = goal_data or table.remove( private.goal_queue[ goal_number ], 1 );

  if ( not goal_data ) then return; end;

  local receiver = public.Goals_GetReceiver( goal_number );

  ui.obj.Set(
    receiver,
    {
      tt_count = private.Goals_Progress( goal_data ),
    }
  );
  
  if ( goal_data.progress >= goal_data.count ) then
    
    private.goals.data[ goal_number ].done = true;
    private.goals.complete = private.goals.complete + 1;

    if ( not private.target_animating                  ) and
       ( private.goals.complete < private.goals.number ) then
    
       common.sound.Play( "half_target" );
    
    end;
    
    private.target_animating = true;
    
    ui.anim.Play(
      receiver,
      "action_goal",
      function ()
        
        ui.anim.Play( receiver, "complete_goal" );

        private.target_animating = nil;
        
        if ( func_end ) then

          func_end();

        end;
        
      end
    );
        
    private.Goals_Remind();

  else

    if ( func_end ) then

      func_end();

    end;
    
  end;

end;
--------------------------------------------------------------------------------
function public.Goals_Receive ( goal_data, func_end )

  local goal_number = goal_data.index;
  local receiver    = public.Goals_GetReceiver( goal_number );
  
  local function func_OnReceived ()

    mediator.Notify(
      "Interface_PanelStageTop_GoalReceived",
      goal_number
    );

    if ( func_end ) then

      func_end();

    end;

  end;
  
  local goal_type = goal_data.type;
  local goal_number = goal_data.index;
  
  -- JK: receiver is not busy
  if ( not private.receiver_animating[ goal_type ] ) and 
    ( not private.goals.data[ goal_number ].done ) then

    private.receiver_animating[ goal_type ] = true;

    ui.anim.Play(
      receiver,
      "receive_goal",
      function ()
    
        ui.anim.Play( receiver, "static_goal" );

        private.receiver_animating[ goal_type ] = nil;

        func_OnReceived();
        
      end
    );

  else

    func_OnReceived();

  end;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Goals_Remind ()
  
  if ( private.goals.number < 2 ) then
    
    -- JK: no need if goal is one
    return;
    
  end;
  
  for g = 1, private.goals.number do
    
    local goal = private.goals.data[ g ];
    
    if ( not goal.done ) then

      local receiver = public.Goals_GetReceiver( g );
      local animfunc = ui.obj.Get( receiver ).animfunc;
        
      ui.anim.Play(
        receiver,
        "remind_goal",
        function ()
       
          ui.anim.Play( receiver, animfunc );
          
        end
      );
      
    end;
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.Goals_Progress ( goal_data )
  
  if ( goal_data.type == "GainScore" ) then
    
    return ui.text.FormatNumber( goal_data.count, true );
    
  end;
  
  return ui.text.FormatNumber( goal_data.progress ).."/"..ui.text.FormatNumber( goal_data.count );
  
end;
--------------------------------------------------------------------------------
function private.Goals_OnStageWon ()

  for goal_index, queue in pairs( private.goal_queue ) do

    if ( not private.goals.data[ goal_index ].done ) and
       ( #queue > 0                                ) then
       
      local goal_data = table.remove( queue, 1 );
      public.Goals_Update( goal_data, goal_index );
      
      Log.w( "Goal queue for "..tostring( goal_data.type ).." is not empty" );
        
    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Goals_Clean ()
  
  ui.obj.Attach(
    "anm_int_panel_stage_top_goals",
    "hub_panel_stage_top"
  );
  
  for g = 1, private.goals.number do
    
    ui.obj.Attach(
      "anm_int_panel_stage_top_goals_goal"..g,
      "hub_panel_stage_top"
    );
    
  end;

  private.goals = nil;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Level_Clean ()

  ui.obj.Attach(
    private.this.level,
    "hub_panel_stage_top"
  );
  
end;
--------------------------------------------------------------------------------
function private.Level_Hide ()
  
  if ( not private.level_standalone ) then
    
    return;
    
  end;
  
  ui.obj.Set(
    private.this.hub,
    {
      pos_z = 0
    }
  );
  
  ui.obj.Attach(
    private.this.hub,
    "int_panel_stage_top"
  );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Adapt ()
  
  local multiset = {};
  
  ui.obj.Set(
    private.this.obj,
    {
      pos_x = private.app_w / 2
    },
    multiset
  );
  
  ui.obj.Set(
    private.this.panel,
    {
      tt_goals = private.strings.goals,
      tt_moves = private.strings.moves
    },
    multiset
  );  
  
  ui.obj.Set(
    private.this.hub,
    {
      pos_x = private.app_w - 35,
      pos_y = private.app_h - 40,
    },
    multiset
  );
  
  ui.obj.Set(
    private.this.goals,
    {
      tt_goals = private.strings.goals
    }
  );
  
  ui.obj.Set(
    private.this.moves,
    {
      animfunc = "static_"..private.app_o.."_moves",
      tt_moves = private.strings.moves
    },
    multiset
  );
  
  ui.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
