-- name=int_popup_cloud
--[[
Universal animated interface pop-up (own interface hub)

Modes: 
- monolog - just text                 - obj's input switching
- hint    - just text                 - obj's input stay same
- counter - counter                   - obj's input switching
- complex - text, scroll to counter   - obj's input switching
- reward  - reward, scroll to counter - obj's input switching
]]
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.this = {
    int         = "int_popup_cloud",
    popup       = "anm_int_popup_cloud",
    message     = "anm_int_popup_cloud_message",
    transparent = "obj_int_popup_cloud_transparent",
    timer       = "tmr_int_popup_cloud_autoclose",
    reward      = "anm_int_popup_cloud_reward"
  };

  private.queue            = {};
  private.block_shift_size = 400;
  private.state            = "close";

  private.close_time =
  {
    [ "monolog" ] = 2.0,
    [ "hint"    ] = 3.0,
    [ "counter" ] = 1.5,
    [ "complex" ] = 3.5,
    [ "reward"  ] = 3.5
  };

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  ui.hub.Create( private.this.int );

  local app_w = ui.screen.Width();
  local app_h = ui.screen.Height();

  ui.obj.Set(
    private.this.transparent,
    { 
      inputrect_x = -app_w, 
      inputrect_y = -app_h,
      inputrect_w =  app_w * 2, 
      inputrect_h =  app_h * 2,
      event_mup   = function() public.Hide(); end
    }
  );

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.Show (
    mode,
    full_obj_name,
    text,
    data,
    offset,
    scale,
    functions,
    disable_sound,
    permanent
  )

  private.disable_sound = disable_sound;
  private.permanent     = permanent or false;

  if ( private.state == "open"    ) or
     ( private.state == "opening" ) then 

    --public.Hide();

    -- JK: Add to queue
    local params = {
      mode          = mode,
      full_obj_name = full_obj_name,
      text          = text,
      data          = data,
      offset        = offset,
      scale         = scale,
      functions     = functions,
      disable_sound = disable_sound,
      permanent     = permanent or false
    };

    private.Queue_Add( params );    

  else

    private.state = "opening";

    functions = functions or {
      func_preopen  = nil,
      func_open     = nil,
      func_close    = nil,
      func_preclose = nil,
      func_scale    = nil
    };

    private.FuncPreopen  = functions.func_preopen;
    private.FuncOpen     = functions.func_open;
    private.FuncPreclose = functions.func_preclose;
    private.FuncClose    = functions.func_close;
    private.FuncScale    = functions.func_scale;

    if ( private.FuncScale ) then

      functions.func_scale = nil;

      private.FuncScale(
        function()

          private.state = "focused";
          
          public.Show(
            mode,
            full_obj_name,
            text,
            data,
            offset,
            scale,
            functions,
            disable_sound,
            permanent
          );

        end
      );

      return;

    end;

    if ( private.FuncPreopen ) then

      private.FuncPreopen();

    end;
    
    if ( not private.disable_sound ) then
      
       common.sound.Play( "create_cloud" );
      
    end;

    local multiset = {};

    ui.input.On( private.this.transparent, multiset );

    offset = offset or { 0, 0 };
    scale  = scale or 1;
    local reverse_scale = 1 / scale;

    ui.obj.Set(
      private.this.popup,
      {
        scale_x = scale,
        scale_y = scale
      },
      multiset
    );
    
    private.Message_SetDefault( mode, text, data, multiset );

    local obj_coords = ui.obj.Position( full_obj_name );

    obj_coords[ 1 ] = obj_coords[ 1 ] + offset[ 1 ];
    obj_coords[ 2 ] = obj_coords[ 2 ] + offset[ 2 ];

    ui.obj.Set(
      private.this.int,
      { 
        pos_x = math.floor( obj_coords[1] ), 
        pos_y = math.floor( obj_coords[2] ), 
        input = false 
      },
      multiset
    );

    ui.obj.Multiset( multiset );
    ui.obj.Attach( private.this.transparent, "ng_interface" );
    ui.hub.AttachContent( private.this.int );

    ui.anim.Play(
      private.this.popup,
      "appear",
      function ()

        private.state = "open";

        ui.anim.Play( private.this.popup, "static" );

        ui.input.On( private.this.int );

        if ( not private.permanent ) then

          private.Autoclose_Start( mode );

        end;        

        if ( private.FuncOpen ) then

          private.FuncOpen();

        end;

      end
    );

    ui.anim.Play( private.this.message, mode );

  end;

end;
--------------------------------------------------------------------------------
function public.Hide ( forced )
  
  if ( private.state == "open" ) or( forced ) then

    private.Autoclose_Stop();

    private.state = "closing";

    if ( private.FuncPreclose ) then

      private.FuncPreclose();

    end;

     if ( not private.disable_sound ) then
      
       common.sound.Play( "crash_cloud" );
      
    end;

    if ( #private.queue < 1 ) then

      local multiset = {};

      ui.input.Off( private.this.transparent, multiset );
      ui.input.Off( private.this.int,         multiset );

      ui.obj.Multiset( multiset );

    end;

    ui.anim.Play(
      private.this.popup,
      "close",
      function ()

        private.state = "close";

        ui.obj.Detach( private.this.transparent );
        ui.hub.DetachContent( private.this.int );

        if ( private.FuncClose ) then

          private.FuncClose();

        end;

        private.Queue_TryShow();

      end
    );

  end;

end;
--------------------------------------------------------------------------------
function public.AddToQueue (
    mode,
    full_obj_name,
    text,
    data,
    offset,
    scale,
    functions,
    disable_sound
  )

  local params = {
    mode          = mode,
    full_obj_name = full_obj_name,
    text          = text,
    data          = data,
    offset        = offset,
    scale         = scale,
    functions     = functions,
    disable_sound = disable_sound 
  };

  private.Queue_Add( params );

end;
--------------------------------------------------------------------------------
function public.TryShowFromQueue ()

  private.Queue_TryShow();

end;
--------------------------------------------------------------------------------
function public.IsOnScreen ()

  return ( ( private.state == "open" ) or ( private.state == "opening" ) );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Message_SetDefault ( mode, text, data, multiset )

  local message_data = {
    text     = "",
    progress = "",
    icon     = "",
    reward   = "str_int_popup_cloud_reward"
  };

  if ( data ) then

    if ( data.name ) then
      message_data.text = ne.strings.Get( "str_"..data.name );
    else
      message_data.text = ne.strings.Get( text );
    end;
    
    if ( mode == "reward" and data.reward ) then
      ui.obj.Enable( private.this.reward, false, multiset );
      
      local count = tern( data.reward.name == "immortality" )( 
                    ui.text.FormatTime( data.reward.count ), data.reward.count );
      
      ui.obj.Set( private.this.reward, {
        animfunc = "reward",
        rt_icon  = ui.obj.Res( data.reward.type, data.reward.name ),
        tt_count = count
      }, multiset );
    else
      ui.obj.Disable( private.this.reward, false, multiset );
    end;

    message_data.icon =
      data.icon or ui.obj.Res( data.target.type, data.target.name );
    message_data.progress = ne.strings.Get(
      "str_int_popup_cloud_progress",
      data.count or #( data.list or {} ),
      data.target.count or #data.target.list or 0
    );
    
  else

    message_data.text = ne.strings.Get( text );

  end;

  ui.obj.Set(
    private.this.message,
    {
      rt_icon     = message_data.icon,
      tt_message  = message_data.text,
      tt_progress = message_data.progress,
      tt_reward   = message_data.reward,
      animfunc    = mode,
      frame       = 0,
      playing     = false
    },
    multiset
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Queue_Add ( params, force_first )

  if ( force_first ) then

    table.insert( private.queue, 1, params );

  else

    table.insert( private.queue, params );

  end;

end;
--------------------------------------------------------------------------------
function private.Queue_TryShow ()

  if ( #private.queue > 0 ) then

    local params = table.remove( private.queue, 1 );

    public.Show(
      params.mode,
      params.full_obj_name,
      params.text,
      params.data,
      params.offset,
      params.scale,
      params.functions,
      params.disable_sound,
      params.permanent
    );

  end;

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function private.Autoclose_Start ( mode )

  ui.obj.Set(
    private.this.timer, 
    {
      playing = true,
      time    = private.close_time[ mode ],
      endtrig = public.Hide
    }
  );

end;
--------------------------------------------------------------------------------
function private.Autoclose_Stop ()

  ui.obj.Set(
    private.this.timer,
    {
      playing = false
    }
  );

end;
--------------------------------------------------------------------------------