-- name=int_popup_flat
--------------------------------------------------------------------------------
function public.__InitObjects ()

  private.queue          = {};
  private.popup_state    = "close";
  private.app_w          = ui.screen.Width();
  private.app_h          = ui.screen.Height();
  private.screen_padding = 10; 

  ObjMultiSet(
  {
    {
      "obj_int_popup_flat",
      {
        input = false
      }
    },
    {
      "obj_int_popup_flat_body",
      {
        scale_x     = 0.0,
        scale_y     = 0.0,
        alp         = 0.0,
        event_mdown = function () public.Hide( false ); end
      }
    },
    { "obj_int_popup_flat_transparent_over",
      {
        inputrect_init = true,
        inputrect_w    = private.app_w,
        inputrect_h    = private.app_h,
        event_mdown    = function () public.Hide( false ); end
      }
    }
  } );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Show ( caption, text, parent, pos, inverse, transparent )

  local tail_border_shift = 2;
  
  if ( inverse ) then

    ObjSet( "anm_int_popup_flat_tail",        { pos_y = -90, scale_y = -1 } );
    ObjSet( "anm_int_popup_flat_tail_border", { pos_y = -88, scale_y = -1 } );

  else

    ObjSet( "anm_int_popup_flat_tail",        { pos_y = 90, scale_y = 1 } );
    ObjSet( "anm_int_popup_flat_tail_border", { pos_y = 88, scale_y = 1 } );

  end;
  
  if ( private.popup_state == "close" ) then

    private.popup_state    = "opening";
    private.current_parent = parent;

    local obj_params = ObjGet( "obj_int_popup_flat_body" );
    local positions  = private.GetPositions( parent, pos, inverse );
    
    ObjMultiSet(
    {
      {
        "txt_int_popup_flat_caption",
        {
          text = caption
        }
      },
      {
        "txt_int_popup_flat_text",
        {
          text     = text,
          fontsize = 30
        }
      },
      {
        "obj_int_popup_flat_body",
        {
          pos_x = positions.bubble[ 1 ],
          pos_y = positions.bubble[ 2 ]
        }
      },
      {
        "anm_int_popup_flat_tail",
        {
          pos_x = positions.tail[ 1 ]
        }
      },
      {
        "anm_int_popup_flat_tail_border",
        {
          pos_x = positions.tail[ 1 ] + tail_border_shift
        }
      }
    } );

    ObjAttach( "obj_int_popup_flat", InterfaceName );

    ui.text.FontsizeByHeight( "txt_int_popup_flat_text", 120, 30 );

    ObjAnimate( "obj_int_popup_flat_body", "alp", 0, 0, "", 
    { 
      0.0, 0, obj_params.alp, 
      0.3, 0, 1.0
    } );

    ObjSet( "obj_int_popup_flat", { input = true } );

    ObjAnimate( "obj_int_popup_flat_body", "scale_xy", 0, 0, function ()

      private.popup_state = "open";

    end, 
    { 
      0.00, 3, obj_params.scale_x, obj_params.scale_y,
      0.20, 3, 1.1,                1.1, 
      0.25, 3, 0.95,               0.95, 
      0.30, 3, 1.0,                1.0
    } );

  elseif (

    ( private.popup_state == "open"    ) or
    ( private.popup_state == "closing" ) 

  ) then

    if ( private.current_parent ~= parent ) then
  
      -- JK: Add to queue
      local params = {
        caption = caption,
        text    = text,
        parent  = parent,
        pos     = pos,
        inverse = inverse
      };

      private.AddToQueue( params );
    
    end;

  else

    public.Hide( false );

  end;

end;
--------------------------------------------------------------------------------
function public.Hide ( instant, func_end )

  private.popup_item_pos = nil;

  if ( instant ) then

    ObjSet( "obj_int_popup_flat_body", { 
      scale_x = 0.0, 
      scale_y = 0.0, 
      alp     = 0.0,
      pos_x   = 0 } );

    ObjDetach( "obj_int_popup_flat" );

    private.popup_state    = "close";
    private.current_parent = nil;
    
    if ( func_end ) then
      
      func_end();
      
    end;

    private.TryShowFromQueue();

  elseif ( private.popup_state == "open" ) then

    private.popup_state = "closing";

    ObjSet( "obj_int_popup_flat", { input = false } );

    local obj_params = ObjGet( "obj_int_popup_flat_body" );

    ObjAnimate( "obj_int_popup_flat_body", "alp", 0, 0, function ()

      ObjDetach( "obj_int_popup_flat" );

      private.popup_state    = "close";
      private.current_parent = nil;
      
      ObjSet( "obj_int_popup_flat_body", { pos_x = 0 } );
      
      if ( func_end ) then
      
        func_end();
      
      end;

      private.TryShowFromQueue();

    end, 
    {
      0.00, 0, obj_params.alp, 
      0.15, 0, 0.0
    } );

    ObjAnimate( "obj_int_popup_flat_body", "scale_xy", 0, 0, "", 
    {
      0.00, 0, obj_params.scale_x, obj_params.scale_y, 
      0.15, 0, 0.0,                0.0
    } );

  end;

end;
--------------------------------------------------------------------------------
function public.AttachedTo ()
  
  return private.current_parent;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.GetPositions ( parent, inner_pos, inverse )

  local parent_pos = ui.obj.Position( parent );
  local modifier   = 1;

  if ( inverse ) then

    modifier = -1;

  end;  

  local popup_pos_x = parent_pos[ 1 ] + inner_pos[ 1 ];
  local popup_pos_y = parent_pos[ 2 ] + ( inner_pos[ 2 ] * modifier );

  local fine_pos_x = popup_pos_x;
  local fine_pos_y = popup_pos_y;

  local overlap = 0;

  if ( fine_pos_x < ( private.screen_padding + 248 ) ) then

    overlap = ( fine_pos_x - ( private.screen_padding + 248 ) ) * -1;

    fine_pos_x = fine_pos_x + overlap;

  end;

  if ( fine_pos_x > ( private.app_w - private.screen_padding - 248 ) ) then

    overlap = ( private.app_w - ( fine_pos_x + 248 + private.screen_padding ) );

    fine_pos_x = fine_pos_x + overlap;

  end;

  return {
    bubble = { fine_pos_x, fine_pos_y },
    tail   = { inner_pos[ 1 ] + overlap * -1 }
  };

end;
--------------------------------------------------------------------------------
function private.AddToQueue ( params )

  table.insert( private.queue, params );

end;
--------------------------------------------------------------------------------
function private.TryShowFromQueue ()

  if ( #private.queue > 0 ) then

    local params = private.queue[ 1 ];

    table.remove( private.queue, 1 );

    public.Show(
      params.caption,
      params.text,
      params.parent,
      params.pos,
      params.inverse
    );

  end;

end;
--------------------------------------------------------------------------------