-- name=int_screen
--------------------------------------------------------------------------------
-- Graphical content switcher view. Used for manipulate screen objects.
-- @module int_screen
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Module constructor.
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.interface = "ng_interface";
  private.obj_level = "ng_level";
  private.anim_time = 0.3;

end;

--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Attach and enable screen object.
-- @param scr_obj screen object name.
--------------------------------------------------------------------------------
function public.Attach ( scr_obj )
  
  ui.obj.Set( scr_obj, { active = true, visible = true, input = false } );
  ui.obj.Attach( scr_obj, private.obj_level );
  
end;

--------------------------------------------------------------------------------
-- Detach and disable screen object.
-- @param scr_obj screen object name.
--------------------------------------------------------------------------------
function public.Detach ( scr_obj )
  
  ui.obj.Set( scr_obj, { active = false, visible = false } );
  ui.obj.Detach( scr_obj );
  
end;

--------------------------------------------------------------------------------
-- Enable input of screen object.
-- @param scr_obj screen object name.
--------------------------------------------------------------------------------
function public.InputOn ( scr_obj )
  
  ui.obj.Set( scr_obj, { input = true } );
  
end;

--------------------------------------------------------------------------------
-- Disable input of screen object.
-- @param scr_obj screen object name.
--------------------------------------------------------------------------------
function public.InputOff ( scr_obj )
  
  ui.obj.Set( scr_obj, { input = false } );
  
end;

--------------------------------------------------------------------------------
-- Shade screen object.
-- @param scr_obj screen object name.
--------------------------------------------------------------------------------
function public.ShadeOn ( scr_obj )
  
  ui.obj.Set( scr_obj, { input = false } );
  
  ObjAnimate( scr_obj, "color_rgb", 0, 0, "",
  {
    0.0,               0, 1.0, 1.0, 1.0,
    private.anim_time, 0, 0.0, 0.0, 0.0
  } );

  ObjAnimate( private.interface, "color_rgb", 0, 0, "",
  {
    0.0,               0, 1.0, 1.0, 1.0,
    private.anim_time, 0, 0.0, 0.0, 0.0
  } );
  
end;

--------------------------------------------------------------------------------
-- Disable shade of screen object.
-- @param scr_obj screen object name.
--------------------------------------------------------------------------------
function public.ShadeOff ( scr_obj )
  
  ObjAnimate( scr_obj, "color_rgb", 0, 0, "",
  {
    0.0,               0, 0.0, 0.0, 0.0,
    private.anim_time, 0, 1.0, 1.0, 1.0
  } );

  ObjAnimate( private.interface, "color_rgb", 0, 0, "",
  {
    0.0,               0, 0.0, 0.0, 0.0,
    private.anim_time, 0, 1.0, 1.0, 1.0
  } );
  
end;

--------------------------------------------------------------------------------