-- name=int_window
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.app_w = ui.screen.Width();
  private.app_h = ui.screen.Height();
  
  private.this = {
    outer    = "obj_int_window_outer",
    map      = "zmv_map",
    black    = "spr_int_window_black"
  };

  ui.hub.Create( "int_window" );

  private.window_data  = {};
  private.window_index = {};
  private.fullscreen   = {};
  private.modal        = {};
  private.overlay      = {};

  private.anmtime = 0.30;
  private.bgalpha = 0.60;
  private.grayrgb = 0.40;

  private.InitQueue();

  mediator.Subscribe( "View_Stage_Pause", private.FreezeWindowQueue );

  mediator.Subscribe(
    "Interface_ButtonPressed",
    public.OuterTap_Disable
  );

  mediator.Subscribe(
    "Interface_ButtonAction",
    public.OuterTap_Enable
  );

  mediator.Subscribe(
    "Interface_ButtonReleased",
    public.OuterTap_Enable
  );
  
  mediator.Subscribe(
    "View_Map_Section_Switched",
    private.OnSectionSwitched
  );

end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  ui.obj.ToCenter( private.this.black );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.RegisterAsFullscreen ( window_name )

  private.fullscreen[ window_name ] = { name = window_name };

end;
--------------------------------------------------------------------------------
function public.RegisterAsModal ( window_name, strict_parent )

  private.modal[ window_name ] = {
    name          = window_name,
    strict_parent = strict_parent
  };

end;
--------------------------------------------------------------------------------
function public.RegisterAsOverlay ( window_name )

  private.overlay[ window_name ] = { name = window_name };

end;
--------------------------------------------------------------------------------
function public.Show ( window_name, func_end, params )

  --DbgTrace( "[ INFO ][ int_window ] Attempt to show window: "..window_name );

  params = params or {};

  local disallow_modal = params.disallow_modal or false;
  local force_modal    = params.force_modal or false;

  -- JK: check if window can be shown before it started to show
  if ( not private.IsAllowedToShow( window_name, disallow_modal, force_modal ) ) then

    --DbgTrace(
    --  "[ INFO ][ int_window ] "..window_name.." is not allowed to show."
    --);
    return;

  end;

  private.Show( window_name, func_end, params );

  ui.obj.Set(
    private.this.map,
    {
      active = ( params.map_active or false ),
      input  = ( params.map_input  or false )
    }
  );

  if ( params.notify ~= false ) then

    mediator.Notify( "Interface_Window_WindowOpened", window_name );

  end;

end;
--------------------------------------------------------------------------------
function public.Hide ( window_name, func_end, skip_map_enable )

  -- JK: check if window can be hidden before it's actually started to hide
  if ( not private.IsAllowedToHide ( window_name ) ) then

    return;

  end;

  mediator.Notify( "Interface_Window_WindowClosed", window_name );

  private.Hide( window_name, func_end );

  if ( #private.window_index < 1 ) then
    
    if ( not skip_map_enable ) then
    
      ObjSet( private.this.map, { active = true, input = true } );
    
    end;
    
  end;

end;
--------------------------------------------------------------------------------
function public.HideAll ( func_end, skip_map_activation )

  if ( #private.window_index > 0 ) then
    
    local func_hide = _G[ private.window_index[ #private.window_index ] ].Hide;
    
    if ( func_hide ) then
      
      func_hide( function ()
        public.HideAll( func_end );
      end );
      
    else
      
      private.Hide(
        private.window_index[ #private.window_index ],
        function () public.HideAll( func_end ); end
      );
      
    end;

  else

    if ( func_end ) then func_end(); end;
    if ( not skip_map_activation ) then ui.obj.Enable( private.this.map ); end;

  end;

end;
--------------------------------------------------------------------------------
function public.Uppermost_Name ()
  
  if ( public.IsAnyWindowOnScreen() ) then
    return private.window_index[ #private.window_index ];
  end;
  
  return false;
  
end;

--------------------------------------------------------------------------------
function public.Lowermost_Name ()
  
  if ( public.IsAnyWindowOnScreen() ) then
    return private.window_index[ 1 ];
  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
function public.TryCloseUpperWindow ()

  if ( public.IsAnyWindowOnScreen() ) then
    
    local window_name = private.window_index[ #private.window_index ];
    local func_close  = private.window_data[ window_name ].close_handler;
    
    if ( private.IsFullscreen( window_name ) and 
         private.window_data[ window_name ].minimized ) then
         
      window_name = private.window_index[ #private.window_index - 1 ];
      func_close  = private.window_data[ window_name ].close_handler;
      
    end;
    
    if ( func_close ) then

      func_close();

      return true, window_name;

    end;
    
    return false, window_name;

  end;

  return false, nil;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Maximize ( window_name )
  
  public.OuterTap_Disable();
  
  private.window_data[ window_name ].minimized = false;
  
  for i = #private.window_index, 1, - 1 do
    
    local lower_window_name = private.window_index[ i ];
    
    if ( window_name ~= lower_window_name ) then
      
      ui.obj.Set( lower_window_name, { visible = false, input = false } );
      
    end;
    
  end;
  
end;
--------------------------------------------------------------------------------
function public.Minimize ( window_name )
  
  public.OuterTap_Enable();
  
  private.window_data[ window_name ].minimized = true;
  
  for i = #private.window_index, 1, - 1 do
    
    local lower_window_name = private.window_index[ i ];
    
    if ( window_name ~= lower_window_name ) then
      
      ui.obj.Set( lower_window_name, { visible = true, input = true } );
      
    end;
    
  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.SetArea ( window_name, area, multiset )

  ui.obj.Set(
    window_name,
    {
      inputrect_init = true,
      inputrect_x    = area[ 1 ],
      inputrect_y    = area[ 2 ],
      inputrect_w    = area[ 3 ],
      inputrect_h    = area[ 4 ]
    },
    multiset
  );

end;
--------------------------------------------------------------------------------
function public.SetAreaByButtons ( obj_window, btn_action, btn_close )
  
  if not ( btn_close ) then
  
    local base_name = ui.obj.GetNameWithoutPrefix( obj_window );
    btn_close  = "btn_"..base_name.."_close";
  
  end;
  
  -- Get positions
  local pos_action = ui.obj.Position( btn_action, obj_window );
  local pos_close  = ui.obj.Position( btn_close,  obj_window );
  
  local left   = -pos_close[1];
  local right  =  pos_close[1];
  local top    =  pos_close[2];
  local bottom =  pos_action[2];
  
  -- Shift from buttons centers
  local params_action = ui.obj.Get( btn_action );
  local params_close  = ui.obj.Get( btn_close  );
  
  local shift_up   = -params_close.inputrect_h * 0.6;
  local shift_down = params_action.inputrect_h * 0.6;
  
  top    = top    + shift_up;
  bottom = bottom + shift_down;
  
  public.SetArea( obj_window, 
    {
      left,
      top,
      right  - left,
      bottom - top
    }
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.OuterTap_Set ( window_name )

  private.OuterTap_Set( window_name );

end;
--------------------------------------------------------------------------------
function public.OuterTap_Clear ()

  private.OuterTap_Clear();

end;
--------------------------------------------------------------------------------
function public.OuterTap_Disable ()

  if ( not private.outertap_active ) then
    
    return;
    
  end;
  
  ui.input.Off( private.this.outer );

end;
--------------------------------------------------------------------------------
function public.OuterTap_Enable ()

  if ( not private.outertap_active ) then
    
    return;
    
  end;
  
  ui.input.On( private.this.outer );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Handler_Close ( window_name )
  
  local close_button = "btn_"..window_name.."_close";

  if ( ui.button.Exists( close_button )    ) and
     ( ui.button.IsEnabled( close_button ) ) then

    local close_handler = function ()

      if ( ui.button.HasInput( close_button ) ) then
      
        local event_handler = ui.button.EventGet( close_button );
        if ( type( event_handler ) == "function" ) then event_handler(); end;

        return true;

      end;

      return false;

    end;

    return close_handler;

  end;

  return false;
  
end;
--------------------------------------------------------------------------------
function public.Handler_Accept ( window_name )
  
  if ( ui.button.Exists( "btn_"..window_name.."_accept" ) ) then

    return ui.button.EventGet( "btn_"..window_name.."_accept" );

  end;

  return false;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.IsFullscreen ( window_name )

  return private.IsFullscreen( window_name );

end;
--------------------------------------------------------------------------------
function public.IsOnScreen ( window_name )

  return private.window_data[ window_name ] ~= nil;

end;
--------------------------------------------------------------------------------
function public.IsAnyWindowOnScreen ()
  
  return ( #private.window_index > 0 );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.IsFullscreen ( window_name )

  return private.fullscreen[ window_name ] ~= nil;

end;
--------------------------------------------------------------------------------
function private.IsModal ( window_name )

  if( private.modal[ window_name ] ~= nil ) then

    return true, private.modal[ window_name ].strict_parent;

  end;

end;
--------------------------------------------------------------------------------
function private.IsOverlay ( window_name )

  if ( not window_name ) then

    return false;

  end;

  return private.overlay[ window_name ] ~= nil;

end;
--------------------------------------------------------------------------------
function private.IsUpperOverlay ()

  return private.IsOverlay( public.Uppermost_Name() );

end;
--------------------------------------------------------------------------------
function private.IsAllowedToShow ( window_name, disallow_modal, force_modal )

  if ( public.IsAnyWindowOnScreen() ) then

    if ( private.IsUpperOverlay()   ) and
       ( #private.window_index == 1 ) then

      return true;

    end;

    local is_modal, strict_parent = private.IsModal( window_name );

    if ( is_modal or force_modal ) then

      if ( strict_parent and not public.IsOnScreen( strict_parent ) ) then

        return false;

      end;

      return not disallow_modal;

    end;

    return false;

  end;

  return true;

end;
--------------------------------------------------------------------------------
function private.IsAllowedToHide ( window_name )

  if ( private.IsModal( window_name )   ) or
     ( private.IsOverlay( window_name ) ) then

    return true;

  end;

  for modal_window, modal_data in pairs( private.modal ) do

    if ( modal_data.strict_parent                ) and
       ( modal_data.strict_parent == window_name ) and
       ( public.IsOnScreen( modal_window )       ) then

      return false;

    end;

  end;

  return true;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Show ( window_name, window_func, params )

  local pos_z = 1;

  if ( #private.window_index < 1 ) then
  
    ui.hub.AttachContent( "int_window" );

  else

    pos_z = private.window_data[ private.window_index[ #private.window_index ] ].pos_z + 2;

  end;

  if ( not private.window_data[ window_name ] ) then

    table.insert( private.window_index, window_name );

    private.window_data[ window_name ] =
    {
      index = #private.window_index;
      name  = window_name,
      pos_z = pos_z
    };

    ui.obj.Set( window_name, { pos_z = pos_z } );
    ui.obj.Attach( window_name, "int_window" );

    --JK: try trigger __InitObjects
    ui.obj.Show( window_name );

  else

    DbgTrace( "[ WARNING ] Dialog window < "..window_name.." > already on the screen!" );
    return;

  end;
  
  if ( private.IsFullscreen( window_name ) and params.open_minimized ) then
    
    private.window_data[ window_name ].minimized = true;
    
  end;

  private.window_data[ window_name ].close_handler = public.Handler_Close(
    window_name 
  );

  if not ( private.IsFullscreen( window_name ) ) then
    
    if ( #private.window_index > 1 ) then

      -- JK: shade previous window
      private.Gray_Out( private.window_index[ #private.window_index - 1 ] );

    end;
    
    private.Background_Show();

    -- JK: now this is set in ui.window.Open via public method
    --private.OuterTap_Set( window_name );

  end;

  if ( window_func ) then

    window_func();

  end;

end;
--------------------------------------------------------------------------------
function private.Hide ( window_name, func_end )

  local prev_window;

  if ( not private.IsFullscreen( window_name ) ) then

    private.OuterTap_Clear();

  end;

  --common.delegate.Clear( window_name );

  ObjDetach( window_name );

  table.remove( private.window_index, private.window_data[ window_name ].index );

  -- JK: shift down if not last
  -- if ( private.window_data[ window_name ].index ~= #private.window_index ) then
  
    for i, n in pairs( private.window_index ) do

      private.window_data[ n ].index = i;

    end;

  -- end;

  private.window_data[ window_name ] = nil;

  if ( #private.window_index > 0 ) then

    prev_window = private.window_index[ #private.window_index ];

  end;

  local event_func = function ()

    mediator.Notify( "Interface_Window_WindowClosedEnd", window_name );
    private.Gray_Reset( window_name );
    
    ( func_end or function () end )();

    if ( #private.window_index < 1 ) then

      ui.hub.DetachContent( "int_window" );

      mediator.Notify( "Interface_Window_NoWindows" );

    end;

  end;

  private.Background_Hide( event_func );

  if ( not private.IsFullscreen( window_name ) ) then

    if ( prev_window ) then

      private.Gray_In(
        prev_window,
        function ()

          if ( not private.IsFullscreen( prev_window ) ) then

            private.OuterTap_Set( prev_window );

          end;

        end
      );

    end;

  end;

  private.ShowWindowFromQueue( window_name );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Background_Show ()

  if ( private.bg_active ) then

    return;

  end;

  private.bg_active  = true;
  private.background = private.this.black;

  ui.obj.Enable( private.background );

  local current = ui.obj.Get( private.background ).alp;

  ObjStopAnimate( private.background, "alp" ); 

  ObjAnimate( private.background, "alp", 0, 0, "",
  {
    ( 0 * private.anmtime ), 0, current,
    ( 1 * private.anmtime ), 0, private.bgalpha
  } );

end;
--------------------------------------------------------------------------------
function private.Background_Hide ( event_func )

  if not ( private.bg_active ) then

    event_func();
    
    return;

  end;

  if ( #private.window_index > 0 ) then

    event_func();

    return;

  end;

  private.bg_active = false;

  ObjStopAnimate( private.background, "alp" );

  local current = ui.obj.Get( private.background ).alp;

  local OnAnimationEnd = function()

    ui.obj.Disable( private.background );
    event_func();

  end;

  if ( current > 0 ) then

    ObjAnimate( private.background, "alp", 0, 0, OnAnimationEnd,
    {
      ( 0 * private.anmtime ), 0, current,
      ( 1 * private.anmtime ), 0, 0.0
    } );
  
  else
    
    OnAnimationEnd();
  
  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Gray_Out ( window_name )

  -- JK: grayed out window has disabled input
  private.OuterTap_Clear();
  ui.input.Off( window_name );

  local current = ui.obj.Get( window_name );

  ObjAnimate( window_name, "color_rgb", 0, 0, "",
  {
    ( 0 * private.anmtime ), 3, current.color_r, current.color_g, current.color_b,
    ( 1 * private.anmtime ), 3, private.grayrgb, private.grayrgb, private.grayrgb
  } );

end;
--------------------------------------------------------------------------------
function private.Gray_In ( window_name, func_end )

  local w = ui.obj.Get( window_name );

  ObjAnimate(
    window_name,
    "color_rgb", 0, 0,
    function () 
  
      ui.input.On( window_name );
      if ( func_end ) then func_end(); end;

    end,
    {
      ( 0 * private.anmtime ), 3, w.color_r, w.color_g, w.color_b,
      ( 1 * private.anmtime ), 3, 1,         1,         1
    }
  );

end;
--------------------------------------------------------------------------------
function private.Gray_Reset ( window_name )

  ui.input.On( window_name );

  ui.obj.Set(
    window_name,
    {
      color_r = 1.0,
      color_g = 1.0,
      color_b = 1.0
    }
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.OuterTap_Set ( window_name )

  private.OuterTap_Clear();

  if ( private.window_data[ window_name ]               ) and
     ( private.window_data[ window_name ].close_handler ) and
     ( not private.IsFullscreen( window_name )          ) then

    ui.control.EventsSet(
      private.this.outer,
      {
        func_on_click = function ()
          if ( private.window_data[ window_name ].close_handler() ) then
            private.OuterTap_Clear();
          end;
        end;
      }
    );

    local pos_z = private.window_data[ window_name ].pos_z - 1;

    ui.obj.Set(
      private.this.outer,
      {
        active  = true,
        visible = true,
        input   = true,
        pos_z   = pos_z
      }
    );
    
    private.outertap_active = true;

  end;

end;
--------------------------------------------------------------------------------
function private.OuterTap_Clear ()

  ui.control.EventsClear( private.this.outer );

  ui.obj.Set(
    private.this.outer,
    {
      active  = false,
      visible = false,
      input   = false,
      pos_z   = 0
    }
  );
  
  private.outertap_active = nil;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.InitQueue ()
  
  private.window_queue             = {};
  private.window_queue.list        = {};
  private.window_queue.last_high   = 0;
  private.window_queue.last_medium = 0;
  private.window_queue.show        = nil;
  private.window_backup            = {};
  
  function public.AddWindowToQueue ( window_name, func_start, params )

    local position = #private.window_queue.list + 1;

    table.insert(
      private.window_queue.list,
      position,
      {
        name   = window_name,
        func   = func_start,
        params = params
      }
    );

    if ( #private.window_index < 1       ) and
       ( #private.window_queue.list == 1 ) then

      private.ShowWindowFromQueue();

    end;

  end;
  
  function public.IsWindowQueueReleased ()

    if ( #private.window_queue.list > 0 ) or ( private.window_queue.show ) then

      return false;

    end;

    return true;

  end;
  
  function public.WindowBackup ( window_name )

    for w = 1, #private.window_queue.list do
   
      if ( private.window_queue.list[ w ].name == window_name ) then

        private.window_backup = table.remove( private.window_queue.list, w );

        break;

      end;

    end;

  end;
  
  function public.WindowRestore ()

    if ( private.window_backup.name ) then

      public.AddWindowToQueue( private.window_backup.name, private.window_backup.func )

      for w = 1, #private.window_queue.list do
     
        if ( private.window_queue.list[ w ].name == window_name ) then

          private.window_backup = table.remove( private.window_queue.list, w );

          break;

        end;

      end;

      private.window_backup = nil;

    end;

  end;

  function private.ShowWindowFromQueue ( last_closed_window )

    local prefix = ui.obj.GetNamePrefix( v_screen.GetCurrentName() );

    if ( "int" == prefix ) or ( private.window_queue.freezed ) then

      --DK> Очищаем текущий, если был выполнен переход.
      if ( private.window_queue.show                            ) and
         ( last_closed_window == private.window_queue.show.name )
      then

        private.window_queue.show = nil;

      end;

      return;

    end;

    if ( private.window_queue.show                                    ) and
       ( not ( last_closed_window == private.window_queue.show.name ) )
    then

      return;

    end;

    if ( #private.window_index < 1 ) and ( #private.window_queue.list > 0 ) then

      private.window_queue.show = table.remove( private.window_queue.list, 1 );

      local window_obj = private.window_queue.show.name;

      public.Show( window_obj, private.window_queue.show.func, private.window_queue.show.params );

    end;

    private.window_queue.show = nil;

    mediator.Notify( "Interface_Window_QueueReleased" );

  end;
  
  function private.FreezeWindowQueue ( param )

    private.window_queue.freezed = param;

    if ( not private.window_queue.freezed ) then
    
      private.ShowWindowFromQueue();
    
    end;

  end;
  
  function private.GetCurrentWindowName ()

    if ( private.window_queue.show ) then

      return private.window_queue.show.name;

    end;

  end;
  
end;
--------------------------------------------------------------------------------
function private.OnSectionSwitched ()
  
  public.HideAll( nil, true );
  
end;
--------------------------------------------------------------------------------