-- name=lib_db
--------------------------------------------------------------------------------

local DB_PATH    = "assets/content/database/";
local LEVEL_PATH = "assets/content/levels/";

appdata = {};
session = {};

private.Player = {};
 public.Player = {};
 
private.settings_variant = 0;
 
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.is_debug    = wiz.IsCheatsEnabled();
  private.driver      = "ram"; --JK: "db" or "ram"
  private.readonly    = false;
  public.Execute      = private.Execute_DB;
  public.Player.Info  = private.Player.Info_DB;
  private.collections = {};
  appdata             = {
    static = {}
  };
  
  private.collections.static = {
    achievements     = { "name", "type" },
    boosters         = { "name", "type" },
    builds           = { "name", "type", "section", "unlock", "deadlock", "unlocklist.min" },
    characters       = { "name", "section", "unlock", "quest_cage", "quest_free" },
    enemies          = { "name", "section", "is_boss" },
    chests           = { "name", "stage", "section" },
    lots             = { "name", "name_ios", "platform" },
    mines            = { "name", "section", "unlock" },
    minievents       = { "name" },
    quests           = { "name", "section", "pregive", "condition.type", "target.name" },
    sections         = { "name", "unlock" },
    skins            = { "name", "platform", "invite" },
    spells           = { "name", "unlock", "section" },
    stages           = { "section", "name", "type" },
    stickers         = { "name" },
    potions          = { "name" },
    dialogs          = { "name" }
  };
  
  private.Register( private.collections.static );
  
  if ( private.driver == "db" ) then
 
    -- JK: static is always cached
    --for name, indices in pairs( private.collections.static ) do
    --  
    --  private.Load( "static", name, indices );
    --  
    --end;
    
    return;
 
  end;
  
  public.Execute = private.Execute_RAM;
  
  -- JK: player/info is retrieved from DB on production
  if ( private.is_debug ) then
  
    public.Player.Info = private.Player.Info_RAM;
  
  end;
  
  for c, collection in pairs( ConfigGetTable( "parts.db" ).writable_collections ) do
    
    local header = common.string.Split( collection.name, "/" );
    local group  = header[ 1 ];
    local name   = header[ 2 ];
    
    if ( not private.IsProtected( group, name ) ) then

      if ( not private.collections[ group ] ) then
        
        private.collections[ group ] = {};
        appdata[ group ]             = {};
        
      end;
      
      private.collections[ group ][ name ] = {};
      
      for i, index in pairs( collection.indices ) do
        
        table.insert( private.collections[ group ][ name ], index );
        
      end;

    end;
    
  end;
  
  for group, collections in pairs( private.collections ) do
    
    for name, indices in pairs( collections ) do
      
      private.Load( group, name, indices );
      
    end;
    
  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Execute ( query )
  
  -- JK: see Init
  
end;
--------------------------------------------------------------------------------
function public.DbExecute ( query )
  
  return DbExecuteQuery( query );
  
end;
--------------------------------------------------------------------------------
function public.Rewrite ( group, name, data )
  
  if ( private.driver == "ram" ) then
  
    local indices = private.collections[ group ][ name ];
    local indexed = private.Index( indices, data );
  
    if ( not indexed ) then
      
      data = data[ 1 ];
      
    end;
    
    appdata[ group ][ name ] = {
      data    = data,
      indices = indexed
    };
    
  end;
  
  -- JK: insert in db immediately
  local query = 
  {
    query      = "insert",
    collection = group.."/"..name,
    value      = data
  };
  
  DbExecuteQuery( query );
  
  return true;
  
end;
--------------------------------------------------------------------------------
function public.Settings ( type_id )
  
  local query =
  {
    query      = "find",
    group      = "static",
    collection = "logic_settings",
    indice     = "type",
    criteria   = { like = type_id }
  };
  
  local result = public.Execute( query ).result;
  
  if ( result ) then
  
    return result.value;
  
  end;
  
  if ( IsEditor() ) then
  
    local msg = "[ ERROR ] Try to get non exists setting '%s'"
    common.DbgCallpoint( msg:format( type_id ) );
  
  end;
  
  return nil;
  
end;
--------------------------------------------------------------------------------
function public.TryReloadSettings ( settings_variant )

  Log.i(
    nil,
    "current settings variant: "..tostring( private.settings_variant )..
    ", reloading to "..settings_variant
  );
  
  if ( private.settings_variant ~= settings_variant ) then
    private.LoadSettings( settings_variant, true );
  else
    Log.i( "no need to reload logic_settings" );
  end;
  
end;
--------------------------------------------------------------------------------
function public.Player.Info ()
  
  -- JK: see Init
  
end;
--------------------------------------------------------------------------------
function public.Player.Set ( key, value )
  
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        = 
    {
      [ key ] = value
    }
  };
  
  local  is_set = public.Execute( query ).result;
  return is_set;
  
end;
--------------------------------------------------------------------------------
function public.AddLevels( source )
  
  if ( type( source ) ~= table ) then
    source = { source };
  end;
  
  local register_list = {};
  for _, filename in pairs( source ) do
    table.insert( register_list, LEVEL_PATH..filename );
  end;

  parts.db.RegisterReadonlyCollections( register_list );
  
  private.Load( "static", "stages", private.collections.static.stages );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Session_Set ( key, value )

  session[ key ] = value;

end;
--------------------------------------------------------------------------------
function public.Session_Get ( key )

  if ( session[ key ] ) then
  
    return session[ key ];
  
  end;
  
  return false;

end;
--------------------------------------------------------------------------------
function public.Session_Clear ( key )

  if ( session[ key ] ) then
  
    session[ key ] = nil;
    
    return true;
    
  end;
  
  return false;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Database_Attach ()

  if ( private.db_method_backup                       ) and
     ( type( private.db_method_backup ) == "function" ) then

    private.database_freeze = false;
       
    _G[ "DbExecuteQuery" ] = private.db_method_backup;

  end;

end;
--------------------------------------------------------------------------------
function public.Database_Detach ()
  
  if ( private.driver == "ram" ) then

    private.db_method_backup = _G[ "DbExecuteQuery" ];
  
    private.database_freeze = true;
    
    _G[ "DbExecuteQuery" ] = function () return { result = 1 }; end;
  
    return true;
    
  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
function public.Database_IsFreeze ()
  
  return private.database_freeze;
  
end;
--------------------------------------------------------------------------------
function public.Database_SetRO ()
  
  private.readonly = true;
  
end;
--------------------------------------------------------------------------------
function public.Database_SetRW ()
  
  private.readonly = false;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Events_ListCallbacks ()
  
  local code, json = parts.storage.data.Read( "parts/time" );

  local function func_error ()

    DbgTrace( "[ ERROR ] [ lib_db ] Failed to read JSON from 'parts/time'." );

  end;

  if ( not json ) or ( code > 0 ) or ( type( json ) ~= "string" ) then

    func_error();
    return nil;

  end;

  local decoded = wiz.json.Decode( json );

  if ( not decoded                ) or
     ( type( decoded ) ~= "table" ) then

    func_error();
    return nil;

  end;

  local callbacks = decoded.callbacks;

  if ( not callbacks                ) or
     ( type( callbacks ) ~= "table" ) then

    func_error();
    return nil;

  end;
  
  return callbacks;
  
end;
--------------------------------------------------------------------------------
function public.Events_ListWatches ()
  
  local _, json = parts.storage.data.Read( "parts/time" );
  local watches = wiz.json.Decode( json ).watches;
  
  return watches;
  
end;
--------------------------------------------------------------------------------
function public.Events_GetCallbackByToken ( token )

  if ( parts.time.event.Exists( token ) ) then

    return parts.time.event.GetData( token );

  end;

  return false;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Execute_DB ( query )
  
  local group, indice = "", "";
  
  if ( query.group ~= "static" ) then
  
    group = query.group.."/";
  
  --else
  --  
  --  -- JK: static is always cached
  --  return private.Execute_RAM( query );
    
  end;
  
  query.group = nil;
  
  if ( query.indice ) then
    
    indice       = ":"..query.indice;
    query.indice = nil;
    
  end;
  
  query.collection = group..query.collection..indice;
  
  local response = DbExecuteQuery( query );
  
  if ( query.query ~= "find"     ) and
     ( query.query ~= "find_all" ) then
     
     return { result = ( response.result == 1 ) };
     
  end;
  
  return response;
  
end;
--------------------------------------------------------------------------------
function private.Execute_RAM ( query )
  
  -- JK: protected data is always retrieved from DB
  if ( private.IsProtected( query.group, query.collection ) ) then

    return private.Execute_DB( query );

  end;

  local methods = {
    find       = private.FindAll,
    find_i     = private.FindByIndex,
    find_all   = private.FindAll,
    find_all_i = private.FindAllByIndex,
    insert     = private.Insert,
    update     = private.Update,
    update_i   = private.UpdateByIndex,
    delete_i   = private.DeleteByIndex,
    delete_all = private.DeleteAll
  };
  
  local method = query.query;
  
  if ( query.indice ) then
    
    method = method .. "_i";
    
  end;
  
  return { result = methods[ method ]( query ) };
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Player.Info_DB ()
  
  local query =
  {
    query      = "find",
    collection = "player/info"
  };
  
  return DbExecuteQuery( query ).result;
  
end;
--------------------------------------------------------------------------------
function private.Player.Info_RAM ()
  
  return private.Copy( appdata.player.info.data );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.IsProtected ( group, name )

  local protected = {
    player = {
      info     = true,
      boosters = true
    }
  };

  if ( private.is_debug ) then

    return false;

  end;

  if ( protected[ group ]         ) and
     ( protected[ group ][ name ] ) then

    return protected[ group ][ name ];

  end;

  return false;

end;
--------------------------------------------------------------------------------
function private.LoadSettings ( settings_variant, replace )
  
  local settings_path = DB_PATH.."logic_settings";
  
  if ( settings_variant     ) and
     ( settings_variant > 0 ) then
    settings_path = settings_path.."_"..settings_variant;
  end;
  
  if ( not replace ) then
    parts.db.RegisterReadonlyCollections( { settings_path } );
  else
    parts.db.ReplaceReadonlyCollection( "logic_settings", { settings_path } );
  end;

  if ( private.driver == "ram" ) then
    private.Load( "static", "logic_settings", { "type" } );
  end;
  
  private.settings_variant = settings_variant;
  
  Log.i( "Settings loaded from "..settings_path );
  
end;
--------------------------------------------------------------------------------
function private.Register ( readonly_collections )

  local path       = DB_PATH;
  local stagespath = LEVEL_PATH;
  
  -- JK: load logic settings first
  local settings_variant = private.GetSettingsVariant();
  private.LoadSettings( settings_variant );

  local by_section_collections = private.RegisterCollectionsBySection();
  
  -- JK: define separated collections, e.g. characters_1, characters_2 ...
  local separated_collections = {
    stages     = { "dailyquest", "minievent", "minievent_section_christmas" },
    minievents = { "tower", "tourney", "chests", "story", "gift", "section" },
    lots       = { "by_platform" },
    dialogs    = { "minievent", "minievent_section_christmas" }
  };
  
  -- JK: add collections to register list
  local register_list = {};
  
  local function add_collection ( key, postfix )
    
    if( key == "stages" ) then
      
      table.insert( register_list, stagespath..key.."_"..postfix );
      
    else
      
      table.insert( register_list, path..key.."_"..postfix );
      
    end;
    
  end;
  
  for key, collection in pairs( readonly_collections ) do
    
    if ( separated_collections[ key ] ) then
        
      local list = separated_collections[ key ];
      
      for i = 1, #list do

        if ( list[ i ] == "by_platform" ) then
        
          -- JK: add by platform
          local platform = GetPlatform();
          platform = tern( platform == "android" )(
            wiz.Android.GetPlatform(),
            platform
          );
          
          add_collection( key, platform );
            
        else
          
          add_collection( key, list[ i ] );
        
        end;
      
      end;

    elseif ( not by_section_collections[ key ] ) then

      table.insert( register_list, path..key );

    end;
    
  end;

  parts.db.RegisterReadonlyCollections( register_list );

end;
--------------------------------------------------------------------------------
function private.Load ( group, name, indices )
  
  local query = {
    query      = "find_all",
    collection = group.."/"..name
  };
  
  if ( group == "static" ) then
    
    query.collection = name;
    
  end;

  local data = DbExecuteQuery( query ).result;
  
  --DbgTrace( "[ STORAGE ] "..group.."/"..name.." ("..( tostring( #data ) )..")" );
  
  local indexed = private.Index( indices, data );
  
  if ( not indexed ) then
    
    data = data[ 1 ];
    
  end;
  
  appdata[ group ][ name ] = {
    data    = data,
    indices = indexed
  };
  
  --DbgTrace( "[STORAGE] RAM consumed: "..math.floor( collectgarbage( "count" ) ).." KB" );
  
end;
--------------------------------------------------------------------------------
function private.Index ( index_names, data )

  if ( #index_names < 1 ) then return nil; end;
  
  local indices = {};
  
  for i, index in pairs( index_names ) do
    
    local indexed = {};
    
    for d, item in pairs( data ) do
      
      local value;
      
      if ( index:find( "%." ) ) then
        
        local s = common.string.Split( index, "." );
        
        if ( item[ s[ 1 ] ] ) then
   
          value = item[ s[ 1 ] ][ s[ 2 ] ];
   
        end;
        
      else
        
        value = item[ index ];
        
      end;
      
      local key = "_index";
      
      if ( value ~= nil ) and ( value ~= "" ) then
        
        if ( type( value ) ~= "table" ) then
        
          key = "_"..tostring( value );
        
        end;
        
        if ( indexed[ key ] ) then
      
          table.insert( indexed[ key ], d );
          
        else
          
          indexed[ key ] = { d };
          
        end;
        
      end;
      
    end;
    
    indices[ index ] = indexed;
    
  end;
  
  return indices;
  
end;
--------------------------------------------------------------------------------
function private.Copy ( object )
  
  if ( type( object ) == "table" ) then
    
    local copy = {};
  
    for key, value in pairs( object ) do
        
      if ( type( value ) == "table" ) then
        
        value = private.Copy( value );
        
      end;
          
      copy[ key ] = value;
        
    end;
    
    return copy;
    
  end;
    
  return object;
  
end;
--------------------------------------------------------------------------------
function private.SelectFields ( fields, data )

  if ( data == nil ) then
    
    return nil;
    
  end;
  
  if ( type( fields ) == "string" ) then
    
    return data[ fields ];
    
  end;
  
  local result = {};
  for key, value in pairs( fields ) do
  
    result[ key ] = data[ value ];
    
  end;
  
  return result;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.FindByIndex ( query )

  local is_index_direct = ( type( query.indice ) == "number" );

  local index;

  if ( is_index_direct ) then

    index = query.indice;

  else

    --JK: -1 is non-existent index
    local indexes = appdata[ query.group ][ query.collection ].indices[ query.indice ][ "_"..tostring( query.criteria.like ) ] or { -1 };
    
    index = indexes[ 1 ];

  end;
  
  local data = appdata[ query.group ][ query.collection ].data[ index ];

  if ( query.result ) then
  
    return private.Copy( private.SelectFields( query.result, data ) );
  
  end;
  
  return private.Copy( data );
  
end;
--------------------------------------------------------------------------------
function private.FindAll ( query )

  local data = appdata[ query.group ][ query.collection ].data;

  if ( query.result and #data > 0 ) then
  
    local result = {};
    
    for _, value in pairs( data ) do
    
      local fields = private.SelectFields( query.result, value );
      table.insert( result, fields );
    
    end;
    
    return private.Copy( result );
  
  end;
  
  return private.Copy( data );
  
end;
--------------------------------------------------------------------------------
function private.FindAllByIndex ( query )
  
  local result  = {};
  local indexes = appdata[ query.group ][ query.collection ].indices[ query.indice ][ "_"..tostring( query.criteria.like ) ] or {};
  local data    = appdata[ query.group ][ query.collection ].data;
  
  for i, index in pairs( indexes ) do
  
    if ( query.result ) then
  
      local fields = private.SelectFields( query.result, data[ index ] );
      table.insert( result, fields );
    
    else
    
      table.insert( result, data[ index ] );
  
    end;
    
  end;
  
  return private.Copy( result );
  
end;
--------------------------------------------------------------------------------
function private.Insert ( query )
  
  -- JK: check for read-only
  if ( query.group == "static" ) or
     ( private.readonly        ) then

    -- JK: if readonly by flag, fake success
    return private.readonly;

  end;
  
  local index = #appdata[ query.group ][ query.collection ].data + 1;
  
  table.insert( appdata[ query.group ][ query.collection ].data, query.value );
  
  -- JK: indexing
  for i, indice in pairs( private.collections[ query.group ][ query.collection ] ) do
    
    local key  = "_index";
    local value = query.value[ indice ];
    
    if ( value ~= nil ) and
       ( value ~= ""  ) then
    
      if ( type( value ) ~= "table" ) then
      
        key = "_"..tostring( query.value[ indice ] );
        
      end;
      
      if ( appdata[ query.group ][ query.collection ].indices[ indice ][ key ] ) then
      
        table.insert( appdata[ query.group ][ query.collection ].indices[ indice ][ key ], index );
      
      else
      
        appdata[ query.group ][ query.collection ].indices[ indice ][ key ] = { index };
      
      end;
      
    end;
    
  end;
  
  -- JK: insert in db immediately
  local query = 
  {
    query      = "insert",
    collection = query.group.."/"..query.collection,
    value      = query.value
  };
  
  DbExecuteQuery( query );
  
  return true;
  
end;
--------------------------------------------------------------------------------
function private.Update ( query )
  
  -- JK: check for read-only
  if ( query.group == "static" ) or
     ( private.readonly        ) then

    -- JK: if readonly by flag, fake success
    return private.readonly;

  end;
  
  local data_root = appdata[ query.group ][ query.collection ].data;

  for key, value in pairs( query.set ) do
  
    data_root[ key ] = value;
  
  end;
  
  --JK: update in db immediately
  local query = 
  {
    query      = "update",
    collection = query.group.."/"..query.collection,
    set        = query.set
  };
  
  DbExecuteQuery( query );
  
  return true;
  
end;
--------------------------------------------------------------------------------
function private.UpdateByIndex ( query )
  
  -- JK: check for read-only
  if ( query.group == "static" ) or
     ( private.readonly        ) then

    -- JK: if readonly by flag, fake success
    return private.readonly;

  end;

  local index_root = appdata[ query.group ][ query.collection ].indices;
  local index_path = "_"..tostring( query.criteria.like ); 
  
  --JK: -1 is non-existent index
  local indexes = index_root[ query.indice ][ index_path ] or { -1 };
  local index   = indexes[ 1 ];
  
  if ( index < 0 ) then return false; end;

  local data_root = appdata[ query.group ][ query.collection ].data;
  
  for key, value in pairs( query.set ) do
  
    local _value = "_"..tostring( value );

    data_root[ index ][ key ] = value;
    
    -- JK: if data is indexed by this key, re-index
    if ( index_root[ key ] ) then
    
      local key_root = index_root[ key ];

      -- JK: remove this index from all paths of this root
      for k, key_root_path in pairs( key_root ) do

        for path_key, path_value in pairs( key_root_path ) do

          if ( index == path_value ) then

            table.remove( key_root_path, path_key );
            break;

          end;

        end;

      end;

      if ( key_root[ _value ] ) then

        table.insert( key_root[ _value ], index );
        
      else
        
        -- JK: create new index
        key_root[ _value ] = { index };
        
      end;
    
    end;
  
  end;
  
  --JK: update in db immediately
  local query = 
  {
    query      = "update",
    collection = query.group.."/"..query.collection..":"..query.indice,
    criteria   = query.criteria,
    set        = query.set
  };
  
  DbExecuteQuery( query );
  
  return true;
  
end;
--------------------------------------------------------------------------------
function private.DeleteByIndex ( query )
  
  -- JK: check for read-only
  if ( query.group == "static" ) or
     ( private.readonly        ) then

    -- JK: if readonly by flag, fake success
    return private.readonly;

  end;
  
  -- JK: -1 is non-existent index
  local indexes = appdata[ query.group ][ query.collection ].indices[ query.indice ][ "_"..tostring( query.criteria.like ) ] or { -1 };
  local index   = indexes[ 1 ];
  
  if ( index < 0 ) then

    -- JK: Object not found by index
    return false;

  end;
  
  table.remove( appdata[ query.group ][ query.collection ].data, index );
  
  -- JK: here we should re-index ALL elements
  appdata[ query.group ][ query.collection ].indices = {};
  
  appdata[ query.group ][ query.collection ].indices = private.Index(
    private.collections[ query.group ][ query.collection ],
    appdata[ query.group ][ query.collection ].data
  );
  
  --JK: delete from db immediately
  local query = 
  {
    query      = "delete",
    collection = query.group.."/"..query.collection..":"..query.indice,
    criteria   = query.criteria
  };
  
  DbExecuteQuery( query );
  
  return true;
  
end;
--------------------------------------------------------------------------------
function private.DeleteAll ( query )
  
  -- JK: check for read-only
  if ( query.group == "static" ) or
     ( private.readonly        ) then

    -- JK: if readonly by flag, fake success
    return private.readonly;

  end;
  
  appdata[ query.group ][ query.collection ].data = {};
  
  --JK: delete indexes
  for i, indice in pairs( private.collections[ query.group ][ query.collection ] ) do
    
    appdata[ query.group ][ query.collection ].indices[ indice ] = {};
    
  end;
  
  --JK: delete from db immediately
  local query = 
  {
    query      = "delete_all",
    collection = query.group.."/"..query.collection
  };
  
  DbExecuteQuery( query );
  
  return true;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.RegisterCollectionsBySection ( section_number )
  
  return private.RegisterCollectionsBySection( section_number );
  
end;
--------------------------------------------------------------------------------
function private.RegisterCollectionsBySection ( section_number )

  section_number = section_number or private.RegisterCollectionsBySection_GetCurrent();
  
  if ( ( private.section or 0 ) >= section_number ) then
    
    return false;
    
  end;
  
  private.section = section_number;

  local path       = "assets/content/database/";
  local stagespath = "assets/content/levels/";

  local separated_collections = {
    characters = true,
    enemies    = true,
    builds     = true,
    chests     = true,
    spells     = true,
    stages     = true,
    quests     = true,
    mines      = true
  };

  local register_list = {};
  
  local function add_collection ( key, postfix )
    
    if ( key == "stages" ) then
      
      table.insert( register_list, stagespath..key.."_"..postfix );
      
    else
      
      table.insert( register_list, path..key.."_"..postfix );
      
    end;
    
  end;
  
  -- FF: this is for achievements! :(
  for i = 1, section_number do
    
    for key, _ in pairs( separated_collections ) do
      
      add_collection( key, i );
    
    end;
    
  end;
  -- FF: end

  parts.db.RegisterReadonlyCollections( register_list );
  
  -- FF: maybe move this part to initialization?
  for name, indices in pairs( separated_collections ) do
    
    private.Load( "static", name, private.collections.static[ name ] );
    
  end;

  return separated_collections;
  
end;
--------------------------------------------------------------------------------
function private.RegisterCollectionsBySection_GetCurrent ()

  local query = 
  {
    query      = "find_all",
    group      = "player",
    collection = "sections"
  };

  local  sections = #public.Execute( query ).result;
  return sections;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.GetSettingsVariant ()
  
  local query = {
    query      = "find",
    group      = "sml",
    collection = "features",
    indice     = "name",
    criteria   = { like = "gold_balance" }
  };

  local feature_data = lib_db.Execute( query ).result;

  if ( feature_data                          ) and
     ( type( feature_data.data ) == "number" ) then

    return feature_data.data;

  end;

  return 0;
  
end;
--------------------------------------------------------------------------------