-- name=m_achievements
--[[
  Types:

  MineCollected
  TurnWheel
  UseBooster
  DestroyChip
  DestroyBlocker
  DestroyBonus
  StageTurns    --commented
  FightTurns    --commented
  WinDiffStage
  AddFriend
  WinFriend
]]

local WELCOME_HOME      = "welcome_home";
local ALWAYS_ONLINE     = "always_online";
local STAGE_CRUSHER     = "stage_crusher";
local MINIEVENT_GRINDER = "minievent_grinder";
local ARENA_MASTER      = "arena_master";

local STAGE_CRUSHER_TIME  = 3600;

--------------------------------------------------------------------------------
function public.__InitModule ()

  private.always_online = 0;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Setup ( stage_stats )
 
  local achievements, associated_ids = private.GetCollection();
  
  private.achievement_ids   = associated_ids;
  private.achievement_types = {};

  for a = 1, #achievements do

    local achievement = achievements[ a ];

    if ( IsMobile()                            ) or
       ( not achievement.type:find( "Friend" ) ) then
    
      query = 
      {
        query      = "find",
        group      = "player",
        collection = "achievements",
        indice     = "name",
        criteria   = { like = achievement.name }
      };

      if ( not lib_db.Execute( query ).result ) then

        local reward = {};
        
        for i = 1, #achievement.levels do
          table.insert( reward, { get = false, done = false } );
        end;
        
        query =
        {
          query      = "insert",
          group      = "player",
          collection = "achievements",
          value      = 
          {
            name   = achievement.name,
            level  = 1,
            count  = 0,
            reward = reward
          }
        };

        lib_db.Execute( query );

        private.Android_IncrementAchievement( achievement.name, 0, true );
        
      end;

      if ( not private.achievement_types[ achievement.type ] ) then

        private.achievement_types[ achievement.type ] = {};

      end;
      
      if ( achievement.type == "WinDiffStage" ) then

        achievement.bonus_count     = stage_stats.bonus_count;
        achievement.completed_count = stage_stats.completed_count;

      end;

      achievement.is_complete = ( achievement.level == #achievement.levels );

      table.insert( private.achievement_types[ achievement.type ], achievement );
      
    end;

  end;

  private.InitSubscribes();
  private.ResetCounters();

end;
--------------------------------------------------------------------------------
function public.SetRewarded ( achievement_name, level )

  local query = 
  {
    query      = "find",
    group      = "player",
    collection = "achievements",
    indice     = "name",
    criteria   = { like = achievement_name }
  };

  local achievement = lib_db.Execute( query ).result;

  local reward = achievement.reward;

  reward[ level ][ "done" ] = true;

  query =
  {
    query      = "update",
    group      = "player",
    collection = "achievements",
    indice     = "name",
    criteria   = { like = achievement_name },
    set        = 
    {
      count  = achievement.count,
      level  = achievement.level,
      name   = achievement.name,
      reward = reward
    }
  };

  lib_db.Execute( query );

  if ( achievement_name == "add_friends_fireman" ) then
    
    if ( level == 4 ) then
      
      reward[ level ][ "done" ] = false;
      reward[ level ][ "get" ] = false;
      
      query =
      {
        query      = "update",
        group      = "player",
        collection = "achievements",
        indice     = "name",
        criteria   = { like = achievement_name },
        set        = 
        {
          reward = reward
        }
      };

      lib_db.Execute( query );
      
    end;
    
    private.TryComplete( achievement_name );
    
  end;
  
  if ( not private.GetAchievementData( achievement_name ).hidden ) then
    mediator.Notify( "Model_Achievements_Rewarded" );
  end;

end;
--------------------------------------------------------------------------------
function public.GetAll ()

  local achievements = private.GetCollection();
  local reward_list  = lib_db.Settings( "achievement_rewards" );
  
  local result = {};
  
  for index, achievement in pairs( achievements ) do

    if ( not achievement.hidden                              ) and 
       ( IsMobile() or not achievement.type:find( "Friend" ) ) then

      query = 
      {
        query      = "find",
        group      = "player",
        collection = "achievements",
        indice     = "name",
        criteria   = { like = achievement.name }
      };    

      local properties = lib_db.Execute( query ).result;

      if ( properties ) then
        achievement.count  = properties.count;
        achievement.level  = properties.level;
        achievement.reward = properties.reward;
      end;
      
      table.insert( result, achievement );
      
    end;

  end;

  return result;

end;
--------------------------------------------------------------------------------
function public.GetRewardData ( achievement_name )

  local query = 
  {
    query      = "find",
    group      = "player",
    collection = "achievements",
    indice     = "name",
    criteria   = { like = achievement_name }
  };    

  return lib_db.Execute( query ).result.reward;

end;
--------------------------------------------------------------------------------
function public.GetAchievementsStat ()

  local achievements = public.GetAll();

  local total_count = 3 * #achievements;

  local completed_count = 0;

  for i = 1, #achievements do

    local achievement = achievements[ i ];

    for r = 1, #achievement.reward do

      if ( achievement.reward[ r ].get ) then

        completed_count = completed_count + 1;

      end; 

    end;

  end;

  return { completed_count = completed_count, total_count = total_count };

end;
--------------------------------------------------------------------------------
function public.Achievement_GetFullData ( achievement_name )

  local static_data = private.GetAchievementData( achievement_name );
  
  local query =
  {
    query      = "find",
    group      = "player",
    collection = "achievements",
    indice     = "name",
    criteria   = { like = achievement_name }
  };    

  local saved_data = lib_db.Execute( query ).result;
  
  for i = 1, #static_data.levels do
    
    saved_data.reward[ i ].name   = static_data.levels[ i ].reward_name;
    saved_data.reward[ i ].type   = static_data.levels[ i ].reward_type;
    saved_data.reward[ i ].count  = static_data.levels[ i ].reward_count;

    if ( i == #static_data.levels     ) and
       ( static_data.levels[ i ].step ) then

      saved_data.reward[ i ].target = static_data.levels[ i - 1 ].value + 
        static_data.levels[ i ].step * math.max( 1, saved_data.level - #static_data.levels + 1 );

    else

      saved_data.reward[ i ].target = static_data.levels[ i ].value;

    end;
    
  end;

  saved_data.levels = static_data.levels;

  return saved_data;
  
end;
--------------------------------------------------------------------------------
function public.Achievement_IncreaseCount ( achievement_name )

  private.ChangeCount( achievement_name, 1 );
  
  mediator.Notify(
    "Model_Achievements_AchievementChanged",
    public.Achievement_GetFullData( achievement_name )
  );
  
end;
--------------------------------------------------------------------------------
function public.CheckFriendsCount ( friends_count )

  for i = 1, #private.achievement_types[ "AddFriend" ] do
    
    local achievement = private.achievement_types[ "AddFriend" ][ i ];
    
    local achievement_name = achievement.name;
    
    if ( achievement.friends_count ~= friends_count ) then
      
      private.achievement_types[ "AddFriend" ][ i ].friends_count = friends_count;
      
      local query =
      {
        query      = "find",
        group      = "player",
        collection = "achievements",
        indice     = "name",
        criteria   = { like = achievement_name }
      };
      local achievement_info  = lib_db.Execute( query ).result or {};
      local achievement_count = achievement_info.count or 0;
      local diff = friends_count - achievement_count;

      if ( diff > 0 ) then
        private.ChangeCount( achievement_name, diff );
      end;
      
    end;
  
  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.GetCollection ()
  
  local query = 
  {
    query      = "find_all",
    group      = "static",
    collection = "achievements"
  };

  local achievements   = lib_db.Execute( query ).result;
  local reward_list    = lib_db.Settings( "achievement_rewards" );
  local associated_ids = {};
  
  -- JK: resolve achievement rewards from settings
  for index, achievement in pairs( achievements ) do
    
    associated_ids[ achievement.name ] = index;
    
    for index, level in pairs( achievement.levels ) do
      local reward = level.reward;
      achievement.levels[ index ].reward_type  = reward_list[ reward ].type;
      achievement.levels[ index ].reward_name  = reward_list[ reward ].name;
      achievement.levels[ index ].reward_count = reward_list[ reward ].count;
      achievement.levels[ index ].reward = nil;
    end;
  end;
  
  return achievements, associated_ids;
  
end;
--------------------------------------------------------------------------------
function private.InitSubscribes ()
  
  mediator.Subscribe(
    "Model_Mines_MineCollected",
    private.CheckMineCollected
  );

  mediator.Subscribe(
    "View_Stage_FortuneWheel_Turn",
    private.IncreaseCounterWheel
  );

  mediator.Subscribe(
    "Model_Boosters_CountChanged",
     private.CheckBoostersUse
  );

  mediator.Subscribe(
    "Model_Stage_StageWon",
    private.OnStageWon
  );

  mediator.Subscribe(
    "Model_Stage_CellDestroyed",
    private.CheckBlocker
  );

  mediator.Subscribe(
    "Model_Stage_ChainActivated",
    function ( params ) private.CheckChips( params.chips_list ); end
  );

  mediator.Subscribe(
    "Model_Stage_BonusesSwapped",
    function ( params ) private.CheckBonus( params.combo_type ); end
  );

  mediator.Subscribe(
    "Model_Match3_BonusRemoved",
    function ( params ) private.CheckBonus( params.type ); end
  );

  --mediator.Subscribe(
  --  "Model_Shop_LotBought",
  --  private.CheckLotBought
  --);

  mediator.Subscribe(
    "View_Stage_FriendsWon",
    private.IncreaseCounterWonFriends
  );

  if ( not private.IsCompleted( WELCOME_HOME ) ) then
    mediator.Subscribe(
      "Model_Player_LaunchDaysRow",
      private.WelcomeHome_Increase
    );
  end;

  if ( not private.IsCompleted( ALWAYS_ONLINE ) ) then
    mediator.Subscribe(
      "Model_Game_SessionTick",
      private.AlwaysOnline_Increase
    );
  end;

  if ( not private.IsCompleted( MINIEVENT_GRINDER ) ) then
    mediator.Subscribe(
      "Model_Minievent_EventComplete",
      private.MinieventGrinder_Increase
    );
  end;

  if ( not private.IsCompleted( ARENA_MASTER ) ) then
    mediator.Subscribe(
      "View_Dailyquest_Finished",
      private.ArenaMaster_Increase
    );
  end;
  
end;
--------------------------------------------------------------------------------
function private.ResetCounters ()

  private.AlwaysOnline_Reset();
  private.StageCrusher_Reset();

end;
--------------------------------------------------------------------------------
function private.Android_IncrementAchievement ( achievement, count, unlock )

  if not ( IsAndroid() and wiz.Android.GetPlatform() == "google" ) then
    return;
  end;

  local query = 
  {
    query      = "find",
    group      = "static",
    collection = "achievements",
    indice     = "name",
    criteria   = { like = achievement }
  };
  
  local data = lib_db.Execute( query ).result;
  
  if ( data and data.google_play_ids ) then
    for _, id in pairs( data.google_play_ids ) do
      wiz.Android.IncrementAchievement( unlock, id, count );
    end;
  end;

end;
--------------------------------------------------------------------------------
function private.GetAchievementData ( achievement_name )
  
  local achievements = private.GetCollection();
  local id           = private.achievement_ids[ achievement_name ];
  
  return achievements[ id ];
  
end;
--------------------------------------------------------------------------------
function private.ChangeCount ( achievement_name, count_delta, reverse )

  local query = 
  {
    query      = "find",
    group      = "player",
    collection = "achievements",
    indice     = "name",
    criteria   = { like = achievement_name }
  };

  local achievement = lib_db.Execute( query ).result;

  local levels = private.GetAchievementData( achievement_name ).levels;

  if ( achievement.level > #levels               ) and
     ( achievement_name ~= "add_friends_fireman" ) then
    return;
  end;

  local new_count;

  if ( not reverse ) then

    new_count = achievement.count + count_delta;

  elseif ( count_delta < achievement.count ) or
         ( achievement.count == 0          ) then

    new_count = count_delta;

  end;

  query =
  {
    query      = "update",
    group      = "player",
    collection = "achievements",
    indice     = "name",
    criteria   = { like = achievement_name },
    set        = 
    {
      count = new_count
    }
  };

  lib_db.Execute( query );
  
  private.Android_IncrementAchievement( achievement_name, count_delta, false );

  private.TryComplete( achievement_name, reverse );

end;
--------------------------------------------------------------------------------
function private.ResetCount ( achievement_name, to_count )

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "achievements",
    indice     = "name",
    criteria   = { like = achievement_name },
    set        = { count = to_count or 0 }
  };

  lib_db.Execute( query );

end;
--------------------------------------------------------------------------------
function private.TryComplete ( achievement_name, reverse )

  local query = 
  {
    query      = "find",
    group      = "player",
    collection = "achievements",
    indice     = "name",
    criteria   = { like = achievement_name }
  };

  local achievement = lib_db.Execute( query ).result;

  local static_data = private.GetAchievementData( achievement_name );
  local levels = static_data.levels;

  if ( achievement_name == "add_friends_fireman" ) then
    if ( achievement.reward[ #levels ][ "get" ] ) then
      return;
    end;
  end;
  
  if ( achievement.level > #levels               ) and
     ( achievement_name ~= "add_friends_fireman" ) then
    return;
  end;

  local count_need;

  if ( achievement_name == "add_friends_fireman" ) then
    if ( achievement.level >= #levels ) then
      count_need = levels[ #levels - 1 ].value + levels[ #levels ].step * ( achievement.level - #levels + 1 );
    else
      count_need = levels[ achievement.level ].value;
    end;
  else
    count_need = levels[ achievement.level ].value;
  end;
  
  if (( not reverse ) and ( achievement.count >= count_need )) or 
  (( reverse ) and ( achievement.count <= count_need ) and ( achievement.count > 0 )) then
  
    local new_count;

    if ( not reverse ) then new_count = achievement.count; --achievement.count - count_need; 
    else                    new_count = achievement.count;
    end;
  
    local new_level = achievement.level + 1;

    if ( new_level > #levels                       ) and
       ( achievement_name ~= "add_friends_fireman" ) then
      new_count = 0;
    end;

    if ( achievement.discard ) then new_count = 0; end;

    local reward = achievement.reward;

    reward[ math.min( achievement.level, #levels ) ][ "get" ] = true;
    
    query =
    {
      query      = "update",
      group      = "player",
      collection = "achievements",
      indice     = "name",
      criteria   = { like = achievement_name },
      set        = 
      {
        level  = new_level,
        count  = new_count,
        reward = reward
      }
    };

    lib_db.Execute( query );

    private.TryComplete( achievement_name, reverse );
    
    local notification_name = tern( static_data.hidden )(
      "Model_Achievements_HiddenCompleted",
      "Model_Achievements_Completed"
    );
      
    mediator.Notify(
      notification_name,
      {
        name  = achievement_name,
        level = achievement.level
      }
    );

  end;

end;
--------------------------------------------------------------------------------
function private.IsCompleted ( achievement_name )
  
  local is_completed =
    private.GetAchievementData( achievement_name ).is_completed;
    
  return is_completed;

end;
--------------------------------------------------------------------------------
function private.IncreaseCounterWheel ( params )

  private.ChangeCount( private.achievement_types[ "TurnWheel" ][ 1 ].name, 1 );

end;
--------------------------------------------------------------------------------
function private.ResetCounterWheel ( params )

  local achievement_name = private.achievement_types[ "TurnWheel" ][ 1 ].name;

  local query = 
  {
    query      = "find",
    group      = "player",
    collection = "achievements",
    indice     = "name",
    criteria   = { like = achievement_name }
  };
  
  local achievement = lib_db.Execute( query ).result;

  for i = 1, #achievement.reward do
    if not ( achievement.reward[ i ].get == achievement.reward[ i ].done ) then
      private.ChangeCount( achievement_name, - achievement.count );
      break;
    end;
  end;

end;
--------------------------------------------------------------------------------
function private.CheckMineCollected ( mine_name )

  for a = 1, #private.achievement_types[ "MineCollected" ] do

    local achievement = private.achievement_types[ "MineCollected" ][ a ];

    if ( achievement.target == "none"    ) or
       ( mine_name == achievement.target ) then

      private.ChangeCount( achievement.name, 1 );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.CheckBoostersUse ( params )

  if ( params.delta >= 0 ) then return; end;

  for a = 1, #private.achievement_types[ "UseBooster" ] do

    local achievement = private.achievement_types[ "UseBooster" ][ a ];

    if ( params.name == achievement.target ) then
      private.ChangeCount( achievement.name, - params.delta );
    end;

  end;

end;
--------------------------------------------------------------------------------
function private.CheckStageTurns ( params )

  local ach_type;

  if ( m_stage.IsCurrentStageFight() ) then
    ach_type = "FightTurns"; 
  else
    return; --ach_type = "StageTurns";
  end;

  local moves = m_stage.MovesDone_GetCount();

  for a = 1, #private.achievement_types[ ach_type ] do

    local achievement = private.achievement_types[ ach_type ][ a ];

    if ( achievement.target == "none" ) or ( params.stage == achievement.target ) then

      private.ChangeCount( private.achievement_types[ ach_type ][ a ].name, moves, true );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.CheckBlocker ( chip )

  for a = 1, #private.achievement_types[ "DestroyBlocker" ] do

    local achievement = private.achievement_types[ "DestroyBlocker" ][ a ];

    if ( chip.name == achievement.target ) then

      private.ChangeCount( achievement.name, 1 );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.CheckChips ( chips )

  for i = 1, #chips do

    local chip = chips[ i ];

    for a = 1, #private.achievement_types[ "DestroyChip" ] do

      local achievement = private.achievement_types[ "DestroyChip" ][ a ];

      if ( achievement.target == "none" ) or ( chip.type == achievement.target ) then

        private.ChangeCount( achievement.name, 1 );

      end;

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.CheckBonus ( chip_type )

  for a = 1, #private.achievement_types[ "DestroyBonus" ] do

    local achievement = private.achievement_types[ "DestroyBonus" ][ a ];

    if ( achievement.target == "none" ) or 
    ( chip_type:find( achievement.target ) and not chip_type:find( "Second" )) then

      private.ChangeCount( achievement.name, 1 );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.CheckDifferentStage ( stage_name )

  if ( stage_name:find( "daily" ) or stage_name:find( "minievent" ) ) then
  
    return;
    
  end;

  local stage_num   = tonumber(stage_name:match( "%d+" ));
  local achievement = private.achievement_types[ "WinDiffStage" ][ 1 ];

  local field;
  if ( stage_name:find( "bonus" )) then
    field = "bonus_count";
  else
    field = "completed_count";
  end;

  local diff = stage_num - achievement[ field ];
  if ( diff > 0 ) then
    private.ChangeCount( achievement.name, diff );
    private.achievement_types[ "WinDiffStage" ][ 1 ][ field ] = stage_num;
  end;

end;
--------------------------------------------------------------------------------
function private.CheckLotBought ( params )

  for a = 1, #private.achievement_types[ "BuyLot" ] do

    local achievement = private.achievement_types[ "BuyLot" ][ a ];

    if ( params.lot_type == achievement.target ) then

      private.ChangeCount( achievement.name, 1 );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.IncreaseCounterWonFriends ( count )

  local achievement = private.achievement_types[ "WinFriend" ][ 1 ].name;
  private.ChangeCount( achievement, count );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.OnStageWon ( params )

  private.CheckDifferentStage( params.name );

  if ( not private.IsCompleted( STAGE_CRUSHER ) ) then
    private.StageCrusher_Increase( params.name, params.state_old );
  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.WelcomeHome_Increase ( count )

  if ( count == 0 ) then
    private.ResetCount( WELCOME_HOME );
  else
    private.ChangeCount( WELCOME_HOME, 1 );
  end;

  mediator.Notify(
    "Model_Achievements_AchievementChanged",
    public.Achievement_GetFullData( WELCOME_HOME )
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.AlwaysOnline_Increase ( params )

  local prev = private.always_online;
  local this = math.floor( params.duration / 60 );
  local diff = ( this - prev );

  private.always_online = this;

  private.ChangeCount( ALWAYS_ONLINE, diff );

  mediator.Notify(
    "Model_Achievements_AchievementChanged",
    public.Achievement_GetFullData( ALWAYS_ONLINE )
  );

end;
--------------------------------------------------------------------------------
function private.AlwaysOnline_Reset ()
  
  private.ResetCount( ALWAYS_ONLINE );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.StageCrusher_Increase ( stage_name, stage_state )

  local time_now = common.time.GetCurrent();
  
  -- JK: allow same stage with different state
  stage_name = stage_name.."_"..stage_state;

  if ( not private.stage_crusher ) then

    private.stage_crusher = {
      start  = time_now,
      finish = time_now + STAGE_CRUSHER_TIME,
      count  = 1,
      stages = { stage_name }
    };

    private.ChangeCount( STAGE_CRUSHER, 1 );
    private.StageCrusher_Start();

  else

    if ( common.IsEntityInList( private.stage_crusher.stages, stage_name ) ) then
      return;
    end;

    if ( time_now <= private.stage_crusher.finish ) then

      table.insert( private.stage_crusher.stages, stage_name );
      private.stage_crusher.count = #private.stage_crusher.stages;
      private.ChangeCount( STAGE_CRUSHER, 1 );

    else

      private.StageCrusher_Reset();
      private.StageCrusher_Increase( stage_name, stage_state );

    end;    

  end;

end;
--------------------------------------------------------------------------------
function private.StageCrusher_Reset ()

  if ( private.stage_crusher ) then
    private.stage_crusher = nil;
  end;

  private.ResetCount( STAGE_CRUSHER );

  mediator.Notify(
    "Model_Achievements_AchievementChanged",
    public.Achievement_GetFullData( STAGE_CRUSHER )
  );

end;
--------------------------------------------------------------------------------
function private.StageCrusher_Start ( params )

  mediator.Subscribe(
    "Model_Game_SessionTick",
    private.StageCrusher_CheckTime
  );

end;
--------------------------------------------------------------------------------
function private.StageCrusher_CheckTime ( params )

  local function func_unsubscribe ()
    mediator.Unsubscribe(
      "Model_Game_SessionTick",
      private.StageCrusher_CheckTime
    );
  end;

  if ( private.stage_crusher ) then

    if ( params.timestamp > private.stage_crusher.finish ) then
      func_unsubscribe();
      private.StageCrusher_Reset();
    end;

  else

    func_unsubscribe();

  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.MinieventGrinder_Increase ()

  private.ChangeCount( MINIEVENT_GRINDER, 1 );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.ArenaMaster_Increase ()

  private.ChangeCount( ARENA_MASTER, 1 );

end;
--------------------------------------------------------------------------------