-- name=m_boosters
--------------------------------------------------------------------------------
function public.__InitModule ()
  
  --Log.i( "Initialized." );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Setup ( funcs )
  
  private.func_get_stats = funcs.func_get_stats;
  
end;
--------------------------------------------------------------------------------
function public.Settings ( key )

  local settings = lib_db.Settings( key );

  return settings;

end;
--------------------------------------------------------------------------------
function private.GetSale ()
  
  if ( private.func_get_stats ) then
    return math.min( private.func_get_stats().sale_boosters, 1 );
  end;
  
  return 0;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.GetStaticData ( booster_name )
  
  local query =
  {
    query      = "find",
    group      = "static",
    collection = "boosters",
    indice     = "name",
    criteria   = { like = booster_name }
  };

  local booster = lib_db.Execute( query ).result;
  
  return booster;
  
end;
--------------------------------------------------------------------------------
function private.GetStorageData ( booster_name )
  
  local query =
  {
    query      = "find",
    group      = "player",
    collection = "boosters",
    indice     = "name",
    criteria   = { like = booster_name }
  };
  
  local booster = lib_db.Execute( query ).result;
  
  return booster;
  
end;
--------------------------------------------------------------------------------
function private.Update ( booster_name, booster_data )
  
  local query;
  local booster = private.GetStorageData( booster_name );
  
  if ( booster ) then
    
    query =
    {
      query      = "update",
      group      = "player",
      collection = "boosters",
      indice     = "name",
      criteria   = { like = booster_name },
      set        = booster_data
    };
    
  else
    
    booster_data.name = booster_name;
    booster_data.count = booster_data.count or 0;
    
    query =
    {
      query      = "insert",
      group      = "player",
      collection = "boosters",
      value      = booster_data
    };
    
  end;
  
  lib_db.Execute( query );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.GetBoostersData ( boosters_list )

  local boosters_data = {}; 

  for b = 1, #boosters_list do

    local booster_name = boosters_list[ b ].name;
    local database_answer = private.GetStaticData( booster_name );
    local booster_data = private.GetStorageData( booster_name );
    local booster = {};
    
    booster.name     = boosters_list[ b ].name;
    booster.packs    = public.GetPacks( booster.name );
    booster.type     = database_answer.type;
    booster.count    = 0;
    booster.unlocked = false;
    booster.priority = database_answer.priority or 0;

    if ( booster_data ) then
      booster.count    = booster_data.count    or booster.count;
      booster.unlocked = booster_data.unlocked or booster.unlocked;
    end;

    table.insert( boosters_data, booster );

  end;

  return boosters_data;

end;
--------------------------------------------------------------------------------
function public.GetBoosterStaticData ( booster_name )
  
  return private.GetStaticData( booster_name );
  
end;
--------------------------------------------------------------------------------
function public.GetList ( type )

  local query;

  if ( type ) then

    query =
    {
      query      = "find_all",
      group      = "static",
      collection = "boosters",
      indice     = "type",
      criteria   = { like = type }
    };

  else

    query =
    {
      query      = "find_all",
      group      = "static",
      collection = "boosters"
    };

  end;

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function public.IsUnlocked ( booster_name )
  
  local booster = private.GetStorageData( booster_name );
  
  if ( booster ) and ( booster.unlocked ) then
    return booster.unlocked;
  end;
  
  return false;

end;
--------------------------------------------------------------------------------
function public.GetPrice ( booster_name, level )

  return public.GetPacks( booster_name )[ level or 1 ].price;

end;
--------------------------------------------------------------------------------
function public.GetCount ( booster_name )

  local booster = private.GetStorageData( booster_name );
  
  if ( booster ) and ( booster.count ) then
    return booster.count;
  end;

  return 0;

end;
--------------------------------------------------------------------------------
function public.GetPacks ( booster_name )
  
  local packs = private.GetStaticData( booster_name ).packs;
      
  if ( parts.features.IsActive( "prices" ) ) then
    
    local feature_data = parts.features.GetFeatureData( "prices" );
    
    if ( feature_data and
         feature_data[ 1 ] and
         feature_data[ 1 ].boosters ) then
         
      local boosters_list = feature_data[ 1 ].boosters;
      local index = common.IsEntityInList( boosters_list, 
                                          { name = booster_name }, 
                                          "name" );
                                          
      if ( index and 
           boosters_list[ index ].new_price and
           boosters_list[ index ].new_price > 0 ) then
        for i = 1, #packs do
          
          if ( packs[ i ].price > boosters_list[ index ].new_price ) then
            packs[ i ].sale = packs[ i ].price - boosters_list[ index ].new_price;
          end;
          
          packs[ i ].price = boosters_list[ index ].new_price;
          
        end;
        
      end;
      
    end;
    
  end;
  
  local sale = private.GetSale();
  local sale_ratio = 1 - sale;
  
  for i = 1, #packs do
    if ( sale > 0 ) then
      packs[ i ].sale = ( packs[ i ].sale or 0 ) + packs[ i ].price * sale;
    end;
    packs[ i ].price = math.ceil( packs[ i ].price * sale_ratio );
  end;
  
  return packs;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.TryChangeCount ( booster_name, count_delta, source, buy )
  
  local booster_data = {
    count = count_delta + public.GetCount( booster_name );
  };
  
  if ( booster_data.count < 0 ) then return false; end;
  
  private.Update( booster_name, booster_data );

  local description = "item_collected";
  
  if ( count_delta < 0 ) then
    description = "item_used";
  end;
  
  if ( source ) and ( string.find( source, "stage" ) ) then
    local _, stage_type = string.match( source, "stage_(%d+)(_%w+)" );
    source = "level"..( stage_type or "" );
  end;  
  
  v_log.Kontagent_LogEvent( "economy", 
  {
    description = description,
    category    = "boosts",
    value       = count_delta,
    name        = booster_name,
    source      = source
  } );  
  
  mediator.Notify( "Model_Boosters_CountChanged",
  {
    name   = booster_name,
    delta  = count_delta,
    count  = booster_data.count,
    source = source 
  } );
  
  return true;

end;
--------------------------------------------------------------------------------
function public.Unlock ( booster_name )
  
  local booster_data = {
    unlocked = true;
  };
  
  private.Update( booster_name, booster_data );
  
end;
--------------------------------------------------------------------------------