-- name=m_builds
--------------------------------------------------------------------------------
function public.__InitModule ()
  
  private.puzzle_unlockers = {};
  
  local puzzles = private.Build_GetStaticData( { type = "puzzle" }, "find_all" );
  
  for p = 1, #puzzles do
    
    if ( puzzles[ p ].reward ) then
      
      private.puzzle_unlockers[ puzzles[ p ].unlock ] = puzzles[ p ];
      
    end;
    
  end;

  --Log.i( "Initialized." );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Builds_GetAllBySection ( section_name, with_saved_data )
  
  local builds = private.Build_GetStaticData( { section = section_name }, "find_all" );

  if ( with_saved_data ) then
  
    for i = 1, #builds do
      
      private.Build_AppendSavedData( builds[ i ] );
      
    end;
  
  end;
  
  return builds;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Build_GetData ( build_name )

  local build_data = private.Build_GetStaticData( { name = build_name } );

  if ( build_data ) then

    private.Build_AppendSavedData( build_data );

  end;

  return build_data;

end;
--------------------------------------------------------------------------------
function public.Build_GetDataByUnlocker ( stage_name, with_saved_data )

  local build_data = private.Build_GetStaticData( { unlock = stage_name } );
  
  if ( build_data and with_saved_data ) then
    
    private.Build_AppendSavedData( build_data );
    
  end;
  
  return build_data;

end;
--------------------------------------------------------------------------------
function public.Build_GetDataByDeadlock ( deadlock )
  
  local build_data = private.Build_GetStaticData( { deadlock = deadlock } );

  if ( build_data ) then

    private.Build_AppendSavedData( build_data );

  end;

  return build_data;

end;
--------------------------------------------------------------------------------
function public.Build_GetNameByFragment ( fragment_name )

  return ( private.Build_GetDataByFragment( fragment_name ) or {} ).name;
  
end;

--------------------------------------------------------------------------------
function public.Build_IsConstructed ( build_name )
  
  return ( private.Build_GetSavedData( build_name ) or {} ).constructed;

end;
--------------------------------------------------------------------------------
function public.Build_IsAutoConstructed ( build_name, build_data ) -- build_data is optional

  if ( not build_data ) then

    build_data = private.Build_GetStaticData( { name = build_name } );
  
  end;
  
  return private.Build_IsAutoConstructed( build_data );

end;
--------------------------------------------------------------------------------
function public.Build_GetUnlocklistByMin ( min_stage )
  
  return private.Build_GetStaticData( { [ "unlocklist.min" ] = min_stage } );

end;
--------------------------------------------------------------------------------
function private.Build_IsAutoConstructed ( build_data )

  if ( not build_data.price     ) and
     ( not build_data.fragments ) then

    return true;

  end;

  return false;

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function public.Build_OnStageCompleted ( stage_name )

  local build_data = private.Build_GetStaticData( { unlock = stage_name } );

  local set_data = {};
  
  if ( build_data ) then
    
    if ( build_data.unlock ) then
      
      set_data.unlocked = true;
      
      if ( private.Build_IsAutoConstructed( build_data ) ) then

        set_data.constructed = true;

      else

        if ( public.Build_IsConstructAccepted( build_data.name, set_data ) ) then
          
          set_data.constructed = true;
          
        end;

      end;
      
      private.Build_SetData( build_data.name, set_data );

    else
      
      --DbgTrace( build_data.type.." < "..build_data.name.." > : no actions." );
      
    end;
    
  end;
  
  return ( build_data or {} ).name, set_data.constructed;

end;
--------------------------------------------------------------------------------
function public.Build_OnFragmentCollected ( fragment_name )

  local build_data = private.Build_GetDataByFragment( fragment_name );
  
  local set_data = {};
  
  if ( build_data ) then
    
    local saved_data = private.Build_GetSavedData( build_data.name ) or {};

    if ( common.IsEntityInList( saved_data.collected_fragments, fragment_name ) ) then

      DbgTrace(
        "[ ERROR ] m_builds -> Build_OnFragmentCollected -> "..
        fragment_name.." already in save for "..build_data.name
      );
      return false, nil;

    end;
    
    set_data.collected_fragments = saved_data.collected_fragments or {};

    table.insert( set_data.collected_fragments, fragment_name );
    
    if ( not saved_data.unlocked and 
         not build_data.unlock ) then
      
      set_data.unlocked = true;
      
    end;

    if ( public.Build_IsConstructAccepted( build_data.name, set_data ) ) then

      set_data.constructed = true;

    end;
    
    private.Build_SetData( build_data.name, set_data );
    
  else

    DbgTrace(
      "[ ERROR ] m_builds -> Build_OnFragmentCollected -> No build for "..
      fragment_name
    );
    return false, nil;
    
  end;
  
  return ( build_data or {} ).name, set_data.constructed;

end;
--------------------------------------------------------------------------------
function public.Build_SetData ( build_name, set_data )
  
  return private.Build_SetData( build_name, set_data );
  
end;
--------------------------------------------------------------------------------
function private.Build_SetData ( build_name, set_data )
  
  if ( not set_data or 
       set_data == {} ) then
    
    DbgTrace( "[ WARNING ][ m_builds ] Build_SetData: set_data is empty!" );
       
  end;
  
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "builds",
    indice     = "name",
    criteria   = { like = build_name },
    set = set_data
  };

  if ( not lib_db.Execute( query ).result ) then
    
    set_data.name = build_name;

    query =
    {
      query      = "insert",
      group      = "player",
      collection = "builds",
      value      = set_data
    };

    lib_db.Execute( query );
    
  end;
  
  if ( set_data.constructed ) then
    
    local build_type = private.Build_GetStaticData( { name = build_name } ).type;
    
    mediator.Notify(
    "Model_Builds_BuildConstructed",
      {
        name = build_name,
        type = build_type
      }
    );

  end;

  --DbgTrace( "Build_SetData: < "..build_name.." >", set_data );
  
  common.Save();
  
end;
--------------------------------------------------------------------------------
function public.Build_IsConstructAccepted ( build_name, set_data )
  
  set_data = set_data or {};
  
  local static_data = private.Build_GetStaticData( { name = build_name } );
  local saved_data  = private.Build_GetSavedData( build_name ) or {};
  local unlocked    = set_data.unlocked or saved_data.unlocked;
  
  if ( not unlocked or static_data.price ) then
    
    --DbgTrace( static_data.type.." < "..static_data.name.." > : can't construct -> not unlocked or manual build." );
    return false, static_data.section, static_data.fragments;
    
  end;

  local frags = set_data.collected_fragments or saved_data.collected_fragments;
  local total = static_data.fragments;
      
  if ( total and frags and ( #frags == #total ) ) then
  
    return true, static_data.section, static_data.fragments;
  
  end;
  
  return false, static_data.section, static_data.fragments;
  
end;
--------------------------------------------------------------------------------
function private.Build_GetDataByFragment ( fragment_name )

  local query = 
  {
    query      = "find_all",
    group      = "static",
    collection = "builds"
  };
  local db_answer = lib_db.Execute( query ).result;

  if ( db_answer ) then

    for i = 1, #db_answer do
      
      local build_name = db_answer[ i ].name;
      local fragments  = db_answer[ i ].fragments;

      if ( fragments                                       ) and
         ( common.IsEntityInList( fragments, fragment_name ) ) then

         return private.Build_GetStaticData( { name = build_name } );

      end;

    end;

  end;

  return nil;
  
end;
--------------------------------------------------------------------------------
function private.Build_GetStaticData ( indice_table, query_type )
  
  query_type = query_type or "find"; 

  for key, value in pairs( indice_table ) do

    local query = 
    {
      query      = query_type,
      group      = "static",
      collection = "builds",
      indice     = key,
      criteria   = { like = value }
    };

    return lib_db.Execute( query ).result;
    
  end;

end;
--------------------------------------------------------------------------------
function private.Build_GetSavedData ( build_name )

  local query = 
  {
    query      = "find",
    group      = "player",
    collection = "builds",
    indice     = "name",
    criteria   = { like = build_name }
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function private.Build_AppendSavedData ( build_data )

  local saved_data = private.Build_GetSavedData( build_data.name );

  if ( saved_data ) then

    for key, value in pairs( saved_data ) do

      if ( not build_data[ key ] ) then

        build_data[ key ] = value;

      end;

    end;

  end;

end;
--------------------------------------------------------------------------------