-- name=m_carriers
--------------------------------------------------------------------------------
function public.__InitModule ()
  
  private.event = {
    ready = "Event_Game_Carriers_CarrierReady"
  };
  
  private.settings = lib_db.Settings( "carrier" );

  ne.msg.Subscribe(
    private.event.ready,
    function ( msg, params )

      private.Stop( params.name );

    end
  );

  private.FixSave();
  
  --Log.i( "Initialized." );

end;
--------------------------------------------------------------------------------
function private.FixSave ()
  
  local carriers = private.DataAll();
  
  if ( #carriers == 0 ) then
    return;
  end;
  
  for i = 1, #carriers do
    
    if ( carriers[ i ].resource ) and ( not private.IsScheduled( carriers[ i ].name ) ) then
      local query = 
      {
        query      = "update",
        group      = "player",
        collection = "carriers",
        indice     = "name",
        criteria   = { like = carriers[ i ].name },
        set        = {
          resource = false
        }
      };
      lib_db.Execute( query );
    end;
    
  end;
  
end;
--------------------------------------------------------------------------------
-- function ...() end;
--------------------------------------------------------------------------------
function public.Cheat ()

  private.settings.time = 20;

  return true;

end;
--------------------------------------------------------------------------------
function public.IsBusy ( carrier_name )

  local carrier_data = private.Data( carrier_name );

  return ( ( carrier_data ) and ( carrier_data.resource ) );

end;
--------------------------------------------------------------------------------
function public.GetRemainingTime ( carrier_name )

  if ( private.IsScheduled( carrier_name ) ) then

    return parts.time.event.GetRemainingTime( "mte_"..carrier_name );

  end;

  return 0;

end;
--------------------------------------------------------------------------------
function public.GetRandomEvent ()

  local busy_list = private.ListBusy();
  
  if ( #busy_list < 1 ) then
    
    return nil;
    
  end;
  
  if ( #busy_list == 1 ) then
    
    return busy_list[ 1 ];
    
  end;
  
  if ( #busy_list > 1 ) then
  
    local random = math.random( 1, #busy_list );
  
    return busy_list[ random ];
    
  end;

end;
--------------------------------------------------------------------------------
function public.GetDefaultDeliveryTime ()

  return private.settings.time;

end;
--------------------------------------------------------------------------------
function public.GetUnlock ()

  return private.settings.unlock;

end;
--------------------------------------------------------------------------------
-- function ...() end;
--------------------------------------------------------------------------------
function public.Data ( carrier_name )
  
  return private.Data( carrier_name );
  
end;
--------------------------------------------------------------------------------
function public.State ( carrier_name )
  
  local carrier_data = private.Data( carrier_name );
  
  if ( carrier_data       ) and
     ( carrier_data.state ) then
    
    return carrier_data.state;
    
  end;
  
  return nil;
  
end;
--------------------------------------------------------------------------------
function public.Capacity ( carrier_name, carrier_state )
  
  carrier_state = carrier_state or private.Data( carrier_name ).state;
  
  if ( carrier_state == 1 ) then
    
    return private.settings.start_capacity;
  
  end;
  
  return private.settings.upgrades[ carrier_state - 1 ].capacity;
  
end;  
--------------------------------------------------------------------------------
-- function ...() end;
--------------------------------------------------------------------------------
function public.Price_Build ( carrier_name )
  
  return private.settings.build_price[ carrier_name ] or
         private.settings.build_price.default;
  
end;
--------------------------------------------------------------------------------
function public.Price_SpeedUp ( carrier_name )
    
  return private.settings.skip_price[ carrier_name ] or
         private.settings.skip_price.default;
  
end;
--------------------------------------------------------------------------------
function public.Price_Upgrade ( carrier_name )
  
  local carrier_state = private.Data( carrier_name ).state;
  
  return private.settings.upgrades[ carrier_state ].price;
  
end;
--------------------------------------------------------------------------------
-- function ...() end;
--------------------------------------------------------------------------------
function public.Build ( carrier_name )
    
  local query =
  {
    query      = "insert",
    group      = "player",
    collection = "carriers",
    value      = {
      name     = carrier_name,
      state    = 1,
      resource = false
    }
  };

  return lib_db.Execute( query ).result;
  
end;
--------------------------------------------------------------------------------
function public.Upgrade ( carrier_name, new_state )
    
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "carriers",
    indice     = "name",
    criteria   = { like = carrier_name },
    set        = {
      state = new_state
    }
  };

  return lib_db.Execute( query ).result;
  
end;
--------------------------------------------------------------------------------
function public.Send (
    carrier_name,
    resource_name,
    resource_count, 
    from_section_number,
    to_section_number,
    delivery_time
  )
  
  return private.Start(
    carrier_name,
    resource_name,
    resource_count,
    from_section_number,
    to_section_number,
    delivery_time
  );
  
end;
--------------------------------------------------------------------------------
function public.SpeedUp ( carrier_name, skip_notification )
  
  return private.Stop( carrier_name, skip_notification );
  
end;
--------------------------------------------------------------------------------
function public.IsScheduled ( carrier_name )

  return private.IsScheduled ( carrier_name );

end;
--------------------------------------------------------------------------------
-- function ...() end;
--------------------------------------------------------------------------------
function private.Data ( carrier_name )
  
  local query = 
  {
    query      = "find",
    group      = "player",
    collection = "carriers",
    indice     = "name",
    criteria   = { like = carrier_name }
  };

  return lib_db.Execute( query ).result;
  
end;
--------------------------------------------------------------------------------
function private.DataAll ()
  
  local query = 
  {
    query      = "find_all",
    group      = "player",
    collection = "carriers"
  };

  return lib_db.Execute( query ).result;
  
end;
--------------------------------------------------------------------------------
function private.Start (
    carrier_name,
    resource_name,
    resource_count,
    from_section_number,
    to_section_number,
    delivery_time
  )

  local time = delivery_time or public.GetDefaultDeliveryTime();

  parts.time.event.Shedule(
    "mte_"..carrier_name,
    "GLOBAL_TIME",
    time,
    private.event.ready,
    {
      name = carrier_name
    }
  );
  
  local query = 
  {
    query      = "update",
    group      = "player",
    collection = "carriers",
    indice     = "name",
    criteria   = { like = carrier_name },
    set        = { 
      resource = {
        name    = resource_name,
        count   = resource_count,
        section = from_section_number,
        time    = delivery_time
      }
    }
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function private.Stop ( carrier_name, skip_notification )

  TimeFinishEvent();
  TimeUnsheduleEvent( "mte_"..carrier_name );
  
  local resource = private.Data( carrier_name ).resource;

  local query = 
  {
    query      = "update",
    group      = "player",
    collection = "carriers",
    indice     = "name",
    criteria   = { like = carrier_name },
    set        = {
      resource = false
    }
  };

  lib_db.Execute( query );

  local data = {
    carrier_name   = carrier_name,
    resource_name  = resource.name,
    resource_count = resource.count,
    section_number = resource.section
  };

  if ( not skip_notification ) then

    mediator.Notify(
      "Model_Carriers_CarrierFinished",
      data
    );

  end;

  return data;
  
end;
--------------------------------------------------------------------------------
function private.IsScheduled ( carrier_name )

  return parts.time.event.Exists( "mte_"..carrier_name );

end;
--------------------------------------------------------------------------------
function private.ListBusy ()

  local busy_list = {};
  local callbacks = lib_db.Events_ListCallbacks();

  if ( not callbacks ) then

    return {};

  end;

  for c, callback in pairs( callbacks ) do

    if ( callback.token:find( "carrier" ) ) and
       ( callback.time > 0                ) then
   
      table.insert(
        busy_list,
        {
          name = callback.data.name,
          time = callback.time
        }
      );
   
    end;
    
  end;
  
  return busy_list;
  
end;
--------------------------------------------------------------------------------