-- name=m_characters
--------------------------------------------------------------------------------

--[[ Char states: 
  locked (for external usage), 
  unlocked, 
  active, 
  inactive
]]

--------------------------------------------------------------------------------
function public.__InitModule ()

  mediator.Subscribe(
    "Model_Sections_SectionEnabled",
    private.OnSectionEnabled
  );
  
  mediator.Subscribe( "Model_Quests_PregiveStarted",   function ( quest_name ) 
    
    local character = public.GetCharacterByQuest( quest_name );
    if  ( character ) then

      mediator.Notify( "Model_Characters_PregiveStarted", character.name );

    end;

  end );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.GetAllInSection ( section_name )

  local query = 
  {
    query      = "find_all",
    group      = "static",
    collection = "characters",
    indice     = "section",
    criteria   = { like = section_name }
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function public.GetQuestByItemFromStage ( item_name, stage_name )

  local query = {
    query      = "find_all",
    group      = "static",
    collection = "quests",
    indice     = "target.name",
    criteria   = { like = item_name }
  };

  local quest_list = lib_db.Execute( query ).result;

  for q = 1, #quest_list do

    if ( common.IsEntityInList( quest_list[ q ].target.list, stage_name ) ) then

      return quest_list[ q ];

    end;

  end;

end;
--------------------------------------------------------------------------------
function public.IsQuestForItemFromStageDone ( item_name, stage_name )

  local quest = public.GetQuestByItemFromStage( item_name, stage_name );

  if ( quest.reward ) then

    local query =
    {
      query      = "find",
      group      = "player",
      collection = "quests_completed",
      indice     = "name",
      criteria   = { like = quest.name }
    };
    
    local quest_completed = lib_db.Execute( query ).result;

    if ( quest_completed ) then

      return true;

    end;

  end;

  return false;

end;
--------------------------------------------------------------------------------
function public.GetCharacterByItemFromStage ( item_name, stage_name )

  local target_quest = public.GetQuestByItemFromStage( item_name, stage_name ).name;

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "characters",
    indice     = "quest_cage",
    criteria   = { like = target_quest }
  };
  
  local character = lib_db.Execute( query ).result;

  if ( character ) and ( character.name ) then return character.name; end;

  query =
  {
    query      = "find",
    group      = "static",
    collection = "characters",
    indice     = "quest_free",
    criteria   = { like = target_quest }
  };
  
  return lib_db.Execute( query ).result.name;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.GetContent ( character_name )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "characters",
    indice     = "name",
    criteria   = { like = character_name }
  };

  return lib_db.Execute( query ).result.content;

end;
--------------------------------------------------------------------------------
function public.GetUnlockedType ( character_name )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "characters",
    indice     = "name",
    criteria   = { like = character_name }
  };

  return lib_db.Execute( query ).result.unlocked_type;

end;
--------------------------------------------------------------------------------
function private.OnSectionEnabled ( params )

  local section_name     = params.section_name;
  local start_characters = params.section_config.characters.start_characters;

  for s = 1, #start_characters do

    local character_name = start_characters[ s ];

    if ( not private.GetCharacterState( character_name ) ) then

      local query =
      {
        query      = "update",
        group      = "player",
        collection = "characters",
        indice     = "name",
        criteria   = { like = character_name },
        set        =
        {
          state = "unlocked"
        }
      };

      if ( not lib_db.Execute( query ).result ) then

        query =
        {
          query      = "insert",
          group      = "player",
          collection = "characters",
          value      = 
          {
            name  = character_name,
            state = "unlocked"
          }
        };
        
        lib_db.Execute( query );

      end;

    end;  

  end;

end;
--------------------------------------------------------------------------------
function public.GetCharacterByQuest ( quest_name )

  local character = nil;

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "characters",
    indice     = "quest_cage",
    criteria   = { like = quest_name }
  };

  character = lib_db.Execute( query ).result;
  
  if ( not character ) then

    query =
    {
      query      = "find",
      group      = "static",
      collection = "characters",
      indice     = "quest_free",
      criteria   = { like = quest_name }
    };
    
    character = lib_db.Execute( query ).result;

  end;
  
  return character;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.GetCharacterState ( character_name )

  return private.GetCharacterState( character_name ) or "locked";

end;
--------------------------------------------------------------------------------
function private.GetCharacterState ( name )

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "characters",
    indice     = "name",
    criteria   = { like = name }
  };
  
  local character = lib_db.Execute( query ).result;
  
  if ( character       ) and
     ( character.state ) then
    return character.state;
  end;

end;
--------------------------------------------------------------------------------
function public.SetCharacterState ( name, state )

  local query;
  
  if ( private.GetCharacterState( name ) ) then

    query =
    {
      query      = "update",
      group      = "player",
      collection = "characters",
      indice     = "name",
      criteria   = { like = name },
      set        = { state = state }
    };
    
    lib_db.Execute( query );

  else

    query =
    {
      query      = "insert",
      group      = "player",
      collection = "characters",
      value      = 
      { 
        name     = name,
        state    = state 
      }
    };
    
    lib_db.Execute( query );

  end;

end;
--------------------------------------------------------------------------------
function private.GetQuestCurrent ( quest_name )

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "quests_current",
    indice     = "name",
    criteria   = { like = quest_name }
  };
  
  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function private.GetQuestComplete ( quest_name )

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "quests_completed",
    indice     = "name",
    criteria   = { like = quest_name }
  };
  
  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function public.GetCharacterData ( character_name )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "characters",
    indice     = "name",
    criteria   = { like = character_name }
  };
  
  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function public.GetQuestPopupData ( quest_data )

  local quest_current = private.GetQuestCurrent( quest_data.name ) or 
                        { count = #quest_data.target.list };

  local data =
  {
    name    = quest_data.name,
    count   = quest_current.count,
    list    = quest_current.list or {},
    target  = quest_data.target,
    content = quest_data.content,
    reward  = quest_data.reward
  };
  
  return data;

end;
--------------------------------------------------------------------------------
function public.GetCurrentQuestName ( character_name )

  local data = public.GetCharacterData( character_name ) or {};

  local quest_cage = private.GetQuestCurrent( data.quest_cage or "" );
  if ( quest_cage ) then return quest_cage.name; end;
  
  local quest_free = private.GetQuestCurrent( data.quest_free or "" );
  if ( quest_free ) then return quest_free.name; end;
  
  return nil;

end;
--------------------------------------------------------------------------------
function public.GetCharacterByBuild ( build_name )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "characters",
    indice     = "unlock",
    criteria   = { like = build_name }
  };
  
  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------