-- name=m_chests
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.section_chests = {};

  --Log.i( "Initialized." );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.List_OnSection ( section_name )

  local query =
  {
    query      = "find_all",
    group      = "static",
    collection = "chests",
    indice     = "section",
    criteria   = { like = section_name }
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function public.List_Unlocked ( section_name )

  local query =
  {
    query      = "find_all",
    group      = "player",
    collection = "chests"
  };
  
  local unlocked_full_list  = lib_db.Execute( query ).result;
  local unlocked_on_section = {};
  
  for c, chest in pairs( unlocked_full_list ) do
  
    if ( private.Origin_Section( chest.name ) == section_name ) then
      
      table.insert( unlocked_on_section, chest );
      
    end;
    
  end;
  
  return unlocked_on_section;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Chest_Status ( chest_name )

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "chests",
    indice     = "name",
    criteria   = {
      like = chest_name
    }
  };

  local chest = lib_db.Execute( query ).result;

  if ( chest and chest.opened ) then

    return "opened";

  end;

  if ( chest and not chest.opened ) then

    return "available";

  end;

  return "not_available";

end;
--------------------------------------------------------------------------------
function public.Chest_Data ( chest_name )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "chests",
    indice     = "name",
    criteria   = { 
      like = chest_name
    }
  };

  return lib_db.Execute( query ).result;  

end;
--------------------------------------------------------------------------------
function public.Chest_NameByStage ( stage_name )

  return private.Origin_Stage( stage_name );

end;
--------------------------------------------------------------------------------
function public.Chest_Unlock( chest_name )
  
  local query =
  {
    query      = "insert",
    group      = "player",
    collection = "chests",
    value      = {
      name   = chest_name,
      opened = false
    }
  };

  lib_db.Execute( query );

end;
--------------------------------------------------------------------------------
function public.Chest_Open ( chest_name )

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "chests",
    indice     = "name",
    criteria   = { like = chest_name },
    set        = { opened = true }
  };

  lib_db.Execute( query );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Origin_Stage ( stage_name )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "chests",
    indice     = "stage",
    criteria   = {
      like = stage_name
    }
  };

  local chest = lib_db.Execute( query ).result;

  if ( chest      ) and
     ( chest.name ) then

    return chest.name;

  end;

  return nil;

end;
--------------------------------------------------------------------------------
function private.Origin_Section ( chest_name )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "chests",
    indice     = "name",
    criteria   = { like = chest_name }
  };

  local chest = lib_db.Execute( query ).result;

  if ( chest ) and ( chest.section ) then

    return chest.section;

  end;

  return false;

end;
--------------------------------------------------------------------------------