-- name=m_dailybonus
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.daily_hour = private.Settings( "dailyhour" );
  private.days_max   = 10;
  private.token      = "mte_dailybonus_refresh";

  if not ( private.Event_IsScheduled() ) then
    private.Event_Schedule();
  end;

  ne.msg.Subscribe(
    "Event_DailyBonus_Refresh",
    function ( msg_type, msg_params )
      private.Refresh( msg_params );
    end
  );

  --Log.i( "Initialized." );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Refresh ()

  local params = {
    expire = TimeServerNowTimestamp()
  };
  
  private.Refresh( params );

end;
--------------------------------------------------------------------------------
function public.Finish ()

  local current_day = private.Day_GetCurrent();
  local next_day    = current_day + 1;

  if ( next_day > private.days_max ) then

    next_day = 1;

  end;

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "dailybonus",
    set        = {
      day   = next_day,
      ready = false
    }
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.IsReady ()

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "dailybonus"
  };

  return lib_db.Execute( query ).result.ready;

end;
--------------------------------------------------------------------------------
function public.Settings ( item )

  return private.Settings( item );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Day_GetMax ()

  return private.days_max;

end;
--------------------------------------------------------------------------------
function public.Day_GetPeriodLimits ( current_day )

  local limits = { 1, private.days_max / 2 };

  if ( current_day > ( private.days_max / 2 ) ) then

    limits = { ( private.days_max / 2 ) + 1, private.days_max };

  end;

  return limits;

end;
--------------------------------------------------------------------------------
function public.Day_ToRelative ( current_day )

  if ( current_day > ( private.days_max / 2 ) ) then

    return current_day - ( private.days_max / 2 );

  end;

  return current_day;

end;
--------------------------------------------------------------------------------
function public.Day_GetCurrent ()

  return private.Day_GetCurrent();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Start ()

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "dailybonus",
    set        = {
      ready = true
    }
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function private.Refresh ( params )
  
  local time = {
    expire = params.expire,
    now    = TimeServerNowTimestamp()
  };
  
  time.diff = time.now - time.expire;
  
  private.Event_Unschedule();
  
  -- JK: set dailybonus ready if player returned back to game in less than 1 day
  -- from last dailybonus expiration time, otherwise reset to day 1 and
  -- schedule a new event
  
  tern( time.diff < 86400 )(
    private.Start,
    function () private.Day_Set( 0 ); end
  )();
  
  private.Event_Schedule();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Day_GetCurrent ()

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "dailybonus"
  };

  return lib_db.Execute( query ).result.day;

end;
--------------------------------------------------------------------------------
function private.Day_Set ( number )

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "dailybonus",
    set        = {
      day = number
    }
  };

  lib_db.Execute( query );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Event_Schedule ()

  local time_now   = TimeServerNow();
  local timestamp  = TimeServerNowTimestamp();
  local hour_end   = private.daily_hour;
  local hour_delta = hour_end - time_now.tm_hour;

  if ( hour_delta <= 0 ) then

    hour_delta = 24 - time_now.tm_hour + hour_end;

  end;

  local new_time    = hour_delta * 60 * 60 - ( time_now.tm_min ) * 60 - ( time_now.tm_sec );
  local expire_time = timestamp + new_time + 86400;

  TimeSheduleEvent(
    private.token,
    "GLOBAL_TIME",
    new_time,
    "Event_DailyBonus_Refresh",
    {
      expire = expire_time
    }
  );

end;
--------------------------------------------------------------------------------
function private.Event_IsScheduled ()

  return TimeIsEventExist( private.token );

end;
--------------------------------------------------------------------------------
function private.Event_Unschedule ()

  TimeFinishEvent();
  TimeUnsheduleEvent( private.token );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Settings ( item )

  return lib_db.Settings( "dailybonus" )[ item ];

end;
--------------------------------------------------------------------------------