-- name=m_dailyquest
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.settings                 = lib_db.Settings( "dailyquest" );
  private.states                   = { "open", "beaten", "done" };
  private.settings.challenge_count = #private.settings.rewards;

  private.tokens = {
    dailyquest = "mte_dailyquest_refresh",
    countdown  = "mte_dailyquest_countdown_refresh",
    challenge  = "mte_dailyquest_challenge_refresh"
  };

  private.events = {
    dailyquest = "Event_DailyQuest_Refresh",
    countdown  = "Event_DailyQuest_CountdownRefresh",
    challenge  = "Event_DailyQuest_ChallengeRefresh"
  };

  ne.msg.Subscribe(
    private.events.countdown,
    function () public.Unschedule( "countdown" ); end
  );

  ne.msg.Subscribe(
    private.events.challenge,
    function () public.Unschedule( "challenge" ); end
  );
  
  --Log.i( "Initialized." );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Setup ( is_unlocked )

  if ( is_unlocked ) then

    if ( not private.IsScheduled( "dailyquest" ) ) then

      public.Schedule( "dailyquest" );

    end;

  end;

end;  
--------------------------------------------------------------------------------
function public.IsReady ()

  return public.Data_Get( "ready" );

end;
--------------------------------------------------------------------------------
function public.GetChallengeCount ()

  return private.settings.challenge_count;

end;
--------------------------------------------------------------------------------
function public.GetStates ()

  return private.states;

end;
--------------------------------------------------------------------------------
function public.GetSkipPrice ()

  return private.settings.skipprice;

end;
--------------------------------------------------------------------------------
function public.GetSettings ()

  return private.settings;

end;
--------------------------------------------------------------------------------
function public.GetStageList ()

  local query =
  {
    query      = "find_all",
    group      = "static",
    collection = "stages",
    indice     = "type",
    criteria   = { like = "dailyquest" }
  };
  
  return lib_db.Execute( query ).result;
  
end;
--------------------------------------------------------------------------------
function public.GetRemainingTime ( timer_type )

  if ( private.IsScheduled( timer_type ) ) then

    return parts.time.event.GetRemainingTime( private.tokens[ timer_type ] );

  end;

  return nil;

end;
--------------------------------------------------------------------------------
function public.GetUnlockStage ()
  
  return private.settings.unlockstage;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Reward_GetSettings ( reward_id )

  return private.settings.rewards[ reward_id ];

end;
--------------------------------------------------------------------------------
function public.Reward_Set ( index )

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "dailyquest",
    set        = { reward = index }
  };
  
  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Countdown_IsActive ()

  return private.IsScheduled( "countdown" );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Schedule ( timer_type, seconds )

  local time;

  if ( timer_type == "dailyquest" ) then

    local time_def   = { tm_hour = 0, tm_min = 0, tm_sec = 0 };
    local time_now   = parts.time.server.Now() or time_def;
    local hour_end   = private.settings.dailyhour;
    local hour_delta = hour_end - time_now.tm_hour;

    if ( hour_delta <= 0 ) then
      
      hour_delta = 24 - time_now.tm_hour + hour_end;
      
    end;

    time = hour_delta * 60 * 60 - ( time_now.tm_min ) * 60 - ( time_now.tm_sec );

  elseif ( timer_type == "countdown" ) then
    
    time = private.settings.unlocktime;

  elseif ( timer_type == "challenge" ) then
    
    time = seconds;

  else

    DbgTrace(
      "[ ERROR ][ m_dailyquest ] Unknown timer: "..tostring( timer_type )
    );

  end;

  if ( not time ) then

    DbgTrace(
      "[ ERROR ][ m_dailyquest ] Intent to set cooldown for "..tostring( timer_type ).." with unspecified time."
    );
    return;

  end;

  parts.time.event.Shedule( 
    private.tokens[ timer_type ], 
    "GLOBAL_TIME", 
    time, 
    private.events[ timer_type ], 
    {} 
  );

  mediator.Notify( "Model_DailyQuest_EventScheduled" );

end;
--------------------------------------------------------------------------------
function public.Unschedule ( timer_type, skip_notification )

  parts.time.event.Finish();
  parts.time.event.Unshedule( private.tokens[ timer_type ] );

  local notification =
  {
    challenge = "Model_DailyQuest_ChallengeCooldownEnded",
    countdown = "Model_DailyQuest_CountdownEnded"
  };

  if ( notification[ timer_type ] ) and
     ( not skip_notification      )then

    mediator.Notify( notification[ timer_type ] );

  end;

end;
--------------------------------------------------------------------------------
function public.Finish ()

  if ( not public.IsReady() ) then

    DbgTrace( "[ ERROR ][ DAILYQUEST ] Repeat reward get" );

    return false;

  end;

  local  is_finished = private.Ready_Set( false );
  return is_finished;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Data_Get ( item )

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "dailyquest"
  };
  
  local dailyquest = lib_db.Execute( query ).result;

  if ( item ) then
  
    return dailyquest[ item ];
  
  end;
  
  return dailyquest;

end;
--------------------------------------------------------------------------------
function public.Data_Set ( data )

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "dailyquest",
    set        = data
  };
  
  if ( not lib_db.Execute( query ).result ) then

    query =
    {
      query      = "insert",
      group      = "player",
      collection = "dailyquest",
      value      = data
    };
    
    lib_db.Execute( query );

  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Challenge_SetState ( challenge_num, state )

  local challenge = "challenge"..challenge_num;

  local update_set = {
    [ challenge ] = public.Data_Get( challenge )
  };

  update_set[ challenge ].state = state;

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "dailyquest",
    set        = update_set
  };
  
  local  result = lib_db.Execute( query );
  return result;

end;
--------------------------------------------------------------------------------
function public.Challenge_IsSheduled ()

  return private.IsScheduled( "challenge" );

end;
--------------------------------------------------------------------------------
function private.Ready_Set ( is_ready )

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "dailyquest",
    set        = { ready = is_ready }
  };
  
  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function private.IsScheduled ( timer_type )

  return parts.time.event.Exists( private.tokens[ timer_type ] );

end;
--------------------------------------------------------------------------------