-- name=m_email
--------------------------------------------------------------------------------
local TAG = "m_email";
--------------------------------------------------------------------------------
-- Fields description
-- - messageId
-- - text
-- - attach
-- - expDate - array
-- - subject
-- - from
--     - name
--     - level
--     - avatar
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.unread_count    = 0;
  private.messages        = {};
  private.msgSection      = "message";
  private.is_data_expired = true;
  private.fake_email      = false;
  private.fast_upd_time   = 10;

  private.FakeEmail_Init();
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Start ( func_IsOnline, func_GetUserId )

  private.func_IsOnline  = func_IsOnline;
  private.func_GetUserId = func_GetUserId;
  
  private.userId = private.func_GetUserId();

  private.updateMessagesToken = "tmr_mail_update_messages";
  private.updateMessagesEvent = "Event_Mail_UpdateMessages";
  
  mediator.Subscribe(
    m_game.msg.ConnectLost,
    function ()

      private.update_time = private.fast_upd_time;

    end
  );
  
  mediator.Subscribe(
    m_game.msg.ConnectRestored,
    function ()

      private.userId      = private.func_GetUserId();
      private.update_time = nil;      

    end
  );

  private.Update_Start();

  -- JK: Broadcast fake emails immediately on start
  private.FakeEmail_Add();
  private.Messages_Broadcast();

end;
--------------------------------------------------------------------------------
function public.Suspend ()

  private.is_suspended = true;

end;
--------------------------------------------------------------------------------
function public.Resume ()

  private.is_suspended = nil;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Messages_RequestUpdate ()

  if ( private.update_time == private.fast_upd_time ) then

    -- JK: we're already polling fast
    return;

  end;

  private.update_time = private.fast_upd_time;
  private.Update_Restart();

end;
--------------------------------------------------------------------------------
function public.Messages_GetViewableCount ()

  local  viewable_count = private.Messages_GetViewableCount();
  return viewable_count;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Message_Send ( userId, senderId, text, attach )

  if ( userId == 1 ) or
     ( userId == 3 ) then
    private.FakeEmail_Send( userId );
    return;
  end;
  
  lib_rpc.Schedule(
    private.msgSection,
    "SendMessage",
    { 
      userId   = userId,
      senderId = senderId,
      text     = text,
      attach   = attach
    }, 
    function () end,
    true,
    true
  );

end;
--------------------------------------------------------------------------------
function public.Message_MarkAsRead ( messageId )

  local message = private.Message_GetById( messageId );

  if ( message ) then

    private.Request_ReadMessage( messageId );

  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.FakeEmail_HeartsAdd ()
  
  --Log.i( "FakeEmail_HeartsAdd" );

  local fake_emails = private.FakeEmail_GetData();
  
  fake_emails.heart_get  = "add";
  fake_emails.heart_send = "add";
  
  private.FakeEmail_Update( fake_emails );
  
  private.fake_email      = false;
  private.is_data_expired = true;
  
end;
--------------------------------------------------------------------------------
function public.FakeEmail_HeartIsHidden ()
  
  local fake_emails = private.FakeEmail_GetData();
  
  return ( fake_emails.heart_get == "hidden" );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Request_DataUpdate ( func_OnAnswer )

  --DbgTrace( "[ INFO ][ m_email ] Request_DataUpdate" );

  if ( not private.func_IsOnline() ) then return; end;

  if ( private.is_data_expired ) then

    private.Request_GetMessages( func_OnAnswer );
    private.is_data_expired = false;

  else

    private.Request_GetUnreadCount( func_OnAnswer );

  end;

end;
--------------------------------------------------------------------------------
function private.Request_GetUnreadCount ( func_OnAnswer )

  lib_rpc.Schedule(
    private.msgSection,
    "GetUnreadMessagesCount",
    { userId = private.userId },
    function ( msg, args )
      private.Messages_OnUnreadCountUpdated( msg, args );
      if ( func_OnAnswer ) then func_OnAnswer(); end;
    end,
    true,
    true
  );

end;
--------------------------------------------------------------------------------
function private.Request_GetMessages ( func_OnAnswer )

  local params = {
    userId         = private.userId,
    limit          = private.Messages_GetViewableCount(),
    startMessageId = 0
  };

  local silent = not func_OnAnswer;

  lib_rpc.Schedule(
    private.msgSection,
    "GetMessagesWithParams",
    params, 
    function ( msg, args )
      private.Messages_OnDataUpdated( msg, args );
      if ( func_OnAnswer ) then func_OnAnswer(); end;
    end, 
    silent,
    true
  );

end;
--------------------------------------------------------------------------------
function private.Request_ReadMessage ( messageId )
  
  -- Make fake email readed.
  if ( private.FakeEmail_Read( messageId ) ) then
    return;
  end;

  local params = {
    msgId          = messageId,
    userId         = private.userId,
    limit          = private.Messages_GetViewableCount(),
    startMessageId = 0
  };  

  lib_rpc.Schedule(
    private.msgSection,
    "ReadMessage",
    params,
    private.Message_OnRead,
    true,
    true
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Message_OnRead ( msg, args )

  local message = private.Message_GetById( args.msgId );

  if ( message ) then

    mediator.Notify(
      "Model_Email_AttachmentProcessingRequired",
      {
        message_id = args.msgId,
        sender_id  = message.from.userId,
        attachment = message.attach,
        text       = message.text
      }
    );

    local msgIdx     = private.Message_GetIndex( args.msgId );
    local removedMsg = table.remove( private.messages, msgIdx );

    private.Messages_UpdateUnreadCount();

  end;

end;
--------------------------------------------------------------------------------
function private.Message_GetById ( messageId )

  if ( #private.messages == 0 ) then return nil; end;

  local message_id = tonumber( messageId );

  for idx = 1, #private.messages do

    local current_id = tonumber( private.messages[ idx ].messageId );

    if ( current_id == message_id ) then

      return private.messages[ idx ];

    end;

  end;

  return nil;

end;
--------------------------------------------------------------------------------
function private.Message_GetIndex ( messageId )

  if ( #private.messages == 0 ) then return nil; end;

  local message_id = tonumber( messageId );

  for idx = 1, #private.messages do

    local current_id = tonumber( private.messages[ idx ].messageId );

    if ( current_id == message_id ) then

      return idx;

    end;

  end;

  return nil;

end;
--------------------------------------------------------------------------------
function private.Message_GetMaxId ()

  local maxMessageId = 0;

  if ( private.messages ) then

    for idx = 1, #private.messages do

      local current_id = tonumber( private.messages[ idx ].messageId );

      if ( current_id > maxMessageId ) then

        maxMessageId = current_id;

      end;

    end;

  end;

  return maxMessageId;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Messages_OnUnreadCountUpdated ( msg, args )

  --DbgTrace( "[ INFO ][ m_email ] Server unread count: "..args.count );

  if ( private.unread_count ~= args.count ) then

    private.is_data_expired = true;
    private.update_time     = private.fast_upd_time;

    return;

  end;

  private.update_time = nil;

end;
--------------------------------------------------------------------------------
function private.Messages_OnDataUpdated ( msg, args )

  --Log.i( "Messages_OnDataUpdated" );

  local messages = args.messages;

  --if ( #messages > 0 ) then

    private.messages = messages;
    private.Messages_UpdateUnreadCount();

    -- JK: fake emails are not included in unread count on m_email level
    -- they count on v_social level
    -- on m_email level only server unread count is used
    private.FakeEmail_Add();

    -- JK: broadcast the messages to v_social
    private.Messages_Broadcast();

  --end;

end;
--------------------------------------------------------------------------------
function private.Messages_Broadcast ()

  mediator.Notify(
    "Model_Email_MessageListUpdated",
    common.table.DeepCopy( private.messages )
  );

end;
--------------------------------------------------------------------------------
function private.Messages_UpdateUnreadCount ()
      
  private.unread_count = private.Messages_GetCurrentCount();

  --DbgTrace( "[ INFO ][ m_email ] Client unread count: "..private.unread_count );

  mediator.Notify( "Model_Email_UpdatedMessageCount" );

end;
--------------------------------------------------------------------------------
function private.Messages_GetCurrentCount ()

  return #private.messages;

end;
--------------------------------------------------------------------------------
function private.Messages_GetViewableCount ()

  return tonumber( lib_db.Settings( "email_settings" ).message_count );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Update_Start ()
  
  if ( not private.is_suspended ) then

    private.Request_DataUpdate();

  end;

  if ( not private.update_time ) then

    private.update_time = lib_db.Settings( "email_settings" ).update_time;

  end;

  ne.msg.Subscribe(
    private.updateMessagesEvent,
    private.Update_Restart
  );
  
  parts.time.event.Shedule(
    private.updateMessagesToken,
    "SESSION_GAME_TIME",
    private.update_time,
    private.updateMessagesEvent,
    {}
  );

end;
--------------------------------------------------------------------------------
function private.Update_Restart ()

  ne.msg.Unsubscribe(
    private.updateMessagesEvent,
    private.Update_Restart
  );

  TimeFinishEvent();
  TimeUnsheduleEvent( private.updateMessagesToken );

  private.Update_Start();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.FakeEmail_Init ()
  
  private.fake_emails = {
    gold = {
      text = { string = "", params = {} },
      attach = {
        name = "gold",
        type = "gold",
        count = 15
      },
      messageId = "0",
      name = "Administrator",
      userId = 1
    },
    heart_get = {
      text = { string = "str_int_dialog_social_mail_message_gift", params = {} },
      attach = {
        name = "heart",
        type = "heart",
        count = 1
      },
      messageId = "-1",
      name = "Administrator",
      userId = 3
    },
    heart_send = {
      text = { string = "str_int_dialog_social_mail_message_help", params = {} },
      attach = {},
      messageId = "-2",
      name = "Administrator",
      userId = 3
    }
  }
  
end;
--------------------------------------------------------------------------------
function private.FakeEmail_GetNameByID ( messageId )
  
  for key, value in pairs( private.fake_emails ) do
    if ( value.messageId == messageId ) then
      return key;
    end;
  end;
  
  return nil;
  
end;
--------------------------------------------------------------------------------
function private.FakeEmail_GetData ()
  
  local info = lib_db.Player.Info();
  local data = {
    gold = "add",
    heart_get = "hidden",
    heart_send = "hidden"
  };
  
  if ( info.fake_emails ) then
    data.gold       = info.fake_emails.gold;
    data.heart_get  = info.fake_emails.heart_get;
    data.heart_send = info.fake_emails.heart_send;
  else
    if ( info.isemailed ) then
      data.gold = "read";
    end;
  end;
  
  return data;
  
end;
--------------------------------------------------------------------------------
function private.FakeEmail_GetCount ()
  
  local fake_emails = private.FakeEmail_GetData();
  local count = 0;
  
  for key, value in pairs( fake_emails ) do
    if ( value == "add" ) then
      count = count + 1;
    end;
  end;
  
  return count;
  
end;
--------------------------------------------------------------------------------
function private.FakeEmail_Add ()
  
  local fake_emails = private.FakeEmail_GetData();
  
  for key, value in pairs( fake_emails ) do
    
    if ( value == "add" ) then
      
      local fake_email = {
        created = "",
        subject = "",
        from = {
          avatar = "",
          level = 0
        },
        expDate = {
          fraction = 0,
          hour = 0,
          error_count = 0,
          minute = 0,
          is_localtime = false,
          year = 0,
          month = 0,
          warnings = {},
          day = 0,
          warning_count = 0,
          errors = {},
          second = 0
        }
      };
      
      fake_email.text = private.fake_emails[ key ].text;
      fake_email.attach = private.fake_emails[ key ].attach;
      fake_email.messageId = private.fake_emails[ key ].messageId;
      fake_email.from.name = private.fake_emails[ key ].name;
      fake_email.from.userId = private.fake_emails[ key ].userId;

      table.insert( private.messages, fake_email );
      
    end;
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.FakeEmail_Read ( messageId )

  local fake_emails  = private.FakeEmail_GetData();
  local fake_message = private.FakeEmail_GetNameByID( messageId );
  
  if ( fake_message                         ) and
     ( fake_emails[ fake_message ] == "add" ) then

    fake_emails[ fake_message ] = "read";
    
    private.FakeEmail_Update( fake_emails );
    private.Message_OnRead( "", { msgId = messageId } );
    
    return true;

  end;
  
  return false;

end;
--------------------------------------------------------------------------------
function private.FakeEmail_Send ( userId )
  
  private.FakeEmail_Read( private.fake_emails.heart_send.messageId );
  
end;
--------------------------------------------------------------------------------
function private.FakeEmail_Update ( fake_emails )
  
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        =
    {
      fake_emails = fake_emails
    }
  };

  if not ( lib_db.Execute( query ).result ) then

    query =
    {
      query      = "insert",
      group      = "player",
      collection = "info",
      value      = 
      {
        fake_emails = fake_emails
      }
    };

    lib_db.Execute( query );

  end;
  
end;
--------------------------------------------------------------------------------