-- name=m_features
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.is_synced = false;
  private.features  = {};

  ne.msg.Subscribe( "Event_HeartBeat", private.RequestHash );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.IsSynced ()
  
  return private.is_synced;
  
end;
--------------------------------------------------------------------------------
function public.RequestList ( bForce, cheat_user_id )
  
  local data = {};
  
  if ( cheat_user_id      ) and
     ( cheat_user_id >= 0 ) then
     
     data.cheatUserId = cheat_user_id;
    
  end;
  
  lib_rpc.Schedule( 
    "feature",
    "GetFeatureList", 
    {}, 
    private.OnListReceived, 
    true,
    bForce
  );
  
end;
--------------------------------------------------------------------------------
function public.GetList ()

  local  features = common.table.DeepCopy( private.features );
  return features;

end;
--------------------------------------------------------------------------------
function public.GetHash ()

  return private.sHash;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Feature_GetState ( feature_name )

  local query = {
    query      = "find",
    group      = "sml",
    collection = "features",
    indice     = "name",
    criteria   = { like = feature_name }
  };

  local feature_data = lib_db.Execute( query ).result;

  if ( feature_data      ) and
     ( feature_data.name ) then

    return ( feature_data.active or false ), true;

  end;

  return false, false;

end;
--------------------------------------------------------------------------------
function public.Feature_SaveState ( feature_name, feature_state )

  local query;
  local state, exists = public.Feature_GetState( feature_name );

  if ( exists ) then

    if ( state == feature_state ) then

      return false;

    end;

    query = {
      query      = "update",
      group      = "sml",
      collection = "features",
      indice     = "name",
      criteria   = { like = feature_name },
      set        = { 
        active = feature_state
      }
    };

    return lib_db.Execute( query ).result;

  end;

  query = {
    query      = "insert",
    group      = "sml",
    collection = "features",
    value      = {
      name   = feature_name,
      active = feature_state
    }
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function public.Feature_GetData ( feature_name )

  local query = {
    query      = "find",
    group      = "sml",
    collection = "features",
    indice     = "name",
    criteria   = { like = feature_name }
  };

  local feature_data = lib_db.Execute( query ).result;

  if ( feature_data ) then

    return feature_data.data or false, true;

  end;

  return false, false;

end;
--------------------------------------------------------------------------------
function public.Feature_SaveData ( feature_name, feature_data )

  local query;
  local data, exists = public.Feature_GetData( feature_name );

  if ( exists ) then

    query = {
      query      = "update",
      group      = "sml",
      collection = "features",
      indice     = "name",
      criteria   = { like = feature_name },
      set        = { 
        data = feature_data
      }
    };

    return lib_db.Execute( query ).result;

  end;

  query = {
    query      = "insert",
    group      = "sml",
    collection = "features",
    value      = {
      name   = feature_name,
      data   = feature_data
    }
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function public.Feature_RemoveData ( feature_name )

  local query = {
    query      = "delete",
    group      = "sml",
    collection = "features",
    indice     = "name",
    criteria   = { like = feature_name }
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function public.Feature_GetGroup ( feature_name )

  local query = {
    query      = "find",
    group      = "sml",
    collection = "features",
    indice     = "name",
    criteria   = { like = feature_name }
  };

  local feature_data = lib_db.Execute( query ).result;

  if ( feature_data ) then

    return feature_data.group or false, true;

  end;

  return false, false;

end;
--------------------------------------------------------------------------------
function public.Feature_SetGroup ( feature_name, feature_group )

  local query;
  local data, exists = public.Feature_GetData( feature_name );

  if ( exists ) then

    query = {
      query      = "update",
      group      = "sml",
      collection = "features",
      indice     = "name",
      criteria   = { like = feature_name },
      set        = { 
        group = feature_group
      }
    };

    return lib_db.Execute( query ).result;

  end;

  query = {
    query      = "insert",
    group      = "sml",
    collection = "features",
    value      = {
      name   = feature_name,
      group  = feature_group
    }
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.RequestHash ()
 
  lib_rpc.Schedule( 
    "feature", 
    "GetFeatureHash", 
    {}, 
    private.OnHashReceived, 
    true
  );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.OnListReceived ( msg, data )
  
  if ( data                             ) and
     ( type( data.features ) == "table" ) then
       
    private.UpdateList( data.features );
    
    --Log.i( "hash: "..tostring( data.hash ) );
    
    if ( not private.sHash ) then
      private.UpdateHash( data.hash );
    end;
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.OnHashReceived ( msg, data )
  
  if ( data and data.hash ) then
    
    if ( private.UpdateHash( tostring( data.hash ) ) ) then

      public.RequestList();

    end;
    
  end;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.UpdateList ( tFeatures )
  
  private.features = {};
  
  for feature, data in pairs( tFeatures ) do
    
    private.features[ feature ] = wiz.json.Decode( data );
    
  end;

  --private.InjectMockFeatures( private.features );
  
  common.MessageSend( "Event_Feature_Updated", {} );
  
  -- JK: should be after message send
  private.is_synced = private.is_synced or true;
  
  Log.i( "synced" );
  
end;
--------------------------------------------------------------------------------
function private.UpdateHash ( sHash )
  
  if ( sHash and 
       sHash ~= "" and 
       sHash ~= private.sHash ) then
       
    Log.i( "hash - "..tostring( sHash ) );
    private.sHash = sHash;
    return true;
    
  end;

  return false;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.InjectMockFeatures ( features )

  --JK: to enable uncomment in private.UpdateList

  local mock_features = {
    feature_name = 'feature_json_here'
  };

  for name, data in pairs( features ) do
    if ( mock_features[ name ] ) then
      features[ name ] = wiz.json.Decode( mock_features[ name ] );
    end;
  end;

end;
--------------------------------------------------------------------------------