-- name=m_fights
--------------------------------------------------------------------------------

local _f = {};

--------------------------------------------------------------------------------
function public.__InitModule ()
  
  mediator.Subscribe(
    "Model_Sections_SectionEnabled",
    private.OnSectionEnabled
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Setup ( delegates )

  _f = delegates;

end;
--------------------------------------------------------------------------------
function public.GetDefaultGoldRewardCount ()

  return lib_db.Settings( "fights" ).reward;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Enemy_SetAbsent ( enemy_name, absent )

  local section_boss = _f.GetCurrentBoss();

  if ( section_boss ) then

    if ( absent == _f.GetFightData( section_boss.stage ).absent ) then

      return;

    end;

    local query =
    {
      query      = "update",
      group      = "player",
      collection = "stages",
      indice     = "name",
      criteria   = { like = section_boss.stage },
      set        =
      {
        absent  = absent
      }
    };

    lib_db.Execute( query );;

  end;

end;
--------------------------------------------------------------------------------
function public.Enemy_GetAbsent ( enemy_name )

  local section_boss = _f.GetCurrentBoss();

  if ( section_boss ) then

    return {
      name   = section_boss.stage,
      enemy  = section_boss.name,
      absent = _f.GetFightData( section_boss.stage ).absent
    };

  end;

  return false;

end;
--------------------------------------------------------------------------------
function public.Enemy_IsBoss ( enemy_name )

  local section_boss = _f.GetCurrentBoss();

  if ( section_boss ) and (
     ( section_boss.name == enemy_name ) or
     ( section_boss.sub  == enemy_name )
  ) then

    return section_boss.name;

  end;

  return false;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Fight_IsBoss ( stage_name )
  
  local section_boss = _f.GetCurrentBoss();

  if ( section_boss ) then

    return ( section_boss.stage == stage_name );

  end;

  return false;
  
end;
--------------------------------------------------------------------------------
function public.Fight_SetVisible ( fight_name )

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "stages",
    indice     = "name",
    criteria   = { like = fight_name },
    set        =
    {
      visible  = true
    }
  };

  if ( not lib_db.Execute( query ).result ) then

    query =
    {
      query      = "insert",
      group      = "player",
      collection = "stages",
      value      = 
      {
        name        = fight_name,
        visible     = true,
        enabled     = false,
        completed   = false,
        state       = 1
      }
    };
    
    lib_db.Execute( query );

  end;

  mediator.Notify( "Model_Fights_FightVisible", fight_name );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.OnSectionEnabled ( params )

  local section_name = params.section_name;
  local start_fights = params.section_config.fights.start_fights;

  for _, fight_name in pairs( start_fights ) do

    local visible = _f.GetFightData( fight_name ).visible;

    if not ( visible ) then

      local query =
      {
        query      = "update",
        group      = "player",
        collection = "stages",
        indice     = "name",
        criteria   = { like = fight_name },
        set        =
        {
          visible = true
        }
      };

      if ( not lib_db.Execute( query ).result ) then

        query =
        {
          query      = "insert",
          group      = "player",
          collection = "stages",
          value      = 
          {
            name      = fight_name,
            visible   = true,
            enabled   = false,
            completed = false,
            state     = 1
          }
        };
        
        lib_db.Execute( query );

      end;
    
    end;

  end;

end;
--------------------------------------------------------------------------------