-- name=m_fireflies
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.sticker_list  = {};
  private.sticker_count = 0;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Settings ()
  
  return lib_db.Settings( "album" );
  
end;
--------------------------------------------------------------------------------
function public.GetBlockerData ()
  
  local settings = private.Settings();
  
  return settings.unlock;
  
end;
--------------------------------------------------------------------------------
function public.GetAlbumReward ()
  
  local reward = lib_db.Settings( "album" ).reward;
  
  return reward;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Firefly_Data ( firefly_name )
  
  return private.Firefly_Data( firefly_name );
  
end;
--------------------------------------------------------------------------------
function public.Firefly_GetAll ()
  
  return private.Firefly_GetAll();
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Stickerpack_Count ()

  return private.Stickerpack_Count();

end;
--------------------------------------------------------------------------------
function public.Stickerpack_Price ()

  return lib_db.Settings( "stickerpack" ).price;

end;
--------------------------------------------------------------------------------
function public.Stickerpack_Obtain ( stickerpack_count, stickerpack_type )
  
  local current_counts = private.Stickerpack_Count();
  local new_count = current_counts[ stickerpack_type ] + stickerpack_count;

  local query  =
  {
    query      = "update",
    group      = "player",
    collection = "fireflies_stickerpacks",
    set        = { [ stickerpack_type ] = new_count }
  };
  
  if ( lib_db.Execute( query ).result ) then
    
    mediator.Notify( "Model_Fireflies_BoosterCountChanged", { 
      change = stickerpack_count,
      count  = current_counts.total + stickerpack_count;
    } );
    
    return true;
    
  end;
  
  return false;

end;
--------------------------------------------------------------------------------
function public.Stickerpack_TryOpen ()
  
  local current_counts = private.Stickerpack_Count();

  if ( current_counts.total < 1 ) then

    return false;

  end;
  
  local bought = ( current_counts.free == 0 );
  
  local stickerpack_type = "free";

  if ( bought ) then

    stickerpack_type = "bought";

  end;
  
  local stickers = private.Stickerpack_Generate( bought );
  
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "fireflies_stickerpacks",
    set        = { [ stickerpack_type ] = current_counts[ stickerpack_type ] - 1 }
  };
  
  if ( lib_db.Execute( query ).result ) then

    mediator.Notify( "Model_Fireflies_BoosterCountChanged", { 
      change = -1,
      count  = current_counts.total - 1
    } );

    return stickers;

  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Cheat_GetAllStickers ()
  
  for _, sticker in pairs( private.sticker_list ) do

    private.Sticker_TryChangeCount( sticker.name, 1 );
    
  end;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Sticker_InitList ()

  private.Sticker_InitList();

end;
--------------------------------------------------------------------------------
function public.Sticker_Count ()
  
  return private.sticker_count;
  
end;
--------------------------------------------------------------------------------
function public.Sticker_Duplicates ()
  
  local dupe_list = {};
  local stickers  = private.Sticker_QueryStorage();
  
  for s, sticker in ipairs( stickers ) do    

    local sticker_data = private.Sticker_Data( sticker.name );
    local dupe_count   = private.Sticker_DuplicateCount( sticker_data );

    if ( dupe_count > 0 ) then

      local dupe_data = sticker_data;
      dupe_data.count = dupe_count;

      table.insert( dupe_list, dupe_data );

    end;

  end;

  return dupe_list;
  
end;
--------------------------------------------------------------------------------
function public.Sticker_TrySell ( sticker_name, sticker_count )

  local gold_count = false;
  
  if ( private.Sticker_TryChangeCount( sticker_name, -sticker_count ) ) then

    gold_count = private.sticker_list[ sticker_name ].price * sticker_count;
    
  end;
  
  return gold_count;
  
end;
--------------------------------------------------------------------------------
function public.Sticker_TryStick ( sticker_name )
  
  local firefly_name = private.Firefly_GetNameBySticker( sticker_name );
  
  if ( private.Firefly_AddSticker( firefly_name, sticker_name ) ) then
    
    local sticker_info = private.Sticker_QueryStorage( sticker_name );

    local query =
    {
      query      = "update",
      group      = "player",
      collection = "fireflies_stickers",
      indice     = "name",
      criteria   = { like  = sticker_name },
      set        = { count = sticker_info.count - 1 }
    };
    
    local result = lib_db.Execute( query ).result;
    
    if ( result ) then

      v_log.Kontagent_LogEvent( "economy", 
      {
        description = "item_used",
        category    = "stickers",
        value       = -1,
        name        = sticker_name
      } );
    
      private.Achievement_TryComplete( sticker_name, firefly_name );
      private.Achievement_TryCompleteAll();
      
    end;
    
    return result;

  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Firefly_Create ( firefly_name )

  local query =
  {
    query      = "insert",
    group      = "player",
    collection = "fireflies",
    value      = {
      name  = firefly_name,
      level = 0,
      exp   = 0
    }
  };
  
  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function private.Firefly_AddSticker ( firefly_name, sticker_name )
  
  local firefly_info = private.Firefly_QueryStorage( firefly_name );
  local firefly_data = private.Firefly_QueryStatic( firefly_name );

  if ( not firefly_info ) then

    private.Firefly_Create( firefly_name );
    firefly_info = {};
    
  end;
  
  if ( common.IsEntityInList( firefly_info.collected_stickers, sticker_name ) ) then

    return false;

  end;
  
  local update_set =
  {
    collected_stickers = firefly_info.collected_stickers or {}
  };

  table.insert( update_set.collected_stickers, sticker_name );
  
  private.Firefly_UpdateStorage( firefly_name, update_set );
  
  return true;
  
end;
--------------------------------------------------------------------------------
function private.Firefly_QueryStorage ( firefly_name )

  return private.GetInfo( "player", "fireflies", "name", firefly_name );

end;
--------------------------------------------------------------------------------
function private.Firefly_QueryStatic ( firefly_name )

  return private.GetInfo( "static", "stickers", "name", firefly_name );

end;
--------------------------------------------------------------------------------
function private.Firefly_GetNameBySticker ( sticker_name )

  return private.sticker_list[ sticker_name ].firefly;

end;
--------------------------------------------------------------------------------
function private.Firefly_UpdateStorage ( firefly_name, update_set )

  if ( firefly_name and update_set ) then
  
    local query =
    {
      query      = "update",
      group      = "player",
      collection = "fireflies",
      indice     = "name",
      criteria   = { like = firefly_name },
      set        = update_set
    };
    
    return lib_db.Execute( query ).result;
    
  end;

end;
--------------------------------------------------------------------------------
function private.Firefly_ListRewards ( firefly_name )

  local rewards = {};
  local firefly = private.Firefly_QueryStatic( firefly_name );
  
  for c, content in pairs( firefly.content ) do

    local reward          = content.reward;
          reward.complete = true;

    for p, part in pairs( content.parts ) do

      if ( not private.Sticker_IsCollected( firefly.name, part ) ) then

        reward.complete = false;

        break;

      end;

    end;

    table.insert( rewards, reward );

  end;

  return rewards;

end;
--------------------------------------------------------------------------------
function private.Firefly_GetAll ()
  
  local firefly_list = private.Firefly_QueryStatic();
  local firefly_data = {};
  
  for _, firefly in ipairs( firefly_list ) do
    
    table.insert( firefly_data, private.Firefly_Data( firefly.name ) );
    
  end;
  
  return firefly_data;
  
end;
--------------------------------------------------------------------------------
function private.Firefly_Data ( firefly_name )
  
  local firefly_data = private.Firefly_QueryStatic( firefly_name );
  local sticker_list = private.Firefly_ListStickers( firefly_data );

  local stickers = {};

  for s, sticker in ipairs( sticker_list ) do
    
    table.insert( stickers, private.Sticker_Data( sticker.name ) );
    
  end;
    
  return {
    name     = firefly_name,
    stickers = stickers,
    rewards  = private.Firefly_ListRewards( firefly_name )
  };
  
end;
--------------------------------------------------------------------------------
function private.Firefly_ListStickers ( firefly_data )

  local sticker_list = {};

  for c, content in ipairs( firefly_data.content ) do

    for p, part in pairs( content.parts ) do

      local sticker = {
        name    = part,
        content = c
      };

      table.insert( sticker_list, sticker );

    end;    

  end;

  return sticker_list;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Sticker_InitList ()

  local fireflies_data = private.Firefly_QueryStatic();
  
  local associated = {};
  local index      = 0;

  local function func_add (
      firefly_name,
      sticker_name,
      sticker_price
    )

    index = index + 1;

    local sticker_data = {
      index   = index,
      name    = sticker_name,
      price   = sticker_price,
      firefly = firefly_name
    };

    associated[ sticker_name ] = sticker_data;

  end;
  
  for f, firefly in ipairs( fireflies_data ) do

    for c, content in ipairs( firefly.content ) do

      for p, part in pairs( content.parts ) do

        func_add( firefly.name, part, content.price[ p ] );

      end;    

    end;
    
  end;
  
  private.sticker_list  = associated;
  private.sticker_count = index;
  
end
--------------------------------------------------------------------------------
function private.Sticker_GetAll ()

  local sticker_data = {};
    
  for _, sticker in pairs( private.sticker_list ) do

    table.insert( sticker_data, private.Sticker_Data( sticker.name ) );

  end;

  return sticker_data;
  
end;
--------------------------------------------------------------------------------
function private.Sticker_Data ( sticker_name )

  local sticker = common.table.DeepCopy(
    private.sticker_list[ sticker_name ]
  );

  local sticker_data = private.Sticker_QueryStorage( sticker.name ) or {};

  sticker.collected = private.Sticker_IsCollected(
    sticker.firefly, sticker.name
  );

  sticker.count     = sticker_data.count or 0;
  sticker.stickable = private.Sticker_IsStickable( sticker );
  sticker.dupes     = private.Sticker_DuplicateCount( sticker );

  return sticker;

end;
--------------------------------------------------------------------------------
function private.Sticker_IsStickable ( sticker_data )

  return ( ( not sticker_data.collected ) and ( sticker_data.count > 0 ) );
  
end;
--------------------------------------------------------------------------------
function private.Sticker_DuplicateCount ( sticker_data )

  if ( sticker_data.collected ) then

    return sticker_data.count;

  else

    return math.max( ( sticker_data.count - 1 ), 0 );

  end;

end;
--------------------------------------------------------------------------------
function private.Sticker_IsCollected ( firefly_name, sticker_name )

  local firefly = private.Firefly_QueryStorage( firefly_name );

  if ( firefly                    ) and
     ( firefly.collected_stickers ) then

    return common.IsEntityInList(
      firefly.collected_stickers,
      sticker_name,
      nil,
      true
    );

  end;

  return false;

end;
--------------------------------------------------------------------------------
function private.Sticker_TryChangeCount ( sticker_name, delta )

  if ( delta == 0 ) then

    return false;

  end;

  local current_count = 0;
  local sticker_info  = private.Sticker_QueryStorage( sticker_name );

  if ( sticker_info ) then

    current_count = sticker_info.count;

  end;

  local new_count = current_count + delta;

  if ( new_count < 0 ) then
    
    return false, new_count;
    
  end;
  
  if ( delta > 0 ) then

    v_log.Kontagent_LogEvent( "economy", 
    {
      description = "item_collected",
      category    = "stickers",
      value       = delta,
      name        = sticker_name
    } );
    
  end;

  if ( sticker_info ) then

    local query =
    {
      query      = "update",
      group      = "player",
      collection = "fireflies_stickers",
      indice     = "name",
      criteria   = { like = sticker_name },
      set        = {
        count = new_count
      }
    };
    
    return lib_db.Execute( query ).result, new_count;

  else
    
    local query =
    {
      query      = "insert",
      group      = "player",
      collection = "fireflies_stickers",
      value      = {
        name  = sticker_name,
        count = new_count
      }
    };
    
    return lib_db.Execute( query ).result, new_count;
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.Sticker_QueryStorage ( sticker_name )
  
  if ( not sticker_name ) then
    
    return private.GetInfo( "player", "fireflies_stickers" );
    
  end;

  return private.GetInfo(
    "player",
    "fireflies_stickers",
    "name",
    sticker_name
  );

end;
--------------------------------------------------------------------------------
function private.Sticker_QueryStatic ( sticker_name )
  
  if ( not sticker_name ) then
    
    return private.GetInfo( "static", "stickers" );
    
  end;

  return private.GetInfo(
    "static",
    "fireflies_stickers",
    "name",
    sticker_name
  );

end;
--------------------------------------------------------------------------------
function private.Sticker_Generate ( ban_list )

  local sticker_names = {};
  local firefly_list  = private.Firefly_QueryStatic();

  for f, firefly in pairs( firefly_list ) do

    local stickers = private.Firefly_ListStickers( firefly );

    for s, sticker in ipairs( stickers ) do

      if ( not common.IsEntityInList( ban_list, sticker.name ) ) then

        table.insert( sticker_names, sticker.name );

      end;

    end;

  end;

  local random = math.random( 1, #sticker_names );
  
  return sticker_names[ random ];
  
end;
--------------------------------------------------------------------------------
function private.Sticker_Index ( sticker_name )

  return private.sticker_list[ sticker_name ].index;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Stickerpack_Count ()
  
  local query =
  {
    query      = "find",
    group      = "player",
    collection = "fireflies_stickerpacks"
  };
  
  local result = lib_db.Execute( query ).result or { 
    free   = 0,
    bought = 0
  };
  
  result.total = result.free + result.bought;

  return result;
  
end;
--------------------------------------------------------------------------------
function private.Stickerpack_Generate ( bought )

  local stickerpack_settings = lib_db.Settings( "stickerpack" );
  local stickers_info        = private.Sticker_QueryStorage();

  local sticker_names = {};
  local stickerpack   = {};
  
  if ( stickers_info      ) and
     ( #stickers_info > 0 ) then

    for s = 1, stickerpack_settings.stickers_count do

      local ban_list;
      
      if ( not bought ) then
        
        ban_list = stickerpack_settings.bought_only;
        
      end;
      
      table.insert( sticker_names, private.Sticker_Generate( ban_list ) );

    end;

  else

    sticker_names = stickerpack_settings.first_stickerpack;

  end;

  for s = 1, #sticker_names do

    local sticker_name = sticker_names[ s ];

    private.Sticker_TryChangeCount( sticker_name, 1 );

    local sticker_data = private.Sticker_Data( sticker_name );

    table.insert( stickerpack, sticker_data );

  end;

  return stickerpack;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Achievement_TryComplete ( sticker_name, firefly_name )
  
  local parts, reward;

  local firefly_data = private.Firefly_QueryStatic( firefly_name );

  for c, content in pairs( firefly_data.content ) do

    if ( common.IsEntityInList( content.parts, sticker_name ) ) then

      parts         = content.parts;
      reward        = content.reward;
      reward.index  = c;
      reward.layout = content.layout;

      break;

    end;   

  end;

  if ( not ( parts or reward ) ) then

    return false;

  end;

  local collected =
    private.Firefly_QueryStorage( firefly_name ).collected_stickers or {};

  for p, part in pairs( parts ) do

    if ( not common.IsEntityInList( collected, part ) ) then

      return false;

    end;    

  end;

  mediator.Notify(
    "Model_Fireflies_AchievementComplete",
    reward
  );

end;
--------------------------------------------------------------------------------
function private.Achievement_TryCompleteAll ()

  local stats = public.Stats();

  if ( stats.stickers.total == stats.stickers.collected ) then

    local reward = lib_db.Settings( "album" ).reward;

    mediator.Notify(
      "Model_Fireflies_AchievementCompleteAll",
      reward
    );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Stats ()

  local stickers  = private.Sticker_GetAll();
  local collected = 0;
  
  for s, sticker in ipairs( stickers ) do
    
    if ( sticker.collected ) then
      
      collected = collected + 1;
      
    end;
    
  end;
  
  local fireflies = private.Firefly_QueryStatic();  

  return {
    stickers  = {
      total     = #stickers,
      collected = collected
    },
    fireflies =
    {
      total = #fireflies,
    }
  };
  
end;
--------------------------------------------------------------------------------
function private.GetInfo ( group, collection, indice, like )

  local query;

  if ( like ) then

    query =
    {
      query      = "find",
      group      = group,
      collection = collection,
      indice     = indice,
      criteria   = { like = like }
    };

  else

    query =
    {
      query      = "find_all",
      group      = group,
      collection = collection
    };

  end;

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------