-- name=m_fortunewheel
--------------------------------------------------------------------------------
function public.__InitModule ()
  
  private.recover = {
    token = "tmr_fortunewheel_recover",
    event = "Event_Fortunewheel_RecoverDone"
  };

  ne.msg.Subscribe( private.recover.event, public.Cooldown_Stop );

  if ( public.Cooldown_IsActive ()             ) and
     ( public.Cooldown_GetRemainingTime() == 0 ) then

    public.Cooldown_Stop();

  end;

  private.is_fortune_wheel_disabled = false;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Setup ()
  
  local fresh_player = lib_db.Player.Info().fortune_wheel_fresh_player_feature;
  local fortune_wheel_fresh_player_server = m_features.Feature_GetState( "fortune_wheel_fresh_player" );
  
  if ( fresh_player ) and ( not fortune_wheel_fresh_player_server ) then
    private.is_fortune_wheel_disabled = true;
  end;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.GetSettings ()
  
  local  settings = lib_db.Settings( "fortunewheel" );
  return settings;
  
end;
--------------------------------------------------------------------------------
function private.Storage_GetData ()
  
  local query =
  {
    query      = "find",
    group      = "player",
    collection = "fortunewheel"
  };

  return lib_db.Execute( query ).result;
  
end;
--------------------------------------------------------------------------------
function private.Storage_UpdateData ( set )
  
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "fortunewheel",
    set        = set
  };

  return lib_db.Execute( query ).result;
  
end;
--------------------------------------------------------------------------------
function public.TurnWheel ( hearts_now, hearts_max )
  
  if ( public.Cooldown_IsActive() ) then
  
    return false;
  
  end;
  
  local result;
  local info = public.GetState();

  if ( info.count_try == 0 ) then

    result = 1; -- Position of ExtraMoves

  else
    
    local smart_chance = false;
    local settings = private.GetSettings();
    
    local prizes = common.table.DeepCopy( settings.prizes );

    if ( hearts_now >= ( hearts_max - 1 ) ) and ( #prizes > 1 ) then

      -- Set heart chance to zero
      local heart_chance = 0;
      for i = 1, #prizes do
        if ( prizes[ i ].name == "heart" ) then

          heart_chance = prizes[ i ].chance;
          prizes[ i ].chance = 0;
          break;

        end;
      end;

      -- Normalize another chances
      local chance_addiction = heart_chance / ( #prizes - 1 );
      for i = 1, #prizes  do
        if ( prizes[ i ].name ~= "heart" ) then

          prizes[ i ].chance = prizes[ i ].chance + chance_addiction;

        end;
      end;
      
    elseif ( hearts_now == 1 ) then

      smart_chance = true;
      
    end;

    result = private.GetRandomPrize( prizes, smart_chance );

  end;

  public.SetState( { count_try = info.count_try + 1 } );
  public.Cooldown_Start();
  
  return result;

end;
--------------------------------------------------------------------------------
function public.GetInfo ()

  local settings = private.GetSettings();
  
  local info = {
    prizes    = settings.prizes,
    spin_cost = settings.spin_price or 0
  };

  return info;

end;
--------------------------------------------------------------------------------
function public.SetState ( set )
  
  return private.Storage_UpdateData( set );

end;
--------------------------------------------------------------------------------
function public.GetState ()
  
  local  data = private.Storage_GetData();
  return data;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Cooldown_Start ()

  if ( public.Cooldown_IsActive() ) then

    return false;

  end;

  local settings = private.GetSettings();
  
  parts.time.event.Shedule(
    private.recover.token,
    "GLOBAL_TIME",
    settings.time_recover,
    private.recover.event,
    {}
  );

  mediator.Notify( "Model_Fortunewheel_TimerStart" );

  return true;

end;
--------------------------------------------------------------------------------
function public.Cooldown_Stop ()

  parts.time.event.Finish();
  parts.time.event.Unshedule( private.recover.token );

  mediator.Notify( "Model_Fortunewheel_TimerStop" );

end;
--------------------------------------------------------------------------------
function public.Cooldown_IsActive ()

  local  is_active = parts.time.event.Exists( private.recover.token );
  return is_active;

end;
--------------------------------------------------------------------------------
function public.Cooldown_GetRemainingTime ()

  local  remaining_time = parts.time.event.GetRemainingTime( private.recover.token );
  return remaining_time;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.GetRandomPrize ( prize_list, smart_chance )

  local random = math.random();
  local chance_counter = 0.0;
  local mode = tern( smart_chance )( "smart_chance", "chance" );

  for i = 1, #prize_list do

    chance_counter = chance_counter + prize_list[ i ][ mode ];

    if ( random <= chance_counter ) then return i; end;

  end;

end;
--------------------------------------------------------------------------------
function public.IsActive ()
  
  local state = not private.is_fortune_wheel_disabled;
  return state;
  
end;
--------------------------------------------------------------------------------