-- name=m_game
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.session_time_launch     = 0;
  private.session_time_seconds    = 0;
  private.session_time_token      = "tmr_session_tick";
  private.session_time_resolution = 60;
  private.session_time_event      = "Event_Session_Tick"

  public.msg =
  {
    WaitingStart       = "Model_Game_WaitingStart",
    WaitingFinish      = "Model_Game_WaitingFinish",
    ConnectLost        = "Model_Game_ConnectLost",
    ConnectRestored    = "Model_Game_ConnectRestored",
    SessionConflict    = "Model_Game_SessionConflict",
    ContentObsolete    = "Model_Game_ContentObsolete",
    StorageConflict    = "Model_Game_StorageConflict",
    VersionOutdated    = "Model_Game_VersionOutdated",
    LostFocus          = "Model_Game_LostFocus",
    GainFocus          = "Model_Game_GainFocus",
    SilentContentReady  = "Model_Game_SilentContentReady",
    SilentContentFailed = "Model_Game_SilentContentFailed",
    Maintenance        = "Model_Game_Maintenance"
  };

  private.connect_retries = 3;
  private.reconnect_retry = 0;
  
  private.error =
  {
    no_server = "ServerTemporarilyUnavailable",
    kicked    = "KickedByOthers"
  }

  if ( IsAndroid() and ( wiz.Android.GetPlatform() == "google" ) ) then
  
    ne.msg.Subscribe(
      "Event_GooglePlay_LoggedStateChanged",
      private.OnGooglepLayLoggedStateChanged
    );
  
  end;
  
  private.css_available = false;
  
  private.game_log = {};
  private.game_log_pos = 1;
  private.game_log_size = 20;
  ne.msg.Subscribe(
    "Event_Application_Closed",
    private.GameLog_Send
  );
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.OnGooglepLayLoggedStateChanged ( msg, params )
  
  if ( params.logged ) then
  
    SetGooglePlayLogged( true );
  
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.SilentContent_IsReady ()

  --if ( private.Content_IsManifestEmpty() ) then
  --
  --  return true;
  --  
  --end;

  local  is_ready = lib_content.Silent_IsReady();
  return is_ready;

end;
--------------------------------------------------------------------------------
function public.SilentContent_HasFreeSpace ()
  
  return lib_content.Silent_HasFreeSpace();
  
end;
--------------------------------------------------------------------------------
function public.SilentContent_Download ()

  if ( not lib_content.Silent_IsDownloading() ) then
  
    lib_content.Silent_StartDownloading();
  
    mediator.SubscribeOnce(
      lib_content.msg.DownloadFinished,
      private.SilentContent_OnDownloadFinished
    );
    
  else
  
    Log.w( "Silent content download not started" );
  
  end;
  
end;
--------------------------------------------------------------------------------
function public.SilentContent_ApplyPackets ()

  lib_content.Silent_ApplyPackets();
  
end;
--------------------------------------------------------------------------------
function private.SilentContent_OnDownloadFinished ( result )

  if ( result ) then
  
    mediator.Notify( public.msg.SilentContentReady );
  
  else
    
    mediator.Notify( public.msg.SilentContentFailed );
    --public.SilentContent_Download();
  
  end;

end;
--------------------------------------------------------------------------------
--function public.Content_IsManifestEmpty ()
--
--  return private.Content_IsManifestEmpty();
--   
--end;
--------------------------------------------------------------------------------
--function private.Content_IsManifestEmpty ()
--
--  local packet_list = parts.content.GetPacketList() or {};
--
--  return ( ( #packet_list == 1 ) and ( packet_list[ 1 ].hash == '0' ) ) or
--         (   #packet_list == 0                                        );
--   
--end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.SetupClient ( online_state )

  private.is_online = online_state;
  
  if ( not private.is_online ) then
    private.offline_reason = "connection_error";
  end;
  
  local lib_rpc_callbacks =
  {
    WaitingShow = function () mediator.Notify( public.msg.WaitingStart  ); end,
    WaitingHide = function () mediator.Notify( public.msg.WaitingFinish ); end
  };
  
  lib_rpc.Setup( lib_rpc_callbacks );
  
  local events =
  {
    HeartBeat       = "Event_HeartBeat",
    InternetError   = "Event_Application_InternetError",
    ContentObsolete = "Event_Content_Obsolete",
    StorageConflict = "Event_Storage_Conflict",
    LostFocus       = "Event_Application_LostFocus",
    GainFocus       = "Event_Application_GainFocus",
    VersionOutdated = "Event_Application_ClientVersionOutdated",
    Maintenance     = "Event_Auth_Maintenance",
    NetResponse     = "Event_Net_Response"
  };
  
  -- Connection lost / restored
  ne.msg.Subscribe( events.InternetError, private.OnConnectLost     );
  ne.msg.Subscribe( events.HeartBeat,     private.OnConnectRestored );
  
  -- Resend messages with changed scope
  ne.msg.Subscribe( events.ContentObsolete,
    function ()
       mediator.Notify( public.msg.ContentObsolete );
    end
  );
  ne.msg.Subscribe( events.StorageConflict,
    function ( message, params )
       mediator.Notify( public.msg.StorageConflict, params );
    end
  );

  ne.msg.Subscribe( events.LostFocus, private.OnFocusLost );
  ne.msg.Subscribe( events.GainFocus, private.OnFocusGained );

  ne.msg.Subscribe( events.VersionOutdated,
    function ()
       mediator.Notify( public.msg.VersionOutdated );
    end
  );
  
  ne.msg.Subscribe( events.Maintenance,
    function ( message, params )
       mediator.Notify( public.msg.Maintenance, params );
    end
  );

  ne.msg.Subscribe( events.NetResponse, private.OnNetResponse );

end;
--------------------------------------------------------------------------------
function private.OnConnectLost ( message, params )

  if ( params ) then
  
    if ( params.error == private.error.no_server ) then
    
      if ( private.is_online                      ) or 
         ( private.offline_reason == "focus_lost" ) then
      
        private.BecomeOffline();
      
      end;
      return;
    
    end;
    
    if ( params.error == private.error.kicked ) then
    
      mediator.Notify( public.msg.SessionConflict );
      return;
    
    end;
  
  end;
  
  if ( private.is_online                      ) or
     ( private.offline_reason == "focus_lost" ) then
  
    if ( private.reconnect_retry < private.connect_retries ) then
    
      private.reconnect_retry = private.reconnect_retry + 1;
      
      lib_rpc.RescheduleAll();
      parts.net.heartbeat.Restart();
    
    else
    
      private.reconnect_retry = 0;
      private.BecomeOffline();
    
    end;
  
  end;

end;
--------------------------------------------------------------------------------
function private.BecomeOffline ()

  private.is_online      = false;
  private.offline_reason = "connection_error";
      
  local is_silent = lib_rpc.IsSilent();
  lib_rpc.ClearScheduled();
  
  mediator.Notify( public.msg.ConnectLost, not is_silent );

end;
--------------------------------------------------------------------------------
function private.OnConnectRestored ( message, params )

  if ( not private.is_online ) then
  
    private.is_online = true;

    if ( private.offline_reason ~= "focus_lost" ) then

      mediator.Notify( public.msg.ConnectRestored );

    end;

    private.offline_reason = nil;
  
  end;

end;
--------------------------------------------------------------------------------
function private.OnFocusLost ()

  private.is_online = false;

  if ( not private.offline_reason ) then
    private.offline_reason = "focus_lost";
  end;

  mediator.Notify( public.msg.LostFocus );

end;
--------------------------------------------------------------------------------
function private.OnFocusGained ()

  mediator.Notify( public.msg.GainFocus );

end;
--------------------------------------------------------------------------------
function private.OnNetResponse ( msg, params )

  local status_code = params.statusCode;

  if ( tonumber( status_code ) == 200 ) then

    if ( private.offline_reason == "focus_lost" ) then
      private.is_online      = true;
      private.offline_reason = nil;
    end;

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.IsOnline ()

  return private.is_online;

end;
--------------------------------------------------------------------------------
function public.Settings ( key )

  local  settings = lib_db.Settings( key );
  return settings;

end;
--------------------------------------------------------------------------------
function public.GetSettingsVariant ()

  local  settings_variant = lib_db.GetSettingsVariant();
  return settings_variant;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.CheckForNewVersion ( func_handle )

  local platform = GetPlatform();
  if ( platform == "android" ) then
      
    platform = wiz.Android.GetPlatform();
      
  end;
  
  local silent = true;

  lib_rpc.Schedule(
    "version",
    "IsNewClientVersionAvailable",
    {
      platform = platform
    },
    func_handle,
    silent,
    true
  );

end;
--------------------------------------------------------------------------------
function public.Update_GetData ()
  
  local update_data = false;
  local player_data = lib_db.Player.Info();
  
  --Uncomment for testing reward
  --player_data.update      = {};
  --player_data.update.type = "reward";
  --player_data.update.reward =
  --{
  --  { type = "booster", name = "hammer",    count = 1   },
  --  { type = "gold",    name = "gold",      count = 150 },
  --  { type = "booster", name = "magicgift", count = 2   }
  --};
  
  --Uncomment for testing replay
  --player_data.update      = {};
  --player_data.update.type = "replay";

  if ( player_data.update ) then

    local version = VersionGetComponentVersion( "game" );

    local message = "";
    local count   = 0;
    
    for key in string.gmatch( version, "%d+" ) do

      if ( count < 3 ) then

        message = message..key;

      end;

      count = count + 1;

    end;

    update_data = {
      type    = player_data.update.type,
      message = message,
      reward  = player_data.update.reward
    };

    local query =
    {
      query      = "update",
      group      = "player",
      collection = "info",
      set        =
      {
        update = false
      }
    };

    lib_db.Execute( query );

  end;

  return update_data;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Section_GetAllowedCount ()

  local result, count = parts.storage.data.Read( "sml/allow_sections" );
  
  if ( result ~= 0 ) then
  
    return nil;
    
  end;
  
  return tonumber( count );

end;
--------------------------------------------------------------------------------
function public.Section_SetAllowedCount ( count )

  local _, value = parts.storage.data.Read( "sml/allow_sections" );

  if ( value and value ~= "" and value ~= "false" and value ~= "true" ) then
    
    if ( tonumber( value ) < count ) then
      
      parts.storage.data.Write( "sml/allow_sections", count );
      
    end;

  end;

end;
--------------------------------------------------------------------------------
function public.GetSkinsContent ()

  local query =
  {
    query      = "find_all",
    group      = "static",
    collection = "skins"
  };

  local skins   = lib_db.Execute( query ).result;
  local content = {};

  for _, skin in pairs( skins ) do

    content[ skin.name ] = skin.content.."/"..skin.name;

  end;

  return content;

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function public.Context_GetUnlockStage ()

  return lib_db.Settings( "context_message" ).unlock;

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function public.IsCCSAvailable ()

  return private.css_available;

end;
--------------------------------------------------------------------------------
function public.RequestIsCCSAvailable ()

  local platform = GetPlatform();
  local silent = true;

  lib_rpc.Schedule( 
    "setting",
    "IsCcsAvailable",
    { platform = platform },
    private.OnIsCcsAvailableReply,
    silent,
    true
  );

end;
--------------------------------------------------------------------------------
function private.OnIsCcsAvailableReply ( msg, params )

  private.css_available = ( params or {} ).available;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.StorageConflict_GetClientPlayerInfo ()

  local key = "parts/cloudstorage/player/info";
  local _, json = parts.storage.data.Read( key );
  local player_info = wiz.json.Decode( json );
  
  return player_info or { lib_db.Player.Info() } or {};

end;
--------------------------------------------------------------------------------
function public.StorageConflict_GetClientSections ()

  local key = "parts/cloudstorage/player/sections";
  local _, json = parts.storage.data.Read( key );
  local sections = wiz.json.Decode( json );
  
  return sections;

end;
--------------------------------------------------------------------------------
function public.StorageConflict_GetServerPlayerInfo ( saveData )

  local key = "player/info";

  if not ( saveData and saveData[ key ] ) then
    return nil;
  end;
  
  return wiz.json.Decode( saveData[ key ] );

end;
--------------------------------------------------------------------------------
function public.StorageConflict_GetServerSections ( saveData )

  local key = "player/sections";

  if not ( saveData and saveData[ key ] ) then
    return nil;
  end;
  
  return wiz.json.Decode( saveData[ key ] );

end;
--------------------------------------------------------------------------------
function public.StorageConflict_GetStageNumber ( stage_name )
  
  local stage_number = tonumber( stage_name:match( "_(%d+)" ) );
  
  if ( stage_name:match( "_bonus" ) ) then
    return "bonus";
  end;
  
  if ( stage_number > 110 ) then
    local section_number, residue = math.modf( stage_number / 110 );
    if ( residue == 0 ) then
      section_number = section_number - 1;
    end;
    stage_number = stage_number - 110 * section_number;
  end;
  
  return stage_number;
  
end;
--------------------------------------------------------------------------------
function public.StorageConflict_GetLastSection ( sections )
  
  local section = "section_1";
  
  for s = #sections, 1, -1  do
    if ( sections[ s ].enabled ) then
      section = sections[ s ].name;
      break;
    end;
  end;
  
  return section;
  
end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function public.FirstLaunch_GetInfo ()

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "stages",
    indice     = "name",
    criteria   = { like = "stage_1" }
  };

  local completed = lib_db.Execute( query ).result.completed;
  
  local stage_last = lib_db.Player.Info().stage_last;

  if ( completed or stage_last ~= "stage_1" ) then
    
    return false;
    
  end;

  return "stage_1";

end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function public.Session_GetDuration ()

  return private.session_time_seconds;

end;
--------------------------------------------------------------------------------
function public.Session_StartTime ( time )

  if ( time ) then
    private.session_time_launch = time;
  end;

  ne.msg.Subscribe(
    private.session_time_event,
    public.Session_OnTimer
  );

  parts.time.event.Shedule(
    private.session_time_token,
    "SESSION_GAME_TIME",
    private.session_time_resolution,
    private.session_time_event,
    {}
  );

end;
--------------------------------------------------------------------------------
function public.Session_OnTimer ()

  ne.msg.Unsubscribe(
    private.session_time_event,
    public.Session_OnTimer
  );

  TimeFinishEvent();
  TimeUnsheduleEvent( private.session_time_token );

  local current_time = common.time.GetCurrent();
  local ingame_time  = current_time - private.session_time_launch;
  local diff         = ingame_time  - private.session_time_seconds;

  private.session_time_seconds = ingame_time;

  --DbgTrace(
  --  "[ INFO ][ m_game ] Time in game: "..ingame_time..", diff: "..diff
  --);

  mediator.Notify(
    "Model_Game_SessionTick",
    {      
      timestamp = current_time,
      duration  = ingame_time,
      diff      = diff
    }
  );

  public.Session_StartTime();

end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function public.LocalNotification_Register ( notification_data )

  local query =
  {
    query      = "insert",
    group      = "sml",
    collection = "notifications",
    value      = notification_data
  };
  
  local is_inserted = lib_db.Execute( query ).result;

  return is_inserted;

end;
--------------------------------------------------------------------------------
function public.LocalNotification_Unregister ( notification_name )

  local query =
  {
    query      = "delete",
    group      = "sml",
    collection = "notifications",
    indice     = "name",
    criteria   = { like = notification_name }
  };
  
  local is_deleted = lib_db.Execute( query ).result;

  return is_deleted;

end;
--------------------------------------------------------------------------------
function public.LocalNotification_GetData ( notification_name )

  local query =
  {
    query      = "find",
    group      = "sml",
    collection = "notifications",
    indice     = "name",
    criteria   = { like = notification_name }
  };
  
  local notification_data = lib_db.Execute( query ).result;

  if ( notification_data ) then
  
    local current_time = parts.time.server.Timestamp();

    -- JK: self-destructing expired notifications, baby
    if ( current_time >= notification_data.time ) then

      public.LocalNotification_Unregister( notification_name );
      return nil;

    end;

  end;

  return notification_data;

end;
--------------------------------------------------------------------------------
--function *** Game log *** () end;
--------------------------------------------------------------------------------
function public.GameLog_AddEvent( sEvent, tEventData )
  private.game_log[ private.game_log_pos ] = {
    event = sEvent,
    data = tEventData,
    created = parts.time.server.Timestamp();
  };
  
  private.game_log_pos = private.game_log_pos + 1;
  if ( private.game_log_pos > private.game_log_size ) then
    private.game_log_pos = 1;
  end;
end;
--------------------------------------------------------------------------------
function private.GameLog_Send()
  public.GameLog_AddEvent( "game_closed", {} );
  lib_rpc.Schedule( 
    "gamelog",
    "send",
    { log = private.game_log },
    nil,
    true,
    true
  );
end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function private.GetNotification ()
  
  local query =
  {
    query      = "find",
    group      = "sml",
    collection = "game_state"
  };
  
  local result = lib_db.Execute( query ).result;
  
  return result.notifications or {};
  
end;
--------------------------------------------------------------------------------
function public.SetNotification ( target, params )
  
  local notifications = private.GetNotification();
  
  notifications[ target ] = params;
  
  local query =
  {
    query      = "update",
    group      = "sml",
    collection = "game_state",
    set        = 
    {
      notifications = notifications
    }
  };

  lib_db.Execute( query );
  
end;
--------------------------------------------------------------------------------
function public.GetNotification ( target )
  
  local notifications = private.GetNotification();
  
  return notifications[ target ];
  
end;
--------------------------------------------------------------------------------