-- name=m_hearts
--------------------------------------------------------------------------------
function public.__InitModule ()
  
  private.blocked_heart = {
    is_was_block = false
  };
  
  private.recover = {
    token = "tmr_hearts_recover",
    time  = lib_db.Settings( "hearts" ).time,
    event = "Event_Hearts_RecoverDone",
    start = false
  };
  
  private.immortality = {
    token = "tmr_hearts_immortality",
    event = "Event_Hearts_Immortality"
  };
  
  private.FixSave();
  private.InitCheck();
  
  ne.msg.Subscribe( private.recover.event, private.Recover_OnFinished );
  ne.msg.Subscribe( private.immortality.event, private.Immortality_EventFinish );
  
  --Log.i( "Initialized." );
  
end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function public.Setup ( funcs )
  
  private.func_get_stats = funcs.func_get_stats;
  private.func_get_remain_time_speedup = funcs.func_get_remain_time_speedup;
  
  if ( ( public.Recover_IsActive() ) and 
       ( not private.recover.start ) and 
       ( private.GetBoostValue() == private.GetDefaultBoostValue() ) ) then
    
    local callback = private.Recover_GetCallback();
    
    private.SynchronizeRecoverHearts( callback );
    
  end;
  
  if ( ( public.Recover_IsActive() ) and 
       ( private.GetBoostValue() ~= private.GetDefaultBoostValue() ) ) then
    
    local callback = private.Recover_GetCallback();
    
    private.SynchronizeRecoverHeartsWithBoost( callback );
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.GetBoostValue ()
  
  if ( private.func_get_stats ) then
    return 1 + private.func_get_stats().health_speed;
  end;
  
  return 1;
  
end;
--------------------------------------------------------------------------------
function private.GetDefaultBoostValue ()
  
  return 1;
  
end;
--------------------------------------------------------------------------------
function private.GetRemainTimeSpeedUp ()
  
  if ( private.func_get_remain_time_speedup ) then
    local time = private.func_get_remain_time_speedup();
    return math.max( 0, time );
  end;
  
  return 0;
  
end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function private.Storage_GetData ()
  
  return lib_db.Player.Info().hearts;
  
end;
--------------------------------------------------------------------------------
function private.Storage_UpdateData ( hearts_info )
  
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        = { hearts = hearts_info }
  };

  return lib_db.Execute( query ).result;
  
end;
--------------------------------------------------------------------------------
function public.GetSettings ()

  local settings = lib_db.Settings( "hearts" );

  return settings;

end;
--------------------------------------------------------------------------------
function public.GetInfo ()
  
  local storage_data = private.Storage_GetData();
  
  return storage_data;
  
end;
--------------------------------------------------------------------------------
function public.GetBonusCount ()

  local info = private.Storage_GetData();
  
  return ( info.facebook or 0 ) + ( info.donate or 0 );

end;
--------------------------------------------------------------------------------
function public.GetCount ()
  
  local info      = private.Storage_GetData();
  local count     = info.count or 0;
  local max_count = public.GetMaxCount();
  
  local is_blocked = public.IsBlock();

  if ( is_blocked ) then
    count = math.min( count + 1, max_count );
  end;
  
  return count, is_blocked;
  
end;
--------------------------------------------------------------------------------
function public.GetMaxCount ()
  
  local basic_count = public.GetSettings().count;
  local bonus_count = public.GetBonusCount();
  
  return ( basic_count + bonus_count );
  
end;
--------------------------------------------------------------------------------
function public.TryChangeCount ( count_delta, reason )

  local is_immortality = public.Immortality_IsActive();
  
  -- JK: check for immortality if delta is negative
  if ( count_delta < 0 ) then
    if ( is_immortality ) then
      return true;
    end;
  end;

  local hearts_info   = private.Storage_GetData();
  local count_max     = public.GetMaxCount();
  local count_current = hearts_info.count;
  local count_new     = count_current + count_delta;

  if ( count_new < 0 ) then
    mediator.Notify( "Model_Hearts_CountChangeFailed" );
    return false, count_new;
  end;

  local delta_real = count_delta;

  if ( count_new > count_max ) then
    delta_real = count_max - count_current;  
    count_new  = count_max;
  end;

  if ( delta_real == 0 ) then
    return false, count_new;
  end;

  hearts_info.count = count_new;

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        = { hearts = hearts_info }
  };

  if ( lib_db.Execute( query ).result ) then

    if ( count_new < count_max ) and ( not public.Recover_IsActive() ) then
      private.Recover_Start();
    end;
    
    if ( count_new == count_max ) and ( public.Recover_IsActive() ) then
      private.Recover_Unschedule();
    end;
    
    mediator.Notify(
    "Model_Hearts_CountChanged",
      {
        count  = count_new,
        max    = count_max,
        delta  = delta_real,
        reason = reason
      }
    );

    return true, count_new;

  end;

  return false, count_new;

end;
--------------------------------------------------------------------------------
function public.AddSocialBonus ()
  
  local hearts_info = private.Storage_GetData();
  hearts_info.facebook = 1;

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        = { hearts = hearts_info }
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function public.AddDonateBonus ( count )

  local hearts_info = private.Storage_GetData();
  hearts_info.donate = hearts_info.donate + count;
  
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        = { hearts = hearts_info }
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function private.FixSave ()

  local base_count = public.GetSettings().count;
  local max_count  = public.GetSettings().count_max;
  
  local facebook_max = tern( IsPC() )( 0, 1 );
  local donation_max = max_count - base_count;
  
  local saved_data = private.Storage_GetData();
  local need_to_fix = false;
  
  if ( saved_data.facebook > facebook_max ) then
    
    need_to_fix = true;
  
    local facebook_over = saved_data.facebook - facebook_max;
    saved_data.count = math.max( 0, saved_data.count - facebook_over );
    saved_data.facebook = facebook_max;
  
  end;
  
  if ( saved_data.donate > donation_max ) then
    
    need_to_fix = true;
  
    local donation_over = saved_data.donate - donation_max;
    saved_data.count = math.max( 0, saved_data.count - donation_over );
    saved_data.donate = donation_max;
  
  end;
  
  if ( saved_data.count > max_count ) then
  
    need_to_fix = true;
    saved_data.count = max_count;
  
  end;
  
  if ( need_to_fix ) then

    local query =
    {
      query      = "update",
      group      = "player",
      collection = "info",
      set        = { hearts = saved_data }
    };

    lib_db.Execute( query );

  end;

end;
--------------------------------------------------------------------------------
function private.InitCheck ()
  
  local is_full = ( public.GetCount() == public.GetMaxCount() );
  
  if ( not is_full ) and ( not public.Recover_IsActive() ) then
  
    local hearts_max = public.GetMaxCount();
    public.TryChangeCount( hearts_max, "recover_fixed" );
  
  end
  
end;
--------------------------------------------------------------------------------
function public.GetFullHeartsRestoreTime ()
  
  local hearts_needed = public.GetMaxCount() - public.GetCount();
  local full_time     = 0;
  local time_remain   = public.Recover_TimeRemains();
  
  if ( hearts_needed == 0 ) then
    return full_time;
  end;
  
  if ( private.GetBoostValue() ~= private.GetDefaultBoostValue() ) then
    
    local time_speedup = private.GetRemainTimeSpeedUp();
    local rate = private.GetBoostValue();
    
    if ( time_speedup < time_remain ) then
      time_remain = math.ceil( rate * ( time_remain - time_speedup ) + time_speedup );
      full_time = time_remain + private.Recover_Time() * ( hearts_needed - 1 );
      return full_time;
    end;
    
    if ( hearts_needed == 1 ) then
      return time_remain;
    end;
    
    full_time = full_time + time_remain;
    time_speedup = time_speedup - time_remain;
    
    local time_nospeedup = private.Recover_Time() * ( hearts_needed - 1 );
    local time_speedup_format = time_speedup * rate;
    
    if ( time_speedup_format > time_nospeedup ) then
      full_time = full_time + math.ceil( time_nospeedup / rate );
      return full_time;
    end;
    
    full_time = full_time + time_speedup + ( time_nospeedup - time_speedup_format );
    return full_time;
    
  end;
  
  full_time = time_remain + private.Recover_Time() * ( hearts_needed - 1 );
  
  return full_time;
  
end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function public.IsBlock ()
  
  return private.blocked_heart.is_was_block;
  
end;
--------------------------------------------------------------------------------
function public.Block ()
  
  private.blocked_heart.is_was_block = true;
  
  public.TryChangeCount( -1 );
  
end;
--------------------------------------------------------------------------------
function public.Unblock ( type_unblock )
  
  if ( not public.IsBlock() ) then
    return;
  end;
  
  private.blocked_heart.is_was_block = false;
  
  if ( public.Immortality_IsActive() ) then
    return;
  end;
  
  if ( type_unblock == "win" ) then
    public.TryChangeCount( 1 );
    return;
  end;
  
  local delta = public.GetMaxCount() - public.GetCount();
  
  if ( delta == 0 ) then
    public.TryChangeCount( -1 );
  end;
  
  if ( delta == 1 ) then
    private.Recover_Stop();
    private.Recover_Start();
  end;
  
  mediator.Notify(
  "Model_Hearts_CountChanged",
    {
      count  = public.GetCount(),
      max    = public.GetMaxCount(),
      delta  = 0,
      reason = ""
    }
  );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.RunActionBoost ()
  
  if ( public.Recover_IsActive() ) then
    private.Recover_SpeedupStart();
  end;
  
end;
--------------------------------------------------------------------------------
function public.CompleteActionBoost ()
  
  if ( public.Recover_IsActive() ) then
    private.Recover_SpeedupComplete();
  end;
  
end;
--------------------------------------------------------------------------------
function private.Recover_SpeedupStart ( time_recover )
  
  local time_remain = time_recover or public.Recover_TimeRemains();
  local rate = private.GetBoostValue();
  
  private.Recover_Stop();
  
  local time_passed = private.Recover_Time() - time_remain;
  local time = math.ceil( time_remain / rate );
  
  local data = {
    time_speedup = private.GetRemainTimeSpeedUp(),
    time = time,
    rate = rate
  };
  
  private.Recover_Schedule( time, data );
  
end;
--------------------------------------------------------------------------------
function private.Recover_SpeedupComplete ()
   
  local callback = private.Recover_GetCallback();
  
  if ( not callback.rate ) then return; end;
  
  local time_remain = public.Recover_TimeRemains();
  local time = math.ceil( time_remain * callback.rate );
  
  private.Recover_Unschedule();
  
  if ( private.GetBoostValue() > private.GetDefaultBoostValue() ) then
    private.Recover_SpeedupStart( time );
  else
    private.Recover_Schedule( time );
  end;
  
end;
--------------------------------------------------------------------------------
function private.Recover_Schedule ( time, data )
  
  TimeSheduleEvent(
    private.recover.token,
    "GLOBAL_TIME",
    time or private.Recover_Time(),
    private.recover.event,
    data or { time = time }
  );

  TimeStartWatch( "wth_hearts_recover", "GLOBAL_TIME" );
  
end;
--------------------------------------------------------------------------------
function private.Recover_Unschedule ()
  
  TimeUnsheduleEvent( private.recover.token );
  TimeStopWatch( "wth_hearts_recover" );
  
end;
--------------------------------------------------------------------------------
function private.Recover_Start ( time )
  
  if ( private.GetBoostValue() > private.GetDefaultBoostValue() ) then
    private.Recover_SpeedupStart( time or private.Recover_Time() );
  else
    private.Recover_Schedule( time );
  end;
  
  mediator.Notify( "Model_Hearts_RecoverStart", { is_block = public.IsBlock() } );
  
end;
--------------------------------------------------------------------------------
function private.Recover_Stop ()

  private.Recover_Unschedule();

end;
--------------------------------------------------------------------------------
function private.Recover_Time ()

  return private.recover.time;

end;
--------------------------------------------------------------------------------
function public.Recover_TimeRemains ()

  return TimeGetEventRemainingTime( private.recover.token );

end;
--------------------------------------------------------------------------------
function public.Recover_IsActive ()

  return TimeIsEventExist( private.recover.token );

end;
--------------------------------------------------------------------------------
function private.Recover_GetCallback ()
  
  local callback = lib_db.Events_GetCallbackByToken( private.recover.token ) or {};
  
  return callback;
  
end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function public.Immortality_Start ( time )

  if ( time <= 0 ) then return false; end;

  -- JK: if immortality is active, increase time
  if ( public.Immortality_IsActive() ) then

    local remains = public.Immortality_TimeRemains();

    public.Immortality_Stop( true );
    private.Immortality_EventStart( time + remains );

    return true;

  end;

  -- JK: check if hearts is restoring, if so - restore instantly
  --if ( public.Recover_IsActive() ) then

    --private.Recover_Unschedule();

    local max_count         = public.GetMaxCount();
    local count, is_blocked = public.GetCount();

    -- JK: when heart is blocked, GetCount reports + 1, we need real count
    if ( is_blocked ) then count = count - 1; end;

    local delta = max_count - count;

    public.TryChangeCount( delta );

  --end;

  private.Immortality_EventStart( time );
  
  return true;
  
end;
--------------------------------------------------------------------------------
function private.Immortality_EventStart ( time, skip_notification )

  TimeSheduleEvent(
    private.immortality.token,
    "GLOBAL_TIME",
    time,
    private.immortality.event,
    {}
  );

  if ( not skip_notification ) then
    mediator.Notify( "Model_Hearts_ImmortalityToggled" );
  end;

end;
--------------------------------------------------------------------------------
function public.Immortality_IsActive ()

  return TimeIsEventExist( private.immortality.token );

end;
--------------------------------------------------------------------------------
function public.Immortality_TimeRemains ()

  return TimeGetEventRemainingTime( private.immortality.token );

end;
--------------------------------------------------------------------------------
function private.Immortality_EventFinish ()
  
  public.Immortality_Stop();
  
end;
--------------------------------------------------------------------------------
function public.Immortality_Stop ( skip_notification )

  TimeFinishEvent();
  TimeUnsheduleEvent( private.immortality.token );

  if ( not skip_notification ) then

    mediator.Notify(
      "Model_Hearts_ImmortalityToggled",
      {
        immortality = false
      }
    );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Lots_GetAvailable ( func_is_purchased, is_fb_connected )

  local lots_available = {};
  local info = public.GetInfo();
  local facebook, donate = info.facebook, info.donate;
  local hearts_settings = public.GetSettings();
  local max_extension = hearts_settings.count_max - hearts_settings.count;
  local extension = facebook + donate;

  local query =
  {
    query      = "find_all",
    group      = "static",
    collection = "lots"
  };

  local lot_data_list = lib_db.Execute( query ).result;
  local max_lots = 2 - extension;
  
  for i = 1, #lot_data_list do

    local lot = lot_data_list[ i ];

    if ( lot.type == "shop"                                              ) and 
       ( lot.currency_type == "hearts"                                   ) and
       ( private.IsLotBelongsPlatform( lot.name_ios )                    ) and
       ( not func_is_purchased( lot.name_ios )                           ) and
       ( ( extension + lot.currency_count ) <= max_extension ) and
       ( #lots_available < max_lots ) then

      table.insert( lots_available, lot );

    end;

  end;

  if ( IsMobile()             ) and 
     ( facebook == 0          ) and
     ( donate < max_extension ) and
     ( not is_fb_connected    ) then
    table.insert( lots_available, 1, { name = "hearts_facebook", currency_count = 1 } );
  end;
  
  local purchase_lots = ( GetLotsInfo() or {} ).lots;

  if ( purchase_lots ) and ( #purchase_lots > 0 ) then

    for p = 1, #purchase_lots do

      local purchase_lot = purchase_lots[ p ];

      for i = 1, #lots_available do

        local lot = lots_available[ i ];

        if ( lot.name_ios == ( purchase_lot.name_ios or purchase_lot.name ) ) then
          lot.price = private.Lot_GetPrice( purchase_lot );
          break;
        end;

      end;

    end;

  end;

  return lots_available;

end;
--------------------------------------------------------------------------------
function private.Lot_GetPrice ( lot_data )
  
  if ( IsPC() ) then
    
    local currency = {
      JPY = "(.*) .",
      USD = ". (.*)",
      EUR = "(.*) ."
    };
    
    local price = ne.strings.Get( "str_dialog_shop_gold_price_"..string.lower( lot_data.currency ), 
                  lot_data.price:match( currency[ lot_data.currency ] ) );

    return price;
    
  end;
  
  return lot_data.price;

end;
--------------------------------------------------------------------------------
function private.IsLotBelongsPlatform ( lot_name )

  local lots =
  {
    ios =
    { 
      "1life",
      "2life",
      "5thlife"
    },
    pc =
    {
      "1life",
      "2lives",
      "5thlife"
    },
    google =
    {
      "1life",
      "2life",
      "5thlife"
    },
    amazon =
    {
      "onelife",
      "twolives",
      "fifthlife"
    }
  };
  
  local platform = GetPlatform();
  platform = tern( platform == "android" )( wiz.Android.GetPlatform(), platform );
  
  if ( common.IsEntityInList( lots[ platform ], lot_name ) ) then
    
    return true;
    
  end;
  
  return false;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Recover_OnFinished ( msg_type, msg_params )
  
  TimeFinishEvent();
  local time_passed_watch  = TimeGetWatchTime( "wth_hearts_recover" );
  private.Recover_Unschedule();
  local complete, count = private.GetCountRecoverHearts( msg_params, time_passed_watch );
  
  if ( not complete ) then
    return;
  end;
  
  public.TryChangeCount( count );
  
  mediator.Notify( "Model_Hearts_RecoverFinished", count );

end;
--------------------------------------------------------------------------------
function private.GetCountRecoverHearts ( callback, time_passed_watch )
  
  local count = 1;
  local complete = true;
  local time_recover = private.Recover_Time();
  local time_cd_on_recover = time_recover;
  
  private.recover.start = true;
  
  if ( callback.rate ) then
    
    local time_speedup = callback.time_speedup;
    local rate = callback.rate;
    local time = callback.time;
  
    local time_real = time * rate;
    local time_speedup_real = time_speedup * rate;
    local time_real_passed_watch = 0;
    
    if ( time_passed_watch > time_speedup ) then
      time_real_passed_watch = time_passed_watch + time_speedup * ( rate - 1 );
    else
      time_real_passed_watch = time_passed_watch * rate;
    end;
    
    if ( time_real > time_real_passed_watch )  then
    
      local time_cd = time_real - time_real_passed_watch;
      
      if ( private.GetBoostValue() > private.GetDefaultBoostValue() ) then
        private.Recover_SpeedupStart( time_cd );
      else
        private.Recover_Schedule( time_cd );
      end;
      
      return false, 0;
      
    end;
    
    time_real_passed_watch = time_real_passed_watch - time_real;
    
    local count_recover = math.floor( time_real_passed_watch / time_recover );
    
    count = count + count_recover;
    
    time_cd_on_recover = time_recover - ( time_real_passed_watch - count_recover * time_recover );
    
  elseif ( callback.time ) then
    
    local time = callback.time;
    
    time_passed_watch = time_passed_watch - time;
    
    if ( time_passed_watch > 0 ) then
      
      local count_recover = math.floor( time_passed_watch / time_recover );
      
      count = count + count_recover;
      
      time_cd_on_recover = time_recover - ( time_passed_watch - count_recover * time_recover );
      
    end;
    
  else
    
    if ( time_passed_watch > time_recover ) then
      
      count = math.floor( time_passed_watch / time_recover );
      
      time_cd_on_recover = time_recover - ( time_passed_watch - count * time_recover );
      
    end;
    
  end;
  
  local max_count = public.GetMaxCount();
  local current_count     = public.GetCount();
  local delta     = max_count - current_count;
  
  if ( time_cd_on_recover > 0 ) and ( count < delta ) then
    private.Recover_Start( time_cd_on_recover );
  end;
  
  return complete, count;
  
end;
--------------------------------------------------------------------------------
function private.SynchronizeRecoverHearts ( callback )
  
  if ( callback.rate ) then
    
    local time_passed_watch = TimeGetWatchTime( "wth_hearts_recover" );
    local time_speedup = callback.time_speedup;
    local rate = callback.rate;
    local time = callback.time;
    
    local time_cd = ( time - time_speedup ) * rate - ( time_passed_watch - time_speedup );
    
    private.Recover_Unschedule();
    
    private.Recover_Schedule( time_cd );
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.SynchronizeRecoverHeartsWithBoost ( callback )
  
  if ( not callback.rate ) then
    private.Recover_SpeedupStart();
  end;
  
end;
--------------------------------------------------------------------------------