-- name=m_migrations
--------------------------------------------------------------------------------

--[[

  Порядок работы с модулем:

  1) Напиши функцию-миграцию private.Migration_XXX_tttttt,
  где XXX - след. ревизия pack build из teamcity,
  tttttt - тип миграции: client, server или др.

  2) Проверь миграцию тестером private.Tester, например:

    private.Tester( private.Migration_404_client )

    путём раскомментирования блока кода в Init

  Он выводит в лог запросы к БД и результаты, но фактически
  не выполняет "update" и "insert"

  3) Внеси миграцию в private.migrations_list.new чтобы она
  заработала у старых пользователей

  4) Внеси миграцию в config.json -> sml/migrations чтобы она
  не заработала у новых пользователей

]]

--------------------------------------------------------------------------------
function public.__InitModule ()

  --

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Run ( is_online )

  private.migrations_list = {
    old = {
      { version = { major = 1, minor = 2 }, handler = private.Migration_1_2 },
      { version = { major = 1, minor = 3 }, handler = private.Migration_1_3 }
    },
    new = {
      --AL: Don't be hurry to add migration to this list, test it on Tester before
      { version = { revision = 360 }, name = "Migration_360_client" },
      { version = { revision = 371 }, name = "Migration_371_client" },
      { version = { revision = 385 }, name = "Migration_385_client" },
      { version = { revision = 393 }, name = "Migration_393_client" },
      { version = { revision = 400 }, name = "Migration_400_client" },
      { version = { revision = 402 }, name = "Migration_402_client" },
      { version = { revision = 403 }, name = "Migration_403_client" },
      { version = { revision = 404 }, name = "Migration_404_client" },
      { version = { revision = 437 }, name = "Migration_437_client" },
      { version = { revision = 438 }, name = "Migration_438_client" },
      { version = { revision = 450 }, name = "Migration_450_client" },
      { version = { revision = 493 }, name = "Migration_493_client" },
      { version = { revision = 494 }, name = "Migration_494_client" },
      { version = { revision = 532 }, name = "Migration_532_client" },
      { version = { revision = 553 }, name = "Migration_553_client" },
      { version = { revision = 560 }, name = "Migration_560_client" },
      { version = { revision = 572 }, name = "Migration_572_client" },
      { version = { revision = 573 }, name = "Migration_573_client" },
      { version = { revision = 574 }, name = "Migration_574_client" },
      { version = { revision = 596 }, name = "Migration_596_client" },
      { version = { revision = 637 }, name = "Migration_637_client" },
      { version = { revision = 763 }, name = "Migration_763_client" },
      { version = { revision = 777 }, name = "Migration_777_client" },
      { version = { revision = 918 }, name = "Migration_918_client" },
      { version = { revision = 931 }, name = "Migration_931_client" },
      { version = { revision = 953 }, name = "Migration_953_client" },
      { version = { revision = 954 }, name = "Migration_954_client" },
      { version = { revision = 931 }, name = "Migration_984_client" },
      { version = { revision = 1106 }, name = "Migration_1106_client" },
      { version = { revision = 1264 }, name = "Migration_1264_client" },
      { version = { revision = 1271 }, name = "Migration_1271_client" },
      { version = { revision = 1312 }, name = "Migration_1312_client" },
      { version = { revision = 1313 }, name = "Migration_1313_client" },
      { version = { revision = 1314 }, name = "Migration_1314_client" },
      { version = { revision = 1467 }, name = "Migration_1467_client" },
      { version = { revision = 1471 }, name = "Migration_1471_client" },
      { version = { revision = 1472 }, name = "Migration_1472_client" },
      { version = { revision = 1624 }, name = "Migration_1624_client" },
      { version = { revision = 1624 }, name = "Migration_1693_client" },
      { version = { revision = 1624 }, name = "Migration_1694_client" },
      { version = { revision = 1624 }, name = "Migration_1695_client" },
      { version = { revision = 1624 }, name = "Migration_1696_client" },
      { version = { revision = 1624 }, name = "Migration_1697_client" },
      { version = { revision = 1624 }, name = "Migration_1698_client" },
      { version = { revision = 1624 }, name = "Migration_1699_client" },
      { version = { revision = 1624 }, name = "Migration_1701_client" },
      { version = { revision = 1624 }, name = "Migration_1702_client" },
      { version = { revision = 1624 }, name = "Migration_1703_client" },
      { version = { revision = 1965 }, name = "Migration_1965_client" },
      { version = { revision = 1984 }, name = "Migration_1984_client" },
      { version = { revision = 1994 }, name = "Migration_1994_client" }
    }
  };

  private.version = private.GetParsedVersion( private.GetVersion() );

  local ignore_saved = private.OldMigrationsStart( private.version );

  private.NewMigrationsStart( is_online, private.version, ignore_saved );

  -- JK: uncomment below to test your migration
  --private.Tester( private.Migration_1703_client );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.GetVersion ()

  local result, storage_version = parts.storage.data.Read( "sml/need_run_migration" );

  if ( storage_version                     ) and
     ( storage_version ~= ""               ) and
     ( storage_version ~= "false"          ) and
     ( not storage_version:find( "0.0.0" ) ) then

    parts.storage.data.Write( "sml/need_run_migration", "false" );
    return storage_version;

  else

    return parts.version.GetSaved( "game" );

  end;

end;
--------------------------------------------------------------------------------
function private.GetParsedVersion ( raw_version )

  local version = {};

  version.major, version.minor, version.build, version.revision =
    string.match( raw_version, "(%d+).(%d+).(%d+).(%d+)" );

  for k, v in pairs( version ) do
    version[ k ] = tonumber( v );
  end;

  return version;

end;
--------------------------------------------------------------------------------
function private.Tester ( migration )

  if ( migration and type( migration ) == "function" ) then

    DbgTrace( "***********************************" );

    local oldDbExecuteQuery = DbExecuteQuery;

    DbExecuteQuery = function ( query )

      DbgTrace( "+++++++++++++++++++++++++++++++++" );

      if ( query.query:find( "find" ) ) then

        local result = oldDbExecuteQuery( query );

        DbgTrace(
          common.table.ToString( query )..
          "-> -> -> -> -> -> -> -> -> -> -> -> ->"..
          common.table.ToString( result )
        );

        return result;

      else

        DbgTrace( common.table.ToString( query ) );

      end;

    end;

    migration();

    DbExecuteQuery = oldDbExecuteQuery;

  end;

end;
--------------------------------------------------------------------------------
-- function *** OLD *** ()
--------------------------------------------------------------------------------
function private.OldMigrationsStart ( version )

  --Log.i( "[ OLD ] Saved game version: "..
  --  version.major.."."..version.minor.."."..version.build );

  if ( version.major == 1 and version.minor < 4 ) then

    for m, migration in ipairs( private.migrations_list.old ) do

      if
      ( version.major < migration.version.major ) or
      ( version.minor < migration.version.minor ) or
      ( migration.version.build and ( version.build < migration.version.build ))
      then

        Log.i( "[ OLD ] Migrating to version: "..
          migration.version.major.."."..
          migration.version.minor.."."..
          ( migration.version.build or "X" )
        );

        migration.handler();

      elseif ( version.major == 1 and version.minor == 3 ) then

        local query =
        {
          query      = "find",
          group      = "player",
          collection = "sections",
          indice     = "name",
          criteria   = { like = "section_1" }
        };
        local find_answer = lib_db.Execute( query ).result.intro;

        if ( not find_answer ) then

          Log.i( "[ OLD ] Using 1.3 migration fix" );
          migration.handler();

        end;

      end;

    end;

    return true;

  else

    --Log.i( "[ OLD ] No need to use old mechanics" );
    return false;

  end;

end;
--------------------------------------------------------------------------------
-- function *** OLD : LIST *** ()
--------------------------------------------------------------------------------
function private.Migration_1_2 ()

  -- Add 100 gold -- -- -- -- -- -- -- --
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        =
    {
      update = true,
      gold   = 100
    }
  };

  lib_db.Execute( query );

end;
--------------------------------------------------------------------------------
function private.Migration_1_3 ()

  -- Set section 1 intro as played -- -- -- -- -- -- -- -- -- -- -- -- -- --
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "sections",
    indice     = "name",
    criteria   = { like = "section_1" },
    set        = { intro = true }
  };

  lib_db.Execute( query );

  -- Fix stonegate quest -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  local quest_name = "qst_collect_stonegate_fragments";
  local quest_criteria = { like = quest_name };
  local item_name = "stonegate_star";

  query =
  {
    query      = "find",
    group      = "player",
    collection = "quests_current",
    indice     = "name",
    criteria   = quest_criteria
  };
  local quest_current = lib_db.Execute( query ).result;

  query =
  {
    query      = "find",
    group      = "static",
    collection = "quests",
    indice     = "name",
    criteria   = quest_criteria
  };

  local quest_data = lib_db.Execute( query ).result;

  if ( quest_current ) then

    quest_current.list = quest_current.list or {};

    if ( not common.IsEntityInList( quest_current.list, item_name )) then

      if ( #quest_current.list == #quest_data.target.list - 1 ) then

        local quest_remove_query =
        {
          query      = "delete",
          group      = "player",
          collection = "quests_current",
          indice     = "name",
          criteria   = quest_criteria
        };
        lib_db.Execute( quest_remove_query );

        local quest_insert_query =
        {
          query      = "insert",
          group      = "player",
          collection = "quests_completed",
          value      = { name = quest_name }
        };
        lib_db.Execute( quest_insert_query );

      else

        local new_list = quest_current.list;
        table.insert( new_list, item_name );

        local quest_update_query =
        {
          query      = "update",
          group      = "player",
          collection = "quests_current",
          indice     = "name",
          criteria   = quest_criteria,
          set        = { list = new_list }
        };
        lib_db.Execute( quest_update_query );

      end;

    end;

  else

    local quest_completed_query =
    {
      query      = "find",
      group      = "player",
      collection = "quests_completed",
      indice     = "name",
      criteria   = quest_criteria
    };
    local quest_completed = lib_db.Execute( quest_completed_query ).result;

    if ( not quest_completed ) then

      local build_info_query =
      {
        query      = "find",
        group      = "player",
        collection = "builds",
        indice     = "name",
        criteria   = { like = "island_1" }
      };
      local build_info = lib_db.Execute( build_info_query ).result;

      if (
      build_info and
      build_info.collected_fragments and (
      #build_info.collected_fragments > 0 )) then

        if ( #build_info.collected_fragments >= #quest_data.target.list - 1 ) then

          local quest_insert_query =
          {
            query      = "insert",
            group      = "player",
            collection = "quests_completed",
            value      = { name = quest_name }
          };
          lib_db.Execute( quest_insert_query );

        else

          local insert_value = {
            name = quest_name,
            list = build_info.collected_fragments
          };

          local insert_query =
          {
            query      = "insert",
            group      = "player",
            collection = "quests_current",
            value      = insert_value
          };
          lib_db.Execute( insert_query );

        end;

      end;

    end;

  end;

end;
--------------------------------------------------------------------------------
-- function *** NEW *** ()
--------------------------------------------------------------------------------
function private.NewMigrationsStart ( is_online, version, ignore_saved )

  local query;

  local function func_RunServerSideMigrations ()

    mediator.Unsubscribe(
      m_game.msg.ConnectRestored,
      func_RunServerSideMigrations
    );

    -- JK: server-side migrations run here

    --private.CheckAllMigrations();

  end;

  if ( not is_online ) then

    mediator.Subscribe(
      m_game.msg.ConnectRestored,
      func_RunServerSideMigrations
    );

  end;

  local migrations = {};

  if ( not ignore_saved ) then

    query =
    {
      query      = "find",
      group      = "sml",
      collection = "migrations"
    };

    migrations = lib_db.Execute( query ).result.migration or {};

  end;

  for m, migration in ipairs( private.migrations_list.new ) do

    if ( migration.name:find( "client" )                   ) or
       ( migration.name:find( "server" ) and ( is_online ) ) then

      migrations = private.NewTryMigrate( migrations, migration.name );

    end;

  end;

  query =
  {
    query      = "update",
    group      = "sml",
    collection = "migrations",
    set        = { migration = migrations }
  };

  lib_db.Execute( query );

end;
--------------------------------------------------------------------------------
function private.NewTryMigrate ( collection, migration_name )

  if ( not common.IsEntityInList( collection, migration_name ) ) then
    Log.i( "[ NEW ] Doing " .. migration_name );
    private[ migration_name ]();
    table.insert( collection, migration_name );
  end;

  return collection;

end;
--------------------------------------------------------------------------------
-- function *** NEW : LIST *** ()
--------------------------------------------------------------------------------
function private.Migration_360_client ()

  -- Reduce swordstone fragments amount. Give 50 gold for each -- -- -- -- -- --
  local find_build_query =
  {
    query      = "find",
    group      = "player",
    collection = "builds",
    indice     = "name",
    criteria   = { like = "puzzle_1" }
  };

  local swordstone_data = lib_db.Execute( find_build_query ).result;

  if (
  swordstone_data and
  swordstone_data.collected_fragments and
  #swordstone_data.collected_fragments >= 2 ) then

    local change_list = {
      [1]  = false, -- не удалять
      [2]  = true,  -- удалять
      [3]  = 2,     -- заменить на 2
      [4]  = true,  -- ...
      [5]  = true,
      [6]  = 3,
      [7]  = true,
      [8]  = true,
      [9]  = true,
      [10] = 4,
      [11] = true,
      [12] = true,
      [13] = 5,
      [14] = true,
      [15] = 6
    };

    local new_fragment_list = {};
    --local new_gold_count    = 0;

    for f = 1, #swordstone_data.collected_fragments do

      local fragment_num = tonumber( swordstone_data.collected_fragments[ f ]:match( "%d+" ) );
      local change       = change_list[ fragment_num ];

      if ( not change ) then
        table.insert( new_fragment_list, "swordstone_part_"..fragment_num );
      elseif ( type( change ) == "number" ) then
        table.insert( new_fragment_list, "swordstone_part_"..change );
        --new_gold_count = new_gold_count + 75;
      else
        --new_gold_count = new_gold_count + 50;
      end;

    end;

    local update_build_query =
    {
      query      = "update",
      group      = "player",
      collection = "builds",
      indice     = "name",
      criteria   = { like = "puzzle_1" },
      set        = {
        collected_fragments = new_fragment_list
      }
    };
    lib_db.Execute( update_build_query );
    -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
    local quest_name = "qst_collect_swordstone_fragments";

    local find_quest_query =
    {
      query      = "find",
      group      = "player",
      collection = "quests_current",
      indice     = "name",
      criteria   = { like = quest_name }
    };
    local quest_data = lib_db.Execute( find_quest_query ).result;

    if ( quest_data ) then

      local update_quest_query =
      {
        query      = "update",
        group      = "player",
        collection = "quests_current",
        indice     = "name",
        criteria   = { like = quest_name },
        set        = {
          list     = new_fragment_list
        }
      };
      lib_db.Execute( update_quest_query );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Migration_371_client ()

  -- Reduce swordstone fragments amount.
  local find_build_query =
  {
    query      = "find",
    group      = "player",
    collection = "builds",
    indice     = "name",
    criteria   = { like = "puzzle_1" }
  };
  local swordstone_data = lib_db.Execute( find_build_query ).result;

  if (
  swordstone_data and
  swordstone_data.collected_fragments and
  #swordstone_data.collected_fragments > 0 ) then

    local change_list = {
      [1] = true, --remove
      [2] = 1,
      [3] = 2,
      [4] = 3,
      [5] = 4,
      [6] = 5,
      [7] = 6
    };

    local new_fragment_list = {};

    for f = 1, #swordstone_data.collected_fragments do

      local fragment_num = tonumber( swordstone_data.collected_fragments[ f ]:match( "%d+" ) );
      local change       = change_list[ fragment_num ];

      if ( not change ) then
        table.insert( new_fragment_list, "swordstone_part_"..fragment_num );
      elseif ( type( change ) == "number" ) then
        table.insert( new_fragment_list, "swordstone_part_"..change );
      end;

    end;

    local update_build_query =
    {
      query      = "update",
      group      = "player",
      collection = "builds",
      indice     = "name",
      criteria   = { like = "puzzle_1" },
      set        = {
        collected_fragments = new_fragment_list
      }
    };
    lib_db.Execute( update_build_query );
    -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
    local quest_name = "qst_collect_swordstone_fragments";

    local find_quest_query =
    {
      query      = "find",
      group      = "player",
      collection = "quests_current",
      indice     = "name",
      criteria   = { like = quest_name }
    };
    local quest_data = lib_db.Execute( find_quest_query ).result;

    if ( quest_data ) then

      local update_quest_query =
      {
        query      = "update",
        group      = "player",
        collection = "quests_current",
        indice     = "name",
        criteria   = { like = quest_name },
        set        = {
          list     = new_fragment_list
        }
      };
      lib_db.Execute( update_quest_query );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Migration_385_client ()

  -- insert 1st swordstone fragment
  local query =
  {
    query      = "find",
    group      = "player",
    collection = "stages",
    indice     = "name",
    criteria   = { like = "stage_1" }
  };

  local stage_1_data = lib_db.Execute( query ).result;

  if ( stage_1_data and stage_1_data.state and stage_1_data.state > 1 ) then

    query =
    {
      query      = "find",
      group      = "player",
      collection = "builds",
      indice     = "name",
      criteria   = { like = "puzzle_1" }
    };
    local swordstone_data = lib_db.Execute( query ).result;

    if ( not swordstone_data ) then

      query =
      {
        query      = "insert",
        group      = "player",
        collection = "builds",
        value      = {
          name                = "puzzle_1",
          constructed         = false,
          unlocked            = false,
          collected_fragments = { "swordstone_part_1" }
        }
      };
      lib_db.Execute( query );

      query =
      {
        query      = "insert",
        group      = "player",
        collection = "quests_current",
        value      = {
          name     = "qst_collect_swordstone_fragments",
          list     = { "swordstone_part_1" }
        }
      };
      lib_db.Execute( query );

    elseif ( swordstone_data.constructed ) then

      local list = swordstone_data.collected_fragments;
      if ( not common.IsEntityInList( list, "swordstone_part_1" )) then

        table.insert( list, "swordstone_part_1" );

        query =
        {
          query      = "update",
          group      = "player",
          collection = "builds",
          indice     = "name",
          criteria   = { like = "puzzle_1" },
          set        = { collected_fragments = list }
        };
        lib_db.Execute( query );

      end;

    elseif ( #swordstone_data.collected_fragments > 0 ) then

      local change_list = {
        [1] = 2,
        [2] = 3,
        [3] = 4,
        [4] = 5,
        [5] = 6
      };

      local list = {};
      table.insert( list, "swordstone_part_1" );

      for f = 1, #swordstone_data.collected_fragments do

        local fragment_num = tonumber( swordstone_data.collected_fragments[ f ]:match( "%d+" ) );
        local change       = change_list[ fragment_num ];
        table.insert( list, "swordstone_part_"..change );

      end;

      query =
      {
        query      = "update",
        group      = "player",
        collection = "builds",
        indice     = "name",
        criteria   = { like = "puzzle_1" },
        set        = { collected_fragments = list }
      };
      lib_db.Execute( query );

      query =
      {
        query      = "update",
        group      = "player",
        collection = "quests_current",
        indice     = "name",
        criteria   = { like = "qst_collect_swordstone_fragments" },
        set        = { list = list }
      };
      lib_db.Execute( query );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Migration_393_client ()

  -- Change resources type to table.
  -- "resource1" : -1 -> "resource1" : [ -1 ]
  local resources_list = { "resource1", "resource2", "resource3" };

  for r = 1, #resources_list do

    local resource = resources_list[ r ];

    local query =
    {
      query      = "find",
      group      = "player",
      collection = "info"
    };

    local count = lib_db.Execute( query ).result[ resource ];

    if ( type( count ) == "number" ) then

      query =
      {
        query      = "update",
        group      = "player",
        collection = "info",
        set = { [ resource ] = { count, count } }
      };

      lib_db.Execute( query );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Migration_400_client ()

  -- SPECIAL FIX FOR THOSE, WHO COME 1.4 FROM 1.3 AND GAIN sml_migrations FROM config

  local query =
  {
    query      = "find",
    group      = "sml",
    collection = "migrations"
  };
  local migrations_answer = lib_db.Execute( query ).result.migration;

  query =
  {
    query      = "find",
    group      = "player",
    collection = "info"
  };
  local resource_answer = lib_db.Execute( query ).result.resource1;

  if (
    migrations_answer and
    common.IsEntityInList( migrations_answer, "Migration_393_client" )
  ) and (
    resource_answer and
    type( resource_answer ) == "number"
  ) then

    private.Migration_360_client();
    private.Migration_371_client();
    private.Migration_385_client();
    private.Migration_393_client();

  end;

end;
--------------------------------------------------------------------------------
function private.Migration_402_client ()

  -- FIX: insert 1st swordstone fragment
  local query =
  {
    query      = "find",
    group      = "player",
    collection = "stages",
    indice     = "name",
    criteria   = { like = "stage_1" }
  };

  local stage_1_data = lib_db.Execute( query ).result;

  query =
  {
    query      = "find",
    group      = "player",
    collection = "builds",
    indice     = "name",
    criteria   = { like = "puzzle_1" }
  };

  local swordstone_data = lib_db.Execute( query ).result;

  if ( stage_1_data and stage_1_data.state and stage_1_data.state > 1 ) and
    ( swordstone_data and
    swordstone_data.collected_fragments and
    #swordstone_data.collected_fragments == 0 ) then

    local list = { "swordstone_part_1"};

    query =
    {
      query      = "update",
      group      = "player",
      collection = "builds",
      indice     = "name",
      criteria   = { like = "puzzle_1" },
      set        = { collected_fragments = list }
    };

    lib_db.Execute( query );

    query =
    {
      query      = "update",
      group      = "player",
      collection = "quests_current",
      indice     = "name",
      criteria   = { like = "qst_collect_swordstone_fragments" },
      set        = { list = list }
    };

    lib_db.Execute( query );

  end;

end;
--------------------------------------------------------------------------------
function private.Migration_403_client ()

  --FIX: goldkeys quest, stars quest
  local find_goldkey_query =
  {
    query      = "find",
    group      = "player",
    collection = "quests_current",
    indice     = "name",
    criteria   = { like = "qst_collect_gold_key" }
  };
  local goldkey_quest_data = lib_db.Execute( find_goldkey_query ).result;

  if ( goldkey_quest_data and goldkey_quest_data.list and
  #goldkey_quest_data.list > 0 ) then

    --"stage_2", "stage_3", "stage_4", "stage_5", "stage_7"
    -- to
    --"stage_2", "stage_4", "stage_6", "stage_8", "stage_10"

    local change_list = {
      [2] = 2,
      [3] = 4,
      [4] = 6,
      [5] = 8,
      [7] = 10
    };

    local list = {};

    for f = 1, #goldkey_quest_data.list do

      local stage_num = tonumber( goldkey_quest_data.list[ f ]:match( "%d+" ) );
      local change    = change_list[ stage_num ];
      table.insert( list, "stage_"..change );

    end;

    local update_goldkey_query =
    {
      query      = "update",
      group      = "player",
      collection = "quests_current",
      indice     = "name",
      criteria   = { like = "qst_collect_gold_key" },
      set        = { list = list }
    };
    lib_db.Execute( update_goldkey_query );

  end;

  local find_stars_query =
  {
    query      = "find",
    group      = "player",
    collection = "quests_current",
    indice     = "name",
    criteria   = { like = "qst_collect_star" }
  };
  local stars_quest_data = lib_db.Execute( find_stars_query ).result;

  if ( stars_quest_data and stars_quest_data.list and
  #stars_quest_data.list > 0 ) then

    --"stage_7", "stage_8", "stage_9", "stage_10", "stage_11"
    -- to
    --"stage_1", "stage_3", "stage_5", "stage_7", "stage_9"

    local change_list = {
      [7]  = 1,
      [8]  = 3,
      [9]  = 5,
      [10] = 7,
      [11] = 9
    };

    local list = {};

    for f = 1, #stars_quest_data.list do

      local stage_num = tonumber( stars_quest_data.list[ f ]:match( "%d+" ) );
      local change    = change_list[ stage_num ];
      table.insert( list, "stage_"..change );

    end;

    local update_stars_query =
    {
      query      = "update",
      group      = "player",
      collection = "quests_current",
      indice     = "name",
      criteria   = { like = "qst_collect_star" },
      set        = { list = list }
    };
    lib_db.Execute( update_stars_query );

  end;


end;
--------------------------------------------------------------------------------
function private.Migration_404_client ()

  -- Remove 2nd swordstone fragment
  local find_build_query =
  {
    query      = "find",
    group      = "player",
    collection = "builds",
    indice     = "name",
    criteria   = { like = "puzzle_1" }
  };
  local swordstone_data = lib_db.Execute( find_build_query ).result;

  if (
  swordstone_data and
  swordstone_data.collected_fragments and
  #swordstone_data.collected_fragments > 0 ) then

    local change_list = {
      [1] = false,
      [2] = true, --remove
      [3] = 2,
      [4] = 3,
      [5] = 4,
      [6] = 5
    };

    local new_fragment_list = {};

    for f = 1, #swordstone_data.collected_fragments do

      local fragment_num = tonumber( swordstone_data.collected_fragments[ f ]:match( "%d+" ) );
      local change       = change_list[ fragment_num ];

      if ( not change ) then
        table.insert( new_fragment_list, "swordstone_part_"..fragment_num );
      elseif ( type( change ) == "number" ) then
        table.insert( new_fragment_list, "swordstone_part_"..change );
      end;

    end;

    local update_build_query =
    {
      query      = "update",
      group      = "player",
      collection = "builds",
      indice     = "name",
      criteria   = { like = "puzzle_1" },
      set        = {
        collected_fragments = new_fragment_list
      }
    };
    lib_db.Execute( update_build_query );
    -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
    local quest_name = "qst_collect_swordstone_fragments";

    local find_quest_query =
    {
      query      = "find",
      group      = "player",
      collection = "quests_current",
      indice     = "name",
      criteria   = { like = quest_name }
    };
    local quest_data = lib_db.Execute( find_quest_query ).result;

    if ( quest_data ) then

      local update_quest_query =
      {
        query      = "update",
        group      = "player",
        collection = "quests_current",
        indice     = "name",
        criteria   = { like = quest_name },
        set        = {
          list     = new_fragment_list
        }
      };
      lib_db.Execute( update_quest_query );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Migration_437_client ()

  -- Remove skin bonuses from save
  local query =
  {
    query      = "find_all",
    group      = "player",
    collection = "skins"
  };

  local skins_list = lib_db.Execute( query ).result;

  local change_list = {};

  for s, skin in ipairs( skins_list ) do

    skin.bonus = nil;
    table.insert( change_list, skin );

  end;

  if ( #change_list > 1 ) then

    local query =
    {
      query      = "delete_all",
      group      = "player",
      collection = "skins"
    };

    lib_db.Execute( query );

    for c, change in ipairs( change_list ) do

      query =
      {
        query      = "insert",
        group      = "player",
        collection = "skins",
        value      = change
      };

      lib_db.Execute( query );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Migration_438_client ()

  -- Give skinpart and skin for complete wizard's quest
  local query =
  {
    query      = "find",
    group      = "player",
    collection = "quests_completed",
    indice     = "name",
    criteria   = { like = "qst_collect_gold_key" }
  };

  local quest_completed = lib_db.Execute( query ).result;

  query =
  {
    query      = "find",
    group      = "player",
    collection = "skins",
    indice     = "name",
    criteria   = { like = "justknight" }
  };

  local skin_info = lib_db.Execute( query ).result;

  if ( quest_completed and ( not skin_info )) then

    query =
    {
      query      = "insert",
      group      = "player",
      collection = "skins",
      value      = {
        name  = "justknight",
        parts = { "justknight_gear" }
      }
    };

    lib_db.Execute( query );

  end;

end;
--------------------------------------------------------------------------------
function private.Migration_450_client ()

  --JK: This migration will fix the hives

  -- UPD 1.8.0: This function is empty,
  -- because Migration_573_client removes
  -- hives from the player's save.

end;
--------------------------------------------------------------------------------
function private.Migration_493_client ()

  -- JK: this migration will fix quest completion for mine 6 construction
  -- Condition: if mine 6 is constructed, the quest must be complete
  -- Action: if the quest is still not complete, this migration will force it
  -- to complete

  local quest_name = "qst_construct_mine_6";
  local mine_name  = "mine_6";

  -- 1. Check if quest is complete
  local query =
  {
    query      = "find",
    group      = "player",
    collection = "quests_completed",
    indice     = "name",
    criteria   = { like = quest_name }
  };

  local quest = lib_db.Execute( query ).result;

  -- 2. If quest is complete, finish migration
  if ( quest ) then return; end;

  -- 3. Quest is not complete, check mine_6 state
  query =
  {
    query      = "find",
    group      = "player",
    collection = "mines",
    indice     = "name",
    criteria   = { like = mine_name }
  };

  local mine = lib_db.Execute( query ).result;

  -- 4. If mine_6 is active, force the quest to complete
  if ( mine ) and ( mine.state == "active" ) then

    query =
    {
      query      = "delete",
      group      = "player",
      collection = "quests_current",
      indice     = "name",
      criteria   = { like = quest_name }
    };

    lib_db.Execute( query );

    query =
    {
      query      = "insert",
      group      = "player",
      collection = "quests_completed",
      value      = { name = quest_name }
    };

    lib_db.Execute( query );

  end;

  -- 5. If mine_6 is not active and quest is not complete,
  -- it's OK, doing nothing :)

end;
--------------------------------------------------------------------------------
function private.Migration_494_client ()

  --AL> Give stage_220-quest to guys, who already done start conditions
  local quest_name = "qst_complete_stage_220";

  local saved_quest_query =
  {
    query      = "find",
    group      = "player",
    collection = "quests_current",
    indice     = "name",
    criteria   = { like = quest_name }
  };
  local saved_quest = lib_db.Execute( saved_quest_query ).result;

  if ( saved_quest ) then return; end; -- RET: Quest already in list

  saved_quest_query =
  {
    query      = "find",
    group      = "player",
    collection = "quests_completed",
    indice     = "name",
    criteria   = { like = quest_name }
  };

  saved_quest = lib_db.Execute( saved_quest_query ).result;

  if ( saved_quest ) then return; end; -- RET: Quest already done
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  local quest_data_query =
  {
    query      = "find",
    group      = "static",
    collection = "quests",
    indice     = "name",
    criteria   = { like = quest_name }
  };
  local quest_data = lib_db.Execute( quest_data_query ).result;

  if ( not quest_data ) then return; end;

  if ( quest_data.target.type == "QT_StageCompleted" ) then

    local saved_stage_query =
    {
      query      = "find",
      group      = "player",
      collection = "stages",
      indice     = "name",
      criteria   = { like = quest_data.target.name }
    };

    local saved_stage = lib_db.Execute( saved_stage_query ).result;

    if ( saved_stage and saved_stage.state and ( saved_stage.state > 1 )) then
      return;     -- RET: Target stage already done
    end;

    saved_stage_query =
    {
      query      = "find",
      group      = "player",
      collection = "stages",
      indice     = "name",
      criteria   = { like = quest_data.condition.name }
    };

    saved_stage = lib_db.Execute( saved_stage_query ).result;

    if ( saved_stage and saved_stage.state and ( saved_stage.state > 1 )) then

      local quest_insert_query =
      {
        query      = "insert",
        group      = "player",
        collection = "quests_current",
        value      = { name = quest_name }
      };
      lib_db.Execute( quest_insert_query );

    end;

  end;

end;
--------------------------------------------------------------------------------
-- Change hearts donation count on PC from 1 to 2
--------------------------------------------------------------------------------
function private.Migration_532_client ()

  if ( IsPC() ) then

    local hearts_info = lib_db.Player.Info().hearts;

    local donation     = 2;
    local donation_now = hearts_info.donate;

    local need_fix = false;

    if ( donation_now > 0 and donation_now < donation ) then

      local delta = donation - donation_now;

      hearts_info.donate = donation;
      hearts_info.count  = hearts_info.count + delta;

      need_fix = true;

    end;

    if ( hearts_info.facebook ~= 0 ) then

      hearts_info.count    = hearts_info.count - hearts_info.facebook;
      hearts_info.facebook = 0;

      need_fix = true;

    end;

    if ( need_fix ) then

      local query =
      {
        query      = "update",
        group      = "player",
        collection = "info",
        set        = { hearts = hearts_info }
      };

      lib_db.Execute( query );

    end;

  end;

end;
--------------------------------------------------------------------------------
-- Change sections content downloading.
--------------------------------------------------------------------------------
function private.Migration_553_client ()

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "sections"
  };

  local sections = lib_db.Execute( query ).result;

  local enabled_count = 0;

  Log.i( "Migration_553_client: < "..#sections.." >." );

  for s = 1, #sections do

    local section_name = sections[ s ].name;

    if ( sections[ s ].enabled ) then

      enabled_count = enabled_count + 1;

    end;

  end;

  local result, value = parts.storage.data.Read( "sml/allow_sections" );

  if ( not value ) or ( value == "" ) or ( value == "false" ) or ( value == "true" ) then

    parts.storage.data.Write( "sml/allow_sections", enabled_count );
    Log.i( "Migration_553_client: set allow_sections to < "..enabled_count.." >." );

  end;

end;
--------------------------------------------------------------------------------
-- 1.8.0 dailybonus
--------------------------------------------------------------------------------
function private.Migration_560_client ()

  -- JK: remove keys, add day

  local query = {
    query      = "delete_all",
    collection = "player/dailybonus"
  };

  lib_db.DbExecute( query );

  query = {
    query      = "insert",
    collection = "player/dailybonus",
    value      = {
      day   = 0,
      ready = false
    }
  };

  lib_db.DbExecute( query );

  lib_db.Database_Detach();

  query =
  {
    query      = "update",
    group      = "player",
    collection = "dailybonus",
    set        = {
      day   = 0,
      ready = false,
      keys  = nil,
      first = nil
    }
  };

  lib_db.Execute( query );

  lib_db.Database_Attach();

end;
--------------------------------------------------------------------------------
-- Change completed misc fights saved state
-- 'completed, fight' -> 'not completed, ride'
--------------------------------------------------------------------------------
function private.Migration_572_client ()

  local query =
  {
    query      = "find_all",
    group      = "static",
    collection = "stages",
    indice     = "type",
    criteria   = { like = "misc" }
  };

  local misc_stages = lib_db.Execute( query ).result or {};

  for i = 1, #misc_stages do

    local name = misc_stages[ i ].name;
    query =
    {
      query      = "find",
      group      = "player",
      collection = "stages",
      indice     = "name",
      criteria   = { like = name }
    };

    local saved = lib_db.Execute( query ).result;

    if ( saved and saved.type == "fight" and saved.completed ) then

      query =
      {
        query      = "update",
        group      = "player",
        collection = "stages",
        indice     = "name",
        criteria   = { like = name },
        set        = { type = "ride", completed = false }
      };

      lib_db.Execute( query );

    end;

  end;

end;
--------------------------------------------------------------------------------
-- Drop player hives data
--------------------------------------------------------------------------------
function private.Migration_573_client ()

  local query =
  {
    query      = "find_all",
    group      = "player",
    collection = "hives"
  };

  local hives = lib_db.Execute( query ).result or {};

  for i = 1, #hives do

    local hive_name = hives[ i ].name;
    local token = string.format( "mte_map_cooldown_hives_%s", hive_name );

    if ( TimeIsEventExist( token ) ) then

      TimeFinishEvent();
      TimeUnsheduleEvent( token );

    end;

  end;

  local query = {
    query      = "delete_all",
    collection = "player/hives"
  };

  lib_db.DbExecute( query );

end;
--------------------------------------------------------------------------------
-- 1.8.0 Update Reward
--------------------------------------------------------------------------------
function private.Migration_574_client ()

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        =
    {
      update = true,
      reward =
      {
        { type = "booster", name = "hammer",    count = 1   },
        { type = "gold",    name = "gold",      count = 150 },
        { type = "booster", name = "magicgift", count = 2   }
      }
    }
  };

  lib_db.Execute( query );

end;
--------------------------------------------------------------------------------
-- Change saved skins data: store skinparts count instead of array
--------------------------------------------------------------------------------
function private.Migration_596_client ()

  local query =
  {
    query      = "find_all",
    group      = "player",
    collection = "skins"
  };

  local skins = lib_db.Execute( query ).result or {};

  for i, skin in pairs( skins ) do

    if ( skin.parts ) then

      skin.parts = #skin.parts;

      local query =
      {
        query      = "update",
        group      = "player",
        collection = "skins",
        indice     = "name",
        criteria   = { like = skin.name },
        set        = skin
      };

      lib_db.Execute( query );

    end;

  end;

end;
--------------------------------------------------------------------------------
-- Fix for worker reward at section 1 stage 44.
--------------------------------------------------------------------------------
function private.Migration_600_client ()

  local update_query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        =
    {
      update = false,
      reward = false
    }
  };

  lib_db.Execute( update_query );

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "stages",
    indice     = "name",
    criteria   = { like = "stage_44" }
  };

  local stage = lib_db.Execute( query ).result;

  if ( stage and stage.noreward ) then

    query =
    {
      query      = "find",
      group      = "player",
      collection = "mines",
      indice     = "name",
      criteria   = { like = "mine_2" }
    };

    local mine = lib_db.Execute( query ).result or {};

    query =
    {
      query      = "find",
      group      = "static",
      collection = "mines",
      indice     = "name",
      criteria   = { like = "mine_2" }
    };

    local progress = lib_db.Execute( query ).result.progress;

    if ( mine ) and ( mine.progress ) and ( mine.progress < #progress ) then

      mine.progress = mine.progress + 1;
      if ( mine.progress == #progress ) then
        mine.state = "wait";
      end;

      query =
      {
        query      = "update",
        group      = "player",
        collection = "mines",
        indice     = "name",
        criteria   = { like = "mine_2" },
        set        = mine
      };

      lib_db.Execute( query );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Migration_637_client ()

  -- JK: 1. dailyquest reworked
  local unlock_stage = lib_db.Settings( "dailyquest" ).unlockstage;

  local query = {
    query      = "find",
    group      = "player",
    collection = "stages",
    indice     = "name",
    criteria   = { like = unlock_stage }
  };

  local unlocked = lib_db.Execute( query ).result.completed;

  if ( not unlocked ) then

    if ( parts.time.event.Exists( "mte_dailyquest_refresh" ) ) then

      parts.time.event.Finish();
      parts.time.event.Unshedule( "mte_dailyquest_refresh" );

    end;

    query =
    {
      query      = "update",
      group      = "player",
      collection = "dailyquest",
      set        = {
        ready  = false,
        reward = false
      }
    };

    lib_db.Execute( query );

  else

    -- JK: set all challenges beaten -> state = 2
    query =
    {
      query      = "find",
      group      = "player",
      collection = "dailyquest"
    };

    local dailyquest = lib_db.Execute( query ).result;

    -- JK: Additional check for dailyquest data
    if ( dailyquest            ) and
       ( dailyquest.challenge1 ) then

      for c = 1, 4 do

        dailyquest[ "challenge"..c ].state = 2;

      end;

    -- JK: 1.2.0 has no dailyquest, but unlock stage may be complete
    -- Set fake challenge data
    else

      dailyquest = {
        challenge1 = {
          enemy = {
            attack = 11,
            health = 155,
            name   = "bandit",
          },
          stage = "stage_1_daily",
          state = 2
        },
        challenge2 = {
          enemy = {
            attack = 10,
            health = 290,
            name   = "pirate",
          },
          stage = "stage_2_daily",
          state = 2
        },
        challenge3 = {
          enemy = {
            attack = 13,
            health = 281,
            name   = "yeti",
          },
          stage = "stage_3_daily",
          state = 2
        },
        challenge4 = {
          enemy = {
            attack = 12,
            health = 176,
            name   = "spider",
          },
          stage = "stage_4_daily",
          state = 2
        }
      };

    end;

    query =
    {
      query      = "update",
      group      = "player",
      collection = "dailyquest",
      set        = {
        challenge1 = dailyquest.challenge1,
        challenge2 = dailyquest.challenge2,
        challenge3 = dailyquest.challenge3,
        challenge4 = dailyquest.challenge4,
        ready      = false,
        reward     = false
      }
    };

    lib_db.Execute( query );

  end;

  if ( parts.time.event.Exists( "mte_dailyquest_countdown_refresh" ) ) then

    parts.time.event.Finish();
    parts.time.event.Unshedule( "mte_dailyquest_countdown_refresh" );

  end;

  if ( parts.time.event.Exists( "mte_dailyquest_challenge_refresh" ) ) then

    parts.time.event.Finish();
    parts.time.event.Unshedule( "mte_dailyquest_challenge_refresh" );

  end;

  query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        =
    {
      update = { type = "replay" },
      reward = false
    }
  };

  lib_db.Execute( query );

end;
--------------------------------------------------------------------------------
-- Set default player name to false (need for localized display)
--------------------------------------------------------------------------------
function private.Migration_763_client ()

  local name = lib_db.Player.Info().name;

  if ( name and ( name == "Player" ) ) then

    local query =
    {
      query      = "update",
      group      = "player",
      collection = "info",
      set        = { name = false }
    };

    lib_db.Execute( query );

  end;

end
--------------------------------------------------------------------------------
function private.Migration_777_client ()

  -- JK: Clear WinSameStage achievement

  local query =
  {
    query      = "delete",
    group      = "player",
    collection = "achievements",
    indice     = "name",
    criteria   = {
      like = "favourite_level"
    }
  };

  lib_db.Execute( query );

end;
--------------------------------------------------------------------------------
-- Add to save missing in 1.10 character 'fisher'
--------------------------------------------------------------------------------
function private.Migration_918_client ()

  -- Get character
  local character_name  = "fisher";
  local query =
  {
    query      = "find",
    group      = "static",
    collection = "characters",
    indice     = "name",
    criteria   = { like = character_name }
  };
  local character_static = lib_db.Execute( query ).result;

  if ( not character_static ) then return; end;

  -- Get unlock build
  query =
  {
    query      = "find",
    group      = "player",
    collection = "builds",
    indice     = "name",
    criteria   = { like = character_static.unlock }
  };
  local build_saved = lib_db.Execute( query ).result;

  if ( not build_saved             ) then return; end;
  if ( not build_saved.constructed ) then return; end;

  -- Get character quest state
  local qst_curr_query =
  {
    query      = "find",
    group      = "player",
    collection = "quests_current",
    indice     = "name",
    criteria   = { like = character_static.quest_free }
  };

  local qst_done_query =
  {
    query      = "find",
    group      = "player",
    collection = "quests_completed",
    indice     = "name",
    criteria   = { like = character_static.quest_free }
  };

  local quest_state = "inactive";

  if ( lib_db.Execute( qst_curr_query ).result ) then

    quest_state = "current";

  elseif ( lib_db.Execute( qst_done_query ).result ) then

    quest_state = "completed";

  end;

  -- Define character state
  local qst_to_char_state =
  {
    inactive  = "unlocked",
    current   = "active",
    completed = "inactive"
  };
  local character_state = qst_to_char_state[ quest_state ];

  -- Save character state
  query =
  {
    query      = "update",
    group      = "player",
    collection = "characters",
    indice     = "name",
    criteria   = { like = character_name },
    set        =
    {
      state = character_state
    }
  };

  if ( not lib_db.Execute( query ).result ) then

    query =
    {
      query      = "insert",
      group      = "player",
      collection = "characters",
      value      =
      {
        name  = character_name,
        state = character_state
      }
    };

    lib_db.Execute( query );

  end;

end;
--------------------------------------------------------------------------------
-- Fix levels in 'qst_collect_coins_50'
--------------------------------------------------------------------------------
function private.Migration_931_client ()

  local quest_name = "qst_collect_coins_50";

  local function get_saved_stage_data ( stage_name )

    local query =
    {
      query      = "find",
      group      = "player",
      collection = "stages",
      indice     = "name",
      criteria   = { like = stage_name }
    };
    return lib_db.Execute( query ).result;

  end;

  -- Get quest data
  local query =
  {
    query      = "find",
    group      = "player",
    collection = "quests_current",
    indice     = "name",
    criteria   = { like = quest_name }
  };
  local quest_current = lib_db.Execute( query ).result;

  query =
  {
    query      = "find",
    group      = "static",
    collection = "quests",
    indice     = "name",
    criteria   = { like = quest_name }
  };
  local quest_data = lib_db.Execute( query ).result;

  if ( quest_current ) then

    -- Drop quest
    query =
    {
      query      = "delete",
      group      = "player",
      collection = "quests_current",
      indice     = "name",
      criteria   = { like = quest_name }
    };
    lib_db.Execute( query );

    -- Just drop quest_current if static data does not exists
    if ( not quest_data ) then
      return;
    end;

    -- Just drop quest_current if start condition not completed
    local start_stage = get_saved_stage_data( quest_data.condition.name );
    if not ( start_stage and start_stage.completed ) then
      return;
    end;

    -- Fix completed stages list
    local valid_list = {};
    local dummy_data = { noreward = false, state = 1 };

    for _, stage_name in pairs( quest_data.target.list ) do

      local stage_data = get_saved_stage_data( stage_name ) or dummy_data;

      if ( stage_data.state >= 3 and stage_data.noreward ) then

        table.insert( valid_list, stage_name );

      end;

    end;

    if ( #valid_list >= #quest_data.target.list ) then

      query =
      {
        query      = "insert",
        group      = "player",
        collection = "quests_completed",
        value      = { name = quest_name }
      };

      lib_db.Execute( query );

    else

      if ( #valid_list > 0 ) then
        quest_current.list = valid_list;
      else
        quest_current.list = nil;
      end;

      -- Insert correct quest_current
      query =
      {
        query      = "insert",
        group      = "player",
        collection = "quests_current",
        value      = quest_current
      };
      lib_db.Execute( query );

    end;

  end;

end;
--------------------------------------------------------------------------------
-- Add 'section' and 'type' fields to 'player/quests_current'
--------------------------------------------------------------------------------
function private.Migration_953_client ()

  local query =
  {
    query      = "find_all",
    group      = "player",
    collection = "quests_current"
  };

  local quests = lib_db.Execute( query ).result or {};

  for _, quest in pairs( quests ) do

    query =
    {
      query      = "find",
      group      = "static",
      collection = "quests",
      indice     = "name",
      criteria   = { like = quest.name }
    };
    local static_data = lib_db.Execute( query ).result;

    if ( static_data ) then

      quest.section = static_data.section;
      quest.type    = static_data.target.type;

      query =
      {
        query      = "update",
        group      = "player",
        collection = "quests_current",
        indice     = "name",
        criteria   = { like = quest.name },
        set        = quest
      };
      lib_db.Execute( query );

    end;

  end;

end;
--------------------------------------------------------------------------------
-- Remove quest 'qst_collect_gold_key' from currents until start condition
--------------------------------------------------------------------------------
function private.Migration_954_client ()

  local quest_name = "qst_collect_gold_key";

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "quests_current",
    indice     = "name",
    criteria   = { like = quest_name }
  };
  local current_data = lib_db.Execute( query ).result;

  if ( not current_data ) then return; end;

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "quests",
    indice     = "name",
    criteria   = { like = quest_name }
  };
  local static_data = lib_db.Execute( query ).result;

  if ( not static_data ) then return; end;

  local stage_name = static_data.condition.name;

  query =
  {
    query      = "find",
    group      = "player",
    collection = "stages",
    indice     = "name",
    criteria   = { like = stage_name }
  };

  local stage_data = lib_db.Execute( query ).result or {};

  if not ( stage_data.completed ) then

    query =
    {
      query      = "delete",
      group      = "player",
      collection = "quests_current",
      indice     = "name",
      criteria   = { like = quest_name }
    };
    lib_db.Execute( query );

    query =
    {
      query      = "update",
      group      = "player",
      collection = "characters",
      indice     = "name",
      criteria   = { like  = "wizard" },
      set        = { state = "unlocked" }
    };
    lib_db.Execute( query );

  end;

end;
--------------------------------------------------------------------------------
-- Fix mines state 'wait'.
--------------------------------------------------------------------------------
function private.Migration_984_client ()

  local query =
  {
    query      = "find_all",
    group      = "player",
    collection = "mines"
  };

  local mines = lib_db.Execute( query ).result;

  for m = 1, #mines do

    local mine = mines[ m ];

    if ( mine.resource and mine.state == "wait" ) then

      mine.state = "active";

      query =
      {
        query      = "update",
        group      = "player",
        collection = "mines",
        indice     = "name",
        criteria   = { like = mine.name },
        set        = mine
      };

      lib_db.Execute( query );

    end;

  end;

end;
--------------------------------------------------------------------------------
-- Fix quests type FragmentCollected not complete.
--------------------------------------------------------------------------------
function private.Migration_1106_client ()

  local list = {
    { quest_name = "qst_collect_stonegate_fragments",  build_name = "island_1" },
    { quest_name = "qst_collect_swordstone_fragments", build_name = "puzzle_1" },
    { quest_name = "qst_collect_saw",                  build_name = "island_3" },
    { quest_name = "qst_collect_pick",                 build_name = "island_4" },
    { quest_name = "qst_collect_sickle",               build_name = "island_5" }
  };

  local func_get_data_quest = function ( quest_name, build_name )

    local data = {};

    local query =
    {
      query      = "find",
      group      = "player",
      collection = "quests_current",
      indice     = "name",
      criteria   = { like = quest_name }
    };

    data.sv_quest = lib_db.Execute( query ).result;

    query =
    {
      query      = "find",
      group      = "player",
      collection = "quests_completed",
      indice     = "name",
      criteria   = { like = quest_name }
    };

    data.sv_completed_quest = lib_db.Execute( query ).result;

    query =
    {
      query      = "find",
      group      = "player",
      collection = "builds",
      indice     = "name",
      criteria   = { like = build_name }
    };

    data.sv_builds = lib_db.Execute( query ).result;

    query =
    {
      query      = "find",
      group      = "static",
      collection = "builds",
      indice     = "name",
      criteria   = { like = build_name }
    };

    data.db_builds = lib_db.Execute( query ).result;

    return data;

  end;

  local func_fix_quest = function ( quest_name )

    local query =
    {
      query      = "delete",
      group      = "player",
      collection = "quests_current",
      indice     = "name",
      criteria   = { like = quest_name }
    };

    lib_db.Execute( query );

    query =
    {
      query      = "insert",
      group      = "player",
      collection = "quests_completed",
      value      = { name = quest_name }
    };

    lib_db.Execute( query );

  end;

  local func_update_quest = function ( quest_name, list )

    local query =
    {
      query      = "update",
      group      = "player",
      collection = "quests_current",
      indice     = "name",
      criteria   = { like = quest_name },
      set        = { list = list }
    };

    lib_db.Execute( query );

  end;

  for i = 1, #list do

    local quest_name = list[ i ].quest_name;
    local build_name = list[ i ].build_name;

    local data = func_get_data_quest( quest_name, build_name );

    if ( not data.sv_completed_quest ) and
       ( data.sv_quest  ) and
       ( data.db_builds ) and
       ( data.sv_builds ) then

       if ( #data.db_builds.fragments == #data.sv_builds.collected_fragments ) then

         func_fix_quest( quest_name );

       elseif ( #data.db_builds.fragments ~= 1 ) then

         func_update_quest( quest_name, data.sv_builds.collected_fragments );

       end;

    end;

  end;

end;
--------------------------------------------------------------------------------
-- Fix quests for fast saga.
--------------------------------------------------------------------------------
function private.Migration_1264_client ()

  local func_get_saved_quest = function ( quest_name )

    local query =
    {
      query      = "find",
      group      = "player",
      collection = "quests_current",
      indice     = "name",
      criteria   = { like = quest_name }
    };

    return lib_db.Execute( query ).result;

  end;

  local func_update_quest = function ( quest_name, list )

    local query =
    {
      query      = "update",
      group      = "player",
      collection = "quests_current",
      indice     = "name",
      criteria   = { like = quest_name },
      set        = { list = list }
    };

    lib_db.Execute( query );

  end;

  local func_fix_key = function ( quest_name, list )

    local query =
    {
      query      = "find",
      group      = "static",
      collection = "quests",
      indice     = "name",
      criteria   = { like = quest_name }
    };
    local static_data = lib_db.Execute( query ).result;

    if ( static_data ) then

      local stage_name = static_data.condition.name;

      query =
      {
        query      = "find",
        group      = "player",
        collection = "stages",
        indice     = "name",
        criteria   = { like = stage_name }
      };

      local stage_data = lib_db.Execute( query ).result or {};

      if not ( stage_data.completed ) then

        query =
        {
          query      = "delete",
          group      = "player",
          collection = "quests_current",
          indice     = "name",
          criteria   = { like = quest_name }
        };
        lib_db.Execute( query );

        query =
        {
          query      = "update",
          group      = "player",
          collection = "characters",
          indice     = "name",
          criteria   = { like  = "wizard" },
          set        = { state = "unlocked" }
        };
        lib_db.Execute( query );

      else

        func_update_quest( quest_name, false );

      end;

    end;

  end;

  local func_fix_swordstone = function ( quest_name, list )

    if ( list ) then

      local new_list = {};

      for i = 1, #list do

        local fragment = tonumber( list[ i ]:match( "%d+" ) );

        if ( fragment > 1 ) then

          table.insert( new_list, "swordstone_part_"..( fragment - 1 ) );

        end;

      end;

      if ( #new_list > 0 ) then

        func_update_quest( quest_name, new_list );

      else

        func_update_quest( quest_name, false );

      end;

    end;

  end;

  local quests = {
    { name = "qst_collect_gold_key",             func = func_fix_key        },
    { name = "qst_collect_swordstone_fragments", func = func_fix_swordstone }
  };

  for i = 1, #quests do

    local quest = quests[ i ];
    local quest_data = func_get_saved_quest( quest.name );

    if ( quest_data ) then

      quest.func( quest.name, quest_data.list );

    end;

  end;

end;
--------------------------------------------------------------------------------
-- Fix starterpacks names in save: starterpack_1_ios -> starterpack_1,
-- delete not use watch
--------------------------------------------------------------------------------
function private.Migration_1271_client ()

  local query =
  {
    query      = "find_all",
    group      = "player",
    collection = "starterpacks"
  };

  local starterpacks = lib_db.Execute( query ).result;

  for _, starterpack in pairs( starterpacks ) do

    local starterpack_number = string.match( starterpack.name, "%d+" );
    local crossplatform_name = "starterpack_"..starterpack_number;

    if ( starterpack.name ~= crossplatform_name ) then

      query =
      {
        query      = "update",
        group      = "player",
        collection = "starterpacks",
        indice     = "name",
        criteria   = { like = starterpack.name },
        set      =
        {
          name = crossplatform_name
        }
      };

      lib_db.Execute( query );

    end;

  end;

  if ( parts.time.watch.Exists( "wtc_hearts_blocker" ) ) then
    parts.time.watch.Stop( "wtc_hearts_blocker" );
  end;

end;
--------------------------------------------------------------------------------
-- Fix swordstone: construct if all fragments collected.
--------------------------------------------------------------------------------
function private.Migration_1312_client ()

  local build_name = "puzzle_1";

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "builds",
    indice     = "name",
    criteria   = { like = build_name }
  };

  local save_data = lib_db.Execute( query ).result;

  query =
  {
    query      = "find",
    group      = "static",
    collection = "builds",
    indice     = "name",
    criteria   = { like = build_name }
  };

  local static_data = lib_db.Execute( query ).result;

  if ( save_data and
       save_data.unlocked and
       not save_data.constructed and
       #( save_data.collected_fragments or {} ) == #static_data.fragments ) then

    -- Constuct build.
    query =
    {
      query      = "update",
      group      = "player",
      collection = "builds",
      indice     = "name",
      criteria   = { like = build_name },
      set      =
      {
        constructed = true
      }
    };

    lib_db.Execute( query );

    -- Complete quest qst_collect_swordstone_fragments.
    local quest_name = "qst_collect_swordstone_fragments";
    query =
    {
      query      = "find",
      group      = "player",
      collection = "quests_current",
      indice     = "name",
      criteria   = { like = quest_name }
    };
    local quest_current = lib_db.Execute( query ).result;

    if ( quest_current ) then

      local quest_remove_query =
      {
        query      = "delete",
        group      = "player",
        collection = "quests_current",
        indice     = "name",
        criteria   = { like = quest_name }
      };
      lib_db.Execute( quest_remove_query );

      local quest_insert_query =
      {
        query      = "insert",
        group      = "player",
        collection = "quests_completed",
        value      = { name = quest_name }
      };
      lib_db.Execute( quest_insert_query );

    end;

    -- Enable next stage.
    local stage_name = "stage_"..static_data.unlocklist.min;
    query =
    {
      query      = "update",
      group      = "player",
      collection = "stages",
      indice     = "name",
      criteria   = { like = stage_name },
      set        =
      {
        enabled = true,
        visible = true
      }
    };

    local result = lib_db.Execute( query ).result;

    if ( not result ) then

      query =
      {
        query      = "insert",
        group      = "player",
        collection = "stages",
        value      =
        {
          name    = stage_name,
          enabled = true,
          visible = true
        }
      };

      result = lib_db.Execute( query ).result;

    end;

    -- Collect skinpart.
    local skin_name = string.match( static_data.reward.name, "%w+" );
    query =
    {
      query      = "update",
      group      = "player",
      collection = "skins",
      indice     = "name",
      criteria   = { like = skin_name },
      set        =
      {
        parts    = static_data.reward.count
      }
    };

    result = lib_db.Execute( query ).result;

    if ( not result ) then

      query =
      {
        query      = "insert",
        group      = "player",
        collection = "skins",
        value      =
        {
          name    = skin_name,
          parts   = static_data.reward.count
        }
      };

      result = lib_db.Execute( query ).result;

    end;

    -- Set skin.
    query =
    {
      query      = "update",
      group      = "player",
      collection = "hero",
      set        = { skin = skin_name }
    };

    result = lib_db.Execute( query ).result;

  end;

end;
--------------------------------------------------------------------------------
function private.Migration_1313_client ()

  -- JK: this migration removes obsolete state flag from player/fortunewheel

  -- JK: get current data
  local query =
  {
    query      = "find",
    group      = "player",
    collection = "fortunewheel"
  };

  local data = lib_db.Execute( query ).result;

  -- JK: clear the state field
  data.state = nil;

  -- JK: remove all that we have
  query =
  {
    query      = "delete_all",
    group      = "player",
    collection = "fortunewheel"
  };

  lib_db.Execute( query );

  -- JK: fill with new data
  lib_db.Rewrite( "player", "fortunewheel", { data } );

end;
--------------------------------------------------------------------------------
-- Fix for previous migration.
--------------------------------------------------------------------------------
function private.Migration_1314_client ()

  local build_name  = "build_9";
  local puzzle_name = "puzzle_1";

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "builds",
    indice     = "name",
    criteria   = { like = puzzle_name }
  };

  local static_data = lib_db.Execute( query ).result;

  query =
  {
    query      = "find",
    group      = "player",
    collection = "builds",
    indice     = "name",
    criteria   = { like = puzzle_name }
  };

  local puzzle_data = lib_db.Execute( query ).result;

  local fragments_list = {};
  local quests_list = {
    "qst_collect_silver_key",
    "qst_collect_red_key",
    "qst_collect_blue_key",
    "qst_collect_pink_key"
  };

  for i = 1, #quests_list do

    query =
    {
      query      = "find",
      group      = "player",
      collection = "quests_completed",
      indice     = "name",
      criteria   = { like = quests_list[ i ] }
    };
    local completed = lib_db.Execute( query ).result;

    if ( completed ) then
      table.insert( fragments_list, "swordstone_part_"..i );
    end;

  end;

  fragments_list = tern( #fragments_list > 0 )( fragments_list, false );

  if ( puzzle_data ) then

    query =
    {
      query      = "update",
      group      = "player",
      collection = "builds",
      indice     = "name",
      criteria   = { like = puzzle_name },
      set      =
      {
        collected_fragments = fragments_list
      }
    };

    lib_db.Execute( query );

    query =
    {
      query      = "find",
      group      = "player",
      collection = "stages",
      indice     = "name",
      criteria   = { like = static_data.unlock }
    };

    local stage_data = lib_db.Execute( query ).result;

    if ( puzzle_data.constructed and
         #fragments_list ~= #static_data.fragments ) then

      -- Fix quest qst_collect_swordstone_fragments.
      local quest_name = "qst_collect_swordstone_fragments";
      query =
      {
        query      = "find",
        group      = "player",
        collection = "quests_completed",
        indice     = "name",
        criteria   = { like = quest_name }
      };
      local quest_completed = lib_db.Execute( query ).result;

      if ( quest_completed ) then

        local quest_remove_query =
        {
          query      = "delete",
          group      = "player",
          collection = "quests_completed",
          indice     = "name",
          criteria   = { like = quest_name }
        };
        lib_db.Execute( quest_remove_query );

        local quest_insert_query =
        {
          query      = "insert",
          group      = "player",
          collection = "quests_current",
          value      = { list = fragments_list,
                         name = quest_name,
                         section = "section_1",
                         type = "QT_FragmentCollected"
          }
        };
        lib_db.Execute( quest_insert_query );

      end;

      -- Fix qst_collect_stonegate_fragments.
      quest_name = "qst_collect_stonegate_fragments";
      query =
      {
        query      = "find",
        group      = "player",
        collection = "quests_current",
        indice     = "name",
        criteria   = { like = quest_name }
      };
      local quest_data = lib_db.Execute( query ).result;

      if ( quest_data ) then

        for i = #quest_data.list, 1, - 1 do

          if ( quest_data.list[ i ] == "stonegate_dragoneye" ) then
            table.remove( quest_data.list, i );
            break;
          end;

        end;

        quest_data.list = tern( #quest_data.list > 0 )( quest_data.list, false );

        local quest_update_query =
        {
          query      = "update",
          group      = "player",
          collection = "quests_current",
          indice     = "name",
          criteria   = { like = quest_name },
          set        = { list = quest_data.list }
        };
        lib_db.Execute( quest_update_query );

        local island_update_query =
        {
          query      = "update",
          group      = "player",
          collection = "builds",
          indice     = "name",
          criteria   = { like = "island_1" },
          set      =
          {
            collected_fragments = quest_data.list
          }
        };
        lib_db.Execute( island_update_query );

      end;

      quest_name = "qst_complete_stage_110";
      query =
      {
        query      = "find",
        group      = "player",
        collection = "quests_completed",
        indice     = "name",
        criteria   = { like = quest_name }
      };
      local quest_completed = lib_db.Execute( query ).result;

      if ( quest_completed ) then

        local quest_remove_query =
        {
          query      = "delete",
          group      = "player",
          collection = "quests_completed",
          indice     = "name",
          criteria   = { like = quest_name }
        };
        lib_db.Execute( quest_remove_query );

        local quest_insert_query =
        {
          query      = "insert",
          group      = "player",
          collection = "quests_current",
          value      = { name = quest_name,
                         section = "section_1",
                         type = "QT_StageCompleted"
          }
        };
        lib_db.Execute( quest_insert_query );

      end;

      local quest_remove_query =
      {
        query      = "delete",
        group      = "player",
        collection = "quests_completed",
        indice     = "name",
        criteria   = { like = "qst_collect_coins_10" }
      };
      lib_db.Execute( quest_remove_query );
      quest_remove_query =
      {
        query      = "delete",
        group      = "player",
        collection = "quests_current",
        indice     = "name",
        criteria   = { like = "qst_collect_coins_10" }
      };
      lib_db.Execute( quest_remove_query );

      -- Fix build.
      query =
      {
        query      = "update",
        group      = "player",
        collection = "builds",
        indice     = "name",
        criteria   = { like = puzzle_name },
        set      =
        {
          constructed = false
        }
      };

      lib_db.Execute( query );

      quest_name = "qst_collect_diamond";
      query =
      {
        query      = "find",
        group      = "player",
        collection = "quests_current",
        indice     = "name",
        criteria   = { like = quest_name }
      };
      local quest_data = lib_db.Execute( query ).result;

      -- Delete stages.
      for i = static_data.unlocklist.min, static_data.unlocklist.max - 1 do
        local stage_name = "stage_"..i;
        query =
        {
          query      = "delete",
          group      = "player",
          collection = "stages",
          indice     = "name",
          criteria   = { like = stage_name }
        };
        lib_db.Execute( query );
        if ( quest_data and quest_data.list ) then
          for i = #quest_data.list, 1, - 1 do
            if ( quest_data.list[ i ] == stage_name ) then
              table.remove( quest_data.list, i );
            end;
          end;
        end;
      end;

      if ( quest_data and quest_data.list ) then

        quest_data.list = tern( #quest_data.list > 0 )( quest_data.list, false );

        local quest_update_query =
        {
          query      = "update",
          group      = "player",
          collection = "quests_current",
          indice     = "name",
          criteria   = { like = quest_name },
          set        = { list = quest_data.list }
        };
        lib_db.Execute( quest_update_query );

      end;

      query =
      {
        query      = "update",
        group      = "player",
        collection = "stages",
        indice     = "name",
        criteria   = { like = "stage_"..static_data.unlocklist.max },
        set        =
        {
          completed = false,
          enabled   = false,
          noreward  = false,
          score     = 0
        }
      };

      lib_db.Execute( query );

      -- Remove skinpart.
      local skin_name = string.match( static_data.reward.name, "%w+" );
      query =
      {
        query      = "update",
        group      = "player",
        collection = "skins",
        indice     = "name",
        criteria   = { like = skin_name },
        set        =
        {
          parts    = 0
        }
      };

      lib_db.Execute( query );

      -- Set skin.
      query =
      {
        query      = "update",
        group      = "player",
        collection = "hero",
        set        = { skin = "default" }
      };

      lib_db.Execute( query );

    end;

  end;

end;
--------------------------------------------------------------------------------
-- fix ach fireman
--------------------------------------------------------------------------------
function private.Migration_1467_client ()

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "achievements",
    indice     = "name",
    criteria   = { like = "add_friends_fireman" }
  };

  local result = lib_db.Execute( query ).result;

  if ( result ) then

    query =
    {
      query      = "find",
      group      = "static",
      collection = "achievements",
      indice     = "name",
      criteria   = { like = "add_friends_fireman" }
    };

    local levels = lib_db.Execute( query ).result.levels;

    if ( result.count == 0 ) and ( result.level > 3 ) then

      query =
      {
        query      = "update",
        group      = "player",
        collection = "achievements",
        indice     = "name",
        criteria   = { like = "add_friends_fireman" },
        set        =
        {
          count = levels[ 1 ].value + levels[ 2 ].value + levels[ 3 ].value
        }
      };

      lib_db.Execute( query );

    end;

    if ( #result.reward == 4 ) then
      return;
    end;

    table.insert( result.reward, { get = false, done = false } );

    query =
    {
      query      = "update",
      group      = "player",
      collection = "achievements",
      indice     = "name",
      criteria   = { like = "add_friends_fireman" },
      set        =
      {
        reward = result.reward
      }
    };

    lib_db.Execute( query );

  end;

end;
--------------------------------------------------------------------------------
-- fix change quests 1 world
--------------------------------------------------------------------------------
function private.Migration_1471_client ()
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  local func_quest_current_sv_data = function ( quest_name )

    local query =
    {
      query      = "find",
      group      = "player",
      collection = "quests_current",
      indice     = "name",
      criteria   = { like = quest_name }
    };
    return lib_db.Execute( query ).result;

  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  local func_quest_completed_sv_data = function ( quest_name )

    local query =
    {
      query      = "find",
      group      = "player",
      collection = "quests_completed",
      indice     = "name",
      criteria   = { like = quest_name }
    };
    return lib_db.Execute( query ).result;

  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  local func_quest_current_delete = function ( quest_name )

    local quest_current = func_quest_current_sv_data( quest_name );

    if ( quest_current ) then

      query =
      {
        query      = "delete",
        group      = "player",
        collection = "quests_current",
        indice     = "name",
        criteria   = { like = quest_name }
      };
      lib_db.Execute( query );

    end;

    return quest_current;

  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  local func_quest_completed_delete = function ( quest_name )

    local quest_completed = func_quest_completed_sv_data( quest_name );

    if ( quest_completed ) then

      query =
      {
        query      = "delete",
        group      = "player",
        collection = "quests_completed",
        indice     = "name",
        criteria   = { like = quest_name }
      };
      lib_db.Execute( query );

    end;

  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  local func_quest_db_data = function ( quest_name )

    local query =
    {
      query      = "find",
      group      = "static",
      collection = "quests",
      indice     = "name",
      criteria   = { like = quest_name }
    };
    return lib_db.Execute( query ).result;

  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  local func_stage_sv_data = function ( stage )

    local query =
    {
      query      = "find",
      group      = "player",
      collection = "stages",
      indice     = "name",
      criteria   = { like = stage }
    };
    return lib_db.Execute( query ).result;

  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  local func_quest_check = function ( quest_name )

    --name
    --section
    --type
    --list

    local db_data = func_quest_db_data( quest_name );

    local sv_data = func_quest_current_delete( quest_name );

    local list = db_data.target.list;
    local item = db_data.target.name;
    local completed_list = {};

    if ( sv_data and sv_data.list ) then

      for i = 1, #list do
        for j = 1, #sv_data.list do
          if ( list[ i ] == sv_data.list[ j ] ) then
            table.insert( completed_list, list[ i ] );
            break;
          end;
        end;
      end;

    end;

    if ( #completed_list == #list ) then

      local quest_insert_query =
      {
        query      = "insert",
        group      = "player",
        collection = "quests_completed",
        value      = {
          name = quest_name
        }
      };
      lib_db.Execute( quest_insert_query );

    else

      local set = {
        name = quest_name,
        section = "section_1",
        type = "QT_StageItemsCollected"
      };

      if ( #completed_list > 0 ) then
        set.list = completed_list;
      end;

      local quest_insert_query =
      {
        query      = "insert",
        group      = "player",
        collection = "quests_current",
        value      = set
      };
      lib_db.Execute( quest_insert_query );

    end;

  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  local func_character_active = function ( character_name )

    local query =
    {
      query      = "update",
      group      = "player",
      collection = "characters",
      indice     = "name",
      criteria   = { like = character_name },
      set        =
      {
        state = "active"
      }
    };
    lib_db.Execute( query );

  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  local func_character_complete = function ( character_name )

    local query =
    {
      query      = "update",
      group      = "player",
      collection = "characters",
      indice     = "name",
      criteria   = { like = character_name },
      set        =
      {
        state = "inactive"
      }
    };
    lib_db.Execute( query );

  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  -- first character
  func_quest_current_delete( "qst_collect_star" );
  func_quest_completed_delete( "qst_collect_star" );

  local completed_sv_data = func_quest_completed_sv_data( "qst_collect_gold_key" );

  if ( completed_sv_data ) then
    func_character_complete( "wizard" );
  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  func_quest_current_delete( "qst_collect_stonegate_fragments" );
  func_quest_completed_delete( "qst_collect_stonegate_fragments" );
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  -- second character
  completed_sv_data = func_quest_completed_sv_data( "qst_collect_silver_key" );

  if ( not completed_sv_data ) then
    local db_data = func_quest_db_data( "qst_collect_silver_key" );
    local stage = db_data.condition.name;
    local sv_stage_data = func_stage_sv_data( stage );

    if ( sv_stage_data and ( ( sv_stage_data.state or 1 ) >= 2 ) ) then
      func_quest_check( "qst_collect_silver_key" );
      func_character_active( "fairy" );
    else
      func_quest_current_delete( "qst_collect_silver_key" );
    end;
  end;

  completed_sv_data = func_quest_completed_sv_data( "qst_collect_silver_key" );

  if ( completed_sv_data ) then
    completed_sv_data = func_quest_completed_sv_data( "qst_collect_flower" );
    if ( not completed_sv_data ) then
      func_quest_check( "qst_collect_flower" );
    end;

    completed_sv_data = func_quest_completed_sv_data( "qst_collect_flower" );
    if ( completed_sv_data ) then
      func_character_complete( "fairy" );
    end;
  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  -- third character
  completed_sv_data = func_quest_completed_sv_data( "qst_collect_red_key" );

  if ( not completed_sv_data ) then
    local db_data = func_quest_db_data( "qst_collect_red_key" );
    local stage = db_data.condition.name;
    local sv_stage_data = func_stage_sv_data( stage );

    if ( sv_stage_data and ( ( sv_stage_data.state or 1 ) >= 2 ) ) then
      func_quest_check( "qst_collect_red_key" );
      func_character_active( "mermaid" );
    else
      func_quest_current_delete( "qst_collect_red_key" );
    end;
  end;

  completed_sv_data = func_quest_completed_sv_data( "qst_collect_red_key" );

  if ( completed_sv_data ) then
    completed_sv_data = func_quest_completed_sv_data( "qst_collect_shell" );
    if ( not completed_sv_data ) then
      func_quest_check( "qst_collect_shell" );
    end;

    completed_sv_data = func_quest_completed_sv_data( "qst_collect_shell" );
    if ( completed_sv_data ) then
      func_character_complete( "mermaid" );
    end;
  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  -- fourth character
  completed_sv_data = func_quest_completed_sv_data( "qst_collect_blue_key" );

  if ( not completed_sv_data ) then
    local db_data = func_quest_db_data( "qst_collect_blue_key" );
    local stage = db_data.condition.name;
    local sv_stage_data = func_stage_sv_data( stage );

    if ( sv_stage_data and ( ( sv_stage_data.state or 1 ) >= 2 ) ) then
      func_quest_check( "qst_collect_blue_key" );
      func_character_active( "sturdy" );
    else
      func_quest_current_delete( "qst_collect_blue_key" );
    end;
  end;

  completed_sv_data = func_quest_completed_sv_data( "qst_collect_blue_key" );

  if ( completed_sv_data ) then
    completed_sv_data = func_quest_completed_sv_data( "qst_collect_eskimo" );
    if ( not completed_sv_data ) then
      func_quest_check( "qst_collect_eskimo" );
    end;

    completed_sv_data = func_quest_completed_sv_data( "qst_collect_eskimo" );
    if ( completed_sv_data ) then
      func_character_complete( "sturdy" );
    end;
  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  -- firth character
  completed_sv_data = func_quest_completed_sv_data( "qst_collect_pink_key" );

  if ( not completed_sv_data ) then
    local db_data = func_quest_db_data( "qst_collect_pink_key" );
    local stage = db_data.condition.name;
    local sv_stage_data = func_stage_sv_data( stage );

    if ( sv_stage_data and ( ( sv_stage_data.state or 1 ) >= 2 ) ) then
      func_quest_check( "qst_collect_pink_key" );
      func_character_active( "princess" );
    else
      func_quest_current_delete( "qst_collect_pink_key" );
    end;
  end;

  completed_sv_data = func_quest_completed_sv_data( "qst_collect_pink_key" );

  if ( completed_sv_data ) then
    func_character_complete( "princess" );
  end;

  func_quest_current_delete( "qst_collect_diamond" );
  func_quest_completed_delete( "qst_collect_diamond" );
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  --build island 1
  local query_build_sv_data =
  {
    query      = "find",
    group      = "player",
    collection = "builds",
    indice     = "name",
    criteria   = { like = "island_1" }
  };

  local build_sv_data = lib_db.Execute( query_build_sv_data ).result;

  if ( build_sv_data and ( not build_sv_data.constructed ) ) then

    completed_sv_data = func_quest_completed_sv_data( "qst_complete_stage_110" );

    local query =
    {
      query      = "delete",
      group      = "player",
      collection = "builds",
      indice     = "name",
      criteria   = { like = "island_1" }
    };
    lib_db.Execute( query );

    local set_data = {
      name = "island_1",
      unlocked = true
    };

    if ( completed_sv_data ) then
      set_data.constructed = true;
      set_data.collected_fragments = { "stonegate_dragoneye" };
    end;

    query =
    {
      query      = "insert",
      group      = "player",
      collection = "builds",
      value      = set_data
    };

    lib_db.Execute( query );

    local stage_sv_data = func_stage_sv_data( "stage_110" );

    query =
    {
      query      = "find",
      group      = "player",
      collection = "info"
    };

    local stage_last = lib_db.Execute( query ).result.stage_last;

    if ( set_data.constructed ) and ( stage_sv_data.completed ) and ( stage_last == "stage_110" ) then

      query =
      {
        query      = "update",
        group      = "player",
        collection = "info",
        set        =
        {
          stage_last = "stage_1_bonus"
        }
      };
      lib_db.Execute( query );

    end;

  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  -- fix 103 level
  local stage_sv_data = func_stage_sv_data( "stage_103" );

  if ( stage_sv_data ) then

    local query =
    {
      query      = "update",
      group      = "player",
      collection = "stages",
      indice     = "name",
      criteria   = { like = "stage_103" },
      set        =
      {
        type = "ride"
      }
    };
    lib_db.Execute( query );

  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
end;
--------------------------------------------------------------------------------
-- fix change mines 1 world
--------------------------------------------------------------------------------
function private.Migration_1472_client ()
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  local func_sv_data = function ( mine )

    local query =
    {
      query      = "find",
      group      = "player",
      collection = "mines",
      indice     = "name",
      criteria   = { like = mine }
    };
    return lib_db.Execute( query ).result;

  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  local func_db_data = function ( mine )

    local query =
    {
      query      = "find",
      group      = "static",
      collection = "mines",
      indice     = "name",
      criteria   = { like = mine }
    };
    return lib_db.Execute( query ).result;

  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  local func_is_collect_worker = function ( stage )

    local query =
    {
      query      = "find",
      group      = "player",
      collection = "stages",
      indice     = "name",
      criteria   = { like = stage }
    };
    local sv_data = lib_db.Execute( query ).result;

    if ( sv_data ) then

      query =
      {
        query      = "find",
        group      = "static",
        collection = "stages",
        indice     = "name",
        criteria   = { like = stage }
      };
      local db_data = lib_db.Execute( query ).result;

      local state = ( sv_data.state or 1 );
      if ( sv_data.noreward ) then
        state = state + 1;
      end;

      for i = 1, #db_data.state do
        if ( db_data.state[ i ].reward ) then

          local reward = db_data.state[ i ].reward;

          for j = 1, #reward do
            if ( reward[ j ].name == "worker" ) then
              if ( state > i ) then
                return true;
              end;
            end;
          end;

        end;
      end;

    end;

    return false;

  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  local func_check_mine = function ( mine_name )

    local sv_data = func_sv_data( mine_name );

    if ( sv_data and ( ( sv_data.state == "none" ) or ( sv_data.state == "wait" ) ) )  then

      local progress = 0;
      local db_mine  = func_db_data( mine_name );
      local stages   = db_mine.progress;
      local state    = "none";

      for i = 1, #stages do
        local is_collect_worker = func_is_collect_worker( stages[ i ] );
        if ( is_collect_worker ) then
          progress = progress + 1;
        end;
      end;

      if ( progress >= #stages ) then
        state = "wait";
      end;

      local query =
      {
        query      = "update",
        group      = "player",
        collection = "mines",
        indice     = "name",
        criteria   = { like = mine_name },
        set        = {
          progress = progress,
          state = state
        }
      };

      lib_db.Execute( query );

    end;

  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  func_check_mine( "mine_1" );
  func_check_mine( "mine_2" );
  func_check_mine( "mine_3" );
  func_check_mine( "mine_4" );

end;
--------------------------------------------------------------------------------
-- fix changes in minievent context message feature data structure
--------------------------------------------------------------------------------
function private.Migration_1624_client ()

  local feature_name = "context_message";

  local query =
  {
    query      = "find",
    group      = "sml",
    collection = "features",
    indice     = "name",
    criteria   = { like = feature_name }
  };

  local feature_data = lib_db.Execute( query ).result;

  if ( feature_data                         ) and
     ( feature_data.data                    ) and
     ( feature_data.data.minievent          ) and
     ( feature_data.data.minievent.event_id ) then

    local old_data = feature_data.data.minievent;

    feature_data.data.minievent = {};
    feature_data.data.minievent[ "id_"..old_data.event_id ] = old_data.enabled;

    query = {
      query      = "update",
      group      = "sml",
      collection = "features",
      indice     = "name",
      criteria   = { like = feature_name },
      set        = {
        data = feature_data.data
      }
    };

    lib_db.Execute( query );

  end;

end;
--------------------------------------------------------------------------------
-- Balance correction for second quests in World 2 and World 3 (potions and
-- scrolls). Correction for qst_collect_sparkle.
--------------------------------------------------------------------------------
function private.Migration_1693_client ()

  local func_get_saved_quest = function ( quest_name )

    local query = {
      query      = "find",
      group      = "player",
      collection = "quests_current",
      indice     = "name",
      criteria   = { like = quest_name }
    };
    return lib_db.Execute( query ).result;

  end;

  local func_get_static_quest = function ( quest_name )

    local query = {
      query      = "find",
      group      = "static",
      collection = "quests",
      indice     = "name",
      criteria   = { like = quest_name }
    };
    return lib_db.Execute( query ).result;

  end;

  local func_get_character = function ( quest_name )

    local query = {
      query      = "find",
      group      = "static",
      collection = "characters",
      indice     = "quest_cage",
      criteria   = { like = quest_name }
    };
    local character_data = lib_db.Execute( query ).result;

    query = {
      query      = "find",
      group      = "static",
      collection = "characters",
      indice     = "quest_free",
      criteria   = { like = quest_name }
    };

    character_data = character_data or lib_db.Execute( query ).result;
    return character_data.name;

  end;

  local func_replace_quest_data = function ( quest_name, saved_data, static_data, completed )

    -- Delete old quest.
    local delete_query = {
      query      = "delete",
      group      = "player",
      collection = "quests_current",
      indice     = "name",
      criteria   = { like = quest_name }
    };
    lib_db.Execute( delete_query );

    if ( completed ) then
      -- Intert in compteted.
      local insert_query = {
        query      = "insert",
        group      = "player",
        collection = "quests_completed",
        value      = { name = quest_name }
      };
      lib_db.Execute( insert_query );

      local character_query = {
        query      = "update",
        group      = "player",
        collection = "characters",
        indice     = "name",
        criteria   = { like = func_get_character( quest_name ) },
        set        = {
          state = "inactive"
        }
      };
      lib_db.Execute( character_query );

      -- Obtain rewards.
      static_data.reward = static_data.reward or {};
      for r = 1, #static_data.reward do
        local reward_data = static_data.reward[ r ];
        reward_data.category = "quest";
        reward_data.quest_name = quest_name;
        insert_query = {
          query      = "insert",
          group      = "player",
          collection = "rewards",
          value      = reward_data
        };
        lib_db.Execute( insert_query );
      end;
      return;
    end;

    -- Intert new data in current.
    local insert_query = {
      query      = "insert",
      group      = "player",
      collection = "quests_current",
      value      = saved_data
    };
    lib_db.Execute( insert_query );

  end;

  local func_main = function ( qst_name )
    local qst_saved_data = func_get_saved_quest( qst_name );
    local qst_static_data = func_get_static_quest( qst_name );

    local new_target_list = {};

    if ( qst_saved_data and qst_saved_data.list ) then
      for i = 1, #qst_saved_data.list do
        local prg_saved = qst_saved_data.list[ i ];
        for j = 1, #qst_static_data.target.list do
          local prg_static = qst_static_data.target.list[ j ];
          if ( prg_saved == prg_static ) then
            table.insert( new_target_list, prg_saved );
          end;
        end;
      end;

      local completed = false;
      qst_saved_data.list = ( #new_target_list > 0 ) and new_target_list or nil;
      if ( #new_target_list == #qst_static_data.target.list ) then
        completed = true;
      end;
      func_replace_quest_data( qst_name, qst_saved_data, qst_static_data, completed );
    end;
  end;

  local qst_list = { "green", "red", "yellow", "blue", "purple" };
  local qst_name_patterns = { "qst_collect_%s_potion", "qst_collect_%s_scroll" };

  for p = 1, #qst_name_patterns do
    local qst_name_pattern = qst_name_patterns[ p ];
    for q = 1, #qst_list do

      local qst_name = qst_name_pattern:format( qst_list[ q ] );
      func_main( qst_name );

    end;
  end;
  func_main( "qst_collect_sparkle" );

end;
--------------------------------------------------------------------------------
-- Fix mines balance in Worlds 2 - 5.
--------------------------------------------------------------------------------
function private.Migration_1694_client ()

  local func_get_saved = function ( collection, name )

    if ( name ) then
      local query =
      {
        query      = "find",
        group      = "player",
        collection = collection,
        indice     = "name",
        criteria   = { like = name }
      };
      return lib_db.Execute( query ).result;
    end;

    local query =
    {
      query      = "find_all",
      group      = "player",
      collection = collection
    };
    return lib_db.Execute( query ).result;

  end;

  local func_get_static = function ( collection, name )

    local query =
    {
      query      = "find",
      group      = "static",
      collection = collection,
      indice     = "name",
      criteria   = { like = name }
    };
    return lib_db.Execute( query ).result;

  end;

  local func_replace = function ( name, data )

    local delete_query =
    {
      query      = "delete",
      group      = "player",
      collection = "mines",
      indice     = "name",
      criteria   = { like = name }
    };
    lib_db.Execute( delete_query );

    local insert_query =
    {
      query      = "insert",
      group      = "player",
      collection = "mines",
      value      = data
    };
    lib_db.Execute( insert_query );

  end;

  local mines_list = func_get_saved( "mines" );

  for m = 1, #mines_list do
    local mine_name = mines_list[ m ].name;
    local saved_data = func_get_saved( "mines", mine_name );
    local static_data = func_get_static( "mines", mine_name );

    if ( saved_data and saved_data.progress and static_data.progress ) then
      local new_progress = 0;
      for j = 1, #static_data.progress do
        local stage_name = static_data.progress[ j ];
        local saved_data = func_get_saved( "stages", stage_name );
        local static_data = func_get_static( "stages", stage_name ).state;
        local worker_state = ( static_data[ 1 ].reward and
        static_data[ 1 ].reward[ 1 ].name == "worker" ) and 1 or 0;
        worker_state = ( static_data[ 2 ].reward and
        static_data[ 2 ].reward[ 1 ].name == "worker" ) and 2 or worker_state;
        worker_state = ( static_data[ 3 ].reward and
        static_data[ 3 ].reward[ 1 ].name == "worker" ) and 3 or worker_state;

        if ( saved_data and ( saved_data.noreward or
          ( saved_data.state and saved_data.state > worker_state ) ) ) then
          new_progress = new_progress + 1;
        end;
      end;
      saved_data.progress = new_progress;
      if ( saved_data.progress == #static_data.progress and
         ( not saved_data.state or saved_data.state == "none" ) ) then
        saved_data.state = "wait";
      end;
      if ( saved_data.progress == 0 ) then
        saved_data.state = "none";
      end;
      func_replace( mine_name, saved_data );
    end;
  end;

end;
--------------------------------------------------------------------------------
-- Fix for fragments in Worlds 2 - 3.
--------------------------------------------------------------------------------
function private.Migration_1695_client ()
  
  local func_get = function ( group, name )

    local query =
    {
      query      = "find",
      group      = group,
      collection = "stages",
      indice     = "name",
      criteria   = { like = name }
    };
    return lib_db.Execute( query ).result;
      
  end;
  
  local stages_list = { 
    { "stage_154",      "saw" },
    { "stage_164",      "pick" },
    { "stage_198",      "sickle" },
    { "stage_7_bonus",  "beermug" },
    { "stage_12_bonus", "steelbone" },
    { "stage_17_bonus", "steeringwheel" },
    { "stage_22_bonus", "ironheart" },
    { "stage_29_bonus", "librarykey" },
    { "stage_34_bonus", "kitchenkey" },
    { "stage_39_bonus", "gardenkey" },
    { "stage_44_bonus", "throneroomkey" }
  };
  
  for i = 1, #stages_list do
    local stage_data = stages_list[ i ];
    local static_data = func_get( "static", stage_data[ 1 ] );
    local saved_data = func_get( "player", stage_data[ 1 ] );
    if ( saved_data and saved_data.state == 3 and saved_data.noreward ) then
      for s = 1, #static_data.state do
        if ( static_data.state[ s ].reward and 
             next( static_data.state[ s ].reward ) and 
             static_data.state[ s ].reward[ 1 ].name == stage_data[ 2 ] ) then
          -- Obtain rewards.
          local reward_data = static_data.state[ s ].reward[ 1 ];
          insert_query = {
            query      = "insert",
            group      = "player",
            collection = "rewards",
            value      = reward_data
          };
          lib_db.Execute( insert_query );
        end; 
      end;
    end;
  end;
  
end;
--------------------------------------------------------------------------------
-- Balance correction for second quests in World 6 - ... 
--------------------------------------------------------------------------------
function private.Migration_1696_client ()
  
  local item_list = {
    "carrot",    "lollipop",      "log",      "walrusfish", "cupoftea",
    "ticket",    "sugarcone",     "plushtoy", "jug",        "dumbbell",
    "grape",     "bale",          "bandage",  "papyrus",    "catstatuette",
    "shuriken",  "bone",          "handbell", "bamboo",     "fan",
    "brush",     "shepherdshorn", "herbs",    "forceps",    "paddle",
    "paperboat", "goldcoin",      "map",      "ball",       "harpoon",
    "loupe",     "berry",         "comb",     "feather",    "lantern",
    "crystalls", "strawberry",    "oldring",  "megabomb",   "freedomkey"
  };
  
  private.MoveQuestItems( item_list );
  
end;
--------------------------------------------------------------------------------
function private.MoveQuestItems ( item_list )
  
  local func_get_saved_quest = function ( quest_name )

    local query = {
      query      = "find",
      group      = "player",
      collection = "quests_current",
      indice     = "name",
      criteria   = { like = quest_name }
    };
    return lib_db.Execute( query ).result;

  end;

  local func_get_static_quest = function ( quest_name )

    local query = {
      query      = "find",
      group      = "static",
      collection = "quests",
      indice     = "name",
      criteria   = { like = quest_name }
    };
    return lib_db.Execute( query ).result;

  end;

  local func_get_character = function ( quest_name )

    local query = {
      query      = "find",
      group      = "static",
      collection = "characters",
      indice     = "quest_cage",
      criteria   = { like = quest_name }
    };
    local character_data = lib_db.Execute( query ).result;

    query = {
      query      = "find",
      group      = "static",
      collection = "characters",
      indice     = "quest_free",
      criteria   = { like = quest_name }
    };

    character_data = character_data or lib_db.Execute( query ).result;
    return character_data.name;

  end;

  local func_replace_quest_data = function ( quest_name, saved_data, static_data, completed )

    -- Delete old quest.
    local delete_query = {
      query      = "delete",
      group      = "player",
      collection = "quests_current",
      indice     = "name",
      criteria   = { like = quest_name }
    };
    lib_db.Execute( delete_query );

    if ( completed ) then
      -- Intert in compteted.
      local insert_query = {
        query      = "insert",
        group      = "player",
        collection = "quests_completed",
        value      = { name = quest_name }
      };
      lib_db.Execute( insert_query );

      local character_query = {
        query      = "update",
        group      = "player",
        collection = "characters",
        indice     = "name",
        criteria   = { like = func_get_character( quest_name ) },
        set        = {
          state = "inactive"
        }
      };
      lib_db.Execute( character_query );

      -- Obtain rewards.
      static_data.reward = static_data.reward or {};
      for r = 1, #static_data.reward do
        local reward_data = static_data.reward[ r ];
        DbgTrace( "Obtained: "..reward_data.name );
        reward_data.category = "quest";
        reward_data.quest_name = quest_name;
        insert_query = {
          query      = "insert",
          group      = "player",
          collection = "rewards",
          value      = reward_data
        };
        lib_db.Execute( insert_query );
      end;
      return;
    end;

    DbgTrace( "New data: ", saved_data );
    -- Intert new data in current.
    local insert_query = {
      query      = "insert",
      group      = "player",
      collection = "quests_current",
      value      = saved_data
    };
    lib_db.Execute( insert_query );

  end;

  local func_main = function ( qst_name )
    local qst_saved_data = func_get_saved_quest( qst_name );
    local qst_static_data = func_get_static_quest( qst_name );

    local new_target_list = {};

    if ( qst_saved_data and qst_saved_data.list ) then
      for i = 1, #qst_saved_data.list do
        local prg_saved = qst_saved_data.list[ i ];
        for j = 1, #qst_static_data.target.list do
          local prg_static = qst_static_data.target.list[ j ];
          if ( prg_saved == prg_static ) then
            table.insert( new_target_list, prg_saved );
          end;
        end;
      end;

      local completed = false;
      qst_saved_data.list = ( #new_target_list > 0 ) and new_target_list or nil;
      if ( #new_target_list == #qst_static_data.target.list ) then
        completed = true;
      end;
      func_replace_quest_data( qst_name, qst_saved_data, qst_static_data, completed );
    end;
  end;
                   
  local qst_name_patterns = { "qst_collect_%s" };

  for p = 1, #qst_name_patterns do
    local qst_name_pattern = qst_name_patterns[ p ];
    for q = 1, #item_list do

      local qst_name = qst_name_pattern:format( item_list[ q ] );
      func_main( qst_name );

    end;
  end;

end;
--------------------------------------------------------------------------------
-- Fix mines balance in Worlds 6 - ...
--------------------------------------------------------------------------------
function private.Migration_1697_client ()

  private.Migration_1694_client();

end;
--------------------------------------------------------------------------------
-- Clean data for removed 'add_friends_fireman' achievement
--------------------------------------------------------------------------------
function private.Migration_1698_client ()

  local query =
  {
    query      = "delete",
    group      = "player",
    collection = "achievements",
    indice     = "name",
    criteria   = { like = "add_friends_fireman" }
  };
  
  lib_db.Execute( query );

end;
--------------------------------------------------------------------------------
-- Fix gold balance and finish abtest from 1.23.1 to 1.24.0
--------------------------------------------------------------------------------
function private.Migration_1699_client ()

  local feature_name = "gold_balance";
  
  -- JK: remove feature data from local storage
  local query = {
    query      = "delete",
    group      = "sml",
    collection = "features",
    indice     = "name",
    criteria   = { like = feature_name }
  };

  lib_db.Execute( query );
  
  -- JK: set abtest to off
  lib_db.Player.Set( "abtest", false );
  
  -- JK: reload settings to default
  lib_db.TryReloadSettings( 0 );

end;
--------------------------------------------------------------------------------
-- Balance correction for second quests in worlds 11 - 18
--------------------------------------------------------------------------------
function private.Migration_1701_client ()
  
  local item_list = {
    "paperboat", "loupe", "crystalls", "caramelstars", "corn", "hotelbell",
    "olivecrown", "dogcookie"
  };
  
  private.MoveQuestItems( item_list );

end;
--------------------------------------------------------------------------------
function private.Migration_1702_client ()
  
  -- JK: repeat mines re-balance
  private.Migration_1694_client ();
  
end;
--------------------------------------------------------------------------------
function private.Migration_1703_client ()
  
  -- JK: fix update for ancients
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        =
    {
      update = false,
      reward = false
    }
  };

  lib_db.Execute( query );
  
end;
--------------------------------------------------------------------------------
function private.Migration_1965_client ()
  
  -- ES: fix build gold mine 
  local query = 
  {
    query      = "find",
    group      = "player",
    collection = "mines",
    indice     = "name",
    criteria   = {
      like = "mine_7"
    }
  };

  local mine_data = lib_db.Execute( query ).result;
  
  if ( ( not mine_data ) or ( not mine_data.resource ) ) then
    query =
    {
      query      = "find",
      group      = "player",
      collection = "stages",
      indice     = "name",
      criteria   = { like = "stage_158" }
    };

    local stage_158_data = lib_db.Execute( query ).result;
    
    if ( stage_158_data and stage_158_data.completed ) then 
      query =
      {
        query      = "update",
        group      = "player",
        collection = "mines",
        indice     = "name",
        criteria   = { like = "mine_7" },
        set        = 
        {
          level    = 1,
          progress = 0,
          resource = "gold",
          state    = "ready",
          unlocked = true
        }
      };

      if ( not lib_db.Execute( query ).result ) then
        query =
        {
          query      = "insert",
          group      = "player",
          collection = "mines",
          value      = 
          {
            name     = "mine_7",
            level    = 1,
            progress = 0,
            resource = "gold",
            state    = "ready",
            unlocked = true
          }
        };

        lib_db.Execute( query );

      end;
    end;
  end;
  
end;
--------------------------------------------------------------------------------
function private.Migration_1984_client ()
  
  local collection_exist = false;
  local query =
  {
    query      = "find",
    group      = "player",
    collection = "dailyquest"
  };
  
  local dailyquest = lib_db.Execute( query ).result;
  
  if ( dailyquest ) then
    collection_exist = true;
    local event_timer = "mte_dailyquest_refresh";
    if ( parts.time.event.Exists( event_timer ) ) then
      return;
    end;
  end;
  
  local data = {
    ready = false,
    reward = false
  };
  
  local unlock_stage = lib_db.Settings( "dailyquest" ).unlockstage;

  query = {
    query      = "find",
    group      = "player",
    collection = "stages",
    indice     = "name",
    criteria   = { like = unlock_stage }
  };

  local unlocked_stage_data = lib_db.Execute( query ).result;
  
  if ( unlocked_stage_data and unlocked_stage_data.completed ) then
    data.refresh = true;
  end;
  
  if ( collection_exist ) then
    query =
    {
      query      = "update",
      group      = "player",
      collection = "dailyquest",
      set        = data
    };
    lib_db.Execute( query );
  else
    lib_db.Rewrite( "player", "dailyquest", { data } );
  end;
  
end;
--------------------------------------------------------------------------------
function private.Migration_1994_client ()
  
  local query =
  {
    query      = "find",
    group      = "player",
    collection = "fireflies_stickerpacks"
  };
  
  local fireflies_stickerpacks = lib_db.Execute( query ).result;
  
  if ( fireflies_stickerpacks ) then
    return;
  end;
  
  query = {
    query      = "find",
    group      = "player",
    collection = "stages",
    indice     = "name",
    criteria   = { like = "stage_12" }
  };

  local unlocked = lib_db.Execute( query ).result;
  
  local data = {
    free = 0,
    bought = 0
  };
  
  if ( unlocked ) then
    data.free = 1;
  end;
  
  lib_db.Rewrite( "player", "fireflies_stickerpacks", { data } );
  
end;
--------------------------------------------------------------------------------