-- name=m_mines
--------------------------------------------------------------------------------
function public.__InitModule ()

  ne.msg.Subscribe(
    "Event_Game_Mines_MineReady",
    function ( msg_type, msg_params )
      private.Cooldown_Stop( msg_params.name, true );
    end
  );

  --Log.i( "Initialized." );

  private.is_gold_mine_disabled = false;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Setup ( funcs )
  
  local fresh_player = lib_db.Player.Info().gold_mine_fresh_player_feature;
  local gold_mine_fresh_player_server = m_features.Feature_GetState( "gold_mine_fresh_player" );
  
  if ( fresh_player ) and ( not gold_mine_fresh_player_server ) then
    private.is_gold_mine_disabled = true;
  end;
  
  private.func_get_stats = funcs.func_get_stats;
  private.func_get_remain_time_speedup = funcs.func_get_remain_time_speedup;
  
end;
--------------------------------------------------------------------------------
function private.GetSpeedUp ()
  
  if ( private.func_get_stats ) then
    return 1 + private.func_get_stats().mines_speed;
  end;
  
  return 1;
  
end;
--------------------------------------------------------------------------------
function private.GetRemainTimeSpeedUp ()
  
  if ( private.func_get_remain_time_speedup ) then
    return private.func_get_remain_time_speedup();
  end;
  
  return 0;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.GetStorageDataAll ()
  
  local query = 
  {
    query      = "find_all",
    group      = "player",
    collection = "mines"
  };

  local data = lib_db.Execute( query ).result;
  
  return data;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Usable_List ( section_name )
  
  local query = 
  {
    query      = "find_all",
    group      = "player",
    collection = "mines",
    indice     = "state",
    criteria   = { like = "active" }
  };

  local usable = lib_db.Execute( query ).result;
  
  if not ( usable ) then return false; end;
  
  local list = {};
  
  for m = 1, #usable do 
    
    if ( private.Mine_Section( usable[ m ].name ) == section_name ) then
      
      table.insert( list, usable[ m ] );
      
    end;
  
  end;
  
  if ( #list < 1 ) then return false; end;

  return list;

end;
--------------------------------------------------------------------------------
function public.ListAll ()

  local query = 
  {
    query      = "find_all",
    group      = "static",
    collection = "mines"
  };

  return lib_db.Execute( query ).result;

end;  
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Progress_GetCount ( mine_name )

  local mine_data  = public.Mine_Data( mine_name );
  local mine_state = mine_data.state;
  local resource   = mine_data.resource;

  local mine_type = public.Mine_Type( mine_name );

  if not ( mine_state == "ready" ) and not ( mine_state == "active" ) then
    return { name = resource, count = 0 };
  end;

  local timer_token = string.format( "mte_map_cooldown_mines_%s", mine_name );

  local rate = 1;
  local callback = lib_db.Events_GetCallbackByToken( timer_token );
  local total_count   = private.Mine_UpgradeData( mine_type )[ mine_data.level ].resources;
  local total_time     = public.GetMineTime( mine_type );
  local current_count = 0;

  local remaining_time = TimeGetEventRemainingTime( timer_token );
  local passed_time    = total_time - remaining_time;
  
  if ( callback and callback.rate ) then
    
    current_count = current_count + callback.count_resource;
    remaining_time = TimeGetEventRemainingTime( timer_token );
    passed_time    = total_time - remaining_time - callback.time_passed;
    rate = callback.rate;
    
    if ( passed_time > callback.time_speedup ) then
      local resources = math.floor( rate * callback.time_speedup * total_count / total_time );
      current_count = current_count + resources;
      passed_time = passed_time - callback.time_speedup;
      rate = 1;
      
    end;
    
  end;
  
  local count = math.floor( rate * passed_time * total_count / total_time );
  
  current_count = current_count + count;
  
  if ( current_count >= total_count ) then
    current_count = total_count;
    if ( mine_state == "active" ) then
      private.Cooldown_Stop( mine_name, true );
    end;
  end;

  if ( current_count < 0 ) then
    current_count = 0;
  end;
  
  return { name = resource, count = current_count, full = total_count };

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Mine_GetNameByUnlock ( stage_name )

  local mine_query = 
  {
    query      = "find",
    group      = "static",
    collection = "mines",
    indice     = "unlock",
    criteria   = { like = stage_name }
  };

  local mine_data = lib_db.Execute( mine_query ).result;
  
  if ( mine_data ) then
    
    return mine_data.name;
    
  end;
  
  return false;

end;
--------------------------------------------------------------------------------
function public.Mine_Unlock ( mine_name )
  
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "mines",
    indice     = "name",
    criteria   = { like = mine_name },
    set        = 
    {
      state    = "none",
      unlocked = true,
      level    = 1,
      progress = 0
    }
  };

  if ( not lib_db.Execute( query ).result ) then

    query =
    {
      query      = "insert",
      group      = "player",
      collection = "mines",
      value      = 
      {
        name     = mine_name,
        unlocked = true,
        state    = "none",
        level    = 1,
        progress = 0
      }
    };

    lib_db.Execute( query );

  end;

  private.HandledAndSendNotify( "Model_Mines_MineUnlocked", mine_name );
  
  return {
    name     = mine_name,
    unlocked = true,
    state    = "none",
    level    = 1,
    progress = 0
  };

end;
--------------------------------------------------------------------------------
function public.Mine_IsUpgradable ( mine_type, mine_level )

  local new_level = mine_level + 1;
  
  if ( new_level > 3 ) then

    return false, nil, nil;

  end;
  
  local price = private.Mine_UpgradeData( mine_type )[ new_level ].price;
  
  return true, price, new_level;
  
end;
--------------------------------------------------------------------------------
function public.Mine_Upgrade ( mine_name, new_level )

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "mines",
    indice     = "name",
    criteria   = { like = mine_name },
    set        = 
    {
      level = new_level
    }
  };

  local is_upgraded = lib_db.Execute( query ).result;
  
  if ( is_upgraded ) then

    private.HandledAndSendNotify( "Model_Mines_MineLevelChanged", mine_name );
  
  end;
  
  return is_upgraded;

end;
--------------------------------------------------------------------------------
function public.Mine_Data ( mine_name )

  local mine_data = {};

  mine_data.name     = mine_name;
  mine_data.unlocked = false;

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "mines",
    indice     = "name",
    criteria   = { like = mine_name }
  };
  
  local price = lib_db.Execute( query ).result.price;
  
  query =
  {
    query      = "find",
    group      = "player",
    collection = "mines",
    indice     = "name",
    criteria   = { like = mine_name }
  };

  local saved_data = lib_db.Execute( query ).result;

  if ( saved_data ) then

    mine_data = saved_data;

    mine_data.level    = mine_data.level    or 1;
    mine_data.state    = mine_data.state    or "none";
    mine_data.progress = mine_data.progress or 0;

  end;
  
  mine_data.progress_max = public.Mine_ProgressMax( mine_name );

  local mine_type = public.Mine_Type( mine_name );

  mine_data.type  = mine_type;
  mine_data.price = price;

  mine_data.produce = {
    resources = private.Mine_UpgradeData( mine_type )[ mine_data.level or 1 ].resources,
    time      = public.GetMineTime( mine_type )
  };

  return mine_data;

end;
--------------------------------------------------------------------------------
function public.Mine_Type ( mine_name )

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "mines",
    indice     = "name",
    criteria   = { like = mine_name }
  };

  local mine = lib_db.Execute( query ).result;

  if ( not mine ) or ( not mine.resource ) then

    query =
    {
      query      = "find",
      group      = "static",
      collection = "mines",
      indice     = "name",
      criteria   = { like = mine_name }
    };

    mine = lib_db.Execute( query ).result;

    if ( mine and mine.progress and #mine.progress > 0 ) then

      return "resource", "resource1";

    end;

    return "gold", "gold";

  end;

  if ( mine.resource ~= "gold" ) then

    return "resource", mine.resource;

  end;

  return "gold", "gold";

end;
--------------------------------------------------------------------------------
function public.Mine_SetProgress ( mine_name, old_progress, new_progress )

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "mines",
    indice     = "name",
    criteria   = { like = mine_name },
    set        = 
    {
      progress = new_progress
    }
  };
  
  lib_db.Execute( query );

  private.HandledAndSendNotify(
    "Model_Mines_MineProgressChanged",
    mine_name,
    {
      name         = mine_name,
      progress     = old_progress,
      progress_new = new_progress
    }
  );

end;
--------------------------------------------------------------------------------
function public.Mine_ProgressMax ( mine_name )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "mines",
    indice     = "name",
    criteria   = { like = mine_name }
  };

  local progress = lib_db.Execute( query ).result.progress;

  if ( progress and #progress > 0 ) then

    --local progress_max = progress[ 2 ] - progress[ 1 ] + 1;
    local progress_max = #progress;

    return progress_max;

  end;

  return 0;

end;
--------------------------------------------------------------------------------
function public.Mine_SetStateWait ( mine_name )

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "mines",
    indice     = "name",
    criteria   = { like = mine_name },
    set        = 
    {
      state = "wait"
    }
  };

  lib_db.Execute( query );

  private.HandledAndSendNotify( "Model_Mines_MineWaiting", mine_name );

end;
--------------------------------------------------------------------------------
function public.Mine_UpgradeData ( mine_name )

  local mine_data    = public.Mine_Data( mine_name );
  local mine_type    = public.Mine_Type( mine_name );
  local upgrade_data = private.Mine_UpgradeData( mine_type );
  
  return {
    price      = upgrade_data[ mine_data.level + 1 ].price,
    count_next = upgrade_data[ mine_data.level + 1 ].resources,
    count      = upgrade_data[ mine_data.level ].resources
  };

end;
--------------------------------------------------------------------------------
function public.Mine_SetResource ( mine_name, mine_type, resource_name )

  local settings   = lib_db.Settings( "mines" );
  local start_full = settings[ mine_type ].upgrades[ 1 ].start_full or false;
  
  local state = tern( start_full )( "ready", "active" );

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "mines",
    indice     = "name",
    criteria   = { like = mine_name },
    set        = 
    {
      state    = state,
      resource = resource_name
    }
  };

  lib_db.Execute( query );
  
  private.HandledAndSendNotify( "Model_Mines_MineConstructed", mine_name );

  if ( start_full ) then
  
    private.HandledAndSendNotify( "Model_Mines_MineReady", mine_name );
    private.Cooldown_Stop( mine_name );

  else

    private.HandledAndSendNotify( "Model_Mines_MineActive", mine_name );
    private.Cooldown_Start( mine_name );

  end;

end;
--------------------------------------------------------------------------------
function public.Mine_OnResourceCollected ( mine_name, mine_type )

  if ( mine_type == "gold" ) then

    private.HandledAndSendNotify( "Model_Mines_GoldMineCollected", mine_name );

  end;

  if ( mine_type == "resource" ) then       

    private.HandledAndSendNotify( "Model_Mines_MineCollected", mine_name );

  end;

  private.Cooldown_Stop( mine_name );
  private.Cooldown_Start( mine_name );

end;
--------------------------------------------------------------------------------
function public.Mine_GetStaticData ( mine_name )

  local query = 
  {
    query      = "find",
    group      = "static",
    collection = "mines",
    indice     = "name",
    criteria   = {
      like = mine_name
    }
  };

  local mine_data = lib_db.Execute( query ).result;
  
  return mine_data;

end;
--------------------------------------------------------------------------------
function public.Mine_GetSavedData ( mine_name )

  local query = 
  {
    query      = "find",
    group      = "player",
    collection = "mines",
    indice     = "name",
    criteria   = {
      like = mine_name
    }
  };

  local saved_data = lib_db.Execute( query ).result;
  
  return saved_data;

end;
--------------------------------------------------------------------------------
function public.Mines_GetAllConstructed ()

  local query = 
  {
    query      = "find_all",
    group      = "player",
    collection = "mines",
    indice     = "state",
    criteria   = {
      like = ( "ready" or "active" )
    }
  };

  local mines_data = lib_db.Execute( query ).result;
  
  return mines_data;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.GetActiveMinesList ()
  
  local mine_data = private.GetStorageDataAll();
  local active = {};

  for i = 1, #mine_data do
    if ( mine_data[ i ].state == "active" ) and ( not private.IsMineUnavailable( mine_data[ i ].name ) ) then
      table.insert( active, mine_data[ i ].name );
    end;
  end;

  return active;

end;
--------------------------------------------------------------------------------
function public.GetResourcesTypes ( mine_name )

  local overrides = lib_db.Settings( "mines" ).resource.overrides;

  if ( overrides[ mine_name ] ) then

    return overrides[ mine_name ];

  end;

  return lib_db.Settings( "mines" ).resource.resources;

end;
--------------------------------------------------------------------------------
function public.GetMineTime ( mine_type )

  return lib_db.Settings( "mines" )[ mine_type ].time;

end;
--------------------------------------------------------------------------------
function private.Mine_UpgradeData ( mine_type )

  return lib_db.Settings( "mines" )[ mine_type ].upgrades;

end;
--------------------------------------------------------------------------------
function private.Mine_Section ( mine_name )
  
  local query =
  {
    query      = "find",
    group      = "static",
    collection = "mines",
    indice     = "name",
    criteria   = { like = mine_name }
  };
  
  return lib_db.Execute( query ).result.section;
  
end;
--------------------------------------------------------------------------------
function private.TrySetMineStateActive ( mine_name )

  if ( public.Mine_Data( mine_name ).state == "active" ) then return; end;

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "mines",
    indice     = "name",
    criteria   = { like = mine_name },
    set        = 
    {
      state = "active"
    }
  };

  lib_db.Execute( query );

  private.HandledAndSendNotify( "Model_Mines_MineActive", mine_name );

end;
--------------------------------------------------------------------------------
function private.TrySetMineStateReady ( mine_name )

  --FF
  -- Подстраховка от редкого бага.
  local active       = ( public.Mine_Data( mine_name ).state == "active" );
  local progress     = public.Mine_Data( mine_name ).progress or 0;
  local progress_max = public.Mine_ProgressMax( mine_name );

  if not ( active ) or not ( progress == progress_max ) then

    return;

  end;
  --FF
  
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "mines",
    indice     = "name",
    criteria   = { like = mine_name },
    set        = 
    {
      state = "ready"
    }
  };

  lib_db.Execute( query );

  private.HandledAndSendNotify( "Model_Mines_MineReady", mine_name );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Cooldown_Start ( mine_name )

  local mine_type  = public.Mine_Type( mine_name );
  local time = public.GetMineTime( mine_type );

  if ( time ) then

    local timer_token = "mte_map_cooldown_mines_"..mine_name; 

    if ( not TimeIsEventExist( timer_token ) ) then

      local data = { name = mine_name, time = time };

      private.Shedule( mine_name, time, data )
      
      if ( private.GetSpeedUp() > 1 ) then
        private.SpeedUpTimer( mine_name );
      end;
      
      private.TrySetMineStateActive( mine_name );

      private.HandledAndSendNotify( "Model_Mines_MineCooldownStart", mine_name );
    
    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Cooldown_Stop ( mine_name, full )

  private.Unshedule( mine_name );

  if ( full ) then

    private.TrySetMineStateReady( mine_name );

  end;

  private.HandledAndSendNotify( "Model_Mines_MineCooldownStop", mine_name );

end;
-------------------------------------------------------------------------------- 
function private.Shedule ( mine_name, time, data )
  
  local timer_token = "mte_map_cooldown_mines_"..mine_name; 
  local watch_token = "mte_map_cooldown_mines_"..mine_name.."_watch";
  
  TimeSheduleEvent(
    timer_token,
    "GLOBAL_TIME",
    time,
    "Event_Game_Mines_MineReady",
    data
  );
  
  TimeStartWatch( watch_token, "GLOBAL_TIME" );
  
end;
--------------------------------------------------------------------------------
function private.Unshedule ( mine_name )
  
  local timer_token = "mte_map_cooldown_mines_"..mine_name;
  local watch_token = "mte_map_cooldown_mines_"..mine_name.."_watch";

  TimeFinishEvent();
  TimeUnsheduleEvent( timer_token );
  TimeStopWatch( watch_token );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.SpeedUpTimer ()
  
  local mines_active = private.GetActiveMinesList();
  
  for i = 1, #mines_active do
    private.SpeedUpTimer( mines_active[ i ] );
  end;
  
end;
--------------------------------------------------------------------------------
function public.SpeedUpTimerFinished ()
  
  local mines_active = private.GetActiveMinesList();
  
  for i = 1, #mines_active do
    private.SpeedUpTimerFinished( mines_active[ i ] );
  end;
  
end;
--------------------------------------------------------------------------------
function private.SpeedUpTimer ( mine_name )
  
  local token = "mte_map_cooldown_mines_"..mine_name;
  
  if ( not TimeIsEventExist( token ) ) then
    return;
  end;
  
  local callback = lib_db.Events_GetCallbackByToken( token );
  
  if ( not callback ) then
    return;
  end;
  
  local rate = callback.rate or 1;
  local time_remain = TimeGetEventRemainingTime( token );
  local progress = public.Progress_GetCount( mine_name );
  
  private.Unshedule( mine_name );
  
  local data = {
    name = mine_name,
    time = callback.time,
    time_speedup = private.GetRemainTimeSpeedUp(),
    time_passed = callback.time - time_remain,
    rate = private.GetSpeedUp(),
    count_resource = progress.count
  };
  
  private.Shedule( mine_name, time_remain, data );
  
end;
--------------------------------------------------------------------------------
function private.SpeedUpTimerFinished ( mine_name )
  
  local token = "mte_map_cooldown_mines_"..mine_name;
  
  if ( not TimeIsEventExist( token ) ) then
    return;
  end;
  
  local callback = lib_db.Events_GetCallbackByToken( token );
  
  if ( not callback ) then
    return;
  end;
  
  local progress = public.Progress_GetCount( mine_name );
  
  if ( progress.count >= progress.full ) then
    return;
  end;
  
  private.Unshedule( mine_name );
  
  local mine_type   = public.Mine_Type( mine_name );
  local total_time  = public.GetMineTime( mine_type );
  local time_remain = total_time * ( progress.full - progress.count ) / progress.full;
  
  local data = {
    name = mine_name,
    time = callback.time
  };
  
  private.Shedule( mine_name, time_remain, data );
  
  if ( private.GetSpeedUp() > 1 ) then
    private.SpeedUpTimer( mine_name );
  end;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Mine_GetRandomCooldown ()
  
  -- For local notifications only.

  local mines_active = private.GetActiveMinesList();

  if ( #mines_active > 0 ) then

    local index = math.random( 1, #mines_active );
    local token = string.format( "mte_map_cooldown_mines_%s", mines_active[ index ] );

    return { name = mines_active[ index ], time = TimeGetEventRemainingTime( token ) };

  end;

  return nil;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.GetMineByUnlockStage( stage )

  local query = 
  {
    query      = "find",
    group      = "static",
    collection = "mines",
    indice     = "unlock",
    criteria   = { like = stage }
  };

  return ( lib_db.Execute( query ).result or {} ).name;

end;
--------------------------------------------------------------------------------
function public.GetAllBySection ( section_name )

  local query = 
  {
    query      = "find_all",
    group      = "static",
    collection = "mines",
    indice     = "section",
    criteria   = { like = section_name }
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function private.HandledAndSendNotify ( msg, mine_name, data )
  
  if ( not private.IsMineUnavailable( mine_name ) ) then
    if ( data ) then
      mediator.Notify( msg, data );
    else
      mediator.Notify( msg, mine_name );
    end;
  end;
  
end;
--------------------------------------------------------------------------------
function private.IsMineUnavailable ( mine_name )
  
  if ( mine_name == "mine_7" ) and ( private.is_gold_mine_disabled ) then
    return true;
  end;
  return false;
  
end;
--------------------------------------------------------------------------------
function public.IsMineUnavailable ( mine_name )
  
  return private.IsMineUnavailable( mine_name );
  
end;
--------------------------------------------------------------------------------