-- name=m_minievent
--------------------------------------------------------------------------------
local MODULE = "m_minievent";

local CHEST_NAMES = {
  wood   = "woodchest",
  silver = "silverchest",
  gold   = "goldchest"
};

local TYPE_GIFT    = "gift";
local TYPE_SECTION = "section";
local TYPE_CHESTS  = "chests";
local TYPE_STORY   = "story";
local TYPE_TOURNEY = "tourney";
local TYPE_TOWER   = "tower";
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.settings = lib_db.Settings( "minievent" );

  private.update =
  {
    scope = "SESSION_GLOBAL_TIME",
    token = "mte_minievent_update",
    event = "Command_UpdateMiniEvent"
  };

  private.finish =
  {
    scope = "SESSION_GLOBAL_TIME",
    token = "mte_minievent_finish",
    event = "Event_Minievent_Finished"
  };

  private.cooldown =
  {
    scope = "GLOBAL_TIME",
    token = "mte_minievent_cooldown",
    event = "Event_Minievent_Cooldown"
  };

  private.event_types        = {};
  private.event_names        = {};
  private.event_scenes       = {};
  private.event_rewards      = {};
  private.reward_buffers     = {};
  private.event_durations    = {};
  private.event_targets      = {};
  private.event_stage_won    = false;
  private.event_cooldowns    = {};
  private.event_content      = {};
  private.event_availability = {};

  private.restart_heartbeat = true;

  private.crown_value = nil;
  private.crown_static_data = private.Event_GetStaticData( "progress_crowns" );
  
  ne.msg.Subscribe( private.finish.event,   private.Timer_Stop );
  ne.msg.Subscribe( private.cooldown.event, private.Cooldown_Stop );

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function public.InitSchedule ()

  private.TryStartSchedule();

  ne.msg.Subscribe(
    Event_Feature_Updated,
    private.TryStartSchedule
  );

end;
--------------------------------------------------------------------------------
function private.TryStartSchedule ( msg, params )

  if ( public.IsFeatureActive() ) then

    if ( not private.schedule_subscribed ) then

      private.schedule_subscribed = true;

      ne.msg.Subscribe(
        private.update.event,
        private.Event_UpdateSchedule_Handler
      );

    end;

    private.Event_UpdateSchedule_Handler();

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Setup ( callbacks )

  private.GetContainerData  = callbacks.func_GetContainerData;
  private.IsPackageReady    = callbacks.func_IsPackageReady;
  private.GetCurrentStage   = callbacks.func_GetCurrentStage;
  private.GetCurrentSection = callbacks.func_GetCurrentSection;
  private.ChestGenerate     = callbacks.func_ChestGenerate;

end;
--------------------------------------------------------------------------------
function public.IsFeatureActive ()

  return parts.features.IsActive( "minievent" );

end;
--------------------------------------------------------------------------------
function public.GetUnlockStage ()

  return private.settings.unlock_stage;

end;
--------------------------------------------------------------------------------
function public.ClearStageWon ( event_id )

  private.event_stage_won = false;

  public.Event_SwitchScene( event_id );

end;
--------------------------------------------------------------------------------
function public.GetEventList ()

  if ( not private.event_list ) then return {}; end;

  local event_list = {};

  for _, event in pairs( private.event_list ) do

    if ( event.id   ) and
       ( event.type ) and
       ( event.name ) then

      local saved_data = private.Event_GetSavedData( event.id );

      if (  ( saved_data                  ) and
            ( not saved_data.is_finished  ) and
            ( not saved_data.is_completed ) ) or
         ( not saved_data                   ) then

        table.insert(
          event_list,
          {
            id   = event.id,
            type = event.type,
            name = event.name
          }
        );

      end;

    end;

  end;

  return event_list;

end;
--------------------------------------------------------------------------------
function public.GetCurrentEventId ( event_type )

  return event_type and private[ event_type.."_event_id" ] or private.current_event;

end;
--------------------------------------------------------------------------------
function public.SetCurrentEventId ( event_id )

  private.current_event = event_id;

end;
--------------------------------------------------------------------------------
function public.GetCurrentCustomEventId ( event_type )

  if ( private[ event_type.."_event_id" ] ) then

    local event_data = private.Event_GetSavedData( private[ event_type.."_event_id" ] );

    if ( event_data                                                ) and
       ( ( event_data.is_completed ) or ( event_data.is_finished ) ) then

      return nil;

    end;

    return private[ event_type.."_event_id" ];

  end;

  return nil;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Event_GetById ( event_id )

  local event_name = private.event_names[ "id_"..event_id ];
  local event_data = private.Event_GetStaticData( event_name );

  if ( not event_data ) then return false; end;

  local is_stage_won = ( type( private.event_stage_won ) == "table" );

  --- ID & default nature ------------------------------------------------------
  event_data.id     = event_id;
  event_data.nature = event_data.nature or "stages";

  --- Saved data ---------------------------------------------------------------
  local saved_data  = private.Event_GetSavedData( event_id ) or {};

  --- Scene --------------------------------------------------------------------
  if ( event_data.type == TYPE_STORY   ) or
     ( event_data.type == TYPE_SECTION ) then

    event_data.scene = tern( private.event_scenes[ "id_"..event_id ] )(
      private.event_scenes[ "id_"..event_id ],
      ( saved_data.scene or 1 )
    );

    event_data.scene_total = #event_data.scenes;

    if ( event_data.type == TYPE_STORY ) then
      event_data.stages = private.Event_GetStagesByScene(
        event_data.stages,
        event_data.scene
      );
    end;

    --- Nature: stages or location ---------------------------------------------
    event_data.nature = event_data.scenes[ event_data.scene ].nature;

    --- Level background -------------------------------------------------------
    event_data.level_background =
      event_data.scenes[ event_data.scene ].level_background;

    --- Chest position variant -------------------------------------------------
    event_data.chest_variant =
      event_data.scenes[ event_data.scene ].chest_variant;

    --- Character & activity ---------------------------------------------------
    event_data.character = event_data.scenes[ event_data.scene ].character;
    event_data.activity  = event_data.scenes[ event_data.scene ].activity;    

    --- Dialogs ----------------------------------------------------------------
    event_data.dialogs = event_data.scenes[ event_data.scene ].dialogs;

    --- Cleanup scenes ---------------------------------------------------------
    event_data.scenes = nil;

  end;

  --- Stages -------------------------------------------------------------------
  local stage_index = ( #( saved_data.stages or {} ) + 1 );

  event_data.is_completed     = saved_data.is_completed or false;
  event_data.stage            = event_data.stages[ stage_index ];
  event_data.stages_completed = saved_data.stages or {};
  event_data.progress         = #event_data.stages_completed;

  --- Reward -------------------------------------------------------------------
  event_data.reward  = public.Event_GetReward( event_id );
  local event_reward = event_data.reward or {};

  local is_all_completed =
    ( #event_data.stages_completed == #event_data.stages );

  local is_reward = {
    point4 = ( #event_data.stages_completed == 4 ) and
             ( event_reward.point4 ~= nil ),
    point7 = ( #event_data.stages_completed == 7 ) and
             ( event_reward.point7 ~= nil ),
    final  = ( is_all_completed ) and
             ( event_reward.final ~= nil ),
    chests = ( event_data.type == TYPE_CHESTS ) or
             ( ( event_reward.chests ~= nil ) and ( is_all_completed ) ),
    current_stage = false
  };

  if ( event_data.type == TYPE_SECTION ) then
    for index, reward in pairs( event_reward ) do
      if ( event_data.progress == reward.stage_number ) then
        is_reward.current_stage = index;
        break;
      end;
    end;
  end;

  event_data.is_reward = (
    is_reward.point4 or
    is_reward.point7 or
    is_reward.final or
    is_reward.chests or
    is_reward.current_stage
  ) and is_stage_won;

  if ( event_data.is_reward ) then

    if ( event_reward.chests ) then
      event_data.reward_group = TYPE_CHESTS;
    elseif ( event_data.type == TYPE_SECTION ) then
      event_data.reward_group = is_reward.current_stage;
    else
      event_data.reward_group = tern( is_all_completed )(
        "final",
        "point"..#event_data.stages_completed
      );
    end;

    -- JK: Collect stage perks
    if ( is_stage_won                  ) and
       ( private.event_stage_won.perks ) then

      local rewards = event_reward[ event_data.reward_group ] or {};
      local perks   = private.event_stage_won.perks;

      for id, reward in pairs( rewards ) do

        if ( perks[ reward.name ] ) then

          local factor = perks[ reward.name ];

          if ( factor > 1 ) then

            rewards[ id ].factor = factor;

            if ( reward.type == "chest" ) then

              for pos, item in pairs( reward.items ) do

                rewards[ id ].items[ pos ].count =
                  rewards[ id ].items[ pos ].count * factor;

              end;

            else

              rewards[ id ].count = rewards[ id ].count * factor;

            end;

          elseif ( factor < 1 ) then

            rewards[ id ].count = 0;

          end;

        end;

      end;

    end;

  end;

  --- Tourney specific data ----------------------------------------------------
  if ( event_data.type == TYPE_TOURNEY ) then

    event_data.reward_visible = {
      point4 = (
                ( #event_data.stages_completed < 4 ) and
                ( event_reward.point4 ~= nil )
               ) or ( event_data.reward_group == "point4" ),
      point7 = (
                ( #event_data.stages_completed < 7 ) and
                ( event_reward.point7 ~= nil )
               ) or ( event_data.reward_group == "point7" )
    };

  end;

  --- Flags --------------------------------------------------------------------
  event_data.is_finished = saved_data.is_finished or false;

  if ( not saved_data.id ) then

    event_data.is_finished = ( not public.Event_IsActive( event_id ) );

  end;

  event_data.is_suspended = private.is_suspended or false;

  --- Cooldown -----------------------------------------------------------------
  if ( event_data.cooldown ) then

    event_data.cooldown = tern( private.event_cooldowns[ "id_"..event_id ] )(
      private.event_cooldowns[ "id_"..event_id ],
      event_data.cooldown
    );

    event_data.is_cooldown = public.Cooldown_IsActive( event_id );

  end;

  --- Targets ------------------------------------------------------------------
  if ( event_data.targets ) then

    if ( private.event_targets[ "id_"..event_id ] ) then

      event_data.targets = common.table.DeepCopy(
        private.event_targets[ "id_"..event_id ],
        { key = "count", type = "number" }
      );

    end;

    local targets = saved_data.targets or {};

    for id, target in pairs( event_data.targets ) do

      local target_save  = targets[ target.type ] or {};
      local progress     = target_save.progress or 0;
      local is_completed = target_save.is_completed or false;
      local goal         = event_data.targets[ id ].goal.count;

      event_data.targets[ id ].progress     = progress;
      event_data.targets[ id ].is_completed = is_completed;

      if ( progress >= goal and not is_completed ) then

        event_data.is_target_reward           = true;
        targets[ target.type ].is_completed   = true;
        event_data.targets[ id ].is_completed = true;

        private.Event_Save( event_id, { targets = targets } );

      end;

    end;

  end;

  return event_data;

end;
--------------------------------------------------------------------------------
function public.Event_IsActive ( event_id )

  return parts.time.event.Exists( private.finish.token.."_"..event_id );

end;
--------------------------------------------------------------------------------
function public.Event_IsCompleted ( event_id )

  local event_data = private.Event_GetSavedData( event_id );

  if ( event_data ) then
    return event_data.is_completed;
  end;

  return false;

end;
--------------------------------------------------------------------------------
function public.Event_IsAvailableForSection ( event_id, section_number )

  local section_list = private.event_availability[ "id_"..event_id ];
  
  if ( section_list ) then
    local is_enlisted = false;
    for i = 1, #section_list do
      if ( tonumber(section_list[i]) == tonumber(section_number) ) then
        is_enlisted = true;
        break;
      end;
    end;
    return is_enlisted;
  end;

  return true;

end;
--------------------------------------------------------------------------------
function public.Event_IsFinishedOrCompleted ( event_id )

  local event_data = private.Event_GetSavedData( event_id );

  if ( event_data ) then
    return ( ( event_data.is_finished  ) or ( event_data.is_completed ) );
  end;

  return false;

end;
--------------------------------------------------------------------------------
function public.Event_GetStages ( event_id )

  local event_name = private.event_names[ "id_"..event_id ];
  local event_data = private.Event_GetStaticData( event_name );

  if ( event_data ) then
    
    if ( event_data.scenes      ) and
       ( #event_data.scenes > 1 ) then
      
      local scene  = private.Event_GetCurrentScene( event_id );
      local stages = private.Event_GetStagesByScene( event_data.stages, scene );
      return stages;
      
    end;
    
    return event_data.stages;
    
  end;

  return nil;

end;
--------------------------------------------------------------------------------
function public.Event_GetRewards ( event_id )

  local rewards = common.table.DeepCopy(
    private.event_rewards[ "id_"..event_id ]
  );

  return rewards;

end;
--------------------------------------------------------------------------------
function public.Event_GetCharacters ( event_name, scene_number )

  scene_number = scene_number or 1;

  local event_data = private.Event_GetStaticData( event_name );

  if ( event_data ) then

    if ( event_data.type == TYPE_SECTION ) then
      return event_data.scenes[ scene_number ].characters;
    end;

    return nil;

  end;

  return nil;

end;
--------------------------------------------------------------------------------
function public.Event_GetEnemies ( event_name, scene_number )

  scene_number = scene_number or 1;

  local event_data = private.Event_GetStaticData( event_name );

  if ( event_data ) then

    if ( event_data.type == TYPE_SECTION ) then
      return event_data.scenes[ scene_number ].enemies;
    end;

    return nil;

  end;

  return nil;

end;
--------------------------------------------------------------------------------
function public.Event_GetQuests ( event_name, scene_number )

  scene_number = scene_number or 1;

  local event_data = private.Event_GetStaticData( event_name );

  if ( event_data ) then

    if ( event_data.type == TYPE_SECTION ) then
      return event_data.scenes[ scene_number ].quests;
    end;

    return nil;

  end;

  return nil;

end;
--------------------------------------------------------------------------------
function public.Event_GetRewardSettings ( event_name, scene_number )

  scene_number = scene_number or 1;

  local event_data = private.Event_GetStaticData( event_name );

  if ( event_data ) then

    if ( event_data.type == TYPE_SECTION ) then
      return event_data.scenes[ scene_number ].reward_settings;
    end;

    return nil;

  end;

  return nil;

end;
--------------------------------------------------------------------------------
function public.Event_GetRemainingTime ( event_id )

  if ( public.Event_IsActive( event_id ) ) then

    local token   = private.finish.token.."_"..event_id;
    local remains = parts.time.event.GetRemainingTime( token );
    local total   = private.event_durations[ "id_"..event_id ];

    return math.min( remains, total );

  end;

  return 0;

end;
--------------------------------------------------------------------------------
function public.Event_GetReward (
    event_id,
    reward_group,
    reward_subgroup,
    extract_items
  )

  --DbgTrace(
  --  "Event_GetReward: id_"..event_id..
  --  ", group: "..tostring( reward_group )..
  --  ", subgroup: "..tostring( reward_subgroup )
  --);

  local reward = common.table.DeepCopy(
    private.event_rewards[ "id_"..event_id ]
  );

  --DbgTrace( reward );

  if ( reward_group           ) and
     ( reward[ reward_group ] ) then

    if ( reward_subgroup                           ) and
       ( reward[ reward_group ][ reward_subgroup ] ) then

      -- JK: extracted items from subgroup container ( chests -> 1 -> items )
      if ( extract_items                                   ) and
         ( reward[ reward_group ][ reward_subgroup ].items ) then

        return reward[ reward_group ][ reward_subgroup ].items;

      end;

      -- JK: non-container reward or container itself from subgroup
      -- ( chests -> 1 )
      return reward[ reward_group ][ reward_subgroup ];

    end;

    -- JK: extracted items from single container reward ( final -> items )
    if ( extract_items                ) and
       ( reward[ reward_group ].items ) then

      return reward[ reward_group ].items;

    end;

    -- JK: non-container reward or container itself from group ( final )
    return reward[ reward_group ];

  end;

  return reward;

end;
--------------------------------------------------------------------------------
function public.Event_ResolveReward ( event_id, stage_index )

  private.Event_ResolveReward( event_id, stage_index );

end;
--------------------------------------------------------------------------------
function public.Event_GetNameById ( event_id )

  return private.event_names[ "id_"..event_id ];

end;
--------------------------------------------------------------------------------
function public.Event_GetTypeById ( event_id )

  return private.event_types[ "id_"..event_id ];

end;
--------------------------------------------------------------------------------
function public.Event_Update ( event_id, stage_data, state )

  --- JK: collecting stage perks -----------------------------------------------
  if ( state == "win" ) then

    private.event_stage_won = {
      name  = stage_data.stage,
      perks = stage_data.perks
    };

  end;

  --- JK: get data for update --------------------------------------------------
  local saved_data       = private.Event_GetSavedData( event_id );
  local event_name       = private.event_names[ "id_"..event_id ];
  local static_data      = private.Event_GetStaticData( event_name );
  local event_type       = static_data.type;
  local scene_index      = saved_data.scene or 1;
  local stages_completed = saved_data.stages or {};
  local is_updated       = false;
  local update           = {};

  --- JK: win ------------------------------------------------------------------
  if ( state == "win" ) then

    table.insert( stages_completed, stage_data.stage );

    if ( event_type == TYPE_STORY ) then

      local scene_stages = private.Event_GetStagesByScene(
        static_data.stages,
        scene_index
      );

      if ( #stages_completed == #scene_stages ) then

        local scene_count = private.Event_GetSceneCount( event_id );
        local next_scene  = math.min( scene_count, scene_index + 1 );

        if ( next_scene <= scene_count ) then

          update.scene = next_scene;

        end;

      end;

    end;

    if ( event_type == TYPE_CHESTS                   ) and
       ( #stages_completed == #static_data.stages ) then

      stages_completed = false;

    end;

    update.stages = stages_completed;

    if ( static_data.targets ) then

      update.targets = saved_data.targets or {};

      local event_targets = static_data.targets;

      if ( private.event_targets[ "id_"..event_id ] ) then

        event_targets = common.table.DeepCopy(
          private.event_targets[ "id_"..event_id ],
          { key = "count", type = "number" }
        );

      end;

      for _, target in pairs( event_targets ) do

        local goal        = target.goal.count;
        local target_data = update.targets[ target.type ] or { is_completed = false };

        if ( not target_data.is_completed ) then

          local progress = target_data.progress or 0;

          if ( target.type == "DropItem" ) then

            for item_name, count in pairs( private.event_stage_won.perks ) do

              progress = progress + count;

            end;

          elseif ( target.type == "CompleteStage" ) then

            progress = progress + 1;

          end;

          update.targets[ target.type ] = update.targets[ target.type ] or {};
          update.targets[ target.type ].progress = progress;

        end;

      end;

    end;

    is_updated = true;

  end;

  --- JK: fail -----------------------------------------------------------------
  if ( state == "fail"       ) and
     ( #stages_completed > 0 ) and
     ( static_data.discard   ) then

    update.stages = false;
    is_updated    = true;

  end;

  --- JK: save data if changed -------------------------------------------------
  if ( is_updated ) then

    private.Event_Save( event_id, update );

  end;

end;
--------------------------------------------------------------------------------
function public.Event_Complete ( event_id )

  --DbgTrace( "[ INFO ][ m_minievent ] Event_Complete" );

  -- JK: if cooldown still exists for this event, stop it
  if ( public.Cooldown_IsActive ( event_id ) ) then
    public.Cooldown_Stop( event_id );
  end;

  private.Event_Save(
    event_id,
    {
      is_completed = true,
      is_finished  = true
    }
  );

  private.Timer_Stop( nil, { event_id = event_id } );

  mediator.Notify(
    "Model_Minievent_EventComplete",
    {
      event_id = event_id
    }
  );

end;
--------------------------------------------------------------------------------
function public.Event_SetProgress ( event_id, value )

  local saved_data = private.Event_GetSavedData( event_id );

  if ( not saved_data ) then

    -- JK: no saved data on current event, add new
    local event_data = {
      id       = event_id,
      progress = value,
      stages   = false
    };

    private.Event_Add( event_data );

    return;

  end;

  private.Event_Save( event_id, { progress = value } );

end;
--------------------------------------------------------------------------------
function public.Event_GetProgress ( event_id )

  local saved_data = private.Event_GetSavedData( event_id );

  if ( saved_data        ) and
     ( saved_data.stages ) then
    return #saved_data.stages, saved_data.stages;
  end;

  return 0;

end;
--------------------------------------------------------------------------------
function public.Event_Suspend ()

  private.is_suspended = true;

end;
--------------------------------------------------------------------------------
function public.Event_Resume ()

  private.is_suspended = nil;

end;
--------------------------------------------------------------------------------
function public.Event_IsSuspended ()

  return private.is_suspended;

end;
--------------------------------------------------------------------------------
function public.Event_Delete ( event_id )

  private.Event_Delete( event_id );

end;
--------------------------------------------------------------------------------
function public.Event_HasTutorial ( event_id, stage_name, minievent_data )

  if ( not minievent_data ) then

    local event_name = public.Event_GetNameById( event_id );
    minievent_data   = private.Event_GetStaticData( event_name );

  end;

  if ( minievent_data.tutorial                           ) and
     ( ( minievent_data.tutorial.stage   == stage_name ) or
       ( minievent_data.tutorial.replace == stage_name ) ) then

    return true, minievent_data.tutorial;

  end;

  return false, false;

end;
--------------------------------------------------------------------------------
function public.Event_HasDecoration ( event_name )

  local event_data = private.Event_GetStaticData( event_name );

  return ( type( event_data.decoration ) == "table" ), event_data.decoration;

end;
--------------------------------------------------------------------------------
function public.Event_SwitchScene ( event_id )

  local event_name  = private.event_names[ "id_"..event_id ];
  local static_data = private.Event_GetStaticData( event_name );

  if ( static_data.type == TYPE_STORY ) then

    -- JK: renew current_scene
    local previous_scene = private.Event_GetCurrentScene( event_id );
    local current_scene  = private.Event_GetCurrentScene( event_id, true );

    -- JK: on scene change, delete stage completed in previous scene
    if ( previous_scene ~= current_scene ) then

      --DbgTrace(
      --  "Scene changed from "..previous_scene..
      --  " to "..private.current_scene
      --);

      private.Event_Save( event_id, { stages = false } );
      private.Event_ResolveReward( event_id, 1 );

    end;

  end;

end;
--------------------------------------------------------------------------------
function public.Event_GetUpcoming ()

  local upcoming_event = private.upcoming_event;

  if ( upcoming_event                    ) and
     ( type( upcoming_event ) == "table" ) then

    local static_data = private.Event_GetStaticData( upcoming_event.name );

    if ( static_data ) then

      upcoming_event.type = static_data.type;

      return upcoming_event;

    end;

    DbgTrace( "[ ERROR ][ m_minievent ] No static data for upcoming event." );
    return false;

  end;

  return false;

end;
--------------------------------------------------------------------------------
function public.Event_GetIntro ( event_name )

  local event_id = public.GetCurrentCustomEventId( event_name:match( "(%w+)_%w+" ) );
  local saved_data = private.Event_GetSavedData( event_id );
  local intro_data = ( private.Event_GetStaticData( event_name ) or {} ).intro;
  intro_data.showed = ( saved_data or {} ).intro;
  return intro_data;

end;
--------------------------------------------------------------------------------
function public.Event_SetIntro ( event_name, value )

  local event_id = public.GetCurrentCustomEventId( event_name:match( "(%w+)_%w+" ) );
  local saved_data = private.Event_GetSavedData( event_id );

  if ( not saved_data ) then

    local event_data = {
      id       = event_id,
      intro    = value,
      stages   = false
    };

    private.Event_Add( event_data );

    return;

  end;

  private.Event_Save( event_id, { intro = value } );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Cooldown_Start ( minievent_data )

  local event_id = minievent_data.id;
  local seconds  = minievent_data.cooldown.time;

  private.Cooldown_Start( event_id, seconds );

end;
--------------------------------------------------------------------------------
function public.Cooldown_Stop ( event_id )

  local params  = {
    event_id = event_id
  };

  private.Cooldown_Stop( nil, params );

  return true;

end;
--------------------------------------------------------------------------------
function public.Cooldown_GetRemainingTime ( event_id )

  event_id = event_id or public.GetCurrentEventId();

  local remaining_time = parts.time.event.GetRemainingTime(
    private.cooldown.token.."_"..event_id
  );

  return remaining_time;

end;
--------------------------------------------------------------------------------
function public.Cooldown_IsActive ( event_id )

  return ( public.Cooldown_GetRemainingTime( event_id ) > 0 );

end;
--------------------------------------------------------------------------------
function public.Cooldown_GetSkipPrice ( event_id )

  if ( private.event_cooldowns[ "id_"..event_id ] ) then

    return private.event_cooldowns[ "id_"..event_id ].price;

  end;

  local event_name  = private.event_names[ "id_"..event_id ];
  local static_data = private.Event_GetStaticData( event_name );

  return static_data.cooldown.price;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Event_UpdateSchedule ()

  local tmr = private.update;

  parts.time.event.Shedule(
    tmr.token,
    tmr.scope,
    private.settings.update_time,
    tmr.event,
    {}
  );

end;
--------------------------------------------------------------------------------
function private.Event_UpdateSchedule_Handler ()

  parts.time.event.Finish();

  local restart_heartbeat = private.restart_heartbeat;

  lib_rpc.Schedule(
    "minievent",
    "GetCurrent",
    {},
    private.Event_OnDataUpdated,
    true,
    restart_heartbeat
  );

  private.restart_heartbeat = false;

  private.Event_UpdateSchedule();

end;
--------------------------------------------------------------------------------
function private.Event_OnDataUpdated ( message, event_data )

  -- JK: for server response debugging
  --private.dbg_event_data = common.table.DeepCopy( event_data );

  if ( not event_data.events                ) or
     ( type( event_data.events ) ~= "table" ) then

    event_data = {
      events = {},
      next   = {}
    };

  end;

  local story_event_id;
  local gift_event_id;
  local section_event_id;

  -- JK: detect clean start ----------------------------------------------------
  local is_clean_start = ( type( private.event_list ) ~= "table" );

  -- JK: backup old event data -------------------------------------------------
  local idx_backup       = private.event_list or {};
  local event_list       = event_data.events;
  private.upcoming_event = event_data.next;

  local assoc_backup = {};
  local assoc_list   = {};
  local assoc_dnld   = {};

  -- JK: process backup data ---------------------------------------------------
  for index, event in pairs( idx_backup ) do
    assoc_backup[ "id_"..event.id ] = event;
  end;

  -- JK: process received data -------------------------------------------------
  for i = #event_list, 1, -1 do

    local event       = event_list[ i ];
    local static_data = private.Event_GetStaticData( event.name );

    if ( static_data ) then

      event.type = static_data.type;

      local is_ready, state, notify_downloading = false, nil, false;

      if ( private.event_content[ event.name ] ~= "ready" ) then

        is_ready, state = private.Event_IsContentReady( event.name );

        notify_downloading =
          ( private.event_content[ event.name ] ~= "downloading" );

        private.event_content[ event.name ] = state;

      end;

      local current_state = private.event_content[ event.name ];

      if ( current_state == "ready" ) then

        private.Event_ProcessData( event, static_data );
        assoc_list[ "id_"..event.id ] = event;

        if ( event.type == TYPE_STORY   ) then story_event_id   = event.id; end;
        if ( event.type == TYPE_GIFT    ) then gift_event_id    = event.id; end;
        if ( event.type == TYPE_SECTION ) then section_event_id = event.id; end;

      elseif ( current_state == "downloading" ) then

        assoc_dnld[ "id_"..event.id ] = true;

        if ( notify_downloading ) then

          mediator.Notify( "Model_Minievent_ContentIsDownloading",
            {
              id   = event.id,
              type = event.type,
              name = event.name
            }
          );

        end;

        table.remove( event_list, i );

      end;

    else

      table.remove( event_list, i );
      Log.e( event.name.." has no static data" );

    end;

  end;

  -- JK: commit the event_list -------------------------------------------------
  private.event_list = event_list;

  -- JK: Current custom event ids ----------------------------------------------
  private.story_event_id   = story_event_id;
  private.gift_event_id    = gift_event_id;
  private.section_event_id = section_event_id;

  local gift_event_data = private.Event_GetSavedData( gift_event_id );

  if ( private.gift_event_id and not gift_event_data ) then

    gift_event_data = {
      id       = gift_event_id,
      progress = false,
      rewards  = private.event_rewards[ "id_"..gift_event_id ].chests
    };

    private.Event_Add( gift_event_data );

  end;

  -- JK: sync with saved events, remove old data -------------------------------
  local saved_events = private.Event_GetAll();

  for _, event in pairs( saved_events ) do
    if ( event.id                          ) and
       ( not assoc_list[ "id_"..event.id ] ) and
       ( not assoc_dnld[ "id_"..event.id ] ) and
       ( event.id ~= private.current_event ) then
      private.Event_Delete( event.id );
    end;
  end;

  -- JK: lookup for finished events --------------------------------------------
  if ( not is_clean_start ) then

    local finished_events = {};

    for _, event in pairs( assoc_backup ) do

      if ( not assoc_list[ "id_"..event.id ] ) and
         ( not assoc_dnld[ "id_"..event.id ] ) then

        table.insert( finished_events, event );

        private.Timer_Stop( nil, { event_id = event.id } );

      end;

    end;

  end;

  -- JK: lookup for new events -------------------------------------------------
  local new_events = {};

  for _, event in pairs( assoc_list ) do

    if ( not assoc_backup[ "id_"..event.id ] ) then

      table.insert( new_events, event );

      mediator.Notify(
        "Model_Minievent_EventStarted",
        {
          event_id   = event.id,
          event_type = event.type
        }
      );

    end;

  end;

  --DbgTrace( "New events: "..tostring( #new_events ) );

  -- JK: Notify that data is updated -------------------------------------------
  mediator.Notify( "Model_Minievent_EventUpdated" );

end;
--------------------------------------------------------------------------------
function private.Event_ProcessData ( event, static_data )

  local saved_data  = private.Event_GetSavedData( event.id );
  local scene_index = 1;
  local stage_index = 1;

  if ( saved_data ) then

    stage_index = ( #( saved_data.stages or {} ) + 1 );

    local stages = {};
    local clear_stages = false;

    if ( static_data.type == TYPE_STORY ) then

      scene_index = tern( private.event_scenes[ "id_"..event.id ] )(
        private.event_scenes[ "id_"..event.id ],
        ( saved_data.scene or 1 )
      );

      stages = private.Event_GetStagesByScene(
        static_data.stages,
        scene_index
      );
      
      if ( not private.event_scenes[ "id_"..event.id ] ) then

        local curr_stage_one = stages[ 1 ].name;
        local prev_stage_one = ( saved_data.stages or {} )[ 1 ];

        --DbgTrace( tostring( curr_stage_one ), tostring( prev_stage_one ) );

        clear_stages = ( prev_stage_one and ( curr_stage_one ~= prev_stage_one ) );
      end;

    else

      stages = static_data.stages;

    end;

    if ( stage_index > #stages ) then

      stage_index = 1;

    end;

    -- JK: game is just started and player has some progress, reset it
    if ( not private.current_event ) and
       ( saved_data.progress       ) then

      if ( static_data.discard ) then

        clear_stages = true;

      end;

    end;
    
    if ( clear_stages ) then
      
      --DbgTrace( "[ m_minievent ] Clearing progress" );

      private.Event_Save( event.id, { stages = false } );
      stage_index = 1;
      clear_stages = nil;
      
    end;

  end;

  --- Cache --------------------------------------------------------------------
  private.event_types[ "id_"..event.id ]     = static_data.type;
  private.event_names[ "id_"..event.id ]     = event.name;
  private.event_cooldowns[ "id_"..event.id ] = event.cooldown;
  private.reward_buffers[ "id_"..event.id ]  = event.reward;

  if ( not private.event_stage_won ) then

    private.Event_ResolveReward( event.id, stage_index );

  end;

  --- Targets ------------------------------------------------------------------
  if ( event.targets ) then

    private.event_targets[ "id_"..event.id ] = event.targets;

  end;

  --- Duration -----------------------------------------------------------------
  private.event_durations[ "id_"..event.id ] =
    ( event.dateEnd - event.dateStart );
    
  --- Availability -------------------------------------------------------------
  if ( event.sections ) then
    private.event_availability[ "id_"..event.id ] = event.sections;
  end;

  --- Timer --------------------------------------------------------------------
  if ( ( not saved_data ) or
       ( ( saved_data                ) and
       ( not saved_data.is_completed ) ) ) then

    private.Timer_Start(
      event.id,
      event.dateEnd - parts.time.server.Timestamp()
    );

  end;

end;
--------------------------------------------------------------------------------
function private.Event_ResolveReward ( event_id, stage_index )

  local event_name  = private.event_names[ "id_"..event_id ];
  local static_data = private.Event_GetStaticData( event_name );
  local event_type  = static_data.type;
  local reward_data = common.table.DeepCopy(
    private.reward_buffers[ "id_"..event_id ] or {}
  );

  --- Default reward -----------------------------------------------------------
  local event_reward = {
    final = reward_data[ #reward_data ]
  };

  --- Tourney reward -----------------------------------------------------------
  if ( event_type == TYPE_TOURNEY ) then

    event_reward.point4 = reward_data[ 4 ];
    event_reward.point7 = reward_data[ 7 ];

  end;

  --- Chests reward ------------------------------------------------------------
  if ( event_type == TYPE_CHESTS ) then

    local chests_index = reward_data[ stage_index ];
    local chests_assoc = {};

    for pos, chest in pairs( chests_index ) do

      local container  = private.GetContainerData( chest.name ) or {};
      local chest_name = CHEST_NAMES[ container.type ];

      chest.type  = "chest";
      chest.name  = chest_name;
      chest.items = container.items;

      chests_assoc[ chest_name ] = chest;

    end;

    event_reward = {
      chests = {
        chests_assoc.woodchest,
        chests_assoc.silverchest,
        chests_assoc.goldchest
      }
    };

  end;

  --- Story reward -------------------------------------------------------------
  if ( event_type == TYPE_STORY ) then

    local scene_index   = private.Event_GetCurrentScene( event_id );
    local event_name    = private.event_names[ "id_"..event_id ];
    local static_data   = private.Event_GetStaticData( event_name );
    local scene_data    = static_data.scenes[ scene_index ];
    local reward_index  = scene_data.reward.stage;
    local scene_rewards = {};

    if ( #reward_data[ reward_index ] == 1 ) then

      local scene_reward = reward_data[ reward_index ][ 1 ];
      local container    = private.GetContainerData( scene_reward.name );

      local chest = {
        type  = "chest",
        name  = CHEST_NAMES[ container.type ],
        count = scene_reward.count,
        items = container.items
      };

      event_reward = {
        final = { chest }
      };

    else

      local chests_assoc = {};

      event_reward = {
        chests = {}
      };

      for i, scene_reward in pairs( reward_data[ reward_index ] ) do

        local scene_reward = reward_data[ reward_index ][ i ];

        if ( scene_reward.type == "container" ) then

          local container = private.GetContainerData( scene_reward.name );

          local chest = {
            type  = "chest",
            name  = CHEST_NAMES[ container.type ],
            count = scene_reward.count,
            items = container.items
          };

          chests_assoc[ chest.name ] = chest;

        end;

      end;

      event_reward = {
        chests = {
          chests_assoc.woodchest,
          chests_assoc.silverchest,
          chests_assoc.goldchest
        }
      };

    end;

  end;

  --- Gift reward --------------------------------------------------------------
  if ( event_type == TYPE_GIFT ) then

    if ( not private.event_rewards[ "id_"..event_id ] ) or
       (
          ( not public.Cooldown_IsActive( event_id )              ) and
          ( #private.event_rewards[ "id_"..event_id ].chests == 0 )

       ) then

      local gifts = private.Gift_GenerateAll( event_id, true );

      if ( gifts ) then

        event_reward = {
          chests = private.Gift_SortByStage( gifts )
        };

      end;

    else

      event_reward = private.event_rewards[ "id_"..event_id ];

    end;

  end;

  --- Section reward -----------------------------------------------------------
  if ( event_type == TYPE_SECTION ) then

    event_reward = {};

    for i = 1, ( #reward_data ) do

      if ( type( reward_data[ i ] ) == "table" ) and
         ( #reward_data[ i ] > 0               ) then

        local rewards     = {};
        local holder_type = TYPE_GIFT;

        for j = 1, #reward_data[ i ] do

          local reward = reward_data[ i ][ j ];
          
          if ( reward.type == "container" ) then

            local container_data = private.GetContainerData( reward.name );

            reward = {
              type  = "chest",
              name  = CHEST_NAMES[ container_data.type ],
              count = reward.count,
              items = container_data.items
            };

            holder_type = "chest";

          end;

          table.insert( rewards, reward );

        end;

        -- JK: sort chest from wood to gold
        if ( holder_type == "chest" ) and
           ( #rewards > 1           ) then
          local sorted = {};
          for _, reward in pairs( rewards ) do
            if ( reward.name == "woodchest" ) then
              table.insert( sorted, 1, reward );
            elseif ( reward.name == "silverchest" ) then
              if ( #sorted > 0                           ) and
                 ( sorted[ #sorted ].name == "goldchest" ) then
                table.insert( sorted, math.max( 1, #sorted - 1 ), reward );
              else
                table.insert( sorted, reward );
              end;
            elseif ( reward.name == "goldchest" ) then
              table.insert( sorted, reward );
            end;
          end;
          rewards = sorted;
        end;

        local stage_reward = {
          stage_number = i,
          holder_type  = holder_type,
          rewards      = rewards
        };

        table.insert( event_reward, stage_reward );

      end;

    end;

  end;

  --- Save to session ----------------------------------------------------------
  private.event_rewards[ "id_"..event_id ] = common.table.DeepCopy(
    event_reward,
    {
      key  = "count",
      type = "number"
    }
  );

end;
--------------------------------------------------------------------------------
function private.Event_Save ( event_id, event_data )

  if ( not event_id ) then

    DbgTrace( "[ ERROR ] m_minievent.Event_Save: event_id is not specified" );

  end;

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "minievents",
    indice     = "id",
    criteria   = { like = event_id },
    set        = event_data
  };

  lib_db.Execute( query );

end;
--------------------------------------------------------------------------------
function private.Event_Add ( event_data )

  local query =
  {
    query      = "insert",
    group      = "player",
    collection = "minievents",
    value      = event_data
  };

  lib_db.Execute( query );

end;
--------------------------------------------------------------------------------
function private.Event_Delete ( event_id )

  local query =
  {
    query      = "delete",
    group      = "player",
    collection = "minievents",
    indice     = "id",
    criteria   = { like = event_id }
  };

  lib_db.Execute( query );

end;
--------------------------------------------------------------------------------
function private.Event_GetSavedData ( event_id )

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "minievents",
    indice     = "id",
    criteria   = { like = event_id }
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function private.Event_GetCurrentScene ( event_id, is_flushed )

  if ( private.event_scenes[ "id_"..event_id ] ) and
     ( not is_flushed                          ) then

    return private.event_scenes[ "id_"..event_id ];

  end;

  local saved_data    = private.Event_GetSavedData( event_id ) or {};
  local current_scene = saved_data.scene or 1;

  private.event_scenes[ "id_"..event_id ] = current_scene;

  return current_scene;

end;
--------------------------------------------------------------------------------
function private.Event_GetStaticData ( event_name )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "minievents",
    indice     = "name",
    criteria   = { like = event_name }
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function private.Event_GetSceneCount ( event_id )

  local event_name = private.event_names[ "id_"..event_id ];
  local event_data = private.Event_GetStaticData( event_name );

  if ( event_data.scenes ) then

    return #event_data.scenes;

  end;

  return false;

end;
--------------------------------------------------------------------------------
function private.Event_GetAll ()

  local query =
  {
    query      = "find_all",
    group      = "player",
    collection = "minievents"
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function private.Event_GetStagesByScene ( stage_list, scene_number )

  local stages = {};

  for _, stage in pairs( stage_list ) do

    if ( stage.scene == scene_number ) then

      table.insert( stages, stage );

    end;

  end;

  return stages;

end;
--------------------------------------------------------------------------------
function private.Event_IsContentReady ( event_name )

  if ( event_name:find( "gift_" ) ) then
    return true, "ready";
  end;

  -- JK: 1st pass to get minievent package name
  local full_name = event_name:match( "^(.*)_%d+$" ) or event_name;

  -- JK: 2nd pass to get minievent package name
  full_name = full_name:match( "^(%w+)$" ) or
              full_name:match( "^(%w+_%w+)$" );

  local package_name = "minievent_"..full_name;

  local  is_ready, state = private.IsPackageReady( package_name );
  return is_ready, state;

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function private.Timer_Start ( event_id, seconds )

  --DbgTrace( "Timer_Start: "..event_id..", seconds - "..tostring( seconds ) );

  local token = private.finish.token.."_"..event_id;

  parts.time.event.Unshedule( token );
  parts.time.event.Shedule(
    token,
    private.finish.scope,
    seconds,
    private.finish.event,
    { event_id = event_id }
  );

end;
--------------------------------------------------------------------------------
function private.Timer_Stop ( message, params )

  local event_id = params.event_id;

  --DbgTrace( "Timer_Stop: "..event_id );

  parts.time.event.Finish();
  parts.time.event.Unshedule( private.finish.token.."_"..event_id );

  local saved_data = private.Event_GetSavedData( event_id );

  if ( saved_data ) then

    private.Event_Save( event_id, { is_finished = true } );

  end;

  mediator.Notify(
    "Model_Minievent_EventFinished",
    {
      event_id   = event_id,
      event_type = public.Event_GetTypeById( event_id )
    }
  );

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function private.Cooldown_Start ( event_id, seconds )

  --DbgTrace(
  --  "[ "..MODULE.." ] Cooldown started: "
  --  ..tostring( event_id )..", "
  --  ..tostring( seconds ).." sec"
  --);

  local token = private.cooldown.token.."_"..event_id;

  if ( private.Cooldown_Exists( event_id ) ) then

    parts.time.event.Finish();
    parts.time.event.Unshedule( token );

  end;

  local is_started = parts.time.event.Shedule(
    token,
    private.cooldown.scope,
    seconds,
    private.cooldown.event,
    {
      event_id = event_id
    }
  );

  return is_started;

end;
--------------------------------------------------------------------------------
function private.Cooldown_Exists ( event_id )

  return parts.time.event.Exists( private.cooldown.token.."_"..event_id );

end;
--------------------------------------------------------------------------------
function private.Cooldown_Stop ( message, params )

  local event_id   = params.event_id;
  local event_type = public.Event_GetTypeById( event_id );

  --DbgTrace(
  --  "[ "..MODULE.." ] Cooldown finished: "
  --  ..tostring( event_id )..", "
  --  ..tostring( event_type )
  --);

  parts.time.event.Finish();
  parts.time.event.Unshedule( private.cooldown.token.."_"..event_id );

  local func_notify = function ( notif_event_id, notif_event_type )

    mediator.Notify(
      "Model_Minievent_CooldownFinished",
      {
        event_id   = notif_event_id,
        event_type = notif_event_type
      }
    );

  end;

  if ( event_type ) then

    func_notify( event_id, event_type );

  else

    mediator.SubscribeOnce(
      "Model_Minievent_EventUpdated",
      function ()
        func_notify( event_id, public.Event_GetTypeById( event_id ) );
      end
    );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.CharacterDialog_GetData (
    event_id,
    scene_number,
    stage_name,
    when
  )

  local event_name = private.event_names[ "id_"..event_id ];

  local dialog_name =
    event_name..
    "_scene_"..scene_number.."_"..
    stage_name.."_"..
    when;

  local event_data = private.Event_GetSavedData( event_id );

  if ( event_data         ) and
     ( event_data.dialogs ) then

    if ( common.IsEntityInList( event_data.dialogs, dialog_name ) ) then
      return nil;
    end;

  end;

  --DbgTrace( dialog_name );

  local query = {
    query      = "find",
    group      = "static",
    collection = "dialogs",
    indice     = "name",
    criteria   = { like = dialog_name }
  };

  local dialog_data = lib_db.Execute( query ).result;

  if ( dialog_data ) then
    dialog_data.minievent = event_name;
    return dialog_data;
  end;

  return nil;

end;
--------------------------------------------------------------------------------
function public.CharacterDialog_Complete ( event_id, dialog_name )

  local event_data = private.Event_GetSavedData( event_id );

  if ( not event_data ) then

    -- JK: no saved data on current event, add new
    event_data = {
      id       = event_id,
      progress = false,
      stages   = false
    };

    private.Event_Add( event_data );

  end;

  local dialog_data = event_data.dialogs or {};

  table.insert( dialog_data, dialog_name );

  local update = {
    dialogs = dialog_data
  };

  private.Event_Save( event_id, update );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Book_GetData ( event_id )

  local event_data = private.Event_GetSavedData( event_id );
  local book_data  = {};

  if ( event_data                           ) and
     ( type( event_data.book ) == "boolean" ) then

    book_data.already_shown = event_data.book;

  else

    book_data.already_shown = false;

  end;

  return book_data;

end;
--------------------------------------------------------------------------------
function public.Book_SetComplete ( event_id )

  local event_data = private.Event_GetSavedData( event_id );

  if ( not event_data ) then

    -- JK: no saved data on current event, add new
    event_data = {
      id       = event_id,
      progress = false,
      stages   = false
    };

    private.Event_Add( event_data );

  end;

  local update = {
    book = true
  };

  private.Event_Save( event_id, update );

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function public.Gift_RefreshObtained ( is_cooldown_ended )

  local event_id = public.GetCurrentCustomEventId( TYPE_GIFT );

  if ( not event_id ) then return; end;

  local event_data = private.Event_GetSavedData( event_id );

  if ( not event_data ) then return; end;

  table.remove( event_data.rewards, 1 );

  if ( #event_data.rewards == 0 ) then

    if ( is_cooldown_ended ) then

      -- JK: cooldown ended, generate new reward
      local gifts = private.Gift_Generate( event_id, nil, true );
      if ( gifts ) then
        event_data.rewards = {
          gifts.woodchest,
          gifts.silverchest,
          gifts.goldchest
        };
      end;

    else

      -- JK: after one cycle played, go to cooldown
      local event_id = event_data.id;
      local seconds  = private.event_cooldowns[ "id_"..event_id ].time;

      private.Cooldown_Start( event_id, seconds );

    end;
    
  end;

  -- JK: sorting by stage
  event_data.rewards = private.Gift_SortByStage( event_data.rewards );

  private.Event_Save( event_id, event_data );
  private.event_rewards[ "id_"..event_id ] = common.table.DeepCopy(
    { chests = event_data.rewards or {} },
    {
      key  = "count",
      type = "number"
    }
  );

end;
--------------------------------------------------------------------------------
function private.Gift_SortByStage ( gifts )

  -- JK: sort the gifts by stage here
  -- they need to be sorted according player's progress not just by type

  local sorted_gifts = {};

  for _, gift in pairs( gifts ) do
    table.insert( sorted_gifts, gift );
  end;

  table.sort(
    sorted_gifts,
    function ( a, b ) return ( tonumber( a.stage ) < tonumber( b.stage ) ); end
  );

  return sorted_gifts;

end;
--------------------------------------------------------------------------------
function private.Gift_GenerateAll ( event_id, force )

  local gifts = {};
  local saved_data = private.Event_GetSavedData( event_id );

  -- Create reward table from saved data.
  if ( saved_data and saved_data.rewards ) then
    for i = 1, #saved_data.rewards do
      local gift = saved_data.rewards[ i ];
      gifts[ gift.name ] = gift;
    end;
  -- Create new reward table.
  else
    gifts = private.Gift_Generate( event_id, nil, force );
  end;

  return gifts;

end;
--------------------------------------------------------------------------------
function private.Gift_Generate ( event_id, gift_data, force )

  local event_name = private.event_names[ "id_"..event_id ];
  local rewards = private.Event_GetStaticData( event_name ).reward;

  local function func_generate_gift ( gift_name, pivot, index )

    local stage = pivot + 2 * index + 1;
    local last_stage = 110 * private.GetCurrentSection() - 1;
    if ( stage >= last_stage and not force ) then
      return nil;
    end;

    local gift = {
      name  = gift_name,
      type  = "chest",
      stage = stage
    };
    gift.items = private.ChestGenerate( gift, rewards[ index ].available )
    
    return gift;

  end;

  if ( not gift_data ) then

    local gifts = {};
    local stage = private.GetCurrentStage();
    local last_stage = 110 * private.GetCurrentSection();
    local pivot = ( force and ( stage + 3 ) >= ( last_stage - 1 ) ) and 
                  last_stage or stage;

    for i = 1, #rewards do
      local gift = func_generate_gift( rewards[ i ].name, pivot, i );
      if ( gift ) then
        pivot = gift.stage;
        gifts[ gift.name ] = gift;
      end;
    end;

    return gifts;

  else

    local gifts_assoc = { woodchest = 1, silverchest = 2, goldchest = 3 };
    local index = gifts_assoc[ gift_data.name ];
    local gift = func_generate_gift( gift_data.name, gift_data.pivot, index );

    return gift;

  end;

end;
--------------------------------------------------------------------------------
function private.Gift_GenerateItems ( gift_data )

  local func_random = function ( items, item )

    item.count = item.count or math.random( item.min, item.max );
    item.min, item.max = nil, nil;
    table.insert( items, item );

  end;

  local items = {};

  if ( gift_data.default ) then

    for i = 1, #gift_data.default do
      local item = gift_data.default[ i ];
      func_random( items, item );
    end;

  end;

  for i = #items + 1, gift_data.count do
    local item = table.remove( gift_data.available, math.random( 1, #gift_data.available ) );
    func_random( items, item );
  end;

  return items;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Crown_GetValue ()
  
  if ( not private.crown_value ) then
    local event_list = public.GetEventList();
    
    private.crown_value = 0;
    
    for index, event in pairs( event_list ) do
      if ( event.type == "progress" ) then
        local saved_data = private.Event_GetSavedData( event.id );
        if ( saved_data and saved_data.progress ) then
          private.crown_value = saved_data.progress;
        else
          private.crown_value = 0;
        end;
      end;
    end;
  end;
  
  return private.crown_value;
  
end;
--------------------------------------------------------------------------------
function public.Crown_SetValueProgress ( event_id, value )
  
  local max_crowns = private.crown_static_data.bonuses.count;
  local crowns = math.min( value, max_crowns );
  private.crown_value = crowns;
  
  m_minievent.Event_SetProgress( event_id, crowns );
  
end;
--------------------------------------------------------------------------------
function public.Crown_GetStatsBonuses ()

  if ( public.Crown_GetValue() == 0 ) then
    return nil;
  end;
  
  local bonuses = {};
  local bonuses_data = private.crown_static_data.bonuses;
  
  for i = 1, public.Crown_GetValue() do
    local bonus = bonuses_data[ "crown_"..i ][ 1 ];
    if ( #bonuses_data[ "crown_"..i ] > 1 ) then
      local rand_index = math.random( #bonuses_data[ "crown_"..i ] );
      bonus = bonuses_data[ "crown_"..i ][ rand_index ];
    end;
    table.insert( bonuses, bonus );
  end;
  
  return bonuses;
  
end;
--------------------------------------------------------------------------------