-- name=m_player
--------------------------------------------------------------------------------
function public.__InitModule ()
  
  private.section = 1;
  
  private.platform = GetPlatform();
  private.platform = tern( private.platform == "android" )( 
    wiz.Android.GetPlatform(), 
    private.platform
  );

  private.default_name = ne.strings.Get( "str_player_name" );
  
  private.attempts_video = {
    token = "tmr_player_attempts_video",
    event = "Event_Player_AttemptsVideo"
  };
  
  private.AttemptsVideo_InitCheck();
  
  ne.msg.Subscribe(
    private.attempts_video.event,
    private.AttemptsVideo_FinishTimer
  );
  
  --Log.i( "Initialized." );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Storage_GetAttemtsVideo ()
  
  return lib_db.Player.Info().video_attempts or 0;
  
end;
--------------------------------------------------------------------------------
function private.Storage_UpdateData ( attempt_count )
  
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        =
    {
      video_attempts = attempt_count
    }
  };

  return lib_db.Execute( query ).result;
  
end;
--------------------------------------------------------------------------------
function public.GetAttemptsVideo ()
  
  local count = private.Storage_GetAttemtsVideo();
  local max_count = public.Settings( "player" ).video_attempts.count or 3;
  
  return count, max_count;
  
end;
--------------------------------------------------------------------------------
function public.UpdateAttemptsVideo ( attempt_count )
  
  local result = private.Storage_UpdateData( attempt_count );
  
  if ( attempt_count > 0 ) and ( not private.AttemptsVideo_IsExistTimer() ) then
    private.AttemptsVideo_StartTimer();
  end;
  
  return result;
  
end;
--------------------------------------------------------------------------------
function private.AttemptsVideo_InitCheck ()
  
  local attempt_count = public.GetAttemptsVideo();
  
  if ( attempt_count > 0 ) and ( not private.AttemptsVideo_IsExistTimer() ) then
    public.UpdateAttemptsVideo( 0 );
  end;
  
end;
--------------------------------------------------------------------------------
function private.AttemptsVideo_StartTimer ()
  
  local time = public.Settings( "player" ).video_attempts.time or 86400;
  local time_now = TimeServerNow();
  local time_recover = time - ( time_now.tm_hour * 3600 + time_now.tm_min * 60 + time_now.tm_sec );
  
  TimeSheduleEvent( 
    private.attempts_video.token,
    "GLOBAL_TIME",
    time_recover,
    private.attempts_video.event,
    {}
  );
  
end;
--------------------------------------------------------------------------------
function private.AttemptsVideo_FinishTimer ()
  
  private.AttemptsVideo_Unshedule();
  public.UpdateAttemptsVideo( 0 );
  private.AttemptsVideo_MsgNotify();
  
end;
--------------------------------------------------------------------------------
function private.AttemptsVideo_Unshedule ()
  
  TimeFinishEvent();
  TimeUnsheduleEvent( private.attempts_video.token );
  
end;
--------------------------------------------------------------------------------
function private.AttemptsVideo_MsgNotify ()
  
  mediator.Notify( "Model_Player_RecoverFinishedAttemptsVideo" );
  
end;
--------------------------------------------------------------------------------
function private.AttemptsVideo_IsExistTimer ()
  
  return TimeIsEventExist( private.attempts_video.token );
  
end;
--------------------------------------------------------------------------------
--function .. () end;
--------------------------------------------------------------------------------
function public.PurchaseExecute ()
  
  local time_now = TimeServerNow();
  
  public.Storage_UpdateVideoDataDays( time_now.tm_yday );
  
end;
--------------------------------------------------------------------------------
function private.Storage_GetVideoTimeoutStartDay ()
  
  return lib_db.Player.Info().day_start;
  
end;
--------------------------------------------------------------------------------
function public.Storage_UpdateVideoDataDays ( day )
  
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        =
    {
      day_start = day
    }
  };

  return lib_db.Execute( query ).result;
  
end;
--------------------------------------------------------------------------------
function public.VideoTimeout_IsFinished ()
  
  local day_start = private.Storage_GetVideoTimeoutStartDay();
  
  if ( not day_start ) then
    return true;
  end;
  
  local days = public.Settings( "player" ).video_timeout_days or 30;
  local time_now = TimeServerNow();
  local now_day = time_now.tm_yday;
  
  if ( now_day < day_start ) then
    now_day = now_day + 365;
  end;
  
  day_start = day_start + days;
  
  if ( now_day >= day_start ) then
    return true;
  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
--function .. () end;
--------------------------------------------------------------------------------
function public.SetSection ( section_number )

  private.section = section_number;
  
end;
--------------------------------------------------------------------------------
function public.Settings ( key )

  local  settings = lib_db.Settings( key );
  return settings;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Name_GetLocal ()

  return lib_db.Player.Info().name or private.default_name;

end;
--------------------------------------------------------------------------------
function public.Name_SetLocal ( player_name )

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        = 
    {
      name = player_name
    }
  };

  if ( lib_db.Execute( query ).result ) then

    mediator.Notify( "Model_PlayerNameChanged", player_name );
  
  end;

end;
--------------------------------------------------------------------------------
function public.Name_TrySetSocial ( player_name, silent )
  
  --DbgTrace( "[ DEBUG ]  start m_player SetPlayerName  " .. name_new);
   
  lib_rpc.Schedule(
    "user",
    "UpdateUserName",
    { name = player_name },
    function ( message_type, message_params )
      
      --DbgTrace( "[ DEBUG ]  SetPlayerName returned" );
      
      if ( not message_params.result ) then
        
         mediator.Notify( "Model_Player_NameChangeFailed" );
         
         return;
         
      end;
       
      private.Name_SetLocal( message_params.name );
      
    end, 
    silent,
    true
  );
   
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Skin_TryObtainPlatformSkins ( platform )

  local query =
  {
    query      = "find_all",
    group      = "static",
    collection = "skins",
    indice     = "platform",
    criteria   = { like = platform }
  };

  local platform_player = lib_db.Execute( query ).result;
  
  for i = 1, #platform_player do
    
    local platform_skin = platform_player[ i ];
    
    if ( not platform_skin.parts ) or ( platform_skin.parts == 0 ) then
      
      query = 
      {
        query      = "find",
        group      = "player",
        collection = "skins",
        indice     = "name",
        criteria   = { like = platform_skin.name }
      };
      
      if not ( lib_db.Execute( query ).result ) then

        query = 
        {
          query      = "insert",
          group      = "player",
          collection = "skins",
          value      = {
            name    = platform_skin.name
          }
        };
        lib_db.Execute( query );
     
        v_log.Kontagent_LogEvent( "skins", 
        {
          description = "skin_collect",
          category    = "",
          value       = 1,
          name        = platform_skin.name
        } );

        mediator.SubscribeOnce(
          "View_Interface_Visible",
          function ()

            -- JK: send messages strictly in this order

            mediator.Notify(
              "Model_Skins_PlatformSkinCollected",
              platform_skin.name
            );

            mediator.Notify(
              "Model_Skins_SkinCollected",
              platform_skin.name
            );

          end
        );
        
      end;

    end;
    
  end;

end;
--------------------------------------------------------------------------------
function public.Skin_GetCurrent ()

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "hero"
  };

  local skin = lib_db.Execute( query ).result.skin;
  local skin_data = public.Skin_QueryStatic( skin );
  
  if ( not skin_data ) then
    
    DbgTrace( "[ ERROR ]: No static data for skin < "..skin.." >. Return default." );
    return "default";
    
  end;

  if ( not skin_data.platform ) then
    if ( skin_data.excluded == private.platform ) then
      return "default";
    end;
    return skin;
  end;
  
  skin = tern( skin_data.platform == private.platform )( skin, "default" );
  
  return skin;
  
end;
--------------------------------------------------------------------------------
function public.Skin_GetPackageName ( skin_name )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "skins",
    indice     = "name",
    criteria   = { like = skin_name }
  };

  return lib_db.Execute( query ).result.content;

end;
--------------------------------------------------------------------------------
function public.Skin_GetByInviteType ( invite_type )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "skins",
    indice     = "invite",
    criteria   = { like = invite_type }
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function public.Skin_TrySet ( skin_name )

  local db_info   = public.Skin_QueryStatic( skin_name );
  local save_info = public.Skin_QueryStorage( skin_name );

  local exists = false;
  if ( db_info ) then exists = true; end;

  local partless = exists and ( not db_info.parts or db_info.parts == 0 );

  local saved = false;
  if ( save_info ) then saved = true; end;

  local completed = exists and saved
                    and db_info.parts and save_info.parts 
                    and save_info.parts >= db_info.parts;

  if ( exists                ) and
     ( partless or completed ) then
    
    v_log.Kontagent_LogEvent( "skins", 
    {
      description = "skin_switch",
      category    = public.Skin_GetCurrent(),
      name        = skin_name
    } );

    local query =
    {
      query      = "update",
      group      = "player",
      collection = "hero",
      set        = { skin = skin_name }
    };
    
    if ( lib_db.Execute( query ).result ) then

      mediator.Notify(
        "Model_Skins_SkinSet",
        skin_name
      );
      
      return true;

    end;

    return false;

  else

    return false;

  end;

end;
--------------------------------------------------------------------------------
function public.Skin_IsPartObtainable ( skin_name, part_name, count )

  local skin_data  = public.Skin_QueryStatic( skin_name );
  local saved_data = public.Skin_QueryStorage( skin_name );

  if ( skin_data ) then

    local current = ( saved_data or {} ).parts or 0;
    local total   = skin_data.parts;

    if ( current >= total ) then return false; end;

    return true;

  end;

  return false;
    
end;
--------------------------------------------------------------------------------
function public.Skin_TryObtainParts (
    skin_name,
    part_name,
    count,
    category,
    silent
  )

  local skin_data  = public.Skin_QueryStatic( skin_name );
  local saved_data = public.Skin_QueryStorage( skin_name );

  if ( skin_data ) then

    local current = ( saved_data or {} ).parts or 0;
    local total   = skin_data.parts;

    if ( current >= total ) then return false; end;

    local new_count = math.min( current + count, total );

    if ( saved_data ) then

      saved_data.parts = new_count;

      local query = 
      {
        query      = "update",
        group      = "player",
        collection = "skins",
        indice     = "name",
        criteria   = { like = skin_name },
        set        = saved_data
      };
      
      lib_db.Execute( query );

    else

      local query = 
      {
        query      = "insert",
        group      = "player",
        collection = "skins",
        value      = {
          name    = skin_name,
          parts   = new_count
        }
      };
      
      lib_db.Execute( query );

    end;

    local kontagent_description = "token_collect";
    local kontagent_category = category or "unknown";
    local kontagent_value = count;
    
    v_log.Kontagent_LogEvent( "skins", 
    {
      description = kontagent_description,
      category    = kontagent_category,
      value       = kontagent_value,
      name        = skin_name,
      tokens      = new_count..", "..total 
    } );

    local is_skin_complete = ( new_count >= total );
    
    if ( is_skin_complete ) then
      
      kontagent_description = "skin_collect";
      kontagent_category = "";
      kontagent_value = 1;
      
      v_log.Kontagent_LogEvent( "skins", 
      {
        description = kontagent_description,
        category    = kontagent_category,
        value       = kontagent_value,
        name        = skin_name
      } );      

    end;

    if ( not silent ) then

      mediator.Notify(
        "Model_Skins_PartCollected",
        {
          skin = skin_name,
          part = part_name
        }
      );

      if ( is_skin_complete ) then
        mediator.Notify( "Model_Skins_SkinCollected", skin_name );
      end;

    end;
   
    return true;

  end;

  return false;

end;
--------------------------------------------------------------------------------
function public.Skin_Data ( skin_name )

  local skin;

  local player_skin = public.Skin_GetCurrent();
  local db_info     = public.Skin_QueryStatic( skin_name );
  local save_info   = public.Skin_QueryStorage( db_info.name );
  
  if ( db_info.parts ) then

    skin = {
      name   = db_info.name,
      bonus  = db_info.bonus,
      parts  = db_info.parts,
      weared = ( player_skin == skin_name )
    };

    if ( save_info ) and ( save_info.parts ) then

      skin.collected_parts = save_info.parts;

    else

      skin.collected_parts = 0;

    end;

    skin.wearable = ( skin.collected_parts >= skin.parts );

  else

    skin = {
      name  = db_info.name,
      bonus = db_info.bonus
    };

    skin.wearable = true;

  end;
  
  skin.price = db_info.price;
  
  if ( skin.price and ( not save_info ) ) then
    
    skin.wearable = false;
    
  end;

  skin.invite = db_info.invite;
  
  if ( skin.invite and ( not save_info ) ) then
    
    skin.wearable = false;
    
  end;

  skin.primary  = db_info.primary;
  skin.animated = db_info.animated;

  return skin;

end;
--------------------------------------------------------------------------------
function public.Skin_GetAllSkins ( func_GetSkinContent )

  local skins          = {};
  local name_list      = {};
  local player_skin    = public.Skin_GetCurrent();
  local skins_database = public.Skin_QueryStatic();

  local result, value = parts.storage.data.Read( "sml/allow_sections" );
  
  for s = 1, #skins_database do

    local db_info = skins_database[ s ];

    if ( db_info.excluded ~= private.platform     ) and 
       ( ( db_info.platform == private.platform ) or
         ( not db_info.platform                 ) ) then

      local content_section = tonumber( db_info.content:match( "%d+" ) );

      if ( content_section                      ) and
         ( content_section <= tonumber( value ) ) or 
         ( db_info.content:find( "minievent_" ) ) then
      
        local content = func_GetSkinContent( db_info.name );
        
        if ( not content ) then

          --DbgTrace(
          --  "[ INFO ][ m_player ] No content available for skin '"
          --  ..db_info.name.."', excluding."
          --);

        else
   
          local skin = {
            name     = db_info.name,
            bonus    = db_info.bonus,
            content  = content,
            wearable = false,
            weared   = ( player_skin == db_info.name ),
            price    = db_info.price,
            invite   = db_info.invite,
            animated = db_info.animated
          };

          db_info.parts = db_info.parts or 0;
          
          if ( db_info.parts ) then

            skin.parts           = db_info.parts;
            skin.collected_parts = 0;

            local save_info = public.Skin_QueryStorage( db_info.name );

            if ( save_info       ) and
               ( save_info.parts ) then
              
              skin.collected_parts = save_info.parts;
              
            end;

            if ( skin.collected_parts >= skin.parts ) then
              
              skin.wearable = true;
              
            end;

            if ( skin.price or skin.invite ) and
               ( not save_info             ) then
              
              skin.wearable = false;
              
            end;
            
          end;

          table.insert( skins, skin );
          table.insert( name_list, skin.name );

        end;

      end;
  
    end;

  end;

  return skins, name_list;

end;
--------------------------------------------------------------------------------
function public.Skin_QueryStatic ( skin_name )

  local query;

  if ( skin_name ) then

    query =
    {
      query      = "find",
      group      = "static",
      collection = "skins",
      indice     = "name",
      criteria   = { like = skin_name }
    };

  else

    query =
    {
      query      = "find_all",
      group      = "static",
      collection = "skins"
    };

  end;

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function public.Skin_QueryStorage ( skin_name )

  local query;

  if ( skin_name ) then

    query =
    {
      query      = "find",
      group      = "player",
      collection = "skins",
      indice     = "name",
      criteria   = { like = skin_name }
    };

  else

    query =
    {
      query      = "find_all",
      group      = "player",
      collection = "skins"
    };

  end;

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function public.Skin_Purchase ( skin_name )
  
  local query = 
  {
    query      = "find",
    group      = "player",
    collection = "skins",
    indice     = "name",
    criteria   = { like = skin_name }
  };
  
  if not ( lib_db.Execute( query ).result ) then

    query = 
    {
      query      = "insert",
      group      = "player",
      collection = "skins",
      value      = {
        name    = skin_name
      }
    };
    lib_db.Execute( query );
 
    v_log.Kontagent_LogEvent( "skins", 
    {
      description = "skin_collect",
      category    = "",
      value       = 1,
      name        = skin_name
    } );
    
  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Gold_GetCount ()

  return lib_db.Player.Info().gold;

end;
--------------------------------------------------------------------------------
function public.Gold_IsEnough ( required_count, notify )
  
  local available_count = public.Gold_GetCount();
  
  if ( available_count >= required_count ) then
  
    return true;
  
  end;
  
  if ( notify ) then
  
    mediator.Notify(
      "Model_Player_GoldCountChangeFailed",
      { 
        count = math.abs( available_count - required_count )
      }
    );
  
  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
function public.Gold_TryChangeCount ( delta, hidden )

  if ( delta == 0 ) then

    return false;

  end;

  local count_current = lib_db.Player.Info().gold;

  local count_new = count_current + delta;

  if ( count_new < 0 ) then

    mediator.Notify(
      "Model_Player_GoldCountChangeFailed",
      { 
        count = math.abs( count_new )
      }
    );

    return false;

  end;

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set = 
    {
      gold = count_new
    }
  };

  lib_db.Execute( query );
  
  mediator.Notify(
    "Model_Player_GoldCountChanged", 
    { 
      count  = count_new,
      delta  = delta,
      hidden = hidden
    }
  );

  return true;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Resource_GetCount ( resource_name, section_number )
  
  section_number = section_number or private.section;
  
  local result = nil;
  
  if ( resource_name == "resource4" ) then
    result = lib_db.Player.Info()[ resource_name ];
  else
    result = lib_db.Player.Info()[ resource_name ][ section_number ];
  end;
  
  if not ( result ) then
    
    local set = {};
    
    if ( resource_name == "resource4" ) then
      set[ resource_name ] = 0;
      result = 0;
    else
      set[ resource_name ] = lib_db.Player.Info()[ resource_name ];
      set[ resource_name ][ section_number ] = -1;
    end;

    local query =
    {
      query      = "update",
      group      = "player",
      collection = "info",
      set        = set
    };

    lib_db.Execute( query );
    
  end;
  
  return result or -1;
    
end;
--------------------------------------------------------------------------------
function public.Resource_IsEnough ( resource_name, required_count, notify )
  
  local available_count = public.Resource_GetCount( resource_name );
  
  if ( available_count >= required_count ) then
  
    return true;
  
  end;
  
  if ( notify ) then
  
    mediator.Notify(
      "Model_Player_ResourceCountChangeFailed",
      {
        name  = resource_name,
        count = math.abs( available_count - required_count )
      }
    );
  
  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
function private.GetCountCurrentResource ( resource, section )
  
  local count_current = 0;
  
  if ( resource ~= "resource4" ) then
    count_current = lib_db.Player.Info()[ resource ][ section ];
  else
    count_current = lib_db.Player.Info()[ resource ] or 0;
  end;
  
  count_current = tern( count_current < 0 )( 0, count_current );
  
  return count_current;
  
end;
--------------------------------------------------------------------------------
function private.SetCountCurrentResource ( resource, section, count_new )
  
  local set = {
    [ resource ] = lib_db.Player.Info()[ resource ]
  };

  if ( resource ~= "resource4" ) then
    set[ resource ][ section ] = count_new;
  else
    set[ resource ] = count_new;
  end;
  
  return set;
  
end;
--------------------------------------------------------------------------------
function public.Resource_TryChangeCount ( resource, delta, hidden, section )

  if ( delta == 0 ) then
    return false;
  end;

  local section_mismatch = tern( section )(
    ( private.section ~= section ),
    false
  );

  section = section or private.section;

  local count_current = private.GetCountCurrentResource( resource, section );

  local count_new = count_current + delta;

  if ( count_new < 0 ) then

    mediator.Notify(
      "Model_Player_ResourceCountChangeFailed",
      {
        name  = resource,
        count = math.abs( count_new )
      }
    );

    return false;

  end;

  local set = private.SetCountCurrentResource( resource, section, count_new );

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        = set
  };

  lib_db.Execute( query );

  -- JK: don't notify and buffer if player is another section
  if ( section_mismatch ) then

    return true;

  end;

  mediator.Notify(
    "Model_Player_ResourceCountChanged",
    {
      count  = count_new,
      name   = resource,
      delta  = delta,
      hidden = hidden
    }
  );

  return true;

end;
--------------------------------------------------------------------------------
function public.Resource_TryChangeMultipleCount ( resource_data )

  local deficient = {};
  
  for i = 1, #resource_data do

    local count_current = public.Resource_GetCount( resource_data[ i ].name );
    
    if ( count_current < 0 ) then

      count_current = 0;

    end;

    local count_new = count_current + resource_data[ i ].delta;

    if ( count_new < 0 ) then

      table.insert(
        deficient,
        {
          name  = resource_data[ i ].name,
          count = math.abs( count_new )
        }
      );

    end;

  end;

  if ( #deficient > 0 ) then

    mediator.Notify(
      "Model_Player_ResourceCountChangeFailed",
      deficient
    );

    return { result = false };

  end;

  for i = 1, #resource_data do

    public.Resource_TryChangeCount(
      resource_data[ i ].name,
      resource_data[ i ].delta
    );

  end;

  return { result = true, resources = resource_data };

end;
--------------------------------------------------------------------------------
function public.Resource_ClearAllForSection ( section_name )
  
  local resource = {};
  
  resource[ "resource1" ] = lib_db.Player.Info()[ "resource1" ];
  resource[ "resource1" ][ section_name ] = -1;
  
  resource[ "resource2" ] = lib_db.Player.Info()[ "resource2" ];
  resource[ "resource2" ][ section_name ] = -1;
  
  resource[ "resource3" ] = lib_db.Player.Info()[ "resource3" ];
  resource[ "resource3" ][ section_name ] = -1;

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        = resource
  };

  lib_db.Execute( query );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Progress_GetLastStageName ()

  return lib_db.Player.Info().stage_last;

end;
--------------------------------------------------------------------------------
function public.Progress_SetLastStageName ( stage_name, section_number, is_stage_completed )

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        =
    {
      section    = section_number,
      stage_last = stage_name
    }
  };

  if ( lib_db.Execute( query ).result ) then
    
    if ( not is_stage_completed ) then
      mediator.Notify(
        "Model_Player_StageLastSet",
        stage_name
      );
    end;

    return true;

  end;
  
  return false;

end;
--------------------------------------------------------------------------------
function public.Progress_SetStageState ( stage_name, stage_data )
  
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "stages",
    indice     = "name",
    criteria   = { like = stage_name },
    set        = stage_data
  };

  lib_db.Execute( query );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.PayingStatus_Set ( price )

  local ispaying = lib_db.Player.Info().ispaying or 0;
  ispaying = tern( ispaying == true )( 0.99, ispaying );
   
  ispaying = ispaying + tonumber( price:match( "$%s*(.+)" ) );
  
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        =
    {
      ispaying = ispaying
    }
  };

  local result = lib_db.Execute( query ).result;

  return result;

end;
--------------------------------------------------------------------------------
function public.PayingStatus_Get ()

  local ispaying = lib_db.Player.Info().ispaying or 0;
  
  ispaying = tern( ispaying == true )( 0.99, ispaying );
  
  local paying_status = 1;
  
  paying_status = tern( ispaying > 0     )( 2, paying_status );
  paying_status = tern( ispaying > 0.99  )( 3, paying_status );
  paying_status = tern( ispaying > 9.99  )( 4, paying_status );
  paying_status = tern( ispaying > 49.99 )( 5, paying_status );
  
  return paying_status;
  
end;
--------------------------------------------------------------------------------
function public.PayingStatus_GetLastPurchaseTime ()

  local last_purchase_time = lib_db.Player.Info().last_purchase or false;
  
  return last_purchase_time;
  
end;
--------------------------------------------------------------------------------
function public.PayingStatus_SetLastPurchaseTime ( seconds )

  if( type( seconds ) ~= "number" ) then return false; end;
  
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        =
    {
      last_purchase = seconds
    }
  };

  local result = lib_db.Execute( query ).result;

  return result;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Metrics_GetLaunchData ()

  local  launch_data = lib_db.Player.Info().game_launch;
  return launch_data;

end;
--------------------------------------------------------------------------------
function public.Metrics_SetLaunchData ( this_launch_time )

  local day = 86400;

  local launch_data = public.Metrics_GetLaunchData();
  local server_now  = parts.time.server.Now();

  if ( not launch_data ) then

    launch_data = {
      time_this = this_launch_time,
      time_last = this_launch_time,
      days_last = this_launch_time,
      days_row  = 0,
      days_yday = server_now.tm_yday
    };

  else

    local last_launch_time = launch_data.time_this;
    local days_diff        = this_launch_time - launch_data.days_last;
    local yday             = server_now.tm_yday;

    launch_data.time_this = this_launch_time;
    launch_data.time_last = last_launch_time;

    if ( days_diff > day               ) or
       ( yday ~= launch_data.days_yday ) then

      local row = launch_data.days_row;

      -- JK: user rewinded time on device for cheating
      if ( this_launch_time < last_launch_time ) then
        launch_data.days_row = 0;
      else
        launch_data.days_last = this_launch_time;
        launch_data.days_row  = tern( days_diff < 2*day )( row + 1, 0 );
      end;

      local is_changed = ( row ~= launch_data.days_row );

      if ( is_changed ) then
        mediator.Notify( "Model_Player_LaunchDaysRow", launch_data.days_row );
      end;

    end;    

    launch_data.days_yday = yday;

  end;
  
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        =
    {
      game_launch = launch_data
    }
  };

  local  is_set = lib_db.Execute( query ).result;
  return is_set;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
-- Функция возвращает код для приглашения другого игрока в игру, либо nil если
-- в данный момент эта функция отключена или не возможна
--------------------------------------------------------------------------------
function public.FriendCode_GetInviteCode ()
  
  local userId = parts.auth.GetUserId();
  
  if ( not userId   ) or
     ( userId == "" ) then
    
    return nil, true;
    
  end;
  
  local code = string.upper( string.format( "%x", tonumber( userId ) ) );
  code = string.gsub( code, "0", "G" );
  code = string.gsub( code, "1", "H" );
  
  if ( IsMobile() ) then
    code = "I"..code;
  else
    code = "E"..code
  end;
  
  return code, true;
  
end;
--------------------------------------------------------------------------------
-- Функция принимает код-строку, и пытается выдать награду для этого кода
-- Возвращает bool - был ли выполнен запрос к серверу
--------------------------------------------------------------------------------
function public.FriendCode_RedeemCode ( sFriendCode )
  
  if ( not sFriendCode or 
           sFriendCode == "" ) then
    
    return false, "code_error";
    
  end;
  
  local code = string.upper( sFriendCode );
  local prefix = string.sub( code, 1, 1 );
  
  if ( prefix == "I" and not IsMobile() ) then
    return false, "code_error";
  end;
  
  if ( prefix == "E" and IsMobile() ) then
    return false, "code_error";
  end;
  
  if ( ( prefix == "I" or prefix == "E" ) and 
         not lib_db.Player.Info().iscaninvited ) then

    return false, "code_used";
    
  end;

  -- JK: look for code applied in this session
  if ( code == private.code_applied ) then

    return false, "code_used";

  end;
  
  local onResponse = function ( message_type, message_data )

    local result = false;
    local reason = "code_error";

    if ( message_data      ) and
       ( message_data.code ) then

      local code = string.upper( message_data.code );

      result = message_data.result;
      reason = tern( result )(
        "code_success",
        "code_error"
      );

      if ( reason == "code_error" ) then

        reason = tern( message_data.reason == "error_already_used" )(
          "code_used",
          "code_error"
        );

      end;

      if ( reason == "code_used" ) or
         ( result                ) then
    
        -- JK: save in this session to reduce server load
        private.code_applied = code;

      end;

      if ( result ) then

        if ( prefix == "I" or prefix == "E" ) then

          private.FriendCode_DisableInvited();

        else
          
          v_log.Kontagent_LogEvent( "promocode", 
          {
            description = "redeem",
            name        = code
          } );
          
        end;

      end;
        
    end;
      
    mediator.Notify(
      "Event_Player_FriendCodeApplied",
      {
        result = result,
        reason = reason
      }
    );
    
  end;

  lib_rpc.Schedule( 
    "Social",
    "RedeemFriendCode",
    { code = code },
    onResponse,
    false,
    true
  );
    
  return true, "code_accepted";
  
end;
--------------------------------------------------------------------------------
function public.FriendCode_GetFeatureUnlocker ()
  
  return lib_db.Settings( "friend_codes" ).unlock_stage;
  
end;
--------------------------------------------------------------------------------
function private.FriendCode_GetSettings ()
  
  return lib_db.Settings( "friend_codes" );
  
end;
--------------------------------------------------------------------------------
function private.FriendCode_DisableInvited ()
  
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        =
    {
      iscaninvited = false
    }
  };

  if not ( lib_db.Execute( query ).result ) then

    query =
    {
      query      = "insert",
      group      = "player",
      collection = "info",
      value      = 
      {
        iscaninvited = false
      }
    };

    lib_db.Execute( query );

  end;
  
end;
--------------------------------------------------------------------------------