-- name=m_potions
--------------------------------------------------------------------------------
function public.__InitModule ()
  
  private.token_name = "mte_potions_";
  
  ne.msg.Subscribe( "Event_Game_Potions_Completed", private.Completed );
  
  --Log.i( "Initialized." );
  
end;
--------------------------------------------------------------------------------
-- function ...()
--------------------------------------------------------------------------------
function private.GetStaticData ( potion_name )
  
  local query;
  
  if ( potion_name ) then
    
    query = 
    {
      query      = "find",
      group      = "static",
      collection = "potions",
      indice     = "name",
      criteria   = { like = potion_name }
    };
    
  else

    query = 
    {
      query      = "find_all",
      group      = "static",
      collection = "potions"
    };
    
  end;
  
  local static_data = lib_db.Execute( query ).result;
  
  return static_data;
  
end;
--------------------------------------------------------------------------------
function private.GetStorageData ( potion_name )
  
  local query = 
  {
    query      = "find",
    group      = "player",
    collection = "potions",
    indice     = "name",
    criteria   = { like = potion_name }
  };

  local storage_data = lib_db.Execute( query ).result;
  
  return storage_data;
  
end;
--------------------------------------------------------------------------------
function private.GetStorageDataOrDefault ( potion_name, default )
  
  local default_data = {
    name = potion_name,
    count = 0
  };
  
  if ( default ) then
    return default_data;
  end;
  
  local storage_data = private.GetStorageData( potion_name );
  
  if ( storage_data and ( not default ) ) then
    return storage_data;
  end;
  
  return default_data;
  
end;
--------------------------------------------------------------------------------
function public.GetFullData ( potion_name )
  
  local static_data = private.GetStaticData( potion_name );
  
  if ( potion_name ) then
    
    local storage_data = private.GetStorageDataOrDefault( potion_name );
    
    return private.GetUnionData( static_data, storage_data );
    
  else
    
    local full_data = {};
    
    for i = 1, #static_data do
      local potion_name = static_data[ i ].name;
      local storage_data = private.GetStorageDataOrDefault( potion_name );
      
      table.insert( full_data, private.GetUnionData( static_data[ i ], storage_data ) );
    end;
    
    return full_data;
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.GetUnionData ( static_data, storage_data )
  
  local union_data = {};
  
  for key, value in pairs( static_data ) do
    union_data[ key ] = value;
  end;
  
  for key, value in pairs( storage_data ) do
    union_data[ key ] = value;
  end;
  
  return union_data;
  
end;
--------------------------------------------------------------------------------
-- function ...()
--------------------------------------------------------------------------------
function public.AddPotion ( potion_name, count )
  
  private.UpdatePotion( potion_name, count );
  
end;
--------------------------------------------------------------------------------
function public.ActivatePotion ( potion_name )
  
  private.UpdatePotion( potion_name, -1 );
  
  private.Run( potion_name );
  
end;
--------------------------------------------------------------------------------
function private.UpdatePotion ( potion_name, count )
  
  local storage_data = private.GetStorageData( potion_name );
  
  if ( not storage_data ) then
    
    storage_data = private.GetStorageDataOrDefault( potion_name, true );
    
    storage_data.count = math.max( count, 0 );
    
    local query = 
    {
      query      = "insert",
      group      = "player",
      collection = "potions",
      value      = storage_data
    };
    lib_db.Execute( query );
    
  else
  
    storage_data.count = storage_data.count + count;
    
    local query =
    {
      query      = "update",
      group      = "player",
      collection = "potions",
      indice     = "name",
      criteria   = { like = potion_name },
      set        = storage_data
    };

    lib_db.Execute( query );
    
  end;
  
  local data = {
    name = storage_data.name,
    count = storage_data.count,
    delta = count
  };
  
  mediator.Notify( "Model_Potion_CountUpdate", data );
  
end;
--------------------------------------------------------------------------------
-- function ...()
--------------------------------------------------------------------------------
function private.Run ( potion_name )
  
  local token = private.token_name..potion_name;
  local static_data = private.GetStaticData( potion_name );
  local time = static_data.time;
  local notify_msg = true;
  
  if ( TimeIsEventExist( token ) ) then
    notify_msg = false;
    time = TimeGetEventRemainingTime( token ) + static_data.time;
    private.Unshedule( token );
  else
    time = static_data.time;
  end;
  
  private.Shedule( token, time, potion_name );
  
  if ( notify_msg ) then
    mediator.Notify( "Model_Potion_RunAction", { name = potion_name } );
  end;
  
end;
--------------------------------------------------------------------------------
function private.Completed ( msg, data )
  
  local token = private.token_name..data.name;
  
  private.Unshedule( token );
  
  mediator.Notify( "Model_Potion_CompleteAction", { name = data.name } );
  
end;
--------------------------------------------------------------------------------
function private.Shedule ( token, time, potion_name )
  
  TimeSheduleEvent(
    token,
    "GLOBAL_TIME",
    time,
    "Event_Game_Potions_Completed",
    { name = potion_name }
  );
  
end;
--------------------------------------------------------------------------------
function private.Unshedule ( token )
  
  TimeFinishEvent();
  TimeUnsheduleEvent( token );
  
end;
--------------------------------------------------------------------------------
function public.IsPotionActive ( potion_name )
  
  local token = private.token_name..potion_name;
  
  if ( TimeIsEventExist( token ) ) then
    return true;
  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
function public.Potions_GetActiveList ()
  
  local potions_list = public.GetFullData();
  
  for i = #potions_list, 1, - 1 do
    
    local potion_name = potions_list[ i ].name;
    if ( not public.IsPotionActive ( potion_name ) ) then
      table.remove( potions_list, i );
    end;
    
  end;

  return potions_list;
  
end;
--------------------------------------------------------------------------------
function public.GetRemainingTime ( potion_name )
  
  local token = private.token_name..potion_name;
  local time = TimeGetEventRemainingTime( token );
  
  return time;
  
end;
--------------------------------------------------------------------------------