-- name=m_quests
--------------------------------------------------------------------------------
function public.__InitModule ()

  public.msg =
  {
    Started        = "Model_Quests_QuestStarted",
    Updated        = "Model_Quests_QuestUpdated",
    Completed      = "Model_Quests_QuestCompleted",
    ReachMaxLevels = "Model_Quests_QuestReachMaxLevelsStarted"
  };
  
  -- See private.Virtual_Setup
  private.virtual =
  {
    names  =
    {
      QT_StageCompleted        = "qst_complete_",
      QT_MineResourceCollected = "qst_mine_collect_resources_",
      QT_SpellCompleted        = "qst_complete_"
    },
    filter  = { stages = {}, mines = {}, spells = {} },
    section = nil,
    quests  = {}
  };

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Current_GetByName ( quest_name )
  
  return private.Current_GetByName( quest_name );

end;
--------------------------------------------------------------------------------
function public.Current_GetAllBySection ( section_name )
  
  return private.Current_GetAllBySection( section_name );

end;
--------------------------------------------------------------------------------
function public.Current_Start ( quest_data )
  
  return private.Current_Start( quest_data );
  
end;
--------------------------------------------------------------------------------
function public.Current_Complete( quest_data )
  
  return private.Current_Complete( quest_data );

end;
--------------------------------------------------------------------------------
function public.Completed_GetByName ( quest_name )
  
  return private.Completed_GetByName( quest_name );

end;
--------------------------------------------------------------------------------
function public.Completed_GetCount ()
  
  return private.Completed_GetCount();

end;
--------------------------------------------------------------------------------
function public.Completed_GetAll ()
  
  return private.Completed_GetAll();

end;
--------------------------------------------------------------------------------
function public.Static_GetByName ( quest_name )
  
  return private.Static_GetByName( quest_name );

end;
--------------------------------------------------------------------------------
function public.Static_GetAllByStartStage ( stage_name )
  
  return private.Static_GetAllByStartStage( stage_name );

end;
--------------------------------------------------------------------------------
function public.Static_GetAllByCompletedQuest ( quest_name )
  
  return private.Static_GetAllByCompletedQuest( quest_name );

end;
--------------------------------------------------------------------------------
function public.Static_GetAllPregive ()

  return private.Static_GetAllPregive();

end;
--------------------------------------------------------------------------------
function public.Static_GetAllBySection ( section_name )

  return private.Static_GetAllBySection( section_name );

end;
--------------------------------------------------------------------------------
function public.Virtual_Setup ( virtual_data, section_name )

  private.Virtual_Setup( virtual_data, section_name );

end;
--------------------------------------------------------------------------------
function public.Virtual_GetByName ( quest_name )

  local id = private.Virtual_GetIdByName( quest_name );
  if ( id ) then
  
    return private.virtual.quests[ id ];
  
  end;
  
  return nil;

end;
--------------------------------------------------------------------------------
function public.Virtual_GetAll ()

  return private.virtual.quests;

end;
--------------------------------------------------------------------------------
function public.Stages_FindByReward ( section_name, reward_type, reward_name )

  local quests = private.Static_GetAllBySection( section_name ) or {};
  
  if ( #quests < 1 ) then return quests; end;
  
  local list = {};
  
  for _, quest in pairs( quests ) do
  
    if ( quest.reward and quest.target.list ) then
    
      local current_data = private.Current_GetByName( quest.name );
      if ( current_data ) then
      
        local current_list = current_data.list or {};
        local targets_list = quest.target.list;
        
        for _, reward in pairs( quest.reward ) do
        
          if ( reward.type == reward_type ) then
          
            local name_is_valid = ( reward.name == reward_name );
            
            if ( reward_type == "fragment" ) then
              name_is_valid = reward.name:find( reward_name );
            end;
            
            if ( name_is_valid ) then
            
              for _, stage_name in pairs( targets_list ) do

                if ( not common.IsEntityInList( current_list, stage_name ) ) then

                  table.insert( list, stage_name );

                end;

              end;
              
            end;
          
          end;
        
        end;
      
      end;
    
    end;
  
  end;
  
  return list;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.HandleQuestEvent ( event, params, section_name )

  -- NOTE: params = { target_name = "target_name", list_item = "item_name" }
  params = params or {};

  local msg = "HandleQuestEvent: '%s'";
  Log.i( msg:format( event ) );
  
  if ( private.Current_CanStartByEvent( event ) ) then
  
    local msg = "HandleQuestEvent: Can start general quests by event '%s'"
    Log.i( msg:format( event ) );
    
    -- NOTE: Cache it?
    local quests_static = private.Static_GetAllBySection( section_name );
    for _, quest in pairs( quests_static ) do
    
      local start_event = private.Static_GetStartEvent( quest );
      if ( start_event and start_event == event ) then
        
        private.Current_TryStart( quest, params );
      
      end;
    
    end;
  
  end;
  
  if ( private.Virtual_CanStartByEvent( event ) ) then
  
    local msg = "HandleQuestEvent: Can start virtual quests by event '%s'"
    Log.i( msg:format( event ) );
  
    private.Virtual_TryStart( event, params );
  
  end;
  
  if ( private.Current_CanUpdateByEvent( event ) ) then
  
    local msg = "HandleQuestEvent: Can update general quests by event '%s'"
    Log.i( msg:format( event ) );
  
    local quests_current = private.Current_GetAllBySection( section_name );
    for _, quest in pairs( quests_current ) do
    
      local update_event = private.Current_GetUpdateEventByType( quest.type );
      if  ( update_event == event ) then
        private.Current_TryUpdate( quest, params );
      end;
    
    end;
  
  end;
  
  if ( private.Virtual_CanUpdateByEvent( event ) ) then
  
    local msg = "HandleQuestEvent: Can update virtual quests by event '%s'"
    Log.i( msg:format( event ) );
    
    local questlist = {};
    for _, quest in pairs( private.virtual.quests ) do
    
      local quest_type = quest.target.type;
      local events = private.Virtual_GetUpdateEventsByType( quest_type );
      
      if  ( common.IsEntityInList( events, event ) ) then
        table.insert( questlist, quest.name );
      end;
    
    end;
    
    private.Virtual_UpdateQuests( questlist, params, event );
  
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Current_CanUpdateByEvent ( event )

  local update_events =
  {
    BuildConstructed    = true,
    FragmentCollected   = true,
    MarketConstructed   = true,
    MineConstructed     = true,
    StageCompletedFull  = true,
    StageCompleted      = true,
    StageItemsCollected = true
  };
  
  return update_events[ event ] or false;

end;
--------------------------------------------------------------------------------
function private.Current_CanStartByEvent ( event )

  local start_events =
  {
    StageCompleted = true,
    QuestCompleted = true
  };
  
  return start_events[ event ] or false;

end;
--------------------------------------------------------------------------------
function private.Current_GetUpdateEventByType ( quest_type )

  local events =
  {
    QT_BuildConstructed    = "BuildConstructed",
    QT_FragmentCollected   = "FragmentCollected",
    QT_MarketConstructed   = "MarketConstructed",
    QT_MineConstructed     = "MineConstructed",
    QT_ReachMaxLevelStages = "StageCompletedFull",
    QT_StageCompleted      = "StageCompleted",
    QT_StageItemsCollected = "StageItemsCollected"
  };
  
  local event = events[ quest_type ];
  
  if ( event ) then
  
    return event;
  
  else
  
    local msg = "Error! Can't define update event for quest type '%s'";
    Log.e( msg:format( quest_type ), true );
    
    return "UnknownEvent";
  
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Current_TryUpdate ( quest, params )

  local condition_types =
  {
    QT_BuildConstructed    = { name = true,  list = false },
    QT_FragmentCollected   = { name = false, list = true  },
    QT_MarketConstructed   = { name = false, list = false },
    QT_MineConstructed     = { name = true,  list = false },
    QT_ReachMaxLevelStages = { name = false, list = true  },
    QT_StageCompleted      = { name = true,  list = false },
    QT_StageItemsCollected = { name = true,  list = true  }
  };
  
  local conditions = condition_types[ quest.type ];
  
  if ( conditions ) then
  
    local msg = "Current_TryUpdate: '%s'";
    Log.i( msg:format( quest.name ) );
  
    private.Current_TryUpdateByConditions( quest, params, conditions );
  
  else
  
    local msg = "Current_TryUpdate: Error! Quest '%s' has unknown type '%s'";
    Log.e( msg:format( quest.name, quest.type ), true );
  
  end;

end;
--------------------------------------------------------------------------------
function private.Current_TryUpdateByConditions ( quest, params, conditions )

  local tag = "Current_TryUpdateByConditions: "

  local quest_data = private.Static_GetByName( quest.name );
  local target     = quest_data.target;

  if ( conditions.list ) then
  
    if ( conditions.name and target.name ~= params.target_name ) then
      local msg = tag.."'%s' target list does not updated, name mismatch";
      Log.w( msg:format( quest.name ) );
      return;
    end;
    
    local static_list  = target.list;
    local current_list = quest.list or {};
    
    if ( common.IsEntityInList( static_list, params.list_item ) ) then
    
      local quest_updated = false;
      if ( not common.IsEntityInList( current_list, params.list_item ) ) then
      
        quest_updated = true;
        table.insert( current_list, params.list_item );
      
      end;
      
      if ( quest_updated ) then
      
        local msg = tag.."'%s' target list updated by '%s'";
        Log.i( msg:format( quest.name, params.list_item ) );
      
        if ( #static_list == #current_list ) then
        
          private.Current_Complete( quest_data );
        
        else
        
          private.Current_Update(
            quest.name, 
            { 
              list = current_list
            }
          );
        
        end;
      
      end;
    
    end
  
  elseif ( conditions.name ) then
  
    if ( target.name == params.target_name ) then
    
      private.Current_Complete( quest_data );
    
    end;
  
  else
  
    private.Current_Complete( quest_data );
  
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Current_TryStart ( quest_data, params )

  local msg = "Current_TryStart: '%s'";
  Log.i( msg:format( quest_data.name ) );

  if ( params.target_name ) then
  
    local condition = quest_data.condition;
    if ( not condition ) then return; end;
    
    if ( condition.name and condition.name == params.target_name ) then
    
      if ( private.Current_CanStart( quest_data.name, true ) ) then
          
        private.Current_Start( quest_data );
      
      end;
    
    elseif ( condition.list ) then
    
      for _, condition_name in pairs( condition.list ) do
      
        if ( condition.name == params.target_name ) then
        
          if ( private.Current_CanStart( quest_data.name, false ) ) then
          
            private.Current_Start( quest_data );
            break;
          
          end;
      
        end;
      
      end;
    
    end;
    
  else
  
    local msg = "Current_TryStart: Event '%s' does not contains 'target_name'";
    Log.w( msg:format( params.target_name ), true );
  
  end;

end;
--------------------------------------------------------------------------------
function private.Current_CanStart ( quest_name, strict )

  if ( private.Current_GetByName( quest_name ) ) then
  
    local msg = "Current_CanStart: Quest '%s' already started";
    Log.w( msg:format( quest_name ), strict );
    return false;
  
  end;
  
  if ( private.Completed_GetByName( quest_name ) ) then
  
    local msg = "Current_CanStart: Quest '%s' already completed";
    Log.w( msg:format( quest_name ), strict );
    return false;
  
  end;
  
  return true;

end;
--------------------------------------------------------------------------------
function private.IsEventMatchCondition ( event, start_condition )

  if not ( event and start_condition ) then
  
    return false;
    
  end;

  local conditions =
  {
    StageCompleted = "QSC_StageCompleted",
    QuestCompleted = "QSC_QuestCompleted"
  };
  
  return ( start_condition == conditions[ event ] );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Current_Start ( quest_data )

  local msg = "Current_Start: '%s'";
  Log.i( msg:format( quest_data.name ) );

  local query =
  {
    query      = "insert",
    group      = "player",
    collection = "quests_current",
    value      =
    {
      name    = quest_data.name,
      section = quest_data.section,
      type    = quest_data.target.type;
    }
  };
  
  lib_db.Execute( query );
  
  mediator.Notify( 
    public.msg.Started,
    {
      name   = quest_data.name,
      silent = false;
    }
  );
  
  if ( quest_data.target.type == "QT_ReachMaxLevelStages" ) then
  
    mediator.Notify( public.msg.ReachMaxLevels, quest_data.target.list );
  
  end;

end;
--------------------------------------------------------------------------------
function private.Current_Update ( quest_name, data )

  local msg = "Current_Update: '%s'";
  Log.i( msg:format( quest_name ) );

  local query = 
  {
    query      = "update",
    group      = "player",
    collection = "quests_current",
    indice     = "name",
    criteria   = { like = quest_name },
    set        = data
  };
  
  if ( not lib_db.Execute( query ).result ) then
  
    local msg = "Current_Update: Error! Can't update '%s'";
    Log.e( msg:format( quest_name ), true );
  
  end;
  
  mediator.Notify( public.msg.Updated, quest_name );

end;
--------------------------------------------------------------------------------
function private.Current_Complete ( quest_data )

  local msg = "Current_Complete: '%s'";
  Log.i( msg:format( quest_data.name ) );

  local query = 
  {
    query      = "delete",
    group      = "player",
    collection = "quests_current",
    indice     = "name",
    criteria   = { like = quest_data.name }
  };
  
  if ( not lib_db.Execute( query ).result ) then
  
    local msg = "Current_Complete: Error! Can't delete '%s' from current";
    Log.e( msg:format( quest_data.name ), true );
  
  end;
  
  if ( not private.Completed_GetByName( quest_data.name ) ) then
    
    query = 
    {
      query      = "insert",
      group      = "player",
      collection = "quests_completed",
      value      = { name = quest_data.name }
    };
    
    if ( not lib_db.Execute( query ).result ) then
    
      local msg = "Current_Complete: Error! Can't insert '%s' to completed";
      Log.e( msg:format( quest_data.name ), true );
    
    end;
    
    mediator.Notify(
      public.msg.Completed,
      { 
        name    = quest_data.name,
        reward  = quest_data.reward,
        section = quest_data.section
      }
    );
    
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Current_GetAllBySection ( section_name )

  local query =
  {
    query      = "find_all",
    group      = "player",
    collection = "quests_current",
    indice     = "section",
    criteria   = { like = section_name }
  };
  
  local current_quests = lib_db.Execute( query ).result or {};
  
  for i = 1, #current_quests do
    
    local quest = current_quests[ i ];
    
    quest.reward = private.Static_GetByName( quest.name ).reward;
    
  end;
  
  return current_quests;

end;
--------------------------------------------------------------------------------
function private.Current_GetByName ( quest_name )

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "quests_current",
    indice     = "name",
    criteria   = { like = quest_name }
  };
  
  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Completed_GetByName ( quest_name )

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "quests_completed",
    indice     = "name",
    criteria   = { like = quest_name }
  };
  
  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function private.Completed_GetAll ()

  local query =
  {
    query      = "find_all",
    group      = "player",
    collection = "quests_completed"
  };
  
  local quest_completed = lib_db.Execute( query ).result;

  return quest_completed;

end;
--------------------------------------------------------------------------------
function private.Completed_GetCount ()

  local quest_completed = private.Completed_GetAll();

  if ( not quest_completed ) then return 0;
  else                            return #quest_completed;
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Static_GetByName ( quest_name )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "quests",
    indice     = "name",
    criteria   = { like = quest_name }
  };
  
  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function private.Static_GetByTarget( target_name )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "quests",
    indice     = "target.name",
    criteria   = { like = target_name }
  };
  
  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function private.Static_GetAllBySection ( section_name )

  local query =
  {
    query      = "find_all",
    group      = "static",
    collection = "quests",
    indice     = "section",
    criteria   = { like = section_name }
  };

  return lib_db.Execute( query ).result or {};

end;
--------------------------------------------------------------------------------
function private.Static_GetAllPregive ()

  local query =
  {
    query      = "find_all",
    group      = "static",
    collection = "quests",
    indice     = "pregive",
    criteria   = { like = true }
  };
    
  return lib_db.Execute( query ).result or {};

end;
--------------------------------------------------------------------------------
function private.Static_GetAllByStartStage ( stage_name )

  local result = {};
  
  local query =
  {
    query      = "find_all",
    group      = "static",
    collection = "quests",
    indice     = "condition.type",
    criteria   = { like = "QSC_StageCompleted" }
  };

  local list = lib_db.Execute( query ).result;
  
  for i, quest in pairs( list ) do
  
    if ( quest.condition.name == stage_name ) then
    
      table.insert( result, quest );
    
    end;
  
  end;
  
  return result;

end;
--------------------------------------------------------------------------------
function private.Static_GetAllByCompletedQuest ( quest_name )

  local result = {};
  
  local query =
  {
    query      = "find_all",
    group      = "static",
    collection = "quests",
    indice     = "condition.type",
    criteria   = { like = "QSC_QuestCompleted" }
  };

  local list = lib_db.Execute( query ).result;
  
  for i, quest in pairs( list ) do
  
    if ( quest.condition.name == quest_name ) then
    
      table.insert( result, quest );
    
    end;
  
  end;
  
  return result;

end;
--------------------------------------------------------------------------------
function private.Static_GetStartEvent ( quest_data )

  if ( not quest_data.condition ) then return nil; end;

  local condition_type = quest_data.condition.type;
  local events =
  {
    QSC_StageCompleted = "StageCompleted",
    QSC_QuestCompleted = "QuestCompleted"
  };
  
  return events[ condition_type ];

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Virtual_Setup ( virtual_data, section_name )

  -- Setup filters
  private.virtual.filter.stages = virtual_data.stages.list;
  private.virtual.filter.mines  = virtual_data.mines.list;
  private.virtual.filter.spells = virtual_data.spells.list;
  
  private.virtual.section = section_name;
  
  -- Clear old
  private.virtual.quests = {};
  
  -- Create complete-next level quest
  local stage_last = virtual_data.stages.active;
  if ( stage_last and not private.Static_GetByTarget( stage_last ) ) then
  
    local quest_type = "QT_StageCompleted";
    local quest_data = private.Virtual_GetQuestData( quest_type, stage_last );
    table.insert( private.virtual.quests, quest_data );
  
  end;

  -- Create collect-resources quests
  for _, mine_name in pairs( virtual_data.mines.active ) do
  
    local quest_type = "QT_MineResourceCollected";
    local quest_data = private.Virtual_GetQuestData( quest_type, mine_name );
    table.insert( private.virtual.quests, quest_data );
  
  end;
  
  -- Create complete-spells quests
  for _, spell_name in pairs( virtual_data.spells.active ) do
  
    local quest_type = "QT_SpellCompleted";
    local quest_data = private.Virtual_GetQuestData( quest_type, spell_name );
    table.insert( private.virtual.quests, quest_data );
  
  end;

end;
--------------------------------------------------------------------------------
function private.Virtual_GetQuestData( quest_type, target_name )

  local name_prefix = private.virtual.names[ quest_type ];
  if ( not name_prefix ) then
  
    local msg = "Virtual_GetQuestData: Error! Unknown type '%s'";
    Log.e( msg:format( quest_type ), true );
    name_prefix = "qst_unknown_";
  
  end;

  local quest_data = 
  {
    name    = name_prefix..target_name,
    section = private.virtual.section,
    target  =
    {
      type = quest_type,
      name = target_name
    }
  };
  
  return quest_data;

end;
--------------------------------------------------------------------------------
function private.Virtual_CanUpdateByEvent ( event )

  local update_events =
  {
    StageCompleted        = true,
    StageCompletedMisc    = true,
    SpellCompleted        = true,
    MineResourceCollected = true
  };
  
  return update_events[ event ] or false;

end;
--------------------------------------------------------------------------------
function private.Virtual_CanStartByEvent ( event )

  local start_events =
  {
    StageEnabled = true,
    SpellEnabled = true,
    MineReady    = true
  };
  
  return start_events[ event ] or false;

end;
--------------------------------------------------------------------------------
function private.Virtual_GetUpdateEventsByType ( quest_type )

  local events_map =
  {
    QT_StageCompleted        = { "StageCompleted", "StageCompletedMisc" },
    QT_MineResourceCollected = { "MineResourceCollected" },
    QT_SpellCompleted        = { "SpellCompleted" }
  };
  
  local events = events_map[ quest_type ];
  
  if ( events ) then
  
    return events;
  
  else
  
    local msg = "Error! Can't define update event for quest type '%s'";
    Log.e( msg:format( quest_type ), true );
    
    return { "UnknownEvent" };
  
  end;

end;
--------------------------------------------------------------------------------
function private.Virtual_UpdateQuests( questlist, params, event )

  for _, quest_name in pairs( questlist ) do
  
    local id = private.Virtual_GetIdByName( quest_name );
    if ( id ) then
    
      local quest  = private.virtual.quests[ id ];
      local target = quest.target;
      
      if ( target.name == params.target_name ) then
      
        if ( target.type == "QT_StageCompleted" ) then
        
          if ( event == "StageCompletedMisc" ) then
            private.Virtual_Update( id );
          else
            private.Virtual_Complete( id );
          end;
        
        else
        
          private.Virtual_Complete( id );
        
        end;
      
      end;
    
    end;
  
  end;

end;
--------------------------------------------------------------------------------
function private.Virtual_TryStart ( event, params )

  local filter = private.virtual.filter;
  local target = params.target_name;

  if     ( event == "StageEnabled" ) then
  
    local stage_valid = common.IsEntityInList( filter.stages, target );
    if ( stage_valid and not private.Static_GetByTarget( target ) ) then
    
      local quest_type = "QT_StageCompleted";
      local quest_data = private.Virtual_GetQuestData( quest_type, target );
      private.Virtual_Start( quest_data, true );
    
    end;
  
  elseif ( event == "MineReady"  ) then
  
    if ( common.IsEntityInList( filter.mines, target ) ) then
    
      local quest_type = "QT_MineResourceCollected";
      local quest_data = private.Virtual_GetQuestData( quest_type, target );
      private.Virtual_Start( quest_data, false );
    
    end;
  
  elseif ( event == "SpellEnabled" ) then
  
    if ( common.IsEntityInList( filter.spells, target ) ) then
    
      local quest_type = "QT_SpellCompleted";
      local quest_data = private.Virtual_GetQuestData( quest_type, target );
      private.Virtual_Start( quest_data, false );
    
    end;
  
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Virtual_Start ( quest_data, silent )

  local msg = "Virtual_Start: '%s'";
  Log.i( msg:format( quest_data.name ) );

  table.insert( private.virtual.quests, quest_data );
  mediator.Notify( 
    public.msg.Started,
    {
      name   = quest_data.name,
      silent = silent;
    }
  );

end;
--------------------------------------------------------------------------------
function private.Virtual_Update ( id )

  local quest_name = private.virtual.quests[ id ].name;

  local msg = "Virtual_Update: '%s'";
  Log.i( msg:format( quest_name ) );

  mediator.Notify( public.msg.Updated, quest_name );

end;
--------------------------------------------------------------------------------
function private.Virtual_Complete ( id )

  local quest_data = private.virtual.quests[ id ];
  
  local msg = "Virtual_Complete: '%s'";
  Log.i( msg:format( quest_data.name ) );

  mediator.Notify(
    public.msg.Completed,
    { 
      name    = quest_data.name,
      reward  = quest_data.reward,
      section = quest_data.section
    }
  );
  
  table.remove( private.virtual.quests, id );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Virtual_GetIdByName ( quest_name )

  for id, quest in pairs( private.virtual.quests ) do
  
    if ( quest.name == quest_name ) then return id; end;
  
  end;
  
  return nil;

end;
--------------------------------------------------------------------------------