-- name=m_sections
--------------------------------------------------------------------------------
function public.__InitModule ()
  
  --Log.i( "Initialized." );
  
end;
--------------------------------------------------------------------------------
-- function ...() end;
--------------------------------------------------------------------------------
function public.GetEnabledCount ()
  
  local  enabled_count = lib_db.Settings( "sections_enable" );
  return enabled_count;

end;
--------------------------------------------------------------------------------
function public.List ()

  local query = 
  {
    query      = "find_all",
    group      = "player",
    collection = "sections"
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function public.StaticList ()

  local query = 
  {
    query      = "find_all",
    group      = "static",
    collection = "sections"
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
-- function ...() end;
--------------------------------------------------------------------------------
function public.Section_GetConfig ( section_name )

  local section_data = public.Section_StaticData( section_name );

  -- JK: split data by scopes
  local section_config = {
    characters = {
      start_characters = section_data.start_characters
    },
    fights = {
      start_fights = section_data.start_fights,
      boss         = section_data.boss,
      boss_blocker = section_data.boss_blocker
    },
    intro = section_data.intro,
    outro = section_data.outro,
    portal = {
      name    = section_data.portal,
      blocker = section_data.portal_blocker,
      bossed  = section_data.portal_bossed or false
    },
    starterpack = {
      unlock = section_data.starterpack_unlock
    }
  };

  return section_config;

end;
--------------------------------------------------------------------------------
function public.Section_StaticData ( section_name )
  
  local query =
  {
    query      = "find",
    group      = "static",
    collection = "sections",
    indice     = "name",
    criteria   = {
      like = section_name
    }
  };
  
  return lib_db.Execute( query ).result;
  
end;
--------------------------------------------------------------------------------
function public.Section_StorageData ( section_name )
  
  local query =
  {
    query      = "find",
    group      = "player",
    collection = "sections",
    indice     = "name",
    criteria   = {
      like = section_name
    }
  };
  
  return lib_db.Execute( query ).result;
  
end;
--------------------------------------------------------------------------------
function public.Section_IsEnabled ( section_name )
  
  local section_data = public.Section_StorageData( section_name ) or {};
  
  return section_data.enabled or false;
  
end;
--------------------------------------------------------------------------------
function public.Section_NeedIntro ( section_name )

  local section_data = public.Section_StorageData( section_name ) or {};
  
  local intro_completed = section_data.intro or false;

  if ( section_name == "section_1" and 
       not intro_completed ) then
    
    public.Section_SetIntro( section_name, true );
    intro_completed = true;
    
  end;

  return ( not intro_completed );

end;
--------------------------------------------------------------------------------
function public.Section_NeedOutro ( section_name )

  local section_data = public.Section_StorageData( section_name ) or {};

  return not section_data.outro or false;

end;
--------------------------------------------------------------------------------
function public.Section_SetIntro ( section_name, value )

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "sections",
    indice     = "name",
    criteria   = { like = section_name },
    set        = { intro = value }
  };

  return lib_db.Execute( query );

end;
--------------------------------------------------------------------------------
function public.Section_SetOutro ( section_name, value )

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "sections",
    indice     = "name",
    criteria   = { like = section_name },
    set        = { outro = value }
  };

  return lib_db.Execute( query );

end;
--------------------------------------------------------------------------------
function public.Section_Enable ( section_name )
    
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "sections",
    indice     = "name",
    criteria = { like = section_name },
    set      = { enabled = true }
  };

  lib_db.Execute( query );
  
  local number  = tonumber( string.match( section_name, "%d+" ) );
  lib_db.RegisterCollectionsBySection( number );

  mediator.Notify(
    "Model_Sections_SectionEnabled",
    {
      section_name   = section_name,
      first_stage    = public.Section_FirstStage( section_name ),
      section_config = public.Section_GetConfig( section_name )
    }
  );

end;
--------------------------------------------------------------------------------
function public.Section_NameByStage ( stage_name )

  local query = {
    query      = "find",
    group      = "static",
    collection = "stages",
    indice     = "name",
    criteria   = { like = stage_name }
  };
  
  return lib_db.Execute( query ).result.section;

end;
--------------------------------------------------------------------------------
function public.Section_Unlock ( section_name, is_enabled )
  
  local number  = tonumber( string.match( section_name, "%d+" ) );
  local enabled = is_enabled;

  local query =
  {
    query      = "insert",
    group      = "player",
    collection = "sections",
    value      = {
      name    = section_name,
      enabled = enabled
    }
  };

  if ( lib_db.Execute( query ).result ) then
  
    if ( enabled ) then
      
      lib_db.RegisterCollectionsBySection( number );
    
      mediator.Notify(
        "Model_Sections_SectionEnabled",
        {
          section_name   = section_name,
          first_stage    = public.Section_FirstStage( section_name ),
          section_config = public.Section_GetConfig( section_name )
        }
      );
    
    end;

    return true;
    
  end;

  return false;

end;
--------------------------------------------------------------------------------
function public.Section_FirstStage ( section_name )

  local query = 
  {
    query      = "find",
    group      = "static",
    collection = "stages",
    indice     = "section",
    criteria   = { like = section_name },
    result     = "name"
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function public.Section_LastStage ( section_name, mainline_only )

  local query = 
  {
    query      = "find_all",
    group      = "static",
    collection = "stages",
    indice     = "section",
    criteria   = { like = section_name }
  };

  local stages = lib_db.Execute( query ).result;
  
  if ( mainline_only ) then
    local main_stages = {};
    for _, stage in pairs( stages ) do
      if ( stage.type ~= "bonus" ) then
        table.insert( main_stages, stage );
      end;
    end;
    stages = main_stages;
  end;

  return stages[ #stages ].name;

end;
--------------------------------------------------------------------------------
function public.Stage_GetLastEnabled ()

  local query  = {
    query      = "find_all",
    group      = "player",
    collection = "stages"
  };
  
  local stages = lib_db.Execute( query ).result;

  for i = #stages, #stages - 5, - 1 do
    
    if ( stages[ i ].enabled ) then
      
      return stages[ i ].name;
      
    end;
    
  end;
  
end;
--------------------------------------------------------------------------------
-- function ...() end;
--------------------------------------------------------------------------------
function public.Player_GetEnabledList ()
  
  local query = 
  {
    query      = "find_all",
    group      = "player",
    collection = "sections"
  };

  local sections = lib_db.Execute( query ).result;
  
  local enabled_count = 0;
    
  for s = 1, #sections do
    
    if ( sections[ s ].enabled ) then
      
      enabled_count = enabled_count + 1;
      
    end;
    
  end;
  
  return enabled_count;

end;
--------------------------------------------------------------------------------