-- name=m_shop
--------------------------------------------------------------------------------
function public.__InitModule ()

  public.msg =
  {
    StarterpackStart  = "Model_Shop_StarterpackStart",
    StarterpackFinish = "Model_Shop_StarterpackFinish",
    MarketUnlocked    = "Model_Shop_MarketUnlocked",
    MarketConstructed = "Model_Shop_MarketConstructed",
    MarketRefresh     = "Model_Shop_MarketRefresh"
  };

  private.platform = GetPlatform();
  if ( private.platform == "android" ) then
    private.platform = wiz.Android.GetPlatform();
  end;

  -- Setup Starter-pack
  private.starterpack_settings = lib_db.Settings( "starterpacks" );

  private.starterpack_timer =
  {
    scope = "GLOBAL_TIME",
    event = "Event_Starterpack_Expired",
    token = "mte_starterpack_cooldown",
    time  = private.starterpack_settings.expire_time
  };
  
  ne.msg.Subscribe(
    private.starterpack_timer.event,
    function ( msg, params )
      public.Starterpack_Finish( params.name );
    end
  );
  
  -- Setup Market
  private.market_settings = lib_db.Settings( "market" );
  
  private.market_timer = 
  {
    scope = "GLOBAL_TIME",
    event = "Event_Market_RefreshOffers",
    token = "mte_market_refresh",
    time  = private.market_settings.refresh
  };
  
  private.feature = "prices";
 
  ne.msg.Subscribe( private.market_timer.event, private.Market_Refresh );
  
  -- Fix for missed market offers timer.
  if ( public.Market_IsBuilt() ) then
    if ( not parts.time.event.Exists( private.market_timer.token ) ) or
       ( private.market_timer.time < 1                             ) then
      private.Market_Refresh();
    end;
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Lot_GetStaticInfo ( lot_id )
  
  local query =
  {
    query      = "find",
    group      = "static",
    collection = "lots",
    indice     = "name",
    criteria   = { like = lot_id.."_"..private.platform }
  };

  local lot_data = lib_db.Execute( query ).result;
  
  lot_data = private.Gold_TryChangePrice( lot_data );
  
  return lot_data;
  
end;
--------------------------------------------------------------------------------
function public.Lot_GetIOSName ( lot_name )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "lots",
    indice     = "name",
    criteria   = { like = lot_name },
    result     = "name_ios"
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function public.Lot_GetNameById ( lot_id )
  
  return ( public.Lot_GetDataByIOSName( lot_id ) or {} ).name;

end;
--------------------------------------------------------------------------------
function public.Lot_GetDataByIOSName ( name_ios )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "lots",
    indice     = "name_ios",
    criteria   = { like = name_ios }
  };

  local lot_data = lib_db.Execute( query ).result;
  
  lot_data = private.Gold_TryChangePrice( lot_data );
  
  return lot_data;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Starterpack_GetLotByIndex ( starterpack_index )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "lots",
    indice     = "name",
    criteria   = { like = "starterpack_"..starterpack_index.."_"..private.platform }
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function public.Starterpack_Get ( lot_id )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "lots",
    indice     = "name_ios",
    criteria   = { like = lot_id }
  };

  local starterpack = lib_db.Execute( query ).result;
  
  if ( starterpack and type( starterpack.currency_count ) == "table" ) then
  
    return starterpack;
    
  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
function public.Starterpack_Start ( starterpack_name )

  --DbgTrace( "Starterpack Start: "..starterpack_name );

  TimeSheduleEvent(
    private.starterpack_timer.token,
    private.starterpack_timer.scope,
    private.starterpack_timer.time,
    private.starterpack_timer.event,
    {
      name = starterpack_name
    }
  );

  mediator.Notify(
    public.msg.StarterpackStart,
    {
      name = starterpack_name
    }
  );

end;
--------------------------------------------------------------------------------
function public.Starterpack_IsTimerExists ()

  return TimeIsEventExist( private.starterpack_timer.token );

end;
--------------------------------------------------------------------------------
function public.Starterpack_GetRemainingTime ()

  if ( not TimeIsEventExist( private.starterpack_timer.token) ) then
    
    return 0;
    
  end;
  
  return TimeGetEventRemainingTime( private.starterpack_timer.token );

end;
--------------------------------------------------------------------------------
function public.Starterpack_Finish ( starterpack_name )

  TimeFinishEvent();
  TimeUnsheduleEvent( private.starterpack_timer.token );
  
  mediator.Notify( public.msg.StarterpackFinish );

  if ( public.Starterpack_IsCompleted( starterpack_name ) ) then

    return;

  end;

  local starterpack_number = string.match( starterpack_name, "%d+" );
  local crossplatform_name = "starterpack_"..starterpack_number;
  
  local query =
  {
    query      = "insert",
    group      = "player",
    collection = "starterpacks",
    value      = 
    {
      name = crossplatform_name
    }
  };

  lib_db.Execute( query );

end;
--------------------------------------------------------------------------------
function public.Starterpack_IsCompleted ( starterpack_name )

  local starterpack_number = string.match( starterpack_name, "%d+" );
  local crossplatform_name = "starterpack_"..starterpack_number;

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "starterpacks",
    indice     = "name",
    criteria   = { like = crossplatform_name }
  };

  local starterpack = lib_db.Execute( query ).result;

  if ( not starterpack ) then

    return false;

  end;
  
  return true;

end;
--------------------------------------------------------------------------------
function public.Starterpack_Clear ( starterpack_name )
  
  local starterpack_number = string.match( starterpack_name, "%d+" );
  local crossplatform_name = "starterpack_"..starterpack_number;

  local query =
  {
    query      = "delete",
    group      = "player",
    collection = "starterpacks",
    indice     = "name",
    criteria   = 
    {
      like = crossplatform_name
    }
  };

  local is_deleted = lib_db.Execute( query ).result;

  return is_deleted;

end;
--------------------------------------------------------------------------------
function public.Starterpack_GetPayerTimeout ()

  return private.starterpack_settings.payer_timeout;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Gold_GetLots ( lot_type )

  local query =
  {
    query      = "find_all",
    group      = "static",
    collection = "lots"
  };

  local lot_data_list = lib_db.Execute( query ).result;

  local lots_list = {};

  for i = 1, #lot_data_list do

    local lot = lot_data_list[ i ];

    if  ( lot.type == "shop" )
    and lot.name:find( lot_type ) then

      lot = private.Gold_TryChangePrice( lot );
      table.insert( lots_list, lot );

    end;

  end;

  return lots_list;

end;
--------------------------------------------------------------------------------
function private.Gold_TryChangePrice ( lot )
  
  if ( parts.features.IsActive( private.feature ) ) then
    
    local feature_data = parts.features.GetFeatureData( private.feature );
    
    if ( feature_data and 
         feature_data[ 1 ] and 
         feature_data[ 1 ].lots ) then
         
      local index = common.IsEntityInList( feature_data[ 1 ].lots, 
                                           { name = lot.name },
                                           "name" );
      if ( index ) then
        local lot_data = feature_data[ 1 ].lots[ index ];
        if ( lot_data.new_count > lot.currency_count ) then
          lot.currency_count_bonus = lot_data.new_count - lot.currency_count;
        end;
      end;
      
    end;
    
  end;

  return lot;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Market_RefreshForced ()

  private.Market_Refresh();

end;
--------------------------------------------------------------------------------
function public.Market_GetSecondsToRefresh ()

  local  seconds = TimeGetEventRemainingTime( private.market_timer.token );
  return seconds;

end;
--------------------------------------------------------------------------------
function public.Market_Settings ( key )

  return private.market_settings[ key ];

end;
--------------------------------------------------------------------------------
function public.Market_IsUnlocked ()

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "market"
  };

  return lib_db.Execute( query ).result.unlock;

end;
--------------------------------------------------------------------------------
function public.Market_Unlock ()

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "market",
    set        = { unlock = true }
  };

  lib_db.Execute( query );
  mediator.Notify( public.msg.MarketUnlocked );

end;
--------------------------------------------------------------------------------
function public.Market_IsBuilt ()

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "market"
  };

  return lib_db.Execute( query ).result.build;

end;
--------------------------------------------------------------------------------
function public.Market_Build ()  

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "market",
    set        = { build = true }
  };

  lib_db.Execute( query );
  mediator.Notify( public.msg.MarketConstructed );

end;
--------------------------------------------------------------------------------
function  public.Market_LoadOffers ()

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "market"
  };

  return lib_db.Execute( query ).result.sell;

end;
--------------------------------------------------------------------------------
function public.Market_SaveOffers ( resources )

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "market",
    set        = { sell = resources }
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
--function .. () end;
--------------------------------------------------------------------------------
function public.Purchase_Register ( lot_id )

  local query;

  if ( public.Purchase_GetRegisteredCount( lot_id ) ) then

    query = 
    {
      query      = "update",
      group      = "player",
      collection = "purchases",
      indice     = "name",
      criteria   = { like = "lot_"..lot_id },
      set        = {
        count = 1
      }
    };

  else

    query = 
    {
      query      = "insert",
      group      = "player",
      collection = "purchases",
      value      = {
        name  = "lot_"..lot_id,
        count = 1
      }
    };

  end;

  lib_db.Execute( query );

end;
--------------------------------------------------------------------------------
function public.Purchase_Unregister ( lot_id )

  local query = 
  {
    query      = "update",
    group      = "player",
    collection = "purchases",
    indice     = "name",
    criteria   = { like = "lot_"..lot_id },
    set        = {
      count = 0
    }
  };

  lib_db.Execute( query );

end;
--------------------------------------------------------------------------------
function public.Purchase_GetRegisteredCount ( lot_id )

  local query = 
  {
    query      = "find",
    group      = "player",
    collection = "purchases",
    indice     = "name",
    criteria   = { like = "lot_"..lot_id }
  };

  local purchase = lib_db.Execute( query ).result;

  if ( purchase ) then
  
    return purchase.count;

  end

  return false;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Market_Refresh ()

  TimeFinishEvent();
  TimeUnsheduleEvent( private.market_timer.token );

  local shares = private.market_settings.sell;
  
  if ( #shares < 1                   ) or 
     ( private.market_timer.time < 1 ) then
    return;
  end;
  
  local random = {};

  for i = 1, #shares do
    random[ "resource"..i ] = math.random( shares[ i ].min, shares[ i ].max );
  end;

  TimeSheduleEvent(
    private.market_timer.token,
    private.market_timer.scope,
    private.market_timer.time,
    private.market_timer.event,
    {}
  );

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "market",
    set        = { sell = random }
  };
  
  lib_db.Execute( query );
  
  mediator.Notify( public.msg.MarketRefresh );

end;
--------------------------------------------------------------------------------