-- name=m_social
--------------------------------------------------------------------------------
function public.__InitModule ()
  
  private.Section = "Social";
  
  if ( IsMobile() ) then

    private.userSection = "user";

    private.update_cooldown = private.GetUpdateCooldown();
    private.update_token    = "tmr_social_update_friends";
    private.friends         = {};
    private.friends_stages  = {};
    private.friends_info    = {};

    mediator.Subscribe( m_game.msg.ConnectLost,     private.UpdateOnline );
    mediator.Subscribe( m_game.msg.ConnectRestored, private.UpdateOnline );
    mediator.Subscribe( "Event_Social_Login_Complete",   private.OnLoginComplete );
    mediator.Subscribe( "Event_Social_Logout_Complete",  private.OnLogoutComplete );

    ne.msg.Subscribe( "Message_FixedUrlReceived_Self", 
      function ( message, reply )
      
        if ( reply.data and reply.data.data and reply.data.data.url ) then
        
          private.fixed_self_avatar_url = 
            reply.data.data.url:gsub( "https", "http" );
            
          private.Server_SetSocialAvatar( private.fixed_self_avatar_url );
        
        end;
    
      end
    );

  end;
  
  private.InitSubscribeAttempts();
  
  --Log.i( "Initialized." );
  
  private.fixed_self_avatar_url = nil;
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Start (
    func_UserId,
    func_IsOnline,
    func_ShowWaiting,
    func_HideWaiting
  )
  
  private.func_UserId        = func_UserId;
  private.func_IsOnline      = func_IsOnline;
  private.func_ShowWaiting   = func_ShowWaiting;
  private.func_HideWaiting   = func_HideWaiting;

  if ( IsMobile() ) then
  
    private.is_online = private.func_IsOnline();

    ne.msg.Subscribe( "Event_Social_FriendsCollectionsCooldownEnd", private.UpdateFriendsCollections );
    ne.msg.Subscribe( Event_Auth_SignIn, function ()
      v_log.Kontagent_LogEvent( "social",
      {
        description = "account_profile",
        name        = "create_raveid"
      } );
    end );

    ne.msg.Subscribe( Event_Social_RaveLoginFail, function ()
      v_log.Kontagent_LogEvent( "social",
      {
        description = "social_login",
        name        = "login_fail"
      } );
    end );

    mediator.Subscribe( "Model_Social_LoginComplete", private.RequestFriendsCollections );

    if ( private.is_online ) then

      if ( public.IsLogged() ) then
        private.GetFriendList();
      end;

    else

      local func_try_get_friends = function ()

        if ( public.IsLogged() ) then
          private.GetFriendList();
        end;

      end;

      mediator.SubscribeOnce( m_game.msg.ConnectRestored, func_try_get_friends );

    end;
    
    private.TryFixSelfAvatar();
    
    local raveId = SocialGetInfo().id;
    if ( raveId         ) and
       ( not IsEditor() ) then
      if ( raveId ~= parts.auth.GetLocalUserId() ) then
        private.LoginReplace(raveId);
      else
        wiz.app.DeleteLocalUserIdBackUp();
      end;
    end;
    
  end;
  

end;
--------------------------------------------------------------------------------
function private.UpdateOnline ()

  private.is_online = private.func_IsOnline();

end;
--------------------------------------------------------------------------------
-- function ***** AUTH ***** ()
--------------------------------------------------------------------------------
function public.IsLogged ()

  local answer = SocialIsLogged();

  return answer.result;

end;
--------------------------------------------------------------------------------
function private.OnLoginComplete()
  if ( IsAndroid() and public.IsLogged() ) then
    private.TryFixSelfAvatar();
  end;
end;
--------------------------------------------------------------------------------
function private.TryFixSelfAvatar()
  if ( not IsAndroid() ) then return; end;
  
  local info = SocialGetInfo();
  
  local fb = "graph.facebook.com";
  
  if ( info.url and info.url:find( fb ) ) then
  
    local get = tern( info.url:find( "?" ) )( "&", "?" ).."redirect=false";
    
    NetSendHttpRequest(
      {
        url = info.url..get
      },
      "Message_FixedUrlReceived_Self"
    );
  
  end;
end;
--------------------------------------------------------------------------------
function private.OnLogoutComplete()
  private.fixed_self_avatar_url = nil;
end;
--------------------------------------------------------------------------------
function public.GetSocialInfo()
  
  if ( public.IsLogged() ) then
  
    local info = SocialGetInfo();
    
    if ( IsAndroid() ) then
    
      info.url = private.fixed_self_avatar_url or info.url;
    
    end;
    
    return info;
  
  end;

  return {
    userId   = v_player.UserId(),
    socialId = 0,
    name     = v_player.Name(),
    score    = 0,
    url      = nil,
    res      = "assets/content/interface/int_dialog_stage_friends/default_0"
  };
  
end;
--------------------------------------------------------------------------------
function public.Login ( callpoint )

  Log.i( "Login" );

  local func_wait;

  func_wait = function ( message_name, message_params )

    ne.msg.Unsubscribe( "Event_Social_Login_Complete", func_wait );
    private.func_HideWaiting();

    if ( not message_params.result ) then

      Log.e( "Login: failed." );

      mediator.Notify( "Model_Social_LoginComplete", message_params.result );

      return;
  
    end;
    
    mediator.Notify( "Model_Social_LoginContinue" );

    v_log.Kontagent_LogEvent( "social",
    {
      description = "social_login",
      name        = "login"
    } );

    if not ( message_params.socialId == message_params.previousSocialId ) then

      Log.i( "Login: replace." );

      private.LoginReplace( message_params.socialId );

      return;
  
    end;

    private.LoginContinue( message_params.socialId );
    
    mediator.Notify( "Model_Social_LoginComplete", message_params.result );

  end;

  ne.msg.Subscribe( "Event_Social_Login_Complete", func_wait );
  private.func_ShowWaiting();
  
  SocialLogin( callpoint );

end;
--------------------------------------------------------------------------------
function public.Logout ( callpoint, func_on_logged_out )

  Log.i( "Logout" );

  local func_wait_logout, func_wait_profile;

  func_wait_logout = function ( message_name, message_params )

    ne.msg.Unsubscribe( "Event_Social_Logout_Complete", func_wait_logout );
    private.func_HideWaiting();

    if ( message_params.result ) then

      -- JK: only here player is actually logged out

      if ( func_on_logged_out ) then func_on_logged_out(); end;

      private.ClearFriends();
      
      if ( message_params.socialId ~= message_params.previousSocialId ) then

        parts.application.RestartWithLocalUserIdReplace( message_params.socialId, true );

      end;
      
      v_log.Kontagent_LogEvent( "social", 
      {
        description = "social_login",
        name        = "logout"
      } );

    end;

  end;

  func_wait_profile = function ( message_name, message_params )

    ne.msg.Unsubscribe( "Event_Social_Profile_Complete", func_wait_profile );
    private.func_HideWaiting();

    if ( message_params.result ) then

      mediator.Notify( "Model_Social_ProfileComplete", message_params.result );

    end;

  end;

  ne.msg.Subscribe( "Event_Social_Logout_Complete", func_wait_logout );
  ne.msg.Subscribe( "Event_Social_Profile_Complete", func_wait_profile );
  private.func_ShowWaiting();
  
  SocialLogout( callpoint );

end;
--------------------------------------------------------------------------------
function private.LoginContinue ( social_id )

  Log.i( "LoginContinue: [ "..social_id.." ]" );

  local social_info = public.GetSocialInfo();

  private.Server_SetSocialId( social_id );
  private.Server_SetSocialName( social_info.name );
  private.Server_SetSocialAvatar( social_info.url );

  private.GetFriendList();

  -- Moved to Login.
  -- mediator.Notify( "Model_Social_LoginContinue" );

end;
--------------------------------------------------------------------------------
function private.LoginReplace ( social_id )

  Log.i( "LoginReplace: [ "..social_id.." ]" );

  private.func_ShowWaiting();
  
  parts.application.release.AutoSave(false);
  parts.application.loseFocus.AutoSave(false);
  wiz.app.BackUpLocalUserId();

  lib_rpc.Schedule( private.userSection, "tryReplaceLocalUserId", { newLocalUserId = social_id }, function ( message_name, message_params )

    Log.i( "LoginReplace: [ "..social_id.." ] "..tostring( message_params.result ) );
  
    if ( not message_params.result ) then
       
      local userIds = {
        oldUserId = message_params.my_profile.userId,
        newUserId = message_params.new_profile.userId
      };
      local conflict_params = {};
      conflict_params.func_local = function () private.LoginPrepareSelect( social_id, true, userIds ); end;
      conflict_params.func_server = function () private.LoginPrepareSelect( social_id, false ); end;
      conflict_params.info_local = message_params.my_profile;
      conflict_params.info_server = message_params.new_profile;
      mediator.Notify( "Event_Social_RaveLoginConflict", conflict_params );

      return;

    end;

    wiz.app.DeleteLocalUserIdBackUp();
    common.MessageSend( "Event_Application_LocalUserIdReplaced", {} );
    
    private.func_HideWaiting();
    private.LoginContinue( social_id );

    end,
   nil,
   true);

end;
--------------------------------------------------------------------------------
function private.LoginPrepareSelect ( social_id, keepLocalSave, userIds )

  if ( keepLocalSave ) then
    
    lib_rpc.Schedule( private.userSection, "migrateUserDataToAnotherUserId", 
      { 
        fromUser = userIds.oldUserId, 
        toUser = userIds.newUserId 
      }, 
      function ( message_name, message_params )

        Log.i( "LoginPrepareSelect: [ "..social_id.." ] "..tostring( message_params.result ) );
        private.LoginSelect( social_id, keepLocalSave );
        
      end,
      nil,
      true);
   
  else
    private.LoginSelect ( social_id, keepLocalSave );
  end;
  
end;
--------------------------------------------------------------------------------
function private.LoginSelect ( social_id, keepLocalSave )

  Log.i( "LoginSelect: restart with id "..social_id );

  private.func_HideWaiting();
  
  mediator.Notify( "Model_Social_LoginComplete", true );
  
  wiz.app.DeleteLocalUserIdBackUp();
  parts.application.RestartWithLocalUserIdReplace( social_id, true, keepLocalSave );

end;
--------------------------------------------------------------------------------
function private.Server_SetSocialId ( socialId )

  socialId = socialId or "";

  Log.i( "Server_SetSocialId: "..socialId );
  lib_rpc.Schedule( private.Section, "SetSocialId", 
    { socialId = socialId }, 
    function () end, true, true );

end;
--------------------------------------------------------------------------------
function private.Server_SetSocialName ( name )

  name = name or "";

  Log.i( "Server_SetSocialName: "..name );
  lib_rpc.Schedule( private.userSection, "updateUserName", 
    { name = name }, 
    function () end, true, true );

end;
--------------------------------------------------------------------------------
function private.Server_SetSocialAvatar( avatar )

  avatar = avatar or "";

  lib_rpc.Schedule( private.userSection, "updateUserAvatar", 
    { avatar = avatar }, 
    function () end, true, true );

end;
--------------------------------------------------------------------------------
-- function ***** FRIENDS ***** ()
--------------------------------------------------------------------------------
function private.UpdateFriendsCollections ()
  
  Log.i( "private.UpdateFriendsCollections" );
  
  if ( private.is_online ) then

    private.UpdateCooldownFinish();
    private.GetFriendList();

  else

    mediator.SubscribeOnce( "Model_Game_ConnectRestored", private.GetFriendList );

  end;

end;
--------------------------------------------------------------------------------
function  public.GetFriendsInZone ( section )

  Log.i( "GetFriendsInZone: " );

  local collection_table = private.friends_info or {};

  local friend_list = {};

  if ( #collection_table > 0 ) then

    for u = 1, #collection_table do

      local user_data = collection_table[ u ];
      if ( user_data.collection and #user_data.collection > 0 ) then

        local stage_max = private.GetStageLast( user_data.collection );
        
        if ( stage_max ) then

          local friend_section = ( m_stage.GetStaticData( stage_max ) or {} ).section;

          if ( section == friend_section ) then

            local friend_data = private.GetFriend( user_data.social_id );
            table.insert( friend_list, { 
              userId   = user_data.user_id,
              socialId = user_data.social_id,
              name     = friend_data.name,
              url      = friend_data.url,
              stage    = stage_max
            } );             
             
          end;

        end;

      end;

    end;

  end;

  return friend_list;

end;
--------------------------------------------------------------------------------
function private.GetStageLast ( user_collection )

  return user_collection[ 1 ].stage_last;

end;
--------------------------------------------------------------------------------
function public.GetFriendsStageScores ( stage_name ) 

  local friends_scores = {};

  if ( IsMobile() ) then
  
    if ( private.is_online and public.IsLogged() ) then
    
      local collection_table = private.friends_stages;
    
      for u = 1, #collection_table do
        local user_data = collection_table[ u ];
        for s = 1, #user_data.collection do
          local stage_data = user_data.collection[ s ];
          if ( not stage_name ) or ( stage_data.name == stage_name and stage_data.score and stage_data.score > 0 ) then
            local friend_data = private.GetFriend( user_data.social_id );
            table.insert( friends_scores, { 
              userId   = user_data.user_id,
              socialId = user_data.social_id,
              name     = friend_data.name,
              url      = friend_data.url,
              score    = user_data.collection[ s ].score
            } );
            break;
          end;
        end;
      end;
      
    end;
    
    local bots = private.GetBotsForStage( stage_name );
    
    for i = 1, #bots do
    
      table.insert( friends_scores, bots[ i ] );
    
    end;

  end;

  return friends_scores;

end;
--------------------------------------------------------------------------------
function private.GetFriendList ()

  Log.i( "GetFriendList" );
  
  ne.msg.Subscribe(
    "Event_Social_Friends_Received",
    private.OnFriendsReceived
  );
  
  SocialGetFriends();

end;
--------------------------------------------------------------------------------
function private.OnFriendsReceived ( msg, params )

  private.rave_list = {};

  Log.i( "OnFriendsReceived: "..os.clock() );

  ne.msg.Unsubscribe( "Event_Social_Friends_Received", private.OnFriendsReceived );
  if ( params and params.friends and #params.friends > 0 ) then
    Log.i( "OnFriendsReceived:"..common.table.ToString( params.friends ) );
    local socialId_list = {};
    for i = 1, #params.friends do
      table.insert( socialId_list, params.friends[i].social_id );
      private.rave_list[ params.friends[i].social_id ] = {
        url = params.friends[i].picture_url,
        name = params.friends[i].display_name
      };
    end;
    private.Server_GetFriends( socialId_list );
  else
    mediator.Notify( "Model_Social_LoginComplete", true );
    mediator.Notify( "Model_Social_FriendsUpdated" );
  end;

end;
--------------------------------------------------------------------------------
function public.GetRandomPlayers()
  Log.i( "GetRandomPlayers" );
  private.Server_GetRandomPlayers();
end;
--------------------------------------------------------------------------------
function private.Server_GetRandomPlayers ()
  Log.i( "Server_GetRandomPlayers" .. private.func_UserId());
  lib_rpc.Schedule(
    private.Section,
    "GetRandomPlayers", 
    { 
      userId  = private.func_UserId()
    },
    private.OnGetRandomPlayers,
    true,
    true
  );
end;
--------------------------------------------------------------------------------
function private.OnGetRandomPlayers ( msg, params )
  local gameFriends = params.players;
  Log.i( "OnGetRandomPlayers: " .. #gameFriends);
  Log.i( "OnGetRandomPlayers:" .. common.table.ToString( gameFriends ) );
end;
--------------------------------------------------------------------------------
function private.Server_GetFriends ( facebookFriends )

  Log.i( "Server_GetFriends: "..private.func_UserId() );

  lib_rpc.Schedule(
    private.Section,
    "GetGameFriends", 
    {
      userId  = private.func_UserId(),
      friends = facebookFriends
    },
    private.OnGetGameFriends,
    true, true
  );

end;
--------------------------------------------------------------------------------
function private.OnGetGameFriends ( msg, params )

  Log.i( "OnGetGameFriends: " );
  private.friends = {};

  local gameFriends = params.friends;
  Log.i( "OnGetGameFriends: " .. #gameFriends );
  if ( type( gameFriends ) == "table" ) then
    
    Log.i( "OnGetGameFriends:"..common.table.ToString( gameFriends ) );
    
    for f, friend in pairs( gameFriends ) do
      
      table.insert( private.friends, friend );
      
      local url = private.rave_list[ friend.socialId ].url;
      
      --For DSprite compability
      if ( url == "" ) then
        url = nil;
      end;
      
      private.friends[ #private.friends ].url = url;
      
      if ( IsAndroid() and url and url:find( "graph.facebook.com" ) ) then
      
        private.friends[ #private.friends ].url = nil;
        private.RequestFixedAvatar( friend.socialId, url );
      
      end;
      
      private.friends[ #private.friends ].name = private.rave_list[ friend.socialId ].name;
    end;
  end;

  private.rave_list = nil;

  Log.i( "Call RequestFriendsCollections: "..os.clock() );
  private.SaveFriends();
  mediator.Notify( "Model_Social_LoginComplete", true );

end;
--------------------------------------------------------------------------------
function private.RequestFixedAvatar( socialId, url )
  
  local message = "Message_FixedUrlReceived_"..socialId;
  ne.msg.Subscribe( message, private.OnFixedUrlReceived );
  
  local get = tern( url:find( "?" ) )( "&", "?" ).."redirect=false";
  
  NetSendHttpRequest( { url = url..get }, message );

end;
--------------------------------------------------------------------------------
function private.OnFixedUrlReceived ( message, reply )

  ne.msg.Unsubscribe( message, private.OnFixedUrlReceived );
  
  if ( reply.data and reply.data.data and reply.data.data.url ) then
  
    local socialId = message:match( "^Message_FixedUrlReceived_(.*)$" );
    local url      = reply.data.data.url:gsub( "https", "http" );

    local query =
    { 
      query      = "update",
      group      = "social",
      collection = "friends",
      indice     = "socialId",
      criteria   = { like = socialId },
      set        = { url  = url      }
    };
    
    lib_db.Execute( query );
  
  end;

end;
--------------------------------------------------------------------------------
-- function ***** FRIENDS - COLLECTIONS ***** ()
--------------------------------------------------------------------------------
function private.RequestFriendsCollections () 

  local friendList = private.GetFriends();

  local idList = {};
  for i = 1, #friendList do
    table.insert( idList, friendList[ i ].userId );
  end;

  private.Server_GetFriendsCollections( idList );

end;

--------------------------------------------------------------------------------
function private.Server_GetFriendsCollections ( friendArray )

  Log.i( "Server_GetFriendsCollections: "..common.table.ToString( friendArray )..": "..os.clock() );
  if(next(friendArray) ~= nil) then
    lib_rpc.Schedule( private.Section, "GetFriendsCollections",
      { friends = friendArray, collection = { "player/info", "player/stages" } },
      private.OnGetFriendsCollections, true, true );
  end;

end;
--------------------------------------------------------------------------------
function private.OnGetFriendsCollections ( msg, params )

  Log.i( "OnGetFriendsCollections: "..#params.collections..": "..os.clock() );

  local collectionTable_stages = {};
  local collectionTable_info = {};

  for i = 1, #params.collections do

      local userData = params.collections[ i ];

      Log.i( "get collection "..userData.userId.." : "..userData.socialId..": "..os.clock() );
      local convertedData_stages = DbConvertCollectionDataToTable( userData.data.stages );
      local convertedData_info = DbConvertCollectionDataToTable( userData.data.info );

      collectionTable_stages[ i ] = {
        user_id    = userData.userId,
        social_id  = userData.socialId,
        user_name  = userData.userName,
        collection = convertedData_stages
      };

      collectionTable_info[ i ] = {
        user_id    = userData.userId,
        social_id  = userData.socialId,
        user_name  = userData.userName,
        collection = convertedData_info
      };

  end;

  private.UpdateCooldownStart();
  private.friends_stages = collectionTable_stages;
  private.friends_info = collectionTable_info;
  mediator.Notify( "Model_Social_FriendsCollectionsUpdated", collectionTable_info );
  mediator.Notify( "Model_Social_FriendsScoresUpdated", collectionTable_stages );

end;
--------------------------------------------------------------------------------
function private.UpdateCooldownStart ()

  TimeSheduleEvent( 
    private.update_token,
    "SESSION_GAME_TIME",
    private.update_cooldown,
    "Event_Social_FriendsCollectionsCooldownEnd",
    {}
  );

end;
--------------------------------------------------------------------------------
function private.UpdateCooldownFinish ()

  TimeFinishEvent();
  TimeUnsheduleEvent( private.update_token );

end;
--------------------------------------------------------------------------------
-- function ***** SAVE ***** ()
--------------------------------------------------------------------------------
function  public.GetFriend ( socialId ) 

  return private.GetFriend( socialId );

end;
--------------------------------------------------------------------------------
function  public.GetFriends () 

  return private.GetFriends();

end;
--------------------------------------------------------------------------------
function private.GetFriend ( socialId ) 

  local query =
  { 
    query      = "find",
    group      = "social",
    collection = "friends",
    indice     = "socialId",
    criteria   = { like = socialId }
  };
  
  local result = lib_db.Execute( query ).result;
  
  return result or {};

end;
--------------------------------------------------------------------------------
function private.GetFriends () 

  local query =
  { 
    query      = "find_all",
    group      = "social",
    collection = "friends"
  };
  
  local  result = lib_db.Execute( query ).result;
  return result or {};

end;
--------------------------------------------------------------------------------
function public.Cheat_AddMockFriend ()

  local socialId = common.time.GetCurrent();

  local query =
  { 
    query      = "insert",
    group      = "social",
    collection = "friends",
    value      = {
      socialId = socialId,
      name     = "friend_"..socialId
    }
  };
  
  if( lib_db.Execute( query ) ) then

    mediator.Notify( "Model_Social_FriendsUpdated" );

  end;

end;
--------------------------------------------------------------------------------
function private.ClearFriends ()

  local query =
  { 
    query      = "delete_all",
    group      = "social",
    collection = "friends"
  };
  
  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function private.SaveFriends ()

  private.ClearFriends();

  if ( next( private.friends ) ) then
  
    lib_db.Rewrite( "social", "friends", private.friends );

  end;

  Log.i( "Friends updated: "..#private.friends );
  mediator.Notify( "Model_Social_FriendsUpdated" );

end;
--------------------------------------------------------------------------------
--function *** BOTS *** ()
--------------------------------------------------------------------------------
function private.GetBotsForStage ( stage_name )

  -- Get bot data
  local bot_data   = lib_db.Settings( "social" ).bot_scores;
  local move_score = lib_db.Settings( "social" ).bot_move_score;
  
  local list = {};
  for i = 1, #bot_data do table.insert( list, i ) end;
  
  -- Shuffle bot list
  local seed = stage_name:match( "%d+" );
  math.randomseed( tonumber( seed ) );
  
  for i = #list, 2, -1 do
  
    local j = math.random( i );
    list[ i ], list[ j ] = list[ j ], list[ i ];
    
  end;
  
  local shifts = {};
  for i = 1, #list do table.insert( shifts, math.random( 100 ) ); end;
  
  math.randomseed( os.time() );
  
  -- Get stage state scores
  local scores = {};
  local states = m_stage.GetStaticData( stage_name ).state;
  
  for s = 1, 3 do
  
    local targets = states[ s ].targets;
    for t = 1, #targets do
    
      local target = targets[ t ];
      if ( target.type == "GainScore" ) then
      
        table.insert( scores, target.count );
        break;
      
      end;
    
    end;
  
  end;
  
  -- Generate bot list
  local bots = {};
  
  for i = 1, #list do
  
    local state = bot_data[ i ].state;
    local moves = bot_data[ i ].moves;
    local id    = list[ i ];
    local score = moves * move_score;
    
    local score_raw = ( scores[ state ] + score + shifts[ i ] );
    local score = 10 * math.ceil( score_raw / 10 );
  
    table.insert( bots,
      { 
        userId   = "bot_"..id,
        socialId = "bot_"..id,
        name     = StringGet( "str_social_bot_name_"..id ),
        score    = score,
        url      = nil,
        res      = "assets/content/interface/social/bot_avatar_"..id,
        is_bot   = true
      }
    );
  
  end;
  
  return bots;

end;
--------------------------------------------------------------------------------
--function ***** MISC ***** ()
--------------------------------------------------------------------------------
--[[function private.Fake_GetScore( stage_name, index )

  if ( private.stage and private.stage.name == stage_name ) then

    return ( private.stage.score + 1000 * index + math.random( 999 ) ); 

  end;

  local state_query = 
  {
    query = "find",
    collection = "player/stages:name",
    criteria = { like = stage_name },
    result = "state"
  };

  local state = lib_db.DbExecute( state_query ).result or 1;

  local score_query = 
  {
    query = "find",
    collection = "stages:name",
    criteria = { like = stage_name },
    result = "state."..( state - 1 )..".targets"
  };

  local targets = lib_db.DbExecute( score_query ).result;

  private.stage = { name = stage_name, score = 1000 };

  for i = 1, #targets do

    if ( targets[ i ].type == "GainScore" ) then

      private.stage.score = targets[ i ].count;

      break;

    end;

  end;

  return ( private.stage.score + 1000 * index + math.random( 999 ) ); 

end;]]
--------------------------------------------------------------------------------
function private.GetUpdateCooldown ()

  return lib_db.Settings( "social" ).friends_collections_update;

end;
--------------------------------------------------------------------------------
function public.GetLoginReward ()

  return lib_db.Settings( "social" ).login_reward;

end;
--------------------------------------------------------------------------------
function public.Overtaken_GetShowChance ()

  return lib_db.Settings( "social" ).overtaken_show_chance;

end;
--------------------------------------------------------------------------------
function public.Settings ()
  
  return lib_db.Settings( "social" );
  
end;
--------------------------------------------------------------------------------
-- function ***** ATTEMPTS ***** ()
--------------------------------------------------------------------------------
function private.Storage_Hearts ()
  
  return lib_db.Player.Info().hearts;
  
end;
--------------------------------------------------------------------------------
function private.UpdateAttempts ( attempt_type, attempt_count )
  
  local hearts_info = private.Storage_Hearts();

  hearts_info[ attempt_type.."_attempts" ] = attempt_count;
  
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "info",
    set        = { hearts = hearts_info }
  };

  lib_db.Execute( query );
  
  if ( attempt_count > 0 ) then
    private.Attempts_StartTimer( attempt_type );
  end;
  
end;
--------------------------------------------------------------------------------
function private.Static_HeartMaxAttempts ( attempt_type )

  local settings = public.Settings();
  local attempts = settings[ attempt_type.."_limit" ].count or 3;
  
  return attempts;
  
end;
--------------------------------------------------------------------------------
function private.Storage_HeartAttempts ( attempt_type )

  local hearts_info = private.Storage_Hearts();
  local attempts = hearts_info[ attempt_type.."_attempts" ] or 0;
  
  return attempts;
  
end;
--------------------------------------------------------------------------------
function public.GetAttempts ( attempt_type )
  
  local attempts = private.Storage_HeartAttempts( attempt_type );
  local max_attempts = private.Static_HeartMaxAttempts( attempt_type );

  return attempts, max_attempts;
  
end;
--------------------------------------------------------------------------------
function public.GetAttemptsAsk ()
  
  local attempts, max_attempts = public.GetAttempts( "ask" );
  
  return attempts, max_attempts;
  
end;
--------------------------------------------------------------------------------
function public.GetAttemptsSend ()
  
  local attempts, max_attempts = public.GetAttempts( "send" );
  
  return attempts, max_attempts;
  
end;
--------------------------------------------------------------------------------
function public.UpdateAttempts ( attempt_type, attempt_count )
  
  private.UpdateAttempts( attempt_type, attempt_count );
  
end;
--------------------------------------------------------------------------------
--function ***** COOLDOWN ATTEMPTS *****()
--------------------------------------------------------------------------------
function private.InitSubscribeAttempts ()
  
  private.heart_ask_token = "tmr_social_heart_cooldown_asked";
  private.heart_ask_event = "Event_Social_AskHeartCooldownEnd";
  private.heart_send_token = "tmr_hearts_sendlimit";
  private.heart_send_event = "Event_Hearts_SendLimitFinished";
  
  private.Attempts_InitCheck();
  
  ne.msg.Subscribe(
    private.heart_ask_event,
    private.Attempts_Ask_FinishTimer
  );
  
  ne.msg.Subscribe(
    private.heart_send_event,
    private.Attempts_Send_FinishTimer
  );
  
end;
--------------------------------------------------------------------------------
function private.Attempts_InitCheck ()
  
  local attempts_ask = private.Storage_HeartAttempts( "ask" );
  local attempts_send = private.Storage_HeartAttempts( "send" );
  
  if ( attempts_ask > 0 ) and ( not public.Attempts_Ask_TimerIsExist() ) then
    private.UpdateAttempts( "ask", 0 );
  end;
  
  if ( attempts_send > 0 ) and ( not public.Attempts_Send_TimerIsExist() ) then
    private.UpdateAttempts( "send", 0 );
  end;
  
end;
--------------------------------------------------------------------------------
function private.Attempts_StartTimer ( attempt_type )
  
  if ( attempt_type == "ask" ) then
    if ( not public.Attempts_Ask_TimerIsExist() ) then
      private.Attempts_Ask_StartTimer();
    end;
  elseif ( attempt_type == "send" ) then
    if ( not public.Attempts_Send_TimerIsExist() ) then
      private.Attempts_Send_StartTimer();
    end;
  end;
  
end;
--------------------------------------------------------------------------------
function private.Attempts_Ask_StartTimer ()
  
  local time = public.Settings().ask_limit.time or 86400;
  
  TimeSheduleEvent( 
    private.heart_ask_token,
    "GLOBAL_TIME",
    time,
    private.heart_ask_event,
    {}
  );
  
  mediator.Notify( "Model_Social_TimerAskStart" );
  
end
--------------------------------------------------------------------------------
function private.Attempts_Ask_FinishTimer ()
  
  TimeFinishEvent();
  TimeUnsheduleEvent( private.heart_ask_token );

  private.UpdateAttempts( "ask", 0 );
  
end
--------------------------------------------------------------------------------
function public.Attempts_Ask_TimerIsExist ()
  
  return TimeIsEventExist( private.heart_ask_token );
  
end;
--------------------------------------------------------------------------------
function public.Attempts_Ask_GetRemainingTime ()
  
  return TimeGetEventRemainingTime( private.heart_ask_token );
  
end;
--------------------------------------------------------------------------------
function public.Attempts_Ask_FinishTimer ()
  
  private.Attempts_Ask_FinishTimer();
  
end;
--------------------------------------------------------------------------------
function private.Attempts_Send_StartTimer ()
  
  local time = public.Settings().send_limit.time or 86400;
  
  TimeSheduleEvent( 
    private.heart_send_token,
    "GLOBAL_TIME",
    time,
    private.heart_send_event,
    {}
  );
  
end;
--------------------------------------------------------------------------------
function private.Attempts_Send_FinishTimer ()
  
  TimeFinishEvent();
  TimeUnsheduleEvent( private.heart_send_token );

  private.UpdateAttempts( "send", 0 );
  
end;
--------------------------------------------------------------------------------
function public.Attempts_Send_TimerIsExist ()
  
  return TimeIsEventExist( private.heart_send_token );
  
end;
--------------------------------------------------------------------------------
function public.Attempts_Send_GetRemainingTime ()
  
  return TimeGetEventRemainingTime( private.heart_send_token );
  
end;
--------------------------------------------------------------------------------
function public.Attempts_Send_FinishTimer ()
  
  private.Attempts_Send_FinishTimer();
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.FacebookShare_GetSectionsPosted ()

  local query =
  {
    query      = "find",
    group      = "sml",
    collection = "game_state"
  };
  
  local result = lib_db.Execute( query ).result;
  
  return result.fb_posted or {};

end;
--------------------------------------------------------------------------------
function public.FacebookShare_MarkSectionAsPosted ( section_name )

  local posted_sections = public.FacebookShare_GetSectionsPosted();

  if ( posted_sections[ section_name ] ) then

    DbgTrace(
      "[ ERROR ][ m_social ] Already posted to facebook: "..section_name
    );
    return;

  end;

  posted_sections[ section_name ] = true;

  local query =
  {
    query      = "update",
    group      = "sml",
    collection = "game_state",
    set        = 
    {
      fb_posted = posted_sections
    }
  };

  lib_db.Execute( query );

end;
--------------------------------------------------------------------------------