-- name=m_spells
--------------------------------------------------------------------------------
function public.__InitModule ()

  --Log.i( "Initialized." );

  private.section_spells = {};

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.TryUnlockSpell ( spell_name )

  local spell_data = public.Spell_Data( spell_name );

  if not ( spell_data.enabled ) then
    
    local query =
    {
      query      = "insert",
      group      = "player",
      collection = "spells",
      value      = 
      {
        name      = spell_name,
        enabled   = true,
        completed = false
      }
    };

    lib_db.Execute( query );

    mediator.Notify( "Model_Spells_SpellEnabled", spell_name );
    
    return true;

  end;
  
  return false;

end;
--------------------------------------------------------------------------------
function public.Spell_GetNameByUnlock ( stage_name )

  local query = 
  {
    query      = "find",
    group      = "static",
    collection = "spells",
    indice     = "unlock",
    criteria   = { like = stage_name }
  };

  local spell = lib_db.Execute( query ).result;
  
  if ( spell ) and ( spell.name ) then
  
    return spell.name;
    
  end;
  
  return nil;

end;
--------------------------------------------------------------------------------
function public.Complete ( spell_name )

  local query =
  {
    query      = "update",
    group      = "player",
    collection = "spells",
    indice     = "name",
    criteria   = { like = spell_name },
    set        =
    {
      completed = true
    }
  };

  lib_db.Execute( query );

  mediator.Notify( "Model_Spells_SpellCompleted", spell_name );

end;
--------------------------------------------------------------------------------
function public.Spell_Data ( spell_name )

  local query = 
  {
    query      = "find",
    group      = "static",
    collection = "spells",
    indice     = "name",
    criteria   = { like = spell_name }
  };

  local spell_data = lib_db.Execute( query ).result;

  if not ( spell_data ) then

    return false;

  end;

  spell_data.enabled   = false;
  spell_data.completed = false;

  query = 
  {
    query      = "find",
    group      = "player",
    collection = "spells",
    indice     = "name",
    criteria   = { like = spell_name }
  };

  local saved_data = lib_db.Execute( query ).result;

  if ( saved_data ) then

    spell_data.enabled   = saved_data.enabled;
    spell_data.completed = saved_data.completed;

  end;

  return spell_data;

end;
--------------------------------------------------------------------------------
function public.GetOutput ( spell_name )

  local query = 
  {
    query      = "find",
    group      = "static",
    collection = "spells",
    indice     = "name",
    criteria   = { like = spell_name }
  };

  local spell_data = lib_db.Execute( query ).result;

  if ( spell_data ) then

    return spell_data.output[ 1 ];

  end;

  return false;

end;
--------------------------------------------------------------------------------
function public.Usable_List ( section_name )
  
  local query = 
  {
    query      = "find_all",
    group      = "player",
    collection = "spells",
    indice     = "completed",
    criteria   = { like = false }
  };

  local usable = lib_db.Execute( query ).result;

  if not ( usable ) then return false; end;
  
  local list = {};
  
  for s = 1, #usable do 
    
    if ( private.Spell_Section( usable[ s ].name ) == section_name ) then
      
      table.insert( list, usable[ s ] );
      
    end;
  
  end;
  
  if ( #list < 1 ) then return false; end;

  return list;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Spell_Section ( spell_name )

  local query = 
  {
    query      = "find",
    group      = "static",
    collection = "spells",
    indice     = "name",
    criteria   = { like = spell_name }
  };

  return lib_db.Execute( query ).result.section;

end;
--------------------------------------------------------------------------------
function public.Static_GetAllBySection ( section_name )

  local query = 
  {
    query      = "find_all",
    group      = "static",
    collection = "spells",
    indice     = "section",
    criteria   = { like = section_name }
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function public.Saved_GetByName ( spell_name )

  local query = 
  {
    query      = "find",
    group      = "player",
    collection = "spells",
    indice     = "name",
    criteria   = { like = spell_name }
  };

  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------