-- name=m_stage
--------------------------------------------------------------------------------
function public.__InitModule ()
  
  private.SubscribeMatch3EngineNotices();
  public.Stateless_CheckStatus();
  private.FastSaga_CheckStatus();
  
end;
--------------------------------------------------------------------------------
-- function ******* Settings ******* ()
--------------------------------------------------------------------------------
function private.GetSettings ( settings_name )
  
  return lib_db.Settings( settings_name );
  
end;
--------------------------------------------------------------------------------
function public.Fight_GetSettings ()
  
  return private.GetSettings( "fights" );
  
end;
--------------------------------------------------------------------------------
function public.Ride_GetSettings ()
  
  return private.GetSettings( "stages" );
  
end;
--------------------------------------------------------------------------------
function private.FastSaga_GetSettings ()
  
  return private.GetSettings( "fast_saga" );
  
end;
--------------------------------------------------------------------------------
function public.BackgroundColors_GetSettings ()
  
  return private.GetSettings( "backgrounds" );
  
end;
--------------------------------------------------------------------------------
function public.GetRewardGoldSettings ()

  local settings = public.Ride_GetSettings();
  return settings.rewards;

end;
--------------------------------------------------------------------------------
function private.GetChainScore ( chain_form, score_type )
  
  local settings = private.GetSettings( "scores" );
  return settings.chain_score[ chain_form ][ score_type ];
  
end;
--------------------------------------------------------------------------------
function private.GetCellScore ( cell_name )

  local settings = private.GetSettings( "scores" );
  return settings.cell_score[ cell_name ];

end;
--------------------------------------------------------------------------------
function private.GetBonusScore ( bonus_name )

  local settings = private.GetSettings( "scores" );
  return settings.bonus_score[ bonus_name ];

end;
--------------------------------------------------------------------------------
function private.GetBonusSwappedScore ( bonus_swapped_name )

  local settings = private.GetSettings( "scores" );
  return settings.bonus_swapped_score[ bonus_swapped_name ];

end;
--------------------------------------------------------------------------------
function private.GetCascadeScore ()

  local settings = private.GetSettings( "scores" );
  return settings.cascade_score;

end;
--------------------------------------------------------------------------------
function private.GetItemdropScore ()

  local settings = private.GetSettings( "scores" );
  return settings.itemdrop_score;

end;
--------------------------------------------------------------------------------
-- function ******* Database ******* ()
--------------------------------------------------------------------------------
function public.GetSavedData ( stage_name )

  if ( not stage_name ) then
    return false;
  end;

  local query = {
    query      = "find",
    group      = "player",
    collection = "stages",
    indice     = "name",
    criteria   = { like = stage_name }
  };
 
  return lib_db.Execute( query ).result;

end;
--------------------------------------------------------------------------------
function private.GetStagesBySection ( section_name )
  
  local query =
  {
    query      = "find_all",
    group      = "static",
    collection = "stages",
    indice     = "section",
    criteria   = { like = section_name }
  };
  
  return lib_db.Execute( query ).result;
  
end;
--------------------------------------------------------------------------------
-- function ******* Temporarily Events ******* ()
--------------------------------------------------------------------------------
function public.Stateless_GetLimit ()

  local settings = public.Ride_GetSettings();
  return settings.stateless_until;

end;
--------------------------------------------------------------------------------
function public.Stateless_CheckStatus () --ES> мб сделать как с фастсагой
  
  local settings = public.Ride_GetSettings();
  local data = public.GetSavedData( "stage_"..settings.stateless_until + 1 );
  private.is_stateless = ( data == nil );
  
end;
--------------------------------------------------------------------------------
function public.Stateless_IsActive ()
  
  return private.is_stateless;
  
end;
--------------------------------------------------------------------------------
function private.FastSaga_CheckStatus ()
  
  local settings = private.FastSaga_GetSettings();
  local condition = settings.condition;
  
  if ( condition.type == "CompleteStage" ) then
    
    local status = not ( ( public.GetSavedData( condition.name ) or {} ).completed );
    private.is_active_fastsaga = status;
    return;
    
  end;
  
  Log.w( "FastSaga_IsActive: unknown condition type!" );
  
  private.is_active_fastsaga = false;
  
end;
--------------------------------------------------------------------------------
function public.FastSaga_IsActive ()
  
  if ( private.is_active_fastsaga ) then
    private.FastSaga_CheckStatus();
  end;
  
  return private.is_active_fastsaga;

end;
--------------------------------------------------------------------------------
function public.FastSaga_Deactive ()
  
  private.is_active_fastsaga = false;
  
end;
--------------------------------------------------------------------------------
-- function ******* Init Engine Subscribes ******* ()
--------------------------------------------------------------------------------
function private.SubscribeMatch3EngineNotices ()
  
  ne.msg.Subscribe( "Command_Match3_ChainActivated", private.OnChainActivated );
  ne.msg.Subscribe( "Event_Match3_FieldActivated", private.OnFieldActivated );
  ne.msg.Subscribe( "Event_Match3_CellChanged", private.OnCellChanged );
  ne.msg.Subscribe( "Event_Match3_BonusesSwapped", private.OnBonusesSwapped );
  ne.msg.Subscribe( "Event_Match3_ChipsRemoved", private.OnChipsRemoved );
  ne.msg.Subscribe( "Event_Match3_CombinationsNotExist", private.OnCombinationsNotExist );
  ne.msg.Subscribe( "Event_Match3_Explosion", private.OnExplosionField );
  ne.msg.Subscribe( "Event_Match3_HiveAdded", private.OnHiveAdded );
  
end;
--------------------------------------------------------------------------------
function private.OnChainActivated ( msg, chain_data )
  
  mediator.Notify( "Model_Stage_ChainActivated", chain_data );

  private.IncChainCount();

  local score_type = "Basic";
  local chain_type = "Unknown";
  local bonus_type = "Unknown";

  for c = 1, #chain_data.chips_list do

    chain_type = chain_data.chips_list[ c ].type;

    if ( chain_data.chips_list[ c ].property ~= "Basic" ) then

      score_type = "Bonus";
      bonus_type = chain_data.chips_list[ c ].property;

      break;

    end;

  end;

  local chain_form = chain_data.chain_form;
  local basic_score = private.GetChainScore( chain_form, score_type );
  local chain_score = basic_score * private.state_data.state;

  if ( private.combination.form ) and 
     ( private.combination.form == chain_data.chain_form ) then

    chain_score = chain_score + private.combination.score;

  end;

  private.combination.form  = chain_form;
  private.combination.score = chain_score;

  local x, y = chain_data.pivot_data.x, chain_data.pivot_data.y;

  private.ChangeScore( x, y, chain_type, chain_score );
  private.AddAttackFactorHero( #chain_data.chips_list );
  private.UpdateRewards( chain_data.chips_list );
  private.CheckActivateBonus( chain_type );
  private.UpdateTargetProgress_DestroyGem( chain_data.chips_list[ 1 ].type, #chain_data.chips_list );

  if ( score_type == "Bonus" ) then

    mediator.Notify( "Model_Match3_BonusRemoved", { type = bonus_type } );

  end;
  
  private.AddMovesFromChip( chain_data.chips_list );
  
end;
--------------------------------------------------------------------------------
function private.OnFieldActivated ( msg, params )
  
  mediator.Notify( "Model_Match3_FieldActivated" );
  
  private.Moves_TryChangeCount( - 1 );
  private.Moves_IncDoneCount();
  private.IncCountMovesFight(); 
  private.ResetChainCount();

end;
--------------------------------------------------------------------------------
function private.OnCellChanged ( msg, cell_data )

  for c = 1, #cell_data.cells_list do

    local cell = cell_data.cells_list[ c ];
    
    local state = private.state_data.state;
    local cell_score = private.GetCellScore( cell.type ) * state;

    private.ChangeScore( cell.x, cell.y, cell.type, cell_score );

    local level = cell.level or 1
    private.UpdateTargetProgress_DestroyCell( cell.type, level, 1 );

    mediator.Notify( "Model_Stage_CellDestroyed", 
      { 
        name  = cell.type,
        x     = cell.x,
        y     = cell.y,
        color = cell.color,
        size  = cell.size  
      } );
  
  end;
  
end;
--------------------------------------------------------------------------------
function private.OnBonusesSwapped ( msg, combo_data )

  mediator.Notify( "Model_Stage_BonusesSwapped", combo_data );

  local combo_type = combo_data.combo_type;
  local chip_type  = combo_data.type;

  if ( combo_type == "HorizontalVertical" ) then
    mediator.Notify( "Model_Match3_BonusRemoved", { type = "Horizontal" } );
  elseif ( combo_type == "NucleusNucleus" ) then
    --
  elseif ( combo_type == "HVNucleus" ) then
    --
  elseif ( combo_type == "BombBomb" ) then
    
    local score = private.GetBonusSwappedScore( "BombBomb" ) * private.state_data.state;
    private.ChangeScore( combo_data.x, combo_data.y, chip_type, score );
    mediator.Notify( "Model_Match3_ComboActivated", { type = chip_type } );
    
  elseif ( combo_type == "NucleusBomb" ) then
    --
  elseif ( combo_type == "BombHV" ) then
    --
  elseif ( combo_type == "BasicBomb" ) then
    
    local score = private.GetBonusSwappedScore( "BasicBomb" ) * private.state_data.state;
    private.ChangeScore( combo_data.x, combo_data.y, chip_type, score );
    mediator.Notify( "Model_Match3_ComboActivated", { type = chip_type } );
    
  end;

end;
--------------------------------------------------------------------------------
function private.OnChipsRemoved ( msg, chips_data )
  
  local chips_list = chips_data.chips_list;
  
  for c = 1, #chips_list do
    
    local chip_data = chips_list[ c ];
    
    if ( chip_data.type == "Item" ) then
      
      private.OnItemDropped( chip_data );
      
    else
      
      local basic_score = private.GetBonusScore( chip_data.property );
      local score = basic_score * private.state_data.state;

      private.ChangeScore( chip_data.x, chip_data.y, chip_data.type, score );

      if ( chip_data.property ~= "Basic" ) then

        mediator.Notify( "Model_Match3_BonusRemoved", 
          { 
            type = chip_data.property
          } );

      end;
        
      private.UpdateTargetProgress_DestroyGem( chip_data.type, 1 );

    end;

  end;

  private.AddAttackFactorHero( #chips_list );

  private.UpdateRewards( chips_list );
  
  private.AddMovesFromChip( chips_list );

end;
--------------------------------------------------------------------------------
function private.OnCombinationsNotExist ( msg, params )

  private.no_combinations = true;

  mediator.Notify( "Model_Match3_CombinationsNotExist" );

end;
--------------------------------------------------------------------------------
function private.OnExplosionField ( msg, params )

  private.explosion_field = true;

  mediator.Notify( "Model_Match3_ExplosionField" );

end;
--------------------------------------------------------------------------------
function private.OnHiveAdded ( msg, params )

  for res, value in pairs( params ) do
    table.insert( private.state_data.chip_types, res );
  end;
  private.InitRewards();

end;
--------------------------------------------------------------------------------
function private.OnItemDropped ( chip_data )
  
  local score = private.GetItemdropScore () * private.state_data.state;
  private.ChangeScore( chip_data.x, chip_data.y, "Item", score );

  private.UpdateTargetProgress_DropItem( chip_data.name, 1 );
  
  mediator.Notify( "Model_Stage_ItemDropped", 
  {
    x = chip_data.x,
    y = chip_data.y,
    name = chip_data.name
  } );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Start (
    player_data,
    stage_name,
    stage_state,
    enemy_data,
    strict_state
  )

  private.need_finalize = false;

  private.stage_name = stage_name;
  private.state_data = private.GetStateData( stage_name, stage_state, strict_state );
  
  private.noreward = nil;

  private.completed         = false;
  private.bonuses_activated = false;
  private.finalized         = false;
  private.no_combinations   = nil;
  private.explosion_field   = nil;
  private.boosters          = {};
  
  private.kontagent = {
    additional_moves = 0,
    extra_moves      = 0
  };
  
  private.InitFight( player_data, enemy_data );
  
  private.InitScore();
  private.Moves_Init();
  private.InitChains();
  private.InitRewards();

  private.InitBonus();

  private.InitTargets();
  
  local saved_data = public.GetSavedData( stage_name ) or {};

  local settings = public.Ride_GetSettings();
  
  private.state_data.attempts     = saved_data.attempts or 0;
  private.state_data.max_attempts = settings.max_attempts;

  return common.table.DeepCopy( private.state_data );

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function  public.IsCanContinue ()

  return ( private.IsConditionFailed() or 
           ( not private.IsCombinationsNotExist() and
           not private.IsExplosionField() )
         );

end;
--------------------------------------------------------------------------------
function  public.TryUseContinue ( continue_name )

  if ( continue_name == "extramoves" ) then

    private.kontagent.extra_moves = private.kontagent.extra_moves + 5;
    DbgTrace("extramoves");
    private.Moves_TryChangeCount( 5, true );
    
  elseif ( continue_name == "extralives" ) then

    private.AddExtraHealthHero();

  end;

  public.Kontagent_SetBoostersUsed( 1, continue_name );

  private.finalized = false;

  mediator.Notify( "Model_Stage_Continue" );

end;
--------------------------------------------------------------------------------
function public.MovesDone_GetCount ()

  return private.Moves_GetDoneCount();

end;
--------------------------------------------------------------------------------
function public.TryContinue ()
  
  if ( private.state_data.type == "fight" ) then

    if ( private.IsHeroFightCharacter() ) then
      if ( private.TryActivateHeroAttack() ) then
        return false;
      end;
    end;

    if ( private.IsEnemyFightCharacter() ) then
      if ( private.TryActivateEnemyAttack() ) then
        return false;
      end;
    end;

    if ( not public.IsWin() ) then
      private.HandleFightBonus();
    end;
    
  end;
  
  private.TryCompleteTargets();

  return true;

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.TryFinalize ( anm_playing )

  if ( private.IsConditionCompleted() and
       private.IsTargetsCompleted() ) then

       --Log.i( "TryFinalize: all completed." );

    if ( not private.completed ) then

      private.completed = true;

      mediator.Notify(
        "Model_Stage_StageComplete",
        { moves = private.Moves_IsAvailable() }
      );

    end;

    if ( private.Moves_IsAvailable() ) then

      if ( not private.moves_remain ) then

        private.moves_remain = private.moves;

      end;
      
      if ( anm_playing ) then

        mediator.SubscribeOnce(
          "View_Stage_AnimationIsFinished",
          private.FinalBonuses_Spawn
        );
      
      else
        
        private.FinalBonuses_Spawn();
        
      end;

      return false;
      
    elseif not ( private.bonuses_activated ) then
      
      private.FinalBonuses_Activate();
      
      return false;

    end;

  end;

  if ( private.completed or 
       private.IsConditionFailed() or 
       private.IsCombinationsNotExist() or
       private.IsExplosionField() ) then

       --Log.i( "TryFinalize: finalize." );
    
    if ( not private.finalized ) then

      private.finalized = true;

      private.Finalize( private.completed );

      return true;

    end;
  
    Log.w( "Finalize recall!" );

  end;

  return false;

end;
--------------------------------------------------------------------------------
function private.Finalize ( is_win )

  private.need_finalize = is_win;
  
  Log.i( "private.Finalize End" );

end;
--------------------------------------------------------------------------------
function public.IsFinalized ()

  return private.finalized;

end;
--------------------------------------------------------------------------------
function public.Finish ( func_Finalize, func_RestoreHearts )
  
  local func_MapPrepare = nil;
  
  if ( private.need_finalize ) then
    
    if ( func_Finalize ) then

      func_MapPrepare = func_Finalize();
      
    end;

    if ( func_RestoreHearts ) then
      
      func_RestoreHearts();
      
    end;
    
  end;

  if ( not func_MapPrepare ) then
    
    loader.Module_LoadChain( true, true, function () end );
    
  else
    
    loader.Module_LoadChain( true, nil, function () end );
    func_MapPrepare();

  end;

  if ( private.Moves_GetDoneCount() == 0 ) then 

    if ( func_RestoreHearts ) then
      
      func_RestoreHearts();
      
    end;

  else

    if ( not private.completed ) then

      private.IncreaseAttempts();

    end;

  end;

  return true;

end;
--------------------------------------------------------------------------------
-- function ******* Current Stage ******* () end
--------------------------------------------------------------------------------
function public.GetCurrentStageData ()

  return private.state_data;

end;
--------------------------------------------------------------------------------
function public.GetCurrentStageName ()

  return private.stage_name;

end;
--------------------------------------------------------------------------------
function public.GetCurrentTargets ()

  return private.targets;

end;
--------------------------------------------------------------------------------
function public.GetCurrentResources ()

  local resources = {};

  for key, value in pairs( private.reward ) do

    table.insert( resources, { name = key, count = value } );

  end;

  return resources;

end;
--------------------------------------------------------------------------------
function public.GetCurrentMoves ()

  if ( private.moves ) then

    return private.moves;

  end;

  return nil;

end;
--------------------------------------------------------------------------------
function public.GetTargetByType ( target_type, target_name )

  return private.GetTargetByType( target_type, target_name );

end;
--------------------------------------------------------------------------------
function public.GetDroppedChests ()

  if ( private.targets_drop_chests ) then
    local chests = {};
    for index, value in pairs( private.targets[ "DropItem" ] ) do
      chests[ value.name ] = value.progress;
    end;
    return chests;
  end;

  return nil;

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.IsWin ()

  return private.completed;

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function private.IncreaseAttempts ()

  local stage_name = private.stage_name;
  local real_type  = private.state_data.realtype;

  if ( real_type ~= "dailyquest" ) then

    local attempts = ( public.GetSavedData( stage_name ) or {} ).attempts or 0;
    attempts = attempts + 1;

    local query =
    {
      query      = "update",
      group      = "player",
      collection = "stages",
      indice     = "name",
      criteria   = { like = stage_name },
      set        =
      {
        attempts = attempts
      }
    };
    
    lib_db.Execute( query ); 
    
    mediator.Notify( "Model_Stage_AttemptsIncreased", attempts );
    
  end;
  
end;
--------------------------------------------------------------------------------
-- function ******* Stage & State Data ******* () end
--------------------------------------------------------------------------------
function public.TryEnable ( stage_name, silent )

  local result = false;
  
  local query =
  {
    query      = "update",
    group      = "player",
    collection = "stages",
    indice     = "name",
    criteria   = { like = stage_name },
    set        =
    {
      enabled = true,
      visible = true
    }
  };
  
  result = lib_db.Execute( query ).result;

  if ( not result ) then

    query =
    {
      query      = "insert",
      group      = "player",
      collection = "stages",
      value      =
      {
        name    = stage_name,
        enabled = true,
        visible = true
      }
    };
    
    result = lib_db.Execute( query ).result;

  end;
  
  if ( result and not silent ) then
  
    mediator.Notify( "Model_Stage_StageEnabled", stage_name );
  
  end;

  return result;

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function  public.GetStageState ( stage_name, stage_data, saved_data )

  -- NOTE: stage_data and saved_data are optional

  return public.Stage_GetState( stage_name, stage_data, saved_data );

end;
--------------------------------------------------------------------------------
function public.Stage_GetState ( stage_name, stage_data, saved_data )

  saved_data = saved_data or public.GetSavedData( stage_name );
  stage_data = stage_data or public.GetStaticData( stage_name, saved_data );

  saved_data  = saved_data or {};
  local state = saved_data.state or 1;

  if ( stage_data.realtype == "misc"  ) and
     ( stage_data.type     == "fight" ) then
    state = 4;
  end;

  return state;

end;
--------------------------------------------------------------------------------
function private.GetStateData ( stage_name, stage_state, strict_state )

  local state_data = {};
  local db_data;
  
  if ( strict_state ) then
    db_data = public.GetStaticData( stage_name, stage_state );
  else
    db_data = public.GetStaticData( stage_name );
  end;

  state_data.name     = stage_name;
  state_data.section  = db_data.section;
  state_data.zone     = db_data.zone;
  state_data.content  = db_data.content;
  state_data.bonus    = db_data.bonus;
  state_data.type     = db_data.type;
  state_data.realtype = db_data.realtype;
  
  if ( not state_data.section ) then
    if ( state_data.realtype == "minievent" ) then
      -- JK: very dirty hack
      local package = state_data.zone.background.content;
      if ( package:find( "section" ) ) then
        state_data.section = package;
      end;
    end;
  end;
  
  if ( public.Stateless_IsActive() ) and
     ( not stage_state             ) then
    stage_state      = 1;
    state_data.state = 1;
  else
    stage_state      = stage_state or public.Stage_GetState( stage_name, db_data );
    state_data.state = stage_state;
  end;
  
  state_data.state_max = math.min( #db_data.state, 3 );

  for key, value in pairs( db_data.state[ stage_state ] ) do
    state_data[ key ] = value;
  end;
  
  return state_data;
  
end;
--------------------------------------------------------------------------------
function public.Stage_IsDeadlock ( stage_name )

  --JK: play with bonus stages only
  if not ( string.find( stage_name, "bonus" ) ) then return false; end;

  local query = 
  {
    query      = "find",
    group      = "static",
    collection = "builds",
    indice     = "deadlock",
    criteria   = { like = tonumber( string.match( stage_name, "%d+" ) ) }
  };

  return ( lib_db.Execute( query ).result ~= nil );

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.GetStaticData ( stage_name, state_number )
  
  local query = {
    query      = "find",
    group      = "static",
    collection = "stages",
    indice     = "name",
    criteria   = {
      like = stage_name
    }
  };

  local stage_data = lib_db.Execute( query ).result;

  if ( stage_data ) then
    private.AppendRealType( stage_data, state_number );
  end;
  
  return stage_data;

end;
--------------------------------------------------------------------------------
function private.AppendRealType ( stage_data, state_number )

  --Log.i( "state: "..tostring( state_number ) );
  
  local realtype      = stage_data.type;
  stage_data.realtype = realtype;

  if     ( realtype == "bonus"      ) then stage_data.type = "ride";
  elseif ( realtype == "minievent"  ) then stage_data.type = "ride";
  elseif ( realtype == "dailyquest" ) then stage_data.type = "fight";
  elseif ( realtype == "misc"       ) then

    if ( state_number ) then
      stage_data.type = tern( state_number == 4 )( "fight", "ride" );
    else
      local saved_data = public.GetSavedData( stage_data.name ) or {};
      stage_data.type  = saved_data.type or "fight";
    end;

  elseif ( realtype ~= "ride" and realtype ~= "fight" ) then

    Log.e( string.format( "Stage '%s' has unknown type: '%s'", stage_data.name, realtype ) );

  end;

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.GetAllBySection ( section_name, less_details )

  local stages = private.GetStagesBySection( section_name );

  if ( stages ) then

    if ( not less_details ) then
  
      for _, stage in pairs( stages ) do

        private.AppendRealType( stage );

      end;

    end;
    
    return stages;
  
  end;
  
  return nil;

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.Boosters_Use_MovesUp ()
  private.Moves_TryChangeCount( 5, true );
end;
--------------------------------------------------------------------------------
function public.Skinbonus_MovesUp ( moves )
  private.Moves_TryChangeCount( moves, true );
end;
--------------------------------------------------------------------------------
function public.Boosters_Use_SuperHit ()
  
  private.ActivateFirstCharacter();
  private.SetHeroAbility( { name = "superhit" } );
  private.ResetAttackFactorHero();
  private.AddAttackFactorHero( private.enemy.health_current / 10 / private.hero.attack );

  mediator.Notify( "Model_Stage_ViewTryContinue" );
  
end;
--------------------------------------------------------------------------------
-- function ******* Moves ******* () end
--------------------------------------------------------------------------------
function private.Moves_Init ()

  private.moves = private.state_data.moves;
  
  if ( private.moves ) then

    --private.moves = private.moves;
    private.moves_remain = nil; -- Счетчик оставшихся ходов, для статистики.

  end;

  -- Общий счетчик ходов, сделанных на уровне.
  private.moves_done = 0;

end;
--------------------------------------------------------------------------------
function private.Moves_TryChangeCount ( moves_delta, ignore )

  -- При отсутствии ходов изменение в любую сторону запрещено.
  if ( not private.Moves_IsAvailable() and not ignore ) then

    return false;

  end;
  DbgTrace("Moves_TryChangeCount new="..(private.moves + moves_delta));
  private.moves = private.moves + moves_delta;

  mediator.Notify(
    "Model_Stage_MovesCountChanged",
    { moves = private.moves, delta = moves_delta }
  );

  return true;

end;
--------------------------------------------------------------------------------
function private.Moves_IsAvailable ()

  if ( not private.moves ) then

    return false;

  end;

  return ( private.moves > 0 );

end;
--------------------------------------------------------------------------------
function private.Moves_GetCount ()

  return private.moves;

end;
--------------------------------------------------------------------------------
function private.Moves_IncDoneCount ()

  private.moves_done = private.moves_done + 1;

end;
--------------------------------------------------------------------------------
function private.Moves_GetDoneCount ()

  return private.moves_done;

end;
--------------------------------------------------------------------------------
function private.AddMovesFromChip ( chips_list )
  
  local move_bonus = 0;
  local chips = {};
  
  for i = 1, #chips_list do
    if ( chips_list[ i ].type == "MoveBonus" ) then
      move_bonus = move_bonus + chips_list[ i ].moves;
      table.insert( chips, {
        x = chips_list[ i ].x,
        y = chips_list[ i ].y,
        color = chips_list[ i ].color,
        move_bonus = chips_list[ i ].moves
      } );
    end;
  end;
  
  if ( move_bonus > 0 ) then
    private.kontagent.additional_moves = private.kontagent.additional_moves + move_bonus;
    DbgTrace("Add move bonus");
    private.Moves_TryChangeCount( move_bonus, true );
    mediator.Notify( "Model_Stage_MoveBonusRemoved", chips );
  end;

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function private.InitChains ()

  private.combination = { form = nil, score = nil };
  private.chain_count = 0;

end;
--------------------------------------------------------------------------------
function private.IncChainCount ()

  private.chain_count = private.chain_count + 1;

  mediator.Notify( "Model_Stage_ChainCountChanged", 
    {
      count = private.chain_count
    } );

end;
--------------------------------------------------------------------------------
function private.ResetChainCount ()

  private.combination.form = nil;

  private.chain_count = 0;

end;
--------------------------------------------------------------------------------
-- function ******* Fight  ******* () end
--------------------------------------------------------------------------------
function private.InitFight ( player_data, enemy_data )
  
  private.InitHero( player_data );
  private.InitEnemy( enemy_data );
  private.DeactivateCharacter();
  
end;
--------------------------------------------------------------------------------
function private.IncCountMovesFight ()

  if ( private.state_data.type ~= "fight" ) then
    return;
  end;

  private.ActivateFirstCharacter();
  private.ChangeCounterBonus();

end;
--------------------------------------------------------------------------------
function public.IsCurrentStageFight ()
  --ES> используется только в модели ачивок (убрать) или сделать для контроллера еще использование метода
  return ( private.state_data.type == "fight" );

end;
--------------------------------------------------------------------------------
-- function ... Hero, Enemy  () end
--------------------------------------------------------------------------------
function private.InitHero ( player_data )

  private.hero                     = player_data;
  private.hero.health_current      = private.hero.health;
  private.ResetAttackFactorHero();
  private.hero.ability = nil;

end;
--------------------------------------------------------------------------------
function private.InitEnemy ( enemy_data )

  private.enemy = enemy_data or private.state_data.enemy;

  if ( private.enemy ) then
  
    local attack = private.hero.enemy_attack;
    private.enemy.attack = private.enemy.attack + attack;
    private.enemy.attack = math.max( private.enemy.attack, 5 );
    
    local health = private.hero.enemy_health;
    private.enemy.health = private.enemy.health + health;
    private.enemy.health = math.max( private.enemy.health, 50 );
    private.enemy.health_current = private.enemy.health;

  end;

end;
--------------------------------------------------------------------------------
function public.GetCurrentHero ()

  return private.hero;

end;
--------------------------------------------------------------------------------
function public.GetCurrentEnemy ()

  return private.enemy;

end;
--------------------------------------------------------------------------------
function private.DeactivateCharacter ()
  
  private.fight_character = nil;
  
end;
--------------------------------------------------------------------------------
function private.ActivateFirstCharacter ()
  
  private.fight_character = "hero";
  
end;
--------------------------------------------------------------------------------
function private.ActivateSecondCharacter ()
  
  private.fight_character = "enemy";
  
end;
--------------------------------------------------------------------------------
function private.IsHeroFightCharacter ()
  
  return private.fight_character == "hero";
  
end;
--------------------------------------------------------------------------------
function private.IsEnemyFightCharacter ()
  
  return private.fight_character == "enemy";
  
end;
--------------------------------------------------------------------------------
function private.IsHeroDead ()

  return ( private.hero.health_current == 0 );

end;
--------------------------------------------------------------------------------
function private.IsEnemyDead ()

  return ( private.enemy.health_current == 0 );

end;
--------------------------------------------------------------------------------
function private.AddAttackFactorHero ( count )

  if ( not private.IsHeroDead() ) then
    private.hero.attack_factor = private.hero.attack_factor + count;
  end;

end;
--------------------------------------------------------------------------------
function private.ResetAttackFactorHero ()
  
  private.hero.attack_factor = 0;

end;
--------------------------------------------------------------------------------
function private.ChangeCurrentHealthHero ( value )
  
  private.hero.health_current = private.hero.health_current + value;

  if ( private.hero.health_current > private.hero.health ) then
    private.hero.health_current = private.hero.health;
  end;
  
  if ( private.hero.health_current < 0 ) then
    private.hero.health_current = 0;
  end;
  
end;
--------------------------------------------------------------------------------
function private.ChangeCurrentHealthEnemy ( value )
  
  private.enemy.health_current = private.enemy.health_current + value;

  if ( private.enemy.health_current < 0 ) then
    private.enemy.health_current = 0;
  end;
  
end;
--------------------------------------------------------------------------------
function private.AddExtraHealthHero ()
  
  local count = math.ceil( 0.5 * private.hero.health )
  private.ChangeCurrentHealthHero( count );
  
  mediator.Notify( "Model_Stage_HeroHealthRecovered",
  {
    delta          = count,
    health_current = private.hero.health_current,
    health_total   = private.hero.health
  } );
  
end;
--------------------------------------------------------------------------------
function private.SetHeroAbility ( ability )
  
  private.hero.ability = ability;
  
end;
--------------------------------------------------------------------------------
function private.TryActivateHeroAttack ()
  
  if ( private.IsEnemyDead() ) then
    return false;
  end;

  local additional_damage = 0;
  local bonus_attack_factor = 1;
  local ability_name = nil;
  
  private.ChangeAbilityCounterBonus();
  private.ActivateSecondCharacter();
  
  if ( private.hero.ability ) then
    ability_name = private.hero.ability.name;
    if ( private.hero.ability.name == "dot" ) then
      additional_damage = private.hero.ability.value;
    elseif ( private.hero.ability.name == "damage" ) then
      bonus_attack_factor = private.hero.ability.factor;
    elseif ( private.hero.ability.name == "superhit" ) then
      private.SetHeroAbility( nil );
      private.DeactivateCharacter();
    elseif ( private.hero.ability.name == "stun" ) then
      private.DeactivateCharacter();
    end;
    
  end;

  local damage = bonus_attack_factor * private.hero.attack_factor *
                 private.hero.attack + additional_damage;

  private.ChangeCurrentHealthEnemy( -damage );
  
  mediator.Notify( "Model_Stage_HeroAttackActivated", 
  { 
    damage        = damage,
    ability       = ability_name, 
    ability_delta = additional_damage,
    enemy_health  = private.enemy.health_current
  } );

  private.ResetAttackFactorHero();
  
  return true;

end;
--------------------------------------------------------------------------------
function private.TryActivateEnemyAttack ()
  
  if ( private.IsEnemyDead() or ( private.hero.ability and private.hero.ability.name == "stun" ) ) then
    return false;
  end;
  
  private.ChangeCurrentHealthHero( -private.enemy.attack );
  private.DeactivateCharacter();
  
  mediator.Notify( "Model_Stage_EnemyAttackActivated",
  {
    damage      = private.enemy.attack,
    hero_health = private.hero.health_current
  } );
  
  return true;

end;
--------------------------------------------------------------------------------
-- function ... Bonus ()
--------------------------------------------------------------------------------
function private.InitBonus ()

  local bonus_data = private.state_data.bonus;

  if ( not bonus_data ) or ( private.state_data.type ~= "fight" ) then
    private.bonus = nil;
    return;
  end;

  private.bonus = {};
  private.bonus.abilities = bonus_data[ "hero" ].ability;
  private.bonus.generate_range = bonus_data.range;
  
  private.GenerateAbilityBonus();   
  private.GenerateMoveBonus();
  private.GenerateChipTypesBonus();

end;
--------------------------------------------------------------------------------
function private.GenerateAbilityBonus ()
  
  local ability_data = private.bonus.abilities;
  private.bonus.ability = ability_data[ math.random( 1, #ability_data ) ];
  private.bonus.ability.counter = private.bonus.ability.period;
  
end;
--------------------------------------------------------------------------------
function private.GenerateMoveBonus ()

  local range = private.bonus.generate_range;
  private.bonus.move = math.random( range[ 1 ], range[ 2 ] );
  private.bonus.counter = 0;

end;
--------------------------------------------------------------------------------
function private.GenerateChipTypesBonus ()

  local req = "Request_Match3_GetPrevalentChipsList"
  private.bonus.chip_type = common.MessageSend( req ).type;

end;
--------------------------------------------------------------------------------
function private.ResetBonus ()
  
  private.GenerateAbilityBonus();
  private.GenerateMoveBonus();
  private.SetHeroAbility( nil );
  
end;
--------------------------------------------------------------------------------
function private.ChangeCounterBonus ()
  
  if ( private.bonus ) then
    private.bonus.counter = private.bonus.counter + 1;
  end;
  
end;
--------------------------------------------------------------------------------
function private.ChangeAbilityCounterBonus ()

  if ( public.IsActivateBonus() ) and ( private.bonus.ability ) then
    private.bonus.ability.counter = private.bonus.ability.counter - 1;
    mediator.Notify( "Model_Stage_BonusCountChanged", private.bonus.ability );
  end;

end;
--------------------------------------------------------------------------------
function private.CheckActivateBonus ( chip_type )

  if ( not private.bonus ) or ( chip_type ~= private.bonus.chip_type ) then
    return;
  end;
  
  if ( private.bonus.enabled ) and ( not public.IsActivateBonus() ) then
    private.bonus.activated = true;
    private.bonus.enabled = false;
    private.SetHeroAbility( private.bonus.ability );
    mediator.Notify( "Model_Stage_BonusCountChanged", private.bonus.ability );
    mediator.Notify( "Model_Stage_BonusDisable" );
  end;

end;
--------------------------------------------------------------------------------
function private.CheckEndingActionAbility ()
  
  if ( private.bonus.ability.counter <= 0 ) then
    private.bonus.activated = false;
    private.ResetBonus();
    mediator.Notify( "Model_Stage_BonusEffectDisabled" );
  end;
  
end;
--------------------------------------------------------------------------------
function public.IsActivateBonus ()
  
  if ( not private.bonus ) then
    return false;
  end;

  return private.bonus.activated;
  
end;
--------------------------------------------------------------------------------
function public.GetDataBonus ()

  if ( not private.bonus ) or ( not private.bonus.enabled ) then
    return nil;
  end;

  return { 
    chip_type = private.bonus.chip_type, 
    ability = private.bonus.ability.name 
  };

end;
--------------------------------------------------------------------------------
function private.HandleFightBonus ()

  if ( not private.bonus ) then 
    return; 
  end;

  if ( public.IsActivateBonus() ) then
    private.CheckEndingActionAbility();
  end;

  if ( private.bonus.counter == private.bonus.move ) and ( not public.IsActivateBonus() ) then
    private.bonus.enabled = true;
    private.GenerateChipTypesBonus();
  end;

end;
--------------------------------------------------------------------------------
-- function ******* Score  ******* () end
--------------------------------------------------------------------------------
function private.InitScore ()

  private.score = 0;

end;
--------------------------------------------------------------------------------
function public.GetCurrentScore ()

  return private.score;

end;
--------------------------------------------------------------------------------
function public.FinalBonuses_ChangeScore ( chip_x, chip_y, chip_type, stage )
  --ES> в одном месте использовалось private.state_data.state надо проверить
  -- есть ли такая переменная и убрать что ниже для получения тира
  local cascade_score = private.GetCascadeScore() * public.Stage_GetState( stage );
  private.ChangeScore( chip_x, chip_y, chip_type, cascade_score );
  
end;
--------------------------------------------------------------------------------
function private.ChangeScore ( chip_x, chip_y, chip_type, chip_score )

  local delta_score = math.floor( chip_score * ( 1 + private.hero.score ) );

  private.score = private.score + delta_score;

  mediator.Notify(
    "Model_Stage_ScoreChanged",
    {
      chip_x     = chip_x,
      chip_y     = chip_y,
      chip_type  = chip_type,
      chip_score = delta_score,
      score      = private.score
    }
  );

  private.UpdateTargetProgress_GainScore( delta_score );

end;
--------------------------------------------------------------------------------
-- function ******* Targets ******* () end
--------------------------------------------------------------------------------
function private.InitTargets ()

  -- GainScore, DropItem, DestroyCell, CollectResource, DestroyGem

  private.targets    = {};
  private.targets_id = {};
  private.targets_drop_chests = false;

  local targets = private.state_data.targets;

  for t = 1, #targets do

    local target_name = targets[ t ].name;
    local target_type = targets[ t ].type;
    
    if ( not private.targets[ target_type ] ) then
      
      private.targets[ target_type ] = {};
      
    end;

    targets[ t ].progress = 0;
    targets[ t ].done = false;
    targets[ t ].index = t;
    
    if ( target_name and target_name:match( "chest" ) ) then
      private.targets_drop_chests = true;
    else
      private.targets_drop_chests = false;
    end; 
    
    private.targets_id[ target_name or target_type ] = t;
    private.targets[ target_type ][ t ] = targets[ t ];
    
    private.ActivateTarget( target_type, targets[ t ] );

  end;

end;
--------------------------------------------------------------------------------
function private.GetTargetByType ( target_type, target_name )
  
  local target_index = private.targets_id[ target_name or target_type ];
  if ( target_index ) then
    return private.targets[ target_type ][ target_index ];
  end;
  
  return nil;

end;
--------------------------------------------------------------------------------
function private.ActivateTarget ( target_type, target_data )

  if ( target_type:match( "GainScore" ) ) then

    private.UpdateTargetProgress_GainScore( 0 );

  elseif ( target_type:match( "DropItem" ) ) then

    private.UpdateTargetProgress_DropItem( target_data.name, 0 );

  elseif ( target_type:match( "DestroyCell" ) ) then

    private.UpdateTargetProgress_DestroyCell( target_data.name, target_data.level, 0 );

  elseif ( target_type:match( "CollectResource" ) ) then

    private.UpdateTargetProgress_CollectResource( target_data.name, 0 );
    
  elseif ( target_type:match( "DestroyGem" ) ) then

    private.UpdateTargetProgress_DestroyGem( target_data.name, 0 );

  end;

end;
--------------------------------------------------------------------------------
function private.UpdateTargetProgress_GainScore ( progress_delta )

  local target = private.GetTargetByType( "GainScore" );

  if ( target ) then  

    target.progress = target.progress + progress_delta;

    private.UpdateTarget( target.type, nil, progress_delta );

    mediator.Notify(
      "Model_Stage_TargetGainScoreUpdated", 
      { 
        delta = progress_delta
      }
    );

  end;

end;
--------------------------------------------------------------------------------
function private.UpdateTargetProgress_DropItem ( item_name, progress_delta )

  local target = private.GetTargetByType( "DropItem", item_name );

  if ( target and not target.done and item_name == target.name ) then

    local progress_allowable = target.count - target.progress;

    if ( progress_delta > progress_allowable ) then

      progress_delta = progress_allowable;

    end;

    target.progress = target.progress + progress_delta;

    private.UpdateTarget( target.type, item_name, progress_delta );

  end;

end;
--------------------------------------------------------------------------------
function private.UpdateTargetProgress_DestroyCell ( cell_name, cell_level, progress_delta )

  local target = private.GetTargetByType( "DestroyCell", cell_name );

  if ( target
       and not target.done
       and target.name == cell_name
       and 1 == cell_level
  ) then

    local progress_allowable = target.count - target.progress;

    if ( progress_delta > progress_allowable ) then

      progress_delta = progress_allowable;

    end;

    target.progress = target.progress + progress_delta;

    private.UpdateTarget( target.type, cell_name, progress_delta );

  end;

end;
--------------------------------------------------------------------------------
function private.UpdateTargetProgress_CollectResource ( resource_name, progress_delta )

  local target = private.GetTargetByType( "CollectResource", resource_name );

  if ( target and target.name == resource_name ) then

    target.progress = target.progress + progress_delta;

    private.UpdateTarget( target.type, resource_name, progress_delta );

  end;

end;
--------------------------------------------------------------------------------
function private.UpdateTargetProgress_DestroyGem ( gem_name, progress_delta )

  local target = private.GetTargetByType( "DestroyGem", gem_name );

  if ( target
       and not target.done
       and target.name == gem_name
  ) then

    local progress_allowable = target.count - target.progress;

    if ( progress_delta > progress_allowable ) then

      progress_delta = progress_allowable;

    end;

    target.progress = target.progress + progress_delta;

    private.UpdateTarget( target.type, gem_name, progress_delta );

  end;

end;
--------------------------------------------------------------------------------
function private.UpdateTarget ( target_type, target_name, progress_delta )

  local target = private.GetTargetByType( target_type, target_name );

  --DbgTrace( "UpdateTarget: < "..target_type.." > "..target.progress.." / "..target.count );

  mediator.Notify( "Model_Stage_TargetUpdated",
  {
    name     = target.name,
    index    = target.index,
    type     = target_type,
    count    = target.count,
    progress = target.progress,
    delta    = progress_delta,
    done     = target.done
  } );

end;
--------------------------------------------------------------------------------
function private.TryCompleteTargets ()

  for target_type, value in pairs( private.targets ) do
    for target_index, target in pairs( value ) do
      if ( not target.done and target.progress >= target.count ) then

        target.done = true;

        private.UpdateTarget( target.type, target.name );
        mediator.Notify( "Model_Stage_TargetComplete", target );

      end;
    end;
  end;

end;
--------------------------------------------------------------------------------
function private.IsConditionCompleted ()

  local stage_type = private.state_data.type;

  if ( stage_type == "ride" ) then

    return ( private.IsTargetsCompleted() );

  elseif ( stage_type == "fight" ) then

    return ( private.IsEnemyDead() );

  end;

end;
--------------------------------------------------------------------------------
function private.IsConditionFailed ()

  local stage_type = private.state_data.type;

  if ( stage_type == "ride" ) then
    return ( not private.Moves_IsAvailable() and not private.IsTargetsCompleted() );
  elseif ( stage_type == "fight" ) then
    return ( private.IsHeroDead() );
  end;

end;
--------------------------------------------------------------------------------
function public.IsExplosionField ()
  
  return private.IsExplosionField();
  
end;
--------------------------------------------------------------------------------
function public.IsCombinationsNotExist ()
  
  return private.IsCombinationsNotExist();
  
end;
--------------------------------------------------------------------------------
function private.IsExplosionField ()

  if ( private.explosion_field ) then

    return true;

  end;

  return false;

end;
--------------------------------------------------------------------------------
function private.IsCombinationsNotExist ()

  if ( private.no_combinations ) then

    return true;

  end;

  return false;

end;
--------------------------------------------------------------------------------
function private.IsTargetsCompleted ()

  local completed = true;

  for target_type, value in pairs( private.targets ) do
  
    for index, target_data in pairs( value ) do

      if ( not target_data.done ) then

        completed = false;

        break;

      end;

    end;

  end;

  return completed;

end;
--------------------------------------------------------------------------------
function private.FinalBonuses_Spawn ()

  private.Moves_TryChangeCount( - 1 );
  local variants = { "Horizontal", "Vertical" };
  local state_bonus = variants[ math.random( 1, #variants ) ];
  local params = {};
  params.state = state_bonus;
  params.delay = 0.5;
  local chip_data = common.MessageSend( Request_Match3_BonusSpawnPlace, params );
  local moves     = private.Moves_GetCount();
  
  mediator.Notify(
    "Model_Stage_FinalBonuses_StartAnimation",
    {
      chip_x    = chip_data.x,
      chip_y    = chip_data.y,
      chip_type = state_bonus,
      stage     = private.stage_name,
      moves     = moves
    }
  );

  if ( moves == 0 ) then

    private.FinalBonuses_Activate();

    return;

  end;
  
  local tme = moves / ( private.moves_remain * 3 );

  ObjSet( "tmr_stage_finalbonuses",
  {
    time    = tme,
    playing = true,
    endtrig = function ()
      
      mediator.Notify( "View_Match3_FieldEstablished" );

    end
  } );

  mediator.Notify( "Model_Match3_BonusSpawn", { type = "Horizontal" } );

end;
--------------------------------------------------------------------------------
function public.CreateBonusesOnField ( bonuses, source )
  
  for i = 1, #bonuses do
    public.CreateBonusOnField( bonuses[ i ], source );
  end;
  
end;
--------------------------------------------------------------------------------
function public.CreateBonusOnField ( bonus, source )
  
  local time = 0.5;
  local variants = { 
    Horizontal = true, 
    Vertical = true, 
    Nucleus = true, 
    Bomb = true 
  };

  if ( source == "minivent_progress_crowns" ) then
    time = 0.0;
  end;
  
  if ( variants[ bonus ] ) then
    local params = {};
    params.state = bonus;
    params.delay = time;
    local chip_data = common.MessageSend( Request_Match3_BonusSpawnPlace, params );
  end;
  
end;
--------------------------------------------------------------------------------
function private.FinalBonuses_Activate ()
  
  private.bonuses_activated = true; 

  common.MessageSend( Command_Match3_ActivateFinalBonuses );

end;
--------------------------------------------------------------------------------
-- function ******* Rewards ******* () end
--------------------------------------------------------------------------------
function private.InitRewards ()

  private.reward = {};

  local chip_types = private.state_data.chip_types;

  for i = 1, #chip_types do
  
    if ( string.lower( chip_types[ i ] ) == "resource1" or
         string.lower( chip_types[ i ] ) == "resource2" or
         string.lower( chip_types[ i ] ) == "resource3" or
         string.lower( chip_types[ i ] ) == "resource4" ) then

      private.reward[ string.lower( chip_types[ i ] ) ] = 0;

    end;

  end; 

end;
--------------------------------------------------------------------------------
function private.UpdateRewards ( chips_list )

  -- FIXME
  local reward_types = {
    Resource1 = "resource1",
    Resource2 = "resource2",
    Resource3 = "resource3",
    Resource4 = "resource4"
  };

  local reward_buffer = {};

  for i = 1, #chips_list do

    local resource_name = reward_types[ chips_list[ i ].type ];
    
    if ( resource_name ) then

      private.IncRewardCount(
        resource_name,
        1,
        chips_list[ i ].x,
        chips_list[ i ].y
      );

      private.UpdateTargetProgress_CollectResource( resource_name, 1 );

      table.insert( reward_buffer,
        {
          name  = resource_name,
          count = private.reward[ resource_name ],
          delta = 1,
          x     = chips_list[ i ].x,
          y     = chips_list[ i ].y
        }
      );

    end;

  end;

  if ( #reward_buffer > 0 ) then

    mediator.Notify( "Model_Stage_RewardCountChanged", reward_buffer );

  end;

end;
--------------------------------------------------------------------------------
function private.IncRewardCount ( resource_name, delta, resource_x, resource_y )

  private.reward[ resource_name ] = private.reward[ resource_name ] or 0;
  private.reward[ resource_name ] = private.reward[ resource_name ] + delta;

  --mediator.Notify( "Model_Stage_RewardCountChanged", { name = resource_name, count = private.reward[ resource_name ], delta = delta, x = resource_x, y = resource_y } );

end;
--------------------------------------------------------------------------------
function public.Stats ()
  
  local func_get_stage_data = function ( stages_data, stage_name ) 
    for i = 1, #stages_data do
      if ( stage_name == stages_data[ i ].name ) then 
        return stages_data[ i ]; 
      end;
    end;
  end;

  local query =
  {
    query      = "find_all",
    group      = "static",
    collection = "stages"
  };
  
  local stages_data = lib_db.Execute( query ).result or {};

  query =
  {
    query      = "find_all",
    group      = "player",
    collection = "stages"
  };
  
  local save_stages = lib_db.Execute( query ).result;

  query =
  {
    query      = "find_all",
    group      = "static",
    collection = "stages",
    indice     = "type",
    criteria   = { like = "bonus" }
  };
  
  local bonus_stages = lib_db.Execute( query ).result;

  local result = {
    total_states       = 0;
    total_count        = #stages_data or 0;
    completed_states   = 0;
    completed_count    = 0;
    bonus_total        = #bonus_stages or 0;
    bonus_count        = 0;
    bonus_states       = 0;
    bonus_total_states = 0;
  };

  -- -- -- count totals -- -- -- 

  for s = 1, #stages_data do

    if ( stages_data[ s ].type ~= "dailyquest" ) and 
       ( stages_data[ s ].type ~= "minievent" ) then

      local section = stages_data[ s ].section;

      if ( not result[ section ] ) then

        result[ section ] = {
          bonus_total        = 0,
          bonus_count        = 0,
          bonus_states       = 0,
          bonus_total_states = 0,
          total_states       = 0,
          total_count        = 0,
          completed_count    = 0,
          completed_states   = 0,
          stages_prev        = 0
        };

      end;

      result[ section ].total_states = result[ section ].total_states + #stages_data[ s ].state;
      result[ section ].total_count  = result[ section ].total_count + 1;
      result.total_states            = result.total_states + #stages_data[ s ].state;

    else

      result.total_count = result.total_count - 1;

    end;

  end;

  -- -- -- count completed -- -- -- 

  for i = 1, #save_stages do

    local stage_data = func_get_stage_data( stages_data, save_stages[ i ].name );
    local section = stage_data.section;

    if ( stage_data.type ~= "bonus" ) then

      if ( save_stages[ i ].completed ) then

        result[ section ].completed_count = result[ section ].completed_count + 1;
        result.completed_count            = result.completed_count + 1;

      end;

      if ( save_stages[ i ].state ) then

        result[ section ].completed_states = result[ section ].completed_states + save_stages[ i ].state - 1;
        result.completed_states            = result.completed_states + save_stages[ i ].state - 1;

      end;

      if ( save_stages[ i ].noreward ) then

        result[ section ].completed_states = result[ section ].completed_states + 1;
        result.completed_states            = result.completed_states + 1;

      end;

    end;

  end;

  -- -- -- bonus stages -- -- -- 

  if ( bonus_stages and #bonus_stages > 0 ) then

    for b = 1, #bonus_stages do

      local saved_stage = func_get_stage_data( save_stages, bonus_stages[ b ].name );
      local section = bonus_stages[ b ].section;

      result[ section ].bonus_total        = result[ section ].bonus_total + 1;
      result[ section ].bonus_total_states = result[ section ].bonus_total_states + #bonus_stages[ b ].state;
      result.bonus_total_states = result.bonus_total_states + #bonus_stages[ b ].state;

      if ( saved_stage ) then

        if ( saved_stage.completed ) then

          result[ section ].bonus_count = result[ section ].bonus_count + 1;
          result.bonus_count            = result.bonus_count + 1;

        end;

        if ( saved_stage.state ) then

          result[ section ].bonus_states = result[ section ].bonus_states + saved_stage.state - 1;
          result.bonus_states            = result.bonus_states + saved_stage.state - 1;

        end;

        if ( saved_stage.noreward ) then

          result[ section ].bonus_states = result[ section ].bonus_states + 1;
          result.bonus_states           = result.bonus_states + 1;

        end;

      end;

    end;

  end;

  -- -- -- stages_prev to section results -- -- -- 
  query =
  {
    query      = "find_all",
    group      = "static",
    collection = "sections"
  };
  
  local section_list = lib_db.Execute( query ).result;

  for s = 2, #section_list do

    local current  = result[ section_list[ s ].name ];
    local previous = result[ section_list[ s - 1 ].name ];

    if ( current ) and ( previous ) then

      current.stages_prev = ( s - 1 ) * ( previous.total_count - previous.bonus_total );

    else

      break;
 
    end;

  end;

  return result;
  
end;
--------------------------------------------------------------------------------
-- function ******* Kontagent ******* () end
--------------------------------------------------------------------------------
function  public.Kontagent_SetBoostersUsed ( value, booster_name )

  private.boosters_used = value;

  if ( value == 0 ) then

    private.continue_used = 0;

  end;

  if ( booster_name ) then

    private.continue_used = private.continue_used or 0;

    private.continue_used = private.continue_used + 1;

  end;

end;
--------------------------------------------------------------------------------
function  public.Kontagent_GetBoostersUsed ()

  return { private.boosters_used, private.continue_used };

end;
--------------------------------------------------------------------------------
function  public.Kontagent_SetStageData ( stage_data )

  private.kontagent_stage_data = stage_data;

end;
--------------------------------------------------------------------------------
function  public.Kontagent_GetStageState ( stage_name )

  local stage_data  = private.kontagent_stage_data or {};
  local stage_state = stage_data.state or 
                      ( public.GetSavedData( stage_name ) or {} ).state or 1;
  
  if ( stage_data.noreward ) then
    stage_state = 4;

  end;

  return stage_state;

end;
--------------------------------------------------------------------------------
function  public.Kontagent_GetCellsDestroyed ()

  local target = private.GetTargetByType( "DestroyCell" );

  if ( target ) then

    return { type = target.name, value = ( target.progress..", "..target.count ) };

  end;

  return nil;

end;
--------------------------------------------------------------------------------
function  public.Kontagent_GetItemsDropped ( target_name )

  local target = private.GetTargetByType( "DropItem", target_name );

  if ( target ) then

    return target.progress..", "..target.count;

  end;

  return nil;

end;
--------------------------------------------------------------------------------
function  public.Kontagent_GetResourceCollected ()

  local target = private.GetTargetByType( "CollectResource" );

  if ( target ) then

    return target.progress..", "..target.count;

  end;

  return nil;

end;
--------------------------------------------------------------------------------
function  public.Kontagent_GetExplosivesDestroyed ()

  local explosives = public.GetCurrentStageData().explosives;

  if ( explosives ) then
    
    local explosives_remain = ( common.MessageSend( Request_Match3_GetExplosivesRemain ) or {} ).explosives;

    if ( not explosives_remain ) then
      return nil;
    end;
    
    if ( private.IsCombinationsNotExist() or private.IsExplosionField() ) then
      
      explosives_remain = explosives_remain + 1;
      
    end;
       
    return ( #explosives - explosives_remain )..", "..#explosives;

  end;

  return nil;

end;
--------------------------------------------------------------------------------
function public.Kontagent_GetCurrentStage ()

  return private.kontagent_stage;

end;
--------------------------------------------------------------------------------
function public.Kontagent_SetCurrentStage ( stage_name )

  private.kontagent_stage = stage_name;

end;
--------------------------------------------------------------------------------
function public.Kontagent_GetExtraMoves ()

  return private.kontagent.extra_moves;

end;
--------------------------------------------------------------------------------
function public.Kontagent_GetAdditionalMoves ()

  return private.kontagent.additional_moves;

end;
--------------------------------------------------------------------------------
-- function ******* Usable Lists ******* () end
--------------------------------------------------------------------------------
function private.Usable_List ( section_name )
  
  local query = 
  {
    query      = "find_all",
    group      = "player",
    collection = "stages",
    indice     = "completed",
    criteria   = { like = true }
  };

  local usable = lib_db.Execute( query ).result;
  
  if not ( usable ) then return false; end;
  
  local list = {};
  
  for s = 1, #usable do 
    
    if ( public.GetStaticData( usable[ s ].name ).section == section_name ) then
      
      table.insert( list, usable[ s ] );
      
    end;
  
  end;
  
  if ( #list < 1 ) then return false; end;

  return list;

end;
--------------------------------------------------------------------------------
function public.Usable_Search ( deficient_name, section_name )

  if ( deficient_name:find( "resource" )) then

    local resource_name = deficient_name;
    local usable_stages = private.Usable_List( section_name );

    if ( usable_stages ) then

      local selected = {};

      for i = 1, #usable_stages do

        local stage_name  = usable_stages[ i ].name;
        local stage_data  = public.GetStaticData( stage_name );
        local stage_info  = public.GetSavedData(  stage_name ) or {};
        local stage_state = stage_info.state or 1;
        local chip_types  = stage_data.state[ stage_state ].chip_types;

        for t = 1, #chip_types do

          if ( string.lower( chip_types[ t ] ) == resource_name ) then
            
            table.insert( selected, usable_stages[ i ]);

          end;

        end;      

      end;

      if ( #selected > 1 ) then

        local random = math.random ( 1, #selected );

        return selected[ random ];

      elseif ( #selected > 0 ) then

        return selected[ 1 ];

      else

        return false;

      end;

    end;

  else

    local item_name = deficient_name;
    local stage_list = private.GetStagesBySection( section_name );

    for i = 1, #stage_list do

      local stage = stage_list[ i ];

      for j = 1, #stage.state do

        local state = stage.state[ j ];

        if ( state.reward ) then

          for k = 1, #state.reward do

            local reward = state.reward[ k ];

            if ( reward.name == item_name ) then

              return stage;

            end;

          end;

        end;

      end;

    end;

  end;

  return false;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Cheat_Win ()

  private.score     = 99999;
  private.completed = true;
  
  if ( private.state_data.type == "fight" ) then
    
    private.moves_done = 1;
    private.enemy.health_current = 0;

  else

    private.moves = 0;

    for target_type, value in pairs( private.targets ) do
      for target_index, target in pairs( value ) do
        target.progress = target.count;
      end;
    end;

  end;
  
  mediator.Notify( "Model_Stage_ViewTryContinue" );

end;
--------------------------------------------------------------------------------
function public.Cheat_Fail ()

  if ( private.Moves_GetDoneCount() == 0 ) then

    private.moves_done = 1;

  end;

  if ( private.state_data.type == "fight" ) then

    private.hero.health_current = 0;

  else

    local moves = private.moves or 0;
    DbgTrace("Cheat_Fail");
    private.Moves_TryChangeCount( - moves, true );

  end;
  
  mediator.Notify( "Model_Stage_ViewTryContinue" );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Stages_FindByReward (
    section_name,
    reward_type,
    reward_name,
    ignore_state
  )

  local stage_list = private.GetStagesBySection( section_name ) or {};

  if ( #stage_list < 1 ) then

    return stage_list;

  end;

  local list = {};

  for _, stage_data in pairs( stage_list ) do

    for state_index, stage_state in pairs( stage_data.state ) do

      if ( stage_state.reward ) then

        for _, reward_data in pairs( stage_state.reward ) do
        
          if ( reward_data.type == reward_type ) then
          
            local is_name_valid = ( reward_data.name == reward_name );
            
            if ( reward_type == "fragment" ) then
              is_name_valid = reward_data.name:find( reward_name );
            end;
            
            if ( is_name_valid ) then
            
              local saved_data  = public.GetSavedData( stage_data.name ) or {};
              local saved_state = saved_data.state or 0;
              local has_reward  = ( not saved_data.noreward );
              
              local do_insert = false;
              
              if ( ignore_state ) then
                do_insert  = true;
              else
                if ( saved_state <= state_index ) and
                   ( has_reward                 ) then
                  do_insert  = true;
                end;
              end;
              
              --Log.i( reward_data.name.." is reward at "..stage_data.name..": "..tostring( do_insert ) );
              
              if ( do_insert ) then
                table.insert( list, stage_data.name );
              end;
            
            end;
          
          end;

        end;

      end;

    end;

  end;
  
  --Log.i( "found "..#list.." stages" );

  return list;

end;
--------------------------------------------------------------------------------
function public.Stages_GetAllCompleted ()

  local query = {
    query      = "find_all",
    group      = "player",
    collection = "stages",
    indice     = "completed",
    criteria   = {
      like = true
    }
  };

  local stages_data = lib_db.Execute( query ).result;
  
  return stages_data;
  
end;
--------------------------------------------------------------------------------