-- name=mediator
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.queue_on = true;
  private.cheat_queue_off = false;

  private.handlers = {};
  private.queue    = {};
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.CheatDisableQueue ()
  private.cheat_queue_off = true;
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Notify ( message_type, message_params )

  -- HACK: Always must be a hash-table
  local packed_params  = {};
  packed_params.params = message_params;

  if ( private.queue_on ) then
 
    public.Queue_Add( message_type, packed_params );

    return;

  end;

  common.MessageSend( message_type, packed_params );

end;
--------------------------------------------------------------------------------
function public.Subscribe ( message_type, handler_func )

  return private.Subscribe( message_type, handler_func, false );

end;
--------------------------------------------------------------------------------
function public.SubscribeOnce ( message_type, handler_func )

  return private.Subscribe( message_type, handler_func, true );
  
end;
--------------------------------------------------------------------------------
function public.Unsubscribe ( message_type, handler_func )

  private.handlers[ message_type ] = private.handlers[ message_type ] or {};
  local msg_handlers = private.handlers[ message_type ];

  local handler = msg_handlers[ handler_func ];

  if ( handler ) then

    ne.msg.Unsubscribe( message_type, handler );
    msg_handlers[ handler_func ] = nil;

    if ( next( msg_handlers ) == nil ) then

      msg_handlers = nil;

    end;

  elseif ( IsEditor() ) then

    local msg = "Warning! Unsubscribe from not handled message '%s'"
    common.DbgCallpoint( msg:format( message_type ) );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Queue_Add ( message_type, message_params )

  local message = {
    type   = message_type,
    params = message_params
  };

  if not private.cheat_queue_off then
    table.insert( private.queue, message );
  end

end;
--------------------------------------------------------------------------------
function public.Queue_Release ()

  private.queue_on = false;

  for _, message in pairs( private.queue ) do

    common.MessageSend( message.type, message.params );

  end;

  if ( not IsEditor() ) then

    private.queue = {};

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Subscribe ( message_type, handler_func, once )

  private.handlers[ message_type ] = private.handlers[ message_type ] or {};
  local handlers = private.handlers[ message_type ];

  if ( handlers[ handler_func ] ) then

    if ( IsEditor() ) then

      local msg = "Warning! Double function subscribe for message '%s'"
      common.DbgCallpoint( msg:format( message_type ), 1 );

    end;
    
    return false;

  end;

  handlers[ handler_func ] = function ( message_type, packed_params )

    local unpacked_params = packed_params.params;
    handler_func( unpacked_params );

    if ( once ) then

      public.Unsubscribe( message_type, handler_func );

    end;

  end;

  ne.msg.Subscribe( message_type, handlers[ handler_func ] );

  return true;

end;
--------------------------------------------------------------------------------