-- name=ui
--------------------------------------------------------------------------------
local CONTENT_PATH     = "assets/content/";
local INTERFACE_PATH   = CONTENT_PATH.."interface/";
local DEFAULT_SKIN_RES = CONTENT_PATH.."skins_1/default/anm_skin_default";
--------------------------------------------------------------------------------
public.obj      = {};
public.hub      = {};
public.button   = {};
public.buttons  = {};
public.tap      = {};
public.control  = {};
public.popup    = {};
public.osd      = {};
public.skin     = {};
public.input    = {};
public.text     = {};
public.swipe    = {};
public.window   = {};
public.screen   = {};
public.scroller = {};
public.timer    = {};
public.anim     = {};
public.grid     = {};
public.skin     = {};
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  GetRes                 = function () end,
  IsFullscreen           = function () end,
  SetOuterTap            = function () end,
  ClearOuterTap          = function () end,
  ShowWindow             = function () end,
  HideWindow             = function () end,
  SetWindowArea          = function () end,
  SetWindowAreaByButtons = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
local _modules = {
  "int_popup_cloud",
  "int_popup_flat",
  "int_osd_effects"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  common.mod.Load( _modules );

  private.screen = {
    width  = GetAppWidth(),
    height = GetAppHeight(),
    topSafeArea = GetAppTopSafeArea(),
    bottomSafeArea = GetAppBottomSafeArea()
  };

  private.screen.aspect_ratio = private.AspectRatio_Detect();

  DbgTrace(
    "[ UI ] Screen: "..
    private.screen.width.."x"..private.screen.height..", "..
    private.screen.aspect_ratio
  );

  private.button = {
    any = {
      pressed = false
    }
  };
  
  private.swipe              = {};
  private.control_event_data = {};

  private.waiting_name        = "anm_waiting";
  private.popup_cloud_enabled = true;
  private.int_pos_z           = 1000;

  private.hub = {
    main  = "int_control_hub",
    stage = "scr_stage"
  };
  
  private.strings = {
    number_separator = ne.strings.Get( "str_number_separator" )
  };
  
  private.table_anim_triggers = {};
  
  public.obj.Create( private.hub.main, "obj" );
  public.obj.Set( private.hub.main, { pos_z = 700 } );
  public.obj.Attach( private.hub.main,  InterfaceName );

  private.TapInit();
  private.LockInit();
  private.WaitingInit();
  private.SkinInit();

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.hub.Name ()

  return private.hub.main;

end;
--------------------------------------------------------------------------------
function public.hub.Create ( control )

  local pos_z =
  {
    int_dialog_info            = 320, -- system messages
    int_dialog_notifications   = 310, -- enable push notifications window
    int_osd_effects            = 300,
    
    tutorial                   = 290,
    int_subtitles              = 280,
    interface_lock             = 270,
    int_popup_cloud            = 260,
    
    int_panel_interface_top    = 250,

    int_panel_social_score     = 240,
    int_panel_hearts_offers    = 231,
    int_window                 = 230,
    
    int_panel_darkhole         = 210, -- highlight quests
    int_panel_interface_left   = 210,
    int_panel_interface_right  = 210,
    int_panel_interface_bottom = 200,
    
    int_panel_stage_top        = 180,
    
    int_panel_stage_win        = 170, -- You Win!
    
    int_panel_stage_settings   = 165,
    
    int_dialog_stage_friends   = 160,
    int_dialog_stage_shuffle   = 160,
    int_popup_stage_combo      = 150,
    int_popup_stage_bonus      = 140,
    int_panel_booster_sorcery  = 130, -- booster variant (v/h)
    int_popup_stage_score      = 120,
    int_panel_stage_boosters   = 120,
    int_panel_stage_resources  = 120,
    int_panel_stage_score      = 120,
    int_panel_worlds           = 120,

    int_panel_minievent_character = 120,
    int_dialog_minievent_crown_continue = 320,
  };
  
  private.HubCreate( control, private.hub.main, pos_z[ control ] );

end;
--------------------------------------------------------------------------------
function public.hub.CreateInStage ( control )

  local pos_z =
  {
   int_panel_stage_progress_fight = 1,
   int_panel_stage_progress_ride  = 1,
  };
  
  private.HubCreate( control, private.hub.stage, pos_z[ control ] );

end;
--------------------------------------------------------------------------------
function public.hub.AttachContent ( control )

  local hub_obj = string.format( "hub_%s", control );

  public.obj.Attach( control, hub_obj );

end;
--------------------------------------------------------------------------------
function public.hub.DetachContent ( control )

  public.obj.Detach( control );

end;
--------------------------------------------------------------------------------
function private.HubCreate ( control, parent, pos_z )

  local hub_obj = string.format( "hub_%s", control );

  public.obj.Create( hub_obj, "obj" );
  
  if ( pos_z ) then
    ObjSet( hub_obj, { pos_z = pos_z } );
  else
    local msg = "[ UI ][ HUB ][ ERROR ] Unknown control '%s'";
    DbgTrace( msg:format( control ) );
  end;
  
  public.obj.Attach( hub_obj, parent );

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.obj.Get ( object_name )

  return ne.obj.Get( object_name );

end;
--------------------------------------------------------------------------------
function public.obj.Set ( object_name, properties, multiset )

  if ( multiset ) then

    table.insert( multiset, { object_name, properties } );

  else

    ne.obj.Set( object_name, properties );

  end;

end;
--------------------------------------------------------------------------------
function public.obj.Create ( object_name, object_type, attach_to, properties )

  object_type = object_type or "obj";

  ne.obj.Create( object_name, object_type, attach_to, properties );

end;
--------------------------------------------------------------------------------
function public.obj.CreateObject ( base_name, attach_to, properties )
  
  local object_type = "obj";

  public.obj.Create(
    object_type.."_"..base_name, object_type, attach_to, properties
  );

  return object_type.."_"..base_name;

end;
--------------------------------------------------------------------------------
function public.obj.CreateSprite ( base_name, attach_to, properties )

  local object_type = "spr";

  public.obj.Create(
    object_type.."_"..base_name, object_type, attach_to, properties
  );

  return object_type.."_"..base_name;

end;
--------------------------------------------------------------------------------
function public.obj.CreateAnimation ( base_name, attach_to, properties )

  local object_type = "anm";

  public.obj.Create(
    object_type.."_"..base_name, "anim", attach_to, properties
  );

  return object_type.."_"..base_name;

end;
--------------------------------------------------------------------------------
function public.obj.Clone ( object_name, to, postfix, with_children )

  postfix = postfix or "_clone";

  local clone_name = object_name..postfix;

  if ( public.obj.Exists( clone_name ) ) then
    
    DbgTrace( "[ ERROR ] "..clone_name.." already exists." );
    
    return;
    
  end;

  with_children = with_children or 0;

  ne.obj.Clone( object_name, postfix, with_children, to );

  return clone_name;

end;
--------------------------------------------------------------------------------
function public.obj.Delete ( object_name )

  ne.obj.Delete( object_name );

end;
--------------------------------------------------------------------------------
function public.obj.Exists ( object_name )
  
  if ( ne.obj.Get( object_name ) ) then

    return true;

  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
function public.obj.GetNamePrefix( object_name )

  local underscore = string.find( object_name, "_" );

  if ( underscore ) then

    return string.sub( object_name, 1, ( underscore - 1 ) );

  end;

  return nil;

end;
--------------------------------------------------------------------------------
function public.obj.GetNamePostfix( object_name )

  object_name  = string.reverse( object_name );
  local prefix = public.obj.GetNamePrefix( object_name );
  
  if ( not prefix ) then
    
    return nil;
    
  end;
  
  return string.reverse( prefix );

end;
--------------------------------------------------------------------------------
function public.obj.GetNameWithoutPrefix( object_name )

  local underscore = string.find( object_name, "_" );

  if ( underscore ) then

    return string.sub( object_name, ( underscore + 1 ) );

  end;

  return object_name;

end;
--------------------------------------------------------------------------------
function public.obj.GetNameWithoutPostfix( object_name )
  
  object_name = string.reverse( object_name );
  
  local name = public.obj.GetNameWithoutPrefix( object_name );
  
  return string.reverse( name );
  
end;
--------------------------------------------------------------------------------
function public.obj.Enable ( object_name, preserve_input, multiset )

  local properties = {
    active  = true,
    visible = true,
    input   = true
  };

  if ( preserve_input ) then

    properties.input = nil;

  end;

  public.obj.Set( object_name, properties, multiset );

end;
--------------------------------------------------------------------------------
function public.obj.Disable ( object_name, preserve_input, multiset )

  local properties =
  {
    active  = false,
    visible = false,
    input   = false
  };

  if ( preserve_input ) then

    properties.input = nil;

  end;

  public.obj.Set( object_name, properties, multiset );

end;
--------------------------------------------------------------------------------
function public.obj.IsEnabled ( object_name )
  
  return public.obj.Get( object_name ).active;
  
end;
--------------------------------------------------------------------------------
function public.obj.HasInput ( object_name )
  
  return public.obj.Get( object_name ).input;
  
end;
--------------------------------------------------------------------------------
function public.obj.Show ( object_name, multiset )

  local properties =
  {
    visible = true
  };

  public.obj.Set( object_name, properties, multiset );

end;
--------------------------------------------------------------------------------
function public.obj.Hide ( object_name, multiset )

  local properties =
  {
    visible = false
  };

  public.obj.Set( object_name, properties, multiset );

end;
--------------------------------------------------------------------------------
function public.obj.IsActive ( object_name )
  
  return public.obj.Get( object_name ).active;
  
end;
--------------------------------------------------------------------------------
function public.obj.IsVisible ( object_name )
  
  return public.obj.Get( object_name ).visible;
  
end;
--------------------------------------------------------------------------------
function public.obj.Multiset ( multiset )

  if ( multiset                    ) and
     ( type( multiset ) == "table" ) and
     ( #multiset > 0               ) then

    ne.obj.MultiSet( multiset );

  else
    
    DbgTrace( "[ OBJECT ] Multiset error" );
    common.DbgCallpoint();
    
  end;

  multiset = nil;

end;
--------------------------------------------------------------------------------
function public.obj.Attach ( object_name, to )

  ne.obj.Attach( object_name, to );

end;
--------------------------------------------------------------------------------
function public.obj.Detach ( object_name )

  ne.obj.Detach( object_name );

end;
--------------------------------------------------------------------------------
function public.obj.Children ( object_name, search_pattern )

  local children = ObjGetRelations( object_name ).childs;

  if ( not search_pattern ) then return children; end;

  local children_by_pattern = {};

  for c, child in pairs( children ) do

    if ( child:find( search_pattern ) ) then

      table.insert( children_by_pattern, child );

    end;

  end;

  return children_by_pattern;

end;
--------------------------------------------------------------------------------
function public.obj.ToCenter ( object_name, shift_x, shift_y, multiset )
  
  shift_x = shift_x or 0;
  shift_y = shift_y or 0;
  
  public.obj.Set(
    object_name,
    {
      pos_x = 0.5 * private.screen.width + shift_x,
      pos_y = 0.5 * private.screen.height + shift_y
    },
    multiset
  );
  
end;
--------------------------------------------------------------------------------
function public.obj.SpriteAdjust ( spr_name, size, spr_params, multiset )

  spr_params = spr_params or ObjGet( spr_name );

  if not ( spr_params and spr_params.draw_width and spr_params.draw_width > 0 ) then 
    return;
  end;

  local smaller_side = spr_params.draw_width;
  if ( spr_params.draw_height < spr_params.draw_width ) then
    smaller_side = spr_params.draw_height;
  end;

  public.obj.Set(
    spr_name,
    {
      scale_x = size / spr_params.draw_width,
      scale_y = size / spr_params.draw_height
    },
    multiset
  );

end;
--------------------------------------------------------------------------------
function public.obj.Position ( object, relative )

  relative = relative or InterfaceName;

  if ( not object ) then

    DbgTrace( "[ ERROR ] Object is not specified ( nil )." );

    return { 0.0, 0.0 };

  end;

  local object_exists   = public.obj.Exists( object );
  local relative_exists = public.obj.Exists( relative );

  if ( not object_exists ) then

    DbgTrace( "[ ERROR ] "..object.." does not exist." );

    return { 0.0, 0.0 };

  end;

  if ( not relative_exists ) then

    DbgTrace( "[ ERROR ] "..relative.." does not exist." );

    return { 0.0, 0.0 };

  end;

  local position = {};
  
  local func_IsValid = function ( position )
  
    return ( type( position ) == "table"                      ) and
           ( #position > 1                                    ) and
           ( ( position[ 1 ] ~= 0 ) or ( position[ 2 ] ~= 0 ) );

  end;

  -- JK: 1 variant
  position = wiz.obj.GetPosByObj( object, relative );
  
  if ( func_IsValid( position ) ) then

    --DbgTrace( "wiz.obj.GetPosByObjOnMatrix: "..object.." is @ "..position[ 1 ].." x "..position[ 2 ] );    
    return position;

  end;
  
  -- JK: 2 variant
  position = GetObjPosByObj( object, relative );
  
  if ( func_IsValid( position ) ) then
 
    --DbgTrace( "GetObjPosByObj: "..object.." is @ "..position[ 1 ].." x "..position[ 2 ] );
    return position;
 
  end;
  
  -- JK: fallback to relative pos
  if ( relative ) then
    
    local rel_obj = ObjGet( relative );
    
    position[ 1 ] = rel_obj.pos_x;
    position[ 2 ] = rel_obj.pos_y;
    
    --DbgTrace( "ObjGet: "..object.." is @ "..position[ 1 ].." x "..position[ 2 ] );
    return position;
    
  end;
  
  -- JK: fallback to 0
  --DbgTrace( "Fallback: "..object.." is @ 0 x 0" );
  return { 0.0, 0.0 };

end;
--------------------------------------------------------------------------------
function public.obj.Res ( item_type, item_name, item_size )

  local  res = _f.GetRes( item_type, item_name, item_size );
  return res;

end;
--------------------------------------------------------------------------------
function public.obj.SetTextTag ( object_name, tag_name, tag_text, multiset )

  local text = tostring( tag_text );

  if ( string.find( tag_text, "str_" ) ) then

    text = ne.strings.Get( tag_text );

  end;
  
  public.obj.Set(
    object_name,
    {
      [ tag_name ] = text
    },
    multiset
  );
  
end;
--------------------------------------------------------------------------------
function public.obj.GetAnimTag ( object_name )

  local  obj = public.obj.Get( object_name );
  return obj.anim_tag;
  
end;
--------------------------------------------------------------------------------
function public.obj.SetAnimTag ( object_name, tag_name, multiset )

  public.obj.Set(
    object_name,
    {
      anim_tag = tag_name
    },
    multiset
  );
  
end;
--------------------------------------------------------------------------------
function public.obj.GetAnimfuncs ( object_name )

  local  object_props = public.obj.Get( object_name ) or {};
  return object_props.anim_funcs;
  
end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.popup.Cloud_Show ( ... )
  --( mode, full_obj_name, text, data, offset, scale,
  -- functions, disable_sound, permanent )
    
  int_popup_cloud.Show( ... );

end;
--------------------------------------------------------------------------------
function public.popup.Cloud_Hide ( forced )
  
  int_popup_cloud.Hide( forced ); 

end;
--------------------------------------------------------------------------------
function public.popup.Cloud_AddToQueue ( ... )

  int_popup_cloud.AddToQueue( ... );

end;
--------------------------------------------------------------------------------
function public.popup.Cloud_TryShowFromQueue ()

  int_popup_cloud.TryShowFromQueue();

end;
--------------------------------------------------------------------------------
function public.popup.Cloud_IsOnScreen ()

  return int_popup_cloud.IsOnScreen();

end;
--------------------------------------------------------------------------------
function public.popup.Flat_Show ( ... ) --caption, text, pos, parent, inverse

  int_popup_flat.Show( ... ); 

end;
--------------------------------------------------------------------------------
function public.popup.Flat_Hide ( instant, func_end )
  
  int_popup_flat.Hide( instant, func_end ); 

end;
--------------------------------------------------------------------------------
function public.popup.Flat_AttachedTo ()
  
  local attached_to = int_popup_flat.AttachedTo();

  return attached_to;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.osd.ClearEffects ()

  int_osd_effects.Clear();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.button.Create ( placeholder, params, func_Exec )
  
  --[[
    USAGE EXAMPLE:
    
    ui.button.Create(
      "btn_int_dialog_fireflies_booster_buy",
      {
        title  = "str_dialog_fireflies_booster_buy",
        color  = "yellow",
        design = "small",
        price  = true
      },
      function()
        private.Tap_Buy();
      end
    );
  ]]

  params = params or {};

  -- JK: button may clear window outer tap or something
  if ( params.clear ) then

    if ( params.clear == "window_outer_tap" ) then

      local func_ExecSrc = func_Exec;

      func_Exec = function ()

        _f.ClearOuterTap();
        func_ExecSrc();

      end;

    end;

  end;

  --JK: button may exist, so just reset funcExec
  if ( private.button[ placeholder ] ) then
    
    ui.button.EventSet( placeholder, func_Exec );
    
    return;
    
  end;
  
  local multiset = {};
  local input    = {};
  local animfunc = "";  
  
  params.res  = params.res or INTERFACE_PATH.."buttons/anm_buttons";
  params.name = placeholder:gsub( "btn", "anm" );
  params.input = false;
  

  if ( params.color ) then
  
    animfunc = params.color.."_"..params.design;
  
  else
    
    animfunc = params.design;
    
  end;
  
  if ( params.title ) then
   
    if ( type( params.title ) == "string" ) then
    
      params.title = StringGet( params.title );
      
    elseif ( type( params.title ) == "number" ) then
      
      params.title = tostring( params.title );
      
    end;
 
  else
    
    params.title = "Button";
    
  end;
  
  ui.button.EventSet( placeholder, func_Exec );
  
  public.obj.Clone(
    "anm_btn",
    placeholder,
    "_"..placeholder
  );
  
  if ( params.design and params.design == "video" ) then
    
    public.obj.Clone(
      "anm_btn_popup",
      "anm_btn_"..placeholder,
      "_"..placeholder
    );
    
    ne.obj.Set( "anm_btn_popup_"..placeholder, 
    {
      playing = true,
      tt_title = params.title
    } );
    
  end;
  
  if ( params.price ) then
    
    public.obj.Clone(
      "anm_btn_price",
      placeholder,
      "_"..placeholder
    );
    
    if ( type( params.price ) == "boolean" ) then

      params.price = "price_gold_top";

    end;

    public.obj.Set(
      "anm_btn_price_"..placeholder,
      {
        name     = placeholder:gsub( "btn", "anm_price" ),
        animfunc = params.price
      },
      multiset
    );

    params.price = nil;
    
  end;

  if ( params.timer ) then
    
    local pos_y = 
    {
      default = 0,
      up   = -35,
      down =  35
    };

    public.obj.Clone(
      "anm_btn_timer",
      placeholder,
      "_"..placeholder
    );
    
    public.obj.Set(
      "anm_btn_timer_"..placeholder,
      {
        name     = placeholder:gsub( "btn", "anm_timer" ),
        animfunc = tern( params.timer == "" or params.timer == "up" )( "timer", "timer_"..params.timer ),
        pos_y    = pos_y[ params.timer ] or pos_y.default
      },
      multiset
    );

    params.timer = nil;
    
  end;
  
  if ( params.info ) then

    public.obj.Clone(
     "anm_btn_info",
      placeholder,
      "_"..placeholder, true
    );

    public.obj.Set(
      "anm_btn_info_"..placeholder,
      {
        name     = placeholder:gsub( "btn", "anm_info" ),
        animfunc = "save_info"
      },
      multiset
    );
    
    public.obj.Set(
      "anm_btn_info_stage_"..placeholder,
      {
        name     = placeholder:gsub( "btn", "anm_info_stage" ),
        animfunc = "save_info"
      },
      multiset
    );    

    params.info = nil;
    
  end;
  
  if ( params.icon ) then
    
    DbgTrace( "[ UI ] No texture for < "..placeholder.." > to set "..params.icon.."!" );

    public.obj.Clone(
      "spr_btn_icon",
      "anm_btn_"..placeholder,
      "_"..placeholder
    );
    
    public.obj.Set(
      "spr_btn_icon_"..placeholder,
      {
        name     = placeholder:gsub( "btn", "spr_icon" ),
        animfunc = params.icon 
      },
      multiset
    );

    params.icon = nil;
    
  end;  
  
  -- JK: Button Down
  local func_BtnDown = function ()
  
    if ( not private.button[ placeholder ] ) then return; end;
  
    if ( private.button.any.pressed             ) or 
       ( private.button[ placeholder ].pressed  ) or 
       ( private.button[ placeholder ].animated ) then
      
      return;
      
    end;
  
    private.button.any.pressed            = placeholder;
    private.button[ placeholder ].pressed = true;
    
    if ( private.button[ placeholder ].idle ) then
      
      AnimStop( params.name );
      
      public.obj.Set(
        params.name,
        {
          animfunc = animfunc,
          playing  = false, 
          frame    = 0
        }
      );
    
    end;
  
    ObjAnimate( params.name, "frame", 0, 0, "",
    {
      0.00, 0, 0,
      0.25, 0, 15
    } );
  
    mediator.Notify(
      "Interface_ButtonPressed",
      placeholder
    );
  
  end;
  
  -- JK: Button Up
  local func_BtnUp = function ()
  
    if ( not private.button[ placeholder ]          ) or
       (     private.button[ placeholder ].animated ) or
       ( not private.button[ placeholder ].pressed  ) then
      
      return;
      
    end;
    
    private.button[ placeholder ].animated = true;
  
    if  ( private.button[ placeholder ].sound ) then
      
      if ( private.button[ placeholder ].sound ~= "no" ) then
      
         common.sound.Play( private.button[ placeholder ].sound );
      
      end;
      
    else
      
       common.sound.Play( "click_wood" );
      
    end;
  
    ObjAnimate( params.name, "frame", 0, 0, function ()
    
      if ( private.button[ placeholder ].idle ) then
        
        public.obj.Set(
          params.name,
          {
            animfunc = animfunc.."_idle",
            playing  = true,
            frame    = 0
          }
        );
      
      end;
      
      private.button[ placeholder ].animated = nil;
      private.button[ placeholder ].pressed  = nil;
      private.button.any.pressed             = false;
      
      private.button[ placeholder ].func_Exec();

      mediator.Notify(
        "Interface_ButtonAction",
        placeholder
      );
  
    end,
    {
      0.00, 0, 15,
      0.05, 0, 30
    } );
  
  end;
  
  -- JK: Button On Leave
  local func_BtnLeave = function () 
  
    if ( IsPC() ) then
      
      SetCursor( CURSOR_DEFAULT );
      
    end;
  
    if ( not private.button[ placeholder ] ) then return; end;
  
    if (     private.button[ placeholder ].pressed  ) and
       ( not private.button[ placeholder ].animated ) then
        
      private.button[ placeholder ].pressed = nil;
      private.button.any.pressed            = false;

      ObjAnimate( params.name, "frame", 0, 0, function ()

        if ( private.button[ placeholder ].idle ) then
          
          public.obj.Set(
            params.name,
            {
              animfunc = animfunc.."_idle",
              playing  = true,
              frame    = 0
            }
          );
        
        end;

      end,
      {
        0.00, 0, 15,
        0.25, 0, 30
      } );
      
      mediator.Notify(
        "Interface_ButtonReleased",
        placeholder
      );
      
    end;
  
  end;
  
  local func_BtnEnter = function ()
  
    if ( IsPC() ) then
      
      SetCursor( CURSOR_HAND );
      
    end;
      
  end;

  input.event_mdown  = function () func_BtnDown();  end;
  input.event_mup    = function () func_BtnUp();    end;
  input.event_mleave = function () func_BtnLeave(); end;
  input.event_menter = function () func_BtnEnter(); end;

  -- JK: auto-inputrect settings
  local input_rects = {
    default           = { 60,  60  },
    close             = { 70,  70  },
    small             = { 120, 84  },
    medium            = { 220, 84  },
    large             = { 320, 84  },
    facebook          = { 220, 84  },
    facebook_large    = { 320, 84  },
    price             = { 220, 84  },
    price_resource4   = { 220, 84  },
    heart             = { 320, 84  },
    delete            = { 58,  58  },
    heart_send        = { 98,  84  },
    heart_receive     = { 98,  84  },
    wheel             = { 320, 84  },
    buy_resource      = { 150, 200 },
    upgrade_resource1 = { 200, 70  },
    upgrade_resource2 = { 200, 70  },
    upgrade_resource3 = { 200, 70  },
    add               = { 120, 120 },
    buy               = { 220, 120 },
    video             = { 100, 100 },
    green_top         = { 134, 64  },
    blue_me           = { 134, 64  }
  };
 
  local input_rect = 
    params.input_rects           or 
    input_rects[ params.design ] or 
    input_rects.default;

  local input_shift   = params.input_shift or {};
  local input_shift_x = input_shift[ 1 ]   or 0;
  local input_shift_y = input_shift[ 2 ]   or 0;
  
  input.inputrect_init =   true;
  input.inputrect_w    =   input_rect[ 1 ];
  input.inputrect_h    =   input_rect[ 2 ];
  input.inputrect_x    = - input_rect[ 1 ] / 2 + input_shift_x;
  input.inputrect_y    = - input_rect[ 2 ] / 2 + input_shift_y;
  
  --JK: position
  if ( params.pos_x ) then
    
    input.pos_x  = params.pos_x;
    params.pos_x = nil;
    
  end;
  
  if ( params.pos_y ) then
    
    input.pos_y  = params.pos_y;
    params.pos_y = nil;
    
  end;
  
  if ( params.pos_z ) then
    
    input.pos_z  = params.pos_z;
    params.pos_z = nil;
    
  end;
  
  -- JK: button may have idle animation
  params.animfunc = animfunc;
  params.playing  = false;
  private.button[ placeholder ].idle = nil;
  
  if ( params.idle ) then
    
    params.animfunc = animfunc.."_idle";
    params.playing  = true;
    private.button[ placeholder ].idle = true;
    
  end;
  
  if ( params.sound ) then
    
    private.button[ placeholder ].sound = params.sound;
    
  end;
  
  -- JK: convert title to text tag
  params.tt_title = params.title;

  -- JK: remove junk properties
  params.title       = nil;
  params.idle        = nil;
  params.color       = nil;
  params.design      = nil;
  params.sound       = nil;
  params.input_rects = nil;
  params.input_shift = nil;
  
  public.obj.Set(
    "anm_btn_"..placeholder,
    params,
    multiset
  );
  
  public.obj.Set(
    placeholder,
    input,
    multiset
  );
  
  public.obj.Multiset( multiset );
  
end;
--------------------------------------------------------------------------------
function public.button.SetPrice ( placeholder, price, hide_zero, multiset )
  
  if ( type( price ) == "number" ) then
    
    price = ui.text.FormatNumber( price );
    
  end;
  
  local anm_name = placeholder:gsub( "btn", "anm_price" );
  
  if ( price == "0" ) and ( hide_zero ) then
    
    public.obj.Disable( anm_name, true, multiset );
    
  else
  
    public.obj.Set(
      anm_name,
      {
        active   = true,
        visible  = true,
        tt_count = price
      },
      multiset
    );
    
  end;
  
end;
--------------------------------------------------------------------------------
function public.button.SetTime ( placeholder, time, multiset )

  local anm_name   = placeholder:gsub( "btn", "anm_timer" );
  local is_enabled = tern( time )( true, false );
  
  public.obj.Set(
    anm_name,
    {
      active  = is_enabled,
      visible = is_enabled,
      time    = time or "",
      tt_time = time or ""
    },
    multiset
  );

end;
--------------------------------------------------------------------------------
function public.button.SetInfo ( placeholder, info_table, multiset )

  local anm_name   = placeholder:gsub( "btn", "anm_info" );
  local is_enabled = tern( info_table )( true, false );
  
  info_table = info_table or {};
  
  public.obj.Set(
    anm_name,
    {
      active     = is_enabled,
      visible    = is_enabled,
      tt_gold    = info_table.gold or "0",
      tt_section = "str_systemmessage_"..( info_table.section or "section_1" )
    },
    multiset
  );
  
  local anm_stage_name = placeholder:gsub( "btn", "anm_info_stage" );
  
  public.obj.Set(
    anm_stage_name,
    {
      active   = is_enabled,
      visible  = is_enabled,
      animfunc = tern( info_table.stage == "bonus" )( "stage_bonus", "stage" ),
      tt_stage = info_table.stage_last
    },
    multiset
  );

end;

--------------------------------------------------------------------------------
function public.button.SetText ( placeholder, text, multiset )

  local title    = tostring( text );
  local anm_name = placeholder:gsub( "btn", "anm" );

  if ( string.find( text, "str_" ) ) then

    title = ne.strings.Get( text );

  end;

  public.obj.Set(
    anm_name,
    {
      tt_title = title
    },
    multiset
  );
  
end;
--------------------------------------------------------------------------------
function public.button.SetTextTag ( placeholder, tag_name, tag_text, multiset )

  local text     = tostring( tag_text );
  local anm_name = placeholder:gsub( "btn", "anm" );

  if ( string.find( tag_text, "str_" ) ) then

    text = ne.strings.Get( tag_text );

  end;
  
  public.obj.Set(
    anm_name,
    {
      [ tag_name ] = text
    },
    multiset
  );
  
end;
--------------------------------------------------------------------------------
function public.button.Enable ( placeholder, multiset )

  public.obj.Enable( placeholder, false, multiset );
  
end;
--------------------------------------------------------------------------------
function public.button.IsEnabled ( placeholder )

  return public.obj.IsEnabled( placeholder );
  
end;
--------------------------------------------------------------------------------
function public.button.HasInput ( placeholder )

  return public.obj.HasInput( placeholder );
  
end;
--------------------------------------------------------------------------------
function public.button.Disable ( placeholder, multiset )
  
  public.obj.Disable( placeholder, false, multiset );
  
end;
--------------------------------------------------------------------------------
function public.button.SetInput ( placeholder, input )
  
  ObjSet( placeholder, { input = input } );
  
end;
--------------------------------------------------------------------------------
function public.button.Delete ( placeholder )
  
  local anm_name = placeholder:gsub( "btn", "anm" );
  
  ui.control.TapClear( placeholder );
  
  ObjDelete( anm_name );
  
  private.button[ placeholder ] = nil;
  
end;
--------------------------------------------------------------------------------
function public.button.EventSet ( placeholder, func_Exec )

  if not ( private.button[ placeholder ] ) then
    
    private.button[ placeholder ] = {};
    
  end;
  
  private.button[ placeholder ].func_Exec = func_Exec;
  
end;
--------------------------------------------------------------------------------
function public.button.EventGet ( placeholder )
  
  return private.button[ placeholder ].func_Exec;
  
end;
--------------------------------------------------------------------------------
function public.button.Exists ( placeholder )
  
  return ( private.button[ placeholder ] ~= nil );
  
end;
--------------------------------------------------------------------------------
function public.button.IsAnyPressed ()
  
  -- JK: returns pressed button placeholder name or false
  return private.button.any.pressed;
  
end;
--------------------------------------------------------------------------------
function public.button.TryReleasePressed ()
    
  local placeholder = private.button.any.pressed;

  if ( placeholder ) then

    local btn_obj = public.obj.Get( placeholder );

    -- JK: button is accidentally deleted while being pressed
    if ( not btn_obj ) then
      
      private.button.any.pressed = false;
      return true;

    end;

    local func_release = public.obj.Get( placeholder ).event_mleave;

    if ( func_release ) then func_release(); end;

    return true;

  end;

  return false;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.buttons.Enable ( placeholders )
  
  local set = {};
  
  for p, placeholder in pairs( placeholders ) do
    
    if ( multiset ) then

      public.obj.Enable( placeholder, false, multiset );

    else

      public.obj.Enable( placeholder, false, set );

    end;
      
  end;

  if ( not multiset ) then
    
    public.obj.Multiset( set );

  end;
  
end;
--------------------------------------------------------------------------------
function public.buttons.Disable ( placeholders, multiset )
  
  local set = {};
  
  for p, placeholder in pairs( placeholders ) do
    
    if ( multiset ) then

      public.obj.Disable( placeholder, false, multiset );

    else

      public.obj.Disable( placeholder, false, set );

    end;
      
  end;

  if ( not multiset ) then
    
    public.obj.Multiset( set );

  end;
  
end;
--------------------------------------------------------------------------------
function public.buttons.Delete ( module_name, whitelist )
  
  for placeholder, func_Exec in pairs( private.button ) do
    
    if ( string.find( placeholder, module_name ) ) then
    
      if ( not common.IsEntityInList( whitelist, placeholder ) ) then

        ui.button.Delete( placeholder );

      end;
      
    end;
    
  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.control.EventsSet ( control_name, events, multiset )

  -- CAUTION! use multiset wisely

  -- allowed events:
  -- func_on_mouse_enter, func_on_mouse_leave
  -- func_on_click_start, func_on_click

  -- TODO: add hold start/end timer

  private.control_event_data[ control_name ] = {
    events = events
  };

  public.obj.Set(
    control_name, 
    { 
      -- input        = true,
      event_menter = function () private.OnMouseIn ( control_name ); end,
      event_mleave = function () private.OnMouseOut( control_name ); end,
      event_mdown  = function () private.ClickStart( control_name ); end,
      event_mup    = function () private.ClickEnd  ( control_name ); end 
    },
    multiset
  );

end;
--------------------------------------------------------------------------------
function public.control.EventsGet ( control_name )

  if ( private.control_event_data[ control_name ] ) then
  
    return private.control_event_data[ control_name ].events;
  
  end;

  return nil;

end;
--------------------------------------------------------------------------------
function public.control.EventsClear ( control_name )

  local data = private.control_event_data[ control_name ] or {};

  private.control_event_data[ control_name ] = nil;

  public.obj.Set(
    control_name,
    {
      event_menter = function () end,
      event_mleave = function () end,
      event_mdown  = function () end,
      event_mup    = function () end 
    }
  );

  -- JK: mouse out didn't trigger before ui.control.EventsClear
  if ( data.mouse_in ) then

    if ( IsPC() ) then SetCursor( CURSOR_DEFAULT ); end;

  end;

end;
--------------------------------------------------------------------------------
function private.OnMouseIn ( control_name )

  local data = private.control_event_data[ control_name ];

  if ( not data ) then

    -- JK: Control events may be cleared at the moment
    DbgTrace( "[ INTERFACE ] Attempt to call empty MouseIn for "..control_name );

    return;

  end;

  data.mouse_in = true;

  if ( IsPC() ) then

    SetCursor( CURSOR_HAND );

  end;

  if ( data.events.func_on_mouse_enter ) then

    data.events.func_on_mouse_enter();

  end;

end;
--------------------------------------------------------------------------------
function private.OnMouseOut ( control_name )

  local data = private.control_event_data[ control_name ];

  if ( not data ) then

    -- JK: Control events may be cleared at the moment
    DbgTrace( "[ INTERFACE ] Attempt to call empty MouseOut for "..control_name );

    return;

  end;

  data.click_started = false;
  data.mouse_in      = false;

  if ( IsPC() ) then

    SetCursor( CURSOR_DEFAULT );

  end;

  if ( data.events.func_on_mouse_leave ) then

    data.events.func_on_mouse_leave();

  end;

end;
--------------------------------------------------------------------------------
function private.ClickStart ( control_name )

  if ( private.control_clicked ) then return; end;

  private.control_clicked = control_name;

  local data = private.control_event_data[ control_name ];
  
  if ( not data ) then

    DbgTrace(
      "[ ERROR ][ ui ] Attempt to call empty ClickStart for "..control_name
    );
    return;

  end;
  
  local events = data.events;

  if ( events.func_on_click_start ) then events.func_on_click_start(); end;
  if ( events.func_on_click       ) then
  
    data.click_started = true;
    data.click_obj_x, data.click_obj_y = private.ClickPos( control_name );
    data.click_cur_x, data.click_cur_y = CursorPosX() or 0, CursorPosY() or 0;
    
  end;

end;
--------------------------------------------------------------------------------
function private.ClickEnd ( control_name )

  if ( not private.control_clicked ) then return; end;

  local data = private.control_event_data[ control_name ];
  
  if ( not data ) then

    DbgTrace( "[ INTERFACE ] Attempt to call empty ClickEnd for "..control_name );

    return;

  end;
  
  local events = data.events;

  if ( events.func_on_click and data.click_started ) then
  
    local obj_x, obj_y = private.ClickPos( control_name );
    local cur_x, cur_y = CursorPosX() or 0, CursorPosY() or 0;
    
    local obj_moved = obj_x ~= data.click_obj_x or obj_y ~= data.click_obj_y;
    local cur_moved = cur_x ~= data.click_cur_x or cur_y ~= data.click_cur_y;
    
    if not ( obj_moved and cur_moved ) then
    
      data.click_obj_x, data.click_obj_y = nil, nil;
      data.click_cur_x, data.click_cur_y = nil, nil;
      events.func_on_click();
    
    end;

  end;

  private.control_clicked = nil;

end;
--------------------------------------------------------------------------------
function private.ClickPos ( control_name )

  local pos = ui.obj.Position( control_name );

  return pos[ 1 ], pos[ 2 ];

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.control.TapClear ( object )

  if ( private.tap.list[ object ] ) then

    private.tap.list[ object ].assigned = false;

  end;

end;
--------------------------------------------------------------------------------
function public.control.TapBreak ( object )

  if ( not private.TapIsCurrent( ( object or private.TapGetCurrent() ) ) ) then

    return;

  end;

  private.TapSetCurrent();

  ObjSet( private.tap.TIMER_HOLD_OBJECT, { playing = false } );

end;
--------------------------------------------------------------------------------
function public.control.TapSet ( object, params )

  private.tap.list[ object ] = private.tap.list[ object ] or {};

  private.tap.list[ object ].anim = params.anim or private.tap.list[ object ].anim;

  private.tap.list[ object ].func_true      = params.func_true      or private.tap.list[ object ].func_true;
  private.tap.list[ object ].func_false     = params.func_false     or private.tap.list[ object ].func_false;
  private.tap.list[ object ].func_menter    = params.func_menter    or private.tap.list[ object ].func_menter;
  private.tap.list[ object ].func_mleave    = params.func_mleave    or private.tap.list[ object ].func_mleave;
  private.tap.list[ object ].func_hold_down = params.func_hold_down or private.tap.list[ object ].func_hold_down;
  private.tap.list[ object ].func_hold_up   = params.func_hold_up   or private.tap.list[ object ].func_hold_up;

  -- DK> Введено для случаев, когда нужно дополнительно обработать mouse up "мимо" целевого объекта ( т.к. ObjWaitForMouseUp перетирает событие ).
  if ( params.func_false ) then

    ObjWaitForMouseUp( object, function ()

      private.TapCallFuncFalse( object );

    end );

  end;

  if not ( private.tap.list[ object ].assigned ) then

    ObjSet( object,
    {
      event_mdown  = function () private.TapMouseDown( object ); end,
      event_mup    = function () private.TapMouseUp( object, true ); end,
      event_menter = function () private.TapMouseEnter( object ); end,
      event_mleave = function () private.TapMouseLeave( object ); end
    } );

    private.tap.list[ object ].assigned = true;

  end;

end;
--------------------------------------------------------------------------------
function public.control.TapGet ( object )

  return private.tap.list[ object ];

end;
--------------------------------------------------------------------------------
function private.TapInit ()

  private.tap = {};

  private.tap.list = {};

  private.tap.object_current = nil;

  private.tap.TIMER_LOCK_OBJECT = "tmr_common_tap_lock";
  private.tap.TIMER_LOCK_TIME   = 0.5;

  ObjCreate( private.tap.TIMER_LOCK_OBJECT, "timer" );
  ObjAttach( private.tap.TIMER_LOCK_OBJECT, InterfaceName );

  private.tap.TIMER_HOLD_OBJECT = "tmr_common_tap_hold";
  private.tap.TIMER_HOLD_TIME   = 1.0;

  ObjCreate( private.tap.TIMER_HOLD_OBJECT, "timer" );
  ObjAttach( private.tap.TIMER_HOLD_OBJECT, InterfaceName );

  private.tap.anim = {};

  private.TapInitAnims();

end;
--------------------------------------------------------------------------------
function private.TapSetCurrent ( object )

  private.tap.object_current = object;

end;
--------------------------------------------------------------------------------
function private.TapGetCurrent ()

  return private.tap.object_current;

end;
--------------------------------------------------------------------------------
function private.TapIsCurrent ( object )

  return ( private.tap.object_current == object );

end;
--------------------------------------------------------------------------------
function private.TapCallFuncFalse ( object )

  if ( private.tap.list[ object ].func_false ) then

    private.tap.list[ object ].func_false( object );

    private.tap.list[ object ].func_false = nil;

  end;

end;
--------------------------------------------------------------------------------
function private.TapHoldDown ( object )

  if ( not private.TapIsCurrent( object ) ) then

    return;

  end;

  private.tap.was_holded = true;

  local func_hold_down = private.tap.list[ object ].func_hold_down;

  if ( func_hold_down ) then

    func_hold_down( object );

  end;

end;
--------------------------------------------------------------------------------
function private.TapHoldUp ( object )

  if ( not private.tap.was_holded ) then

    return;

  end;

  private.tap.was_holded = false;

  local func_hold_up = private.tap.list[ object ].func_hold_up;

  if ( func_hold_up ) then

    func_hold_up( object );

  end;

end;
--------------------------------------------------------------------------------
function private.TapMouseDown ( object )

  if ( private.TapGetCurrent() ) then

    return;

  end;

  if ( ObjGet( private.tap.TIMER_LOCK_OBJECT ).playing ) then

    return;

  end;

  private.TapSetCurrent( object );

  local func = function ()

    private.TapMouseUp( object );

    private.TapCallFuncFalse( object );

  end;

  ObjWaitForMouseUp( object, func );

  private.TapMouseDownAnim( object );

  ObjSet( private.tap.TIMER_HOLD_OBJECT,
  {
    time = private.tap.TIMER_HOLD_TIME,
    endtrig = function () private.TapHoldDown( object ); end,
    playing = true
  } );

end;
--------------------------------------------------------------------------------
function private.TapMouseUp ( object, is_true )

  if ( not private.TapIsCurrent( object ) ) then

    return;

  end;

  private.TapSetCurrent();

  private.TapHoldUp( object );

  if ( is_true ) then

    ObjSet( private.tap.TIMER_LOCK_OBJECT, { time = private.tap.TIMER_LOCK_TIME, endtrig = "", playing = true } );

    local func_true = public.control.TapGet( object ).func_true;

    if ( func_true ) then

      if ( "string" == type( func_true ) ) then

        DbgTrace( string.format( "Warning! Using code string in TAP for [ %s ] ( code: '%s' )!", object, func_true ) );

      elseif ( "function" == type( func_true ) ) then

        func_true( object );

      end;

    else

      DbgTrace( string.format( "Warning! Trigger for < %s > doesn't exist!", object ) );

    end;

  end;

  private.TapMouseUpAnim( object );

end;
--------------------------------------------------------------------------------
function private.TapMouseEnter ( object )

  if ( GetPlatform() == "ios" ) then

    return;

  end;
  
  SetCursor( CURSOR_HAND );

  private.TapMouseEnterAnim( object );

  if ( private.tap.list[ object ].func_menter ) then

    private.tap.list[ object ].func_menter( object );

  end;
  
end;
--------------------------------------------------------------------------------
function private.TapMouseLeave ( object )

  -- Снятие фокуса с объекта может быть после нажатия, и до отпускания.
  private.TapMouseUp( object );

  if ( GetPlatform() == "ios" ) then

    return;

  end;
  
  SetCursor( CURSOR_DEFAULT );

  private.TapMouseLeaveAnim( object );

  if ( private.tap.list[ object ].func_mleave ) then

    private.tap.list[ object ].func_mleave( object );

  end;

end;
--------------------------------------------------------------------------------
function private.TapRegisterAnim ( anim_type, func_down, func_up, func_enter, func_leave )

  local func_empty = function () end;

  private.tap.anim[ anim_type ] =
  {
    MouseDown  = func_down  or func_empty,
    MouseUp    = func_up    or func_empty,
    MouseEnter = func_enter or func_empty,
    MouseLeave = func_leave or func_empty
  };

end;
--------------------------------------------------------------------------------
function private.TapMouseDownAnim ( object )

  local anim = public.control.TapGet( object ).anim;

  if not ( anim ) then

    return;

  end;

  private.tap.anim[ anim.type ].MouseDown( object, anim.value );

end;
--------------------------------------------------------------------------------
function private.TapMouseUpAnim ( object )

  local anim = public.control.TapGet( object ).anim;

  if not ( anim ) then

    return;

  end;

  private.tap.anim[ anim.type ].MouseUp( object, anim.value );

end;
--------------------------------------------------------------------------------
function private.TapMouseEnterAnim ( object )

  local anim = public.control.TapGet( object ).anim;

  if not ( anim ) then

    return;

  end;

  private.tap.anim[ anim.type ].MouseEnter( object, anim.value );

end;
--------------------------------------------------------------------------------
function private.TapMouseLeaveAnim ( object )

  local anim = public.control.TapGet( object ).anim;

  if not ( anim ) then

    return;

  end;

  private.tap.anim[ anim.type ].MouseLeave( object, anim.value );

end;
--------------------------------------------------------------------------------
function private.TapInitAnims ()

  private.TapRegisterAnim( "focus_alpha",
    private.AnimFocusAlpShow,
    private.AnimFocusAlpHide,
    private.AnimFocusAlpShow,
    private.AnimFocusAlpHide
  );

  private.TapRegisterAnim( "object_scale",
    private.AnimObjectScaleDown,
    private.AnimObjectScaleUp,
    --[[TEMP]]private.AnimObjectScaleDown,--private.AnimObjectScaleEnter,
    --[[TEMP]]private.AnimObjectScaleUp--private.AnimObjectScaleLeave
  );

  private.TapRegisterAnim( "icon_scale",
    private.AnimIconScaleDown,
    private.AnimIconScaleUp,
    private.AnimIconScaleDown,
    private.AnimIconScaleLeave
  );
  
  private.TapRegisterAnim( "button",
    private.AnimButtonDown,
    private.AnimButtonLeave,
    private.AnimButtonDown,
    private.AnimButtonLeave
  );

  private.TapRegisterAnim( "button_color",
    private.AnimButtonColorDown,
    private.AnimButtonColorLeave,
    private.AnimButtonColorDown,
    private.AnimButtonColorLeave
  );

end;
--------------------------------------------------------------------------------
function private.AnimFocusAlpShow ( sender, anim_value, trg )

  local sender_name = ui.obj.GetNameWithoutPrefix( sender );

  local obj = "fcs_"..sender_name;

  if ( ObjGet( obj ) ) then

    local alp = ObjGet( obj ).alp;
    local tme = 0.3 * ( 1 - alp );

    ObjAnimate( obj, "alp", 0, 0, trg,
    {
      0.0, 2, alp,
      tme, 2, 1.0
    } );

  end;

end;
--------------------------------------------------------------------------------
function private.AnimFocusAlpHide ( sender, anim_value, trg )

  local sender_name = ui.obj.GetNameWithoutPrefix( sender );

  local obj = "fcs_"..sender_name;

  if ( ObjGet( obj ) ) then

    local alp = ObjGet( obj ).alp;
    local tme = 0.3 * alp;

    ObjAnimate( obj, "alp", 0, 0, trg,
    {
      0.0, 1, alp,
      tme, 1, 0.0
    } );

  end;

end;
--------------------------------------------------------------------------------
function private.AnimObjectScaleDown ( sender, anim_value, trg )

  local sender_name = ui.obj.GetNameWithoutPrefix( sender );

  local obj = "btn_"..sender_name;
  local param_obj = ObjGet( obj );

  ObjAnimate( obj, "scale_xy", 0, 0, trg,
  {
    0.0, 2, param_obj.scale_x, param_obj.scale_y,
    0.3, 2, anim_value * 0.8, anim_value * 0.8
  } );

end;
--------------------------------------------------------------------------------
function private.AnimObjectScaleUp ( sender, anim_value, trg )

  local sender_name = ui.obj.GetNameWithoutPrefix( sender );

  local obj = "btn_"..sender_name;

  ObjAnimate( obj, "scale_xy", 0, 0, trg,
  {
    0.00, 3, anim_value * 0.80, anim_value * 0.80,
    0.30, 3, anim_value * 1.00, anim_value * 1.00
    --0.30, 3, anim_value * 1.20, anim_value * 1.20,
    --0.60, 3, anim_value * 0.90, anim_value * 0.90,
    --0.90, 3, anim_value * 1.00, anim_value * 1.00
  } );

end;
--------------------------------------------------------------------------------
function private.AnimObjectScaleEnter ( sender, anim_value, trg )

  local sender_name = ui.obj.GetNameWithoutPrefix( sender );

  local obj = "btn_"..sender_name;

  local param_obj = ObjGet( obj );

  ObjAnimate( obj, "scale_xy", 0, 0, trg,
  {
    0.0, 2, param_obj.scale_x, param_obj.scale_y,
    0.3, 2, anim_value * 1.15, anim_value * 1.15
  } );

end;
--------------------------------------------------------------------------------
function private.AnimObjectScaleLeave ( sender, anim_value, trg )

  local sender_name = ui.obj.GetNameWithoutPrefix( sender );

  local obj = "btn_"..sender_name;

  local param_obj = ObjGet( obj );

  ObjAnimate( obj, "scale_xy", 0, 0, trg,
  {
    0.0, 2, param_obj.scale_x, param_obj.scale_y,
    0.3, 2, anim_value * 1.0, anim_value * 1.0
  } );

end;
--------------------------------------------------------------------------------
function private.AnimIconScaleDown ( sender, anim_value, trg )

  local sender_name = ui.obj.GetNameWithoutPrefix( sender );

  local obj = "spr_"..sender_name.."_icon";
  local param_obj = ObjGet( obj );

  ObjAnimate( obj, "scale_xy", 0, 0, trg,
  {
    0.0, 2, param_obj.scale_x, param_obj.scale_y,
    0.3, 2, anim_value * 0.8, anim_value * 0.8
  } );

end;
--------------------------------------------------------------------------------
function private.AnimIconScaleUp ( sender, anim_value, trg )

  local sender_name = ui.obj.GetNameWithoutPrefix( sender );

  local obj = "spr_"..sender_name.."_icon";

  ObjAnimate( obj, "scale_xy", 0, 0, trg,
  {
    0.00, 3, anim_value * 0.80, anim_value * 0.80,
    0.30, 3, anim_value * 1.20, anim_value * 1.20,
    0.60, 3, anim_value * 0.90, anim_value * 0.90,
    0.90, 3, anim_value * 1.00, anim_value * 1.00
  } );

end;
--------------------------------------------------------------------------------
function private.AnimIconScaleEnter ( sender, anim_value, trg )

  local sender_name = ui.obj.GetNameWithoutPrefix( sender );

  local obj = "btn_"..sender_name.."_icon";

  local param_obj = ObjGet( obj );

  ObjAnimate( obj, "scale_xy", 0, 0, trg,
  {
    0.0, 2, param_obj.scale_x, param_obj.scale_y,
    0.3, 2, anim_value * 1.15, anim_value * 1.15
  } );

end;
--------------------------------------------------------------------------------
function private.AnimIconScaleLeave ( sender, anim_value, trg )

  local sender_name = ui.obj.GetNameWithoutPrefix( sender );

  local obj = "spr_"..sender_name.."_icon";

  local param_obj = ObjGet( obj );

  ObjAnimate( obj, "scale_xy", 0, 0, trg,
  {
    0.0, 2, param_obj.scale_x, param_obj.scale_y,
    0.3, 2, anim_value * 1.0, anim_value * 1.0
  } );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.window.Show ( window_name, func_end, params )

  _f.ShowWindow( window_name, func_end, params );

end;
--------------------------------------------------------------------------------
function public.window.Hide ( window_name, func_end, skip_map_enable )

  _f.HideWindow( window_name, func_end, skip_map_enable );

end;
--------------------------------------------------------------------------------
function public.window.Open ( dialog_name, func_end, params, sound_name )

  public.screen.Lock( true );

  sound_name = sound_name or "window_open";
  
  common.sound.Play( sound_name );
  
  local int_name         = "int_dialog_"..dialog_name;
  local is_fullscreen    = _f.IsFullscreen( int_name );
  local animfunc_postfix = tern( is_fullscreen )( "fullscreen", "medium" );

  public.input.On( int_name );
  
  public.anim.Play(
    "anm_"..int_name,
    "open_"..animfunc_postfix,
    function ()

      public.anim.Play( "anm_"..int_name, "static_"..animfunc_postfix );

      if ( not is_fullscreen and not ( params or {} ).outer_disable ) then

        _f.SetOuterTap( int_name );

      end;

      public.screen.Lock( false );
      
      if ( func_end ) then

        func_end();

      end;

    end
  );

end;
--------------------------------------------------------------------------------
function public.window.Close ( dialog_name, func_end, sound_name )

  public.screen.Lock( true );

  sound_name = sound_name or "window_close";
  
  common.sound.Play( sound_name );

  local int_name         = "int_dialog_"..dialog_name;
  local is_fullscreen    = _f.IsFullscreen( int_name );
  local animfunc_postfix = tern( is_fullscreen )( "fullscreen", "medium" );

  public.input.Off( int_name );
  
  if ( not is_fullscreen ) then

    _f.ClearOuterTap();

  end;

  public.anim.Play(
    "anm_"..int_name,
    "close_"..animfunc_postfix,
    function ()

      public.screen.Lock( false );

      if ( func_end ) then

        func_end();

      end;

    end
  );

end;
--------------------------------------------------------------------------------
function public.window.SetArea ( window_name, area, multiset )

  _f.SetWindowArea( window_name, area, multiset );

end;
--------------------------------------------------------------------------------
function public.window.SetAreaByButtons ( obj_window, btn_action, btn_close )

  _f.SetWindowAreaByButtons( obj_window, btn_action, btn_close );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.scroller.Next ( scroller, max, loop )
  
  local scrollpos = ne.obj.Get( scroller ).scrollpos;
  local next_pos  = math.min( scrollpos + 1, max - 1 );
  
  if ( next_pos == scrollpos ) then
    
    if ( not loop ) then return next_pos; end;
    
    next_pos = 0;
    
  end;
  
  ne.obj.Set( scroller, { scrollpos = next_pos } );
  
  return next_pos;
  
end;
--------------------------------------------------------------------------------
function public.scroller.Prev ( scroller, max, loop )
  
  local scrollpos = ne.obj.Get( scroller ).scrollpos;
  local prev_pos  = math.max( scrollpos - 1, 0 );
  
  if ( prev_pos == scrollpos ) then
    
    if ( not loop ) then return prev_pos; end;
    
    prev_pos = max - 1;
    
  end;
  
  ne.obj.Set( scroller, { scrollpos = prev_pos } );
  
  return prev_pos;
  
end;
--------------------------------------------------------------------------------
function public.scroller.ResetPos ( scroller )
  
  ne.obj.Set( scroller, { scrollpos = 0 } );
  
end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function public.timer.Show (
    timer_name,
    seconds,
    func_exec,
    custom_obj
  )

  if ( not seconds  ) or
     ( seconds <= 0 ) then
    public.timer.Hide( timer_name, true, custom_obj );
    return;
  end;

  local multiset        = {};
  local tmr_timer       = "tmr_"..timer_name;
  local anm_timer       = "anm_"..timer_name.."_timer";
  local time_text       = public.text.FormatTime( seconds );
  local text_tag        = "timer";
  local animfunc_show   = "timer_show";
  local animfunc_static = "timer_static";

  local resolution = 10;

  -- JK: 1 h + 2 mins & 1 d + 2 mins
  if ( seconds <= 3720  ) then resolution = 1;  end;
  if ( seconds >= 86520 ) then resolution = 60; end;

  if ( custom_obj ) then
    anm_timer = custom_obj.anm_timer;
    text_tag  = custom_obj.text_tag;
  end;

  public.obj.SetTextTag( anm_timer, text_tag, time_text, multiset );
  common.Timer_Start( tmr_timer, resolution, func_exec, multiset );

  public.obj.Multiset( multiset );

  -- JK: if custom object supports animfuncs, use them
  if ( custom_obj ) then
    if ( custom_obj.animfunc ) then
      animfunc_show   = custom_obj.animfunc.show;
      animfunc_static = custom_obj.animfunc.static;
    else
      --JK: otherwise, just set
      return;
    end;
  end;

  if ( public.obj.Get( anm_timer ).animfunc == animfunc_static ) then
    return;
  end;

  public.anim.Play( anm_timer, animfunc_show, nil, animfunc_static );

end;
--------------------------------------------------------------------------------
function public.timer.Hide ( timer_name, animated, custom_obj )

  local multiset      = {};  
  local tmr_timer     = "tmr_"..timer_name;
  local anm_timer     = "anm_"..timer_name.."_timer";
  local animfunc_hide = "timer_hide";

  common.Timer_Stop( tmr_timer, multiset );

  if ( custom_obj ) then

    anm_timer = custom_obj.anm_timer;

    if ( custom_obj.animfunc ) then
      animfunc_hide = custom_obj.animfunc.hide;
    else
      public.obj.Multiset( multiset );
      return;
    end;

  end;

  if ( public.obj.Get( anm_timer ).animfunc == animfunc_hide ) then
    public.obj.Multiset( multiset );
    return;
  end;
  
  if ( animated ) then
    public.anim.Play( anm_timer, animfunc_hide );
  else
    public.obj.Set(
      anm_timer,
      {
        playing  = false,
        animfunc = animfunc_hide,
        frame    = 24
      },
      multiset
    );
  end;
  
  public.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.swipe.Create ( placeholder, params )

  private.swipe[ placeholder ] = params;

  ObjSet( placeholder,
    {
      drag            = true,
      realdrag        = true,
      event_startdrag = function () private.Swipe_Start( placeholder ); end,
      event_dragdrop  = function () private.Swipe_Stop(  placeholder ); end
    }
  );

end;
--------------------------------------------------------------------------------
function private.Swipe_Start ( placeholder )

  private.swipe[ placeholder ].start = GetGameCursorPos();

end;
--------------------------------------------------------------------------------
function private.Swipe_Stop ( placeholder )

  local swipe_start = private.swipe[ placeholder ].start;
  
  if ( not swipe_start ) then
   
    DbgTrace( "[ ERROR ] Broken swipe occured." );
    
    return;
   
  end;
  
  private.swipe[ placeholder ].start = nil;
  
  local swipe_stop  = GetGameCursorPos();
  local sensitivity = private.screen.width / 6;
  
  local delta_x = swipe_start[ 1 ] - swipe_stop[ 1 ];
  local delta_y = swipe_start[ 2 ] - swipe_stop[ 2 ];

  if ( math.abs( delta_x ) >= math.abs( delta_y ) ) then

    if ( math.abs( delta_x ) < sensitivity ) then

      return;

    end;

    if ( delta_x < 0 ) then

      -- JK: swipe RIGHT

      if ( private.swipe[ placeholder ].right ) then

        private.swipe[ placeholder ].right();

      end;

    else

      -- JK: swipe LEFT

      if ( private.swipe[ placeholder ].left ) then

        private.swipe[ placeholder ].left();

      end;

    end;

  else

    if ( math.abs( delta_y ) < sensitivity ) then

      return;

    end;

    if ( delta_y < 0 ) then

      -- JK: swipe DOWN

      if ( private.swipe[ placeholder ].down ) then

        private.swipe[ placeholder ].down();

      end;

    else

      -- JK: swipe UP

      if ( private.swipe[ placeholder ].up ) then

        private.swipe[ placeholder ].up();

      end;

    end;    

  end;

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.screen.Lock( state )

  -- common.DbgCallpoint( tostring( state ) );
  
  ObjMultiSet( {
    { "ng_interface_lock",     { input   = state } },
    { "ng_interface_lock_spr", { visible = state } }
  } );
  
  private.screen_locked = state;

end;
--------------------------------------------------------------------------------
function public.screen.IsLocked ()

  return private.screen_locked;

end;
--------------------------------------------------------------------------------
function public.screen.Waiting_Show ()

  --ShowLoadingWheel( true );
  public.screen.Lock( true );

  ObjAttach( private.waiting_name, InterfaceName );

  ObjStopAnimate( private.waiting_name, "alp" );
  ObjAnimate(  private.waiting_name, "alp", 0, 0, "",
  {
    0.0, 0, 0.0,
    0.5, 0, 0.0,
    0.8, 0, 1.0
  } );

end;
--------------------------------------------------------------------------------
function public.screen.Waiting_Hide ()
  
  --ShowLoadingWheel( false );
  public.screen.Lock( false );

  local alp = ObjGet( private.waiting_name ).alp;

  ObjStopAnimate( private.waiting_name, "alp" );
  ObjAnimate(  private.waiting_name, "alp", 0, 0,
  function ()

    ObjDetach( private.waiting_name );

  end,
  {
    0.0, 0, alp,
    0.3, 0, 0.0
  } );

end;
--------------------------------------------------------------------------------
function public.screen.AspectRatio ()
  
  return private.screen.aspect_ratio;
  
end;
--------------------------------------------------------------------------------
function public.screen.IsLandscape ()
  
  return ( private.screen.width > private.screen.height );
  
end;
--------------------------------------------------------------------------------
function public.screen.Orientation ()
  
  if ( public.screen.IsLandscape() ) then
    
    return "landscape";
    
  end;
  
  return "portrait";
  
end;
--------------------------------------------------------------------------------
function public.screen.Width ()
  
  return private.screen.width;
  
end;
--------------------------------------------------------------------------------
function public.screen.Height ()
  
  return private.screen.height;
  
end;
--------------------------------------------------------------------------------
function public.screen.TopSafeArea ()

  return private.screen.topSafeArea;

end;
--------------------------------------------------------------------------------
function public.screen.BottomSafeArea ()

  return private.screen.bottomSafeArea;

end;
--------------------------------------------------------------------------------
function public.screen.MaxDimension ()

  return math.max( private.screen.width, private.screen.height );

end;
--------------------------------------------------------------------------------
function public.screen.MinDimension ()

  return math.min( private.screen.width, private.screen.height );

end;
--------------------------------------------------------------------------------
function public.screen.WidthPercent ()
  
  return private.screen.width / 100;
  
end;
--------------------------------------------------------------------------------
function public.screen.HeightPercent ()
  
  return private.screen.height / 100;
  
end;
--------------------------------------------------------------------------------
function public.screen.Center ()
  
  return {
    pos_x = 0.5 * private.screen.width,
    pos_y = 0.5 * private.screen.height
  };
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.input.On( object_name, multiset )

  private.Input_Set( object_name, true, multiset );

end;
--------------------------------------------------------------------------------
function public.input.Off( object_name, multiset )

  private.Input_Set( object_name, false, multiset );

end;
--------------------------------------------------------------------------------
function public.input.Set( object_name, flag, multiset )

  private.Input_Set( object_name, flag, multiset );

end;
--------------------------------------------------------------------------------
function private.Input_Set ( object_name, flag, multiset )

  local properties =
  {
    input = flag
  };

  if ( type( object_name ) == "string" ) then

    public.obj.Set( object_name, properties, multiset );

  else

    local autoset = false;

    if ( not multiset ) then

      multiset = {};
      autoset  = true;

    end;

    for _, list_item in pairs( object_name ) do

      public.obj.Set( list_item, properties, multiset );

    end;

    if ( autoset ) then

      public.obj.Multiset( multiset );

    end;

  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.text.FontsizeByHeight ( object, text_height, default_fontsize )

  local text_parms = ObjGet( object );
  local fontsize   = default_fontsize;

  if ( text_parms.draw_height > text_height ) then

    fontsize =
      math.ceil( text_parms.fontsize * text_height / text_parms.draw_height );

    ObjSet( object, { fontsize = fontsize } );

  end;

end;
--------------------------------------------------------------------------------
function public.text.CenterIconized ( spr_name, txt_name, spr_width, margin )

  local sprite_w = spr_width or 48;
  local margin_w = margin    or 0;
  local text_w   = ObjGet( txt_name ).draw_width;

  local half_width = ( sprite_w + text_w + margin_w ) / 2;

  local sprite_pos = - half_width + sprite_w / 2;
  local text_pos   = - half_width + sprite_w + margin_w + text_w / 2;

  ObjMultiSet( {
    { spr_name, { pos_x = sprite_pos } },
    { txt_name, { pos_x = text_pos   } }
  } );

end;
--------------------------------------------------------------------------------
function public.text.FormatTime ( seconds )

  local time = common.TimeFromSeconds( seconds );

  if ( time.d > 0 ) then

    if ( time.h > 0 ) then
      
      return ne.strings.Get(
        "str_time_dh",
        string.format( "%.1d", time.d ),
        string.format( "%.1d", time.h )
      );
      
    end;
    
    if ( time.m > 0 ) then
      
      return ne.strings.Get(
        "str_time_dm",
        string.format( "%.1d", time.d ),
        string.format( "%.2d", time.m )
      );
      
    end;
    
    return ne.strings.Get(
      "str_time_d",
      string.format( "%.1d", time.d )
    );

  end;

  if ( time.h > 0 ) then

    return ne.strings.Get(
      "str_time_hm",
      string.format( "%.1d", time.h ),
      string.format( "%.2d", time.m )
    );

  end;
  
  if ( time.m > 0 ) then

    return ne.strings.Get(
      "str_time_ms",
      string.format( "%.2d", time.m ),
      string.format( "%.2d", time.s )
    );

  end;

  return ne.strings.Get(
    "str_time_s",
    string.format( "%.2d", time.s )
  );

end;
--------------------------------------------------------------------------------
function public.text.FormatNumberBooster ( number )
  
  if ( number < 100 ) then
    
    return tostring( number );
    
  end;
  
  return "99+";
  
end;
--------------------------------------------------------------------------------
function public.text.FormatNumber ( number, unconverted )

  local postfix = "";
  
  if ( not unconverted ) then
    
    if ( number > 999999 ) then

      number = math.floor( number / 1000000 );

      postfix = "M";

    elseif ( number > 99999 ) then

      number = math.floor( number / 1000 );

      postfix = "K";

    end;
    
  end;

  local text_number  = tostring( math.floor( number ) );
  local first_symbol = string.sub( text_number, 1, 1 );
  local sign         = "";

  if ( first_symbol == "-" ) then

    sign        = "-";
    text_number = string.sub( text_number, 2 );

  end;

  local count_add_symbol  = math.floor( #text_number / 3 );
  local count_digit_begin = math.fmod( #text_number, 3 );

  local symbol_separator = private.strings.number_separator;
  local new_text         = "";

  if ( count_digit_begin > 0 ) then

    new_text    = string.sub( text_number, 1, count_digit_begin );
    text_number = string.sub( text_number, count_digit_begin + 1 );

  end;

  for i = 1, count_add_symbol do
    
    local part_text = string.sub( text_number, 1, 3 );
    text_number     = string.sub( text_number, 4 );
    new_text        = new_text..symbol_separator..part_text;

  end;

  local first = string.sub( new_text, 1, 1 );

  if ( first == symbol_separator ) then

    new_text = string.sub( new_text, 2 );

  end;

  return sign..new_text..postfix;

end;
--------------------------------------------------------------------------------
function public.text.FormatNumberToDigits ( value )
  
  local result = common.math.NumberToDigits(value);
  
  if ( value > 999 ) then
    
    table.insert( result, #result - 2, private.strings.number_separator );
    
  end;
  
  return result;
  
end;
--------------------------------------------------------------------------------
function public.text.InfinitySign ()

  return "~";

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.anim.Receive ( receiver, object_name, func_end, ext_pos )

  local position = ext_pos or ui.obj.Position( receiver, object_name );
  local obj      = ne.obj.Get( object_name );

  if ( not obj ) then

    if ( func_end ) then func_end(); end; return;

  end;

  local func_OnAnimationEnd = function ()

    ne.obj.Set(
      object_name,
      { 
        pos_x = obj.pos_x,
        pos_y = obj.pos_y
      } 
    );

    if ( func_end ) then func_end(); end;

  end;

  ne.obj.Animate(
    object_name, "pos_xy", 0, 0, func_OnAnimationEnd,
    { 
      0.0, 3, obj.pos_x,     obj.pos_y,
      1.0, 3, position[ 1 ], position[ 2 ]
    }
  );

end;
--------------------------------------------------------------------------------
function public.anim.Receive_Gold ( object, func_end )

  ui.anim.Receive ( "obj_int_panel_interface_top_gold", object, func_end );

end;
--------------------------------------------------------------------------------
function public.anim.Bounce ( object_name, func_end )

  local obj = ne.obj.Get( object_name );

  if ( not obj ) then
    if ( func_end ) then func_end(); end; return;
  end;

  local scale_x = obj.scale_x or 1;
  local scale_y = obj.scale_y or 1;

  ne.obj.Animate(
    object_name, "scale_xy", 0, 0, func_end,
    {
      0.00, 0, scale_x * 1.0, scale_y * 1.0,
      0.10, 0, scale_x * 1.5, scale_y * 1.0,
      0.20, 0, scale_x * 1.0, scale_y * 1.5,
      0.30, 0, scale_x * 0.8, scale_y * 0.8,
      0.35, 0, scale_x * 1.2, scale_y * 1.2,
      0.40, 0, scale_x * 1.0, scale_y * 1.0
    }
  );

end;
--------------------------------------------------------------------------------
function public.anim.Shake ( object_name, func_end )

  local obj = ne.obj.Get( object_name );

  if ( not obj ) then

    if ( func_end ) then func_end(); end; return;

  end;

  local pos_x = obj.pos_x or 0;

  ne.obj.Animate(
    object_name, "pos_x", 0, 0, func_end,
    {
      0.00, 0, pos_x,
      0.05, 0, pos_x - 30,
      0.15, 0, pos_x + 30,
      0.22, 0, pos_x - 15,
      0.27, 0, pos_x + 10,
      0.30, 0, pos_x
    }
  );

end;
--------------------------------------------------------------------------------
function public.anim.FadeIn ( time, object_name, reset, func_end )

  local obj = ne.obj.Get( object_name );

  if ( not obj ) then

    if ( func_end ) then func_end(); end; return;

  end;

  local alp = obj.alp or 1;

  if ( alp > 0 ) and ( reset ) then

    ne.obj.Set( object_name, { alp = 0.0 } );

  end;

  ne.obj.Animate(
    object_name, "alp", 0, 0, func_end,
    {
      0.0,  0, alp,
      time, 0, 1.0
    }
  );

end;
--------------------------------------------------------------------------------
function public.anim.FadeOut ( time, object_name, func_end )

  local obj = ne.obj.Get( object_name );

  if ( not obj ) then

    if ( func_end ) then func_end(); end; return;

  end;

  local alp = obj.alp or 1;

  ne.obj.Animate(
    object_name, "alp", 0, 0, func_end,
    {
      0.0,  0, alp,
      time, 0, 0.0
    }
  );

end;
--------------------------------------------------------------------------------
function public.anim.Fake ( time, object_name, func_end )

  local obj = ne.obj.Get( object_name );

  if ( not obj ) then

    if ( func_end ) then func_end(); end; return;

  end;

  local alp = obj.alp or 1;

  ne.obj.Animate(
    object_name, "alp", 0, 0, func_end,
    {
      0.0,  0, alp,
      time, 0, alp
    }
  );

end;
--------------------------------------------------------------------------------
function public.anim.Pulse ( object_name, func_end, linear, time_factor, scale_factor )

  time_factor  = time_factor  or 1;
  scale_factor = scale_factor or 1;

  local obj = ne.obj.Get( object_name );

  if ( not obj ) then

    if ( func_end ) then func_end(); end; return;

  end;

  local base_scale  = obj.scale_x or 1;
  local upper_scale = base_scale + 0.05 * scale_factor * base_scale;
  
  if ( linear ) then

    local lower_scale = math.max( 0, base_scale - 0.05 * scale_factor * base_scale );
  
    ne.obj.Animate(
      object_name, "scale_xy", 0, 0, func_end, 
      {
        0.00 * time_factor, 0, base_scale,  base_scale,
        0.07 * time_factor, 0, upper_scale, upper_scale,
        0.15 * time_factor, 0, base_scale,  base_scale,
        0.22 * time_factor, 0, lower_scale, lower_scale,
        0.30 * time_factor, 0, base_scale,  base_scale
      }
    );

  else

    local lower_scale = math.max( 0, base_scale - 0.03 * scale_factor * base_scale );

    ne.obj.Animate(
      object_name, "scale_xy", 0, 0, func_end, 
      {
        0.00 * time_factor, 0, base_scale,  base_scale,
        0.15 * time_factor, 2, upper_scale, upper_scale,
        0.25 * time_factor, 2, lower_scale, lower_scale,
        0.28 * time_factor, 1, base_scale,  base_scale
      }
    );

  end;

end;
--------------------------------------------------------------------------------
function public.anim.Scale ( object_name, time, to_scale, func_end )

  local obj = ne.obj.Get( object_name );

  ne.obj.Animate(
    object_name, "scale_xy", 0, 0, func_end, 
    {
      0.00, 0, obj.scale_x, obj.scale_y,
      time, 0, to_scale,    to_scale
    }
  );

end;
--------------------------------------------------------------------------------
function public.anim.Spin ( object_name, func_end, func_change )

  local obj = ne.obj.Get( object_name );

  if ( not obj ) then
    if ( func_end ) then func_end(); end; return;
  end;

  local scale_x     = obj.scale_x or 1;
  local inv_scale_x = -1 * scale_x;

  ne.obj.Animate(
    object_name, "scale_x", 0, 0, function ()

      if ( func_change ) then func_change(); end;

      ne.obj.Animate(
          object_name, "scale_x", 0, 0, func_end,
          {
            0.00,  0, 0,
            0.08, 0, scale_x,
            0.28,  0, inv_scale_x,
            0.48,  0, scale_x
          }
        );

    end,
    {
      0.00,  0, scale_x,
      0.07,  0, inv_scale_x,
      0.16,  0, scale_x,
      0.28,  0, inv_scale_x,
      0.38, 0, 0
    }
  );

end;
--------------------------------------------------------------------------------
function public.anim.Rotate ( object_name, func_end, func_change )

  local obj = ne.obj.Get( object_name );

  if ( not obj ) then
    if ( func_end ) then func_end(); end; return;
  end;

  local scale_x     = obj.scale_x or 1;
  local inv_scale_x = -1 * scale_x;

  ne.obj.Animate(
    object_name, "scale_x", 0, 0, function ()

      if ( func_change ) then func_change(); end;

      ne.obj.Animate(
          object_name, "scale_x", 0, 0, func_end,
          {
            0.00,  0, 0,
            0.12, 0, scale_x
          }
        );

    end,
    {
      0.00,  0, scale_x,
      0.06,  0, inv_scale_x,
      0.09, 0, 0
    }
  );

end;
--------------------------------------------------------------------------------
function public.anim.Play ( anm_name, main_animfunc, func_end, trailing_animfunc )

  if ( type( func_end ) ~= "function" ) then
    func_end = nil;
  end;
  
  local func_OnAnimationEnd = "";

  if ( ( func_end ) and ( trailing_animfunc ) ) then

    func_OnAnimationEnd = function ()

      private.Anim_PlayWithTrigger( anm_name, trailing_animfunc, "" );

      func_end( { sender = anm_name } );

    end;

  elseif ( trailing_animfunc ) then

    func_OnAnimationEnd = function ()

      private.Anim_PlayWithTrigger( anm_name, trailing_animfunc, "" );

    end;

  elseif ( func_end ) then

    func_OnAnimationEnd = function ()

      local func_exec = func_end;

      public.obj.Set( anm_name, { endtrig = "" } );

      func_exec( { sender = anm_name } );

    end;

  end;

  private.Anim_PlayWithTrigger( anm_name, main_animfunc, func_OnAnimationEnd );

end;
--------------------------------------------------------------------------------
function public.anim.Func ( anm_name )

  local obj = ne.obj.Get( anm_name );

  if ( not obj ) then

    return;

  end;

  local  animfunc = obj.animfunc;
  return animfunc;

end;
--------------------------------------------------------------------------------
function public.anim.Stop ( anm_name, multiset )

  private.Anim_RemoveTrigger( anm_name );
  
  public.obj.Set(
    anm_name,
    {
      playing = false,
      endtrig = function () end
    },
    multiset
  );

end;
--------------------------------------------------------------------------------
function private.Anim_PlayWithTrigger ( anm_name, main_animfunc, func_end )
  
  private.Anim_AddTrigger( anm_name, main_animfunc, func_end );
  
  AnimPlay( 
    anm_name, 
    main_animfunc, 
    function () 
      private.Anim_ExecuteTrigger( anm_name, main_animfunc );
    end
  );
  
end;
--------------------------------------------------------------------------------
function private.Anim_AddTrigger ( anim_obj, anim_func, trigger )
  
  if ( private.table_anim_triggers[ anim_obj ] ) then
    public.anim.Stop( anim_obj );
  end;
  
  private.table_anim_triggers[ anim_obj ] = {
    anim_func = anim_func,
    trigger = trigger
  };
  
end;
--------------------------------------------------------------------------------
function private.Anim_RemoveTrigger ( anim_obj )
  
  if ( private.table_anim_triggers[ anim_obj ] ) then
    private.table_anim_triggers[ anim_obj ] = nil;
  end;
  
end;
--------------------------------------------------------------------------------
function private.Anim_ExecuteTrigger ( anim_obj, anim_func )
  
  if ( private.table_anim_triggers[ anim_obj ] ) then
    if ( private.table_anim_triggers[ anim_obj ].anim_func == anim_func ) then
      local trigger = private.table_anim_triggers[ anim_obj ].trigger;
      private.Anim_RemoveTrigger( anim_obj );
      if ( trigger and ( type( trigger ) == "function" ) ) then
        trigger();
      end;
    end;
  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.grid.Resize ( name, size, border, frame )
  
  frame = frame or 0;
  
  if ( border ) then
    
    GridSet( name, frame,
    {
       0,  { pos_x = -size.left,                 pos_y = -size.top },
       1,  { pos_x = -size.left  + border.left,  pos_y = -size.top },
       2,  { pos_x =  size.right - border.right, pos_y = -size.top },
       3,  { pos_x =  size.right,                pos_y = -size.top },
      
       4,  { pos_x = -size.left,                 pos_y = -size.top + border.top },
       5,  { pos_x = -size.left  + border.left,  pos_y = -size.top + border.top },
       6,  { pos_x =  size.right - border.right, pos_y = -size.top + border.top },
       7,  { pos_x =  size.right,                pos_y = -size.top + border.top },
      
       8,  { pos_x = -size.left,                 pos_y = size.bottom - border.bottom },
       9,  { pos_x = -size.left  + border.left,  pos_y = size.bottom - border.bottom },
      10,  { pos_x =  size.right - border.right, pos_y = size.bottom - border.bottom },
      11,  { pos_x =  size.right,                pos_y = size.bottom - border.bottom },
      
      12,  { pos_x = -size.left,                 pos_y = size.bottom },
      13,  { pos_x = -size.left  + border.left,  pos_y = size.bottom },
      14,  { pos_x =  size.right - border.right, pos_y = size.bottom },
      15,  { pos_x =  size.right,                pos_y = size.bottom },
    } );
    
    return;
    
  end;
  
  GridSet( name, frame,
  {
    0,  { pos_x = -size.left,  pos_y = -size.top },
    1,  { pos_x =  size.right, pos_y = -size.top },
    2,  { pos_x = -size.left,  pos_y =  size.bottom },
    3,  { pos_x =  size.right, pos_y =  size.bottom }
  } );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.skin.Attach ( destination, skin_name, skin_type )
  
  skin_type = skin_type or "ride";

  local multiset = {};
  local postfix  = "_"..destination;
  local source   = "anm_ui_skin_"..skin_type;
  local skin_res = _f.GetRes( "skin", skin_name ) or DEFAULT_SKIN_RES;

  public.obj.Clone(
    source,
    destination,
    postfix,
    1
  );

  local dest_name = string.sub( destination, 5 );
  local anm_name  = "anm_"..dest_name.."_hero";

  public.obj.Set(
    source..postfix,
    {
      name = anm_name
    },
    multiset
  );

  for p, part_name in pairs( private.skin_parts[ skin_type ] ) do

    public.obj.Set(
      source.."_"..part_name..postfix,
      {
        name     = anm_name.."_"..part_name,
        res      = skin_res, 
        animfunc = part_name, 
        playing  = true
      },
      multiset
    );
    
  end;

  public.obj.Multiset( multiset );

  return anm_name;

end;
--------------------------------------------------------------------------------
function public.skin.Detach ( anm_name )

  public.obj.Delete( anm_name );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.AspectRatio_Detect ()

  local ratio = tern( public.screen.IsLandscape() )(
    ( private.screen.width / private.screen.height ),
    ( private.screen.height / private.screen.width )
  );
  
  ratio = math.floor( ratio * 100 );
  
  local asp;
  
  asp = tern( ratio == 100 )( "1:1",  nil ); if ( asp ) then return asp; end;
  asp = tern( ratio <= 126 )( "5:4",  nil ); if ( asp ) then return asp; end;
  asp = tern( ratio <= 134 )( "4:3",  nil ); if ( asp ) then return asp; end;
  asp = tern( ratio <= 151 )( "3:2",  nil ); if ( asp ) then return asp; end;
  asp = tern( ratio <= 161 )( "8:5",  nil ); if ( asp ) then return asp; end;
  asp = tern( ratio <= 167 )( "5:3",  nil ); if ( asp ) then return asp; end;
  asp = tern( ratio <= 178 )( "16:9", nil ); if ( asp ) then return asp; end;
  asp = tern( ratio <= 189 )( "17:9", nil ); if ( asp ) then return asp; end;
  asp = tern( ratio <= 237 )( "21:9", nil ); if ( asp ) then return asp; end;
  
  return asp;
  
end;
--------------------------------------------------------------------------------
function private.WaitingInit ()

  ObjCreate( private.waiting_name, "anim" );
  ObjSet(    private.waiting_name, {
    res      = INTERFACE_PATH.."anm_waiting",
    animfunc = "static",
    playing  = true,
    pos_x    = private.screen.width / 2,
    pos_y    = private.screen.height / 2,
    pos_z    = private.int_pos_z + 1,
    tt_wait  = "str_loader_message_wait"
  } );
  ObjAttach( private.waiting_name, "ne_storage" );
  
  local bg_name = string.format( "spr_%s_black", 
                  ui.obj.GetNameWithoutPrefix( private.waiting_name ) );
  
  ObjCreate( bg_name, "spr" );
  ObjSet(    bg_name, {
    res      = "assets/resources/default",
    anim_tag = "bg",
    scale_x  = 32,
    scale_y  = 36
  } );
  ObjAttach( bg_name, private.waiting_name );

end;
--------------------------------------------------------------------------------
function private.LockInit ()

  ui.hub.Create( "interface_lock" );

  ObjCreate( "ng_interface_lock", "obj" );
  ObjAttach( "ng_interface_lock", "hub_interface_lock" );

  ObjSet( "ng_interface_lock",
  { 
    pos_x          = ( private.screen.width / 2 ),
    pos_y          = ( private.screen.height / 2 ),
    pos_z          = 1,
    input          = false,
    inputrect_init = true,
    inputrect_x    = - ( private.screen.width / 2 ),
    inputrect_y    = - ( private.screen.height / 2 ),
    inputrect_w    = private.screen.width,
    inputrect_h    = private.screen.height
  } );
  
  ObjCreate( "ng_interface_lock_spr", "grid" );
  ObjAttach( "ng_interface_lock_spr", "ng_interface_lock" );

  ObjSet( "ng_interface_lock_spr",
  { 
    res     = INTERFACE_PATH.."grd_stage_background",
    active  = false,
    visible = false,
    input   = false,
    pos_x   = - ( private.screen.width / 2 ),
    pos_y   = - ( private.screen.height / 2 ),
    scale_x = 10,
    scale_y = 10,
    alp     = 0.0,
    color_r = 1,
    color_g = 0.5,
    color_b = 0.8
  } );  

  local grid_params = 
  {
     0,  { pos_x = 0 },
     1,  { pos_x = 1 },
     2,  { pos_x = private.screen.width - 1 },
     3,  { pos_x = private.screen.width },

     4,  { pos_x = 0 },
     5,  { pos_x = 1 },
     6,  { pos_x = private.screen.width - 1 },
     7,  { pos_x = private.screen.width },

     8,  { pos_x = 0, pos_y = private.screen.height - 1 },
     9,  { pos_x = 1, pos_y = private.screen.height - 1 },
     10, { pos_x = private.screen.width - 1, pos_y = private.screen.height - 1 },
     11, { pos_x = private.screen.width,     pos_y = private.screen.height - 1 },

     12, { pos_x = 0, pos_y = private.screen.height },
     13, { pos_x = 1, pos_y = private.screen.height },
     14, { pos_x = private.screen.width - 1, pos_y = private.screen.height },
     15, { pos_x = private.screen.width,     pos_y = private.screen.height }
  };

  GridSet( "ng_interface_lock_spr", 0, grid_params );  
  
end;
--------------------------------------------------------------------------------
function private.SkinInit ()

  private.skin_parts = {
  
    fight = {
      "sword",
      "palm_r",
      "arm_r",
      "shoulder_r",
      "helm_ft",
      "head",
      "helm_bk",
      "feather",
      "foot_r",
      "leg_r",
      "body",
      "palm_l",
      "palm2_l",
      "shoulder_l",
      "foot_l",
      "leg_l",
      "arm_r_fun",
      "belt",
      "nose"
      },
      
    ride ={
      "sword",
      "palm_r",
      "arm_r",
      "shoulder_r",
      "helm_ft",
      "head",
      "helm_bk",
      "feather",
      "leg_r",
      "body",
      "palm_l",
      "shoulder_l",
      "leg_l",
      "foot2_r",
      "foot2_l",
      "foot3_r",
      "foot3_l",
      "palm2_l",
      "palm3_l",
      "palm_l_fun",
      "arm_r_fun",
      "belt",
      "nose"
    }
  };

end;
--------------------------------------------------------------------------------
