-- name=v_boosters
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_boosters",
  "int_dialog_booster_buy",
  "int_panel_booster_sorcery"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()
  
  common.mod.Load( _modules );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Init ()

  local funcs = {
    func_get_stats = function ()
      return v_player.Stats_GetPlayerData();
    end;
  }
  
  m_boosters.Setup( funcs );
  
  private.UseCrownBonus = false;
  
  v_interface.window.RegisterAsModal( "int_dialog_booster_buy" );

end;
--------------------------------------------------------------------------------
function public.TryChangeCount ( ... )

  local is_changed = m_boosters.TryChangeCount( ... );
  
  return is_changed;

end;
--------------------------------------------------------------------------------
function public.TryBuy ( booster_name, pack_level, source )

  local pack_data     = m_boosters.GetPacks( booster_name )[ pack_level ];
  local booster_price = pack_data.price;
  local booster_count = pack_data.count;

  if ( v_player.Gold_TryChangeCount( - booster_price ) ) then
    
    if ( source ) and ( string.find( source, "stage" ) ) then
      local _, stage_type = string.match( source, "stage_(%d+)(_%w+)" );
      source = "level"..( stage_type or "" );
    end;

    if ( booster_name == "extramoves" or 
         booster_name == "extralives" ) then
      v_log.Kontagent_LogEvent( "economy", 
      {
        name = booster_name,
        description = "extra_moves_offer",
        purchase = 1,
        level = 1,
        value = - booster_price
      } );
    end;

    v_log.Kontagent_LogEvent( "economy", 
    {
      description = "premium_currency_spend",
      category    = "boosts",
      value       = - booster_price,
      name        = booster_name,
      source      = source
    } );

    v_game.GameLog_Add( "booster_buy", { name = booster_name, count = booster_count } );
    
    return m_boosters.TryChangeCount( booster_name, booster_count, source, true );

  end;

  return false;

end;
--------------------------------------------------------------------------------
function public.TryUse ( booster_name, source )

  local is_changed = m_boosters.TryChangeCount( booster_name, - 1, source );
  
  if ( is_changed ) then
    v_game.GameLog_Add( "booster_use", { name = booster_name } );
  end;
  
  return is_changed;

end;
--------------------------------------------------------------------------------
function public.IsUnlocked ( booster_name )

  local is_unlocked = m_boosters.IsUnlocked( booster_name );
  
  return is_unlocked;

end;
--------------------------------------------------------------------------------
function public.Unlock ( booster_name )

  m_boosters.Unlock( booster_name );

end;
--------------------------------------------------------------------------------
function public.Tap ( callpoint, booster_name, stage_name, funcs )

  return private.Tap( callpoint, booster_name, stage_name, funcs );

end;
--------------------------------------------------------------------------------
function public.Use ( params, stage_name )

  private.Use( params, stage_name );

end;
--------------------------------------------------------------------------------
function public.List ()

  local booster_list = m_boosters.GetList();
  
  return booster_list;

end;
--------------------------------------------------------------------------------
function public.GetLots ()

  local booster_list = public.List();
  local booster_data = m_boosters.GetBoostersData( booster_list );
  local stats = v_player.Stats_GetPlayerData();
  
  for i = #booster_data, 1, - 1 do

    booster_data[ i ].phase = public.GetPowerPhase( booster_data[ i ].name, stats );
    
    if ( booster_data[ i ].type == "continue" ) then

      table.remove( booster_data, i );

    end;

  end;

  return booster_data;

end;
--------------------------------------------------------------------------------
function public.SetSelected ( boosters_list )

  private.selected = boosters_list;

end;
--------------------------------------------------------------------------------
function public.GetSelected ()

  if ( private.selected ) then

    local selected = {};

    for i = 1, #private.selected do

      table.insert( selected, private.selected[ i ] );

    end;

    private.selected = nil;

    return selected;

  end;

  return {};

end;
--------------------------------------------------------------------------------
function public.GetCount ( booster_name )

  local count = m_boosters.GetCount( booster_name );
  
  return count;

end;
--------------------------------------------------------------------------------
function public.GetPrice ( booster_name, level )

  local price = m_boosters.GetPrice( booster_name, level );
  
  return price;

end;
--------------------------------------------------------------------------------
function public.GetBoostersData ( boosters_list )

  local booster_data = m_boosters.GetBoostersData( boosters_list );
  
  return booster_data;

end;
--------------------------------------------------------------------------------
function public.GetBoosterStaticData ( booster_name )

  local static_data = m_boosters.GetBoosterStaticData( booster_name );
  
  return static_data;

end;
--------------------------------------------------------------------------------
function public.Booster_GetType ( booster_name )
  
  local static_data = m_boosters.GetBoosterStaticData( booster_name );
  
  return static_data.type;
  
end;
--------------------------------------------------------------------------------
function public.GetPacks ( booster_name )
  
  local packs = m_boosters.GetPacks( booster_name );
  local booster_pack = {};

  for b = 1, #packs do

    if ( packs[ b ].count >= 3 ) then

      booster_pack = packs[ b ];
      break;

    end;

  end;
  
  return booster_pack;
  
end;
--------------------------------------------------------------------------------
function public.Boosters_GetInfo ( booster_name )
  
  local booster = 
  {
    name     = booster_name,
    count    = public.GetCount( booster_name ),
    sale     = public.GetPacks( booster_name ).sale,
    phase    = v_boosters.GetPowerPhase( booster_name )
  };
  
  return booster;
  
end;
--------------------------------------------------------------------------------
function public.GetPowerPhase ( booster_name, statistic )
  
  local stats = statistic or v_player.Stats_GetPlayerData();
  
  if ( booster_name == "hammer" ) and ( stats.super_hammer > 0 ) then
    return "super_hammer";
  end;
  
  return nil;
  
end;
--------------------------------------------------------------------------------
function public.GetRevertPowerPhase ( booster_name )
  
  if ( booster_name == "super_hammer" ) then
    return "hammer";
  end;
  
  return booster_name;
  
end;
--------------------------------------------------------------------------------
function public.IsActive ( booster_name )
  
  if ( booster_name ) then
    return ( private.active_booster == booster_name );
  end;
  
  return private.active_booster;
  
end;
--------------------------------------------------------------------------------
function public.Reset ()

  if ( private.active_booster == "sorcery" ) then
    int_panel_booster_sorcery.__GetDelegateScheme().Hide();
    int_panel_booster_sorcery.__GetDelegateScheme().Reject();
  end;

  local params = { result = false, name = private.active_booster };
  mediator.Notify( "Event_Match3_BoosterUsed", params );
  common.MessageSend( Event_Match3_BoosterDeactivated );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Tap ( callpoint, booster_name, stage_name, funcs )
  
  funcs = funcs or {};
  
  local func_end = funcs.func_end or function () end;
  
  local func_OnTap_Shop = function ()

    local func_top_panel_show = function () end;
    local func_top_panel_hide = function () end;

    private.Dialog_Show(
      booster_name,
      nil,
      func_end,
      func_top_panel_show,
      func_top_panel_hide,
      true
    );
  
    return true;

  end;

  local func_OnTap_Other = function ()

    if ( m_boosters.GetCount( booster_name ) == 0 ) then

      local func_top_panel_show = function ()
        v_interface.panel.top.Highlight( "gold" );
        v_interface.panel.top.Show();
      end;

      local func_top_panel_hide = function ()
        v_stage.StageTopPanel_SetVisible();
        v_interface.panel.top.Hide();
      end;
      
      v_stage.StageTopPanel_SetInvisible();

      private.Dialog_Show(
        booster_name,
        stage_name,
        func_end,
        func_top_panel_show,
        func_top_panel_hide
      );

      return false;

    end;

    local booster_phase = public.GetPowerPhase( booster_name );
    
    if ( booster_phase ) then
      booster_name = booster_phase;
    end;
    
    private.active_booster = booster_name;
    
    common.MessageSend(
      Event_Match3_BoosterActivated,
      {
        name     = booster_name,
        row      = nil,
        col      = nil,
        vertical = nil 
      }
    );

    return true;

  end;

  local callpoint_funcs = 
  {
    shop   = func_OnTap_Shop,
    stage  = func_OnTap_Other,
    dialog = func_OnTap_Other
  };

  return callpoint_funcs[ callpoint ]();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.UseCrownBooster ()
  
  local stage_name = m_stage.GetCurrentStageName();
  local params = { name =  "crowns", result = true };
  private.Use ( params, stage_name );
  
end;
--------------------------------------------------------------------------------
function private.Use ( params, stage_name )

  local booster_name = params.name;
  if ( not params.result ) then

    mediator.Notify( "View_Boosters_BoosterUsageFailed", { name = booster_name } );

    return;

  end;

  local use_functions = 
  {
    hammer     =   { func = private.Use_ToChip,  effect = true,  use = true  },
    sorcery    =   { func = private.Use_Sorcery, effect = true,  use = false },
    sorcery_last = { func =                nil,  effect = false, use = true  },
    wave       =   { func = private.Use_ToChip,  effect = false, use = true  },
    bomb       =   { func = private.Use_ToChip,  effect = false, use = true  },
    tornado    =   { func = private.Use_ToArea,  effect = true,  use = true  },
    wind       =   { func = private.Use_ToPair,  effect = true,  use = true  },
    cascade    =   { func = private.Use_ToChip,  effect = false, use = true  },
    fireburst  =   { func = private.Use_ToList,  effect = true,  use = true  },
    magicgift  =   { func = private.Use_ToArea,  effect = true,  use = true  },
    heavenhelp =   { func = private.Use_ToArea,  effect = true,  use = true  },
    downright  =   { func =                nil,  effect = false, use = true  },
    nolimits   =   { func = private.Use_ToArea,  effect = true,  use = true  },
    superbomb  =   { func = private.Use_ToArea,  effect = false, use = true  },
    swatter    =   { func = private.Use_ToArea,  effect = true,  use = true  },
    shuffle    =   { func = private.Use_ToArea,  effect = true,  use = true  },
    movesup    =   { func = private.Use_ToArea,  effect = true,  use = true  },
    anticrystal=   { func = private.Use_ToChip,  effect = false, use = true  },
    superhit   =   { func = private.Use_ToArea,  effect = false, use = true  },
    crowns     =   { func = private.Use_ToCrowns,effect = true,  use = false }
  };

  local func_end = function ()
    
    private.active_booster = nil;

  end;

  if ( use_functions[ booster_name ].use ) then
    
    public.TryUse( ui.obj.GetNamePrefix( booster_name ) or booster_name, stage_name );
 
  end;
  
  if ( use_functions[ booster_name ].func ) then
    
    func_end = function ()
    
      private.active_booster = nil;

      if ( params.func_end ) then params.func_end(); end;

      ObjDelete( "anm_boosters_action_"..booster_name );
      
      mediator.Notify(
        "View_Boosters_BoosterUsedAnimEnd",
        { 
          name = booster_name,
          crown = tern( booster_name == "crowns" )( true, false )
        }
      );

    end;
    
    ObjClone( "anm_boosters_action", "_"..booster_name, 0, InterfaceName );

    ObjSet( "anm_boosters_action_"..booster_name, { pos_z = 899 } );
    
    local effect = tern( private.UseCrownBonus )( false, use_functions[ booster_name ].effect );
    
    use_functions[ booster_name ].func(
      params,
      func_end,
      effect
    );
    
  else
    func_end();
  end;
  
  mediator.Notify(
    "View_Boosters_BoosterUsed",
    { name = booster_name }
  );
  
end;
--------------------------------------------------------------------------------
function private.Use_ToChip ( params, func_end, effect )

  local booster_name   = params.name;
  local booster_object = "anm_boosters_action_"..booster_name;

  local func_end_anim = function ()

    ObjSet( booster_object,
    {
      active  = false,
      visible = false } );

    if ( func_end ) then

      func_end();

    end;

  end;

  if ( effect ) then

    local animfunc = booster_name;

    if ( params.property ) then

      animfunc = booster_name.."_"..string.lower( params.property );

    end;

    ObjSet( booster_object, 
    { 
      active   = true,
      visible  = true,
      res      = "assets/content/match3/fx/anm_"..booster_name, 
      animfunc = animfunc,
      pos_x    = params.x + 35,
      pos_y    = params.y + 35 
    } );

    ui.anim.Play( booster_object, animfunc, func_end_anim );

  else

    func_end_anim();

  end;
  
  private.Sound_Play( booster_name );

end;
--------------------------------------------------------------------------------
function private.Use_ToList ( params, func_end, effect )

  local booster_name = params.name;
  local res_name     = booster_name;

  local func_end_anim = function ( data )

    ObjDelete( data.sender );

    if ( tonumber( string.match( data.sender, "%d+" ) ) == #params.chips_list and func_end ) then

      func_end();

    end;

  end;

  for i = 1, #params.chips_list do

    ObjClone(
      "anm_boosters_action",
      "_"..booster_name.."_"..i, 1,
      InterfaceName
    );

    ObjSet( "anm_boosters_action_"..booster_name.."_"..i, 
    { 
      active   = true,
      visible  = true,
      res      = "assets/content/match3/fx/anm_"..res_name, 
      animfunc = booster_name,
      pos_x    = params.chips_list[ i ].x + 35,
      pos_y    = params.chips_list[ i ].y + 35 
    } );

    ui.anim.Play(
      "anm_boosters_action_"..booster_name.."_"..i,
      booster_name,
      func_end_anim
    );

  end;

  private.Sound_Play( booster_name );

end;
--------------------------------------------------------------------------------
function private.Use_ToArea ( params, func_end, effect )
  
  if ( effect ) then
    
    local booster_name   = params.name;
    local booster_object = "anm_boosters_action_"..booster_name;

    local x, y = v_match3.GetFieldCenter();

    ObjSet( booster_object, 
    { 
      active   = true,
      visible  = true,
      res      = "assets/content/match3/fx/anm_"..booster_name, 
      animfunc = booster_name,
      pos_x    = x,
      pos_y    = y
    } );

    local func_end_anim = function ()

      ObjSet( booster_object,
      {
        active  = false,
        visible = false } );

      if ( func_end ) then

        func_end();

      end;

    end;

    ui.anim.Play( booster_object, booster_name, func_end_anim );

    private.Sound_Play( booster_name );
  
  elseif ( func_end ) then

    func_end();
        
  end;
  
end;
--------------------------------------------------------------------------------
function private.Use_ToPair ( params, func_end, effect )

  local booster_name = params.name;

  private.Sound_Play( booster_name );

  if ( func_end ) then

    func_end();

  end;

end;
--------------------------------------------------------------------------------
function private.Use_Sorcery ( params, func_end, effect )

  private.Sorcery_Show( params.x, params.y, func_end );

end;
--------------------------------------------------------------------------------
function private.Use_ToCrowns ( params, func_end, effect )
  
  if ( effect ) then
    
    private.UseCrownBonus = true;
    
    local booster_name   = params.name;
    local booster_object = "anm_boosters_action_"..booster_name;

    local x, y = v_match3.GetFieldCenter();

    ObjSet( booster_object, 
    { 
      active   = true,
      visible  = true,
      res      = "assets/content/match3/fx/anm_"..booster_name, 
      animfunc = booster_name,
      pos_x    = x,
      pos_y    = y
    } );

    local func_end_anim = function ()
      
      private.UseCrownBonus = false;
      private.active_booster = nil;
      
      if ( func_end ) then func_end(); end;

    end;
    
    ui.anim.Play( booster_object, booster_name, func_end_anim );

    private.Sound_Play( booster_name );
  
  elseif ( func_end ) then

    func_end();
        
  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Sorcery_Show ( x, y, func_end )

  local func_hide = function ( direction )

    private.Sorcery_Hide( x, y, direction, func_end );

  end;

  local func_reject = function ()

    common.MessageSend( Event_Match3_BoosterDeactivated );

    if ( func_end ) then

      func_end();

    end;

  end;
  
  int_panel_booster_sorcery.Show( x, y, func_hide, func_reject ); 

end;
--------------------------------------------------------------------------------
function private.Sorcery_Hide ( x, y, direction, func_end )

  local booster_name   = "sorcery";
  local booster_object = "anm_boosters_action_"..booster_name;

  common.MessageSend(
    Event_Match3_BoosterActivated,
    { name = "sorcery_last", direction = direction }
  );

  local func_end_anim = function ()

    ObjSet( booster_object, { active = false, visible = false } );

    if ( func_end ) then

      func_end();

    end;

  end;

  local animfunc = booster_name.."_"..string.lower( direction );

  ObjSet( booster_object, 
  { 
    active   = true, visible = true,
    res      = "assets/content/match3/fx/anm_"..booster_name, 
    animfunc = animfunc,
    pos_x    = x + 35,
    pos_y    = y + 35 
  } );

  ui.anim.Play( booster_object, animfunc, func_end_anim );

  private.Sound_Play( booster_name );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Dialog_Show (
    booster_name,
    stage_name,
    func_end,
    func_top_panel_show,
    func_top_panel_hide,
    hide_on_deficient
  )

  local boosters = m_boosters.GetPacks( booster_name );  
  local booster_pack;
  local booster = {
    name = booster_name,
    phase = public.GetPowerPhase( booster_name )
  };
  
  for b = 1, #boosters do
    if ( boosters[ b ].count >= 3 ) then
      booster_pack = boosters[ b ];
      break;
    end;
  end;

  local func_Close = function ()
    func_top_panel_hide();
    private.Dialog_Hide();
  end;

  local func_buy = function ()

    if ( public.TryBuy( booster_name, 1, stage_name ) ) then
      return true;
    end;

    if ( hide_on_deficient ) then
      func_Close();
    end;

    return false;

  end;

  local func_OnHide = function ()
    v_interface.window.Hide( "int_dialog_booster_buy", func_end );
    v_log.Playhaven_LogPlacement( "boost_ingame_exit" );
  end;

  v_log.Playhaven_LogPlacement( "boost_ingame_purchase" );
  
  v_interface.window.Show(
    "int_dialog_booster_buy",
    function ()

      int_dialog_booster_buy.Show( booster, booster_pack );

    end,
    nil,
    {
      Buy    = func_buy,
      Close  = func_Close,
      OnShow = func_top_panel_show,
      OnHide = func_OnHide
    }
  );

end;
--------------------------------------------------------------------------------
function private.Dialog_Hide ()

  int_dialog_booster_buy.Hide();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Sound_Play ( booster_name )
  
  local path = "assets/content/boosters/aud_booster_"..booster_name;
  
  common.sound.Play( path );
  
end;
--------------------------------------------------------------------------------