-- name=v_builds
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_builds",
  "int_dialog_map_build_construct",
  "int_dialog_map_build_gift"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  common.mod.Load( _modules );

  private.fragments = {};
  private.visible   = {};

  --Log.i( "Initialized." );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.GetAllInSection ( section_name, only_static_data )

  local builds = m_builds.Builds_GetAllBySection( section_name, ( not only_static_data ) );

  return builds;

end;
--------------------------------------------------------------------------------
function public.Create ( build_data )

  local build_name = build_data.name;
  local obj_name   = "obj_map_builds_"..build_name;
  
  build_data = private.AutoFix( build_data );
  
  ObjSet( obj_name, { anim_tag = build_name } );
  ObjClone( "btn_map_builds", "_"..build_name, 1, obj_name );
  
  if ( not ( ( build_data.constructed ) and ( build_data.vanish ) ) ) then

    ui.control.EventsSet( obj_name,
      { 
        func_on_click = function () private.Tap( build_name ); end
      }
    );
    
  else

    ui.input.Off( obj_name );

  end;
  
  if ( not build_data.constructed ) and ( build_data.gift ) then
    local build_gift_name_obj = "obj_map_builds_gift_"..build_name;
    ObjCreate( build_gift_name_obj, "obj" );
    ObjAttach( build_gift_name_obj, "anm_map_"..build_data.section );
    ObjClone( "btn_map_builds_gift", "_"..build_name, 1, build_gift_name_obj );
    local build_gift_name_btn = "btn_map_builds_gift_"..build_name;
    local build_gift_name_anm = "anm_map_builds_gift_"..build_name;
    ObjSet( build_gift_name_obj, 
    { 
      anim_tag = "at_"..build_name.."_gift";
    } );
    ObjSet( build_gift_name_anm, 
    { 
      visible = true,
      playing  = true,
      animfunc = "static"
    } );
    ui.control.EventsSet( build_gift_name_btn,
      {
        func_on_click = function () private.TapGift( build_name ); end;
      }
    );
    
  end;
  
  if ( build_data.type == "puzzle" ) then
    
    ObjSet( "anm_map_builds_icon_state_"..build_name, 
    { 
      res = "assets/content/"..build_data.section.."/map/anm_"..build_data.icon.."_states",
      anim_tag = "states"
    } );

    ui.obj.Attach(
      "anm_map_builds_icon_state_"..build_name,
      "anm_map_builds_icon_"..build_name
    );
    
    ui.input.On( "anm_map_builds_icon_"..build_name );
    ui.input.Off( "obj_map_builds_indicator_"..build_name );

    ui.control.EventsSet( "anm_map_builds_icon_state_"..build_name,
      { 
        func_on_click = function () private.Tap( build_name ); end
      }
    );
    
    if ( build_data.fragments ) then

      for p = 1, #build_data.fragments do

        local fragment_name = "anm_map_builds_"..build_name.."_part_"..p;
        if ( ui.obj.Exists( fragment_name ) ) then
          
          ObjSet( fragment_name, 
          { 
            res = "assets/content/"..build_data.section.."/map/anm_"..build_data.icon.."_states",
            anim_tag = "defeat"..p,
            animfunc = "defeat"
          } );
        
          ui.control.EventsSet( fragment_name,
            { 
              func_on_click = function ()
                private.Tap( build_name, fragment_name );
              end
            }
          );

        end;

      end;
      
      if ( ObjGet( "anm_map_builds_"..build_name.."_dummy" ) ) then
        
        ObjSet( "anm_map_builds_"..build_name.."_dummy", 
        { 
          res = "assets/content/"..build_data.section.."/map/anm_"..build_data.icon.."_states",
          animfunc = "dummy"
        } );
        
      end;

    end;

  elseif ( ( build_data.type == "island" ) and build_data.fragments ) then

    for i = 1, #build_data.fragments do

      local fragment_object = "obj_map_builds_" .. build_name ..
                              "_fragment_" .. build_data.fragments[ i ];
      
      if ( ui.obj.Exists( fragment_object ) ) then
        ObjAttach( fragment_object, "anm_map_builds_icon_"..build_name );
      end;

    end;

  elseif ( build_data.type == "build" ) then

    -- JK: turn off input of the indicator at creation,
    -- as it may interfere with the parent object input
    ui.input.Off( "obj_map_builds_indicator_"..build_name );

  end;

  private.Set( build_data );

end;
--------------------------------------------------------------------------------
function private.Set ( build_data )

  local name    = build_data.name;
  local section = build_data.section;
  local icon    = build_data.icon;

  private.visible[ name ] = true;

  local obj_name = "obj_map_builds_"..name;
  local anm_name = "anm_map_builds_icon_"..name;
  
  local anm_res = tern( icon == "dummy" )(
    "assets/content/map/builds/anm_build_"..icon,
    "assets/content/"..section.."/map/anm_"..icon
  );
  
  local build_type  = build_data.type;
  local fragments   = build_data.fragments or {};
  local unlocked    = build_data.unlocked or false;
  local constructed = build_data.constructed or false;
  local collected   = build_data.collected_fragments or {};

  local set = {};
  local animfunc = tern( constructed )( "normal", "corrupted" );

  ui.obj.Show( obj_name, set );
  ui.obj.Set(
    anm_name, 
    {
      res      = anm_res,
      animfunc = animfunc,
      playing  = tern( animfunc == "normal" )( false, true )
    },
    set
  );

  -- Inputs
  if ( build_type == "island" ) or
     ( build_type == "puzzle" ) then
   
    ui.obj.Set(
      obj_name,
      {
        input = not constructed
      },
      set
    );
    
  end;
  
  -- Fragments
  if ( #collected > 0 ) then

    private.fragments[ name ] = collected;

  end;

  if ( build_type == "island" ) then

    for i = 1, #fragments do
      
      local fragment_object = "obj_map_builds_"..name .. 
                              "_fragment_"..fragments[ i ];
      
      if ( ui.obj.Exists( fragment_object ) ) then

        if ( common.IsEntityInList( collected, fragments[ i ] ) ) then

          ui.obj.Enable(
            obj_name.."_fragment_"..fragments[ i ],
            true,
            set
          );
          
        else

          ui.obj.Disable(
            obj_name.."_fragment_"..fragments[ i ],
            true,
            set
          );

        end;

      end;

    end;

  elseif ( build_type == "puzzle" ) then

    local states_name = "anm_map_builds_icon_state_"..name;
    local state = #collected;
    
    -- ui.control.TapClear( "anm_map_builds_"..name.."_part_"..state );

    table.insert( set, { states_name,
    {
      active   = true,
      visible  = true,
      res      = "assets/content/"..section.."/map/anm_"..icon.."_states",
      animfunc = "state"..state,
      playing  = true
    } } );
    
    for i = 1, #fragments do

      if ( i <= #collected ) then
      
        local fragment_name = "anm_map_builds_"..name.."_part_"..i;
        if ( ui.obj.Exists( fragment_name ) ) then
        
          ui.input.Off( fragment_name, set );
        
        end;

      end;

    end;

  -- Build indicator
  elseif ( build_type == "build" ) then

    local obj_indicator = "obj_map_builds_indicator_"..name;
    local spr_indicator = "spr_map_build_indicator_"..name;

    if ( unlocked and not constructed ) then
      if ( not ObjGet( spr_indicator ) ) then

        ObjClone( "spr_map_build_indicator", "_"..name, 1, obj_indicator );
        ui.obj.Enable( obj_indicator, nil, set );
        ui.control.EventsSet( obj_indicator, { 
          func_on_click = function () private.Tap( name ); end
        } );
        
      end;
    end;

  end;

  ui.obj.Multiset( set );

end;
--------------------------------------------------------------------------------
-- function ... end
--------------------------------------------------------------------------------
function private.Tap ( build_name, build_part )
  
  local build_data = m_builds.Build_GetData( build_name );
  
  -- Play tap sound
  local sounds =
  {
    build  = "tap_build",
    island = "tap_build",
    puzzle = "tap_build"
  };
  
  common.sound.Play( sounds[ build_data.type ] );
  
  -- Animate click
  private.Animate_Click( build_data );
  
  -- Try show construct dialog or popup
  if not ( build_data.constructed ) then
  
    if ( build_data.type == "build" and build_data.unlocked ) then
  
      private.Dialog_Build_Show( build_name, function ()
      
        local contructed_data = m_builds.Build_GetData( build_name );
      
        private.Build_HideIndicator( build_name );
        
        local animation = private.GetAnimData_Construct( contructed_data );
        local func_anim = v_map.Callback_Animate( animation );
        
        v_map.Builds_UpdateStages( contructed_data, func_anim );
        
      end );
    
    else
    
      local func_after;
    
      if ( build_data.type == "puzzle" ) then

        func_after = function ()

          private.Puzzle_HighlightStage( build_data.fragments[ 1 ] );

        end;
        
      elseif ( build_data.type == "island" ) then

        local is_autoconstructed = m_builds.Build_IsAutoConstructed(
          build_data.name,
          build_data
        );
        
        if ( not is_autoconstructed ) then
        
          func_after = function ()
          
            public.Build_HighlightStage( build_data.name );
          
          end;
        
        end;
        
      end;

      private.Build_ShowPopup( build_data, build_part, func_after );
    
    end;
    
  else
  
    private.Focus_Build( build_name );
  
  end;
  
end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function private.TapGift ( build_name )
  
  common.sound.Play( "button_click" );
  
  v_map.ShowObjectPopup(
    "monolog",
    "obj_map_builds_gift_indicator_"..build_name,
    ne.strings.Get( "str_int_popup_cloud_builds_gift_"..build_name ),
    nil,
    nil,
    nil
  );
  
end;
--------------------------------------------------------------------------------
function public.Callback_AnimateGiftOnBuild ( build_name )
  
  local build_data = m_builds.Build_GetData( build_name );
  
  if ( build_data and build_data.gift ) then
    
    return function ( func_after )
      
      ui.screen.Lock( true );
      ObjSet( "btn_map_builds_gift_"..build_name, { input = false } );
      
      local gift = build_data.gift;
      
      if ( gift.type ) and ( gift.type == "booster" ) then
        gift.phase = v_boosters.GetPowerPhase( gift.name );
      end;
      
      local func_dialog = function ()
        
        private.DialogGiftShow( build_name, gift,
          function ()
            ui.screen.Lock( false );
          end,
          function ()
            v_items.Reward_Show(
              gift,
              function ()
                if ( func_after ) then func_after(); end;
              end
            );
          end
        );
        
      end;
      
      private.OpenGift( build_name, func_dialog );
      
    end;
    
  end;
  
  return nil;
  
end;
--------------------------------------------------------------------------------
function private.OpenGift ( build_name, func_end )

  common.sound.Play( "click_present" );
  ui.anim.Play( "anm_map_builds_gift_"..build_name, "get", func_end );

end;
--------------------------------------------------------------------------------
function private.DialogGiftShow ( build_name, gift, func_open, func_hide )

  local delegates = {
    Open  = func_open,
    Close = function () private.DialogGiftHide( func_hide ); end
  };

  common.delegate.Set( "int_dialog_map_build_gift", delegates );

  v_interface.window.AddToQueue(
    "int_dialog_map_build_gift",
    function ()
      int_dialog_map_build_gift.Show( build_name, gift );
    end
  );

end;
--------------------------------------------------------------------------------
function private.DialogGiftHide ( func_OnHide )

  int_dialog_map_build_gift.Hide(
    func_OnHide,
    function ()
      v_interface.window.Hide( "int_dialog_map_build_gift" );
    end
  );

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function private.Animate_Click ( build_data, state )

  state = state or tern( build_data.constructed )( "normal", "corrupted" );
  
  local anm_build = "anm_map_builds_icon_"..build_data.name;

  if ( ui.obj.Get( anm_build ).animfunc ~= "click_"..state ) then

    ui.anim.Play( anm_build, "click_"..state, function ()
    
      if ( state ~= "normal" ) then
        ui.anim.Play( anm_build, state );
      else
        ui.obj.Set( anm_build, { animfunc = state, playing = false } );
      end;
      
    end );

  end;

end;
--------------------------------------------------------------------------------
function private.Animate_Construct ( build_data, func_ShowPortal )

  local build_name = build_data.name;
  local build_type = build_data.type;

  local obj_name = "obj_map_builds_"..build_name;
  local anm_name = "anm_map_builds_icon_"..build_name;

  ui.anim.Stop( anm_name );

  if ( build_type == "puzzle" ) or
     ( build_type == "island" ) or
     ( build_data.vanish      ) then

    ui.input.Off( obj_name );
    ui.control.EventsClear( obj_name );

  else

    ui.input.On( obj_name );

  end;

  ui.obj.Set( anm_name, { animfunc = "normal", playing = false } );

  if ( func_ShowPortal ) then func_ShowPortal(); end;

end;
--------------------------------------------------------------------------------
function private.GetAnimData_Obtain ( build_data, fragment )

  local anim = {};
  
  anim.name   = "obj_map_builds_"..build_data.name;
  anim.coords = { pos_x = 0, pos_y = 0 };
  
  local anm_name = "anm_map_builds_icon_"..build_data.name;
  
  if     ( build_data.type == "build" ) then
  
    anim.func_action = function ( func_after )
    
      ui.anim.Play( anm_name, "click_corrupted", function ()
      
        private.Set( build_data );
        ui.anim.Play( anm_name, "corrupted" );
        if ( func_after ) then func_after(); end;
        
      end );
    
    end;
  
  elseif ( build_data.type == "island" ) then
  
    anim.func_action = function ( func_after )
    
      ui.anim.Play( anm_name, "action", function ()
      
        private.Set( build_data );
        ui.anim.Play( anm_name, "corrupted" );
        
      end );
    
      -- Fadeout item indicator if exists
      local obj_indicator = "obj_map_builds_indicator_"..build_data.name;
      local spr_item_icon = "spr_map_build_island_item_icon_"..build_data.name;
      local obj_item = "obj_map_builds_"..build_data.name.."_item";
        
      if ( ObjGet( obj_item ) ) then

        ObjSet( spr_item_icon,
        {
          color_r = 1.0, 
          color_g = 1.0,
          color_b = 1.0
        } );

        local function func_fade_indicator ()

          ObjAnimate( obj_indicator, "alp", 0, 0, "",
          {
            0.0, 0, 1.0,
            0.3, 0, 0.0
          } )

        end;

        ui.anim.Bounce( obj_indicator, func_fade_indicator );

      end;
        
      -- Hide sign of fork or animate fragment on place
      if ( ObjGet( spr_item_icon) ) then
          
        ObjAnimate( obj_indicator, "alp", 0, 0,
        function ()
          ObjSet( obj_indicator, { active  = false, visible = false } );
        end,
        {
          0.0, 3, 1.0,
          0.3, 3, 0.0
        } );
          
      else
      
        local obj_fragment = anim.name.."_fragment_"..fragment.name;

        ObjSet( obj_fragment,
        {
          active  = true,
          visible = true,
          alp     = 0.0
        } );

        ObjAnimate( obj_fragment, "alp", 0, 0, "",
        {
          0.0, 3, 0.0,
          0.3, 3, 1.0
        } );
      
      end;
      
      if ( func_after ) then func_after(); end;
      
    end;
  
  elseif ( build_data.type == "puzzle" ) then
  
    anim.func_action = function ( func_after )
    
      private.Set( build_data );
      ui.anim.Play( anm_name, "action", function ()
      
        ui.anim.Play( anm_name, "corrupted" );
        
      end );
      
      local pat_part  = "anm_map_builds_%s_part_%s";
      local pat_dummy = "anm_map_builds_%s_dummy";
      
      local count = #build_data.collected_fragments;
      
      local anm_part  = pat_part:format( build_data.name, count );
      local anm_dummy = pat_dummy:format( build_data.name );

      if ( ObjGet( anm_part ) ) then

        ui.anim.Play( anm_dummy, "dummy" );
        ui.anim.Play( anm_part, "defeat" );

      end;
      
      if ( func_after ) then func_after(); end;
    
    end;
  
  else
  
    local msg = "Error! Can't animate obtain for build %q of type %q!"
    DbgTrace( msg:format( build_data.name, build_data.type ) );
  
  end;
  
  return anim;

end;
--------------------------------------------------------------------------------
function private.GetAnimData_Construct ( build_data, func_ShowPortal )

  local anim = {};
  anim.name  = "obj_map_builds_"..build_data.name;
  
  local anm_name = "anm_map_builds_icon_"..build_data.name;
  
  if ( m_builds.Build_IsAutoConstructed( build_data.name ) ) then
  
    anim.skip_focus  = true;
    anim.func_action = function ( func_after )
    
      local anm_name = "anm_map_builds_icon_"..build_data.name;

      if ( build_data.vanish ) then
        
        ui.input.Off( anim.name );
        ui.control.EventsClear( anim.name );

      end;

      ui.anim.Play( anm_name, "appear", func_after, "normal" );
    
    end;
    
  else
  
    anim.res    = "assets/content/map/effects/anm_appearance_build";
    anim.coords = { pos_x = 0, pos_y = 0, scale_x = 0.5, scale_y = 0.5 };
    
    anim.func_preset = function ()
    
      private.Animate_Click( build_data, "corrupted" );
    
    end;
    
    anim.func_dummy = function ()
    
      common.sound.Play( "unspell" );
      
      private.Set( build_data );
      private.Animate_Construct( build_data, func_ShowPortal );
    
    end;
    
    anim.func_action = function ( func_after )
  
      if ( func_after ) then func_after(); end;
    
    end;
  
  end;
  
  return anim;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
--[[
  Build reaction format:
  {
    -- Basic data:
    name       = "build_name",
    func_popup = function ( func_after ) end,
    func_anim  = function ( func_after ) end,
    
    -- Only for reward drop:
    object          = "obj_build_name", -- receiver-object
    func_drop       = function () end,  -- on-drop function
    call_drop_first = true, -- true : func_drop on focus, func_anim on drop
                            -- false: func_anim on focus, func_drop on drop
    
    -- Rewards:
    skinpart = "skn_part_name",
    
    -- Other:
    show_stages = { min = 1, max = 3, type = "bonus" },
    move_marker = true
  }
]]
--------------------------------------------------------------------------------
function public.GetRewardReaction ( rewards, func_ShowPortal )

  local reaction = nil;
  
  if not ( rewards ) then return reaction; end;
  
  -- Find fragment
  local fragment = nil;
  for i, reward in pairs( rewards ) do
  
    if ( reward.type == "fragment" ) then
    
      fragment = reward;
      break;
    
    end;
  
  end;
  
  if not ( fragment ) then return reaction; end;
  
  -- Try setup reaction
  local build_name = m_builds.Build_GetNameByFragment( fragment.name );

  if ( build_name ) then
  
    local build_data  = m_builds.Build_GetData( build_name );
    local constructed = build_data.constructed or false;
    local collected   = build_data.collected_fragments or {};
    
    reaction        = {};
    reaction.name   = build_name;
    reaction.object = "obj_map_builds_"..build_name;
    
    local anim_skip_focus = false;
    
    if ( build_data.type == "puzzle" ) then

      local idx_part = #collected;
      local obj_part = "anm_map_builds_"..build_name.."_part_"..idx_part;
      
      if ( ui.obj.Exists( obj_part ) ) then
        reaction.object = obj_part;
        anim_skip_focus = true; 
      end
    
    end;
    
    reaction.call_drop_first = true;
    reaction.func_drop       = nil;
    
    if ( constructed ) then
    
      local animation = private.GetAnimData_Construct(
        build_data,
        func_ShowPortal
      );
      animation.skip_focus = anim_skip_focus;
      reaction.func_anim   = v_map.Callback_Animate( animation );
      
      if ( build_data.reward                    ) and
         ( build_data.reward.type == "skinpart" ) then
      
        reaction.skinpart = build_data.reward.name;
      
      end;
      
      reaction.show_stages = build_data.unlocklist;
      reaction.move_marker = build_data.last or 
                             ( build_data.unlocklist.type == "common" );
      
    else
      
      if ( build_data.type ~= "build" ) then
    
        reaction.func_popup = function ( func_after )
          private.Build_OnItemReceived( build_data, fragment, func_after );
        end;
      
      end;
      
      local animation = private.GetAnimData_Obtain(
        build_data,
        fragment
      );
      animation.skip_focus = anim_skip_focus;
      reaction.func_anim   = v_map.Callback_Animate( animation );
    
    end;
  
  end;
  
  return reaction;

end;
--------------------------------------------------------------------------------
function public.GetStageCompleteReaction ( stage_name )

  local reaction   = nil;
  local build_data = m_builds.Build_GetDataByUnlocker( stage_name, true );
  
  if ( build_data ) then
  
    if ( build_data.constructed ) then
    
      reaction           = {};
      reaction.name      = build_data.name;
      
      local animation    = private.GetAnimData_Construct( build_data );
      reaction.func_anim = v_map.Callback_Animate( animation );
      
      if ( build_data.reward and build_data.reward.type == "skinpart" ) then
      
        reaction.skinpart = build_data.reward.name;
      
      end;
      
      reaction.show_stages = build_data.unlocklist;
      reaction.move_marker = build_data.last or 
                             ( build_data.unlocklist.type == "common" );
    
    elseif ( build_data.type == "build" and build_data.unlocked ) then
    
      reaction           = {};
      reaction.name      = build_data.name;
      reaction.func_anim = function ( func_after )
          
        private.Build_ShowIndicator( build_data.name );
        if ( func_after ) then func_after(); end;
      
      end;
    
    end;
  
  end;
  
  return reaction;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Dialog_Build_Show ( build_name, func_construct )

  local build_data        = m_builds.Build_GetData( build_name );
  local deficient         = private.Deficient_Detect( build_data.price );
  local build_content     = build_data.section;
  local construct_price   = build_data.price;
  local construct_rewards = build_data.reward;

  local construct_fragments = nil;

  if ( build_data.fragments and ( #build_data.fragments > 0 ) ) then
  
    local collected = build_data.collected_fragments or {};

    construct_fragments = {};

    for f = 1, #build_data.fragments do

      local fragment  = build_data.fragments[ f ];
      
      table.insert( construct_fragments, {
        name      = fragment,
        collected = common.IsEntityInList( collected, fragment );
      } );

    end;

  end;

  local delegates = {

    Accept = function ()

      if ( private.Build_TryConstruct( "manual", build_name ) ) then
 
        private.Dialog_Build_Hide();
        func_construct();
        
      end;
      
    end,

    Reject = private.Dialog_Build_Hide,

    Find = function ( fragment )

      local stage_data = v_stage.GetUsableStages( fragment );
      if  ( stage_data ) then
      
        private.Focus_Stage( stage_data.name );
        return;
        
      end;

      local character = v_characters.GetCharacterByFragment( fragment );
      if  ( character ) then
        
        private.Focus_Character( character.name );
        return;

      end;

    end,

    ResolveDeficient = function ()
      
      return private.Deficient_TryResolve( deficient, build_name );
      
    end,

    ShowTopPanel = v_interface.panel.top.Show,
    HideTopPanel = v_interface.panel.top.Hide,
    LockTopPanel = v_interface.panel.top.Lock

  };

  common.delegate.Set( "int_dialog_map_build_construct", delegates );

  v_interface.window.AddToQueue(
    "int_dialog_map_build_construct",
    function ()

      int_dialog_map_build_construct.Show(
        build_name,
        build_content,
        construct_price,
        deficient,
        construct_fragments,
        construct_rewards
      );

    end
  );

end;
--------------------------------------------------------------------------------
function private.Dialog_Build_Hide ()

  --local window = "int_dialog_map_build_construct"
  --local func_hide = function () v_interface.window.Hide( window ); end

  int_dialog_map_build_construct.Hide( --[[func_hide]] );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Focus_Build ( build_name, func_after )

  v_map.Focus( "obj_map_builds_"..build_name, { func_after = func_after } );

end;
--------------------------------------------------------------------------------
function private.Focus_Stage ( target_name )

  local func_get_blocker_name = function ( stage_name )

    local stage_data    = v_stage.Stage_StaticData( stage_name );
    local stage_type    = stage_data.realtype;
    local stage_section = stage_data.section;
    local stage_num     = tonumber(stage_name:match( "%d+" ));
    local build_data    = m_builds.Builds_GetAllBySection( stage_section );

    for b = 1, #build_data do

      local list = build_data[ b ].unlocklist;

      if ( list and ( ( list.type == stage_type ) or 
                      ( list.type == "common"   ) ) ) then

        if ( stage_num >= list.min ) and ( stage_num <= list.max ) then

          return build_data[ b ].name;

        end;

      end;

    end;

  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- 
  local func_focus_stage = function ( stage_name ) 

    local obj_name = "obj_map_point_"..stage_name;
    
    v_map.Focus( obj_name, { func_after = function ()
    
      ui.anim.Spin( obj_name, function()

        local focus_name = string.format( "fcs_map_point_%s", stage_name );

        ObjAnimate( focus_name, "alp", 0, 0, "", 
        {
          0.0, 2, 0,
          0.3, 2, 1,
          0.6, 2, 0,
          0.9, 2, 1,
          1.2, 2, 0
        } );

      end );
        
    end } );

  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  local func_focus_build = function ( name )
  
    private.Focus_Build( name, function () private.Tap( name ); end );

  end;
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- 
  v_interface.window.HideAll( function ()

    local build_name = func_get_blocker_name( target_name );

    if ( build_name and not m_builds.Build_IsConstructed( build_name ) ) then

      func_focus_build( build_name );

    else

      func_focus_stage( target_name );

    end;

  end );

end;
--------------------------------------------------------------------------------
function private.Focus_Character ( character )
  
  local obj_name  = "obj_map_characters_"..character;
  
  v_interface.window.HideAll( function ()

    v_map.Focus( obj_name, 
      { 
        func_after = function () v_characters.Tap( character ); end
      }
    );

  end );  
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Build_TryConstruct ( event_type, ... )

  local build_name, constructed = nil, nil;
  
  if ( event_type == "stage" ) then
    
    build_name, constructed = m_builds.Build_OnStageCompleted( ... );
    
  elseif ( event_type == "fragment" ) then
    
    build_name, constructed = m_builds.Build_OnFragmentCollected( ... );
    
  elseif ( event_type == "manual" ) then
    
    build_name, constructed = ..., false;
    
  end;

  if ( build_name ) then
    
    local build_data = m_builds.Build_GetData( build_name );
    
    local function onConstruct ()
      
      mediator.Notify( "View_Builds_BuildConstructed", build_name );
      
      local stage_name = "stage_"..build_data.unlocklist.min;

      if ( build_data.unlocklist.type == "bonus" ) then 

        stage_name = "stage_"..build_data.unlocklist.min.."_bonus";

      end;

      v_stage.Stage_TryEnable( stage_name );

      -- NOTE: Build rewards has no unified format!
      if ( build_data.reward ) then
      
        local reward;
      
        if     ( build_data.type == "build"  ) then reward = build_data.reward[ 1 ];
        elseif ( build_data.type == "puzzle" ) then reward = build_data.reward;
        elseif ( IsEditor()                  ) then
        
          local msg = "Error! Can't obtain reward for build %q of type %q";
          DbgTrace( msg:format( build_data.name, build_data.type ) );
        
        end;
        
        if ( reward ) then 

          reward.category = "build";
          v_items.TryObtain( reward );
        
        end;
      
      end;
      
      if ( build_data.gift ) then
        local gift = build_data.gift;
        gift.category = "build";
        v_items.TryObtain( gift );
      end;
      
    end;

    if ( constructed ) then
      
      onConstruct();
      
      return build_name;
      
    else
      
      if ( event_type == "manual" ) then
        
        local price        = build_data.price;
        local fragments    = build_data.fragments;
        local unlock_stage = build_data.unlock;

        -- if unlock_stage
        if ( unlock_stage       ) and 
           ( unlock_stage ~= "" ) then

          local is_stage_completed = v_stage.Stage_IsCompleted( unlock_stage );

          if not ( is_stage_completed ) then

            return false;

          end;

        end;
        -- if unlock_stage end
     
        -- if fragments
        if ( fragments ) then

          if ( not build_data.collected_fragments           ) or
             ( #build_data.collected_fragments < #fragments ) then

            for f = 1, #fragments do

              local lost_fragment = fragments[ f ];

              if ( not common.IsEntityInList( build_data.collected_fragments, lost_fragment ) ) then

                local params = { 
                  name      = build_name,
                  deficient = {
                    name    = lost_fragment,
                    count   = 1,
                    section = build_data.section
                  }
                };

                private.Deficient_Show( params );

                return false;

              end;

            end;

          end;

        end;
        -- if fragments end

        -- if price
        if ( price ) then

          local resources_check_table = {};

          for i = 1, #price do

            table.insert(
              resources_check_table,
              {
                name  = price[ i ].resource,
                delta = - price[ i ].count
              }
            );

          end;

          local resource_sufficient = v_player.Resource_TryChangeMultipleCount(
            resources_check_table
          );

          if ( not resource_sufficient.result ) then

            return false;

          end;

          for i = 1, #resource_sufficient.resources do

            v_log.Kontagent_LogEvent( "economy", 
            {
              description = "soft_currency_spend",
              category    = "build",
              value       = resource_sufficient.resources[ i ].delta,
              name        = build_name,
              type        = resource_sufficient.resources[ i ].name
            } );

          end;
          
          m_builds.Build_SetData( build_name, { constructed = true } );
          onConstruct();
          
          return build_name;

        end;
        -- if price end
        
      end;
      
    end;
  
  end;

  return build_name;

end;
--------------------------------------------------------------------------------
function private.Build_ShowPopup ( build_data, build_part, func_after )

  local target, mode, data;
  local functions = { func_preclose = func_after };
  
  local obj_ind = build_part or "obj_map_builds_indicator_"..build_data.name;
  local obj_bld = "obj_map_builds_"..build_data.name;
  
  target = tern( ui.obj.Exists( obj_ind ) )( obj_ind, obj_bld );
  
  local fragments = #( build_data.fragments or {} );
  local collected = #( build_data.collected_fragments or {} );
  
  if ( fragments > 0 ) then
  
    data =
    {
      count  = collected,
      target =
      {
        type  = "fragment",
        name  = build_data.fragments[ 1 ]:gsub( "_%d+$", "" ),
        count = fragments
      }
    };

  else

    -- JK: no fragments - nothing to show
    if ( func_after ) then func_after(); end;
    return;
  
  end;
  
  v_map.ShowObjectPopup( "counter", target, nil, data, nil, functions );

end;
--------------------------------------------------------------------------------
function private.Build_OnItemReceived ( build_data, fragment, func_after )

  private.Build_ShowPopup( build_data, nil, func_after );

end;
--------------------------------------------------------------------------------
function private.Build_ShowIndicator ( build_name )

  local parent = "obj_map_builds_indicator_"..build_name;
  local sprite = "spr_map_build_indicator_"..build_name;
  
  if ui.obj.Exists( sprite ) then return; end;
  
  ui.obj.Enable( parent );
  ui.control.EventsSet( parent, { 
    func_on_click = function () private.Tap( build_name ); end
  } );
  
  ObjClone( "spr_map_build_indicator", "_"..build_name, 1, parent );
  ObjSet( sprite, { active = true, visible = true } );

  ObjAnimate( sprite, "scale_xy", 0, 0, "",
  { 
    0.0, 0, 0.0, 0.0,
    0.3, 0, 1.0, 1.0
  } );

end;
--------------------------------------------------------------------------------
function private.Build_HideIndicator ( build_name )

  local sprite = "spr_map_build_indicator_"..build_name;
  
  if not ui.obj.Exists( sprite ) then return; end;

  ObjAnimate( sprite, "scale_xy", 0, 0, function () ObjDelete( sprite ); end,
  { 
    0.0, 0, 1.0, 1.0,
    0.3, 0, 0.0, 0.0
  } );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Deficient_Show ( params )

  local build_name = params.name;
  local deficient  = params.deficient;

  if ( #deficient == 0 ) then

    deficient = { deficient };

  end;
  
  local first = deficient[ 1 ];

  local func_solution  = function () end;
  local solution       = "stage";
  local stage_data     = v_stage.GetUsableStages( first.name );
  local character_data = v_characters.GetCharacterByFragment( first.name );

  if ( stage_data ) then

    func_solution = function ()

      private.Focus_Stage( stage_data.name );

    end;
    
  elseif ( character_data ) then
    
    func_solution = function ()

      private.Focus_Character( character_data.name );

    end;

  end;
  
  local func_accept = function ()
  
    v_game.dialog.Deficient_Hide( func_solution );
  
  end;

  local func_reject = function ()

    v_game.dialog.Deficient_Hide();
  
  end;

  v_game.dialog.Deficient_Show(
    deficient,
    func_accept,
    func_reject,
    solution
  );

end;
--------------------------------------------------------------------------------
function private.Deficient_Detect ( build_price )
  
  local deficient =
  {
    list  = {},
    delta = {},
    gold  = 0
  };
  
  local market_prices   = {};
  local market_settings = v_shop.Market_Settings( "buy" );
  
  for l, lot in pairs( market_settings ) do
    
    market_prices[ lot.name ] = { value = lot.value, price = lot.price };
    
  end;
  
  for p, price in pairs( build_price ) do
    
    local available_count = v_player.Resource_GetCount( price.resource );
    
    if ( available_count < price.count ) then
      
      local delta = price.count - available_count;
      
      table.insert( deficient.list,  price.resource );
      table.insert( deficient.delta, delta );
      
      deficient.gold = deficient.gold + delta / market_prices[ price.resource ].value * market_prices[ price.resource ].price;
      
    end;
    
  end;
  
  deficient.gold = math.ceil( deficient.gold );
   
  return tern( #deficient.list < 1 )( false, deficient );
  
end;
--------------------------------------------------------------------------------
function private.Deficient_TryResolve ( deficient, build_name )
  
  if ( v_player.Gold_TryChangeCount( - deficient.gold ) ) then
    
    v_log.Kontagent_LogEvent( "economy", 
    {
      description = "premium_currency_spend",
      category    = "build",
      value       = - deficient.gold,
      name        = build_name
    } );    
    
    for d, delta in pairs( deficient.delta ) do
      
      v_player.Resource_TryChangeCount( deficient.list[ d ], delta );
      
      v_log.Kontagent_LogEvent( "economy", 
      {
        description = "soft_currency_collect",
        category    = "build",
        value       = delta,
        name        = build_name,
        type        = deficient.list[ d ]
      } );
      
    end;
    
    return true;
    
  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Build_TryObtainFragment ( ... )

  return private.Build_TryConstruct( "fragment", ... );

end;
--------------------------------------------------------------------------------
function public.Build_TryUnlockByStage ( stage_name )
  
  return private.Build_TryConstruct( "stage", stage_name );

end;
--------------------------------------------------------------------------------
function public.Build_GetData ( build_name )

  local build_data = m_builds.Build_GetData( build_name );
  
  return build_data;

end;
--------------------------------------------------------------------------------
function public.Build_GetNameByFragment ( fragment_name )

  local build_name = m_builds.Build_GetNameByFragment( fragment_name );
  
  return build_name;

end;
--------------------------------------------------------------------------------
function public.Build_GetDataByUnlocker ( stage_name, with_saved_data )

  local build_data = m_builds.Build_GetDataByUnlocker( stage_name, with_saved_data );
  
  return build_data;

end;
--------------------------------------------------------------------------------
function public.Build_GetDataByDeadlock ( deadlock )

  local build_data = m_builds.Build_GetDataByDeadlock( deadlock );
  
  return build_data;

end;
--------------------------------------------------------------------------------
function public.Build_GetUnlockList ( build_data, construct )
  
  local stages_list = {};

  local unlocked    = build_data.unlocked or false;
  local constructed = build_data.constructed or construct;
  local build_type  = build_data.type;

  if ( build_data.first ) then

    stages_list.stages_previous =
    { 
      build_data.unlocklist.min - 11,
      build_data.unlocklist.min - 1 
    };

  end;

  if ( build_type == "island" ) or ( build_type == "puzzle" ) then

    if ( constructed and build_data.unlocklist ) then

      stages_list.stages_type = build_data.unlocklist.type;
      stages_list.stages_next =
      { 
        build_data.unlocklist.min,
        build_data.unlocklist.max
      };

    end;

  elseif ( unlocked and constructed and build_data.unlocklist ) then

    stages_list.stages_type = build_data.unlocklist.type;
    stages_list.stages_next =
    { 
      build_data.unlocklist.min,
      build_data.unlocklist.max
    };

  end;

  -- Cheater stages displaying hook
  if not ( stages_list.stages_next ) then

    local unlock_min   = build_data.unlocklist.min;
    local unlock_max   = build_data.unlocklist.max;
    local unlock_type  = build_data.unlocklist.type;

    local stage_name = "stage_"..unlock_max;

    if ( unlock_type == "bonus" ) then
      stage_name = stage_name.."_bonus"
    end;

    local stage_info = v_stage.Stage_StorageData( stage_name );

    if ( stage_info and stage_info.enabled ) then
      stages_list.stages_type = unlock_type;
      stages_list.stages_next = { unlock_min, unlock_max };
    end;

  end;
  
  return stages_list;
  
end;
--------------------------------------------------------------------------------
function public.Build_IsConstructed ( build_name )

  local  constructed = m_builds.Build_IsConstructed( build_name );
  return constructed;

end;
--------------------------------------------------------------------------------
function public.Build_IsAutoConstructed ( build_name )
  
  local  is_autoconstructed = m_builds.Build_IsAutoConstructed( build_name );
  return is_autoconstructed;
  
end;
--------------------------------------------------------------------------------
function public.Build_HighlightStage ( build_name )

  local build_data = m_builds.Build_GetData( build_name );
  local fragments  = build_data.fragments;

  if ( not fragments ) then

    return false;

  end;

  -- Puzzle --------------------------------------------------------------------
  if ( build_data.type == "puzzle" ) then

    local fragment_name = fragments[ 1 ];

    private.Puzzle_HighlightStage( fragment_name );

    return;

  end;

  -- Island --------------------------------------------------------------------
  local current = {};
  
  for _, fragment_name in pairs( fragments ) do
  
    if ( v_quests.Quest_GetNameByItem( "fragment", fragment_name ) ) then
      
      table.insert( current, fragment_name );
    
    end;
    
  end;
  
  if ( #current > 0 ) then
    
    local fragment_id   = math.random( 1, #current );
    local fragment_name = current[ fragment_id ];
  
    v_map.Point_SuggestByReward( "fragment", fragment_name );
    
  else
  
    for pos, fragment_name in pairs( fragments ) do
  
      local quest_name = v_quests.Quest_GetNameByItem(
        "fragment", fragment_name
      );
      
      if ( quest_name                              ) and
         ( v_quests.Quest_IsComplete( quest_name ) ) then
        
         table.remove( fragments, pos );
      
      end;
      
    end;
    
    local fragment_id   = math.random( 1, #fragments );
    local fragment_name = fragments[ fragment_id ];
  
    v_map.Point_SuggestByReward( "fragment", fragment_name );
    
  end;

end;
--------------------------------------------------------------------------------
function public.Build_GetUnlocklistByMin ( min_stage )
  
  local  build_name = m_builds.Build_GetUnlocklistByMin( min_stage );
  return build_name;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Puzzle_HighlightStage ( fragment_name )

  local pattern = fragment_name:match( "^(.*)_part_%d+$" );
  
  if ( not pattern ) then return; end;
  
  v_map.Point_SuggestByReward( "fragment", pattern );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.AutoFix ( build_data )

  local build_name = build_data.name;
  local msg = build_name.." / ";
  local set_data = {};
  local stage_to_enable;
  
  local unlock_name = build_data.unlock;
  local min_stage_name = "stage_"..build_data.unlocklist.min;
  
  if ( build_data.unlocklist.type == "bonus" ) then
    min_stage_name = "stage_"..build_data.unlocklist.min.."_bonus";
  end;  

  -- Unlock build if condition met.
  if ( not build_data.unlocked and 
       unlock_name ) then
    
    local stage_completed = v_stage.Stage_IsCompleted( unlock_name );
    
    if ( stage_completed ) then
      
      msg = msg.."unlock by "..unlock_name.." / ";
      
      set_data.unlocked = true;

      if ( m_builds.Build_IsAutoConstructed( build_name, build_data ) ) then
        
        msg = msg.."autoconstruct / ";
        set_data.constructed = true;
        
      end;

      build_data.unlocked = set_data.unlocked;
      build_data.constructed = set_data.constructed;
      
    end;
    
  end;
  
  -- Construct build if min stage is enabled.
  if ( not build_data.constructed ) then
    
    if ( private.Autofix_IsBuildConstructAccepted( build_name, set_data ) ) then

      msg = msg.."construct by fragments / ";
      set_data.unlocked    = true;
      set_data.constructed = true;
      build_data.unlocked = set_data.unlocked;
      build_data.constructed = set_data.constructed;
      
    end;
    
    local min_stage_completed = v_stage.Stage_IsCompleted( min_stage_name ) or 
                                v_stage.Stage_IsEnabled( min_stage_name );
 
    
    local completed_unlock_level = true;
    if ( unlock_name ) then
      completed_unlock_level = v_stage.Stage_IsCompleted( unlock_name );
    end;
    
    if ( min_stage_completed and completed_unlock_level ) then
      
      msg = msg.."construct by min "..min_stage_name.." / ";

      set_data.unlocked    = true;
      set_data.constructed = true;
      set_data.collected_fragments = build_data.fragments;
      
      build_data.unlocked = set_data.unlocked;
      build_data.constructed = set_data.constructed;
      build_data.collected_fragments = set_data.collected_fragments;
      
    end;

  end;
  
  -- Enable stage if build is constructed.
  if ( build_data.constructed ) then
    
    local stage_enabled = v_stage.Stage_IsEnabled( min_stage_name );
    
    if ( not stage_enabled ) then
      
      msg = msg.."enable stage "..min_stage_name.." / ";
      stage_to_enable = min_stage_name;
      set_data.constructed = true;
      
    end;

  end;  
  
  if ( next( set_data ) ) then
    
    --private.AutoFix_TrySend();

    m_builds.Build_SetData( build_name, set_data );
    if ( stage_to_enable ) then
      v_stage.Stage_TryEnable( min_stage_name );
    end;
    
  end;
  
  return build_data;
  
end;

--------------------------------------------------------------------------------
function private.Autofix_IsBuildConstructAccepted ( build_name, set_data )
  
  local is_accepted, section, fragments =
    m_builds.Build_IsConstructAccepted( build_name, set_data );
    
  -- JK: additional check based on fragments
  if ( not is_accepted ) and
     ( fragments       ) then
    
    for _, fragment in pairs( fragments ) do
      
      local stages = v_stage.Stages_FindByReward(
        "fragment",
        fragment,
        section
      );
      
      if ( #stages == 0 ) then
        
        -- JK: search in quests
        local quest_name = v_quests.Quest_GetNameByItem(
          "fragment",
          fragment,
          true
        );
        
        local quest_done = false;
        
        if ( quest_name ) then
          quest_done = v_quests.Quest_IsComplete( quest_name );
        end;
        
        if ( not quest_name            ) or
           ( quest_name and quest_done ) then
        
          local item = {
            type  = "fragment",
            name  = fragment,
            count = 1
          };
          
          v_items.TryObtain( item );
        
        end;
        
      end;
      
    end;
    
  end;
  
  return is_accepted;
  
end;
--------------------------------------------------------------------------------
function private.AutoFix_TrySend ()

  if ( not private.autofix_send ) then
    
    v_game.AutoFix_Send( "builds" );
    private.autofix_send = true;
    
  end;

end;
--------------------------------------------------------------------------------