-- name=v_carriers
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_carriers",
  "int_dialog_carrier_main",
  "int_dialog_carrier_upgrade",
  "int_dialog_carrier_send",
  "int_dialog_carrier_confirm"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()
  
  private.this = {
    dialog = {
      main    = "int_dialog_carrier_main",
      upgrade = "int_dialog_carrier_upgrade",
      send    = "int_dialog_carrier_send",
      confirm = "int_dialog_carrier_confirm"
    }
  };
  
  common.mod.Load( _modules );

  mediator.Subscribe(
    "Model_Carriers_CarrierFinished",
    function ( params )

      private.OnFinished(
        params.carrier_name,
        params.resource_name,
        params.resource_count,
        params.section_number
      );

    end
  );
  
  private.created_carriers = {};

  --Log.i( "Initialized." );

end;
--------------------------------------------------------------------------------
-- function ...() end;
--------------------------------------------------------------------------------
function public.Init ()

  v_interface.window.RegisterAsModal( private.this.dialog.main );
  v_interface.window.RegisterAsModal( private.this.dialog.upgrade );
  v_interface.window.RegisterAsModal( private.this.dialog.confirm );

  -- JK: set delivery time once at init
  int_dialog_carrier_send.Init(
    private.DeliveryTime_GetString()
  );

  v_game.RegisterFeature( 
  { 
    entity = "v_carriers", 
    feature = "carrier",
    func_save = private.Feature_SaveData
  } );
  
end;
--------------------------------------------------------------------------------
function public.Cheat ()

  return m_carriers.Cheat();

end;
--------------------------------------------------------------------------------
function public.Create ()
  
  local section_number = v_sections.Section_NumberCurrent();

  local multiset = {};

  local carrier_name  = "carrier_"..section_number;
  local obj_name      = "obj_map_"..carrier_name;
  local anm_name      = "anm_map_"..carrier_name;
  local carrier_state = m_carriers.State( carrier_name );
  local carrier_busy  = m_carriers.IsBusy( carrier_name );
  
  ui.obj.Clone(
    "anm_map_carrier_platform",
    obj_name,
    "_"..section_number,
    1
  );
  
  ui.obj.Clone(
    "anm_map_carrier",
    obj_name,
    "_"..section_number,
    1
  );

  ui.obj.Set(
    obj_name,
    {
      inputrect_init = true,
      inputrect_x    = -100,
      inputrect_y    = -200,
      inputrect_w    =  200,
      inputrect_h    =  300
    },
    multiset
  );
    
  local animfunc = tern( ( carrier_state ) and ( not carrier_busy ) )(
    "state_"..tostring( carrier_state ),
    "ladder"
  );

  private.SetAppearance( carrier_name, animfunc );

  ui.control.EventsSet(
    obj_name,
    {
      func_on_click = function ()
        
        private.Tap( carrier_name );
        
      end
    },
    multiset
  );
  
  if ( public.IsBuilt( section_number ) ) then
  
    private.Indicator_Hide( section_number, multiset );
  
  end;
  
  if ( public.IsSuspended() ) then
    
    ui.obj.Set(
      obj_name,
      {
        active = false,
        input = false,
        visible = false
      },
      multiset
    );
    
  end;

  ui.obj.Multiset( multiset );
  
  table.insert( private.created_carriers, obj_name );
  
end;
--------------------------------------------------------------------------------
function public.OnIntroStart ()
  
  local section_number = v_sections.Section_NumberCurrent();

  private.Indicator_Hide( section_number );
  
end;
--------------------------------------------------------------------------------
function public.OnIntroFinish ()
    
  if ( public.IsSuspended() ) then

    return;

  end;

  local section_number = v_sections.Section_NumberCurrent();

  private.Indicator_Show( section_number );
  
end;
--------------------------------------------------------------------------------
function public.GetRandomEvent ()

  local random_event = m_carriers.GetRandomEvent();

  return random_event;

end;
--------------------------------------------------------------------------------
function public.Data ( carrier_name )

  return m_carriers.Data( carrier_name );

end;
--------------------------------------------------------------------------------
function public.GetCurrentMapObjectName ()

  local section_number = v_sections.Section_NumberCurrent();

  return "obj_map_carrier_"..section_number;

end;
--------------------------------------------------------------------------------
function public.Dialog_GetBuildButtonName ()

  return "btn_int_dialog_carrier_main_build";

end;
--------------------------------------------------------------------------------
function public.IsUnlocked ()

  local unlock      = m_carriers.GetUnlock();
  local is_unlocked = v_sections.Section_IsEnabled( unlock );

  return is_unlocked;

end;
--------------------------------------------------------------------------------
function public.IsSuspended ()

  if ( not public.IsUnlocked() ) then

    return true;

  end;

  local is_active = v_features.GetFeatureState( "carrier" );

  return ( not is_active );

end;
--------------------------------------------------------------------------------
function public.IsAvailable ()

  local is_suspended = public.IsSuspended();

  --- JK: check for suspended --------------------------------------------------
  if ( is_suspended ) then

    return false;

  end;

  --- JK: check for availability in current section ----------------------------
  local current    = v_sections.Section_NumberCurrent();
  local discovered = v_sections.DiscoveredCount();

  if ( current == discovered ) then

    return true;

  end;

  --- JK: check for legacy feature setting -------------------------------------
  local feature_data = private.Feature_GetData();

  if ( feature_data ) then

    return feature_data.legacy;

  end;

  return false;

end;
--------------------------------------------------------------------------------
function public.IsBuilt ( section_number )

  section_number = section_number or v_sections.Section_NumberCurrent();

  local carrier_name = "carrier_"..section_number;
  local carrier_data = m_carriers.Data( carrier_name );
  local is_built     = ( ( carrier_data ) and ( carrier_data.name ) );

  return is_built;

end;
--------------------------------------------------------------------------------
function public.GetNumberByName ( carrier_name )
  
  return tonumber( string.match( carrier_name, "%d+" ) );
  
end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function private.Feature_GetData ()

  local data, exists = v_features.GetFeatureData( "carrier" );

  if ( exists ) and ( data ) then
    return data;
  end;

  return false;

end;
--------------------------------------------------------------------------------
function private.Feature_SaveData ()
  
  v_features.Feature_SaveState( "carrier" );
  
  if ( v_features.IsActive( "carrier" ) ) then

    local data = parts.features.GetFeatureData( "carrier" ) or {};

    if ( data[ 1 ] ) then
      v_features.SaveFeatureData( "carrier", data[ 1 ] );
    end;

  end;
  
end;
--------------------------------------------------------------------------------
function public.FeatureUpdated ()
  
  local is_suspended = public.IsSuspended();
  local multiset = {};
  
  for _, obj_name in pairs( private.created_carriers ) do
  
    if ( is_suspended ) then
      ui.obj.Disable( obj_name, false, multiset );
    else
      ui.obj.Enable( obj_name, false, multiset );
    end;
  
  end;
  
  if ( #multiset > 0 ) then
    ui.obj.Multiset( multiset );
  end;

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function private.Dialog_Main (
    carrier_name,
    section_number,
    is_carrier_available
  )
  
  local delegates    = {};
  local tab_number   = 1;
  local tab1_variant = 1;
  local carrier_data = m_carriers.Data( carrier_name );
  local is_built     = public.IsBuilt();
  local is_available = is_carrier_available;
  local is_suspended = public.IsSuspended();
  
  v_interface.panel.top.Show();
  v_interface.panel.top.Lock();

  delegates.OnShow = function () end;

  if ( is_built     ) and
     ( is_available ) then
  
    if ( m_carriers.IsBusy( carrier_name ) ) then
      
      tab1_variant = 3;
      
      delegates.GetTab1Data = function ()
      
        return private.Data_Speedup( carrier_name );
        
      end;
      
    else
  
      -- JK: check if player has discovered at least 2 sections
      -- to be able to transfer resources from one to another
      if ( v_player.Progress_SectionCount() > 1 ) then
      
        tab1_variant = 2;
        
        delegates.GetTab1Data = function ()
      
          return private.Data_Send( carrier_name, section_number );
          
        end;
        
      else
        
        tab1_variant = 4;
        
        delegates.GetTab1Data = function ()
      
          return private.Data_Blocked( carrier_name, not is_available );
          
        end;
        
      end;
      
    end;
    
  elseif ( not is_built ) and ( not is_suspended ) then
    
    delegates.GetTab1Data = function ()
    
      return private.Data_Build( carrier_name );
      
    end;

  else

    tab1_variant = 4;
        
    delegates.GetTab1Data = function ()
  
      return private.Data_Blocked( carrier_name, not is_available );
      
    end;
      
  end;
  
  delegates.Close = function ()

    _G[ private.this.dialog.main ].Hide();
    
  end;

  delegates.OnClose = function ()

    v_interface.panel.top.Unlock();
    v_interface.window.Hide( private.this.dialog.main );
    
  end;
  
  v_interface.window.Show(
    private.this.dialog.main, 
    function ()
  
      int_dialog_carrier_main.Show( tab_number, tab1_variant );
      
    end,
    nil,
    delegates
  );
  
end;
--------------------------------------------------------------------------------
function private.Dialog_Upgrade ( carrier_name, carrier_state )
  
  local old_state = carrier_state;
  local new_state = carrier_state + 1;
  
  local old_capacity  = m_carriers.Capacity( carrier_name, old_state );
  local new_capacity  = m_carriers.Capacity( carrier_name, new_state );
  local upgrade_price = m_carriers.Price_Upgrade( carrier_name, new_state );
  
  local delegates = {};
  
  delegates.Accept = function ()
    
    private.TryUpgrade( carrier_name, old_state, new_state );
    
  end;
  
  delegates.OnOpen = function ()

    --

  end;

  delegates.Close = function ()

    _G[ private.this.dialog.upgrade ].Hide();
    
  end;
  
  delegates.OnClose = function ()

    v_interface.window.Hide( private.this.dialog.upgrade );
    
  end;
  
  v_interface.window.Show(
    private.this.dialog.upgrade, 
    function ()
  
      int_dialog_carrier_upgrade.Show(
        old_state,
        new_state,
        old_capacity,
        new_capacity,
        upgrade_price
      );
      
    end,
    nil,
    delegates
  );
  
end;
--------------------------------------------------------------------------------
function private.Dialog_Send ( carrier_name )
 
  v_interface.panel.top.Lock();
 
  local section_data  = private.GetSectionData();
  local delivery_time = private.DeliveryTime_GetString();

  local delegates = {};

  delegates.Select = function ( section_number )

    private.Dialog_Main( carrier_name, section_number, true );

    return true;

  end;
  
  delegates.Close = function ()

    _G[ private.this.dialog.send ].Hide();
        
  end;

  delegates.OnClose = function ()
          
    v_interface.panel.top.Unlock();
    v_interface.window.Hide( private.this.dialog.send );
        
  end;

  v_interface.window.Show(
    private.this.dialog.send, 
    function ()
  
      int_dialog_carrier_send.Show( section_data, delivery_time );
      
    end,
    nil,
    delegates
  );
   
end;
--------------------------------------------------------------------------------
function private.Dialog_Confirm ( 
    message_type, 
    resource_name, 
    resource_count, 
    func_execute )

  local delegates = {
    Accept = function ()

      return func_execute();

    end,
    Close = function ()

      _G[ private.this.dialog.confirm ].Hide();

    end,
    OnHide = function ()

      v_interface.window.Hide( private.this.dialog.confirm );

    end
  };

  v_interface.window.Show(
    private.this.dialog.confirm, 
    function ()
  
      int_dialog_carrier_confirm.Show( message_type, resource_name, resource_count );
      
    end,
    nil,
    delegates
  );
   
end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function private.Indicator_Show ( indicator_number, multiset )

  local object_name = "obj_map_carrier_indicator_"..indicator_number;

  if ( not ui.obj.Exists( object_name ) ) then return; end;

  ui.obj.Enable( object_name, false, multiset );
  
end;
--------------------------------------------------------------------------------
function private.Indicator_Hide ( indicator_number, multiset )

  local object_name = "obj_map_carrier_indicator_"..indicator_number;

  if ( not ui.obj.Exists( object_name ) ) then return; end;

  ui.obj.Disable( object_name, false, multiset );
  
end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function private.Tap ( carrier_name )
  
  local obj_name = "obj_map_"..carrier_name;
  
  ui.input.Off( obj_name );
  
  v_map.Focus(
    obj_name,
    {
      func_after = function ()

        private.OnTap( carrier_name );
        ui.input.On( obj_name );

      end
    }
  );

  local is_busy = m_carriers.IsBusy( carrier_name );

  if ( is_busy ) then return; end;

  local is_suspended = public.IsSuspended();

  if ( is_suspended ) then return; end;

  local carrier_state = m_carriers.State( carrier_name );

  if ( carrier_state ) then

    ui.anim.Play(
      "anm_map_"..carrier_name,
      "state_"..carrier_state.."_click",
      nil,
      "state_"..carrier_state
    );

  end;
  
end;
--------------------------------------------------------------------------------
function private.OnTap ( carrier_name )

  local is_available = public.IsAvailable();

  if ( not is_available ) then

    private.Dialog_Main( carrier_name, nil, is_available );
    return;

  end;

  local is_built      = public.IsBuilt();
  local is_busy       = m_carriers.IsBusy( carrier_name );
  local is_send_ready = ( ( is_built ) and ( not is_busy ) );

  local func_dialog = tern( is_send_ready )(
    private.Dialog_Send,
    private.Dialog_Main
  );

  func_dialog( carrier_name, nil, is_available );

end;
--------------------------------------------------------------------------------
function private.SetAppearance ( carrier_name, animfunc )
  
  ui.anim.Play( "anm_map_"..carrier_name, animfunc );
  
end;
--------------------------------------------------------------------------------
function private.TryBuild ( carrier_name )
  
  if ( private.is_building ) then return false; end;

  local carrier_data = m_carriers.Data( carrier_name );

  if ( carrier_data ) then

    -- JK: already built, do nothing
    return false;

  end;

  local build_price = m_carriers.Price_Build( carrier_name );
  local build_paid  = v_player.Gold_TryChangeCount( - build_price );
  
  if ( not build_paid ) then 

    return false;

  end;
  
  v_log.Kontagent_LogEvent( "economy",
  {
    description = "premium_currency_spend",
    category    = "carrier_build",
    value       = - build_price,
    name        = carrier_name
  } );

  v_log.Kontagent_LogEvent( "carrier",
  {
    name        = carrier_name,
    description = "build",
    level       = 1
  } );

  if ( m_carriers.Build( carrier_name ) ) then
    
    private.is_building = true;

    local delegates = {
      OnPostClose = function ()
    
        private.is_building = false;

        local obj_name         = "obj_map_"..carrier_name;
        local anm_name         = "anm_map_"..carrier_name;
        local indicator_number = public.GetNumberByName( carrier_name );
        
        ui.input.Off( obj_name );
        
        private.Indicator_Hide( indicator_number );
        
        ui.anim.Play(
          anm_name,
          "state_1_build",
          function ()

            ui.input.On( obj_name );
            
          end,
          "state_1"
        );
      
      end
    };

    common.delegate.Set( "int_dialog_carrier_main", delegates );

    int_dialog_carrier_main.Hide();

    return true;
    
  end;

  return false;
  
end;
--------------------------------------------------------------------------------
function private.TryUpgrade ( carrier_name, old_state, new_state )
  
  if ( private.is_upgrading ) then return false; end;

  local carrier_data = m_carriers.Data( carrier_name );
  
  if ( not carrier_data ) then
    
    -- JK: carrier is not built, can't upgrade
    return false;
    
  end;

  local upgrade_price = m_carriers.Price_Upgrade( carrier_name );
  local upgrade_paid  = v_player.Gold_TryChangeCount( - upgrade_price );
  
  if ( not upgrade_paid ) then 

    return;

  end;

  local new_state = carrier_data.state + 1;

  v_log.Kontagent_LogEvent( "carrier",
  {
    name        = carrier_name,
    description = "upgrade",
    level       = new_state
  } );

  if ( m_carriers.Upgrade( carrier_name, new_state ) ) then
    
    private.is_upgrading = true;

    local delegates = {

      OnPostClose = function()

        private.is_upgrading = false;

        local obj_name = "obj_map_"..carrier_name;
        local anm_name = "anm_map_"..carrier_name;
        
        ui.input.Off( obj_name );
        
        ui.anim.Play(
          anm_name,
          "state_"..old_state.."_upgrade",
          function ()
            
            ui.input.On( obj_name );
            
          end,
          "state_"..new_state
        );
        
      end
    };

    common.delegate.Set( "int_dialog_carrier_main", delegates );

    delegates = {

      OnClose = function()

        v_interface.window.Hide( private.this.dialog.upgrade );
        int_dialog_carrier_main.Hide();

      end;

    };

    common.delegate.Set( "int_dialog_carrier_upgrade", delegates );

    int_dialog_carrier_upgrade.Hide();

    return true;
    
  end;

  return false;
  
end;
--------------------------------------------------------------------------------
function private.TrySend (
    carrier_name,
    resource_name,
    resource_count,
    from_section_number,
    to_section_number
  )
 
  if ( private.is_sending ) then return false; end;

  local send_paid = v_player.Resource_TryChangeCount(
    resource_name,
    - resource_count,
    false,
    from_section_number
  );

  if ( not send_paid ) then

    return false;

  end;

  local delivery_time = private.DeliveryTime_GetSeconds();

  local sent = m_carriers.Send(
    carrier_name,
    resource_name,
    resource_count,
    from_section_number,
    to_section_number,
    delivery_time
  );

  v_log.Kontagent_LogEvent( "carrier",
  {
    name        = resource_name,
    category    = carrier_name,
    description = "use",
    value       = resource_count,
    level       = m_carriers.State( carrier_name ),
    from        = "section_"..from_section_number
  } );

  if ( not sent ) then

    return false;

  end;

  private.is_sending = true;

  local carrier_state = m_carriers.State( carrier_name );

  local delegates = {
    
    OnPostClose = function ()

      private.is_sending = false;

      local obj_name = "obj_map_"..carrier_name;
      local anm_name = "anm_map_"..carrier_name;
      
      ui.input.Off( obj_name );
      
      ui.anim.Play(
        anm_name,
        "state_"..carrier_state.."_disappear",
        function ()
          
          ui.input.On( obj_name );
          
        end,
        "ladder"
      );

    end
  };

  common.delegate.Set( "int_dialog_carrier_main", delegates );

  int_dialog_carrier_main.Hide();

  return true;
  
end;
--------------------------------------------------------------------------------
function private.TrySpeedUp ( carrier_name )

  if ( private.is_skipping ) then return false; end;

  if ( not m_carriers.IsScheduled( carrier_name ) ) then
    return false;
  end;
  
  local speedup_price = private.GetSpeedUpPrice( carrier_name );
  local speedup_paid  = v_player.Gold_TryChangeCount( - speedup_price )

  if ( not speedup_paid ) then

    return false;

  end;

  local speedup_result = m_carriers.SpeedUp( carrier_name, true );
  
  if ( speedup_result ) then

    private.is_skipping = true;

    v_log.Kontagent_LogEvent( "economy", 
    {
      description = "premium_currency_spend",
      category    = "carrier_speedup",
      value       = - speedup_price,
      name        = carrier_name
    } );

    ui.screen.Lock( true );

    int_dialog_carrier_main.Speedup_OnAccept(
      function ()

        private.is_skipping = false;

        ui.screen.Lock( false );

        private.OnFinished(
          speedup_result.carrier_name,
          speedup_result.resource_name,
          speedup_result.resource_count,
          speedup_result.section_number
        );

      end
    );

    return true;

  end;
  
  return false;

end;
--------------------------------------------------------------------------------
function private.OnFinished (
    carrier_name,
    resource_name,
    resource_count,
    from_section_number
  )

  local current_section_number = v_sections.Section_NumberCurrent();
  local carrier_number         = public.GetNumberByName( carrier_name );
  local to_section_number      = carrier_number;
  local carrier_state          = m_carriers.State( carrier_name );
  local is_section_created     = v_map.IsSectionCreated( current_section_number );
  local silent_enrollment      = ( current_section_number ~= to_section_number );

  v_player.Resource_TryChangeCount(
    resource_name,
    resource_count,
    silent_enrollment,
    to_section_number
  );

  local func_notification = function ()

    local section_name = ne.strings.Get(
      "str_int_dialog_worldmap_section_"..to_section_number
    );
    
    local text = section_name..": +"..resource_count;
    local icon = resource_name;
    
    v_interface.ShowNotification( text, icon );

  end;

  local function func_arrival ()

    -- JK: if player is in other section than carrier origin
    if ( current_section_number ~= carrier_number ) then

      ui.input.On( "obj_map_"..carrier_name );

      private.SetAppearance( carrier_name, "state_"..carrier_state );

      return;

    end;

    local obj_name = "obj_map_"..carrier_name;
    local anm_name = "anm_map_"..carrier_name;

    ui.input.Off( obj_name );
    
    ui.anim.Play(
      anm_name,
      "state_"..carrier_state.."_appear",
      function ()
        
        private.SetAppearance( carrier_name, "state_"..carrier_state );
        
        ui.input.On( obj_name );
        
      end
    );

  end;

  if ( not is_section_created ) then

    mediator.SubscribeOnce(
      "View_Map_Update_Finished",
      func_notification
    );

    return;

  end;

  if ( v_interface.window.IsOnScreen( private.this.dialog.main ) ) then
    int_dialog_carrier_main.Hide();
  end;
  
  if ( v_interface.window.IsOnScreen( private.this.dialog.confirm ) ) then
    int_dialog_carrier_confirm.Hide();
  end; 

  func_notification();
  func_arrival();

end;
--------------------------------------------------------------------------------
function private.DeliveryTime_GetString ()

  local seconds = private.DeliveryTime_GetSeconds();
  local string  = ui.text.FormatTime( seconds );

  return string;

end;
--------------------------------------------------------------------------------
function private.DeliveryTime_GetSeconds ()

  local data = private.Feature_GetData();
  
  if ( data and data.time ) then
    return data.time;
  end;
  
  local  default_time = m_carriers.GetDefaultDeliveryTime();
  return default_time;

end;
--------------------------------------------------------------------------------
function private.RemainingTime_GetString ( carrier_name )

  local seconds = m_carriers.GetRemainingTime( carrier_name );
  local string  = ui.text.FormatTime( seconds );

  return string;

end;
--------------------------------------------------------------------------------
function private.GetSectionData ()

  local section_data    = {};  
  local sections   = v_sections.DetailedList();
  local section_current = v_sections.Section_NumberCurrent();
  local prefix = "str_int_dialog_worldmap_section_";

  for section_number = 1, #sections do 

    if ( section_number ~= section_current ) then

      local section_resources = private.GetSectionResource( section_number );

      table.insert(
        section_data,
        {
          section_number = section_number,
          section_string = ne.strings.Get( prefix..section_number ),
          resource1      = section_resources.resource1,
          resource2      = section_resources.resource2,
          resource3      = section_resources.resource3,
          color          = v_sections.Section_StaticData( "section_"..section_number ).color,
          unlock         = sections[ section_number ].enabled
        }
      );

    end;
    
  end;

  return section_data;

end;
--------------------------------------------------------------------------------
function private.GetSectionResource ( section_number, resource_name )

  if ( resource_name ) then

    return v_player.Resource_GetCount( resource_name, section_number );

  end;

  local resource1 = v_player.Resource_GetCount( "resource1", section_number );
  local resource2 = v_player.Resource_GetCount( "resource2", section_number );
  local resource3 = v_player.Resource_GetCount( "resource3", section_number );

  return {
    resource1 = tern( ( resource1 ) and ( resource1 > 0 ) )( resource1, 0 );
    resource2 = tern( ( resource2 ) and ( resource2 > 0 ) )( resource2, 0 );
    resource3 = tern( ( resource3 ) and ( resource3 > 0 ) )( resource3, 0 );
  };

end;
--------------------------------------------------------------------------------
function private.GetSpeedUpPrice ( carrier_name )

  local data = private.Feature_GetData();
  
  if ( data and data.skip_price ) then
    return data.skip_price;
  end;

  local  skip_price = m_carriers.Price_SpeedUp( carrier_name );
  return skip_price;

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function private.Data_Build ( carrier_name )

  common.delegate.Set(
    "int_dialog_carrier_main",
    {
      Build = function () return private.TryBuild( carrier_name ); end
    }
  );

  return {
    price = m_carriers.Price_Build( carrier_name ),
  };
  
end;
--------------------------------------------------------------------------------
function private.Data_Blocked ( carrier_name, is_suspended )
  
  return {
    carrier_state = m_carriers.State( carrier_name ),
    has_carrier   = true,
    is_suspended  = is_suspended
  };
  
end;
--------------------------------------------------------------------------------
function private.Data_Send ( carrier_name, from_section_number )
  
  local carrier_state       = m_carriers.State( carrier_name );
  local section_resources   = private.GetSectionResource( from_section_number );
  local available_resources = {};
  
  for r = 1, 3 do
    
    local resource_name  = "resource"..r;
    local resource_count = section_resources[ resource_name ];

    if ( resource_count > 0 ) then
          
      table.insert( available_resources, resource_name );

    end;
    
  end;

  local delegates = {

    Upgrade = function ()
      
      if ( carrier_state < 6 ) then
      
        private.Dialog_Upgrade( carrier_name, carrier_state );
      
      end;
      
    end,

    Send = function ( resource_name )

      local available_count = private.GetSectionResource(
        from_section_number,
        resource_name
      );

      local carrier_capacity  = m_carriers.Capacity( carrier_name );
      local resource_count    = math.min( carrier_capacity, available_count );
      local to_section_number = v_sections.Section_NumberCurrent();
      local func_TrySend = function ()
        return private.TrySend(
          carrier_name,
          resource_name,
          resource_count,
          from_section_number,
          to_section_number
        );
      end;
      
      if ( available_count < carrier_capacity ) then
        private.Dialog_Confirm( 
          "send", 
          resource_name, 
          resource_count, 
          func_TrySend
        );
        return false;
      end;

      return func_TrySend();
    
    end

  };

  common.delegate.Set( "int_dialog_carrier_main", delegates );
  
  return {
    carrier_state = carrier_state,
    capacity      = m_carriers.Capacity( carrier_name ),
    resources     = available_resources
  };
  
end;
--------------------------------------------------------------------------------
function private.Data_Speedup ( carrier_name )
  
  local carrier_data   = m_carriers.Data( carrier_name );
  local carrier_origin = "section_"..public.GetNumberByName( carrier_name );
  local skip_price     = private.GetSpeedUpPrice( carrier_name );

  local function get_time_sec ()
    local  seconds = m_carriers.GetRemainingTime( carrier_name );
    return seconds;
  end;

  local function get_time_str ()
    local  time_string = private.RemainingTime_GetString( carrier_name );
    return time_string;
  end;

  local function get_frame ()
    
    local delivery_time  = carrier_data.resource.time;
    local remaining_time = get_time_sec();
    local progress_ratio = 1 - ( remaining_time / delivery_time );
    local progress_ratio = math.max( 0, progress_ratio );
    local progress_frame = math.floor( 120 * progress_ratio );

    return progress_frame;

  end;

  local delegates = {

    Speedup = function ()

      local func_TrySpeedUp = function ()
        return private.TrySpeedUp( carrier_name );
      end;
      
      private.Dialog_Confirm( 
        "speedup", 
        carrier_data.resource.name,
        carrier_data.resource.count,
        func_TrySpeedUp
      );

    end,

    SpeedupGetFrame   = get_frame,
    SpeedupGetTimeSec = get_time_sec,
    SpeedupGetTimeStr = get_time_str,
  };

  common.delegate.Set( "int_dialog_carrier_main", delegates );

  return {
    carrier_state  = carrier_data.state,
    resource_name  = carrier_data.resource.name,
    resource_count = carrier_data.resource.count,
    origin         = "section_"..carrier_data.resource.section,
    destination    = carrier_origin,
    time           = get_time_str(),
    frame          = get_frame(),
    skip_price     = skip_price
  };
  
end;
--------------------------------------------------------------------------------