-- name=v_chests
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_chests",
  "int_dialog_map_chest"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.this = {
    dialog = "int_dialog_map_chest"
  };

  common.mod.Load( _modules );

  private.app_w            = ui.screen.Width();
  private.chests_available = {};
  
  --Log.i( "Initialized." );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.List_OnSection ( section_name )

  local chests = m_chests.List_OnSection( section_name );
  
  return chests;

end;
--------------------------------------------------------------------------------
function public.Chest_Will ( stage_name )
  
  local chest_name = m_chests.Chest_NameByStage( stage_name );
  
  if ( not chest_name ) then
    return false;
  end;
  
  local state = m_chests.Chest_Status( chest_name );
  
  if ( state ~= "opened" ) then
    return true;
  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Create ( chest_name, section_name )
  
  private.Autofix( chest_name );

  local obj_name = "obj_map_chests_"..chest_name;
  local btn_name = "btn_map_chests_"..chest_name;
  
  ObjCreate( obj_name, "obj" );
  ObjSet(    obj_name, { anim_tag = chest_name } );
  ObjAttach( obj_name, "anm_map_"..section_name );

  ObjClone( "btn_map_chests", "_"..chest_name, 1, obj_name );
  
  ui.control.EventsSet( btn_name,
    {
      func_on_click = function () private.Chest_Open( chest_name ); end;
    }
  );
  
  private.Chest_SetHidden( chest_name );
  
end;
--------------------------------------------------------------------------------
function private.Chest_Open ( chest_name, func_after )
  
  local state = private.Chest_State( chest_name );
  
  if ( not state.ready ) then
    
    local section = v_sections.Section_NumberCurrent();
    
    v_map.ShowObjectPopup(
      "monolog",
      "obj_map_chests_indicator_"..chest_name,
      ne.strings.Get( "str_int_popup_cloud_chests", state.value - 110 * ( section - 1 ) ),
      nil,
      nil,
      nil
    );

    return;
    
  end;
  
  ui.input.Off( "btn_map_chests_"..chest_name );
  mediator.Notify( "View_Chests_ChestCollected", chest_name );

  local gift = private.Chest_Content( chest_name );
  v_items.TryObtain( gift );
  
  if ( gift.type ) and ( gift.type == "booster" ) then
    gift.phase = v_boosters.GetPowerPhase( gift.name );
  end;
  
  m_chests.Chest_Open( chest_name );

  local func_after_hide = function ()

    private.Dialog_Show( chest_name, gift,
      function ()
      
        ui.screen.Lock( false );
        
      end,
      function ()
      
        v_items.Reward_Show(
          gift,
          function ()
      
            ObjSet( "obj_map_chests_"..chest_name, { active = true } );
            ui.anim.Play(
              "anm_map_chests_paw_"..chest_name,
              "dynamic",
              function ()
                private.Chest_SetOpened( chest_name );
              end
            );
            
            if ( func_after ) then func_after(); end;
      
          end
        );
        
      end
    );

  end;
  
  ui.screen.Lock( true );
   common.sound.Play( "button_click" );
  private.Chest_Hide( chest_name, func_after_hide );

end
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.TryUnlockByStage ( stage_name )

  local chest_name = m_chests.Chest_NameByStage( stage_name );

  -- JK: if nothing found to unlock
  if ( not chest_name ) then return; end;

  local chest_status = m_chests.Chest_Status( chest_name );

  -- JK: check if already unlocked
  if ( chest_status ~= "not_available" ) then return; end;

  m_chests.Chest_Unlock( chest_name );

  if ( private.Chest_Content( chest_name ).type == "booster" ) then

    local chest_content = private.Chest_Content( chest_name );

    v_boosters.Unlock( chest_content.name );

  end;

end;
--------------------------------------------------------------------------------
function public.Callback_AnimatePlace ( stage, fast_saga )

  local chest_name = m_chests.Chest_NameByStage( stage );
  
  if ( chest_name ) then

    if ( m_chests.Chest_Status( chest_name ) == "opened" ) then return nil; end;
  
    return function ( func_after )
    
      ui.screen.Lock( true );
      
      if ( fast_saga ) then
        
        private.Chest_SetActive( chest_name );
        private.Chest_Show( chest_name, function ()
          ui.screen.Lock( false );
          private.Chest_Open( chest_name, func_after );
        end );
        
      else
        
        private.Fox_Show( chest_name, function ()
      
          ui.screen.Lock( false );
          private.Chest_Open( chest_name, func_after );
      
        end );
        
      end;
    
    end;
  
  end;
  
  return nil;

end;
--------------------------------------------------------------------------------
function public.Callback_AnimateShow ( stage )

  local chest_name = m_chests.Chest_NameByStage( stage );
  if ( chest_name ) then
  
    return function ( func_after )
    
      private.Chest_SetActive( chest_name );
      private.Chest_Show( chest_name, func_after );
      
    end;
  
  end;
  
  return nil;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Chest_SetOnMap ( stage_name )
  
  local chest_name = m_chests.Chest_NameByStage( stage_name );
  
  if ( not chest_name ) then
    return;
  end;
  
  local state = m_chests.Chest_Status( chest_name );

  if ( state == "opened" ) then 
    private.Chest_SetOpened( chest_name );
  else
    private.Chest_SetActive( chest_name );
  end;
  
end;
--------------------------------------------------------------------------------
function private.Chest_SetActive ( chest_name )

  local btn_gift = "btn_map_chests_"..chest_name;
  local anm_gift = "anm_map_chests_gift_"..chest_name;
  local anm_paw  = "anm_map_chests_paw_"..chest_name;
  
  local set = {};
  
  ui.obj.Set( btn_gift, { input = true }, set );
  
  ui.obj.Set( anm_gift,
    {
      visible  = true,
      animfunc = "static",
      playing  = true
    }, 
    set
  );
  
  ui.obj.Set( anm_paw,
    { 
      visible  = true,
      animfunc = "static",
      playing  = true
    },
    set
  );

  ui.obj.Multiset( set );

end;
--------------------------------------------------------------------------------
function private.Chest_SetOpened ( chest_name )

  local btn_gift = "btn_map_chests_"..chest_name;
  local anm_gift = "anm_map_chests_gift_"..chest_name;
  local anm_paw  = "anm_map_chests_paw_"..chest_name;
  
  local set = {};
  
  ui.obj.Set( btn_gift, { input   = false                  }, set );
  ui.obj.Set( anm_gift, { visible = false, playing = false }, set );
  ui.obj.Set( anm_paw,  { visible = false, playing = false }, set );

  ui.obj.Multiset( set );

end;
--------------------------------------------------------------------------------
function private.Chest_SetHidden ( chest_name )

  local btn_gift = "btn_map_chests_"..chest_name;
  local anm_gift = "anm_map_chests_gift_"..chest_name;
  local anm_paw  = "anm_map_chests_paw_"..chest_name;
  
  local set = {};
  
  ui.obj.Set( btn_gift, { input   = false                  }, set );
  ui.obj.Set( anm_gift, { visible = false, playing = false }, set );
  ui.obj.Set( anm_paw,
    {
      visible  = true,
      playing  = true,
      animfunc = "static"
    },
    set
  );

  ui.obj.Multiset( set );

end;
--------------------------------------------------------------------------------
function private.Chest_Show ( chest_name, func_after )

  local anm_gift = "anm_map_chests_gift_"..chest_name;

  ObjSet( "obj_map_chests_"..chest_name, { active = true } );
  ui.anim.Play( anm_gift, "appear", func_after, "static" );

end;
--------------------------------------------------------------------------------
function private.Chest_Hide ( chest_name, func_after )

  common.sound.Play( "click_present" );
  ObjSet( "obj_map_chests_"..chest_name, { active = true } );
  ui.anim.Play( "anm_map_chests_gift_"..chest_name, "get", func_after );

end;
--------------------------------------------------------------------------------
function private.Chest_Content ( chest_name )

  local chest_data = m_chests.Chest_Data( chest_name );

  if ( chest_data         ) and
     ( chest_data.content ) then

    local fox_bonus = v_player.Stats_GetPlayerData().fox;
    chest_data.content.count    = chest_data.content.count + fox_bonus;
    chest_data.content.category = "gift";

    return chest_data.content;

  end;

  return false;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Dialog_Show ( chest_name, gift, func_open, func_hide )

  local delegates = {

    Open  = func_open,
    Close = function () private.Dialog_Hide( func_hide ); end

  };

  common.delegate.Set( private.this.dialog, delegates );

  v_interface.window.AddToQueue(
    private.this.dialog,
    function ()

      int_dialog_map_chest.Show( chest_name, gift );

    end
  );

end;
--------------------------------------------------------------------------------
function private.Dialog_Hide ( func_OnHide )

  int_dialog_map_chest.Hide(
    func_OnHide,
    function ()

      v_interface.window.Hide( private.this.dialog );

    end
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Fox_Show ( chest_name, func_after )

  local obj_chest = "obj_map_chests_"..chest_name;

   common.sound.Play( "flying_fox" );

  private.ZoomFocus( obj_chest, "out", 1.0, { x = 0, y = 200 } );

  ObjSet( "obj_map_foreground_fox", { visible = true, active = true } );
  ObjAttach( "obj_map_foreground", InterfaceName );

  local func_end = function()

    ObjSet( "obj_map_foreground_fox", { visible = false, active = false } );
    ObjAttach( "obj_map_foreground", "obj_map_body" );

  end;

  ui.anim.Play( "anm_map_foreground_fox_actor", "action", func_end );
  ui.anim.Play( "anm_map_foreground_fox_dummy", "dummy",  function()

    private.Fox_Sparks( chest_name, func_after );

  end );
  
  common.Defer( "foreground_fox", 1.3, function () 
  
    private.ZoomFocus( obj_chest, "in", 2.8, { x = 0, y = 200 } );
  
  end );

end;
--------------------------------------------------------------------------------
function private.Fox_Blast ( chest_name, pos )

  local anm_blast  = "anm_map_foreground_fox_blast";
  local anm_sparks = "anm_map_foreground_fox_sparks";

  ui.obj.Set( anm_blast,  { pos_x = pos[ 1 ], pos_y = pos[ 2 ]            } );
  ui.obj.Set( anm_sparks, { pos_x = 0.5 * private.app_w - 50, pos_y = 530 } );

  ui.anim.Play( anm_blast, "blast" );

end;
--------------------------------------------------------------------------------
function private.Fox_Sparks ( chest_name, func_after )

  ui.anim.Play( "anm_map_foreground_fox_sparks", "sparks" );

  local chest_pos = ui.obj.Position( "btn_map_chests_"..chest_name );

  ObjAnimate( "anm_map_foreground_fox_sparks", "pos_xy", 0, 0, function()

    private.Fox_Blast( chest_name, chest_pos );
    private.ZoomFocus( "obj_map_chests_"..chest_name, "in", 1.0 );
    
    private.Chest_SetActive( chest_name );
    private.Chest_Show( chest_name, func_after );

  end,
  {
    0, 0, 0.5 * private.app_w - 50, 530,
    1, 0, chest_pos[ 1 ], chest_pos[ 2 ]
  } );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.ZoomFocus( object, zoom, time, offset )

  time   = time or 0.6;

  if ( zoom == "in" ) then

    v_map.Focus( object, { scale = 1.0, time = time, offset = offset } );

  elseif ( zoom == "out" ) then

    v_map.Focus( object, { scale = 0.5, time = time, offset = offset } );

  end;

end;
--------------------------------------------------------------------------------
function public.HideChestsInSection ( section_name )
  
  local chests = m_chests.List_OnSection( section_name );
  
  for i = 1, #chests do
    private.Chest_SetHidden( chests[ i ].name );
  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Chest_State ( chest_name )

  local stage_name = m_chests.Chest_Data( chest_name ).stage;

  return { 
    ready = v_stage.Stage_IsCompleted( stage_name ), 
    value = tonumber( string.match( stage_name, "%d+" ) )
  };

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Autofix ( chest_name )

  local chest_state = private.Chest_State( chest_name )
  
  if ( chest_state.ready and
       m_chests.Chest_Status( chest_name ) == "not_available" ) then

    m_chests.Chest_Unlock( chest_name );

    if ( private.Chest_Content( chest_name ).type == "booster" ) then

      local chest_content = private.Chest_Content( chest_name );

      v_boosters.Unlock( chest_content.name );

    end;
    
  end; 

end;
--------------------------------------------------------------------------------