-- name=v_dailyquest
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_dailyquest",
  "int_dialog_dailyquest"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  common.mod.Load( _modules );
  
  private.dailyquest_data = {};
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Init ()

  private.PanelButton_Init();

  mediator.Subscribe(
    "Model_DailyQuest_ChallengeCooldownEnded",
    function () private.Cooldown_OnFinished( "challenge" ); end
  );

  mediator.Subscribe(
    "Model_Stage_StageWon",
    private.DailyQuest_TryStart
  );

  ne.msg.Subscribe(
    "Event_DailyQuest_Refresh",
    private.Refresh
  );

  private.challenge_states = m_dailyquest.GetStates();
  private.challenge_count  = m_dailyquest.GetChallengeCount();

  --ES> if need force refresh data or not data 
  if ( m_dailyquest.Data_Get( "refresh" ) ) then
    m_dailyquest.Data_Set( { refresh = false } );
    private.Refresh();
  end;
  
  v_game.RegisterFeature( 
  { 
    entity = "v_dailyquest", 
    feature = "dailyquest",
    func_save = private.Feature_SaveData
  } );
  
  if ( v_game.IsOnline() ) then
    private.PanelButton_Enable();
  end;
  
  local is_unlocked = private.IsUnlocked();
  
  m_dailyquest.Setup( is_unlocked );
  
  --Log.i( "Initialized." );

end;
--------------------------------------------------------------------------------
function public.Cheater_Refresh ()

  if ( m_dailyquest.Challenge_IsSheduled() ) then

    m_dailyquest.Unschedule( "challenge" );
    
    return;

  end;

  if not ( m_dailyquest.IsReady() ) then

    private.Refresh();

  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.OnMapUpdate ()

  return private.OnMapUpdate;

end;
--------------------------------------------------------------------------------
function public.IsReady ()

  local  is_ready = m_dailyquest.IsReady();
  return is_ready;

end;
--------------------------------------------------------------------------------
function public.IsSuspended ()

  if ( not private.IsUnlocked() ) then

    return true;

  end;

  local is_active = private.Feature_GetState();

  return ( not is_active );

end;
--------------------------------------------------------------------------------
function public.GetRemainingTime ( timer_type )

  local  remaining_time = m_dailyquest.GetRemainingTime( timer_type );
  return remaining_time;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Countdown_IsActive ()

  local is_active = m_dailyquest.Countdown_IsActive();

  return is_active;

end;
--------------------------------------------------------------------------------
function public.Countdown_IsNeeded ()

  return ( not private.IsUnlocked() );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.PanelButton_Init ()

  private.PanelButton_Enable();
  
  mediator.Subscribe(
    "Model_DailyQuest_EventScheduled",
    private.PanelButton_TryEnableTimer
  );
  
  mediator.Subscribe(
    "Model_DailyQuest_ChallengeCooldownEnded",
    private.PanelButton_SetNotifications
  );
  
  mediator.Subscribe(
    "Model_DailyQuest_CountdownEnded",
    private.PanelButton_SetNotifications
  );

end;
--------------------------------------------------------------------------------
function private.PanelButton_Enable ()

  if ( public.IsSuspended() ) then
    private.PanelButton_Disable();
    return;
  end;
  
  v_interface.panel.left.SetDelegates( {
    Arena_OnTap = function ()
      private.Dialog_Show();
      v_log.Playhaven_LogPlacement( "arena_open" );
    end
  } );

  v_interface.Scheme_AddButton( "map", "left", { name = "dailyquest" } );

  private.PanelButton_TryEnableTimer();

end;
--------------------------------------------------------------------------------
function private.PanelButton_TryEnableTimer ()

  local func_seconds = private.PanelButton_GetCooldown;
  v_interface.panel.left.DailyquestEnableTimer( func_seconds );
  private.PanelButton_SetNotifications();

end;
--------------------------------------------------------------------------------
function private.PanelButton_Disable ()

  v_interface.Scheme_DeleteButton( "map", "left", { name = "dailyquest" } );

end;
--------------------------------------------------------------------------------
function private.PanelButton_SetNotifications ()

  local cooldown = private.PanelButton_GetCooldown();
  local count = 0;
  
  if ( cooldown == 0 ) then
    count = public.Challenge_GetOpenedCount();
  end;
  
  local data    = private.dailyquest_data;
  local reward  = data.reward_id or m_dailyquest.Data_Get().reward;
  
  if ( reward ) then
    count = 1;
  end;

  v_interface.panel.left.DailyquestSetNotifications( count );
  
end;
--------------------------------------------------------------------------------
function private.PanelButton_GetCooldown ()

  if ( m_dailyquest.Countdown_IsActive() ) then
  
    return m_dailyquest.GetRemainingTime( "countdown" ) or 0;

  elseif ( not m_dailyquest.IsReady()              ) and
         ( not m_dailyquest.Challenge_IsSheduled() ) then
    
    return m_dailyquest.GetRemainingTime( "dailyquest" ) or 0;

  elseif ( m_dailyquest.Challenge_IsSheduled() ) then

    return m_dailyquest.GetRemainingTime( "challenge" ) or 0;

  end;
  
  return 0;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Dialog_Show ( winned_stage_name, func_after )
  
  private.func_ContinueMapUpdate = func_after;
  
  private.Update();
  
  if ( public.IsSuspended() and ( not private.dailyquest_data.reward_id ) ) then
    private.Dialog_FeatureDisable( func_after );
    return;
  end;

  local dailyquest = private.dailyquest_data;
  local delegates  = {};

  delegates.RewardTap    = private.Reward_Tap;
  delegates.RewardNotify = function ( challenge_num )

    mediator.Notify( "View_DailyQuest_RewardShown", challenge_num );

  end;

  delegates.ChallengeTap = function ( challenge_id )

    if ( public.IsSuspended() ) then
      private.Dialog_FeatureDisable();
      return;
    end;
    
    local challenge = dailyquest[ "challenge"..challenge_id ];

    -- JK: open
    if ( challenge.state == private.challenge_states[ 1 ] ) then
      
      if ( v_stage.IsCanPlay() ) then
        
        private.Dialog_Hide(
        function ()
          v_stage.TryStart( challenge.stage, challenge.enemy );
          private.Dialog_Wait( challenge_id, challenge.stage );
        end );
        
      end;

    -- JK: beaten
    elseif ( challenge.state == private.challenge_states[ 2 ] ) then

      --

    end;

  end;

  delegates.CooldownGetTime = function ( timer_type )

    local remaining_time = m_dailyquest.GetRemainingTime( timer_type );

    return remaining_time;

  end;
  
  delegates.CooldownSkip = function ( timer_type )

    private.Cooldown_Skip( timer_type );

  end;

  delegates.Close = function () 
    private.Dialog_Hide( func_after );
    v_log.Playhaven_LogPlacement( "arena_close" );
  end;

  local func_AnimateChallenge = function ( challenge_id )    

    local old_dailyquest = private.dailyquest_data;

    private.Update();

    local new_dailyquest = private.dailyquest_data;
    
    int_dialog_dailyquest.Challenge_Set( 
      challenge_id, 
      new_dailyquest,
      old_dailyquest
    );

  end;

  local func_OnStageComplete = function ( stage_name )

    local challenge_num;

    for num = 1, private.challenge_count do

      if ( dailyquest[ "challenge"..num ].stage == stage_name ) then
        challenge_num = num;
        break;
      end;
    end;

    if ( challenge_num ) then

      func_AnimateChallenge( challenge_num, private.challenge_states[ 2 ] );

    end;

  end;
  
  local func_OnShow = function ()

    private.dialog_active = true;
  
    int_dialog_dailyquest.Show(
      private.challenge_states,
      dailyquest
    );

    if ( winned_stage_name ) then

      func_OnStageComplete( winned_stage_name );

    end;
    
  end;
  
  v_interface.panel.top.Lock();
  
  if ( not v_interface.panel.top.IsVisible() ) then
  
    v_interface.panel.top.Show();
    v_interface.panel.top.Unlock();
    
  end;
    
  v_interface.window.Show(
    "int_dialog_dailyquest",
    func_OnShow,
    nil,
    delegates
  );

end;
--------------------------------------------------------------------------------
function private.Dialog_Hide ( func_end )
  
  private.wait_map_opened = nil;
  
  v_interface.panel.top.Hide( true );
      
  int_dialog_dailyquest.Hide( function ()

    private.dialog_active = false;

    v_interface.window.Hide( "int_dialog_dailyquest" );

    if ( func_end ) then
   
      func_end();
    
    end;

  end );

end;
--------------------------------------------------------------------------------
function private.Dialog_Wait ( challenge_id, stage_name )
  
  local func_OnStageWon;
  local func_OnStageFailed;
  local func_OnStageAbandon;
  
  local func_OnMapUpdate = function ( stage_name )
    if not ( private.wait_map_opened ) then
      private.OnMapUpdate = function ( func_after )
      
        private.wait_map_opened = nil;
        mediator.Unsubscribe( "View_Stage_StageFailed",  func_OnStageFailed );
        mediator.Unsubscribe( "View_Stage_StageAbandon", func_OnStageAbandon );
        mediator.Unsubscribe( "Model_Stage_StageWon", func_OnStageWon );
        private.Dialog_Show( stage_name, func_after );
        private.OnMapUpdate = nil;

      end;
    end;
    private.wait_map_opened = true;
  end;
  
  func_OnStageFailed = function ()

    mediator.Unsubscribe( "View_Stage_StageFailed", func_OnStageFailed );
    v_log.Kontagent_LogEvent( "arena", 
    {
      name  = "fail",
      level = ( m_dailyquest.GetChallengeCount() - 
                public.Challenge_GetOpenedCount() + 1 )
    } );
    func_OnMapUpdate();

  end;

  func_OnStageAbandon = function ()

    mediator.Unsubscribe( "View_Stage_StageAbandon", func_OnStageAbandon );
    v_log.Kontagent_LogEvent( "arena",
    {
      name  = "abandon",
      level = ( m_dailyquest.GetChallengeCount() -
                public.Challenge_GetOpenedCount() + 1 )
    } );
    func_OnMapUpdate();

  end;

  func_OnStageWon = function ( stage_data )

    if ( stage_name ~= stage_data.name ) then

      return;

    end;

    mediator.Unsubscribe( "Model_Stage_StageWon", func_OnStageWon );

    func_OnMapUpdate( stage_data.name );

    private.Challenge_SetBeaten( challenge_id );
    private.Reward_Activate();

    int_dialog_dailyquest.Reward_Enable();

  end;

  mediator.Subscribe( "Model_Stage_StageWon", func_OnStageWon );
  mediator.Subscribe( "View_Stage_StageFailed", func_OnStageFailed );
  mediator.Subscribe( "View_Stage_StageAbandon", func_OnStageAbandon );

end;
--------------------------------------------------------------------------------
function private.Dialog_FeatureDisable ( func_after )
  
  if ( not int_dialog_dailyquest.IsActive() ) then
    if ( func_after ) then func_after(); end;
    return;
  end;
  
  local func_hide = function ()
    private.Dialog_Hide( func_after );
  end;
  
  v_game.dialog.Info_Show(
    "dailyquest_feature_disabled",
    func_hide,
    func_hide
  );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.IsUnlocked ()
  
  local stage_name = m_dailyquest.GetUnlockStage();
  
  return v_stage.Stage_StorageData( stage_name ).completed;
  
end;
--------------------------------------------------------------------------------
function private.Start ( enemies_packages )

  local stage_ban_list = {};
  local enemy_ban_list = {};

  local dailyquest = {
    challenge1 = private.Challenge_Create( enemies_packages, stage_ban_list, enemy_ban_list ),
    challenge2 = private.Challenge_Create( enemies_packages, stage_ban_list, enemy_ban_list ),
    challenge3 = private.Challenge_Create( enemies_packages, stage_ban_list, enemy_ban_list ),
    challenge4 = private.Challenge_Create( enemies_packages, stage_ban_list, enemy_ban_list ),
    ready      = true
  };

  m_dailyquest.Data_Set( dailyquest );

end;
--------------------------------------------------------------------------------
function private.Update ()

  local dailyquest = m_dailyquest.Data_Get();
  local enemies    = v_enemies.ListCommon();

  dailyquest.reward_id = dailyquest.reward;
  
  for c = 1, private.challenge_count do

    for e = 1, #enemies do

      if ( enemies[ e ].name == dailyquest[ "challenge"..c ].enemy.name ) then

        dailyquest[ "challenge"..c ].enemy.content = enemies[ e ].content;

        break;

      end;

    end;

  end;

  dailyquest.reward    = private.Reward_Data();
  dailyquest.skipprice = m_dailyquest.GetSkipPrice();

  private.dailyquest_data = dailyquest;

end;
--------------------------------------------------------------------------------
function private.Refresh ()

  if ( private.IsUnlocked() ) then
 
    local enemy_packages = private.GetEnemies_GetEnabledPackeges();
    m_dailyquest.Unschedule( "dailyquest" );
    private.Start( enemy_packages );
    m_dailyquest.Schedule( "dailyquest" );

    mediator.Notify( "View_DailyQuest_Ready" );

    private.Cooldown_OnFinished( "dailyquest" );

  else

    -- JK: delete save does not clear timers
    m_dailyquest.Unschedule( "dailyquest", true );

  end;
  
end;
--------------------------------------------------------------------------------
function private.GetEnemies ( enemies_packages )

  local by_section = true;
  local enemies    = v_enemies.ListCommon();

  if ( by_section ) then

    local result   = {};

    for e = 1, #enemies do
      for s = 1, #enemies_packages do
        if ( enemies[ e ].content == enemies_packages[ s ] ) then
          table.insert( result, enemies[ e ] );
          break;
        end;
      end;
    end;

    return result;

  else

    return enemies;

  end;

end;
--------------------------------------------------------------------------------
function private.GetStages ()
  
  local stages = m_dailyquest.GetStageList();
  
  for s = #stages, 1, - 1 do
    
    local stage_data = v_stage.Stage_StorageData( stages[ s ].unlock ) or {};

    if ( not stage_data.completed ) then
       
      table.remove( stages, s );
      
    end;
    
  end;

  return stages;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Challenge_GetTotalCount ()

  local total = m_dailyquest.GetChallengeCount();
  
  return total;

end;
--------------------------------------------------------------------------------
function public.Challenge_GetCompletedCount ()

  local total    = m_dailyquest.GetChallengeCount();
  local complete = ( total - public.Challenge_GetOpenedCount() );
  
  return complete;

end;
--------------------------------------------------------------------------------
function public.Challenge_GetOpenedCount ()

  local open_count = 0;
  local challenges = m_dailyquest.Data_Get();
  
  for _, challenge in pairs( challenges ) do

    if ( type( challenge ) == "table" ) and
       ( challenge.state   == "open"  ) then
      
      open_count = open_count + 1;
      
    end;
    
  end;
    
  return open_count;

end;
--------------------------------------------------------------------------------
function private.Challenge_Create ( enemies_packages, stage_ban_list, enemy_ban_list ) 

  local challenge = {};

  local enemy = private.Challenge_CreateUniqueObject(
    private.GetEnemies( enemies_packages ),
    enemy_ban_list
  );

  table.insert( enemy_ban_list, enemy.name );

  challenge.enemy = {
    name   = enemy.name;
    health = math.random( enemy.health.min, enemy.health.max );
    attack = math.random( enemy.attack.min, enemy.attack.max );
  };

  local stage = private.Challenge_CreateUniqueObject(
    private.GetStages(),
    stage_ban_list
  );

  table.insert( stage_ban_list, stage.name );
  challenge.stage = stage.name;

  challenge.state = private.challenge_states[ 1 ];

  return challenge;

end;
--------------------------------------------------------------------------------
function private.Challenge_CreateUniqueObject ( list, ban_list ) 

  local ok = false;
  local obj;

  while ( not ok ) do
    obj = list[ math.random( #list ) ];
    if ( not common.IsEntityInList( ban_list, obj.name ) ) then
      ok = true;
    end;
  end;

  return obj;

end;
--------------------------------------------------------------------------------
function private.Challenge_SetBeaten ( challenge_id )
  
  if ( m_dailyquest.Challenge_SetState( challenge_id, 2 ) ) then

    -- JK: check if this challenge was last
    local complete = public.Challenge_GetCompletedCount();
    local total    = m_dailyquest.GetChallengeCount();
    
    v_log.Kontagent_LogEvent( "arena", 
    {
      name  = "success",
      level = complete
    } );

    if ( complete == total ) then

      local is_finished = m_dailyquest.Finish();
      
      if ( is_finished ) then
        mediator.Notify( "View_Dailyquest_Finished" );
      end;

    end;

    return true;

  end;

  return false;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Cooldown_OnFinished ( timer_type )

  if ( not int_dialog_dailyquest.IsActive() ) then
    
    return;
    
  end;
  
  if ( timer_type == "challenge" ) then
    
    int_dialog_dailyquest.Timer_Hide();
    
  elseif( timer_type == "dailyquest" ) then
    
    ui.screen.Lock( true );
    
    private.Dialog_Hide(
      function ()
     
        private.Dialog_Show( nil, private.func_ContinueMapUpdate );
        
        ui.screen.Lock( false );
     
      end
    );
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.Cooldown_Skip ( timer_type )
  
  local price = m_dailyquest.GetSkipPrice();
  
  if ( timer_type == "dailyquest" ) then
    
    if ( v_player.Gold_TryChangeCount( - price ) ) then

      v_log.Kontagent_LogEvent( "economy", 
      {
        description = "premium_currency_spend",
        category    = "dailyquest",
        value       = - price,
        name        = "skip_cooldown"
      } );
      
      private.Refresh();

    end;
    
    return;
    
  end;

  m_dailyquest.Unschedule( timer_type );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Reward_Data ()
  
  local reward_id = m_dailyquest.Data_Get( "reward" );

  if ( not reward_id ) then

    -- JK: show next reward
    reward_id = public.Challenge_GetCompletedCount() + 1;

  end;

  local reward = m_dailyquest.Reward_GetSettings( reward_id );

  if ( reward ) then

    return reward;

  end;

  return false;
  
end;
--------------------------------------------------------------------------------
function private.Reward_Tap ()

  if ( not private.dailyquest_data.reward_id ) then

    int_dialog_dailyquest.Reward_ShowPopup();
    
  else

    private.Reward_Show();

  end;

end;
--------------------------------------------------------------------------------
function private.Reward_Show ()

  local reward = private.dailyquest_data.reward;

  local chest = {
    type    = reward.type,
    name    = reward.name,
    count   = reward.count,
    extra_rewards = reward.extra_rewards
  };

  local params = {
    category   = "dailyquest",
    stage_name = chest.name,
    silent     = true
  };

  v_items.Chest_Show(
    chest,
    params,
    nil,
    function()

      if ( public.Challenge_GetOpenedCount() < 1 ) then

        v_log.Playhaven_LogPlacement( "arena_complete" );
        int_dialog_dailyquest.Timer_TryShow( "dailyquest", m_dailyquest.GetSkipPrice() );

      else

        int_dialog_dailyquest.Timer_TryShow( "challenge" );
        int_dialog_dailyquest.Reward_Disable();
        
        local multiset = {};
        
        for id = 1, private.challenge_count do

          int_dialog_dailyquest.Challenge_Update( id, multiset );

        end;
        if ( #multiset > 0 ) then
          ui.obj.Multiset( multiset );
        end;

      end;

      if ( public.IsSuspended() ) then
        private.Dialog_FeatureDisable( private.func_ContinueMapUpdate );
        return;
      end;
      
    end,
    function()

      local reward_id = private.dailyquest_data.reward_id;
      local reward    = private.dailyquest_data.reward;

      int_dialog_dailyquest.Reward_Set( reward_id, reward );

    end
  );
  
  if ( private.dailyquest_data.reward_id == 4 ) then
    v_log.Playhaven_LogPlacement( "arena_chest_open" );
  end;
   
  v_log.Kontagent_LogEvent( "arena", 
  {
    name  = "collect_reward",
    level = private.dailyquest_data.reward_id
  } );

  int_dialog_dailyquest.Reward_Disable( true );

  private.Reward_Deactivate();

  private.Update();
  private.PanelButton_SetNotifications();

end;
--------------------------------------------------------------------------------
function private.Reward_Activate ()

  -- JK: Set reward index
  local reward_index = public.Challenge_GetCompletedCount();
  m_dailyquest.Reward_Set( reward_index );

  -- JK: get reward data by index set above
  local reward_data = private.Reward_Data() or {};

  -- JK: Schedule challenge cooldown
  m_dailyquest.Schedule( "challenge", reward_data.cooldown );

end;
--------------------------------------------------------------------------------
function private.Reward_Deactivate ()

  m_dailyquest.Reward_Set( false );

end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function private.DailyQuest_TryStart ( stage_data )

  if ( m_dailyquest.GetSettings().unlockstage == stage_data.name ) then

    mediator.Unsubscribe(
      "Model_Stage_StageWon",
      private.DailyQuest_TryStart
    );

    m_dailyquest.Schedule( "countdown" );
    private.Refresh();

    private.PanelButton_Enable();

    mediator.Notify( "View_DailyQuest_CountdownStarted" );
    
  end;

end;
--------------------------------------------------------------------------------
function private.GetEnemies_GetEnabledPackeges ()
  
  local enabled_sections = v_sections.List();
  local enabled_packeges = {};
  
  for i = 1, #enabled_sections do
    
    local section = enabled_sections[ i ];
    if ( section.enabled ) then

      local enemy_package = "enemies_"..section.name:match( "%d+" );

      if ( v_game.Content_IsPackageReady( enemy_package ) ) then
        table.insert( enabled_packeges, enemy_package );
      end;

    end;
    
  end;
  
  if ( #enabled_packeges == 0 ) then
    enabled_packeges = { "enemies_1" };
  end;

  return enabled_packeges;
  
end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function public.FeatureUpdated ()
  
  local is_suspended = public.IsSuspended();

  private.PanelButton_Enable();
  
  mediator.Notify(
    "View_Dailyquest_FeatureUpdate",
    {
      active = not is_suspended
    }
  );
  
end;
--------------------------------------------------------------------------------
function public.ConnectionRestored ()
  
  private.PanelButton_Enable();
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Feature_GetState ()

  local  feature_state = v_features.GetFeatureState( "dailyquest" );
  return feature_state;

end;
--------------------------------------------------------------------------------
function private.Feature_SaveData ()
  
  v_features.Feature_SaveState( "dailyquest" );
  
end;
--------------------------------------------------------------------------------