-- name=v_features
--------------------------------------------------------------------------------

local _f = {};

--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_features"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()
  
  common.mod.Load( _modules );

  private.iCheatUserId = -1;
  
  if ( not parts ) then
    
    parts = {};
    
  end;

  if ( not parts.features ) then
    
    parts.features = {};
    
  end;
  
  private.registered_features = {};
  
  parts.features.IsActive       = public.IsActive;
  parts.features.GetFeatureData = private.Feature_GetData;
  parts.features.GetFeatureList = private.Feature_GetNameList;

  ne.msg.Subscribe(
    "Event_Feature_Updated",
    private.Feature_OnUpdated
  ); 
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Init ()

  m_features.RequestList( true, private.iCheatUserId );

end;
--------------------------------------------------------------------------------
function public.IsSynced ()
  
  local  is_synced = m_features.IsSynced();
  return is_synced;
  
end;
--------------------------------------------------------------------------------
function public.RegisterFeature ( data, is_online )
  
  table.insert( private.registered_features, data );
  
  if ( is_online ) then
    if ( data.func_save ) then
      data.func_save();
    end;
  end;
  
end;
--------------------------------------------------------------------------------
function public.GetRegisterFeatures ()
  
  return private.registered_features;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.SetCheatUserGroup( for_feature, group )
  local max_group, count_result = public.GetFeatureGroupCount( for_feature ) ;
  if ( not count_result ) then 
    return false;
  end;
  if ( group > max_group ) then
    common.DbgCallpoint( "[ ERROR ] SetCheatUserGroup: group "..tostring(group)..">max_group" );
    return false;
  end;
  private.iCheatUserId = 1;
  local user_group = 0;
  local result = false;
  for i = 1, max_group do
    user_group, result = public.GetFeatureGroup( for_feature );
    if ( not result or user_group == group ) then
      break;
    end;
    private.iCheatUserId = private.iCheatUserId + 1;
  end;
  local result = result and (user_group == group);
  if ( result ) then
    m_features.RequestList( true, private.iCheatUserId );
  end;
  return result;
end;
--------------------------------------------------------------------------------
function public.GetFeatureGroupCount ( feature_name )
  
  if ( not feature_name ) then

    common.DbgCallpoint( "[ ERROR ] Feature has no name." );
    return false, false;

  end;

  local features = m_features.GetList();

  if ( features.abtest_groups and 
       features.abtest_groups[ feature_name ] ) then
    
    local group_count = features.abtest_groups[ feature_name ];
      
    return group_count, true;
      
  end;

  return false, false;

end;
--------------------------------------------------------------------------------
function public.GetFeatureGroup ( feature_name )
  
  local group_count, result = public.GetFeatureGroupCount( feature_name );
  
  if ( not result ) then 
    return false, false;
  end;
  
  local userId = private.GetUserId();
  
  if ( ( group_count ) and ( userId ~= 0 ) ) then
    
    local group = math.fmod( userId, group_count );
    
    return group, true;
    
  end;

  return false, false;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.GetFeatureState ( feature_name )

  if ( not feature_name ) then

    common.DbgCallpoint( "[ ERROR ] Feature has no name." );
    return false, false;

  end;

  local  state, exists = m_features.Feature_GetState( feature_name );
  return state, exists;

end;
--------------------------------------------------------------------------------
function public.SaveFeatureState ( feature_name, feature_state )
  --ES> make private after refactor and delete private.Feature_SaveState
  if ( not feature_name ) then

    common.DbgCallpoint( "[ ERROR ] Feature has no name." );
    return false;

  end;

  local  is_saved = m_features.Feature_SaveState( feature_name, feature_state );
  return is_saved;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.GetFeatureData ( feature_name )

  if ( not feature_name ) then

    common.DbgCallpoint( "[ ERROR ] Feature has no name." );
    return false, false;

  end;

  local  data, exists = m_features.Feature_GetData( feature_name );
  return data, exists;

end;
--------------------------------------------------------------------------------
function public.SaveFeatureData ( feature_name, feature_data )

  if ( not feature_name ) then

    common.DbgCallpoint( "[ ERROR ] Feature has no name." );
    return false;

  end;

  local  is_saved = m_features.Feature_SaveData( feature_name, feature_data );
  return is_saved;

end;
--------------------------------------------------------------------------------
function public.RemoveFeatureData ( feature_name )

  if ( not feature_name ) then

    common.DbgCallpoint( "[ ERROR ] Feature has no name." );
    return false;

  end;

  local  is_deleted = m_features.Feature_RemoveData( feature_name );
  return is_deleted;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.GetSavedFeatureGroup ( feature_name )

  if ( not feature_name ) then

    common.DbgCallpoint( "[ ERROR ] Feature has no name." );
    return false, false;

  end;

  local  group, exists = m_features.Feature_GetGroup( feature_name );
  return group, exists;

end;
--------------------------------------------------------------------------------
function public.SaveFeatureGroup ( feature_name, feature_group )

  if ( not feature_name ) then

    common.DbgCallpoint( "[ ERROR ] Feature has no name." );
    return false;

  end;

  local  is_saved = m_features.Feature_SetGroup( feature_name, feature_group );
  return is_saved;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Feature_OnUpdated ()

  if ( parts.features.IsActive( "parts" ) ) then
  
    local parts_data = parts.features.GetFeatureData( "parts" )[ 1 ];
    local net_data = parts_data[ "net" ] or nil;
    
    if ( net_data and net_data.heartBeatPeriod ) then
    
      local period = math.tointeger( net_data.heartBeatPeriod );
      if  ( period > 0 ) then
      
        parts.net.heartbeat.SetPeriod( period );
      
      end;
      
    end;
  
  end;
  
  if ( parts.features.IsActive( "containers" ) ) then
    
    local data = parts.features.GetFeatureData( "containers" );
    if ( data ) then
      public.SaveFeatureData( "containers", data );
    end;
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.Feature_GetData ( feature_name )
  
  local features = m_features.GetList();

  return features[ feature_name ] or {};
  
end;
--------------------------------------------------------------------------------
function private.Feature_GetNameList ()
  
  local names    = {};
  local features = m_features.GetList();
  
  for feature, data in pairs( features ) do
    
    table.insert( names, feature );
    
  end;
  
  return names;
  
end;
--------------------------------------------------------------------------------
function public.IsActive ( feature_name )
  
  local features = m_features.GetList();

  if ( features[ feature_name ] ) then
    
    return true;
    
  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.GetUserId ()
  
  local user_id = AuthGetUserId();
  if ( private.iCheatUserId >= 0 ) then 
    user_id = private.iCheatUserId;
  end;
  
  if ( user_id ~= "" ) then
    return tonumber( user_id );
  end;

  return 0;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Feature_SaveState ( feature_name )
  
  private.Feature_SaveState( feature_name );
  
end;
--------------------------------------------------------------------------------
function private.Feature_SaveState ( feature_name )

  local is_active = public.IsActive( feature_name );

  public.SaveFeatureState( feature_name, is_active );

end;
--------------------------------------------------------------------------------