-- name=v_fights
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_fights"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  common.mod.Load( _modules );
  
  mediator.Subscribe(
    "Model_Builds_BuildConstructed",
    private.OnBuildConstructed
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Init ()

  local delegates = {
    GetCurrentBoss = function ()
      
      local current_boss = private.Boss_GetCurrent();
      return current_boss;
      
    end,
    GetFightData = function ( stage_name )
      
      local fight_data = public.Fight_Data( stage_name );
      return fight_data;
      
    end
  };
  
  m_fights.Setup( delegates );

end;
--------------------------------------------------------------------------------
function public.GetDefaultGoldRewardCount ()

  local count = m_fights.GetDefaultGoldRewardCount();
  
  return count;

end;
--------------------------------------------------------------------------------
function public.Create ( fight_data, section )

  local fight    = fight_data.name;
  local enemy    = fight_data.enemy;
  local realtype = fight_data.realtype;
  
  -- Create fight object
  local obj_fight = ( "obj_map_enemy_%s" ):format( fight );
  
  ObjCreate( obj_fight, "obj" );
  ObjSet   ( obj_fight, { anim_tag = fight.."_enemy" } );
  ObjAttach( obj_fight, "anm_map_"..section );
  
  -- Clone enemy object
  local hub_enemy = ( "obj_map_enemy_%s"    ):format( enemy );
  local obj_enemy = ( "obj_map_enemy_%s_%s" ):format( enemy, fight );
  
  ObjClone( hub_enemy, "_"..fight, 1, obj_fight );
  
  -- Create indicator object
  local indicator = ( "obj_map_enemy_%s_%s_indicator" ):format( enemy, fight );
  local anm_point = ( "anm_map_point_enemy_%s_%s"     ):format( enemy, fight );
  
  ObjCreate(anm_point, "anim");
  ObjAttach(anm_point, obj_enemy );
  
  local section_number = section:match( "%d+" );
  local boss_name = m_fights.Enemy_IsBoss( enemy );
  
  ObjSet( anm_point, {
    input = false,
    res = "assets/content/"..fight_data.content.."/"..
          ( boss_name or enemy ).."/anm_map_"..( boss_name or enemy )
  } );
  
  ObjCreate( indicator, "obj" );
  ObjSet   ( indicator, { anim_tag = "indicator" } );
  ObjAttach( indicator, anm_point );
  
  -- Set input events
  if     ( realtype == "fight" ) then
  
    local func_click = function () private.Tap( fight ); end
    ui.control.EventsSet( obj_enemy, { func_on_click = func_click } );
  
  elseif ( realtype == "misc" ) then
  
    local params = ObjGet( obj_enemy );
    
    ObjSet( obj_enemy,
    {
      input       = false,
      inputrect_x = 0.5 * params.inputrect_x,
      inputrect_y = 0.5 * params.inputrect_y,
      inputrect_w = 0.5 * params.inputrect_w,
      inputrect_h = 0.5 * params.inputrect_h
    } );
  
  elseif ( IsEditor() ) then
  
    local msg = "Error! Try to create fight %q of type %q";
    common.DbgCallpoint( msg:format( fight, realtype ) );
  
  end;
  
  private.Set( fight_data );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Set ( fight_data )

  local fight    = fight_data.name;
  local enemy    = fight_data.enemy;
  local realtype = fight_data.realtype;
  
  local visible   = fight_data.visible;
  local enabled   = fight_data.enabled;
  local completed = fight_data.completed;
  local absent    = fight_data.absent;
  
  if ( visible and fight_data.type ~= "ride" ) then

    private.Fight_Show( fight, enemy, realtype );
    private.Boss_SetAbsent( enemy, fight, absent );
    if ( enabled and not completed ) then
      
      private.Fight_Enable( fight );
      
    elseif ( not enabled and completed ) then

      private.Fight_Disable( fight );

    end;

  else

    private.Fight_Hide( fight, enemy );

  end;

  if ( completed ) then

    private.Fight_Complete( fight, enemy );

  end;

end;
--------------------------------------------------------------------------------
function private.Tap ( fight )

  local fight_data = public.Fight_Data( fight );
  
  local fight   = fight_data.name;
  local enemy   = fight_data.enemy;
  local content = fight_data.content;
  
  local obj_name = ( "obj_map_enemy_%s"          ):format( fight );
  local anm_name = ( "anm_map_point_enemy_%s_%s" ):format( enemy, fight );
  
  if ( ObjGet( anm_name ).animfunc ~= "click" ) then
  
    ui.anim.Play( anm_name, "click", nil, "static" );
  
  end;

  if ( fight_data.visible and fight_data.enabled ) then
    
    v_map.Focus( obj_name,
      {
        scale      = 1.0,
        func_after = function ()
          v_stage.DialogStart_Show( fight, "stages" );
        end
      }
    );

  else

    local is_boss = m_fights.Fight_IsBoss( fight );
    local str_pop = ( "str_int_popup_cloud_fights_%s" ):format( fight );
    local obj_ind = ( "obj_map_enemy_%s_%s_indicator" ):format( enemy, fight );

    local func_preclose;

    if ( is_boss ) then

      func_preclose = function ()

        local section_config = v_sections.Section_GetConfig();
        local boss_blocker   = section_config.fights.boss_blocker;
        
        if ( boss_blocker ) then

          v_builds.Build_HighlightStage( boss_blocker );

        else

          DbgTrace( "[ ERROR ] Boss has no associated blocker." );

        end;

      end;

    end;

    v_map.ShowObjectPopup(
      "monolog",
      obj_ind,
      str_pop,
      nil,
      nil,
      {
        func_preclose = func_preclose
      }
    );
    
    local sound = tern( is_boss )(
      "assets/content/"..content.."/"..enemy.."/aud_"..enemy.."_tap_map",
      "assets/content/"..content.."/"..enemy.."/aud_"..enemy.."_tap"
    );
    local alt_sound = is_boss and
      "assets/content/"..content.."/"..enemy.."/aud_"..enemy.."_tap"
      or
      "tap_monster";

    common.sound.Play( sound, nil, nil, alt_sound );
    
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Fight_Data ( fight_name )

  local fight_data = v_stage.Stage_StorageData( fight_name ) or {};
  local stage_data = v_stage.Stage_StaticData( fight_name );

  fight_data.name     = fight_name;
  fight_data.type     = stage_data.type;
  fight_data.realtype = stage_data.realtype;

  fight_data.state = fight_data.state or 1; -- FIXME: always 1?

  if ( fight_data.realtype == "misc" ) then

    fight_data.state = 4;

  end;

  local enemy = stage_data.state[ fight_data.state ].enemy;

  if ( enemy ) then

    fight_data.enemy   = enemy.name;
    fight_data.content = v_enemies.Enemy_GetContentPackageName( enemy.name );
    
  end;
  
  return fight_data;

end;
--------------------------------------------------------------------------------
function public.Fight_GetUnlockStage ( stages_list, stages_type )

  for s = stages_list[ 1 ], stages_list[ 2 ] do

    local stage_name = "stage_"..s;

    if not ( stages_type == "common" ) then

      stage_name = "stage_"..s.."_"..stages_type;

    end;

    if ( v_stage.Stage_IsFight( stage_name ) ) then

      return s;

    end;

  end;

  return false;

end;
--------------------------------------------------------------------------------
function private.Fight_Enable ( fight_name )

  local objects = {};

  table.insert( objects, { "anm_map_point_label_"..fight_name,
  {
    animfunc = "label_fight",
    playing  = true 
  } } );
    
  table.insert( objects, { "anm_map_point_bg_"..fight_name,
  {
    animfunc = "bg_active",
    playing  = true 
  } } );

  ObjMultiSet( objects );

end;
--------------------------------------------------------------------------------
function private.Fight_Disable ( fight_name, fight_data )
  
  fight_data = fight_data or public.Fight_Data( fight_name );

  local objects  = {};
  local fight_id = "_enemy_"..fight_data.enemy.."_"..fight_name;
  local anm_name = "anm_map_point"..fight_id;

  if ( fight_data ) and ( fight_data.realtype == "misc" ) then

    ui.anim.Play( anm_name, "disappear_fight" );

    local boss_fight = m_fights.Enemy_GetAbsent( fight_data.enemy );

    if ( boss_fight ) and not ( boss_fight.absent ) then

      private.Boss_SetAbsent( boss_fight.enemy, boss_fight.name, false );

    end;

  else

    fight_data = fight_data or public.Fight_Data( fight_name );

    if ( m_fights.Enemy_IsBoss( fight_data.enemy ) ) then

      ui.anim.Play( anm_name, "fail" );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Fight_Show ( fight_name, fight_enemy, realtype )

  local objects  = {};
  local fight_id = "_enemy_"..fight_enemy.."_"..fight_name;
  local obj_name = "obj_map"..fight_id;

  table.insert( objects, { obj_name,
  {
    visible = true,
    input   = true } } );

  local anm_name = "anm_map_point"..fight_id;

  if ( realtype == "misc" ) then
  
    table.insert( objects, { anm_name,
    {
      animfunc = "static_fight",
      playing  = true } } );

  else

    table.insert( objects, { anm_name,
    {
      animfunc = "static",
      playing  = true } } );
  
  end;

  ObjMultiSet( objects );

end;
--------------------------------------------------------------------------------
function private.Fight_Hide ( fight_name, fight_enemy )

  local fight_id = "_enemy_"..fight_enemy.."_"..fight_name;
  local obj_name = "obj_map"..fight_id;

  local fight_data = public.Fight_Data( fight_name );

  if ( fight_data ) and ( fight_data.realtype == "misc" ) then

    ObjSet( obj_name,
    {
      visible = false,
      input   = false } );

  else 

    if ( m_fights.Enemy_IsBoss( fight_enemy ) ) then

      local anm_name = "anm_map_point"..fight_id;

      ui.anim.Play( anm_name, "fail_static" );

      ObjSet( obj_name,
      {
        visible = true,
        input   = false } );

    else

      ObjSet( obj_name,
      {
        visible = false,
        input   = false } );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Fight_Complete ( fight_name, fight_enemy )

  local fight_id = "_enemy_"..fight_enemy.."_"..fight_name;
  local obj_name = "obj_map"..fight_id;

  local fight_data = public.Fight_Data( fight_name );

  if ( fight_data ) and ( fight_data.realtype == "misc" ) then

    ObjSet( obj_name,
    {
      visible = false,
      input   = false } );

  else

    if ( m_fights.Enemy_IsBoss( fight_enemy ) ) then

      local anm_name = "anm_map_point"..fight_id;

      ui.anim.Play( anm_name, "fail_static" );

      ObjSet( obj_name,
      {
        visible = true,
        input   = false } );

    else

      ObjSet( obj_name,
      {
        visible = false,
        input   = false } );

    end;

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Boss_SetAbsent (
    enemy_name,
    fight_name,
    absent,
    func_end,
    animate
  )

  if ( public.Fight_Data( fight_name ).realtype == "misc" ) then
    return;
  end;

  local anm_name = "anm_map_point_enemy_"..enemy_name.."_"..fight_name;
  local input    = false;

  if ( absent ) then

    if ( animate ) then

      ObjClone( anm_name, "_dummy", 1, "obj_map_point_"..fight_name );

      ui.anim.Play( anm_name.."_dummy", "disappear_dummy", function ()

        ObjDelete( anm_name.."_dummy" );

        if ( func_end ) then func_end(); end;

      end );

      ui.anim.Play( anm_name, "disappear", nil, "absent" );

    else

      ui.anim.Play( anm_name, "absent" );

    end;

  else

    input = true;
    
    ui.anim.Play( anm_name, "static" );

    if ( func_end ) then func_end(); end;

  end;
  
  ObjSet( "obj_map_enemy_"..fight_name, { input = input } );

end;
--------------------------------------------------------------------------------
function private.Boss_GetCurrent ()

  local section_name   = v_sections.Section_NameCurrent();
  local section_config = v_sections.Section_GetConfig( section_name );

  return section_config.fights.boss;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Callback_AnimateEnable ( stage )

  local data = public.Fight_Data( stage );
  
  if ( data.type == "fight" and data.visible and data.enabled ) then
  
    local enemy = data.enemy;
  
    -- Create object for misc fight
    if ( data.realtype == "misc" ) then
    
      local obj_enemy = ( "obj_map_enemy_%s_%s" ):format( enemy, stage );
      if ( not ui.obj.Exists( obj_enemy ) ) then
      
        ObjClone( 
          "obj_map_enemy_"..enemy, "_"..stage, 1, "obj_map_enemy_"..stage
        );
        
        local params = ObjGet( obj_enemy );
        ObjSet( obj_enemy,
          {
            input       = false,
            inputrect_x = 0.5 * params.inputrect_x,
            inputrect_y = 0.5 * params.inputrect_y,
            inputrect_w = 0.5 * params.inputrect_w,
            inputrect_h = 0.5 * params.inputrect_h
          }
        );
      
      end;
    
    end;
    
    -- Set animation
    local obj_name = ( "obj_map_point_%s" ):format( stage );
    local anm_name = ( "anm_map_point_enemy_%s_%s" ):format( enemy, stage );
    
    local res = 
    {
      fight = "assets/content/map/effects/anm_enemy_unlock",
      misc  = "assets/content/map/effects/anm_boss_appear"
    }
  
    local anim = 
    {
      name   = obj_name,
      res    = res[ data.realtype ],
      coords = { pos_x = 0, pos_y = 0 },
      
      func_dummy = function ()

        private.Sound_Play( "battle_ready_icon" );
        private.Fight_Enable( stage );

        if ( data.realtype == "misc" ) then

          local boss = m_fights.Enemy_GetAbsent( enemy );
          if  ( boss and boss.absent ) then
          
            local obj_boss = "obj_map_enemy_"..boss.name;
            local obj_stage = "obj_map_enemy_"..stage;
          
            v_map.Focus( obj_boss, { time = 1.5, func_after = function ()
                
              private.Sound_Play( "boss_magic" );
              private.Boss_SetAbsent( boss.enemy, boss.name, true, 
              function ()

                v_map.Focus( obj_stage, { time = 1.5, func_after = function ()
                    
                  private.Fight_Show( stage, enemy );
                  
                  ui.anim.Play( anm_name, "appear_fight", nil, "static_fight" );
                    
                end } );

              end, true );
            
            end } );

          end;

        end;

      end,
      func_action = function ( func_after )      
        
        if ( func_after ) then func_after(); end;
      
      end
    };
    
    return v_map.Callback_Animate( anim );
  
  end;
  
  return nil;

end;
--------------------------------------------------------------------------------
function public.Callback_AnimateDisable ( stage )
  
  local is_boss  = m_fights.Fight_IsBoss( stage );
  
  if ( is_boss ) then
    return nil;
  end;
  
  local data  = public.Fight_Data( stage );
  
  local real_fight = ( data.realtype == "fight" );
  local misc_ride  = ( data.realtype == "misc" and data.type == "ride" );
  
  if ( real_fight and     data.completed ) or
     ( misc_ride  and not data.completed ) then
  
    local enemy = data.enemy;
  
    local obj_name = ( "obj_map_enemy_%s_%s" ):format( enemy, stage );
    local anm_name = ( "anm_map_point_enemy_%s_%s" ):format( enemy, stage );

    local resource = "assets/content/map/effects/anm_appearance";
    local animfunc = "action";
  
    local anim = 
    {
      name     = obj_name,
      coords   = { pos_x = 0, pos_y = 0 },
      res      = resource,
      animfunc = animfunc,
      
      func_preset = function ()

        private.Fight_Disable( stage, data );

      end,
      func_dummy  = function ()
        
        ObjAnimate( obj_name, "alp", 0, 0, "",
        {
          0.0, 0, 1.0, 
          0.4, 0, 1.0, 
          0.7, 0, 0.0
        } );

        private.Sound_Play( "enemy_destroy" );

      end,
      func_action = function ( func_after )

        private.Fight_Hide( stage, enemy );
        if ( func_after ) then func_after(); end;

      end
    };
    
    return v_map.Callback_Animate( anim );
  
  end;
  
  return nil;

end;
--------------------------------------------------------------------------------
function public.Callback_AnimateBossDisable ( stage )
  
  local is_boss  = m_fights.Fight_IsBoss( stage );
  
  if not ( is_boss ) then
    return nil;
  end;

  local data  = public.Fight_Data( stage );
  
  local real_fight = ( data.realtype == "fight" );
  local misc_ride  = ( data.realtype == "misc" and data.type == "ride" );
  
  if ( real_fight and     data.completed ) or
     ( misc_ride  and not data.completed ) then
  
    local enemy = data.enemy;
  
    local obj_name = ( "obj_map_enemy_%s_%s" ):format( enemy, stage );
    local anm_name = ( "anm_map_point_enemy_%s_%s" ):format( enemy, stage );

    local resource = "assets/content/map/effects/anm_enemy_disappear";
    local animfunc = "boss_disappear";
  
    local anim = 
    {
      name     = obj_name,
      coords   = { pos_x = 0, pos_y = 0 },
      res      = resource,
      animfunc = animfunc,
      
      func_preset = function ()

        private.Fight_Disable( stage, data );

      end,
      func_dummy  = function ()

        private.Fight_Hide( stage, enemy );
        private.Sound_Play( "enemy_destroy" );

      end,
      func_action = function ( func_after )

        if ( func_after ) then func_after(); end;

      end
    };
    
    return v_map.Callback_Animate( anim );
  
  end;
  
  return nil;

end;
--------------------------------------------------------------------------------
function public.Callback_AnimateAppear ( stage )
  
  local data = public.Fight_Data( stage );
  
  if ( data.type == "fight" ) then
  
    local enemy   = data.enemy;
    local is_boss = m_fights.Fight_IsBoss( stage );
  
    local obj_name = ( "obj_map_enemy_%s" ):format( stage );
    local anm_name = ( "anm_map_point_enemy_%s_%s" ):format( enemy, stage );
    
    local anim  = {};
    anim.name   = obj_name;
    anim.res    = "assets/content/map/effects/anm_appearance";
    anim.coords = { pos_x = 0, pos_y = 0 };
    
    anim.func_preset = function ()
    
      if not ( is_boss ) then
      
        private.Fight_Show( stage, enemy );
        ObjSet( obj_name, { alp = 0.0 } );
      
      end;
    
    end;
    
    anim.func_dummy = function ()
    
      private.Sound_Play( "lighting_app" );
      
      if not ( is_boss ) then
      
        ObjAnimate( obj_name, "alp", 0, 0, "",
        {
          0.0, 0, 0, 
          0.3, 0, 1
        } );
      
      else
      
        ui.anim.Play( anm_name, "click", nil, "static" );
      
      end;

    end;
    
    anim.func_action = function ( func_after )
    
      if ( func_after ) then func_after(); end;
    
    end;
    
    return v_map.Callback_Animate( anim );
  
  end;
  
  return nil;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Sound_Play ( name )

  local sect = v_sections.Section_NameCurrent();
  
  local sound_generic = "assets/content/map/sounds/aud_"..name;
  local sound_section = "assets/content/"..sect.."/aud_"..name;
  
  local sound = tern( name == "boss_magic" )( sound_section, sound_generic );
  
  common.sound.Play( sound );
  
end;
--------------------------------------------------------------------------------
function private.OnBuildConstructed ( params )
 
  local build_data = v_builds.Build_GetData( params.name );
  
  if ( build_data ) and ( build_data.unlocklist ) then
    
    local fight_name = private.Zone_Fight( build_data.unlocklist );
    
    if ( fight_name ) then
      
      m_fights.Fight_SetVisible( fight_name );
      
    end;
    
  end;
 
end;
--------------------------------------------------------------------------------
function private.Zone_Fight ( stage_list )

  for s = stage_list.min, stage_list.max do

    local stage_name = "stage_"..s;

    if not ( stage_list.type == "common" ) then

      stage_name = "stage_"..s.."_"..stage_list.type;

    end;

    if ( v_stage.Stage_IsFight( stage_name ) ) then

      return stage_name;

    end;

  end;    

  return false;

end;
--------------------------------------------------------------------------------