-- name=v_fireflies
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_fireflies",
  "int_dialog_fireflies"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  common.mod.Load( _modules );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Init ()
  
  private.Init_Subscribes();

  m_fireflies.Sticker_InitList();

  private.stickerpack_price = m_fireflies.Stickerpack_Price();
  
  if ( private.Unlocked() ) then

    private.PanelButton_Enable();
  
  else

    local blocker_data = private.GetBlockerData();
    
    private.messages = {
      stage = "Model_Stage_StageWon",
      build = "Model_Builds_BuildConstructed"
    };
    
    mediator.Subscribe(
      private.messages[ blocker_data.type ],
      private.PanelButton_TryEnable
    );
    
  end;

end;
--------------------------------------------------------------------------------
function public.Kontagent_GetStickersTotal ()
  
  local total = m_fireflies.Stats().stickers.collected;
  
  return total;
  
end;
--------------------------------------------------------------------------------
function public.Cheat_GetAllStickers ()

  m_fireflies.Cheat_GetAllStickers();

  if ( v_interface.window.IsOnScreen( "int_dialog_fireflies" ) ) then

    int_dialog_fireflies.Tab_SetupAlbum();

  end;  

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.PanelButton_Enable ()
  
  common.delegate.Set( "int_panel_interface_bottom", 
  {
    StickersClickHandler = function ()
    
      private.Dialog_Show();
      common.sound.Play( "bottom_button" );
      v_log.Playhaven_LogPlacement( "stickers_open" );
      private.Panel_UpdateCount( 0 );
    
    end;
  } );
  
  v_interface.Scheme_AddButton( "map", "bottom", { name = "stickers" } );
  private.Panel_UpdateCount( m_fireflies.Stickerpack_Count().total );
  
end;
--------------------------------------------------------------------------------
function private.PanelButton_TryEnable ( unlock_data )
  
  local blocker_name = private.GetBlockerData().name;

  if ( unlock_data.name == blocker_name ) then

    local blocker_type = unlock_data.name:match( "(%w+)_%d+" );
    
    mediator.Unsubscribe(
      private.messages[ blocker_type ],
      private.PanelButton_TryEnable
    );
    
    private.PanelButton_Enable();
  
  end

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Init_Subscribes ()
  
  mediator.Subscribe(
    "Model_Fireflies_BoosterCountChanged",
    function ( params )
  
      private.Panel_UpdateCount( params.count );
    
    end
  );
  
  mediator.Subscribe(
    "Model_Fireflies_AchievementComplete",
    function ( reward )
  
      private.Achievement_Complete( reward );
    
    end
  );

  mediator.Subscribe(
    "Model_Fireflies_AchievementCompleteAll",
    function ( reward )
  
      private.Achievement_CompleteAll( reward );
    
    end
  );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Dialog_Show ( tab_number )

  private.Dialog_Show( tab_number );

end;
--------------------------------------------------------------------------------
function private.Dialog_Show ( tab_number )

  tab_number = tab_number or 1;

  ------------------------------------------------------------------------------
  local func_StickerpackGetCount = function ()
 
    local total = m_fireflies.Stickerpack_Count().total;
    
    return total;
 
  end;
  ------------------------------------------------------------------------------
  local func_top_panel_show = function ()
  
    if ( not v_interface.panel.top.IsVisible() ) then

      v_interface.panel.top.Highlight( "gold" );
      v_interface.panel.top.Show();

    end;
  
  end;
  ------------------------------------------------------------------------------
  local func_top_panel_hide = function ()

    if ( v_interface.panel.top.IsVisible() ) then

      v_interface.panel.top.Hide( false );

    end;
  
  end;
  ------------------------------------------------------------------------------
  local func_show = function ()
  
    int_dialog_fireflies.Show(
      tab_number,
      function ( firefly_name )
        local firefly_data = m_fireflies.Firefly_Data( firefly_name );
        return firefly_data;
      end,
      function ()
        local fireflies = m_fireflies.Firefly_GetAll();
        return fireflies;
      end,
      private.Stickerpack_Data,
      private.Duplicates_Data,
      func_StickerpackGetCount,
      function ( sticker_name )
        local is_sticked = m_fireflies.Sticker_TryStick( sticker_name );
        return is_sticked;
      end,
      function ()
        local stats = m_fireflies.Stats();
        return stats;
      end,
      private.Dialog_Hide,
      func_top_panel_show,
      func_top_panel_hide,
      private.Reward_TryShow,
      private.Reward_IsPending,
      private.Reward_GetData,
      function ()
        local reward = m_fireflies.GetAlbumReward();
        return reward;
      end
    );
    
  end;
    
  v_interface.window.Show( "int_dialog_fireflies", func_show );
  
end;
--------------------------------------------------------------------------------
function private.Dialog_Hide ( func_end )
  
  v_log.Playhaven_LogPlacement( "stickers_close" );
  
  int_dialog_fireflies.Hide(
    function ()

      v_interface.window.Hide( "int_dialog_fireflies" );

      if ( func_end ) then
  
        func_end();

      end;

    end
  );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Dialog_GoToSticker ( sticker )

  int_dialog_fireflies.Tutorial_GoToSticker( sticker );

end;
--------------------------------------------------------------------------------
function public.Dialog_GetStickerPoolPos ( sticker_name )

  local position = int_dialog_fireflies.Tutorial_GetStickerPoolPos(
    sticker_name
  );

  return position;

end;
--------------------------------------------------------------------------------
function public.Dialog_StickerPosition ( sticker_index )

  local position = int_dialog_fireflies.Tutorial_StickerPosition(
    sticker_index
  );

  return position;

end;
--------------------------------------------------------------------------------
function public.Dialog_SetButtonsInput ( value )

  int_dialog_fireflies.Tutorial_SetButtonsInput( value );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Duplicates_Data ()

  return {
    duplicates      = m_fireflies.Sticker_Duplicates(),
    func_TrySell    = private.Sticker_TrySell,
    func_TrySellAll = private.Sticker_TrySellAll
  };  
  
end;
--------------------------------------------------------------------------------
function private.Sticker_TrySell ( sticker_name, count )

  local gold_count = m_fireflies.Sticker_TrySell( sticker_name, count );
  
  if ( gold_count ) then
  
    v_items.TryObtain(
      { 
        name       = "gold", 
        type       = "gold", 
        count      = gold_count,
        category   = "stickers", 
        stage_name = "duplicate",
        silent     = false
      }
    );
    
    return true;
  
  end;
  
  return false;

end;
--------------------------------------------------------------------------------
function private.Sticker_TrySellAll ( stickers )

  local unsold = {};

  for _, sticker in ipairs( stickers ) do

    if not ( private.Sticker_TrySell( sticker.name, sticker.count ) ) then

      table.insert( unsold, sticker );

    end;

  end;

  return unsold;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Stickerpack_TryObtain ( count, bought, source )

  if not ( private.Unlocked() ) then
    
    return false;
    
  end;
  
  count = count or 1;
  
  local stickerpack_type = tern( bought )( "bought", "free" );

  if ( m_fireflies.Stickerpack_Obtain( count, stickerpack_type ) ) then
  
    v_log.Kontagent_LogEvent( 
      "economy", 
      {
        description = "item_collected",
        category    = "stickers",
        value       = count,
        name        = "stickerpack",
        source      = source
      }
    );
    
    return true;
  
  end;
  
  return false;

end;
--------------------------------------------------------------------------------
function public.Stickerpack_Count ()
  
  local  count = m_fireflies.Stickerpack_Count();
  return count;
  
end;
--------------------------------------------------------------------------------
function private.Stickerpack_Data ()
  
  local price = private.stickerpack_price;

  local func_Open = function ()

    local stickers = private.Stickerpack_TryOpen();

    if ( stickers ) then

      return stickers;

    end;

    return false;

  end;
  ------------------------------------------------------------------------------
  local func_Get = function ()

    return func_Open();

  end;
  ------------------------------------------------------------------------------
  local func_Buy = function ()
  
    if ( v_player.Gold_TryChangeCount( - price ) ) then
      
      v_log.Kontagent_LogEvent( "economy", 
      {
        description = "premium_currency_spend",
        category    = "fireflies_album",
        value       = - price,
        name        = "stickerpack"
      } );
  
      if ( public.Stickerpack_TryObtain( 1, true, "album" ) ) then
      
        return func_Open();

      end;

      return false;
      
    end;
    
    return false;
  
  end;
  ------------------------------------------------------------------------------
  return {
    price       = price,
    packs_total = m_fireflies.Stickerpack_Count(),
    func_Get    = func_Get,
    func_Buy    = func_Buy
  };
  
end;
--------------------------------------------------------------------------------
function private.Stickerpack_TryOpen ()
  
  local is_opened = m_fireflies.Stickerpack_TryOpen();
  
  return is_opened;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Panel_UpdateCount ( count )
  
  local data = {
    target = "stickers",
    data = {
      type = "common",
      entity = "v_fireflies",
      count = count
    }
  };
  
  v_interface.panel.bottom.Notification( data );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Achievement_Complete ( reward )

  private.reward = reward;
  
  if ( reward.type ~= "chest" ) then
  
    v_items.TryObtain( {
      name       = reward.name, 
      type       = reward.type, 
      count      = reward.count,
      category   = "stickers",
      stage_name = "stickers",
      silent     = true
    } );
  
  elseif ( reward.type == "chest" ) then
    
    local chest_data =
    {
      type  = reward.type,
      name  = reward.name,
      count = 1
    };
    
    local rewards = v_items.Chest_Generate( chest_data );
    
    private.reward.custom = rewards;
    
    local obtain_params = 
    {
      category   = "stickers",
      stage_name = reward.name,
      silent     = true
    };
    
    v_items.TryObtainMultiple( rewards, obtain_params );
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.Achievement_CompleteAll ( reward )

  private.reward_all = reward;

  local obtain_params = 
  {
    category   = "album",
    stage_name = "album",
    silent     = true
  };
  
  v_items.TryObtainMultiple( reward.custom, obtain_params );
  
end;
--------------------------------------------------------------------------------
function private.Reward_TryShowForAll ()
  
  if ( not private.reward_all ) then

    return;

  end;

  ui.screen.Lock( true );

  int_dialog_fireflies.Tab_Show( 4 );
  
  local reward = private.reward_all.custom;
  local chest  =
  {
    type  = private.reward_all.type,
    name  = private.reward_all.name,
    count = private.reward_all.count
  };

  private.reward_all = nil;

  common.Defer(
    "album_reward",
    1,
    function()

      v_items.Chest_Show(
        chest,
        nil,
        reward,
        nil,
        function ()

          ui.screen.Lock( false );

        end
      );

    end
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Reward_TryShow ( func_exec )

  if ( not private.reward ) then

    return;

  end;

  local reward   = private.reward;
  private.reward = nil;

  if ( reward.type == "chest" ) then
    
    local chest_data =
    {
      type  = reward.type,
      name  = reward.name,
      count = 1
    };

    v_items.Chest_Show(
      chest_data,
      nil,
      reward.custom,
      private.Reward_TryShowForAll,
      func_exec
    );

    return;

  end;

  v_items.Reward_Show( reward );

end;
--------------------------------------------------------------------------------
function private.Reward_IsPending ()

  return private.reward ~= nil;

end;
--------------------------------------------------------------------------------
function private.Reward_GetData ()

  return {
    index  = private.reward.index,
    layout = private.reward.layout
  };

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Unlocked ()

  return private.Unlocked();

end;
--------------------------------------------------------------------------------
function private.Unlocked ()
  
  return not private.GetBlockerData().locked;

end;
--------------------------------------------------------------------------------
function private.GetBlockerData ()
  
  local blocker = m_fireflies.GetBlockerData();
  local blocker_type = blocker:match( "(%w+)_%d+" );
  
  local func_isLocked = {
    build = function ( build_name )
      return not v_builds.Build_IsConstructed( build_name );
    end,
    stage = function ( stage_name )
      return not v_stage.IsStageComplete( stage_name );
    end
  };
  
  local locked = ( func_isLocked[ blocker_type ] or function () return false; end )( blocker );
  
  return { name = blocker, type = blocker_type, locked = locked };

end;
--------------------------------------------------------------------------------