-- name=v_friends_new
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  if ( IsMobile() ) then

    mediator.SubscribeOnce( "View_Map_MapOpened", private.Start );
    
    private.started       = false;
    private.popup_timeout = 3.5;
    private.hub_offset_x  = 40;
    private.obj_parent    = "obj_map_friends";
    private.obj_hub_base  = "obj_map_friend_";
    private.point_base    = "obj_map_point_";
    private.anm_icon      = "anm_map_friend_icon_";

    private.active_popups = {};
    private.saved_friends = {};
    private.section = nil;
    
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Start ()
  
  private.started = true;
  
  mediator.Subscribe( "View_Map_MapClosed", private.OnMapClosed );
  mediator.Subscribe( "Model_Social_FriendsCollectionsUpdated", private.OnFriendsRecieved );
  mediator.Subscribe( "View_Map_GroundLoaded", private.Update );
  
  private.Update();

end;
--------------------------------------------------------------------------------
function public.ConnectionRestored ()
  
  if ( IsMobile() ) then
    
    private.Update();
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.Update ()
  
  local section = v_sections.Section_NameCurrent();

  if ( section ~= private.section ) then
    private.section = section;
    private.UpdateFriendsList();
  end;

  if ( private.execute_lists ) then
    local create_list = private.execute_lists.create_list;
    local update_list = private.execute_lists.update_list;
    local delete_list = private.execute_lists.delete_list;

    if ( create_list and #create_list > 0 ) then
      for i = 1, #create_list do
        local friend = create_list[ i ];
        private.saved_friends[ tostring(friend.userId) ] = friend.stage;
        private.FriendCreate( friend );
      end;
    end;

    if ( delete_list and #delete_list > 0 ) then
      for i = 1, #delete_list do
        local friend = delete_list[ i ];
        private.saved_friends[ tostring(friend.userId) ] = nil;
        private.FriendDelete( friend );
      end;
    end;

    if ( update_list and #update_list > 0 ) then
      for i = 1, #update_list do
        local friend = update_list[ i ];
        private.saved_friends[ tostring(friend.userId) ] = friend.stage;
        private.FriendUpdate( friend );
      end;
    end;

    private.execute_lists = nil;
  end;

end;
--------------------------------------------------------------------------------
function private.UpdateFriendsList ()
  
  --Important: you can't use RAW 'friend.userId' as index of associated array coz it's number

  local friend_list = v_social.Friends_ListInZone( private.section );
  
  local create_list = {}; -- Created: NO,  Point: EMPTY
  local update_list = {}; -- Created: YES, Point: EMPTY, CHANGED
  local delete_list = {}; -- Created: YES, Point: FILLED

  -- Remove friends, that gone too far
  for friend_id in pairs( private.saved_friends ) do
    local userId = tonumber( friend_id );
    if ( not private.IsValueInList( friend_list, "userId", userId ) ) then
      table.insert( delete_list, { userId = userId } );
    end;
  end;

  local counter = 1;
  while counter <= #friend_list do 
    local friend = friend_list[ counter ];
    -- Point: EMPTY
    if ( not private.IsValueInList( friend_list, "stage", friend.stage, "userId", friend.userId )) then

      -- Created: NO
      if ( not private.saved_friends[ tostring(friend.userId) ] ) then
        table.insert( create_list, friend );

      -- Created: YES
      else

        -- Point: CHANGED
        if ( private.saved_friends[ tostring(friend.userId) ] ~= friend.stage ) then
          table.insert( update_list, friend );
        end;
      end;
      counter = counter + 1; 

    -- Point: FILLED
    else

      -- Created: YES
      if ( private.saved_friends[ tostring(friend.userId) ] ) then
        table.insert( delete_list, friend );
      end;
      table.remove( friend_list, counter );

    end;
  end;

  private.execute_lists = {
    create_list = create_list,
    update_list = update_list,
    delete_list = delete_list
  };
  
end;
--------------------------------------------------------------------------------
function private.OnFriendsRecieved ()

  if ( not private.section ) then
    return;
  end;
  
  private.UpdateFriendsList();
  
  if ( v_map.IsActive() ) then
    private.Update();
  end;

end;
--------------------------------------------------------------------------------
--function ...()
--------------------------------------------------------------------------------
function private.FriendCreate ( friend )

  local obj_hub    = private.obj_hub_base..friend.userId;
  local anm_icon   = private.anm_icon..friend.userId;
  local obj_parent = private.obj_parent;
  local tad_name   = "tad_map_friend_name_"..friend.userId;
  local grd_bg     = "grd_map_friend_name_bg_"..friend.userId;
  local spr_icon   = "spr_map_friend_icon_"..friend.userId;
  local obj_popup  = "obj_map_friend_name_"..friend.userId;

  ObjClone( "obj_map_friend", "_"..friend.userId, 1, obj_parent );

  local obj_pos = private.GetPosByStage( private.point_base..friend.stage, obj_parent );

  if ( friend.url == '' ) then
    friend.url = nil;
  end;

  ObjMultiSet( {
    { obj_hub, { 
      pos_x = obj_pos.pos_x,
      pos_y = obj_pos.pos_y
    } },
    { tad_name, { text = friend.name } },
    { spr_icon, {
      url = friend.url,
      res = "assets/content/interface/int_dialog_stage_friends/default_0",
      default_res = "assets/content/interface/int_dialog_stage_friends/default_0",
      trg_resloaded = function ()
        ui.obj.SpriteAdjust( spr_icon, 34 );
        private.AnimateAppear( anm_icon );
      end;
    } }
  } );

  if ( not friend.url ) then
    private.AnimateAppear( anm_icon );
  end;

  ui.obj.SpriteAdjust( spr_icon, 34 );

  local tad_params = ObjGet( tad_name );
  local draw_width = tad_params.draw_width + 10;
  local rounded_width = draw_width - 10;
  GridSet( grd_bg, 0, {
    2,  { pos_x = rounded_width },
    3,  { pos_x = draw_width },
    6,  { pos_x = rounded_width },
    7,  { pos_x = draw_width },
    10, { pos_x = rounded_width },
    11, { pos_x = draw_width },
    14, { pos_x = rounded_width },
    15, { pos_x = draw_width }
  } );

  private.PopupHide( obj_popup, true );
  ui.control.TapClear( obj_hub );
  ui.control.TapSet( obj_hub,
  {
    func_true = function ()
      private.PopupShow( obj_popup );
    end
  } );

end;
--------------------------------------------------------------------------------
function private.FriendUpdate ( friend )

  local obj_hub    = private.obj_hub_base..friend.userId;
  local obj_parent = private.obj_parent;
  local old_pos    = ObjGet( obj_hub );
  local new_pos    = private.GetPosByStage ( private.point_base..friend.stage, obj_parent );

  private.AnimateMove( obj_hub, old_pos, new_pos );

end;
--------------------------------------------------------------------------------
function private.FriendDelete ( friend )

  local obj_hub  = private.obj_hub_base..friend.userId;
  local anm_icon = private.anm_icon..friend.userId;

  private.AnimateDisappear( anm_icon, function ()
    ObjDelete( obj_hub );
  end );

end;
--------------------------------------------------------------------------------
function private.AnimateMove ( obj_hub, old_pos, new_pos )

  ObjAnimate( obj_hub, "pos_xy", 0, 0, "", 
  {
    0.00, 3, old_pos.pos_x, old_pos.pos_y,
    0.40, 3, new_pos.pos_x, new_pos.pos_y
  } );

end;
--------------------------------------------------------------------------------
function private.AnimateAppear ( anm_icon )

  ui.anim.Play( anm_icon, "appear", nil, "static" );

end;
--------------------------------------------------------------------------------
function private.AnimateDisappear ( anm_icon, func_end )

  ui.anim.Play( anm_icon, "disappear", func_end );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.OnMapClosed ()

  for obj_popup, is_active in pairs( private.active_popups ) do
    private.PopupHide( obj_popup, true );
  end;

end;
--------------------------------------------------------------------------------
function private.PopupShow ( obj_popup )

  ObjStopAnimate( obj_popup, "alp" );

  if ( private.active_popups[ obj_popup ] ) then
    private.PopupHide( obj_popup );
  else
    private.active_popups[ obj_popup ] = true;
    ObjSet( obj_popup, { active = true, visible = true } );
    ObjAnimate( obj_popup, "alp", 0, 0, "", 
    {
      0.0, 3, 0.0,
      0.2, 3, 1.0
    } );
    private.TimerStart( obj_popup );
  end;

end;
--------------------------------------------------------------------------------
function private.PopupHide ( obj_popup, instant )

  ObjStopAnimate( obj_popup, "alp" );
  private.TimerStop( obj_popup );

  if ( instant ) then
    ObjSet( obj_popup, { alp = 0.0, active = false, visible = false } );
  else
    ObjAnimate( obj_popup, "alp", 0, 0, function ()
      ObjSet( obj_popup, { active = false, visible = false } );
    end, 
    {
      0.0, 3, 1.0,
      0.2, 3, 0.0
    } );
  end;

  private.active_popups[ obj_popup ] = nil;

end;
--------------------------------------------------------------------------------
function private.TimerStart ( obj_popup )

  local tmr_popup = "tmr_"..obj_popup;

  ObjCreate( tmr_popup, "timer" );
  ObjAttach( tmr_popup, obj_popup );
  ObjSet( tmr_popup, {
    time = private.popup_timeout,
    playing = true,
    endtrig = function ()
      private.PopupHide( obj_popup );
    end
  } );

end;
--------------------------------------------------------------------------------
function private.TimerStop ( obj_popup )

  local tmr_popup = "tmr_"..obj_popup;

  ObjDelete( tmr_popup );

end;
--------------------------------------------------------------------------------
-- function ...()
--------------------------------------------------------------------------------
function private.GetPosByStage ( stage_name, parent_name )

  local obj_pos = ui.obj.Position( stage_name, parent_name );
  return {
    pos_x = obj_pos[1] + private.hub_offset_x, 
    pos_y = obj_pos[2] 
  };

end;
--------------------------------------------------------------------------------
function private.IsValueInList ( list, target_field, target_value, ignore_field, ignore_value )
  for i = 1, #list do
    if ( not ( ignore_field and list[ i ][ ignore_field ] == ignore_value )) and
    ( list[ i ][ target_field ] == target_value ) then
      return i;
    end;
  end;
  return false;
end;
--------------------------------------------------------------------------------