-- name=v_game
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_game",
  "m_migrations",
  "int_game_start",
  "int_dialog_info",
  "int_dialog_update",
  "int_dialog_deficient",
  "int_dialog_notifications",
  "int_dialog_saveconflict"
};
--------------------------------------------------------------------------------

public.dialog   = {};
public.features = {};

--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  common.mod.Load( _modules );

  private.app_url =
  {
    ios    = "https://itunes.apple.com/app/id905121703",
    google = "https://play.google.com/store/apps/details?id=com.bigfishgames.sirmatchgoogle",
    amazon = "http://www.amazon.com/gp/mas/dl/android?p=com.bigfishgames.sirmatchamazon"
  };

  private.splash_screen_active = false;

  ne.msg.Subscribe( Event_LocalNotifications_GetInfo, function ()

    private.LocalNotifications_AddData();

  end );

  ne.msg.Subscribe( Event_Application_MemoryWarning, private.OnMemoryWarning );

  private.controllers = {
    "v_friends_new",
    "v_dailyquest",
    "v_social",
    "v_minievent"
  };

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Init ( params )

  private.is_debug    = params.is_debug;
  private.is_online   = params.is_online;
  private.sysmsg_type = params.sysmsg_type;
  private.has_free_space = true;
  private.total_packages_size = nil;
  private.ccs_mustbe_showed = false;
  private.ccs_freeze = false;

  m_migrations.Run( private.is_online );
  private.Subscribe();
  --private.empty_manifest = m_game.Content_IsManifestEmpty();

  v_screen.SetInitial( params.screen.name, params.screen.callbacks );

  m_game.SetupClient( private.is_online );
  private.Setup_UI();

  if ( IsMobile() ) then

    if ( m_game.IsOnline() ) then

      m_game.RequestIsCCSAvailable();

    else

      mediator.SubscribeOnce(
        m_game.msg.ConnectRestored,
        m_game.RequestIsCCSAvailable
      );

    end;

  end;

  v_interface.window.RegisterAsModal( "int_dialog_info" );
  v_interface.window.RegisterAsModal( "int_dialog_update" );
  v_interface.window.RegisterAsModal( "int_dialog_deficient" );
  v_interface.window.RegisterAsModal( "int_dialog_notifications" );
  v_interface.window.RegisterAsModal( "int_dialog_saveconflict" );

  --if ( IsPC() ) then

    local value = #v_sections.StaticList() - 1;

    public.Content_LoadStrings( value );

  --end;

  private.first_install = m_game.FirstLaunch_GetInfo();

end;
--------------------------------------------------------------------------------
function private.Subscribe ()

  -- Lock screen with waiting animation
  mediator.Subscribe( m_game.msg.WaitingStart,  private.OnWaitingStart  );
  mediator.Subscribe( m_game.msg.WaitingFinish, private.OnWaitingFinish );

  -- Connect lost/restore handlers
  mediator.Subscribe( m_game.msg.ConnectLost,     private.OnConnectLost    );
  mediator.Subscribe( m_game.msg.ConnectRestored, private.OnConnectRestore );

  -- System messages handlers
  mediator.Subscribe( m_game.msg.ContentObsolete, private.Content_OnObsolete );
  mediator.Subscribe( m_game.msg.StorageConflict, private.OnStorageConflict );
  mediator.Subscribe( m_game.msg.SessionConflict, private.OnSessionConflict );
  mediator.Subscribe( m_game.msg.VersionOutdated, private.OnVersionOutdated );
  mediator.Subscribe( m_game.msg.Maintenance,     private.OnMaintenance );

  -- Silent content handlers
  mediator.Subscribe( m_game.msg.SilentContentFailed, private.SilentContent_OnFailed );
  mediator.Subscribe( m_game.msg.SilentContentReady,  private.SilentContent_OnReady );

  -- Charboost location
  mediator.Subscribe( m_game.msg.GainFocus, private.OnGainFocus );

  ne.msg.Subscribe( Event_Feature_Updated, private.OnFeatureUpdated );

  -- Rave login conflict
  mediator.Subscribe(
    "Event_Social_RaveLoginConflict",
    private.OnRaveLoginConflict
  );

  if ( IsAndroid() or ( IsPC() and IsEditor() and debug ) ) then

    ne.msg.Subscribe(
      "Event_Android_BackPressed",
      private.OnAndroidBackButtonPressed
    );

    mediator.Subscribe(
      "View_Stage_DialogWin_NextButtonPressed",
      private.OnDialogWinNextButtonPressed
    );

  end;

  if ( IsMobile() ) then

    mediator.Subscribe(
      "Interface_RightPanel_Visible",
      private.CCS_OnPanelShow
    );

    mediator.Subscribe(
      "Interface_RightPanel_Invisible",
      private.CCS_OnPanelHide
    );

    mediator.Subscribe(
      "Interface_RightPanel_Updated",
      private.CCS_OnPanelUpdate
    );

    mediator.Subscribe(
      "Interface_Dialog_Tutorial_Hide",
      private.CCS_OnTutorialHide
    );

    mediator.Subscribe(
      "Interface_Dialog_Tutorial_Open",
      private.CCS_OnTutorialShow
    );

  end;

  if ( IsAndroid() and ( wiz.Android.GetPlatform() == "google" ) ) then

    local is_connected = wiz.Android.GooglePlayIsConnected();

    if ( not is_connected ) then

      v_interface.panel.right.SetDelegates( {
        Google_OnTap = private.GooglePlayTap
      } );
      
      v_interface.Scheme_AddButton(
        "map",
        "right",
        { name = "google" },
        not public.IsOnline()
      );

    end;

    v_settings.GooglePlay_Setup( is_connected, private.GooglePlayTap );

    ne.msg.Subscribe(
      "Event_GooglePlay_LoggedStateChanged",
      private.OnGooglePlayLoggedStateChanged
    );

  end;

  v_interface.Scheme_AddButton( "map", "right", { name = "ccs_dummy" } );

end;
--------------------------------------------------------------------------------
function public.Splash_Show ( obj_splash )

  if ( wiz.IsCheatsEnabled() ) then

    ModLoad( "assets/cheater/mod_cheater" );
    mediator.SubscribeOnce( "Screen_Open", cheater.Show );

    local ready = m_game.SilentContent_IsReady();
    cheater.ShowSilentContent( ready );

  end;

  v_tutorial.Start();

  local time = common.time.GetCurrent();
  v_player.Metrics_SetLaunchData( time );
  m_game.Session_StartTime( time );

  local firstlaunch_info = m_game.FirstLaunch_GetInfo();

  if ( firstlaunch_info ) then

    private.Splash_ShowFirst( firstlaunch_info );

  else

    private.Splash_ShowNext( obj_splash );

  end;

end;
--------------------------------------------------------------------------------
function private.Splash_ShowFirst ( stage_last )

  v_social.Start();

  if ( v_hearts.GetCount() < 1 ) then

    -- JK: if player can't pass further 1 stage and have lost all hearts,
    -- restore a heart to them to avoid forcing them to pay for the heart

    --DbgTrace( "[ INFO ][ v_game ] Restore 1 heart for the weak." );
    v_hearts.TryChangeCount( 1, "only_for_the_weak" );

  end;

  v_stage.Stage_ForceStart( stage_last, true );

end;
--------------------------------------------------------------------------------
function private.Splash_ShowNext ( obj_splash )

  private.splash_screen_active = true;

  v_social.Start();
  int_game_start.Attach( obj_splash );
  int_game_start.BtnPlay_Show( private.Play );

  if ( IsPC() ) then

    int_game_start.BtnExit_Show( private.TryQuit );

  end;

  if ( IsMobile() ) then

    int_game_start.Social_Show();
    private.Splash_SetSocial( v_social.IsLogged() );

    mediator.Subscribe(
      "Model_Social_LoginContinue",
      private.Splash_OnSocialLogin
    );

  end;

  if ( IsAndroid() and wiz.Android.GetPlatform() == "google" ) then

    int_game_start.GooglePlay_Show();
    private.Splash_SetGooglePlay( wiz.Android.GooglePlayIsConnected() );

    ne.msg.Subscribe(
      "Event_GooglePlay_LoggedStateChanged",
      private.Splash_OnGoogleLoggedStateChanged
    );

    if ( GetGooglePlayLogged() ) then private.GooglePlaySignIn(); end;

  end;

end;
--------------------------------------------------------------------------------
function private.Splash_OnSocialLogin ()

  private.Splash_SetSocial( true );

end;
--------------------------------------------------------------------------------
function private.Splash_SetSocial ( is_logged )

  if ( is_logged ) then

    int_game_start.Social_SetFormLogout(
      function () v_social.Logout( "title_screen" ); end
    );

  else

    int_game_start.Social_SetFormLogin(
      v_social.GetLoginReward(),
      function () v_social.Login( "title_screen" ); end
    );

  end;

end;
--------------------------------------------------------------------------------
function private.Splash_OnGoogleLoggedStateChanged ( msg, params )

  private.Splash_SetGooglePlay( params.logged );

end;
--------------------------------------------------------------------------------
function private.Splash_SetGooglePlay ( is_logged )

  if ( is_logged ) then

    int_game_start.GooglePlay_SetFormLogout(
      wiz.Android.GooglePlayGetPlayerName(),
      private.GooglePlayTap
    );

  else

    int_game_start.GooglePlay_SetFormLogin( private.GooglePlayTap );

  end;

end;
--------------------------------------------------------------------------------
function private.Play ()

  if ( IsMobile() ) then

    mediator.Unsubscribe(
      "Model_Social_LoginContinue",
      private.Splash_OnSocialLogin
    );

  end;

  if ( IsAndroid() and wiz.Android.GetPlatform() == "google" ) then

    ne.msg.Unsubscribe(
      "Event_GooglePlay_LoggedStateChanged",
      private.Splash_OnGoogleLoggedStateChanged
    );

  end;

  private.splash_screen_active = false;

  v_screen.Switch( "scr_map" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Setup_UI ()

  local default   = "default";
  local base_path = "assets/content/";

  local function func_GetRes ( item_type, item_data, item_size, section )

    item_type = item_type or default;
    item_data = item_data or default;
    item_size = item_size or "";
    section   = section   or v_sections.Section_NameCurrent();

    if ( item_size ~= "" ) then
      item_size = "_"..item_size;
    end;

    if ( item_type == default ) and
       ( item_data == default ) then
      return "assets/resources/"..default;
    end;

    local content =
    {
      default     = private.ContentPath_GetItemRes,
      gold        = private.ContentPath_GetGoldRes,
      resource    = private.ContentPath_GetItemRes,
      booster     = private.ContentPath_GetBoosterRes,
      skin        = private.ContentPath_GetSkinRes,
      skinpart    = private.ContentPath_GetSkinpartRes,
      quest       = private.ContentPath_GetQuestRes,
      fragment    = private.ContentPath_GetFragmentRes,
      sticker     = private.ContentPath_GetStickerRes,
      chest       = private.ContentPath_GetItemRes,
      stickerpack = private.ContentPath_GetItemRes,
      worker      = private.ContentPath_GetItemRes,
      item        = private.ContentPath_GetItemRes,
      mine        = private.ContentPath_GetMineRes,
      heart       = private.ContentPath_GetItemRes,
      container   = private.ContentPath_GetItemRes,
      character   = private.ContentPath_GetCharacterRes,
      enemy_icon  = private.ContentPath_GetEnemyIconRes,
      potion      = private.ContentPath_GetPotionRes,
      goal        = private.ContentPath_GetGoalRes,

      --FF
      QT_StageItemsCollected = private.ContentPath_GetQuestRes
    };

    return base_path..content[ item_type ]( section, item_data, item_size );

  end;

  local delegates = {
    GetRes                 = func_GetRes,
    IsFullscreen           = v_interface.window.IsFullscreen,
    SetOuterTap            = v_interface.window.SetOuterTap,
    ClearOuterTap          = v_interface.window.ClearOuterTap,
    ShowWindow             = v_interface.window.Show,
    HideWindow             = v_interface.window.Hide,
    SetWindowArea          = v_interface.window.SetArea,
    SetWindowAreaByButtons = v_interface.window.SetAreaByButtons
  };

  common.delegate.Set( "ui", delegates );

end;
--------------------------------------------------------------------------------
function public.RegisterFeature ( data )

  local is_online = public.IsOnline();

  v_features.RegisterFeature( data, is_online );

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function private.Content_OnObsolete ()

  if ( not private.Content_IsReady() ) then

    private.Content_ObsoleteDialog_Show();

  elseif ( not m_game.SilentContent_IsReady() and
           not m_game.FirstLaunch_GetInfo() ) then

    private.SilentContent_TryDownload( true );

  end;

end;
--------------------------------------------------------------------------------
function private.Content_IsReady ()

  local packet_list   = ContentGetPacketList();
  local section_count = m_game.Section_GetAllowedCount();

  if ( not section_count ) then

    section_count = v_player.Progress_SectionCount();

    m_game.Section_SetAllowedCount( section_count );

  end;

  for p, packet in pairs( packet_list ) do

    if ( packet.name and
         packet.force_update and
         packet.state and
         packet.state == "downloading" ) then

      return false;

    end;

  end;

  return true;

end;
--------------------------------------------------------------------------------
function public.Content_ObsoleteDialog_Show ( force_message )

  private.Content_ObsoleteDialog_Show( force_message );

end;
--------------------------------------------------------------------------------
function private.Content_ObsoleteDialog_Show ( force_message )

  -- Show "obsolete" message only on MAP.
  if ( not force_message and ( v_interface.window.IsAnyOnScreen() or
    v_screen.GetCurrentName() == "scr_stage" ) ) then

    return;

  end;

  public.SystemDialog_Show(
    "contentobsolete",
    function ()
      common.Return();
    end
  );

end;
--------------------------------------------------------------------------------
function public.SilentContent_TryDownload ( is_on_obsolete )

  return private.SilentContent_TryDownload ( is_on_obsolete );

end;
--------------------------------------------------------------------------------
function private.SilentContent_TryDownload ( is_on_obsolete )

  -- Already downloaded.
  if ( m_game.SilentContent_IsReady() ) then

    return false;

  end;

  private.has_free_space, private.total_packages_size = m_game.SilentContent_HasFreeSpace();

  if ( not private.has_free_space ) then

    if ( is_on_obsolete ) then

      if ( v_interface.window.IsAnyOnScreen() or
          v_screen.GetCurrentName() == "scr_stage" ) then

        return;

      end;

      public.SystemDialog_Show( "nofreespace_silent", function ()

        private.SilentContent_TryDownload();

      end );

    end;

  else

    m_game.SilentContent_Download();

  end;

end;
--------------------------------------------------------------------------------
function private.SilentContent_OnReady ()

  private.need_restart = true;

end;
--------------------------------------------------------------------------------
function private.SilentContent_OnFailed ()

  private.SilentContent_TryDownload( true );

end;
--------------------------------------------------------------------------------
function private.SilentContent_NeedRestart_Show ()

  if ( v_interface.window.IsAnyOnScreen() or
       v_screen.GetCurrentName() == "scr_stage" ) then

    return;

  end;

  public.SystemDialog_Show(
    "needrestart",
    function ()
      private.need_restart = false;
      common.Save( true );
      m_game.SilentContent_ApplyPackets();
      common.Return();
    end
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.ContentPath_GetItemRes ( section, name, size )

  local package_path = "items/";

  return package_path..name..size;

end;
--------------------------------------------------------------------------------
function private.ContentPath_GetMineRes ( section, name, size )

  local package_path = "map/mines/";

  return package_path.."anm_mine_"..name;

end;
--------------------------------------------------------------------------------
function private.ContentPath_GetGoldRes ( section, name, size )

  name = tern( name == "gold" )( "gold_2", name );

  local package_path = "items/";

  return package_path..name;

end;
--------------------------------------------------------------------------------
function private.ContentPath_GetBoosterRes ( section, name, size )

  local package_path = "boosters/";

  return package_path..name;

end;
--------------------------------------------------------------------------------
function private.ContentPath_GetPotionRes ( section, name, size )

  local package_path = "items/";

  return package_path..name;

end;
--------------------------------------------------------------------------------
function private.ContentPath_GetGoalRes ( section, goal, size )
  
  if ( goal.type == "GainScore" ) then
     return "items/icon_score";
  end;
  
  local res = {
    default  = "items/icon_%s",
    DropItem = section.."/icon_%s_%s"
  };
  
  local resource = res[ goal.type ] or res.default;
  
  if ( goal.name:match( "chest" ) ) then
    resource = res.default;
  end;
  
  return string.format(
    resource,
    string.lower( goal.name ),
    section:match( "minievent_(.*)" ) or section:match( "%d+" )
  );

end;
--------------------------------------------------------------------------------
function private.ContentPath_GetSkinRes ( section, name )

  local skin_packages = m_game.GetSkinsContent();
  local package_path  = skin_packages[ name ];

  return package_path.."/anm_skin_"..name;

end;
--------------------------------------------------------------------------------
function private.ContentPath_GetSkinpartRes ( section, name, size )

  -- JK: omit item postfix for skin name
  local skin_name = name:gsub( "_item", "" );

  local skin_packages = m_game.GetSkinsContent();
  local package_path  = skin_packages[ skin_name ];

  return package_path.."/"..name..size;

end;
--------------------------------------------------------------------------------
function private.ContentPath_GetQuestRes ( section, name, size )

  local package_path = section.."/quests/";

  return package_path..name..size;

end;
--------------------------------------------------------------------------------
function private.ContentPath_GetFragmentRes ( section, name, size )

  -- JK: omit number, e.g. "_1"
  name = name:gsub( "_%d+$", "" );

  local package_path = section.."/fragments/";

  return package_path..name..size;

end;
--------------------------------------------------------------------------------
function private.ContentPath_GetStickerRes ( section, name, size )

  local package_path = "stickers/";

  return package_path..name;

end;
--------------------------------------------------------------------------------
function private.ContentPath_GetCharacterRes ( section, name, size )

  local character_data = v_characters.Character_GetData( name );

  if ( not character_data ) then

    return nil;

  end;

  local package_path = character_data.content.."/";

  return package_path.."anm_character_"..name;

end;
--------------------------------------------------------------------------------
function private.ContentPath_GetEnemyIconRes ( section, name, size )

  local enemy_data = v_enemies.Enemy_GetStaticData( name );

  if ( not enemy_data ) then return nil; end;

  local  package_path = enemy_data.content.."/"..enemy_data.name.."/";
  return package_path.."icon_"..enemy_data.name;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Section_SetAllowedCount ( section_number )

  m_game.Section_SetAllowedCount( section_number );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.OnWaitingStart ()

  ui.screen.Waiting_Show();

end;
--------------------------------------------------------------------------------
function private.OnWaitingFinish ()

  ui.screen.Waiting_Hide();

end;
--------------------------------------------------------------------------------
function public.WaitingShow ()

  private.fake_waiting = true; -- NOTE: used by social
  private.OnWaitingStart();

end;
--------------------------------------------------------------------------------
function public.WaitingHide ()

  private.fake_waiting = false;
  private.OnWaitingFinish();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.OnMemoryWarning ()

  wiz.ClearUsedResource();

  if ( not IsAndroid() ) then
    DbgTrace( "OnMemoryWarning: memory in use by Lua: "..
            collectgarbage( "count" ) );
    collectgarbage( "collect" );
    DbgTrace( "OnMemoryWarning: memory in use by Lua after collectgarbage: "..
              collectgarbage( "count" ) );
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.OnGainFocus ()

  if ( private.fake_waiting ) then return; end;

  public.TryReconnect( false );

  m_game.CheckForNewVersion( private.OnNewVersionAvailable );

  if ( private.need_restart ) then

    -- Don't show restart message, apply content in loader.
    -- private.SilentContent_NeedRestart_Show();

  end;

end;
--------------------------------------------------------------------------------
function private.OnNewVersionAvailable ( msg, params )

  if ( params.available ) then

    public.SystemDialog_Show( "newversion", function ()

      local platform = GetPlatform();

      if ( platform == "pc" ) then
        common.Save( true );
        wiz.Quit();
        return;
      else
        if ( platform == "android" ) then
          platform = wiz.Android.GetPlatform();
        end;
        wiz.bfg.OpenUrl( private.app_url[ platform ] );
      end;

    end, function () end );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.OnMaintenance ( params )

  local maintenance = params.maintenance;
  if ( maintenance == 0 ) then return; end;

  if ( private.maintenance ) then return; end;

  private.maintenance = true;

  if ( maintenance ~= 3 ) then
    public.SystemDialog_Show(
      "maintenance_"..maintenance
    );
  else
    public.SystemDialog_Show(
      "maintenance_"..maintenance,
      function ()
        common.Return();
      end
    );
  end;

end;
--------------------------------------------------------------------------------
function private.OnStorageConflict ( params )

  local clientTime, serverTime = params.clientSaveTime, params.serverSaveTime;

  local func_message_main, func_message_local, func_message_server;

  func_message_main = function ()

    local clientInfoData = private.StorageConflict_GetClientInfo();
    local serverInfoData = private.StorageConflict_GetServerInfo();

    if ( not serverInfoData ) then

      public.SystemDialog_Show(
        "conflictstorage_force_local",
        function ()
          parts.storage.ResolveConflict( "local" );
        end
      );
      return;

    end;

    -- JK: if client time is -1, reset it to default
    clientTime = tern( clientTime < 0 )( 0, clientTime );

    local clientTimeFormatted = os.date( "%Y-%m-%d", clientTime );
    local serverTimeFormatted = os.date( "%Y-%m-%d", serverTime );

    if clientTimeFormatted == serverTimeFormatted then
      clientTimeFormatted = os.date( "%H:%M:%S", clientTime );
      serverTimeFormatted = os.date( "%H:%M:%S", serverTime );
    end;

    common.delegate.Set( "int_dialog_saveconflict", {
      OnFuncLocal = func_message_local,
      OnFuncServer = func_message_server
    } );

    int_dialog_saveconflict.Show(
    { clientTimeFormatted = clientTimeFormatted, clientInfoData = clientInfoData },
    { serverTimeFormatted = serverTimeFormatted, serverInfoData = serverInfoData }
    );

  end;

  func_message_local = function ()

    public.SystemDialog_Show(
      "conflictstorage_local",
      function ()
        parts.storage.ResolveConflict( "local" );
      end,
      function ()
        func_message_main();
      end
    );

  end;

  func_message_server = function ()

    public.SystemDialog_Show(
      "conflictstorage_server",
      function ()
        parts.storage.ResolveConflict( "server" );
        common.Return();
      end,
      function ()
        func_message_main();
      end
    );

  end;

  func_message_main();

end;
--------------------------------------------------------------------------------
function private.StorageConflict_GetClientInfo ()

  local player_info = m_game.StorageConflict_GetClientPlayerInfo();
  local stage_last  = player_info[ 1 ].stage_last;

  local sections = m_game.StorageConflict_GetClientSections();
  if not ( sections ) then
    sections = v_sections.List() or {};
  end;

  local client_data =
  {
    stage_last = m_game.StorageConflict_GetStageNumber( stage_last );
    section    = m_game.StorageConflict_GetLastSection( sections );
    gold       = player_info[ 1 ].gold;
  };

  client_data.data_section = v_sections.Section_StaticData( client_data.section );

  return client_data;

end;
--------------------------------------------------------------------------------
function private.StorageConflict_GetServerInfo ()

  local data = parts.storage.GetConflictSave();

  local player_info = m_game.StorageConflict_GetServerPlayerInfo( data );
  if ( not player_info ) then return nil; end;

  local sections = m_game.StorageConflict_GetServerSections( data );
  if ( not sections ) then return nil; end;

  local stage_last  = player_info[ 1 ].stage_last;
  local server_data =
  {
    stage_last = m_game.StorageConflict_GetStageNumber( stage_last );
    section    = m_game.StorageConflict_GetLastSection( sections );
    gold       = player_info[ 1 ].gold;
  };

  server_data.data_section = v_sections.Section_StaticData( server_data.section );

  return server_data;

end;
--------------------------------------------------------------------------------
function private.OnSessionConflict ()

  parts.net.heartbeat.Stop();
  public.SystemDialog_Show(
    "prioritysession",
    function ()
      common.Return();
    end
  );

end;
--------------------------------------------------------------------------------
function private.OnVersionOutdated ()

  public.SystemDialog_Show(
    "oldclientversion",
    function ()

      local platform = GetPlatform();
      if ( platform == "pc" ) then
        common.Save( true );
        wiz.Quit();
        return;
      else
        if ( platform == "android" ) then
          platform = wiz.Android.GetPlatform();
        end;
        wiz.bfg.OpenUrl( private.app_url[ platform ] );
        common.Return();
      end;

    end
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.OnRaveLoginConflict ( params )

  local func_message_main, func_message_local, func_message_server;

  func_message_main = function ()

    local clientTimeFormatted = os.date( "%Y-%m-%d", params.info_local.saveTime or clientTime );
    local serverTimeFormatted = os.date( "%Y-%m-%d", params.info_server.saveTime or serverTime );

    if clientTimeFormatted == serverTimeFormatted then
      clientTimeFormatted = os.date( "%H:%M:%S", params.info_local.saveTime or clientTime );
      serverTimeFormatted = os.date( "%H:%M:%S", params.info_server.saveTime or serverTime );
    end;

    local clientInfoData = {
      stage_last = 1,
      section = "section_1",
      gold = 10
    };

    if ( params.info_local.collections[ "player/info" ] ) then
      clientInfoData.stage_last = m_game.StorageConflict_GetStageNumber( params.info_local.collections[ "player/info" ][ 1 ].stage_last );
      clientInfoData.gold = params.info_local.collections[ "player/info" ][ 1 ].gold;
    end;
    if ( params.info_local.collections[ "player/sections" ] ) then
      clientInfoData.section = m_game.StorageConflict_GetLastSection( params.info_local.collections[ "player/sections" ] );
    end;

    local serverInfoData = {
      stage_last = 1,
      section = "section_1",
      gold = 10
    };

    if ( params.info_server.collections[ "player/info" ] ) then
      serverInfoData.stage_last = m_game.StorageConflict_GetStageNumber( params.info_server.collections[ "player/info" ][ 1 ].stage_last );
      serverInfoData.gold = params.info_server.collections[ "player/info" ][ 1 ].gold;
    end;
    if ( params.info_server.collections[ "player/sections" ] ) then
      serverInfoData.section = m_game.StorageConflict_GetLastSection( params.info_server.collections[ "player/sections" ] );
    end;

    public.SystemDialog_Show( "conflictstorage",
      func_message_local,
      func_message_server,
      { clientTimeFormatted, clientInfoData },
      { serverTimeFormatted, serverInfoData }
    );

  end;

  func_message_local = function ()

    public.SystemDialog_Show(
      "conflictstorage_local",
      function ()
        params.func_local();
      end,
      function ()
        func_message_main();
      end
    );

  end;

  func_message_server = function ()

    public.SystemDialog_Show(
    "conflictstorage_server",
      function ()
        params.func_server();
        common.Return();
      end,
      function ()
        func_message_main();
      end
    );

  end;

  func_message_main();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.OnConnectLost ( not_silent )

  public.WaitingHide();

  if ( not_silent ) then

    private.ShowConnectionLostDialog();

  end;
  
  v_interface.panel.right.DisableButton( "google" );

  v_interface.ShowNotification(
    ne.strings.Get( "str_ingame_notification_connection_lost" ),
    "offline",
    "fail"
  );

  private.NotifyControllers( private.controllers, "ConnectionLost" );

end;
--------------------------------------------------------------------------------
function private.ShowConnectionLostDialog ()

  public.SystemDialog_Show( "connectionlost", function () end );

end;
--------------------------------------------------------------------------------
function private.OnConnectRestore ()

  v_interface.panel.right.EnableButton( "google" );
  
  v_interface.ShowNotification(
    ne.strings.Get( "str_ingame_notification_connection_restored" ),
    "online",
    "success"
  );

  private.NotifyControllers( private.controllers, "ConnectionRestored" );

end;
--------------------------------------------------------------------------------
function private.OnFeatureUpdated ()

  if ( not public.IsOnline() ) then
    return;
  end;

  local features = v_features.GetRegisterFeatures();
  local controllers = {};

  for i = 1, #features do
    table.insert( controllers, features[ i ].entity );
    if ( features[ i ].func_save ) then
      features[ i ].func_save();
    end;
  end;

  private.NotifyControllers( controllers, "FeatureUpdated" );

end;
--------------------------------------------------------------------------------
function private.NotifyControllers ( group, method_name )

  for i = 1, #group do

    local controller = group[ i ];

    if ( _G[ controller ] and _G[ controller ][ method_name ] ) then
      _G[ controller ][ method_name ]();
    end;

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.TryReconnect ( force_reconnect )

  if ( force_reconnect and not m_game.IsOnline() ) then

    parts.net.heartbeat.Restart();

  end;

end;
--------------------------------------------------------------------------------
function public.Platform ()

  return GetPlatform();

end;
--------------------------------------------------------------------------------
function public.IsOnline ()

  local client_state = m_game.IsOnline();

  if ( client_state == nil ) then

    return private.is_online;

  end;

  return client_state;

end;
--------------------------------------------------------------------------------
function public.IsDebug ()

  return private.is_debug;

end;
--------------------------------------------------------------------------------
function public.IsWidescreen ()

  local  is_widescreen = GetWideScreen();
  return is_widescreen;

end;
--------------------------------------------------------------------------------
function public.GetSettingsVariant ()

  local  settings_variant = m_game.GetSettingsVariant();
  return settings_variant;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Update_GetData ()

  local update_data = m_game.Update_GetData();

  return update_data;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Context_GetUnlockStage ()

  local  unlock_stage = m_game.Context_GetUnlockStage();
  return unlock_stage;

end;
--------------------------------------------------------------------------------
function public.Context_GetShowFunction ( is_reopen, reopen_value )

  local data = parts.features.GetFeatureData( "context_message" );

  if ( data ) then

    local funcs = {
      facebook    = {
        IsAvailable = v_social.LoginDialog_IsAvailable,
        Show        = function ( func_close ) 
          v_social.LoginDialog_Show( "context_message", func_close ); 
        end
      },
      minievent   = {
        IsAvailable = function ()
          local is_available = false;
          if ( _G.v_minievent ) then
            is_available = v_minievent.Context_IsAvailable();
          end;
          return is_available;
        end,
        Show = function ( func_close )
          if ( _G.v_minievent ) then
            v_minievent.Context_TryShowDialog( func_close );
          end;
        end
      },
      starterpack = {
        IsAvailable = v_shop.Starterpack_IsAvailable,
        Show        = v_shop.Starterpack_DialogShow
      },
      shop        = {
        IsAvailable = function () return true; end,
        Show        = function ( func_close ) 
          v_shop.Show( "gold", func_close ); 
        end
      },
      boosters = {
        IsAvailable = function () return true; end,
        Show        = function ( func_close ) 
          v_shop.Show( "boosters", func_close ); 
        end
      },
      send_hearts = {
        IsAvailable = v_social.FriendsHearts.Send.IsAvailable,
        Show        = v_social.FriendsHearts.Send.Show
      }
    };
    
    for i = 1, #data do
      
      local message = data[ i ];

      local is_available = 
        ( message.visible                     ) and
        ( funcs[ message.name ]               ) and
        ( funcs[ message.name ].IsAvailable() );
      
      --DbgTrace( message, is_available, is_reopen );

      if ( is_available ) and
         ( is_reopen    ) then
        if ( not message.reopen            ) or
           ( message.reopen < 1            ) or
           ( reopen_value < message.reopen ) then
          return nil;
        end;
      end;

      if ( is_available               ) and
         ( funcs[ message.name ]      ) and
         ( funcs[ message.name ].Show ) then
        return funcs[ message.name ].Show;
      end;
      
    end;
    
  end;

  return nil;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Content_IsSectionReady ( section_name )

  if ( not section_name ) then

    DbgTrace( "[ V_GAME ][ ERROR ] Try to check unknown section content!" );
    return false;

  end;

  local section_number = tonumber( string.match( section_name, "%d+" ) );

  m_game.Section_SetAllowedCount( section_number );

  if ( public.Content_IsPackageReady( section_name ) ) then

    return true;

  end;

  if ( IsMobile() and not public.IsOnline() ) then

    private.ShowConnectionLostDialog();

  end;

  private.Content_ObsoleteDialog_Show( true );
  return false;

end;
--------------------------------------------------------------------------------
function public.Content_IsPackageReady ( package_name )

  local packet_list = parts.content.GetPacketList();

  for _, packet in pairs( packet_list ) do

    if ( packet.name and packet.name == package_name ) then

      if ( packet.state and packet.state:match( "download" ) ) then

        if ( packet.name:match( "^(minievent_).*$" ) ~= nil ) then

          -- JK: minievents are unavailable on downloading
          return false, packet.state;

        else

          -- JK: other packets are available on downloading if they exists
          return parts.content.IsPacketReady( package_name ), packet.state;

        end;

      end;

      return ( packet.state == "ready" ), packet.state;

    end;

  end;

  -- JK: not found in manifest
  local is_ready = parts.content.IsPacketReady( package_name );

  return is_ready, tern( is_ready )( "ready", nil );

end;
--------------------------------------------------------------------------------
function public.Content_LoadStrings ( section_number )

  private.strings_loaded = private.strings_loaded or {};

  for s = 1, section_number do

    if ( not private.strings_loaded[ s ] ) then

      private.strings_loaded[ s ] = true;
      ne.strings.Load( "assets/content/section_"..s.."/strings_section_"..s );

    end;

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.SilentContent_IsReady ()

  return m_game.SilentContent_IsReady();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.dialog.Info_Show (
    message_type,
    func_accept,
    func_reject,
    params
  )

  local delegates = {
    Accept = func_accept,
    Reject = func_reject
  };

  common.delegate.Set( "int_dialog_info", delegates );

  int_dialog_info.Show( message_type, params );

end;
--------------------------------------------------------------------------------
function public.dialog.Info_Hide ( ... )

  int_dialog_info.Hide( ... );

end;
--------------------------------------------------------------------------------
function public.dialog.Notifications_Show ( reward, func_accept, func_close )

  local delegates = {
    Accept = func_accept,
    Close  = func_close
  };

  common.delegate.Set( "int_dialog_notifications", delegates );

  int_dialog_notifications.Show( reward );

end;
--------------------------------------------------------------------------------
function public.dialog.Notifications_Hide ( ... )

  int_dialog_notifications.Hide( ... );

end;
--------------------------------------------------------------------------------
function public.dialog.Update_Show (
    dialog_type,
    message,
    skin_current,
    func_OnHide
  )

  v_interface.window.Show(
    "int_dialog_update",
    function ()

      int_dialog_update.Show( dialog_type, message, skin_current );

    end,
    nil,
    { OnHide = func_OnHide }
  );

end;
--------------------------------------------------------------------------------
function public.dialog.Deficient_Show (
    deficient_data,
    func_accept,
    func_reject,
    solution
  )

  local delegates = {
    Accept = func_accept,
    Reject = func_reject
  };

  common.delegate.Set( "int_dialog_deficient", delegates );

  v_interface.window.Show(
    "int_dialog_deficient",
    function()

      int_dialog_deficient.Show(
        deficient_data,
        solution,
        v_player.Skin_GetCurrent()
      );

    end
  );

end;
--------------------------------------------------------------------------------
function public.dialog.Deficient_Hide ( func_after )

  int_dialog_deficient.Hide(
    function()
      v_interface.window.Hide( "int_dialog_deficient", func_after );
    end
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
-- Функция пытается найти и возвратить контейнер с именем 'container_name'
-- Возвращает таблицу { name : string, type: string, items : item_array }
-- предметов либо nil, если контейнера с заданным именем не существует.
-- item_array = { name : string, count : int, type : string }
--------------------------------------------------------------------------------
function public.features.Container_GetData ( container_name )

  local data;

  if ( parts.features.IsActive( "containers" ) ) then

    data = parts.features.GetFeatureData( "containers" );

  else

    data = v_features.GetFeatureData( "containers" );

  end;

  if ( data and type( data ) == "table" ) then

    for i = 1, #data do

      if ( data[ i ].name == container_name ) then

        return data[ i ];

      end;

    end;

  end;

  return nil;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.SystemDialog_Show (
    dialog_type,
    func_btn1,
    func_btn2,
    text_params_btn1,
    text_params_btn2
  )

  local is_ingame = ( private.sysmsg_type == "ingame" ) or
                    ( dialog_type == "conflictstorage" ) or
                    ( string.find( dialog_type, "maintenance" ) );

  local is_pc = ( GetPlatform() == "pc" );

  local buttons = {};
  local postfix = tern( is_ingame )( "_ingame", "" );
  postfix = tern( is_pc )( "_pc", postfix );

  if ( func_btn1 ) then

    text_params_btn1 = text_params_btn1 or {};

    table.insert( buttons, {
      title = ne.strings.Get(
        "str_systemmessage_"..dialog_type.."_button1"..postfix,
        text_params_btn1[ 1 ],
        text_params_btn1[ 2 ]
      ),
      action = func_btn1,
      params = text_params_btn1
    } );

  end;

  if ( func_btn2 ) then

    text_params_btn2 = text_params_btn2 or {};

    table.insert( buttons, {

      title = ne.strings.Get(
        "str_systemmessage_"..dialog_type.."_button2"..postfix,
        text_params_btn2[ 1 ],
        text_params_btn2[ 2 ]
      ),
      action = func_btn2,
      params = text_params_btn2
    } );

  end;

  local str     = "str_systemmessage_"..dialog_type;
  local user_id = v_player.UserId();
  local user_id_message = "";
  if ( user_id ~= 0 ) then
    user_id_message = ne.strings.Get( "str_loader_userid", user_id );
  end;

  if ( dialog_type == "nofreespace_silent" ) then
    user_id_message = private.total_packages_size or 50;
  end;

  local params =
  {
    title   = ne.strings.Get( str.."_title" ),
    body    = ne.strings.Get( str.."_body", user_id_message ),
    buttons = buttons
  };

  if ( is_ingame ) then

    params.dialog_type = dialog_type;
    params.id = v_player.UserId();

    if ( is_pc and
       ( dialog_type == "oldclientversion" or
         dialog_type == "newversion" ) ) then

      params.body = ne.strings.Get( str.."_body"..postfix, user_id_message );

    end;

    int_dialog_info.ShowLikeSysMsg( params );

  else

    wiz.ShowSystemMessage( params );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.LocalNotifications_SetData ( notification_type, value )

  -- Fires up when the application is about to move from active to inactive state.
  -- This can occur for certain types of temporary interruptions (such as an
  -- incoming phone call or SMS message) or when the user quits the application
  -- and it begins the transition to the background state.

  common.Trace( "LocalNotifications_SetData" );

  if ( "comeback_level" == notification_type ) then

    private.comeback_level = value;

  end;

end;
--------------------------------------------------------------------------------
function public.SetFakeDataLocalNotification ( notifications )

  private.fake_local_notification = notifications;

end;
--------------------------------------------------------------------------------
function private.LocalNotifications_AddData ()

  common.Trace( "LocalNotifications_AddData" );

  local events = {};

  if ( private.fake_local_notification ) then

    local count = 0;

    for key, value in pairs( private.fake_local_notification ) do
      private[ "LocalNotification_"..key ]( events );
      if ( #events > count ) then
        events[ #events ].time = value.time;
      else
        table.insert( events, value );
      end;
      count = count + 1;
    end;
    private.fake_local_notification = nil;
    wiz.localnotifications.Update( { events = events } );
    return;
  end;

  --- Hearts -------------------------------------------------------------------
  private.LocalNotification_Hearts( events );
  --- Mines --------------------------------------------------------------------
  private.LocalNotification_Mines( events );
  --- Daily Quest --------------------------------------------------------------
  private.LocalNotification_Dailyquest( events );
  --- Level --------------------------------------------------------------------
  private.LocalNotification_ComebackLevel( events );
  --- 12 Hrs timeout -----------------------------------------------------------
  private.LocalNotification_ComebackGame( events );
  --- Carrier ------------------------------------------------------------------
  private.LocalNotification_Carriers( events );
  --- Minievent ----------------------------------------------------------------
  private.LocalNotification_Minievent( events );

  wiz.localnotifications.Update( { events = events } );

end;
--------------------------------------------------------------------------------
function private.LocalNotification_Hearts ( events )

  if ( _G.v_hearts ) then

    local hearts_restore_time = v_hearts.GetFullHeartsRestoreTime();

    if ( hearts_restore_time > 0 ) then

      table.insert( events, { type = "hearts_restored",
                              time = hearts_restore_time } );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.LocalNotification_Mines ( events )

  if ( _G.v_mines ) then

    local mine_ready = v_mines.Mine_GetRandomCooldown();

    if ( mine_ready ) then

      if ( mine_ready.time and mine_ready.time > 0 ) then

        table.insert( events, {
          type  = "mine_ready",
          time  = mine_ready.time,
          param = StringGet( "str_mine_name_notification" )
        } );

      end;

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.LocalNotification_Dailyquest ( events )

  if ( _G.v_dailyquest ) then

    local dailyquest_cooldown = v_dailyquest.GetRemainingTime( "dailyquest" );

    if ( dailyquest_cooldown ) and ( dailyquest_cooldown > 0 ) then

      table.insert( events, { type = "dailyquest_ready",
                              time = dailyquest_cooldown } );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.LocalNotification_ComebackLevel ( events )

  if ( private.comeback_level ) then

    local stage_name = private.comeback_level;
    local param_string;

    if ( stage_name:find( "stage_" ) ) or ( stage_name:find( "_daily" ) ) then
      param_string = common.string.FormatNameStage( stage_name );
    else
      param_string = StringGet( "str_int_dialog_minievent_"..stage_name.."_title" );
    end;

    table.insert( events, {
      type  = "comeback_level",
      time  = 3600,
      param = param_string
    } );

  end;

end;
--------------------------------------------------------------------------------
function private.LocalNotification_ComebackGame ( events )

  table.insert(
    events,
    {
      type = "comeback_over12hours",
      time = 43200
    }
  );

end;
--------------------------------------------------------------------------------
function private.LocalNotification_Carriers ( events )

  if ( _G.v_carriers ) then

    local carrier_event = v_carriers.GetRandomEvent();

    if ( carrier_event ) then

      local carrier_work = v_carriers.Data( carrier_event.name ).resource;

      if ( not carrier_work ) then

        return;

      end;

      local resource_name  = ne.strings.Get( "str_resource_name_"..carrier_work.name );
      local resource_count = carrier_work.count;
      local section_number = v_carriers.GetNumberByName( carrier_event.name );
      local section_name   = ne.strings.Get( "str_int_dialog_worldmap_section_"..section_number );

      local param = section_name..", +"..resource_count.." "..resource_name;

      table.insert(
        events,
        {
          type  = "carrier_ready",
          time  = carrier_event.time,
          param = param
        }
      );

      --DbgTrace( "Register push notification for carrier: "..param );

    else

      --DbgTrace( "Can't register push notification for carrier: no data" );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.LocalNotification_Minievent ( events )

  if ( not _G.v_minievent ) then return; end;

  local event_list = v_minievent.GetEventList();

  --- JK: existing minievent ends soon -----------------------------------------
  for _, event in pairs( event_list ) do

    local notification_name = "minievent_"..event.type.."_ends";
    local notification_data = m_game.LocalNotification_GetData(
      notification_name
    );

    if ( not notification_data ) and
       ( event.type ~= "story" ) then

      local timeout_1day   = 86400;
      local timeout_12hr   = 43200;
      local timeout_1hr    = 3600;
      local current_time   = common.time.GetCurrent();
      local remaining_time = v_minievent.GetEventRemainingTime( event.id );

      if ( remaining_time > timeout_1hr ) then

        local selected_timeout = timeout_1day;

        selected_timeout = tern( remaining_time < timeout_1day )(
          timeout_12hr,
          selected_timeout
        );

        selected_timeout = tern( remaining_time < timeout_12hr )(
          timeout_1hr,
          selected_timeout
        );

        local notification_time    = remaining_time - selected_timeout + 60; -- 1 min overlap
        local notification_expires = notification_time + current_time;

        local param_string = ne.strings.Get(
          "str_int_dialog_minievent_"..event.name.."_title"
        );

        table.insert(
          events,
          {
            type  = notification_name,
            time  = notification_time,
            param = param_string
          }
        );

        local register_data = {
          name = notification_name,
          time = notification_expires
        };

        m_game.LocalNotification_Register( register_data );

      end;

    end;

  end;

  -- JK: new minievent begins soon ---------------------------------------------
  local upcoming_event = v_minievent.GetUpcomingEvent();

  if ( upcoming_event ) then

    local notification_name = "minievent_"..upcoming_event.type.."_begins";
    local notification_data = m_game.LocalNotification_GetData(
      notification_name
    );

    if ( not notification_data ) then

      local current_time         = parts.time.server.Timestamp();
      local notification_time    = upcoming_event.dateStart - current_time;
      local notification_expires = upcoming_event.dateStart;

      local param_string = ne.strings.Get(
        "str_int_dialog_minievent_"..upcoming_event.name.."_title"
      );

      table.insert(
        events,
        {
          type  = notification_name,
          time  = notification_time,
          param = param_string
        }
      );

      local register_data = {
        name = notification_name,
        time = notification_expires
      };

      m_game.LocalNotification_Register( register_data );

    end;

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.OnAndroidBackButtonPressed ()

  if ( private.android_back_button_disabled ) then

    --DbgTrace( "[ INFO ] Android back button is disabled by flag." );
    return;

  end;

  if ( ui.screen.IsLocked() ) then

    --DbgTrace( "[ INFO ] Android back button is disabled by screen lock." );
    --private.Minimize();
    return;

  end;

  if ( int_dialog_info.IsActive() ) then

    local func_close = int_dialog_info.GetCloseFunc();
    local func_ok    = int_dialog_info.GetOkFunc();

    if ( func_close ) then
      func_close();
    elseif ( func_ok ) then
      func_ok();
    else
      private.Minimize();
    end;
    return;

  end;

  local function func_unsubscribe ()

    ne.msg.Unsubscribe(
      "Event_Android_BackPressed",
      private.OnAndroidBackButtonPressed
    );

  end;

  local function func_subscribe ()

    ne.msg.Subscribe(
      "Event_Android_BackPressed",
      private.OnAndroidBackButtonPressed
    );

  end;

  if ( not v_interface.window.IsAnyOnScreen() ) then

    if ( private.splash_screen_active ) then

      private.TryQuit();
      return;

    end;

    if ( v_stage.IsActive() ) then
      
      if ( v_boosters.IsActive() ) then
        v_boosters.Reset();
        return;
      end;

      if ( not v_stage.IsPaused() ) then
        v_stage.Pause();
      end;

      v_stage.DialogExit_TryShow();
      return;

    end;

    if ( v_map.IsActive() and v_map.IsOpened() ) then

        private.TryQuit();

      return;

    end;

    return;

  end;

  local closed, window_name = v_interface.window.TryCloseUpper();

  if ( closed ) then

    func_unsubscribe();

    mediator.SubscribeOnce(
      "Interface_Window_WindowClosedEnd",
      function ()

        func_subscribe();

      end
    );

    return;

  end;

  -- JK: Default action is ignore
  --private.TryQuit();

end;
--------------------------------------------------------------------------------
function private.TryQuit ()

  local delegates = {

    Accept = function ()

      if ( IsAndroid() ) then

        ne.msg.Unsubscribe(
          "Event_Android_BackPressed",
          private.OnAndroidBackButtonPressed
        );
        int_dialog_info.Hide(
          function ()
            private.Minimize();
            ne.msg.Subscribe(
              "Event_Application_GainFocus",
              private.OnFocusGainedAfterMinimize
            );
          end
        );

      elseif ( IsPC() ) then

        int_dialog_info.Hide();
        wiz.Quit();

      elseif ( IsEditor() ) then

        common.DbgCallpoint( "ERROR: try to quit at: "..GetPlatform() )

      end;

    end,

    Reject = function ()

      int_dialog_info.Hide();

    end

  };

  common.delegate.Set( "int_dialog_info", delegates );

  int_dialog_info.Show( "exitconfirm" );

end;
--------------------------------------------------------------------------------
function private.Minimize ()

  wiz.app.Minimize();

end;
--------------------------------------------------------------------------------
function private.OnFocusGainedAfterMinimize ( message, params )

  ne.msg.Unsubscribe( message, private.OnFocusGainedAfterMinimize );
  ne.msg.Subscribe(
    "Event_Android_BackPressed",
    private.OnAndroidBackButtonPressed
  );

end;
--------------------------------------------------------------------------------
function private.OnDialogWinNextButtonPressed ()

  private.android_back_button_disabled = true;

  local function func_ClearFlag ()

    private.android_back_button_disabled = nil;
    --DbgTrace( "[ INFO ] Android back button disabling flag cleared." );

  end;

  local func_OnMapOpened, func_OnMinieventReady;

  func_OnMapOpened = function ()

    mediator.Unsubscribe(
      "View_Map_MapOpened",
      func_OnMapOpened
    );

    mediator.Unsubscribe(
      "Interface_Dialog_Minievent_Ready",
      func_OnMinieventReady
    );

    func_ClearFlag();

  end;

  func_OnMinieventReady = function ()

    mediator.Unsubscribe(
      "View_Map_MapOpened",
      func_OnMapOpened
    );

    mediator.Unsubscribe(
      "Interface_Dialog_Minievent_Ready",
      func_OnMinieventReady
    );

    func_ClearFlag();

  end;

  mediator.Subscribe(
    "View_Map_MapOpened",
    func_OnMapOpened
  );

  mediator.Subscribe(
    "Interface_Dialog_Minievent_Ready",
    func_OnMinieventReady
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
function private.CCS_OnPanelShow ( params )

  if ( not v_tutorial.Dialog_IsActive() and params.ccs_dummy.enabled ) then
    private.ccs_mustbe_showed = true;
    private.CCS_Show();

  end;

end;
--------------------------------------------------------------------------------
function private.CCS_OnPanelHide ( params )

  private.ccs_mustbe_showed = false;
  private.CCS_Hide();

end;
--------------------------------------------------------------------------------
function private.CCS_OnPanelUpdate ( params )

  if ( not params.visible           ) then return; end;
  if ( v_tutorial.Dialog_IsActive() ) then return; end;

  private.ccs_mustbe_showed = true;
  private.CCS_Hide();
  private.CCS_Show();

end;
--------------------------------------------------------------------------------
function private.CCS_OnTutorialShow ()

  private.ccs_mustbe_showed = false;
  private.CCS_Hide();

end;
--------------------------------------------------------------------------------
function private.CCS_OnTutorialHide ()

  if ( v_interface.panel.right.IsVisible() ) then

    --private.CCS_Show();

  end;

end;
--------------------------------------------------------------------------------
function public.CCS_Freeze ( state )
  
  private.ccs_freeze = state;
  
  if ( not private.ccs_freeze ) then
    if ( private.ccs_mustbe_showed ) then
      private.CCS_Show();
    else
      private.CCS_Hide();
    end;
  end;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.CCS_Show ( forced_pos )

  if ( private.ccs_freeze ) then return; end;
  
  if ( not m_game.IsCCSAvailable() or private.first_install ) then return; end;

  local screen = {
    width  = ui.screen.Width(),
    height = ui.screen.Height()
  };

  forced_pos = { screen.width - 50, screen.height - 24 * ui.screen.HeightPercent() };

  local target_pos = forced_pos or ui.obj.Position( "dmy_int_panel_interface_right_ccs" );

  local button = {
    width  = 100,
    height = 100,
    pos_x  = target_pos[ 1 ],
    pos_y  = target_pos[ 2 ]
  };

  local def_params = { scale_x = 1, scale_y = 1 };
  local app_params = ui.obj.Get( "ng_application" ) or def_params;

  wiz.bfg.ccs.ShowButton(
    {
      x = ( button.pos_x - button.width  / 2 ) * app_params.scale_x,
      y = ( button.pos_y - button.height / 2 ) * app_params.scale_y,
      w = button.width  * app_params.scale_x,
      h = button.height * app_params.scale_y
    }
  );

  if ( wiz.IsCheatsEnabled() ) then
    ui.obj.Multiset( {
      {
        "obj_loader_ccs_placeholder",
        {
          active  = true,
          visible = true,
          pos_x   = button.pos_x * app_params.scale_x,
          pos_y   = button.pos_y * app_params.scale_y,
          pos_z   = 60
        }
      },
      { "spr_loader_ccs_placeholder",
        {
          scale_x = button.width  / 2,
          scale_y = button.height / 2
        }
      }
    } );

    ui.obj.Attach( "obj_loader_ccs_placeholder", "ne_origin" );
  end;

  private.ccs_show = true;

end;
--------------------------------------------------------------------------------
function private.CCS_Hide ()

  if ( private.ccs_freeze ) then return; end;
  
  if ( wiz.IsCheatsEnabled() ) then

    ui.obj.Set(
      "obj_loader_ccs_placeholder",
      {
        active  = false,
        visible = false
      }
    );

    ui.obj.Attach( "obj_loader_ccs_placeholder", "scr_loader" );

  end;

  wiz.bfg.ccs.HideButton();

  private.ccs_show = nil;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Cheat_ShowCCS ()

  DbgTrace( "Cheat_ShowCCS" );
  local forced_pos = { ui.screen.Width() * 0.5, ui.screen.Height() * 0.1 };
  private.CCS_Show( forced_pos );

end
--------------------------------------------------------------------------------
function public.Cheat_HideCCS ()

  DbgTrace( "Cheat_HideCCS" );
  private.CCS_Hide();

end
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.GooglePlayTap ()

  if ( wiz.Android.GooglePlayIsConnected() ) then

    private.GooglePlaySignOut();

  else

    private.GooglePlaySignIn();

  end;

end;
--------------------------------------------------------------------------------
function private.GooglePlaySignIn ()

  mediator.Notify( "View_Game_GooglePlayLoginStart" );
  wiz.Android.GooglePlaySignIn();

end;
--------------------------------------------------------------------------------
function private.GooglePlaySignOut ()

  wiz.Android.GooglePlaySignOut();
  SetGooglePlayLogged( false );

end;
--------------------------------------------------------------------------------
function private.OnGooglePlayLoggedStateChanged ( msg, params )

  local is_connected = params.logged;

  if ( is_connected ) then

    v_interface.Scheme_DeleteButton( "map", "right", { name = "google" } );
    mediator.Notify( "View_Game_GooglePlayLoggedIn" ); 

  else

    v_interface.Scheme_AddButton( "map", "right", { name = "google" } );
    mediator.Notify( "View_Game_GooglePlayLoggedOut" );

  end;

  v_settings.GooglePlay_Setup( is_connected, private.GooglePlayTap );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.IsVideoAvailable ()

  local result = wiz.vungle.IsPlayable();

  return result;

end;
--------------------------------------------------------------------------------
function public.ShowVideo ( point, callback )

  ne.msg.Subscribe( "Event_Vungle_AdEnd", private.ShowVideoFinished );

  private.show_video_callback = { point = point, callback = callback };

  ui.screen.Lock( true );

  private.music_effects = {};

  if ( common.music.IsEnabled() ) then
    common.music.Toggle();
    private.music_effects.music = function ()
      common.music.Toggle();
    end;
  end;

  if ( common.sound.IsEnabled() ) then
    common.sound.Toggle();
    private.music_effects.sound = function ()
      common.sound.Toggle();
    end;
  end;

  wiz.vungle.Play();

end;
--------------------------------------------------------------------------------
function private.ShowVideoFinished ( msg, params )

  ne.msg.Unsubscribe( "Event_Vungle_AdEnd", private.ShowVideoFinished );

  ui.screen.Lock( false );

  if ( private.music_effects ) then
    if ( private.music_effects.music ) then
      private.music_effects.music();
    end;
    if ( private.music_effects.sound ) then
      private.music_effects.sound();
    end;
  end;

  private.music_effects = nil;

  if ( params.result ) then
    if ( private.show_video_callback and
         private.show_video_callback.callback ) then
      private.show_video_callback.callback();
    end;
  end;

  local rvwatched, rvpossible = v_player.GetAttemptsVideo();
  local category = ( private.show_video_callback or {} ).point or "unknown";

  wiz.bfg.LogRewardedVideoSeen( category );

  v_log.Kontagent_LogEvent( "video",
  {
    name = "end",
    category = category,
    rvselection = params.result,
    rvwatched   = rvwatched,
    rvpossible  = rvpossible
  } );

  private.show_video_callback = nil;

end;
--------------------------------------------------------------------------------
function public.Kontagent_VideoRefused ( point )

  if ( public.IsVideoAvailable() and
       v_player.IsAvailableAttemptsVideo() and
       private.video_can_watch ) then

    local rvwatched, rvpossible = v_player.GetAttemptsVideo();

    v_log.Kontagent_LogEvent( "video",
    {
      name = "end",
      category = point,
      rvselection = false,
      rvwatched   = rvwatched,
      rvpossible  = rvpossible
    } );

    public.Kontagent_VideoCanWatch( false );

  end;

end;
--------------------------------------------------------------------------------
function public.Kontagent_VideoCanWatch ( value )

  private.video_can_watch = value;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.GameLog_Add ( sEvent, tEventData )

  m_game.GameLog_AddEvent( sEvent, tEventData );

end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function public.Session_GetDuration ()

  local  duration = m_game.Session_GetDuration();
  return duration;

end;
--------------------------------------------------------------------------------
function public.Session_ForceTick ()

  m_game.Session_OnTimer();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.SetNotification ( target, params )

  m_game.SetNotification( target, params );

end;
--------------------------------------------------------------------------------
function public.GetNotification ( target )

  local params = m_game.GetNotification( target );

  return params;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.ShareToFacebook ( title, description, icon, handler, params )

  ne.msg.Subscribe( Event_Social_FacebookShare, private.OnFacebookShare );

  local protocol = ConfigGetInfo( "parts.net.protocol" ) or "http";
  local host = ConfigGetInfo( "parts.content.contentMainHost" ) or "cdnf2p-a.akamaihd.net/FairySaga";
  local image_path = protocol .. "://" ..host.."/files/fbshare/"..icon..".png";
  local category = icon:match( "icon_(.*)" );
  local categories = {
    event_complete = { name = "eventcomplete", link = "event-complete" },
    lvl_1_place    = { name = "highscore", link = "level-highest" },
    lvl_complete   = { name = "level", link = "level-complete" },
    world_complete = { name = "worldcomplete", link = "world-complete" }
  };
  local link = "http://www.bigfishgames.com/daily/sir-match-a-lot/share/?action="..
               categories[ category ].link;
  if ( category == "lvl_1_place" or
       category == "lvl_complete" ) then
    for key, value in pairs( params ) do
      link = link.."&"..key.."="..value;
    end;
  end;

  if ( handler ) then

    private.facebook_share_handler = handler;

  end;

  mediator.Notify( "View_Game_ShareToFacebook" );

  v_log.Kontagent_LogEvent( "social",
  {
    name        = "send",
    description = "share",
    category    = categories[ category ].name
  } );

  wiz.social.ShareFacebook( title, description, image_path, link );

end;
--------------------------------------------------------------------------------
function private.OnFacebookShare ( msg, params )

  ne.msg.Unsubscribe( Event_Social_FacebookShare, private.OnFacebookShare );

  if ( params.result ) then

    if ( private.facebook_share_handler ) then

      private.facebook_share_handler();

    end;

  end;

  private.facebook_share_handler = nil;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.InviteToFacebook ()

  mediator.Notify( "View_Game_InviteToFacebook" );

  wiz.social.SendInvitesToFacebook();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.AutoFix_Send ( msg )

  local onSetSaveCorrupted;
  onSetSaveCorrupted = function ()

    ne.msg.Unsubscribe( "Server_save_onSetSaveCorrupted", onSetSaveCorrupted );

  end;

  ne.msg.Subscribe( "Server_save_onSetSaveCorrupted", onSetSaveCorrupted );
  NetInvokeServerMethod(
    "save",
    "setSaveCorrupted",
    { info = { [ "Auto fix" ] = msg } },
    1
  );

end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function public.GetSettingsVariant ( )
  
  local  settings_variant = v_features.GetFeatureData( "gold_balance" ) or 0;
  return settings_variant;
  
end;
--------------------------------------------------------------------------------