-- name=v_hearts
--------------------------------------------------------------------------------
local TAG = "v_hearts";
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_hearts",
  "int_dialog_hearts",
  "int_dialog_hearts_send",
  "int_panel_hearts_offers"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()
  
  private.this = {
    dialog = {
      main = "int_dialog_hearts"
    }
  };

  private.immortality_feature = {
    name = "prices"
  };

  common.mod.Load( _modules );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Init ()
  
  local funcs = {
    func_get_stats = function ()
      return v_player.Stats_GetPlayerData();
    end,
    func_get_remain_time_speedup = function ()
      return v_potions.GetRemainingTime( "potion_health_speed" );
    end;
  }
  
  m_hearts.Setup( funcs );
  
  private.immortality_feature = {
    name = "prices"
  };

  public.FeatureUpdated();

  v_game.RegisterFeature( 
  { 
    entity = "v_hearts", 
    feature = private.immortality_feature.name
  } );
  
  private.Panel_Init();

  mediator.Subscribe(
    "Model_Hearts_CountChanged",
    private.Update
  );

  mediator.Subscribe(
    "Model_Hearts_CountChangeFailed",
    public.DialogMain_Show
  );

  mediator.Subscribe(
    "Model_Hearts_ImmortalityToggled",
    private.Immortality_OnToggled
  );

  mediator.Subscribe(
    "Model_Potion_RunAction",
    private.RunActionBoost
  );
  
  mediator.Subscribe(
    "Model_Potion_CompleteAction",
    private.CompleteActionBoost
  );
  
  mediator.Subscribe(
    "Model_Player_RecoverFinishedAttemptsVideo",
    private.RecoverAttemptsVideo
  );
  
  v_interface.window.RegisterAsModal( private.this.dialog.main );
  
end;
--------------------------------------------------------------------------------
function private.Panel_Init ()

  common.delegate.Set( "int_panel_interface_left", 
  {
    HeartsEventTimeRemaining = function ()
    
      local time_remain, type_event = public.EventTimeRemaining();
      return time_remain, type_event;
      
    end,
    HeartsIsImmortality = function ()
      
      return public.Immortality_IsActive();
      
    end, 
    HeartsGetCount = function ()
      
      local count = public.GetCount();
      return count;
      
    end,
    HeartsDialogMain = function ()
    
      public.DialogMain_Show( nil, true );
      
    end
  } );
  
  private.Panel_Enable();
  private.Panel_SetCount();
  public.Panel_EnableTimer();
  
  mediator.Subscribe( "Model_Hearts_RecoverStart", public.Panel_EnableTimer );
  mediator.Subscribe( "Model_Hearts_CountChanged", private.Panel_SetCount );

end;
--------------------------------------------------------------------------------
function private.Panel_Enable ()
  
  v_interface.Scheme_AddButton( "map", "left", { name = "hearts" } );
  
end;
--------------------------------------------------------------------------------
function private.Panel_SetCount ()

  v_interface.panel.left.HeartsSetCount();

end;
--------------------------------------------------------------------------------
function public.Panel_EnableTimer ()
  
  v_interface.panel.left.HeartsEnableTimer();
  
end;
--------------------------------------------------------------------------------
function public.Panel_DisableTimer ()
  
  v_interface.panel.left.HeartsDisableTimer();
  
end;
--------------------------------------------------------------------------------
function public.Panel_ButtonName ()
  
 return v_interface.panel.left.HeartsButtonName();
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.RunActionBoost ()
  
  m_hearts.RunActionBoost();
  
end;
--------------------------------------------------------------------------------
function private.CompleteActionBoost ()
  
  m_hearts.CompleteActionBoost();
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.TryChangeCount ( count_delta, reason )
  
  local is_immortality = public.Immortality_IsActive();
  local result, count_new = m_hearts.TryChangeCount( count_delta, reason );

  if ( is_immortality ) then return result; end;

  if ( result ) then
    if ( count_new == 0 ) then
      v_log.Playhaven_LogPlacement( "lives_zero" );
    end;
  end;

  return result;
    
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.GetInfo ()
  
  local info = m_hearts.GetInfo();
  
  return info;
  
end;
--------------------------------------------------------------------------------
function public.GetMaxCount ()

  local  max_count = m_hearts.GetMaxCount();
  return max_count;
  
end;
--------------------------------------------------------------------------------
function public.GetCount ()
  
  local  heart_count = m_hearts.GetCount();
  return heart_count;
  
end;
--------------------------------------------------------------------------------
function public.IsFull ()

  return ( public.GetCount() == public.GetMaxCount() );
  
end;
--------------------------------------------------------------------------------
function public.IsEnough ( count )
  
  local heart_count = public.GetCount();
  
  if ( count > heart_count ) then
    public.DialogMain_Show();
    return false;
  end;
  
  return true;
  
end;
--------------------------------------------------------------------------------
function public.GetSetting_HeartsBasic ()

  local  count = m_hearts.GetSettings().count;
  return count;

end;
--------------------------------------------------------------------------------
function public.GetSetting_HeartsMax ()

  local  count_max = m_hearts.GetSettings().count_max;
  return count_max;

end;
--------------------------------------------------------------------------------
function public.Block ()
  
  m_hearts.Block();
  
end;
--------------------------------------------------------------------------------
function public.Unblock ( type_unblock )

  m_hearts.Unblock( type_unblock );

end;
--------------------------------------------------------------------------------
function public.IsBlock ()
  
  local  is_block = m_hearts.IsBlock();
  return is_block;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.GetFullHeartsRestoreTime ()
  
  local time_full = m_hearts.GetFullHeartsRestoreTime();

  return time_full;

end;
--------------------------------------------------------------------------------
function private.GetFullRestorePrice ()

  local  heart_price = m_hearts.GetSettings().price;
  return heart_price;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.DialogMain_Show ( func_OnHide, disallow_modal )
  
  local func_show_top_panel = function ()
    if ( not v_interface.panel.top.IsVisible() ) then
      v_interface.panel.top.Highlight( "gold" );
      v_interface.panel.top.Show();
    end;
  end;
  
  local func_close_dialog_hearts = function ()
    if ( IsMobile() ) then
      v_social.Score.VisualShow();
    end;
    public.HeartsOffers_Hide( true );
    int_dialog_hearts.Hide();
    v_interface.panel.top.Hide();
  end;
  
  local delegates = {
    ShowTopPanel = function () func_show_top_panel(); end,
    Accept = function ()

      v_game.Kontagent_VideoRefused( "hearts" );
      if ( private.TryBuy() ) then
        func_close_dialog_hearts();
      end;

    end,
    ShowAskDialog = function ()
      
      v_social.FriendsHearts.Ask.Show( func_show_top_panel );
      v_interface.panel.top.Hide();
      
    end,
    ShowSendDialog = function ()

      v_social.FriendsHearts.Send.Show( func_show_top_panel );
      v_interface.panel.top.Hide();
      
    end,
    Close = function ()

      v_game.Kontagent_VideoRefused( "hearts" );
      func_close_dialog_hearts();

    end,
    OnClose = function ()
      
      local func_end = function ()
        local count = public.GetCount();
        local result = ( count > 0 );
        
        if ( func_OnHide ) then
          func_OnHide( result );
        end;
      end;
      
      v_interface.window.Hide( private.this.dialog.main, func_end );
      
    end;
    GetAttemptsVideo = function ()
    
      local count, max_count = v_player.GetAttemptsVideo();
      return count, max_count;
      
    end,
    ShowVideo = function ()
      
      private.ShowVideo();
      
    end,
    IsImmortality = function ()
      
      return public.Immortality_IsActive();
      
    end,
    IsFacebook = function ()
      
      return ( v_game.IsOnline()            ) and
             ( v_social.IsLogged()          ) and
             ( v_social.Friends_IsUpdated() );
      
    end,
    IsInvite = function ()

      return ( v_social.Friends_IsUpdated()  ) and
             ( v_social.Friends_Count() == 0 );
      
    end,
    GetHeartsData = function ()
      
      local count     = public.GetCount();
      local max_count = public.GetMaxCount();
      local price     = private.GetFullRestorePrice();
      
      return count, max_count, price;
      
    end,
    IsEventExist = function ()
      
      return public.IsEventExist();
      
    end,
    IsEventTimeRemaining = function ()
      
      return public.EventTimeRemaining();
      
    end,
    IsVideoAvailable = function ()
      
      local result = v_game.IsVideoAvailable() and v_player.IsAvailableAttemptsVideo();
      return result;
      
    end,
    VideoCanWatch = function ( value )
      
      v_game.Kontagent_VideoCanWatch( value );
      
    end
  };

  common.delegate.Set( private.this.dialog.main, delegates );
  
  local func_start = function ()

    int_dialog_hearts.Show( v_player.Skin_GetCurrent() );

    public.HeartsOffers_Show( true );
    
  end;

  if ( IsMobile() ) then
    v_social.Score.VisualHide();
  end;
  
  v_interface.window.Show(
    private.this.dialog.main,
    func_start,
    { disallow_modal = disallow_modal }
  );

  common.sound.Play( "click_energy" );

end;
--------------------------------------------------------------------------------
function private.DialogMain_Update ( reason )
  
  _G[ private.this.dialog.main ].Update( reason );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.IsEventExist ()
  
  local is_exist = public.Immortality_IsActive() or public.Recover_IsActive();
  
  return is_exist;
  
end;
--------------------------------------------------------------------------------
function public.EventTimeRemaining ()
  
  if ( public.Immortality_IsActive() ) then
    local time_remains = public.Immortality_TimeRemains();
    return time_remains, "immortality";
  end;
  
  if ( public.Recover_IsActive() ) then
    local time_remains = public.Recover_TimeRemains();
    return time_remains, "recover";
  end;
  
  return 0;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.ShowVideo ()
  
  private.ShowVideo();
  
end;
--------------------------------------------------------------------------------
function private.ShowVideo ()
  
  v_game.ShowVideo( "hearts", private.ShowVideoFinished );
  
end;
--------------------------------------------------------------------------------
function private.ShowVideoFinished ()
  
  local count, max_count = v_player.GetAttemptsVideo();
  local attempts = math.min( count + 1, max_count );
  
  public.UpdateAttemptsVideo( attempts );
  
  local heart_data = {
    category = "hearts",
    count    = 1,
    name     = "heart",
    source   = "video_reward"
  };
  
  public.Heart_TryObtain( heart_data );
  
end;
--------------------------------------------------------------------------------
function public.UpdateAttemptsVideo ( attempts )
  
  v_player.UpdateAttemptsVideo( attempts );
  
end;
--------------------------------------------------------------------------------
function public.GetAttemptsVideo ()
  
  local count, max_count = v_player.GetAttemptsVideo();
  return count, max_count;
  
end;
--------------------------------------------------------------------------------
function private.RecoverAttemptsVideo ()
  
  private.DialogMain_Update();
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Recover_IsActive ()
  
  local is_exist = m_hearts.Recover_IsActive();
  
  return is_exist;
  
end;
--------------------------------------------------------------------------------
function public.Recover_TimeRemains ()
  
  local time_remains = m_hearts.Recover_TimeRemains();
  
  return time_remains;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Immortality_Cheat ()

  if ( m_hearts.Immortality_IsActive() ) then

    m_hearts.Immortality_Stop();

  else

    public.Immortality_Start( "cheat" );

  end;

end;
--------------------------------------------------------------------------------
function public.Immortality_IsFeatureActive ()

  return v_features.IsActive( private.immortality_feature.name );

end;
--------------------------------------------------------------------------------
function public.Immortality_IsActive ()

  local is_active = m_hearts.Immortality_IsActive();
  
  return is_active;

end;
--------------------------------------------------------------------------------
function public.Immortality_TimeByLot ( lot_name )

  if ( private.immortality_feature.data and 
       private.immortality_feature.data[ 1 ] and
       private.immortality_feature.data[ 1 ].lots ) then
       
  
    local index = common.IsEntityInList( private.immortality_feature.data[ 1 ].lots, 
                                         { name = lot_name },
                                         "name" );
    if ( index ) then
      return private.immortality_feature.data[ 1 ].lots[ index ].immortality or 0;
    end;
    
  end;

  return 0;

end;
--------------------------------------------------------------------------------
function public.Immortality_Start ( lot_name, lot_id )

  if ( not lot_name ) then return false; end;

  if ( not public.Immortality_IsFeatureActive() ) and
     ( lot_name ~= "cheat"                        ) then

    return false;

  end;

  local time = public.Immortality_TimeByLot( lot_name );
  
  if ( time > 0 ) then
    
    local heart_data = {
      category = "store",
      count    = time,
      name     = "immortality",
      source   = lot_id
    };
    
    public.Heart_TryObtain( heart_data );
    
  end;

end;
--------------------------------------------------------------------------------
function public.Immortality_TimeRemains ()
  
  local time_remains = m_hearts.Immortality_TimeRemains();
  
  return time_remains;
  
end;
--------------------------------------------------------------------------------
function public.Heart_TryObtain ( heart_data )
  
  local result = false;
  
  if ( heart_data.name == "immortality" ) then
    
    result = m_hearts.Immortality_Start( heart_data.count );
    
  else
    
    result = public.TryChangeCount( heart_data.count, heart_data.category );
    
  end;
  
  if ( result ) then

    v_log.Kontagent_LogEvent( "heart", 
    {
      name        = "collect",
      category    = heart_data.category,
      value       = heart_data.count,
      description = heart_data.name,
      source      = heart_data.source
    } );
    
  end;

  return result;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Update ( params )    

  if ( not public.Immortality_IsActive() ) then
    private.DialogMain_Update( params.reason );
  end;

end;
--------------------------------------------------------------------------------
function private.TryBuy ()
  
  local current_count = public.GetCount();
  local maximum_count = public.GetMaxCount();

  if ( current_count >= maximum_count ) then

    return false;

  end;

  local pack_price = private.GetFullRestorePrice();

  if ( v_player.Gold_TryChangeCount( - pack_price ) ) then

    v_log.Kontagent_LogEvent( "economy", 
    {
      description = "premium_currency_spend",
      category    = "hearts",
      value       = - pack_price,
      name        = "restore"
    } );

    local heart_data = {
      category = "hearts",
      count    = maximum_count,
      source   = "restore",
      name     = "heart"
    };
    public.Heart_TryObtain( heart_data );

    return true;

  end;

  return false;
  
end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function public.FeatureUpdated ()

  private.immortality_feature.data = {};

  if ( public.Immortality_IsFeatureActive() ) then

    local data = parts.features.GetFeatureData(
      private.immortality_feature.name
    );

    private.immortality_feature.data = data;

    if ( not private.immortality_feature.subscribed ) then

      mediator.Subscribe(
        "View_Shop_PremiumPurchaseMade",
        function ( params )
          public.Immortality_Start( params.name, params.id );
        end
      );

      private.immortality_feature.subscribed = true;

    end;

  else

    --DbgTrace( "[ VIEW ][ HEARTS ] Immortality feature is disabled." );

  end;

  -- JK: Immortality cheat
  if ( wiz.IsCheatsEnabled() ) then

    local data_lot_cheat = { name = "cheat", immortality = 86400 };
    
    if ( #private.immortality_feature.data == 0 ) then
      private.immortality_feature.data = { { lots = { data_lot_cheat } } };
    else
      if ( private.immortality_feature.data[ 1 ].lots ) then
        table.insert( private.immortality_feature.data[ 1 ].lots, data_lot_cheat );
      else
        private.immortality_feature.data[ 1 ].lots = { data_lot_cheat };
      end;
    end;
      
  end;

end;
--------------------------------------------------------------------------------
function private.Immortality_OnToggled ()

  public.Panel_EnableTimer();
  private.Panel_SetCount();
  private.DialogMain_Update();
  
end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function public.TryAddSocialBonus ()

  if ( public.CanAddSocialBonus() ) then
    
    if ( m_hearts.AddSocialBonus() ) then
    
      v_log.Kontagent_LogEvent( "social", 
      {
        description = "social_login",
        category    = "hearts",
        name        = "increase"
      } );
      
      v_hearts.TryChangeCount( 1, "facebook" );
      
      return true;
    
    end;
  
  end;
  
  return false;

end;
--------------------------------------------------------------------------------
function public.CanAddSocialBonus ()

  if ( IsPC() ) then return false; end;
  
  local hearts_info = public.GetInfo();
  
  local hearts_facebook = hearts_info.facebook;
  local hearts_donated  = hearts_info.donate;
  
  if ( hearts_facebook > 0 ) then return false; end;

  local hearts_base = public.GetSetting_HeartsBasic();
  local hearts_max  = public.GetSetting_HeartsMax();
  
  if ( hearts_base + hearts_donated >= hearts_max ) then return false end;
  
  return true;

end;
--------------------------------------------------------------------------------
function public.TryAddDonateBonus ( count )

  if ( private.CanAddDonateBonus( count ) ) then
  
    if ( m_hearts.AddDonateBonus( count ) ) then
      
      v_hearts.TryChangeCount( count, "donate" );
      
      return true;
    
    end;
  
  end;
  
  return false;

end;
--------------------------------------------------------------------------------
function private.CanAddDonateBonus ( donate )

  local hearts_info = public.GetInfo();

  local facebook = hearts_info.facebook;
  local donated  = hearts_info.donate;

  local base_count = public.GetSetting_HeartsBasic();
  local max_count  = public.GetSetting_HeartsMax();
  
  if ( base_count + facebook + donated + donate > max_count ) then 
  
    return false;
  
  end;
  
  return true;

end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function public.HeartsOffers_Show ( subscribe )

  local lots_available = m_hearts.Lots_GetAvailable(
    v_shop.Purchase_IsRegistered,
    v_social.IsLogged()
  );
    
  if ( #lots_available == 0 ) then
    return;
  end;
  
  local func_obtain = function ( lot_name )
    
    mediator.Subscribe( "Model_Hearts_CountChanged", 
                        public.HeartsOffers_Update );
    mediator.SubscribeOnce( "View_Shop_Purchase_Failed", function () 
      mediator.Unsubscribe( "Model_Hearts_CountChanged", 
                            public.HeartsOffers_Update );
    end );
    
    if ( lot_name:find( "facebook" ) ) then
      v_social.Login( "hearts_offer" );
    else
      v_shop.Hearts_TryBuy( lot_name );
    end;
 
  end;
  
  if ( subscribe ) then
    mediator.Subscribe(
      "Interface_Window_WindowOpened",
      private.HeartsOffers_OnWindowOpened
    );
    
    mediator.Subscribe(
      "Interface_Window_WindowClosedEnd",
      private.HeartsOffers_OnWindowClosed
    );
  end;

  common.delegate.Set( "int_panel_hearts_offers", { Obtain = func_obtain } );

  int_panel_hearts_offers.Show( lots_available );
  
end;
--------------------------------------------------------------------------------
function public.HeartsOffers_Hide ( unsubscribe )
  
  if ( unsubscribe ) then
    mediator.Unsubscribe(
      "Interface_Window_WindowOpened",
      private.HeartsOffers_OnWindowOpened
    );
    
    mediator.Unsubscribe(
      "Interface_Window_WindowClosedEnd",
      private.HeartsOffers_OnWindowClosed
    );
  end;

  int_panel_hearts_offers.Hide();
  
end;
--------------------------------------------------------------------------------
function public.HeartsOffers_Update ()
  
  mediator.Unsubscribe( "Model_Hearts_CountChanged", public.HeartsOffers_Update );

  local lots_available = m_hearts.Lots_GetAvailable(
    v_shop.Purchase_IsRegistered,
    v_social.IsLogged()
  );

  int_panel_hearts_offers.Update( lots_available );
  
end;
--------------------------------------------------------------------------------
function private.HeartsOffers_OnWindowOpened ()

  local upper_name = v_interface.window.UpperName();

  if ( upper_name ~= "int_dialog_stage_fail" ) and
     ( upper_name ~= "int_dialog_hearts" ) then

    public.HeartsOffers_Hide();

  end;

end;
--------------------------------------------------------------------------------
function private.HeartsOffers_OnWindowClosed ()

  local upper_name = v_interface.window.UpperName();

  if ( upper_name == "int_dialog_stage_fail" ) or
     ( upper_name == "int_dialog_hearts" ) then 

    public.HeartsOffers_Show();

  end;

end;
--------------------------------------------------------------------------------