-- name=v_interface
--------------------------------------------------------------------------------
local MODULE = "v_interface";
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------

public.window = {};

public.panel = {
  top = {},
  right = {},
  bottom = {},
  left = {}
};

local _modules = {
  "int_window",
  "int_loading",
  "int_panel_interface_top",
  "int_panel_interface_left",
  "int_panel_interface_right",
  "int_panel_interface_bottom",
  "int_subtitles"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  common.mod.Load( _modules );

  private.interface = {
    enabled = true,
    anm_tme = 0.6,
    bot_tme = 0.6 -- Bottom panel animation time
  };
  
  -- Resource counters
  private.resources =
  {
    gold      = { count =  0, buffer = {} },
    resource1 = { count = -1, buffer = {} },
    resource2 = { count = -1, buffer = {} },
    resource3 = { count = -1, buffer = {} },
    resource4 = { count = -1, buffer = {} }
  };
  
  private.notifications = {
    left = {
    
    },
    right = {
    
    },
    top = {
    
    },
    bottom = {
      player = {
        counter = {},
        handler = private.BottomPanel_PlayerNotification
      },
      shop = {
        counter = {},
        handler = private.BottomPanel_ShopNotification
      },
      stickers = {
        counter = {},
        handler = private.BottomPanel_StickersNotification
      },
      worlds = {
        counter = {},
        handler = private.BottomPanel_WorldsNotification
      }
    }
  };
  
  private.scheme = {
    map = {
      top    = {},
      left   = {},
      right  = {},
      bottom = {}
    },
    stage = {
      top    = {},
      left   = {},
      right  = {},
      bottom = {}
    },
    minievent_section = {
      top    = {},
      left   = {},
      right  = {},
      bottom = {}
    }
  };

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.Init ()

  local delegates = {
    GetCount = function ( resource_name )
      local  count = private.resources[ resource_name ].count;
      return count;
    end
  };
  
  common.delegate.Set( "int_panel_interface_top", delegates );
  
  mediator.Subscribe(
    v_screen.Notifications.ScreenPreOpen,
    private.OnScreenPreOpen
  );  

  mediator.Subscribe(
    v_screen.Notifications.ScreenOpen,
    private.OnScreenOpen
  );
  
  mediator.Subscribe(
    v_screen.Notifications.ScreenClose,
    private.OnScreenClose
  );
  
  mediator.Subscribe(
    "Model_Player_GoldCountChanged",
    private.OnGoldCountChanged
  );

  mediator.Subscribe(
    "Model_Player_ResourceCountChanged",
    private.Resource_OnCountChanged
  );
  
end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.Scheme_AddButton ( screen, panel, button, is_disabled )
  
  --left   = { "hearts", "dailyquest", "minievent" };
  --right  = { "mail", "google", "social", "sale", "ccs_dummy" };
  --bottom = { "options", "player", "shop", "stickers", "globalmap" };
  --top    = { "none", "default", "gold_only" };

  local button_name = tern( button.type )( button.type, button.name );

  if ( not private.scheme[ screen ][ panel ][ button_name ] ) then
    private.scheme[ screen ][ panel ][ button_name ] = button;
  end;

  if ( is_disabled ) then return; end;
  
  if ( v_screen.GetCurrentName() == "scr_"..screen                  ) and
     ( not private.scheme[ screen ][ panel ][ button_name ].enabled ) then
    public.panel[ panel ].EnableButton( button.name, button.type );
    private.scheme[ screen ][ panel ][ button_name ].enabled = true;
  end;

end;
--------------------------------------------------------------------------------
function private.IsAddedToScheme ( panel, button_name )
  
  local screen = v_screen.GetCurrentName():gsub( "scr_", "" );
  if ( not private.scheme[ screen ] ) then
    return false;
  end;
  return private.scheme[ screen ][ panel ][ button_name ];
  
end;
--------------------------------------------------------------------------------
function public.Scheme_DeleteButton ( screen, panel, button )
  
  local button_name = tern( button.type )( button.type, button.name );
  
  if ( private.scheme[ screen ][ panel ][ button_name ] ) then
    
    if ( v_screen.GetCurrentName() == "scr_"..screen              ) and
       ( private.scheme[ screen ][ panel ][ button_name ].enabled ) then
      public.panel[ panel ].DisableButton( button.name, button.type );
      private.scheme[ screen ][ panel ][ button_name ].enabled = false;
    end;
    
    private.scheme[ screen ][ panel ][ button_name ] = nil;
    
  end;

end;
--------------------------------------------------------------------------------
function private.Scheme_Show ( func_after )

  private.TopPanel_Show();
  private.BottomPanel_Show( func_after );
  private.SidePanels_Show();

end;
--------------------------------------------------------------------------------
function private.Scheme_Hide ( func_after )

  private.TopPanel_Hide();
  private.BottomPanel_Hide( func_after );
  private.SidePanels_Hide();

end;
--------------------------------------------------------------------------------
function public.Show ( params, func_after )
  
  params = params or {};
  
  if ( params.freeze ~= nil ) then private.freeze = params.freeze; end;

  if ( private.freeze ) then
    
    --common.DbgCallpoint( "[ ERROR ] Interface is locked! Can't Show." );
    return false;
    
  end;

  if ( not private.interface.visible ) then
    
    --common.DbgCallpoint( "set visible" );
    
    private.interface.visible = true;

    local func_end = function ()

      mediator.Notify( "View_Interface_Visible" );
      ui.screen.Lock( false );
      
      if ( func_after ) then func_after(); end;

    end;

    private.Scheme_Show( func_end );
    return true;

  elseif ( func_after ) then

    --common.DbgCallpoint( "[ WARNING ][ v_interface ] Already visible!" );
    if ( func_after ) then func_after(); end;
    return false;

  end;

  return false;

end;
--------------------------------------------------------------------------------
function public.Hide ( params, func_after )
  
  params = params or {};
  local screen = params.screen;
  local freeze = params.freeze;
  
  if ( freeze ~= nil ) then
    
    private.freeze = freeze;
    
  end;  
  
  if ( private.freeze ) then
    
    --common.DbgCallpoint( "[ ERROR ] Interface is locked! Can't Hide." );
    return;
    
  end;

  if ( private.interface.visible ) then
    
    --common.DbgCallpoint( "set invisible" );

    private.interface.visible = false;

    local func_end = function ()

      if ( func_after ) then func_after(); end;
    
      mediator.Notify( "View_Interface_Invisible" );

    end;
    
    private.Scheme_Hide( func_end );
    
  elseif ( func_after ) then

    --common.DbgCallpoint( "[ ERROR ] Interface is already invisible!" );
    if ( func_after ) then func_after(); end;

  end;

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.IsVisible ()
  
  return private.interface.visible;
  
end;
--------------------------------------------------------------------------------
function public.Enable ( show, point, func_end )
  
  --DbgTrace( "Interface_Enable: "..tostring( point ) );

  if ( not v_map.IsOpened()                      ) or
     ( v_map.IsPending() == "ConstructBuildAnim" ) then

    --DbgTrace( "Interface_Enable: blocked" );

    return;

  end;

  private.interface.enabled = true;
  
  if ( private.interface.visible ) then
    
    ui.screen.Lock( false );
    
  end;
  
  if ( show ) then

    public.Show( private.TopPanel_Unlock, func_end );

  end;

end;
--------------------------------------------------------------------------------
function public.Disable ( hide )

  if ( hide ) then

    public.Hide();

  end;

  private.interface.enabled = false;

end;
--------------------------------------------------------------------------------
function public._Show ( func_OnStart, func_OnEnd )

  if ( private.interface.enabled ) then
    
    if not ( private.interface.visible ) then

      private.interface.visible = true;

      if ( func_OnStart ) then

        func_OnStart();

      end;

      local func_End = function ()

        mediator.Notify( "View_Map_InterfaceVisible" );
        ui.screen.Lock( false );
        
        if ( func_OnEnd ) then func_OnEnd(); end;

      end;

      private.TopPanel_Show();
      private.BottomPanel_Show( func_End );
      private.SidePanels_Show();
    
    else
    
      -- Check this!
      ( func_OnEnd or function () end )();
        
    end;
      
  elseif ( IsEditor() and func_OnEnd ) then

    common.DbgCallpoint( "Interface disabled, but tries to show!" );
      
  end;

end;
--------------------------------------------------------------------------------
function public._Hide ( func_OnStart, func_OnEnd )

  if --[[( private.interface.enabled ) and]] ( private.interface.visible ) then

    private.interface.visible = false;
    
    if ( func_OnStart ) then

      func_OnStart();

    end;

    local func_End = function ()

      if( func_OnEnd ) then func_OnEnd(); end;
    
      mediator.Notify( "View_Map_InterfaceInvisible" );

    end;

    private.TopPanel_Hide();
    private.BottomPanel_Hide( func_End );

    private.SidePanels_Hide();

  end;

end;
--------------------------------------------------------------------------------
function public.Off_TopPanel ()
  
  int_panel_interface_top.Off();
  
end;
--------------------------------------------------------------------------------
function public.Off ()

  int_panel_interface_top.Off();
  int_panel_interface_left.Off();
  int_panel_interface_bottom.Off();
  int_panel_interface_right.Off();

  public.Disable();

end;
--------------------------------------------------------------------------------
function public.Input_On ()

  int_panel_interface_top.Input_On();
  int_panel_interface_left.Input_On();
  int_panel_interface_bottom.Input_On();
  int_panel_interface_right.Input_On();

end;
--------------------------------------------------------------------------------
function public.Input_Off ()

  int_panel_interface_top.Input_Off();
  int_panel_interface_left.Input_Off();
  int_panel_interface_bottom.Input_Off();
  int_panel_interface_right.Input_Off();

end;
--------------------------------------------------------------------------------
function public.Lock( state )

  ui.screen.Lock( state );

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.window.Show ( window_name, func_end, params, delegates )

  if ( delegates ) then

    common.delegate.Set( window_name, delegates );

  end;

  int_window.Show( window_name, func_end, params );

  if ( ui.popup.Cloud_IsOnScreen() ) then
    ui.popup.Cloud_Hide ( true );
  end;
  
end;
--------------------------------------------------------------------------------
function public.window.AddToQueue ( window_name, func_end, params, delegates )

  if ( delegates ) then

    common.delegate.Set( window_name, delegates );

  end;

  int_window.AddWindowToQueue( window_name, func_end, params );

  if ( ui.popup.Cloud_IsOnScreen() ) then
    ui.popup.Cloud_Hide ( true );
  end;
  
end;
--------------------------------------------------------------------------------
function public.window.Hide ( window_name, func_end, skip_map_enable )

  int_window.Hide( window_name, func_end, skip_map_enable );

end;
--------------------------------------------------------------------------------
function public.window.HideAll ( func_end, skip_map_activation )

  int_window.HideAll( func_end, skip_map_activation );

end;
--------------------------------------------------------------------------------
function public.window.IsAnyOnScreen ()

  local is_on_screen = int_window.IsAnyWindowOnScreen();

  return is_on_screen;

end;
--------------------------------------------------------------------------------
function public.window.IsOnScreen ( window_name )

  local is_on_screen = int_window.IsOnScreen( window_name );

  return is_on_screen;

end;
--------------------------------------------------------------------------------
function public.window.IsFullscreen ( window_name )

  local is_fullscreen = int_window.IsFullscreen( window_name );

  return is_fullscreen;

end;
--------------------------------------------------------------------------------
function public.window.UpperName ()

  local  uppermost_window_name = int_window.Uppermost_Name();
  return uppermost_window_name;

end;

--------------------------------------------------------------------------------
function public.window.LowermostName ()

  local  lowermost_window_name = int_window.Lowermost_Name();
  return lowermost_window_name;

end;
--------------------------------------------------------------------------------
function public.window.TryCloseUpper ()

  local is_closed = int_window.TryCloseUpperWindow();

  return is_closed;

end;
--------------------------------------------------------------------------------
function public.window.RegisterAsFullscreen ( window_name )

  int_window.RegisterAsFullscreen( window_name );

  if ( window_name ~= v_tutorial.Dialog_Name() and
       window_name ~= "int_dialog_stage_fortunewheel" ) then

    local max = ui.screen.MaxDimension();

    int_window.SetArea(
      "obj_"..window_name,
      { -0.5 * max, -0.5 * max, max, max }
    );

  end;

end;
--------------------------------------------------------------------------------
function public.window.RegisterAsModal ( window_name )

  int_window.RegisterAsModal( window_name );

end;
--------------------------------------------------------------------------------
function public.window.RegisterAsOverlay ( window_name )

  int_window.RegisterAsOverlay( window_name );

end;
--------------------------------------------------------------------------------
function public.window.SetOuterTap ( window_name )

  int_window.OuterTap_Set( window_name );

end;
--------------------------------------------------------------------------------
function public.window.ClearOuterTap ()

  int_window.OuterTap_Clear();

end;
--------------------------------------------------------------------------------
function public.window.EnableOuterTap ()

  int_window.OuterTap_Enable();

end;
--------------------------------------------------------------------------------
function public.window.DisableOuterTap ()

  int_window.OuterTap_Disable();

end;
--------------------------------------------------------------------------------
function public.window.SetArea ( window_name, area, multiset )

  int_window.SetArea( window_name, area, multiset );

end;
--------------------------------------------------------------------------------
function public.window.SetAreaByButtons ( obj_window, btn_action, btn_close )

  int_window.SetAreaByButtons( obj_window, btn_action, btn_close );

end;
--------------------------------------------------------------------------------
function public.window.GetAcceptHandler ( window_name )

  local  handler = int_window.Handler_Accept( window_name );
  return handler;

end;
--------------------------------------------------------------------------------
function public.window.GetCloseHandler ( window_name )

  local  handler = int_window.Handler_Close( window_name );
  return handler;

end;
--------------------------------------------------------------------------------
function public.window.Maximize ( window_name )

  int_window.Maximize( window_name );
  
end;
--------------------------------------------------------------------------------
function public.window.Minimize ( window_name )

  int_window.Minimize( window_name );

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.ShowNotification ( ... )

  int_panel_interface_top.ShowNotification( ... );

end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.Loading_Show ( need_fade, func_continue )

  local func_notify = function ()
    
    mediator.Notify( "View_Interface_LoadingSplashIsShown" );
    
  end;
  
  int_loading.Show( need_fade, func_continue, func_notify );

end;
--------------------------------------------------------------------------------
function public.Loading_Hide ( need_fade, func_after )

  int_loading.Hide( need_fade, func_after );

end;
--------------------------------------------------------------------------------
function public.Loading_SetProgress ( relative, from_abs, to_abs )

  int_loading.SetProgressBar( relative, from_abs, to_abs );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.panel.left.GetName ()

  return "int_panel_interface_left";

end;
--------------------------------------------------------------------------------
function public.panel.left.Show ( highlight_hearts )

  private.LeftPanel_Show( highlight_hearts );

end;
--------------------------------------------------------------------------------
function public.panel.left.Hide ( func_end )

  private.LeftPanel_Hide( func_end );

end;
--------------------------------------------------------------------------------
function public.panel.left.SetDelegates ( delegates )

  common.delegate.Set( "int_panel_interface_left", delegates );

end;
--------------------------------------------------------------------------------
function public.panel.left.EnableButton ( button_name, button_type )

  local enabler = {
    hearts     = int_panel_interface_left.Hearts_Enable,
    dailyquest = int_panel_interface_left.Dailyquest_Enable,
    minievent  = int_panel_interface_left.Minievent_Enable,
    countdown  = int_panel_interface_left.Countdown_Enable,
  };

  if ( enabler[ button_name ] ) then
    enabler[ button_name ]( button_type );
  else
    DbgTrace(
      "[ ERROR ][ v_interface ] No such button: "..tostring( button_name )
    );
  end;

end;
--------------------------------------------------------------------------------
function public.panel.left.DisableButton ( button_name, button_type )

  local disabler = {
    hearts     = int_panel_interface_left.Hearts_Disable,
    dailyquest = int_panel_interface_left.Dailyquest_Disable,
    minievent  = int_panel_interface_left.Minievent_Disable,
    countdown  = int_panel_interface_left.Countdown_Disable,
  };

  if ( disabler[ button_name ] ) then
    disabler[ button_name ]( button_type );
  else
    DbgTrace(
      "[ ERROR ][ v_interface ] No such button: "..tostring( button_name )
    );
  end;

end;
--------------------------------------------------------------------------------
function public.panel.left.HeartsSetCount ()

  int_panel_interface_left.Hearts_SetCount();

end;
--------------------------------------------------------------------------------
function public.panel.left.HeartsEnableTimer ()

  int_panel_interface_left.Hearts_EnableTimer();

end;
--------------------------------------------------------------------------------
function public.panel.left.HeartsDisableTimer ()

  int_panel_interface_left.Hearts_DisableTimer();

end;
--------------------------------------------------------------------------------
function public.panel.left.HeartsButtonName ()

  return int_panel_interface_left.Hearts_GetButtonObjName();

end;
--------------------------------------------------------------------------------
function public.panel.left.DailyquestSetNotifications ( count )

  int_panel_interface_left.Dailyquest_SetNotifications( count );

end;
--------------------------------------------------------------------------------
function public.panel.left.DailyquestEnableTimer ( func_seconds )

  int_panel_interface_left.Dailyquest_EnableTimer( func_seconds );

end;
--------------------------------------------------------------------------------
function public.panel.left.MinieventToggleTray ()

  int_panel_interface_left.Minievent_ToggleTray();

end;
--------------------------------------------------------------------------------
function public.panel.left.MinieventEnableSlot (
    slot_index,
    icon_name,
    str_title,
    func_tap,
    func_seconds
  )

  int_panel_interface_left.Minievent_EnableSlot(
    slot_index,
    icon_name,
    str_title,
    func_tap,
    func_seconds
  );

end;
--------------------------------------------------------------------------------
function public.panel.left.MinieventDisableSlot ( slot_index )

  int_panel_interface_left.Minievent_DisableSlot( slot_index );

end;
--------------------------------------------------------------------------------
function public.panel.left.DecorationEnable ( entity_name )
  
  int_panel_interface_left.Decoration_Enable( entity_name );
  
end;
--------------------------------------------------------------------------------
function public.panel.left.DecorationDisable ()
  
  int_panel_interface_left.Decoration_Disable();
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.panel.right.GetName ()

  return "int_panel_interface_right";

end;
--------------------------------------------------------------------------------
function public.panel.right.Show ()

  private.RightPanel_Show();

end;
--------------------------------------------------------------------------------
function public.panel.right.Hide ( func_end )

  private.RightPanel_Hide( func_end );

end;
--------------------------------------------------------------------------------
function public.panel.right.SetDelegates ( delegates )

  common.delegate.Set( "int_panel_interface_right", delegates );

end;
--------------------------------------------------------------------------------
function public.panel.right.EnableButton ( button_name, button_type )

  local enabler = {
    sale   = int_panel_interface_right.Sale_Enable,
    google = int_panel_interface_right.Google_Enable,
    ccs_dummy = int_panel_interface_right.CCS_Enable,
    mail   = int_panel_interface_right.Mail_Enable,
    social = int_panel_interface_right.Social_Enable,
  };

  if ( enabler[ button_name ] ) and ( private.IsAddedToScheme( "right", button_name ) ) then
    enabler[ button_name ]( button_type );
  else
    Log.e( "No button found: "..tostring( button_name ) );
  end;

end;
--------------------------------------------------------------------------------
function public.panel.right.DisableButton ( button_name, button_type )

  local disabler = {
    sale   = int_panel_interface_right.Sale_Disable,
    google = int_panel_interface_right.Google_Disable,
    ccs_dummy = int_panel_interface_right.CCS_Disable,
    mail   = int_panel_interface_right.Mail_Disable,
    social = int_panel_interface_right.Social_Disable,
  };

  if ( disabler[ button_name ] ) and ( private.IsAddedToScheme( "right", button_name ) ) then
    disabler[ button_name ]( button_type );
  else
    Log.e( "No button found: "..tostring( button_name ) );
  end;

end;
--------------------------------------------------------------------------------
function public.panel.right.IsVisible ()

  local  is_visible = int_panel_interface_right.IsVisible();
  return is_visible

end;
--------------------------------------------------------------------------------
function public.panel.right.QuestsButtonName ()

  local  button_name = int_panel_interface_right.Quests_GetButtonObjName();
  return button_name;

end;
--------------------------------------------------------------------------------
function public.panel.right.SaleWarn ( count )

  int_panel_interface_right.Sale_Warn( count );

end;
--------------------------------------------------------------------------------
function public.panel.right.DecorationEnable ( entity_name )
  
  int_panel_interface_right.Decoration_Enable( entity_name );
  
end;
--------------------------------------------------------------------------------
function public.panel.right.DecorationDisable ()
  
  int_panel_interface_right.Decoration_Disable();
  
end;
--------------------------------------------------------------------------------
function public.panel.right.MailSetNotifications ( count )

  int_panel_interface_right.Mail_SetNotifications( count );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.panel.top.GetName ()

  return "int_panel_interface_top";

end;
--------------------------------------------------------------------------------
function public.panel.top.Show ( lock, highlight_resource )

  private.TopPanel_Show( lock, highlight_resource );

end;
--------------------------------------------------------------------------------
function public.panel.top.Hide ( unlock, func_exec )

  private.TopPanel_Hide( unlock, func_exec );

end;
--------------------------------------------------------------------------------
function public.panel.top.Lock ()

  private.TopPanel_Lock();

end;
--------------------------------------------------------------------------------
function public.panel.top.Unlock ()

  private.TopPanel_Unlock();

end;
--------------------------------------------------------------------------------
function public.panel.top.Highlight ( resource )
  
  private.TopPanel_Highlight( resource );
  
end;
--------------------------------------------------------------------------------
function public.panel.top.IsVisible ()
  
  return private.TopPanel_IsVisible();
  
end;
--------------------------------------------------------------------------------
function public.panel.top.IsLocked ()
  
  return private.TopPanel_IsLocked();
  
end;
--------------------------------------------------------------------------------
function public.panel.top.GoldCounterName ()
  
  return "obj_int_panel_interface_top_gold";
  
end;
--------------------------------------------------------------------------------
function public.panel.top.DecorationEnable ( entity_name )
  
  int_panel_interface_top.Decoration_Enable( entity_name );
  
end;
--------------------------------------------------------------------------------
function public.panel.top.DecorationDisable ()
  
  int_panel_interface_top.Decoration_Disable();
  
end;
--------------------------------------------------------------------------------
function public.panel.top.SetView ( view_type )

  int_panel_interface_top.View_Set( view_type );

end;
--------------------------------------------------------------------------------
function public.panel.top.EnableButton ( view_type )

  int_panel_interface_top.View_Set( view_type );

end;
--------------------------------------------------------------------------------
function public.panel.top.DisableButton ()

  int_panel_interface_top.View_Reset();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.panel.bottom.GetName ()

  return "int_panel_interface_bottom";

end;
--------------------------------------------------------------------------------
function public.panel.bottom.EnableButton ( button_name )

  local enabler = {
    options  = int_panel_interface_bottom.Options_Enable,
    player   = int_panel_interface_bottom.Player_Enable,
    shop     = int_panel_interface_bottom.Shop_Enable,
    stickers = int_panel_interface_bottom.Stickers_Enable,
    worlds   = int_panel_interface_bottom.Worlds_Enable
  };

  if ( enabler[ button_name ] ) then
    enabler[ button_name ]();
  else
    DbgTrace(
      "[ ERROR ][ v_interface ] No such button: "..tostring( button_name )
    );
  end;

end;
--------------------------------------------------------------------------------
function public.panel.bottom.DisableButton ( button_name )

  local disabler = {
    options  = int_panel_interface_bottom.Options_Disable,
    player   = int_panel_interface_bottom.Player_Disable,
    shop     = int_panel_interface_bottom.Shop_Disable,
    stickers = int_panel_interface_bottom.Stickers_Disable,
    worlds   = int_panel_interface_bottom.Worlds_Disable
  };

  if ( disabler[ button_name ] ) then
    disabler[ button_name ]();
  else
    DbgTrace(
      "[ ERROR ][ v_interface ] No such button: "..tostring( button_name )
    );
  end;

end;
--------------------------------------------------------------------------------
function public.panel.bottom.WorldsWarn ( count, icon_path )

  int_panel_interface_bottom.Worlds_Warn( count, icon_path );

end;
--------------------------------------------------------------------------------
function public.panel.bottom.Notification ( params )
  
  private.Panel_Notification( "bottom", params );
  
end;
--------------------------------------------------------------------------------
function public.panel.bottom.DecorationEnable ( entity_name )
  
  int_panel_interface_bottom.Decoration_Enable( entity_name );
  
end;
--------------------------------------------------------------------------------
function public.panel.bottom.DecorationDisable ()
  
  int_panel_interface_bottom.Decoration_Disable();
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Gold_ShowPanel ()
  
  int_panel_interface_top.Gold_Show( private.resources.gold.count );

end;
--------------------------------------------------------------------------------
function public.Gold_SetCount ( count )

  private.resources.gold.count = count;
  
  public.Gold_ShowPanel();

end;
--------------------------------------------------------------------------------
function public.Gold_Buffer_Out ()

  private.Gold_Buffer_Out();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Resource_ShowPanel ( resource, count )
  
  count = count or private.resources[ resource ].count;
  
  int_panel_interface_top.Resource_Show( resource, count );

end;
--------------------------------------------------------------------------------
function public.Resource_IsPanelActive ( resource )

  int_panel_interface_top.IsActive( resource );

end;
--------------------------------------------------------------------------------
function public.Resource4_SetCount ( count )
  
  private.resources.resource4.count  = count;
  private.resources.resource4.buffer = {};
  
  public.Resource_ShowPanel( "resource4" );
  
end;
--------------------------------------------------------------------------------
function public.Resource_Reload ( counts )

  for id, count in pairs( counts ) do
  
    local resource = "resource"..id;

    private.resources[ resource ].count  = count;
    private.resources[ resource ].buffer = {};
    
    if ( count > 0 ) then
      private.Resource_UpdatePanel( resource, 0, count, nil, true );
    else
      private.Resource_HidePanel( resource );
    end;
  
  end;

end;
--------------------------------------------------------------------------------
function public.Resource_Buffer_Out ( resource, func_end )

  private.Resource_Buffer_Out( resource, func_end );
  
end;
--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.Subtitles_Prepare ()

  int_subtitles.Prepare();

end;
--------------------------------------------------------------------------------
function public.Subtitles_Show (
    section_name,
    subtitle_name,
    subtitles_config,
    func_skip,
    func_after
  )

  local delegates = {
    Skip     = func_skip,
    Continue = func_after
  };
  
  common.delegate.Set( "int_subtitles", delegates );
  
  int_subtitles.Show( 
    section_name.."_"..subtitle_name, 
    subtitles_config
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.OnScreenPreOpen ( screen_name )
  
  --DbgTrace( "["..MODULE.."] OnScreenPreOpen: "..screen_name );

  local screen = screen_name:match( "scr_(%w+)" );

  if ( private.scheme[ screen ] ) then
  
    for panel, buttons in pairs( private.scheme[ screen ] ) do 

      for button, params in pairs( buttons ) do

        v_interface.panel[ panel ].EnableButton( params.name, params.type );
        private.scheme[ screen ][ panel ][ button ].enabled = true;
        
      end;
      
    end;
  
  end;
  
  if ( screen:find( "map" ) ) then
    
    mediator.Subscribe(
      "Interface_Window_WindowOpened",
      private.OnWindowOpen
    );

    mediator.Subscribe(
      "Interface_Window_NoWindows",
      private.OnWindowsClose
    );
    
    mediator.Subscribe(
      "View_Map_Update_Started",
      private.OnUpdateStarted
    );
    
    mediator.Subscribe(
      "View_Map_Update_Finished",
      private.OnUpdateFinished
    );
    
    public.Gold_ShowPanel();
    public.Resource_ShowPanel( "resource1" );
    public.Resource_ShowPanel( "resource2" );
    public.Resource_ShowPanel( "resource3" );
      
  end;

  if ( screen_name == "scr_minievent_section" ) then

    mediator.Subscribe(
      "Interface_Window_WindowOpened",
      private.OnWindowOpen
    );

    mediator.Subscribe(
      "Interface_Window_NoWindows",
      private.OnWindowsClose
    );

    mediator.Subscribe(
      "View_Minievent_Section_Update_Started",
      private.OnUpdateStarted
    );
    
    mediator.Subscribe(
      "View_Minievent_Section_Update_Finished",
      private.OnUpdateFinished
    );

  end;

end;
--------------------------------------------------------------------------------
function private.OnScreenOpen ( screen_name )
  
  if ( screen_name:find( "map" )              ) or
     ( screen_name == "scr_minievent_section" ) then
    
    private.map_opened = true;

  end;

end;
--------------------------------------------------------------------------------
function private.OnScreenClose ( screen_name )
  
  local screen = screen_name:match( "scr_(.*)" );

  if ( private.scheme[ screen ] ) then
    
    for panel, buttons in pairs( private.scheme[ screen ] ) do 

      for button, params in pairs( buttons ) do

        v_interface.panel[ panel ].DisableButton( params.name, params.type );
        private.scheme[ screen ][ panel ][ button ].enabled = false;
        
      end;
      
    end;
    
  end;

  if ( screen:find( "map" ) ) then
    
    private.map_opened = false;

    mediator.Unsubscribe(
      "Interface_Window_WindowOpened",
      private.OnWindowOpen
    );

    mediator.Unsubscribe(
      "Interface_Window_NoWindows",
      private.OnWindowsClose
    );
    
    mediator.Unsubscribe(
      "View_Map_Update_Started",
      private.OnUpdateStarted
    );

    mediator.Unsubscribe(
      "View_Map_Update_Finished",
      private.OnUpdateFinished
    );  

    public.Off();
      
  end;

  if ( screen_name == "scr_minievent_section" ) then

    private.map_opened = false;

    mediator.Unsubscribe(
      "Interface_Window_WindowOpened",
      private.OnWindowOpen
    );

    mediator.Unsubscribe(
      "Interface_Window_NoWindows",
      private.OnWindowsClose
    );

    mediator.Unsubscribe(
      "View_Minievent_Section_Update_Started",
      private.OnUpdateStarted
    );

    mediator.Unsubscribe(
      "View_Minievent_Section_Update_Finished",
      private.OnUpdateFinished
    );

    public.Off();

  end;

end;
--------------------------------------------------------------------------------
function private.OnWindowOpen ( window_name )
  
  public.Hide();

end;
--------------------------------------------------------------------------------
function private.OnWindowsClose ()

  if ( private.map_opened        ) and
     ( not private.map_update    ) and 
     ( not ( v_map.IsPending() == "ConstructBuildAnim" ) ) and
     ( not v_map.IsIntroActive() ) then --optional expression

    private.TopPanel_Unlock();
    public.Show();

  end;

end;
--------------------------------------------------------------------------------
function private.OnUpdateStarted ()

  private.map_update = true;

end;
--------------------------------------------------------------------------------
function private.OnUpdateFinished ()

  private.map_update = nil;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.SidePanels_Show ()

  private.LeftPanel_Show();
  private.RightPanel_Show();

end;
--------------------------------------------------------------------------------
function private.SidePanels_Hide ()

  private.LeftPanel_Hide();
  private.RightPanel_Hide();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.TopPanel_Show ( lock, highlight_resource )
  
  if ( not private.TopPanel_IsVisible() ) then

    if ( highlight_resource ) then
      int_panel_interface_top.Highlight( highlight_resource );
    end;
    
    if ( not v_stage.StageTopPanel_IsVisible() )then
      int_panel_interface_top.Show();
    end;

    if ( lock ) then
      private.TopPanel_Lock();
    end;

  end;

end;
--------------------------------------------------------------------------------
function private.TopPanel_Hide ( unlock, func_exec )

  if ( unlock ) then
    private.TopPanel_Unlock();
  end;

  int_panel_interface_top.Hide( func_exec );
  
end;
--------------------------------------------------------------------------------
function private.TopPanel_Lock ()

  int_panel_interface_top.Lock( true );

end;
--------------------------------------------------------------------------------
function private.TopPanel_Unlock ()

  int_panel_interface_top.Lock( false );

end;
--------------------------------------------------------------------------------
function private.TopPanel_Highlight ( resource )
  
  int_panel_interface_top.Highlight( resource );
  
end;
--------------------------------------------------------------------------------
function private.TopPanel_IsVisible ()
  
  local is_visible = int_panel_interface_top.IsVisible();

  return is_visible;
  
end;
--------------------------------------------------------------------------------
function private.TopPanel_IsLocked ()
  
  local  is_locked = int_panel_interface_top.IsLocked();
  return is_locked;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.LeftPanel_Show ( highlight_button )

  int_panel_interface_left.Show( highlight_button );

end;
--------------------------------------------------------------------------------
function private.LeftPanel_Hide ( func_end )

  int_panel_interface_left.Hide( func_end );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.BottomPanel_Show ( func_end )
  
  int_panel_interface_bottom.Show( func_end );
  
end;
--------------------------------------------------------------------------------
function private.BottomPanel_Hide ( func_end )

  int_panel_interface_bottom.Hide( func_end );

  v_achievements.HidePanel();
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.RightPanel_Show ()
  
  int_panel_interface_right.Show();

end;
--------------------------------------------------------------------------------
function private.RightPanel_Hide ( func_end )

  int_panel_interface_right.Hide( func_end );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Gold_HidePanel ()

  int_panel_interface_top.Gold_Hide();

end;
--------------------------------------------------------------------------------
function private.Gold_UpdatePanel ( count_old, count_new )

  int_panel_interface_top.Gold_Update( count_old, count_new );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.OnGoldCountChanged ( params )

  if ( params.hidden ) then

    table.insert( private.resources.gold.buffer, params.delta );
  
  else
  
    local count_old = private.resources.gold.count;
    local count_new = private.resources.gold.count + params.delta;

    private.resources.gold.count = count_new;
    private.Gold_UpdatePanel( count_old, count_new );
  
  end;

end;
--------------------------------------------------------------------------------
function private.Gold_Buffer_Out ()

  local delta = table.remove( private.resources.gold.buffer, 1 );

  if ( delta ) then
  
    local count_old = private.resources.gold.count;
    local count_new = private.resources.gold.count + delta;
    
    private.resources.gold.count = count_new;
    private.Gold_UpdatePanel( count_old, count_new );
  
  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Resource_HidePanel ( resource )

  int_panel_interface_top.Resource_Hide( resource );

end;
--------------------------------------------------------------------------------
function private.Resource_UpdatePanel (
    resource,
    count_old,
    count_new,
    func_after,
    skip_animation
  )

  if ( not int_panel_interface_top.IsActive( resource ) ) then

    public.Resource_ShowPanel( resource, count_old );

  end;

  local func_end;

  if ( int_panel_interface_top.IsVisible()         ) and
     ( not int_panel_interface_top.IsHighlighted() ) and
     ( resource ~= "resource4"                     ) and
     ( not skip_animation                          ) then
  
    for i = 1, 3 do

      int_panel_interface_top.Resource_GrayIn( "resource"..i );

    end;
    
    func_end = function ()
    
      for i = 1, 3 do

        int_panel_interface_top.Resource_GrayOut( "resource"..i );

      end;
    
    end;
  
  end;
  
  if ( int_panel_interface_top.IsVisible()     ) and
     ( resource == "resource4"                 ) and
     ( int_panel_interface_top.IsHighlighted() ) and
     ( not skip_animation                      ) then

    int_panel_interface_top.Resource_GrayIn( "resource4" );
    func_end = function ()
      int_panel_interface_top.Resource_GrayOut( "resource4" );
      if ( func_after ) then func_after(); end;
    end;
  end;
  
  int_panel_interface_top.Resource_Update(
    resource,
    count_old,
    count_new,
    func_end,
    skip_animation
  );

end;
--------------------------------------------------------------------------------
function private.Resource_OnCountChanged ( params )

  local resource = params.name;

  if ( params.hidden ) then
  
    table.insert( private.resources[ resource ].buffer, params.delta );
  
  else
  
    local count_old = math.max( private.resources[ resource ].count, 0 );
    local count_new = count_old + params.delta;
    private.resources[ resource ].count = count_new;
    
    private.Resource_UpdatePanel( resource, count_old, count_new );
  
  end;

end;
--------------------------------------------------------------------------------
function private.Resource_Buffer_Out ( resource, func_end )

  local delta = table.remove( private.resources[ resource ].buffer, 1 );
  
  if ( delta ) then
  
    local count_old = math.max( private.resources[ resource ].count, 0 );
    local count_new = count_old + delta;
    private.resources[ resource ].count = count_new;
    
    private.Resource_UpdatePanel( resource, count_old, count_new, func_end );
    
  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Panel_NotificationNeedful ( 
  data, 
  type_notification, 
  func_common, 
  func_custom )
  
  if ( type_notification == "common" ) then
    
    local count = 0;
    for key, value in pairs( data ) do
      if ( value.type == "common" ) then
        count = count + value.count;
      end;
    end;
    func_common( count );
    
  else
    
    local customs = {};
    for key, value in pairs( data ) do
      if ( value.type == "custom" ) then
        customs[ value.entity..value.style ] = value;
      end;
    end;
    func_custom( customs );
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.Panel_Notification ( panel, params )
  
  local counter = private.notifications[ panel ][ params.target ].counter;
  local handler = private.notifications[ panel ][ params.target ].handler;
  
  params.data.style = params.data.style or "";
  
  counter[ params.data.entity..params.data.style ] = params.data;
  
  handler( counter, params.data.type );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.BottomPanel_PlayerNotification ( data, type_notification )
  
  private.Panel_NotificationNeedful( 
    data, 
    type_notification,
    int_panel_interface_bottom.Player_SetNotification,
    int_panel_interface_bottom.Player_SetCustomNotification
  );
  
end;
--------------------------------------------------------------------------------
function private.BottomPanel_ShopNotification ( data, type_notification )
  
  private.Panel_NotificationNeedful( 
    data, 
    type_notification,
    int_panel_interface_bottom.Shop_SetNotification,
    int_panel_interface_bottom.Shop_SetCustomNotification
  );
  
end;
--------------------------------------------------------------------------------
function private.BottomPanel_StickersNotification ( data, type_notification )
  
  private.Panel_NotificationNeedful( 
    data, 
    type_notification,
    int_panel_interface_bottom.Stickers_SetNotification,
    function () end
  );
  
end;
--------------------------------------------------------------------------------
function private.BottomPanel_WorldsNotification ( data, type_notification )
  
  private.Panel_NotificationNeedful( 
    data, 
    type_notification,
    int_panel_interface_bottom.Worlds_SetNotification,
    int_panel_interface_bottom.Worlds_SetCustomNotification
  );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------