-- name=v_items
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_items",
  "int_dialog_items_chest",
  "int_dialog_items_gift",
  "int_dialog_reward"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.dialog_chest  = "int_dialog_items_chest";
  private.dialog_gift   = "int_dialog_items_gift";
  private.dialog_reward = "int_dialog_reward";

  common.mod.Load( _modules );

  public.msg =
  {
    ItemObtained = "View_Items_ItemObtain"
  };

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Init ()

  private.chest_reward_func = {
    random = {
      gold        = private.Chest_Random_Gold,
      booster     = private.Chest_Random_Booster,
      skinpart    = private.Chest_Random_Skinpart,
      stickerpack = private.Chest_Random_Stickerpack,
      heart       = private.Chest_Random_Heart,
      resource    = private.Chest_Random_Resource
    },
    check = {
      stickerpack = private.Chest_Check_Stickerpack
    }
  };
  
  v_interface.window.RegisterAsModal( private.dialog_chest );
  v_interface.window.RegisterAsModal( private.dialog_gift );
  v_interface.window.RegisterAsModal( private.dialog_reward );

  private.Reward_TryResolveAll();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.TryObtainMultiple ( rewards, obtain_params )
  
  obtain_params = obtain_params or {};

  for index, reward in pairs( rewards ) do

    public.TryObtain( {
      name       = reward.name,
      type       = reward.type,
      count      = reward.count,
      category   = obtain_params.category,
      stage_name = obtain_params.stage_name,
      silent     = obtain_params.silent,
      source     = obtain_params.source
    } );

  end;
  
end;
--------------------------------------------------------------------------------
function public.TryObtain ( ... )

  local item;
  local args = { ... };
  
  --DbgTrace( "v_items.TryObtain:" );
  --DbgTrace( args );

  -- Check arguments format
  if     ( #args == 1 ) then
  
    item = args[ 1 ];
  
  elseif ( #args == 3 ) then
  
    item = 
    {
      name  = args[ 1 ],
      type  = args[ 2 ],
      count = args[ 3 ]
    };
  
  else
  
    common.DbgCallpoint( "[ ERROR ] Bad arguments!" );
    DbgTrace( args );
    return false;
  
  end;

  if ( item.count     ) and 
     ( item.count < 1 ) then
    Log.w( "Can't obtain "..item.name..": count is < 1" );
    return false;
  end;
  
  -- Try obtain
  local funcs =
  {
    gold        = private.TryObtainGold,
    resource    = private.TryObtainResource,
    skinpart    = private.TryObtainSkinpart,
    booster     = private.TryObtainBooster,
    fragment    = private.TryObtainFragment,
    stickerpack = private.TryObtainStickerpack,
    heart       = private.TryObtainHeart,
    chest       = private.TryObtainChest,
    container   = private.TryObtainContainer,
    potion      = private.TryObtainPotion
  };

  local func_try_obtain = funcs[ item.type ];
  
  if ( func_try_obtain ) then
  
    local is_obtained = func_try_obtain( item );
  
    if ( is_obtained ) then
      --Log.i( "Item obtained:", item );
      mediator.Notify( public.msg.ItemObtained, item );
    else
      Log.w( "Can't obtain item:", item );
    end;
    
    return is_obtained;
  
  else

    common.DbgCallpoint( "[ ERROR ] Bad item type: "..tostring( item.type ) );
    return false;
  
  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Chest_Show (
    chest_data,
    obtain_params,
    ext_rewards,
    func_after,
    func_before,
    ignore_panel
  )

  private.Chest_Show (
    private.dialog_chest,
    chest_data,
    obtain_params,
    ext_rewards,
    func_after,
    func_before,
    ignore_panel
  );

end;

--------------------------------------------------------------------------------
function public.Chest_Generate ( chest_data, content )
  
  return private.Chest_Generate( chest_data, content );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Gift_Show (
    chest_data,
    obtain_params,
    ext_rewards,
    func_after,
    func_before,
    ignore_panel
  )

  private.Chest_Show (
    private.dialog_gift,
    chest_data,
    obtain_params,
    ext_rewards,
    func_after,
    func_before,
    ignore_panel
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Reward_Show ( reward, func_end )

  v_interface.window.AddToQueue(
    private.dialog_reward,
    function () int_dialog_reward.Show( reward ); end,
    nil,
    {
      Close  = function () int_dialog_reward.Hide(); end,
      OnHide = function ()
        v_interface.window.Hide( private.dialog_reward, func_end );
      end
    }
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Chest_Show (
    dialog,
    chest_data,
    obtain_params,
    ext_rewards,
    func_after,
    func_before,
    ignore_panel
  )

  -- JK: default is wood chest
  chest_data = chest_data or { name = "woodchest" };
  
  chest_data.type  = "chest";
  chest_data.count = chest_data.count or 1;

  private.Chest_CheckCorrect( ext_rewards );
  
  local rewards = ext_rewards or private.Chest_Generate( chest_data );

  if ( obtain_params ) then

    -- JK: obtain all the rewards immediately
    public.TryObtainMultiple( rewards, obtain_params );

    if ( ext_rewards ) then

      -- JK: remove the chest from player/rewards collection

      local fingerprint = private.Chest_GetFingerprint(
        chest_data.name,
        ext_rewards
      );

      m_items.Reward_DeleteByFingerprint( fingerprint );

    end;

  end;
  
  local func_show_top_panel = function ( highlight_gold )
    if ( highlight_gold ) then  
      v_interface.panel.top.Highlight( "gold" );
    end;  
    v_interface.panel.top.Show();
  end;
  
  local func_hide_top_panel = function ()
    if ( v_interface.panel.top.IsVisible() ) then
      v_interface.panel.top.Hide( false );
    end;
  end;
  
  local delegates = {

    GetRewards = function ()
  
      local highlight_gold = false;
      local show_top_panel = false;

      for r, reward in ipairs( rewards ) do

        if ( reward.type == "gold" ) then
          show_top_panel = true;
          highlight_gold = true;
        elseif ( reward.type == "resource" ) then
          show_top_panel = true;
          highlight_gold = false;
        end;
        
      end;

      if ( show_top_panel ) then
        func_show_top_panel( highlight_gold );
      end;
      
      return rewards;
  
    end,
  
    CountRewards = function ()
       
      for r, reward in pairs( rewards ) do
      
        if ( reward.type == "gold" ) then
          
          v_interface.Gold_Buffer_Out();
          
        end;
        
        if ( reward.type == "resource" ) then
          
          v_interface.Resource_Buffer_Out( reward.name );
          
        end;
        
      end;
  
    end,
  
    OnShow = func_before,

    OnHide = function ()

      private.Chest_Hide(
        dialog,
        function ()      
          if ( not ignore_panel ) then func_hide_top_panel(); end;
          if ( func_after       ) then func_after();          end;
        end
      );
  
    end
  
  };

  v_interface.window.Show(
    dialog,
    function ()
      _G[ dialog ].Show( chest_data.name, chest_data.factor );
    end,
    nil,
    delegates
  );

end;
--------------------------------------------------------------------------------
function private.Chest_Hide ( dialog, func_end )

  common.delegate.Set(
    dialog,
    {
      OnHide = function () v_interface.window.Hide( dialog, func_end ); end
    }
  );

  _G[ dialog ].Hide();

end;
--------------------------------------------------------------------------------
function private.Chest_Generate ( chest_data, content )

  if ( chest_data.rewards ) then

    for r, reward in pairs( chest_data.rewards ) do

      if ( reward.type == "booster" ) then

        content = content or m_items.Chest_GetSettings( chest_data.name );

        chest_data.rewards[ r ] = 
          private.Chest_Random_OneReward( content, "booster" )

      end;

    end;

    return chest_data.rewards;

  end;

  return private.Chest_Rewards( chest_data, content );

end;
--------------------------------------------------------------------------------
function private.Chest_Rewards ( chest_data, content )

  content = content or m_items.Chest_GetSettings( chest_data.name );
  local rewards = {};

  for reward_type in pairs( content ) do
    
    table.insert( rewards, private.Chest_Random_OneReward( content, reward_type ) );
    
  end;
  
  if ( chest_data.extra_rewards ) then
    
    for i = 1, #chest_data.extra_rewards do
    
      private.Chest_Check_OneReward( chest_data.extra_rewards[ i ], rewards );
      
    end;
    
  end;
  
  return rewards;

end;
--------------------------------------------------------------------------------
function private.Chest_CheckCorrect ( rewards )
  
  if ( not rewards ) then return; end;
  
  local is_unlocked = false;
  
  if ( v_fireflies ) then is_unlocked = v_fireflies.Unlocked(); end;
  if ( is_unlocked ) then return; end;
  
  for i = 1, #rewards do
    
    local reward = rewards[ i ];
    
    if ( reward.type == "stickerpack" ) then
      local reward_instead = m_items.Reward_InsteadStickerpack();
      reward.name  = reward_instead.name;
      reward.type  = reward_instead.type;
      reward.count = reward_instead.count;
      rewards[ i ] = reward;
    end;
      
  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Chest_Random_OneReward ( data_chest, type )
  
  if ( data_chest[ type ] ) and ( #data_chest[ type ] > 0 ) then
    
    return private.chest_reward_func.random[ type ]( data_chest[ type ] ) or 
           private.chest_reward_func.random[ "gold" ]( data_chest[ "gold" ] );
    
  end;
  
  return private.chest_reward_func.random[ "gold" ]( data_chest[ "gold" ] );
  
end;
--------------------------------------------------------------------------------
function private.Chest_Random_Gold ( list )

  local reward = list[ math.random( 1, #list ) ];
  reward.count = reward.count or math.random( reward.min, reward.max );
  reward.min, reward.max = nil, nil;
  return reward;

end;
--------------------------------------------------------------------------------
function private.Chest_Random_Booster ( list )

  local reward = list[ math.random( 1, #list ) ];
  reward.count = reward.count or math.random( reward.min, reward.max );
  reward.min, reward.max = nil, nil;
  
  if ( v_boosters.IsUnlocked( reward.name ) ) then

    return reward;

  end;
  
  return nil;
  
end;
--------------------------------------------------------------------------------
function private.Chest_Random_Skinpart ( list )

  local reward = list[ math.random( 1, #list ) ];
  reward.count = reward.count or math.random( reward.min, reward.max );
  reward.min, reward.max = nil, nil;

  local collectable_count = v_player.Skin_CollectablePartCount( reward.name );

  if ( collectable_count > 0 ) then

    reward.count = math.min( reward.count, collectable_count );
    return reward;

  end;

  return nil;
  
end;
--------------------------------------------------------------------------------
function private.Chest_Random_Stickerpack ( list )

  if ( v_fireflies            ) and
     ( v_fireflies.Unlocked() ) then

    local reward = list[ math.random( 1, #list ) ];
    reward.count = reward.count or math.random( reward.min, reward.max );
    reward.min, reward.max = nil, nil;
    return reward;

  end;
  
  return nil;
  
end;
--------------------------------------------------------------------------------
function private.Chest_Random_Heart ( list )

  local reward = list[ math.random( 1, #list ) ];
  reward.count = reward.count or math.random( reward.min, reward.max );
  reward.min, reward.max = nil, nil;
  return reward;
  
end;
--------------------------------------------------------------------------------
function private.Chest_Random_Resource ( list )

  local reward = list[ math.random( 1, #list ) ];
  reward.count = reward.count or math.random( reward.min, reward.max );
  reward.min, reward.max = nil, nil;
  return reward;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Chest_Check_OneReward ( chest_reward, rewards )
  
  private.chest_reward_func.check[ chest_reward.type ]( chest_reward, rewards );
  
end;
--------------------------------------------------------------------------------
function private.Chest_Check_Stickerpack ( chest_reward, rewards )
  
  if ( not v_fireflies            ) then return; end;
  if ( not v_fireflies.Unlocked() ) then return; end;
  
  local rand = math.random( 1, 100 );
  local rewards_is_gold = true;
  
  for i = 1, #rewards do
    if ( rewards[ i ].type ~= "gold" ) then 
      rewards_is_gold = false;
      break; 
    end;
  end;
  
  if ( rand <= ( chest_reward.chance or 1 ) * 100 ) and ( rewards_is_gold ) then
    
    local rand_pos = math.random( 1, #rewards );
    rewards[ rand_pos ] = chest_reward;
    
  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Chest_GetFingerprint ( chest_name, chest_items )

  local fingerprint = chest_name;

  for _, item in pairs( chest_items ) do

    fingerprint = fingerprint..item.name..item.count;

  end;

  return fingerprint;

end;
--------------------------------------------------------------------------------
function private.Chest_TryObtain ( obtain_data )

  local chest_data = 
  {
    type  = "chest",
    name  = obtain_data.name,
    count = obtain_data.count or 1,
    items = obtain_data.items or false,
    category   = obtain_data.category,
    source     = obtain_data.source,
    stage_name = obtain_data.stage_name
  };

  if ( obtain_data.factor     ) and
     ( obtain_data.factor > 1 ) then

    chest_data.factor = obtain_data.factor;

  end;

  if ( chest_data.items ) then

    chest_data.fingerprint = private.Chest_GetFingerprint(
      chest_data.name,
      chest_data.items
    );

  end;

  return m_items.Reward_Add( chest_data );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Reward_TryResolveAll ()

  local rewards = m_items.Reward_GetAll();

  if ( #rewards < 1 ) then

    return;

  end;

  for _, reward in pairs( rewards ) do

    reward.silent = reward.silent or false;
    if ( reward.type == "chest" ) then

      public.TryObtainMultiple( reward.items, reward );

    else

      public.TryObtain( reward );

    end;

  end;

  m_items.Reward_DeleteAll();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.TryObtainGold ( item )

  local silent = tern( item.silent == nil )( true, item.silent );

  v_log.Kontagent_LogEvent(
    "economy", 
    {
      description = "premium_currency_collect",
      category    = item.category or "reward",
      value       = item.count,
      name        = item.stage_name or item.quest_name or item.source or "reward",
      state       = item.stage_state
    }
  );
  
  return v_player.Gold_TryChangeCount( item.count, silent );

end;
--------------------------------------------------------------------------------
function private.TryObtainResource ( item )

  v_log.Kontagent_LogEvent(
    "economy", 
    {
      description = "soft_currency_collect",
      category    = item.category or "reward",
      value       = item.count,
      name        = item.stage_name or item.source or "reward",
      type        = item.name,
      state       = item.stage_state,
      source      = item.source,
      scene       = item.scene,
      other_stage = item.other_stage
    }
  );
  
  return v_player.Resource_TryChangeCount( item.name, item.count, true );

end;
--------------------------------------------------------------------------------
function private.TryObtainSkinpart ( item )

  item.category = tern( item.category == "reward" )(
    item.stage_name,
    item.category
  );

  return v_player.Skin_TryObtainParts( item.name, item.count, item.category );

end;
--------------------------------------------------------------------------------
function private.TryObtainBooster ( item )

  return v_boosters.TryChangeCount( item.name, item.count, item.category );

end;
--------------------------------------------------------------------------------
function private.TryObtainFragment ( item )

  return v_builds.Build_TryObtainFragment( item.name, item.stage_name );

end;
--------------------------------------------------------------------------------
function private.TryObtainStickerpack ( item )

  if ( v_fireflies ) then

    local source = item.source or
                   item.stage_name or
                   item.category;

    local is_obtained = v_fireflies.Stickerpack_TryObtain(
      item.count,
      false,
      source
    );

    return is_obtained;

  else

    DbgTrace(
      "[ ERROR ][ v_items ] Failed to obtain stickerpack: module is not loaded."
    );

  end;

end;
--------------------------------------------------------------------------------
function private.TryObtainHeart ( item )

  return v_hearts.Heart_TryObtain( item );

end;
--------------------------------------------------------------------------------
function private.TryObtainChest ( item )

  return private.Chest_TryObtain( item );

end;
--------------------------------------------------------------------------------
function private.TryObtainContainer ( obtain_data )
  
  local container_name  = obtain_data.name;
  local container_count = obtain_data.count or 1;
  
  if ( not container_name ) then
    DbgTrace( "[ VIEW ][ ITEMS ][ ERROR ] Container name not set!" );
    return false;
  end;
  
  local chest_data = {
    type   = "chest",
    name   = "goldchest",
    factor = container_count
  };

  local container_items = ( v_game.features.Container_GetData( container_name ) or {} ).items;
  
  if ( not container_items ) then
 
    DbgTrace( "[ ERROR ] Container < "..container_name.." > or items not exist! Generate new container." );

    container_items = private.Chest_Generate( chest_data );
    
  end;
  
  if ( container_count > 1 ) then
    
    for i = 1, #container_items do

      container_items[ i ].count = container_items[ i ].count * container_count;
    
    end;
  
  end;

  public.Chest_Show(
    chest_data,
    obtain_data,
    container_items
  );

  return true;
  
end;
--------------------------------------------------------------------------------
function private.TryObtainPotion ( item )

  return v_potions.TryObtain( item );

end;
--------------------------------------------------------------------------------