-- name=v_log
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_log"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  common.mod.Load( _modules );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Playhaven_LogPlacement ( string_id )

  common.Trace( "Playhaven_LogPlacement: "..string_id );

  wiz.bfg.playhaven.LogPlacement( string_id );

end;
--------------------------------------------------------------------------------
function public.Kontagent_LogEvent ( event_type, event_data )

  local event_handlers = 
  {
    economy   = private.Kontagent_EconomyEvent,
    tutorial  = private.Kontagent_TutorialEvent,
    stage     = private.Kontagent_StageEvent,
    progress  = private.Kontagent_ProgressEvent,
    quest     = private.Kontagent_QuestEvent,
    skins     = private.Kontagent_SkinsEvent,
    social    = private.Kontagent_SocialEvent,
    arena     = private.Kontagent_ArenaEvent,
    minievent = private.Kontagent_MinieventEvent,
    carrier   = private.Kontagent_CarrierEvent,
    abtest    = private.Kontagent_ABTestEvent,
    promocode = private.Kontagent_PromoCodeEvent,
    heart     = private.Kontagent_HeartEvent,
    video     = private.Kontagent_VideoEvent
  };  

  local events_table = event_handlers[ event_type ]( event_data );

  events_table.params = private.Kontagent_AdditionalDetails( events_table or {} );
 
  common.Trace( "Kt_LogEvent: "..
                events_table.default1..", "..
                events_table.default2..", "..
                tostring( events_table.default3 )..
                ", name = "..events_table.name..
                ", value = "..tostring( events_table.params.v )..
                ", level = "..tostring( events_table.params.l )..
                ", source = "..tostring( events_table.params.source ) );
  
  wiz.bfg.kontagent.LogEvent( events_table );

end;
--------------------------------------------------------------------------------
function public.HasOffers_LogEvent ( stage_data )

  local hasoffers_levels = {
    stage_7  = "complete_level_7",
    stage_12 = "complete_level_12",
    stage_19 = "complete_level_19",
    stage_23 = "complete_level_23",
    stage_31 = "complete_level_31",
    stage_34 = "complete_level_34",
  };
  
  if ( hasoffers_levels[ stage_data.name ] ) then
    if ( stage_data.state == 1 ) then
      wiz.bfg.hasoffers.LogEvent( hasoffers_levels[ stage_data.name ] );
    end;
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Kontagent_EconomyEvent ( event_data )

  local default1 = "economy";
  local default2 = event_data.description;
  local default3 = event_data.category or "";
  
  local value = event_data.value;
  local level = event_data.level or 0;
  local name  = event_data.name;

  local stagename = function ( stage_name )

    if ( not stage_name ) then

      return nil;

    end;

    if ( stage_name:find( "_daily" ) ) then

      return "dailyquest";

    end;
    
    if ( stage_name:find( "_minievent" ) ) then

      return "minievent_level_"..( ( ( event_data.scene or 1 ) * 100 ) + ( event_data.other_stage or 1 ) );

    end;
    
    local prefix = "";

    local stage_data     = m_log.Stage_GetStaticData( stage_name );
    local section_name   = stage_data.section;
    local section_number = tonumber( string.match( section_name, "%d+" ) );
    local stage_number   = tonumber( string.match( stage_name, "%d+" ) );

    if ( section_number < 2 ) then
      
      if ( stage_name:find( "_bonus" ) ) then

        prefix = "bonus_";

      end;

      return prefix.."level_"..math.ceil( section_number * 1000 ) + tonumber( stage_name:match( "%d+" ) ); --stage_name;

    end;

    local stages_list = m_log.Stage_ListBySection( section_name );

    local bonus_stages_list = {};

    for i = #stages_list, 1, - 1 do

      if ( string.find( stages_list[ i ].name, "_bonus" ) ) then

        table.insert( bonus_stages_list, 1, table.remove( stages_list, i ) );

      end;

    end;

    for i = 1, #stages_list do

      if ( stages_list[ i ].name == stage_name ) then

        stage_number = i;

        break;

      end;

    end;

    for i = 1, #bonus_stages_list do

      if ( bonus_stages_list[ i ].name == stage_name ) then
        
        prefix = "bonus_";

        stage_number = i;

        break;

      end;

    end;

    return prefix.."level_"..math.ceil( section_number * 1000 ) + tonumber( stage_number );--"stage_"..stage_number;

  end;

  local stage = nil;
  
  if ( ui.obj.GetNamePrefix( name ) == "stage" ) and
     ( name:match( "%d+" ) ~= nil              ) then

    stage = event_data.name;
    name = stagename( name );

  end;

  if ( event_data.type ) then

    local resources = 
    {
      resource1 = "yellowflies",
      resource2 = "blueflies",
      resource3 = "purpleflies",
      resource4 = "redflies"
    };

    default2 = string.gsub( default2, "currency", resources[ event_data.type ] );

    event_data.type = nil;

  end;
  
  --[[DbgTrace( { 
    name = name,
    default1 = default1,
    default2 = default2,
    default3 = default3,
    params = { 
      v = value, 
      l = level,
      transaction_id = event_data.transaction_id,
      source = event_data.source,
      stage = stage,
      state = event_data.state,
      purchase = event_data.purchase
    } } );]]

  return { 
    name = name,
    default1 = default1,
    default2 = default2,
    default3 = default3,
    params = { 
      v = value, 
      l = level,
      transaction_id = event_data.transaction_id,
      source = event_data.source,
      stage = stage,
      state = event_data.state,
      purchase = event_data.purchase
    } };

end;
--------------------------------------------------------------------------------
function private.Kontagent_TutorialEvent ( event_data )

  local default1 = "tutorial";
  local default2 = event_data.group or "";
  local default3 = "";
  
  local value = event_data.value;
  local level = event_data.level;
  local name  = event_data.name;

  return { 
    name     = name,
    default1 = default1,
    default2 = default2,
    default3 = default3,
    params   = { v = value, l = level }
  };

end;
--------------------------------------------------------------------------------
function private.Kontagent_StageEvent ( event_data )

  local getdefault2 = function ( stage_name )

    if ( string.find( stage_name, "_bonus" ) ) then

      return "bonus_level_complete";

    end;

    if ( string.find( stage_name, "_daily" ) ) then

      return "daily_level_complete";

    end;
    
    if ( string.find( stage_name, "_minievent" ) ) then

      return "minievent_level_complete";

    end;

    return "level_complete";

  end;

  local getlevel = function ( event_name, stage_state, noreward )
    
    if ( event_data.scene ) then
      return event_data.scene;
    end;

    if ( event_name == "success" ) then

      if ( stage_state == 4 ) or ( stage_state == 3 and noreward ) then
  
        return stage_state;

      end;

      local state = stage_state;
  
      if ( state == 0 ) then

        state = 1;

      end;

      return state;

    end;

    return stage_state;

  end;

  local getvalue = function ( section_name, stage_name, other_stage )

    if ( other_stage ) then
      
      return other_stage;

    end;

    local section_number = tonumber( string.match( section_name, "%d+" ) );
    local stage_number   = tonumber( string.match( stage_name, "%d+" ) );

    if ( section_number < 2 ) then

      return stage_number;

    end;

    local stages_list = m_log.Stage_ListBySection( section_name );

    local bonus_stages_list = {};

    for i = #stages_list, 1, - 1 do

      if ( string.find( stages_list[ i ].name, "_bonus" ) ) then

        table.insert( bonus_stages_list, 1, table.remove( stages_list, i ) );

      end;

    end;

    for i = 1, #stages_list do

      if ( stages_list[ i ].name == stage_name ) then

        stage_number = i;

        break;

      end;

    end;

    for i = 1, #bonus_stages_list do

      if ( bonus_stages_list[ i ].name == stage_name ) then

        stage_number = i;

        break;

      end;

    end;

    return stage_number;

  end;

  local tiles, butterflies, bubbles;

  local getcellsdestroyed = function ( cells_data )

    if ( cells_data ) then

      if ( cells_data.type == "Dirt" ) then

        tiles = cells_data.value;

      end;

      if ( cells_data.type == "Ice" ) then

        tiles = cells_data.value;

      end;


      if ( cells_data.type == "Bug" ) then

        butterflies = cells_data.value;

      end;
      
      if ( cells_data.type == "Bubble" ) then

        bubbles = cells_data.value;

      end;

    end;

  end;

  getcellsdestroyed( event_data.cells );

  local default1 = "progression";
  local default2 = getdefault2( event_data.stage );
  local default3 = event_data.section or "";
  
  local value = getvalue( event_data.section, event_data.stage, event_data.other_stage ) or 0;
  local level = getlevel( event_data.name, event_data.state, event_data.noreward );
  local name  = event_data.name;

  local score   = event_data.score;
  local moves   = event_data.moves;
  local boost   = event_data.boost[ 1 ];
  local power   = event_data.boost[ 2 ];
  local items   = event_data.items;
  local flies   = event_data.flies;
  local bombs   = event_data.bombs;
  local is_boss = event_data.is_boss;  
  local woodchests   = event_data.woodchests;
  local silverchests = event_data.silverchests;
  local goldchests   = event_data.goldchests;
  local additional_moves = event_data.additional_moves;
  
  event_data.state    = nil;
  event_data.noreward = nil;
  event_data.score    = nil;
  event_data.moves    = nil;
  event_data.boost    = nil;
  event_data.cells    = nil;
  event_data.items    = nil;
  event_data.flies    = nil;
  event_data.bombs    = nil;
  event_data.bubbles  = nil;
  event_data.is_boss  = nil;
  event_data.woodchests    = nil;
  event_data.silverchests  = nil;
  event_data.goldchests    = nil;
  event_data.additional_moves  = nil;
  event_data.scene  = nil;

  --[[DbgTrace( { 
    name = name, 
    default1 = default1, 
    default2 = default2, 
    default3 = default3, 
    params = { 
      v = value, 
      l = level, 
      score = score, 
      moves = moves, 
      boost = boost, 
      power = power, 
      tiles = tiles, 
      butterflies = butterflies, 
      items = items, 
      flies = flies, 
      bubbles = bubbles, 
      bombs = bombs, 
      is_boss = is_boss,
      woodchests   = woodchests,
      silverchests = silverchests,
      goldchests   = goldchests,
      additional_moves = additional_moves
    } 
  } );]]
  
  return { 
    name = name, 
    default1 = default1, 
    default2 = default2, 
    default3 = default3, 
    params = { 
      v = value, 
      l = level, 
      score = score, 
      moves = moves, 
      boost = boost, 
      power = power, 
      tiles = tiles, 
      butterflies = butterflies, 
      items = items, 
      flies = flies, 
      bubbles = bubbles, 
      bombs = bombs, 
      is_boss = is_boss,
      woodchests   = woodchests,
      silverchests = silverchests,
      goldchests   = goldchests,
      additional_moves = additional_moves,
      state = level
    } 
  };

end;
--------------------------------------------------------------------------------
function private.Kontagent_ProgressEvent ( event_data )

  -- Reset progress option / section complete event.
  local default1 = "progression";
  local default2 = tern( event_data.name:find( "section_" ) )( "world_complete", "" );
  local default3 = "";

  return {
    name     = event_data.name, 
    default1 = default1, 
    default2 = default2, 
    default3 = default3, 
    params   = { v = 0, l = 0 }
  };

end;
--------------------------------------------------------------------------------
function private.Kontagent_QuestEvent ( event_data )

  local default1 = "progression";
  local default2 = "quest_complete";
  local default3 = event_data.section;

  return {
    name     = event_data.name,
    default1 = default1,
    default2 = default2,
    default3 = default3,
    params   = { v = 0, l = 0 }
  };

end;
--------------------------------------------------------------------------------
function private.Kontagent_SkinsEvent ( event_data )

  local default1 = "skins";
  local default2 = event_data.description;
  local default3 = event_data.category;

  return { 
    name = event_data.name, 
    default1 = default1, 
    default2 = default2, 
    default3 = default3, 
    params = { v = event_data.value or 0, l = 0, tokens = event_data.tokens } 
  };

end;
--------------------------------------------------------------------------------
function private.Kontagent_SocialEvent ( event_data )

  local default1 = "social";
  local default2 = event_data.description;
  local default3 = event_data.category or "";

  return {
    name = event_data.name,
    default1 = default1,
    default2 = default2,
    default3 = default3,
    params = { v = 0, l = 1, 
      login_channel = ( wiz.social.GetChannel() or {} ).login_channel }
  };

end;
--------------------------------------------------------------------------------
function private.Kontagent_ArenaEvent ( event_data )

  local default1 = "arena";
  local default2 = event_data.description or "";
  local default3 = event_data.category or "";
  
  return {
    name = event_data.name,
    default1 = default1,
    default2 = default2,
    default3 = default3,
    params = { v = 0, l = event_data.level }
  };

end;
--------------------------------------------------------------------------------
function private.Kontagent_MinieventEvent ( event_data )

  local default1 = "minievent";
  local default2 = event_data.description or "";
  local default3 = event_data.category or "";

  --[[DbgTrace( {
    name = event_data.name,
    default1 = default1,
    default2 = default2,
    default3 = default3,
    params = { v = 0, l = event_data.level }
  } );]]
  
  return {
    name = event_data.name,
    default1 = default1,
    default2 = default2,
    default3 = default3,
    params = { v = event_data.value or 0, l = event_data.level or 1 }
  };

end;
--------------------------------------------------------------------------------
function private.Kontagent_CarrierEvent ( event_data )

  local default1 = "carrier";
  local default2 = event_data.description or "";
  local default3 = event_data.category or "";
  local level    = event_data.level or 0;
  local value    = event_data.value or 0;  
  
  --[[DbgTrace( {
    name = event_data.name,
    default1 = default1,
    default2 = default2,
    default3 = default3,
    params = { 
      v = event_data.value or 0, 
      l = event_data.level or 0,
      from = event_data.from
    }
  } );]]
  
  return {
    name = event_data.name,
    default1 = default1,
    default2 = default2,
    default3 = default3,
    params = { 
      v = event_data.value or 0, 
      l = event_data.level or 0,
      from = event_data.from
    }
  };

end;
--------------------------------------------------------------------------------
function private.Kontagent_ABTestEvent ( event_data )

  local default1 = "abtest";
  local default2 = event_data.description or "";
  local default3 = event_data.category or "";
  
  --[[DbgTrace( {
    name = event_data.name,
    default1 = default1,
    default2 = default2,
    default3 = default3,
    params = { 
      v = event_data.value or 0, 
      l = event_data.level or 0
    }
  } );]]
  
  return {
    name = event_data.name,
    default1 = default1,
    default2 = default2,
    default3 = default3,
    params = { 
      v = event_data.value or 0, 
      l = event_data.level or 0
    }
  };

end;
--------------------------------------------------------------------------------
function private.Kontagent_PromoCodeEvent ( event_data )

  local default1 = "promocode";
  local default2 = event_data.description or "";
  local default3 = event_data.category or "";
  
  return {
    name = event_data.name,
    default1 = default1,
    default2 = default2,
    default3 = default3,
    params = { 
      v = event_data.value or 0, 
      l = event_data.level or 0
    }
  };

end;
--------------------------------------------------------------------------------
function private.Kontagent_HeartEvent ( event_data )

  local default1 = "heart";
  local default2 = event_data.description or "";
  local default3 = event_data.category or "";
  
  return {
    name     = event_data.name,
    default1 = default1,
    default2 = default2,
    default3 = default3,
    params   = { v = event_data.value or 0, l = 0, source = event_data.source }
  };

end;
--------------------------------------------------------------------------------
function private.Kontagent_VideoEvent ( event_data )

  local default1 = "rewarded_video";
  local default2 = event_data.description or "vungle";
  local default3 = event_data.category or "";
  
  --[[DbgTrace( {
    name     = event_data.name,
    default1 = default1,
    default2 = default2,
    default3 = default3,
    params   = { 
      v = 0,
      l = 0,
      source = event_data.source,
      rvselection = tern( event_data.rvselection )( 1, 0 ),
      rvwatched   = event_data.rvwatched,
      rvpossible  = event_data.rvpossible
    }
  } );]]
  
  return {
    name     = event_data.name,
    default1 = default1,
    default2 = default2,
    default3 = default3,
    params   = { 
      v = 0,
      l = 0,
      source = event_data.source,
      rvselection = tern( event_data.rvselection )( 1, 0 ),
      rvwatched   = event_data.rvwatched,
      rvpossible  = event_data.rvpossible
    }
  };

end;
--------------------------------------------------------------------------------
function private.Kontagent_AdditionalDetails ( event_data )

  local additional_data = event_data.params;
  
  local timeplayed = function ()

    local token = "wtc_ingame_time";

    if ( TimeIsWatchExist( token ) ) then

      local time = TimeGetWatchTime( token );

      return tostring( time );

    end;

    return tostring( 0 );

  end;

  local find_last_generic_stage = function ( stage_name )
    
    local stages_list = m_log.Stage_ListFromStorage();

    if ( stages_list ) then

      table.sort( stages_list, function ( a, b )
        return tonumber( a.name:match( "%d+" ) ) < tonumber( b.name:match( "%d+" ) );
      end );
      
      for i = #stages_list, 1, -1 do

        if not ( stages_list[ i ].name:find( "_bonus" ) ) then

          local stage_number = tonumber( stages_list[ i ].name:match( "%d+" ) );

          if ( stages_list[ i ].enabled or stages_list[ i ].completed ) then
            return "stage_"..stage_number;
          end;

        end;

      end;

    end;

    DbgTrace( "[ ERROR ] Last generic stage not found. Returned 'stage_0'." );
    return "stage_0";

  end;

  local section_name = v_sections.Section_NameCurrent();
  
  local maxlevel = function ()

    local stage_last = m_log.Stage_GetLast();
    local stage_data = m_log.Stage_GetStaticData( stage_last );
    
    local section_name = stage_data.section;

    local section_number = tonumber( section_name:match( "%d+" ) );

    if ( stage_last:find( "_bonus" ) ) then

      stage_last = find_last_generic_stage( stage_last ); 

    end;

    local latest = m_log.Stage_GetStorageData( stage_last );

    if ( latest and latest.completed ) then

      local stage_last_next = "stage_"..( tonumber( stage_last:match( "%d+" ) ) + 1 );
      
      local enabled = m_log.Stage_GetStorageData( stage_last_next );
      
      if ( enabled ) then

        stage_last = stage_last_next;

      end;

    end;

    stage_last = math.ceil( section_number * 1000 ) + tonumber( stage_last:match( "%d+" ) );

    if ( section_number > 1 ) then

      stage_last = stage_last - 110 * ( section_number - 1 );

    end;

    return tonumber( stage_last );

  end;

  local userlvl = function ( stage_name )

    if ( not stage_name ) then

      return nil;

    end;
    
    local stage_number = tonumber( string.match( stage_name, "%d+" ) );
    if ( stage_name:find( "_daily" ) ) then
      return stage_number.."_d";
    end;
    
    if ( stage_name:find( "_minievent" ) ) then
      return stage_number.."_me";
    end;    

    local stage_data     = m_log.Stage_GetStaticData( stage_name );
    local section_name   = stage_data.section;
    local section_number = tonumber( string.match( section_name, "%d+" ) );
    
    if ( stage_name:find( "_bonus" ) ) then

      local stages_list = m_log.Stage_ListBySection( section_name );

      local bonus_stages_list = {};

      for i = #stages_list, 1, - 1 do
        if ( string.find( stages_list[ i ].name, "_bonus" ) ) then
          table.insert( bonus_stages_list, 1, table.remove( stages_list, i ) );
        end;
      end;

      for i = 1, #bonus_stages_list do
        if ( bonus_stages_list[ i ].name == stage_name ) then
          stage_number = i;
          break;
        end;
      end;
      
      return ( math.ceil( section_number * 1000 ) + stage_number ).."_b";
    end;    
    
    if ( section_number > 1 ) then
      stage_number = stage_number - 110 * ( section_number - 1 );
    end;

    return math.ceil( section_number * 1000 ) + stage_number;

  end;

  local currency = function ( resource_name, section_name )
    
    local currency = m_log.Player_GetResourceData( resource_name );
    
    if not ( resource_name == "gold" ) then

      if ( section_name ) then
        -- JK: fake -1 if the currency is nil by that index
        currency = currency[ tonumber( section_name:match( "%d+" ) ) ] or -1;
      else
        currency = currency;
      end;
      
    end;
    
    if ( currency < 0 ) then
      
      currency = 0;
      
    end;
    
    return tostring( currency );

  end;

  local ispaying = function ()

    local result = m_log.Player_IsPaying();
    result = tern( result == 0 )( 0, 1 );

    return result;

  end;

  local isfb = function ()

    if ( ( wiz.social.GetChannel() or {} ).login_channel == "Facebook" ) then

      return 1;

    end;

    return 0;

  end;

  local userlvltier = function ( stage_name )

    if ( stage_name ) then

      return v_stage.Kontagent_GetStageState( stage_name );

    end;

    return nil;

  end;

  local is_googleplay_connected = function ()

    if ( wiz.Android.GooglePlayIsConnected() ) then

      return 1;

    end;

    return 0;

  end;
  
  local isbfg = function ()

    if ( ( wiz.social.GetChannel() or {} ).login_channel == "BFG" ) then

      return 1;

    end;

    return 0;

  end;
  
  local getstickers = function ()

    if ( _G.v_fireflies ) then

      return tostring( v_fireflies.Kontagent_GetStickersTotal() );

    end;
    
    return "0";

  end;
  
  local getlives = function ()

    if ( _G.v_hearts ) then

      return tostring( v_hearts.GetInfo().count );

    end;
    
    return nil;

  end;

  local getpotions = function ( params_table )

    if ( _G.v_potions ) then
      
      local list = v_potions.Potions_GetActiveList();

      for i = 1, #list do
        
        params_table[ list[ i ].name ] = "1";
        
      end;

    end;

  end;

  additional_data.appVersion = string.match( 
                        VersionGetComponentVersion( "game" ), "(%d+.%d+.%d+)" );
  additional_data.host       = private.Kontagent_GetHostAddress();
  additional_data.cntVersion = string.match( parts.version.GetCurrent( "content_main" ), 
                               "%d+.%d+.%d+.(%d+)" );
  additional_data.lvlVersion = string.match( parts.version.GetCurrent( "content_levels" ), 
                               "%d+.%d+.%d+.(%d+)" );
  additional_data.timeplayed = timeplayed();
  additional_data.maxlevel   = maxlevel();
  additional_data.userlvl    = tostring( 
                              userlvl( additional_data.stage ) or 
                              userlvl( v_stage.Kontagent_GetCurrentStage() ) or 0 );
  additional_data.userlvltier = tostring( 
                             additional_data.state or
                             userlvltier( v_stage.Kontagent_GetCurrentStage() ) or 0 );
  additional_data.softbal_yellowflies = currency( "resource1", section_name );
  additional_data.softbal_blueflies   = currency( "resource2", section_name );
  additional_data.softbal_purpleflies = currency( "resource3", section_name );
  additional_data.softbal_redflies    = currency( "resource4" );
  additional_data.premiumbal  = currency( "gold" );
  additional_data.isfb        = isfb();
  additional_data.isgc        = is_googleplay_connected();
  additional_data.isbfg       = isbfg();
  additional_data.ispaying    = ispaying();
  additional_data.stickers    = getstickers();
  additional_data.currentskin = v_player.Skin_GetCurrent() or "default";
  additional_data.unlimited_lives = tern( v_hearts.Immortality_IsActive() )( 1, 0 );
  additional_data.abtest_tutorial = v_features.GetFeatureGroup( "tutorial" ) or nil;
  additional_data.abtest_prices   = v_features.GetFeatureGroup( "prices" ) or nil;
  additional_data.lives = getlives();

  if ( v_player.IsAbtestParticipant() ) then
    additional_data.gb = v_game.GetSettingsVariant();
  else
    additional_data.gb = -1;
  end;
  
  getpotions( additional_data );
  
  if ( IsMobile() ) then
    
    additional_data.friends   = tostring( v_social.Friends_Count() );
    
  else
    
    additional_data.friends   = "0";
    
    local bfginfo = wiz.bfg.lot.GetBigFishInfo();
    
    additional_data.wrappingID = bfginfo.wrappingID;
    additional_data.gmguid     = bfginfo.gmguid;
    additional_data.bfgudID    = bfginfo.bfgudID;
    additional_data.userID     = bfginfo.userID;
    
  end;
  
  additional_data.stage = nil;
  additional_data.state = nil;
  
  --DbgTrace( additional_data );

  return additional_data;

end;
--------------------------------------------------------------------------------
function private.Kontagent_GetHostAddress ()

  local host = ConfigGetInfo( "server.host" ) or "UNKNOWN";

  local hosts_list = 
  {
    { host = "http://dev1.fairysaga.elephant-games.com/", type = "EG_DEV1" },
    { host = "http://dev2.fairysaga.elephant-games.com/", type = "EG_DEV2" },
    { host = "http://dev3.fairysaga.elephant-games.com/", type = "EG_DEV3" },
    { host = "http://f2p-qa.bigfishgames.com/FairySaga1/", type = "BFG_QA1" },
    { host = "http://f2p-qa.bigfishgames.com/FairySaga2/", type = "BFG_QA2" },
    { host = "http://f2p-st.bigfishgames.com/FairySaga/", type = "BFG_STAGE" },
    { host = "http://f2p.bigfishgames.com/FairySaga/", type = "BFG_PROD" }
  };

  for i = 1, #hosts_list do

    if ( hosts_list[ i ].host:find( host, 1, true ) ) then

      host = hosts_list[ i ].type;

    end;

  end;
  
  return host;

end;
--------------------------------------------------------------------------------