-- name=v_map
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_map",
  "int_dialog_map_build_market",
  "int_dialog_map_comingsoon",
  "int_dialog_map_reward",
  "int_dialog_worldmap",
  "int_splash_map_overtaken"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  common.mod.Load( _modules );

  private.Init_Variables();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Init ()

  private.Init_Objects();
  private.Init_Subscribes();

  -- Save data for 'sections_enable' feature.
  v_sections.SaveFeatureState();
  --v_sections.SetCurrentNumber( #v_sections.List() --[[v_sections.Section_NumberLast()]] );
  local section_number = v_sections.Player_GetEnabledList();
  v_sections.SetCurrentNumber( section_number );
  local prev_section = "section_"..( section_number - 1 );
  if ( v_sections.Section_NeedOutro( prev_section ) ) then
    v_sections.Section_SetOutro( prev_section, true );
  end;

  v_interface.Scheme_AddButton( "map", "bottom", { name = "worlds" } );
  v_interface.Scheme_AddButton( "map", "top",    { name = "default" } );

  if ( #private.worldmap_notification.new_worlds == 0 ) then
    local params = v_game.GetNotification( "worlds" );
    if ( params ) and ( #params ~= 0 ) then
      private.Worldmap_AddNewWorlds( params );
    end;
  end;

  --Log.i( "Initialized." );

  private.DEV_GetPointPos();

  -- common.DbgWatchTable( private, "v_map", "marker_point_current", false );

end;
--------------------------------------------------------------------------------
function private.Init_Variables ()

  -- Первый запуск.
  private.is_clean_start = true;

  -- JK: ask for screen dimensions once
  private.app_h = ui.screen.Height();
  private.app_w = ui.screen.Width();
  private.app_r = ui.screen.AspectRatio();

  -- Доп. обработчики состояния.
  private.point_handler = {};

  -- Созданные секции.
  private.sections = {};

  private.map_scale_current = 1.0;
  private.map_scale_max = 1.0;
  private.map_scale_min = 0.5;

  if ( private.app_h < 960 ) then

    private.map_scale_current = 0.75;
    private.map_scale_max = 0.75;

  end;

  private.scale_modifier = 2;
  private.offset         = 256;
  private.map_w          = 2048;
  private.map_h          = 1280;

  private.create_at_once = 10;

  private.reward          = {};
  private.resource_buffer = {};

  private.popup_shift = { 0, - 100 };

  private.opened = false;

  private.context_message = {
    unlocked    = false,
    triggered   = false,
    stage_count = 0
  };

  private.appear_queue = {};

  private.this = {
    background = "obj_map_background",
    body       = "obj_map_body",
    marker     = "obj_map_marker"
  };

  private.worldmap_notification = {
    new_worlds = {}
  };

  private.me_gift = {
    created = false,
    anm_stage   = "anm_map_point_stage",
    anm_section = "anm_map_section",
    anm_gift    = "anm_map_chest_minievent",
    anm_dummy   = "anm_map_chest",
    chests      = {},
    build_gift  = {},
    stage_build = {},
    subscribed  = false
  };

end;
--------------------------------------------------------------------------------
function private.Init_Objects ()

  --[[
  ObjSet( "obj_map_drag",
  {
    event_startdrag = private.DragStart,
    event_drag      = private.DragUpdate,
    event_dragdrop  = private.DragDrop,
  } );
  ]]

  local screen_w = private.app_w;
  local screen_h = private.app_h;
  local scroll_h = private.scale_modifier * private.map_h;

  local multiset = {};

  ui.obj.Set(
    "obj_map_foreground_intro",
    {
      pos_x = 0.5 * screen_w,
      pos_y = 0.5 * screen_h
    },
    multiset
  );

  ui.obj.Set(
    "zmv_map",
    {
      scale         = private.map_scale_current,
      min_scale     = private.map_scale_min,
      max_scale     = private.map_scale_max,
      scrollrect_x  = 0,
      scrollrect_y  = - scroll_h,
      scrollrect_w  = private.scale_modifier * private.map_w,
      scrollrect_h  = scroll_h,
      screen_w      = screen_w,
      screen_h      = screen_h,
      begindragtrig = private.DragStart,
      dragtrig      = private.DragUpdate,
      enddragtrig   = private.DragDrop
    },
    multiset
  );

  GridSet( "grd_map_foreground_movie_vignette", 0,
  {
    2,  { pos_x = screen_w - 1 },
    3,  { pos_x = screen_w     },
    6,  { pos_x = screen_w - 1 },
    7,  { pos_x = screen_w     },

    8,  { pos_y = screen_h - 4 },
    9,  { pos_y = screen_h - 4 },
    10, { pos_x = screen_w - 1, pos_y = screen_h - 4 },
    11, { pos_x = screen_w,     pos_y = screen_h - 4 },

    12, { pos_y = screen_h },
    13, { pos_y = screen_h },
    14, { pos_x = screen_w - 1, pos_y = screen_h },
    15, { pos_x = screen_w,     pos_y = screen_h }
  } );

  v_screen.Setup( "scr_map",
  {
    func_pre_open    = private.PreOpen,
    func_open        = private.Open,
    func_close       = private.Close,
    func_after_close = private.AfterClose
  } );

  ui.obj.Attach( "scr_map", "ne_storage" );

  if ( wiz.IsCheatsEnabled() ) then

    ui.obj.Enable( "obj_map_point_DEV", false, multiset );

  end;

  ui.obj.Set(
    private.this.background,
    {
      pos_x = 0,
      pos_y = 0
    },
    multiset
  );

  ui.obj.Set(
    "anm_map_foreground_fox_actor",
    {
      pos_x = 0.5 * private.app_w
    },
    multiset
  );

  ui.obj.Set(
    "obj_map_anims",
    {
      pos_x =  private.offset,
      pos_y = -private.offset
    },
    multiset
  );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.Init_Subscribes ()

  v_sections.Init_Subscribes();

  mediator.Subscribe(
    "Model_Skins_SkinSet",
    private.Marker_SetSkin
  );

  mediator.Subscribe(
    "View_Sections_CurrentDisabled",
    private.Section_OnCurrentDisabled
  );

  mediator.Subscribe(
    "View_Sections_EnabledList",
    private.Worldmap_AddNewWorlds
  );

  mediator.Subscribe(
    "View_Stage_Started",
    private.Context_CountStageStart
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.IsOpened ()

  return private.opened;

end;
--------------------------------------------------------------------------------
function public.IsActive ()

  return v_screen.GetCurrentName() == "scr_map";

end;
--------------------------------------------------------------------------------
function public.IsSectionCreated ( section_number )

  return private.sections[ "section_"..section_number ];

end;
--------------------------------------------------------------------------------
function public.IsIntroActive ()

  return private.intro_active;

end;
--------------------------------------------------------------------------------
function public.FocusAtCurrentPoint ( func_after )

  if ( private.last_point_selected  ) or
     ( private.marker_point_current ) then

    local target = private.Marker_Target(
      private.last_point_selected or private.marker_point_current,
      ( private.last_point_selected ~= nil )
    );

    public.Focus( target, { func_after = func_after } );

  elseif ( func_after ) then

    func_after();

  end;

end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function public.Scale_GetMax ()

  return private.map_scale_max;

end;
--------------------------------------------------------------------------------
function public.Scale_GetMin ()

  return private.map_scale_min;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Hero_ShowPopup ( string_id, stage_name, funcs )

  if ( stage_name ) then

    mediator.Notify(
      "View_Map_HeroPopupActivated",
      stage_name
    );

  end;

  public.ShowObjectPopup(
    "monolog",
    "anm_map_marker_icon",
    string_id,
    nil,
    { 0, -50 },
    funcs,
    true
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Dialog_DailyBonus_TryShow ( func_close )

  v_dailybonus.Dialog_TryShow( func_close );

end;
--------------------------------------------------------------------------------
function private.Dialog_Notifications_Show ( func_after )

  local push_reward = v_player.Playhaven_GetPushNotificationsReward();

  v_game.dialog.Notifications_Show(
    push_reward,
    function()

      return v_player.Playhaven_RegisterPushNotifications( push_reward );

    end,
    function()

      v_game.dialog.Notifications_Hide();
      if ( func_after ) then func_after() end;

    end
  );

end;
--------------------------------------------------------------------------------
function private.Dialog_SectionDisabled_Show ( func_on_hide )

  v_game.SystemDialog_Show(
    "section_disabled",
    func_on_hide
  );

end;
--------------------------------------------------------------------------------
function private.WorldsButtonClick ()

  private.Dialog_Worldmap_Show( false );
  private.Worldmap_CheckNotification( true );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Worldmap_AddNewWorlds ( data )

  private.worldmap_notification.new_worlds = data;

  v_game.SetNotification( "worlds", data );

  private.Worldmap_CheckNotification();

end;
--------------------------------------------------------------------------------
function private.Worldmap_CheckNotification ( clear )

  if ( clear ) then
    if ( #private.worldmap_notification.new_worlds ~= 0 ) then
      v_game.SetNotification( "worlds", false );
    end;
    private.worldmap_notification.new_worlds = {};
  end;

  local count = #private.worldmap_notification.new_worlds;

  local data = {
    target = "worlds",
    data = {
      type = "common",
      entity = "v_map",
      count = count
    }
  };

  v_interface.panel.bottom.Notification( data );

end;
--------------------------------------------------------------------------------
function private.Dialog_Worldmap_Show ( unlock )

  local section_name  = v_sections.Section_NameCurrent();
  local sections_list = v_sections.DetailedList();
  local skin_name     = v_player.Skin_GetCurrent();
  local skin_cont     = v_player.Skin_GetContent( skin_name );
  local new_worlds    = private.worldmap_notification.new_worlds;
  local chidren = ui.obj.Children( public.Dialog_Worldmap_GetFeaturedPlaceholderName() );
  local is_featured = ( #chidren > 0 );

  if ( is_featured ) then

    mediator.Subscribe(
      "View_Minievent_StoryStarted",
      private.Worlds_ShowFeaturedInPanel
    );

    mediator.Subscribe(
      "View_Minievent_StoryFinished",
      private.Worlds_HideFeaturedInPanel
    );

    mediator.Subscribe(
      "View_Minievent_Resumed",
      private.Worlds_ShowFeaturedInPanel
    );

    mediator.Subscribe(
      "View_Minievent_Suspended",
      private.Worlds_HideFeaturedInPanel
    );

  end;
  
  local delegates = {

    IsContentReady = v_game.Content_IsSectionReady,

    GoTo = function ( destination, func_hide ) -- OnButtonClick

      if ( destination == v_sections.Section_NumberCurrent() ) then

        func_hide();

      else

        v_interface.Loading_Show( true,
        function ()
        
          v_interface.Disable( true );
          v_game.CCS_Freeze( true );
          func_hide( function ()
            
            local new_section_number, current_section_number = v_sections.Section_NumberCurrent();
            
            private.Section_Switch( current_section_number, destination, 
            function()

              v_interface.Loading_Hide( true );
              v_game.CCS_Freeze( false );
              private.Update();

            end );
            
          end );
        
        end );

      end;
    
    end,

    OnClose = function ()

      if ( is_featured ) then

        mediator.Unsubscribe(
          "View_Minievent_StoryStarted",
          private.Worlds_ShowFeaturedInPanel
        );

        mediator.Unsubscribe(
          "View_Minievent_StoryFinished",
          private.Worlds_HideFeaturedInPanel
        );

        mediator.Unsubscribe(
          "View_Minievent_Resumed",
          private.Worlds_ShowFeaturedInPanel
        );

        mediator.Unsubscribe(
          "View_Minievent_Suspended",
          private.Worlds_HideFeaturedInPanel
        );

        mediator.Notify( "View_Map_DialogWorldmap_ClosedWithFeatured" );
        
      end;

    end,
    Share = function ()
      v_social.FacebookShare();
    end,
    GetSharingWorld = function ()
      return v_social.FacebookShare_GetAvailableSectionName();
    end
    
  };

  common.delegate.Set( "int_dialog_worldmap", delegates );

  v_interface.window.Show(
    "int_dialog_worldmap",
    function ()

      int_dialog_worldmap.Show(
        section_name,
        sections_list,
        is_featured,
        { 
          name    = skin_name,
          content = skin_cont
        },
        unlock,
        new_worlds
      );
  
    end
  );

  if ( is_featured ) then

    mediator.Notify( "View_Map_DialogWorldmap_OpenedWithFeatured" );

  end;

end;
--------------------------------------------------------------------------------
function private.Worlds_ShowFeaturedInPanel ()

  if ( int_dialog_worldmap.IsActive() ) then
    int_dialog_worldmap.Featured_Show();
  end;

end;
--------------------------------------------------------------------------------
function private.Worlds_HideFeaturedInPanel ()

  if ( int_dialog_worldmap.IsActive() ) then
    int_dialog_worldmap.Featured_Hide();
  end;

end;
--------------------------------------------------------------------------------
function public.Dialog_Worldmap_Hide ( func_end )

  int_dialog_worldmap.Hide( func_end );

end;
--------------------------------------------------------------------------------
function public.Dialog_Worldmap_IsActive ()

  local is_active = int_dialog_worldmap.IsActive();
  return is_active;

end;
--------------------------------------------------------------------------------
function public.Dialog_Worldmap_GetFeaturedPlaceholderName ()

  return "obj_int_dialog_worldmap_featured";

end;
--------------------------------------------------------------------------------
function public.Dialog_Worldmap_ShareComplete ( section_name )

  if ( public.Dialog_Worldmap_IsActive() ) then
    int_dialog_worldmap.ShareComplete( section_name );
  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Context_Show ( func_show, func_close )

  func_show( func_close );

  private.context_message.stage_count = 0;
  private.context_message.triggered   = true;

end;
--------------------------------------------------------------------------------
function private.Context_CountStageStart ( params )

  local stage_name = params.stage_name;

  if ( stage_name:find( "daily" )     ) or
     ( stage_name:find( "minievent" ) ) then

    return;

  end;

  private.context_message.stage_count = private.context_message.stage_count + 1;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Market_Update ()

  v_shop.Market_Update( private.Market_SetOnMap, private.Market_SetTap );

end;
--------------------------------------------------------------------------------
function private.Market_SetOnMap ( unlocked, built )

  local objects = {};

  if ( unlocked ) and not ( built ) then

    table.insert( objects, { "obj_map_builds_market_sign",
    {
      active  = true,
      visible = true } } );

  else

    table.insert( objects, { "obj_map_builds_market_sign",
    {
      active  = false,
      visible = false } } );

  end;

  if ( built ) then

    table.insert( objects, { "anm_map_builds_market",
    {
      animfunc = "static",
      playing  = true } } );

  else

    table.insert( objects, { "anm_map_builds_market",
    {
      animfunc = "base",
      playing  = true } } );

  end;

  ObjMultiSet( objects );

end;
--------------------------------------------------------------------------------
function private.Market_Dialog_Build_Show ()

  local gold_count = v_shop.Market_Settings( "build" ).gold;

  local delegates = {

    Accept = function()

      v_shop.Market_TryBuild(
        gold_count,
        private.Market_SetOnMap,
        private.Market_SetTap
      );

    end,

    Close = private.Market_Dialog_Build_Hide,

    ShowTopPanel = function ()

      if ( gold_count > 0 ) then

        v_interface.panel.top.Highlight( "gold" );
        v_interface.panel.top.Show();

      end;

    end,

    HideTopPanel = function ()

      if ( gold_count > 0 ) then

        v_interface.panel.top.Hide();

      end;

    end

  };

  common.delegate.Set( "int_dialog_map_build_market", delegates );

  v_interface.window.Show(
    "int_dialog_map_build_market",
    function ()

      int_dialog_map_build_market.Show( gold_count );

    end
  );

   common.sound.Play( "bottom_button" );

end;
--------------------------------------------------------------------------------
function private.Market_Dialog_Build_Hide ()

  int_dialog_map_build_market.Hide( function ()

    v_interface.window.Hide( "int_dialog_map_build_market" );

  end );

end;
--------------------------------------------------------------------------------
function private.Market_SetTap ( action ) -- "popup", "build" or "open"

  if ( action == "popup" ) then

    ui.control.TapSet( "obj_map_builds_market",
    {
    func_true = function ( sender, skip_focus )

        public.ShowObjectPopup(
          "monolog",
          "obj_map_builds_market",
          "str_int_popup_cloud_builds_market",
          nil,
          nil,
          nil,
          skip_focus
        );

      end
    } );

  end;

  if ( action == "build" ) then

    ui.control.TapSet( "obj_map_builds_market",
    {
      func_true = function()

        private.Market_Dialog_Build_Show();

      end
    } );

  end;

  if ( action == "open" ) then

    ui.control.TapSet( "obj_map_builds_market",
    {
      func_true = function ()

        v_interface.panel.top.Lock();
        v_shop.Show( "market" );

      end
    } );

  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.ViewPort_Contract ( panel_width, tme )

  ui.input.Off( "scr_map" );

  ObjAnimate(
    "zmv_map",
    "screen_w", 0, 0, "",
    {
      0.0, 0, ui.obj.Get( "zmv_map" ).screen_w,
      tme, 0, private.app_w - panel_width
    }
  );

end;
--------------------------------------------------------------------------------
function public.ViewPort_Restore ( tme )

  ObjAnimate(
    "zmv_map",
    "screen_w", 0, 0,
    function ()

      ui.input.On( "scr_map" );

    end,
    {
      0.0, 0, ui.obj.Get( "zmv_map" ).screen_w,
      tme, 0, private.app_w
    }
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.PreOpen ( func_after )

  mediator.Notify( "View_Map_PreOpen" );

  -- PreOpen end callback
  local function func_preopened ( skip_counters_reload )

    --- First load -------------------------------------------------------------
    if ( private.is_clean_start ) then

      v_game.SilentContent_TryDownload();

      private.is_clean_start = false;

      local ignore_stage_info = false;
      if ( private.changelist_active ) then
        ignore_stage_info = true;
      end;
      private.Marker_SetAtLastPoint( ignore_stage_info );

      if ( v_sections.Section_NeedIntro() ) then
        local section_name = v_sections.Section_NameCurrent();
        public.Focus( "obj_map_portal_"..section_name.."_prev",
          {
            instant = true,
            scale   = 1.0
          }
        );
      end;

    end;

    if ( not skip_counters_reload ) then
      v_interface.Gold_Buffer_Out();
      for i = 1, 4 do
        v_interface.Resource_Buffer_Out( "resource"..i );
      end;
    end;

    --- Intro ------------------------------------------------------------------
    if ( v_sections.Section_NeedIntro() ) then

      private.Intro_Prepare();

    end;

    --- Minievent --------------------------------------------------------------
    local func_minievent;

    if ( _G.v_minievent ) then
      private.Minievent_OnGiftUpdated();
      func_minievent = v_minievent.OnMapUpdate();
    end;

    if ( func_minievent ) then
      func_minievent( func_after );
    else
      if ( func_after ) then func_after(); end;
    end;

  end;

  local current_section_number = v_sections.Section_NumberCurrent();
  local loaded_section_number  = v_sections.Section_GetLoaded() or 0;

  if ( current_section_number == loaded_section_number ) then

    func_preopened( true );
    return;

  end;

  private.Section_Switch(
    loaded_section_number,
    current_section_number,
    func_preopened
  );

end;
--------------------------------------------------------------------------------
function private.Open ( func_after )

  mediator.Subscribe(
    "View_Minievent_GiftStarted",
    private.Minievent_OnGiftUpdated
  );

  mediator.Subscribe(
    "View_Minievent_GiftUpdated",
    private.Minievent_OnGiftUpdated
  );

  mediator.Subscribe(
    "View_Minievent_GiftFinished",
    private.Minievent_OnGiftFinished
  );

  common.sound.Stop( "assets/content/hero/aud_hero_win_ride" );
  common.sound.Stop( "level_completed" );

  if ( not v_sections.Section_NeedIntro() ) then

    private.Music_Start();

  end;
  
  common.event.Set(
    "int_panel_interface_bottom",
    "WorldsClickHandler",
    {
      name = "v_map",
      handler = function ()
        private.WorldsButtonClick();
        common.sound.Play( "bottom_button" );
      end
    }
  );

  private.opened = true;

  local function func_update ()

    -- JK: map is closing or map is closed -> drop the update
    if ( v_screen.IsClosing()  ) or
       ( not public.IsActive() ) then
      return;
    end;

    local portal_unlocked = v_sections.Portal_TryUnlock();
    if ( portal_unlocked ) then
      m_map.ChangeList_AddEvent( portal_unlocked );
    end;

    private.Update( true );

    v_log.Playhaven_LogPlacement( "map_open" );

    v_game.LocalNotifications_SetData( "comeback_level", nil );

    mediator.Notify( "View_Map_MapOpened" );

    if ( func_after ) then func_after() end;

  end;

  if ( v_interface.window.IsAnyOnScreen() ) then
    mediator.SubscribeOnce(
      "Interface_Window_NoWindows",
      func_update
    );
  else
    func_update();
  end;

end;
--------------------------------------------------------------------------------
function private.Close ( func_after )

  mediator.Unsubscribe(
    "View_Minievent_GiftStarted",
    private.Minievent_OnGiftUpdated
  );

  mediator.Unsubscribe(
    "View_Minievent_GiftUpdated",
    private.Minievent_OnGiftUpdated
  );

  mediator.Unsubscribe(
    "View_Minievent_GiftFinished",
    private.Minievent_OnGiftFinished
  );

  v_interface.Hide( { screen = "map" } );

  common.event.Delete(
    "int_panel_interface_bottom",
    "WorldsClickHandler",
    {
      name = "v_map"
    }
  );
  
  -- JK: cancel any running focus
  ObjStopAnimate( private.this.body, "pos_xy" );

  mediator.Notify( "View_Map_Close" );

  private.opened = false;
  private.FPS_StopFrameRateUpdate();

  if ( func_after ) then

    func_after();

  end;

end;
--------------------------------------------------------------------------------
function private.AfterClose ( func_after )

  private.Music_Stop();

  mediator.Notify( "View_Map_MapClosed" );

  private.Reward_Collect();

  -- Detach splash if not finished
  int_splash_map_overtaken.Hide();

  if ( func_after ) then

    func_after();

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.IsCleanStart ()

  return private.is_clean_start;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.FPS_FrameRateUpdate ()

  ObjSet( "tmr_update_level_fps", {
    time = 1.0,
    playing = true,
    endtrig = private.FPS_FrameRateUpdate
  } );

  if ( common.fps.Get() < 30 and not ui.screen.IsLocked() ) then

    private.FPS_DisableEnvironment();
    private.FPS_SetChests( false );

  else

    private.FPS_EnableEnvironment();
    private.FPS_SetChests( true );

  end;
end;
--------------------------------------------------------------------------------
function private.FPS_SheduleFrameRateUpdate ( func_after )

  if ( not ObjGet( "tmr_update_level_fps" ) ) then
    ObjCreate( "tmr_update_level_fps", "timer" );
  end;

  ObjSet( "tmr_update_level_fps", {
    time = 1.0,
    playing = true,
    endtrig = private.FPS_FrameRateUpdate
  } );
  ObjAttach( "tmr_update_level_fps", "ng_application" );

  local section = v_sections.Section_NameCurrent();
  private.fps = {
    section = section,
    chests  = v_chests.List_OnSection( section ),
    mines   = v_mines.GetAllBySection( section ),
    builds  = v_builds.GetAllInSection( section, true ),
    stages  = v_stage.Stage_GetAllBySection( section, true )
  };
  private.fps_env_enabled = false;

  if ( func_after ) then func_after(); end;

end;
--------------------------------------------------------------------------------
function private.FPS_StopFrameRateUpdate ()

  ObjDelete( "tmr_update_level_fps" );

end;
--------------------------------------------------------------------------------
function private.FPS_DisableEnvironment ()

  if ( private.fps_env_enabled == true ) then

    for z = 1, 5 do

      local env_name = "anm_map_"..private.fps.section.."_environment_"..z;

      ObjAnimate( env_name, "alp", 0, 0,
      function()
        ObjSet( env_name, { active = false, visible = false } );
        ObjDetach( env_name );
        private.fps_env_enabled = false;
      end,
      {
        0.0,  0,  1,
        0.2,  0,  0
      } );
    end;

  end;

end;
--------------------------------------------------------------------------------
function private.FPS_EnableEnvironment ()

  if ( private.fps_env_enabled == false ) then

    for z = 1, 5 do

      local env_name = "anm_map_"..private.fps.section.."_environment_"..z;

      ObjAttach( env_name, "anm_map_"..private.fps.section );
      ObjSet( env_name, { active = true, visible = true } );
      ObjAnimate( env_name, "alp", 0, 0,
      function()
        private.fps_env_enabled = true;
      end,
      {
        0.0,  0,  0,
        0.2,  0,  1
      } );

    end;
  end;

end;
--------------------------------------------------------------------------------
function private.FPS_SetChests ( enabled )

  for i = 1, #private.fps.chests do
    ObjSet( "obj_map_chests_"..private.fps.chests[ i ].name, { active = enabled } );
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Section_OnCurrentDisabled ()

  local new_section_number,
        current_section_number = v_sections.Section_NumberCurrent();

  if ( new_section_number ~= current_section_number ) then

    if ( public.IsOpened() ) then

      local func_on_hide = function ()

        v_interface.Loading_Show( true,
        function ()

          v_interface.Hide();

          private.Section_Switch(
            current_section_number,
            new_section_number,
            function ()

              v_interface.Loading_Hide( true );
              private.Update();

            end
          );

        end );

      end;

      private.Dialog_SectionDisabled_Show( func_on_hide );

    else

      private.section_disabled_message_delayed = true;

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Section_TryShowDisabled ()

  if ( private.section_disabled_message_delayed ) then

    private.Section_OnCurrentDisabled();

    private.section_disabled_message_delayed = nil;

  end;

end;
--------------------------------------------------------------------------------
function private.Section_Switch ( prev_number, destination, func_after_switch )

  local section_name = v_sections.Section_NameCurrent();

  if ( not v_game.Content_IsPackageReady( section_name ) ) then
    destination = destination - 1;
  end;

  local function func_switch_finished ( func_after )

    mediator.Notify( "View_Map_Section_Switched", destination );

    if ( v_sections.Section_NeedIntro() ) then private.Intro_Prepare(); end;

    if ( func_after_switch ) then func_after_switch(); end;
    if ( func_after )        then func_after();        end;

  end;

  -- Do not switch to not enabled sections
  local section_enabled = v_sections.Section_IsEnabled( "section_"..destination );

  if ( not section_enabled ) then

    private.open_map_deferred = false;

    if ( func_after_switch ) then func_after_switch(); end;
    return;

  end;

  private.open_map_deferred = true;

  -- Detach old objects, if any section already loaded
  if ( private.sections[ "section_"..prev_number ] ) then

    local anm_map = "anm_map_section_"..prev_number;
    ui.obj.Set( anm_map, { active = false, visible = false } );
    ui.obj.Detach( anm_map );
    --wiz.ClearUsedResource();

  end;

  -- Update current section number and load section
  v_sections.SetCurrentNumber( destination );

  local section_number = v_sections.Section_NumberCurrent();

  v_game.Content_LoadStrings( section_number );
  v_player.Progress_SetSection( section_number );

  -- private.Section_LoadCurrent using timer for processing long cycles,
  -- all next operations passed as callbacks (prevention of EventLoop locking)

  local function func_marker_transfer ( func_after )

    if ( prev_number ~= section_number ) then

      private.last_point_selected = nil;

      private.Marker_Transfer( prev_number, section_number );

    end;

    func_after();

  end;

  local function func_after_load_actions()
    common.Defer(
      "map_create_env",
      1,
      function()
        private.Create_Environment( private.FPS_SheduleFrameRateUpdate );
      end
    );
  end;

  if ( prev_number > 0 ) then
    -- JK: clear the changelist: no updates as we are just landed to a new land
    m_map.ChangeList_Clear();
  end;

  common.RunCallbackList({
    private.Section_LoadCurrent,
    func_marker_transfer,
    func_switch_finished,
    func_after_load_actions
  });

end;
--------------------------------------------------------------------------------
function private.Section_LoadCurrent ( func_after_load )

  local section_number = v_sections.Section_NumberCurrent();
  local section_name   = v_sections.Section_NameCurrent();

  v_sections.Section_CheckPrevious( section_number );
  v_sections.Section_EnableNext(); -- WUT?
  v_sections.Section_SetLoaded( section_number );

  local function func_load_ground ( func_after )

    ObjSet( "spr_map_background",
    {
      res   = "assets/content/section_"..section_number.."/background",
      pos_x =   0.5 * ( private.map_w ),
      pos_y = - 0.5 * ( private.map_h )
    } );

    mediator.Notify( "View_Map_GroundLoaded", section_name );

    func_after();

  end;

  common.RunCallbackList({
    private.Create_Objects,
    func_load_ground,
    func_after_load
  });

end;
--------------------------------------------------------------------------------
function private.Create_Objects ( func_after )

  local section_name = v_sections.Section_NameCurrent();

  -- Just reattach animation if section already created
  if ( private.sections[ section_name ] ) then

    -- NOTE: Attach map and wait for object placing
    local anm_map = "anm_map_"..section_name;
    ObjAttach( anm_map, private.this.body );
    ui.obj.Set( anm_map, { active = true, visible = true } );

    common.Defer( "map_attach_wait", 0.1, func_after );
    return;

  end;

  private.sections[ section_name ] = true;

  local function func_attach_map ()

    ObjSet( "anm_map_"..section_name, {
      res      = "assets/content/"..section_name.."/anm_map_"..section_name,
      animfunc = section_name,
      playing  = true,
      pos_x    =  private.offset,
      pos_y    = -private.offset,
      pos_z    = 1000
    } );

    ObjAttach( "anm_map_"..section_name, private.this.body );

  end;

  local function func_create_other ( func_after )

    local creation_funcs =
    {
      private.Create_Carrier,
      private.Portal_Show,
      private.Portal_Next_SetTap,
      private.Create_PortalSign,
      func_attach_map,
      private.Marker_SetSkin
    };

    local function func_exec ( func, progress )

      func();
      v_interface.Loading_SetProgress( progress, 0.9, 1.00 );

    end;

    common.ProcessArrayByTimer(
      creation_funcs, func_exec, func_after, private.create_at_once
    );

  end;

  common.RunCallbackList({
    private.Create_Stages,      -- 00 - 10%
    private.Create_Builds,      -- 10 - 30%
    private.Create_Characters,  -- 30 - 40%
    private.Create_Fights,      -- 40 - 50%
    private.Create_Chests,      -- 50 - 60%
    private.Create_Spells,      -- 60 - 70%
    private.Create_Mines,       -- 70 - 80%
    private.Create_SetZones,    -- 80 - 90%
    func_create_other,          -- 90 - 100%
    function () common.Defer( "map_attach_wait", 0.1, func_after ); end
  });

end;
--------------------------------------------------------------------------------
function private.Create_Environment ( func_after )

  local section_name     = v_sections.Section_NameCurrent();
  local environment_list = {};

  -- JK: 5 environment animations per section
  for z = 1, 5 do

    table.insert( environment_list, "environment_"..z );

  end;

  -- Create and attach environment animations
  local function func_create ( env_name, progress )

    local map_name = "anm_map_"..section_name;
    local anm_name = map_name.."_"..env_name;
    local res_name = "assets/content/"..section_name.."/map/anm_"..env_name;

    if ( not ui.obj.Exists( anm_name ) ) then

      ui.obj.Create( anm_name, "anim", map_name );

    end;

    ui.obj.Set(
      anm_name,
      {
       res      = res_name,
       animfunc = "static",
       playing  = true,
       anim_tag = env_name,
       input    = false,
       active   = false,
       visible  = false
      }
    );

  end;

  common.ProcessArrayByTimer(
    environment_list,
    func_create,
    func_after,
    private.create_at_once
  );

end;
--------------------------------------------------------------------------------
function private.Create_Stages ( func_after )

  local multiset     = {};
  local section_name = v_sections.Section_NameCurrent();
  local stage_list   = v_stage.Stage_GetAllBySection( section_name, true );

  -- Prepare default point object
  local multiset = {};

  local res = "assets/content/"..section_name.."/map/anm_point";

  ui.obj.Set(
    "anm_map_point",
    { res = res, animfunc = "point_static" },
    multiset
  );
  ui.obj.Set( "anm_map_point_bg", { res = res }, multiset );
  ui.obj.Set( "anm_map_point_label", { res = res }, multiset );
  ui.obj.Set( "anm_map_point_label_number", { res = res }, multiset );
  ui.obj.Set( "anm_map_point_label_number_digit_1", { res = res }, multiset );
  ui.obj.Set( "anm_map_point_label_number_digit_2", { res = res }, multiset );
  ui.obj.Set( "anm_map_point_label_number_digit_3", { res = res }, multiset );
  ui.obj.Set( "anm_map_point_state", { res = res }, multiset );
  ui.obj.Set(
    "fcs_map_point",
    { res = res, animfunc = "bg_default" },
    multiset
  );
  ui.obj.Set( "anm_map_point_reward_icon", { res = res }, multiset );
  ui.obj.Set(
    "anm_map_point_event",
    { res = res, animfunc = "event" },
    multiset
  );

  ui.obj.Multiset( multiset );

  local function func_create_stage ( stage, progress )

    private.Stages_PointCreate( stage.name, section_name, multiset );
    v_interface.Loading_SetProgress( progress, 0.0, 0.1 );

  end;

  common.ProcessArrayByTimer(
    stage_list,
    func_create_stage,
    function ()

      ui.obj.Multiset( multiset );
      func_after();

    end,
    private.create_at_once
  );

end;
--------------------------------------------------------------------------------
function private.Create_Builds ( func_after )

  local section_name = v_sections.Section_NameCurrent();
  local builds_list  = v_builds.GetAllInSection( section_name );

  local function func_create_build ( build, progress )

    v_builds.Create( build );
    v_interface.Loading_SetProgress( progress, 0.1, 0.2 );

  end;

  common.ProcessArrayByTimer(
    builds_list,
    func_create_build,
    func_after,
    private.create_at_once
  );

end;
--------------------------------------------------------------------------------
function private.Create_SetZones ( func_after )

  local section_name = v_sections.Section_NameCurrent();
  local builds_list  = v_builds.GetAllInSection( section_name );

  local function func_set_zone ( build, progress )

    private.Stages_SetByBuild( build );
    v_interface.Loading_SetProgress( progress, 0.8, 0.9 );

  end;

  common.ProcessArrayByTimer(
    builds_list,
    func_set_zone,
    func_after,
    private.create_at_once
  );

end;
--------------------------------------------------------------------------------
function private.Create_Mines ( func_after )

  local section_name = v_sections.Section_NameCurrent();
  local mines_list   = v_mines.GetAllBySection( section_name );

  local function func_create_mine ( mine, progress )

    v_mines.Mine_Create( mine );
    v_interface.Loading_SetProgress( progress, 0.7, 0.8 );

  end;

  common.ProcessArrayByTimer(
    mines_list, func_create_mine, func_after, private.create_at_once
  );

end;
--------------------------------------------------------------------------------
function private.Create_Fights ( func_after )

  local section_name = v_sections.Section_NameCurrent();
  local fights_list  = v_stage.Stage_ListByType(
    section_name,
    { "fight", "misc" }
  );

  local function func_create_fight ( fight_data, progress )

    fight_data = v_fights.Fight_Data( fight_data.name );

    v_fights.Create( fight_data, section_name );
    v_interface.Loading_SetProgress( progress, 0.4, 0.5 );

  end;

  common.ProcessArrayByTimer(
    fights_list, func_create_fight, func_after, private.create_at_once
  );

end;
--------------------------------------------------------------------------------
function private.Create_Characters ( func_after )

  local section_name    = v_sections.Section_NameCurrent();
  local characters_list = v_characters.GetAllInSection( section_name );

  local function func_char_create ( character, progress )

    v_characters.Create( character, section_name );
    v_interface.Loading_SetProgress( progress, 0.3, 0.4 );

  end;

  common.ProcessArrayByTimer(
    characters_list, func_char_create, func_after, private.create_at_once
  );

end;
--------------------------------------------------------------------------------
function private.Create_Spells ( func_after )

  local section_name = v_sections.Section_NameCurrent();
  local spells_list  = v_spells.Static_GetAllBySection( section_name );

  local function func_spell_create ( spell, progress )

    v_spells.Spell_Create( spell.name, section_name );
    v_interface.Loading_SetProgress( progress, 0.6, 0.7 );

  end;

  common.ProcessArrayByTimer(
    spells_list, func_spell_create, func_after, private.create_at_once
  );

end;
--------------------------------------------------------------------------------
function private.Create_Chests ( func_after )

  local section_name = v_sections.Section_NameCurrent();
  local chest_list   = v_chests.List_OnSection( section_name );

  local function func_create_chest ( chest, progress )

    v_chests.Create( chest.name, section_name );
    ObjSet( "obj_map_chests_"..chest.name, { active = false } );
    v_interface.Loading_SetProgress( progress, 0.5, 0.6 );

  end;

  common.ProcessArrayByTimer(
    chest_list, func_create_chest, func_after, private.create_at_once
  );

end;
--------------------------------------------------------------------------------
function private.Create_Carrier ()

  v_carriers.Create();

end;
--------------------------------------------------------------------------------
function private.Create_PortalSign ()

  local section = v_sections.Section_NameCurrent();

  local hub_obj = "anm_map_portal_next";
  local suffix  = "_"..section;
  local parent  = "obj_map_portal_"..section.."_sign";

  ObjClone( hub_obj, suffix, true, parent );

  ui.obj.Set( hub_obj..suffix, { label = "str_map_portal_next_sign" } );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Marker_CurrentPoint ()

  return private.marker_point_current;

end;
--------------------------------------------------------------------------------
function private.Marker_SetSkin ( skin_name )

  skin_name     = skin_name or v_player.Skin_GetCurrent();
  local content = v_player.Skin_GetContent( skin_name );

  ui.obj.Set(
    "anm_map_marker_icon_skin",
    {
      res      = "assets/content/"..content.."/anm_skin_"..skin_name,
      animfunc = "marker",
      playing  = true
    }
  );

end;
--------------------------------------------------------------------------------
function private.Marker_Transfer ( from_screen, to_screen )

  local section_name = "section_"..to_screen;

  local marker_at_stage;
  local marker_is_visible = true;

  if ( from_screen <= to_screen ) then

    local current_stage_name   = v_player.Progress_GetLastStageName();
    local check_section_name   = v_sections.Section_NameByStage( current_stage_name );
    local check_section_number = tonumber( string.match( check_section_name, "%d+" ) );

    if ( check_section_number == to_screen ) then

      marker_at_stage = current_stage_name;

    else

      marker_at_stage = private.Marker_DetectLastPoint( section_name );

    end;

  elseif ( from_screen > to_screen ) then

    marker_at_stage = v_sections.Section_LastStage( section_name );

  end;

  private.Marker_SetAtPoint( marker_at_stage );

  ui.obj.Set( private.this.marker, { visible = marker_is_visible } );

end;
--------------------------------------------------------------------------------
function private.Marker_SetAtLastPoint ( ignore_stage_info )

  local section_name = v_sections.Section_NameCurrent();
  local last_stage   = private.Marker_DetectLastPoint( section_name );

  private.Marker_SetAtPoint( last_stage, true, ignore_stage_info );

end;
--------------------------------------------------------------------------------
function private.Marker_DetectLastPoint ( section_name )

  section_name = section_name or v_sections.Section_NameCurrent();

  -- Try set marker at LAST section point.
  --[[local last_point = v_sections.Section_LastStage( section_name );

  if ( not v_stage.Stage_GetInfo( last_point ).completed ) then
    -- Try set marker at FIRST section point.
    last_point = v_sections.Section_FirstStage( section_name );

    if ( not v_stage.Stage_GetInfo( last_point ).completed ) then
      v_player.Progress_SetLastStageName( last_point, section_name );
    else
      -- Try set marker at LAST ENABLED section point.
      last_point = v_sections.Stage_GetLastEnabled();
      v_player.Progress_SetLastStageName( last_point, section_name );
    end;

  end;]]

  -- Last point is the 'stage_last' from 'player/info'.
  local last_point = v_player.Progress_GetLastStageName();

  local is_section_completed = not v_sections.Section_NeedOutro( section_name );

  if ( is_section_completed ) then
    -- Last point is the LAST section stage.
    last_point = v_sections.Section_LastStage( section_name );
    return last_point;
  end;

  local last_point_section  = v_stage.Stage_GetSection( last_point );
  local is_section_mismatch = ( section_name ~= last_point_section );

  if ( is_section_mismatch ) then
    --JK: select first stage, as the last may be locked
    last_point = v_sections.Section_FirstStage( section_name );
    return last_point;
  end;

  return last_point;

end;
--------------------------------------------------------------------------------
function private.Marker_Target ( stage_name, ignore_stage_info )

  local target_point = "obj_map_point_"..stage_name;

  if ( ignore_stage_info ) then
    return target_point, stage_name;
  end;

  local stage_info = v_stage.Stage_GetInfo( stage_name );

  if ( stage_info.completed ) then

    if ( stage_info.realtype ~= "misc"                                ) or
       ( stage_info.realtype == "misc" and stage_info.type ~= "fight" ) then

      stage_name   = v_stage.Stage_NameNext( stage_name );
      target_point = "obj_map_point_"..stage_name;
      
      -- JK: Autofix for stuck at disabled point,
      -- excluding last stage in section
      
      local last_stage = v_sections.Section_LastStage(
        v_sections.Section_NameCurrent(),
        true
      );
      
      if ( stage_name ~= last_stage ) then
      
        local is_enabled, was_enabled = v_stage.Stage_TryEnable( stage_name );
        
        if ( not was_enabled ) then
          Log.w( "Stage force-enabled: "..stage_name );
          private.Stages_PointSet( stage_name );
        end;
      
      end;

    end;

  end;

  return target_point, stage_name;

end;
--------------------------------------------------------------------------------
function private.Marker_SetAtPoint ( stage_name, visible, ignore_stage_info )

  local target_point;
  local multiset = {};

  target_point, stage_name     = private.Marker_Target( stage_name, ignore_stage_info );
  private.marker_point_current = stage_name;

  public.Focus( target_point, { instant = true } );

  local point_params = private.Object_Position( target_point );

  ui.input.On( target_point, multiset );

  ui.obj.Set(
    private.this.marker,
    {
      pos_x = point_params[ 1 ] + private.offset,
      pos_y = point_params[ 2 ] - private.offset,
      visible = visible
    },
    multiset
  );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.Marker_MoveToPoint ( stage_name, execute_tap, no_focus, func_end )

  if ( not no_focus ) then

    public.Focus( private.Marker_Target( stage_name ), { time = 0.5 } );

  end;

  private.marker_point_current  = stage_name;
  private.marker_target_execute = execute_tap;

  local target_object = private.Marker_Target( stage_name );
  local target_params = private.Object_Position( target_object );

  local button = "btn_"..ui.obj.GetNameWithoutPrefix( target_object );
  local events = ui.control.EventsGet( button );

  local func_tap = function () end;
  if ( events ) then func_tap = events.func_on_click; end;

  ObjSet( target_object, { input = true } );

  private.marker_target_params =
  {
    pos_x    = target_params[ 1 ],
    pos_y    = target_params[ 2 ],
    func_tap = func_tap,
    name     = target_object
  };

  private.Marker_MoveToPoint_Animate( func_end );

end;
--------------------------------------------------------------------------------
function private.Marker_MoveToPoint_Animate ( func_end )

  local func_OnEnd = function ()

    private.Marker_MoveToPoint_Done( func_end );

  end;

  local marker_object = private.this.marker;

  ObjStopAnimate( marker_object, "pos_xy" );

  local marker_params = ObjGet( marker_object );
  local target_params = private.marker_target_params;

  ObjAnimate( marker_object, "pos_xy", 0, 0, func_OnEnd,
  {
    0.0, 3, marker_params.pos_x, marker_params.pos_y,
    0.2, 3, marker_params.pos_x, marker_params.pos_y,
    0.5, 3, target_params.pos_x + private.offset, target_params.pos_y - private.offset
  } );

end;
--------------------------------------------------------------------------------
function private.Marker_MoveToPoint_Done ( func_end )

  -- Turned off: GD's request.
  -- PartSysRestart( "pfx_map_marker_done" );

  common.sound.Play( "open_new_level" );

  if ( func_end ) then

    func_end();

  end;

  if ( private.marker_target_execute ) then

    private.marker_target_params.func_tap();

  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Point_Tap ( stage_name, skip_focus )

  local stage_info = v_stage.Stage_GetInfo( stage_name );

  if ( not stage_info.enabled ) then

    return;

  end;

  if ( stage_name ~= private.marker_point_current ) and
     ( not stage_info.completed                   ) then

    private.Marker_MoveToPoint( stage_name, true );

    return;

  end;

  local function func_dialog ()

    v_stage.DialogStart_Show( stage_name, "stages" );

  end;

  if ( not skip_focus ) then

    -- JK: map scale normalization
    public.Focus(
      "obj_map_point_"..stage_name,
      {
        scale      = 1.0,
        func_after = func_dialog
      }
    );

  else

    func_dialog();

  end;

  private.last_point_selected = stage_name;

end;
--------------------------------------------------------------------------------
function public.Point_Enable ( stage_name )

  private.Stages_PointEnable( stage_name );

end;
--------------------------------------------------------------------------------
function public.Point_SuggestByReward ( reward_type, reward_name )

  local stages_by_reward = v_stage.Stages_FindByReward( reward_type, reward_name );
  local stages_by_quests = v_quests.Stages_FindByReward( reward_type, reward_name );

  local stage_list = {};

  for _, stage_name in pairs( stages_by_reward ) do
    table.insert( stage_list, stage_name )
  end;

  for _, stage_name in pairs( stages_by_quests ) do
    table.insert( stage_list, stage_name )
  end;

  public.Point_SuggestRandom( stage_list );

end;
--------------------------------------------------------------------------------
function public.Point_SuggestRandom ( stage_list )

  if ( v_interface.window.IsAnyOnScreen() ) then
    return;
  end;

  if ( #stage_list < 1 ) then return; end;

  local index      = math.random( 1, #stage_list );
  local stage_name = stage_list[ index ];
  local obj_name   = "obj_map_point_"..stage_name;
  local btn_name   = "btn_map_point_"..stage_name;

  if ( not ui.obj.IsVisible( btn_name ) ) then

    return;

  end;

  ui.screen.Lock( true );

  public.Focus(
    obj_name,
    {
      func_after = function ()

        ui.anim.Spin( obj_name, function () ui.screen.Lock( false ); end );

      end,
      scale = public.Scale_GetMax()
    }
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.IsBeingDragged ()

  return private.is_dragged;

end;
--------------------------------------------------------------------------------
function private.DragStart ()

  private.is_dragged = true;

  --mediator.Notify( "View_Map_DragStarted" );

  ui.control.TapBreak();
  ui.popup.Cloud_Hide();
  ui.button.TryReleasePressed();

  --[[
  local tLocPosMap = ObjGet( private.this.body );
  local tLocCursor = GetGameCursorPos();

  private.tPosObjMap = { tLocPosMap.pos_x,   tLocPosMap.pos_y };
  private.tPosCursor = { tLocCursor[ 1 ], tLocCursor[ 2 ] };
  ]]

end;
--------------------------------------------------------------------------------
function private.DragUpdate ()

  --[[
  local tLocCursor = GetGameCursorPos();
  local tDelta = {
    private.tPosCursor[ 1 ] - tLocCursor[ 1 ],
    private.tPosCursor[ 2 ] - tLocCursor[ 2 ]
  };

  private.tPosCursor = { tLocCursor[ 1 ], tLocCursor[ 2 ] };

  private.tPosObjMap = private.CheckPosition( { private.tPosObjMap[ 1 ] - tDelta[ 1 ], private.tPosObjMap[ 2 ] - tDelta[ 2 ] }, private.map_scale_current );

  ObjSet( private.this.body, { pos_x = private.tPosObjMap[ 1 ], pos_y = private.tPosObjMap[ 2 ] } );
  ]]

  --private.GroundUpdate();

end;
--------------------------------------------------------------------------------
function private.DragDrop ()

  private.map_scale_current = ObjGet( private.this.body ).scale_x;

  --mediator.Notify( "View_Map_DragDropped" );

  private.is_dragged = nil;

  --ObjSet( "obj_map_drag", { pos_x = 0, pos_y = 0 } );

end;
--------------------------------------------------------------------------------
function private.CheckPosition ( position, scale )

  if ( position[ 1 ] > 0 ) then

    position[ 1 ] = 0;

  end;

  if ( position[ 1 ] < ( - private.scale_modifier * private.map_w * scale ) + private.app_w ) then

    position[ 1 ] = ( - private.scale_modifier * private.map_w * scale ) + private.app_w;

  end;

  local screen_height = private.scale_modifier * private.map_h;

  if ( position[ 2 ] > screen_height ) then

    position[ 2 ] = screen_height;

  end;

  local app_height = private.app_h;

  if ( position[ 2 ] < app_height ) then

    position[ 2 ] = app_height;

  end;

  return position;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.AppearInsertToQueue ( obj_type, animation )

  if ( not private.appear_queue[ obj_type ] ) then
    private.appear_queue[ obj_type ] = {};
  end;

  table.insert( private.appear_queue[ obj_type ], animation );

end;
--------------------------------------------------------------------------------
function public.AppearClearQueue ( obj_type )

  if ( obj_type ) then
    private.appear_queue[ obj_type ] = {};
  end;

end;
--------------------------------------------------------------------------------
function public.AppearClearAllQueue ()

  private.appear_queue = {};

end;
--------------------------------------------------------------------------------
-- function ******* Map Scale ******* () end
--------------------------------------------------------------------------------
function private.ShowFocusPoint ( stage_name )

  local stage_info = v_stage.Stage_GetInfo( stage_name );

  if ( stage_info.enabled ) then

    local focus_name = string.format( "fcs_map_point_%s", stage_name );

    local tme = 0.3;

    local alp = ObjGet( focus_name ).alp;

    ObjSet( focus_name, { active = true, visible = true } );
    ObjAnimate( focus_name, "alp", 0, 0, "",
    {
      0.0, 2, alp,
      tme, 2, 1
    } );

  end

end;
--------------------------------------------------------------------------------
function private.HideFocusPoint ( stage_name )

  local focus_name = string.format( "fcs_map_point_%s", stage_name );

  local tme = 0.3;

  local alp = ObjGet( focus_name ).alp;

  ObjAnimate( focus_name, "alp", 0, 0,
  function()
    ObjSet( focus_name, { active = false, visible = false } );
  end,
  {
    0.0, 2, alp,
    tme, 2, 0
  } );

end;
--------------------------------------------------------------------------------
function private.Object_Position ( object, relative )

  relative = relative or "anm_map_"..v_sections.Section_NameCurrent();

  return ui.obj.Position( object, relative );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Builds_UpdateStages ( build_data, func_anim, func_after )

  common.sound.Play( "unspell_button" );

  local function func_end ()

    v_interface.Show( { freeze = false } );
    if ( func_after ) then func_after(); end;

  end;

  v_interface.Hide( { freeze = true } );

  if ( build_data.type == "build" ) then

    private.Builds_TakeResources( build_data, function ()

      common.sound.Play( "catch_glow" );

      if ( func_anim ) then

        func_anim( function ()
          private.Stages_UpdateByBuild( build_data, func_end );
        end );

      else

        private.Stages_UpdateByBuild( build_data, func_end );

      end;

    end );

  else

    private.Stages_UpdateByBuild( build_data, func_end );

  end;

end;
--------------------------------------------------------------------------------
function private.Builds_TakeResources ( build_data, func_end )

  local build_name = build_data.name;
  local price      = build_data.price;

  if ( not price ) then

    func_end();
    return;

  end;

  ui.screen.Lock( true, "Builds_TakeResources" );

  local build_pos = private.Object_Position(
    "obj_map_builds_"..build_name,
    "obj_map_anims"
  );

  local flying_objects = 0;

  for j = 1, #price do

    local max_objects = math.min( price[ j ].count, 15 );
    local current_obj = 0;
    local resource    = price[ j ].resource;

    local res_pos = private.Object_Position(
      "obj_int_panel_interface_top_"..resource,
      "obj_map_anims"
    );

    for i = 1, max_objects do

      current_obj = current_obj + 1;

      flying_objects = flying_objects + 1;

      local obj_name = "obj_map_flying_resources_"..resource.."_"..current_obj;
      local anm_name = "anm_map_flying_resources_"..resource.."_"..current_obj;
      local res_name = "anm_map_flying_resources_icon_"..resource.."_"..current_obj;
      local pfx_name = "pfx_map_flying_resources_"..resource.."_"..current_obj;

      ObjClone( "obj_map_flying_resources", "_"..resource.."_"..current_obj, 1, "obj_map_anims" );

      ObjSet( obj_name, { pos_x = res_pos[ 1 ], pos_y = res_pos[ 2 ], pos_z = ( 10 + current_obj ) } );
      ObjSet( res_name, { animfunc = resource, playing = true } );

      local animfunc = math.random( 1, 8 );

      local resource_type = "resource";

      if ( string.find( resource, "gold" ) ) then

        resource_type = "gold";

      end;

      ui.anim.Play( anm_name, resource_type..animfunc );

      local shift = math.random( 1, max_objects );
      local type  = math.random( 0, 2 );
      local delay = math.random( 1, max_objects ) / 100;

      local func_OnAnimationEnd = function ()

        ObjDelete( obj_name );

        flying_objects = flying_objects - 1;

        if ( flying_objects < 1 ) and ( func_end ) then

          ui.screen.Lock( false, "Builds_TakeResources" );
          func_end();

        end;

      end;

      ObjSet( pfx_name, { visible = true, active = true } );

      PartSysStart( pfx_name );

      ObjAnimate( obj_name, "pos_x", 0, 0, func_OnAnimationEnd,
      {
        0.0, type, res_pos[ 1 ],
        0.5, type, res_pos[ 1 ] + 15 * shift,
        1.0 + delay, type, res_pos[ 1 ] + 15 * shift,
        1.5 + delay, type, build_pos[ 1 ]
      } );

      shift = math.random( 1, max_objects );
      type  = math.random( 0, 2 );

      ObjAnimate( obj_name, "pos_y", 0, 0, "",
      {
        0.0, type, res_pos[ 2 ],
        0.5, type, res_pos[ 2 ] + 150 + 15 * shift,
        1.0 + delay, type, res_pos[ 2 ] + 150 + 15 * shift,
        1.5 + delay, type, build_pos[ 2 ]
      } );

      ObjAnimate( obj_name, "alp", 0, 0, "",
      {
        0.0, 0, 1,
        0.5, 0, 1,
        1.0 + delay, 0, 1,
        1.5 + delay, 1, 0
      } );

      ObjAnimate( obj_name, "scale_xy", 0, 0, "",
      {
        0.0,         0, 1.0, 1.0,
        0.5,         0, 1.0, 1.0,
        1.0 + delay, 0, 1.0, 1.0,
        1.5 + delay, 1, 0.5, 0.5
      } );

    end;

  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Stages_PointCreate ( stage_name, section_name, multiset )

  -- Create and attach base object
  local point_name = "obj_map_point_"..stage_name;

  ObjCreate( point_name, "obj" );
  ui.obj.Set( point_name, { anim_tag = stage_name }, multiset );

  ui.obj.Attach( point_name, "anm_map_"..section_name );
  ui.obj.Clone( "btn_map_point", point_name, "_"..stage_name, 1 );

  local funcs =
  {
    func_on_click = function ()
      private.Point_Tap( stage_name );
    end
  };

  if ( IsPC() ) then

    funcs.func_on_mouse_enter = function ()
      private.ShowFocusPoint( stage_name );
    end;

    funcs.func_on_mouse_leave = function ()
      private.HideFocusPoint( stage_name );
    end;

  end;

  ui.control.EventsSet( "btn_map_point_"..stage_name, funcs );

  -- Set text label
  if not ( string.match( stage_name, "bonus" ) ) then

    private.Stages_PointNumberSet( stage_name, multiset );

  end;

end;
--------------------------------------------------------------------------------
function private.Stages_PointSet ( stage_name, hide_quest_reward, animated )

  local obj_point = "btn_map_point_"..stage_name;
  local active    = {
    [ true  ] = { active  = true,  visible = true  },
    [ false ] = { active  = false, visible = false }
  };

  local set = {};

  ui.obj.Enable( obj_point, false, set );

  local stage_info = v_stage.Stage_GetInfo( stage_name );
  local enabled    = stage_info.enabled or false;
  local ride       = ( stage_info.type     == "ride" );
  local fight      = ( stage_info.type     == "fight" );
  local realfight  = ( stage_info.realtype == "fight" );
  local realbonus  = ( stage_info.realtype == "bonus" );
  local state      = ( stage_info.state or 1 ) - 1;
  local completed  = ( stage_info.noreward or false );
  local disabled   = ( stage_info.completed and realfight );

  -- Setup input
  table.insert( set, { obj_point, { input = enabled } } );

  -- Setup background
  local anm_bg = "anm_map_point_bg_"..stage_name;
  local bg_af  =
  {
    [ "default"   ] = "bg_default",
    [ "active"    ] = "bg_active",
    [ "completed" ] = "bg_completed",
    [ "disabled"  ] = "bg_disabled"
  };

  local bg_animfunc = bg_af.active;

  if ( disabled ) then

    bg_animfunc = bg_af.disabled;

  elseif ( completed ) then

    bg_animfunc = bg_af.completed;

  elseif ( not enabled or ( state <= 0 ) ) then

    bg_animfunc = bg_af.default;

  end;

  ui.obj.Set(
    anm_bg,
    { animfunc = bg_animfunc },
    set
  );

  -- Setup label
  local anm_label = "anm_map_point_label_"..stage_name;
  local label_af  =
  {
    [ "empty"  ] = "label_empty",
    [ "bonus"  ] = "label_bonus",
    [ "ride"   ] = "label_number",
    [ "fight"  ] = "label_fight"
  };

  local label_animfunc = label_af[ stage_info.type ];

  if ( disabled ) then

    label_animfunc = label_af.fight;

  elseif not ( enabled ) then

    label_animfunc = tern( fight and realfight )( label_af.fight, label_af.empty );

  elseif ( realbonus ) then

    label_animfunc = label_af.bonus;

  end;

  ui.obj.Set(
    anm_label,
    { animfunc = label_animfunc },
    set
  );

  -- Setup state
  local anm_state = "anm_map_point_state_"..stage_name;
  local state_af  =
  {
    [ 0 ] = "state_0",
    [ 1 ] = "state_1",
    [ 2 ] = "state_2",
    [ 3 ] = "state_3"
  };

  local state_animfunc = state_af[ state ];

  if ( fight or not enabled ) then

    state_animfunc = state_af[ 0 ];

  elseif ( completed ) then

    state_animfunc = state_af[ 3 ];

  end;

  ui.obj.Set(
    anm_state,
    { animfunc = state_animfunc },
    set
  );

  -- Setup reward
  local obj_reward = "obj_map_point_reward_"..stage_name;

  local reward   = stage_info.reward;
  local is_shown = ( not ( hide_quest_reward and reward and reward.type == "quest" ) );

  if ( stage_info.reward ) and
     ( is_shown          ) then

    local anm_reward      = "anm_map_point_reward_"..stage_name;
    local anm_reward_icon = "anm_map_point_reward_icon_"..stage_name;

    ui.anim.Play(
      anm_reward_icon,
      "reward_"..stage_info.reward.name:gsub( "_%d+$", "" )
    );

    ui.obj.Enable( obj_reward, false, set );

    if ( animated ) then

      ui.anim.Play(
        "anm_map_point_reward_"..stage_name,
        "appear",
        nil,
        "static"
      );

    else

      ui.anim.Play( "anm_map_point_reward_"..stage_name, "static" );

    end;

  else

    ui.obj.Disable( obj_reward, false, set );

  end;

  ui.obj.Multiset( set );

end;
--------------------------------------------------------------------------------
function private.Stages_PointUpdate (
    stage_name,
    animated,
    hide_quest_reward,
    func_after
  )

  if ( not ( ObjGet( "btn_map_point_"..stage_name ) or {} ).visible ) then

    if ( func_after ) then func_after(); end;
    return;

  end;

  local function func_update ()

    private.Stages_PointSet( stage_name, hide_quest_reward, animated.reward );
    if ( func_after ) then func_after(); end;

  end;

  if ( animated.point ) then

    local function func_swap ()

      ObjAnimate( "btn_map_point_"..stage_name, "scale_x", 0, 0, "",
      {
        0.0, 0, 0.0,
        0.3, 0, 1.0
      } );

      ObjAnimate( "anm_map_point_state_"..stage_name, "alp", 0, 0, "",
      {
        0.0, 0, 0,
        0.3, 0, 0,
        0.6, 0, 1
      } );

      ObjAnimate( "anm_map_point_state_"..stage_name, "scale_xy", 0, 0, "",
      {
        0.0, 0, 1.5, 1.5,
        0.3, 0, 1.5, 1.5,
        0.6, 0, 1.0, 1.0
      } );

      func_update();

    end;

    ObjAnimate( "btn_map_point_"..stage_name, "scale_x", 0, 0, func_swap,
    {
      0.0, 0, 1.0,
      0.3, 0, 0.0
    } );

  else

    func_update();

  end;

end;
--------------------------------------------------------------------------------
function private.Stages_PointNumberSet ( stage_name, multiset )

  local number = private.Stages_RelativeNumber( stage_name );
  local digits = common.math.NumberToDigits( number );

  ui.obj.Set(
    "anm_map_point_label_number_"..stage_name,
    {
      animfunc = "number_"..(#digits)
    },
    multiset
  );

  for i = 1, #digits do

    ui.obj.Set(
      "anm_map_point_label_number_digit_"..i.."_"..stage_name,
      {
        animfunc = "digit_"..digits[ i ]
      },
      multiset
    );

  end;

end;
--------------------------------------------------------------------------------
function private.Stages_PointEnable ( stage_name )

  local multiset  = {};
  local anm_label = "anm_map_point_label_"..stage_name;

  ui.obj.Enable( "btn_map_point_"..stage_name, false, multiset );

  local stage_data = v_stage.Stage_StaticData( stage_name );

  if ( stage_data.realtype == "bonus" ) then

    ui.obj.Set( anm_label, { animfunc = "label_bonus" }, multiset );

  elseif ( stage_data.type == "ride" ) then

    ui.obj.Set( anm_label, { animfunc = "label_number" }, multiset );

    private.Stages_PointNumberSet( stage_name, multiset );

  end;

  ui.obj.Multiset( multiset );

  ui.anim.FadeIn( 0.15, anm_label );

end;
--------------------------------------------------------------------------------
function private.Stages_GetName ( stage_num, stage_type )

  local result = "stage_"..stage_num;
  if ( stage_type == "bonus" ) then
    result = result.."_bonus";
  end;
  return result;

end;
--------------------------------------------------------------------------------
function private.Stages_RelativeNumber ( stage_name )

  local section_current = v_sections.Section_NumberCurrent();

  return tonumber( string.match( stage_name, "%d+" ) ) - 110 * ( section_current - 1 );

end;
--------------------------------------------------------------------------------
function public.Stages_UpdateByBuild ( build_data, func_after )

  private.Stages_UpdateByBuild( build_data, func_after );

end;
--------------------------------------------------------------------------------
function private.Stages_UpdateByBuild ( build_data, func_after )

  local list = v_builds.Build_GetUnlockList( build_data, true );

  if ( list and list.stages_next and #list.stages_next > 0 ) then

    private.AnimateZone(
      build_data.name,
      list.stages_next,
      list.stages_type,
      function ()

        --private.Stages_Show_Instant(
        --  list.stages_next[ 1 ],
        --  list.stages_next[ 2 ],
        --  list.stages_type
        --);

        if ( func_after ) then func_after(); end;

      end
    );

  elseif ( func_after ) then

    func_after();

  end;

end;
--------------------------------------------------------------------------------
function private.Stages_SetByBuild ( build_data )

  local list = v_builds.Build_GetUnlockList( build_data, false );

  if ( list ) then

    if ( list.stages_previous and #list.stages_previous > 0 ) then

      private.Stages_Show_Instant( list.stages_previous[ 1 ],
                                   list.stages_previous[ 2 ] );

    end;

    if ( list.stages_next and #list.stages_next > 1 ) then

      private.Stages_Show_Instant( list.stages_next[ 1 ],
                                   list.stages_next[ 2 ],
                                   list.stages_type );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Stages_TrySetWorkerSign ( stage_name, objects )

  local worker_state = v_stage.Stage_WorkerState( stage_name );

  if not ( worker_state ) then

    return;

  end;

  local stage_data = v_stage.Stage_GetInfo( stage_name );

  if ( stage_data.state <= worker_state ) and not ( stage_data.noreward ) then

    table.insert( objects, { "obj_map_point_reward_"..stage_name,
    {
      active  = true,
      visible = true } } );

    table.insert( objects, { "anm_map_point_reward_"..stage_name,
    {
      animfunc = "static",
      playing  = true } } );

    table.insert( objects, { "anm_map_point_reward_icon_"..stage_name,
    {
      animfunc = "reward_worker" } } );

  end;

end;
--------------------------------------------------------------------------------
function private.Stages_Show_Instant ( start_id, stop_id, stage_type )

  for i = start_id, stop_id do

    local stage_name = private.Stages_GetName( i, stage_type );
    private.Stages_PointSet( stage_name );
    v_chests.Chest_SetOnMap( stage_name );

  end;

end;
--------------------------------------------------------------------------------
function public.Zone_ShowCascade (
    start_id,
    stop_id,
    func_end,
    stages_type
  )

  private.Stages_Show_Cascade (
    start_id,
    stop_id,
    func_end,
    stages_type
  );

end;
--------------------------------------------------------------------------------
function private.Stages_Show_Cascade (
    start_id,
    stop_id,
    func_end,
    stages_type
  )

  local stage_type = stages_type or "common";
  local stage_name = private.Stages_GetName( start_id, stage_type );

  local func_OnSwitch = function ()

    if ( start_id == stop_id ) then

      if ( func_end ) then func_end(); end;
      return;

    end;

    private.Stages_Show_Cascade( start_id + 1, stop_id, func_end, stages_type );

  end;

  local func_OnAnimationEnd = function ()

    local func = v_chests.Callback_AnimateShow( stage_name );

    if ( func ) then
      func( func_OnSwitch );
    else
      func_OnSwitch();
    end;

  end;

  local multiset = {};

  private.Stages_TrySetWorkerSign( stage_name, multiset );

  ui.obj.Set(
    "btn_map_point_"..stage_name,
    {
      active  = true,
      visible = true,
      alp     = 0
    },
    multiset
  );

  ui.obj.Multiset( multiset );

  private.Stages_PointUpdate( stage_name, { point = false, reward = true } );

  ObjAnimate( "btn_map_point_"..stage_name, "alp", 0, 0, func_OnAnimationEnd,
  {
    0.0, 0, 0,
    0.2, 0, 1
  } );

  ObjAnimate( "btn_map_point_"..stage_name, "scale_x", 0, 0, "",
  {
    0.0, 0, 0,
    0.2, 0, 1
  } );

  local zone_id  = math.floor( start_id / 11 );
  local sound_id = start_id - zone_id * 11;

  sound_id = tern( sound_id == 0 )( 11, sound_id );

  common.sound.Play( "level_app"..sound_id );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Portal_Unlock ( func_after )

  local section_name = v_sections.Section_NameCurrent();

  ui.screen.Lock( true );
  ObjAnimate( private.this.marker, "alp", 0, 0, function()

    ui.obj.Hide( private.this.marker );

    ui.anim.Play(
      "anm_map_portal_"..section_name.."_next",
      "ready_start",
      function ()

        ui.anim.Play(
          "anm_map_portal_"..section_name.."_next",
          "ready"
        );

        -- JK: workaround for section 3
        if ( section_name == "section_3" ) then

          ObjSet( "btn_map_portal_"..section_name.."_next",
          {
            inputrect_h = 240,
            inputrect_w = 240,
            inputrect_y = -200,
            inputrect_x = -120
          } );

        end;

      end
    );

    public.ShowObjectPopup(
      "hint",
      "obj_map_portal_"..section_name.."_next_indicator",
      "str_int_popup_portal_"..section_name.."_next_hero",
      nil,
      nil,
      {
        func_preclose = function ()

          ui.screen.Lock( false );
          if ( func_after ) then func_after(); end;

        end
      }
    );

  end,
  {
    0.0, 0, 1,
    1.0, 0, 0
  } );

end;
--------------------------------------------------------------------------------
function private.Portal_Show ()

  local section_name      = v_sections.Section_NameCurrent();
  local next_section_name = v_sections.Section_NameNext( section_name );
  local status            = v_sections.Section_Status( next_section_name );

  local multiset = {};

  local portal_icon = v_sections.Portal_Name( section_name );

  if ( portal_icon ) then
    ui.obj.Set(
      "anm_map_portal_"..section_name.."_next",
      {
        res = "assets/content/"..section_name.."/map/anm_"..portal_icon
      },
      multiset
    );
  else
    DbgTrace( "[ WARNING ] Portal icon is not set!" );
  end;

  if ( status == "available" ) or
     ( status == "enabled"   ) then

    ui.obj.Hide( private.this.marker, multiset );

    if ( v_sections.Section_NeedOutro( section_name ) ) then

      ui.obj.Set(
        "anm_map_portal_"..section_name.."_next",
        {
          animfunc = "ready",
          playing  = true
        },
        multiset
      );

      -- JK: workaround for section 3
      if ( section_name == "section_3" ) then

        ui.obj.Set(
          "btn_map_portal_"..section_name.."_next",
          {
            inputrect_h = 240,
            inputrect_w = 240,
            inputrect_y = -200,
            inputrect_x = -120
          },
          multiset
        );

      end;

    else

      ui.obj.Disable(
        "btn_map_portal_"..section_name.."_next",
        false,
        multiset
      );

    end;

  elseif ( status == "prepare" ) then

    ui.obj.Set(
      "anm_map_portal_"..section_name.."_next",
      {
        animfunc = "prepare",
        playing  = true
      },
      multiset
    );

  else

    ui.obj.Set(
      "anm_map_portal_"..section_name.."_next",
      {
        animfunc = "normal",
        playing  = true
      },
      multiset
    );

    ui.obj.Set(
      private.this.marker,
      {
        active  = true,
        visible = true
      },
      multiset
    );

  end;

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.Portal_Prepare ( func_after )

  local section_name = v_sections.Section_NameCurrent();
  local multiset = {};

  ui.obj.Set(
    "anm_map_portal_"..section_name.."_next",
    {
      animfunc = "prepare",
      playing  = true
    },
    multiset
  );

  ui.obj.Multiset( multiset );

  if ( func_after ) then func_after(); end;

end;
--------------------------------------------------------------------------------
function private.Portal_ShowComingSoon ()

  local section_name = v_sections.Section_NameCurrent();

  local delegates = {

    Accept = int_dialog_map_comingsoon.Hide,

    OnHide = function ()

      v_interface.window.Hide( "int_dialog_map_comingsoon" );

    end

  };

  v_interface.window.Show(
    "int_dialog_map_comingsoon",
    function ()

      int_dialog_map_comingsoon.Show( section_name );

    end,
    nil,
    delegates
  );

end;
--------------------------------------------------------------------------------
function private.Portal_GoToNextSection ()

  private.Dialog_Worldmap_Show( true );

end;
--------------------------------------------------------------------------------
function private.Portal_Next_Tap ()

  -- JK: debug coming soon
  --private.Portal_ShowComingSoon();

  local curr_section_name   = v_sections.Section_NameCurrent();
  local next_section_name   = v_sections.Section_NameNext();
  local next_section_status = v_sections.Section_Status( next_section_name );

  local anm_portal = "anm_map_portal_"..curr_section_name.."_next";
  local btn_portal = "btn_map_portal_"..curr_section_name.."_next";

  ui.input.Off( btn_portal );

  local sections_list = v_sections.ListEnabledWithPlayableProperty();
  local disabled_next_section = true;
  
  for i = 1, #sections_list do
    if ( sections_list[ i ].name == next_section_name ) then
      disabled_next_section = false;
    end;
  end;
  
  if ( disabled_next_section ) then
    next_section_status = "available";
  end;
  
  if ( next_section_status == "available" ) then

    ui.anim.Play(
      anm_portal,
      "ready_click",
      function()

        private.Portal_ShowComingSoon();
        ui.input.On( btn_portal );

      end,
      "ready"
    );

  elseif ( next_section_status == "enabled" ) then

    if ( v_sections.Section_NeedOutro( curr_section_name ) ) then

      public.Focus( btn_portal, {
        scale = private.map_scale_max,
        func_after = function ()
          ui.obj.Disable( btn_portal );
          private.Outro_Show( private.Portal_GoToNextSection );
          v_sections.Section_SetOutro( curr_section_name, true );
          ui.input.On( btn_portal );
        end
      } );

    end;

  elseif ( next_section_status == "prepare" ) then

    private.Portal_Next_ShowPopup( private.Portal_Next_Hint );

    ui.anim.Play(
      anm_portal,
      "prepare_click",
      function() ui.input.On( btn_portal ); end,
      "prepare"
    );

  else

    private.Portal_Next_ShowPopup( private.Portal_Next_Hint );

    ui.anim.Play(
      anm_portal,
      "normal_click",
      function() ui.input.On( btn_portal ); end,
      "normal"
    );

  end;

end;
--------------------------------------------------------------------------------
function private.Portal_Next_SetTap ()

  local section_name   = v_sections.Section_NameCurrent();
  local section_config = v_sections.Section_GetConfig( section_name );

  local func_tap = private.Portal_Next_Tap;

  if ( section_config.portal.bossed ) then

    func_tap = function ()

      local boss                    = section_config.fights.boss;
      local is_boss_stage_completed = v_stage.Stage_IsCompleted( boss.stage );

      if ( not is_boss_stage_completed ) then

        local obj_name      = "obj_map_enemy_"..boss.name.."_"..boss.stage;
        local func_on_click = ui.control.EventsGet( obj_name ).func_on_click;

        if ( func_on_click ) then func_on_click(); end;

      else

        private.Portal_Next_Tap();

      end;

    end;

  end;

  ui.control.TapSet(
    "btn_map_portal_"..section_name.."_next",
    {
      func_true = function () func_tap(); end
    }
  );

end;
--------------------------------------------------------------------------------
function private.Portal_Next_ShowPopup ( func_preclose )

  local section_name = v_sections.Section_NameCurrent();

  public.ShowObjectPopup(
    "monolog",
    "obj_map_portal_"..section_name.."_next_indicator",
    "str_int_popup_portal_"..section_name.."_next",
    nil,
    nil,
    {
      func_preclose = func_preclose
    }
  );

end;
--------------------------------------------------------------------------------
function private.Portal_Next_Hint ()

  local section_config = v_sections.Section_GetConfig();
  local portal_blocker = section_config.portal.blocker;

  if ( not portal_blocker ) then

    DbgTrace( "[ ERROR ] Portal has no associated blocker." );

    return;

  end;

  if ( portal_blocker.type == "build" ) then

    v_builds.Build_HighlightStage( portal_blocker.name );

  elseif( portal_blocker.type == "quest" ) then

    local quest_name = portal_blocker.name;

    if ( v_quests.Quest_IsCurrent( quest_name ) ) then

      local quest_data = v_quests.Quest_Data( quest_name );

      public.Point_SuggestByReward( "quest", quest_data.target.name );

    end;

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Pending_Start ( pending_token, func_exec )

  private.Pending_Start( pending_token, func_exec );

end;
--------------------------------------------------------------------------------
function public.IsPending ()

  return private.pending;

end;
--------------------------------------------------------------------------------
function public.Pending_Complete ( pending_token )

  private.Pending_Complete( pending_token );

end;
--------------------------------------------------------------------------------
function private.Pending_Start ( pending_token, func_exec )

  -- DbgTrace( "[ VIEW ][ MAP ] Pending started: "..pending_token );

  private.pending = pending_token;

  if ( func_exec ) then

    private.Pending_AddOnComplete( pending_token, func_exec );

  end;

end;
--------------------------------------------------------------------------------
function private.Pending_AddOnComplete ( pending_token, func_exec )

  if not ( func_exec ) then

    return;

  end;

  if not ( private.pending_oncomplete ) then

    private.pending_oncomplete = {};

  end;

  if not ( private.pending_oncomplete[ pending_token ] ) then

    private.pending_oncomplete[ pending_token ] = {};

  end;

  table.insert( private.pending_oncomplete[ pending_token ], func_exec );

end;
--------------------------------------------------------------------------------
function private.Pending_Complete ( pending_token )

  -- DbgTrace( "[ VIEW ][ MAP ] Pending complete: "..pending_token );

  private.pending = nil;

  --ES> fix no show ach by tutorial
  --v_player.Achievements_TryShowPanel();

  if ( not private.pending_oncomplete                   ) or
     ( not private.pending_oncomplete[ pending_token ]  ) or
     ( #private.pending_oncomplete[ pending_token ] < 1 ) then

     -- DbgTrace( "[ VIEW ][ MAP ] Pending complete: nothing executed" );

    return;

  end;

  for f = 1, #private.pending_oncomplete[ pending_token ] do

    private.pending_oncomplete[ pending_token ][ f ]();

  end;

  private.pending_oncomplete[ pending_token ] = nil;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Intro_Prepare ()

  local section_name   = v_sections.Section_NameCurrent();
  local section_config = v_sections.Section_GetConfig( section_name );
  local intro_settings = section_config.intro;

  v_interface.Subtitles_Prepare();

  local multiset = {};

  private.Intro_PresetMap       ( section_name, intro_settings, multiset );
  private.Intro_PresetBoss      ( section_name, intro_settings, multiset );
  private.Intro_PresetForeground( section_name, intro_settings, multiset );
  private.Intro_PresetBuilds    ( section_name, intro_settings, multiset );
  private.Intro_PresetPuzzle    ( section_name, intro_settings, multiset );
  private.Intro_PresetPoints    ( section_name, intro_settings, multiset );
  private.Intro_PresetChests    ( section_name, intro_settings, multiset );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.Intro_PresetMap ( section_name, intro_settings, multiset )

  if ( section_name == "section_1" ) then

    local pos_x =
    {
      [ 640  ] = - 1152,
      [ 768  ] = - 1024,
      [ 1024 ] = - 800
    };

    ui.obj.Set(
      private.this.body,
      {
        pos_x   = pos_x[ ui.screen.Width() ],
        pos_y   = ui.screen.Height(),
        scale_x = 0.5,
        scale_y = 0.5
      },
      multiset
    );

  end;

end;
--------------------------------------------------------------------------------
function private.Intro_PresetForeground ( section_name, intro_settings, multiset )

  if ( section_name == "section_1" ) then

    ui.obj.Attach( "obj_map_foreground", InterfaceName );
    ui.obj.Enable( "obj_map_foreground_intro", true, multiset );

  end;

end;
--------------------------------------------------------------------------------
function private.Intro_PresetBuilds ( section_name, intro_settings, multiset )

  if ( section_name == "section_1" ) then

    for i = 1, 9 do

      local anm_build = "anm_map_builds_icon_build_"..i;

      ui.obj.Set(
        anm_build,
        {
          animfunc = "normal",
          playing  = true,
          frame    = 0
        },
        multiset
      );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Intro_PresetPuzzle ( section_name, intro_settings, multiset )

  if ( section_name == "section_1" ) then

    ui.obj.Set(
      "anm_map_builds_icon_state_puzzle_1",
      {
        animfunc = "intro",
        playing  = false
      },
      multiset
    );

  end;

end;
--------------------------------------------------------------------------------
function private.Intro_PresetPoints ( section_name, intro_settings, multiset )

  local stage_first = tonumber(
    string.match( v_sections.Section_FirstStage( section_name ), "%d+" )
  );

  local stage_last = stage_first + 10;

  ui.obj.Set(
    private.this.marker,
    {
      alp = 0
    },
    multiset
  );

  for i = stage_first, stage_last do

    ui.obj.Set(
      "btn_map_point_stage_"..i,
      {
        alp     = 0,
        scale_x = 0
      },
      multiset
    );

  end;

end;
--------------------------------------------------------------------------------
function private.Intro_PresetChests ( section_name, intro_settings,  multiset )

  v_chests.HideChestsInSection( section_name );

end;
--------------------------------------------------------------------------------
function private.Intro_PresetBoss ( section_name, intro_settings,  multiset )

  if ( section_name == "section_1" ) then

    local boss_name  = intro_settings.target_name;
    local boss_stage = intro_settings.target_stage;

    ui.obj.Set(
      "anm_map_point_enemy_"..boss_name.."_"..boss_stage,
      {
        animfunc = "action",
        playing  = false,
        frame    = 0
      },
      multiset
    );

  end;

end;
--------------------------------------------------------------------------------
function private.Intro_Show ( func_OnEnd )

  private.intro_active = true;

  local section_name   = v_sections.Section_NameCurrent();
  local section_config = v_sections.Section_GetConfig( section_name );
  local intro_settings = section_config.intro;

  --[[if ( section_name == "section_1" ) then

    v_tutorial.Kontagent_LogEvent( "tutorial",
    {
      name  = "tut_intro"
    } );

  end;]]

  -- JK: Hide carrier indicator
  v_carriers.OnIntroStart();

  ui.screen.Lock( true, "Intro_Show" );
  v_interface.Hide();

  v_sections.Section_SetIntro( section_name, true );

  local stage_first = tonumber(
    string.match( v_sections.Section_FirstStage( section_name ), "%d+" )
  );

  local stage_last  = stage_first + 10;

  local objects = {};

  local func_CascadeShow = function ()

    ObjAttach( "obj_map_foreground_movie", "obj_map_foreground" );
    ObjAttach( "obj_map_spawn", "hub_map_anims" );

    ObjSet( "obj_map_foreground_movie", { active = false, visible = false } );
    ObjSet( "obj_map_spawn",            { active = false, visible = false } );

    local func_OnCascadeEnd = function ()

      public.Focus( "obj_map_point_stage_"..stage_first,
        {
          time = 1.5,
          scale = 1.0,
          func_after = function ()

            private.Stages_PointEnable( "stage_"..stage_first );
            private.intro_active = false;
            if ( func_OnEnd ) then func_OnEnd(); end;

          end
        }
      );

    end;

    private.Stages_Show_Cascade( stage_first, stage_last, func_OnCascadeEnd );

    ObjAnimate( private.this.marker, "alp", 0, 0, "",
    {
      0.0, 0, 0,
      0.3, 0, 1
    } );

    public.Focus( "obj_map_point_stage_"..stage_last,
      {
        time  = 1.5,
        scale = 0.6
      }
    );

  end;

  ObjAttach( "obj_map_spawn", "obj_map_anims" );
  ObjAttach( "obj_map_foreground_movie", InterfaceName );

  local portal_pos =
    private.Object_Position( "obj_map_portal_"..section_name.."_prev" );

  if ( section_name ~= "section_1" ) then

    public.Focus( "obj_map_portal_"..section_name.."_prev",
      {
        instant = true,
        scale   = 1.0
      }
    );

  end;

  table.insert( objects, { "obj_map_foreground_movie",
  {
    active  = true,
    visible = true
  } } );

  table.insert( objects, { "anm_map_foreground_movie",
  {
    active  = false,
    visible = false,
    res     = "assets/content/"..section_name.."/movie/anm_intro",
    pos_x   = portal_pos[ 1 ],
    pos_y   = portal_pos[ 2 ]
  } } );

  table.insert( objects, { "obj_map_spawn",
  {
    active  = true,
    visible = true,
    pos_x   = portal_pos[ 1 ],
    pos_y   = portal_pos[ 2 ]
  } } );

  table.insert( objects, { "anm_map_spawn_hero",
  {
    active  = false,
    visible = false,
    res     = "assets/content/"..section_name.."/movie/anm_intro"
  } } );

  ObjMultiSet( objects );

  local func_skip = function ( func_OnSkipEnd )

    private.Intro_Skip( section_name, intro_settings, func_OnSkipEnd );

  end;

  local func_after_skip = function ()

    private.intro_active = false;

    if ( func_OnEnd ) then func_OnEnd(); end;

  end;

  v_interface.Subtitles_Show(
    section_name,
    "intro",
    intro_settings.subs,
    func_skip,
    func_after_skip
  );

  private.Intro_Play(
    section_name,
    1,
    intro_settings,
    func_CascadeShow
  );

end;
--------------------------------------------------------------------------------
function private.Intro_Play (
    section_name,
    intro_step,
    intro_settings,
    func_end
  )

  local intro_steps;

  if ( intro_settings.type == "default" ) then

    local target_type  = intro_settings.target_type;
    local target_stage = intro_settings.target_stage;
    local target_name  = intro_settings.target_name;
    local hero_stage   = intro_settings.hero_stage;

    local step_1_obj = "anm_map_spawn_hero";

    local step_2_obj = tern( target_type == "enemy" )(
      "obj_map_enemy_"..target_stage,
      "obj_map_characters_"..target_name
    );

    local step_3_obj = tern( target_type == "enemy" )(
      "anm_map_point_enemy_"..target_name.."_"..target_stage,
      "anm_map_characters_icon_"..target_name
    );

    local step_4_obj = "obj_map_point_"..hero_stage;
    local step_5_obj = "anm_map_spawn_hero";

    local step_3_params = tern( target_type == "enemy" )(
      {
        animfunc_start = "action",
        animfunc_end   = "static",
        animfunc_dummy = "dummy"
      },
      {
        animfunc_start = "action_"..target_name,
        animfunc_end   = "static_"..target_name
      }
    );

    intro_steps = {
      {
        func   = private.Intro_Animate,
        object = step_1_obj,
        params = {
          animfunc_start = "intro_1"
        }
      },
      {
        func   = private.Intro_Scale,
        object = step_2_obj,
        params = {
          time  = 1.0,
          scale = 1.0
        }
      },
      {
        func   = private.Intro_Animate,
        object = step_3_obj,
        params = step_3_params
      },
      {
        func   = private.Intro_Scale,
        object = step_4_obj,
        params = {
          time  = 1.0,
          scale = 1.0
        }
      },
      {
        func   = private.Intro_Animate,
        object = step_5_obj,
        params = {
          animfunc_start = "intro_2"
        }
      }
    };

  elseif ( intro_settings.type == "section_1" ) then

    intro_steps = {
      {
        func = private.Intro_AnimateForeground,
      },
      {
        func = private.Intro_AnimateHero,
      }
    };

  end;

  if ( intro_settings.type ) then

    if ( intro_step == 1 ) then

      private.Music_Stop();
      common.sound.Play( "assets/content/"..section_name.."/movie/aud_intro" );

    end;

    if ( intro_steps[ intro_step ] ) then

      intro_steps[ intro_step ].func(
        intro_step,
        intro_steps[ intro_step ].object,
        intro_steps[ intro_step ].params,
        function ()

          private.Intro_Play(
            section_name,
            ( intro_step + 1 ),
            intro_settings,
            func_end
          );

        end
      );

    else

      -- JK: Show carrier indicator
      v_carriers.OnIntroFinish();

      private.Music_Start();
      func_end();

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Intro_Animate ( intro_step, object_name, params, func_end )

  if not ( object_name ) then

    object_name = "anm_map_foreground_movie";

  end;

  --DbgTrace( "Intro_Animate: "..intro_step..", "..object_name );

  local animfunc_start = "intro_"..intro_step;
  local animfunc_end = nil;
  local animfunc_dummy = nil;

  if ( params ) then

    if ( params.animfunc_start ) then

      animfunc_start = params.animfunc_start;

    end;

    if ( params.animfunc_dummy ) then

      animfunc_dummy = params.animfunc_dummy;

    end;

    if ( params.animfunc_end ) then

      animfunc_end = params.animfunc_end;

    end;

  end;

  local func_end_anim = func_end;

  if ( animfunc_end ) then

    func_end_anim = function ()

      ui.anim.Play( object_name, animfunc_end );

      if not ( animfunc_dummy ) then

        func_end();

      end;

    end;

  end;

  if ( animfunc_dummy ) then

    ObjClone( object_name, "_dummy", 1, ObjGetRelations( object_name ).parent );

    ui.anim.Play(
      object_name.."_dummy",
      animfunc_dummy,
      function ()

        ObjDelete( object_name.."_dummy" );

        func_end();

      end
    );

  end;

  ObjSet( object_name, { active = true, visible = true } );

  ui.anim.Play( object_name, animfunc_start, func_end_anim );

end;
--------------------------------------------------------------------------------
function private.Intro_AnimateForeground (
    intro_step,
    object_name,
    params,
    func_next
  )

  ui.screen.Lock( true );

  ui.obj.Enable( "obj_map_foreground_intro", true );

  local func_OnAnimationEnd = function ()

    local func_anim_end = function ()

      ui.obj.Disable(
        "obj_map_foreground_intro", true
      );

      ui.obj.Attach( "obj_map_foreground", private.this.body );

      func_next();

    end;

    public.Focus(
      "obj_map_builds_puzzle_1",
      {
        time   = 1.0,
        scale  = 0.8
      }
    );

    ObjAnimate( "obj_map_foreground_intro", "alp", 0, 0, func_anim_end,
    {
      0.0, 0, 1.0,
      2.0, 0, 0.0
    } );

  end;

  local func_OnDummyEnd = function ()

    local indexes = {
      { 4, 1 },   { 3, 2 }, { 2, 3 }, { 8, 4 },
      { 9, 4.3 }, { 1, 2 }, { 5, 2 }, { 6, 2 }, { 7, 2 }
    };

    local func_anim_end = function ( index )

      private.Intro_AnimateBuilds( index );

    end;

    for i = 1, #indexes do

      ObjAnimate(
        "anm_map_builds_icon_build_"..indexes[ i ][ 1 ],
        "alp", 0, 0,
        function () func_anim_end( indexes[ i ][ 1 ] ) end,
        {
          0.0                    , 0, 1,
          0.0 + indexes[ i ][ 2 ], 0, 1,
          0.3 + indexes[ i ][ 2 ], 0, 1
        }
      );

    end;

    local offset =
    {
      default    = 160,
      [ "16:9" ] = 70
    };

    local offset_y = offset[ private.app_r ] or offset.default;

    public.Focus( "obj_map_point_stage_110",
      {
        time       = 9.0,
        scale      = 1.0,
        offset     = { x = 0, y = offset_y },
        func_after = func_OnAnimationEnd
      }
    );

  end;

  public.Focus( "obj_map_point_stage_25", { time = 18.0, scale = 0.8 } );

  ui.anim.Play(
    "anm_map_foreground_dragon",
    "action",
    function ()

      ui.anim.Play(
        "anm_map_point_enemy_reddragon_stage_110",
        "action",
        function ()

          ui.anim.Play(
            "anm_map_builds_icon_state_puzzle_1",
            "intro",
            nil,
            "state0"
          );

        end,
        "static"
      );

      local fight_data = v_fights.Fight_Data( "stage_110" );
      local enemy      = fight_data.enemy;
      local content    = fight_data.content;

      local sound = "assets/content/"..content.."/"..enemy.."/aud_"..enemy.."_tap";

      common.sound.Play( sound );

    end
  );

  ui.anim.Play( "anm_map_foreground_dragon_dummy", "dummy", func_OnDummyEnd );

end;
--------------------------------------------------------------------------------
function private.Intro_AnimateBuilds ( index )

  local anm_action = "anm_map_builds_icon_build_"..index.."_action";
  local anm_dummy  = "anm_map_builds_icon_build_"..index.."_dummy";

  local multiset = {};

  ui.obj.Create(
    anm_action,
    "anim",
    "obj_map_builds_build_"..index,
    {
      pos_x = 0,
      pos_y = 0,
      pos_z = 10,
      playing  = false,
      visible = false
    }
  );

  ui.obj.Create(
    anm_dummy,
    "anim",
    "obj_map_builds_build_"..index,
    {
      pos_x = 0,
      pos_y = 0,
      pos_z = 10,
      playing  = false,
      visible = false
    }
  );

  ui.obj.Set(
    anm_action,
    {
      res      = "assets/content/map/effects/anm_appearance_build",
      animfunc = "action",
      color_r  = 0.7,
      color_g  = 0.2,
      color_b  = 0.8,
      playing  = true,
      visible = true,
      endtrig  = function ()

        ui.obj.Delete( anm_action );

      end
    },
    multiset
  );

  ui.obj.Set(
    anm_dummy,
    {
      res      = "assets/content/map/effects/anm_appearance_build",
      animfunc = "dummy",
      playing  = true,
      visible = true,
      endtrig  = function ()

        ui.obj.Delete( anm_dummy );

        ui.obj.Set(
          "anm_map_builds_icon_build_"..index,
          {
            animfunc = "corrupted",
            playing  = true
          }
        );

      end
    },
    multiset
  );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.Intro_AnimateHero ( intro_step, object_name, params, func_next )

  public.Focus(
    "obj_map_portal_section_1_prev",
    {
      time = 1.0
    }
  );

  local multiset = {};

  ui.obj.Set(
    "anm_map_spawn_hero",
    {
      res     = "assets/content/map/effects/anm_appearance_hero",
      active  = true,
      visible = true
    }
  );

  ui.obj.Enable( "obj_map_spawn", true, multiset );

  ui.obj.Multiset( multiset );

  ui.anim.Play(
    "anm_map_spawn_hero",
    "action",
    function ()

      public.HideObjectPopup();

      func_next();

    end
  );

  ui.anim.Play(
    "anm_map_spawn_hero_dummy",
    "dummy",
    function ()

      private.Hero_ShowPopup( "str_int_popup_cloud_intro" );

    end
  );

  common.sound.Play( "intro_hero" );

end;
--------------------------------------------------------------------------------
function private.Intro_Scale ( intro_step, object_name, params, func_end )

  --DbgTrace( "Intro_Scale: "..intro_step..", "..object_name );

  local scale = params.scale or 1.0;
  local time  = params.time or 1.0;

  public.Focus(
    object_name,
    {
      time       = time,
      scale      = scale,
      func_after = func_end
    }
  );

end;
--------------------------------------------------------------------------------
function private.Intro_Skip ( section_name, intro_settings, func_OnSkipEnd )

  local skip_steps;

  if ( intro_settings.type == "default" ) then

    local target_type  = intro_settings.target_type;
    local target_stage = intro_settings.target_stage;
    local target_name  = intro_settings.target_name;
    local hero_stage   = intro_settings.hero_stage;

    local step_1_obj = "anm_map_spawn_hero";

    local step_2_obj = tern( target_type == "enemy" )(
      "anm_map_point_enemy_"..target_name.."_"..target_stage,
      "anm_map_characters_icon_"..target_name
    );

    local step_2_params = tern( target_type == "enemy" )(
      { animfunc_end = "static" },
      { animfunc_end = "static_"..target_name }
    );

    local step_3_obj = "anm_map_spawn_hero";

    skip_steps = {
      { func   = private.Intro_StopAnimate,
        object = step_1_obj,
        params = {}
      },
      { func   = private.Intro_StopAnimate,
        object = step_2_obj,
        params = step_2_params
      },
      { func   = private.Intro_StopAnimate,
        object = step_3_obj,
        params = {}
      }
    };

  elseif( intro_settings.type == "section_1" ) then

    skip_steps = {
      { func   = private.Intro_StopAnimate,
        object = "anm_map_spawn_hero",
        params = {}
      },
      { func   = private.Intro_StopForeground,
        object = false,
        params = {}
      },
      { func   = private.Intro_StopBuilds,
        object = false,
        params = {}
      },
      { func   = private.Intro_StopAnimate,
        object = "anm_map_builds_icon_state_puzzle_1",
        params = { animfunc_end = "state0" }
      }
    };

  end;

  for i = 1, #skip_steps do

    local skip_step = skip_steps[ i ];
    local func      = skip_step.func;
    local object    = skip_step.object;
    local animfunc  = skip_step.params.animfunc_end;

    func( object, animfunc );

  end;

  local stage_first = tonumber(
    string.match( v_sections.Section_FirstStage( section_name ), "%d+" ) );

  local stage_last = stage_first + 10;

  local multiset = {};

  for i = stage_first, stage_last do

    ui.obj.Set(
      "btn_map_point_stage_"..i,
      {
        alp     = 1,
        scale_x = 1
      },
      multiset
    );

  end;

  ui.obj.Disable( "obj_map_foreground_movie", true, multiset );
  ui.obj.Disable( "obj_map_spawn", true, multiset );

  ui.obj.Multiset( multiset );

  private.Stages_Show_Instant( stage_first, stage_last, "common" );

  public.Focus( "obj_map_point_stage_"..stage_first,
    {
      time       = 0.3,
      scale      = 1.0,
      func_after = function ()
        private.Stages_PointEnable( "stage_"..stage_first );
      end
    }
  );

  ui.obj.Attach( "obj_map_foreground_movie", "obj_map_foreground" );
  ui.obj.Attach( "obj_map_spawn", "hub_map_anims" );

  -- JK: Show carrier indicator
  v_carriers.OnIntroFinish();

  common.sound.Stop( "assets/content/"..section_name.."/movie/aud_intro" );
  private.Music_Start();

  ObjAnimate(
    private.this.marker, "alp", 0, 0,
    func_OnSkipEnd,
    {
      0.0, 0, 0,
      0.6, 0, 1
    }
  );

end;
--------------------------------------------------------------------------------
function private.Intro_StopAnimate ( object_name, animfunc_end )

  --DbgTrace( "Intro_StopAnimate: "..object_name );

  ui.anim.Stop( object_name );

  if ( animfunc_end ) then

    ui.anim.Play( object_name, animfunc_end );

  end;

end;
--------------------------------------------------------------------------------
function private.Intro_StopForeground ( anm_name, animfunc_end )

  local multiset = {};

  ui.anim.Stop( "anm_map_foreground_dragon", multiset );
  ui.obj.Disable( "obj_map_foreground_intro", true, multiset );

  ui.obj.Multiset( multiset );

  ui.obj.Attach( "obj_map_foreground", private.this.body );

  ui.anim.Stop( "anm_map_point_enemy_reddragon_stage_110" );
  ui.anim.Play( "anm_map_point_enemy_reddragon_stage_110", "static" );

end;
--------------------------------------------------------------------------------
function private.Intro_StopBuilds ( anm_name, animfunc_end )

  local multiset = {};

  for i = 1, 9 do

    local anm_build  = "anm_map_builds_icon_build_"..i;
    local anm_action = "anm_map_builds_icon_build_"..i.."_action";
    local anm_dummy  = "anm_map_builds_icon_build_"..i.."_dummy";

    ui.obj.Set(
      anm_build,
      {
        animfunc = "corrupted",
        playing  = true
      },
      multiset
    );

    ObjStopAnimate( anm_build, "alp" );

    ui.obj.Delete( anm_action );
    ui.obj.Delete( anm_dummy );

  end;

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Outro_Show ( func_OnEnd )

  local section_name   = v_sections.Section_NameCurrent();
  local section_config = v_sections.Section_GetConfig( section_name );
  local outro_settings = section_config.outro;
  local portal_obj     = "obj_map_portal_"..section_name.."_next";
  local portal_pos     = ui.obj.Position( portal_obj );

  ui.screen.Lock( true );
  ui.obj.Attach( "obj_map_foreground_movie", InterfaceName );

  local multiset = {};

  ui.obj.Enable( "obj_map_foreground_movie", true, multiset );

  ui.obj.Set(
    "anm_map_foreground_movie",
    {
      active  = true,
      visible = true,
      res     = "assets/content/"..section_name.."/movie/anm_outro",
      pos_x   = portal_pos[ 1 ],
      pos_y   = portal_pos[ 2 ]
    },
    multiset
  );

  ui.obj.Multiset( multiset );

  local function func_OnMovieEnd ()

    ui.screen.Lock( false );
    ui.obj.Disable( "obj_map_foreground_movie", true );
    ui.obj.Attach( "obj_map_foreground_movie", "obj_map_foreground" );

    if ( func_OnEnd ) then func_OnEnd(); end;

  end;

  ui.anim.Play(
    "anm_map_foreground_movie",
    "outro",
    func_OnMovieEnd
  );

  private.Music_Stop();
  common.sound.Play( "assets/content/"..section_name.."/aud_leave_section" );

  if ( outro_settings ) then

    local focus_target = "enemy";

    if ( outro_settings.target_stage ) then

      focus_target = tern( outro_settings.target_type )( outro_settings.target_type, focus_target );

      private.Outro_Scale(
        "obj_map_"..focus_target.."_"..outro_settings.target_stage,
        outro_settings.time,
        outro_settings.scale
      );

    end;

    local func_skip = function ( func_after )

      private.Outro_Skip(
        func_after,
        function ()

          if ( outro_settings.target_stage ) then

            private.Outro_Scale(
              "obj_map_"..focus_target.."_"..outro_settings.target_stage,
              0.1,
              outro_settings.scale
            );

          end;

        end
      );

    end;

    local func_after_skip = function () func_OnMovieEnd(); end;

    v_interface.Subtitles_Prepare();

    v_interface.Subtitles_Show(
      section_name,
      "outro",
      outro_settings.subs,
      func_skip,
      func_after_skip
    );

  end;

  v_interface.Hide();

end;
--------------------------------------------------------------------------------
function private.Outro_Scale ( object_name, time, scale )

  public.Focus(
    object_name,
    {
      time  = time or 1.0,
      scale = scale or 1.0
    }
  );

end;
--------------------------------------------------------------------------------
function private.Outro_Skip ( func_OnSkipEnd, func_Skip )

  local section_name = v_sections.Section_NameCurrent();

  common.sound.Stop( "leave_"..section_name );

  local multiset = {};

  ui.anim.Stop( "anm_map_foreground_movie", multiset );
  ui.obj.Disable( "obj_map_foreground_movie", true, multiset );

  ui.obj.Multiset( multiset );

  ui.obj.Attach( "obj_map_foreground_movie", "obj_map_foreground" );

  func_Skip();
  func_OnSkipEnd();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.ShowObjectPopup (
    mode,
    indicator_name,
    text,
    data,
    offset,
    functions,
    skip_focus
  )

  offset    = offset or { 0, 0 };
  functions = functions or {};

  local object_name = indicator_name:gsub( "_indicator", "" );
  local obj_pos     = private.Object_Position( object_name );

  if ( not skip_focus ) then

    local scale   = nil;
    local panel_h = 100;
    local popup_h = 240;
    local map_top = -( private.map_w + 2 * private.offset );

    local space = obj_pos[2] - ( map_top + ( panel_h + popup_h ) / private.map_scale_current );

    if ( space < popup_h + panel_h ) then

      scale = math.abs( ( popup_h / private.map_scale_current ) / space );

    end;

    functions.func_scale = function( end_func )

      public.Focus(
        object_name,
        {
          scale      = scale,
          func_after = end_func,
          offset = { x = 0, y = 100 }
        }
      );

    end;

  end;

  ui.popup.Cloud_Show(
    mode,
    indicator_name,
    text,
    data,
    offset,
    1.0,
    functions
  );

end;
--------------------------------------------------------------------------------
function  public.HideObjectPopup ()

  ui.popup.Cloud_Hide();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Sound_Play ( sound_id )

  local path = "assets/content/map/sounds/aud_"..sound_id;

   common.sound.Play( path );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Music_Start ()

  if ( not private.music ) then
    common.music.Play( "assets/content/audio/music/aud_map_mus",
                    nil, nil, true );
    private.music = true;
  end;

end;
--------------------------------------------------------------------------------
function private.Music_Stop ()

  if ( private.music ) then
    common.music.Stop( "assets/content/audio/music/aud_map_mus" );
    private.music = false;
  end;

end;
--------------------------------------------------------------------------------
-- function ******* DEV ******* () end
--------------------------------------------------------------------------------
function  public.Cheater_Update ()

  private.Update();

end;
--------------------------------------------------------------------------------
function private.DEV_GetPointPos ()

  ObjSet( "obj_map_point_DEV", { event_dragdrop = function ()

    local pos = private.Object_Position(
      "obj_map_point_DEV",
      private.this.background
    );

    DbgTrace( "pos_x = "..tostring( private.scale_modifier * pos[ 1 ] ) );
    DbgTrace( "pos_y = "..tostring( private.scale_modifier * pos[ 2 ] ) );
    DbgTrace( "" );

    ObjSet( "obj_map_point_DEV",
    {
      pos_x = private.scale_modifier * pos[ 1 ] + private.offset,
      pos_y = private.scale_modifier * pos[ 2 ] - private.offset
    } );

  end; } );

end;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- function *** NEW FOCUSING *** ()
--------------------------------------------------------------------------------
function public.Focus ( obj, params )

  -- common.DbgCallpoint( obj );

  params = params or {};

  -- NOTE: all parameters are optional:
  -- time       - focus time, default = ( distance / map_w )^( 1/3 )
  -- speed      - focus speed factor, default = 1
  -- instant    - instant focus flag, default = false
  -- scale      - new scale, default = current scale
  -- func_after - after focus callback
  -- offset     - target position offset, default = { x = 0, y = 0 }

  -- Warn about breaking solid focus
  if ( IsEditor() and private.focusing_solid ) then

    local msg = "Warning! Solid focusing at '%s' breaked by '%s'";
    common.DbgCallpoint( msg:format( private.focusing_solid, obj ) );

  end;

  local func_after = params.func_after;

  private.focusing_solid = tern( func_after )( obj, false );

  local function func_focus_end ()

    if ( v_interface.IsVisible() ) then
      v_interface.Input_On();
    end

    ui.input.On( "zmv_map" );

    if ( func_after ) then
      private.focusing_solid = false;
      func_after();
    end;

  end;

  -- Stop map animations
  ZoomViewStopVelocity( "zmv_map" );
  ObjStopAnimate( private.this.body, "pos_xy" );
  ObjStopAnimate( private.this.body, "scale_xy" );

  -- Define map state changes
  local scale_now = private.Focus_ScaleCurrent();
  local scale_new = params.scale or scale_now;

  scale_new = math.min( scale_new, private.map_scale_max );
  scale_new = math.max( scale_new, private.map_scale_min );

  local scale_factor = scale_now / scale_new;

  local pos_now  = private.Focus_PositionCurrent();
  local offset   = params.offset;
  local pos_new  = private.Focus_PositionTarget( obj, offset, scale_new );

  local distance = private.Focus_DistanceTo( pos_now, pos_new );
  local speed    = params.speed or 1;
  local time_opt = ( distance / private.map_w )^( 1/3 ) * scale_factor / speed;
  local time     = params.time or math.max( time_opt, 0.2 );

  local hold     = ( distance < 30 ) and not params.time;
  local instant  = params.instant or ( time < 1 / 25 );

  -- local tr = "\tp = %.2f,%.2f -> %.2f,%.2f; d = %.2f; "..
             -- "s = %.2f -> %.2f; t = %.2f; i = %s; h = %s";
  -- DbgTrace( tr:format(
    -- pos_now.x, pos_now.y, pos_new.x, pos_new.y, distance,
    -- scale_now, scale_new, time, instant, hold
  -- ) );

  if ( hold ) then

    pos_new = pos_now;
    instant = true;

  end;

  if ( instant ) then

    local map_params =
    {
      pos_x = pos_new.x,
      pos_y = pos_new.y
    };

    if ( scale_now ~= scale_new ) then

      map_params.scale_x = scale_new;
      map_params.scale_y = scale_new;

    end;

    ObjSet( private.this.body, map_params );
    func_focus_end();

  else

    v_interface.Input_Off();
    ui.input.Off( "zmv_map" );

    if ( scale_now ~= scale_new ) then

      ObjAnimate( private.this.body, "scale_xy", 0, 0, "",
      {
        0.00, 3, scale_now, scale_now,
        time, 3, scale_new, scale_new
      } );

    end;

    ObjAnimate( private.this.body, "pos_xy", 0, 0, func_focus_end,
    {
      0.00, 3, pos_now.x, pos_now.y,
      time, 3, pos_new.x, pos_new.y
    } );

  end;

end;
--------------------------------------------------------------------------------
function private.Focus_ScaleCurrent ()

  local params = ObjGet( private.this.body ) or {};
  return params.scale_x;

end;
--------------------------------------------------------------------------------
function private.Focus_PositionCurrent ()

  local params = ObjGet( private.this.body ) or {};
  return { x = params.pos_x or 0, y = params.pos_y or 0 };

end;
--------------------------------------------------------------------------------
function private.Focus_PositionTarget ( obj, offset, scale_new )

  -- Get raw position
  local pos_raw = ui.obj.Position( obj, private.this.body );
  local pos = { x = pos_raw[ 1 ] or 0, y = pos_raw[ 2 ] or 0 };

  -- Get zoomview params
  local params = ui.obj.Get( "zmv_map" );
  local scr_x, scr_y = params.screen_x, params.screen_y;
  local scr_w, scr_h = params.screen_w, params.screen_h;

  -- Shift by offset and move to screen center
  offset = offset or { x = 0, y = 0 };
  local center = { x = scr_w / 2 + scr_x, y = scr_h / 2 + scr_y };

  -- DbgTrace( ("Focus center: %s, %s"):format( center.x, center.y ) );

  pos.x = center.x + offset.x - pos.x * scale_new;
  pos.y = center.y + offset.y - pos.y * scale_new;

  -- Check bounds
  local map_s        = private.scale_modifier;
  local map_h, map_w = private.map_h, private.map_w;
  -- local app_h, app_w = private.app_h, private.app_w;

  local x_min, x_max = scr_w - map_s * map_w * scale_new, 0;
  local y_min, y_max = scr_h,  map_s * map_h * scale_new;

  pos.x = math.min( pos.x, x_max );
  pos.x = math.max( pos.x, x_min );

  pos.y = math.min( pos.y, y_max );
  pos.y = math.max( pos.y, y_min );

  return pos;

end;
--------------------------------------------------------------------------------
function private.Focus_DistanceTo ( a, b )

  local dx, dy = b.x - a.x, b.y - a.y;
  return math.sqrt( dx^2 + dy^2 );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Position_ScreenToMap ( scr_pos )

  local obj_pos = private.Focus_PositionCurrent();
  local obj_scl = private.Focus_ScaleCurrent();

  -- Translate coordinates to crazy zmv-coordinates, then back
  local zmv_scl = 2 - obj_scl;
  local map_pos =
  {
    x = -( obj_pos.x * zmv_scl - scr_pos.x / obj_scl ),
    y = -( obj_pos.y * zmv_scl - scr_pos.y / obj_scl )
  };

  return map_pos;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.ChangeList_Prepare (
    stage_name,
    is_first,
    reward_data,
    noreward,
    enable_next
  )

  private.changelist_active = true;

  if ( is_first ) then

    v_chests.TryUnlockByStage( stage_name );
    v_builds.Build_TryUnlockByStage( stage_name );
    v_spells.TryUnlockSpellByStage( stage_name );
    v_shop.Market_TryUnlock( stage_name );

  end;

  v_mines.Mine_TryUnlockByStage( stage_name, reward_data );

  local portal_unlocked = v_sections.Portal_TryUnlock( stage_name );

  if ( portal_unlocked ) then

    m_map.ChangeList_AddEvent( portal_unlocked );

  end;

  if ( #reward_data > 0 ) then

    --

  else

    reward_data = nil;

  end;

  m_map.ChangeList_AddEvent(
    {
      type        = "StageCompleted",
      name        = stage_name,
      is_first    = is_first,
      reward      = reward_data,
      noreward    = noreward,
      enable_next = enable_next
    }
  );

  local fight_data = v_fights.Fight_Data( stage_name );
  if ( fight_data.realtype == "misc" ) then

    m_fights.Enemy_SetAbsent( fight_data.enemy, false );

  end;

  if ( enable_next ) then

    local next_stage_name = v_stage.Stage_NameNext( stage_name );
    local next_info       = v_stage.Stage_StorageData( next_stage_name );

    if ( not next_info                                     ) or
       ( not next_info.completed and not next_info.enabled ) then

      if ( v_stage.Stage_TryEnable( next_stage_name ) ) then

        v_stage.Stateless_Check();

        m_map.ChangeList_AddEvent(
          {
            type = "StageEnabled",
            name = next_stage_name
          }
        );

        local next_fight_data = v_fights.Fight_Data( next_stage_name );
        if ( next_fight_data.realtype == "misc" ) then

          m_fights.Enemy_SetAbsent( next_fight_data.enemy, true );

        end;

      end;

    end;

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Update ( screen_switch )

  mediator.Notify( "View_Map_Update_Triggered" );

  local changes = m_map.ChangeList_GetData();
  local update  = {}; -- callback list

  --DbgTrace( "[ v_map ] Update: "..tostring( #changes ).." changes" );

  -- Notify about update started
  local begin_msg = "View_Map_Update_Started";
  local func_notify = private.Callback_NotifyForPending( begin_msg );
  table.insert( update, func_notify );

  private.Update_TryShowIntro( update );
  private.Update_TryShowUpdateReward( update );

  -- Internal update flags
  local enable_notifications = false;
  local unlock_portal = false;
  local playhaven_placement = false;
  local stage_completed = false;

  -- Begin changes
  local ch_begin_msg = "View_Map_Update_ChangesBegin";
  local func_ch_begin = private.Callback_NotifyForPending( ch_begin_msg );
  table.insert( update, func_ch_begin );

  for id, change in pairs( changes ) do

    local stage   = change.name;
    local rewards = change.reward;

    if ( change.type == "StageCompleted" ) then

      stage_completed     = true;
      playhaven_placement = true;

      -- Try set notifications flag
      local need_push = wiz.bfg.playhaven.NeedRegisterPushNotifications();
      if ( need_push.result and change.is_first ) then

        local num = tonumber( stage:match( "%d+" ) );
        if ( num >= 6 ) and ( math.fmod( num - 10, 4 ) == 0 ) then
          enable_notifications = true;
        end;

      end;

      -- Setup reactions
      local reactions =
      {
        stage_reward = {},
        stage_done   = {},
        quest_reward = {},
        quest_done   = {} -- DG: Used for start second character quest
      };

      reactions.stage_reward =
      {
        build     = v_builds.GetRewardReaction(
          rewards,
          private.Portal_Show
        ),
        character = v_characters.GetRewardReaction( stage, rewards ),
        mine      = v_mines.GetRewardReaction( stage, rewards )
      };

      if ( change.is_first ) then

        reactions.stage_done =
        {
          build     = v_builds.GetStageCompleteReaction( stage );
          character = v_characters.GetStageCompleteReaction( stage );
          mine      = v_mines.GetStageCompleteReaction( stage );
        };

        if ( reactions.stage_reward.build ) and
           ( reactions.stage_done.build   ) then

          local build_name = {
            reactions.stage_reward.build.name,
            reactions.stage_done.build.name
          };

          if ( build_name[ 1 ] == build_name[ 2 ] ) then

            reactions.stage_done.build = nil;

          end;

        end;

      end;

      if ( reactions.stage_reward.character ) then

        local reward = reactions.stage_reward.character.reward;
        local quest  = reactions.stage_reward.character.quest_done;

        reactions.quest_reward.build = v_builds.GetRewardReaction(
          reward,
          private.Portal_Show
        );

        reactions.quest_done.character =
          v_characters.GetQuestCompleteReaction( quest );

      end;

      --DbgTrace( reactions );

      -- Check for characters will update stages
      local hide_quest_item = false;
      for _, reaction in pairs( reactions ) do

        if ( reaction.character and reaction.character.update_stages ) then

          local list = reaction.character.update_stages;
          if ( common.IsEntityInList( list, stage ) ) then

            hide_quest_item = true;
            break;

          end;

        end;

      end;

      -- Process reactions
      private.Update_FocusAtCompletedPoint(
        update, stage, hide_quest_item
      );
      private.Update_TryPreparePortal( update, stage );
      private.Update_TryDisableBossFight( update, stage );
      private.Update_TryShowStageReward(
        update, stage, rewards, reactions.stage_reward
      );
      private.Update_TryCollectMinieventGift( update, stage );

      if ( reactions.stage_reward.build ) then

        local reaction = reactions.stage_reward.build;

        private.Update_Build_TryAnimate( update, reaction );
        private.Update_Build_TryShowPopup( update, reaction );
        private.Update_Build_TryShowStages( update, reaction );
        private.Update_Build_TryChangeSkin( update, reaction );

      end;

      if ( reactions.stage_reward.mine ) then

        local reaction = reactions.stage_reward.mine;

        private.Update_Mine_TryAnimate( update, reaction );
        private.Update_Mine_TryShowPopup( update, reaction );

      end;

      if ( reactions.stage_reward.character ) then

        local reaction    = reactions.stage_reward.character;
        local quest_react = reactions.quest_reward;

        private.Update_Character_TryShowPopup( update, reaction );
        private.Update_Character_TryAnimate( update, reaction );
        private.Update_Character_TryShowStory( update, reaction );
        private.Update_Character_TryShowReward( update, reaction, quest_react );
        private.Update_Character_TryChangeSkin( update, reaction, stage );

      end;

      if ( reactions.quest_reward.build ) then

        local reaction = reactions.quest_reward.build;

        private.Update_Build_TryAnimate( update, reaction );
        private.Update_Build_TryShowPopup( update, reaction );
        private.Update_Build_TryShowStages( update, reaction );
        private.Update_Build_TryChangeSkin( update, reaction );

      end;

      if ( reactions.quest_done.character ) then

        local reaction = reactions.quest_done.character;

        private.Update_Character_TryAnimate( update, reaction );
        private.Update_Character_TryUpdateStages( update, reaction );
        private.Update_Character_TryShowPopup( update, reaction );

      end;

      if ( reactions.stage_done.build ) then

        local reaction = reactions.stage_done.build;

        private.Update_Build_TryAnimate( update, reaction );
        private.Update_Build_TryShowStages( update, reaction );
        private.Update_Build_TryChangeSkin( update, reaction );

      end;

      if ( reactions.stage_done.mine ) then

        local reaction = reactions.stage_done.mine;

        private.Update_Mine_TryAnimate( update, reaction );
        private.Update_Mine_TryShowPopup( update, reaction );

      end;

      if ( reactions.stage_done.character ) then

        local reaction = reactions.stage_done.character;

        private.Update_Character_TryAnimate( update, reaction );
        private.Update_Character_TryUpdateStages( update, reaction );
        private.Update_Character_TryShowPopup( update, reaction );

      end;

      private.Update_TryShowResourcesReward( update, stage, rewards );

      if ( change.is_first ) then

        private.Update_TryDisableFight( update, stage );
        private.Update_TryAppearSpell(  update, stage );

        private.Update_TryPlaceChest( update, stage );

      end;

    elseif ( change.type == "StageEnabled" ) then

      -- JK: purge last_point_selected flag, we move to next point
      private.last_point_selected = nil;

      private.Update_FocusAtEnabledPoint( update, stage );
      private.Update_TryEnableFight( update, stage );
      private.Update_TryShowHint( update, stage );

    elseif ( change.type == "PortalUnlocked" ) then

      unlock_portal = true;

    end;

  end;

  local ch_done_msg = "View_Map_Update_ChangesDone";
  local func_ch_done = private.Callback_NotifyForPending( ch_done_msg );
  table.insert( update, func_ch_done );

  private.Update_TryEnableNotifications( update, enable_notifications );
  private.Update_TryShowDailyQuest( update );
  private.Update_TryShowDailyBonus( update );
  private.Update_TryShowContextMessage( update, unlock_portal );
  private.Update_TryShowOvertaken( update );

  private.Update_TryUnlockPortal( update, unlock_portal );
  
  private.Update_Finish( update, {
    screen_switch = screen_switch,
    unlock_portal = unlock_portal,
    stage_completed = stage_completed },
    playhaven_placement );

  private.Update_TryMinieventCrown( update );
    
  common.RunCallbackList( update );

  private.Market_Update(); -- FIXME

end;
--------------------------------------------------------------------------------
function private.Callback_NotifyForPending ( notification, params )

  return function ( func_after )

    mediator.Notify( notification, params );

    if ( func_after ) then

      local pending_token = public.IsPending();
      if  ( pending_token ) then

        local function func_after_pending ()

          if ( v_interface.window.IsAnyOnScreen() ) then
            v_interface.Hide();
            mediator.SubscribeOnce( "Interface_Window_NoWindows", func_after );
          else
            func_after();
          end;

        end;

        private.Pending_AddOnComplete( pending_token, func_after_pending );

      else

        func_after();

      end;

    end;

  end;

end;
--------------------------------------------------------------------------------
--[[
  Animation format
  {
    res = "resource", -- optional
    coords = {}, -- parameters for anim_action
    name = "parent_name",
    func_preset = function () end,
    func_dummy  = function () end,
    func_action = function ( func_after ) end
  }
]]
--------------------------------------------------------------------------------
function public.Callback_Animate ( anim )

  if not ( anim ) then return false; end;

  local function func_anim ( func_after )

    if ( anim.res and anim.res ~= "" ) then

      local anm_dummy  = string.format( "anm_%s_dummy", anim.name );
      local anm_action = string.format( "anm_%s_action", anim.name );
      local anm_coords =
      {
        pos_x = anim.coords.pos_x,
        pos_y = anim.coords.pos_y,
        pos_z = 10,
        scale_x = anim.coords.scale_x or 1,
        scale_y = anim.coords.scale_y or 1,
      };

      -- Set animations
      ObjCreate( anm_dummy,  "anim" );
      ObjCreate( anm_action, "anim" );
      ObjSet(    anm_action, anm_coords );
      ObjAttach( anm_dummy,  anim.name );
      ObjAttach( anm_action, anim.name );

      ObjSet( anm_dummy,  { res = anim.res } );
      ObjSet( anm_action, { res = anim.res } );

      -- Run animations
      ui.anim.Play( anm_dummy, "dummy", function ()

        anim.func_dummy();
        ObjDelete( anm_dummy );

      end );

      local animfunc = anim.animfunc or "action";
      ui.anim.Play( anm_action, animfunc, function ()

        anim.func_action( func_after );
        ObjDelete( anm_action );

      end );

    else

      if ( anim.func_dummy ) then

        anim.func_dummy();

      end;

      anim.func_action( func_after );

    end;

  end;

  return function ( func_after )

    ui.screen.Lock( true );

    local function func_end ()
      ui.screen.Lock( false );
      if ( func_after ) then func_after(); end;
    end;

    if ( anim.func_preset ) then

      anim.func_preset();

    end;

    if ( anim.skip_focus ) then

      func_anim( func_end );

    else

      public.Focus( anim.name,
        {
          func_after = function () func_anim( func_end ) end
        }
      );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Update_TryAddStep ( update, func_step, notification, params )

  if ( func_step ) then

    local func_notify =
      private.Callback_NotifyForPending( notification, params );

    -- DbgTrace( "Add to update: " .. notification, params );

    table.insert( update, func_step );
    table.insert( update, func_notify );

  end;

end;
--------------------------------------------------------------------------------
function private.Update_TryShowIntro ( update )

  local func_intro = tern( v_sections.Section_NeedIntro() )(
    private.Intro_Show,
    nil
  );

  local notification = "View_Map_Update_IntroDisplayed";
  private.Update_TryAddStep( update, func_intro, notification );

end;
--------------------------------------------------------------------------------
function private.Update_TryShowUpdateReward ( update )

  local func_reward = nil;

  local update_data = v_game.Update_GetData();
  if  ( update_data ) then

    func_reward = function ( func_after )

      if ( update_data.type == "replay" ) then

        v_game.dialog.Update_Show(
          update_data.type,
          update_data.message,
          nil,
          func_after
        );

      else

        local chest =
        {
          type  = "chest",
          name  = "goldchest",
          count = 1
        };

        local params =
        {
          category   = "reward",
          stage_name = "update",
          silent     = true
        };

        local function func_obtain ()
          v_items.Chest_Show( chest, params, update_data.reward, func_after );
        end;

        v_game.dialog.Update_Show(
          update_data.type,
          update_data.message,
          v_player.Skin_GetCurrent(),
          func_obtain
        );

      end;

    end;

  end;

  local notification = "View_Map_Update_UpdateRewardDisplayed";
  private.Update_TryAddStep( update, func_reward, notification, update_data );

end;
--------------------------------------------------------------------------------
function private.Update_FocusAtCompletedPoint (
    update,
    stage,
    hide_quest_reward
  )

  local function func_focus ( func_after )

    public.Focus( "obj_map_point_"..stage,
      {
        instant    = true,
        func_after = function ()

          private.Stages_PointUpdate(
            stage,
            { point = true, reward = true },
            hide_quest_reward,
            func_after
          );

        end
      }
    );

  end;

  local notification = "View_Map_Update_CompletedStageFocused";
  private.Update_TryAddStep( update, func_focus, notification, stage );

end;
--------------------------------------------------------------------------------
function private.Update_TryShowStageReward ( update, stage, rewards, reactions )

  local func_reward = nil;
  local has_drops = false;

  if ( rewards ) then

    for r, reward in pairs( rewards ) do
      if ( reward.type ~= "gold" and reward.type ~= "resource" ) then
        has_drops = true;
        break;
      end;
    end;
    
    local func_end = function ( func_after )
      if ( stage == "stage_6" ) then
        v_log.Playhaven_LogPlacement( "rate_prompt_postlvl6" );
      end;
      if ( func_after ) then func_after(); end;
    end;
    
    if ( v_stage.FastSaga_IsActive() ) then

      func_reward = function ( func_after )

        v_interface.panel.top.Show();
        public.Reward_Show( true, rewards, reactions, function() func_end( func_after ); end );

      end;

    else

      func_reward = function ( func_after )

        local function func_reward_accept ()

          if ( has_drops ) then

            int_dialog_map_reward.Hide(
              function ()
                public.Reward_Show( true, rewards, reactions, function() func_end( func_after ); end );
              end,
              function ()
                v_interface.window.Hide( "int_dialog_map_reward" );
              end );

          else

            int_dialog_map_reward.Hide(
              function ()
                public.Reward_Show( true, rewards ); -- Just for gold
              end,
              function ()
                v_interface.window.Hide( "int_dialog_map_reward", function() func_end( func_after ); end );
              end );

          end;

          mediator.Notify( "View_Map_Update_StageRewardCollected", stage );

        end; -- end func_reward_accept

        v_interface.panel.top.Show();

        v_interface.window.Show(
          "int_dialog_map_reward",
          function () int_dialog_map_reward.Show( rewards ); end,
          { map_active = true },
          { Accept = func_reward_accept }
        );

      end; -- end func_reward

    end;

  end;

  local notification = "View_Map_Update_StageRewardDisplayed";
  private.Update_TryAddStep( update, func_reward, notification, stage );

end;
--------------------------------------------------------------------------------
function private.Update_Build_TryShowPopup ( update, reaction )

  local notification = "View_Map_Update_Build_PopupDisplayed";
  local build = reaction.name;

  private.Update_TryAddStep( update, reaction.func_popup, notification, build );

end;
--------------------------------------------------------------------------------
function private.Update_Build_TryAnimate ( update, reaction )

  local func_anim    = reaction.func_anim;
  local notification = "View_Map_Update_Build_Animated";
  local build        = reaction.name;

  -- JK: used to cutoff context message window
  private.is_build_animated = true;
  
  private.Update_TryAddStep( update, func_anim, notification, build );

end;
--------------------------------------------------------------------------------
function private.Update_Build_TryChangeSkin ( update, reaction )

  local func_skin = nil;

  if ( reaction.skinpart ) then

    func_skin = function ( func_after )
      v_player.DialogSkin_Show( reaction.skinpart, func_after );
    end;

  end;

  local notification = "View_Map_Update_Build_SkinChanged";
  local params =
  {
    build = reaction.name,
    skinpart = reaction.skinpart
  }

  private.Update_TryAddStep( update, func_skin, notification, params );

end;
--------------------------------------------------------------------------------
function private.Update_Build_TryShowStages ( update, reaction )

  local func_stages = nil;

  if ( reaction.show_stages ) then

    local stage_min  = reaction.show_stages.min;
    local stage_max  = reaction.show_stages.max;
    local stage_type = reaction.show_stages.type;

    func_stages = function ( func_after )

      local func_move_marker = function ()

        if ( reaction.move_marker ) then
          private.last_point_selected = nil;
          private.Marker_MoveToPoint( private.Stages_GetName( stage_min,
                                                              stage_type ),
                                      false,
                                      false,
                                      func_after );
        elseif ( func_after ) then
          func_after();
        end;

      end;

      private.Stages_Show_Cascade(
        stage_min, stage_max, func_move_marker, stage_type
      );
    end;

  end;

  local notification = "View_Map_Update_Build_StagesAppeared";
  local build = reaction.name;

  private.Update_TryAddStep( update, func_stages, notification, build );

end;
--------------------------------------------------------------------------------
function private.Update_Mine_TryShowPopup ( update, reaction )

  local notification = "View_Map_Update_Mine_PopupDisplayed";
  local func_popup   = reaction.func_popup;
  local mine         = reaction.name;

  private.Update_TryAddStep( update, func_popup, notification, mine );

end;
--------------------------------------------------------------------------------
function private.Update_Mine_TryAnimate ( update, reaction )

  local func_anim    = reaction.func_anim;
  local notification = "View_Map_Update_Mine_Animated";
  local mine         = reaction.name;

  private.Update_TryAddStep( update, func_anim, notification, mine );

end;
--------------------------------------------------------------------------------
function private.Update_Character_TryShowPopup ( update, reaction )

  local notification = "View_Map_Update_Character_PopupDisplayed";
  local func_popup   = reaction.func_popup;
  local character    = reaction.name;

  private.Update_TryAddStep( update, func_popup, notification, character );

end;
--------------------------------------------------------------------------------
function private.Update_Character_TryAnimate ( update, reaction )

  local func_anim    = reaction.func_anim;
  local notification = "View_Map_Update_Character_Animated";
  local character    = reaction.name;

  private.Update_TryAddStep( update, func_anim, notification, character );

end;
--------------------------------------------------------------------------------
function private.Update_Character_TryShowReward ( update, reaction, reactions )

  local func_reward;

  if ( reaction.func_reward ) then

    func_reward = function ( func_after )
      reaction.func_reward( reactions, func_after );
    end;

  end;

  local quest = reaction.quest_done;

  local notification = "View_Map_Update_Quest_RewardDisplayed";
  private.Update_TryAddStep( update, func_reward, notification, quest );

end;
--------------------------------------------------------------------------------
function private.Update_Character_TryChangeSkin ( update, reaction, stage )

  local func_skin = nil;

  if ( reaction.skinpart ) then

    func_skin = function ( func_after )
      local func_end = function ()
        if ( stage == "stage_10" ) then
          v_log.Playhaven_LogPlacement( "rate_prompt_postlvl10" );
        end;
        if ( func_after ) then func_after() end;
      end 
      v_player.DialogSkin_Show( reaction.skinpart, func_end );
    end;

  end;

  local notification = "View_Map_Update_Character_SkinChanged";
  local params =
  {
    character = reaction.name,
    skinpart  = reaction.skinpart
  }

  private.Update_TryAddStep( update, func_skin, notification, params );

end;
--------------------------------------------------------------------------------
function private.Update_Character_TryShowStory ( update, reaction )

  local func_story;

  local notification = "View_Map_Update_Character_StoryDisplayed";
  local quest = reaction.quest_done;

  if ( quest ) then

    local section    = v_sections.Section_NameCurrent();
    local condition  = "CompleteQuest";
    
    if ( v_story.IsExistStory( section, condition, quest ) ) then
      func_story = function ( func_after )
        if ( common.sound.IsEnabled() ) then
          private.Music_Stop();
        end;
        v_story.ShowStory(  
        section, 
        condition, 
        quest, 
        "map", 
        function  ()
          if ( func_after ) then func_after(); end;
          if ( common.sound.IsEnabled() ) then
            private.Music_Start();
          end;
        end
        )
      end;
    end;

  end;

  private.Update_TryAddStep( update, func_story, notification, quest );

end;
--------------------------------------------------------------------------------
function private.Update_Character_TryUpdateStages ( update, reaction )

  local func_update = nil;

  if ( reaction.update_stages ) then

    func_update = function ( func_after )

      for _, stage in pairs( reaction.update_stages ) do
        private.Stages_PointUpdate( stage, { point = false, reward = true } );
      end;

      if ( func_after ) then func_after(); end;

    end;

  end;

  local notification = "View_Map_Update_Character_StagesUpdated";
  local character    = reaction.name;

  private.Update_TryAddStep( update, func_update, notification, character );

end;
--------------------------------------------------------------------------------
function private.Update_TryShowResourcesReward ( update, stage, rewards )

  local func_res = nil;

  if ( rewards ) then

    local resources = {};

    for r, reward in pairs( rewards ) do

      if ( reward.type == "resource" ) then
        table.insert( resources, reward );
      end;

    end;

    if ( #resources > 0 ) then

      if ( v_stage.FastSaga_IsActive() ) then

        func_res = function ( func_after )

          v_interface.panel.top.Show();

          for r = 1, #resources do
            local reward = resources[ r ];
            v_interface.Resource_Buffer_Out( reward.name );
          end;

          func_after();

        end;

      else

        func_res = function ( func_after )

          public.Focus( "obj_map_point_"..stage,
            {
              func_after = function ()

                v_interface.panel.top.Show();
                ui.input.Off( "zmv_map" );

                for r = 1, #resources do

                  local reward = resources[ r ];
                  private.Reward_Resource_EmitResources( stage, reward,
                    function ()

                      v_interface.Resource_Buffer_Out( reward.name );
                      if ( r == #resources ) then
                        ui.input.On( "zmv_map" );
                        func_after();
                      end;

                    end
                  );

                end;

              end
            }
          );

        end;

      end;

    end;

  end;

  local notification = "View_Map_Update_ResourcesRewardDisplayed";
  private.Update_TryAddStep( update, func_res, notification, stage );

end;
--------------------------------------------------------------------------------
function private.Update_TryDisableFight ( update, stage )

  local func_fight = v_fights.Callback_AnimateDisable( stage );

  local notification = "View_Map_Update_FightDisabled";
  private.Update_TryAddStep( update, func_fight, notification, stage );

end;
--------------------------------------------------------------------------------
function private.Update_TryCollectMinieventGift ( update, stage )

  if ( _G.v_minievent ) then

    local func_reward = function ( func_after )

      private.Minievent_GiftUpdate();
      local func_gift = v_minievent.Callback_GiftShowReward( stage );
      if ( func_gift ) then
        func_gift( func_after );
        return;
      end;

      func_after();

    end;

    local notification = "View_Map_Update_MinieventGiftCollected";
    private.Update_TryAddStep( update, func_reward, notification, stage );

  end;

end;
--------------------------------------------------------------------------------
function private.Update_TryDisableBossFight ( update, stage )

  local func_fight = v_fights.Callback_AnimateBossDisable( stage );

  local notification = "View_Map_Update_BossFightDisabled";
  private.Update_TryAddStep( update, func_fight, notification, stage );

end;
--------------------------------------------------------------------------------
function private.Update_TryAppearSpell ( update, stage )

  local func_spell = v_spells.Callback_AnimateAppear( stage );

  local notification = "View_Map_Update_SpellAppeared";
  private.Update_TryAddStep( update, func_spell, notification, stage );

end;
--------------------------------------------------------------------------------
function private.Update_FocusAtEnabledPoint ( update, stage )

  local function func_focus ( func_after )

    private.Marker_MoveToPoint( stage, false, true,
      function ()
        private.Stages_PointEnable( stage );
      end
    );

    public.Focus( "obj_map_point_"..stage,
      {
        time       = 0.6,
        func_after = func_after
      }
    );

  end;

  local notification = "View_Map_Update_EnabledStageFocused";
  private.Update_TryAddStep( update, func_focus, notification, stage );

end;
--------------------------------------------------------------------------------
function private.Update_TryEnableFight ( update, stage )

  local func_fight = v_fights.Callback_AnimateEnable( stage );

  local notification = "View_Map_Update_FightEnabled";
  private.Update_TryAddStep( update, func_fight, notification, stage );

end;
--------------------------------------------------------------------------------
function private.Update_TryPlaceChest ( update, stage )

  local func_chest = v_chests.Callback_AnimatePlace( stage, v_stage.FastSaga_IsActive() );
  local notification = "View_Map_Update_ChestPlaced";

  private.Update_TryAddStep( update, func_chest, notification, stage );

end;
--------------------------------------------------------------------------------
function private.Update_TryShowHint ( update, stage )

  local stage_info = v_stage.Stage_GetInfo( stage ) or {};
  local func_hint = nil;

  if ( stage_info.hint and  stage_info.hint == stage_info.state ) then

    func_hint = function ( func_after )

      local function func_popup ()
        local str = "str_int_popup_cloud_"..stage;
        private.Hero_ShowPopup( str, stage, { func_preclose = func_after } );
      end;

      public.Focus( "obj_map_point_"..stage, { func_after = func_popup } );

    end;

  end;

  local notification = "View_Map_Update_HintDisplayed";
  private.Update_TryAddStep( update, func_hint, notification, stage );

end;
--------------------------------------------------------------------------------
function private.Update_TryEnableNotifications ( update, show_dialog )

  local func_push = nil;

  if ( show_dialog ) then
    func_push = private.Dialog_Notifications_Show;
  end;

  local notification = "View_Map_Update_NotificationsDisplayed";
  private.Update_TryAddStep( update, func_push, notification );

end;
--------------------------------------------------------------------------------
function private.Update_TryShowDailyQuest ( update )

  local func_daily = nil;

  if ( _G.v_dailyquest ) then
    func_daily = v_dailyquest.OnMapUpdate();
  end;

  local notification = "View_Map_Update_DailyQuestDisplayed";
  private.Update_TryAddStep( update, func_daily, notification );

end;
--------------------------------------------------------------------------------
function private.Update_TryShowDailyBonus ( update )

  -- Uncomment for testing
  -- v_dailybonus.Refresh();

  local func_daily = nil;

  if ( v_dailybonus.IsReady() ) then

    func_daily = private.Dialog_DailyBonus_TryShow;

  end;

  local notification = "View_Map_Update_DailyBonusDisplayed";
  private.Update_TryAddStep( update, func_daily, notification );

end;
--------------------------------------------------------------------------------
function private.Update_TryShowContextMessage ( update, unlock_portal )

  --JK: don't show context window if build is animated in this update
  if ( private.is_build_animated ) then return; end;

  -- JK: don't show context window if portal is unlocked and outro is soon
  if ( unlock_portal ) then return; end;

  if ( _G.v_minievent and v_minievent.OnMapUpdateFinish() ) then
    return;
  end;
  
  private.context_message.unlocked =
    v_stage.Stage_GetInfo( v_game.Context_GetUnlockStage() ).completed;

  local func_context = nil;

  if ( parts.features.IsActive( "context_message" ) ) and
     ( private.context_message.unlocked             ) then

    local func_show = v_game.Context_GetShowFunction(
      private.context_message.triggered,
      private.context_message.stage_count
    );

    if ( func_show ) then
      func_context = function ( func_close )
        private.Context_Show( func_show, func_close );
      end;
    end;

  end;

  local notification = "View_Map_Update_ContextMessageDisplayed";
  private.Update_TryAddStep( update, func_context, notification );

end;
--------------------------------------------------------------------------------
function private.Update_TryShowOvertaken ( update )

  local func_splash = nil;
  local player = nil;

  local section_name = v_sections.Section_NameCurrent();

  if ( --[[true or]] v_social.Overtaken_IsNeedToShow( section_name ) ) then

    player = v_social.Overtaken_GetRandomPlayer( section_name );
    -- player = player or { name = "Ashot", stage = "stage_1" };

    func_splash = function ( func_after )

      local scr_pos =
      {
        x = ui.screen.Width()  * 0.50,
        y = ui.screen.Height() * 0.20
      };
      local map_pos = private.Position_ScreenToMap( scr_pos );
      local map_obj = private.this.body;

      int_splash_map_overtaken.Show( map_obj, map_pos, player, func_after );

    end;

  end;

  local notification = "View_Map_Update_OvertakenSplashDisplayed";
  private.Update_TryAddStep( update, func_splash, notification, player );

end;
--------------------------------------------------------------------------------
function private.Update_TryUnlockPortal ( update, unlock_portal )

  if ( unlock_portal ) then

    local section_name = v_sections.Section_NameCurrent();
    local notification = "View_Map_Update_PortalUnlocked";

    private.Update_TryAddStep(
      update,
      private.Portal_Unlock,
      notification,
      section_name
    );

  end;

end;
--------------------------------------------------------------------------------
function private.Update_TryPreparePortal ( update, stage_name )

  local section_name   = v_sections.Section_NameCurrent();
  local portal_prepare = v_sections.Section_StaticData( section_name ).portal_prepare;

  if ( portal_prepare                    ) and
     ( portal_prepare.type == "stage"    ) and
     ( portal_prepare.name == stage_name ) then

    local notification = "View_Map_Update_PortalPrepared";

    private.Update_TryAddStep(
      update,
      private.Portal_Prepare,
      notification,
      section_name
    );

  end;

end;
--------------------------------------------------------------------------------
function private.Update_TryMinieventCrown ( update )
  
  if ( _G.v_minievent ) then
    local notification = "View_Map_Update_MinieventCrown";
    
    local func_crown = function ( func_after )

      local func_update = v_minievent.OnMapUpdateFinish();
      if ( func_update ) then
        func_update( func_after );
        return;
      end;

      func_after();

    end;
    
    private.Update_TryAddStep( update, func_crown, notification );
  end;

end;
--------------------------------------------------------------------------------
function private.Update_Finish ( update, params, playhaven_placement )

  local unlock_portal   = params.unlock_portal;
  local stage_completed = params.stage_completed;

  local function func_finish ( func_after )

    local function func_end ()

      if ( playhaven_placement ) then
        v_log.Playhaven_LogPlacement( "map_level_success" );
      end;

      private.Section_TryShowDisabled();

      ui.screen.Lock( true );

      local func_after_interface = function ()

        v_player.Achievements_TryShowPanel();

        if ( func_after ) then func_after(); end;

      end;

      if ( v_stage.FastSaga_IsActive() ) then

        func_after_interface = function ()

          if ( func_after ) then func_after(); end;

          if ( stage_completed ) then

            local stage_name = private.marker_point_current;
            local stage_data = v_stage.Stage_GetInfo( stage_name );

            if ( stage_data and stage_data.state == 1 ) then
              private.Point_Tap( stage_name );
            end;

          end;

        end;

      end;

      if ( not v_interface.Show( nil, func_after_interface ) ) then

        ui.screen.Lock( false );

      end;

      --JK: clean update flags
      private.is_build_animated = nil;
    
    end;

    if ( not unlock_portal ) then

      public.FocusAtCurrentPoint( func_end );

    else

      func_end();

    end;

  end;

  local notification = "View_Map_Update_Finished";
  private.Update_TryAddStep( update, func_finish, notification );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Reward_ShowCollected ( emitter, count, name, func_after )

  local reward =
  {
    name  = name,
    count = count,
    type  = tern( name == "gold" )( "gold", "resource" );
  };

  local funcs_buffer_out =
  {
    gold     = function() v_interface.Gold_Buffer_Out(); end,
    resource = function() v_interface.Resource_Buffer_Out( name ); end
  };

  local function func_end ()

    funcs_buffer_out[ reward.type ]();
    if ( func_after ) then func_after(); end;

  end;

  private.Reward_Resource_EmitResources( emitter, reward, func_end );

end;
--------------------------------------------------------------------------------
function public.Reward_Show ( is_map_reward, rewards, reactions, func_after )

  -- Set drop sender
  local stage_sender = "spr_int_dialog_map_reward_item_icon";
  local quest_sender = "spr_int_dialog_map_quest_reward_item_icon";

  local sender = tern( is_map_reward )( stage_sender, quest_sender );

  -- Prepare for animation
  ui.screen.Lock( true );
  local function func_end ()

    ui.screen.Lock( false );
    if ( func_after ) then func_after(); end;

  end;

  -- Show gold reward
  for i = 1, #rewards do
    if ( rewards[ i ].type == "gold" ) then

      v_interface.Gold_Buffer_Out();
      break;

    end;
  end;

  local func = {
    quest = function ( reward_name )

      private.Reward_DropToMap( sender, reactions.character, func_end );

    end,
    worker = function ()

      private.Reward_DropToMap( sender, reactions.mine, func_end );

    end,
    fragment = function ()

      private.Reward_DropToMap( sender, reactions.build, func_end );

    end,
    skinpart = function ()

      private.Reward_DropToPanelBottom( sender, "player", func_end );

    end,
    stickerpack = function ()

      private.Reward_DropToPanelBottom( sender, "stickers", func_end );

    end,
    heart = function ()

      v_hearts.DialogMain_Show( func_end );

    end
  };

  -- Show first item reward
  for i = 1, #rewards do

    local reward_type = rewards[ i ].type;

    if ( func[ reward_type ] ) then

      func[ reward_type ]();

      return;

    end;

  end;

  func_end();

end;
--------------------------------------------------------------------------------
function private.Reward_DropToMap ( sender, reaction, func_after )

  local target    = reaction.object;
  local func_drop = reaction.func_drop;

  if ( reaction.call_drop_first ) then

    private.Reward_AnimateDrop( sender, target, func_drop, func_after );

  else

    private.Reward_AnimateDrop( sender, target, func_after, func_drop );

  end;

end;
--------------------------------------------------------------------------------
function private.Reward_DropToPanelBottom ( sender, button, func_after )

  local receiver = "btn_int_panel_interface_bottom_"..button;

  ui.screen.Lock( true );
  v_interface.Show();

  local function func_end ()

    ui.anim.Bounce( receiver, function ()

      ui.screen.Lock( false );
      if ( func_after ) then func_after(); end;

    end );

  end;

  private.Reward_AnimateDrop( sender, receiver, nil, func_end, true );

end;
--------------------------------------------------------------------------------
function private.Reward_AnimateDrop (
    sender,
    receiver,
    func_start,
    func_end,
    skip_focus
  )

  if ( not ui.obj.IsActive( sender ) ) then
    if ( func_end ) then func_end(); end;
    return;
  end;

  -- Setup sound
  local sound = "reward_drop_common";

  if     ( string.find( receiver, "puzzle" ) ) then

    sound = "reward_drop_puzzle";

  elseif ( string.find( receiver, "island" ) ) then

    sound = "reward_drop_island";

  end;

  private.Sound_Play( sound );

  -- Setup animation
  ObjClone( sender, "_drop", 0, "obj_map_drop" );

  local sender_pos = private.Object_Position( sender, InterfaceName );

  ObjSet( "anm_map_drop",
  {
    pos_x = sender_pos[ 1 ],
    pos_y = sender_pos[ 2 ]
  } );

  ObjAttach( "anm_map_drop", InterfaceName );

  local func_OnFocusEnd = function ()

    if ( func_start ) then func_start(); end;

    local receiver_pos;

    sender_pos   = private.Object_Position( "anm_map_drop", InterfaceName );
    receiver_pos = private.Object_Position( receiver,       InterfaceName );

    local func_OnAnimationEnd = function ()

      local func_OnAnimationEnd = function ()

        ObjAttach( "anm_map_drop", "hub_map_drop" );
        ObjDelete( sender.."_drop" );

        if ( func_end ) then func_end(); end;

      end;

      ObjAnimate( sender.."_drop", "alp", 0, 0, func_OnAnimationEnd,
      {
        0.0, 0, 1,
        0.3, 0, 0
      } );

      ObjAnimate( sender.."_drop", "scale_xy", 0, 0, "",
      {
        0.0, 0, 0.5, 0.5,
        0.3, 0, 0.0, 0.0
      } );

    end;

    local sender_props = ObjGet( sender.."_drop" );

    ObjAnimate( sender.."_drop", "scale_xy", 0, 0, "",
    {
      0.0, 0, sender_props.scale_x, sender_props.scale_y,
      0.5, 0, 0.5,                  0.5
    } );

    ObjAnimate( "anm_map_drop", "pos_xy", 0, 0, func_OnAnimationEnd,
    {
      0.0, 0, sender_pos[ 1 ],   sender_pos[ 2 ],
      0.5, 0, receiver_pos[ 1 ], receiver_pos[ 2 ]
    } );

  end;

  -- Play animation
  if not ( skip_focus ) then

    ui.anim.Play( "anm_map_drop", "item_apply" );
    public.Focus( receiver,
      {
        scale      = 1.0,
        func_after = func_OnFocusEnd
      }
    );

  else

    ui.anim.Play( "anm_map_drop", "item_apply", func_OnFocusEnd );

  end;

end;
--------------------------------------------------------------------------------
function private.Reward_Resource_EmitResources ( emitter, reward, func_end )

  local obj_fly   = "obj_map_flying_resources";
  local anm_fly   = "anm_map_flying_resources";
  local anm_res   = "anm_map_flying_resources_icon";
  local obj_anims = "obj_map_anims";
  local tmr_fly   = "tmr_map_stage_reward";

  -- default: point
  local emitter_obj = "obj_map_point_"..emitter;

  -- spell
  if ( string.find( emitter, "spell" ) ) then

    emitter_obj = "obj_map_spells_"..emitter;

  end;

  -- mine
  if ( string.find( emitter, "mine" ) ) then

    emitter_obj = "obj_map_mines_"..emitter;

    local indicator = "obj_map_mines_indicator_"..emitter;
    local ind_pos   = private.Object_Position( indicator, InterfaceName );

    mediator.Notify(
      "View_Map_MineGathered",
      {
        resource    = reward.name,
        delta       = reward.count,
        posx        = ind_pos[ 1 ],
        posy        = ind_pos[ 2 ],
        fontcolor_r = 1,
        fontcolor_g = 1,
        fontcolor_b = 1
      }
    );

  end;

  local emitter_pos  = private.Object_Position( emitter_obj );
  local count        = 0;
  local flight_times = {};
  local max_count    = math.min( reward.count, 15 );
  local tme          = { 0.45, 0.47, 0.51, 0.54, 0.58, 0.60, 0.64, 0.66 };

  local pattern = "_%s_%s_%s";

  for j = 1, max_count do

    count = count + 1;

    local suffix = pattern:format( emitter, reward.name, count );

    local obj_name = obj_fly..suffix;
    local anm_name = anm_fly..suffix;
    local res_name = anm_res..suffix;
    local tmr_name = tmr_fly..suffix;

    private.reward[ obj_name ] = true;

    local index = math.random( 1, 8 );
    local pos_x = math.random( - 150, 150 );
    local pos_y = math.random( - 80, 80 );

    ui.obj.Clone(
      "obj_map_flying_resources",
      "obj_map_anims",
      suffix,
      true
    );

    local multiset = {};

    ui.obj.Set(
      obj_name,
      {
        pos_x   = emitter_pos[ 1 ],
        pos_y   = emitter_pos[ 2 ],
        pos_z   = ( 10 + index ),
        scale_x = 0,
        scale_y = 0
      },
      multiset
    );

    ui.obj.Set(
      res_name,
      {
        animfunc = reward.name,
        playing  = true
      },
      multiset
    );

    local delay = math.random( 100, 500 ) / 1000;

    ui.obj.Create( tmr_name, "timer"  );
    ui.obj.Attach( tmr_name, obj_name );

    local res_type = tern( reward.name:find( "gold" ) )( "gold", "resource" );

    ui.obj.Set(
      tmr_name,
      {
        time    = delay,
        playing = true,
        endtrig = function() ui.anim.Play( anm_name, res_type..index ); end
      },
      multiset
    );

    ui.obj.Multiset( multiset );

    local fly_time = tme[ index ] + delay + 0.5;

    table.insert( flight_times, fly_time );

    ObjAnimate(
      obj_name, "pos_xy", 0, 0,
      function ( params )

        local index = ui.obj.GetNamePostfix( params.sender );

        private.Reward_Resource_ApplyToTopPanel( emitter, reward, index );

      end,
      {
        0.0,            1, emitter_pos[ 1 ],         emitter_pos[ 2 ],
        delay,          1, emitter_pos[ 1 ],         emitter_pos[ 2 ],
        fly_time - 0.5, 1, emitter_pos[ 1 ] + pos_x, emitter_pos[ 2 ] + pos_y,
        fly_time,       1, emitter_pos[ 1 ] + pos_x, emitter_pos[ 2 ] + pos_y
      }
    );

    ObjAnimate( obj_name, "scale_xy", 0, 0, "",
      {
        0.00, 0, 0.0, 0.0,
        0.10, 0, 1.2, 1.2,
        0.20, 0, 0.8, 0.8,
        0.25, 0, 1.1, 1.1,
        0.3,  0, 1.0, 1.0
      }
    );

  end;

  table.sort( flight_times );

  local max_flight_time = flight_times[ #flight_times ] + 0.5;
  local tmr_reward      = "tmr_map_reward_"..reward.name

  ui.obj.Attach( tmr_reward, obj_anims );

  local func_OnTimerEnd = function()

    ObjDetach( tmr_reward );
    if ( func_end ) then func_end(); end;

  end;

  ui.obj.Set(
    tmr_reward,
    {
      time    = max_flight_time,
      playing = true,
      endtrig = func_OnTimerEnd
    }
  );

  private.Sound_Play( "fireflies" );

end;
--------------------------------------------------------------------------------
function private.Reward_Resource_ApplyToTopPanel ( emitter, reward_data, index )

  local resource_name  = reward_data.name;
  private.reward_count = private.reward_count or {};

  private.reward_count[ resource_name ] =
    ( private.reward_count[ resource_name ] or 0 ) + 1;

  local postfix =
    "_map_flying_resources_"..emitter.."_"..resource_name.."_"..index;

  local obj_name = "obj"..postfix;
  local pfx_name = "pfx"..postfix;
  local anm_name = "anm"..postfix;

  if ( not private.reward[ obj_name ] ) then return; end;

  private.reward[ obj_name ] = nil;

  ObjStopAnimate( obj_name, "pos_x" );

  local start_pos = ObjGet( obj_name );

  if ( not v_interface.Resource_IsPanelActive( resource_name ) ) then

    if ( string.find( reward_data.name, "gold" ) ) then

      v_interface.Gold_ShowPanel();

    else

      v_interface.Resource_ShowPanel( resource_name );

    end;

  end;

  local stop_pos  = private.Object_Position(
    "obj_int_panel_interface_top_"..resource_name,
    "obj_map_anims"
  );

  if ( string.find( resource_name, "resource" ) ) then

    stop_pos[ 2 ] = stop_pos[ 2 ] + 140;

  end;

  local func_end = function ()

    ObjDelete( obj_name );

  end;

  ObjSet( pfx_name, { visible = true, active = true } );

  PartSysStart( pfx_name );

  ObjAnimate( obj_name, "pos_xy", 0, 0, func_end,
  {
    0.0, 0, start_pos.pos_x, start_pos.pos_y,
    0.5, 0, stop_pos[ 1 ], stop_pos[ 2 ]
  } );

  ObjAnimate( obj_name, "alp", 0, 0, "",
  {
    0.00, 1, 1,
    0.45, 1, 1,
    0.50, 1, 0,
  } );

end;
--------------------------------------------------------------------------------
function private.Reward_Collect ()

  if ( private.reward ) then

    for key, value in pairs( private.reward ) do

      ObjDelete( key );

      private.reward[ key ] = nil;

    end;

  end;

  if ( private.reward_count ) then

    private.reward_count = {};

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Minievent_OnGiftUpdated ()

  private.Minievent_GiftUpdate();

end;
--------------------------------------------------------------------------------
function private.Minievent_OnGiftFinished ()

  if ( private.me_gift.created ) then

    for gift_name, _ in pairs( private.me_gift.chests ) do

      local anm_gift = private.me_gift.anm_gift.."_"..gift_name;
      ui.obj.Delete( anm_gift );
      private.me_gift.chests[ gift_name ] = nil;

    end;

    private.me_gift.created = false;

  end;

end;
--------------------------------------------------------------------------------
function private.Minievent_GiftUpdate ()

  -- JK: gift minievent is not available if player has reached the furthest section
  local section_number = v_player.Progress_SectionCount();
  if ( v_sections.Section_IsLastAvailable( section_number ) ) then
    return;
  end;
  
  local gifts_data = v_minievent.GetCurrentCustomEvent( "gift" );
  
  if ( gifts_data ) then
    
    local is_available = v_minievent.Event_IsAvailableForSection(
      gifts_data.id, v_sections.Section_NumberCurrent()
    );
    
    if ( not is_available ) then return; end;

    local gifts_to_remove = common.table.DeepCopy( private.me_gift.chests );

    for i = 1, #gifts_data.reward.chests do

      local gift = gifts_data.reward.chests[ i ];
      local section, residue = math.modf( gift.stage / 110 );
      section = ( gift.stage > 110 and residue == 0 ) and section or section + 1;

      gifts_to_remove[ gift.name ] = nil;

      local anm_section  = private.me_gift.anm_section.."_"..section;
      local anm_gift     = private.me_gift.anm_gift.."_"..gift.name;
      local anm_stage    = private.me_gift.anm_stage.."_"..( gift.stage + 1 );

      if ( section == v_sections.GetCurrentNumber() ) then

        local position = ui.obj.Position( anm_stage, anm_section );

        local set_params = {
          pos_x = position[ 1 ],
          pos_y = position[ 2 ]
        };

        if ( not private.me_gift.stage_build[ gift.stage ] ) then
          local build = v_builds.Build_GetUnlocklistByMin( gift.stage );
          if ( build ) then
            private.me_gift.stage_build[ gift.stage ] = build;
            private.me_gift.build_gift[ build.name ] = gift.name;
            set_params.input = false; 
            if ( not private.me_gift.subscribed ) then
              mediator.Subscribe( "View_Builds_BuildConstructed", private.Minievent_GiftBuildConstructed );
              private.me_gift.subscribed = true;
            end;
          end;
        end;

        if ( not private.me_gift.chests[ gift.name ] ) then

          ui.obj.Clone( private.me_gift.anm_dummy, "obj_map_event_gift_"..section, "_minievent_"..gift.name, true );

          set_params.res      = "assets/content/items/anm_event_gift";
          set_params.animfunc = gift.name.."_gift";
          set_params.playing  = true;
          set_params.pos_z    = 10;
          set_params.inputrect_x = -60;
          set_params.inputrect_y = -70;
          set_params.inputrect_w = 120;
          set_params.inputrect_h = 120;

          private.me_gift.chests[ gift.name ] = true;
          private.me_gift.created = true;

        end;

        ui.obj.Set( anm_gift, set_params );

        ui.control.EventsSet( anm_gift, {
          func_on_click = function ()

            ui.input.Off( anm_gift );
            v_minievent.Reward_Preview( "chests", i, function ()
              ui.input.On( anm_gift );
            end, "gift" );

          end
        } );

        private.Minievent_GiftTimerStart( gift.name, gifts_data.id );

      else

        if ( private.me_gift.chests[ gift.name ] ) then

          ui.obj.Delete( anm_gift );
          private.me_gift.chests[ gift.name ] = nil;

        end;

      end;

    end;

    for gift_name, _ in pairs( gifts_to_remove ) do

      local anm_gift = private.me_gift.anm_gift.."_"..gift_name;

      ui.obj.Delete( anm_gift );
      private.me_gift.chests[ gift_name ] = nil;

    end;

    gifts_to_remove = nil;

  else

    private.Minievent_OnGiftFinished();

  end;

end;
--------------------------------------------------------------------------------
function private.Minievent_GiftBuildConstructed( build_name )
  if ( private.me_gift.build_gift[ build_name ] ) then
    local anm_gift = private.me_gift.anm_gift.."_"..private.me_gift.build_gift[ build_name ];
    ui.obj.Set( anm_gift, { input = true } );
    private.me_gift.build_gift[ build_name ] = nil;
  end;
  if ( not next( private.me_gift.build_gift ) and private.me_gift.subscribed ) then
    mediator.Unsubscribe( "View_Builds_BuildConstructed", private.Minievent_GiftBuildConstructed );
    private.me_gift.stage_build = {};
    private.me_gift.subscribed = false;
  end;
end;
--------------------------------------------------------------------------------
function private.Minievent_GiftTimerStart ( chest_name, id )

  local custom_obj = {
    anm_timer = "anm_map_chest_minievent_"..chest_name,
    text_tag  = "tt_counter"
  };

  if ( ui.obj.Get( custom_obj.anm_timer ) ) then

    ui.obj.SetTextTag(
      custom_obj.anm_timer,
      custom_obj.text_tag,
      ui.text.FormatTime( v_minievent.GetEventRemainingTime( id ) )
    );

    local function func_tick ()

      ui.timer.Show(
        "map_chest_minievent_"..chest_name,
        v_minievent.GetEventRemainingTime( id ),
        func_tick,
        custom_obj
      );

    end;

    func_tick();

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.AnimateZone (
    build_name,
    stages_list,
    stages_type,
    func_end
  )

  -- Animation callbacks
  local function get_func_appear_stages ( stage_from, stage_to )

    return function ( func_after )

      private.Stages_Show_Cascade( stage_from, stage_to, nil, stages_type );

      local tme = 0.3 * ( stage_to - stage_from );

      local stage = private.Stages_GetName( stage_to, stages_type );
      local point = "obj_map_point_"..stage;

      public.Focus( point, { time = tme, func_after = func_after } );

    end;

  end;

  -- Build animations list
  local anims = {};

  local stage_fight = v_fights.Fight_GetUnlockStage( stages_list, stages_type );
  local stage_char  = v_characters.GetStageUnlocker( build_name );

  local stage_min = stages_list[ 1 ];
  local stage_max = stages_list[ 2 ];
  local stage_cur = stage_min;

  for s = stage_min, stage_max do

    if ( s == stage_fight ) or ( s == stage_char ) then

      table.insert( anims, get_func_appear_stages( stage_cur, s ) );
      stage_cur = s + 1;

      if ( s == stage_fight ) then

        local stage = private.Stages_GetName( s, stages_type );
        local func_appear = v_fights.Callback_AnimateAppear( stage );
        table.insert( anims, func_appear );

      end;

      if ( s == stage_char ) then

        local func_appear = v_characters.Callback_AnimateAppear( build_name );
        table.insert( anims, func_appear );

      end;

    elseif ( s == stage_max ) then

      table.insert( anims, get_func_appear_stages( stage_cur, s ) );

    end;

  end;

  -- End animations func
  local function func_finish_anims ( func_after )

    local stage = private.Stages_GetName( stage_min, stages_type );
    local autoconstruct = v_builds.Build_IsAutoConstructed( build_name );

    if not ( autoconstruct ) then

      local tap, no_focus = false, true;
      private.Marker_MoveToPoint( stage, tap, no_focus );

      public.Focus( "obj_map_point_"..stage, { func_after = func_after } );

    elseif ( func_after ) then

      func_after();

    end;

  end;
  
  local function func_build_focus ( func_after )
    
    public.Focus( "anm_map_builds_gift_"..build_name, { func_after = func_after } );
    
  end;
  
  local func_gift_on_build = v_builds.Callback_AnimateGiftOnBuild( build_name );
    
  if ( func_gift_on_build ) then
    
    table.insert( anims, func_build_focus ); --AS add in queue focus on the build animation
    table.insert( anims, func_gift_on_build );
  end;
  
  table.insert( anims, func_finish_anims );

  if ( func_end ) then
    table.insert( anims, func_end );
  end;

  common.RunCallbackList( anims );

end;
--------------------------------------------------------------------------------
