-- name=v_match3
--------------------------------------------------------------------------------
function public.__InitModule ()

  common.mod.Load( "m_match3" );

  private.app_w = ui.screen.Width();
  private.app_h = ui.screen.Height();
  
  private.pos_x = private.app_w / 2;
  private.pos_y = private.app_h / 2;
  
  private.max_size = 500;

  private.active = false;

  --Log.i( "Initialized." );
  
  if ( IsEditor() ) then
    
    ne.msg.Subscribe( "Cheat_Stage_Reload", function ()
      
      local sections_enabled = v_sections.GetEnabledList();
      local prefix = "assets/content/levels/";
      local list = {};
      
      for i = 1, #sections_enabled do
        local section_index = sections_enabled[ i ]:match( "%d+" );
        table.insert( list, prefix.."stages_"..section_index );
      end;
      
      parts.db.ReplaceReadonlyCollection( "stages", list );
      ObjDelete( "match" );
      private.Match3_Create( private.editor.stage_data );
      
    end );
    
  end;
  
end;
--------------------------------------------------------------------------------
function public.__InitObjects ()

  --Log.i( "Objects initialized." );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.SetPos ( x, y )
  
  private.pos_x = x;
  private.pos_y = y;
  
end;
--------------------------------------------------------------------------------
function public.SetMaxSize ( size )
  
  private.max_size = size;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.SetStartData ( start_data )
  
  m_match3.StartData_Set( start_data );
  
end;
--------------------------------------------------------------------------------
function public.Start ()

  if ( private.active ) then

    public.Finish();

  end;

  private.active = true;

  wiz.ClearUsedResource();

  local data = m_match3.StartData_Get();
  
  private.field_width  = 70 * data.field[ 2 ];
  private.field_height = 70 * data.field[ 1 ];;
  
  private.scale = private.CalcFieldScale( private.field_width, private.field_height );

  ObjAttach( "obj_match3", v_screen.GetCurrentName() );

  private.Border_Create( );
  private.Match3_Create( data );

end;
--------------------------------------------------------------------------------
function public.Finish ()

  ObjDetach( "grd_match3_bg" );
  ObjDelete( "obj_match3_field_cells_border" );
  ObjDelete( "match" );

  ObjDetach( "obj_match3" );

  private.active = false;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.GetFieldCenter ()

  return private.pos_x, private.pos_y;

end;
--------------------------------------------------------------------------------
function public.GetFieldScale ()

  return private.scale;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Match3_Create( stage_data )

  if( IsEditor() ) then
    private.editor = { stage_data = stage_data };
  end;
  
  local field_width  = private.field_width;
  local field_height = private.field_height;
  
  local scale = private.scale;

  local skin_name       = "sirtuxedo";
  local no_limits_value = 10;

  if ( v_player.Skin_GetCurrent() == skin_name ) then

    local skin_data = v_player.Skin_Data( skin_name );

    if ( skin_data                ) and
       ( skin_data.bonus.nolimits ) then

      no_limits_value = no_limits_value + skin_data.bonus.nolimits;

    end;

  end;

  ObjCreate( "match", "matchbox" );
  ObjSet( "match", 
  { 
    pos_x           = common.math.Round( private.pos_x - ( 0.5 * field_width  * scale ) ),
    pos_y           = common.math.Round( private.pos_y - ( 0.5 * field_height * scale ) ),
    scale_x         = scale,
    scale_y         = scale,
    stage           = stage_data.stage, 
    state           = stage_data.state,
    preset          = stage_data.preset,
    boosters        = stage_data.boosters,
    section         = stage_data.section,
    attempts        = stage_data.attempts,
    max_attempts    = stage_data.max_attempts,
    no_limits_value = no_limits_value
  } );
  
  ObjAttach( "match", "obj_match3" );

end;
--------------------------------------------------------------------------------
function private.Border_Create ( )

  local field_width  = private.field_width;
  local field_height = private.field_height;
  
  local scale = private.scale;

  ObjCreate( "obj_match3_field_cells_border", "obj" );
  
  local properties = ObjGet( "obj_match3_field_cells_default" );
  properties.name = "obj_match3_field_cells_border";
  
  local multiset_table = {};
  
  table.insert( multiset_table, { "obj_match3_field_cells_border", properties } );
  table.insert( multiset_table, { "obj_match3_field_cells_border", 
  { 
    pos_x   = common.math.Round( private.pos_x - ( 0.5 * field_width  * scale ) ),
    pos_y   = common.math.Round( private.pos_y - ( 0.5 * field_height * scale ) ),
    scale_x = scale,
    scale_y = scale
  } } );
  table.insert( multiset_table, { "grd_match3_bg", { pos_x = - 5, pos_y = - 5 } } );
  
  ObjMultiSet( multiset_table );
  
  ObjAttach( "obj_match3_field_cells_border", "obj_match3" );
  ObjAttach( "grd_match3_bg", "obj_match3_field_cells_border" );
  
  local grid_params = 
  {
     0,  { pos_x = 0 },
     1,  { pos_x = 10 },
     2,  { pos_x = field_width },
     3,  { pos_x = field_width + 10 },

     4,  { pos_x = 0 },
     5,  { pos_x = 10 },
     6,  { pos_x = field_width },
     7,  { pos_x = field_width + 10 },

     8,  { pos_x = 0,                pos_y = field_height },
     9,  { pos_x = 10,               pos_y = field_height },
     10, { pos_x = field_width,      pos_y = field_height },
     11, { pos_x = field_width + 10, pos_y = field_height },

     12, { pos_x = 0,                pos_y = field_height + 10 },
     13, { pos_x = 10,               pos_y = field_height + 10 },
     14, { pos_x = field_width,      pos_y = field_height + 10 },
     15, { pos_x = field_width + 10, pos_y = field_height + 10 }
  };

  GridSet( "grd_match3_bg", 0, grid_params );  
end;
--------------------------------------------------------------------------------
function private.CalcFieldScale ( w, h )
  
  local max_dim = math.max( w, h );
  local scale = private.max_size / max_dim;
  
  scale = private.Round( scale, 10 );
  
  return math.min( scale, 1.0 );
  
end;
--------------------------------------------------------------------------------
function private.Round ( value, precision )

  precision = precision or 1;
  return math.floor( ( value * precision ) + 0.5 ) / precision;
  
end;
--------------------------------------------------------------------------------