-- name=v_minievent
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_minievent",
  "int_dialog_minievent",
  "int_dialog_minievent_start",
  "int_dialog_minievent_reward",
  "int_dialog_minievent_notification",
  "int_dialog_minievent_share",
  "int_dialog_minievent_crown_continue",
  "int_panel_minievent_worldmap",
  "int_panel_minievent_character"
};

local TYPE_GIFT     = "gift";
local TYPE_SECTION  = "section";
local TYPE_CHESTS   = "chests";
local TYPE_STORY    = "story";
local TYPE_TOURNEY  = "tourney";
local TYPE_TOWER    = "tower";
local TYPE_PROGRESS = "progress";
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.this = {
    dialog              = "int_dialog_minievent",
    dialog_start        = "int_dialog_minievent_start",
    dialog_reward       = "int_dialog_minievent_reward",
    dialog_notification = "int_dialog_minievent_notification",
    dialog_share        = "int_dialog_minievent_share",
    dialog_crown_continue = "int_dialog_minievent_crown_continue"
  };

  common.mod.Load( _modules );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Init ()

  -- DG: Init then Start
  -- Init  condition: content is ready
  -- Start condition: stage is unlocked

  local callbacks = {
    func_GetContainerData = function ( container_name )

      local  container_data = v_game.features.Container_GetData( container_name );
      return container_data;

    end,
    func_IsPackageReady = function ( package_name )

      local  is_ready, state = v_game.Content_IsPackageReady( package_name );
      return is_ready, state;

    end,
    func_GetCurrentStage = function ()

      local stage_name = v_player.Progress_GetLastStageName();

      if ( stage_name:find( "_bonus" ) ) then

        stage_name = v_stage.Stages_FindLastCompletedSagaStage( stage_name );

      end;

      local number = tonumber( stage_name:match( "%d+" ) );
      if ( not v_stage.Stage_IsCompleted( stage_name ) ) then
        return math.max( number - 1, 1 );
      end;
      return number;

    end,
    func_GetCurrentSection = function ()

      return v_sections.GetCurrentNumber();

    end,
    func_ChestGenerate = function ( ... )

      return v_items.Chest_Generate( ... );

    end
  };

  m_minievent.Setup( callbacks );

  --if ( v_game.SilentContent_IsReady() ) then

    private.Init();

  --else
  --
  --  mediator.SubscribeOnce( m_game.msg.SilentContentReady, private.Init );
  --
  --end;

end;
--------------------------------------------------------------------------------
function private.Init ()

  Log.i( "Initialized" );

  if ( public.IsUnlocked() ) then

    private.Start();

  else

    mediator.Subscribe( "View_Stage_StageWon", private.TryStart );

  end;

  v_interface.window.RegisterAsFullscreen( private.this.dialog );
  v_interface.window.RegisterAsModal( private.this.dialog_start, private.this.dialog );
  v_interface.window.RegisterAsModal( private.this.dialog_reward, private.this.dialog );
  v_interface.window.RegisterAsModal( private.this.dialog_notification );
  v_interface.window.RegisterAsModal( private.this.dialog_share, private.this.dialog );
  v_interface.window.RegisterAsModal( private.this.dialog_crown_continue, private.this.dialog );

  private.reward_obtained = {};
  private.notification_data = {
    event_exist = false,
    click = false
  };
  
  private.Crown_Init();
  
  common.event.Set(
    "int_panel_interface_bottom",
    "WorldsClickHandler",
    {
      name = "v_minievent",
      handler = function ()
        private.WorldsButtonClick();
      end
    }
  );

end;
--------------------------------------------------------------------------------
function private.TryStart ( stage_data )

  if ( stage_data.name == m_minievent.GetUnlockStage() ) then

    mediator.Unsubscribe( "Model_Stage_StageWon", private.TryStart );

    private.Crown_SetStartValue();
    
    private.Start();

  end;

end;
--------------------------------------------------------------------------------
function private.Start ()

  mediator.Subscribe(
    "Model_Minievent_EventUpdated",
    private.OnEventUpdated
  );

  ne.msg.Subscribe(
    "Event_Feature_Updated",
    private.OnFeatureUpdated
  );

  mediator.Subscribe(
    "Model_Minievent_ContentIsDownloading",
    private.OnEventContentIsDownloading
  );

  mediator.Subscribe(
    "Model_Minievent_EventStarted",
    private.OnEventStarted
  );

  mediator.Subscribe(
    "Model_Minievent_EventFinished",
    private.OnEventFinished
  );

  mediator.Subscribe(
    "Model_Minievent_CooldownFinished",
    private.OnEventCooldownFinished
  );

  mediator.Subscribe(
    "Model_Potion_RunAction",
    private.HandleActionSaleBoosters
  );

  mediator.Subscribe(
    "Model_Potion_CompleteAction",
    private.HandleActionSaleBoosters
  );

  mediator.Subscribe(
    "View_Map_MapOpened",
    private.OnMapOpened
  );

  m_minievent.InitSchedule();
  
end;
--------------------------------------------------------------------------------
function private.WorldsButtonClick ()

  local data = {
    target = "worlds",
    data = {
      type = "common",
      entity = "v_minievent",
      count = 0
    }
  };

  v_interface.panel.bottom.Notification( data );

  if ( private.notification_data.event_exist ) then
    private.notification_data.click = true;
  end;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Cheat_ResetEventProgress ()

  local event_id = m_minievent.GetCurrentEventId();

  if ( not event_id ) then return false; end;

  local event_type = m_minievent.Event_GetTypeById( event_id );

  local func_OnReload = function ()
    private.Cooldown_Stop( event_id );
    m_minievent.Event_Delete( event_id );
    m_minievent.ClearStageWon( event_id );
    m_minievent.Event_ResolveReward( event_id, 1 );
  end;

  if ( event_type == TYPE_SECTION ) then
    private.Section_Reload( event_id, func_OnReload );
  else
    private.Dialog_Reload( event_id, func_OnReload );
  end;

  return true;

end;
--------------------------------------------------------------------------------
function public.Cheat_ResetNotification ()

  local  is_removed = v_features.RemoveFeatureData( "context_message" );
  return is_removed;

end;
--------------------------------------------------------------------------------
function public.Cheat_CompleteLevel ()

  local event_id = m_minievent.GetCurrentEventId();

  if ( not event_id or not private.stage_selected ) then return false; end;
  
  local event_type = m_minievent.Event_GetTypeById( event_id );

  private.DialogStart_Hide( function ()

    if ( m_minievent.Event_IsSuspended() ) then
      return;
    end;

    private.Event_WaitForStageResult( event_type, event_id );
    
    local minievent_data = m_minievent.Event_GetById( event_id );

    if ( minievent_data.cooldown         ) and
       ( minievent_data.type ~= TYPE_CHESTS ) then

      m_minievent.Cooldown_Start( minievent_data );

    end;
    
    mediator.Notify( "View_Stage_StageWon", { 
      stage = private.stage_selected
    } );
    
    if ( v_minievent_section ) then
      v_minievent_section.Cheat_Update();
    end;
    
    private.stage_selected = nil;

  end, true );

  return true;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.IsUnlocked ()

  local stage_name  = m_minievent.GetUnlockStage();
  local is_unlocked = v_stage.Stage_IsCompleted( stage_name );

  return is_unlocked;

end;
--------------------------------------------------------------------------------
function public.IsAvailable ()

  if ( not public.IsUnlocked() ) then

    return false;

  end;

  local event_list = m_minievent.GetEventList();

  for _, event_id in pairs( event_list ) do

    local minievent_data = m_minievent.Event_GetById( event_id );

    if ( minievent_data                  ) and
       ( not minievent_data.is_completed ) then

      return true;

    end;

  end;

  return false;

end;
--------------------------------------------------------------------------------
function public.OnMapUpdate ()

  return private.OnMapUpdate;

end;
--------------------------------------------------------------------------------
function public.OnMapUpdateFinish ()

  return private.OnMapUpdateFinish;

end;
--------------------------------------------------------------------------------
function public.GetEventList ()

  local  event_list = m_minievent.GetEventList();
  return event_list;

end;
--------------------------------------------------------------------------------
function public.GetUpcomingEvent ()

  local  upcoming_event = m_minievent.Event_GetUpcoming();
  return upcoming_event;

end;
--------------------------------------------------------------------------------
function public.GetEventRemainingTime ( event_id )

  local  remaining_time = m_minievent.Event_GetRemainingTime( event_id );
  return remaining_time;

end;
--------------------------------------------------------------------------------
function public.GetCurrentCustomEventId ( event_type )

  local  event_id = m_minievent.GetCurrentCustomEventId( event_type );
  return event_id;

end;
--------------------------------------------------------------------------------
function public.GetCurrentCustomEvent ( event_type )

  local event_id = m_minievent.GetCurrentCustomEventId( event_type );

  if ( not event_id ) then

    return nil;

  end;

  local event_data = m_minievent.Event_GetById( event_id );

  if ( event_data                                                ) and
     ( ( event_data.is_finished ) or ( event_data.is_completed ) ) then

    return nil;

  end;

  return event_data;

end;
--------------------------------------------------------------------------------
function public.GetStageNumber ( stage_name )

  local event_id       = m_minievent.GetCurrentEventId();
  local minievent_data = m_minievent.Event_GetById( event_id );

  local is_tutorial, tutorial_data =
    m_minievent.Event_HasTutorial( event_id, stage_name, minievent_data );

  if ( is_tutorial ) then

    stage_name = tutorial_data.stage;

  end;

  for number, stage in pairs( minievent_data.stages ) do

    if ( stage_name == stage.name ) then

      return number;

    end;

  end;

  DbgTrace( "[ ERROR ] Number is not found for "..stage_name.."." );

  return 0;

end;
--------------------------------------------------------------------------------
function public.GetRelatedEventData ( event_type )

  local event_id = m_minievent.GetCurrentEventId( event_type );

  if ( not event_id ) then
    return false;
  end;

  local  event_data = m_minievent.Event_GetById( event_id );
  return event_data;

end;
--------------------------------------------------------------------------------
function public.Stage_IsRetryAllowed ()

  local event_id   = m_minievent.GetCurrentEventId();
  local event_type = m_minievent.Event_GetTypeById( event_id );

  return ( event_type == TYPE_STORY ) or ( event_type == TYPE_TOURNEY )
    or ( event_type == TYPE_CHESTS ) or ( event_type == TYPE_SECTION );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Event_IsFinished ( event_id )

  local is_active = m_minievent.Event_IsActive( event_id );
  return not is_active;

end;
--------------------------------------------------------------------------------
function public.Event_IsAvailableForSection ( event_id, section_number )

  local is_available = m_minievent.Event_IsAvailableForSection(
    event_id, section_number
  );
  
  return is_available;

end;
--------------------------------------------------------------------------------
function public.Event_GetData ( event_id )

  local  event_data = m_minievent.Event_GetById( event_id );
  return event_data;

end;
--------------------------------------------------------------------------------
function public.Event_GetStages ( event_id )

  local  stages = m_minievent.Event_GetStages( event_id );
  return stages;

end;
--------------------------------------------------------------------------------
function public.Event_GetProgress ( event_id )

  local  progress, stages = m_minievent.Event_GetProgress( event_id );
  return progress, stages;

end;
--------------------------------------------------------------------------------
function public.Event_GetActiveStageIndex ( event_id )

  local stage_list  = m_minievent.Event_GetStages( event_id );
  local stage_max   = #stage_list;
  local stage_index = m_minievent.Event_GetProgress( event_id ) + 1;

  if ( stage_index <= stage_max ) then
    return stage_index;
  end;

  return nil, stage_max;

end;
--------------------------------------------------------------------------------
function public.Event_GetRewards ( event_id )

  local  rewards = m_minievent.Event_GetRewards( event_id );
  return rewards;

end;
--------------------------------------------------------------------------------
function public.Event_GetRewardSettings ( event_name )

  local  reward_settings = m_minievent.Event_GetRewardSettings( event_name );
  return reward_settings;

end;
--------------------------------------------------------------------------------
function public.Event_GetSectionStageRewards ( event_id, stage_number )

  local reward_group;
  local holder_type;

  local rewards = m_minievent.Event_GetReward( event_id );

  if ( type( rewards ) ~= "table" ) then return; end;

  for index, reward in pairs( rewards ) do
    if ( stage_number == reward.stage_number ) then
      reward_group = index;
      holder_type  = reward.holder_type;
      break;
    end;
  end;

  if ( not reward_group ) then return; end;

  local reward_data = private.Reward_FilterGroup(
    "stage_"..stage_number,
    rewards[ reward_group ].rewards
  );

  return reward_data, reward_group, holder_type;

end;
--------------------------------------------------------------------------------
function public.Event_GetCharacters ( event_name )

  local  characters = m_minievent.Event_GetCharacters( event_name );
  return characters;

end;
--------------------------------------------------------------------------------
function public.Event_GetEnemies ( event_name )

  local  enemies = m_minievent.Event_GetEnemies( event_name );
  return enemies;

end;
--------------------------------------------------------------------------------
function public.Event_GetQuests ( event_name )

  local  quests = m_minievent.Event_GetQuests( event_name );
  return quests;

end;
--------------------------------------------------------------------------------
function public.Event_GetQuestByCharacter ( event_name, character_name )

  local quests = m_minievent.Event_GetQuests( event_name );

  for _, quest in pairs( quests ) do
    if ( quest.giver == character_name ) then
      return quest;
    end;
  end;

  return nil;

end;
--------------------------------------------------------------------------------
function public.Event_GetDialogData ( stage_number, when, event_id, event_scene )

  local event_id    = event_id    or m_minievent.GetCurrentEventId();
  local event_scene = event_scene or 1;

  local dialog_data = m_minievent.CharacterDialog_GetData(
    event_id,
    event_scene,
    "stage_"..stage_number,
    when
  );

  return dialog_data;

end;
--------------------------------------------------------------------------------
function public.Event_SetDialogComplete ( dialog_name, event_id )

  local event_id = event_id or m_minievent.GetCurrentEventId();

  m_minievent.CharacterDialog_Complete( event_id, dialog_name );

end;
--------------------------------------------------------------------------------
function public.Event_GetIntro ( event_name )

  local  intro_data = m_minievent.Event_GetIntro( event_name );
  return intro_data;

end;
--------------------------------------------------------------------------------
function public.Event_SetIntro ( event_name, value )

  m_minievent.Event_SetIntro( event_name, value );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.DialogStart_Show ( stage_index, func_before, force )

  local event_id = m_minievent.GetCurrentEventId();

  return private.DialogStart_Show( event_id, stage_index, func_before, force );

end;
--------------------------------------------------------------------------------
function public.DialogShare_Show ( func_after )

  private.DialogShare_Show( func_after );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Context_TryShowDialog ( func_on_end )

  local event_list = m_minievent.GetEventList();

  -- JK: no events, do nothing
  if ( #event_list < 1 ) then

    if ( func_on_end ) then func_on_end(); end;
    return;

  end;

  --JK: throw away gift event
  --for index, event in pairs( event_list ) do
  --  if ( event.type == TYPE_GIFT ) then
  --    table.remove( event_list, index );
  --  end;
  --end;

  -- JK: check story first
  local story_event_id = m_minievent.GetCurrentCustomEventId( TYPE_STORY );

  -- JK: story event id or first other event id
  local event_id = tern( story_event_id and story_event_id ~= event_list[ 1 ].id )(
                         event_list[ 1 ].id,
                         ( event_list[ 2 ] or event_list[ 1 ] ).id );

  local minievent_data = m_minievent.Event_GetById( event_id );
  local is_context     = private.Context_GetState( event_id );

  if ( story_event_id ) then

    local story_minievent_data = m_minievent.Event_GetById( story_event_id );
    local story_is_context     = private.Context_GetState( story_event_id );

    if ( story_minievent_data          ) and
       ( story_is_context              ) and
       ( not story_minievent_data.is_suspended ) then

      event_id = story_event_id;
      minievent_data = story_minievent_data;
      is_context     = story_is_context;

    end;

  end;

  if ( not minievent_data            ) or
     ( not is_context                ) or
     ( minievent_data.is_suspended   ) then

    if ( func_on_end ) then func_on_end(); end;
    return;

  end;

  if ( minievent_data.type == TYPE_GIFT ) then
    local stage_name    = v_sections.Stage_GetLastEnabled();
    minievent_data.zone = v_stage.Zone_Data( stage_name );
  end;

  if ( minievent_data.type == TYPE_SECTION ) then
    local stages   = public.Event_GetStages( minievent_data.id );
    local progress = math.max( public.Event_GetProgress( minievent_data.id ), 1 );
    minievent_data.zone = v_stage.Zone_Data( stages[ progress ].name );
  end;

  local function func_CloseDialog ( func_after )

    int_dialog_minievent_notification.Hide(
      function ()
        v_interface.window.Hide(
          private.this.dialog_notification,
          func_after
        );
      end
    );

  end;

  local delegates = {
    Close = function () func_CloseDialog( func_on_end ); end,
    Play = function ()

      if ( minievent_data.type == TYPE_GIFT ) then

        func_CloseDialog( func_on_end );

      elseif ( minievent_data.type == TYPE_SECTION ) then

        func_CloseDialog(
          function ()
            private.Section_Open( minievent_data.id, minievent_data.name );
          end
        );

      else

        v_interface.Hide();
        v_interface.Off();

        func_CloseDialog(
          function () private.Dialog_Show( event_id, nil, func_on_end ); end
        );

      end;

    end,
    Toggle = function ()
      local  context_state = private.Context_ToggleState( event_id );
      return context_state;
    end
  };

  v_interface.window.Show(
    private.this.dialog_notification,
    function () int_dialog_minievent_notification.Show( minievent_data ); end,
    nil,
    delegates
  );

end;
--------------------------------------------------------------------------------
function public.Context_IsAvailable ()

  local is_available = false;

  if ( not m_minievent.IsFeatureActive() ) then return is_available; end;
  if ( m_minievent.Event_IsSuspended() )   then return is_available; end;
  if ( not public.IsUnlocked() )           then return is_available; end;

  -- JK: check story first
  local story_event_id = m_minievent.GetCurrentCustomEventId( TYPE_STORY );

  if ( story_event_id ) then
    is_available = private.Context_GetState( story_event_id );
    if ( is_available ) then return is_available; end;
  end;

  -- JK: check other events
  local event_list = m_minievent.GetEventList();

  for _, event in pairs( event_list ) do
    is_available = private.Context_GetState( event.id );
    
    --JK: check if gift is on cooldown
    if ( event.type == TYPE_GIFT ) then
      local is_cooldown         = m_minievent.Cooldown_IsActive( event.id );
      local is_furthest_reached = v_sections.Section_IsLastAvailable(
        v_player.Progress_SectionCount()
      );
      local is_enlisted = m_minievent.Event_IsAvailableForSection(
        event.id, v_sections.Section_NumberCurrent()
      );
      is_available = ( not is_cooldown ) and
                     ( not is_furthest_reached ) and
                     ( is_enlisted );
    end;
    
    if ( is_available ) then return is_available; end;
  end;

  return is_available;

end;
--------------------------------------------------------------------------------
function private.Context_ToggleState ( event_id )

  local current_state = private.Context_GetState( event_id );
  local new_state     = not current_state;

  private.Context_SetState( event_id, new_state );

  return new_state;

end;
--------------------------------------------------------------------------------
function private.Context_GetState ( event_id )

  local data, result = v_features.GetFeatureData( "context_message" );

  if ( data           ) and
     ( data.minievent ) then

    return ( data.minievent[ "id_"..event_id ] ~= false );

  end;

  return true;

end;
--------------------------------------------------------------------------------
function private.Context_SetState ( event_id, state )

  local data, result = v_features.GetFeatureData( "context_message" );

  local cleanup = true;

  if ( not data ) then data = {}; cleanup = false; end;
  if ( not data.minievent ) then data.minievent = {}; cleanup = false; end;

  --JK: cleanup obsolete flags
  if ( cleanup ) then
    for id, flag in pairs( data.minievent ) do
      if ( not m_minievent.Event_IsActive( id:gsub( "id_", "" ) ) ) then
        data.minievent[ id ] = nil;
      end;
    end;
  end;

  data.minievent[ "id_"..event_id ] = state;

  v_features.SaveFeatureData( "context_message", data );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.OnFeatureUpdated ()

  private.HandleActionSaleBoosters();

end;
--------------------------------------------------------------------------------
function private.OnMapOpened ()
  
  private.Crown_SetStartValue();
  
  if ( not m_minievent.Event_IsSuspended() ) then

    local event_list  = m_minievent.GetEventList();
    local event_count = #event_list;

    for _, event in pairs( event_list ) do
      if ( event.type == TYPE_STORY   ) or
         ( event.type == TYPE_SECTION ) then
        private.SetWorldsOverlayIcon( event.name, event.type, true, event_count );
      end;
    end;

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.FeaturedPanel_Enable ( event_data, delegates )

  common.delegate.Set( "int_panel_minievent_worldmap", delegates );

  local placeholder = v_map.Dialog_Worldmap_GetFeaturedPlaceholderName();

  int_panel_minievent_worldmap.Setup( "started", event_data );
  int_panel_minievent_worldmap.Attach( placeholder );

  if ( v_map.Dialog_Worldmap_IsActive() ) then
    int_panel_minievent_worldmap.Timer_Start();
  end;

  mediator.Subscribe(
    "View_Map_DialogWorldmap_OpenedWithFeatured",
    int_panel_minievent_worldmap.Timer_Start
  );

  mediator.Subscribe(
    "View_Map_DialogWorldmap_ClosedWithFeatured",
    int_panel_minievent_worldmap.Timer_Stop
  );

  mediator.Notify( "View_Minievent_FeaturedPanelEnabled" );

end;
--------------------------------------------------------------------------------
function private.FeaturedPanel_Disable ()

  mediator.Unsubscribe(
    "View_Map_DialogWorldmap_OpenedWithFeatured",
    int_panel_minievent_worldmap.Timer_Start
  );

  mediator.Unsubscribe(
    "View_Map_DialogWorldmap_ClosedWithFeatured",
    int_panel_minievent_worldmap.Timer_Stop
  );

  if ( v_map.Dialog_Worldmap_IsActive() ) then
    mediator.SubscribeOnce(
      "View_Map_DialogWorldmap_ClosedWithFeatured",
      function ()
        int_panel_minievent_worldmap.Detach();
      end
    );
  else
    int_panel_minievent_worldmap.Detach();
  end;

  mediator.Notify( "View_Minievent_FeaturedPanelDisabled" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.GetDataBooster ( booster_name )

  return v_boosters.Boosters_GetInfo( booster_name );

end;
--------------------------------------------------------------------------------
function private.HandleActionSaleBoosters ()

  if ( int_dialog_minievent_start.IsActive() ) then
    int_dialog_minievent_start.UpdateSaleBoosters();
  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.ConnectionLost ()

  m_minievent.Event_Suspend();

  mediator.Notify( "View_Minievent_Suspended" );

  if ( private.DialogStart_IsActive()           ) and
     ( v_screen.GetCurrentName() ~= "scr_stage" ) then

    private.DialogStart_Hide();

  end;

  private.Dialog_Update();
  private.Launcher_Disable();
  private.TrySuspendStory();
  private.Section_OnInactive( nil, "suspended" );

end;
--------------------------------------------------------------------------------
function public.ConnectionRestored ()

  m_minievent.Event_Resume();

  mediator.Notify( "View_Minievent_Resumed" );

  private.Dialog_Update();
  private.Launcher_Update();
  private.TryResumeStory();
  private.Section_OnActive();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.OnEventContentIsDownloading ( params )

  local event_id   = params.id;
  local event_type = params.type;

  if ( event_type == TYPE_STORY ) then

    private.OnStoryContentIsDownloading( params );

  end;

end;
--------------------------------------------------------------------------------
function private.OnEventUpdated ()

  private.Launcher_Update();
  private.Decoration_TryShow();

end;
--------------------------------------------------------------------------------
function private.OnEventStarted ( params )

  local event_id   = params.event_id;
  local event_type = params.event_type;

  if ( event_type == TYPE_STORY ) then
    private.OnStoryStarted( event_id );
  end;

  if ( event_type == TYPE_GIFT ) then
    private.OnGiftStarted( event_id );
  end;

  if ( event_type == TYPE_SECTION ) then
    private.Section_OnActive( event_id );
  end;

end;
--------------------------------------------------------------------------------
function private.OnEventFinished ( params )

  local event_id   = params.event_id;
  local event_type = params.event_type;

  -- JK: that's story finished
  if ( event_type == TYPE_STORY ) then
    private.OnStoryFinished( event_id );
  end;

  if ( event_type == TYPE_GIFT ) then
    private.OnGiftFinished( event_id );
  end;

  if ( event_type == TYPE_SECTION ) then
    private.Section_OnInactive( event_id, "finished" );
    --v_minievent_section = nil;
  end;

end;
--------------------------------------------------------------------------------
function private.OnEventCooldownFinished ( params )

  local event_type = params.event_type;

  local handler = {
    gift = private.OnGiftCooldownFinished
  };

  if ( handler[ event_type ] ) then
    handler[ event_type ]( params.event_id );
  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.OnStoryContentIsDownloading ( params )

  local placeholder = v_map.Dialog_Worldmap_GetFeaturedPlaceholderName();

  int_panel_minievent_worldmap.Setup( "downloading", params );
  int_panel_minievent_worldmap.Attach( placeholder );

end;
--------------------------------------------------------------------------------
function private.OnStoryStarted ( event_id, skip_notify )

  local event_data = m_minievent.Event_GetById( event_id );

  if ( event_data.is_finished  ) or
     ( event_data.is_completed ) then

    return;

  end;

  private.SetWorldsOverlayIcon( event_data.name, event_data.type, true );

  local delegates = {
    Play = function ()

      v_map.Dialog_Worldmap_Hide( function ()
        private.Dialog_Show(
          event_id,
          nil,
          nil,
          function () v_interface.Loading_Hide(); end
        );
      end );

    end,
    GetRemainingTime = function ()

      local  seconds = m_minievent.Event_GetRemainingTime( event_id );
      return seconds;

    end
  };

  private.FeaturedPanel_Enable( event_data, delegates );

  if ( not skip_notify ) then
    mediator.Notify( "View_Minievent_StoryStarted" );
  end;

end;
--------------------------------------------------------------------------------
function private.OnStoryFinished ( event_id, skip_notify )

  private.FeaturedPanel_Disable();
  private.Decoration_Hide();

  if ( not skip_notify ) then
    mediator.Notify( "View_Minievent_StoryFinished" );
  end;

end;
--------------------------------------------------------------------------------
function private.TryResumeStory ()

  local story_event_id = m_minievent.GetCurrentCustomEventId( TYPE_STORY );

  if ( story_event_id ) then
    private.OnStoryStarted( story_event_id, true );
  end;

end;
--------------------------------------------------------------------------------
function private.TrySuspendStory ()

  local story_event_id = m_minievent.GetCurrentCustomEventId( TYPE_STORY );

  if ( story_event_id ) then
    private.OnStoryFinished( story_event_id, true );
  end;

end;
--------------------------------------------------------------------------------
function private.SetWorldsOverlayIcon ( event_name, event_type, state, count )

  count = count or #m_minievent.GetEventList();

  if ( m_minievent.GetCurrentCustomEventId( TYPE_GIFT ) ) then
    count = count - 1;
  else
    private.notification_data.event_exist = true;
  end;

  --DbgTrace(
  --  "SetStoryNotification: "..tostring( state )..", count: "..tostring( count )
  --);

  local icon_path, icon_anm;

  if ( state and event_name ) then
    local dir = "assets/content/minievent_"..event_name.."/";
    icon_path = dir.."overlay_icon";
    icon_anm  = dir.."anm_icon_overlay";
  end;
  
  local data_custom = {
    target = "worlds",
    data = {
      type      = "custom",
      entity    = "v_minievent",
      count     = count,
      style     = "minievent",
      icon_path = icon_path
    }
  };

  local right_panel_count = tern( state )( 1, 0 );
  
  --if ( event_type ~= TYPE_SECTION ) then
    v_interface.panel.bottom.Notification( data_custom );
  --end;

  --v_interface.panel.bottom.WorldsWarn( right_panel_count, icon_anm );
  
  if ( private.notification_data.event_exist and private.notification_data.click ) then
    private.WorldsButtonClick();
  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Section_OnActive ( event_id )

  local event_id = event_id or m_minievent.GetCurrentCustomEventId( TYPE_SECTION );  
  
  if ( event_id ) then

    if ( m_minievent.Event_IsFinishedOrCompleted( event_id ) ) then
      return;
    end;

    local event_name = m_minievent.Event_GetNameById( event_id );
    local event_data = m_minievent.Event_GetById( event_id );

    local func_on_click = function ()
      private.Section_Open( event_id, event_name );
    end;

    local delegates = {
      Play = function ()
        private.Section_Open(
          event_id,
          event_name,
          function ( func_after )
            v_map.Dialog_Worldmap_Hide( func_after );
          end
        );
      end,
      GetRemainingTime = function ()
        local  seconds = m_minievent.Event_GetRemainingTime( event_id );
        return seconds;
      end
    };

    private.FeaturedPanel_Enable( event_data, delegates );
    private.SetWorldsOverlayIcon( event_data.name, event_data.type, true );

  end;

end;
--------------------------------------------------------------------------------
function private.Section_OnInactive ( event_id, reason )

  local event_id = event_id or m_minievent.GetCurrentCustomEventId( TYPE_SECTION );

  if ( event_id ) then

    local event_name = m_minievent.Event_GetNameById( event_id );
    local item_name  = "minievent_"..event_name;

    private.FeaturedPanel_Disable();
    private.Decoration_Hide();
    
    if ( reason == "finished" ) then
      mediator.Notify(
        "View_Minievent_SectionFinished",
        {
          event_id     = event_id,
          is_completed = m_minievent.Event_IsCompleted( event_id )
        }
      );
    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Section_Open ( event_id, event_name, func_on_loading )

  local function func_OnShow ()
    
    if ( not v_minievent_section ) then
      common.mod.Load( "v_minievent_section" );
      v_minievent_section.Init( event_id, event_name );
    end;
    m_minievent.SetCurrentEventId( event_id );


    if ( func_on_loading ) then
      func_on_loading(
        function ()
          v_screen.Switch( "scr_minievent_section" );
        end
      );
    else
      v_screen.Switch( "scr_minievent_section" );
    end;

  end;

  v_interface.Loading_Show( true, func_OnShow );

end;
--------------------------------------------------------------------------------
function private.Section_Reload ( event_id, func_OnReload )

  local function func_OnShow ()

    v_minievent_section.Section_Destroy(
      function ()

        if ( func_OnReload ) then func_OnReload(); end;

        local event_name = m_minievent.Event_GetNameById( event_id );

        v_minievent_section.Init( event_id, event_name );
        v_minievent_section.Section_Create(
          function ()
            v_interface.Loading_Hide();
          end
        );

      end
    );

  end;

  v_interface.Loading_Show( true, func_OnShow );

end;
--------------------------------------------------------------------------------
function public.Section_Unload ()

  local function func_OnShow ()
    v_minievent_section.Section_Destroy(
      function ()
        v_interface.Off();
        v_screen.Switch( "scr_map" );
        v_minievent_section = nil;
      end
    );
  end;

  v_interface.Loading_Show( true, func_OnShow );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Callback_GiftShowReward ( stage_name )

  local event_type = TYPE_GIFT;
  local event_data = public.GetCurrentCustomEvent( event_type );

  if ( event_data ) then

    if ( private.ready_gift ) then

      return function ( func_after )

        private.Reward_ShowContainer (
          event_data.name,
          private.ready_gift,
          nil,
          func_after,
          event_type
        );

        private.ready_gift = nil;

      end;

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.OnGiftStarted ( event_id, skip_notify )

  local event_data = m_minievent.Event_GetById( event_id );

  if ( event_data.is_finished  ) or
     ( event_data.is_completed ) then

    return;

  end;

  mediator.Subscribe( "Model_Stage_StageWon", private.OnGiftStageWon );

  if ( not skip_notify ) then
    mediator.Notify( "View_Minievent_GiftStarted" );
  end;

end;
--------------------------------------------------------------------------------
function private.OnGiftStageWon ( stage_data )

  local event_data = public.GetCurrentCustomEvent( TYPE_GIFT );
  local stage_name = stage_data.name;

  if ( event_data ) then
    local index = common.IsEntityInList( event_data.reward.chests,
    { stage = tonumber( stage_name:match( "%d+" ) ) }, "stage" );

    if ( index and not private.ready_gift ) then

      private.ready_gift = event_data.reward.chests[ index ];

      -- JK: immediately obtain the chest
      v_items.TryObtain( private.ready_gift );

      m_minievent.Gift_RefreshObtained();

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.OnGiftFinished ( event_id, skip_notify )

  mediator.Unsubscribe( "Model_Stage_StageWon", private.OnGiftStageWon );

  if ( not skip_notify ) then
    mediator.Notify( "View_Minievent_GiftFinished" );
  end;

end;
--------------------------------------------------------------------------------
function private.OnGiftUpdated ( event_id, skip_notify )

  if ( not skip_notify ) then
    mediator.Notify( "View_Minievent_GiftUpdated" );
  end;

end;
--------------------------------------------------------------------------------
function private.OnGiftCooldownFinished ( event_id )

  m_minievent.Gift_RefreshObtained( true );
  private.OnGiftUpdated( event_id )

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Launcher_Update ()

  local event_list, excluded_list = private.Launcher_FilteredEventList();

  if ( #excluded_list > 0 ) then

    for i = 1, #excluded_list do

      if ( excluded_list[ i ].type == TYPE_GIFT ) then

        private.OnGiftUpdated( excluded_list[ i ].id );
        break;

      end;

    end;

  end;

  if ( #event_list < 1 ) then

    private.Launcher_Disable();
    return;

  end;

  --- JK: subscribe the button in case it's still not subscribed
  if ( not private.launcher_subscribed ) then

    mediator.Subscribe(
      "Model_Minievent_EventFinished",
      private.Launcher_Update
    );

    private.launcher_subscribed = true;

  end;

  if ( not private.launcher_enabled ) then

    -- JK: enable minievent launcher
    private.Launcher_Enable();

  end;

  --- JK: update buttons in launcher
  for index = 1, 2 do

    local event = event_list[ index ];

    if ( event ) then

      local function func_LaunchEvent ()
        private.Dialog_Show( event.id );
        v_log.Playhaven_LogPlacement( "event_open" );
      end;

      local function func_GetRemainingTime ()
        local  seconds = m_minievent.Event_GetRemainingTime( event.id );
        return seconds;
      end;

      local icon_name = event.type;
      local str_title = "str_int_dialog_minievent_"..event.type;

      v_interface.panel.left.MinieventEnableSlot(
        index,
        icon_name,
        str_title,
        func_LaunchEvent,
        func_GetRemainingTime
      );

    else
      
      v_interface.panel.left.MinieventDisableSlot( index );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Launcher_FilteredEventList ()

  local event_list = m_minievent.GetEventList();
  local excluded_list = {};

  for pos = #event_list, 1, -1 do

    -- JK: story, gift and section are not shown in launcher
    if ( event_list[ pos ].type == TYPE_STORY   ) or
       ( event_list[ pos ].type == TYPE_GIFT    ) or
       ( event_list[ pos ].type == TYPE_SECTION ) then

      table.insert( excluded_list, table.remove( event_list, pos ) );

    end;

  end;

  return event_list, excluded_list;

end;
--------------------------------------------------------------------------------
function private.Launcher_Enable ()

  v_interface.panel.left.SetDelegates( {
    Minievent_OnTap = function ()

      local event_list = private.Launcher_FilteredEventList();

      if ( #event_list > 1 ) then

        v_interface.panel.left.MinieventToggleTray();

      else

        local event_id = event_list[ 1 ].id;

        private.Dialog_Show( event_id );
        v_log.Playhaven_LogPlacement( "event_open" );

      end;

    end
  } );

  v_interface.Scheme_AddButton( "map", "left", { name = "minievent" } );

  private.launcher_enabled = true;

end;
--------------------------------------------------------------------------------
function private.Launcher_Disable ()

  if ( private.launcher_subscribed ) then

    mediator.Unsubscribe(
      "Model_Minievent_EventFinished",
      private.Launcher_Update
    );

    private.launcher_subscribed = nil;

  end;

  v_interface.Scheme_DeleteButton( "map", "left", { name = "minievent" } );

  private.launcher_enabled = nil;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Decoration_TryShow ()

  local event_list = m_minievent.GetEventList();

  local decoration_found = false;

  for i = #event_list, 1, -1 do

    local event_name = event_list[ i ].name;

    local has_decoration, decoration_settings =
      m_minievent.Event_HasDecoration( event_name );

    if ( has_decoration ) then

      if ( decoration_settings.top ) then
        v_interface.panel.top.DecorationEnable( "minievent_"..event_name );
      else
        v_interface.panel.top.DecorationDisable();
      end;

      if ( decoration_settings.bottom ) then
        v_interface.panel.bottom.DecorationEnable( "minievent_"..event_name );
      else
        v_interface.panel.bottom.DecorationDisable();
      end;

      if ( decoration_settings.left ) then
        v_interface.panel.left.DecorationEnable( "minievent_"..event_name );
      else
        v_interface.panel.left.DecorationDisable();
      end;

      if ( decoration_settings.right ) then
        v_interface.panel.right.DecorationEnable( "minievent_"..event_name );
      else
        v_interface.panel.right.DecorationDisable();
      end;

      decoration_found         = true;
      private.decoration_shown = true;

      break;

    end;

  end;

  if ( not decoration_found     ) and
     ( private.decoration_shown ) then

    private.Decoration_Hide();

  end;

end;
--------------------------------------------------------------------------------
function private.Decoration_Hide ()

  v_interface.panel.top.DecorationDisable();
  v_interface.panel.bottom.DecorationDisable();
  v_interface.panel.left.DecorationDisable();
  v_interface.panel.right.DecorationDisable();

  private.SetWorldsOverlayIcon( nil, nil, false, 0 );
  
  private.decoration_shown = nil;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Event_WaitForStageResult ( event_type, event_id )

  local func_OnStageWon;
  local func_OnStageFailed;
  local func_OnStageAbandon;

  local func_OnMapUpdate = function ( state )

    mediator.Unsubscribe( "View_Stage_StageFailed",  func_OnStageFailed );
    mediator.Unsubscribe( "View_Stage_StageAbandon", func_OnStageAbandon );
    mediator.Unsubscribe( "View_Stage_StageWon",     func_OnStageWon );

    if ( event_type ~= TYPE_SECTION ) then

      private.OnMapUpdate = function ( func_after )

        local minievent_data = m_minievent.Event_GetById( event_id );
        
        if ( minievent_data.cooldown            ) and
           ( minievent_data.type ~= TYPE_CHESTS ) then
          m_minievent.Cooldown_Start( minievent_data );
        end;
        
        private.Dialog_Show( event_id, state, nil, func_after );
        private.OnMapUpdate = nil;

      end;

    end;

  end;

  func_OnStageFailed = function ( stage_data )

    m_minievent.Event_SetProgress( event_id, false );

    local minievent_data = m_minievent.Event_GetById( event_id );

    v_log.Kontagent_LogEvent( "minievent",
    {
      name     = "fail",
      category = minievent_data.name,
      value    = #minievent_data.stages_completed + 1,
      level    = minievent_data.scene
    });

    v_game.GameLog_Add( "minievent_fail",
    {
      category = minievent_data.name,
      level    = #minievent_data.stages_completed + 1
    });

    m_minievent.Event_Update( event_id, stage_data.name, "fail" );

    func_OnMapUpdate( "fail" );

  end;

  func_OnStageAbandon = function ( stage_data )

    if ( stage_data.moves_done > 0 ) then
      m_minievent.Event_SetProgress( event_id, false );
    end;

    local minievent_data = m_minievent.Event_GetById( event_id );

    v_log.Kontagent_LogEvent( "minievent",
    {
      name     = "abandon",
      category = minievent_data.name,
      value    = #minievent_data.stages_completed + 1,
      level    = minievent_data.scene
    });

    v_game.GameLog_Add( "minievent_abandon",
    {
      category = minievent_data.name,
      level    = #minievent_data.stages_completed + 1
    });

    if ( stage_data.moves_done > 0 ) then
      m_minievent.Event_Update( event_id, stage_data.name, "fail" );
      func_OnMapUpdate( "fail" );
      return;
    end;

    func_OnMapUpdate( "abandon" );

  end;

  func_OnStageWon = function ( stage_data )
    
    local index = public.Event_GetActiveStageIndex( event_id );
    if ( not index ) then
      Log.i( "func_OnStageWon: no active stage index, aborting" );
      return;
    end;

    local stages = public.Event_GetStages( event_id );
    local stage  = stages[ index ];
    
    if ( not stage ) then
      Log.e( "func_OnStageWon: no stage found, aborting" );
      return;
    end;
    
    if ( stage.name ~= stage_data.stage ) then
      
      local stage_name = stage.name;
      
      local is_tutorial, tutorial_data =
        m_minievent.Event_HasTutorial( event_id, stage_name );

      if ( is_tutorial ) then
        stage_name = tutorial_data.replace;
      end;
      
      if ( stage_name ~= stage_data.stage ) then
        Log.e( "func_OnStageWon: stage name does not match, aborting" );
        Log.e( stage_name.." ~= "..stage_data.stage );
        return;
      end;
      
    end;

    m_minievent.Event_SetProgress( event_id, false );
    m_minievent.Event_Update( event_id, stage_data, "win" );

    local minievent_data = m_minievent.Event_GetById( event_id );

    if ( minievent_data.cooldown             ) and
       ( minievent_data.type == TYPE_CHESTS  ) and
       ( not minievent_data.is_target_reward ) then

      m_minievent.Cooldown_Start( minievent_data );

    end;

    v_log.Kontagent_LogEvent( "minievent",
    {
      name     = "success",
      category = minievent_data.name,
      value    = #minievent_data.stages_completed,
      level    = minievent_data.scene
    });

    v_game.GameLog_Add( "minievent_success",
    {
      category = minievent_data.name,
      level    = #minievent_data.stages_completed + 1
    });

    --- JK: target reward ------------------------------------------------------
    if ( minievent_data.is_target_reward ) then

      --DbgTrace( "[ INFO ][ v_minievent ] Target reward." );

      private.is_target_reward = true;

      local reward = minievent_data.targets[ 1 ].reward;
      reward.category   = "minievent_"..minievent_data.name;
      reward.source     = "minievent_"..minievent_data.name;
      reward.stage_name = "minievent_"..minievent_data.name;

      v_items.TryObtain( reward );

    end;

    --- JK: progress reward ----------------------------------------------------
    if ( minievent_data.is_reward ) then

      private.Reward_TryObtain( event_id, minievent_data );

      v_log.Kontagent_LogEvent( "minievent",
      {
        name     = "collect_reward",
        category = minievent_data.name,
        level    = minievent_data.scene
      });

    end;

    func_OnMapUpdate( "win" );

  end;

  mediator.Subscribe( "View_Stage_StageWon",     func_OnStageWon );
  mediator.Subscribe( "View_Stage_StageFailed",  func_OnStageFailed );
  mediator.Subscribe( "View_Stage_StageAbandon", func_OnStageAbandon );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Dialog_Show ( event_id, state, func_on_end, func_on_show )

  m_minievent.SetCurrentEventId( event_id );

  local minievent_data = m_minievent.Event_GetById( event_id );

  if ( minievent_data ) then

    minievent_data.reward = private.Reward_Filter( minievent_data.reward );
    minievent_data.is_target_reward = minievent_data.is_target_reward or
                                      private.is_target_reward;

    mediator.Subscribe(
      "Model_Minievent_EventUpdated",
      private.Dialog_Update
    );

    mediator.Subscribe(
      "Model_Minievent_EventFinished",
      private.Dialog_Update
    );

    minievent_data.animate = state;

    local skin_name = v_player.Skin_GetCurrent();
    local skin_data = {
      name    = skin_name,
      content = v_player.Skin_GetContent( skin_name )
    };

    local delegates = {
      OnShow = function ()

        if ( func_on_show ) then func_on_show(); end;

      end,
      ShowStartDialog   = function ()

        if ( minievent_data.type == "progress" ) then
          private.Crown_HighlightStage( event_id, func_on_end );
        else
          local is_shown, reason = private.DialogStart_Show( event_id );

          if ( not is_shown ) then
            if ( reason == "suspended" ) then private.Dialog_Update(); end;
          end;
        end;

      end,
      ShowRewardDialog  = private.Reward_Show,
      ShowPreviewDialog = private.Reward_Preview,
      Close             = function ( func_after_close )

        private.Dialog_Hide(
          nil,
          function ()

            if ( func_after_close ) then func_after_close() end;
            if ( func_on_end ) then func_on_end(); end;

          end
        );
        v_log.Playhaven_LogPlacement( "event_close" );

      end,
      CheckShare = function ( func_after )

        if ( not v_social.IsFBShareActive() ) then
          func_after();
        elseif ( minievent_data.is_completed ) then
          private.DialogShare_Show( func_after );
        elseif ( minievent_data.is_target_reward ) then
          private.DialogShare_Show( func_after );
        else
          func_after();
        end;

      end,
      GetSeconds = function ()

        local seconds = m_minievent.Event_GetRemainingTime( event_id );
        return seconds;

      end,
      ShowTopPanel = function ( updates )
        
        if ( updates                     ) and
           ( #updates == 1               ) and
           ( updates[ 1 ]:find( "gold" ) ) then
          v_interface.panel.top.Show( false, "gold" );
        else
          v_interface.panel.top.Show( false );
        end;
        
      end,
      UpdateTopPanel = function ( updates )

        for _, resource_name in pairs( updates ) do
          if ( resource_name:find( "gold" ) ) then
            v_interface.Gold_Buffer_Out();
          else
            v_interface.Resource_Buffer_Out( resource_name );
          end;
        end;

      end,
      HideTopPanel = function ( func_end )

        v_interface.panel.top.Hide(
          false,
          function ()
            if ( func_end ) then func_end(); end;
          end
        );

      end,
      IsTopPanelVisible = function ()

        local  is_visible = v_interface.panel.top.IsVisible();
        return is_visible;

      end,
      GetCooldownSeconds = function ()

        local  remaining_time = m_minievent.Cooldown_GetRemainingTime();
        return remaining_time;

      end,
      SkipCooldown = function ()

        local  is_skipped = private.Cooldown_TrySkip( minievent_data.id );
        return is_skipped;

      end,
      StopCooldown = function ()

        local  is_stopped = private.Cooldown_Stop( minievent_data.id );
        return is_stopped;

      end,
      ShowChestReward = function ( chest_id, func_after )

        private.Reward_ShowContainer(
          minievent_data.name,
          minievent_data.reward,
          chest_id,
          func_after
        );

      end,
      GetNextReward = function ()

        local  next_reward = private.Reward_GetNext( minievent_data.id );
        return next_reward;

      end,
      GetDialogData = function ( stage_index, when )

        local dialog_data = m_minievent.CharacterDialog_GetData(
          minievent_data.id,
          minievent_data.scene,
          "stage_"..stage_index,
          when
        );

        return dialog_data;

      end,
      SetDialogComplete = function ( dialog_name )

        m_minievent.CharacterDialog_Complete(
          m_minievent.GetCurrentEventId(),
          dialog_name
        );

      end,
      SwitchScene = function ()

        private.Dialog_Reload( minievent_data.id );

      end,
      GetBookData = function ()

        local  book_data = m_minievent.Book_GetData( minievent_data.id );
        return book_data;

      end,
      SetBookComplete = function ()

        m_minievent.Book_SetComplete(
          m_minievent.GetCurrentEventId()
        );

      end,
      CheckRewardStage = function ( func_after, point_index )

        private.Update_TryShowStageReward( func_after, point_index );

      end,
      GetCurrentSkin = function ()
        return v_player.Skin_GetCurrent();
      end,
      GetCrownDataChanged = function ()
        local new_progress, old_progress = private.Crown_GetDataChanged();
        return new_progress, old_progress;
      end
    };

    v_interface.window.Show(
      private.this.dialog,
      function ()

        int_dialog_minievent.Show( minievent_data, skin_data );

      end,
      nil,
      delegates
    );

    if ( minievent_data.is_reward ) then
      private.reward_obtained = {};
    end;

    m_minievent.ClearStageWon( event_id );
    private.is_target_reward = false;

  end;

end;
--------------------------------------------------------------------------------
function private.Dialog_Hide ( is_instantly, func_end )

  mediator.Unsubscribe(
    "Model_Minievent_EventUpdated",
    private.Dialog_Update
  );

  mediator.Unsubscribe(
    "Model_Minievent_EventFinished",
    private.Dialog_Update
  );

  int_dialog_minievent.Hide(
    function ()

      v_interface.window.Hide( private.this.dialog, func_end );

    end,
    is_instantly
  );

end;
--------------------------------------------------------------------------------
function private.Dialog_Update ()

  if ( v_interface.window.IsOnScreen( private.this.dialog ) ) then

    local event_id        = m_minievent.GetCurrentEventId();
    local minievent_data  = m_minievent.Event_GetById( event_id );
    minievent_data.reward = private.Reward_Filter( minievent_data.reward );

    int_dialog_minievent.Update(
      minievent_data,
      private.Dialog_Update
    );

  end;

end;
--------------------------------------------------------------------------------
function private.Dialog_Reload ( event_id, func_OnReopen )

  event_id = event_id or m_minievent.GetCurrentEventId();

  local function func_OnShow ()

    private.Dialog_Hide(
      true,
      function ()

        if ( func_OnReopen ) then func_OnReopen(); end;

        if ( not m_minievent.Event_IsSuspended() ) then

          private.Dialog_Show(
            event_id,
            nil,
            nil,
            function () v_interface.Loading_Hide(); end
          );

        else

          v_interface.Loading_Hide();

        end;

      end
    );

  end;

  v_interface.Loading_Show( true, func_OnShow );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.DialogShare_Show ( func_after )

  if ( not ( IsMobile() and m_social.IsLogged() ) ) then
    func_after();
    return;
  end;

  local delegates = {
    Share = function()

      private.Share();

    end,
    Close  = function ()

      private.DialogShare_Hide();

    end,
    OnClose = function ()

      v_interface.window.Hide( private.this.dialog_share, func_after );

    end
  };

  v_interface.window.Show(
  private.this.dialog_share,
    function ()

      _G[ private.this.dialog_share ].Show();

    end,
    nil,
    delegates
  );

end;
--------------------------------------------------------------------------------
function private.DialogShare_Hide ()

  _G[ private.this.dialog_share ].Hide();

end;
--------------------------------------------------------------------------------
function private.Share ()

  if ( v_game.IsOnline() ) then

    local icon = "icon_event_complete";
    local title = ne.strings.Get( "str_fbshare_event_complete_title" );
    local description = ne.strings.Get( "str_fbshare_event_complete_description" );

    v_game.ShareToFacebook( title, description, icon, private.DialogShare_Hide );

  else

    v_game.SystemDialog_Show( "nointernetconnection", function () end );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.DialogStart_Show ( event_id, stage_index, func_before, force )

  -- JK: check for event suspended
  if ( m_minievent.Event_IsSuspended() ) then return false, "suspended"; end;

  local minievent_data = m_minievent.Event_GetById( event_id );

  -- JK: check for data exists -------------------------------------------------
  if ( not minievent_data ) then
    return false, "no_data";
  end;

  -- JK: check for event is active ---------------------------------------------
  if ( minievent_data.is_finished ) then
    return false, "finished";
  end;

  -- JK: check for proper stage index ------------------------------------------
  if ( stage_index and not force ) then

    local current_index = #minievent_data.stages_completed + 1;

    -- JK: only next stage is eligible
    if ( stage_index ~= current_index ) then
      return false, "progress_mismatch";
    end;

  end;

  -- JK: check for stage data is available -------------------------------------
  local stage_name     = ( force and stage_index ) and 
                          public.Event_GetStages( event_id )[ stage_index ].name or 
                          minievent_data.stage.name;
  local stage_progress = stage_name;

  private.stage_selected = stage_name;
  
  local is_tutorial, tutorial_data =
    m_minievent.Event_HasTutorial( event_id, stage_name, minievent_data );

  if ( is_tutorial ) then

    local tutorial_name = tutorial_data.name;
    local replace_stage = tutorial_data.replace;

    if ( not v_tutorial.IsCompleted( tutorial_name ) ) then

      stage_name = replace_stage;

    end;

  end;

  local stage_data = private.GetStartData( stage_name );

  if ( not stage_data ) then

    Log.e( "No data found for "..tostring( stage_name ) );
    return false, "no_stage_data";

  end;
  
  if ( minievent_data.type == TYPE_SECTION ) then
    -- JK: dirty hack
    stage_data.section = stage_data.zone.background.content;
  end;

  local func_boost_buy = function ( booster_name )

    return v_boosters.Tap(
      "dialog",
      booster_name,
      "minievent_"..minievent_data.name
    );

  end;

  local function func_accept ( booster_list )

    local boosters = {};

    for key, value in pairs( booster_list ) do

      table.insert( boosters, key );

    end;

    v_boosters.SetSelected( boosters );

    if ( v_stage.IsCanPlay() ) then

      private.DialogStart_Hide( function ()

        if ( m_minievent.Event_IsSuspended() ) then
          return;
        end;

        if ( v_stage.TryStart( stage_name ) ) then

          mediator.SubscribeOnce(
            "Screen_AfterClose",
            function () private.Dialog_Hide( true ); end
          );

          mediator.SubscribeOnce(
            "Model_Match3_FieldActivated",
            function ()
              m_minievent.Event_SetProgress( minievent_data.id, stage_progress );
            end
          );

          private.Event_WaitForStageResult( minievent_data.type, minievent_data.id );

          if ( minievent_data.cooldown         ) and
             ( minievent_data.type ~= TYPE_CHESTS ) then

            m_minievent.Cooldown_Start( minievent_data );

          end;

        else

          private.Dialog_Hide();

        end;

      end, true );

    end;

  end;

  local func_reject = private.DialogStart_Hide;

  mediator.Subscribe(
    "Model_Boosters_CountChanged",
    private.DialogStart_Update
  );

  local delegates = {
    DataBooster = function ( booster_name )

      return private.GetDataBooster( booster_name );

    end
  };

  common.delegate.Set( "int_dialog_minievent_start", delegates );

  local func_dialog = function ()

    v_interface.window.Show(
      private.this.dialog_start,
      function ()

        int_dialog_minievent_start.Show(
          stage_name,
          stage_data,
          func_accept,
          func_reject,
          func_boost_buy
        );

      end
    );

  end;

  if ( func_before ) then
    func_before( func_dialog );
  else
    func_dialog();
  end;

  return true, "ok";

end;
--------------------------------------------------------------------------------
function private.DialogStart_Update ( booster_data )

  int_dialog_minievent_start.Update( booster_data );

end;
--------------------------------------------------------------------------------
function private.DialogStart_IsActive ()

  return int_dialog_minievent_start.IsActive();

end;
--------------------------------------------------------------------------------
function private.DialogStart_Hide ( func_on_end, start )

  mediator.Unsubscribe(
    "Model_Boosters_CountChanged",
    private.DialogStart_Update
  );

  local func_after = function ()
    if ( func_on_end ) then func_on_end(); end;
    if ( not start ) then
      v_stage.Finish();
    end;
  end;

  int_dialog_minievent_start.Hide(
    function ()

      v_interface.window.Hide( private.this.dialog_start, func_after );

    end
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.DialogReward_Show (
    reward_group,
    reward_data,
    is_reward,
    func_end,
    stage_number
  )

  local is_gold   = false;
  local is_heart  = false;
  local resources = {};

  -- JK: Disable close button on reward mode
  if ( is_reward ) then

    ui.button.Disable( "btn_"..private.this.dialog_reward.."_close" );

  end;

  for _, reward in pairs( reward_data ) do

    if ( reward.type == "gold"     ) then is_gold  = true; end;
    if ( reward.type == "heart"    ) then is_heart = true; end;
    if ( reward.type == "resource" ) then
      table.insert( resources, reward.name );
    end;

  end;
  
  local is_resource = ( #resources > 0 );

  local delegates = {
    OnShow = function()
      
      if ( is_reward ) then
        if ( is_resource ) then
          v_interface.panel.top.Show();
        elseif ( is_gold ) then
          v_interface.panel.top.Show( nil, "gold" );
        end;
      end;

    end,
    Accept = function()

      if ( is_gold ) then
        v_interface.Gold_Buffer_Out();
      end;
      
      if ( is_resource ) then
        for i = 1, #resources do
          v_interface.Resource_Buffer_Out( resources[ i ] );
        end;
      end;

      return true;

    end,
    Close  = private.DialogReward_Hide,
    OnHide = function ()

      if ( v_interface.panel.top.IsVisible()   ) and
         ( not m_minievent.Cooldown_IsActive() ) then

        v_interface.panel.top.Hide( nil, nil );

      end;

      local func_hide = func_end;

      if ( is_reward ) then

        func_hide = function ()

          ui.button.Enable( "btn_"..private.this.dialog_reward.."_close" );

          if ( func_end ) then func_end() end;

        end;

      end;

      v_interface.window.Hide( private.this.dialog_reward, func_hide );

    end
  };

  v_interface.window.Show(
    private.this.dialog_reward,
    function ()

      _G[ private.this.dialog_reward ].Show(
        reward_group,
        reward_data,
        is_reward,
        is_heart,
        stage_number
      );

    end,
    nil,
    delegates
  );

end;
--------------------------------------------------------------------------------
function private.DialogReward_Hide ()

  _G[ private.this.dialog_reward ].Hide();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Reward_Preview (
    reward_group,
    reward_subgroup,
    func_end,
    event_type,
    stage_number
  )

  private.Reward_Preview(
    reward_group,
    reward_subgroup,
    func_end,
    event_type,
    stage_number
  );

end;
--------------------------------------------------------------------------------
function public.Reward_Show ( reward_group, reward_subgroup, func_end )
  
  private.Reward_Show( reward_group, reward_subgroup, func_end );

end;
--------------------------------------------------------------------------------
function private.Reward_TryObtain ( event_id, minievent_data )

  --DbgTrace( "[ INFO ][ v_minievent ] Reward_TryObtain" );

  local reward_group = minievent_data.reward_group;
  local reward_data  = m_minievent.Event_GetReward( event_id, reward_group );

  if ( minievent_data.type == TYPE_SECTION ) then
    reward_data = reward_data.rewards;
  end;

  private.reward_obtained = {};

  for _, reward in pairs( reward_data ) do

    reward.category   = "minievent_"..minievent_data.name;
    reward.source     = "minievent_"..minievent_data.name;
    reward.stage_name = "minievent_"..minievent_data.name;
    reward.silent     = true;

    local is_obtained = v_items.TryObtain( reward );
    local obtain_id   = reward_group.."_"..reward.name;

    private.reward_obtained[ obtain_id ] = is_obtained;

  end;

  if ( minievent_data.reward_group == "final"  ) or
     ( minievent_data.reward_group == TYPE_CHESTS ) or
     (
        ( minievent_data.type == TYPE_SECTION                           ) and
        ( #minievent_data.stages == #minievent_data.stages_completed )
     ) then

    if ( minievent_data.type == TYPE_CHESTS     ) and
       ( not minievent_data.is_target_reward ) then
      return;
    end;

    if ( minievent_data.type == TYPE_STORY                    ) and
       ( minievent_data.scene < minievent_data.scene_total ) then
      return;
    end;

    m_minievent.Event_Complete( event_id );

  end;

end;
--------------------------------------------------------------------------------
function private.Reward_Filter ( reward_data )

  for group, data in pairs( reward_data ) do

    reward_data[ group ] = private.Reward_FilterGroup( group, data );

  end;

  return reward_data;

end;

--------------------------------------------------------------------------------
function private.Reward_FilterGroup ( reward_group, reward_data )

  for i = #reward_data, 1, - 1 do

    local reward    = reward_data[ i ];
    local obtain_id = reward_group.."_"..reward.name;

    if ( reward.type == "skinpart"                         ) and
       ( not private.reward_obtained[ obtain_id ]          ) and
       ( not v_player.Skin_IsPartObtainable( reward.name ) ) then

      -- JK: remove reward if it's not obtained and not obtainable
      table.remove( reward_data, i );

    end;

  end;

  return reward_data;

end;
--------------------------------------------------------------------------------
function private.Reward_ShowContainer (
    event_name,
    reward_data,
    chest_id,
    func_after,
    view
  )

  view = view or "chest";

  local chest_data = reward_data;

  -- JK: bare rewards or chest? ------------------------------------------------
  if ( not chest_data.items ) then

    -- JK: default chest with bare rewards -------------------------------------
    local chest_data = {
      type  = "chest",
      name  = "woodchest",
      count = 1,
      items = reward_data
    };

  end;

  -- JK: try to find chests in reward_data -------------------------------------
  if ( reward_data.chests ) then
    chest_data = reward_data.chests[ chest_id ];
  end;

  -- JK: final reward is given through chest -----------------------------------
  if ( reward_data.final ) then
    chest_data = reward_data.final[ 1 ];
  end;

  local ext_rewards = chest_data.items;
  chest_data.items  = nil;

  local obtain_params =
  {
    category   = "minievent_"..event_name,
    source     = "minievent_"..event_name,
    stage_name = "minievent_"..event_name,
    silent     = true
  };

  local has_resource_reward = false;
  local ignore_panel_state  = v_interface.panel.top.IsVisible();

  for _, reward in pairs( ext_rewards ) do
    if ( reward.name:find( "resource" ) ) then
      has_resource_reward = true;
      break;
    end;
  end;

  if ( not has_resource_reward ) and
     ( not ignore_panel_state  ) then
    v_interface.panel.top.Highlight( "gold" );
  end;

  local handler = {
    chest = v_items.Chest_Show,
    gift  = v_items.Gift_Show
  };

  handler[ view ](
    chest_data,
    obtain_params,
    ext_rewards,
    func_after,
    nil,
    ignore_panel_state
  );

end;
--------------------------------------------------------------------------------
function private.Reward_GetNext ( event_id )

  local progress     = m_minievent.Event_GetProgress( event_id );
  local stage_index  = progress + 1;
  local event_stages = m_minievent.Event_GetStages( event_id );

  if ( stage_index > #event_stages ) then
    stage_index = 1;
  end;

  m_minievent.Event_ResolveReward( event_id, stage_index );

  local  reward = m_minievent.Event_GetReward( event_id );
  return reward;

end;
--------------------------------------------------------------------------------
function private.Reward_Preview ( 
    reward_group,
    reward_subgroup,
    func_end,
    event_type,
    stage_number
  )

  if ( m_minievent.Event_IsSuspended() ) then return; end;

  local event_id = m_minievent.GetCurrentEventId();
  if ( event_type ) then
    event_id = m_minievent.GetCurrentCustomEventId( event_type );
  end;

  local reward_data = private.Reward_FilterGroup(
    reward_group,
    m_minievent.Event_GetReward( event_id, reward_group, reward_subgroup, true )
  );

  if ( #reward_data < 1 ) then

    DbgTrace( "[ ERROR ][ v_minievent ] No reward found for preview." );

    if ( func_end ) then func_end(); end;

    return;

  end;

  private.DialogReward_Show(
    reward_group,
    reward_data,
    false,
    func_end,
    stage_number
  );

end;
--------------------------------------------------------------------------------
function private.Reward_Show ( reward_group, reward_subgroup, func_end )

  local event_id    = m_minievent.GetCurrentEventId();
  local reward_data = m_minievent.Event_GetReward(
    event_id,
    reward_group,
    reward_subgroup,
    true
  );

  if ( #reward_data < 1 ) then

    DbgTrace( "[ ERROR ][ v_minievent ] No reward found to show." );

    if ( func_end ) then func_end(); end;

    return;

  end;

  private.DialogReward_Show(
    reward_group,
    reward_data,
    true,
    func_end
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Cooldown_TrySkip ( event_id )

  if ( m_minievent.Cooldown_IsActive( event_id ) ) then

    local price   = m_minievent.Cooldown_GetSkipPrice( event_id );
    local is_paid = v_player.Gold_TryChangeCount( -price );

    if ( not is_paid ) then return false; end;

    v_log.Kontagent_LogEvent( "economy",
    {
      description = "premium_currency_spend",
      category    = "minievent",
      value       = - price,
      name        = "skip_cooldown"
    } );

    return m_minievent.Cooldown_Stop( event_id );

  end;

  return false;

end;
--------------------------------------------------------------------------------
function private.Cooldown_Stop ( event_id )

  if ( m_minievent.Cooldown_IsActive( event_id ) ) then

    return m_minievent.Cooldown_Stop( event_id );

  end

  return false;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.ChangeList_Prepare ( stage_name, reward_data )

  private.reward_data_stage = nil;

  local rewards = {};

  for r, reward in pairs( reward_data ) do
    if ( reward.name == "resource4" ) then
      table.insert( rewards, reward );
    end;
  end;

  if ( #rewards ~= 0 ) then
    private.reward_data_stage = {
      stage_name = stage_name,
      rewards    = rewards
    };
  end;

end;
--------------------------------------------------------------------------------
function private.Update_TryShowStageReward ( func_after, point_index )

  if ( not private.reward_data_stage ) then
    func_after();
    return;
  end;

  local func_reward_accept = function ()

    int_dialog_map_reward.Hide(
      nil,
      function ()
        v_interface.window.Hide( "int_dialog_map_reward",
        function ()
          private.Update_TryShowResourcesReward( func_after, point_index );
        end
        );
      end
    );

  end;

  v_interface.window.Show(
    "int_dialog_map_reward",
    function ()

      int_dialog_map_reward.Show( private.reward_data_stage.rewards );

    end,
    { force_modal = true },
    { Accept = func_reward_accept }
  );

end;
--------------------------------------------------------------------------------
function private.Update_TryShowResourcesReward ( func_after, point_index )

  local resources = private.reward_data_stage.rewards;

  private.reward_data_stage = nil;

  v_interface.panel.top.Highlight( "resource4" );
  v_interface.panel.top.Show();

  local func_hide_interface = function ()
    v_interface.panel.top.Hide( true );
  end;

  for r = 1, #resources do

    local reward = resources[ r ];

    if ( not point_index ) then

      if ( r == #resources ) then
        v_interface.Resource_Buffer_Out( reward.name, func_hide_interface );
        func_after();
      else
        v_interface.Resource_Buffer_Out( reward.name );
      end;

    else

      int_dialog_minievent.Point_Reward_ShowFireflies( point_index, reward,
        function ()

          if ( r == #resources ) then
            v_interface.Resource_Buffer_Out( reward.name, func_hide_interface );
            func_after();
          else
            v_interface.Resource_Buffer_Out( reward.name );
          end;

        end
      );

    end;

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.GetStartData ( stage_name )

  local db_data = v_stage.Stage_StaticData( stage_name );

  if ( not db_data ) then return false; end;

  local stage_state = 1;
  local db_index    = stage_state;

  local start_data = {
    type     = db_data.type,
    realtype = db_data.realtype,
    zone     = db_data.zone,
    content  = db_data.content,
    moves    = db_data.state[ db_index ].moves,
    time     = db_data.state[ db_index ].time,
    targets  = db_data.state[ db_index ].targets,
    boosters = v_stage.Boosters_GetStageList( stage_name, "start" ),
    color    = private.GetDialogBackgroundColor(
                 db_data.zone,
                 db_data.type,
                 ( db_data.enemy or {} ).name
                 ),
    resources =  db_data.state[ db_index ].chip_types
  };

  -- Trim not resources from chip_types
  for i = #start_data.resources, 1, - 1 do

    if not ( string.find( start_data.resources[ i ], "Resource" ) ) then

      table.remove( start_data.resources, i );

    end;

  end;

  -- Add resource form targets
  if ( #start_data.resources > 0 ) then
    for i = 1, #start_data.targets do
      if ( start_data.targets[ i ].type == "CollectResource" ) then
        local name = start_data.targets[ i ].name;
        local exist = false;
        for r = 1, #start_data.resources do
          if ( string.lower( start_data.resources[ r ] ) == name ) then
            exist = true;
          end;
        end;
        if ( not exist ) and ( string.find( name, "resource" ) ) then
          table.insert( start_data.resources, name );
        end;
      end;
    end;
  end;

  return start_data;

end;
--------------------------------------------------------------------------------
function private.GetDialogBackgroundColor ( zone, stage_type, enemy_name )

  zone = zone.background.name;
  
  local colors =
  {
    default = {
      ride  = { r = 0.5484, g = 0.6617, b = 0.1562 }
    },
    chests  = {
      ride  = { r = 0.9607, g = 0.8392, b = 0.4549 }
    },
    christmas  = {
      ride  = { r = 0.2901, g = 0.6745, b = 0.8392 }
    },
    tourney  = {
      ride  = { r = 0.5921, g = 0.7372, b = 0.4980 }
    },
    tourney_valentine  = {
      ride  = { r = 0.9843, g = 0.7019, b = 0.5764 }
    },
    tower  = {
      ride  = { r = 0.6000, g = 0.8862, b = 0.7137 }
    },
    tower_valentine  = {
      ride  = { r = 0.9843, g = 0.7019, b = 0.5764 }
    },
    treasury  = {
      ride  = { r = 0.9607, g = 0.8392, b = 0.4549 }
    },
    patrick  = {
      ride  = { r = 0.4745, g = 0.8000, b = 0.3882 }
    },
    tourney_foolsday = {
      ride  = { r = 0.4470, g = 0.7254, b = 0.3960 }
    },
    piratesday_tropic = {
      ride  = { r = 0.4431, g = 0.7216, b = 0.3922 }
    },
    piratesday_beach = {
      ride  = { r = 0.2784, g = 0.6510, b = 0.5686 }
    },
    tourney_anniversary = {
      ride  = { r = 0.5255, g = 0.8039, b = 0.5137 }
    },
    halloween_forest = {
      ride  = { r = 0.5019, g = 0.3254, b = 0.698 }
    },
    snowvillage = {
      ride  = { r = 0.2196, g = 0.6510, b = 0.7765 }
    },
    icecastle = {
      ride  = { r = 0.5176, g = 0.8980, b = 0.9373 }
    },
    toyplains = {
      ride  = { r = 0.8667, g = 0.9804, b = 0.9922 }
    },
    santahouse = {
      ride  = { r = 0.9765, g = 0.7255, b = 0.3608 }
    }
  };

  if not ( colors[ zone ] ) then

    zone = "default";

  end;

  local color = colors[ zone ][ stage_type ];

  if ( enemy_name ) then

    color = colors[ zone ][ enemy_name ];

  end;

  if not ( colors[ zone ][ enemy_name ] ) then

    color = colors[ zone ][ stage_type ];

  end;

  return color;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Crown_Init ()
  
  private.crown_progress_saved = 0;
  private.crown_progress_changed = {
    new = 0,
    old = 0
  };
  
  private.crown_progress_set_start_data = false;
  
end;
--------------------------------------------------------------------------------
function private.Crown_SetStartValue ()
  
  if ( private.crown_progress_set_start_data ) then
    return;
  end;
  
  private.crown_progress_set_start_data = true;
  
  private.crown_progress_saved = 0;
  
  local value = m_minievent.Crown_GetValue();
  private.crown_progress_changed = {
    new = value,
    old = value
  };
  
end;
--------------------------------------------------------------------------------
function public.Crown_IsAvailableMinievent ()
 
  local event_list = m_minievent.GetEventList();
  
  if ( #event_list < 1 ) then
    return false;
  end;

  local event_exist = false;
  
  for index, event in pairs( event_list ) do
    if ( event.type == TYPE_PROGRESS ) then
      return true;
    end;
  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
function public.Crown_DialogContinue_Show ( func_accept, func_reject )
  
  local delegates = { 
    Accept = function () 
      private.Crown_DialogContinue_Hide( func_accept );
    end,
    Reject = function () 
      private.Crown_DialogContinue_Hide( func_reject );
    end
  }
  common.delegate.Set( "int_dialog_minievent_crown_continue", delegates );
  
  local crowns_value = public.Crown_GetCurrentValue();
  
  v_interface.window.Show(
    "int_dialog_minievent_crown_continue",
    function ()
      int_dialog_minievent_crown_continue.Show( crowns_value );
    end
  );
  
end;
--------------------------------------------------------------------------------
function private.Crown_DialogContinue_Hide ( func_end )
  
  int_dialog_minievent_crown_continue.Hide( function ()
    v_interface.window.Hide( "int_dialog_minievent_crown_continue", func_end );
  end );
  
end;
--------------------------------------------------------------------------------
function public.Crown_GetCurrentValue ()
  
  return private.crown_progress_saved;
  
end;
--------------------------------------------------------------------------------
function private.Crown_GetDataChanged ()
  
  local old_progress = private.crown_progress_changed.old;
  local new_progress = private.crown_progress_changed.new;
  
  private.crown_progress_changed.old = private.crown_progress_changed.new;
  
  return new_progress, old_progress;
  
end;
--------------------------------------------------------------------------------
function public.Crown_LevelStart ()
  
  local event_list = m_minievent.GetEventList();
  
  for index, event in pairs( event_list ) do
    if ( event.type == TYPE_PROGRESS ) then
      private.crown_progress_saved = m_minievent.Crown_GetValue();
      m_minievent.Crown_SetValueProgress( event.id, 0 );
      private.Crown_WaitForStageResult( event.type, event.id );
    end;
  end;
  
end;
--------------------------------------------------------------------------------
function public.Crown_GetStatsBonuses ()

  local bonuses = m_minievent.Crown_GetStatsBonuses();
  return bonuses;
  
end;
--------------------------------------------------------------------------------
function private.Crown_HighlightStage ( event_id, func_on_end )
  
  local scale  = 1.5;
  local bounce = true;
  local func_action = function() end;
    
  local section = v_sections.Section_NameCurrent();
  local stages  = v_stage.Stage_GetAllBySection( section, true );

  local object = "";
  local prefix = "btn_map_point_";
  
  for i = 1, #stages do
    
    local info = v_stage.Stage_IsEnabled( stages[i].name );
    if( v_stage.Stage_IsEnabled( stages[i].name ) ) then
      
      object = prefix..stages[i].name;
      
    end;
    
  end;
  
  local func_end = function ()
    if ( func_on_end ) then
      func_on_end();
    end;
    v_quests.Highlight_ShowHole( object, func_action, scale, bounce );
  end;
    
  private.Dialog_Hide( nil, func_end );
  
end;
--------------------------------------------------------------------------------
function private.Crown_WaitForStageResult ( event_type, event_id )

  local func_OnStageWon;
  local func_OnStageFailed;
  local func_OnStageAbandon;

  local func_OnMapUpdate = function ( update )

    mediator.Unsubscribe( "View_Stage_StageFailed",  func_OnStageFailed );
    mediator.Unsubscribe( "View_Stage_StageAbandon", func_OnStageAbandon );
    mediator.Unsubscribe( "View_Stage_StageWon",     func_OnStageWon );
    
    if ( update ) then
      private.OnMapUpdateFinish = function ( func_after )
        private.Dialog_Show( event_id, nil, nil, func_after );
        private.OnMapUpdateFinish = nil;
      end;
    end;

  end;

  func_OnStageFailed = function ( stage_data )

    local update = tern( private.crown_progress_saved == 0 )( false, true );
    
    m_minievent.Crown_SetValueProgress( event_id, 0 );
    
    private.crown_progress_saved = 0;
    private.crown_progress_changed.new = 0;
    
    func_OnMapUpdate( update );

  end;

  func_OnStageAbandon = function ( stage_data )

    local update = tern( private.crown_progress_saved == 0 )( false, true );
    
    m_minievent.Crown_SetValueProgress( event_id, 0 );
    
    private.crown_progress_saved = 0;
    private.crown_progress_changed.new = 0;
  
    func_OnMapUpdate( update );

  end;

  func_OnStageWon = function ( stage_data )
    
    local new_progress = private.crown_progress_saved + 1;
    
    m_minievent.Crown_SetValueProgress( event_id, new_progress );
    
    local update = tern( private.crown_progress_saved == m_minievent.Crown_GetValue() )( false, true );
    
    private.crown_progress_changed.new = m_minievent.Crown_GetValue();
    private.crown_progress_saved = 0;
    
    func_OnMapUpdate( update );

  end;

  mediator.Subscribe( "View_Stage_StageWon",     func_OnStageWon );
  mediator.Subscribe( "View_Stage_StageFailed",  func_OnStageFailed );
  mediator.Subscribe( "View_Stage_StageAbandon", func_OnStageAbandon );

end;
--------------------------------------------------------------------------------