-- name=v_minievent_section
--------------------------------------------------------------------------------

local NAMESPACE   = "minievent_section";
local MODULE      = "v_"..NAMESPACE;
local SCREEN      = "scr_"..NAMESPACE;
local ZOOM_VIEW   = "zmv_"..NAMESPACE;
local MAP         = "anm_"..NAMESPACE.."_map";
local PORTAL_PREV = "obj_"..NAMESPACE.."_portal_prev";
local DIR;

--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
--function public.__InitModule ()
--
--  common.mod.Load( _modules );  
--
--end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Init ( event_id, event_name )

  private.Init_Variables( event_id, event_name );
  private.Init_Objects();
  private.Init_Subscribes();

end;
--------------------------------------------------------------------------------
function private.Init_Variables ( event_id, event_name )

  private.event = {
    id   = event_id,
    name = event_name
  };

  DIR = "assets/content/minievent_"..event_name.."/";

  -- Первый запуск.
  private.is_clean_start = true;

  -- JK: ask for screen dimensions once
  private.app_h = ui.screen.Height();
  private.app_w = ui.screen.Width();
  private.app_r = ui.screen.AspectRatio();

  private.map_scale_current = 1.0;
  private.map_scale_max = 1.0;
  private.map_scale_min = 0.5;

  if ( private.app_h < 960 ) then
    private.map_scale_current = 0.75;
    private.map_scale_max = 0.75;
  end;

  private.scale_modifier = 2;
  private.offset         = 256;
  private.map_w          = 2048;
  private.map_h          = 1280;

  private.create_at_once = 10;

  private.reward          = {};
  private.resource_buffer = {};

  private.popup_shift = { 0, - 100 };

  private.opened = false;

  private.this = {
    background = "obj_"..NAMESPACE.."_background",
    body       = "obj_"..NAMESPACE.."_body",
    marker     = "obj_"..NAMESPACE.."_marker"
  };

  private.worldmap_notification = {
    new_worlds = {}
  };

end;
--------------------------------------------------------------------------------
function private.Init_Objects ()

  local screen_w = private.app_w;
  local screen_h = private.app_h;
  local scroll_h = private.scale_modifier * private.map_h;

  local multiset = {};

  ui.obj.Set(
    ZOOM_VIEW,
    {
      scale         = private.map_scale_current,
      min_scale     = private.map_scale_min,
      max_scale     = private.map_scale_max,
      scrollrect_x  = 0,
      scrollrect_y  = - scroll_h,
      scrollrect_w  = private.scale_modifier * private.map_w,
      scrollrect_h  = scroll_h,
      screen_w      = screen_w,
      screen_h      = screen_h,
      begindragtrig = private.DragStart,
      dragtrig      = private.DragUpdate,
      enddragtrig   = private.DragDrop
    },
    multiset
  );

  GridSet( "grd_"..NAMESPACE.."_foreground_movie_vignette", 0,
  {
    2,  { pos_x = screen_w - 1 },
    3,  { pos_x = screen_w     },
    6,  { pos_x = screen_w - 1 },
    7,  { pos_x = screen_w     },

    8,  { pos_y = screen_h - 4 },
    9,  { pos_y = screen_h - 4 },
    10, { pos_x = screen_w - 1, pos_y = screen_h - 4 },
    11, { pos_x = screen_w,     pos_y = screen_h - 4 },

    12, { pos_y = screen_h },
    13, { pos_y = screen_h },
    14, { pos_x = screen_w - 1, pos_y = screen_h },
    15, { pos_x = screen_w,     pos_y = screen_h }
  } );

  v_screen.Setup(
    SCREEN,
    {
      func_pre_open    = private.PreOpen,
      func_open        = private.Open,
      func_close       = private.Close,
      func_after_close = private.AfterClose
    }
  );

  ui.obj.Attach( SCREEN, "ne_storage" );

  ui.obj.Set(
    private.this.background,
    {
      pos_x = 0,
      pos_y = 0
    },
    multiset
  );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.Init_Subscribes ()

  mediator.Subscribe(
    "Model_Skins_SkinSet",
    private.Marker_SetSkin
  );

  mediator.Subscribe(
    "View_Minievent_SectionFinished",
    private.Section_OnFinished
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Cheat_Update ()

  private.Update();

end;
--------------------------------------------------------------------------------
function public.IsOpened ()

  return private.opened;

end;
--------------------------------------------------------------------------------
function public.IsActive ()

  return ( v_screen.GetCurrentName() == SCREEN );

end;
--------------------------------------------------------------------------------
function public.IsCleanStart ()

  return private.is_clean_start;

end;
--------------------------------------------------------------------------------
function public.IsBeingDragged ()

  return private.is_dragged;

end;

--------------------------------------------------------------------------------
function public.IsPending ()

  return private.pending;

end;
--------------------------------------------------------------------------------
function public.FocusAtCurrentPoint ( func_after )

  if ( private.last_point_selected  ) or
     ( private.marker_point_current ) then

    local target = "obj_"..NAMESPACE.."_point_"..( private.last_point_selected or private.marker_point_current );

    public.Focus( target, { func_after = func_after } );

  elseif ( func_after ) then

    func_after();

  end;

end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function public.Scale_GetMax ()

  return private.map_scale_max;

end;
--------------------------------------------------------------------------------
function public.Scale_GetMin ()

  return private.map_scale_min;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Dialog_SectionDisabled_Show ( func_on_hide )

  v_game.SystemDialog_Show(
    "section_disabled",
    func_on_hide
  );

end;
--------------------------------------------------------------------------------
function private.WorldsButtonClick ()

  if ( v_minievent.Event_IsFinished( private.event.id ) ) then
    v_minievent.Section_Unload();
  else
    v_screen.Switch( "scr_map" );
  end;

  private.Worldmap_CheckNotification( true );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Worldmap_AddNewWorlds ( data )

  private.worldmap_notification.new_worlds = data;

  v_game.SetNotification( "worlds", data );

  private.Worldmap_CheckNotification();

end;
--------------------------------------------------------------------------------
function private.Worldmap_CheckNotification ( clear )

  if ( clear ) then
    if ( #private.worldmap_notification.new_worlds ~= 0 ) then
      v_game.SetNotification( "worlds", false );
    end;
    private.worldmap_notification.new_worlds = {};
  end;

  local count = #private.worldmap_notification.new_worlds;

  local data = {
    target = "worlds",
    data = {
      type = "common",
      entity = "v_map",
      count = count
    }
  };

  v_interface.panel.bottom.Notification( data );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.ViewPort_Contract ( panel_width, tme )

  ui.input.Off( SCREEN );

  ObjAnimate(
    ZOOM_VIEW,
    "screen_w", 0, 0, "",
    {
      0.0, 0, ui.obj.Get( ZOOM_VIEW ).screen_w,
      tme, 0, private.app_w - panel_width
    }
  );

end;
--------------------------------------------------------------------------------
function public.ViewPort_Restore ( tme )

  ObjAnimate(
    ZOOM_VIEW,
    "screen_w", 0, 0,
    function ()

      ui.input.On( SCREEN );

    end,
    {
      0.0, 0, ui.obj.Get( ZOOM_VIEW ).screen_w,
      tme, 0, private.app_w
    }
  );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.PreOpen ( func_after )

  private.is_updating = true;

  --DbgTrace( "[ "..MODULE.."] PreOpen" );

  --mediator.Notify( "View_Map_PreOpen" );

  -- PreOpen end callback
  local function func_preopened ( skip_resource_reload )

    --- First load -------------------------------------------------------------
    if ( private.is_clean_start ) then

      private.Marker_SetAtLastPoint();

    end;

    if ( not skip_resource_reload ) then

      -- Display buffered resources.
      v_interface.Gold_Buffer_Out();

    end;

    --- Intro ------------------------------------------------------------------
    if ( private.Intro_NeedToShow() ) then

      private.Intro_Prepare();

    end;

    func_after();

  end;

  if ( private.is_clean_start ) then

    private.event.progress = v_minievent.Event_GetProgress( private.event.id );
    
    common.RunCallbackList({
      private.Section_SetBackground,
      private.Section_CreateEnvironment,
      function ( func_next )
        private.FPS_StartManager(); func_next();
      end,
      private.Section_CreateObjects,
      func_preopened
    });

  else
    private.FPS_StartManager();
    func_preopened( true );
  end;

end;
--------------------------------------------------------------------------------
function private.Open ( func_after )

  --DbgTrace( "[ "..MODULE.." ] Open" );

  common.sound.Stop( "assets/content/hero/aud_hero_win_ride" );
  common.sound.Stop( "level_completed" );

  if ( not v_sections.Section_NeedIntro() ) then
    private.Music_Start();
  end;

  private.opened = true;

  common.delegate.Set(
    v_interface.panel.left.GetName(),
    {
      Countdown_GetSeconds = function ()
        local  seconds = v_minievent.GetEventRemainingTime( private.event.id );
        return seconds;
      end,
      Countdown_GetLabel = function ()
        local  button_label = "str_int_dialog_minievent_end"
        return button_label;
      end,
      Countdown_GetHint = function ()
        return {
          caption = "str_countdown_minievent_caption_"..private.event.name,
          text    = "str_countdown_minievent_text_"..private.event.name
        };
      end
    }
  );

  v_interface.Scheme_AddButton( NAMESPACE, "left",   { name = "hearts" } );
  v_interface.Scheme_AddButton( NAMESPACE, "left",   { name = "countdown" } );
  v_interface.Scheme_AddButton( NAMESPACE, "bottom", { name = "options" } );
  v_interface.Scheme_AddButton( NAMESPACE, "bottom", { name = "player" } );
  v_interface.Scheme_AddButton( NAMESPACE, "bottom", { name = "shop" } );
  v_interface.Scheme_AddButton( NAMESPACE, "bottom", { name = "worlds" } );
  v_interface.Scheme_AddButton( NAMESPACE, "top",    { name = "gold_only" } );

  common.event.Set(
    v_interface.panel.bottom.GetName(),
    "WorldsClickHandler",
    {
      name = "v_minievent_section",
      handler = function ()
        private.WorldsButtonClick();
        common.sound.Play( "bottom_button" );
      end
    }
  );

  local function func_update ()

    -- JK: map is closing or map is closed -> drop the update
    if ( v_screen.IsClosing()  ) or
       ( not public.IsActive() ) then
      return;
    end;    

    private.Update( func_after );

  end;

  if ( v_interface.window.IsAnyOnScreen() ) then
    mediator.SubscribeOnce( "Interface_Window_NoWindows", func_update );
  else
    func_update();
  end;

end;
--------------------------------------------------------------------------------
function private.Close ( func_after )

  --DbgTrace( "[ "..MODULE.."] Close" );

  v_interface.Hide( { screen = "minievent_section" }, function () end );

  common.event.Delete(
    v_interface.panel.bottom.GetName(),
    "WorldsClickHandler",
    {
      name = "v_minievent_section"
    }
  );
  
  -- JK: cancel any running focus
  ObjStopAnimate( private.this.body, "pos_xy" );

  private.opened = false;
  private.FPS_StopManager();
  private.fps_env_enabled = nil;

  if ( func_after ) then func_after(); end;

end;
--------------------------------------------------------------------------------
function private.AfterClose ( func_after )

  --DbgTrace( "[ "..MODULE.."] AfterClose" );

  private.Music_Stop();

  if ( func_after ) then func_after(); end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.FPS_StartManager ()

  common.Timer_Start( "tmr_fps_manager", 1.0, private.FPS_StartManager );

  if ( common.fps.Get() < 30         ) and 
     ( not ui.screen.IsLocked() ) then
    private.Section_DisableEnvironment();
  else
    private.Section_EnableEnvironment();
  end;

end;

--------------------------------------------------------------------------------
function private.FPS_StopManager ()

  common.Timer_Stop( "tmr_fps_manager" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Section_Create ( func_after )

  if ( not private.is_clean_start ) then return; end;

  ui.obj.Attach( SCREEN, "ng_level" );

  private.PreOpen(
    function ()
      private.Open();
      if ( func_after ) then func_after(); end;
    end
  );

end;
--------------------------------------------------------------------------------
function public.Section_Destroy ( func_after )

  mediator.Unsubscribe(
    "Model_Skins_SkinSet",
    private.Marker_SetSkin
  );

  mediator.Unsubscribe(
    "View_Minievent_SectionFinished",
    private.Section_OnFinished
  );

  ui.obj.Detach( PORTAL_PREV );
  ui.obj.Detach( MAP );

  local children = ui.obj.Children( MAP );

  for _, child in pairs( children ) do
    ui.obj.Delete( child );
  end;

  private.is_clean_start = true;

  private.Close( function () private.AfterClose( func_after ); end );

end;
--------------------------------------------------------------------------------
function private.Section_SetBackground ( func_after )

  ui.obj.Set( "spr_"..NAMESPACE.."_background",
  {
    res   = DIR.."background",
    pos_x = ( private.map_w ) *   0.5,
    pos_y = ( private.map_h ) * - 0.5 
  } );

  func_after();

end;
--------------------------------------------------------------------------------
function private.Section_CreateObjects ( func_after )

  local function func_attach_map ()

    ui.obj.Set(
      MAP,
      {
        res      = DIR.."anm_map_minievent_"..private.event.name,
        animfunc = private.event.name,
        playing  = true,
        pos_x    =  private.offset,
        pos_y    = -private.offset,
        pos_z    = 1000
      }
    );

    ui.obj.Attach( PORTAL_PREV, MAP );
    ui.obj.Attach( MAP, private.this.body );

  end;

  local function func_create_other ( func_after )

    local creation_funcs =
    {
      func_attach_map,
      private.Marker_SetSkin
    };

    local function func_exec ( func, progress )

      func();
      v_interface.Loading_SetProgress( progress, 0.9, 1.00 );

    end;

    common.ProcessArrayByTimer(
      creation_funcs, func_exec, func_after, private.create_at_once
    );

  end;

  common.RunCallbackList({
    private.Section_CreateStages,      -- 00 - 50%
    private.Section_CreateCharacters,  -- 50 - 60%
    private.Section_CreateEnemies,     -- 60 - 70%
    private.Section_CreateRewards,     -- 70 - 80%
    private.Section_CreateQuests,      -- 80 - 90%
    func_create_other,                 -- 90 - 100%
    function () common.Defer( "map_attach_wait", 0.1, func_after ); end
  });

end;
--------------------------------------------------------------------------------
function private.Section_CreateEnvironment ( func_after )

  local environment_list = {};

  -- JK: 5 environment animations per section
  for z = 1, 5 do
    table.insert( environment_list, "environment_"..z );
  end;

  -- Create and attach environment animations
  local function func_create ( env_name, progress )

    local anm_name = NAMESPACE.."_"..env_name;
    local res_name = DIR.."map/anm_"..env_name;

    if ( not ui.obj.Exists( anm_name ) ) then
      ui.obj.Create( anm_name, "anim", MAP );
    end;

    ui.obj.Set(
      anm_name,
      {
       res      = res_name,
       animfunc = "static",
       playing  = true,
       anim_tag = env_name,
       input    = false,
       active   = false,
       visible  = false
      }
    );

  end;

  common.ProcessArrayByTimer(
    environment_list,
    func_create,
    func_after,
    private.create_at_once
  );

end;
--------------------------------------------------------------------------------
function private.Section_DisableEnvironment ()

  if ( private.fps_env_enabled ) then

    for z = 1, 5 do

      local env_name = NAMESPACE.."_environment_"..z;

      ObjAnimate( env_name, "alp", 0, 0,
      function()
        ui.obj.Set( env_name, { active = false, visible = false } );
        ui.obj.Detach( env_name );
        private.fps_env_enabled = nil;
      end,
      {
        0.0,  0,  1,
        0.2,  0,  0
      } );
    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Section_EnableEnvironment ()

  if ( not private.fps_env_enabled ) then

    for z = 1, 5 do

      local env_name = NAMESPACE.."_environment_"..z;

      ui.obj.Attach( env_name, MAP );
      ui.obj.Set( env_name, { active = true, visible = true } );
      ObjAnimate( env_name, "alp", 0, 0,
      function()
        private.fps_env_enabled = true;
      end,
      {
        0.0,  0,  0,
        0.2,  0,  1
      } );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Section_CreateStages ( func_after )

  local multiset           = {};
  local stage_list         = v_minievent.Event_GetStages( private.event.id );
  local active_stage_index, max_index = v_minievent.Event_GetActiveStageIndex( private.event.id );

  -- Prepare default point object
  local multiset = {};

  local res   = DIR.."map/anm_point";
  local point = "anm_"..NAMESPACE.."_point";

  ui.obj.Set( point, { res = res, animfunc = "point_static" }, multiset );
  ui.obj.Set( point.."_bg", { res = res, animfunc = "bg_default" }, multiset );
  ui.obj.Set( point.."_label", { res = res }, multiset );
  ui.obj.Set( point.."_label_number", { res = res }, multiset );
  ui.obj.Set( point.."_label_number_digit_1", { res = res }, multiset );
  ui.obj.Set( point.."_label_number_digit_2", { res = res }, multiset );
  ui.obj.Set( point.."_label_number_digit_3", { res = res }, multiset );
  ui.obj.Set( point.."_state", { res = res }, multiset );
  ui.obj.Set( point.."_reward_icon", { res = res }, multiset );
  ui.obj.Set( point.."_event", { res = res, animfunc = "event" }, multiset );
  ui.obj.Set( "fcs_"..NAMESPACE.."_point", { res = res, animfunc = "bg_default" }, multiset );  

  ui.obj.Multiset( multiset );

  local function func_create_stage ( stage, progress, current_index )

    private.Stage_CreatePoint( current_index, active_stage_index, max_index, multiset );
    v_interface.Loading_SetProgress( progress, 0.0, 0.5 );

  end;

  common.ProcessArrayByTimer(
    stage_list,
    func_create_stage,
    function ()
      ui.obj.Multiset( multiset ); func_after();
    end,
    private.create_at_once
  );

end;
--------------------------------------------------------------------------------
function private.Section_CreateCharacters ( func_after )

  local characters = v_minievent.Event_GetCharacters( private.event.name );

  local function func_create ( character, progress )

    local anim_tag = character.name;
    local speech   = "str_minievent_"..private.event.name..
                     "_"..character.name.."_speech";

    if ( character.stages ) then

      local character_stage_number =
        private.Character_GetStageNumberByProgress(
          character,
          private.event.progress
        );

      anim_tag = character.name.."_"..character_stage_number;

    end;

    local base_name = NAMESPACE.."_character_"..character.name;

    local character_holder = ui.obj.CreateObject(
      base_name, MAP, { anim_tag = anim_tag }
    );    

    local character_btn = "btn_"..base_name;

    ui.obj.Clone(
      "btn_"..NAMESPACE.."_character",
      character_holder,
      "_"..character.name,
      1
    );

    ui.obj.Set(
      character_btn,
      {
        inputrect_init = true,
        inputrect_x    = character.area[ 1 ],
        inputrect_y    = character.area[ 2 ],
        inputrect_w    = character.area[ 3 ],
        inputrect_h    = character.area[ 4 ]
      }
    );

    local character_anm = "anm_"..base_name;

    ui.obj.Set(
      character_anm,
      {
        res      = DIR.."characters/anm_character_"..character.name,
        animfunc = "static",
        playing  = true
      }
    );

    local indicator = "obj_"..NAMESPACE.."_character_indicator_"..character.name;

    local quest = v_minievent.Event_GetQuestByCharacter(
      private.event.name, character.name
    );

    local func_tap = function ( func_end )

      local speech_str  = speech;
      local speech_type = "monolog";
      local speech_data;

      if ( character.stages ) then

        local character_stage_number =
          private.Character_GetStageNumberByProgress(
            character,
            private.event.progress
          );

        speech_str = "str_minievent_"..private.event.name.."_"..character.name..
                     "_"..character_stage_number.."_speech";
      end;

      common.sound.Play( 
        DIR.."characters/aud_"..character.name.."_tap_map", 
        nil, nil, "crystal" );
      
      if ( quest ) then

        local quest_index = common.IsEntityInList(
          quest.stages, private.event.progress
        );

        local is_completed =
          ( private.event.progress >= quest.stages[ #quest.stages ] );

        if ( not is_completed                        ) and
           ( private.active_quests[ character.name ] ) then

          local quest_progress = 0;

          for _, stage_number in pairs( quest.stages ) do
            if ( stage_number <= private.event.progress ) then
              quest_progress = quest_progress + 1;
            end;
          end;

          speech_type = "complex";
          speech_data = {
            icon  = DIR.."items/item_"..quest.item,
            count = quest_progress,
            target = {
              count = #quest.stages
            }
          };

        elseif ( is_completed ) then

          speech_str = "str_int_popup_cloud_characters_completed";

        end;
        
      end;

      ui.anim.Play(
        character_anm,
        "click",
        function ()
          ui.popup.Cloud_Show(
            speech_type,
            indicator,
            speech_str,
            speech_data,
            nil,
            nil,
            { func_close = func_end }
          );
          ui.input.On( character_btn );
        end,
        "static"
      );

    end;

    ui.control.EventsSet(
      character_btn,
      {
        func_on_click = function ( func_end )

          ui.input.Off( character_btn );

          public.Focus(
            character_holder,
            {
              scale      = 1.0,
              offset     = { x = 0, y = 110 },
              func_after = function () func_tap( func_end ); end
            }
          );

        end
      }
    );

    v_interface.Loading_SetProgress( progress, 0.5, 0.6 );

  end;

  common.ProcessArrayByTimer(
    characters, func_create, func_after, private.create_at_once
  );

end;
--------------------------------------------------------------------------------
function private.Section_CreateEnemies ( func_after )

  local enemies = v_minievent.Event_GetEnemies( private.event.name );

  local function func_create ( enemy, progress )

    local anim_tag = enemy.name;
    local speech   = "str_minievent_"..private.event.name..
                     "_"..enemy.name.."_speech";

    local enemy_holder = "obj_"..NAMESPACE.."_enemy_"..enemy.name;
    ui.obj.CreateObject(
      NAMESPACE.."_enemy_"..enemy.name,
      MAP,
      {
        anim_tag = anim_tag,
        inputrect_init = true,
        inputrect_x = enemy.area[ 1 ],
        inputrect_y = enemy.area[ 2 ],
        inputrect_w = enemy.area[ 3 ],
        inputrect_h = enemy.area[ 4 ]
      }
    );
    
    local enemy_btn = "btn_"..NAMESPACE.."_enemy_"..enemy.name;
    ui.obj.Clone(
      "btn_"..NAMESPACE.."_enemy",
      enemy_holder,
      "_"..enemy.name,
      1
    );

    local enemy_anm = "anm_"..NAMESPACE.."_enemy_"..enemy.name;
    local animfunc  = "static";
    local func_tap  = function ()
      
      local indicator = "obj_"..NAMESPACE.."_enemy_indicator_"..enemy.name;
      
      common.sound.Play( DIR.."enemies/aud_"..enemy.name.."_tap_map", nil, nil, "enemy_click" );

      ui.anim.Play(
        enemy_anm,
        "click",
        function ()
          ui.popup.Cloud_Show( "monolog", indicator, speech );
          ui.input.On( enemy_holder );
        end,
        "static"
      );

    end;
    
    local _, stages = v_minievent.Event_GetProgress( private.event.id );
    if ( common.IsEntityInList( stages, enemy.stage ) ) then
      animfunc = "fail_static";
      func_tap = function ()
        common.sound.Play( DIR.."enemies/aud_"..enemy.name.."_tap_map" ); 
        ui.input.On( enemy_holder ); 
      end;
    end;

    ui.obj.Set(
      enemy_anm,
      {
        res      = DIR.."enemies/anm_enemy_"..enemy.name,
        animfunc = animfunc,
        playing  = true
      }
    );

    ui.control.EventsSet(
      enemy_holder,
      {
        func_on_click = function ()

          ui.input.Off( enemy_holder );

          public.Focus(
            enemy_holder,
            {
              scale      = public.Scale_GetMax(),
              offset     = { x = 0, y = 110 },
              func_after = func_tap
            }
          );

        end
      }
    );

    v_interface.Loading_SetProgress( progress, 0.6, 0.7 );

  end;

  common.ProcessArrayByTimer(
    enemies, func_create, func_after, private.create_at_once
  );

end;
--------------------------------------------------------------------------------
function private.Section_CreateRewards ( func_after )

  local area = {
    gift  = { -50, -50, 100, 100 },
    chest = { -40, -90, 100, 100 }
  };

  local rewards  = v_minievent.Event_GetRewards( private.event.id );
  local settings = v_minievent.Event_GetRewardSettings( private.event.name );

  local function func_create ( reward_data, progress, current_index )

    if ( reward_data.stage_number > private.event.progress ) then

      local r = 0;

      local function func_create_reward ( data, item, multiset )

        local number = data.stage_number;

        r = r + 1;

        local postfix = tern(
          ( data.holder_type == "chest" ) and 
          ( #data.rewards > 1           )
        )( "_"..r, "" );

        local base_name = NAMESPACE.."_reward_"..number..postfix;
        local anim_tag  = data.holder_type.."_"..number..postfix;

        local holder = ui.obj.CreateObject(
          base_name, MAP, { anim_tag = anim_tag }
        );
        
        local button = "btn_"..base_name;

        ui.obj.Clone(
          "btn_"..NAMESPACE.."_reward",
          holder,
          "_"..number..postfix,
          1
        );

        ui.obj.Set(
          button,
          {
            inputrect_init = true,
            inputrect_x    = area[ data.holder_type ][ 1 ],
            inputrect_y    = area[ data.holder_type ][ 2 ],
            inputrect_w    = area[ data.holder_type ][ 3 ],
            inputrect_h    = area[ data.holder_type ][ 4 ]
          },
          multiset
        );

        local reward_anm = "anm_"..base_name;

        local func_tap;

        if ( data.holder_type == "gift" ) then

          ui.obj.Set(
            reward_anm,
            {
              res      = "assets/content/map/chests/anm_gift",
              animfunc = "static",
              playing  = true
            },
            multiset
          );

          func_tap = function ()

            v_minievent.Reward_Preview(
              current_index,
              "rewards",
              function () ui.input.On( holder ); end,
              "section",
              data.stage_number
            );

          end;

        elseif ( data.holder_type == "chest" ) then

          local indicator =
            "obj_"..NAMESPACE.."_reward_indicator_"..number..postfix;

          local speech = "str_"..NAMESPACE.."_chest_condition";

          if ( settings[ "stage_"..number ] ) then
            speech = speech.."_help_"..settings[ "stage_"..number ].owner;
          end;

          ui.obj.Set(
            reward_anm,
            {
              res      = "assets/content/map/anm_character_chest",
              animfunc = "static_"..item.name,
              rt_chest = ui.obj.Res( "item", item.name ),
              playing  = true
            },
            multiset
          );

          func_tap = function ()

            ui.anim.Play(
              reward_anm,
              "click_"..item.name,
              function ()

                ui.popup.Cloud_Show(
                  "monolog",
                  indicator,
                  speech
                );

                ui.input.On( holder );
              end,
              "static_"..item.name
            );

          end;

        end;

        ui.control.EventsSet(
          button,
          {
            func_on_click = function ()

              ui.input.Off( holder );

              public.Focus(
                holder,
                {
                  scale      = 1.0,
                  func_after = func_tap
                }
              );

            end
          }
        );

        if ( settings[ "stage_"..number ]                                 ) and
           ( settings[ "stage_"..number ].appear                          ) and
           ( settings[ "stage_"..number ].appear > private.event.progress ) then
          ui.obj.Disable( holder, false, multiset );
        end;

      end;

      local multiset = {};

      if ( reward_data.holder_type == "chest" ) then
        for _, reward in pairs( reward_data.rewards ) do
          func_create_reward( reward_data, reward, multiset );
        end;
      else
        func_create_reward( reward_data, nil, multiset );
      end;

      ui.obj.Multiset( multiset );

    end;

    v_interface.Loading_SetProgress( progress, 0.7, 0.8 );

  end;

  common.ProcessArrayByTimer(
    rewards, func_create, func_after, private.create_at_once
  );

end;
--------------------------------------------------------------------------------
function private.Section_CreateQuests ( func_after )

  local quests = v_minievent.Event_GetQuests( private.event.name );

  if ( type( quests ) ~= "table" ) then return; end;

  private.active_quests = {};

  for _, quest in pairs( quests ) do
    if ( quest.start <= private.event.progress + 1               ) and
       ( quest.stages[ #quest.stages ] >= private.event.progress ) then
      private.active_quests[ quest.giver ] = true;
    end;
  end;

  local function func_create ( quest, progress, current_index )

    local multiset = {};
    local step     = 1;
    local max_step = #quest.stages;

    for index, stage_number in pairs( quest.stages ) do

      if ( stage_number >= private.event.progress + 1 ) and
         ( private.active_quests[ quest.giver ]       ) then

        local base_name = NAMESPACE.."_point_reward";

        ui.obj.Set(
          "anm_"..base_name.."_stage_"..stage_number,
          {
            animfunc = "static"
          },
          multiset
        );

        ui.obj.Set(
          "anm_"..base_name.."_icon_stage_"..stage_number,
          {
            animfunc = "reward_"..quest.item
          },
          multiset
        );

        ui.obj.Enable(
          "obj_"..base_name.."_stage_"..stage_number, true, multiset
        );

      elseif ( stage_number <= private.event.progress ) then

        step = step + 1;

      end;

    end;

    if ( #multiset > 0 ) then
      ui.obj.Multiset( multiset );
    end;

    local animfunc = tern( step <= max_step )(
      "step_"..step,
      "step_finished"
    );

    local playing = ( step > max_step );

    step = math.min( step, max_step );

    local activity_anm = ui.obj.CreateAnimation(
      NAMESPACE.."_activity_"..quest.giver,
      MAP,
      {
        res      = DIR.."activities/anm_"..quest.giver.."_activity",
        animfunc = animfunc,
        anim_tag = quest.giver.."_activity",
        frame    = 0,
        playing  = playing
      }
    );

    v_interface.Loading_SetProgress( progress, 0.8, 0.9 );

  end;

  common.ProcessArrayByTimer(
    quests, func_create, func_after, private.create_at_once
  );

end;
--------------------------------------------------------------------------------
function private.Section_OnFinished ( params )

  v_interface.panel.left.DisableButton( "countdown" );

  if ( params.is_completed ) then return; end;

  local func_dialog = function ()
    v_game.SystemDialog_Show( 
      "event_finished",
      function ()
        v_minievent.Section_Unload();
      end
    );
  end;

  if ( v_screen.GetCurrentName() == SCREEN ) then
    if ( private.is_updating ) then
      mediator.SubscribeOnce(
        "View_Minievent_Section_Update_Finished",
        func_dialog
      );
    else
      func_dialog();
    end;
  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Character_GetStageNumberByProgress ( character, progress )

  local character_stage_number = character.stages[ 1 ];

  for _, stage_number in pairs( character.stages ) do
    if ( stage_number <= ( progress + 1 ) ) then
      character_stage_number = stage_number;
    end;
  end;

  return character_stage_number;

end;
--------------------------------------------------------------------------------
function private.Character_TryRelocate ( character_name, stage_number, func_after )

  local anim_tag         = character_name.."_"..stage_number;
  local character_holder = "obj_"..NAMESPACE.."_character_"..character_name;

  if ( ui.obj.GetAnimTag( character_holder ) ~= anim_tag ) then

    public.Focus(
      character_holder,
      {
        scale      = public.Scale_GetMin(),
        func_after = function ()

          ui.anim.FadeOut(
            0.3, character_holder,
            function()
              ui.obj.SetAnimTag( character_holder, anim_tag );

              public.Focus(
                character_holder,
                {
                  scale      = public.Scale_GetMax(),
                  func_after = function ()
                    ui.anim.FadeIn( 0.3, character_holder, false, func_after );
                  end
                }
              );

            end
          );

        end
      }
    );    

  elseif ( func_after ) then func_after(); end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Marker_SetSkin ( skin_name )

  skin_name = skin_name or v_player.Skin_GetCurrent();

  ui.obj.Set(
    "anm_"..NAMESPACE.."_marker_icon_skin",
    {
      res      = ui.obj.Res( "skin", skin_name ),
      animfunc = "marker",
      playing  = true
    }
  );

end;
--------------------------------------------------------------------------------
function private.Marker_SetAtLastPoint ()

  local active, max = v_minievent.Event_GetActiveStageIndex( private.event.id );
  local stage_name = "stage_"..( active or max );

  private.Marker_SetAtPoint( stage_name, true );

end;
--------------------------------------------------------------------------------
function private.Marker_Target ( stage_name, ignore_stage_info )

  local target_point = "obj_"..NAMESPACE.."_point_"..stage_name;

  if ( ignore_stage_info ) then

    return target_point, stage_name;

  end;

  local stage_info = v_stage.Stage_GetInfo( stage_name );

  if ( stage_info.completed ) then

    if ( stage_info.realtype ~= "misc"                                ) or
       ( stage_info.realtype == "misc" and stage_info.type ~= "fight" ) then

      stage_name   = v_stage.Stage_NameNext( stage_name );
      target_point = "obj_"..NAMESPACE.."_point_"..stage_name;

    end;

  end;

  return target_point, stage_name;

end;
--------------------------------------------------------------------------------
function private.Marker_SetAtPoint ( stage_name, is_visible )

  local multiset = {};
  local point    = "obj_"..NAMESPACE.."_point_"..stage_name;

  private.marker_point_current = stage_name;

  public.Focus( point, { instant = true } );

  local point_pos = ui.obj.Position( point, MAP );

  ui.obj.Set(
    private.this.marker,
    {
      pos_x   = point_pos[ 1 ] + private.offset,
      pos_y   = point_pos[ 2 ] - private.offset,
      visible = is_visible
    },
    multiset
  );

  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.Marker_MoveToPoint ( stage_name, execute_tap, no_focus, func_end )

  if ( not no_focus ) then

    public.Focus( private.Marker_Target( stage_name ), { time = 0.5 } );

  end;

  private.marker_point_current  = stage_name;
  private.marker_target_execute = execute_tap;

  local target_object = "obj_"..NAMESPACE.."_point_"..stage_name;
  local target_params = ui.obj.Position( target_object, MAP );

  local button = "btn_"..ui.obj.GetNameWithoutPrefix( target_object );
  local events = ui.control.EventsGet( button );

  local func_tap = function () end;
  if ( events ) then func_tap = events.func_on_click; end;

  ObjSet( target_object, { input = true } );

  private.marker_target_params =
  {
    pos_x    = target_params[ 1 ],
    pos_y    = target_params[ 2 ],
    func_tap = func_tap,
    name     = target_object
  };

  private.Marker_MoveToPoint_Animate( func_end );

end;
--------------------------------------------------------------------------------
function private.Marker_MoveToPoint_Animate ( func_end )

  local func_OnEnd = function ()

    private.Marker_MoveToPoint_Done( func_end );

  end;

  local marker_object = private.this.marker;

  ObjStopAnimate( marker_object, "pos_xy" );

  local marker_params = ObjGet( marker_object );
  local target_params = private.marker_target_params;

  ObjAnimate( marker_object, "pos_xy", 0, 0, func_OnEnd,
  {
    0.0, 3, marker_params.pos_x, marker_params.pos_y,
    0.2, 3, marker_params.pos_x, marker_params.pos_y,
    0.5, 3, target_params.pos_x + private.offset, target_params.pos_y - private.offset
  } );

end;
--------------------------------------------------------------------------------
function private.Marker_MoveToPoint_Done ( func_end )

  -- Turned off: GD's request.
  -- PartSysRestart( "pfx_map_marker_done" );

  common.sound.Play( "open_new_level" );

  if ( func_end ) then

    func_end();

  end;

  if ( private.marker_target_execute ) then

    private.marker_target_params.func_tap();

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.DragStart ()

  private.is_dragged = true;

  ui.control.TapBreak();
  ui.popup.Cloud_Hide();
  ui.button.TryReleasePressed();

end;
--------------------------------------------------------------------------------
function private.DragUpdate ()

  --

end;
--------------------------------------------------------------------------------
function private.DragDrop ()

  --private.map_scale_current = ui.obj.Get( private.this.body ).scale_x;

  private.is_dragged = nil;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Stage_CreatePoint ( stage_number, active_stage_index, max_index, multiset )

  local stage_name = "stage_"..stage_number;
  local point_name = "obj_"..NAMESPACE.."_point_"..stage_name;
  local btn_name   = "btn_"..NAMESPACE.."_point_"..stage_name;
  local anm_bg     = "anm_"..NAMESPACE.."_point_bg_"..stage_name;
  local anm_state  = "anm_"..NAMESPACE.."_point_state_"..stage_name;

  local reward, group, holder =
    v_minievent.Event_GetSectionStageRewards( private.event.id, stage_number );

  ui.obj.Create( point_name, "obj" );
  ui.obj.SetAnimTag( point_name, stage_name, multiset );
  ui.obj.Attach( point_name, MAP );
  ui.obj.Clone( "btn_"..NAMESPACE.."_point", point_name, "_"..stage_name, 1 );
  ui.obj.Enable( "btn_"..NAMESPACE.."_point_"..stage_name );

  local is_gift_point = ( reward and holder == "gift" );
  local is_completed  = ( stage_number < ( active_stage_index or max_index ) );
  is_completed  = ( not active_stage_index ) and true or is_completed;
  active_stage_index = active_stage_index or max_index;

  if ( stage_number <= active_stage_index ) or
     ( is_gift_point                      ) then
    private.Stage_SetPointNumber( stage_number, multiset );
  end;

  local animfunc = "bg_default";

  if ( is_completed ) then
    animfunc = "bg_completed";
    ui.obj.Set( anm_state, { animfunc = "state_3" },      multiset );
  end;

  if ( stage_number == active_stage_index and not is_completed ) then
    local funcs =
      private.Stage_GetPointEventHandlers( stage_number, is_gift_point );
    ui.control.EventsSet( btn_name, funcs );
    animfunc = "bg_active";
  end;

  if ( is_gift_point    ) and
     ( not is_completed ) then
    animfunc = "bg_reward";
  end;

  ui.obj.Set( anm_bg, { animfunc = animfunc }, multiset );

end;
--------------------------------------------------------------------------------
function private.Stage_GetPointEventHandlers ( stage_number, is_gift_point )

  local stage_name = "stage_"..stage_number;

  local funcs =
  {
    func_on_click = function () private.Stage_OnPointTap( stage_number ); end
  };  

  if ( IsPC()            ) and
     ( not is_gift_point ) then
    funcs.func_on_mouse_enter = function ()
      private.Stage_ShowPointFocus( stage_name );
    end;
    funcs.func_on_mouse_leave = function ()
      private.Stage_HidePointFocus( stage_name );
    end;
  end;

  return funcs;

end;
--------------------------------------------------------------------------------
function private.Stage_SetPointNumber ( stage_number, multiset )

  local digits     = common.math.NumberToDigits( stage_number );
  local stage_name = "stage_"..stage_number;
  local anm_label  = "anm_"..NAMESPACE.."_point_label_"..stage_name;
  local anm_number = "anm_"..NAMESPACE.."_point_label_number_"..stage_name;

  ui.obj.Set( anm_label,  { animfunc = "label_number" },         multiset );
  ui.obj.Set( anm_number, { animfunc = "number_"..( #digits ) }, multiset );

  for i = 1, #digits do

    ui.obj.Set(
      "anm_"..NAMESPACE.."_point_label_number_digit_"..i.."_"..stage_name,
      {
        animfunc = "digit_"..digits[ i ]
      },
      multiset
    );

  end;

end;
--------------------------------------------------------------------------------
function private.Stage_SetPointCompleted ( stage_number )

  local multiset = {};

  local stage_name = "stage_"..stage_number;
  local point_name = "obj_"..NAMESPACE.."_point_"..stage_name;
  local btn_name   = "btn_"..NAMESPACE.."_point_"..stage_name;
  local anm_bg     = "anm_"..NAMESPACE.."_point_bg_"..stage_name;
  local anm_state  = "anm_"..NAMESPACE.."_point_state_"..stage_name;

  ui.obj.Set( anm_bg,    { animfunc = "bg_completed" }, multiset );
  ui.obj.Set( anm_state, { animfunc = "state_3" },      multiset );

  ui.obj.Multiset( multiset );

  ui.control.EventsClear( btn_name );

end;
--------------------------------------------------------------------------------
function private.Stage_SetPointEnabled ( stage_number )

  local stage_name = "stage_"..stage_number;
  local point_name = "obj_"..NAMESPACE.."_point_"..stage_name;
  local btn_name   = "btn_"..NAMESPACE.."_point_"..stage_name;
  local anm_bg     = "anm_"..NAMESPACE.."_point_bg_"..stage_name;
  local anm_state  = "anm_"..NAMESPACE.."_point_state_"..stage_name;
  local anm_label  = "anm_"..NAMESPACE.."_point_label_"..stage_name;

  local is_gift_point = ( ui.obj.Get( anm_bg ).animfunc == "bg_reward" );

  local funcs = private.Stage_GetPointEventHandlers(
    stage_number, is_gift_point
  );
  ui.control.EventsSet( btn_name, funcs );

  if ( is_gift_point ) then return; end;

  local multiset = {};

  private.Stage_SetPointNumber( stage_number, multiset );

  ui.obj.Set( anm_bg, { animfunc = "bg_active" }, multiset );

  ui.anim.FadeIn( 0.6, anm_label, true );
  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.Stage_AnimatePointChange (
    anim_type,
    stage_number,
    func_change,
    func_after
  )

  local handler = {
    rotate = ui.anim.Rotate,
    bounce = ui.anim.Bounce
  }

  local point_name = "obj_"..NAMESPACE.."_point_stage_"..stage_number;
  handler[ anim_type ]( point_name, func_after, func_change );

end;
--------------------------------------------------------------------------------
function private.Stage_OnPointTap ( stage_number, skip_focus )

  local stage_name = "stage_"..stage_number;

  local func_before;

  if ( not skip_focus ) then
    func_before = function ( func_next )
      -- JK: map scale normalization
      public.Focus(
        "obj_"..NAMESPACE.."_point_"..stage_name,
        {
          scale      = 1.0,
          func_after = func_next
        }
      );
    end;
  end;

  local is_started, reason = v_minievent.DialogStart_Show(
    stage_number, func_before, private.cheat_force
  );

  if ( is_started ) then
    private.last_point_selected = stage_name;
  elseif ( reason == "suspended" ) then
    v_game.SystemDialog_Show( 
      "connectionlost",
      function () end
    );
  elseif ( reason == "finished" ) then
    v_game.SystemDialog_Show( 
      "event_finished",
      function ()
        v_minievent.Section_Unload();
      end
    );
  end;

end;
--------------------------------------------------------------------------------
function private.Stage_ShowPointFocus ( stage_name )

  local focus_name = string.format( "fcs_"..NAMESPACE.."_point_%s", stage_name );
  local alp = ObjGet( focus_name ).alp;

  ObjSet( focus_name, { active = true, visible = true } );
  ObjAnimate( focus_name, "alp", 0, 0, "",
  {
    0.0, 2, alp,
    0.3, 2, 1
  } );

end;
--------------------------------------------------------------------------------
function private.Stage_HidePointFocus ( stage_name )

  local focus_name = string.format( "fcs_"..NAMESPACE.."_point_%s", stage_name );
  local alp = ObjGet( focus_name ).alp;

  ObjAnimate( focus_name, "alp", 0, 0,
  function()
    ObjSet( focus_name, { active = false, visible = false } );
  end,
  {
    0.0, 2, alp,
    0.3, 2, 0
  } );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Pending_Start ( pending_token, func_exec )

  private.Pending_Start( pending_token, func_exec );

end;
--------------------------------------------------------------------------------
function public.Pending_Complete ( pending_token )

  private.Pending_Complete( pending_token );

end;
--------------------------------------------------------------------------------
function private.Pending_Start ( pending_token, func_exec )

  -- DbgTrace( "[ VIEW ][ MAP ] Pending started: "..pending_token );

  private.pending = pending_token;

  if ( func_exec ) then

    private.Pending_AddOnComplete( pending_token, func_exec );

  end;

end;
--------------------------------------------------------------------------------
function private.Pending_AddOnComplete ( pending_token, func_exec )

  if not ( func_exec ) then

    return;

  end;

  if not ( private.pending_oncomplete ) then

    private.pending_oncomplete = {};

  end;

  if not ( private.pending_oncomplete[ pending_token ] ) then

    private.pending_oncomplete[ pending_token ] = {};

  end;

  table.insert( private.pending_oncomplete[ pending_token ], func_exec );

end;
--------------------------------------------------------------------------------
function private.Pending_Complete ( pending_token )

  -- DbgTrace( "[ VIEW ][ MAP ] Pending complete: "..pending_token );

  private.pending = nil;

  if ( not private.pending_oncomplete                   ) or
     ( not private.pending_oncomplete[ pending_token ]  ) or
     ( #private.pending_oncomplete[ pending_token ] < 1 ) then

     -- DbgTrace( "[ VIEW ][ MAP ] Pending complete: nothing executed" );

    return;

  end;

  for f = 1, #private.pending_oncomplete[ pending_token ] do

    private.pending_oncomplete[ pending_token ][ f ]();

  end;

  private.pending_oncomplete[ pending_token ] = nil;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Intro_NeedToShow ()

  return not v_minievent.Event_GetIntro( private.event.name ).showed;

end;
--------------------------------------------------------------------------------
function private.Intro_Prepare ()

  local multiset = {};

  ui.obj.Set(
    private.this.marker,
    {
      alp = 0
    },
    multiset
  );
  
  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
function private.Intro_Show ( func_OnEnd )

  private.intro_active = true;

  local intro_settings = v_minievent.Event_GetIntro( private.event.name );
  local full_name = "minievent_"..private.event.name;

  ui.screen.Lock( true, "Intro_Show" );
  v_interface.Hide();

  v_minievent.Event_SetIntro( private.event.name, true );

  local objects = {};

  local function func_end ()
    
    ObjAttach( "obj_minievent_section_foreground_movie", "obj_minievent_section_foreground" );
    ObjAttach( "obj_map_spawn", "hub_map_anims" );

    ObjSet( "obj_minievent_section_foreground_movie", { active = false, visible = false } );
    ObjSet( "obj_map_spawn",            { active = false, visible = false } );

    local func_anim_end = function ()

      private.intro_active = false;
      ui.screen.Lock( false );
      if ( func_OnEnd ) then func_OnEnd(); end;

    end;

    ObjAnimate( private.this.marker, "alp", 0, 0, func_anim_end,
    {
      0.0, 0, 0,
      0.3, 0, 1
    } );
    
  end;
  
  ObjAttach( "obj_map_spawn", "obj_minievent_section_anims" );
  ObjAttach( "obj_minievent_section_foreground_movie", InterfaceName );

  table.insert( objects, { "obj_minievent_section_foreground_movie",
  {
    active  = true,
    visible = true
  } } );

  table.insert( objects, { "anm_minievent_section_foreground_movie",
  {
    active  = false,
    visible = false,
    res     = "assets/content/"..full_name.."/movie/anm_intro",
  } } );

  table.insert( objects, { "obj_map_spawn",
  {
    active  = true,
    visible = true,
  } } );

  table.insert( objects, { "anm_map_spawn_hero",
  {
    active  = false,
    visible = false,
    res     = "assets/content/"..full_name.."/movie/anm_intro"
  } } );

  ObjMultiSet( objects );

  private.Intro_Play(
    full_name,
    1,
    intro_settings,
    func_end
  );

end;
--------------------------------------------------------------------------------
function private.Intro_Play (
    section_name,
    intro_step,
    intro_settings,
    func_end
  )

  local intro_steps;

  if ( intro_settings.type == "default" ) then

    local step_1_obj = "anm_map_spawn_hero";
    local step_5_obj = "anm_map_spawn_hero";

    intro_steps = {
      {
        func   = private.Intro_Animate,
        object = step_1_obj,
        params = {
          animfunc_start = "intro_1"
        }
      },
      {
        func   = private.Intro_Animate,
        object = step_5_obj,
        params = {
          animfunc_start = "intro_2"
        }
      }
    };

  end;

  if ( intro_settings.type ) then

    if ( intro_step == 1 ) then

      private.Music_Stop();
      common.sound.Play( "assets/content/"..section_name.."/movie/aud_intro" );

    end;

    if ( intro_steps[ intro_step ] ) then

      intro_steps[ intro_step ].func(
        intro_step,
        intro_steps[ intro_step ].object,
        intro_steps[ intro_step ].params,
        function ()

          private.Intro_Play(
            section_name,
            ( intro_step + 1 ),
            intro_settings,
            func_end
          );

        end
      );

    else

      private.Music_Start();
      func_end();

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Intro_Animate ( intro_step, object_name, params, func_end )

  if not ( object_name ) then

    object_name = "anm_map_foreground_movie";

  end;

  --DbgTrace( "Intro_Animate: "..intro_step..", "..object_name );

  local animfunc_start = "intro_"..intro_step;
  local animfunc_end = nil;
  local animfunc_dummy = nil;

  if ( params ) then

    if ( params.animfunc_start ) then

      animfunc_start = params.animfunc_start;

    end;

    if ( params.animfunc_dummy ) then

      animfunc_dummy = params.animfunc_dummy;

    end;

    if ( params.animfunc_end ) then

      animfunc_end = params.animfunc_end;

    end;

  end;

  local func_end_anim = func_end;

  if ( animfunc_end ) then

    func_end_anim = function ()

      ui.anim.Play( object_name, animfunc_end );

      if not ( animfunc_dummy ) then

        func_end();

      end;

    end;

  end;

  if ( animfunc_dummy ) then

    ObjClone( object_name, "_dummy", 1, ObjGetRelations( object_name ).parent );

    ui.anim.Play(
      object_name.."_dummy",
      animfunc_dummy,
      function ()

        ObjDelete( object_name.."_dummy" );

        func_end();

      end
    );

  end;

  ObjSet( object_name, { active = true, visible = true } );

  ui.anim.Play( object_name, animfunc_start, func_end_anim );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Sound_Play ( sound_id )

  local path = DIR.."audio/aud_"..sound_id;

   common.sound.Play( path );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Music_Start ()

  if ( not private.music ) then
    common.music.Play(
      DIR.."audio/aud_minievent_"..private.event.name.."_mus", nil, nil, true
    );
    private.music = true;
  end;

end;
--------------------------------------------------------------------------------
function private.Music_Stop ()

  if ( private.music ) then
    common.music.Stop(
      DIR.."audio/aud_minievent_"..private.event.name.."_mus"
    );
    private.music = false;
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Focus ( obj, params )

  -- common.DbgCallpoint( obj );

  params = params or {};

  -- NOTE: all parameters are optional:
  -- time       - focus time, default = ( distance / map_w )^( 1/3 )
  -- speed      - focus speed factor, default = 1
  -- instant    - instant focus flag, default = false
  -- scale      - new scale, default = current scale
  -- func_after - after focus callback
  -- offset     - target position offset, default = { x = 0, y = 0 }

  -- Warn about breaking solid focus
  if ( IsEditor() and private.focusing_solid ) then

    local msg = "Warning! Solid focusing at '%s' breaked by '%s'";
    common.DbgCallpoint( msg:format( private.focusing_solid, obj ) );

  end;

  local func_after = params.func_after;

  private.focusing_solid = tern( func_after )( obj, false );

  local function func_focus_end ()

    if ( v_interface.IsVisible() ) then
      v_interface.Input_On();
    end

    ui.input.On( ZOOM_VIEW );

    if ( func_after ) then
      private.focusing_solid = false;
      func_after();
    end;

  end;

  -- Stop map animations
  ZoomViewStopVelocity( ZOOM_VIEW );
  ObjStopAnimate( private.this.body, "pos_xy" );
  ObjStopAnimate( private.this.body, "scale_xy" );

  -- Define map state changes
  local scale_now = private.Focus_ScaleCurrent();
  local scale_new = params.scale or scale_now;

  scale_new = math.min( scale_new, private.map_scale_max );
  scale_new = math.max( scale_new, private.map_scale_min );

  local scale_factor = scale_now / scale_new;

  local pos_now  = private.Focus_PositionCurrent();
  local offset   = params.offset;
  local pos_new  = private.Focus_PositionTarget( obj, offset, scale_new );

  local distance = private.Focus_DistanceTo( pos_now, pos_new );
  local speed    = params.speed or 1;
  local time_opt = ( distance / private.map_w )^( 1/3 ) * scale_factor / speed;
  local time     = params.time or math.max( time_opt, 0.2 );

  local hold     = ( distance < 30 ) and not params.time;
  local instant  = params.instant or ( time < 1 / 25 );

  -- local tr = "\tp = %.2f,%.2f -> %.2f,%.2f; d = %.2f; "..
             -- "s = %.2f -> %.2f; t = %.2f; i = %s; h = %s";
  -- DbgTrace( tr:format(
    -- pos_now.x, pos_now.y, pos_new.x, pos_new.y, distance,
    -- scale_now, scale_new, time, instant, hold
  -- ) );

  if ( hold ) then

    pos_new = pos_now;
    instant = true;

  end;

  if ( instant ) then

    local map_params =
    {
      pos_x = pos_new.x,
      pos_y = pos_new.y
    };

    if ( scale_now ~= scale_new ) then

      map_params.scale_x = scale_new;
      map_params.scale_y = scale_new;

    end;

    ObjSet( private.this.body, map_params );
    func_focus_end();

  else

    v_interface.Input_Off();
    ui.input.Off( ZOOM_VIEW );

    if ( scale_now ~= scale_new ) then

      ObjAnimate( private.this.body, "scale_xy", 0, 0, "",
      {
        0.00, 3, scale_now, scale_now,
        time, 3, scale_new, scale_new
      } );

    end;

    ObjAnimate( private.this.body, "pos_xy", 0, 0, func_focus_end,
    {
      0.00, 3, pos_now.x, pos_now.y,
      time, 3, pos_new.x, pos_new.y
    } );

  end;

end;
--------------------------------------------------------------------------------
function private.Focus_ScaleCurrent ()

  local params = ObjGet( private.this.body ) or {};
  return params.scale_x;

end;
--------------------------------------------------------------------------------
function private.Focus_PositionCurrent ()

  local params = ObjGet( private.this.body ) or {};
  return { x = params.pos_x or 0, y = params.pos_y or 0 };

end;
--------------------------------------------------------------------------------
function private.Focus_PositionTarget ( obj, offset, scale_new )

  -- Get raw position
  local pos_raw = ui.obj.Position( obj, private.this.body );
  local pos = { x = pos_raw[ 1 ] or 0, y = pos_raw[ 2 ] or 0 };

  -- Get zoomview params
  local params = ui.obj.Get( ZOOM_VIEW );
  local scr_x, scr_y = params.screen_x, params.screen_y;
  local scr_w, scr_h = params.screen_w, params.screen_h;

  -- Shift by offset and move to screen center
  offset = offset or { x = 0, y = 0 };
  local center = { x = scr_w / 2 + scr_x, y = scr_h / 2 + scr_y };

  -- DbgTrace( ("Focus center: %s, %s"):format( center.x, center.y ) );

  pos.x = center.x + offset.x - pos.x * scale_new;
  pos.y = center.y + offset.y - pos.y * scale_new;

  -- Check bounds
  local map_s        = private.scale_modifier;
  local map_h, map_w = private.map_h, private.map_w;
  -- local app_h, app_w = private.app_h, private.app_w;

  local x_min, x_max = scr_w - map_s * map_w * scale_new, 0;
  local y_min, y_max = scr_h,  map_s * map_h * scale_new;

  pos.x = math.min( pos.x, x_max );
  pos.x = math.max( pos.x, x_min );

  pos.y = math.min( pos.y, y_max );
  pos.y = math.max( pos.y, y_min );

  return pos;

end;
--------------------------------------------------------------------------------
function private.Focus_DistanceTo ( a, b )

  local dx, dy = b.x - a.x, b.y - a.y;
  return math.sqrt( dx^2 + dy^2 );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Update ( func_after )

  ui.screen.Lock( true );

  local update = {};

  local event_id   = private.event.id;
  local event_data = v_minievent.Event_GetData( event_id );

  private.event.progress = private.event.progress or 0;
  local progress         = v_minievent.Event_GetProgress( event_id );
  
  local function func_update_finish ()

    if ( event_data.is_completed ) then

      local function func_complete_event ()
        v_game.SystemDialog_Show(
          "event_completed",
          function ()
            v_minievent.Section_Unload();
          end
        );
      end;

      table.insert( update, func_complete_event );

    else

      private.Update_Finish( update );
      
      local function func_drop_clean_start ( func_after )
        if ( private.is_clean_start ) then
          private.is_clean_start = false;
        end;
        if ( func_after ) then func_after(); end;
      end;

      table.insert( update, func_drop_clean_start );
      table.insert( update, func_after );

    end;

    common.RunCallbackList( update );
    
  end;

  local stages = v_minievent.Event_GetStages( event_id );

  if ( private.last_point_selected and 
    private.last_point_selected ~= stages[ math.min( private.event.progress + 1, #stages ) ].name:match( "(stage_%d+).*" ) ) then
    func_update_finish();
    return;
  end;
  
  local progress_changed = false;
  if ( private.event.progress < progress ) then
    private.event.progress = progress;
    progress_changed = true;
  end;
  
  local active_stage_index = v_minievent.Event_GetActiveStageIndex( event_id );
  
  private.Update_Start( update );

  if ( private.is_clean_start ) then
    private.Update_TryShowIntro( update );
  end;

  if ( not private.is_clean_start and progress_changed ) then

    private.Update_FocusAtCompletedPoint( update, progress );

    if ( event_data.is_finished      ) and
       ( not event_data.is_completed ) then
      private.Update_ShowEventIsFinished( update );
    end;

    if ( event_data.is_completed ) then
      private.Update_TryCompleteBoss( update, progress );
      private.Update_TryShowCharacterPanel( update, event_data, progress, "win" );
    end;

    private.Update_TryApplyQuestItem( update, progress );
    private.Update_TryFinishQuests( update, progress );
    private.Update_TryShowStageReward( update, event_data, progress );

    if ( not event_data.is_completed ) then
      private.Update_TryShowCharacterPanel( update, event_data, progress, "win" );
      private.Update_TryRelocateCharacters( update, event_data, progress );
    end;

    if ( active_stage_index ) then
      private.Update_FocusAtEnabledPoint( update, active_stage_index );
    end;

  end;

  if ( active_stage_index ) then

    private.Update_TryShowCharacterPanel(
      update, event_data, active_stage_index, "start"
    );

    if ( not private.is_clean_start and progress_changed ) then
      private.Update_TryStartQuests( update, active_stage_index );
      private.Update_TryPlaceRewardsAtStages( update, active_stage_index );
    end;

  end;

  if ( event_data.is_completed ) then
    private.Update_ShowShareDialog( update );
  end;

  func_update_finish();

end;
--------------------------------------------------------------------------------
function private.Callback_NotifyForPending ( notification, params )

  return function ( func_after )

    mediator.Notify( notification, params );

    if ( func_after ) then func_after(); end;

  end;

end;
--------------------------------------------------------------------------------
function private.Update_TryAddStep ( update, func_step, notification, params )

  if ( func_step ) then

    local func_notify =
      private.Callback_NotifyForPending( notification, params );

    -- DbgTrace( "Add to update: " .. notification, params );

    table.insert( update, func_step );
    table.insert( update, func_notify );

  end;

end;
--------------------------------------------------------------------------------
function private.Update_Start ( update )

  local func_update = function ( func_after ) func_after(); end;

  local notification = "View_Minievent_Section_Update_Started";
  private.Update_TryAddStep( update, func_update, notification );

end;
--------------------------------------------------------------------------------
function private.Update_TryShowIntro ( update )

  local func_intro = ( private.Intro_NeedToShow() ) 
                       and private.Intro_Show
                       or nil;

  local notification = "View_Minievent_Section_Update_IntroDisplayed";
  private.Update_TryAddStep( update, func_intro, notification );

end;
--------------------------------------------------------------------------------
function private.Update_FocusAtCompletedPoint ( update, stage_number )

  local function func_focus ( func_after )

    public.Focus( "obj_"..NAMESPACE.."_point_stage_"..stage_number,
      {
        time       = 0.3,
        func_after = function ()

          private.Stage_AnimatePointChange(
            "rotate",
            stage_number,
            private.Stage_SetPointCompleted( stage_number ),
            func_after
          );

        end
      }
    );

  end;

  local notification = "View_Minievent_Section_Update_CompletedStageFocused";
  private.Update_TryAddStep( update, func_focus, notification, stage_number );

end;
--------------------------------------------------------------------------------
function private.Update_TryCompleteBoss ( update, stage_number )
  
  local completion = {};
  local enemies = v_minievent.Event_GetEnemies( private.event.name );
  local stages = v_minievent.Event_GetStages( private.event.id );
  
  for i = 1, #enemies do
    if ( enemies[ i ].stage == stages[ stage_number ].name ) then
      local function func_anim ( func_after )
        ui.anim.Play( "anm_minievent_section_enemy_"..enemies[ i ].name, "fail", func_after, "fail_static" );
      end;
      table.insert( completion, func_anim );
    end;
  end;

  local func_complete = function ( func_after )
    table.insert( completion, func_after );
    common.RunCallbackList( completion );
  end;  

  local notification = "View_Minievent_Section_Update_BossCompleted";
  private.Update_TryAddStep( update, func_complete, notification );

end;
--------------------------------------------------------------------------------
function private.Update_TryShowCharacterPanel (
    update,
    event_data,
    stage_index,
    when
  )

  if ( event_data.dialogs                                  ) and
     ( event_data.dialogs[ "stage_"..stage_index ]         ) and
     ( event_data.dialogs[ "stage_"..stage_index ] == when ) then

    local dialog_data = v_minievent.Event_GetDialogData( stage_index, when );

    if ( dialog_data ) then

      local func_show = function ( func_after )

        local delegates = {
          Complete = function ()
            v_minievent.Event_SetDialogComplete(
              dialog_data.name,
              private.event.id
            );
          end,
          OnShow = function ()
            ui.input.Off( MAP );
            ui.screen.Lock( false );
          end,
          OnHide = function ()
            ui.screen.Lock( true );
            ui.input.On( MAP );
            func_after();
          end
        };

        common.delegate.Set( "int_panel_minievent_character", delegates );

        int_panel_minievent_character.Show( dialog_data );

      end;

      local notification = "View_Minievent_Section_Update_CharacterPanelShown";
      private.Update_TryAddStep( update, func_show, notification );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Update_TryShowStageReward ( update, event_data, progress )

  if ( event_data.is_reward ) then

    local rewards, reward_group, holder_type =
      v_minievent.Event_GetSectionStageRewards( private.event.id, progress );

    if ( not type( rewards ) == "table" ) then return; end;

    --DbgTrace( rewards );

    if ( holder_type == "chest" ) then

      for i, reward in pairs( rewards ) do

        local postfix = tern( #rewards > 1 )( "_"..i, "" );

        local ext_rewards = reward.items;
        reward.items      = nil;

        local obtain_params =
        {
          category   = "minievent_"..private.event.name,
          source     = "minievent_"..private.event.name,
          stage_name = "minievent_"..private.event.name,
          silent     = true
        };

        local base_name     = NAMESPACE.."_reward_"..progress..postfix;
        local reward_anm    = "anm_"..base_name;
        local reward_holder = "obj_"..base_name;

        local func_show = function ( func_after )

          ui.screen.Lock( true );

          public.Focus( reward_holder,
            {
              time       = 0.6,
              func_after = function ()

                ui.anim.Play(
                  reward_anm,
                  "get_"..reward.name,
                  function ()

                    v_items.Chest_Show( 
                      reward,
                      obtain_params,
                      ext_rewards,
                      function ()
                        ui.screen.Lock( true );
                        public.FocusAtCurrentPoint( func_after );
                      end,
                      function () ui.screen.Lock( false ); end
                    );

                    ui.obj.Delete( reward_holder );

                  end
                );

              end
            }
          );

        end;

        local notification = "View_Minievent_Section_Update_StageRewardShown";
        private.Update_TryAddStep( update, func_show, notification );

      end;

    elseif ( holder_type == "gift" ) then

      local base_name     = NAMESPACE.."_reward_"..progress;
      local reward_anm    = "anm_"..base_name;
      local reward_holder = "obj_"..base_name;

      local func_show = function ( func_after )

        public.Focus( reward_holder,
          {
            time       = 0.6,
            func_after = function ()

              ui.anim.Play(
                reward_anm,
                "get",
                function ()

                  v_minievent.Reward_Show(
                    reward_group,
                    "rewards",
                    func_after
                  );

                  ui.obj.Delete( reward_holder );

                end
              );

            end
          }
        );

      end;

      local notification = "View_Minievent_Section_Update_StageRewardShown";
      private.Update_TryAddStep( update, func_show, notification );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Update_TryRelocateCharacters ( update, event_data, progress )

  local characters = v_minievent.Event_GetCharacters( private.event.name );
  local relocation = {};

  for _, character in pairs( characters ) do
    if ( character.stages ) then
      local stage_number =
        private.Character_GetStageNumberByProgress( character, progress );

      local func_relocate = function ( func_after )
        private.Character_TryRelocate( character.name, stage_number, func_after );
      end;

      table.insert( relocation, func_relocate );

    end;
  end;

  if ( #relocation > 0 ) then

    local func_relocate = function ( func_after )
      table.insert( relocation, func_after );
      common.RunCallbackList( relocation );
    end;

    local notification = "View_Minievent_Section_Update_CharacterRelocation";
    private.Update_TryAddStep( update, func_relocate, notification );

  end;

end;
--------------------------------------------------------------------------------
function private.Update_FocusAtEnabledPoint ( update, stage_number )

  local function func_focus ( func_after )

    public.Focus( "obj_"..NAMESPACE.."_point_stage_"..stage_number,
      {
        time       = 0.6,
        func_after = function ()

          private.Marker_MoveToPoint( "stage_"..stage_number, false, true,
            function ()

              private.Stage_AnimatePointChange(
                "bounce",
                stage_number,
                private.Stage_SetPointEnabled( stage_number ),
                func_after
              );

            end
          );

        end
      }
    );

  end;

  local notification = "View_Minievent_Section_Update_EnabledStageFocused";
  private.Update_TryAddStep( update, func_focus, notification, stage_number );

end;
--------------------------------------------------------------------------------
function private.Update_TryStartQuests ( update, stage_number )

  local quests = v_minievent.Event_GetQuests( private.event.name );

  if ( type( quests ) ~= "table" ) then return; end;

  local started_quests = {};

  for _, quest in pairs( quests ) do
    if ( quest.start == stage_number ) then
      table.insert( started_quests, quest );
      private.active_quests[ quest.giver ] = true;
    end;
  end;

  if ( #started_quests < 1 ) then return; end;

  for _, quest in pairs( started_quests ) do

    local function func_update ( func_after )

      local function func_reveal_items ()

        local is_func_after_set;      

        for _, number in pairs( quest.stages ) do

          local multiset  = {};
          local base_name = NAMESPACE.."_point_reward";

          ui.obj.Set(
            "anm_"..base_name.."_icon_stage_"..number,
            {
              animfunc = "reward_"..quest.item
            },
            multiset
          );

          ui.obj.Enable(
            "obj_"..base_name.."_stage_"..number, true, multiset
          );

          ui.obj.Multiset( multiset );

          local func_end;
          if ( not is_func_after_set ) then
            func_end = function ()
              public.Focus( "obj_"..NAMESPACE.."_point_stage_"..stage_number,
                {
                  time       = 0.6,
                  func_after = func_after
                }
              );
            end;
            is_func_after_set = true;
          end;

          ui.anim.Play(
            "anm_"..base_name.."_stage_"..number,
            "appear",
            func_end,
            "static"
          );

        end;

      end;

      local base_name = NAMESPACE.."_character_"..quest.giver;
      local giver_anm = "anm_"..base_name;
      local giver_btn = "btn_"..base_name;

      public.Focus(
        giver_anm,
        {
          time       = 0.6,
          func_after = function ()
            ui.control.EventsGet( giver_btn ).func_on_click( func_reveal_items );
          end
        }
      );

    end;

    local notification = "View_Minievent_Section_Update_QuestStarted";
    private.Update_TryAddStep( update, func_update, notification );

  end;

end;
--------------------------------------------------------------------------------
function private.Update_TryApplyQuestItem ( update, stage_number )

  local base_name = NAMESPACE.."_point_reward";
  local icon_anm  = "anm_"..base_name.."_icon_stage_"..stage_number;
  local holder    = "obj_"..base_name.."_stage_"..stage_number;

  if ( ui.obj.IsEnabled( holder ) ) then

    local quests = v_minievent.Event_GetQuests( private.event.name );

    if ( type( quests ) ~= "table" ) then return; end;

    local current_quest;
    local progress = {};

    for _, quest in pairs( quests ) do
      local current_index = common.IsEntityInList( quest.stages, stage_number );
      if ( current_index ) then
        current_quest  = quest;
        progress.index = current_index;
        progress.total = #current_quest.stages;
        break;
      end;
    end;

    local character = current_quest.giver;
    local receiver  = "obj_"..NAMESPACE.."_character_"..character;

    if ( not ui.obj.Exists( receiver ) ) then return; end;

    local indicator = "obj_"..NAMESPACE.."_character_indicator_"..character;
    local speech    = "str_minievent_"..private.event.name.."_"..character.."_item_received";

    local pos = ui.obj.Position( icon_anm );

    local icon_spr = ui.obj.CreateSprite(
      base_name,
      InterfaceName,
      {
        res     = DIR.."items/item_"..current_quest.item,
        alp     = 0.0,
        pos_x   = pos[ 1 ],
        pos_y   = pos[ 2 ],
        scale_y = 0.3,
        scale_x = 0.3
      }
    );

    local center = ui.screen.Center();

    local speech_data = {
      icon  = DIR.."items/item_"..current_quest.item,
      count = progress.index,
      target = {
        count = progress.total
      }
    };

    local func_update = function ( func_after )

      ui.anim.FadeIn( 0.3, icon_spr );
      ui.anim.FadeOut(
        0.3, icon_anm, function ()

          ui.anim.Receive(
            receiver,
            icon_spr,
            function ()
              
              ui.obj.Delete( icon_spr );
              ui.obj.Disable( holder );

              ui.anim.Play(
                "anm_"..NAMESPACE.."_activity_"..character,
                "step_"..progress.index
              );

              ui.popup.Cloud_Show(
                "counter",
                indicator,
                speech,
                speech_data,
                nil,
                nil,
                {
                  func_close = function ()
                    --public.FocusAtCurrentPoint( func_after );
                    func_after();
                  end
                }
              );

            end,
            { center.pos_x, center.pos_y }
          );

          ui.anim.Scale(
            icon_spr, 0.6, 1.5, function ()
              ui.anim.Scale( icon_spr, 0.3, 0.25 );
               ui.anim.FadeOut( 0.3, icon_spr );
            end
          );

          public.Focus( receiver, { time = 1.0 } );

        end
      );

    end;

    local notification = "View_Minievent_Section_Update_QuestItemApplied";
    private.Update_TryAddStep( update, func_update, notification );

  end;

end;
--------------------------------------------------------------------------------
function private.Update_TryFinishQuests ( update, stage_number )

  local quests = v_minievent.Event_GetQuests( private.event.name );

  if ( type( quests ) ~= "table" ) then return; end;

  local finished_quests = {};

  for _, quest in pairs( quests ) do
    local index = common.IsEntityInList( quest.stages, stage_number );
    if ( index == #quest.stages ) then
      table.insert( finished_quests, quest );
      private.active_quests[ quest.giver ] = nil;
    end;
  end;

  if ( #finished_quests < 1 ) then return; end;

  for _, quest in pairs( finished_quests ) do

    local giver_anm = "anm_"..NAMESPACE.."_character_"..quest.giver;
    local indicator = "obj_"..NAMESPACE.."_character_indicator_"..quest.giver;
    local speech    = "str_int_popup_cloud_characters_completed";

    local func_update = function ( func_after )

      public.Focus(
        giver_anm,
        {
          time       = 0.6,
          func_after = function ()
            ui.popup.Cloud_Show(
              "monolog",
              indicator,
              speech,
              nil,
              nil,
              nil,
              {
                func_close = function ()
                  --public.FocusAtCurrentPoint( func_after );
                  func_after();
                end
              }
            );

          end
        }
      );

    end;

    local notification = "View_Minievent_Section_Update_QuestFinished";
    private.Update_TryAddStep( update, func_update, notification );

  end;

end;
--------------------------------------------------------------------------------
function private.Update_TryPlaceRewardsAtStages ( update, stage_number )

  local stage_list = {};
  local settings   = v_minievent.Event_GetRewardSettings( private.event.name );

  for stage_name, data in pairs( settings ) do
    if ( data.appear                 ) and
       ( data.appear == stage_number ) then
      table.insert( stage_list, stage_name ); 
    end;
  end;

  for _, stage_name in pairs( stage_list ) do

    local number  = tonumber( stage_name:match( "_(%d+)" ) );
    local postfix = "";
    local holder  = "obj_"..NAMESPACE.."_reward_"..number..postfix;

    local func_update = function ( func_after )

      ui.obj.Hide( holder );
      ui.obj.Enable( holder );
      ui.anim.FadeIn( 0.6, holder, false, func_after );
      ui.anim.Bounce( holder );

    end;

    local notification = "View_Minievent_Section_Update_RewardPlacedAtStage";
    private.Update_TryAddStep( update, func_update, notification );

  end;

end;
--------------------------------------------------------------------------------
function private.Update_ShowShareDialog ( update ) 

  --DbgTrace( "Update_ShowShareDialog" );

  local func_update  = v_minievent.DialogShare_Show;
  local notification = "View_Minievent_Section_Update_ShareDialogShown";
  private.Update_TryAddStep( update, func_update, notification );

end;
--------------------------------------------------------------------------------
function private.Update_ShowEventIsFinished ( update )

  local func_update = function ( func_after )
    v_game.SystemDialog_Show( 
      "event_finished",
      function ()
        v_minievent.Section_Unload();
      end
    );
    ui.screen.Lock( false );
  end;

  local notification = "View_Minievent_Section_Update_EventIsFinished";
  private.Update_TryAddStep( update, func_update, notification );

end;
--------------------------------------------------------------------------------
function private.Update_Finish ( update )

  local function func_finish ( func_after )    

    ui.screen.Lock( true );

    if ( not v_interface.Show( nil, func_after ) ) then
      ui.screen.Lock( false );
    end;

    private.is_updating = nil;

  end;

  local notification = "View_Minievent_Section_Update_Finished";
  private.Update_TryAddStep( update, func_finish, notification );

end;
--------------------------------------------------------------------------------
function public.Cheat_EnableAllLevels ()
  
  local _, stages = m_minievent.Event_GetProgress( private.event.id );
  local result = false;
  
  private.cheat_force = true;
  
  for i = 1, 110 do
    
    if ( not common.IsEntityInList( stages, "stage_"..i.."_minievent_"..private.event.name ) ) then
      result = true;
      private.Stage_SetPointEnabled( i );
    end;

  end;
  
  return result;

end;
--------------------------------------------------------------------------------