-- name=v_player
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_player",
  "int_dialog_player_main",
  "int_dialog_player_skin",
  "int_dialog_player_code_enter",
  "int_dialog_player_code_info"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  common.mod.Load( _modules );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Init ()
  
  private.count_notifications_potions = 0;
  
  private.Stats_StartIngameWatch();
  private.Subscribe();
  private.TryObtainPlatformSkins();
  private.Notification_Reset( true );
  private.DialogMain_SetHandlers();

  v_interface.window.RegisterAsModal( "int_dialog_player_skin" );
  v_interface.window.RegisterAsModal( "int_dialog_player_code_enter" );
  v_interface.window.RegisterAsModal( "int_dialog_player_code_info" );
  
  private.PanelButton_Enable();
  private.Panel_Update();
  
  v_interface.Gold_SetCount( m_player.Gold_GetCount() );
  v_interface.Resource4_SetCount( m_player.Resource_GetCount( "resource4" ) );
  
  --Log.i( "Initialized." );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.PanelButton_Enable ()
  
  common.delegate.Set( "int_panel_interface_bottom", 
  {
    PlayerClickHandler = function ()
    
      v_interface.panel.top.Lock();
      private.DialogMain_Show();
      common.sound.Play( "bottom_button" );
      private.Notification_CustomClear();
      
    end
  } );
  
  v_interface.Scheme_AddButton( "map", "bottom", { name = "player" } );

  local count_potions = v_game.GetNotification( "potions" );
  if ( count_potions ) and ( count_potions > 0 ) then
    private.count_notifications_potions = count_potions;
  end;
  
  local count_skins = v_game.GetNotification( "skins" );
  if ( count_skins ) and ( count_skins > 0 ) then
    private.notifications[ "skins_counter" ] = count_skins;
  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.GetAttemptsVideo ()
  
  local  count, max_count = m_player.GetAttemptsVideo();
  return count, max_count;
  
end;
--------------------------------------------------------------------------------
function public.UpdateAttemptsVideo ( attempt_count )
  
  m_player.UpdateAttemptsVideo( attempt_count );
  
end;
--------------------------------------------------------------------------------
function public.IsAvailableAttemptsVideo ()
  
  local  state = m_player.VideoTimeout_IsFinished();
  return state;
  
end;
--------------------------------------------------------------------------------
function public.PurchaseExecute ()
  
  m_player.PurchaseExecute();
  
end;
--------------------------------------------------------------------------------
function public.UpdateVideoDataDays ( day )
  
  m_player.Storage_UpdateVideoDataDays( day );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.DialogMain_FocusOnSkin ( skin_name, func_end )

  int_dialog_player_main.Skins_Focus( skin_name, func_end );

end;
--------------------------------------------------------------------------------
function public.DialogMain_HasSkinSelectButton ()

  local  is_button_available = int_dialog_player_main.HasSelectButton();
  return is_button_available;

end;
--------------------------------------------------------------------------------
function public.DialogSkin_Show ( new_skin_part, func_OnClose, force )

  local old_skin      = m_player.Skin_GetCurrent(); 
  local new_skin      = public.Skin_GetByPart( new_skin_part );
  local new_skin_info = m_player.Skin_Data( new_skin );
  
  if ( not m_player.Skin_Data( new_skin ).primary ) and ( not force ) then
    
    ( func_OnClose or function () end )();
    return;
    
  end;

  local delegates = {

    Claim = function ()
    
      m_player.Skin_TrySet( new_skin ); 
      --ES> translate to Hide
      ( func_OnClose or function () end )();
      
      private.DialogSkin_Hide();

    end
  };

  common.delegate.Set( "int_dialog_player_skin", delegates );

  if ( force ) then
    
    v_interface.window.Show(
      "int_dialog_player_skin",
      function ()

        v_interface.panel.top.Hide();

        int_dialog_player_skin.Show(
          old_skin,
          new_skin,
          new_skin_info
        );

      end
    );
    
  else
  
    v_interface.window.AddToQueue(
      "int_dialog_player_skin",
      function ()

        v_interface.panel.top.Hide();

        int_dialog_player_skin.Show(
          old_skin,
          new_skin,
          new_skin_info
        );

        end,
        { disallow_modal = true }
    ); 
  
  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.DialogCode_Enter_Show ()

  local func_OnCodeApplied = function ( params )
    
    if ( params.result ) then
      
      int_dialog_player_code_enter.OnCodeAccepted();
      v_social.Mail_RequestUpdate();
      
    else
      
      int_dialog_player_code_enter.OnCodeRejected( params.reason );
        
    end;
    
  end;
  
  mediator.Subscribe(
    "Event_Player_FriendCodeApplied",
    func_OnCodeApplied
  );
  
  local delegates = {
    
    EnterCode = function ( code )

      local result, reason = m_player.FriendCode_RedeemCode( code, v_social );
      
      if ( not result ) then

        int_dialog_player_code_enter.OnCodeRejected( reason );

      end;
      
    end,
    
    Close = function ()
      
      int_dialog_player_code_enter.Hide();
      
    end,
    
    OnHide = function ()
      
      mediator.Unsubscribe(
        "Event_Player_FriendCodeApplied",
        func_OnCodeApplied
      );
      
      private.Panel_Update();
      
      v_interface.window.Hide( "int_dialog_player_code_enter" );
      
    end
    
  };
  
  common.delegate.Set( "int_dialog_player_code_enter", delegates );

  v_interface.window.Show(
    "int_dialog_player_code_enter",
    function ()
    
      int_dialog_player_code_enter.Show();

    end
  );
  
end;
--------------------------------------------------------------------------------
function public.DialogCode_Enter_Hide ()
  
  int_dialog_player_code_enter.Hide();
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.UserId ()
  
  local user_id = AuthGetUserId();
  
  if ( user_id ~= "" ) then
    
    return tonumber( user_id );
    
  end;

  return 0;
  
end;
--------------------------------------------------------------------------------
function public.IsAbtestParticipant ()
  
  local  is_participant = lib_db.Player.Info().abtest;
  return is_participant;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Name ()

  local  local_name = m_player.Name_GetLocal();
  return local_name;

end;
--------------------------------------------------------------------------------
function public.Name_SetLocal ( player_name )

  local  is_set = m_player.Name_SetLocal( player_name );
  return is_set;

end;
--------------------------------------------------------------------------------
function public.Name_TrySetSocial ( player_name, silent )

  local  is_set = m_player.Name_TrySetSocial( player_name, silent );
  return is_set;

end;
--------------------------------------------------------------------------------
function public.Name_GenerateSocial ()
  
  local player_name = m_player.Name_GetLocal();
  
  if ( not public.Name_IsSocial( player_name ) ) then
     
     local first_name  = ne.strings.Get( "str_player_name_"..math.random( 10 ) );
     local second_name = ne.strings.Get( "str_player_surname_"..math.random( 10 ) );
    
    player_name = first_name.." "..second_name;
    
  end;
  
  return player_name;
  
end;
--------------------------------------------------------------------------------
function public.Name_IsSocial ( player_name )
  
  player_name = player_name or m_player.Name_GetLocal();
  
  return ( not player_name ) or ( player_name == "Player" or player_name == "" );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Skin_GetCurrent ()

  local  current_skin = m_player.Skin_GetCurrent();
  return current_skin;

end;
--------------------------------------------------------------------------------
function public.Skin_GetAllSkins ()
  
  local function func_GetSkinContent ( skin_name )
  
    local package_name = m_player.Skin_GetPackageName( skin_name );

    package_name = package_name:match( "(.*)/.*" ) or package_name;

    if ( not v_game.Content_IsPackageReady( package_name ) ) then
      return false;
    end;
    
    --[[local states = loader.Content_GetPacketsState();

    if ( next( states ) and
         ( not states[ package_name ] or 
               states[ package_name ] ~= "ready" ) ) then
      return false;
    end;]]

    return public.Skin_GetContent( skin_name );

  end;
  
  local  skins, name_list = m_player.Skin_GetAllSkins( func_GetSkinContent );
  return skins, name_list;
  
end;
--------------------------------------------------------------------------------
function public.Skin_TryObtainParts ( part_name, count, category, silent )
  
  local skin_name   = public.Skin_GetByPart( part_name );
  local is_obtained = m_player.Skin_TryObtainParts(
    skin_name,
    part_name,
    count,
    category,
    silent
  );

  return is_obtained;
  
end;
--------------------------------------------------------------------------------
function public.Skin_IsPartObtainable ( part_name, count )
  
  local skin_name = public.Skin_GetByPart( part_name );

  local is_obtainable = m_player.Skin_IsPartObtainable(
    skin_name,
    part_name,
    count
  );
  
  --DbgTrace( "[ V_PLAYER ] "..part_name.." is obtainable: "..tostring( is_obtainable ) );
  
  return is_obtainable;
  
end;
--------------------------------------------------------------------------------
function public.Skin_GetByPart ( part_name )
  
  -- Part name prefix "braveheart" for "braveheart_boots" 
  return string.match( part_name, "%w+" );
  
end;
--------------------------------------------------------------------------------
function public.Skin_GetContent ( skin_name )
    
  local package_name = m_player.Skin_GetPackageName( skin_name );

  return package_name.."/"..skin_name;
  
end;
--------------------------------------------------------------------------------
function public.Skin_CollectablePartCount ( part_name )

  local skin_name  = public.Skin_GetByPart( part_name );
  local skin_data  = m_player.Skin_QueryStatic( skin_name );
  local saved_data = m_player.Skin_QueryStorage( skin_name ) or {};

  if ( skin_data ) then

    local current = saved_data.parts or 0;
    local total   = skin_data.parts  or 0;

    return ( total - math.min( current, total ) );

  end;

  return 0;

end;
--------------------------------------------------------------------------------
function public.Skin_Data ( skin_name )

  local  skin_data = m_player.Skin_Data( skin_name );
  return skin_data;

end;
--------------------------------------------------------------------------------
function public.Skin_TryAdd ( lot_id )
  
  local skins = public.Skin_GetAllSkins();
  
  for i = 1, #skins do
    
    local skin = skins[ i ];
    
    if ( skin.price and skin.price.lot and ( skin.price.lot == lot_id ) ) then
      private.Skin_ShowBougth( skin.name );
      break;
    end;
    
  end;

end;
--------------------------------------------------------------------------------
function private.Skin_GetFriendCodeSkin ()

  local  skin_data = m_player.Skin_GetByInviteType( "friendcode" );
  return skin_data.name;

end;
--------------------------------------------------------------------------------
function private.Skin_Purchase ( skin_name )
  
  local skin = m_player.Skin_Data( skin_name );
  
  if ( not skin.price ) then
    return;
  end;
  
  if ( skin.price.gold ) then
  
    if ( public.Gold_TryChangeCount( - skin.price.gold ) ) then
      
      v_log.Kontagent_LogEvent( "economy", 
      {
        description = "premium_currency_spend",
        category    = "skin_purchase",
        value       = - skin.price.gold,
        name        = skin_name
      } );
      
      private.Skin_ShowBougth( skin_name );
      
    end;
    
  elseif ( skin.price.lot ) then
    
    v_shop.Skin_TryBuy( skin.price.lot );
    
  end;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.FriendCode_Show ()
  
  if ( private.FriendCode_IsAvailable() ) then
    
    local skin_name = private.Skin_GetFriendCodeSkin();

    local onHide = function ( onEnd )
      v_interface.window.Hide( "int_dialog_player_code_info", onEnd );
    end;
    
    local delegates = {
      Hide = function () 
        int_dialog_player_code_info.Hide( onHide );
      end
    };
    common.delegate.Set( "int_dialog_player_code_info", delegates );    
    
    local invite_code, is_available = m_player.FriendCode_GetInviteCode();
    
    if ( not invite_code ) then
      
      mediator.SubscribeOnce(
        m_game.msg.ConnectRestored,
        function ()
          
          local invite_code, is_available = m_player.FriendCode_GetInviteCode();
          
          if ( invite_code ) then
          
            int_dialog_player_code_info.UpdateInviteCode( invite_code );
          
          end;
          
        end
      );
      
    end;
    
    local onShow = function ()
   
      int_dialog_player_code_info.Show(
        { code = invite_code },
        skin_name
      );

    end;
    v_interface.window.Show( "int_dialog_player_code_info", onShow );    
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.FriendCode_IsAvailable ()
 
  local unlock_stage = m_player.FriendCode_GetFeatureUnlocker() or 1;
  unlock_stage = "stage_"..unlock_stage;
  
  if ( not v_stage.Stage_IsCompleted( unlock_stage ) ) then
    
    return false;
    
  end;
  
  return true;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Skin_GetInvite ( skin_name )
  
  local skin_data = public.Skin_Data( skin_name );
  
  if ( skin_data.invite ) then
    
    local invite_funcs = {
      friendcode = {
        show = private.FriendCode_Show
      }
    };
    
    if ( invite_funcs[ skin_data.invite ]            ) and
       ( private.Skin_IsInviteAvailable( skin_name ) ) then
      ( invite_funcs[ skin_data.invite ].show or function () end )();
    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Skin_IsInviteAvailable ( skin_name )
  
  local skin_data = public.Skin_Data( skin_name );
  
  if ( skin_data.invite ) then
    
    local invite_funcs = {
      friendcode = {
        IsAvailable = private.FriendCode_IsAvailable,
      }
    };
    
    if ( invite_funcs[ skin_data.invite ] ) then
      local  is_available = invite_funcs[ skin_data.invite ].IsAvailable();
      return is_available
    end;

  end;
  
  return false;

end;
--------------------------------------------------------------------------------
function private.Skin_ShowBougth ( skin_name )
  
  m_player.Skin_Purchase( skin_name );
  
  if ( not private.DialogPlayerMain ) then
    return;
  end;
  
  local func_OnClose = function ()
  
    local skins = public.Skin_GetAllSkins();
    
    int_dialog_player_main.Skins_Update( skins );
    
  end;

  public.DialogSkin_Show( skin_name, func_OnClose, true );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Gold_GetCount ()
  
  local  count = m_player.Gold_GetCount();
  return count;
  
end;
--------------------------------------------------------------------------------
function public.Gold_TryChangeCount ( delta, hidden )
  
  local  is_changed = m_player.Gold_TryChangeCount( delta, hidden );
  return is_changed;
  
end;
--------------------------------------------------------------------------------
function public.Gold_IsEnough ( required_count, notify )
  
  local  is_enough = m_player.Gold_IsEnough( required_count, notify );
  return is_enough;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Resource_GetCount ( resource_name, section_number )
  
  section_number = section_number or v_sections.Section_NumberCurrent();
  local count    = m_player.Resource_GetCount( resource_name, section_number );
  
  return count;
  
end;
--------------------------------------------------------------------------------
function public.Resource_TryChangeCount ( resource_name, delta, hidden, section )
  
  local is_changed = m_player.Resource_TryChangeCount(
    resource_name,
    delta,
    hidden,
    section
  );
  
  return is_changed;
  
end;
--------------------------------------------------------------------------------
function public.Resource_TryChangeMultipleCount ( resource_data )
  
  local  is_changed = m_player.Resource_TryChangeMultipleCount( resource_data );
  return is_changed;
  
end;
--------------------------------------------------------------------------------
function public.Resource_IsEnough ( resource_name, required_count, notify )
  
  local is_enough = m_player.Resource_IsEnough(
    resource_name,
    required_count,
    notify
  );
  
  return is_enough;
  
end;
--------------------------------------------------------------------------------
function public.Resource_ClearAllForSection ( section_name )
  
  m_player.Resource_ClearAllForSection( section_name );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Achievements_TryShowPanel ()

  if ( #private.notifications[ "achievements" ] > 0 ) then
    
    local achievement_level = private.notifications[ "achievements" ][ #private.notifications[ "achievements" ] ]; 

    private.notifications[ "achievements" ] = {};
   
    local functions = {
      OnHide = function () end,
      OnTap  = function ()
  
        v_interface.panel.top.Lock();
        private.DialogMain_Show();
    
      end
    };
    
    v_achievements.ShowPanel(
      achievement_level,
      functions
    );

  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.PayingStatus_Set ( price )

  local  is_set = m_player.PayingStatus_Set( price );
  return is_set;
  
end;
--------------------------------------------------------------------------------
function public.PayingStatus_Get ()

  local  paying_status = m_player.PayingStatus_Get();
  return paying_status;
  
end;
--------------------------------------------------------------------------------
function public.PayingStatus_GetLastPurchaseTime ()

  local  last_purchase_time = m_player.PayingStatus_GetLastPurchaseTime();
  return last_purchase_time;
  
end;
--------------------------------------------------------------------------------
function public.PayingStatus_SetLastPurchaseTime ( seconds )

  local  is_set = m_player.PayingStatus_SetLastPurchaseTime( seconds );
  return is_set;
  
end;
--------------------------------------------------------------------------------
--function .. () end;
--------------------------------------------------------------------------------
function public.Playhaven_GetPushNotificationsReward ()

  return m_player.Settings( "notifications_reward" );

end;
--------------------------------------------------------------------------------
function public.Playhaven_RegisterPushNotifications ( reward )

  wiz.bfg.playhaven.RegisterPushNotifications();
  
  if ( reward ) then
    
    m_player.Gold_TryChangeCount( reward );
    
  end;

  return true;
  
end;
--------------------------------------------------------------------------------
function public.GetNotification ()
  
  return wiz.settings.GetNotifications();
  
end;
--------------------------------------------------------------------------------
function public.EnableNotification ( enabled )
  
  wiz.settings.SetNotifications( enabled );
  
  return public.GetNotification();
  
end;
--------------------------------------------------------------------------------
--function .. () end;
--------------------------------------------------------------------------------
function public.Progress_SetStageState ( stage_name, stage_data )
  
  if ( stage_data ~= {} ) then
  
    m_player.Progress_SetStageState( stage_name, stage_data );
  
  end;
  
end;
--------------------------------------------------------------------------------
function public.Progress_SetSection ( section_number )
  
  m_player.SetSection( section_number );
  
end;
--------------------------------------------------------------------------------
function public.Progress_SectionCount ()
  
  return v_sections.DiscoveredCount();
  
end;
--------------------------------------------------------------------------------
function public.Progress_SetLastStageName ( stage_name, section_name, is_stage_completed )
  
  local stage_last_name    = m_player.Progress_GetLastStageName();
  local stage_last_section = v_stage.Stage_StaticData( stage_last_name ).section;
  local section_number = tonumber( section_name:match( "%d+" ) );
  local last_section_number = tonumber( stage_last_section:match( "%d+" ) );

  if ( last_section_number == section_number and not is_stage_completed ) or
     ( last_section_number > section_number ) then

    return false;

  end;

  local stage_name = m_player.Progress_SetLastStageName(
    stage_name,
    section_number,
    is_stage_completed
  );

  return stage_name;
  
end;
--------------------------------------------------------------------------------
function public.Progress_GetLastStageName ()

  local stage_name = m_player.Progress_GetLastStageName();

  return stage_name;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.FriendCode_GetFeatureUnlocker ()

  local  feature_unlocker = m_player.FriendCode_GetFeatureUnlocker();
  return feature_unlocker;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Subscribe ()
  
  ne.msg.Subscribe(
    Event_AddPlayhavenReward,
    function ( msg_type, msg_params )

      private.Playhaven_AddReward( msg_params );

    end
  );

  mediator.Subscribe(
    "Model_Skins_SkinCollected",
    function ( skin_name )
      
      private.Notification_Skins_Add( skin_name );
      private.Panel_Update();
    
    end
  );
  
  mediator.Subscribe(
    "Model_Skins_PartCollected",
    function ( params )
  
      private.Notification_Skins_AddPart( params.skin );
      private.Panel_Update();

    end
  );
  
  mediator.Subscribe(
    "Model_Potion_CountUpdate",
    function ( params )
  
      private.Notification_Potions_Add( params );
      private.PotionChangeValue( params );
      private.Panel_Update();

    end
  );

  mediator.Subscribe(
    "Model_Achievements_Rewarded",
    private.Panel_Update
  );
  
  mediator.Subscribe(
    "Model_Achievements_Completed",
    function ( achievement_data )

      private.Notification_Achievements_Add( achievement_data.level );
      private.Panel_Update();
    
    end
  );
  
  mediator.Subscribe(
    "Model_Skins_PlatformSkinCollected",
    function ( skin_name )
  
      private.Notification_CustomAdd( skin_name );

    end
  );

  mediator.Subscribe(
    "Model_Sections_SectionEnabled",
    private.OnSectionEnabled
  );
  
  mediator.Subscribe(
    "View_Map_Section_Switched",
    function ()
    
      local counts = {};
      
      for i = 1, 3 do
        local count = m_player.Resource_GetCount( "resource"..i );
        table.insert( counts, count );
      end;
      
      v_interface.Resource_Reload( counts );
    
    end
  );
  
  mediator.Subscribe( "Model_Potion_RunAction", private.PotionRunAction );
  mediator.Subscribe( "Model_Potion_CompleteAction", private.PotionCompleteAction );

end;
--------------------------------------------------------------------------------
function private.TryObtainPlatformSkins ()

  local platform = v_game.Platform();

  if ( platform == "android" ) then

    platform = wiz.Android.GetPlatform();

  end;

  m_player.Skin_TryObtainPlatformSkins( platform );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Panel_Update ()
  
  local count = private.Notification_Total_Count();
  
  local data = {
    target = "player",
    data = {
      type = "common",
      entity = "v_player",
      count = count
    }
  };
  
  v_interface.panel.bottom.Notification( data );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.DialogMain_GetCurrentTab ()

  local  current_tab = int_dialog_player_main.GetCurrentTab();
  return current_tab

end;
--------------------------------------------------------------------------------
function public.DialogMain_EnableAchievementsScrolling ()

  int_dialog_player_main.AchievementsTab_EnableScrolling();

end;
--------------------------------------------------------------------------------
function public.DialogMain_DisableAchievementsScrolling ()

  int_dialog_player_main.AchievementsTab_DisableScrolling();

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.DialogMain_SetHandlers ()

  local delegates = {

    ----------------------------------------------------------------------------
    PurchaseSkin = function ( skin_name )
      
      private.Skin_Purchase( skin_name );
      
    end,

    ----------------------------------------------------------------------------
    GetLotInfo = function ( lot_id )
      
      return v_shop.Lot_GetDataByIOSName( lot_id );
      
    end,
    ----------------------------------------------------------------------------
    Close = function ()
      
      if ( private.current_tab ) and ( private.current_tab == "potions" ) then
        v_interface.Off_TopPanel();
        private.current_tab = nil;
      end;

      mediator.Unsubscribe(
        "Model_Achievements_AchievementChanged",
        int_dialog_player_main.AchievementsTab_UpdateItem
      );
      
      private.DialogPlayerMain = false;
      
    end;
    ----------------------------------------------------------------------------
    Show = function ()
      
      private.DialogPlayerMain = true;
      
    end;
    ----------------------------------------------------------------------------
    ChangeSkin = function ( skin_name )

      local skins = public.Skin_GetAllSkins();

      if ( m_player.Skin_TrySet( skin_name ) ) then
        skins = public.Skin_GetAllSkins();
        int_dialog_player_main.Skins_Update( skins );
        v_log.Playhaven_LogPlacement( "outfits_change" );
      end;

    end,
    
    ----------------------------------------------------------------------------
    ToggleTopPanel = function ( tab_name )
      
      local func_tab_potions = function ()
        v_interface.Off_TopPanel();
        v_interface.panel.top.Highlight( "resource4" );
        v_interface.panel.top.Show();
      end;
      
      if ( not private.current_tab ) then
        if ( tab_name == "potions" ) then
          func_tab_potions();
        end;
      else
        if ( tab_name == "potions" ) then
          func_tab_potions();
        elseif ( private.current_tab == "potions" ) then
          if ( tab_name == "achievements" or tab_name == "skins" ) then
            v_interface.Off_TopPanel();
            v_interface.panel.top.Show();
          else
            v_interface.panel.top.Hide( true );
          end;
        else
          if ( tab_name == "achievements" or tab_name == "skins" ) then
            v_interface.panel.top.Show();
          else
            v_interface.panel.top.Hide( true );
          end;
        end;

      end;
      
      private.current_tab = tab_name;

    end,

    ----------------------------------------------------------------------------
    IsNeedLogin = function ()

      return ( IsMobile() and v_game.IsOnline() and not v_social.IsLogged() );

    end,
    
    ----------------------------------------------------------------------------
    IsInviteSkinAvailable           = private.Skin_IsInviteAvailable,
    GetInviteSkin                   = private.Skin_GetInvite,
    UpdatePanel                     = private.Panel_Update,
    Login                           = function () 
      v_social.Login( "hero_info" );
    end,
    GetAchievementList              = function ()
      local  list = v_achievements.GetAll();
      return list;
    end,
    GetAchievementReward            = private.Achievement_TryGetReward,
    GetAchievementRewardData        = m_achievements.GetRewardData,
    GetAchievementNotificationCount = private.Notification_Achievements_Count,
    OpenGoogleAchievements          = private.OpenGooglePlayAchievements,
    PotionsNotification             = private.Notification_Potions,
    PotionUse                       = private.UsePotion,
    PotionBuy                       = private.BuyPotion,
    PotionGet                       = private.PotionGet,
    PotionGetRemainingTime          = private.PotionGetRemainingTime,
    PotionIsActive                  = private.PotionIsActive,
    IsSkinTutorialStarted           = function ()
      local  is_started = v_tutorial.IsSkinTutorialStarted();
      return is_started;
    end
  };

  common.delegate.Set( "int_dialog_player_main", delegates );

end;
--------------------------------------------------------------------------------
function private.UsePotion ( potion_name )
  
  v_potions.UsePotion( potion_name );
  
end;
--------------------------------------------------------------------------------
function private.BuyPotion ( potion_name )
  
  local potion_price = private.PotionGet( potion_name ).price;
  
  if ( public.Resource_TryChangeCount( "resource4", - potion_price.count ) ) then
    
    v_log.Kontagent_LogEvent( "economy", 
    {
      description = "soft_currency_spend",
      category    = "potions",
      value       = - potion_price.count,
      name        = potion_name,
      type        = potion_price.resource
    } );
    
    v_potions.AddPotion( potion_name, 1, "shop" );
    
  end;
    
end;
--------------------------------------------------------------------------------
function private.PotionGet ( potion_name )
  
  local data = v_potions.GetPotionsData( potion_name );
  
  return data;
  
end;
--------------------------------------------------------------------------------
function private.PotionGetRemainingTime ( potion_name )
  
  local time = v_potions.GetRemainingTime( potion_name );
  
  return time;
  
end;
--------------------------------------------------------------------------------
function private.PotionIsActive ( potion_name )
  
  local state = v_potions.IsPotionActive( potion_name );
  
  return state;
  
end;
--------------------------------------------------------------------------------
function private.Achievement_TryGetReward (
    achievement_name,
    reward_type,
    reward_count,
    level
  )
  
  if ( reward_type ~= "gold" ) then return; end;

  if ( m_player.Gold_TryChangeCount( reward_count ) ) then
  
    v_achievements.SetRewarded( achievement_name, level );
    
    v_log.Kontagent_LogEvent( "economy", 
    {
      description = "premium_currency_collect",
      category    = "achievement",
      value       = reward_count,
      name        = achievement_name
    } );
  
  end;

end;
--------------------------------------------------------------------------------
function private.DialogMain_Show ( force_tab )

  local statistic        = private.Stats_GetGameData(); 
  local skins, name_list = public.Skin_GetAllSkins();
  local achievements     = v_achievements.GetAll();
  local potions          = v_potions.GetPotionsData();
  
  local tab_name = "skins";

  if ( private.Notification_Achievements_Count( achievements ) > 0 ) then

    tab_name = "achievements";

  end;
  
  if ( private.Notification_Potions_Count() > 0 ) then

    tab_name = "potions";

  end;
  
  if ( private.Notification_Skins_Count() > 0 ) then

    tab_name = "skins";

  end;
  
  if ( force_tab ) then

    tab_name = force_tab;

  end;
  
  -- JK: check if focused skin is available
  local focused_skin = private.notifications[ "skins" ][ 1 ] or
                       m_player.Skin_GetCurrent();
 
  if ( not common.IsEntityInList( name_list, focused_skin, nil, true ) ) then
    
    -- JK: fallback to first skin available
    focused_skin = name_list[ 1 ];
    
  end;

  v_interface.window.Show(
    "int_dialog_player_main",
    function ()

      int_dialog_player_main.Show(
        tab_name,
        skins,
        potions,
        achievements,
        statistic,
        focused_skin,
        private.notifications[ "skins" ],
        private.notifications[ "skinparts" ]
      );

    end
  );

  private.Notification_Reset();
  v_achievements.HidePanel();

  mediator.Subscribe(
    "Model_Achievements_AchievementChanged",
    int_dialog_player_main.AchievementsTab_UpdateItem
  );

end;
--------------------------------------------------------------------------------
function private.DialogSkin_Hide ( func_end )

  int_dialog_player_skin.Hide( function ()

    v_interface.window.Hide( "int_dialog_player_skin", func_end );

  end );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.PotionRunAction ( params )
  
  private.PotionAction( params.name, "run" );
  
end;
--------------------------------------------------------------------------------
function private.PotionCompleteAction ( params )
  
  private.PotionAction( params.name, "complete" );
  
end;
--------------------------------------------------------------------------------
function private.PotionAction ( potion_name, action )
  
  local potion = v_potions.GetPotionsData( potion_name );
  local potions = v_potions.GetPotionsData();
  
  int_dialog_player_main.PotionsUpdate( potions, potion, action );
  
end;
--------------------------------------------------------------------------------
function private.PotionChangeValue ( potion )
  
  private.PotionAction( potion.name, "update" );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Notification_Total_Count ()

  local count = 
    private.Notification_Achievements_Count() +
    private.Notification_Skins_Count() + 
    private.Notification_Potions_Count();
    
  return count;

end;
--------------------------------------------------------------------------------
function private.Notification_Potions_Add ( params )
  
  if ( params.delta > 0 ) then
    private.count_notifications_potions = private.count_notifications_potions + params.delta;
    v_game.SetNotification( "potions", private.count_notifications_potions );
  end;
  
end;
--------------------------------------------------------------------------------
function private.Notification_Potions_Count ()
  
  return private.count_notifications_potions;
  
end;
--------------------------------------------------------------------------------
function private.Notification_Potions ( clear )
  
  if ( clear ) then
    private.Notification_Potions_Clear();
  end;
  
  return private.Notification_Potions_Count();
  
end;
--------------------------------------------------------------------------------
function private.Notification_Potions_Clear ()
  
  if ( private.count_notifications_potions > 0 ) then
    v_game.SetNotification( "potions", 0 );
  end;
  private.count_notifications_potions = 0;
  private.Panel_Update();
  
end;
--------------------------------------------------------------------------------
function private.Notification_Skins_Count ()
  
  return private.notifications[ "skins_counter" ];
  
end;
--------------------------------------------------------------------------------
function private.Notification_Skins_Add ( skin_name )

  local skins = private.notifications[ "skins" ];

  if not ( common.IsEntityInList( skins, skin_name ) ) then

    table.insert( skins, skin_name );

    private.notifications[ "skins_counter" ] = #skins;
    
    v_game.SetNotification( "skins", #skins );
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.Notification_Skins_AddPart ( skin_name )

  local parts = private.notifications[ "skinparts" ];

  if not ( parts[ skin_name ] ) then

    parts[ skin_name ] = 1;

  else

    parts[ skin_name ] = parts[ skin_name ] + 1;

  end;

  private.Notification_Skins_Add( skin_name );
  
end;
--------------------------------------------------------------------------------
function private.Notification_Achievements_Count ( ext_achievement_list )

  local achievements = ext_achievement_list or v_achievements.GetAll();
  local count = 0;

  for a = 1, #achievements do

    for r = 1, #achievements[ a ].reward do

      if  ( achievements[ a ].reward[ r ].get      )
      and ( not achievements[ a ].reward[ r ].done ) then

        count = count + 1;
        break;

      end;

    end;

  end;

  return count;

end;
--------------------------------------------------------------------------------
function private.Notification_Achievements_Add ( achievement_level )
  
  table.insert( private.notifications[ "achievements" ], achievement_level );
  
end;
--------------------------------------------------------------------------------
function private.Notification_Reset ( start )
  
  if ( not start ) then
    if ( private.notifications.skins_counter > 0 ) then
      v_game.SetNotification( "skins", 0 );
    end;
  end;
  
  private.notifications = {
    skins         = {},
    skinparts     = {},
    achievements  = {},
    skins_counter = 0
  };
  
end;
--------------------------------------------------------------------------------
function private.Notification_CustomAdd ( name )
  
  if ( not private.notification_custom ) then
    private.notification_custom = {};
  end;
  
  table.insert( private.notification_custom, name );
  
  private.Notification_CustomCall( name, 1 );
  
end;
--------------------------------------------------------------------------------
function private.Notification_CustomClear ()
  
  if ( not private.notification_custom ) then return; end;
  
  for i = 1, #private.notification_custom do
    private.Notification_CustomCall( private.notification_custom[ i ], 0 );
  end;
  
  private.notification_custom = nil;
  
end;
--------------------------------------------------------------------------------
function private.Notification_CustomCall ( name, count )
  
  local funcs = {
    android = function ( count )
      
      local data_custom = {
        target = "player",
        data = {
          type = "custom",
          entity = "v_player",
          count = count,
          style = "google_notification",
          icon_path = ""
        }
      };
      
      v_interface.panel.bottom.Notification( data_custom );
      
    end
  };
  
  funcs[ name ]( count );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Stats_GetPlayerData ()

  local hero_stats    = private.Stats_GetHeroData();
  local skins_stats   = private.Stats_GetSkinData();
  local potions_stats = private.Stats_GetPotionsData();
  local union_stats   = private.Stats_Union( hero_stats, skins_stats, potions_stats );
  
  union_stats.name = m_player.Name_GetLocal();
  
  private.Stats_GetMinieventProgressData( union_stats );
  
  return union_stats;

end;
--------------------------------------------------------------------------------
function private.Stats_StartIngameWatch ()

  local token = "wtc_ingame_time";

  if ( not TimeIsWatchExist( token ) ) then

    TimeStartWatch( token, "GAME_TIME" );

  end;

end;
--------------------------------------------------------------------------------
function private.Stats_GetGameData ()

  local stats = {
    id           = public.UserId(),
    achievements = v_achievements.Stats(),
    quests       = v_quests.GetCompleteQuestCount(),
    stages       = v_stage.Stats()
  };

  return stats;

end;
--------------------------------------------------------------------------------
function private.Stats_GetSkinData ()

  local skin_name  = m_player.Skin_GetCurrent();
  local skin_bonus = m_player.Skin_QueryStatic( skin_name ).bonus or {};
  local stats = private.Stats_Fill( skin_bonus );

  return stats;
  
end;
--------------------------------------------------------------------------------
function private.Stats_GetPotionsData ()

  local potions = v_potions.GetStatsBonusPotions() or {};
  local stats = private.Stats_Fill( potions );

  return stats;
  
end;
--------------------------------------------------------------------------------
function private.Stats_GetHeroData ()

  local hero  = m_player.Settings( "hero" ) or {};
  local stats = private.Stats_Fill( hero );

  return stats;
  
end;
--------------------------------------------------------------------------------
function private.Stats_GetMinieventProgressData ( stats )
  
  stats.minievent_progress_crowns = nil;
  
  if ( _G.v_minievent ) and ( v_minievent.Crown_IsAvailableMinievent() ) then
    stats.minievent_progress_crowns = v_minievent.Crown_GetStatsBonuses();
  end;
  
end;
--------------------------------------------------------------------------------
function private.Stats_Fill ( data )
  
  local stats =  {
    gold          = data.gold          or 0,
    moves         = data.moves         or 0,
    score         = data.score         or 0,
    attack        = data.attack        or 0,
    health        = data.health        or 0,
    resource1     = data.resource1     or 0,
    resource2     = data.resource2     or 0,
    resource3     = data.resource3     or 0,
    resource4     = data.resource4     or 0,
    enemy_attack  = data.enemy_attack  or 0,
    enemy_health  = data.enemy_health  or 0,
    event_moves   = data.event_moves   or 0,
    fox           = data.fox           or 0,
    mines_speed   = data.mines_speed   or 0,
    saga_moves    = data.saga_moves    or 0,
    health_speed  = data.health_speed  or 0,
    super_hammer  = data.super_hammer  or 0,
    sale_boosters = data.sale_boosters or 0,
  };

  if ( data.resources ) then
    stats.resource1 = stats.resource1 + data.resources;
    stats.resource2 = stats.resource2 + data.resources;
    stats.resource3 = stats.resource3 + data.resources;
    stats.resource4 = stats.resource4 + data.resources;
  end;
  
  return stats;
  
end;
--------------------------------------------------------------------------------
function private.Stats_Union ( ... )
  
  local args = { ... };
  local stats_union = {};
  
  for key, value in pairs( args[ 1 ] ) do
    local value_param = 0;
    for i = 1, #args do
      value_param = value_param + args[ i ][ key ];
    end;
    stats_union[ key ] = value_param;
  end;
  
  return stats_union;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Metrics_GetLaunchData ()

  local  game_launch = m_player.Metrics_GetLastLaunchTime( seconds );
  return game_launch;

end;
--------------------------------------------------------------------------------
function public.Metrics_SetLaunchData ( seconds )

  local  is_set = m_player.Metrics_SetLaunchData( seconds );
  return is_set;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Playhaven_AddReward ( params )

  local func_handler = 
  {
    gold_coin = m_player.Gold_TryChangeCount
  };

  if ( params ) then

    if ( func_handler[ params.name ] ) then

      func_handler[ params.name ]( params.count );

    end;

  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.OpenGooglePlayAchievements ()

  if not ( IsAndroid() and ( wiz.Android.GetPlatform() == "google" ) ) then
  
    return;
  
  end;

  if ( wiz.Android.GooglePlayIsConnected() ) then
    wiz.Android.AndroidShowAchievements();
  else
    wiz.Android.GooglePlaySignIn();
  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.OnSectionEnabled ( params )
  
  public.Progress_SetLastStageName(
    params.first_stage,
    params.section_name
  );

end;
--------------------------------------------------------------------------------