-- name=v_quests
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_quests",
  "int_panel_darkhole"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------

public.Questlist  = {};

--------------------------------------------------------------------------------
function public.__InitModule ()
  
  common.mod.Load( _modules );
  
  private.notifications = {};
  
end;
--------------------------------------------------------------------------------
function public.Init ()

  private.AutoFix_TryStart();
  
  private.QuestList_Init();
  private.Subscribe_QuestEvents();
  
  mediator.Subscribe(
    "View_Map_Section_Switched",
    function ()
      
      private.AutoFix_TryStart();
    
      local section_name = v_sections.Section_NameCurrent();
      private.SetupVirtualQuests( section_name );
      private.Questlist_Set( section_name );
    
    end
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.AutoFix_TryStart ()
  
  Log.i( "AutoFix: TryStart" );

  local section_name = v_sections.Section_NameCurrent();

  private.autofix = {
    quests_static  = m_quests.Static_GetAllBySection( section_name ),
    stages = v_stage.Stages_GetAllCompleted(),
    builds = v_builds.GetAllInSection( section_name ),
    mines  = v_mines.Mines_GetAllConstructed()
  };
  
  local stages = {};
  
  for i = 1, #private.autofix.stages do
    
    stages[ private.autofix.stages[ i ].name ] = private.autofix.stages[ i ];
    
  end;
  
  local builds = {};
    
  for i = 1, #private.autofix.builds do
    
    builds[ private.autofix.builds[ i ].name ] = private.autofix.builds[ i ];
    
  end;
  
  local mines = {};
    
  for i = 1, #private.autofix.mines do
    
    mines[ private.autofix.mines[ i ].name ] = private.autofix.mines[ i ];
    
  end;   
  
  for i = 1, #private.autofix.quests_static do
    
    local quest     = private.autofix.quests_static[ i ];
    local condition = quest.condition;
    local target    = quest.target;
    
    if ( condition and condition.type == "QSC_StageCompleted" and
         stages[ condition.name ] ) then
         
      if ( not m_quests.Completed_GetByName( quest.name ) and
           not m_quests.Current_GetByName( quest.name ) ) then
           
        local event_params_target = { target_name = condition.name };
        private.HandleQuestEvent( "StageCompleted", event_params_target );
        
      end;
      
      private.AutoFix_TryHandle( quest, stages, builds, mines );
      
    elseif ( condition and condition.type == "QSC_QuestCompleted" ) then
      
      if ( condition.name and m_quests.Completed_GetByName( condition.name ) ) or
         ( condition.list and private.AutoFix_IsListCompleted( condition.list ) ) then
        
        private.AutoFix_TryHandle( quest, stages, builds, mines );
         
      end;

    end;
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.AutoFix_TryHandle ( quest, stages, builds, mines )
  
  if ( private.autofix.need_reload_builds ) then

    local section_name = v_sections.Section_NameCurrent();
    private.autofix.builds = v_builds.GetAllInSection( section_name );
    
    builds = {};
    
    for i = 1, #private.autofix.builds do
      
      builds[ private.autofix.builds[ i ].name ] = private.autofix.builds[ i ];
      
    end;
    
    private.autofix.need_reload_builds = nil;
    
  end;
  
  local condition = quest.condition;
  local target    = quest.target;

  Log.i( "AutoFix_TryHandle: "..quest.name );
  
  if ( target.type == "QT_StageCompleted" and
       stages[ target.name ] ) then
    
     private.AutoFix_TryHandleStageCompleted( quest );
     
  elseif ( target.type == "QT_ReachMaxLevelStages" ) then
     
    private.AutoFix_TryHandleReachMaxLevelStages( quest, stages );
    
  elseif ( target.type == "QT_BuildConstructed" and
           builds[ target.name ] and
           builds[ target.name ].constructed ) then
     
    private.AutoFix_TryHandleBuildConstructed( quest );
    
  elseif ( target.type == "QT_MineConstructed" ) then
     
    private.AutoFix_TryHandleMineConstructed( quest, mines, stages );
 
  elseif ( target.type == "QT_StageItemsCollected" ) then
     
    private.AutoFix_TryHandleStageItemsCollected( quest, stages );

  elseif ( target.type == "QT_FragmentCollected" ) then
     
    private.AutoFix_TryHandleFragmentCollected( quest, builds );
    
  elseif ( target.type == "QT_MarketConstructed" and
    v_shop.Market_IsBuilt() ) then
     
    private.AutoFix_TryHandleBuildConstructed( quest );

  end;
  
end;
--------------------------------------------------------------------------------
function private.AutoFix_TryHandleStageCompleted ( quest )
  
  if ( private.AutoFix_Condition( quest.name ) ) then
    
    Log.i( "HandleQuestEvent: StageCompleted: COMPLETE < "..quest.name.." > by < "..quest.target.name.." >." );
    m_quests.Current_Complete( quest );
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.AutoFix_TryHandleBuildConstructed ( quest )
  
  if ( private.AutoFix_Condition( quest.name ) ) then
    
    Log.i( "HandleQuestEvent: BuildConstructed : COMPLETE < "..quest.name.." > by < "..tostring( quest.target.name ).." >." );
    m_quests.Current_Complete( quest );
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.AutoFix_TryHandleMineConstructed ( quest, mines, stages )
  
  if ( private.AutoFix_Condition( quest.name ) ) then
    
    if ( mines[ quest.target.name ] ) then
      
      Log.i( "HandleQuestEvent: MineConstructed : COMPLETE < "..quest.name.." > by < "..quest.target.name.." >." );   
      m_quests.Current_Complete( quest );
      
    else
      
      local mine_data  = v_mines.Mine_GetStaticData( quest.target.name );
      local saved_data = v_mines.Mine_GetSavedData( quest.target.name );
      
      if ( mine_data.progress ) then
        
        local progress = 0;
        
        for i = 1, #mine_data.progress do
          
          local stage_name = mine_data.progress[ i ];
          
          if ( stages[ stage_name ] and
               stages[ stage_name ].noreward ) then
               
            progress = progress + 1;
            
          end;
          
        end;

        if ( saved_data and ( saved_data.progress or 0 ) < progress ) then

          v_mines.Mine_SetProgress( mine_data.name, 0, progress );
          Log.i( "HandleQuestEvent: MineConstructed : SET PROGRESS < "..mine_data.name.." > to < "..progress.." >." );
         
          if ( progress == #mine_data.progress and
               saved_data.state ~= "ready" and 
               saved_data.state ~= "active" ) then
            v_mines.Mine_SetStateWait( mine_data.name );
          end;
          
        end;

      end;
    
    end;
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.AutoFix_TryHandleReachMaxLevelStages ( quest, stages )
  
  local completed = true;  
  
  for j = 1, #quest.target.list do
     
    if ( not stages[ quest.target.list[ j ] ] or
         not stages[ quest.target.list[ j ] ].noreward ) then
      
      completed = false;
      break;
      
    end;
    
  end;  
  
  if ( completed and 
       private.AutoFix_Condition( quest.name ) ) then
    
    Log.i( "HandleQuestEvent: ReachMaxLevelStages: COMPLETE < "..quest.name.." >." );
    m_quests.Current_Complete( quest );
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.AutoFix_TryHandleStageItemsCollected ( quest, stages )
  
  local completed = true;
  
  for j = 1, #quest.target.list do
     
    if ( not stages[ quest.target.list[ j ] ] or
         not stages[ quest.target.list[ j ] ].noreward ) then
      
      completed = false;
      break;
      
    end;
    
  end;
  
  local current_data = m_quests.Current_GetByName( quest.name );
  
  if ( completed and 
       not m_quests.Completed_GetByName( quest.name ) and
       current_data and
       current_data.list and
       #current_data.list == #quest.target.list ) then
    
    Log.i( "HandleQuestEvent: StageItemsCollected: COMPLETE < "..quest.name.." >." );
    m_quests.Current_Complete( quest );
    
  elseif ( not completed and 
           not m_quests.Completed_GetByName( quest.name ) and 
           not current_data ) then
    
    Log.i( "HandleQuestEvent: StageItemsCollected: START < "..quest.name.." >." );
    m_quests.Current_Start( quest );
    
  elseif ( m_quests.Completed_GetByName( quest.name ) and quest.reward ) then
    
    for i = 1, #quest.reward do
      
      local reward_type = quest.reward[ i ].type;
      
      if ( reward_type == "fragment" or 
           reward_type == "skinpart" ) then
        
        Log.i( "HandleQuestEvent: StageItemsCollected: TRY OBTAIN < "..quest.reward[ i ].name.." > for < "..quest.name.." >." );
        private.ObtainQuestRewards( { quest.reward[ i ] }, quest.name );

        private.autofix.need_reload_builds = true;
    
      end;
 
    end;
       
  end;

end;
--------------------------------------------------------------------------------
function private.AutoFix_TryHandleFragmentCollected ( quest, builds )
  
  local is_completed = true;
  local build_name   = v_builds.Build_GetNameByFragment( quest.target.list[ 1 ] );
  local current_data = m_quests.Current_GetByName( quest.name ); 

  if ( build_name ) then

    for j = 1, #quest.target.list do
      
      local build_data = builds[ build_name ];
      local fragment   = quest.target.list[ j ];
      
      if ( not build_data ) then
        is_completed = false;
        break;
      end;
      
      local is_collected = common.IsEntityInList(
        build_data.collected_fragments, fragment
      );
      
      if ( not is_collected ) then
        
        -- JK: additional check if fragment is collected
        local stage_list = v_stage.Stages_FindByReward(
          "fragment", fragment, build_data.section, true
        );
        
        if ( v_stage.Stage_IsFinished( stage_list[ 1 ] ) ) then
          is_collected = true;
          
          local item = {
            type  = "fragment",
            name  = fragment,
            count = 1
          };
          
          v_items.TryObtain( item );
        end;
        
      end;
        
      if ( not is_collected ) then
        is_completed = false;
        break;
      end;
      
    end;
    
  end;

  if ( is_completed                            ) and 
     ( private.AutoFix_Condition( quest.name ) ) then
    
    Log.i( "HandleQuestEvent: FragmentCollected: for < "..quest.name.." >." );
    m_quests.Current_Complete( quest );
    
  end;  

end;
--------------------------------------------------------------------------------
function private.AutoFix_Condition ( quest_name )
  
  local completed = m_quests.Completed_GetByName( quest_name );
  local current = m_quests.Current_GetByName( quest_name );

  return ( not completed or current or ( not completed and not current ) );
  
end;
--------------------------------------------------------------------------------
function private.AutoFix_IsListCompleted ( condition_list )
  
  local completed = true;
  
  for i = 1, #condition_list do
    
    local quest_name = condition_list[ i ];
    
    if ( not m_quests.Completed_GetByName( quest_name ) ) then
      
      completed = false;
      break;
      
    end;
    
  end;
  
  return completed;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.QuestList_Init ()
  
  local delegate =
  {
    ViewPort_Restore  = v_map.ViewPort_Restore,
    ViewPort_Contract = v_map.ViewPort_Contract,
    Interface_Hide    = function () v_interface.Hide(); end,
    Interface_Show    = function () v_interface.Show(); end,
    Highlight_Show    = private.Highlight_Show,
    Highlight_Hide    = private.Highlight_Hide
  }
  mediator.Subscribe( m_quests.msg.Started,
    function ( params )
    
      if not ( params.silent ) then
      
        table.insert( private.notifications, params.name );
      
      end;
      
      local item = private.Questlist_GetItemDataByName( params.name );
    
    end
  );
  
  mediator.Subscribe( m_quests.msg.Updated,
    function ( quest_name )
    
      local item = private.Questlist_GetItemDataByName( quest_name );
    
    end
  );
  
  mediator.Subscribe( m_quests.msg.Completed,
    function ( params )
      
      for i = 1, #private.notifications do
        if ( private.notifications[i] == params.name ) then
          table.remove( private.notifications, i );
          break;
        end;
      end;
    
    end
  );

end;
--------------------------------------------------------------------------------
function private.Subscribe_QuestEvents ()
  
  mediator.Subscribe( "Model_Builds_BuildConstructed",
    function ( params )
    
      local event_params = { target_name = params.name };
      private.HandleQuestEvent( "BuildConstructed", event_params );
    
    end
  );
  
  mediator.Subscribe( v_items.msg.ItemObtained,
    function ( params )
      if ( params.type == "fragment" ) then
      
        local event_params = { list_item = params.name };
        private.HandleQuestEvent( "FragmentCollected", event_params );
      
      end;
    end
  );
  
  mediator.Subscribe( m_shop.msg.MarketConstructed,
    function ( params )
      private.HandleQuestEvent( "MarketConstructed" );
    end
  );
  
  mediator.Subscribe( "Model_Stage_StageWon",
    function ( params )
    
      local event_params_target = { target_name = params.name };
      local event_params_item   = { list_item = params.name };
    
      if ( params.state_old == 1 ) then
      
        private.HandleQuestEvent( "StageCompleted", event_params_target );
        
      elseif ( params.state_old == 4 and params.state_new == 1 ) then
      
        private.HandleQuestEvent( "StageCompletedMisc", event_params_target );
      
      elseif ( params.state_old == 3 and not params.noreward ) then
      
        private.HandleQuestEvent( "StageCompletedFull", event_params_item );
      
      end;
    
    end
  );
  
  mediator.Subscribe( "Model_Stage_StageEnabled",
    function ( stage_name )
    
      local event_params = { target_name = stage_name };
      private.HandleQuestEvent( "StageEnabled", event_params );
    
    end
  );
  
  mediator.Subscribe( "Model_Stage_QuestItemDropped",
    function ( params )
    
      private.HandleQuestEvent( 
        "StageItemsCollected", 
        { target_name = params.name, list_item = params.stage }
      );
    
    end
  );
  
  mediator.Subscribe( m_quests.msg.Completed,
    function ( params )
    
      if ( params.reward ) then
        private.ObtainQuestRewards( params.reward, params.name );
      end;
    
      local event_params = { target_name = params.name };
      private.HandleQuestEvent( "QuestCompleted", event_params );
    
    end
  );
  
  mediator.Subscribe( "Model_Mines_MineConstructed",
    function ( mine_name )
      local event_params = { target_name = mine_name };
      private.HandleQuestEvent( "MineConstructed", event_params );
    end
  );
  
  mediator.Subscribe( "Model_Mines_MineReady",
    function ( mine_name )
      local event_params = { target_name = mine_name };
      private.HandleQuestEvent( "MineReady", event_params );
    end
  );
  
  mediator.Subscribe( "Model_Mines_MineCollected",
    function ( mine_name )
      local event_params = { target_name = mine_name };
      private.HandleQuestEvent( "MineResourceCollected", event_params );
    end
  );
  
  mediator.Subscribe( "Model_Mines_GoldMineCollected",
    function ( mine_name )
      local event_params = { target_name = mine_name };
      private.HandleQuestEvent( "MineResourceCollected", event_params );
    end
  );
  
  mediator.Subscribe( "Model_Spells_SpellEnabled",
    function ( spell_name )
    
      local event_params = { target_name = spell_name };
      private.HandleQuestEvent( "SpellEnabled", event_params );
      
    end
  );
  
  mediator.Subscribe( "Model_Spells_SpellCompleted",
    function ( spell_name )
    
      local event_params = { target_name = spell_name };
      private.HandleQuestEvent( "SpellCompleted", event_params );
      
    end
  );
  
  mediator.Subscribe( m_quests.msg.ReachMaxLevels,
    function ( stages )
    
      for _, stage_name in pairs( stages ) do
      
        local stage_data = v_stage.Stage_StorageData( stage_name );
        if ( stage_data ) then
        
          local state    = stage_data.state    or 1;
          local noreward = stage_data.noreward or false;
          
          if ( state == 3 and noreward ) then
          
            local event_params = { list_item = stage_name };
            private.HandleQuestEvent( "StageCompletedFull", event_params );
          
          end;
        
        end;
      
      end;
    
    end
  );

end;
--------------------------------------------------------------------------------
function private.HandleQuestEvent ( event, params )

  local section_name = v_sections.Section_NameCurrent();
  m_quests.HandleQuestEvent( event, params, section_name );

end;
--------------------------------------------------------------------------------
function private.ObtainQuestRewards ( rewards, quest_name )

  for _, reward in pairs( rewards ) do

    reward.category = "quest";
    reward.quest_name = quest_name;
    
    v_items.TryObtain( reward );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.SetupVirtualQuests ( section_name )

  local virtual_data =
  {
    -- DG: list for filter events, active for start quests immediately
    stages = { list = {}, active = nil },
    mines  = { list = {}, active = { } },
    spells = { list = {}, active = { } }
  };
  
  -- Setup stages
  local section_number = v_sections.Section_NumberCurrent();
  local stage_max = 110 * section_number;
  local stage_min = stage_max - 109;
  
  -- NOTE: 110 requests in bad case, need to fix it?
  -- Probably need to add "section" index into "player/stages"
  for stage = stage_max, stage_min, -1 do
  
    local stage_name = "stage_"..stage;
    
    if ( v_stage.Stage_IsCompleted( stage_name ) ) then
    
      break;
    
    else
    
      table.insert( virtual_data.stages.list, stage_name );
    
    end;
    
  end;
  
  local stage_count = #virtual_data.stages.list;
  if ( stage_count > 0 ) then
  
    local stage_last = virtual_data.stages.list[ stage_count ];
    local stage_data = v_stage.Stage_StorageData( stage_last );
    
    if ( stage_data and stage_data.enabled ) then
    
      virtual_data.stages.active = stage_last;
    
    end;
  
  end;
  
  -- Setup mines
  local mines = v_mines.GetAllBySection( section_name );
  for _, mine in pairs( mines or {} ) do
  
    if ( not v_mines.IsMineUnavailable( mine.name )) then
    
      table.insert( virtual_data.mines.list, mine.name );
      
      local mine_data = v_mines.Mine_Data( mine.name );
      if  ( mine_data.state == "ready" ) then
      
        table.insert( virtual_data.mines.active, mine.name );
      
      end;
      
    end;
  
  end;
  
  -- Setup spells
  local spells = v_spells.Static_GetAllBySection( section_name );
  for _, spell in pairs( spells or {} ) do
  
    table.insert( virtual_data.spells.list, spell.name );
    
    local saved = v_spells.Saved_GetByName( spell.name );
    if  ( saved and not saved.completed ) then
    
      table.insert( virtual_data.spells.active, spell.name );
    
    end;
  
  end;
  
  m_quests.Virtual_Setup( virtual_data, section_name );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Questlist_Set ( section_name )
  
  local list = {};
  
  local quests_general = m_quests.Current_GetAllBySection( section_name );
  for _, quest in pairs( quests_general ) do
  
    table.insert( list, private.Questlist_GetItemData( quest, false ) );
  
  end;
  
  local quests_virtual = m_quests.Virtual_GetAll();
  for _, quest in pairs( quests_virtual ) do
  
    table.insert( list, private.Questlist_GetItemData( quest, true ) );
  
  end;
  
  private.notifications = {};

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Questlist_GetItemDataByName ( quest_name )

  local virtual_data = m_quests.Virtual_GetByName( quest_name );
  if ( virtual_data ) then
  
    local item_data = private.Questlist_GetItemData( virtual_data, true );

    if ( item_data ) then return item_data; end;

    DbgTrace( "[ ERROR ] No data found for item: "..virtual_data.name );
    return false;
  
  end;
  
  local general_data = m_quests.Current_GetByName( quest_name );
  if ( general_data ) then
  
    local item_data = private.Questlist_GetItemData( general_data, false );

    if ( item_data ) then return item_data; end;

    DbgTrace( "[ ERROR ] No data found for item: "..general_data.name );
    return false;

  end;

  DbgTrace( "[ ERROR ] No data found for quest: "..quest_name );
  return false;

end;
--------------------------------------------------------------------------------
function private.Questlist_GetItemData ( quest, is_virtual )

  local quest_data;
  if ( is_virtual ) then quest_data = quest;
  else                   quest_data = m_quests.Static_GetByName( quest.name );
  end;
  
  local stage_data = nil;
  local target     = quest_data.target;
  
  if ( target and target.type == "QT_StageCompleted" ) then
  
    stage_data = v_stage.Stage_StaticData( target.name );
  
  end;
  
  local current = quest.list or {}
  local item =
  {
    name     = quest_data.name;
    priority = private.QuestPriority( target, stage_data ),
    target   = private.QuestTarget( quest_data, current, stage_data ),
    focus    = private.QuestFocus( quest_data, current, stage_data ),
    reward   = private.QuestReward( quest_data.reward ),
    progress = private.QuestProgress( quest_data, current ),
    pointer  = private.QuestPointer( quest_data )
  };
  
  return item;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.QuestPointer ( quest_data )
  
  local target = quest_data.target;
  local condition = quest_data.condition;
  
  if ( target.type == "QT_ReachMaxLevelStages" ) then
  
    return { target.list[ 1 ], target.list[ #target.list ] };
    
  elseif ( target.type == "QT_BuildConstructed" ) then
    
    return { condition.name, "stage_1" };
    
  elseif ( target.type == "QT_SpellCompleted" ) then
    
    local spell_data = v_spells.GetSpellData( target.name );
    
    return { spell_data.unlock, "stage_1" };
    
  end;
  
  return nil;
  
end;
--------------------------------------------------------------------------------
function private.QuestPriority ( target, stage_data )

  if ( target.type == "QT_StageCompleted" and stage_data ) then
  
    local stage_priority =
    {
      fight = 0,
      ride  = 1,
      bonus = 2
    };
    
    return stage_priority[ stage_data.type ] or 2;
  
  end;

  local type_priority =
  {
    QT_StageItemsCollected   = 3,
    QT_FragmentCollected     = 5,
    QT_BuildConstructed      = 5,
    QT_MarketConstructed     = 6,
    QT_MineConstructed       = 7,
    QT_MineResourceCollected = 8,
    QT_SpellCompleted        = 9,
    QT_ReachMaxLevelStages   = 10,
  };
  
  return type_priority[ target.type ] or 20;

end;
--------------------------------------------------------------------------------
function private.QuestTarget ( quest_data, current_list, stage_data )

  local target = quest_data.target;

  if ( target.type == "QT_StageCompleted" ) then
  
    if ( stage_data and stage_data.type == "fight" ) then
    
      local state = tern( stage_data.realtype == "misc" )( 4, 1 );
      local target_data =
      {
        type  = "enemy",
        name  = stage_data.state[ state ].enemy.name,
        stage = stage_data.name
      };

      target_data.content = v_enemies.Enemy_GetContentPackageName(
        target_data.name
      );

      return target_data;
    
    else
    
      return { type = "stage", name = target.name };
    
    end;
    
  elseif ( target.type == "QT_BuildConstructed"    ) then
  
    return { type = "build", name = target.name };
  
  elseif ( target.type == "QT_FragmentCollected"   ) then
  
    local list = private.GetTargetList( target.list, current_list );
    return { type = "fragment", name = list[ 1 ] };
  
  elseif ( target.type == "QT_MarketConstructed"   ) then
  
    return { type = "market", name = "market" };
  
  elseif ( target.type == "QT_MineConstructed"     ) then
  
    return { type = "mine", name = target.name };
  
  elseif ( target.type == "QT_ReachMaxLevelStages" ) then
  
    local list = private.GetTargetList( target.list, current_list );
    return { type = "stagelist", name = list[ 1 ] };
  
  elseif ( target.type == "QT_StageItemsCollected" ) then
  
    return { type = "quest", name = target.name };
  
  elseif ( target.type == "QT_SpellCompleted" ) then
  
    return { type = "spell", name = target.name };
    
  elseif ( target.type == "QT_MineResourceCollected" ) then
  
    local _, res_name = v_mines.Mine_Type( target.name );
    return { type = "resource", name = res_name };
  
  else
  
    return { type = "unknown", name = "unknown" };
  
  end;

end;
--------------------------------------------------------------------------------
function private.QuestFocus ( quest_data, current_list, stage_data )
  
  local focus = { targets = {}, current = 1 };
  
  local target = quest_data.target;
  
  if ( target.type == "QT_StageCompleted" ) then
  
    if ( stage_data and stage_data.realtype == "fight" ) then
    
      local enemy = stage_data.state[ 1 ].enemy.name;
      local data  = { type = "fight", name = enemy, stage = target.name };
      table.insert( focus.targets, data );
    
    else
    
      local data = { type = "stage", name = target.name, stage = target.name };
      table.insert( focus.targets, data );
    
    end;
    
  elseif ( target.type == "QT_BuildConstructed"    ) then
  
    table.insert( focus.targets, { type = "build", name = target.name } );
  
  elseif ( target.type == "QT_FragmentCollected"   ) then
  
    local build_name = v_builds.Build_GetNameByFragment( target.list[ 1 ] );
    table.insert( focus.targets, { type = "build", name = build_name } );
  
  elseif ( target.type == "QT_MarketConstructed"   ) then
  
    table.insert( focus.targets, { type = "build", name = "market" } );
  
  elseif ( target.type == "QT_MineConstructed"     ) then
  
    table.insert( focus.targets, { type = "mine", name = target.name } );
  
  elseif ( target.type == "QT_ReachMaxLevelStages" ) or
         ( target.type == "QT_StageItemsCollected" ) then
  
    local  list = private.GetTargetList ( target.list, current_list );
    for _, item in pairs( list ) do
    
      local data = { type = "stage", name = item, stage = item };
      table.insert( focus.targets, data );
    
    end;
  
  elseif ( target.type == "QT_SpellCompleted" ) then
  
    table.insert( focus.targets, { type = "spell", name = target.name } );
    
  elseif ( target.type == "QT_MineResourceCollected" ) then
  
    table.insert( focus.targets, { type = "mine", name = target.name } );
    
  else
  
    local data = { type = "unknown", name = "unknown" };
    table.insert( focus.targets, data );
  
  end;
  
  return focus;

end;
--------------------------------------------------------------------------------
function private.QuestReward ( rewards )

  if ( rewards and rewards[ 1 ] ) then
  
    local reward = rewards[ 1 ];
    return { type = reward.type, name = reward.name, count = reward.count };
  
  end;
  
  return nil;

end;
--------------------------------------------------------------------------------
function private.QuestProgress ( quest_data, current_list )

  if ( quest_data.target and quest_data.target.list ) then
    
    return { current = #current_list, total = #quest_data.target.list };
  
  end;

end;
--------------------------------------------------------------------------------
function private.GetTargetList ( total, collected )

  local list = {}
  
    for _, item in pairs( total or {} ) do
    
      if ( not common.IsEntityInList( collected, item ) ) then
      
        table.insert( list, item );
      
      end;
    
    end;
  
  return list;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Highlight_FilterTargtets ( targets )

  local list = {};
  
  for _, target in pairs ( targets ) do
  
    -- Filter invisible stages
      if ( target.type == "stage" ) then
      
        local point = ui.obj.Get( "obj_map_point_"..target.name );
        if  ( point and point.visible ) then
        
          table.insert( list, target );
        
        end;
      
      else
      
        table.insert( list, target );
      
      end;
  
  end;
  
  return list;

end;
--------------------------------------------------------------------------------
function private.Highlight_Show ( focus_data )

  -- Filter invisible stages
  local targets = private.Highlight_FilterTargtets( focus_data.targets );
  focus_data.current = math.min( focus_data.current, #targets );

  local target = targets[ focus_data.current ];
  
  if ( focus_data.current >= #targets ) then
  
    focus_data.current = 1;
  
  else
  
    focus_data.current = focus_data.current + 1;
  
  end;
  
  local object, func_action;
  local scale  = 3.0;
  local bounce = false;
  
  if     ( target.type == "stage" ) then
  
    scale  = 1.5;
    bounce = true;
    object = "btn_map_point_"..target.name;
    
  
  elseif ( target.type == "fight" ) then
  
    object = "obj_map_enemy_"..target.name.."_"..target.stage;
    
    local btn_object = "btn_map_point_"..target.stage;
    local btn_params = ui.obj.Get( btn_object );
    
    if ( btn_params and btn_params.active and btn_params.input ) then
    
      func_action = ui.control.EventsGet( btn_object ).func_on_click;
    
    else
    
      func_action = ui.control.EventsGet( object ).func_on_click;
    
    end;
  
  elseif ( target.type == "build" ) then -- NOTE: market also here
  
    object = "obj_map_builds_"..target.name;
  
  elseif ( target.type == "mine"  ) then
  
    object = "obj_map_mines_"..target.name;
  
  elseif ( target.type == "spell" ) then
  
    object = "obj_map_spells_"..target.name;
  
  else
  
    return;
  
  end;
  
  -- Determinate func_action for non-constant objects
  if ( not func_action ) then
  
    local obj = object;
    local btn = "btn"..obj:sub( 4 );
    
    local events = ui.control.EventsGet( obj ) or ui.control.EventsGet( btn );
    if not ( events ) then return; end;
    
    func_action = events.func_on_click;
    
  end;
  
  public.Highlight_ShowHole( object, func_action, scale, bounce );

end;
--------------------------------------------------------------------------------
function public.Highlight_ShowHole ( object, func_action, scale, bounce )

  local obj_params = ObjGet( object );
  if ( not obj_params.visible ) then return; end;
  
  private.can_show_hole = true;
  
  common.sound.Play( "desire" );
  v_map.Focus( object, 
    {
    time       = 0.2,
    func_after = function ()
      
      if ( bounce ) then ui.anim.Bounce( object ); end;
      
      if ( private.can_show_hole ) then
        
        local anm_map = "anm_map_"..v_sections.Section_NameCurrent();
        local body_params = ObjGet( "obj_map_body" );
        local map_params  = GetObjPosByObj( anm_map );
        local obj_pos     = ui.obj.Position( object, anm_map );
        obj_params.pos_x  = map_params[1] + obj_pos[1] * body_params.scale_x;
        obj_params.pos_y  = map_params[2] + obj_pos[2] * body_params.scale_y;
        scale             = scale * body_params.scale_x;
        
        int_panel_darkhole.Show( obj_params, func_action, scale );
        
      end;
      
      private.can_show_hole = false;

    end
    }
  );

end;
--------------------------------------------------------------------------------
function private.Highlight_Hide ()

  private.can_show_hole = false;
  
  int_panel_darkhole.Hide();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.IsStageItemActive ( item_name, stage_name )

  local section_name = v_sections.Section_NameCurrent();
  
  -- Current quests
  local quests       = m_quests.Current_GetAllBySection( section_name );
  for _, quest in pairs( quests ) do
  
    if ( quest.type == "QT_StageItemsCollected" ) then
    
      local static_data = m_quests.Static_GetByName( quest.name );
      
      if ( static_data.target.name == item_name ) then
      
        local current_list = quest.list or {};
        local targets_list = static_data.target.list or {};
        
        if ( common.IsEntityInList( targets_list, stage_name ) ) then
        
          if ( common.IsEntityInList( current_list, stage_name ) ) then
          
            return false;
          
          end;
          
          return true;
        
        end;
      
      end;
    
    end;
  
  end;
  
  -- Pregive items quests
  local pregived = m_quests.Static_GetAllPregive();
  
  for _, quest in pairs( pregived ) do
  
    if ( quest.target.type == "QT_StageItemsCollected") then
    
      if ( not m_quests.Current_GetByName( quest.name )   ) and
         ( not m_quests.Completed_GetByName( quest.name ) ) then
    
        if ( common.IsEntityInList( quest.target.list, stage_name ) ) then
        
          return true;
        
        end;
    
      end;
    
    end;
  
  end;
  
  return false;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Quest_IsCurrent ( quest_name )
  
  return ( m_quests.Current_GetByName( quest_name ) ~= nil );
  
end;
--------------------------------------------------------------------------------
function public.Quest_IsComplete ( quest_name )
  
  return ( m_quests.Completed_GetByName( quest_name ) ~= nil );
  
end;
--------------------------------------------------------------------------------
function public.Quest_Data ( quest_name )
  
  return m_quests.Static_GetByName( quest_name );
  
end;
--------------------------------------------------------------------------------
function public.Quest_CurrentData ( quest_name )
  
  return m_quests.Current_GetByName( quest_name );
  
end;
--------------------------------------------------------------------------------
function public.Quest_GetNameByItem ( item_type, item_name, total_search )

  local section_name = v_sections.Section_NameCurrent();
  local quests;
  
  if ( total_search ) then
    quests = m_quests.Static_GetAllBySection( section_name );
  else
    quests = m_quests.Current_GetAllBySection( section_name );
  end;

  for i, quest in pairs( quests ) do
  
    if ( quest.reward ) then
    
      for r, reward in pairs( quest.reward ) do

        local same_type = ( reward.type == item_type );
        local same_name = ( reward.name == item_name );

        if ( same_type and same_name ) then 

          return quest.name;

        end;

      end;
    
    end;

  end;

  return false;
  
end;
--------------------------------------------------------------------------------
function public.GetCompleteQuestCount ()
  
  local quest_count = m_quests.Completed_GetCount();

  return quest_count;
  
end;
--------------------------------------------------------------------------------
function public.GetQuestsOnStageComplete ( stage_name )

  local quests = m_quests.Static_GetAllByStartStage( stage_name );

  return quests;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Questlist_IsActive ()

  return false;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Stages_FindByReward ( reward_type, reward_name )

  local section = v_sections.Section_NameCurrent();
  return m_quests.Stages_FindByReward( section, reward_type, reward_name );

end;
--------------------------------------------------------------------------------