-- name=v_sections
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_sections"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  common.mod.Load( _modules );

  private.feature = "sections_enable";
  private.section = { current = {} };
  
  ne.msg.Subscribe( "Event_Feature_Updated", private.OnFeatureUpdated );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Init_Subscribes ()

  --

end;
--------------------------------------------------------------------------------
function public.SaveFeatureState ()

  if ( v_game.IsOnline() ) then

    --Log.i( "SaveFeatureState" );
    private.OnFeatureUpdated();
    
  end;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Cheat_Unlock ( section_name )
  
  local is_enabled = private.IsEnabled( section_name );
  m_sections.Section_Unlock( section_name, is_enabled );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.IsFeatureActive ()

  local  is_active = ( parts.features.IsActive( private.feature ) or false );
  return is_active;

end;
--------------------------------------------------------------------------------
function public.GetEnabledList ()
  
  local  enabled_list = private.GetEnabledList();
  return enabled_list;
  
end;
--------------------------------------------------------------------------------
function public.List ()

  local  section_list = m_sections.List();
  return section_list;

end;
--------------------------------------------------------------------------------
function public.StaticList ()

  local  static_list = m_sections.StaticList();
  return static_list;

end;
--------------------------------------------------------------------------------
function public.ListEnabledWithPlayableProperty ()

  local static_list      = m_sections.StaticList();
  local discovered_count = public.DiscoveredCount();
  local enabled_list     = private.GetEnabledList();
  local result_list      = {};

  for index, section in pairs( static_list ) do

    if ( private.IsEnabled( section.name ) ) then

      local section_data = {
        number   = tonumber( section.name:match( "%d+" )),
        name     = section.name,
        playable = ( index <= discovered_count )
      };

      table.insert( result_list, section_data );

    end;

  end;

  return result_list;

end;
--------------------------------------------------------------------------------
function public.DetailedList ()

  local list = {};

  local static_list  = m_sections.StaticList();
  local enabled_list = private.GetEnabledList();
  
  for i = 1, #static_list do
    
    local section = static_list[ i ];
    
    if ( common.IsEntityInList( enabled_list, section.name ) ) then
      
      local saved = m_sections.Section_StorageData( section.name ) or {};
      
      for key, value in pairs( saved ) do
        section[ key ] = value;
      end;
      
      local stages = v_stage.Stage_GetAllBySection( section.name ) or {};
      
      section.stages_count = #stages;
      section.stages_done  = 0;
      
      for _, stage in pairs( stages ) do
      
        local stage_data = v_stage.Stage_StorageData( stage.name ) or {};
      
        if ( stage_data.completed ) then
          section.stages_done = section.stages_done + 1;
        end;
      
      end;
      
      table.insert( list, section );
      
    end;
  
  end;

  return list;

end;
--------------------------------------------------------------------------------
function public.SetCurrentNumber ( section_number )
  
  private.section.current.number = section_number;

  local section = {};

  section.current_name = public.Section_NameCurrent();
  section.next_name    = public.Section_NameNext( section.current_name );

  mediator.Notify(
    "View_Sections_ConfigLoaded",
    {
      section_name   = "section_"..section_number,
      section_config = m_sections.Section_GetConfig( section.current_name )
    }
  );
  
end;
--------------------------------------------------------------------------------
function public.GetCurrentNumber ()
  
  return private.section.current.number;
  
end;
--------------------------------------------------------------------------------
function public.DiscoveredCount ()
  
  local section_list = m_sections.List();  
  return #section_list;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Section_GetConfig ( section_name )

  section_name         = section_name or public.Section_NameCurrent();
  local section_config = m_sections.Section_GetConfig( section_name );

  return section_config;
  
end;
--------------------------------------------------------------------------------
function public.Section_IsEnabled ( section_name )

  local  is_enabled = m_sections.Section_IsEnabled( section_name );
  return is_enabled;
  
end;
--------------------------------------------------------------------------------
function public.Section_IsPlayable ( section_number )

  local discovered_count = public.DiscoveredCount();
  local is_playable      = ( section_number <= discovered_count );

  return is_playable;
  
end;
--------------------------------------------------------------------------------
function public.Section_IsLastAvailable ( section_number )
  
  local  is_last_available = ( section_number == #private.GetEnabledList() );
  return is_last_available
  
end;
--------------------------------------------------------------------------------
function public.Section_NameByStage ( stage_name )
  
  local section_name = m_sections.Section_NameByStage( stage_name );
  
  return section_name;
  
end;
--------------------------------------------------------------------------------
function public.Section_GetNameByNumber ( section_number )
  
  return "section_"..section_number;
  
end;
--------------------------------------------------------------------------------
function public.Section_Status ( next_section_name )

  local section_data = m_sections.Section_StorageData( next_section_name );

  if ( section_data ) then

    return tern( section_data.enabled )( "enabled", "available" );
    
  end;

  -- JK: check portal blocker in current section
  local current_section_name = public.Section_NameCurrent();

  local portal_blocker =
    m_sections.Section_StaticData( current_section_name ).portal_blocker;
  local portal_prepare =
    m_sections.Section_StaticData( current_section_name ).portal_prepare;
    
  local state;
   
  if ( portal_prepare ) then
    
    if ( portal_prepare.type == "stage" ) then
    
      if ( v_stage.Stage_IsCompleted( portal_prepare.name ) ) then
      
        state = "prepare";
      
      end;

    end;
    
  end;

  if ( portal_blocker ) then

    if ( portal_blocker.type == "stage" ) then
    
      if ( v_stage.Stage_IsCompleted( portal_blocker.name ) ) then
      
        state = "prepare";
      
      end;

    end;

    if ( portal_blocker.type == "build" ) then
    
      if ( v_builds.Build_IsConstructed( portal_blocker.name ) ) then
      
        state = "prepare";
      
      end;

    end;

    if ( portal_blocker.type == "quest" ) then
      
      if ( v_quests.Quest_IsComplete( portal_blocker.name ) ) then
      
        state = "prepare";
      
      end;

    end;
    
  end;

  return state;
  
end;
--------------------------------------------------------------------------------
function public.Section_EnableNext ()
  
  local sections      = m_sections.List();
  local enabled_list  = private.GetEnabledList();
  local enabled_count = m_sections.GetEnabledCount();

  for s = 1, #sections do

    if ( s <= enabled_count                                        ) and
       ( common.IsEntityInList( enabled_list, sections[ s ].name ) ) and
       ( not sections[ s ].enabled                                 ) then

      m_sections.Section_Enable( sections[ s ].name );

    end;

  end;
  
end;
--------------------------------------------------------------------------------
function public.Section_NameNext ( section_name )

  section_name = section_name or public.Section_NameCurrent();
  
  local section_number    = tonumber( string.match( section_name, "%d+" ) );
  local next_section_name = "section_"..( section_number + 1 );

  return next_section_name;

end;
--------------------------------------------------------------------------------
function public.Section_NamePrev ( section_name )

  local section_number = tonumber( string.match( section_name, "%d+" ) );
  local prev_number    = section_number - 1;

  if ( prev_number > 0 ) then

    return "section_"..prev_number;

  end;

  return false;

end;
--------------------------------------------------------------------------------
function public.Section_NameLast ()

  local last_stage_name   = v_player.Progress_GetLastStageName();
  local last_section_name = m_sections.Section_NameByStage( last_stage_name );

  return last_section_name;

end;
--------------------------------------------------------------------------------
function public.Section_NameCurrent ()
  
  local section_number, real_current = public.Section_NumberCurrent();
  
  return "section_"..section_number, "section_"..real_current;

end;
--------------------------------------------------------------------------------
function public.Section_NumberLast ()
  
  local last_section_name   = public.Section_NameLast();
  local last_section_number = string.match( last_section_name, "%d+" );

  return tonumber( last_section_number );

end;
--------------------------------------------------------------------------------
function public.Section_NumberCurrent ()
  
  local section_number = private.section.current.number or 
                         public.Section_NumberLast();
                         
  local real_current = section_number;
                         
  if ( not private.IsEnabled( "section_"..section_number ) ) then
    
    local enabled = private.GetEnabledList();
    
    for i = #enabled, 1, - 1 do
      
      local section_name = enabled[ i ];
      
      if ( public.Section_IsEnabled( section_name ) ) then
        
        section_number = tonumber( section_name:match( "%d+" ) );
        
        break;
        
      end;
      
    end;
    
  end;
  
  return section_number, real_current;

end;
--------------------------------------------------------------------------------
function public.Section_NeedIntro ( section_name )

  section_name = section_name or public.Section_NameCurrent();
  
  local is_intro_needed = m_sections.Section_NeedIntro( section_name );
  
  return is_intro_needed;

end;
--------------------------------------------------------------------------------
function public.Section_SetIntro ( section_name, value )

  local is_set = m_sections.Section_SetIntro( section_name, value );
  
  return is_set;

end;
--------------------------------------------------------------------------------
function public.Section_NeedOutro ( section_name )

  section_name = section_name or public.Section_NameCurrent();

  local is_outro_needed = m_sections.Section_NeedOutro( section_name );

  return is_outro_needed;

end;
--------------------------------------------------------------------------------
function public.Section_SetOutro ( section_name, value )

  m_sections.Section_SetOutro( section_name, value );

end;
--------------------------------------------------------------------------------
function public.Section_FirstStage ( section_name )

  section_name     = section_name or public.Section_NameCurrent();
  local stage_name = m_sections.Section_FirstStage( section_name );
  
  return stage_name;

end;
--------------------------------------------------------------------------------
function public.Section_LastStage ( section_name, mainline_only )

  section_name     = section_name or public.Section_NameCurrent();
  local stage_name = m_sections.Section_LastStage(
    section_name,
    mainline_only
  );
  
  return stage_name;

end;
--------------------------------------------------------------------------------
function public.Section_StaticData ( section_name )
  
  local static_data = m_sections.Section_StaticData( section_name );
  
  return static_data;
  
end;
--------------------------------------------------------------------------------
function public.Section_ReadyList ()
  
  local sections     = {};
  local list         = m_sections.List();
  local list_enabled = private.GetEnabledList();
  
  for i = 1, #list do
    if ( not ( list[ i ].intro or list[ i ].outro ) and 
       ( common.IsEntityInList( list_enabled, list[ i ].name ) ) ) then
      table.insert( sections, list[ i ] );
    end;
  end;
  
  return sections;
  
end;
--------------------------------------------------------------------------------
function public.Section_CheckPrevious ( section_number )
  
  local prev_section_number = section_number - 1;
  local data = m_sections.Section_StorageData( "section_"..prev_section_number );
  
  if ( data ) then
    
    if ( not data.outro ) then
      
      public.Section_SetOutro( "section_"..prev_section_number, true );
      
    end;
    
  end;
  
end;
--------------------------------------------------------------------------------
function public.Section_SetLoaded ( section_number )
  
  private.section_loaded = section_number;
  
end;
--------------------------------------------------------------------------------
function public.Section_GetLoaded ()
  
  return private.section_loaded;
  
end;
--------------------------------------------------------------------------------
function public.Stage_GetLastEnabled ()
  
  local  stage_name = m_sections.Stage_GetLastEnabled();
  return stage_name;
  
end;
--------------------------------------------------------------------------------
-- function ...() end;
--------------------------------------------------------------------------------
function public.Player_GetEnabledList ()
  
  local  enabled_list = m_sections.Player_GetEnabledList();
  return enabled_list;

end;
--------------------------------------------------------------------------------
-- function ...() end;
--------------------------------------------------------------------------------
function public.Portal_Name ( section_name )

  local  portal_name = m_sections.Section_StaticData( section_name ).portal;
  return portal_name;

end;
--------------------------------------------------------------------------------
function public.Portal_TryUnlock ( stage_name )
  
  if ( wiz.IsCheatsEnabled() ) then
    if ( lib_db.Database_IsFreeze() ) then
      v_interface.ShowNotification( "Can't unlock portal because of freezing game progress." );
      return;
    end;
  end;
  
  local section_name = public.Section_NameNext();
  local section_data = m_sections.Section_StorageData( section_name );
  
  -- JK: must be already unlocked, do nothing
  if ( section_data ) then return false; end;

  local unlock = m_sections.Section_StaticData( section_name ).unlock;

  if ( unlock.type == "stage" ) then
    
    if ( unlock.name ~= stage_name ) then

      return false;
      
    end;
    
  elseif ( unlock.type == "quest" ) then
      
    if ( not v_quests.Quest_IsComplete( unlock.name ) ) then

      return false;

    end;

  elseif ( unlock.type == "build" ) then
      
    if ( not v_builds.Build_IsConstructed( unlock.name ) ) then

      return false;

    end;

  else

    return false;
    
  end;
  
  if ( IsPC() ) then
  
    local section_number = tonumber( string.match( section_name, "%d+" ) );
    
    v_game.Section_SetAllowedCount( section_number );
    
  end;
  
  local is_enabled = private.IsEnabled( section_name );
  m_sections.Section_Unlock( section_name, is_enabled );

  v_social.FacebookShare_SetShareWorld();
  
  v_log.Playhaven_LogPlacement( "land_uncursed" );
  
  v_log.Kontagent_LogEvent(
    "progress", 
    {
      name = public.Section_NamePrev( section_name )
    }
  );

  return { type = "PortalUnlocked" };
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.OnFeatureUpdated ()

  if ( not public.IsFeatureActive() ) then return; end;

  local data     = parts.features.GetFeatureData( private.feature ) or { {} };
  local new_list = data[ 1 ].enabled;

  if ( not new_list ) then return; end;
  
  local old_list = v_features.GetFeatureData( private.feature );
  
  if ( old_list ) then

    local enabled  = {};
    local disabled = {};

    -- JK: check for added sections
    for i = 1, #new_list do

      local is_enabled = true;

      for j = 1, #old_list do
        if ( new_list[ i ] == old_list[ j ] ) then
          is_enabled = false; break;
        end;
      end;

      if ( is_enabled ) then
        table.insert( enabled, new_list[ i ] );
      end;

    end;

    -- JK: check for removed sections
    for j = 1, #old_list do

      local is_disabled = true;

      for i = 1, #new_list do
        if ( old_list[ j ] == new_list[ i ] ) then
          is_disabled = false; break;
        end;
      end;

      if ( is_disabled ) then
        table.insert( disabled, old_list[ j ] );
      end;

    end;

    if ( #enabled > 0 ) then
      mediator.Notify( "View_Sections_EnabledList", enabled );
    end;

    if ( #disabled > 0 ) then
      mediator.Notify( "View_Sections_DisabledList", disabled );
    end;

  end;
  
  v_features.SaveFeatureData( private.feature, new_list );
  
  private.TryEnableSections( new_list );
  
  local new_section_number, current_section_number =
    v_sections.Section_NumberCurrent();

  if ( new_section_number ~= current_section_number ) then
    mediator.Notify( "View_Sections_CurrentDisabled" );
  end;
  
end;
--------------------------------------------------------------------------------
function private.IsEnabled ( section_name )
  
  local enabled_list = private.GetEnabledList();

  for i = 1, #enabled_list do
    
    if ( enabled_list[ i ] == section_name ) then
      
      --Log.i( "IsEnabled: "..tostring( section_name ) );
      
      return true;
      
    end;
    
  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
function private.GetDefaultEnabledList ()
  
  -- Returns default value from logic_settings.
  if ( not private.default_enabled ) then
    
    private.default_enabled = {};
    
    local enabled_count = m_sections.GetEnabledCount();
    
    for i = 1, enabled_count do
      table.insert( private.default_enabled, "section_"..i );
    end;
    
  end;
  
  return private.default_enabled;
  
end;
--------------------------------------------------------------------------------
function private.GetEnabledList ()
  
  local enabled = {};
  
  if ( public.IsFeatureActive() ) then
    
    --Log.i( "GetEnabledList: get server data." );
    enabled = ( parts.features.GetFeatureData( private.feature ) or { {} } )[ 1 ].enabled;
    
  else
    
    --Log.i( "GetEnabledList: get saved data." );
    enabled = v_features.GetFeatureData( private.feature );
    
  end;
  
  if ( enabled ) then
    local enabled_count = m_sections.GetEnabledCount();
    for s = #enabled, enabled_count + 1, - 1 do
      
      table.remove( enabled, s );
      
    end;
    return enabled;
    
  end;

  return private.GetDefaultEnabledList();
  
end;  

--------------------------------------------------------------------------------
function private.TryEnableSections ( enabled_list )

  local saved_list = m_sections.List();
  
  for s = 1, #saved_list do
    
    local section = saved_list[ s ];
    local enabled_count = m_sections.GetEnabledCount();
    
    if ( s <= enabled_count                                  ) and
       ( not section.enabled                                 ) and
       ( common.IsEntityInList( enabled_list, section.name ) ) then
      
      m_sections.Section_Enable( section.name );
      
    end;
    
  end;

end;
--------------------------------------------------------------------------------