-- name=v_settings
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "int_dialog_options",
  "int_dialog_language"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.this = {
    dialog = {
      options  = "int_dialog_options",
      language = "int_dialog_language"
    }
  };

  common.mod.Load( _modules );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Init ()

  if ( private.googleplay_data ) then

    private.DialogOptions_SetupGooglePlay();

  end;

  ne.msg.Subscribe(
    "Event_Application_ToggleFullScreen",
    function ( msg, params )

      private.Toggle_FullScreen(params);

    end
  );

  ne.msg.Subscribe(
    "Command_OptionsMenuShow",
    function ( msg )

      if ( v_interface.IsVisible() ) then

        private.DialogOptions_Show();

      end;

    end
  );

  mediator.Subscribe(
    "Model_Social_LoginComplete",
    int_dialog_options.OnLogin
  );
  mediator.Subscribe(
    "Model_Social_ProfileComplete",
    int_dialog_options.OnLogin
  );

  v_interface.window.RegisterAsModal( private.this.dialog.language );

  common.delegate.Set( "int_panel_interface_bottom",
  {
    OptionsClickHandler = function ()

      public.Dialog_Show();

    end
  } );

  v_interface.Scheme_AddButton( "map", "bottom", { name = "options" } );

  private.initialized = true;

  --Log.i( "Initialized." );

end;
--------------------------------------------------------------------------------
function public.Dialog_Show ( func_on_hide )

  private.DialogOptions_Show( func_on_hide );

end;
--------------------------------------------------------------------------------
function public.GooglePlay_Setup ( is_connected, func_action )

  private.googleplay_data = {
    is_connected = is_connected,
    func_action  = func_action
  };

  if ( private.initialized ) then

    private.DialogOptions_SetupGooglePlay();

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.DialogOptions_Show ( func_on_hide )

  --- Fullscreen toggle
  local func_fullscreenToggle = function ()

    local is_fullscreen = GetFullScreen();

    SetFullScreen( not is_fullscreen );
    return not is_fullscreen;

  end;

  --- Widescreen toggle
  local func_widescreenToggle = function ()

    local func_ws_yes = function()

      v_game.dialog.Info_Hide();
      SetWideScreen( not GetWideScreen() );

    end;

    local func_ws_no = function()

      v_game.dialog.Info_Hide();

    end;

    local current_screen = v_screen.GetCurrentName();

    local message_type = tern( current_screen:find( "map" ) )(
      "widescreen_map",
      "widescreen_stage"
    );

    if ( current_screen == "scr_stage" ) then
      local stage_data = v_stage.Stage_GetCurrentData();
      if ( stage_data.name == "stage_1" ) then
        message_type = "widescreen_map";
      end;
    end;

    v_game.dialog.Info_Show( message_type, func_ws_yes, func_ws_no );

  end;

  --- Cursor toggle
  local func_cursorToggle = function ()

    SetSysCursor( not GetSysCursor() );
    return not GetSysCursor();

  end;

  --- Notification toggle
  local func_notificationToggle = function ()

    local enabled = v_player.GetNotification();
    return v_player.EnableNotification( not enabled );

  end;

  --- Quit
  local func_Quit = function ()

    v_game.dialog.Info_Show(
      "exitconfirm",
      wiz.Quit,
      v_game.dialog.Info_Hide
    );

  end;

  -- Is online
  local func_IsOnline = function ()

    if ( v_game.IsOnline() ) then return true; end;

    v_game.SystemDialog_Show( "connectionlost", function () end );
    return false;

  end;

  local delegates = {

    Language             = private.Language_Show,
    IsLogged             = v_social.IsLogged,
    Login                = function () v_social.Login( "options_menu" ); end,
    Logout               = function ()
                             v_log.Playhaven_LogPlacement( "facebook_manage" );
                             v_social.Logout( "options_menu" );
                           end,
    GetSocialInfo        = m_social.GetSocialInfo,
    RestorePurchases     = private.RestorePurchases_Start,
    Close                = function () int_dialog_options.Hide( func_on_hide ); end,
    Quit                 = func_Quit,
    IsOnline             = func_IsOnline,
    musicGetState        = common.music.IsEnabled,
    musicToggle          = function ()
                            ne.snd.PauseAll( false );
                            local is_enabled= common.music.Toggle();
                            return is_enabled;
                           end,
    soundGetState        = common.sound.IsEnabled,
    soundToggle          = function ()
                            ne.snd.PauseAll( false );
                            local is_enabled = common.sound.Toggle();
                            return is_enabled;
                           end,
    fullscreenGetState   = function () return GetFullScreen(); end,
    fullscreenToggle     = func_fullscreenToggle,
    widescreenGetState   = function () return GetWideScreen(); end,
    widescreenToggle     = func_widescreenToggle,
    cursorGetState       = function () return not GetSysCursor(); end,
    cursorToggle         = func_cursorToggle,
    notificationGetState = v_player.GetNotification,
    notificationToggle   = func_notificationToggle,
    supportShow          = function ()
                              if ( IsMobile() ) then
                                
                                wiz.bfg.SetReportingUserId( v_player.UserId() );
                                
                                -- JK: off sound on iOS
                                if ( IsIOS() ) then
                                  
                                  ne.snd.PauseAll( true );
                                  
                                  --[[local is_music_toggled = false;
                                  local is_sound_toggled = false;
                                  
                                  if ( common.music.IsEnabled() ) then
                                    is_music_toggled = true;
                                    common.music.Toggle();
                                  end;
                                  
                                  if ( common.sound.IsEnabled() ) then
                                    is_sound_toggled = true;
                                    common.sound.Toggle();
                                  end;]]
                                  
                                  --ne.msg.Subscribe()
                                
                                  mediator.SubscribeOnce(
                                    "Interface_Window_WindowClosed",
                                    function ()
                                      
                                      ne.snd.PauseAll( false );
                                      
                                      --[[if ( is_music_toggled ) then
                                        common.music.Toggle();
                                      end;
                                      
                                      if ( is_sound_toggled ) then
                                        common.sound.Toggle();
                                      end;]]
                                      
                                    end
                                  );
                                  
                                end;
                                
                                ShowSupport();
                                
                              else
                                OpenUrl( "https://www.bigfishgames.com/help/sirmatchalot" );
                              end;
                           end
  };

  v_interface.window.Show(
    "int_dialog_options",
    int_dialog_options.Show,
    nil,
    delegates
  );

end;
--------------------------------------------------------------------------------
function private.DialogOptions_SetupGooglePlay ()

  common.delegate.Set(
    "int_dialog_options",
    {
      GooglePlayAction = private.googleplay_data.func_action
    }
  );

  int_dialog_options.GooglePlay_Setup( private.googleplay_data.is_connected );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.RestorePurchases_Start ()

  if ( private.restore_purchases_started ) then

    v_game.SystemDialog_Show( "purchases_restore_empty", function () end );

    return;

  end;

  if ( wiz.bfg.lot.RestorePurchases ) then

    ne.msg.Subscribe(
      "Event_Purchase_Restore",
      private.RestorePurchases_OnFinished
    );

    mediator.Notify( "View_Settings_RestorePurchases", { status = "start" } );

    wiz.bfg.lot.RestorePurchases();

    private.restore_purchases_started = true;

  else

    v_game.SystemDialog_Show( "purchases_restore_empty", function () end );

  end;

end;
--------------------------------------------------------------------------------
function private.RestorePurchases_OnFinished ( msg, params )

  ne.msg.Unsubscribe(
    "Event_Purchase_Restore",
    private.RestorePurchases_OnFinished
  );

  mediator.Notify( "View_Settings_RestorePurchases", { status = "finish" } );
  if ( params.result == "failed" ) then
    v_game.SystemDialog_Show( "purchases_restore_failed", function () end );
  elseif ( params.result == "empty" ) then
    v_game.SystemDialog_Show( "purchases_restore_empty", function () end );
  else
    v_game.SystemDialog_Show( "purchases_restored_ok", function () end );
  end;
  private.restore_purchases_started = nil;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Toggle_FullScreen ( params )

  local togglefullscreen = params.togglefullscreen;
  local togglewidescreen = params.togglewidescreen;

  if ( togglefullscreen == false ) and
     ( togglewidescreen == false ) then

    Log.e( "Failed to change screen mode" );

    v_game.dialog.Info_Show( "togglescreenmode" );

  elseif ( togglewidescreen == true ) then

    ObjSet( "ng_application", { color_r = 0.0, color_g = 0.0, color_b = 0.0 } );
    common.Return( true );

  else

    int_dialog_options.OnFullScreenToggle( params );

  end;

end;
--------------------------------------------------------------------------------
function private.Language_Show ()

  local language = wiz.settings.GetLanguage();
  local list = { "fr", "it", "de", "es", "ja", "ru", "nl", "pt", "ko", "en" };

  local func_select = function ( language )

    v_game.dialog.Info_Show(
      "languageconfirm",
      function ()
        wiz.settings.SetLanguage( language );
        common.Return( true );
      end
    )

  end;

  v_interface.window.Show(
    "int_dialog_language",
    function () int_dialog_language.Show( language, list ); end,
    nil,
    { Select = func_select }
  );

end;
--------------------------------------------------------------------------------
