-- name=v_shop
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_shop",
  "int_dialog_shop",
  "int_dialog_starterpack",
  "int_dialog_minibank"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  common.mod.Load( _modules );

  private.dialog_notify = false;
  private.feature = "prices";
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Init ()
  
  v_interface.window.RegisterAsModal( "int_dialog_minibank" );

  private.SetDialogHandlers();

  v_game.RegisterFeature( 
  { 
    entity = "v_shop", 
    feature = private.feature,
    func_save = private.Feature_Save
  } );
  
  private.Subscribe();

  private.performed_restore_purchases = true;
  
  wiz.bfg.lot.InitializePurchases();

  private.performed_restore_purchases = false;
  
  --local is_sale = GetSaleInfo().sale;  -- Sales on PC

  local starterpack = private.Starterpack_Data();
  private.Starterpack_Init( starterpack.name );

  private.PanelButton_Enable();
  private.PanelButton_UpdateSaleNotification();

  --Log.i( "Initialized." );

end;
--------------------------------------------------------------------------------
function private.PanelButton_Enable ()
  
  common.delegate.Set( "int_panel_interface_bottom", 
  {
    ShopClickHandler = function ()

      public.Show();
      common.sound.Play( "bottom_button" );

    end;
  } );
  
  v_interface.Scheme_AddButton( "map", "bottom", { name = "shop" } );
  private.CheckButtonNotification();
  
end;
--------------------------------------------------------------------------------
function public.Notify ( flag )

  private.dialog_notify = flag;

end;
--------------------------------------------------------------------------------
function private.Subscribe ()

  mediator.Subscribe(
    "Model_Player_GoldCountChangeFailed",
    private.Gold_OnDeficient
  );
  
  mediator.Subscribe(
    "Model_Player_ResourceCountChangeFailed",
    private.Resources_OnDeficient
  );

  ne.msg.Subscribe( Event_Purchase_Finished, private.Purchase_OnFinished );
  ne.msg.Subscribe( Event_Purchase_Disabled, private.Purchase_OnDisable );
  ne.msg.Subscribe( 
    "Event_Purchase_ProductInformationReady",
    private.OnProductInformationReady
  );

  mediator.Subscribe( m_shop.msg.MarketRefresh, function()
 
    if ( int_dialog_shop.IsActive() ) then
      
      if ( int_dialog_shop.CurrentTab() == "market" ) then

        int_dialog_shop.Market_Sell_SetDefaults();

      else
        
        int_dialog_shop.Market_SetNotification( true );
        
      end;
    
    else
    
      private.CheckButtonNotification();
      int_dialog_shop.Market_SetNotification( true );
    
    end;
 
  end );

  mediator.Subscribe(
    "Model_Shop_StarterpackStart",
    private.Starterpack_OnStart
  );

  mediator.Subscribe(
    "Model_Shop_StarterpackFinish",
    private.Starterpack_OnFinish
  );

  mediator.Subscribe(
    "Model_Potion_RunAction",
    private.HandleActionSaleBoosters
  );
  
  mediator.Subscribe(
    "Model_Potion_CompleteAction",
    private.HandleActionSaleBoosters
  );
  
  mediator.Subscribe(
    "View_Settings_RestorePurchases",
    private.CheckRestorePurchases
  );
  
end;
--------------------------------------------------------------------------------
function private.SetDialogHandlers ()

  local delegates = {

    BuyGold = function ( lot_name )

      if ( v_game.IsOnline() ) then

        local lot_id = m_shop.Lot_GetIOSName( lot_name );

        private.PurchaseLot( lot_id );

      else

        v_game.TryReconnect( false );
        v_game.SystemDialog_Show( "nointernetconnection", function () end );
        mediator.Notify( "View_Shop_Purchase_Failed" );

      end;

    end,

    BuyBooster = function ( booster_name, pack_level )

      if ( v_boosters.TryBuy( booster_name, pack_level, "shop" ) ) then

         common.sound.Play( "buy" );
        return true;

      end;

      return false;

    end,

    TapBooster = function ( booster_name ) 

      return v_boosters.Tap( "shop", booster_name );

    end,

    MarketShareValue          = private.Market_Share_Value,
    MarketSharePrice          = private.Market_Share_Price,
    MarketShareLimit          = private.Market_Share_Limit,
    MarketGetSharesSell       = private.Market_GetShares_Sell,
    MarketTryBuy              = private.Market_TryBuy,
    MarketTrySell             = private.Market_TrySell,
    MarketGetSecondsToRefresh = private.Market_GetSecondsToRefresh,
    PanelLock                 = v_interface.panel.top.Lock,
    PanelUnlock               = v_interface.panel.top.Unlock,

    OnTabOpen = function ( tab_name )
    
      local events = 
      {
        gold     = "shop_enter",
        boosters = "boosts_enter",
        --market   = "market_enter"
      };
      
      if ( events[ tab_name ] ) then
        
        v_log.Playhaven_LogPlacement( events[ tab_name ] );
        
      end;

      mediator.Notify( "View_Shop_TabOpened", tab_name );

    end,

    OnTabClose = function ( tab_name )

      local events = 
      {
        gold     = "shop_close",
        boosters = "boosts_close",
        --market   = "market_close"
      };
      
      if ( events[ tab_name ] ) then
        
        v_log.Playhaven_LogPlacement( events[ tab_name ] );
        
      end;

    end,

    IsMarketAvailable = function ()

      if ( v_screen.GetCurrentName():find( "minievent" ) ) then
        return false, "block";
      end;
      
      local is_built =
        ( m_shop.Market_IsUnlocked() and m_shop.Market_IsBuilt() );
      
      return is_built, "unlock";

    end,

    --DialogClose = private.Hide

  };

  common.delegate.Set( "int_dialog_shop", delegates );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Show ( tab, func_on_close )

  tab = tab or "gold";
  
  local answer = GetLotsInfo();

  local purchase_lots = nil;

  if ( answer ) then purchase_lots = answer.lots; end;

  local gold_lots = m_shop.Gold_GetLots( "coins" );

  if ( purchase_lots ) and ( #purchase_lots > 0 ) then

    for p = 1, #purchase_lots do

      local purchase_lot = purchase_lots[ p ];

      for i = 1, #gold_lots do

        local lot = gold_lots[ i ];

        if ( lot.name_ios == ( purchase_lot.name_ios or purchase_lot.name ) ) then

          lot.price = private.Lot_GetPrice( purchase_lot );

          break;

        end;

      end;

    end;

  end;

  private.Hearts_TryAddImmortality( gold_lots );

  local boosters_list = v_boosters.GetLots();
  
  v_game.TryReconnect( false );
  
  local func_close = function ()
    
    private.CheckButtonNotification();
    private.Hide( func_on_close );
    
  end;

  v_interface.window.Show(
    "int_dialog_shop",
    function ()

      int_dialog_shop.Show( tab, gold_lots, boosters_list, func_close );

    end
  );

end;
--------------------------------------------------------------------------------
function private.Hide ( func_on_close )
  
  int_dialog_shop.Hide( func_on_close );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Lot_GetDataByIOSName ( lot_id )
  
  local lot = m_shop.Lot_GetDataByIOSName( lot_id );
  
  local answer = GetLotsInfo();

  if ( answer and answer.lots ) then
    
    local purchase_lots = answer.lots;
    
    for p = 1, #purchase_lots do

      local purchase_lot = purchase_lots[ p ];
      
      if ( lot.name_ios == ( purchase_lot.name_ios or purchase_lot.name ) ) then
        
        lot.price = private.Lot_GetPrice( purchase_lot );

        break;
        
      end;

    end;
    
  end;
  
  return lot;
  
end;
--------------------------------------------------------------------------------
function private.Lot_GetPrice ( lot_data )
  
  if ( IsPC() ) then
    
    local currency = {
      JPY = "(.*) .",
      USD = ". (.*)",
      EUR = "(.*) ."
    };
    
    local price = ne.strings.Get( "str_dialog_shop_gold_price_"..string.lower( lot_data.currency ), 
                  lot_data.price:match( currency[ lot_data.currency ] ) );

    return price;
    
  end;
  
  return lot_data.price or "";

end;
--------------------------------------------------------------------------------
function private.Lot_GetResourceExtras ( lot_name )

  local feature_data = parts.features.GetFeatureData( private.feature );
  
  local resource_extras = {
    resource1 = 0,
    resource2 = 0,
    resource3 = 0
  };

  if ( feature_data           ) and
     ( feature_data[ 1 ]      ) and
     ( feature_data[ 1 ].lots ) then
       
    local lots  = feature_data[ 1 ].lots;
    local index = common.IsEntityInList( lots, { name = lot_name }, "name" );

    if ( index ) then
      for key, _ in pairs( resource_extras ) do
        resource_extras[ key ] = lots[ index ][ key ] or 0;
      end;
    end;
    
  end;

  return resource_extras;

end;
--------------------------------------------------------------------------------
function private.Lot_ApplyResourceExtras ( lot_name )

  local resource_extras = private.Lot_GetResourceExtras( lot_name );
  local resource_data   = {};

  for key, value in pairs( resource_extras ) do
    if ( value > 0 ) then
      table.insert( resource_data, { name = key, delta = value } );
    end;
  end;

  v_player.Resource_TryChangeMultipleCount( resource_data );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.PurchaseLot ( lot_id )

  mediator.Notify(
    "View_Shop_PurchaseStart",
    lot_id
  );

  private.connection_pending = true;
  private.purchase_started = true;
  
  ui.screen.Waiting_Show();
  
  private.LogPurchase( lot_id, "start" );

  wiz.bfg.lot.Buy( lot_id );

end;
--------------------------------------------------------------------------------
function private.Purchase_OnFinished ( msg, params )

  if ( private.connection_pending ) then

    private.connection_pending = false;

    ui.screen.Waiting_Hide();

  end;

  if ( private.Purchase_IsRegistered( params.id ) ) then

    return false;

  end;

  local is_purchased = private.Premium_TryBuyLot( params );

  if ( is_purchased ) then

    -- JK: track last purchase UNIX time here
    local seconds = parts.time.server.Timestamp();
    local is_set  = v_player.PayingStatus_SetLastPurchaseTime( seconds );

    if ( not is_set ) then

      DbgTrace( "[ ERROR ][ v_shop ] Failed to set last purchase time." );

    end;

  end;

  mediator.Notify(
    "View_Shop_PurchaseFinish",
    params
  );

end;
--------------------------------------------------------------------------------
function private.Purchase_OnDisable ( msg, params )

  private.connection_pending = false;

  local str   = "str_systemmessage_inapp_purchase_disabled_";

  wiz.ShowSystemMessage(
  {
    title   = StringGet( str.."title" ),
    body    = StringGet( str.."body" ),
    buttons =
    {
      {
        title  = StringGet( str.."button1" ),
        action = function () ui.screen.Waiting_Hide(); end;
      }
    }
  } );

end;
--------------------------------------------------------------------------------
function private.Purchase_IsNonConsumable ( lot_id )

  if ( not lot_id ) then return false; end;

  return ( lot_id:find( "life" ) or 
           lot_id:find( "live" ) or 
           lot_id:find( "speedy" ) );

end;
--------------------------------------------------------------------------------
function private.Purchase_Register ( lot_id )

  if ( not lot_id ) then return false; end;

  if ( private.Purchase_IsNonConsumable( lot_id ) ) then

    m_shop.Purchase_Register( lot_id );

  end;

end;
--------------------------------------------------------------------------------
function public.Purchase_IsRegistered ( lot_id )

  return private.Purchase_IsRegistered ( lot_id )

end;
--------------------------------------------------------------------------------
function private.Purchase_IsRegistered ( lot_id )

  if ( not lot_id   ) or
     ( lot_id == "" ) then

    return false;

  end;

  local count = m_shop.Purchase_GetRegisteredCount( lot_id );

  return ( ( count ) and ( count > 0 ) );

end;
--------------------------------------------------------------------------------
function private.OnProductInformationReady ( msg, params )

  private.product_info_ready = true;

  private.UpdateAllLotsInfo();

end;
--------------------------------------------------------------------------------
function private.UpdateAllLotsInfo ()

  local purchase_lots = GetLotsInfo().lots;
  local database_lots = m_shop.Gold_GetLots( "coins" );

  for _, database_lot in pairs( database_lots ) do
    for _, purchase_lot in pairs( purchase_lots ) do

      if ( database_lot.name_ios == ( purchase_lot.name_ios or purchase_lot.name ) ) then

        database_lot.price = private.Lot_GetPrice( purchase_lot );
        break;

      end;

    end;
  end;

  private.Hearts_TryAddImmortality( database_lots );

  int_dialog_shop.UpdateLots( database_lots );
  int_dialog_minibank.UpdateLots( database_lots );

  local starterpack = private.Starterpack_Data();
  local sp_name_ios = starterpack.name_ios;

  for _, purchase_lot in pairs( purchase_lots ) do
    if ( sp_name_ios == ( purchase_lot.name_ios or purchase_lot.name ) ) then

      starterpack.price = private.Lot_GetPrice( purchase_lot );
      break;

    end;
  end;

  int_dialog_starterpack.UpdatePrice( starterpack.price );

  return database_lots;

end;
--------------------------------------------------------------------------------
function public.FeatureUpdated ()
  
  private.HandleActionSale();
  
end;
--------------------------------------------------------------------------------
function private.HandleActionSale ()
  
  local lots_list = private.OnProductInformationReady();
  local boosters_list = v_boosters.GetLots();
  
  int_dialog_shop.UpdateBoosters( boosters_list );

  private.PanelButton_UpdateSaleNotification();
  
end;
--------------------------------------------------------------------------------
function private.PanelButton_UpdateSaleNotification ()

  local lots_list     = private.UpdateAllLotsInfo();
  local boosters_list = v_boosters.GetLots();
  local value         = private.IsSaleExist( lots_list, boosters_list );
  local count         = tern( value )( 1, 0 );
  
  local data_custom = {
    target = "shop",
    data = {
      type = "custom",
      entity = "v_shop",
      count = count,
      style = "sale",
      icon_path = tern( count > 0 )( "assets/content/items/icon_sale", "" )
    }
  };
  
  v_interface.panel.bottom.Notification( data_custom );
  
end;
--------------------------------------------------------------------------------
function private.IsSaleExist ( lots, boosters )
  
  local is_lots_sale = false;
  local is_boosters_sale = false;
  
  for _, lot in pairs( lots ) do
    if ( lot.currency_count_bonus > 0 ) then
      is_lots_sale = true;
      break;
    end;
  end;
  
  for _, booster in pairs( boosters ) do
    if ( booster.packs[ 1 ].sale ) then
      is_boosters_sale = true;
      break;
    end;
  end;
  
  return is_lots_sale or is_boosters_sale;
  
end;
--------------------------------------------------------------------------------
function private.HandleActionSaleBoosters ()
  
  private.HandleActionSale();
  
end;
--------------------------------------------------------------------------------
function private.CheckButtonNotification ( forced_count )
  
  local count = forced_count or 0;
  
  if ( not forced_count ) then
    
    if ( m_shop.Market_IsUnlocked() and 
         m_shop.Market_IsBuilt() ) then
         
      local sells = private.Market_GetShares_Sell();

      for res_name, res_count in pairs( sells ) do

        if ( res_count > 0 ) then
          count = 1;
          break;
        end; 
        
      end;
    end;
    
  end;
  
  local data = {
    target = "shop",
    data = {
      type = "common",
      entity = "v_shop",
      count = count
    }
  };
  
  v_interface.panel.bottom.Notification( data );
  
end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function private.LogPurchase ( lot_id, action )
  
  if ( not lot_id or lot_id == "" ) then
    return;
  end;
  
  local Server_PurchaseLogAdded = function ( params )
    
    if ( not params.result ) then 
      DbgTrace("[ Logging ] log purchase end failed"); 
    end;
    
  end;
  
  local lot_data = m_shop.Lot_GetDataByIOSName( lot_id );
  if ( not lot_data or not lot_data.price or lot_data.price == "" ) then
    return;
  end;

  local lot_price = tonumber( string.match(lot_data.price, "%d+.%d+") ) * 100;
  if ( lot_price == 0 ) then
    return;
  end;

  lib_rpc.Schedule( "user", "logPurchase", { 
      action = action, 
      lotId = lot_id, 
      price = lot_price,
    },
    Server_PurchaseLogAdded, true, true );
    
end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function private.CheckRestorePurchases ( params )
  
  if ( params.status == "start" ) then
    private.performed_restore_purchases = true;
  elseif ( params.status == "finish" ) then
    private.performed_restore_purchases = false;
  end;
  
end;
--------------------------------------------------------------------------------
function private.Premium_TryBuyLot ( params )
  
  local source = { name = "shop" };

  if ( private.minibank_purchase ) then
    
    source = private.minibank_purchase;
    source.name = "minibank";

    if ( params.result ) then

      int_dialog_minibank.Hide();

    end;

  end;

  private.LogPurchase( params.id or "unknown", "end_"..tostring( params.result ) );
      
  if ( not params.result ) or ( not params.id ) or ( params.id == "" ) then

    private.purchase_started = nil;
    
    -- PC: params.canceled always false.
    if ( IsMobile() ) then
      
      local purchase_message = tern( params.canceled )( "purchasecanceled", 
                                                        "purchasefailed" );

      if not ( private.product_info_ready or loader.IsProductInfoReady() ) then
        purchase_message = "nointernetconnection";
      end;

      v_game.SystemDialog_Show( purchase_message, function () end );

    end;
      
    mediator.Notify( "View_Shop_Purchase_Failed" );

    return false;

  end;

  local lot_name = m_shop.Lot_GetNameById( params.id );

  -- JK: notification to start immortality
  mediator.Notify( "View_Shop_PremiumPurchaseMade",
  { 
    id   = params.id, 
    name = lot_name
  } );

  -- JK: try to reward player with resource extras specified in lot
  private.Lot_ApplyResourceExtras( lot_name );
  
  if ( private.performed_restore_purchases ) then
    
    if ( not ( params.id:find( "life" ) or 
               params.id:find( "live" ) or 
               params.id:find( "speedy" ) ) ) then
               
      v_player.PurchaseExecute();
      
    end;
    
  else
    
    v_player.PurchaseExecute();
    
  end;
  
  if ( private.purchase_started ) then
    
    common.sound.Play( "buy" );
    private.purchase_started = nil;
    
  end;
  
  if ( params.id:find( "life" ) or params.id:find( "live" ) ) then

    private.Purchase_Register( params.id );

    private.Hearts_Buy( params.id, params.transaction_id );

    return true;
    
  end;
  
  if ( params.id:find( "speedy" ) ) then

    private.Purchase_Register( params.id );

    private.Skin_Buy( params.id, params.transaction_id );

    return true;
    
  end;

  if ( params.id:find( "starterpack" ) ) then
    
    private.Starterpack_Buy( params.id, params.transaction_id );

    return true;
    
  end;
  
  private.Gold_TryBuyLot( params.id, params.transaction_id, source );

  private.minibank_purchase = false;

  return true;

end;
--------------------------------------------------------------------------------
function private.Gold_TryBuyLot ( lot_id, transaction_id, source )

  local lot_data = m_shop.Lot_GetDataByIOSName( lot_id );
  
  if ( lot_data and lot_data.currency_type == "gold" ) then
    
    local count = lot_data.currency_count;
    local bonus = lot_data.currency_count_bonus;
    local price = lot_data.price;
    
    local name_ios = lot_data.name_ios;
    
    v_player.PayingStatus_Set( price );
    v_player.Gold_TryChangeCount( count + bonus );
  
    v_log.Playhaven_LogPlacement( name_ios );
    
    if ( source.name == "shop" ) then
      
      local is_playhaven = ( lot_data.type == "playhaven" );
  
      v_log.Kontagent_LogEvent(
        "economy", 
        {
          description    = "premium_currency_collect",
          category       = tern( is_playhaven )( "playhaven", "store" ),
          value          = count + bonus,
          name           = name_ios,
          transaction_id = transaction_id
        }
      );      
      
    elseif ( source.name == "minibank" ) then

      local minibank_name = name_ios;
      
      if ( type( source ) == "table"      ) and
         ( type( source.lots ) == "table" ) and
         ( #source.lots > 0               ) then
        minibank_name = source.lots[ 1 ].name_ios or name_ios;
      end;
      
      v_log.Kontagent_LogEvent(
        "economy", 
        {
          description    = "context_offer",
          category       = name_ios,
          value          = count + bonus,
          name           = minibank_name,
          level          = 1,
          transaction_id = transaction_id
        }
      );
      
    end;
  
  end;
  
  common.Save( true );
  
  -- NOTE: Is it really need?
  NetRestartHeartBeats();

end;
--------------------------------------------------------------------------------
function private.Starterpack_Buy ( lot_id, transaction_id )

  local starterpack = m_shop.Starterpack_Get( lot_id );
  
  if ( starterpack ) then
  
    v_player.PayingStatus_Set( starterpack.price );
  
    v_log.Kontagent_LogEvent( "economy", 
    {
      description = "starterpack_collect",
      category    = "store",
      value       = 1,
      name        = starterpack.name_ios,
      transaction_id = transaction_id
    } );  
  
    for i = 1, #starterpack.currency_count do
      
      local starterpack_content = starterpack.currency_count[ i ];
      
      v_items.TryObtain( {
        name       = starterpack_content.name,
        type       = starterpack_content.type,
        count      = starterpack_content.count,
        category   = "starterpack",
        stage_name = "starterpack",
        silent     = false
      } );
 
    end;
    
    m_shop.Starterpack_Finish( starterpack.name );
  
  end;

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function public.Skin_TryBuy ( lot_id )
  
  if ( v_game.IsOnline() ) then

    private.PurchaseLot( lot_id );

  else

    v_game.TryReconnect( false );
    v_game.SystemDialog_Show( "nointernetconnection", function () end );
    mediator.Notify( "View_Shop_Purchase_Failed" );

  end;
  
end;
--------------------------------------------------------------------------------
function private.Skin_Buy ( lot_id, transaction_id )

  local lot_data = m_shop.Lot_GetDataByIOSName( lot_id );
  
  if ( lot_data ) then
  
    if ( lot_data.currency_count ) then
    
      v_log.Kontagent_LogEvent( "economy", 
      {
        description    = "item_collected",
        category       = "store",
        value          = lot_data.currency_count,
        name           = lot_data.name_ios,
        transaction_id = transaction_id
      } );
    
      v_player.Skin_TryAdd( lot_data.name_ios );
    
    end;
    
  end;

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function public.Hearts_TryBuy ( lot_name )
  
  if ( v_game.IsOnline() ) then

    local lot_id = m_shop.Lot_GetIOSName( lot_name );

    private.PurchaseLot( lot_id );

  else

    v_game.TryReconnect( false );
    v_game.SystemDialog_Show( "nointernetconnection", function () end );
    mediator.Notify( "View_Shop_Purchase_Failed" );

  end;
  
end;
--------------------------------------------------------------------------------
function private.Hearts_TryAddImmortality ( lots )

  if ( v_hearts.Immortality_IsFeatureActive() ) then

    for index, lot in pairs( lots ) do

      lots[ index ].immortality = v_hearts.Immortality_TimeByLot( lot.name );
      local resource_extras     = private.Lot_GetResourceExtras( lot.name );

      lots[ index ].resource1 = resource_extras.resource1;
      lots[ index ].resource2 = resource_extras.resource2;
      lots[ index ].resource3 = resource_extras.resource3;

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Hearts_Buy ( lot_id, transaction_id )

  local lot_data = m_shop.Lot_GetDataByIOSName( lot_id );
  
  if ( lot_data ) then
  
    if ( lot_data.price ) then
    
      v_player.PayingStatus_Set( lot_data.price );
    
    end;
  
    if ( lot_data.currency_count ) then
    
      v_log.Kontagent_LogEvent( "economy", 
      {
        description    = "item_collected",
        category       = "store",
        value          = lot_data.currency_count,
        name           = lot_data.name_ios,
        transaction_id = transaction_id
      } );
    
      v_hearts.TryAddDonateBonus( lot_data.currency_count );
    
    end;
    
  end;

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function private.Gold_OnDeficient ( params )

  local deficient = params;
  deficient.name  = "gold";

  local gold_lots     = m_shop.Gold_GetLots( "coins" );
  local purchase_lots = ( GetLotsInfo() or {} ).lots or gold_lots;
  
  deficient.lots = {};
  
  if ( purchase_lots      ) and
     ( #purchase_lots > 0 ) then
     
     for _, purchase_lot in pairs( purchase_lots ) do
      for index, lot in pairs( gold_lots ) do
        
        if ( lot.currency_count >= deficient.count ) and
           ( #deficient.lots < 2                   ) then
          if ( not deficient.lots[ 1 ] ) or
             ( deficient.lots[ 1 ] and deficient.lots[ 1 ].currency_count < lot.currency_count ) then
            lot.number = index;
            table.insert( deficient.lots, lot );
          end;
        end;

        if ( lot.name_ios == ( purchase_lot.name_ios or purchase_lot.name ) ) then
          lot.price = private.Lot_GetPrice( purchase_lot );
          break;
        end;
        
      end;
    end;
  end;
  
  if ( #deficient.lots == 1 ) or
     ( 
       ( #deficient.lots == 2                       ) and
       ( deficient.lots[ 1 ] == deficient.lots[ 2 ] )
     ) then
     
     local index = #gold_lots - 1;
     
     deficient.lots[ 2 ]        = gold_lots[ index ];
     deficient.lots[ 2 ].number = index;
  end;

  private.Hearts_TryAddImmortality( deficient.lots );

  local func_accept = function ( lot_name )
    
    if ( v_game.IsOnline() ) then

      local lot_id = m_shop.Lot_GetIOSName( lot_name );

      private.PurchaseLot( lot_id );

    else

      v_game.TryReconnect( false );
      v_game.SystemDialog_Show( "nointernetconnection", function () end );
      mediator.Notify( "View_Shop_Purchase_Failed" );

    end;

  end;

  local func_reject = function () 

    if ( private.minibank_purchase ) then

      local name_ios = "";
      
      if ( type( private.minibank_purchase ) == "table"      ) and
         ( type( private.minibank_purchase.lots ) == "table" ) and
         ( #private.minibank_purchase.lots > 0               ) then
        name_ios = private.minibank_purchase.lots[ 1 ].name_ios or "";
      end;
      
      v_log.Kontagent_LogEvent( "economy", 
      {
        description = "context_offer",
        category    = "",
        value       = 0,
        name        = name_ios,
        level       = 0
      } );
      
      private.minibank_purchase = false;

    end;
    
  end;

  private.minibank_purchase = { lots = deficient.lots };
  
  v_game.TryReconnect( false );
  
  v_interface.window.Show(
    "int_dialog_minibank",
    function ()

      int_dialog_minibank.Show( deficient );

    end,
    nil,
    {
      Accept = func_accept,
      Reject = func_reject
    }
  );  

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Resources_OnDeficient ( params )

  local deficient = params;

  if ( #params == 0 ) then

    deficient = { params };

  end;

  local func_solution = function () end;
  local solution      = "stage";

  local solved = false;

  -- JK: Solution 1. Market
  -- GD: Hided because of possibility to purchase 
  -- resources in the BuildConstruct window.
  --[[if ( m_shop.Market_IsBuilt() ) then

    solved        = true;
    solution      = "market";
    func_solution = function () 

      public.Show( "market" );

    end;

  end;]]

  -- JK: Solution red fireflies.
  if ( deficient[ 1 ].name == "resource4" ) then
    solved = true;
    solution = "minievent";
  end;
  
  -- JK: Solution 2. Spell
  if not ( solved ) then

    local spell_data = v_spells.Usable_Search( deficient[ 1 ].name );

    if ( spell_data ) then

      solved        = true;
      solution      = "spell";
      func_solution = function ()

        private.Resources_FocusOnSolution( "obj_map_spells_"..spell_data.name, 
          function()

            ui.anim.Bounce( "obj_map_spells_"..spell_data.name,
              function()
             
                v_map.ShowObjectPopup(
                  "monolog",
                  "obj_map_spells_"..spell_data.name,
                  "str_int_popup_cloud_spell_usable",
                  nil,
                  { 0, -100 }
                );
              
              end );

          end );

      end;

    end;

  end;

  -- JK: Solution 3. Mine
  if not ( solved ) then

    local mine_data = v_mines.Usable_Search( deficient[ 1 ].name );

    if ( mine_data ) then

      solved        = true;
      solution      = "mine";
      func_solution = function ()

        private.Resources_FocusOnSolution( "obj_map_mines_"..mine_data.name );

      end;

    end;

  end;

  -- JK: Solution 4. Stage
  if not ( solved ) then

    local stage_data = v_stage.GetUsableStages( deficient[ 1 ].name );

    if ( stage_data ) then

      solved        = true;
      solution      = "stage";
      func_solution = function ()

        private.Resources_FocusOnSolution( "obj_map_point_"..stage_data.name,
          function ()

            ui.anim.Spin( "obj_map_point_"..stage_data.name,
              function()

                local focus_name = string.format( "fcs_map_point_%s",
                                                  stage_data.name );

                ObjAnimate( focus_name, "alp", 0, 0, "", 
                {
                  0.0, 2, 0, 
                  0.3, 2, 1,
                  0.6, 2, 0,
                  0.9, 2, 1,
                  1.2, 2, 0
                } );

            end );

        end );

      end;

    end;

  end;

  if not ( solved ) then

    DbgTrace( "[ VIEW SHOP ] Deficient solution not found!" );

  end;

  local func_accept = function ()
  
    v_game.dialog.Deficient_Hide( func_solution );
  
  end;

  local func_reject = function ()

    v_game.dialog.Deficient_Hide();

  end;

  v_game.dialog.Deficient_Show(
    deficient,
    func_accept,
    func_reject,
    solution
  );

end;
--------------------------------------------------------------------------------
function private.Resources_FocusOnSolution ( obj_name, func_end )

  v_interface.window.HideAll( function ()

    v_map.Focus( obj_name, { func_after = func_end } );

  end );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Market_Settings( key )

  local  settings = m_shop.Market_Settings( key );
  return settings;

end;
--------------------------------------------------------------------------------
function public.Market_TryUnlock ( stage )

  if ( stage == m_shop.Market_Settings( "unlock" ) ) then
  
    if ( not m_shop.Market_IsUnlocked() ) then
    
      m_shop.Market_Unlock();
      return true;
    
    end;
  
  end;
  
  return false;

end;
--------------------------------------------------------------------------------
function public.Market_IsUnlocked ()

  local is_unlocked = m_shop.Market_IsUnlocked();
  
  return is_unlocked;

end;
--------------------------------------------------------------------------------
function public.Market_TryBuild ( gold_count, func_SetObject, func_SetTap )

  if ( m_shop.Market_IsBuilt() ) then return false; end;

  -- JK: i was like omg market is free now
  if ( gold_count < 1 ) then
    
    m_shop.Market_Build();
    m_shop.Market_RefreshForced();
    private.Market_Update( func_SetObject, func_SetTap );
    return true;
    
  end;

  if ( v_player.Gold_TryChangeCount( - gold_count ) ) then

    v_log.Kontagent_LogEvent( "economy", 
    {
      description = "premium_currency_spend",
      category    = "market",
      value       = - gold_count,
      name        = "build"
    } );

    m_shop.Market_Build();
    m_shop.Market_RefreshForced();
    private.Market_Update( func_SetObject, func_SetTap );
    return true;

  end;

  return false;

end;
--------------------------------------------------------------------------------
function public.Market_IsBuilt ()

  local is_built = m_shop.Market_IsBuilt();
  
  return is_built;

end;
--------------------------------------------------------------------------------
function public.Market_Update ( func_SetObject, func_SetTap )

  private.Market_Update( func_SetObject, func_SetTap );

end;
--------------------------------------------------------------------------------
function public.Market_CheaterRefresh ()

  m_shop.Market_RefreshForced();

end;
--------------------------------------------------------------------------------
function public.Market_IsSellDisabled ()

  return ( m_shop.Market_Settings( "refresh" ) < 1 );
  
end;

--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Market_Update ( func_SetObject, func_SetTap )

  local unlocked = m_shop.Market_IsUnlocked();
  local built    = m_shop.Market_IsBuilt();

  if ( unlocked ) then

    func_SetObject( unlocked, built );

  else

    func_SetObject( unlocked, built );
    func_SetTap( "popup" );

    return;

  end;

  if ( built ) then

    func_SetTap( "open" );

  else

    func_SetTap( "build" );

  end;

end;
--------------------------------------------------------------------------------
function private.Market_TryBuy ( resources, subtotal )

  if ( v_player.Gold_TryChangeCount( - subtotal ) ) then

    v_log.Kontagent_LogEvent( "economy", 
    {
      description = "premium_currency_spend",
      category    = "market",
      value       = - subtotal,
      name        = "buy"
    } );

    for i = 1, #resources do

      local share_value = private.Market_Share_Value( "buy", i );

      if ( v_player.Resource_TryChangeCount( "resource"..i, resources[ i ] ) ) then

        v_log.Kontagent_LogEvent( "economy", 
        {
          description = "soft_currency_collect",
          category    = "market",
          value       = resources[ i ],
          name        = "buy",
          type        = "resource"..i
        } );

      end;
  
    end;    

    return true;

  end;

  return false;

end;
--------------------------------------------------------------------------------
function private.Market_TrySell ( resource_index )

  local resources = private.Market_GetShares_Sell();

  if ( v_player.Resource_TryChangeCount( "resource"..resource_index, - resources[ "resource"..resource_index ] ) ) then

    v_log.Kontagent_LogEvent( "economy", 
    {
      description = "soft_currency_spend",
      category    = "market",
      value       = - resources[ "resource"..resource_index ],
      name        = "sell",
      type        = "resource"..resource_index
    } );

    local share_value = private.Market_Share_Value( "sell", resource_index );
    local share_price = private.Market_Share_Price( "sell", resource_index );
    local gold_amount = math.floor( resources[ "resource"..resource_index ] / share_value * share_price );

    v_player.Gold_TryChangeCount( gold_amount );

    v_log.Kontagent_LogEvent( "economy", 
    {
      description = "premium_currency_collect",
      category    = "market",
      value       = gold_amount,
      name        = "sell"
    } );

    resources[ "resource"..resource_index ] = 0;

    m_shop.Market_SaveOffers( resources );

    return true;

  else

    --DbgTrace( "not enough res: "..resources[ "resource"..resource_index ] );
    return false;

  end;

end;
--------------------------------------------------------------------------------
function private.Market_Share_Value ( type, index )

  local shares = m_shop.Market_Settings( type );
  
  if ( shares[ index ] ) then
    return shares[ index ].value or 0;
  end;
  
  return 0;

end;
--------------------------------------------------------------------------------
function private.Market_Share_Price ( type, index )

  local shares = m_shop.Market_Settings( type );
  
  if ( shares[ index ] ) then
    return shares[ index ].price or 0;
  end;
  
  return 0;

end;
--------------------------------------------------------------------------------
function private.Market_Share_Limit ( type, index )

  local shares = m_shop.Market_Settings( type );

  return shares[ index ].max;

end;
--------------------------------------------------------------------------------
function private.Market_GetShares_Sell ()

  local  sell_offers = m_shop.Market_LoadOffers();
  return sell_offers;

end;
--------------------------------------------------------------------------------
function private.Market_GetSecondsToRefresh ()

  local  seconds = m_shop.Market_GetSecondsToRefresh();
  return seconds;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Cheat_Starterpack ()

  local starterpack = private.Starterpack_Data();

  if ( starterpack ) then

    local is_active = m_shop.Starterpack_IsTimerExists();

    -- JK: toggle OFF ----------------------------------------------------------
    if ( is_active ) then

      m_shop.Starterpack_Finish( starterpack.name );
      return;

    end;

    -- JK: toggle ON -----------------------------------------------------------
    local is_completed = m_shop.Starterpack_IsCompleted( starterpack.name );

    if ( is_completed ) then

      m_shop.Starterpack_Clear( starterpack.name );

    end;
    
    m_shop.Starterpack_Start( starterpack.name );

  else

    v_interface.ShowNotification( "Starterpack lot is not available." );

  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Starterpack_DialogShow ( func_on_close )

  private.Starterpack_DialogShow( func_on_close );

end;
--------------------------------------------------------------------------------
function public.Starterpack_IsAvailable ()
  
  return ( private.Starterpack_IsUnlocked() and
           m_shop.Starterpack_IsTimerExists() );

end;
--------------------------------------------------------------------------------
function private.Starterpack_Init ( starterpack_name )

  mediator.Subscribe(
    "Model_Sections_SectionEnabled",
    private.Starterpack_OnSectionEnabled
  );
  
  if ( private.Starterpack_IsUnlocked() ) then
  
    if ( m_shop.Starterpack_IsTimerExists() ) then
      
      private.Starterpack_EnableButton();
      
    elseif ( not m_shop.Starterpack_IsCompleted( starterpack_name ) ) then
      
      m_shop.Starterpack_Start( starterpack_name );
      
    end;

  else

    mediator.Subscribe(
      "Model_Stage_StageWon",
      private.Starterpack_OnStageWon
    );

    private.is_staterpack_subscribed = true;
  
  end;

end;
--------------------------------------------------------------------------------
function private.Starterpack_GetUnlockStage ()
  
  local stage_name = v_sections.Section_GetConfig().starterpack.unlock;
  
  return stage_name;
  
end;
--------------------------------------------------------------------------------
function private.Starterpack_IsUnlocked ()

  local unlock_stage = private.Starterpack_GetUnlockStage();
  local stage_data   = v_stage.Stage_StorageData( unlock_stage ) or {};
  
  return stage_data.completed or false;

end;
--------------------------------------------------------------------------------
function private.Starterpack_OnStageWon ( params )
  
  local stage_name = private.Starterpack_GetUnlockStage();
  
  if ( params.name == stage_name ) then

    mediator.Unsubscribe(
      "Model_Stage_StageWon",
      private.Starterpack_OnStageWon
    );
    
    private.is_staterpack_subscribed = false;
    
    local starterpack = private.Starterpack_Data();

    if ( starterpack ) then
  
      local is_active    = m_shop.Starterpack_IsTimerExists();
      local is_completed = m_shop.Starterpack_IsCompleted( starterpack.name );

      --DbgTrace( "is completed: "..tostring( is_completed ) );

      -- JK: check if player is paying and purchased starterpack once
      -- if last purchase was less than payer_timeout, don't offer starterpack
      if ( is_completed ) then

        local server_time        = parts.time.server.Timestamp();
        local last_purchase_time = v_player.PayingStatus_GetLastPurchaseTime();

        --DbgTrace( "last_purchase: "..tostring( last_purchase_time ) );

        if ( last_purchase_time ) then

          local payer_timeout   = m_shop.Starterpack_GetPayerTimeout();
          local current_timeout = server_time - last_purchase_time;
          local diff_timeout    = payer_timeout - current_timeout;

          --DbgTrace( "Diff timeout: "..tostring( diff_timeout ) );
          
          if ( diff_timeout > 0 ) then

            return;

          end;

        end;

      end;

      if ( is_active ) then
        
        m_shop.Starterpack_Finish( starterpack.name );
        
      end;

      if ( is_completed ) then

        m_shop.Starterpack_Clear( starterpack.name );

      end;
      
      m_shop.Starterpack_Start( starterpack.name );

    else

      DbgTrace( "[ ERROR ][ v_shop ] Starterpack lot is not available." );

    end;
  
  end;

end;
--------------------------------------------------------------------------------
function private.Starterpack_OnSectionEnabled ()
  
  if ( not private.is_staterpack_subscribed ) then
  
    mediator.Subscribe(
      "Model_Stage_StageWon",
      private.Starterpack_OnStageWon
    );

    private.is_staterpack_subscribed = true;
  
  end;
  
end;
--------------------------------------------------------------------------------
function private.Starterpack_Data ()

  -- NOTE: Do not check paying status until we have just single lot
  local paying_status = 1; -- v_player.PayingStatus_Get();
  local starterpack   = m_shop.Starterpack_GetLotByIndex( paying_status );

  return starterpack;

end;
--------------------------------------------------------------------------------
function private.Starterpack_DialogShow ( func_on_close )

  local starterpack = private.Starterpack_Data();
 
  local lots = ( GetLotsInfo() or {} ).lots or {};

  for l = 1, #lots do
    
    local purchase_lot = lots[ l ];
    
    if ( starterpack.name_ios == ( purchase_lot.name_ios or purchase_lot.name ) ) then
      starterpack.price = private.Lot_GetPrice( purchase_lot );
    end;
  end;
  
  local func_accept = function ()
    
    if ( v_game.IsOnline() ) then

      local lot_id = m_shop.Lot_GetIOSName( starterpack.name );

      private.PurchaseLot( lot_id );

    else
      
      v_game.TryReconnect( false );
      v_game.SystemDialog_Show( "nointernetconnection", function () end );
      mediator.Notify( "View_Shop_Purchase_Failed" );

    end;

  end;
  
  local func_reject = function () private.Starterpack_DialogHide(); end;
  local func_close  = function ()
    if (func_on_close) then
      func_on_close();
    end;
  end;
    
  v_game.TryReconnect( false );
  
  v_interface.window.Show(
    "int_dialog_starterpack",
    function ()
      
      int_dialog_starterpack.Show( starterpack );

    end,
    nil,
    {
      Accept = func_accept,
      Reject = func_reject,
      OnClose = func_close
    }
  );

end;
--------------------------------------------------------------------------------
function private.Starterpack_DialogHide ()

  local func_window_hide = function ( func_on_close )
    
    v_interface.window.Hide( "int_dialog_starterpack", func_on_close );
    
  end;
  
  int_dialog_starterpack.Hide( func_window_hide );
  
end;
--------------------------------------------------------------------------------
function private.Starterpack_OnStart ()

  private.Starterpack_EnableButton();

end;
--------------------------------------------------------------------------------
function private.Starterpack_OnFinish ()

  v_interface.Scheme_DeleteButton( "map", "right", { name = "sale" } );
  
  if ( v_interface.window.IsOnScreen( "int_dialog_starterpack" ) ) then

    private.Starterpack_DialogHide();

  end;

end;
--------------------------------------------------------------------------------
function private.Starterpack_EnableButton ()
  
  v_interface.panel.right.SetDelegates( {
    Sale_OnTap = private.Starterpack_DialogShow,
    Sale_GetTime = function ()
      local time = m_shop.Starterpack_GetRemainingTime();
      return time;
    end
  } );

  v_interface.Scheme_AddButton( "map", "right", { name = "sale" } );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Feature_Save ()
  
  if ( v_features.IsActive( private.feature ) ) then

    local abtest_group  = v_features.GetFeatureGroup( private.feature );
    local feature_group = v_features.GetSavedFeatureGroup( private.feature );

    if ( abtest_group and
         abtest_group ~= feature_group ) then
      
      v_log.Kontagent_LogEvent( "abtest", 
      {
        description = private.feature,
        name        = "add",
        value       = abtest_group
      } );
      
      v_features.SaveFeatureGroup( private.feature, abtest_group );
      
    elseif ( not abtest_group and feature_group ) then
      
      v_features.SaveFeatureGroup( private.feature, abtest_group );
      
    end;

  end;
  
end;
--------------------------------------------------------------------------------