-- name=v_social
--------------------------------------------------------------------------------
local TAG = "v_social";
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  base = {
    "m_social",
    "m_email",
    "int_dialog_social_mail",
    "int_dialog_hearts_send"
  },
  mobile = {
    "int_dialog_social_friends_choose",
    "int_dialog_social_friends_hearts_send",
    "int_dialog_social_friends_not_attempts",
    "int_panel_social_score",
    "int_dialog_social_login"
  }
};
--------------------------------------------------------------------------------

public.Mail   = {};
public.Hearts = {};

public.Panel = {
  Mail   = {},
  Social = {}
};

public.Score = {};

public.FriendsHearts = {
  Ask         = {},
  Send        = {},
  NotAttempts = {}
};

private.Mail   = {};
private.Hearts = {};

private.FriendsHearts = {
  Ask         = {},
  Send        = {},
  NotAttempts = {};
};

private.Panel = {};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  local modules = _modules.base;

  for _, module_name in pairs( _modules.mobile ) do

    table.insert( modules, module_name );

  end;

  return modules;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.__InitModule ()

  common.mod.Load( _modules.base );

  if ( IsMobile() ) then

    common.mod.Load( _modules.mobile );

  end;
    
end;
--------------------------------------------------------------------------------
function public.Init ()

  v_interface.window.RegisterAsModal( "int_dialog_social_login" );

  private.userId = v_player.UserId();
  private.fb_share_section_name = false;
    
  if ( IsMobile() ) then
      
    v_interface.window.RegisterAsModal( "int_dialog_social_friends_choose" );
    v_interface.window.RegisterAsModal( "int_dialog_social_friends_hearts_send" );
    v_interface.window.RegisterAsModal( "int_dialog_hearts_send" );
    v_interface.window.RegisterAsModal( "int_dialog_social_friends_not_attempts" );
    
  end;

  mediator.Subscribe(
    "Model_Email_AttachmentProcessingRequired",
    private.Attachment_Obtain
  );
  
  if ( m_email.FakeEmail_HeartIsHidden() ) then
    mediator.Subscribe(
      "Model_Hearts_RecoverStart",
      private.Hearts.WaitForWriteOff
    );
  end;
  
  private.Overtaken_Init();
  
  --Log.i( "Initialized." );
  
end;
--------------------------------------------------------------------------------
function public.Start ()

  m_social.Start(
    v_player.UserId,
    v_game.IsOnline,
    v_game.WaitingShow,
    v_game.WaitingHide
  );
  
  private.is_logged          = m_social.IsLogged();
  private.is_friends_updated = false;

  private.Mail.messages = {};
  private.Mail.removed  = {};
  
  mediator.Subscribe(
    "Model_Email_MessageListUpdated",
    private.Mail_UpdateList
  );
  
  v_game.RegisterFeature(
  { 
    entity = "v_social",
    feature = "fbshare",
    func_save = private.Feature_SaveData
  } );
  
  private.UpdateOnline();
  
  if ( IsMobile() ) then

    mediator.Subscribe(
      "Model_Social_LoginComplete",
      private.OnLoginComplete
    );
    
    mediator.Subscribe(
      "Model_Social_LoginContinue",
      private.OnLoginContinue
    );

    mediator.Subscribe( 
      "Model_Social_FriendsUpdated",
      private.OnFriendsUpdated
    );
    
  end;
  
  m_email.Start(
    v_game.IsOnline,
    v_player.UserId
  );
  
  public.FacebookShare_SetShareWorld();
  
  --private.Mail_UpdateUnreadCount();
  
--  local function func_prepare ()
--
--    int_panel_social_score.Prepare( m_social.GetFriends() );
--
--  end;
--
--  if ( v_game.IsOnline() ) then
--    
--    local message = tern( private.is_logged )(
--      "View_Map_MapOpened",
--      "Model_Social_LoginComplete"
--    );
--
--    mediator.SubscribeOnce( message, func_prepare );
--    
--  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.IsLogged ()
  
  local  is_logged = m_social.IsLogged();
  return is_logged;

end;
--------------------------------------------------------------------------------
function public.Logout ( callpoint )

  if ( IsMobile() ) then
    
    -- fix FSFTP-1577
    m_social.Logout(
      callpoint,
      function ()

        private.is_logged = m_social.IsLogged();

        if ( not private.is_logged ) then
          private.Social_DisableButton();
          private.Social_EnableButton();
        end;

      end
    );
    
  end;

end;
--------------------------------------------------------------------------------
function public.Login ( callpoint )

  private.LoginDialog_Show( callpoint );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Friends_Cheat_AddMockFriend ()

  m_social.Cheat_AddMockFriend();

end;
--------------------------------------------------------------------------------
function public.Friends_Count ()
  
  return #m_social.GetFriends() or 0;
  
end;
--------------------------------------------------------------------------------
function public.Friends_Scores ( stage_name )
  
  local  friend_score = m_social.GetFriendsStageScores( stage_name );
  return friend_score;
  
end;
--------------------------------------------------------------------------------
function public.Friends_ListInZone ( section_name )
  
  local  friend_list = m_social.GetFriendsInZone( section_name );
  return friend_list;
  
end;
--------------------------------------------------------------------------------
function public.Friends_IsUpdated ()

  return private.is_friends_updated;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.FacebookShare_SetShareWorld ()
  
  local list = v_sections.List();
  
  if ( #list > 1 ) then
    private.fb_share_section_name = v_sections.Section_GetNameByNumber( #list - 1 );
  end;
  
  if ( private.FacebookShare_IsPosted() ) then
    private.fb_share_section_name = false;
  end;
  
end;
--------------------------------------------------------------------------------
function public.FacebookShare ()

  if ( not private.fb_share_section_name ) then
    return;
  end;
  
  local str_prefix  = "str_fbshare_world_complete_";
  local icon        = "icon_world_complete";
  local title       = ne.strings.Get( str_prefix.."title" );
  local description = ne.strings.Get( str_prefix.."description" );
  
  v_game.ShareToFacebook(
    title,
    description,
    icon,
    function ()
      v_map.Dialog_Worldmap_ShareComplete( private.fb_share_section_name );
      private.FacebookShare_MarkSectionAsPosted();
    end
  );

end;
--------------------------------------------------------------------------------
function public.FacebookShare_GetAvailableSectionName ()
  
  if ( not private.fb_share_section_name ) then
    return nil;
  end;
  
  if not ( IsMobile() and v_game.IsOnline() and 
       m_social.IsLogged() and v_social.IsFBShareActive() and
       not private.FacebookShare_IsPosted() ) then
    return nil;
  end;
  
  return private.fb_share_section_name;
  
end;
--------------------------------------------------------------------------------
function private.FacebookShare_MarkSectionAsPosted ()

  if ( not private.fb_share_section_name ) then
    return;
  end;
  
  m_social.FacebookShare_MarkSectionAsPosted( private.fb_share_section_name );

  private.fb_share_section_name = false;
  
end;
--------------------------------------------------------------------------------
function private.FacebookShare_IsPosted ()
  
  if ( not private.fb_share_section_name ) then
    return false;
  end;
  
  local posted_sections = m_social.FacebookShare_GetSectionsPosted();
  local is_posted       = ( posted_sections[ private.fb_share_section_name ] == true );

  return is_posted;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Social_EnableButton ()
  
  --common.DbgCallpoint( "Social_EnableButton", 1 );
  local button_type = nil;
  local button_func = nil;
  
  if ( not public.IsLogged() ) then
    
    button_type = "social";
    button_func = function ()
      private.LoginDialog_Show( "worlds_map", true );
    end;
    
  end;

  if ( button_type ) then
  
    --DbgTrace( "Enable social button: "..button_type );

    v_interface.panel.right.SetDelegates( { Social_OnTap = button_func } );
    v_interface.Scheme_AddButton( "map", "right", { name = "social", type = button_type } );
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.Social_DisableButton ( button_type )
  
  --common.DbgCallpoint( "Social_DisableButton: "..tostring( button_type ), 1 );

  if ( not button_type ) then
    v_interface.Scheme_DeleteButton( "map", "right", { name = "social", type = "social" } );
    return;
  end;
  
  v_interface.Scheme_DeleteButton( "map", "right", { name = "social", type = button_type } );
  
end;
--------------------------------------------------------------------------------
function private.Login ( callpoint )

  if ( v_game.IsOnline() ) then

    if ( not private.is_logged ) then

      mediator.Notify( "View_Social_LoginStart" );
      private.Social_DisableButton();
      m_social.Login( callpoint );

    end;

  else

    private.ShowOfflineNote();

  end;

end;
--------------------------------------------------------------------------------
function private.OnLoginComplete ( success )
  
  if ( success ) then
    private.is_logged = true;
    private.Social_DisableButton();
  end;
  
  private.Social_EnableButton();
  
end;
--------------------------------------------------------------------------------
function private.OnLoginContinue ()

  if ( not v_hearts.TryAddSocialBonus() ) then
  
    v_items.TryObtain( m_social.GetLoginReward() );
  
  end;

end;
--------------------------------------------------------------------------------
function private.OnFriendsUpdated ()

  --Log.i( "Friends updated" );

  private.is_friends_updated = true;

  private.Social_EnableButton();
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.GetLoginReward ()

  if ( v_hearts.CanAddSocialBonus() ) then
  
    return { name = "heart", type = "heart", count = 1 };
  
  end;
  
  return m_social.GetLoginReward();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.LoginDialog_Show ( callpoint, func_on_close )
  
  private.LoginDialog_Show( callpoint, false, func_on_close );
  
end;
--------------------------------------------------------------------------------
function public.LoginDialog_IsAvailable ()
  
  return ( IsMobile() and not private.is_logged );
  
end;
--------------------------------------------------------------------------------
function private.LoginDialog_Show ( callpoint, disallow_modal, func_on_close )

  local reward = public.GetLoginReward();
  
  local func_close = function ()
    if (func_on_close) then
      func_on_close();
    end;
  end;
  
  v_interface.window.Show(
    "int_dialog_social_login",
    function ()

       int_dialog_social_login.Show( reward.name, reward.count );

    end,
    { disallow_modal = disallow_modal },
    {
      Action = function () private.LoginDialog_Login( callpoint ); end,
      Close  = function ()
        private.LoginDialog_Hide();
      end,
      OnClose = func_close
    }
  );

end;
--------------------------------------------------------------------------------
function private.LoginDialog_Hide ()

  int_dialog_social_login.Hide(
    function ( func_on_close )
      v_interface.window.Hide( "int_dialog_social_login", func_on_close );
    end
  );

end;
--------------------------------------------------------------------------------
function private.LoginDialog_Login ( callpoint )
  
  mediator.SubscribeOnce( "Event_Social_Login_Complete", private.LoginDialog_Hide );
  
  private.Login( callpoint );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.ButtonMail_Show ()
  
  v_interface.panel.right.SetDelegates( {
    Mail_OnTap = function ()
      private.Mail_ShowDialog();
      v_log.Playhaven_LogPlacement( "mail_open" );
    end
  } );

  v_interface.Scheme_AddButton( "map", "right", { name = "mail" } );
  
end;
--------------------------------------------------------------------------------
function private.ButtonMail_Hide ()
  
  v_interface.Scheme_DeleteButton( "map", "right", { name = "mail" } );
  
end;
--------------------------------------------------------------------------------
function public.Mail_RequestUpdate ()
  
  m_email.Messages_RequestUpdate();
  
end;
--------------------------------------------------------------------------------
function private.Mail_ShowDialog ()

  if ( v_game.IsOnline() ) then

    m_email.Suspend();

    local mail_list  = private.Mail_ListMessages( 1, 6 );
    local mail_total = private.Mail_GetCount();

    local delegates = {};

    delegates.ReadMessage = function ( message_id )
      m_email.Message_MarkAsRead( message_id );
      private.Mail_MarkForRemoval( message_id );
    end;

    delegates.SendHeart = function ( targetId )
      private.Hearts.SendHeart( targetId, private.userId, true );
    end;

    delegates.IsAttachmentObtainable = private.Attachment_IsObtainable;

    delegates.ShowTopPanel = function ()
      v_interface.panel.top.Show( true, "gold" );
    end;

    delegates.EnterPromocode = function ()
      v_player.DialogCode_Enter_Show();
    end;

    delegates.Hide = function ()
      private.Mail_PurgeRemovedMessages();
      v_interface.panel.top.Hide( true );
      v_log.Playhaven_LogPlacement( "mail_close" );
      m_email.Resume();
    end;

    delegates.ShowAskDialog = function ()
      public.FriendsHearts.Ask.Show();
    end;

    delegates.ShowSendDialog = function ()
      public.FriendsHearts.Send.Show();
    end;

    delegates.GetMessages = function ( from, limit )
      local  messages = private.Mail_ListMessages( from, limit );
      return messages;
    end;

    delegates.OnHide = function ()
      v_interface.window.Hide( "int_dialog_social_mail" );
    end;

    local stage_number = v_player.FriendCode_GetFeatureUnlocker();
    local is_promocode = v_stage.Stage_IsCompleted( "stage_"..stage_number );
    local friend_list  = m_social.GetFriends();

    local is_invite = ( public.Friends_IsUpdated() ) and
                      ( #friend_list == 0          );

    local is_logged = ( public.Friends_IsUpdated() ) and
                      ( private.is_logged          );

    v_interface.window.Show(
      "int_dialog_social_mail",
      function ()

        int_dialog_social_mail.Show(
          mail_list,
          mail_total,
          is_logged,
          is_promocode,
          is_invite
        );

      end,
      nil,
      delegates
    );

  else

    private.ShowOfflineNote();

  end;

end;
--------------------------------------------------------------------------------
function private.Mail_UpdateList ( messages )

  local important = {};
  local sorted    = {};

  for _, message in pairs( messages ) do

    if ( message.text.string:find( "help" )   ) or
       ( message.text.string:find( "thanks" ) ) or
       ( message.text.string:find( "gift" )   ) then

      table.insert( sorted, 1, message );

    else

      table.insert( important, message );

    end;

  end;

  for _, message in pairs( important ) do

    table.insert( sorted, 1, message );

  end;

  local viewable_count = m_email.Messages_GetViewableCount();

  -- JK: trim message list
  for i = #sorted, 1, -1 do

    if ( #sorted > viewable_count ) then

      table.remove( sorted, i );

    else break; end;

  end;

  private.Mail.messages = sorted;

  private.Mail_UpdateUnreadCount();

end;
--------------------------------------------------------------------------------
function private.Mail_GetCount ()

  return #private.Mail.messages;

end;
--------------------------------------------------------------------------------
function private.Mail_GetList ()

  return common.table.DeepCopy( private.Mail.messages );

end;
--------------------------------------------------------------------------------
function private.Mail_ListMessages ( from, limit )

  local friend_list = m_social.GetFriends();
  local mail_list   = private.Mail_GetList();
  local messages    = {};

  for i = from, ( from + limit ) do

    local message = mail_list[ i ];

    if ( message ) then

      local sender = message.from or {
        name   = "Administrator",
        avatar = "",
        level  = 0,
        userId = 1
      };

      local friend_data = friend_list[
        common.IsEntityInList( 
          friend_list, 
          { socialId = sender.socialId },
          "socialId"
        )
      ];
    
      if ( friend_data ) then
        message.from.avatar = friend_data.url;
        message.from.name   = friend_data.name or sender.name;
      end;

      if ( not message.from.name   ) or
         ( message.from.name == "" ) then
      
        message.from.name = "id"..sender.userId;
        
      end;

      table.insert( messages, message );

    end;

  end;

  return messages;

end;
--------------------------------------------------------------------------------
function private.Mail_MarkForRemoval ( message_id )

  private.Mail.removed[ "id_"..message_id ] = true;

end;
--------------------------------------------------------------------------------
function private.Mail_PurgeRemovedMessages ()

  if ( next( private.Mail.removed ) ) then

    for i = #private.Mail.messages, 1, -1  do 

      local id = private.Mail.messages[ i ].messageId;

      if ( private.Mail.removed[ "id_"..id ] ) then

        table.remove( private.Mail.messages, i );
        private.Mail.removed[ "id_"..id ] = nil;

        if ( not next( private.Mail.removed ) ) then  break; end;

      end;

    end;

  end;

  private.Mail.removed = {};

  private.Mail_UpdateUnreadCount();

end;
--------------------------------------------------------------------------------
function private.Mail_UpdateUnreadCount ()
  
  local mail_count = #private.Mail.messages;

  v_interface.panel.right.MailSetNotifications( mail_count );

  if ( not v_game.IsOnline() ) then
    private.ButtonMail_Hide();
  else
    private.ButtonMail_Show();
  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.DialogSend_Show (
    func_accept,
    friend_name,
    friend_res,
    friend_url,
    pos_obj_friend
  )

  local attempts, max_attempts = m_social.GetAttemptsSend();
  local is_limit_reached = ( ( max_attempts - attempts ) <= 0 );

  if ( is_limit_reached ) then
    private.FriendsHearts.NotAttempts.Show( "send" );
    return;
  end; 
  
  local delegates = {
    Accept = func_accept,
    Close  = function ()
      int_dialog_hearts_send.Hide();
    end,
    OnClose = function ()
      v_interface.window.Hide( "int_dialog_hearts_send" );
    end;
  };

  v_interface.window.Show(
    "int_dialog_hearts_send",
    function ()

      int_dialog_hearts_send.Show(
        is_limit_reached,
        friend_name,
        friend_res,
        friend_url,
        pos_obj_friend
      );

    end,
    nil,
    delegates
  );  

end;
--------------------------------------------------------------------------------
function private.FriendsHearts.NotAttempts.Show ( attempt_type, func_OnClose )
  
  local delegates = {

    Close = function ()
      int_dialog_social_friends_not_attempts.Hide();
    end,
    
    OnClose = function ()
      v_interface.window.Hide( "int_dialog_social_friends_not_attempts", func_OnClose );
    end
  };

  v_interface.window.Show(
    "int_dialog_social_friends_not_attempts",
    function ()

      int_dialog_social_friends_not_attempts.Show( attempt_type );

    end,
    nil,
    delegates
  );
  
end;
--------------------------------------------------------------------------------
function public.FriendsHearts.Ask.Show ( func_OnClose )
  
  if ( v_game.IsOnline() ) and
     ( private.is_logged ) then

    local friend_list = m_social.GetFriends();
    local attempts, max_attempts = m_social.GetAttemptsAsk();
    local left_attempts = math.max( 0, max_attempts - attempts );
    
    if ( left_attempts == 0 ) then
      private.FriendsHearts.NotAttempts.Show( "ask", func_OnClose );
      return;
    end;
    
    local delegates = {
  
      Send = private.Hearts.Ask,

      Close = function ()
        int_dialog_social_friends_choose.Hide();
        if ( func_OnClose ) then func_OnClose(); end;
      end,
      
      OnClose = function ()
        v_interface.window.Hide( "int_dialog_social_friends_choose" );
      end;  
    };

    v_interface.window.Show(
      "int_dialog_social_friends_choose",
      function ()

        int_dialog_social_friends_choose.Show( friend_list, left_attempts );

      end,
      nil,
      delegates
    );

  else

    private.ShowOfflineNote();

  end;
  
end;
--------------------------------------------------------------------------------
function public.FriendsHearts.Send.IsAvailable ()
  
  local attempts, max_attempts = m_social.GetAttemptsSend();
  local left_attempts = math.max( 0, max_attempts - attempts );
  local friend_list = m_social.GetFriends();
  
  if ( v_game.IsOnline() ) and ( private.is_logged ) and ( left_attempts > 0 ) and 
    ( #friend_list > 0 ) then
    return true;
  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
function public.FriendsHearts.Send.Show ( func_OnClose )
    
  if ( v_game.IsOnline() ) and ( private.is_logged ) then

    local friend_list = m_social.GetFriends();
    local attempts, max_attempts = m_social.GetAttemptsSend();
    local left_attempts = math.max( 0, max_attempts - attempts );

    if ( left_attempts == 0 ) then
      private.FriendsHearts.NotAttempts.Show( "send", func_OnClose );
      return;
    end;
    
    local delegates = {
  
      Send = private.Hearts.Send,

      Close = function ()
        int_dialog_social_friends_hearts_send.Hide();
      end,
      
      OnClose = function ()
        v_interface.window.Hide( "int_dialog_social_friends_hearts_send", func_OnClose );
      end;
    };

    v_interface.window.Show(
    "int_dialog_social_friends_hearts_send",
      function ()

        int_dialog_social_friends_hearts_send.Show( friend_list, left_attempts );

      end,
      nil,
      delegates
    );

  else

    private.ShowOfflineNote();

  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Score.TryShow ( point, stage_name, stage_score, score_old )

  if ( not IsMobile() ) then return; end;

  local func_show_panel = function ()

    local friends_scores = m_social.GetFriendsStageScores( stage_name );

    if ( friends_scores and #friends_scores > 0 ) then
      
      local func_send_heart = function ( friend_id, friend_name, friend_res, friend_url, pos_obj_friend )
        
        local func_accept = function ()

          private.Hearts.SendHeart( friend_id, private.userId );

          local message_id = private.Hearts.GetAskingMessageId( friend_id );

          m_email.Message_MarkAsRead( message_id );

        end;

        private.DialogSend_Show(
          func_accept,
          friend_name,
          friend_res,
          friend_url,
          pos_obj_friend
        );

      end;

      local player_score = stage_score;

      score_old = score_old or 0;

      if ( point == "win"          ) and
         ( score_old > stage_score ) then

        point        = "win_low";
        player_score = score_old;

      end;

      local social_info = m_social.GetSocialInfo();

      table.insert(
        friends_scores,
        { 
          userId   = private.userId,
          socialId = social_info.id,
          name     = social_info.name,
          url      = social_info.url,
          score    = player_score,
          heart    = false,
          res      = "assets/content/interface/social/"..v_player.Skin_GetCurrent().."_avatar"
        }
      );

      table.sort(
        friends_scores,
        function( a, b ) return a.score > b.score; end
      );

      local is_logged = public.IsLogged();

      if ( point ~= "start" ) and ( not public.IsFBShareActive() ) then
        is_logged = false;
      end;
      
      local delegates = {
        SendHeart = func_send_heart,
        Share     = function ()
          private.Share( private.userId, friends_scores, stage_name );
        end
      };

      common.delegate.Set( "int_panel_social_score", delegates );
    
      int_panel_social_score.Show( 
        point,
        friends_scores,
        private.userId,
        is_logged
      );
      
    end;

  end;

  func_show_panel();

end;
--------------------------------------------------------------------------------
function public.Score.IsActive ()

  return ( int_panel_social_score and int_panel_social_score.IsActive() );
  
end;
--------------------------------------------------------------------------------
function public.Score.Hide ()

  if ( IsMobile() ) then
    
    int_panel_social_score.Hide();

  end;
  
end;
--------------------------------------------------------------------------------
function public.Score.VisualShow ()

  int_panel_social_score.VisualShow();

end;
--------------------------------------------------------------------------------
function public.Score.VisualHide ()

  int_panel_social_score.VisualHide();

end;
--------------------------------------------------------------------------------
function private.Share ( userId, friends_scores, stage_name )
  
  if ( v_game.IsOnline() ) then
    
    local stage = common.string.FormatNameStage( stage_name );
    local icon = "icon_lvl_complete";
    local title = ne.strings.Get( "str_fbshare_level_complete_title" );
    local description = ne.strings.Get( "str_fbshare_level_complete_description", stage );
    local place = 1;
    
    for f = 1, #friends_scores do
      if ( tonumber( friends_scores[ f ].userId ) == userId ) then
        place = f;
        break;
      end;
    end;

    local params = { level = stage:match( "%d+" ), place = place, position = place };

    if ( place == 1 ) then

      param = stage:match( "%d+" );
      icon = "icon_lvl_1_place";
      title = ne.strings.Get( "str_fbshare_level_complete_first_place_title" );
      description = ne.strings.Get( "str_fbshare_level_complete_first_place_description", stage );
      
    end;
    
    v_game.ShareToFacebook( title, description, icon, int_panel_social_score.ShareComplete, params );
    
  else
    
    v_game.SystemDialog_Show( "nointernetconnection", function () end );
    
  end;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Hearts.Send ( friend_list )
  
  if ( v_game.IsOnline() ) then

    local attempts, max_attempts = m_social.GetAttemptsSend();
    local left_attempts = math.max( 0, max_attempts - attempts );
    
    if ( left_attempts == 0 ) then
      return;
    end;
    
    if ( friend_list and #friend_list > 0 ) then

      for i = 1, math.min( #friend_list, left_attempts ) do
        private.Hearts.SendHeart( friend_list[ i ], private.userId );
      end;
      
    end;

  else

    private.ShowOfflineNote();

  end;
  
end;
--------------------------------------------------------------------------------
function private.Hearts.Ask ( friend_list )

  if ( v_game.IsOnline() ) then

    local attempts, max_attempts = m_social.GetAttemptsAsk();
    local left_attempts = math.max( 0, max_attempts - attempts );
    
    if ( left_attempts == 0 ) then
      return;
    end;
    
    if ( friend_list and #friend_list > 0 ) then

      for i = 1, math.min( #friend_list, left_attempts ) do
        private.Hearts.SendAsk( friend_list[ i ], private.userId );
      end;
      
    end;

  else

    private.ShowOfflineNote();

  end;

end;
--------------------------------------------------------------------------------
function private.Hearts.Add ( friend_list )

  local out_list = {};
  local mail_list = private.Mail_GetList();

  for f = 1, #friend_list do

    local friend = friend_list[f];
    friend.heart = ( private.Hearts.GetAskingMessageId( friend.userId, mail_list ) ~= false );
    table.insert( out_list, friend );

  end;

  return out_list;

end;
--------------------------------------------------------------------------------
function private.Hearts.GetAskingMessageId ( targetId, mail_list )

  mail_list = mail_list or private.Mail_GetList();
  if ( mail_list and #mail_list > 0 ) then
    for m = 1, #mail_list do
      if ( mail_list[m].text == "str_int_dialog_social_mail_message_help" ) and
         ( mail_list[m].from.userId == targetId ) then
        return mail_list[m].messageId;
      end;
    end;
  end;
  return false; 

end;
--------------------------------------------------------------------------------
function private.Hearts.SendAsk ( targetId, userId, skip_update )

  m_email.Message_Send( 
    targetId, 
    userId, 
    "str_int_dialog_social_mail_message_help", 
    { name = "heart", type = "heart", count = 1 } 
  );

  if ( not skip_update ) then
    private.Attempts_TryChangeCount( "ask", 1 );
  end;
  
  local source = tern( targetId < 1000 )( "bot", "user" );
  
  v_log.Kontagent_LogEvent( "heart", 
  {
    name        = "request",
    category    = "mail",
    value       = 1,
    description = "heart",
    source      = source
  } );
  
end;
--------------------------------------------------------------------------------
function private.Hearts.SendHeart ( targetId, userId, skip_update )

  m_email.Message_Send( 
    targetId, 
    userId, 
    "str_int_dialog_social_mail_message_gift", 
    { name = "heart", type = "heart", count = 1 } 
  );

  if ( not skip_update ) then
    private.Attempts_TryChangeCount( "send", 1 );
  end;
  
  local source = tern( targetId < 1000 )( "bot", "user" );

  v_log.Kontagent_LogEvent( "heart", 
  {
    name        = "send",
    category    = "mail",
    value       = 1,
    description = "heart",
    source      = source
  } );

  --[[
    GD: Если игроку на почту приходит письмо с запросом Сердец, 
        и он отправляет сердце в ответ,
        то помимо отправки сердца другу, 
        он оправляет сообщение с запросом Сердца.
  
    UPD: Убрано по просьбе издателя (FSFTP-605). 
    private.Hearts.SendAsk( targetId, userId );
  ]]

end;
--------------------------------------------------------------------------------
function private.Hearts.SendThanks ( targetId, userId )

  m_email.Message_Send( 
    targetId, 
    userId, 
    "str_int_dialog_social_mail_message_thanks", 
    { name = "heart", type = "heart", count = 1 } 
  );

end;
--------------------------------------------------------------------------------
function private.Hearts.WaitForWriteOff ( data )
  
  if ( data.is_block ) then return; end;
  
  mediator.Unsubscribe(
    "Model_Hearts_RecoverStart",
    private.Hearts.WaitForWriteOff
  );
  
  m_email.FakeEmail_HeartsAdd();
  m_email.Messages_RequestUpdate();
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.FriendsLimitReset_Cheat ( attempt_type )
  
  if ( attempt_type == "ask" ) then
    if ( m_social.Attempts_Ask_TimerIsExist() ) then
      m_social.Attempts_Ask_FinishTimer();
    end;
  elseif ( attempt_type == "send" ) then
    if ( m_social.Attempts_Send_TimerIsExist() ) then
      m_social.Attempts_Send_FinishTimer();
    end;
  end;
  
  m_social.UpdateAttempts( attempt_type, 0 );
  
end;
--------------------------------------------------------------------------------
function private.Attempts_TryChangeCount ( attempt_type, count )
  
  local attempts, max_attempts = m_social.GetAttempts( attempt_type );
  local attempt_count = math.min( attempts + count, max_attempts );
  
  m_social.UpdateAttempts( attempt_type, attempt_count );
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Attachment_IsObtainable ( attachment )

  if not ( attachment ) then return false; end;

  if ( attachment.type == "stickerpack" ) then
    if ( not _G.v_fireflies         ) or 
       ( not v_fireflies.Unlocked() ) then
      return false, "wait";
    end;
  end;

  if ( attachment.type == "heart"       ) and
     ( attachment.name ~= "immortality" ) and
     ( v_hearts.IsFull()                ) then
    return false, "hearts_full";
  end;

  return true;

end;
--------------------------------------------------------------------------------
function private.Attachment_Obtain ( params )

  local attachment  = params.attachment;
  local is_obtained = false;
  
  local text = ( params.text or {} ).string or "";
  local message_type = text:match( "str_int_dialog_social_mail_message_(%w+)" ) or "mail";
  message_type = text:match( "str_social_reward_for_(%w+[_a-z]*)" ) or message_type;

  local category, source = "mail", message_type; 
  source = tern( params.sender_id and params.sender_id < 1000 )( "bot_"..source, "user_"..source );

  if ( attachment.name ) then

    if ( attachment.type ~= "chest" ) then
    
      if ( message_type ~= "help" and 
           message_type ~= "thanks" ) then
           
        local item =
        {
          name       = attachment.name,
          type       = attachment.type,
          count      = attachment.count,
          category   = category,
          source     = source,
          silent     = false
        };
        
        if ( attachment.type == "container" ) then
          item.silent = true;
        end;

        return v_items.TryObtain( item );
        
      end;

      return true;
      
    end;

    if ( attachment.name:find( "event" ) ) then
    
      -- TODO: Fix it in 1.12
      attachment.name = "woodchest";
    
    end;
    
    local chest = 
    { 
      type  = "chest",
      name  = attachment.name,
      count = 1
    };
    
    local params =
    {
      category   = category,
      stage_name = source,
      silent     = true
    };
    
    v_items.Chest_Show( chest, params, attachment.content );
    
    return true;
      
  end;
  return false;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.ConnectionLost ()
  
  private.UpdateOnline();
  
end;
--------------------------------------------------------------------------------
function public.ConnectionRestored ()
  
  private.UpdateOnline();
  
end;
--------------------------------------------------------------------------------
function private.UpdateOnline ()

  if ( not v_game.IsOnline() ) then
    -- JK: bots are available even offline, comment below
    --public.Score.Hide();
    if ( v_interface.window.LowermostName() == "int_dialog_social_mail" ) then
      v_interface.window.HideAll();
    end;
    if ( IsMobile() ) then
      private.Social_DisableButton();
    end;
  else
    private.userId = v_player.UserId();
    if ( IsMobile() ) then
      private.Social_EnableButton();
    end;
  end;
  
  private.Mail_UpdateUnreadCount();
  
end;
--------------------------------------------------------------------------------
function private.ShowOfflineNote ( func_close )

  func_close = func_close or function () end;

  v_game.SystemDialog_Show( "connectionlost", func_close );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Feature_SaveData ()
  
  v_features.Feature_SaveState( "fbshare" );
  
end;
--------------------------------------------------------------------------------
function private.Feature_GetState ()

  local  feature_state = v_features.GetFeatureState( "fbshare" );
  return feature_state;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.IsFBShareActive ()
  
  local  is_active = private.Feature_GetState();
  return is_active;
  
end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Overtaken_IsNeedToShow ( section_name )
  
  if ( math.random() > m_social.Overtaken_GetShowChance() ) then
  
    return false;
  
  end;

  for _, leader in pairs( private.overtake.known_leads ) do
  
    if ( leader.section == section_name ) then
    
      return true;
    
    end;
  
  end;
  
  return false;

end;
--------------------------------------------------------------------------------
function public.Overtaken_GetRandomPlayer ( section_name )

  local ids = {};
  
  for id, leader in pairs( private.overtake.known_leads ) do
  
    if ( leader.section == section_name ) then
    
      table.insert( ids, id );
    
    end;
  
  end;
  
  if ( #ids == 0 ) then
  
    return nil;
    
  end;
  
  local id = math.random( 1, #ids );
  return table.remove( private.overtake.known_leads, id );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Overtaken_Init ()

  private.overtake =
  {
    score_cache = nil,
    known_leads = { }
  };
  
  -- Insert new known_leads by comparison of score_cache and new_scores
  mediator.Subscribe( 
    "Model_Social_FriendsScoresUpdated",
    private.Overtaken_OnScoresUpdated
  );
  
  -- Remove known leads by completed stage
  mediator.Subscribe( 
    "Model_Stage_StageWon",
    private.Overtaken_OnStageWon
  );
  
end;
--------------------------------------------------------------------------------
function private.Overtaken_OnScoresUpdated ( scores_new )

  if ( not private.overtake.score_cache ) then
  
    -- Do not look for outrun players at first scores receive
    private.overtake.score_cache = scores_new;
    return;
  
  end;

  local section_name = v_sections.Section_NameCurrent();
  local scores_old   = private.overtake.score_cache;

  local section_number = v_sections.Section_NumberCurrent();
  local stage_max = section_number * 110;
  local stage_min = stage_max - 109;
  
  for stage_num = stage_min, stage_max do
  
    local stage_name = "stage_"..stage_num;
    local stage_data = v_stage.Stage_StorageData( stage_name );
    
    if not ( stage_data and stage_data.score ) then break; end;
    
    local leaders_new = private.Overtaken_GetLeaders( scores_new, stage_data );
    local leaders_old = private.Overtaken_GetLeaders( scores_old, stage_data );
    
    for _, leader_new in pairs( leaders_new ) do
    
      if ( not common.IsEntityInList( leaders_old, leader_new, "id" ) )  then
      
        local leader = 
        {
          id      = leader_new.id,
          score   = leader_new.score,
          section = section_name,
          stage   = stage_name,
          name    = m_social.GetFriend( leader_new.social_id ).name,
        };
        table.insert( private.overtake.known_leads, leader );
      
      end;
    
    end;
  
  end;
  
  private.overtake.score_cache = scores_new;

end;
--------------------------------------------------------------------------------
function private.Overtaken_GetLeaders( scores_new, stage_data )

  local stage_score = stage_data.score;
  local stage_name  = stage_data.name; 

  local leaders = {};
  
  for _, player in pairs( scores_new ) do
    
    for _, stage in pairs( player.collection ) do
    
      if ( stage.score ) then
      
        if ( stage.name == stage_name and stage.score > stage_score ) then
        
          local leader = 
          {
            id        = player.user_id,
            social_id = player.social_id,
            score     = stage.score
          };
          table.insert( leaders, leader );
        
        end;
      
      end;
    
    end;
  
  end;
  
  return leaders;

end;
--------------------------------------------------------------------------------
function private.Overtaken_OnStageWon ( params )

  local stage_name = params.name;
  local stage_data = v_stage.Stage_StorageData( stage_name );
  
  if ( stage_data and stage_data.score ) then
  
    local score = stage_data.score;
    
    while true do
    
      local id = private.Overtaken_FindObsolteLeader( stage_name, score );
      if  ( id ) then
      
        table.remove( private.overtake.known_leads, id );
      
      else
      
        break;
      
      end;
    
    end;
  
  end;

end;
--------------------------------------------------------------------------------
function private.Overtaken_FindObsolteLeader ( stage_name, score )

  for id, leader in pairs( private.overtake.known_leads ) do
  
    if ( stage_name == leader.stage and leader.score <= score ) then
    
      return id;
    
    end;
  
  end;
  
  return nil;

end;
--------------------------------------------------------------------------------
