-- name=v_spells
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_spells",
  "int_dialog_map_spell"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  common.mod.Load( _modules );
  
  --Log.i( "Initialized." );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Static_GetAllBySection ( section_name )

  local spells = m_spells.Static_GetAllBySection( section_name );
  
  return spells;

end;
--------------------------------------------------------------------------------
function public.Saved_GetByName ( spell_name )

  local spell = m_spells.Saved_GetByName( spell_name );
  
  return spell;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Spell_Create ( spell_name, section_name )

  local obj_name = ( "obj_map_spells_%s" ):format( spell_name );
  local btn_name = ( "btn_map_spells_%s" ):format( spell_name );
  
  ObjCreate( obj_name, "obj" );
  ObjSet(    obj_name, { anim_tag = spell_name } );
  ObjAttach( obj_name, "anm_map_"..section_name );
  
  ObjClone( "btn_map_spells", "_"..spell_name, 1, obj_name );
  
  ui.control.EventsSet( btn_name,
    { 
      func_on_click = function () private.Dialog_Show( spell_name ) end
    }
  );
  
  private.Spell_Set( spell_name );

end;
--------------------------------------------------------------------------------
function public.Callback_AnimateAppear ( stage_name )

  local spell_name = m_spells.Spell_GetNameByUnlock( stage_name );
  
  if not ( spell_name ) then 
  
    return nil;
    
  end;
  
  local obj_name = "obj_map_spells_"..spell_name;
  
  local anim =
  {
    name   = obj_name,
    res    = "assets/content/map/effects/anm_appearance",
    coords = { pos_x = 0, pos_y = 0, scale_x = 0.5, scale_y = 0.5 },

    func_preset = function () 
      ObjSet( obj_name, { visible = true, alp = 0.0 } );
    end,

    func_dummy = function () 
      
      private.Sound_Play( "new_bandit" );
      
      ObjAnimate( obj_name, "alp", 0, 0, function ()
        private.Spell_Show( spell_name );
      end, 
      { 
        0.0, 0, 0.0, 
        0.2, 0, 1.0
      } );
    end,

    func_action = function ( func_after )

      mediator.Notify( "View_Spells_SpellbookAppear", spell_name );
      if ( func_after ) then func_after(); end;

    end
  };
  
  return v_map.Callback_Animate( anim );

end;
--------------------------------------------------------------------------------
function public.Usable_Search ( resource_name )

  local section_name  = v_sections.Section_NameCurrent();
  local usable_spells = m_spells.Usable_List( section_name );

  if ( usable_spells ) then

    for i = 1, #usable_spells do

      local output = m_spells.GetOutput( usable_spells[ i ].name );

      if ( output.name == resource_name ) then

        return usable_spells[ i ];

      end;

    end;

  end;

  return false;

end;
--------------------------------------------------------------------------------
function public.Spell_GetNameByUnlock ( stage_name )

  local spell_name = m_spells.Spell_GetNameByUnlock( stage_name );
  
  return spell_name;

end;
--------------------------------------------------------------------------------
function public.TryUnlockSpellByStage ( stage_name )

  local spell_name = m_spells.Spell_GetNameByUnlock( stage_name );
  if  ( spell_name ) then
  
    local is_unlocked = m_spells.TryUnlockSpell( spell_name );
    
    return is_unlocked;
  
  end;
  
  return false;

end;
--------------------------------------------------------------------------------
function public.GetSpellData ( spell_name )
  
  local spell_data = m_spells.Spell_Data( spell_name );
  
  return spell_data;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Spell_Set ( spell_name )

  local spell_data = m_spells.Spell_Data( spell_name );

  if ( spell_data.enabled ) then

    private.Spell_Show( spell_name );

  else

    private.Spell_Hide( spell_name );

  end;

  if ( spell_data.completed ) then

    private.Spell_Complete( spell_name );

  end;

end;
--------------------------------------------------------------------------------
function private.Spell_Show ( spell_name )

  ui.obj.Enable( "btn_map_spells_"..spell_name );

end;
--------------------------------------------------------------------------------
function private.Spell_Hide ( spell_name )

  ui.obj.Disable( "btn_map_spells_"..spell_name );

end;
--------------------------------------------------------------------------------
function private.Spell_Complete ( spell_name )

  Log.i( "CompleteSpell: "..spell_name );

  local btn_name = "btn_map_spells_"..spell_name;
  local anm_name = "anm_map_spells_"..spell_name;
  
  ui.obj.Set( btn_name, { input = false } );
  ui.obj.Set( anm_name, { animfunc = "static", playing = true } );

end;
--------------------------------------------------------------------------------
function private.Spell_TryExecute ( spell_name )

  local data = m_spells.Spell_Data( spell_name );
  
  -- Check inputs
  for _, input in pairs( data.input ) do
  
    local notify = true; -- DG: Send "can't change count" notification
    local name   = input.name;
    local count  = input.count;
  
    if ( name == "gold" ) then
    
      if ( not v_player.Gold_IsEnough( count, notify ) ) then
        return false;
      end;
    
    else
    
      if ( not v_player.Resource_IsEnough( name, count, notify ) ) then
        return false;
      end;
    
    end;
    
  end;
  
  -- Process inputs
  for _, input in pairs( data.input ) do
  
    local name  = input.name;
    local count = input.count;
  
    if ( name == "gold" ) then
    
      v_player.Gold_TryChangeCount( -count );
      
      v_log.Kontagent_LogEvent( "economy", 
      {
        description = "premium_currency_spend",
        category    = "spellbook",
        value       = -count,
        name        = spell_name
      } );
    
    else
    
      v_player.Resource_TryChangeCount( name, -count );
      
      v_log.Kontagent_LogEvent( "economy", 
      {
        description = "soft_currency_spend",
        category    = "spellbook",
        value       = -count,
        name        = spell_name,
        type        = name
      } );
    
    end;
    
  end;
  
  -- Process outputs
  for _, output in pairs( data.output ) do
  
    local name  = output.name;
    local count = output.count;
    
    v_items.TryObtain( 
      {
        name       = name, 
        type       = tern( name == "gold" )( "gold", "resource" ),
        count      = count,
        category   = "spellbook",
        stage_name = spell_name,
        silent     = true
      }
    );
  
  end;
  
  -- Complete spell
  m_spells.Complete( spell_name );
  
  return data;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Dialog_Show ( spell_name )

  local delegates = {

    Accept = function ()
    
      return private.Spell_TryExecute( spell_name );
    
    end,

    Close = function ( spell_data )

      int_dialog_map_spell.Hide( function ()

        if ( spell_data ) then
        
          private.Spell_Complete( spell_data.name );
          
          if ( spell_data.output ) then
          
            for _, reward in pairs( spell_data.output ) do
            
              if     ( reward.name:find( "gold"     ) ) then
              
                v_interface.Gold_Buffer_Out();
              
              elseif ( reward.name:find( "resource" ) ) then
              
                v_map.Reward_ShowCollected( 
                  spell_data.name, reward.count, reward.name
                );
              
              elseif ( IsEditor()                     ) then
              
                local msg = "Warning! No visual reward (%s) for spell %q!"
                DbgTrace( msg:format( reward.name, spell_data.name ) );
              
              end;
            
            end;
          
          end;

        end;
      
      end );
      
    end,

    ShowTopPanel = v_interface.panel.top.Show,
    HideTopPanel = v_interface.panel.top.Hide

  };

  v_interface.window.AddToQueue(
    "int_dialog_map_spell",
    function ()

      local spell_data = m_spells.Spell_Data( spell_name );
      local zone_data  = v_stage.Zone_Data( spell_data.unlock );

      int_dialog_map_spell.Show( spell_data, zone_data );

    end,
    nil,
    delegates
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Sound_Play ( sound_id )
  
  common.sound.Play( "assets/content/map/sounds/aud_"..sound_id );
  
end;
--------------------------------------------------------------------------------