-- name=v_stage
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_stage",
  "m_fortunewheel",
  "int_panel_stage_top",
  "int_panel_stage_progress_ride",
  "int_panel_stage_progress_fight",
  "int_panel_stage_score",
  "int_panel_stage_resources",
  "int_panel_stage_boosters",
  "int_panel_stage_settings",
  "int_panel_stage_win",
  "int_popup_stage_score",
  "int_popup_stage_combo",
  "int_popup_stage_bonus",
  "int_dialog_stage_shuffle",
  "int_dialog_stage_start",
  "int_dialog_stage_exit",
  "int_dialog_stage_win",
  "int_dialog_stage_fail",
  "int_dialog_stage_continue",
  "int_dialog_stage_fortunewheel",
  "int_dialog_stage_targets",
  "int_dialog_stage_friends"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.is_active                  = false;
  private.popup_show                 = false;
  private.prev_popup_stage           = nil;
  private.fortunewheel_popup_enabled = true;
  
  private.this = {
    model = {
      stage = "m_stage",
      wheel = "m_fortunewheel"
    },
    dialog = {
      shuffle      = "int_dialog_stage_shuffle",
      start        = "int_dialog_stage_start",
      exit         = "int_dialog_stage_exit",
      win          = "int_dialog_stage_win",
      fail         = "int_dialog_stage_fail",
      continue     = "int_dialog_stage_continue",
      fortunewheel = "int_dialog_stage_fortunewheel",
      targets      = "int_dialog_stage_targets",
      friends      = "int_dialog_stage_friends"
    },
    panel = {},
    popup = {},
    hub = {
      animations = "hub_stage_animations"
    },
    effect = {
      cell_destroyed    = "anm_cell_destroyed",
      resource_removed  = "anm_resource_remove",
      movebonus_removed = "anm_movebonus_remove",
      item_dropped      = "anm_item_drop_effect",
      snowflake_found   = "anm_snowflake_found"
    }
  };

  common.mod.Load( _modules );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Unsubscribe ()
  
  mediator.Unsubscribe( "Model_Player_StageLastSet", private.Player_StageLastSet );

  ne.msg.Unsubscribe( "Event_Match3_FieldEstablished", private.Match3_FieldEstablished );

  ne.msg.Unsubscribe(
    "Event_Match3_BoosterUsed",
    private.Match3_BoosterUsed
  );

  mediator.Unsubscribe(
    "Model_Fortunewheel_TimerStart",
    private.Fortunewheel_TimerStart
  );

  mediator.Unsubscribe(
    "Model_Fortunewheel_TimerStop",
    private.Fortunewheel_TimerStop
  );
  
  mediator.Unsubscribe(
    "View_Boosters_BoosterUsed",
    private.View_Boosters_BoosterUsed
  );
  
  mediator.Unsubscribe(
    "View_Boosters_BoosterUsedAnimEnd",
    private.PanelBoosters_HideAnim
  );

  mediator.Unsubscribe(
    "View_Match3_FieldEstablished",
    private.TryContinue
  );

  ne.msg.Unsubscribe(
    "Event_Match3_FieldNeedShuffle",
    private.OnFieldNeedShuffle
  );

  mediator.Unsubscribe( "Model_Match3_FieldActivated",  private.OnFieldActivated );
  mediator.Unsubscribe( "View_Match3_FieldEstablished", private.OnFieldEstablished );
  mediator.Unsubscribe( "Model_Match3_BonusRemoved",    private.OnBonusRemoved );
  mediator.Unsubscribe( "Model_Match3_BonusSpawn",      private.OnBonusSpawn );
  --mediator.Unsubscribe( "Model_Match3_ComboActivated",        private.OnComboActivated );
  mediator.Unsubscribe( "Model_Match3_CombinationsNotExist",  private.TryContinue );
  mediator.Unsubscribe( "Model_Match3_ExplosionField",        private.TryContinue );
  
  mediator.Unsubscribe( "Model_Stage_BonusCountChanged", private.OnBonusCountChanged );

  mediator.Unsubscribe( "Model_Stage_ScoreChanged", private.Stage_ScoreChanged  );

  mediator.Unsubscribe( "Model_Stage_TargetUpdated",     private.TopPanel_UpdateGoals );
  mediator.Unsubscribe( "Model_Stage_ChainCountChanged", private.PopupCombo_Show );

  mediator.Unsubscribe( "Model_Stage_MovesCountChanged", private.OnMovesCountChanged );
  mediator.Unsubscribe( "Model_Stage_CellDestroyed",     private.OnCellDestroyed );
  mediator.Unsubscribe( "Model_Stage_ItemDropped",       private.OnItemDropped );
  mediator.Unsubscribe( "Model_Stage_TargetUpdated",     private.OnTargetUpdated );
  mediator.Unsubscribe( "Model_Stage_TargetComplete",    private.OnTargetComplete );
  mediator.Unsubscribe( "Model_Stage_StageComplete",     private.OnStageComplete );

  mediator.Unsubscribe( "Model_Stage_HeroAttackActivated",  private.OnHeroAttackActivated );
  mediator.Unsubscribe( "Model_Stage_EnemyAttackActivated", private.OnEnemyAttackActivated );

  mediator.Unsubscribe( "Model_Stage_MoveBonusRemoved", private.OnMoveBonusRemoved );
  
  mediator.Unsubscribe( "Model_Stage_RewardCountChanged", private.OnResourceRewardCountChanged );
  mediator.Unsubscribe( "Model_Stage_RewardCountChanged", private.PanelResources_Show );

  -- Panel score.
  mediator.Unsubscribe( "Model_Stage_ScoreChanged", private.OnScoreChanged );

  mediator.Unsubscribe( "Model_Stage_BonusDisable",        private.PopupBonus_Hide );
  mediator.Unsubscribe( "Model_Stage_BonusEffectDisabled", private.BonusAbility_Hide );
  
  mediator.Unsubscribe(
    "Model_Stage_ViewTryContinue",
    private.TryContinue
  );
  
  ne.msg.Unsubscribe(
    "Event_Feature_Updated",
    private.OnFeatureUpdated
  );
  
  mediator.Unsubscribe(
    "Model_Potion_RunAction",
    private.HandleActionPotions
  );
  
  mediator.Unsubscribe(
    "Model_Potion_CompleteAction",
    private.HandleActionPotions
  );
  
  mediator.Unsubscribe(
    "Model_Hearts_CountChanged",
    private.UpdateHeartsCount
  );
  
  mediator.Unsubscribe(
    "Model_Hearts_ImmortalityToggled",
    private.Immortality_OnToggled
  );
  
  mediator.Unsubscribe(
    "Model_Player_RecoverFinishedAttemptsVideo",
    private.Video_RecoverFinishedAttempts
  );

  mediator.Unsubscribe(
    "View_Sections_CurrentDisabled",
    private.OnSectionDisabled
  );
end;
--------------------------------------------------------------------------------
function private.Player_StageLastSet ( stage_name )
  m_stage.TryEnable( stage_name );
end
--------------------------------------------------------------------------------
function private.Match3_FieldEstablished ( msg, params )
  private.TryContinueNEW( params );
end
--------------------------------------------------------------------------------
function private.Match3_BoosterUsed ( msg, params )
  private.PanelBoosters_Use( params );
end
--------------------------------------------------------------------------------
function private.Fortunewheel_TimerStart ( params )
  int_dialog_stage_fortunewheel.Cooldown_OnStarted();
end
--------------------------------------------------------------------------------
function private.Fortunewheel_TimerStop ( params )
  int_dialog_stage_fortunewheel.Cooldown_OnFinished();
end
--------------------------------------------------------------------------------
function private.View_Boosters_BoosterUsed ( params )
  if ( params                   ) and
     ( params.name == "swatter" ) then
    private.TryContinue();
  end
end
--------------------------------------------------------------------------------
function private.Stage_ScoreChanged( params )
  private.PopupScore_Show( params.chip_x, params.chip_y, params.chip_type, params.chip_score );
end
--------------------------------------------------------------------------------
function public.Init ()
  
  v_interface.window.RegisterAsModal(      private.this.dialog.fortunewheel );
  v_interface.window.RegisterAsFullscreen( private.this.dialog.fortunewheel );

  v_game.RegisterFeature( 
  { 
    entity = "v_stage", 
    feature = "fortune_wheel_fresh_player",
    func_save = private.Feature_SaveData
  } );
  
  m_fortunewheel.Setup();
  
  mediator.Subscribe( "Model_Player_StageLastSet", private.Player_StageLastSet );

  ne.msg.Subscribe( "Event_Match3_FieldEstablished", private.Match3_FieldEstablished );

  ne.msg.Subscribe(
    "Event_Match3_BoosterUsed",
    private.Match3_BoosterUsed
  );
  
  mediator.Subscribe(
    "View_Boosters_BoosterUsed",
    private.View_Boosters_BoosterUsed
  );

  mediator.Subscribe(
    "Model_Fortunewheel_TimerStart",
    private.Fortunewheel_TimerStart
  );

  mediator.Subscribe(
    "Model_Fortunewheel_TimerStop",
    private.Fortunewheel_TimerStop
  );
  
  mediator.Subscribe(
    "View_Boosters_BoosterUsedAnimEnd",
    private.PanelBoosters_HideAnim
  );

  mediator.Subscribe(
    "View_Match3_FieldEstablished",
    private.TryContinue
  );

  ne.msg.Subscribe(
    "Event_Match3_FieldNeedShuffle",
    private.OnFieldNeedShuffle
  );

  mediator.Subscribe( "Model_Match3_FieldActivated",  private.OnFieldActivated );
  mediator.Subscribe( "View_Match3_FieldEstablished", private.OnFieldEstablished );
  mediator.Subscribe( "Model_Match3_BonusRemoved",    private.OnBonusRemoved );
  mediator.Subscribe( "Model_Match3_BonusSpawn",      private.OnBonusSpawn );
  --mediator.Subscribe( "Model_Match3_ComboActivated",        private.OnComboActivated );
  mediator.Subscribe( "Model_Match3_CombinationsNotExist",  private.TryContinue );
  mediator.Subscribe( "Model_Match3_ExplosionField",        private.TryContinue );
  
  mediator.Subscribe( "Model_Stage_BonusCountChanged", private.OnBonusCountChanged );

  mediator.Subscribe( "Model_Stage_ScoreChanged", private.Stage_ScoreChanged  );

  mediator.Subscribe( "Model_Stage_TargetUpdated",     private.TopPanel_UpdateGoals );
  mediator.Subscribe( "Model_Stage_ChainCountChanged", private.PopupCombo_Show );

  mediator.Subscribe( "Model_Stage_MovesCountChanged", private.OnMovesCountChanged );
  mediator.Subscribe( "Model_Stage_CellDestroyed",     private.OnCellDestroyed );
  mediator.Subscribe( "Model_Stage_ItemDropped",       private.OnItemDropped );
  mediator.Subscribe( "Model_Stage_TargetUpdated",     private.OnTargetUpdated );
  mediator.Subscribe( "Model_Stage_TargetComplete",    private.OnTargetComplete );
  mediator.Subscribe( "Model_Stage_StageComplete",     private.OnStageComplete );

  mediator.Subscribe( "Model_Stage_HeroAttackActivated",  private.OnHeroAttackActivated );
  mediator.Subscribe( "Model_Stage_EnemyAttackActivated", private.OnEnemyAttackActivated );

  mediator.Subscribe( "Model_Stage_MoveBonusRemoved", private.OnMoveBonusRemoved );
  
  mediator.Subscribe( "Model_Stage_RewardCountChanged", private.OnResourceRewardCountChanged );
  mediator.Subscribe( "Model_Stage_RewardCountChanged", private.PanelResources_Show );

  -- Panel score.
  mediator.Subscribe( "Model_Stage_ScoreChanged", private.OnScoreChanged );

  mediator.Subscribe( "Model_Stage_BonusDisable",        private.PopupBonus_Hide );
  mediator.Subscribe( "Model_Stage_BonusEffectDisabled", private.BonusAbility_Hide );
  
  mediator.Subscribe(
    "Model_Stage_ViewTryContinue",
    private.TryContinue
  );
  
  ne.msg.Subscribe(
    "Event_Feature_Updated",
    private.OnFeatureUpdated
  );
  
  mediator.Subscribe(
    "Model_Potion_RunAction",
    private.HandleActionPotions
  );
  
  mediator.Subscribe(
    "Model_Potion_CompleteAction",
    private.HandleActionPotions
  );
  
  mediator.Subscribe(
    "Model_Hearts_CountChanged",
    private.UpdateHeartsCount
  );
  
  mediator.Subscribe(
    "Model_Hearts_ImmortalityToggled",
    private.Immortality_OnToggled
  );
  
  mediator.Subscribe(
    "Model_Player_RecoverFinishedAttemptsVideo",
    private.Video_RecoverFinishedAttempts
  );

  mediator.Subscribe(
    "View_Sections_CurrentDisabled",
    private.OnSectionDisabled
  );
  
  v_screen.Setup( "scr_stage",
  {
    func_pre_open    = private.PreOpen,
    func_open        = private.Open,
    func_close       = private.Close,
    func_after_close = private.AfterClose
  } );

  ObjAttach( "scr_stage", "ne_storage" );

  ObjAttach( "tmr_stage_finalize",     InterfaceName );
  ObjAttach( "tmr_stage_finalbonuses", InterfaceName );
  ObjAttach( "tmr_stage_resources",    InterfaceName );

  private.item_dropped_target_offset_x = 0;
  private.item_dropped_target_offset_y = -100;

  if( ui.screen.IsLandscape() ) then

    private.item_dropped_target_offset_x = 100;
    private.item_dropped_target_offset_y = -70;

  end;

  --Log.i( "Initialized." );

end;
--------------------------------------------------------------------------------
function public.Finish ()

  private.Finish();

end;
--------------------------------------------------------------------------------
function public.TryStart ( stage_name, enemy_data, stage_state )

  local stage_data = m_stage.GetStaticData( stage_name );

  private.score_old = 0;
  private.match3_shades = false;

  local stage_started;

  if ( stage_data.realtype == "dailyquest" ) then

    private.enemy = enemy_data;

    stage_started = private.PreStart(
      stage_name,
      enemy_data
    );

  elseif ( stage_data.realtype == "fight" ) then

    stage_started = private.PreStart(
      stage_name,
      nil,
      1
    );

  elseif ( stage_data.realtype == "misc" ) then

    stage_state = tern( stage_data.type == "ride" )( stage_state, 4 );

    stage_started = private.PreStart(
      stage_name,
      nil,
      stage_state
    );

  else

    stage_started = private.PreStart(
      stage_name,
      nil,
      stage_state
    );

  end;

  if ( stage_started ) then

    ui.input.Off( "zmv_map" );

    v_log.Playhaven_LogPlacement( "load_level" ); 

    private.Start();
    
    return true;

  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
function public.IsActive ()
  
  return private.is_active;
  
end;
--------------------------------------------------------------------------------
function public.IsStageComplete ()
  
  local is_win = m_stage.IsWin();
  
  return is_win;
  
end;
--------------------------------------------------------------------------------
function public.Stateless_Limit ()
  
  local limit = m_stage.Stateless_GetLimit();
  
  return limit;
  
end;
--------------------------------------------------------------------------------
function public.Stateless_Check ()
  
  m_stage.Stateless_CheckStatus();
  
end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function private.Feature_SaveData ()
  
  v_features.Feature_SaveState( "fortune_wheel_fresh_player" );
  
end;
--------------------------------------------------------------------------------
function public.FeatureUpdated ()
  
  --
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.StageTopPanel_IsVisible ()
  
  local  is_visible = int_panel_stage_top.IsVisible();
  return is_visible;
  
end;
--------------------------------------------------------------------------------
function public.StageTopPanel_SetVisible ()
  
  int_panel_stage_top.SetVisible();
  
end;
--------------------------------------------------------------------------------
function public.StageTopPanel_SetInvisible ()
  
  int_panel_stage_top.SetInvisible();
  
end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function public.IsPaused ()
  
  return int_panel_stage_settings.IsOpened();
  
end;
--------------------------------------------------------------------------------
function public.Pause ()

  int_panel_stage_settings.MenuOpen();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Match3_ShowEffect ( effect_name, loop )
  
  private.Match3_ShowEffect( effect_name, loop );
  
end;
--------------------------------------------------------------------------------
function public.Match3_HideEffect ()
  
  private.Match3_HideEffect();
  
end;
--------------------------------------------------------------------------------
function private.Match3_SetInput ( value )

  ObjSet( "obj_match3", { input = value } );

end;
--------------------------------------------------------------------------------
function private.Match3_ShadesOn ()

  private.match3_shades_forced = private.match3_shades_forced or false;

  if ( private.match3_shades ) then

    return;

  end;

  private.match3_shades = true;
  
  local function func_notify ()
  
    mediator.Notify( "View_Stage_Match_ShadesOn" );
  
  end;
  
  local tme = 0.3;
  
  int_panel_stage_progress_fight.Move_Near( tme );

  ObjAnimate( "obj_match3", "color_rgb", 0, 0, func_notify, 
  { 
    0.0, 0, 1.0, 1.0, 1.0, 
    tme, 0, 0.5, 0.5, 0.5 
  } );
  
  if ( private.match_custom_shade ) then
    
    --local cur_y = ui.obj.Position( "obj_match3" )[2];
    --local shift = private.match_custom_shade_shift or 100;
    --
    --ObjAnimate( "obj_match3", "pos_y", 0, 0, "", 
    --{ 
    --  0.0, 0, cur_y, 
    --  tme, 0, cur_y + shift
    --} );
    --
    --private.PanelBoosters_Hide();
    
    ObjAnimate( "obj_match3", "alp", 0, 0, "",
    { 
      0.0, 0, 1.0,
      tme, 0, 0.0
    } );
    
  end;

end;
--------------------------------------------------------------------------------
function private.Match3_ShadesOff ( func_after )

  func_after = func_after or function () end;

  if not ( private.match3_shades ) then

    func_after();
    return;

  end;

  if ( private.match3_shades_forced ) then

    func_after();
    return;

  end;
  
  local function func_notify ()

    func_after();
    mediator.Notify( "View_Stage_Match_ShadesOff" );
  
  end;

  private.match3_shades_forced = false;
  private.match3_shades = false;
  
  local tme = 0.3;
  
  int_panel_stage_progress_fight.Move_Far( tme );

  ObjAnimate( "obj_match3", "color_rgb", 0, 0, func_notify, 
  { 
    0.0, 0, 0.5, 0.5, 0.5,
    tme, 0, 1.0, 1.0, 1.0
  } );
  
  if ( private.match_custom_shade ) then
    
    --local cur_y = ui.obj.Position( "obj_match3" )[2];
    --local shift = private.match_custom_shade_shift or 100;
    --
    --ObjAnimate( "obj_match3", "pos_y", 0, 0, "", 
    --{ 
    --  0.0, 0, cur_y, 
    --  tme, 0, cur_y - shift
    --} );
    --
    --private.PanelBoosters_Show();
    
    ObjAnimate( "obj_match3", "alp", 0, 0, "",
    { 
      0.0, 0, 0.0,
      tme, 0, 1.0
    } );
    
  end;

end;
--------------------------------------------------------------------------------
function private.Match3_ShadesReset ()

  ObjStopAnimate( "obj_match3", "color_rgb" );
  ObjStopAnimate( "obj_match3", "alp" );
  ObjSet( "obj_match3", { alp = 1, color_r = 1, color_g = 1, color_b = 1 } );
  
end;
--------------------------------------------------------------------------------
function private.Match3_ShowEffect ( effect_name, loop, fx_params, func_after )

  private.is_match3_effect_active = true;

  local x, y       = v_match3.GetFieldCenter();
  local anm_effect = "anm_match3_effect";

  ui.obj.Attach( 
    anm_effect,
    "obj_match3"
  );
  
  local params =
  {
    res   = "assets/content/match3/fx/anm_"..effect_name,
    pos_x = x,
    pos_y = y
  };
  
  if ( fx_params ) then
    for key, value in pairs( fx_params ) do
    
      params[ key ] = value;
    
    end;
  end;
  
  ui.obj.Set(
    anm_effect,
    params
  );
  
  local function func_loop ()

    ui.anim.Play(
      anm_effect,
      effect_name,
      function()
        
        func_loop();
        
      end
    );
    if ( func_after ) then func_after(); end;
  
  end;
  
  local function func_once ()

    ui.anim.Play(
      anm_effect,
      effect_name,
      function()
        
        private.Match3_HideEffect();
        if ( func_after ) then func_after(); end;
        
      end
    );
  
  end;
  

  if ( loop ) then func_loop();
  else             func_once();
  end;
  
end;
--------------------------------------------------------------------------------
function private.Match3_HideEffect ()

  private.is_match3_effect_active = nil;

  local anm_effect = "anm_match3_effect";
  
  AnimStop( anm_effect );

  ui.obj.Attach( anm_effect, "hub_stage" );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.OnFeatureUpdated ()
  --ES> refactor registered how feature
  private.HandleActionPotions();
  
end;
--------------------------------------------------------------------------------
function private.GetDataBooster ( booster_name )
  
  local data = v_boosters.Boosters_GetInfo( booster_name );
  
  return data;
  
end;
--------------------------------------------------------------------------------
function private.HandleActionPotions ()
  
  if ( int_dialog_stage_start.IsActive() ) then
    int_dialog_stage_start.OnUpdateBoosters();
  end;
  
  if ( int_panel_stage_boosters.IsActive() ) then
    int_panel_stage_boosters.UpdateInfo( private.GetDataBooster );
  end;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Stage_ForceStart ( stage_name, first )
  
  local start_data = private.GetStartData( stage_name );
  private.score_old = start_data.score_old;

  if ( private.PreStart( stage_name, start_data.enemy ) ) then

    private.Start();
    
    if ( not first ) then
      
      private.PreOpen( private.Open );
      
    end;

    mediator.Notify(
      "View_Stage_DialogStartDeactivated",
      {
        name  = stage_name,
        start = true
      }
    );

  end;
  
end;
--------------------------------------------------------------------------------
function public.DialogStart_Show ( stage_name, database, dailyquest )

  mediator.Subscribe(
    "Model_Boosters_CountChanged",
    private.DialogStart_Update
  );

  local start_data = private.GetStartData( stage_name, database, dailyquest );

  private.score_old = start_data.score_old;

  local delegates = {

    Accept = function ( booster_list )

      local boosters = {};

      for key, value in pairs( booster_list ) do

        table.insert( boosters, key );

      end;

      v_boosters.SetSelected( boosters );

      if ( private.PreStart( stage_name, start_data.enemy ) ) then

        local func_StartStage = function ( is_instantly )

          local func_start = function () private.Start(); end;

          private.DialogStart_Hide( func_start, true, is_instantly );

          ui.input.Off( "zmv_map" );

          mediator.Notify(
            "View_Stage_DialogStartDeactivated",
            {
              name  = stage_name,
              start = true
            }
          );

        end;

        if ( not private.started ) then
          v_interface.Loading_Show(
            true, function () func_StartStage( true ); end
          );
        else
          func_StartStage();
        end;

      end;

    end,

    Reject = function ()

      mediator.Notify(
        "View_Stage_DialogStartDeactivated",
        {
          name  = stage_name,
          start = false
        }
      );

      private.DialogStart_Hide( private.Finish );

    end,

    BuyBooster = function ( booster_name )

      return v_boosters.Tap( "dialog", booster_name, stage_name );

    end,
  
    DataBooster = function ( booster_name )
    
      return private.GetDataBooster( booster_name );
      
    end,
    
    OnShow = function ()
    
      private.Social_Score_TryShow(
        "start",
        start_data.type,
        stage_name, 
        start_data.score_old
      );    

      local boosters = {};
      
      for b, booster in pairs( start_data.boosters ) do
        
        table.insert( boosters, booster.name );
        
      end;
    
      mediator.Notify(
        "View_Stage_DialogStartActivated",
        {
          name     = stage_name,
          boosters = boosters
        }
      );

    end
  };  

  common.delegate.Set( "int_dialog_stage_start", delegates );

  start_data.reward_chest = v_chests.Chest_Will( stage_name );
  
  local func_start = function ()
  
    -- TODO: Fix it in 1.12
    local show_event_marker = false;
    
    int_dialog_stage_start.Show( 
      stage_name,
      start_data,
      show_event_marker
    );

    v_log.Playhaven_LogPlacement( "load_level" );

  end; 

  v_interface.window.AddToQueue( "int_dialog_stage_start", func_start );

  common.sound.Play( "level_click" );

end;
--------------------------------------------------------------------------------
function private.DialogStart_Update ( booster_data )

  int_dialog_stage_start.Update( booster_data );

end;
--------------------------------------------------------------------------------
function private.DialogStart_Hide ( func_end, skip_map_enable )

  mediator.Unsubscribe( "Model_Boosters_CountChanged", private.DialogStart_Update );
    
  private.Social_Score_Hide();
  int_dialog_stage_start.Hide( func_end, skip_map_enable );

end;
--------------------------------------------------------------------------------
function private.GetStartData (
    stage_name,
    database,
    dailyquest,
    stage_state
  )

  local saved_data = m_stage.GetSavedData(  stage_name ) or {};
  local db_data    = m_stage.GetStaticData( stage_name );

  stage_state = stage_state or m_stage.GetStageState(
    stage_name,
    db_data,
    saved_data
  );

  local stateless = m_stage.Stateless_IsActive();
  local db_index;
  
  if ( stateless ) then db_index = 1;
  else                  db_index = stage_state;
  end
  
  -- Data for stage start dialog
  local start_data = {
    section     =  db_data.section,
    type        =  db_data.type,
    realtype    =  db_data.realtype,
    zone        =  db_data.zone,
    content     =  db_data.content,

    moves       =  db_data.state[ db_index ].moves,
    time        =  db_data.state[ db_index ].time,
    targets     =  db_data.state[ db_index ].targets,
    resources   =  db_data.state[ db_index ].chip_types,
    enemy       =  db_data.state[ db_index ].enemy,

    boosters    =  public.Boosters_GetStageList( stage_name, "start" ),

    states      = #db_data.state,
    state       =  stage_state,
    stateless   =  stateless,

    noreward    =  saved_data.noreward,

    score_old   =  nil,
    attempts    =  nil,
    color       =  nil, -- Dialog bg color

    reward      =  {}   -- List of reward lists for each state
  };

  -- FIXME: Is it used?
  if not ( dailyquest ) then 

    start_data.score_old = saved_data.score    or 0;
    start_data.attempts  = saved_data.attempts or 0;

  end;

  -- State and noreward
  if ( stateless and saved_data.completed ) then

    start_data.state     = 1;
    start_data.noreward  = true;

  end;

  if ( start_data.type == "fight" ) then

    start_data.state         = 1;
    start_data.enemy.content = v_enemies.Enemy_GetContentPackageName(
      start_data.enemy.name
    );

  end;

  -- Dialog BG color
  local enemy    = ( start_data.enemy or {} ).name;
  local zone     =   start_data.zone;
  local stg_type =   start_data.type;

  start_data.color = private.GetBackgroundColor( zone, stg_type );

  -- Trim not resources from chip_types
  for i = #start_data.resources, 1, - 1 do

    if not ( string.find( start_data.resources[ i ], "Resource" ) ) then

      table.remove( start_data.resources, i );

    end;

  end;

  -- Add resource form targets
  if ( ( start_data.targets ) and ( #start_data.targets > 0 ) and 
    ( #start_data.resources > 0 ) ) then
    for i = 1, #start_data.targets do
      if ( start_data.targets[ i ].type == "CollectResource" ) then
        local name = start_data.targets[ i ].name;
        local exist = false;
        for r = 1, #start_data.resources do
          if ( string.lower( start_data.resources[ r ] ) == name ) then
            exist = true;
          end;
        end;
        if ( not exist ) and ( string.find( name, "resource" ) ) then
          table.insert( start_data.resources, name );
        end;
      end;
    end;
  end;
  
  -- Generate rewards list
  local state_max = math.min( start_data.states, 3 );

  for i = 1, state_max do
    table.insert( start_data.reward, {} );
  end;

  -- Get unique rewards
  for s = 1, state_max do

    local state_rewards = db_data.state[ s ].reward or {};

    for r = 1, #state_rewards do

      local reward = state_rewards[ r ];

      if     ( reward.type == "quest" ) then

        local index = math.max( s, start_data.state );
        if ( _G.v_quests and
             v_quests.IsStageItemActive( reward.name, stage_name, "stage", start_data.section ) ) then

          table.insert( start_data.reward[ index ], reward );

        end;
        
      elseif ( reward.type ~= "gold"  ) then
        
        if ( reward.type == "worker" ) then
          local mine_name = v_mines.Mine_GetNameByStage( stage_name );
          local mine_data = v_mines.Mine_Data( mine_name );
          if ( ( mine_data.progress or 0 ) < mine_data.progress_max ) then
            table.insert( start_data.reward[ s ], reward );
          end;
        else
          table.insert( start_data.reward[ s ], reward );
        end;

      end;

    end;
    
  end;

  -- Append generic rewards
  local gold_bonus = v_player.Stats_GetPlayerData().gold;

  if ( db_data.type == "ride" ) then

    local rewards = m_stage.GetRewardGoldSettings();

    for r = 1, #rewards do

      if ( start_data.reward[ r ] ) then

        local gold_count = rewards[ r ].count;

        table.insert( start_data.reward[ r ], 
        { 
          name  = "gold", 
          type  = "gold", 
          count = gold_count + gold_bonus
        } );

      end;

    end;

  elseif ( db_data.type == "fight" ) then

    if ( db_data.realtype == "misc" ) then

      -- Replace misc fight reward by generic
      start_data.reward[ start_data.state ] = {};

    end;

    local default_fight_reward;

    if ( #start_data.reward[ start_data.state ] == 0 ) then

      local fight_gold_count = v_fights.GetDefaultGoldRewardCount();

      default_fight_reward = {
        type  = "gold", 
        name  = "gold",
        count = fight_gold_count + gold_bonus
      };

    end;

    local is_spell = v_spells.Spell_GetNameByUnlock( stage_name );

    default_fight_reward = tern( is_spell )(
      {
        type = "item",
        name = "spellbook"
      },
      default_fight_reward
    );

    if ( default_fight_reward ) then

      table.insert(
        start_data.reward[ start_data.state ], default_fight_reward
      );

    end;

  end;
  
  m_stage.Kontagent_SetStageData( start_data );
  
  return start_data;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.DialogExit_TryShow ()

  -- TODO: check for final animations

  private.DialogExit_Show(
    int_panel_stage_settings.MenuClose,
    int_panel_stage_settings.MenuClose
  );

end;
--------------------------------------------------------------------------------
function private.DialogExit_Show ( func_on_reject, func_on_accept )
  
  local stage_data = m_stage.GetCurrentStageData();
  local targets    = m_stage.GetCurrentTargets();
  local moves_done = m_stage.MovesDone_GetCount();
  local hearts_cnt = v_hearts.GetCount();
  local hearts_max = v_hearts.GetMaxCount();

  local func_accept = function ()
    
    local func_reject_dialog = function ()
    
      if ( func_on_accept ) then 
        func_on_accept();
      end;

      private.Match3_SetInput( false );
      private.PanelBoosters_SetInput( false, "DialogExit_Show" );
      --public.PanelSettings_SetInput( false );

      private.DialogFriends_Hide();

      private.DialogExit_Hide( function ()

        v_hearts.Unblock( "abandon" );
        mediator.Notify( "View_Stage_StageAbandon", { stage = stage_data.name, moves_done = moves_done } );
        private.Finish();

      end );

      local score_target = 0;
      local moves_target = 0;

      if ( stage_data.type == "ride" ) or ( stage_data.type == "bonus" ) then

        score_target = ( m_stage.GetTargetByType( "GainScore" ) or {} ).count or score_target;
        moves_target = stage_data.moves + m_stage.Kontagent_GetAdditionalMoves();

      end;

      local getscore = function ( stage_type )

        if ( stage_type == "fight" ) then
          
          local hero_data = m_stage.GetCurrentHero();

          return { ( hero_data.health_current..", "..hero_data.health ), 1 };

        end;

        return { ( m_stage.GetCurrentScore()..", "..score_target ), 0 };

      end;
      
      local scoredata  = getscore( stage_data.type );
      local other_stage = false;
      local section = stage_data.section;
      local scene;
      
      if ( _G.v_dailyquest ) then
        other_stage = tern( stage_data.realtype == "dailyquest" )( 
            ( v_dailyquest.Challenge_GetTotalCount() - v_dailyquest.Challenge_GetOpenedCount() + 1 ), 
            false );
      end;
      if ( _G.v_minievent ) then
        other_stage = tern( stage_data.realtype == "minievent" )( 
            ( #( v_minievent.GetRelatedEventData() or { stages_completed = {} } ).stages_completed + 1 ), 
            other_stage );
        section = tern( stage_data.realtype == "minievent" )(
            ( v_minievent.GetRelatedEventData() or {} ).name, section );
        scene = tern( stage_data.realtype == "minievent" )( 
            ( v_minievent.GetRelatedEventData() or {} ).scene, nil );
      end;
            
      v_log.Kontagent_LogEvent( "stage",
        {
          section = section,
          stage   = stage_data.name,
          other_stage = other_stage,
          scene   = scene,
          name    = "abandon",
          state   = m_stage.Kontagent_GetStageState( stage_data.name ),
          score   = scoredata[ 1 ],
          is_boss = scoredata[ 2 ],
          moves   = moves_done..", "..moves_target,
          boost   = m_stage.Kontagent_GetBoostersUsed(),
          cells   = m_stage.Kontagent_GetCellsDestroyed(),
          items   = m_stage.Kontagent_GetItemsDropped( "Key" ),
          flies   = m_stage.Kontagent_GetResourceCollected(),
          bombs   = m_stage.Kontagent_GetExplosivesDestroyed(),
          woodchests   = m_stage.Kontagent_GetItemsDropped( "woodchest" ),
          silverchests = m_stage.Kontagent_GetItemsDropped( "silverchest" ),
          goldchests   = m_stage.Kontagent_GetItemsDropped( "goldchest" ),
          additional_moves = m_stage.Kontagent_GetAdditionalMoves()
      } );

      v_game.GameLog_Add( "stage_abandon",
       {
         stage   = stage_data.name
       } );
        
      v_log.Playhaven_LogPlacement( "level_quit" );

    end;
      
    if (_G.v_minievent) and ( v_minievent.Crown_IsAvailableMinievent() ) and
      ( v_minievent.Crown_GetCurrentValue() > 0 ) then
      v_minievent.Crown_DialogContinue_Show( func_reject_dialog, function () end );
      return;
    end;
    
    func_reject_dialog();
    
  end;

  local func_reject = function ()

    private.DialogExit_Hide( func_on_reject, true );

  end;

  local state_window_failed = ( moves_done > 0 ) or ( v_hearts.IsBlock() );
  
  v_interface.window.Show(
    "int_dialog_stage_exit",
    function ()

      int_dialog_stage_exit.Show(
        stage_data,
        state_window_failed,
        hearts_cnt,
        hearts_max,
        v_player.Skin_GetCurrent(),
        v_hearts.Immortality_IsActive()
      );

    end,
    nil,
    {
      Accept = func_accept,
      Reject = func_reject
    }
  );

end;
--------------------------------------------------------------------------------
function private.DialogExit_Hide ( func_end, no_animation )

  int_dialog_stage_exit.Hide( function ()

    v_interface.window.Hide( "int_dialog_stage_exit", func_end );

  end, no_animation );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.DialogComplete_Show ()
  
  private.finalized = false;
  
  if ( m_stage.IsWin() ) then

    private.FinalizeWin();
    private.DialogWin_Show();
    private.DialogFriends_Hide();

  else

    if ( m_stage.IsCanContinue() ) then

      private.DialogContinue_Show();

    else

      private.Dialog_Fail_Show();
      private.DialogFriends_Hide();

    end;

  end;
  
  private.PopupBonus_Hide();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.DialogTargets_Show ( state, targets, enemy_name )

  local func_show = function ()
  
    int_dialog_stage_targets.Show(
      state, 
      targets, 
      enemy_name,
      ( m_stage.GetCurrentStageData().section or "section_1" ),
      private.DialogTargets_Hide
    );

  end;

  v_interface.window.Show(
    "int_dialog_stage_targets",
    func_show
  );

end;
--------------------------------------------------------------------------------
function private.DialogTargets_Hide ()

  local func_hide = function ()

    private.ShowStartEffects();

  end;

  v_interface.window.Hide( "int_dialog_stage_targets", func_hide );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.SetPresetEffect ( func_preset )
  
  private.preset_effects = func_preset;
  
end;
--------------------------------------------------------------------------------
function private.ShowStartEffects ()
  
  v_interface.Lock( true );
  
  local stage_data = m_stage.GetCurrentStageData();
  local stage_name = stage_data.name;
  
  if ( private.popup_show                     ) and
     ( private.prev_popup_stage ~= stage_name ) then
    
    int_panel_stage_boosters.Popup_Show();
    
    private.prev_popup_stage = stage_name;
    
    private.popup_show = false;
  
  end;
  
  local effects = {};

  table.insert( effects, private.SkinsEffects( stage_data ) );
  table.insert( effects, private.BoostersEffects( stage_data ) );
  table.insert( effects, private.LastEffects( stage_data ) );
  
  local func_run_callback_list = function ()
    if ( #effects > 0 ) then
      common.RunCallbackList( effects );
    end;
  end;
  
  if ( private.preset_effects ) then
    
    local func_preset_effects = private.preset_effects;
    private.preset_effects = nil;
    func_preset_effects( func_run_callback_list );
    return;
    
  end;
  
  func_run_callback_list();
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.BoostersEffects ( stage_data )

  return function ( func_after )
  
    if ( private.Boosters_SuperhitExist() ) then
      ne.msg.Subscribe( "Event_Match3_FieldEstablished", private.Match3FieldEstablished );
      private.applies_start_boosters_fight = { count = 1 };
      private.AddStartBoosterOnField();
    end;
    
    common.MessageSend( Command_Match3_DidBecomeActive );
    
    local stats_hero  = v_player.Stats_GetPlayerData();
    if ( stats_hero.minievent_progress_crowns ) and 
        ( stage_data.realtype ~= "minievent" ) and 
        ( stage_data.realtype ~= "dailyquest" ) then
      m_stage.CreateBonusesOnField( 
        stats_hero.minievent_progress_crowns, 
        "minievent_progress_crowns" 
      );
      private.AddStartBoosterOnField();
      v_boosters.UseCrownBooster();
    end;
    
    local stage_name = m_stage.GetCurrentStageName();
    
    local source = stage_name;
    if ( stage_name:find( "_minievent" ) ) then
      source = "minievent_"..v_minievent.GetRelatedEventData().name;
    end;

    private.Boosters_TryUse( v_boosters.Use, source );
  
    if ( func_after ) then func_after(); end;
  end;
  
end;
--------------------------------------------------------------------------------
function private.Boosters_TryUse ( func_execute, source )

  local boosters_list = private.selected_boosters;

  local boosters = 
  {
    movesup  = m_stage.Boosters_Use_MovesUp,
    superhit = m_stage.Boosters_Use_SuperHit
  };
  
  if ( boosters_list ) then
    
    for b = 1, #boosters_list do

      local booster_name = boosters_list[ b ];

      if ( boosters[ booster_name ] ) then

        if ( func_execute ) then

          func_execute( 
            { 
              result = true,
              name = booster_name, 
              func_end = boosters[ booster_name ]
            }, source );

        end;

      end;

    end;

    private.selected_boosters = nil;

  end;

end;
--------------------------------------------------------------------------------
function private.Boosters_SuperhitExist ()
  
  local boosters_list = private.selected_boosters;
  
  for i = 1, #boosters_list do
    if ( boosters_list[ i ] == "superhit" ) then
      return true;
    end;
  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Match3FieldEstablished ( msg, params )
  
  ne.msg.Unsubscribe( "Event_Match3_FieldEstablished", private.Match3FieldEstablished );
  
  private.ExecuteStartBoosterOnField();
  
end;
--------------------------------------------------------------------------------
function private.AddStartBoosterOnField ()
  
  if ( private.applies_start_boosters_fight ) then
    
    private.applies_start_boosters_fight.count = private.applies_start_boosters_fight.count + 1;
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.ExecuteStartBoosterOnField ()
  
  if ( private.applies_start_boosters_fight ) then
    
    private.applies_start_boosters_fight.count = private.applies_start_boosters_fight.count - 1;
    
    private.StartBoostersApplied();
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.StartBoostersApplied ()
  
  if ( private.applies_start_boosters_fight ) and ( private.applies_start_boosters_fight.count <= 0 ) then
    
    private.applies_start_boosters_fight = nil;
    
    mediator.Notify( "View_Match3_FieldEstablished" );
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.SkinsEffects ( stage_data )
  
  if ( stage_data.type ~= "fight" ) then
    
    local stats_hero  = v_player.Stats_GetPlayerData();
    local extra_moves = stats_hero.moves + stats_hero.event_moves;
    
    if ( stage_data.realtype ~= "minievent" ) then
      extra_moves = stats_hero.moves + stats_hero.saga_moves;
    else
      extra_moves = stats_hero.moves + stats_hero.event_moves;
    end;
    
    if ( extra_moves ~= 0 ) then
    
      local skin_name = v_player.Skin_GetCurrent();
      local skin_dir  = v_player.Skin_GetContent( skin_name );
      local resource  = "assets/content/%s/%s_item"
      local description = "str_skinbonus_moves_description"
      local sign = tern( extra_moves > 0 )( "+", "-" )
      
      if ( ( extra_moves == 1 ) or ( extra_moves == -1 ) ) then
        description = "str_skinbonus_moves_description_one";
      end;
      
      return function ( func_after )
        
          local params =
          {
            tt_title       = "str_skinbonus_moves_title",
            tt_bonus_value = StringGet( description, sign, math.abs( extra_moves ) ),
            rt_item_skin   = resource:format( skin_dir, skin_name ),
            rt_item_potion = tern( stats_hero.saga_moves ~= 0 )( "assets/content/items/potion_moves", "" )
          };
        
          private.Match3_ShowEffect(
            "skinbonus", false, params,
            function ()
          
              m_stage.Skinbonus_MovesUp( extra_moves );
              if ( func_after ) then func_after(); end;
          
            end
          );
        end;
    
     end;
     
  end;
  
  return function ( func_after )
    if ( func_after ) then func_after(); end;
  end;
    
end;
--------------------------------------------------------------------------------
function private.LastEffects ( stage_data )
  
  local func_end = function ()
    v_interface.Lock( false );
    mediator.Notify( "View_Stage_DialogTargetsHided" );
  end;
  
  if ( stage_data.windflaw                            ) and
     ( not public.Pending_IsActive( "tutorial_wind" ) ) then
     
     return function ( func_after )
       func_end();
       private.Match3_ShowEffect( "windflaw", false, nil, func_after );
     end;

  end;
  
  return function ( func_after )
    func_end();
    if ( func_after ) then func_after(); end;
  end;
   
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Video_RecoverFinishedAttempts ()

  private.RecoveredHeartsAttemptsVideo();
  private.FortuneWheel_RecoverAttemptsVideo();
  private.DialogContinue_RecoverAttemptsVideo();
  
end;
--------------------------------------------------------------------------------
-- function ... ()  
--------------------------------------------------------------------------------
function private.DialogContinue_Show ()

  local stage_name = m_stage.GetCurrentStageName();
  local stage_type = private.GetCurrentStageType();
  
  local continue_name = tern( stage_type == "fight" )( "extralives", "extramoves" );
  local continue_data = {
    name  = continue_name,
    price = v_boosters.GetPrice( continue_name, 1 ),
    count = v_boosters.GetCount( continue_name )
  };

  local func_use_continue_booster = function ()
    
    v_game.Kontagent_VideoRefused( "extramoves" );
    private.FortuneWheel_Hide();
    private.DialogContinue_Hide();
    int_panel_stage_top.SetVisible();

    private.PopupBonus_Show();

    private.Match3_SetInput( true );
    v_boosters.TryUse( continue_name, stage_name );
    m_stage.TryUseContinue( continue_name );
    private.PanelBoosters_SetInput( true, "Use Continue" );

    if ( stage_type == "fight" ) then
      int_panel_stage_progress_fight.GameContinue();
    end;

  end;

  local func_pause_check = function ()
    if ( public.IsPaused() ) then
      int_panel_stage_settings.MenuClose();
    end;
  end;

  local delegates = {

    OnShow = function ()

      func_pause_check();

    end,
    BuyBooster = function ()

      if ( v_boosters.GetCount( continue_name ) <= 0 ) then
        if ( v_boosters.TryBuy( continue_name, 1, stage_name ) ) then
          func_use_continue_booster();
        end;
      else
        func_use_continue_booster();
      end;
    
    end,
    Reject = function ()
    
      local func_reject_dialog = function ()
        v_game.Kontagent_VideoRefused( "extramoves" );
        private.FortuneWheel_Hide();
        int_panel_stage_top.SetVisible();

        private.DialogContinue_Hide( private.Dialog_Fail_Show );
        
        -- Send message only if purchase available.
        if ( v_boosters.GetCount( continue_name ) <= 0 ) then
          
          v_log.Kontagent_LogEvent( "economy", 
          {
            name = continue_name,
            description = "extra_moves_offer",
            level = 1,
            value = 0,
            purchase = 0
          } );
          
        end;
      end;
      
      if (_G.v_minievent) and ( v_minievent.Crown_IsAvailableMinievent() ) and
        ( v_minievent.Crown_GetCurrentValue() > 0 ) then
        v_minievent.Crown_DialogContinue_Show( func_reject_dialog, function () end );
        return;
      end;
      
      func_reject_dialog();

    end,
    ShowVideo = function ()
      
      private.DialogContinue_ShowVideo();
      
    end,
    GetAttemptsVideo = function ()
    
      local count, max_count = v_player.GetAttemptsVideo();
      return count, max_count;
      
    end,
    IsVideoAvailable = function ()
      
      local result = v_game.IsVideoAvailable() and v_player.IsAvailableAttemptsVideo();
      return result;
      
    end,
    VideoCanWatch = function ( value )
      
      v_game.Kontagent_VideoCanWatch( value );
      
    end,
    GetTargets = function ()
      local  targets = m_stage.GetCurrentTargets();
      return targets;
    end  
  };

  common.delegate.Set( "int_dialog_stage_continue", delegates );
  
  local func_start = function ()

    func_pause_check();
    
    if ( public.IsActiveFortuneWheel() ) then
      private.FortuneWheel_Show();
    end;
    
    int_panel_stage_top.SetInvisible();
    
    if ( continue_data.count == 0 ) then

      v_interface.panel.top.Highlight( "gold" );
      v_interface.panel.top.Show();

    end;
    
    common.delegate.Set( "int_dialog_stage_continue", delegates );
    
    int_dialog_stage_continue.Show( continue_data );

  end;

  if ( stage_type == "ride"  ) or 
     ( stage_type == "bonus" ) then

    int_panel_stage_progress_ride.Update_StageFailed();

  end;

  v_interface.window.AddToQueue(
    "int_dialog_stage_continue",
    func_start
  );

end;
--------------------------------------------------------------------------------
function private.DialogContinue_Hide ( func_end )

  local stage_type = private.GetCurrentStageType();
  
  if ( stage_type == "ride"  ) or 
     ( stage_type == "bonus" )then

    int_panel_stage_progress_ride.Update_HeroRespawn();
        
    private.popup_show = true;

  end;

  v_interface.panel.top.Hide( true );
  
  int_dialog_stage_continue.Hide( function ()

    v_interface.window.Hide( "int_dialog_stage_continue", func_end );

  end );

end;
--------------------------------------------------------------------------------
function private.DialogContinue_Update ()
  
  int_dialog_stage_continue.Update();
  
end;
--------------------------------------------------------------------------------
function private.DialogContinue_RecoverAttemptsVideo ()
  
  private.DialogContinue_Update();
  
end;
--------------------------------------------------------------------------------
function private.DialogContinue_ShowVideo ()
  
  v_game.ShowVideo( "extramoves", private.DialogContinue_ShowVideoFinished );
  
end;
--------------------------------------------------------------------------------
function private.DialogContinue_ShowVideoFinished ()
  
  local count, max_count = v_player.GetAttemptsVideo();
  local attempts = math.min( count + 1, max_count );
  
  v_player.UpdateAttemptsVideo( attempts );
  
  private.DialogContinue_Update();
  
  local stage_name    = m_stage.GetCurrentStageName();
  local stage_type    = private.GetCurrentStageType();
  local continue_name = tern( stage_type == "fight" )( "extralives", "extramoves" );
  
  private.FortuneWheel_Hide();
  private.DialogContinue_Hide();
  int_panel_stage_top.SetVisible();
  
  private.PopupBonus_Show();
  
  private.Match3_SetInput( true );
  v_boosters.TryChangeCount( continue_name, 1, "video_reward" );
  v_boosters.TryUse( continue_name, "video_reward" );
  m_stage.TryUseContinue( continue_name );
  private.PanelBoosters_SetInput( true, "Use Continue" );
  
  if ( stage_type == "fight" ) then
    int_panel_stage_progress_fight.GameContinue();
  end;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.IsActiveFortuneWheel ()
  
  local state = m_fortunewheel.IsActive();
  return state;
  
end;
--------------------------------------------------------------------------------
function public.FortuneWheel_SetPopupEnabled ( value )
  
  private.fortunewheel_popup_enabled = value;
  
end;
--------------------------------------------------------------------------------
function public.FortuneWheel_IsOnCooldown ()

  local  is_cooldown = m_fortunewheel.Cooldown_IsActive();
  return is_cooldown;

end;
--------------------------------------------------------------------------------
function private.FortuneWheel_Show ()
  
  local stage_name = m_stage.GetCurrentStageName();
  
  local wheel_data = m_fortunewheel.GetInfo();
  wheel_data.state = tern( m_fortunewheel.Cooldown_IsActive() )( "recover", "ready" );
  wheel_data.notify_opened = function ()
    mediator.Notify( "View_Stage_FortuneWheel_Opened" );
  end;
  wheel_data.notify_ready = function ()
    mediator.Notify( "View_Stage_FortuneWheel_ReadyToSpin" );
  end;
  
  local stage_type = private.GetCurrentStageType();  
  local continue_name = tern( stage_type == "fight" )( "extralives", "extramoves" );
  local booster_count = v_boosters.GetCount( continue_name );
  
  wheel_data.continue_name = continue_name;
  
  local func_turn = function ()
  
    mediator.Notify( "View_Stage_FortuneWheel_Turn" );
    
    local hearts_now = v_hearts.GetCount();
    local hearts_max = v_hearts.GetMaxCount();
    local reward     = m_fortunewheel.TurnWheel( hearts_now, hearts_max );
    
    return reward;
    
  end;

  local func_reset_cooldown = function ()
  
    local is_cooldown = m_fortunewheel.Cooldown_IsActive();
    local price       = m_fortunewheel.GetInfo().spin_cost;
    
    if ( is_cooldown                             ) and 
       ( v_player.Gold_TryChangeCount( - price ) ) then
    
      m_fortunewheel.Cooldown_Stop();

      v_log.Kontagent_LogEvent( "economy", 
      {
        description = "premium_currency_spend",
        category    = "fortunewheel",
        value       = - price,
        name        = "reset"
      } );

      return true;
    
    end;
    
    return false;
  
  end;
  
  local func_use_continue_booster = function ( booster_name )

    private.PopupBonus_Show();

    private.Match3_SetInput( true );
  
    v_boosters.TryUse( booster_name, stage_name );

    m_stage.TryUseContinue( continue_name );

    private.PanelBoosters_SetInput( true, "Use Continue" );

  end;

  local func_get_reward = function ( item, func_end )

    if ( item.name == "heart" ) then

      v_hearts.TryChangeCount( item.count or 1 );
      
      common.Defer(
        "hide_left_panel",
        0.7,
        function()
          
          v_interface.panel.left.Hide(
            function ()
              func_end();
              v_hearts.Panel_EnableTimer();
            end
          );
         
        end
      );

    elseif ( item.name == "extramoves" ) then

      v_boosters.TryChangeCount( item.name, item.count or 1, stage_name );
      private.DialogContinue_Hide();
      private.FortuneWheel_Hide();
      func_use_continue_booster( item.name );

    elseif ( item.name == "gold" ) then

      if ( booster_count > 0 ) then
      
        v_interface.panel.top.Hide( nil, func_end );
        
      else
        
        if ( func_end ) then
          
          func_end();
          
        end;
        
      end;
      
      v_player.Gold_TryChangeCount( item.count or 1 );

      local stage_data = m_stage.GetCurrentStageData();
      local stage_name = stage_data.name;

      v_log.Kontagent_LogEvent( "economy", 
      {
        description = "premium_currency_collect",
        category    = "fortunewheel",
        value       = item.count or 1,
        name        = "reward"
      } );

    end;

    mediator.Notify( "View_Stage_FortuneWheel_Reward" );

  end;

  local func_detach_res_panel = function ()

    v_interface.panel.top.Hide( true );

  end;

  local func_attach_res_panel = function ()

    v_interface.panel.top.Highlight( "gold" );
    v_interface.panel.top.Show();

  end;

  local func_get_flight_target = function ( reward_name )

    if ( reward_name == "extramoves" ) or ( reward_name == "extralives" ) then 

      func_detach_res_panel();

      if ( continue_name == "extralives" ) then

        return "anm_int_panel_stage_progress_fight_healthbars_hero_icon"; 

      else

        return "anm_int_panel_stage_top_moves"; 

      end;

    elseif ( reward_name == "gold" ) then

      func_attach_res_panel();

      return v_interface.panel.top.GoldCounterName();

    elseif ( reward_name == "heart" ) then

      func_detach_res_panel();

      v_hearts.Panel_DisableTimer();
      v_interface.panel.left.Show( "hearts" );

      return v_hearts.Panel_ButtonName();

    end;

  end;

  local popup_funcs = {
    show = function ( source, offset, scale ) 
      if ( private.fortunewheel_popup_enabled ) then
        ui.popup.Cloud_Show(
          "hint", 
          "anm_int_dialog_stage_fortunewheel",
          "str_int_popup_cloud_wheel_"..source,
          nil,
          offset,
          scale,
          nil,
          false,
          true
        );
        
      end;
    end,
    hide = function ()
      ui.popup.Cloud_Hide();
    end
  
  };

  local func_hero_fun = function ()
  
    if ( stage_type == "fight" ) then
      int_panel_stage_progress_fight.Hero_Fun_Wheel();
    else
      int_panel_stage_progress_ride.Hero_Fun_Wheel();
    end;

  end;

  local function func_show_panel_stage_top ()
    
    int_panel_stage_top.SetVisible();
    
  end;
  
  local function func_hide_panel_stage_top ()
    
    int_panel_stage_top.SetInvisible();
    
  end;
  
  local function func_maximize ()
    
    v_interface.window.Maximize( "int_dialog_stage_fortunewheel" );
    
  end;
  
  local function func_minimize ()
    
    v_interface.window.Minimize( "int_dialog_stage_fortunewheel" );
    
  end;

  local delegates = {

    ShowVideo = function ()
      
      private.FortuneWheel_ShowVideo();
      
    end,
    GetAttemptsVideo = function ()
    
      local  count, max_count = v_player.GetAttemptsVideo();
      return count, max_count;
      
    end,
    IsRecover = function ()
    
      local  is_recovering = m_fortunewheel.Cooldown_IsActive();
      return is_recovering;
      
    end,
    IsVideoAvailable = function ()
      
      local  result = v_game.IsVideoAvailable() and v_player.IsAvailableAttemptsVideo();
      return result;
      
    end,
    GetCooldownSeconds = function ()

      local  seconds = m_fortunewheel.Cooldown_GetRemainingTime();
      return seconds;

    end,
    VideoCanWatch = function ( value )
      
      v_game.Kontagent_VideoCanWatch( value );
      
    end,
    VideoRefused = function ()
      
      v_game.Kontagent_VideoRefused( "fortunewheel" );
      
    end
  };

  common.delegate.Set( "int_dialog_stage_fortunewheel", delegates );
  
  local func_start = function ()

    int_dialog_stage_fortunewheel.Show( 
      wheel_data,
      func_maximize,
      func_minimize,
      func_turn,
      func_reset_cooldown,
      func_get_reward,
      func_get_flight_target,
      popup_funcs,
      func_hero_fun,
      func_show_panel_stage_top,
      func_hide_panel_stage_top,
      func_attach_res_panel,
      func_detach_res_panel
    );

  end;
  
  v_interface.window.Show(
    "int_dialog_stage_fortunewheel",
    func_start,
    { open_minimized = true }
  );  

end;
--------------------------------------------------------------------------------
function private.FortuneWheel_Hide ( func_end )
  
  int_dialog_stage_fortunewheel.Hide( function ()

    v_interface.panel.top.Hide( true );

    v_interface.window.Hide( "int_dialog_stage_fortunewheel", func_end );

  end );
  
end;
--------------------------------------------------------------------------------
function private.FortuneWheel_Update ( reason )
  
  int_dialog_stage_fortunewheel.Update( reason );
  
end;
--------------------------------------------------------------------------------
function private.FortuneWheel_RecoverAttemptsVideo ()
  
  private.FortuneWheel_Update( "video_attempts" );
  
end;
--------------------------------------------------------------------------------
function private.FortuneWheel_ShowVideo ()
  
  v_game.ShowVideo( "fortunewheel", private.FortuneWheel_ShowVideoFinished );
  
end;
--------------------------------------------------------------------------------
function private.FortuneWheel_ShowVideoFinished ()
  
  local count, max_count = v_player.GetAttemptsVideo();
  local attempts = math.min( count + 1, max_count );
  
  v_player.UpdateAttemptsVideo( attempts );
  m_fortunewheel.Cooldown_Stop();
  
  private.DialogContinue_Update();
  
  v_log.Kontagent_LogEvent( "economy", 
  {
    description = "premium_currency_spend",
    category    = "fortunewheel",
    value       = 0,
    name        = "reset",
    source      = "video_reward"
  } );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Dialog_Fail_Show ()
  
  local stage_data = m_stage.GetCurrentStageData();
  local stage_name = stage_data.name;

  stage_data.explosion_field = m_stage.IsExplosionField();
  
  local database = "stages";

  if ( string.find( stage_name, "fight" ) ) then

    database = "fights";

  end;

  local targets     = m_stage.GetCurrentTargets();
  local heart_count = v_hearts.GetCount();
  local heart_max   = v_hearts.GetMaxCount();
  
  v_hearts.Unblock( "fail" );
  
  local func_try_show_dialog_start = function ()

    if ( string.find( stage_name, "daily" ) ) then
      public.TryStart( stage_name, m_stage.GetCurrentEnemy() );
    else
      public.DialogStart_Show( stage_name, database ); 
    end;

  end;
  
  local func_try_close = function ()
    v_game.Kontagent_VideoRefused( "hearts" );
    v_hearts.HeartsOffers_Hide( true );
    private.Dialog_Fail_Hide( private.Finish );
  end;

  local function func_check_retry ( stage_name )

    local is_retry_allowed = true;

    if ( string.find( stage_name, "minievent" ) ) then

      is_retry_allowed = v_minievent.Stage_IsRetryAllowed();

    end;

    if ( v_hearts.GetCount() < 1 ) then

      is_retry_allowed = false;

    end;

    return is_retry_allowed;

  end;
  
  local delegates = {
  
    Retry = function ()

      if ( string.find( stage_name, "minievent" ) ) then
        
        local is_retry_allowed = func_check_retry( stage_name );

        if ( not is_retry_allowed ) then

          func_try_close();

        else
          v_game.Kontagent_VideoRefused( "hearts" );
          private.Dialog_Fail_Hide(
            function ()

              private.OnFinish();
              v_minievent.DialogStart_Show();

            end
          );

        end;

      else
        v_game.Kontagent_VideoRefused( "hearts" );
        private.Dialog_Fail_Hide( function () 
          private.OnFinish();
          func_try_show_dialog_start();
        end );
      end;

      common.sound.Play( "simply_click" );

    end,
    Restore = function ()
      
      if ( string.find( stage_name, "minievent" ) ) then
        func_try_close();
      else
        v_game.Kontagent_VideoRefused( "hearts" );
        private.Dialog_Fail_Hide( function ()

          local func_OnHide = function ( result )

            private.FinalizeWin();
          
            if ( string.find( stage_name, "daily" ) ) and not ( result ) then
              private.Finish();
              return;
            end;
            func_try_show_dialog_start();

          end;

          v_hearts.DialogMain_Show( func_OnHide );

        end );
      end;

      common.sound.Play( "simply_click" );
      
    end,
    Close = function ()

      func_try_close();

    end,
    ShowVideo = function ()
      
      v_hearts.ShowVideo();
      
    end,
    IsImmortality = function ()
      
      local is_immortality = v_hearts.Immortality_IsActive();
      return is_immortality;
      
    end,
    GetHeartsData = function ()
      
      local count     = v_hearts.GetCount();
      local max_count = v_hearts.GetMaxCount();
      return count, max_count;
      
    end,
    GetAttemptsVideo = function ()
    
      local count, max_count = v_hearts.GetAttemptsVideo();
      return count, max_count;
      
    end,
    EventTimeRemaining = function ()
      
      local time_remains, type_timer = v_hearts.EventTimeRemaining();
      return time_remains, type_timer;
      
    end,
    IsEventExist = function ()
      
      return v_hearts.IsEventExist();
      
    end,
    IsVideoAvailable = function ()
      
      local result = v_game.IsVideoAvailable() and v_player.IsAvailableAttemptsVideo();
      return result;
      
    end,
    VideoCanWatch = function ( value )
      
      v_game.Kontagent_VideoCanWatch( value );
      
    end
  };

  common.delegate.Set( "int_dialog_stage_fail", delegates );

  local func_start = function ()
    
    local placement = "level_fail";
    placement = tern( stage_data.realtype == "minievent" )( "event_fail", placement );
    placement = tern( stage_data.realtype == "dailyquest" )( "arena_fail", placement );

    v_log.Playhaven_LogPlacement( placement );

    local is_retry_allowed = func_check_retry( stage_name );

    int_dialog_stage_fail.Show( 
      stage_data, 
      targets, 
      heart_count,
      heart_max,
      v_player.Skin_GetCurrent(),
      is_retry_allowed
    );
    
    v_hearts.HeartsOffers_Show( true );
    
  end;
  ------------------------------------------------------------------------------

  int_panel_stage_boosters.StopAnim();
  public.Soundtrack_Stop();
  
  common.sound.Play( "assets/content/hero/aud_hero_win_ride" );

  local score_target = 0;
  local moves_target = 0;

  if ( stage_data.type == "ride" ) then

    score_target = ( m_stage.GetTargetByType( "GainScore" ) or {} ).count or score_target;
    moves_target = stage_data.moves + m_stage.Kontagent_GetAdditionalMoves();

  end;

  local getscore = function ( stage_type )

    if ( stage_type == "fight" ) then
      
      local hero_data = m_stage.GetCurrentHero();

      return { ( hero_data.health_current..", "..hero_data.health ), 1 };

    end;

    return { ( m_stage.GetCurrentScore()..", "..score_target ), 0 };

  end;
  
  local scoredata = getscore( stage_data.type );
  local other_stage = false;
  local section = stage_data.section;
  local scene;
  
  if ( _G.v_dailyquest ) then
    other_stage = tern( stage_data.realtype == "dailyquest" )( 
      ( v_dailyquest.Challenge_GetTotalCount() - v_dailyquest.Challenge_GetOpenedCount() + 1 ), 
      false );
  end;
  if ( _G.v_minievent ) then
    other_stage = tern( stage_data.realtype == "minievent" )( 
        ( #( v_minievent.GetRelatedEventData() or { stages_completed = {} } ).stages_completed + 1 ), 
        other_stage );
    section = tern( stage_data.realtype == "minievent" )(
        ( v_minievent.GetRelatedEventData() or {} ).name, section );
    scene = tern( stage_data.realtype == "minievent" )( 
        ( v_minievent.GetRelatedEventData() or {} ).scene, nil );
  end;
      
  v_log.Kontagent_LogEvent( "stage",
  {
    section = section,
    stage   = stage_data.name,
    other_stage = other_stage,
    scene   = scene,
    name    = "fail",
    state   = m_stage.Kontagent_GetStageState( stage_data.name ),
    score   = scoredata[ 1 ],
    is_boss = scoredata[ 2 ],
    moves   = m_stage.MovesDone_GetCount()..", "..moves_target,
    boost   = m_stage.Kontagent_GetBoostersUsed(),
    cells   = m_stage.Kontagent_GetCellsDestroyed(),
    items   = m_stage.Kontagent_GetItemsDropped( "Key" ),
    flies   = m_stage.Kontagent_GetResourceCollected(),
    bombs   = m_stage.Kontagent_GetExplosivesDestroyed(),
    woodchests   = m_stage.Kontagent_GetItemsDropped( "woodchest" ),
    silverchests = m_stage.Kontagent_GetItemsDropped( "silverchest" ),
    goldchests   = m_stage.Kontagent_GetItemsDropped( "goldchest" ),
    additional_moves = m_stage.Kontagent_GetAdditionalMoves()
  } );

  v_game.GameLog_Add( "stage_fail",
   {
     stage   = stage_data.name
   } );
     
  local section_name = m_stage.GetStaticData( stage_name ).section;
  
  if (_G.v_minievent) and ( v_minievent.Crown_IsAvailableMinievent() ) and 
    ( stage_data.realtype ~= "minievent" ) and 
    ( stage_data.realtype ~= "dailyquest" ) then

    if ( v_minievent.Crown_GetCurrentValue() > 0 ) then
      mediator.Notify( "View_Stage_StageFailed", { name = stage_name, section = section_name } );
      private.Finish();
      return;
    end;
  end;
  
  mediator.Notify( "View_Stage_StageFailed", { name = stage_name, section = section_name } );
  
  v_interface.window.Show( "int_dialog_stage_fail", func_start );
  
end;
--------------------------------------------------------------------------------
function private.Dialog_Fail_Hide ( func_end )

  int_dialog_stage_fail.Hide( function ()

    v_interface.window.Hide( "int_dialog_stage_fail", func_end );

  end );

end;
--------------------------------------------------------------------------------
function private.Dialog_Fail_Update ( reason )
  
  int_dialog_stage_fail.Update( reason );
  
end;
--------------------------------------------------------------------------------
function private.RecoveredHeartsAttemptsVideo ()
  
  private.Dialog_Fail_Update( "video_attempts_recovered" );
  
end;
--------------------------------------------------------------------------------
function private.UpdateHeartsCount ( params )
  
  if ( params.reason == "video" ) then
    private.Dialog_Fail_Update( "video_hearts_add" );
  elseif ( params.reason ~= "video" ) and ( params.delta > 0 ) then
    private.Dialog_Fail_Update( "hearts_add" );
  end;
  
end;
--------------------------------------------------------------------------------
function private.Immortality_OnToggled ()
  
  private.Dialog_Fail_Update( "immortality" );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.DialogFriends_Prepare ( stage_name )
  
  local stage_data = m_stage.GetCurrentStageData();
  
  if ( stage_data.type == "fight" ) or 
     ( stage_data.realtype == "minievent" ) or ( IsPC() ) then
    
    return;
    
  end;
  
  local friends_data = v_social.Friends_Scores( stage_name );
  
  int_dialog_stage_friends.Prepare( friends_data );

end;
--------------------------------------------------------------------------------
function private.DialogFriends_Update ( score, is_win )
  
  local stage_data = m_stage.GetCurrentStageData();
  
  if ( stage_data.type == "fight" ) or 
     ( stage_data.realtype == "minievent" ) or ( IsPC() ) then
    
    return;
    
  end;

  int_dialog_stage_friends.Update( score, is_win );

end;
--------------------------------------------------------------------------------
function private.DialogFriends_Hide ()

  int_dialog_stage_friends.Hide();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.DialogWin_Show ()
  
  local stage_data  = m_stage.GetCurrentStageData();
  local stage_score = m_stage.GetCurrentScore();
  local stage_moves = m_stage.MovesDone_GetCount() or 0;
  local score_old   = private.score_old;
  private.score_old = nil;

  v_log.HasOffers_LogEvent( stage_data );
  
  local func_show_hero = function ( stage_type )

    int_dialog_stage_win.ShowHero( stage_type );

  end;

  local func_accept = function ()

     common.sound.Play( "simply_click" );

    -- JK: disable back button on Android after stage won
    -- when game is minimized and restored right after Next button pressed
    -- level is stuck and player looses life to get out despite the win
    -- back button is enabled back on map opened
    mediator.Notify( "View_Stage_DialogWin_NextButtonPressed" );

    private.DialogWin_Hide( function ()

      local section_name = ( stage_data.realtype ~= "minievent" ) and
                             stage_data.section or 
                             stage_data.name:match( "stage_%d+_minievent(.*)" );
      section_name = ( section_name ) and section_name:match( "^_(.*)" ) or section_name;

      if ( v_story.IsExistStory( section_name, "CompleteStage", stage_data.name ) ) then
        private.Finish(true);
        private.TopPanel_Hide();
        v_story.ShowStory( 
          section_name, 
          "CompleteStage", 
          stage_data.name, 
          "stage", 
          private.SwitchToPreviousScreen
        );
      else
        private.Finish();
      end;

    end );

  end;

  local func_start = function ()

    local social_panel = false;
    
    if ( stage_data.type     ~= "fight"     ) and 
       ( stage_data.realtype ~= "minievent" ) then
      
      social_panel = v_social.IsLogged();
      
      local friends_data = v_social.Friends_Scores( stage_data.name );
      
      if ( #friends_data > 0 ) then
        
        local outrun_count = 0;
      
        for f = 1, #friends_data do
        
          local friend_score = friends_data[ f ].score;
          local friend_id    = tostring( friends_data[ f ].userId );
          
          local friend_was_ahead  = friend_score > score_old;
          local friend_now_behind = stage_score  > friend_score;
          
          local friend_real = ( friend_id:match("^bot_.*$") == nil );
          
          if ( friend_real and friend_was_ahead and friend_now_behind ) then
            outrun_count = outrun_count + 1;
          end;
        
        end;
        
        if ( outrun_count > 0 ) then
          mediator.Notify( "View_Stage_FriendsWon", outrun_count );
        end;
      
      end;
      
    end;
    
    local display_data = {
      score  = stage_score,
      moves  = stage_moves,
      chests = m_stage.GetDroppedChests()
    };
    
    if ( private.FastSaga_IsActive() ) then
      display_data.rewards = private.win_reward_data;
      private.win_reward_data = nil;
    end;
    
    int_dialog_stage_win.Show(
      stage_data.name,
      stage_data.type,
      display_data,
      v_player.Skin_GetCurrent(),
      social_panel
    );
    
    private.Social_Score_TryShow( 
      "win", 
      stage_data.type, 
      stage_data.name, 
      stage_score, 
      score_old 
    );

    local is_win_anim_ended = false;

    if ( stage_data.type == "ride" ) then

      is_win_anim_ended = int_panel_stage_progress_ride.IsWinAnimEnded();

    elseif ( stage_data.type == "fight" ) then

      is_win_anim_ended = int_panel_stage_progress_fight.IsWinAnimEnded();

    end;

    if ( is_win_anim_ended ) then

      func_show_hero( stage_data.type );

    else

      mediator.SubscribeOnce(
        "Interface_PanelStageProgress_WinAnimEnd",
        func_show_hero
      );

    end;

  end;
  
  local func_WinDialogShow = function() 
    
    local placement = "level_success";
    placement = tern( stage_data.realtype == "minievent" )( "event_success", placement );
    placement = tern( stage_data.realtype == "dailyquest" )( "arena_success", placement );

    v_log.Playhaven_LogPlacement( placement );
    
    v_interface.window.AddToQueue(
      "int_dialog_stage_win",
      func_start,
      nil,
      { Accept = func_accept }
    );
    
  end;
  
  if ( stage_data.type == "ride" ) then

    int_panel_stage_progress_ride.Update_StageComplete( function()
  
      if ( not v_interface.window.IsAnyOnScreen() ) then

        int_panel_stage_win.Show(
          "level_complete",
          func_WinDialogShow,
          v_player.Skin_GetCurrent()
        );

      else

        func_WinDialogShow();

      end;

    end );
  
  else
    
    func_WinDialogShow();
    
  end;

  public.Soundtrack_Stop();
  common.music.Play( "splash" );

  if ( stage_data.type == "ride" ) then

    common.sound.Play( "level_win" );

  elseif ( stage_data.type == "fight" ) then

    common.sound.Play( "assets/content/hero/aud_hero_win_fight" );

  end;

  local getnoreward = function ()

    local stage_info = m_stage.GetSavedData( stage_data.name ) or {};
    return stage_info.noreward;

  end;

  local score_target = 0;
  local moves_target = 0;

  if ( stage_data.type == "ride" ) then

    score_target = ( m_stage.GetTargetByType( "GainScore" ) or {} ).count or score_target;
    moves_target = stage_data.moves + m_stage.Kontagent_GetAdditionalMoves();

  end;

  local getscore = function ( stage_type )

    if ( stage_type == "fight" ) then
      
      local hero_data = m_stage.GetCurrentHero();

      return { ( hero_data.health_current..", "..hero_data.health ), 1 };

    end;

    return { ( m_stage.GetCurrentScore()..", "..score_target ), 0 };

  end;

  local scoredata = getscore( stage_data.type );
  local other_stage = false;
  local section = stage_data.section;
  local scene;
  
  if ( _G.v_dailyquest ) then
    other_stage = tern( stage_data.realtype == "dailyquest" )( 
                        ( v_dailyquest.Challenge_GetCompletedCount() + 1 ), 
                        false );
  end;
  if ( _G.v_minievent ) then
    other_stage = tern( stage_data.realtype == "minievent" )( 
          ( #( v_minievent.GetRelatedEventData() or { stages_completed = {} } ).stages_completed + 1 ), 
          other_stage );
    section = tern( stage_data.realtype == "minievent" )(
      ( v_minievent.GetRelatedEventData() or {} ).name, section );
    scene = tern( stage_data.realtype == "minievent" )( 
      ( v_minievent.GetRelatedEventData() or {} ).scene, nil );
  end;

  v_log.Kontagent_LogEvent( "stage",
  {
    section  = section,
    stage    = stage_data.name,
    other_stage = other_stage,
    scene    = scene,
    name     = "success",
    state    = m_stage.Kontagent_GetStageState( stage_data.name ),
    noreward = getnoreward(),
    score    = scoredata[ 1 ],
    is_boss  = scoredata[ 2 ],
    moves    = m_stage.MovesDone_GetCount()..", "..moves_target,
    boost    = m_stage.Kontagent_GetBoostersUsed(),
    cells    = m_stage.Kontagent_GetCellsDestroyed(),
    items    = m_stage.Kontagent_GetItemsDropped( "Key" ),
    flies    = m_stage.Kontagent_GetResourceCollected(),
    bombs    = m_stage.Kontagent_GetExplosivesDestroyed(),
    woodchests   = m_stage.Kontagent_GetItemsDropped( "woodchest" ),
    silverchests = m_stage.Kontagent_GetItemsDropped( "silverchest" ),
    goldchests   = m_stage.Kontagent_GetItemsDropped( "goldchest" ),
    additional_moves = m_stage.Kontagent_GetAdditionalMoves()
  } );

  v_game.GameLog_Add( "stage_win",
     {
       stage   = stage_data.name
     } );
  
  local stage_perks = m_stage.GetDroppedChests();

  v_hearts.Unblock( "win" );
  
  mediator.Notify( "View_Stage_StageWon", 
    { 
      stage = stage_data.name,
      perks = stage_perks
    }
  );

end;
--------------------------------------------------------------------------------
function private.DialogWin_Hide ( func_end )

  if ( IsMobile() ) then
    
    private.Social_Score_Hide();
    
  end;

  common.sound.Stop( "level_win" );
  common.music.Stop( "splash" );

  --v_interface.window.Hide( "int_dialog_stage_win", function ()
  int_dialog_stage_win.Hide( func_end );
  --end );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Social_Score_TryShow (
    point,
    stage_type,
    stage_name,
    stage_score,
    score_old
  )

  if ( stage_type == "fight" ) or 
     ( m_stage.GetStaticData( stage_name ).realtype == "minievent" ) or 
     ( IsPC() ) then 
    
    return; 

  end;

  private.score_data   = {
    point       = point,
    stage_name  = stage_name,
    stage_score = stage_score,
    score_old   = score_old
  };

  mediator.Subscribe(
    "Interface_Window_WindowOpened",
    private.Social_Score_WindowOpened
  );
  
  mediator.Subscribe(
    "Interface_Window_WindowClosedEnd",
    private.Social_Score_WindowClosed
  );

  v_social.Score.TryShow( point, stage_name, stage_score, score_old );

end;
--------------------------------------------------------------------------------
function private.Social_Score_Hide ()

  if ( not v_social.Score.IsActive() ) then return; end;

  private.score_data = nil;

  mediator.Unsubscribe(
    "Interface_Window_WindowOpened",
    private.Social_Score_WindowOpened
  );
  
  mediator.Unsubscribe(
    "Interface_Window_WindowClosedEnd",
    private.Social_Score_WindowClosed
  );

  v_social.Score.Hide();

end;
--------------------------------------------------------------------------------
function private.Social_Score_WindowOpened ( name )

  local upper_name = v_interface.window.UpperName();

  if ( upper_name ~= "int_dialog_stage_start" ) and
     ( upper_name ~= "int_dialog_stage_win" ) then

     v_social.Score.VisualHide();

  end;

end;
--------------------------------------------------------------------------------
function private.Social_Score_WindowClosed ( name )

  if ( v_interface.window.UpperName() == "int_dialog_stage_start" ) then

    v_social.Score.VisualShow();

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.SetupLayout ( is_fight, has_res )
  
  local round = common.math.Round;
  
  local h, w         = ui.screen.Height(), ui.screen.Width();
  local mid_x, mid_y = round( 0.50 * w ), round( 0.50 * h );
  
  -- ride interface position
  local hero_x     = mid_x;
  local hero_y     = round( 0.23 * h );
  local hero_scale = round( hero_y / 350, 2 );
  
  local hero_smooth_border = false;
  
  -- fight interface positions
  local fight_x     = hero_x;
  local fight_y     = hero_y;
  
  local fight_offset = hero_y;
  local fight_scale  = hero_scale;
  
  local fight_near = 350; -- will be scaled
  local fight_far  = fight_near;
  
  private.bg_offset = hero_y;
  
  -- set bonus popup position
  local bonus_x = mid_x;
  local bonus_y = math.max( round( fight_y - 330 * fight_scale ), 75 );
  
  -- booster panel size and position
  local boost_scale  = 1.0; --round( h / 1536, 2 );
  
  local boost_x      = mid_x;
  local boost_y      = round( 0.94 * h );
  local boost_hide_x = mid_x;
  local boost_hide_y = round( 1.06 * h );
  
  local boost_hideable = true;
  
  local boost_space  = round( boost_scale * 70 * 1.3 );
  local boost_margin = round( h - boost_y + boost_space / 2 );
  
  -- matchbox position
  local match_x = mid_x;
  local match_y = round( hero_y + ( ( h - hero_y - boost_margin ) / 2 ) );
  
  local match_size_h = round( 0.90 * w );
  local match_size_v = round( 0.90 * ( h - hero_y - boost_margin ) );
  
  local match_size   = math.min( match_size_h, match_size_v );
  
  -- score visible/hidden position
  local score_scale  =  hero_scale;
  local score_x      =  mid_x;
  local score_y      =  hero_y - 10;
  
  -- resources panel position
  local res_x      =  w - 45;
  local res_y      =  135;
  local res_hide_x =  w + 55;
  local res_hide_y =  res_y;
  
  -- top panels
  local is_landscape = false;
  
  -- stage number
  local num_x      = w - 35;
  local num_y      = h - 40;
  local num_hide_x = num_x;
  local num_hide_y = num_y + 50;
  
  local num_scale = 1.0;
  
  local num_hideable = true;
  
  -- custom shading flag for mtach-3 field
  private.match_custom_shade = false;
  
  if ( ui.screen.IsLandscape() ) then
    
    local match_space = round( 0.9 * h );
    
    match_size = round( 0.90 * match_space );
    match_x    = round( w - ( match_space / 2 ) );
    match_y    = round( mid_y + ( ( h - match_space ) / 2 ) - 10 );
    
    hero_scale = 1.2;
    hero_x     = round( ( w - match_space ) / 2 );
    hero_y     = round( 0.65 * h );
    
    hero_smooth_border = true;
    
    private.bg_offset = hero_y;
    
    score_scale  = 1.0;
    score_x      = hero_x;
    score_y      = hero_y - 10;
    
    boost_scale  = 1.0;--0.7;
    boost_x      = hero_x;
    boost_y      = round( 0.75 * h );
    
    boost_hide_x = boost_x;
    boost_hide_y = boost_y;
    
    boost_hideable = false;
    
    -- Shift boost pos if resources are exist
    if ( has_res ) then
      
      boost_y      = round( 0.84 * h );
      boost_hide_y = boost_y;
      
    end;
    
    res_x      =  hero_x;
    res_y      =  round( 0.72 * h );
    res_hide_x =  hero_x;
    res_hide_y =  round( 0.72 * h );
    
    is_landscape = true;
    
    num_x      =  35;
    num_y      =  35;
    num_hide_x =  num_x;
    num_hide_y = -num_y;
    
    num_scale = 0.8;
    
    num_hideable = boost_hideable;
    
    if ( is_fight ) then
      
      match_x = mid_x;
      match_y = round( 0.47 * h );
      
      fight_y     = hero_y;
      fight_scale = 1.0;
      
      fight_near = round( 0.35 * w / fight_scale );
      fight_far  = round( 0.90 * w / fight_scale );
      
      private.bg_offset = fight_y;
      
      bonus_x = tern( v_game.IsWidescreen() )( round( 0.2 * w ), round( 0.18 * w ) );
      bonus_y = round( fight_y - 330 * fight_scale );
      
      boost_x      = match_x;
      boost_y      = round( 0.92 * h ) - ui.screen.BottomSafeArea();
      boost_hide_x = boost_x;
      boost_hide_y = round( 1.10 * h ) - ui.screen.BottomSafeArea();
      
      boost_hideable = true;
      
      num_x      = w - 35;
      num_y      = h - 40;
      num_hide_x = num_x;
      num_hide_y = num_y + 50;
      
      num_scale = 1.0;
      
      num_hideable = true;
      
      -- See private.Match3_ShadesOn, private.Match3_ShadesOff
      private.match_custom_shade       = true;
      private.match_custom_shade_shift = h - match_y;
      
    end;
      
  end;
  
  v_match3.SetMaxSize( match_size );
  v_match3.SetPos( match_x,  match_y );
  
  int_panel_stage_progress_ride.SetScale( hero_scale );
  int_panel_stage_progress_ride.SetFgPos( hero_x, hero_y );
  int_panel_stage_progress_ride.SetBgOffset( hero_y );
  int_panel_stage_progress_ride.SetSmoothBorder( hero_smooth_border );
  
  int_panel_stage_progress_fight.SetScale( fight_scale );
  int_panel_stage_progress_fight.SetOffset( fight_y );
  int_panel_stage_progress_fight.SetDistances( fight_near, fight_far );
  int_panel_stage_progress_fight.SetSmoothBorder( hero_smooth_border );
  
  int_popup_stage_bonus.SetPos( bonus_x, bonus_y );
  
  int_panel_stage_score.SetScale( score_scale );
  int_panel_stage_score.SetPos( score_x, score_y );
  
  int_panel_stage_boosters.SetScale( boost_scale );
  int_panel_stage_boosters.SetPosVisible(      boost_x,      boost_y );
  int_panel_stage_boosters.SetPosHidden(  boost_hide_x, boost_hide_y );
  int_panel_stage_boosters.SetHideable( boost_hideable );
  
  int_panel_stage_resources.SetPosVisible(      res_x,      res_y );
  int_panel_stage_resources.SetPosHidden(  res_hide_x, res_hide_y );
  int_panel_stage_resources.SetLandscape( is_landscape );
  
end;
--------------------------------------------------------------------------------
function public.IsCanPlay ()
  
  return v_hearts.IsEnough( 1 );
  
end;
--------------------------------------------------------------------------------
function public.GetCurrentTargets ()
  
  local  targets = m_stage.GetCurrentTargets();
  return targets;
  
end;
--------------------------------------------------------------------------------
function private.PreStart ( stage_name, enemy_data, stage_state )

  if ( v_hearts.IsEnough( 1 ) ) then

    m_stage.Kontagent_SetCurrentStage( stage_name );
    
    --- JK: boosters selected at start -----------------------------------------
    private.selected_boosters = v_boosters.GetSelected();
    local booster_metric      = tern( #private.selected_boosters > 0 )( 1, 0 );

    m_stage.Kontagent_SetBoostersUsed( booster_metric );

    --- JK: model start --------------------------------------------------------
    local start_data = m_stage.Start(
      v_player.Stats_GetPlayerData(),
      stage_name,
      stage_state,
      enemy_data
    );
    
    --- JK: match3 -------------------------------------------------------------
    if (start_data.type == "fight") then
      for i = 1, #private.selected_boosters do
        if (private.selected_boosters[i] == "movesup") then
          table.remove(private.selected_boosters, i);
        end;
      end;
    end;
    
    start_data.boosters = private.selected_boosters;
    if ( start_data.realtype == "minievent" ) then
      if ( stage_name:find( "section" ) ) then
        local event_data  = v_minievent.GetRelatedEventData( "section" );
        start_data.pet_id = event_data.pet_id or 1;
      end;
    end;
    
    v_match3.SetStartData( start_data );
    
    --- JK: Chain activated once subscription ----------------------------------
    if ( private.chain_activated_once_subscribed ) then
      DbgTrace(
        "[ ERROR ][ v_stage ] OnChainActivatedOnce is subscribed while should not."
      );
    end;

    private.chain_activated_once_subscribed = mediator.SubscribeOnce(
      "Model_Stage_ChainActivated",
      private.OnChainActivatedOnce
    );

    return true;

  end;

  return false;

end;
--------------------------------------------------------------------------------
function private.Start ()

  mediator.Subscribe(
    "Model_Stage_FinalBonuses_StartAnimation",
    private.FinalBonuses_StartAnimation
  );
  
  private.Match3_SetInput( true );

  if ( private.started ) then

    private.PreOpen();

    return;

  end;

  -- JK: Create hub for match3 effect animations
  ui.obj.Create(
    private.this.hub.animations,
    "obj",
    ui.hub.Name()
  );

  private.started   = true;
  private.finalized = false;

  private.screen_prev = v_screen.GetCurrentName();

  if ( private.screen_prev == "scr_loader" ) then
    private.screen_prev = "scr_map";
  end;

  local stage_name = m_stage.GetCurrentStageName();

  v_screen.Switch( "scr_stage" );
  v_log.Playhaven_LogPlacement( "level_start" );
  v_game.GameLog_Add( "stage_start", { name = stage_name } );

  mediator.Notify( "View_Stage_Started", { stage_name = stage_name } );

end;
--------------------------------------------------------------------------------
function private.OnFinish ()

  if ( private.chain_activated_once_subscribed ) then

    mediator.Unsubscribe(
      "Model_Stage_ChainActivated",
      private.OnChainActivatedOnce
    );

    private.chain_activated_once_subscribed = nil;

  end;

  private.is_shuffle_required = nil;

  private.FinalizeWin();

end;
--------------------------------------------------------------------------------
function private.Finish ( not_switch_to_room )
  
  --DbgTrace( "Finish" );
  
  m_stage.Kontagent_SetCurrentStage( nil );
  
  if ( private.started ) then

    private.OnFinish();
  
    private.started = false;
    
    mediator.Unsubscribe(
      "Model_Stage_FinalBonuses_StartAnimation",
      private.FinalBonuses_StartAnimation
    );

    public.Soundtrack_Stop();
    if (not not_switch_to_room) then
      private.SwitchToPreviousScreen();
    end;

  end;

end;
--------------------------------------------------------------------------------
function private.SwitchToPreviousScreen ()
  if (private.screen_prev) then
    v_screen.Switch( private.screen_prev );
    private.screen_prev = nil;
  end;
end;
--------------------------------------------------------------------------------
function private.FinalizeWin ()
  
  if ( not private.finalized ) then
  
    m_stage.Finish( private._FinalizeWin );
    private.finalized = true;
    
  end;
 
end;
--------------------------------------------------------------------------------
function private.PreOpen ( func_after )

  private.finalized = false;

  local stage_data = m_stage.GetCurrentStageData(); 
  local is_fight   = ( stage_data.type == "fight" );
  local has_res    = ( #m_stage.GetCurrentResources() > 0 );

  -- HACK: add enemy name to dailyquest
  if ( is_fight ) then

    if ( not stage_data.enemy ) then

      stage_data.enemy = private.enemy; -- see public.TryStart

    end;

  end;

  local notification_param = stage_data.name;

  if ( stage_data.realtype == "minievent" ) then

    -- JK: publisher wants to exclude minievent levels from push notifications
    -- 0046360: FSFTP-1639 [Local Notifications] Event levels should not be
    -- included in "comeback" notifications

    notification_param = nil;

  end;
  
  v_game.LocalNotifications_SetData( "comeback_level", notification_param );

  private.DialogFriends_Prepare( stage_data.name );

  public.Soundtrack_Start();

  private.SetupLayout( is_fight, has_res );
  private.SetStageBackgroundColor( stage_data.zone, stage_data.type );

  v_match3.Start();
  private.Match3_ShadesReset();

  int_popup_stage_score.Attach();
  int_popup_stage_combo.Attach();

  private.PanelProgress_Show();
  private.PanelScore_Show();

  private.PanelSettings_Show();
  --public.PanelSettings_SetInput( true );

  private.PanelResources_Attach();
  private.PanelBoosters_Show();
  
  private.TopPanel_TryShow();

  local enemy = stage_data.enemy or {};
  
  private.DialogTargets_Show(
    stage_data.state,
    stage_data.targets,
    enemy.name
  );
  
  if ( func_after ) then
 
    func_after();

  end;

end;
--------------------------------------------------------------------------------
function private.Open ( func_after ) -- After splash fade

  private.is_active = true;

  if ( func_after ) then 
  
    func_after();
 
  end;

end;
--------------------------------------------------------------------------------
function private.Close ( func_after )

  if ( private.boosters_panel_shown ) then

    private.PanelBoosters_Hide();

  end;

  int_popup_stage_combo.Detach();

  private.PopupBonus_Hide();
  private.TopPanel_Hide();
  
  if ( func_after ) then 
  
    func_after();
 
  end;

end;
--------------------------------------------------------------------------------
function private.Cleanup ()

  m_stage.Kontagent_SetStageData( nil );

  -- JK: force hide shuffle dialog
  if ( int_dialog_stage_shuffle.IsActive() ) then

    int_dialog_stage_shuffle.Hide();

  end;

  -- JK: force hide match3 effect if it's active
  if ( private.is_match3_effect_active ) then

    private.Match3_HideEffect();

  end;

  --JK: Destroy animations hub with all contents
  ui.obj.Delete( private.this.hub.animations );

  --JK: Clear OSD effects
  ui.osd.ClearEffects();
  
  --ES: Clean used start boosters
  private.applies_start_boosters_fight = nil;
  ne.msg.Unsubscribe( "Event_Match3_FieldEstablished", private.Match3FieldEstablished );
  
  if ( private.is_shuffle_required ) then
    mediator.Unsubscribe( "Model_Stage_Continue", private.Shuffle_TryShow);
  end;
  
end;
--------------------------------------------------------------------------------
function private.AfterClose ( func_after )
  
  v_match3.Finish();
  
  private.Cleanup();
  
  if ( private.panel_score_shown ) then
    private.PanelScore_Hide();
  end;

  int_popup_stage_score.Detach();
  
  private.PanelSettings_Hide();
  private.PanelProgress_Hide();

  private.is_active = false;
  
  if ( func_after ) then
    
    func_after();
    
  end;
  
  private.popup_show = false;

end;
--------------------------------------------------------------------------------
function private.TryContinueNEW ()

  --Log.i( "TryContinueNEW" );

  mediator.Notify( "View_Match3_FieldEstablished" );
                           
end;
--------------------------------------------------------------------------------
function private.TryContinue ()

  if ( private.finalized ) then

    --Log.i( "TryContinue: finalized" );
    return;

  end;
  
  local stage_type = private.GetCurrentStageType();
  
  -- JK: wait for attacks finish
  if ( stage_type == "fight"                                    ) and
     ( not int_panel_stage_progress_fight.Attack_IsQueueEmpty() ) then
     
     return;
     
  end;

  if ( not m_stage.TryContinue() ) then

    private.Match3_SetInput( false );
    
    if ( stage_type == "fight" ) then

      private.Match3_ShadesOn();

    end;

    return;

  end;
  
  local anm_playing = public.IsAnimationPlaying();
  local finalize = m_stage.TryFinalize( anm_playing );
  
  -- Unlock field.
  private.Match3_ShadesOff( function ()
    if ( not finalize ) then
      private.Match3_SetInput( true );
    end;
  end );

  if ( finalize ) then

    private.Match3_SetInput( false );

    -- Lock field.
    private.PanelBoosters_SetInput( false, "TryContinue" );

    private.finalized = true;

    local time_finalize = 0.6;

    if ( not m_stage.IsCanContinue() ) then

      time_finalize = 2.0;

    end;

    ObjSet( "tmr_stage_finalize",
    {
      time    = time_finalize,
      playing = true,
      endtrig = function ()

        private.DialogComplete_Show();

      end,
    } );

  else

    if ( m_stage.IsWin() ) then

      private.Match3_SetInput( false );
      --public.PanelSettings_SetInput( false );
      private.PanelBoosters_SetInput( false );

      private.PopupBonus_Hide();

    else

      private.PopupBonus_Show();

    end;

  end;

  if ( stage_type == "fight" ) and ( m_stage.GetDataBonus() ) then
    private.PopupBonus_Show();
  end;
  
  --DbgTrace( "TryContinue Finish" );

end;
--------------------------------------------------------------------------------
function private._FinalizeWin ()
  
  local stage_name = m_stage.GetCurrentStageName();
  local stage_data = m_stage.GetCurrentStageData();
  local state_max  = stage_data.state_max;
  
  --JK: check stateless
  m_stage.Stateless_CheckStatus( stage_name );

  if ( stage_data.realtype ~= "dailyquest" ) and
     ( stage_data.realtype ~= "minievent" ) then

    local saved_info = m_stage.GetSavedData( stage_name ) or {};
    local stage_info = {
      completed = saved_info.completed or false,
      enabled   = saved_info.enabled   or false,
      noreward  = saved_info.noreward  or false,
      state     = saved_info.state     or 1,
      score     = saved_info.score     or 0,
      attempts  = saved_info.attempts  or 0,
      type      = stage_data.type,
      realtype  = stage_data.realtype
    };
    
    if ( m_stage.Stateless_IsActive() ) then
    
      if ( stage_info.state > 1 ) then
        stage_info.noreward = true;
      end;
      
      state_max = 2;
    
    end;

    private.noreward = stage_info.noreward;
    
    local update_set = {};

    local state_new   = math.min( stage_info.state + 1, state_max );
    local enabled     = true;
    local completed   = true;
    local enable_next = true;

    if ( stage_info.realtype == "misc"  ) and
       ( stage_info.type     == "fight" ) then

      stage_info.state = 4;
      state_new = 1;

      enable_next = false;
      completed   = false;

      update_set.type = "ride";

    elseif ( stage_info.realtype == "fight" ) then

      state_new = 1;
      enabled = false;

    end;

    if ( stage_info.attempts ~= 0 ) then

      update_set.attempts = 0;

    end;

    if ( stage_info.state == state_max and not stage_info.noreward ) then

      update_set.noreward = true;

    end;

    if ( completed ~= stage_info.completed ) then

      update_set.completed = completed;

    end;

    local score_new = m_stage.GetCurrentScore();

    if ( score_new > stage_info.score ) then

      update_set.score = score_new;

    end;

    if ( state_new ~= stage_info.state ) then

      update_set.state = state_new;

    end;

    if ( enabled ~= stage_info.enabled ) then

      update_set.enabled = enabled;

    end;

    v_player.Progress_SetStageState( stage_name, update_set );
    
    -- DG: NOTE: Say to v_map update that is first stage completion
    --  * for generic levels first complete marked by completed flag
    --  * for misc fight levels first complete marked by type = "ride"
    local is_first = update_set.completed or ( update_set.type == "ride" );
    
    if ( is_first ) then
      
      v_player.Progress_SetLastStageName(
        stage_name,
        v_sections.Section_NameCurrent(),
        true
      );

    end;

    if ( stage_info.realtype == "fight" ) then

      mediator.Notify(
        "View_Stage_EnemyDefeated",
        {
          name = stage_data.enemy.name
        }
      );

    end;

    return function ()
      
      local data_for_reward = {
        name        = stage_name,
        type        = stage_info.realtype,
        state_old   = stage_info.state,
        state_new   = state_new,
        noreward    = stage_info.noreward,
        enable_next = enable_next
      };
      
      local reward_data = private.Reward_TryGet( stage_data.name, data_for_reward );
      
      private.win_reward_data = reward_data;

      v_map.ChangeList_Prepare(
        stage_name,
        is_first,
        reward_data,
        update_set.noreward,
        enable_next
      );
      
      local notify_data = {
        name        = stage_name,
        type        = stage_info.realtype,
        state_old   = stage_info.state,
        state_new   = state_new,
        noreward    = stage_info.noreward,
        enable_next = enable_next
      };
      
      mediator.Notify( "Model_Stage_StageWon", notify_data );
      
    end;

  else

    if ( stage_data.realtype == "minievent" ) and ( _G.v_minievent ) then
      
      local data_for_reward = {
        name        = stage_name,
        state_old   = 1
      };
      
      local reward_resources = private.Reward_TryGet( stage_name, data_for_reward, true );
      
      v_minievent.ChangeList_Prepare(
        stage_name,
        reward_resources
      );
      
    end;
    
    local notify_data =
    {
      name = stage_data.name,
      type = stage_data.type,
      state_old = 1,
      state_new = 1,
      noreward = true
    };

    mediator.Notify( "Model_Stage_StageWon", notify_data );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Reward_TryGet ( stage_name, data_for_reward, resource )

  local reward_data = nil;
  
  if ( resource ) then
    reward_data = private.Reward_GetListResource();
  else
    reward_data = private.Reward_GetList( data_for_reward );
  end;
  
  local drop_reward = {}; 

  for i = 1, #reward_data do

    local reward_name  = reward_data[ i ].name;
    local reward_type  = reward_data[ i ].type;
    local reward_count = reward_data[ i ].count;

    if ( "quest" == reward_type ) then

      mediator.Notify(
        "Model_Stage_QuestItemDropped",
        {
          name  = reward_name,
          stage = stage_name
        }
      );

    elseif ( "worker" == reward_type ) then

      --

    else
  
      reward_data[ i ].stage_name  = stage_name;
      reward_data[ i ].stage_state = data_for_reward.state_old;
      reward_data[ i ].category    = "stage";
      
      local scene;
      local other_stage;
      local source;

      if ( _G.v_minievent ) then
        other_stage = ( #( v_minievent.GetRelatedEventData() or { stages_completed = {} } ).stages_completed + 1 );
        scene = ( v_minievent.GetRelatedEventData() or {} ).scene;
        source = "minievent_"..( ( v_minievent.GetRelatedEventData() or {} ).name or "unknown" );
      end;
      reward_data[ i ].scene       = scene;
      reward_data[ i ].other_stage = other_stage;
      reward_data[ i ].source      = source;

      v_items.TryObtain( reward_data[ i ] );

    end;

    table.insert( drop_reward, reward_data[ i ] );

  end;

  return drop_reward;

end;
--------------------------------------------------------------------------------
-- Возвращает данные только для начисления награды!
function private.Reward_GetList ( data_for_reward )

  local stage_name = m_stage.GetCurrentStageName();
  
  if ( stage_name ) then

    local reward_list = {};

    reward_list = private.Reward_GetGold( reward_list, stage_name, data_for_reward );
    reward_list = private.Reward_GetWorker( reward_list, stage_name, data_for_reward );
    reward_list = private.Reward_GetQuestItem( reward_list, stage_name, data_for_reward );
    reward_list = private.Reward_GetFragment( reward_list, stage_name, data_for_reward );
    reward_list = private.Reward_GetSkinpart( reward_list, stage_name, data_for_reward );
    reward_list = private.Reward_GetStickerpack( reward_list, stage_name, data_for_reward );
    reward_list = private.Reward_GetHeart( reward_list, stage_name, data_for_reward );
    reward_list = private.Reward_GetResource( reward_list );

    return reward_list;

  end;

  return nil;

end;
--------------------------------------------------------------------------------
function private.Reward_GetListResource ()

  local stage_name = m_stage.GetCurrentStageName();
  
  if ( stage_name ) then

    local reward_list = {};

    reward_list = private.Reward_GetResource( reward_list );

    return reward_list;

  end;

  return nil;

end;
--------------------------------------------------------------------------------
function private.Reward_GetGold ( reward_list, stage_name, first )

  if ( first and first.noreward ) then

    return reward_list;

  end;

  local stage_state = first.state_old or m_stage.Stage_GetState( stage_name );
  local stage_data  = m_stage.GetCurrentStageData();
  local stage_type  = stage_data.type;
  local gold_count  = 0;

  if ( stage_type == "fight" ) then

    local spell_name = v_spells.Spell_GetNameByUnlock( stage_name );

    if ( spell_name ) then return reward_list; end;

    -- JK: give default gold if no reward for fight is specified
    if ( not stage_data.reward   ) or
       ( #stage_data.reward == 0 ) then
    
      gold_count = m_stage.Fight_GetSettings().reward or 0;

    end;

  else

    local ride_settings = m_stage.Ride_GetSettings();
    
    if ( ride_settings.rewards[ stage_state ] ) then
    
      gold_count = ride_settings.rewards[ stage_state ].count;
    
    end;
    
  end;

  if ( gold_count < 1 ) then return reward_list; end;
  
  local gold_bonus = m_stage.GetCurrentHero().gold;
  local gold_count = gold_count + gold_bonus;
  
  if ( gold_count < 1 ) then return reward_list; end;
  
  table.insert(
    reward_list,
    {
      name  = "gold",
      type  = "gold",
      count = gold_count
    }
  );

  return reward_list;

end;
--------------------------------------------------------------------------------
function private.Reward_GetWorker ( reward_list, stage_name, first )

  if ( first and first.noreward ) then

    return reward_list;

  end;

  local stage_data  = m_stage.GetStaticData( stage_name );
  local stage_state = first.state_old or m_stage.Stage_GetState( stage_name );

  if ( stage_data.state[ stage_state ].reward ) then

    for i = 1, #stage_data.state[ stage_state ].reward do

      local reward = stage_data.state[ stage_state ].reward[ i ];

      if ( "worker" == reward.type ) then
        
        local mine_name = v_mines.Mine_GetNameByStage( stage_name );
        local mine_data = v_mines.Mine_Data( mine_name );
        if ( ( mine_data.progress or 0 ) < mine_data.progress_max ) then
          table.insert( reward_list, reward );
        end;

      end;

    end;

  end;

  return reward_list;

end;
--------------------------------------------------------------------------------
function private.Reward_GetQuestItem ( reward_list, stage_name, first )

  local stage_data  = m_stage.GetStaticData( stage_name );
  local stage_state = first.state_old or m_stage.Stage_GetState( stage_name );

  for s = 1, stage_state do

    if ( stage_data.state[ s ].reward ) then

      for r = 1, #stage_data.state[ s ].reward do

        local reward = stage_data.state[ s ].reward[ r ];

        if ( reward.type == "quest" ) then

          local is_active = v_quests.IsStageItemActive(
            reward.name,
            stage_name,
            "stage",
            stage_data.section
          );
          
          if ( is_active ) then

            table.insert( reward_list, reward );

          end;

        end;

      end;

    end;

  end;

  return reward_list;

end;
--------------------------------------------------------------------------------
function private.Reward_GetFragment ( reward_list, stage_name, first )

  local stage_data  = m_stage.GetStaticData( stage_name );
  local stage_state = first.state_old or m_stage.Stage_GetState( stage_name );

  if ( stage_data.state[ stage_state ].reward ) then

    for r = 1, #stage_data.state[ stage_state ].reward do

      local reward = stage_data.state[ stage_state ].reward[ r ];

      if ( reward.type == "fragment" ) then
          
        local build_name  = v_builds.Build_GetNameByFragment( reward.name )
        local build_data  = v_builds.Build_GetData( build_name );
        local is_collected = common.IsEntityInList(
          build_data.collected_fragments,
          reward.name
        );
        
        if ( not is_collected ) then

          table.insert( reward_list, reward );

        end;

      end;

    end;

  end;

  return reward_list;

end;
--------------------------------------------------------------------------------
function private.Reward_GetSkinpart ( reward_list, stage_name, first )

  if ( first and first.noreward ) then

    return reward_list;

  end;

  local stage_data  = m_stage.GetStaticData( stage_name );
  local stage_state = first.state_old or m_stage.Stage_GetState( stage_name );

  if ( stage_data.state[ stage_state ].reward ) then

    for r = 1, #stage_data.state[ stage_state ].reward do
      
      local reward = stage_data.state[ stage_state ].reward[ r ];
      
      if ( reward.type == "skinpart" ) then

        local is_obtainable = v_player.Skin_IsPartObtainable(
          reward.name,
          reward.count
        );
      
        if ( is_obtainable ) then
        
          table.insert( reward_list, reward );
        
        end;

      end;

    end;

  end;

  return reward_list;

end;
--------------------------------------------------------------------------------
function private.Reward_GetStickerpack ( reward_list, stage_name, first )

  if ( first and first.noreward ) then

    return reward_list;

  end;
  
  local stage_data  = m_stage.GetStaticData( stage_name );
  local stage_state = first.state_old or m_stage.Stage_GetState( stage_name );

  if ( stage_data.state[ stage_state ].reward ) then

    for r = 1, #stage_data.state[ stage_state ].reward do

      local reward = stage_data.state[ stage_state ].reward[ r ];

      if ( reward.type == "stickerpack" ) then

        table.insert( reward_list, reward );

      end;

    end;

  end;

  return reward_list;

end;
--------------------------------------------------------------------------------
function private.Reward_GetHeart ( reward_list, stage_name, data_for_reward )

  if ( data_for_reward          ) and
     ( data_for_reward.noreward ) then

    return reward_list;

  end;
  
  local stage_data  = m_stage.GetStaticData( stage_name );
  local stage_state = data_for_reward.state_old or m_stage.Stage_GetState( stage_name );

  if ( stage_data.state[ stage_state ].reward ) then

    for r = 1, #stage_data.state[ stage_state ].reward do

      local reward = stage_data.state[ stage_state ].reward[ r ];

      if ( reward.type == "heart" ) then

        table.insert( reward_list, reward );

      end;

    end;

  end;

  return reward_list;

end;
--------------------------------------------------------------------------------
function private.Reward_GetResource ( reward_list )

  local resources = m_stage.GetCurrentResources();
  
  if ( #resources > 0 ) then

    for _, resource in pairs( resources ) do

      if ( resource.count > 0 ) then

        local bonus = m_stage.GetCurrentHero()[ resource.name ];
        local count = resource.count + bonus;

        if ( count > 0 ) then
        
          local reward = {
            type  = "resource",
            name  = resource.name,
            count = count
          };
          
          --DbgTrace( "Bonus: "..bonus );
          --DbgTrace( "Reward: ", reward );
          
          table.insert( reward_list, reward );
        
        end;

      end;

    end;

  end;

  return reward_list;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.IsAnimationPlaying ()
  
  return int_panel_stage_win.IsActive();
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.OnSectionDisabled ()

  if ( int_dialog_stage_start.IsActive() ) then
    private.DialogStart_Hide();
  end;

end;
--------------------------------------------------------------------------------
function private.OnScoreChanged ( params )

  local score_current = params.score;
  local action_data   = {
    type          = "HeroRide",
    score_current = score_current
  };
  
  int_panel_stage_progress_ride.Update( action_data );

  private.PanelScore_Update( score_current );

end;
--------------------------------------------------------------------------------
function private.OnStageComplete ( params )

  if ( params.moves ) then
    
    local func_end = function ()

      int_panel_stage_progress_ride.Update_BonusRun();

      mediator.Notify( "View_Stage_AnimationIsFinished" );

    end;

    if ( not v_interface.window.IsAnyOnScreen() ) then

      int_panel_stage_win.Show( "fox_appear", func_end );
      
       common.sound.Play( "you_win" );

    else

      func_end();

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.OnTargetUpdated ( params )

 if ( params.type == "GainScore" ) then
    private.DialogFriends_Update( params.progress, m_stage.IsWin() );
  end;

end;
--------------------------------------------------------------------------------
function private.OnTargetComplete ( target )

  int_panel_stage_progress_ride.Update_TargetComplete();

end;
--------------------------------------------------------------------------------
function private.OnFieldActivated ()
  
  --

end;
--------------------------------------------------------------------------------
function public.ChainActivatedOnce ()
  
  private.OnChainActivatedOnce();
  
end;
--------------------------------------------------------------------------------
function private.OnChainActivatedOnce ()

  if ( not private.chain_activated_once_subscribed ) then
    return;
  end;
  
  private.chain_activated_once_subscribed = false;

  v_hearts.Block();
  
  local stage_data = m_stage.GetCurrentStageData();
  
  if (_G.v_minievent) and ( v_minievent.Crown_IsAvailableMinievent() ) and 
    ( stage_data.realtype ~= "minievent" ) and 
    ( stage_data.realtype ~= "dailyquest" ) then
    v_minievent.Crown_LevelStart();
  end;
  
end;
--------------------------------------------------------------------------------
function private.OnFieldEstablished ()

  --

end;
--------------------------------------------------------------------------------
function private.OnFieldNeedShuffle ()
  if ( m_stage.IsWin() ) then
    return;
  end;
  
  private.is_shuffle_required = true;
  
  local function func_shuffle ()

    if ( v_interface.window.IsAnyOnScreen() ) then
      mediator.SubscribeOnce(
        "Interface_Window_NoWindows",
        private.Shuffle_TryShow
      );
      return;
    end;  
    
    private.Shuffle_TryShow();
    
  end;
  
  if ( private.match3_shades ) then
    
    mediator.SubscribeOnce(
      "View_Stage_Match_ShadesOff",
      func_shuffle
    );
    
  else

    func_shuffle();
    
  end;

end;
--------------------------------------------------------------------------------
function private.Shuffle_TryShow ()
  if ( not private.is_shuffle_required ) then return; end;
    
  if ( m_stage.IsFinalized() ) then
    mediator.SubscribeOnce(
      "Model_Stage_Continue",
      private.Shuffle_TryShow);
    return;
  end;
  
  local x, y = v_match3.GetFieldCenter();
      
  int_dialog_stage_shuffle.Show(
    x,
    y,
    function ()
  
      common.MessageSend( "Command_Match3_FieldShuffle" );
      private.Match3_ShowEffect( "shuffle" );
      private.is_shuffle_required = nil;
  
    end
  );
    
end;
--------------------------------------------------------------------------------
function private.OnCellDestroyed ( params )

  -- Bug
  if ( params.name == "Bug" ) then

    local colors = 
    {
      Magenta = 4,
      Orange  = 5,
      Yellow  = 6,
      Blue    = 3,
      Red     = 1,
      Green   = 2     
    };

    params.animfunc = "bug_landing_"..colors[ params.color ];
    
    private.OnCellDestroyedReceive( params );

  end;
  
  -- Snowflake
  if ( params.name == "Snowflake" ) then

    private.OnSnowflakeFound( params );

  end;

end;
--------------------------------------------------------------------------------
function private.OnBonusRemoved ( params )

  if ( params.type ) then

    local sound_name = "chip_"..string.lower( params.type ).."_remove";

    if ( params.type == "Vertical" ) then

      sound_name = "chip_horizontal_remove";

    end;

    if ( params.type == "NucleusSecond" ) then

      sound_name = "chip_nucleus_remove";

    end;

     common.sound.Play( sound_name );

  end;

end;
--------------------------------------------------------------------------------
function private.OnBonusSpawn ( params )

  if ( params.type ) then

    local sound_name = "chip_"..string.lower( params.type ).."_create";

    if ( params.type == "Vertical" ) then

      sound_name = "chip_horizontal_create";

    end;

     common.sound.Play( sound_name );

  end;

end;
--------------------------------------------------------------------------------
function private.OnBonusCountChanged ( ability )
  
  local ability_count = ability.counter;
  local left_count  = 0;
  local right_count = 0;
  
  if ( ability.name == "stun" ) or 
     ( ability.name == "dot"  ) then
    right_count = ability_count;
  else
    left_count = ability_count;
  end;
  
  int_panel_stage_progress_fight.ShowDots( left_count, right_count );
  
end;
--------------------------------------------------------------------------------
function private.OnComboActivated ( params )

  --

end;
--------------------------------------------------------------------------------
function private.OnCellDestroyedReceive ( params )

  local target_list = m_stage.GetCurrentTargets();
  local target_data = m_stage.GetTargetByType( "DestroyCell", params.name );

  int_panel_stage_top.Goals_AddToQueue( target_data );

  local effect_index = target_data.progress;
  local effect_name  = private.this.effect.cell_destroyed.."_"..effect_index;
  
  ui.obj.Clone(
    private.this.effect.cell_destroyed,
    private.this.hub.animations,
    "_"..effect_index
  );

  local function func_end ()

    int_panel_stage_top.Goals_Receive(
      target_data,
      function ()

        int_panel_stage_top.Goals_Update(
          nil, -- JK: use queue
          target_data.index,
          function ()

            ui.obj.Delete( effect_name );

          end
        );

      end
    );

  end;

  local receiver_name = int_panel_stage_top.Goals_GetReceiver( target_data.index );
  local receiver_pos  = ui.obj.Position( receiver_name );

  ui.obj.Set(
    effect_name, 
    {
      animfunc = params.animfunc,
      playing  = true,
      pos_x    = params.x,
      pos_y    = params.y,
      pos_z    = 20  
    }
  );
  
  ObjAnimate( effect_name, "pos_xy", 0, 0, func_end,
  { 
    0.0, 3, params.x + 35,          params.y + 35,
    2.0, 3, receiver_pos[ 1 ] + 50, receiver_pos[ 2 ]
  } );

  ObjAnimate( effect_name, "scale_xy", 0, 0, "", 
  { 
    0.0, 3, 1.0, 1.0,
    2.0, 3, 0.5, 0.5
  } );

  ObjAnimate( effect_name, "alp", 0, 0, "", 
  { 
    0.0, 3, 1.0,
    1.5, 3, 1.0,
    2.0, 3, 0.0,      
  } );

end;
--------------------------------------------------------------------------------
function private.OnItemDropped ( params )

  Log.i( "Item dropped: "..params.name );
  
  local section = m_stage.GetCurrentStageData().section or "section_1";
  local section_number = section:match( "minievent_(.*)" ) or section:match( "%d+" );
  local target_name = params.name;
  local target_data = m_stage.GetTargetByType( "DropItem", target_name );
  
  int_panel_stage_top.Goals_Update( target_data, target_data.index );
  
  local index = private.EffectIndex_Get();

  local default_object  = "anm_item_drop_icon";
  local item_name       = string.lower( target_name ).."_"..section_number;
  local item_animfunc   = "item_"..item_name;
  local effect_animfunc = "appear";
  local resource        = "assets/content/"..section.."/anm_"..item_name;
  local resource_tag    = "";
  local sound_name      = "pet_fly";
  
  if ( params.name:match( "chest" ) ) then
    item_animfunc   = "item_chest";
    effect_animfunc = "appear_chest";
    resource        = "assets/content/match3/items/anm_item_key";
    resource_tag    = "assets/content/items/"..target_name;
    sound_name      = "chest_fly";
  end;
  
  local effect_name = private.this.effect.item_dropped.."_"..index;

  ui.obj.Clone(
    private.this.effect.item_dropped,
    private.this.hub.animations,
    "_"..index
  );

  local item_name = string.format( default_object.."_%s", index );
  
  ObjClone( default_object, "_"..index, 1, private.this.hub.animations );

  ObjSet( effect_name, 
  { 
    active  = true, 
    visible = true,
    animfunc = effect_animfunc,
    playing = false,
    pos_x   = params.x + 35,
    pos_y   = params.y + 35,
    pos_z   = 899
  } );
  
  ObjSet( item_name, 
  { 
    active   = true, 
    visible  = true,
    pos_x    = params.x + 35,
    pos_y    = params.y + 35,
    animfunc = item_animfunc,
    playing  = false,
    res      = resource,
    pos_z    = 898,
    rt_chest = resource_tag
  } );

  local func_anim_end = function ()

    local func_end = function ( effect_name )

      ObjDelete( effect_name );

      ui.anim.Play( item_name, item_animfunc, function () 

        ObjDelete( item_name );
        
      end );

      local target_x = 0.5 * ui.screen.Width() + private.item_dropped_target_offset_x;
      local target_y = 0.5 * ui.screen.Height() + private.item_dropped_target_offset_y;

      ObjAnimate( item_name, "pos_xy", 0, 0, "", 
      { 
        0.0, 3, params.x + 35, params.y + 130,
        1.0, 3, target_x,      target_y
      } );

    end;

    ui.anim.Play( effect_name, effect_animfunc, function () func_end( effect_name ) end );

    common.sound.Play( sound_name );

  end;

  ObjAnimate( effect_name, "pos_xy", 0, 0, func_anim_end, 
  { 
    0.0, 3, params.x + 35, params.y + 35,
    0.2, 3, params.x + 35, params.y + 95,
    0.3, 3, params.x + 35, params.y + 95
  } );

  ObjAnimate( item_name, "pos_xy", 0, 0, "", 
  { 
    0.0, 3, params.x + 35, params.y + 35,
    0.2, 3, params.x + 35, params.y + 95,
    0.3, 3, params.x + 35, params.y + 95
  } );

end;
--------------------------------------------------------------------------------
function private.OnSnowflakeFound ( params )

  Log.i( "Snowflake found: "..params.name );

  local target_data = m_stage.GetTargetByType( "DestroyCell", params.name );
  
  int_panel_stage_top.Goals_AddToQueue( target_data );
  
  --int_panel_stage_top.Goals_Update( target_data, target_data.type );
  
  local index = private.EffectIndex_Get();

  local item_name = string.format( "anm_snowflake_found_%s", index );

  ObjClone( "anm_snowflake_found", "_"..index, 1, private.this.hub.animations );

  local size = {
    small = 1,
    big   = 2
  };
  
  local start_x = params.x + 35 * size[ params.size ];
  local start_y = params.y + 35 * size[ params.size ];

  ObjSet( item_name, 
  { 
    active   = true, 
    visible  = true,
    pos_x    = start_x,
    pos_y    = start_y,
    animfunc = "object_"..size[ params.size ],
    playing  = false,
    res      = "assets/content/match3/items/anm_collect_objects",
    pos_z    = 898
  } );
  
  local function func_end ()

    int_panel_stage_top.Goals_Receive(
      target_data,
      function ()

        int_panel_stage_top.Goals_Update(
          nil,
          target_data.index
        );

      end
    );
    
  end;
  
  local receiver_name = int_panel_stage_top.Goals_GetReceiver( target_data.index );
  local receiver_pos  = ui.obj.Position( receiver_name );  

  ui.anim.Play( item_name, "object_"..size[ params.size ], function () 

    ObjDelete( item_name );
    func_end();
    
  end );

  local field_center_x, field_center_y = v_match3.GetFieldCenter();

  ObjAnimate( item_name, "pos_xy", 0, 0, "", 
  { 
    0.00,  3, start_x,                start_y,
    1.00,  3, field_center_x,         field_center_y,
    1.50,  3, field_center_x,         field_center_y,
    2.25,  3, receiver_pos[ 1 ] + 30, receiver_pos[ 2 ]
  } );

end;
--------------------------------------------------------------------------------
function private.OnMoveBonusRemoved ( chips_list )
  
  local animfuncs = 
  {
    Red = 1,
    Green = 2,
    Blue = 3,
    Magenta = 4,
    Orange = 5,
    Yellow = 6,
    Resource1 = 6,
    Resource2 = 3,
    Resource3 = 4,
    Resource4 = 1
  };
  
  for j = 1, #chips_list do

    local index = private.EffectIndex_Get();
    local pfx_name = "pfx_spawn_bonuses_move_"..index;
    ui.obj.Clone(
      "pfx_spawn_bonuses",
      private.this.hub.animations,
      "_move_"..index,
      1
    );
    
    local anim_name = private.this.effect.movebonus_removed.."_"..index;
    ui.obj.Clone(
      private.this.effect.movebonus_removed,
      private.this.hub.animations,
      "_"..index
    );
    ObjSet( anim_name, { 
      tt_add_move = "+"..chips_list[ j ].move_bonus,
      active  = true, 
      visible = true,
      pos_x   = chips_list[ j ].x,
      pos_y   = chips_list[ j ].y,
      animfunc = "add_move_match_"..animfuncs[ chips_list[ j ].color ],
      pos_z   = 20  
    } );
    ObjStopAnimate( anim_name, "pos_y" );

    local func_end = function ( anim_name )
      ObjDelete( anim_name );
      ObjDelete( pfx_name );
    end;
    private.MoveBonus_FlyToPanel(
      anim_name,
      pfx_name,
      function ()
        func_end( anim_name )
      end
    );

  end;
  
end;
--------------------------------------------------------------------------------
function private.MoveBonus_FlyToPanel ( anim_name, pfx_name, func_end )

  local current_pos     = GetObjPosByObj( anim_name );
  current_pos = { current_pos[1] + 60, current_pos[2] + 60 };
  local target_pos      = int_panel_stage_top.Moves_GetPosition();
  local tme             = 0.6 + math.random( 1, 3 ) / 10;
  local target_offset_x = 0;
  local target_offset_y = 10;

  ObjSet( anim_name, {
    pos_x = current_pos[ 1 ],
    pos_y = current_pos[ 2 ],
    pos_z = 20
  } );
  ObjAnimate( anim_name, "scale_xy", 0, 0, "", {
    0.0,       3, 1.0, 1.0,
    tme - 0.1, 3, 2.0, 2.0,
    tme,       3, 0.3, 0.3
  } );
  ObjAnimate( anim_name, "alp", 0, 0, "", {
    0.0,       1, 1.0,
    tme - 0.1, 1, 1.0,
    tme,       1, 0.0
  } );
  
  local animtable_pos_xy = {
    0.0, 0, current_pos[ 1 ], current_pos[ 2 ],
    tme, 0, target_pos[ 1 ] + target_offset_x, target_pos[ 2 ] + target_offset_y
  };

  ObjAnimate( anim_name, "pos_xy", 0, 0, func_end, animtable_pos_xy );
  ObjAnimate( pfx_name, "pos_xy", 0, 0, "", animtable_pos_xy );

  ui.obj.Enable( pfx_name, true );
  PartSysStart( pfx_name );

end;
--------------------------------------------------------------------------------
function private.OnResourceRewardCountChanged ( resources_list )

  local resources_animfuncs = 
  {
    resource1 = { "match_7", "collect_7" },
    resource2 = { "match_8", "collect_8" },
    resource3 = { "match_9", "collect_9" },
    resource4 = { "match_10", "collect_10" }
  };
  
   common.sound.Play( "glowworms_match" );

  for j = 1, #resources_list do

    local index     = private.EffectIndex_Get();
    local anim_name = private.this.effect.resource_removed.."_"..index;

    ui.obj.Clone(
      private.this.effect.resource_removed,
      private.this.hub.animations,
      "_"..index
    );

    ObjSet( anim_name, 
    { 
      active  = true, 
      visible = true,
      pos_x   = resources_list[ j ].x,
      pos_y   = resources_list[ j ].y,
      pos_z   = 20  
    } );

    local func_anim_end = function ( anim_name )

      if ( not ui.obj.Exists( anim_name ) ) then
        return;
      end;
      
      ObjStopAnimate( anim_name, "pos_y" );

      local resources_data = m_stage.GetCurrentResources();

      local func_end = function ( anim_name, resources_data )

        for i = 1, #resources_data do

          if ( resources_data[ i ].name == resources_list[ j ].name ) then

            private.PanelResources_Update( resources_data[ i ] );

          end;

        end;
  
        ObjDelete( anim_name );

      end;

      private.Resource_FlyToPanel(
        anim_name,
        resources_list[ j ].name,
        function ()
          func_end( anim_name, resources_data )
        end
      );

    end;

    ui.anim.Play( anim_name, resources_animfuncs[ resources_list[ j ].name ][ 1 ], function () 

      ui.anim.Play( anim_name, resources_animfuncs[ resources_list[ j ].name ][ 2 ] );

      local timer_name = "tmr_stage_resources_"..index;

      ObjClone( "tmr_stage_resources", "_"..index, 1, InterfaceName );

      ObjSet( timer_name,
      {
        time    = 0.01,
        playing = true,
        endtrig = function ()

          ObjDelete( timer_name );

          func_anim_end( anim_name );

        end
      } );

    end );

  end;

end;
--------------------------------------------------------------------------------
function private.Resource_FlyToPanel ( anim_name, resource_name, func_end )

  local target_name = 
    "spr_int_panel_stage_resources_resource_"..
    string.match( resource_name, "%d+" ).."_icon";

  local current_pos     = GetObjPosByObj( anim_name );
  local target_pos      = int_panel_stage_resources.GetPosVisible();
  local offset_x        = math.random( -80, 80 );
  local tme             = 0.6 + math.random( 1, 3 ) / 10;
  local target_offset_x = 0;
  local target_offset_y = 10;

  ObjSet( anim_name,
  {
    pos_x = current_pos[ 1 ],
    pos_y = current_pos[ 2 ],
    pos_z = 20
  } );

  ObjAnimate( anim_name, "scale_xy", 0, 0, "", 
  {
    0.0,       3, 1.0, 1.0,
    tme - 0.1, 3, 1.0, 1.0,
    tme,       3, 0.3, 0.3
  } );

  ObjAnimate( anim_name, "alp", 0, 0, "", 
  {
    0.0,       1, 1.0,
    tme - 0.1, 1, 1.0,
    tme,       1, 0.0
  } );

  ObjAnimate( anim_name, "pos_y", 0, 0, func_end, 
  {
    0.0, 1, current_pos[ 2 ],
    tme, 1, target_pos[ 2 ] + target_offset_y
  } );

  ObjAnimate( anim_name, "pos_x", 0, 0, "", 
  {
    0.0,       1, current_pos[ 1 ],
    0.5 * tme, 1, current_pos[ 1 ] + offset_x,
    tme,       1, target_pos[ 1 ] + target_offset_x
  } );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.PanelScore_Show ( stage_name )  

  if ( private.GetCurrentStageType() == "fight" ) then

    return;

  end;

  local color = private.GetStageBackgroundColor();

  int_panel_stage_score.Show( color );

  private.panel_score_shown = true;

end;
--------------------------------------------------------------------------------
function private.PanelScore_Update ( score )

  if ( private.GetCurrentStageType() == "fight" ) then

    return;

  end;

  int_panel_stage_score.Update( score );

end;
--------------------------------------------------------------------------------
function private.PanelScore_Hide ()

  if ( private.GetCurrentStageType() == "fight" ) then

    return;

  end;

  int_panel_stage_score.Hide();

  private.panel_score_shown = false;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.TopPanel_TryShow ()
  
  local stage_type = m_stage.GetCurrentStageData().type;
  
  private.TopPanel_SetupLevel();
  
  if ( stage_type == "fight" ) then
    
    private.TopPanel_ShowLevelStandalone();
    
    return;
    
  end;
  
  if ( not ui.screen.IsLandscape() ) then
    
    private.TopPanel_ShowLevelStandalone();
    
  end;
  
  private.TopPanel_SetupMoves();
  private.TopPanel_SetupGoals();
  
  int_panel_stage_top.Show();

end;
--------------------------------------------------------------------------------
function private.TopPanel_ShowLevelStandalone ()
  
  int_panel_stage_top.Level_Show();
  
end;
--------------------------------------------------------------------------------
function private.TopPanel_Hide ()

  int_panel_stage_top.Hide();

  private.PanelResources_Hide();

end;
--------------------------------------------------------------------------------
function private.TopPanel_SetupLevel ()

  local stage_name = m_stage.GetCurrentStageName();
  local stage_data = m_stage.GetCurrentStageData();
  
  if ( stage_data.realtype == "dailyquest" ) then

    int_panel_stage_top.Level_Disable();
    
    return;

  end;

  local stage_type = tern( stage_data.realtype == "bonus" )(
    stage_data.realtype,
    stage_data.type
  );

  local stage_number, is_stateless, is_minievent;

  if ( stage_data.realtype == "minievent" ) then

    stage_number = v_minievent.GetStageNumber( stage_name );
    is_stateless = true;
    is_minievent = true;

  else

    stage_number = public.Number_Relative( stage_data.section, stage_name );
    is_stateless = m_stage.Stateless_IsActive();

  end;

  int_panel_stage_top.Level_Setup(
    stage_number,
    stage_data.state,
    stage_type,
    is_stateless,
    is_minievent
  );

end;
--------------------------------------------------------------------------------
function private.TopPanel_SetupMoves ()

  local moves = m_stage.GetCurrentMoves() or
                m_stage.GetCurrentStageData().moves;

  if ( moves ) then

    int_panel_stage_top.Moves_Update( moves, 0, false );

  end;

end;
--------------------------------------------------------------------------------
function private.TopPanel_UpdateMoves ( moves, delta )

  int_panel_stage_top.Moves_Update( moves, delta, m_stage.IsWin() );

end;
--------------------------------------------------------------------------------
function private.TopPanel_SetupGoals ()

  local goals   = m_stage.GetCurrentTargets();
  local package = ( m_stage.GetCurrentStageData().section or "section_1" );
  
  int_panel_stage_top.Goals_Setup( goals, package );

end;
--------------------------------------------------------------------------------
function private.TopPanel_UpdateGoals ( goal_data )

  if ( goal_data.type ~= "DropItem"  ) and
     ( goal_data.name ~= "Bug"       ) and
     ( goal_data.name ~= "Snowflake" ) then

     int_panel_stage_top.Goals_Update( goal_data, goal_data.index );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.OnMovesCountChanged ( params )

  local moves = params.moves;
  local delta = params.delta;
  
  private.TopPanel_UpdateMoves( moves, delta );
  private.PanelBoosters_Appearance( moves );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.PanelResources_Attach ()

  local resources_data = m_stage.GetCurrentResources();

  if ( resources_data ) and ( #resources_data > 0 ) then

    int_panel_stage_resources.Attach( resources_data );

  end;

end;
--------------------------------------------------------------------------------
function private.PanelResources_Show ()

  int_panel_stage_resources.Show();

end;
--------------------------------------------------------------------------------
function private.PanelResources_Update ( resource_data )

  int_panel_stage_resources.Update( resource_data );

end;
--------------------------------------------------------------------------------
function private.PanelResources_Hide ()

  int_panel_stage_resources.Hide();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.PanelProgress_Show ()

  local stage_data   = m_stage.GetCurrentStageData();
  local hero_data    = m_stage.GetCurrentHero();
  local enemy_data   = m_stage.GetCurrentEnemy();
  local skin_current = v_player.Skin_GetCurrent();

  stage_data.color = private.GetStageBackgroundColor();
  
  if ( stage_data.type == "ride"  ) or
     ( stage_data.type == "bonus" ) then

    int_panel_stage_progress_ride.Show( stage_data, skin_current );

  end;

  if ( stage_data.type == "fight" ) then
    
    hero_data.player_name = v_player.Name();
    hero_data.skin_res    = v_player.Skin_GetContent( skin_current );
    hero_data.skin_name   = skin_current;
    
    int_panel_stage_progress_fight.Show( stage_data, hero_data, enemy_data );

  end;

end;
--------------------------------------------------------------------------------
function private.PanelProgress_Hide ()

  if ( private.GetCurrentStageType() == "ride"  ) or
     ( private.GetCurrentStageType() == "bonus" ) then

    int_panel_stage_progress_ride.Hide();

  end;

  if ( private.GetCurrentStageType() == "fight" ) then

    int_panel_stage_progress_fight.Hide();

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.PanelBoosters_StartAnim ( booster_name )
  
  int_panel_stage_boosters.StartAnim( booster_name );

end;
--------------------------------------------------------------------------------
function public.PanelBoosters_SetInput ( value, point )
  
  private.PanelBoosters_SetInput( value, point );
  
end;
--------------------------------------------------------------------------------
function private.PanelBoosters_SetInput ( value, point )

  --DbgTrace( "Booster panel input: "..tostring( value )..", point: "..tostring( point ) );
  
  int_panel_stage_boosters.SetInput( value );

end;
--------------------------------------------------------------------------------
function private.PanelBoosters_Show ()

  local stage         = m_stage.GetCurrentStageData();
  local boosters_list = public.Boosters_GetStageList( stage.name, "stage" );

  if ( #boosters_list < 1 ) then

    private.boosters_panel_shown = false;
    return;

  end;

  private.boosters_panel_shown = true;

  mediator.Subscribe(
    "Model_Boosters_CountChanged",
    private.PanelBoosters_Update
  );

  local func_tap = function ( booster_name )

    private.PanelBoosters_SetInput( false, "Tap" );
    public.PanelSettings_SetInput( false );
    
    local func_end = function ()
      private.PanelBoosters_SetInput( true, "Canceled Buy" );
      public.PanelSettings_SetInput( true );
    end
    
    local source = stage.name;
    if ( stage.name:find( "_minievent" ) ) then
      source = "minievent_"..v_minievent.GetRelatedEventData().name;
    end;

    local tap = v_boosters.Tap( "stage", booster_name, source, { func_end = func_end } );

    return tap;

  end;

  private.PanelBoosters_SetInput( true, "PanelBoosters_Show" );

  int_panel_stage_boosters.Show( boosters_list, func_tap );

end;
--------------------------------------------------------------------------------
function private.PanelBoosters_Update ()

  local stage = m_stage.GetCurrentStageData();

  local boosters_list = public.Boosters_GetStageList( stage.name, "stage" );
  local boosters_data = v_boosters.GetBoostersData( boosters_list );

  int_panel_stage_boosters.Update( boosters_data );

end;
--------------------------------------------------------------------------------
function private.PanelBoosters_Hide ()

  mediator.Unsubscribe(
    "Model_Boosters_CountChanged",
    private.PanelBoosters_Update
  );

  int_panel_stage_boosters.Hide();

end;
--------------------------------------------------------------------------------
function private.PanelBoosters_Use ( params )

  params.name = v_boosters.GetRevertPowerPhase( params.name );
  
  local booster_name = params.name;
  local source = m_stage.GetCurrentStageName();
  if ( source:find( "_minievent" ) ) then
    source = "minievent_"..v_minievent.GetRelatedEventData().name;
  end;
  
  local booster_data = v_boosters.GetBoosterStaticData( booster_name );
  if ( booster_data ) and ( booster_data.type == "start" ) then
    private.AddStartBoosterOnField();
  end;
  
  v_boosters.Use( params, source );

  if not ( params.result ) then

    int_panel_stage_boosters.StopAnim( function ()
      private.PanelBoosters_SetInput( true, "Not Use" );
      public.PanelSettings_SetInput( true );
    end );

  else
    
    int_panel_stage_boosters.StopAnim();
    m_stage.Kontagent_SetBoostersUsed( 1 );

  end;

end;
--------------------------------------------------------------------------------
function private.PanelBoosters_HideAnim ( params )

  if ( params.crown ) then
    private.ExecuteStartBoosterOnField();
    return;
  end;
  
  private.PanelBoosters_SetInput( true, "HideAnim" );
  public.PanelSettings_SetInput( true );

  local booster_data = v_boosters.GetBoosterStaticData( params.name );
  if ( booster_data.type == "start" ) then
    private.ExecuteStartBoosterOnField();
  end;
  
end;
--------------------------------------------------------------------------------
function private.PanelBoosters_Appearance ( moves )
  
  if  ( moves < 5 and not m_stage.IsWin() ) then
  
    int_panel_stage_boosters.EmergencyMode_TryStart();
  
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.PanelSettings_SetInput ( value )

  int_panel_stage_settings.SetInput( value );

end;
--------------------------------------------------------------------------------
function private.PanelSettings_Show ()

  local function func_hide_others ()

    v_log.Playhaven_LogPlacement( "game_paused" );
    ObjSet( "scr_stage", { active = false } );
    --mediator.Notify( "View_Stage_Pause", true );

  end;

  local function func_show_others ()

    ObjSet( "scr_stage", { active = true } );
    --mediator.Notify( "View_Stage_Pause", false );

  end;

  local function func_show_options ( func_after )

    v_settings.Dialog_Show( func_after );

  end;

  local function func_exit_stage ( func_reject, func_accept )

    private.DialogExit_Show( func_reject, func_accept );

  end;

  int_panel_stage_settings.SetHandlers( func_show_options, func_exit_stage,
                                        func_show_others, func_hide_others );
                                        
  int_panel_stage_settings.Show(); 

end;
--------------------------------------------------------------------------------
function private.PanelSettings_Hide ()
  
  int_panel_stage_settings.Hide();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.PopupBonus_Show ()

  local stage_type = private.GetCurrentStageType(); 
  local bonus_data = m_stage.GetDataBonus();

  if ( bonus_data and not private.finalized ) then
    
    local func_OnShow = function ()
      mediator.Notify( "View_Stage_PopupBonus_Activated" );
    end;

    int_popup_stage_bonus.Show(
      bonus_data.chip_type,
      stage_type,
      bonus_data.ability,
      func_OnShow
    );

  end;

end;
--------------------------------------------------------------------------------
function private.PopupBonus_Hide ()

  local activated = m_stage.IsActivateBonus();

  if ( activated ) then

    int_panel_stage_progress_ride.Update_PopupHit();

  end;

  if not ( activated ) then

    activated = nil;

  end;

  int_popup_stage_bonus.Hide( activated );

end;
--------------------------------------------------------------------------------
function private.BonusAbility_Hide ()

  int_panel_stage_progress_fight.Update_Ability();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.OnHeroAttackActivated ( params )

  Log.i( "Hero attack: damage = "..params.damage..", health = "..params.enemy_health );

  local action_data =
  {
    type           = "HeroAttackActivated",
    delta          = - params.damage,
    ability        = params.ability,
    ability_delta  = - params.ability_delta,
    health_current = params.enemy_health,
    health_total   = m_stage.GetCurrentEnemy().health
  };

  int_panel_stage_progress_fight.Update( action_data, private.TryContinue );

  private.PopupBonus_Hide();

end;
--------------------------------------------------------------------------------
function private.OnEnemyAttackActivated ( params )

  Log.i( "Enemy attac: damage = "..params.damage..", hero_health = "..params.hero_health );

  local action_data = 
  {
    type           = "EnemyAttackActivated",
    delta          = - params.damage,
    health_current = params.hero_health,
    health_total   = m_stage.GetCurrentHero().health
  };  

  int_panel_stage_progress_fight.Update( action_data, private.TryContinue );

  private.PopupBonus_Hide();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.PopupScore_Show ( chip_x, chip_y, chip_type, chip_score )

  local popup_color =
  {
    Unknown = { r = 0.500, g = 0.500, b = 0.500 },
    Magenta = { r = 0.922, g = 0.000, b = 0.922 },
    Orange  = { r = 1.000, g = 0.432, b = 0.040 },
    Yellow  = { r = 1.000, g = 0.863, b = 0.000 },
    Blue    = { r = 0.118, g = 0.412, b = 1.000 },
    Red     = { r = 1.000, g = 0.000, b = 0.000 },
    Green   = { r = 0.000, g = 1.000, b = 0.197 },
    Heart   = { r = 1.000, g = 1.000, b = 1.000 },
    Sword   = { r = 1.000, g = 0.000, b = 0.000 },
    Bomb    = { r = 0.197, g = 0.000, b = 0.197 },
    Dirt    = { r = 0.500, g = 0.500, b = 0.560 },
    Ice     = { r = 0.500, g = 0.500, b = 0.560 },
    Cage    = { r = 1.000, g = 1.000, b = 1.000 },
    Rock    = { r = 1.000, g = 1.000, b = 1.000 },
    Resource1    = { r = 1.000, g = 0.863, b = 0.000 },
    Resource2    = { r = 0.118, g = 0.412, b = 1.000 },
    Resource3    = { r = 0.922, g = 0.000, b = 0.922 },
    Resource4    = { r = 1.000, g = 0.000, b = 0.000 }
  };
  
  if ( not chip_x ) and ( not chip_y ) then
    
    chip_x, chip_y = v_match3.GetFieldCenter();   
    
  end;

  local text_color = popup_color[ chip_type ] or popup_color[ "Unknown" ];

  int_popup_stage_score.Show( chip_x, chip_y, text_color, chip_score );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.FinalBonuses_StartAnimation ( params )

  if ( params.moves == 0 ) then
    
     common.sound.Play( "count_moves_off" );
    
  end;
  
  local pfx_name  = "pfx_spawn_bonuses_"..params.moves;

  ui.obj.Clone(
    "pfx_spawn_bonuses",
    ui.hub.Name(),
    "_"..params.moves,
    1
  );
  
  local func_end = function ()
  
    m_stage.FinalBonuses_ChangeScore(
      params.chip_x,
      params.chip_y,
      params.chip_type,
      params.stage
    );
    
    ui.obj.Delete( pfx_name );

  end;

  ui.obj.Enable( pfx_name, true );

  PartSysStart( pfx_name );
  
  local pos = ui.obj.Position( "anm_int_panel_stage_top_moves" );
  
  --DbgTrace( pos );

  ObjAnimate( pfx_name, "pos_xy", 0, 0, func_end, 
  {
    0.0, 0, pos[ 1 ],      pos[ 2 ],
    0.5, 0, params.chip_x, params.chip_y
  } );

  ObjAnimate( pfx_name, "alp", 0, 0, "", 
  {
    0.00, 1, 1,
    0.45, 1, 1,
    0.50, 1, 0,
  } );
  
  local sound_id = ( 7 - params.moves % 7 );
    
  common.sound.Play( "part_"..sound_id );

end;
--------------------------------------------------------------------------------
function private.PopupCombo_Show ( chain_data )

  local stage = m_stage.GetCurrentStageData();

  local combo =
  {
    [ "ride" ] =
    {
      [ 3 ] = 1,
      [ 5 ] = 2,
      [ 7 ] = 3,
      [ 9 ] = 4
    },
    [ "bonus" ] =
    {
      [ 3 ] = 1,
      [ 5 ] = 2,
      [ 7 ] = 3,
      [ 9 ] = 4
    },
    [ "fight" ] =
    {
      [ 2 ] = 1,
      [ 3 ] = 2,
      [ 5 ] = 3
    }
  };

  if ( combo[ stage.type ][ chain_data.count ] ) then
    
    local x, y = v_match3.GetFieldCenter();

    int_popup_stage_combo.Show( x, y, combo[ stage.type ][ chain_data.count ] );

    if ( stage.type == "ride" ) or ( stage.type == "bonus" ) then

      int_panel_stage_progress_ride.Update_ComboSpeed(
        combo[ stage.type ][ chain_data.count ]
      );

    elseif ( stage.type == "fight" ) then
      
      int_panel_stage_progress_fight.Update_ComboHit(
        combo[ stage.type ][ chain_data.count ]
      );
      
    end;

  end;

   common.sound.Play( "combo_sound"..math.min( chain_data.count, 12 ) );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Soundtrack_Start ( stage_data )

  private.soundtrack = m_stage.GetCurrentStageData().type;

  common.music.Play( private.soundtrack, 1 );

end;
--------------------------------------------------------------------------------
function public.Soundtrack_Name ()

  return private.soundtrack;

end;
--------------------------------------------------------------------------------
function public.Soundtrack_Stop ()

  common.music.Stop( private.soundtrack );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Number_Absolute ( stage_name )

  local stage_number = string.match( stage_name, "%d+" ) or nil;

  return stage_number;

end;
--------------------------------------------------------------------------------
function public.Number_Relative ( section_name, stage_name )

  local section_number  = tonumber( string.match( section_name, "%d+" ) );
  local relative_number = tonumber( string.match( stage_name, "%d+" ) ) - 110 * ( section_number - 1 );

  return relative_number;

end;
--------------------------------------------------------------------------------
function public.Number_RelativeToAbsolute ( section_name, stage_number )

  local section_number  = tonumber( string.match( section_name, "%d+" ) );

  return ( stage_number + ( ( section_number - 1 ) * 110 ) );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Stats ()
  
  local  stats = m_stage.Stats();
  return stats;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Pending_Start ( pending_name )
  
  --DbgTrace( "Pending_Start: "..pending_name );
  
  private.pending = pending_name;
  
  public.PanelSettings_SetInput( false );
  
end;
--------------------------------------------------------------------------------
function public.Pending_Complete ( pending_name )
  
  --DbgTrace( "Pending_Complete: "..pending_name );
  
  private.pending = nil;
  
  public.PanelSettings_SetInput( true );
  
end;
--------------------------------------------------------------------------------
function public.Pending_IsActive ( pending_name ) -- JK: pass nil for any pending
  
  --DbgTrace( "Pending_IsActive: "..private.pending );
  
  if ( pending_name ) then
  
    return ( private.pending == pending_name );
  
  end;
  
  return ( private.pending ~= nil );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.GetCurrentStageType ()

  return m_stage.GetCurrentStageData().type;

end;
--------------------------------------------------------------------------------
function private.GetBackgroundColor ( zone, stage_type, is_stage )
  
  local zone_name = tern( is_stage )( zone.objects.name, zone.background.name );
  local backgrounds = m_stage.BackgroundColors_GetSettings();
  
  if not ( backgrounds[ zone_name ] ) then
    zone_name = "default";
  end;
  
  if ( stage_type == "fight" ) then
    zone_name = "arena";
  end;
  
  local background = backgrounds[ zone_name ];
  
  local color = {
    r = background.r,
    g = background.g,
    b = background.b
  };
  
  if ( is_stage ) then
    color.decoration = background.decor;
    color.angle = background.angle;
    color.r = background.r_stage or background.r;
    color.g = background.g_stage or background.g;
    color.b = background.b_stage or background.b;
  end;
  
  return color;

end;
--------------------------------------------------------------------------------
function private.GetStageBackgroundColor ()

  return private.color;

end;
--------------------------------------------------------------------------------
function private.SetStageBackgroundColor ( zone, stage_type )
  
  private.color = private.GetBackgroundColor( zone, stage_type, true );
  
  local multiset_table = {};

  -- Setup stage BG grid
  table.insert( multiset_table, 
  {
    "grd_stage_background", { color_r = private.color.r, 
                              color_g = private.color.g, 
                              color_b = private.color.b }
  } );
  
  local h, w = GetAppHeight(), GetAppWidth();
  
  local grd_size = {
    top    = 0,
    left   = 0,
    bottom = h,
    right  = w
  };
  
  local grd_border = {
    top    = 0,
    left   = 0,
    bottom = 0,
    right  = 0
  };
  
  ui.grid.Resize( "grd_stage_background", grd_size, grd_border );
  
  -- Setup background objects
  local deco_h = 85;
  local deco_w = 450;
  
  local offset = private.bg_offset or 0;
  
  local rows = math.floor( ( h - offset ) / deco_h );
  local cols = math.ceil( w / deco_w );
  
  for row = 1, rows  do
    
    local cols_here = cols;
    
    if ( ( row % 2 ) == 0 ) then
      
      cols_here = cols - 1;
      
    end;
    
    for col = 1, cols_here do
      
      local deco_name = "anm_stage_background_decoration_"..col.."_"..row;
      
      if not ( ObjGet( deco_name ) ) then

        ObjCreate( deco_name, "anim", "obj_stage_background_decoration" );

      end;
      
      local x_shift = common.math.Round( 0.5 * ( w - cols_here * deco_w ) );
      local y_shift = common.math.Round( offset - 0.15 * deco_h );
      
      local x_rand = common.math.Round( 0.015 * deco_w * math.random( -10, 10 ) );
      local y_rand = common.math.Round( 0.020 * deco_h * math.random( -10, 10 ) );
      
      if ( row == 1 ) then 
        
        y_rand = common.math.Round( 0.020 * deco_h * math.random( 0, 10 ) );
      
      end;
      
      local x = common.math.Round( x_shift + ( ( col - 0.5 ) * deco_w ) + x_rand );
      local y = common.math.Round( y_shift + ( ( row - 0.5 ) * deco_h ) + y_rand);
      
      local scale_pm = ( - 1 ) ^ math.random( 1, 2 );
      local scale    = math.random( 6, 10 ) / 10;
      
      local ang_pm = ( - 1 ) ^ math.random( 1, 2 );
      
      table.insert( multiset_table, 
      {
        deco_name, 
        { 
          res      = "assets/content/interface/tex_stage_a",
          animfunc = "bg_decoration_"..private.color.decoration,
          playing  = true,
          input    = false,
          pos_x    = x,
          pos_y    = y,
          scale_x  = scale_pm * scale,
          scale_y  = scale,
          ang      = ang_pm * 0.01 * math.random( private.color.angle[ 1 ], private.color.angle[ 2 ] ) or 0,
          color_r  = ( math.abs( private.color.r - 0.01 * math.random( 10, 15 ) ) ),
          color_g  = ( math.abs( private.color.g - 0.01 * math.random( 10, 15 ) ) ),
          color_b  = ( math.abs( private.color.b - 0.01 * math.random( 10, 15 ) ) )
        }
      } );
      
    end;
    
  end;

  table.insert( multiset_table, 
  {
    "obj_stage_background_decoration", { pos_x = 0, pos_y = 0 }
  } );

  ObjMultiSet( multiset_table );

end;
--------------------------------------------------------------------------------
function private.EffectIndex_Get ()

  private.effect_index = ( private.effect_index or 0 ) + 1;

  return private.effect_index;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Cheater_GetStartData (
    stage_name,
    database,
    dailyquest,
    stage_state
  )

  return private.GetStartData( stage_name, nil, nil, stage_state );

end;
--------------------------------------------------------------------------------
function public.Cheater_DialogStart_Hide ( func_end, skip_map_enable )

  return private.DialogStart_Hide( func_end, skip_map_enable );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Stage_GetCurrentData ()
  
  local current_stage_data = m_stage.GetCurrentStageData();
  
  return current_stage_data;
  
end;
--------------------------------------------------------------------------------
function public.Stage_IsCompleted ( stage_name )

  local stage_data = m_stage.GetSavedData( stage_name );
  
  return ( stage_data and stage_data.completed );

end;
--------------------------------------------------------------------------------
function public.Stage_IsFinished ( stage_name )

  local stage_data = m_stage.GetSavedData( stage_name );
  
  return ( stage_data and stage_data.noreward );

end;
--------------------------------------------------------------------------------
function public.Stage_IsEnabled ( stage_name )

  local stage_data = m_stage.GetSavedData( stage_name );
  
  return ( stage_data and stage_data.enabled );

end;
--------------------------------------------------------------------------------
function public.Stage_TryEnable ( stage_name, silent )
  
  local is_enabled  = public.Stage_IsEnabled( stage_name );
  local was_enabled = true;
  
  if ( not is_enabled ) then
    is_enabled  = m_stage.TryEnable( stage_name, silent );
    was_enabled = false;
  end;
  
  return is_enabled, was_enabled;

end;
--------------------------------------------------------------------------------
function public.Stage_GetInfo ( stage_name )

  local saved_data = m_stage.GetSavedData( stage_name ) or {};
  local stage_data = m_stage.GetStaticData( stage_name );
  
  if not ( stage_data ) then
    return false;
  end;

  -- Map point info
  local stage_info =
  {
    type     = stage_data.type,
    realtype = stage_data.realtype,
    hint     = stage_data.hint or false,

    state    = m_stage.GetStageState( stage_name, stage_data, saved_data ),

    enabled   = saved_data.enabled,
    completed = saved_data.completed,
    visible   = saved_data.visible,
    noreward  = saved_data.noreward, -- All states completed flag

    reward    = nil  -- displaying item reward
  };

  if ( m_stage.Stateless_IsActive() and stage_info.state > 1 ) then

    stage_info.state    = 1;
    stage_info.noreward = true;

  end;

  if ( stage_info.type == "fight" ) and
     ( stage_info.completed       ) then
    
    stage_info.noreward = true;

  end;

  -- Find first active reward item
  local states = #stage_data.state;

  for state = 1, states  do

    local rewards = stage_data.state[ state ].reward or {};

    for r = 1, #rewards do

      local reward = rewards[ r ];

      if ( reward.type == "quest" ) then

        if ( v_quests.IsStageItemActive( reward.name, stage_name, "map", stage_data.section ) ) then

          stage_info.reward = reward;
          break;

        end;

      elseif ( reward.type == "fragment" ) then

        local build_name = v_builds.Build_GetNameByFragment( reward.name );

        if ( build_name ) then

          local build_data = v_builds.Build_GetData( build_name );

          if ( not common.IsEntityInList( build_data.collected_fragments, reward.name ) ) then

            stage_info.reward = reward;
            break;

          end;

        end;

      elseif not ( stage_info.noreward ) and ( stage_info.state <= state ) and ( reward.type ~= "gold" ) then

        if ( reward.type == "worker" ) then
          local mine_name = v_mines.Mine_GetNameByStage( stage_name );
          local mine_data = v_mines.Mine_Data( mine_name );
          if ( ( mine_data.progress or 0 ) < mine_data.progress_max ) then
            stage_info.reward = reward;
            break;
          end;
        else
          stage_info.reward = reward;
          break;
        end;

      end;

    end;

  end;

  return stage_info;

end;
--------------------------------------------------------------------------------
function public.Stage_StaticData ( stage_name )

  local  static_data = m_stage.GetStaticData( stage_name );
  return static_data;

end;
--------------------------------------------------------------------------------
function public.Stage_GetSection ( stage_name )

  local  static_data = m_stage.GetStaticData( stage_name );
  return static_data.section;

end;
--------------------------------------------------------------------------------
function public.Stage_StorageData ( stage_name )

  local  saved_data = m_stage.GetSavedData( stage_name );
  return saved_data;

end;
--------------------------------------------------------------------------------
function public.Stage_GetAllBySection ( section_name, less_details )

  local  stages = m_stage.GetAllBySection( section_name, less_details );
  return stages;

end;
--------------------------------------------------------------------------------
function public.Stage_ListByType ( section_name, stage_types )

  local stage_list  = public.Stage_GetAllBySection( section_name, true ) or {};
  local filter_list = {};

  for _, stage_data in pairs( stage_list ) do

    if ( common.IsEntityInList( stage_types, stage_data.type ) ) then

      table.insert( filter_list, stage_data );

    end;

  end;

  return filter_list;

end;
--------------------------------------------------------------------------------
function public.Stage_WorkerState ( stage_name )
  
  local state = m_stage.GetStaticData( stage_name ).state;

  for i = 1, #state do

    if ( state[ i ].reward ) then

      for j = 1, #state[ i ].reward do

         if ( state[ i ].reward[ j ].name == "worker" ) then

          return i;

         end;

      end;

    end;

  end;

  return false;

end;
--------------------------------------------------------------------------------
function public.Stage_Exists ( stage_name, section_name )
  
  local stage_data = m_stage.GetStaticData( stage_name );
  
  return ( ( stage_data ) and ( stage_data.section == section_name ) );

end;
--------------------------------------------------------------------------------
function public.Stage_IsDeadlock ( stage_name )

  local is_deadlock = m_stage.Stage_IsDeadlock( stage_name );
  
  return is_deadlock;

end;
--------------------------------------------------------------------------------
function public.Stage_IsFight ( stage_name )

  local stage_realtype = m_stage.GetStaticData( stage_name ).realtype;
  
  return ( stage_realtype == "fight" );

end;
--------------------------------------------------------------------------------
function public.Stage_NameNext ( stage_name )

  -- JK: check if current stage is last in the section
  local last_stage_name = v_sections.Section_LastStage();

  if ( stage_name == last_stage_name ) then

    return stage_name;

  end;

  --JK: check if current stage is a deadlock
  if ( public.Stage_IsDeadlock( stage_name ) ) then

    return stage_name;

  end;

  local this_number = string.match( stage_name, "%d+" );
  local next_number = tonumber( this_number ) + 1;
  local stage_next  = string.gsub( stage_name, this_number, next_number );
  local build_data  = v_builds.Build_GetDataByUnlocker( stage_name, true );

  if (     build_data and
       not build_data.constructed ) then

    return stage_name;

  end;

  local stage_exists = public.Stage_Exists(
    stage_next,
    v_sections.Section_NameCurrent()
  );
  
  return tern( stage_exists )( stage_next, stage_name );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Zone_Data ( stage_name )

  local stage_data = m_stage.GetStaticData( stage_name );

  return {
    name    = stage_data.zone.background.name,
    content = stage_data.zone.background.content
  };

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Kontagent_GetCurrentStage ()

  local  current_stage = m_stage.Kontagent_GetCurrentStage();
  return current_stage;

end;
--------------------------------------------------------------------------------
function public.Kontagent_GetStageState ( stage_name )

  local  stage_state = m_stage.Kontagent_GetStageState( stage_name );
  return stage_state;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Boosters_GetStageList ( stage_name, load_booster_type )

  local func_BoostersOnStage = function ( stage_name )

    local saved_data = m_stage.GetSavedData(  stage_name ) or {};
    local db_data    = m_stage.GetStaticData( stage_name );

    local state_number = m_stage.GetStageState(
      stage_name,
      db_data,
      saved_data
    );

    state_number = tern( m_stage.Stateless_IsActive() )( 1, state_number );
    
    local stage = m_stage.GetStaticData( stage_name );

    if ( stage.state[ state_number ].boosters ) then

      return stage.state[ state_number ].boosters;

    end;

    return {};

  end;

  local boosters_on_stage = func_BoostersOnStage( stage_name );
  local boosters_list     = {};
  
  if ( boosters_on_stage ) and ( #boosters_on_stage > 0 ) then

    for i = 1, #boosters_on_stage do

      local is_minievent = 
        tern( string.find( stage_name, "minievent" ) )( true, false );
      
      local booster = 
      {
        name     = boosters_on_stage[ i ],
        count    = v_boosters.GetCount( boosters_on_stage[ i ] ),
        unlocked = v_boosters.IsUnlocked( boosters_on_stage[ i ] ) or is_minievent,
        sale     = v_boosters.GetPacks( boosters_on_stage[ i ] ).sale,
        phase    = v_boosters.GetPowerPhase( boosters_on_stage[ i ] )
      };
      
      local booster_type = v_boosters.Booster_GetType( booster.name );
      
      if ( load_booster_type == "all"        ) or 
         ( load_booster_type == booster_type ) then

        table.insert( boosters_list, booster );

      end;

    end;

  end;

  return boosters_list;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.GetUsableStages ( deficient_name )
  
  local section_name = v_sections.Section_NameCurrent();
  
  local usable_stages = m_stage.Usable_Search(
    deficient_name,
    section_name
  );
  
  return usable_stages;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Cheat_Win ()

  m_stage.Cheat_Win();

end;
--------------------------------------------------------------------------------
function public.Cheat_Fail ()

  m_stage.Cheat_Fail();
  
end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function public.Stages_FindByReward (
    reward_type,
    reward_name,
    section,
    ignore_state
  )

  section = section or v_sections.Section_NameCurrent();
  
  local stages = m_stage.Stages_FindByReward(
    section, reward_type, reward_name, ignore_state
  );
  
  return stages;

end;
--------------------------------------------------------------------------------
function public.Stages_GetAllCompleted ()
  
  local  completed = m_stage.Stages_GetAllCompleted();
  return completed;
  
end;
--------------------------------------------------------------------------------
function public.Stages_FindLastCompletedSagaStage ( stage_name )
  
  local section_name = v_sections.Section_NameCurrent();
  local stage_list   = v_stage.Stage_GetAllBySection( section_name, true );
  
  for i = 1, #stage_list do
    
    if ( stage_name == stage_list[ i ].name ) then
      
      return stage_list[ i - 1 ].name;
      
    end;
 
  end;
  
  return nil;
 
end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function public.FastSaga_IsActive ()

  local status = private.FastSaga_IsActive();
  
  return status;

end;
--------------------------------------------------------------------------------
function public.FastSaga_Deactive ()
  
  m_stage.FastSaga_Deactive();
  
end;
--------------------------------------------------------------------------------
function private.FastSaga_IsActive ()

  local is_active = m_stage.FastSaga_IsActive();
  return is_active;

end;
--------------------------------------------------------------------------------
