-- name=v_tutorial
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_tutorial",
  "int_dialog_tutorial",
  "int_dialog_tutorial_animation"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.this = {
    dialog = "int_dialog_tutorial",
    dialog_animation = "int_dialog_tutorial_animation"
  };

  common.mod.Load( _modules );
  
  private.app_h     = ui.screen.Height();
  private.app_w     = ui.screen.Width();
  private.landscape = ui.screen.IsLandscape();
  private.feature   = "tutorial";
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Init ()

  v_interface.window.RegisterAsFullscreen( private.this.dialog );
  v_interface.window.RegisterAsModal( private.this.dialog );
  v_interface.window.RegisterAsOverlay( private.this.dialog );
  v_interface.window.RegisterAsModal( private.this.dialog_animation );
  
  ui.hub.Create( "tutorial" );

  private.enabled = m_tutorial.IsEnabled();

  private.list   = {};
  private.BACKUP = {};

  v_game.RegisterFeature( 
  { 
    entity = "v_tutorial", 
    feature = private.feature,
    func_save = private.Feature_SaveData
  } );
  
  private.InitTutorialSteps();
  private.InitStartTutorials();

  private.Add( "stickerpack",     private.stickerpack );
  private.Add( "stickerpool",     private.stickerpool );
  --private.Add( "stickerpack_buy", private.stickerpack_buy );
  private.Add( "booster_shop", private.booster_shop );
  private.Add( "booster_buy",  private.booster_buy );
  --private.Add( "questlist_quest", private.questlist_quest );  
  --private.Add( "magicgift",      private.magicgift );--ES> disable temp
  private.Add( "skin",           private.skin );
  private.Add( "build",          private.build );
  private.Add( "mine",           private.mine );
  private.Add( "stage_14",       private.stage_14 );
  private.Add( "stage_16_tap",   private.stage_16_tap );  
  private.Add( "stage_16",       private.stage_16 );
  private.Add( "spellbook",      private.spellbook );
  private.Add( "heavenhelp",     private.heavenhelp );
  private.Add( "mine_construct", private.mine_construct, "v_mines" );
  private.Add( "stage_26",       private.stage_26 );
  private.Add( "nolimits",       private.nolimits );
  private.Add( "stage_44",       private.stage_44 );
  private.Add( "stage_45",       private.stage_45 );
  private.Add( "market_construct", private.market_construct );  
  private.Add( "stage_69",       private.stage_69 );
  private.Add( "downright",      private.downright );  
  private.Add( "stage_79",       private.stage_79 );
  private.Add( "stage_80",       private.stage_80 );
  private.Add( "dragonspeech",   private.dragonspeech );
  private.Add( "worldmap",       private.worldmap );
  private.Add( "stage_111",      private.stage_111 );
  private.Add( "stage_125_tap",  private.stage_125_tap );  
  private.Add( "stage_125",      private.stage_125 );
  private.Add( "stage_221",      private.stage_221 );
  
  private.Add( "count_bomb",     private.count_bomb );
  private.Add( "spiders",        private.spiders );
  private.Add( "fans",           private.fans );
  private.Add( "anticrystal",    private.anticrystal );
  
  --private.Add( "candy",          private.candy   );
  private.Add( "bottles",        private.bottles );
  private.Add( "paint",          private.paint   );

  private.Add( "wind",           private.wind );
  private.Add( "snowflakes",     private.snowflakes );

  private.Add( "market",         private.market );
  private.Add( "dailyquest",     private.dailyquest, "v_dailyquest" );
  private.Add( "dailyquest_win", private.dailyquest_win );
  private.Add( "stickers_duplicates", private.stickers_duplicates );
  private.Add( "bonuspopup",     private.bonuspopup );
  private.Add( "chest",          private.chest );
  private.Add( "fortunewheel",   private.fortunewheel, nil,
   function ()
     return v_stage.IsActiveFortuneWheel();
   end );
  private.Add( "fortunewheel_activate", private.fortunewheel_activate, nil,
   function ()
     return v_stage.IsActiveFortuneWheel();
   end );
  private.Add( "fortunewheel_win",   private.fortunewheel_win, nil,
   function ()
     return v_stage.IsActiveFortuneWheel();
   end );
  private.Add( "carrier",        private.carrier );
  private.Add( "stage_661",      private.stage_661 );
  private.Add( "stage_683",      private.stage_683 );
  private.Add( "stage_tutorial_1_minievent", private.stage_tutorial_1_minievent );
  private.Add( "stage_771",      private.stage_771 );
  private.Add( "stage_793",      private.stage_793 );
  private.Add( "stage_903",      private.stage_903 );
  private.Add( "stage_1211",     private.stage_1211 );
  
  --private.Add( "robber_cloud",      private.robber_cloud );  
  --private.Add( "resources",         private.resources );  
  --private.Add( "swordstone_cloud",  private.swordstone_cloud );
  --private.Add( "prisoner_cloud",    private.prisoner_cloud );
  --private.Add( "replay",            private.replay );
    
  --Log.i( "Initialized." );

end;
--------------------------------------------------------------------------------
function private.NewModulesLoad ( modules )
  
  for t = 1, #private.list do

    local tutorial = private.list[ t ];

    if ( not m_tutorial.IsCompleted( tutorial.name ) ) then
      if ( tutorial.depend ) then
        for m = 1, #modules do
          if ( modules[ m ] == tutorial.depend ) and ( not tutorial.processed ) then
            if ( tutorial.func ) then
              private.list[ t ].processed = true;
              tutorial.func();
            end;
          end;
        end;
      end;
    end;

  end;
  
end;
--------------------------------------------------------------------------------
function private.InitTutorialSteps ()

  private.tutorial_step = 1;
  local completed_list  = m_tutorial.GetCompletedList();
  
  for i = 1, #completed_list do
    if ( completed_list[ i ].name:find( "tut_" ) ) then
      private.tutorial_step = private.tutorial_step + 1;
    end;
  end;
 
end;
--------------------------------------------------------------------------------
function private.InitStartTutorials ()
  
  local tutorials = {
    --stage_1_tap = { name = "stage_1_tap", default = true },--ES> disable temp
    --stage_1_start_dialog = { name = "stage_1_start_dialog" },--ES> disable temp
    stage_1 = { name = "stage_1", default = true, relations = {
      "stage_1_play", "stage_1_win", "stage_1_reward"
    } },
    stage_1_modification_1 = { name = "stage_1_modification_1", relations = {
      "stage_1_play", "stage_1_win", "stage_1_reward"
    } },
    --swordstone = { name = "swordstone", default = true },--ES> disable temp
    stage_2 = { name = "stage_2", default = true, relations = {
      "stage_2_tap", "stage_2_play", "stage_2_win", "stage_2_reward"
    } },
    stage_2_modification_1 = { name = "stage_2_modification_1", relations = {
      "stage_2_tap", "stage_2_play", "stage_2_win", "stage_2_reward"
    } },
    stage_2_modification_2 = { name = "stage_2_modification_2", relations = {
      "stage_2_tap", "stage_2_play", "stage_2_win", "stage_2_reward"
    } },
    --quest = { name = "quest", default = true },--ES> disable temp
    stage_3 = { name = "stage_3", default = true, relations = {
      "stage_3_tap", "stage_3_play", "stage_3_win"
    } },
    stage_3_modification_1 = { name = "stage_3_modification_1", relations = {
      "stage_3_tap", "stage_3_play", "stage_3_win"
    } },
    stage_3_modification_2 = { name = "stage_3_modification_2", relations = {
      "stage_3_tap", "stage_3_play", "stage_3_win"
    } },
    stage_4 = { name = "stage_4", default = true, relations = {
      "stage_4_tap", "stage_4_play", "stage_4_win"
    } },
    stage_4_modification_1 = { name = "stage_4_modification_1", relations = {
      "stage_4_tap", "stage_4_play", "stage_4_win"
    } },
    stage_5 = { name = "stage_5", default = true, relations = {
      "stage_5_tap", "stage_5_play", "stage_5_win"
    } },
    stage_5_modification_1 = { name = "stage_5_modification_1", relations = {
      "stage_5_tap", "stage_5_play", "stage_5_win"
    } },
    stage_6 = { name = "stage_6", default = true, relations = {
      "stage_6_tap", "stage_6_play", "stage_6_win"
    } },
    stage_6_modification_1 = { name = "stage_6_modification_1", relations = {
      "stage_6_tap", "stage_6_play", "stage_6_win"
    } },
    --dailyquest_countdown = { name = "dailyquest_countdown", default = true },--ES> disable temp
    --achievements = { name = "achievements", default = true },--ES> disable temp
    --questlist = { name = "questlist", default = true },--ES> disable temp
    stage_9 = { name = "stage_9", default = true }
  };
  
  local list, state = private.GetListActiveStartTutorials();
  
  if ( not state or #list == 0 ) then
    private.InitDefaultTutorials( tutorials );
    return;
  end;
  
  for i = 1, #list do
    
    if ( tutorials[ list[ i ] ] ) then
      local name = tutorials[ list[ i ] ].name;
      private.Add( name, private[ name ] );
      if ( tutorials[ list[ i ] ].relations ) then
        local relations = tutorials[ list[ i ] ].relations;
        for i = 1, #relations do
          private.Add( relations[ i ], private[ relations[ i ] ] );
        end;
      end;
    end;
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.GetListActiveStartTutorials ()
  
  local state = private.Feature_GetState();
  
  if ( state ) then
    local data = private.Feature_GetData();
    if ( data ) then
      return data, true;
    end;
  end;
  
  return {}, false;
  
end;
--------------------------------------------------------------------------------
function private.InitDefaultTutorials ( tutorials )
  
  for key, value in pairs( tutorials ) do
    
    if ( value.default ) then
      private.Add( value.name, private[ value.name ] );
      if ( value.relations ) then
        local relations = value.relations;
        for i = 1, #relations do
          private.Add( relations[ i ], private[ relations[ i ] ] );
        end;
      end;
    end;
    
  end;
  
end;
--------------------------------------------------------------------------------
function public.Start ()
  
  local is_enabled = m_tutorial.IsEnabled();

  --DbgTrace( "[ INFO ][ v_tutorial ] Enabled: "..tostring( is_enabled ) );

  if ( not is_enabled ) then
    
    DbgTrace( "[ INFO ][ v_tutorial ] Disabled from launcher settings." );
    return;
    
  end;
  
  mediator.Subscribe( "Module_Load", private.NewModulesLoad );

  local stages = 
  {
    --stage_1_tap = "stage_1",--ES> disable temp
    --stage_1_start_dialog = "stage_1",--ES> disable temp
    stage_1_play = "stage_1",
    stage_1 = "stage_1",
    stage_1_modification_1 = "stage_1",
    stage_1_win = "stage_1",
    stage_1_reward = "stage_1",
    --swordstone = "stage_1",--ES> disable temp
    stage_2_tap = "stage_2",
    stage_2_play = "stage_2",
    stage_2_win = "stage_2",
    stage_2_reward = "stage_2",
    stage_2 = "stage_2",
    stage_2_modification_1 = "stage_2",
    stage_2_modification_2 = "stage_2",
    --quest = "stage_2",--ES> disable temp
    stage_3_tap = "stage_3",
    stage_3_play = "stage_3",
    stage_3 = "stage_3",
    stage_3_modification_1 = "stage_3",
    stage_3_modification_2 = "stage_3",
    stage_3_win = "stage_3",
    chest = "stage_3",
    stage_4_tap = "stage_4",
    stage_4_play = "stage_4",
    stage_4 = "stage_4",
    stage_4_modification_1 = "stage_4",
    stage_4_win = "stage_4",
    stage_5_tap = "stage_5",
    stage_5_play = "stage_5",
    stage_5 = "stage_5",
    stage_5_modification_1 = "stage_5",
    stage_5_win = "stage_5",
    stage_6_tap = "stage_6",
    stage_6_play = "stage_6",
    stage_6 = "stage_6",
    stage_6_modification_1 = "stage_6",
    stage_6_win = "stage_6",
    --dailyquest_countdown = "stage_6",--ES> disable temp
    --achievements = "stage_7",--ES> disable temp
    --questlist = "stage_8",--ES> disable temp
    stage_9 = "stage_9",
    --magicgift = "stage_10",--ES> disable temp
    skin = "stage_10",
    build = "stage_11",
    mine = "stage_12",
    stage_14 = "stage_14",
    booster_shop = "stage_14",
    stage_16_tap = "stage_16",
    stage_16 = "stage_16",
    spellbook = "stage_18",
    heavenhelp = "stage_20",
    mine_construct = "stage_22",
    stage_26 = "stage_26",
    stage_44 = "stage_44",
    stage_45 = "stage_45",
    market_construct = "stage_49",
    stage_69 = "stage_69",
    downright = "stage_76",
    stage_79 = "stage_79",
    stage_80 = "stage_80",
    dragonspeech = "stage_108",
    stage_111 = "stage_111",
    carrier = "stage_121",
    stage_125_tap = "stage_125",
    stage_125 = "stage_125",
    count_bomb = "stage_230",
    stage_661 = "stage_661",
    stage_683 = "stage_683",
    stage_tutorial_1_minievent = "stage_tutorial_1_minievent",
    stage_771 = "stage_771",
    stage_793 = "stage_793",
    stage_903 = "stage_903",
    stage_1211 = "stage_1211"
  };  
  
  private.started = true;

  for t = 1, #private.list do

    local tutorial = private.list[ t ];

    if ( stages[ tutorial.name ] ) then

      if ( ( v_stage.Stage_StorageData( stages[ tutorial.name ] ) or {} ).completed ) then
        if ( not m_tutorial.IsCompleted( tutorial.name ) ) then
          private.list[ t ].processed = true;
          private.Complete( tutorial.name );
        end;
      end;
      
      if ( m_tutorial.IsCompleted( tutorial.name ) ) then
        private.Kontagent_Check( tutorial.name );
      end;

    end;

    if ( not m_tutorial.IsCompleted( tutorial.name ) ) then
      if not ( ( tutorial.depend ) and ( not _G[ tutorial.depend ] ) ) then
        if ( tutorial.func ) then
          private.list[ t ].processed = true;
          tutorial.func();
        end;
      end;
    end;

  end;

  -- Add hook after screen closed
  local function func_closed_hook ( screen )
    if ( screen == "scr_stage" ) then
      mediator.Notify( "View_Tutorial_StageClosed" );
    end;
  end;

  mediator.Subscribe( v_screen.Notifications.ScreenClosed, func_closed_hook );

  -- Add hook after screen opened
  local function func_opened_hook ( screen )

    if     ( screen == "scr_map" ) then

      mediator.Notify( "View_Tutorial_MapOpened" );
     
    elseif ( screen == "scr_stage" ) then

      -- JK: tutorials are on first state only
      local stage = v_stage.Stage_GetCurrentData();
      
      if ( stage.state == 1 ) then
     
        private.cell_scale = v_match3.GetFieldScale();
        private.cell_size  = 0.5 * 70 * ( private.cell_scale or 1.0 );

        mediator.Notify(
          "View_Tutorial_StageOpened",
          stage.name
        );

      end;
     
    end;

  end;

  mediator.Subscribe( v_screen.Notifications.ScreenOpened, func_opened_hook );

  --mediator.Subscribe( "Model_Stage_StageWon", private.UnsubscribeOld );

  private.stage_animation();
  
end;
--------------------------------------------------------------------------------
function public.Enable ()
  
  if ( private.enabled ) then return; end;
  
  for e, event in pairs( private.events ) do
    for i = 1, #event do
      mediator.Subscribe( event[ i ].name, event[ i ].handler );
    end;
  end;

  mediator.Subscribe( "Module_Load", private.NewModulesLoad );
  
  private.enabled = true;
  
end;
--------------------------------------------------------------------------------
function public.Disable ()
  
  if ( not private.enabled ) then return; end;
  
  for e, event in pairs( private.events ) do
    for i = 1, #event do
      mediator.Unsubscribe( event[ i ].name, event[ i ].handler );
    end;
  end;

  mediator.Unsubscribe( "Module_Load", private.NewModulesLoad );
  
  private.enabled = false;
  
end;
--------------------------------------------------------------------------------
function public.IsStarted ()

  return private.started;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.IsCompleted ( tutorial )

  local  is_completed = m_tutorial.IsCompleted( tutorial );  
  return is_completed;

end;
--------------------------------------------------------------------------------
function public.Complete ( tutorial )

  m_tutorial.Complete( tutorial );

end;
--------------------------------------------------------------------------------
function public.Dialog_IsActive ()

  local  is_active = int_dialog_tutorial.IsActive();
  return is_active;

end;
--------------------------------------------------------------------------------
function public.Dialog_Name ()

  return private.this.dialog;

end;
--------------------------------------------------------------------------------
function public.IsSkinTutorialStarted ()

  if ( int_tutorial_skin ) then

    local  is_started = int_tutorial_skin.IsStarted();
    return is_started;

  end;

  return false;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Feature_GetData ()

  local data, exists = v_features.GetFeatureData( private.feature );

  if ( exists ) and ( data ) then
    return data[ 1 ];
  end;

  return false;

end;
--------------------------------------------------------------------------------
function private.Feature_GetState ()

  local state, exists = v_features.GetFeatureState( private.feature );

  if ( exists ) then
    return state;
  end;

  return false;

end;
--------------------------------------------------------------------------------
function private.Feature_SaveData ()
  
  local is_active = v_features.IsActive( private.feature );

  v_features.SaveFeatureState( private.feature, is_active );
  
  if ( is_active ) then
    
    local data = parts.features.GetFeatureData( private.feature ) or { {} };

    local abtest_group  = v_features.GetFeatureGroup( private.feature );
    local feature_group = v_features.GetSavedFeatureGroup( private.feature );

    if ( abtest_group and
         abtest_group ~= feature_group ) then
      
      v_log.Kontagent_LogEvent( "abtest", 
      {
        description = private.feature,
        name        = "add",
        value       = abtest_group
      } );
      
    end;

    v_features.SaveFeatureGroup( private.feature, abtest_group );
    
    if ( not next( data ) ) then
      data = false;
    end;
    v_features.SaveFeatureData( private.feature, data );

  end;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Kontagent_Check ( tutorial_name )
  
  local notifications = 
  {
    stage_1_tap            = { list = { "tut_level_1",
                                        "tut_level_1_tap" } },
    stage_1_start_dialog   = { list = { "tut_level_1_play" } },
    stage_1_play           = { list = { "tut_level_1_playdone" } },
    stage_1                = { list = { "tut_level_1_goal",
                                        "tut_level_1_swipe" } },
    stage_1_win            = { list = { "tut_level_1_win" } },
    stage_1_reward         = { list = { "tut_level_1_reward" } },
    stage_1_modification_1 = { list = { "tut_level_1_tiles", 
                                        "tut_level_1_swipe",
                                        "tut_level_1_swipedone",
                                        "tut_level_1_match3",
                                        "tut_level_1_match3done",
                                        "tut_level_1_goal" } },
    swordstone             = { list = { "tut_swordstone" } },
    stage_2_tap            = { list = { "tut_level_2" } },
    stage_2_play           = { list = { "tut_level_2_play" } },
    stage_2                = { list = { "tut_level_2_match4",
                                        "tut_level_2_match4done" } },
    stage_2_win            = { list = { "tut_level_2_win" } },
    stage_2_reward         = { list = { "tut_level_2_reward" } }, 
    stage_2_modification_1 = { list = { "tut_level_2_match4", 
                                        "tut_level_2_match4done",
                                        "tut_level_2_bonus",
                                        "tut_level_2_bonusdone",
                                        "tut_level_2_goal" } },
    stage_2_modification_2 = { list = { "tut_level_2_match4", 
                                        "tut_level_2_match4done",
                                        "tut_level_2_bonus",
                                        "tut_level_2_bonusdone",
                                        "tut_level_2_match4blue",
                                        "tut_level_2_match4bluedone",
                                        "tut_level_2_bonusblue",
                                        "tut_level_2_bonusbluedone" } },
    quest                  = { list = { "tut_cage" } },    
    stage_3_tap            = { list = { "tut_level_3" } },
    stage_3_play           = { list = { "tut_level_3_play" } },
    stage_3                = { list = { "tut_level_3_matchl",
                                        "tut_level_3_matchldone" } },
    stage_3_win            = { list = { "tut_level_3_win" } },
    stage_3_modification_1 = { list = { "tut_level_3_matchl", 
                                        "tut_level_3_matchldone",
                                        "tut_level_3_bonus",
                                        "tut_level_3_bonusdone" } },
    stage_3_modification_2 = { list = { "tut_level_3_matchl", 
                                        "tut_level_3_matchldone",
                                        "tut_level_3_matchltwo",
                                        "tut_level_3_matchltwodone",
                                        "tut_level_3_bonus",
                                        "tut_level_3_bonusdone" } }, 
    chest                  = { list = { "tut_present" } },
    stage_4_tap            = { list = { "tut_level_4" } },
    stage_4_play           = { list = { "tut_level_4_play" } },
    stage_4                = { list = { "tut_level_4_fireflies", 
                                        "tut_level_4_hammer",
                                        "tut_level_4_hammerdone",
                                        "tut_level_4_hammeruse" } }, 
    stage_4_win            = { list = { "tut_level_4_win" } },    
    stage_4_modification_1 = { list = { "tut_level_4_fireflies", 
                                        "tut_level_4_match3one",
                                        "tut_level_4_match3onedone",
                                        "tut_level_4_match3two",
                                        "tut_level_4_match3twodone" } },
    stage_5_tap            = { list = { "tut_level_5" } },
    stage_5_play           = { list = { "tut_level_5_play" } },
    stage_5                = { list = { "tut_level_5_matchbigcross",
                                        "tut_level_5_matchbigcrossdone" } },
    stage_5_win            = { list = { "tut_level_5_win" } },
    stage_5_modification_1 = { list = { "tut_level_5_fireflies", 
                                        "tut_level_5_matchbigcross",
                                        "tut_level_5_matchbigcrossdone" } },
    stage_6_tap            = { list = { "tut_level_6" } },
    stage_6_play           = { list = { "tut_level_6_play" } },
    stage_6                = { list = { "tut_level_6_match5",
                                        "tut_level_6_match5done" } },
    stage_6_win            = { list = { "tut_level_6_win" } },
    stage_6_modification_1 = { list = { "tut_level_6_match5", 
                                        "tut_level_6_match5done",
                                        "tut_level_6_fireworks",
                                        "tut_level_6_fireworksdone" } },
    dailyquest_countdown   = { list = { "tut_arena_countdown" } },
    achievements           = { list = { "tut_acheivements_tap",
                                        "tut_acheivements_tapdone",
                                        "tut_acheivements_reward",
                                        "tut_acheivements_rewarddone" } },
    questlist              = { list = { "tut_questlist" } },
    stage_9                = { list = { "tut_level_9_stoneblocks",
                                        "tut_level_9_stoneblocksdone",
                                        "tut_level_9_match3",
                                        "tut_level_9_match3done" } },
    magicgift              = { list = { "tut_magicgift" } }
  };    
  
  if ( notifications[ tutorial_name ] ) then
    
    for n = 1, #notifications[ tutorial_name ].list do

      private.Kontagent_LogEvent( "tutorial", 
      {
        name  = notifications[ tutorial_name ].list[ n ]
      } );
      
    end;
    
  end
  
end;
--------------------------------------------------------------------------------
function public.Kontagent_LogEvent ( event_type, params )

  private.Kontagent_LogEvent( event_type, params );
  
end;
--------------------------------------------------------------------------------
function private.Kontagent_LogEvent ( event_type, params )

  if ( not m_tutorial.IsCompleted( params.name ) ) then

    local group = v_features.GetFeatureGroup( "tutorial" );
    
    params.group = tern( group )( "tutorial_"..tostring( group ), "" );
    params.value = tern( params.level and params.level > 0 )( 0, private.tutorial_step );
    params.level = params.level or 0;
    
    v_log.Kontagent_LogEvent( event_type, params );
    private.Complete( params.name );
    
    if ( params.level == 0 ) then
      private.tutorial_step = private.tutorial_step + 1;
    end;
    
  end;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.PendingBegin ( tutorial_name, screen )

  --DbgTrace( "[ INFO ][ v_tutorial ] Start tutorial pending: "..tutorial_name );

  local pending_token = "tutorial_"..tutorial_name;
  
  local func = {
    map   = function () v_map.Pending_Start( pending_token ); end,
    stage = function () v_stage.Pending_Start( pending_token ); end
  };
  
  func[ screen ]();
  
end;
--------------------------------------------------------------------------------
function private.PendingEnd ( tutorial_name, screen )

  --DbgTrace( "[ INFO ][ v_tutorial ] End tutorial pending: "..tutorial_name );

  local pending_token = "tutorial_"..tutorial_name;
  
  local func = {
    map   = function () v_map.Pending_Complete( pending_token ); end,
    stage = function () v_stage.Pending_Complete( pending_token ); end
  };
  
  func[ screen ]();
  
end;
--------------------------------------------------------------------------------
function private.IsAutoPending ( event_name )
    
  if ( event_name == "View_Tutorial_StageOpened" ) then
    return "stage";
  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Subscribe ( tutorial_name, event_name, event_handler )

  private.events                  = private.events or {};
  private.events[ tutorial_name ] = private.events[ tutorial_name ] or {};

  table.insert(
    private.events[ tutorial_name ],
    { name = event_name, handler = event_handler }
  );

  mediator.Subscribe( event_name, event_handler );

end;
--------------------------------------------------------------------------------
function private.Unsubscribe ( tutorial_name, event_name, event_handler )

  if ( private.events[ tutorial_name ] ) then

    --JK: if tutorial is unsubscribing from StageOpened, it means it is pending
    local screen = private.IsAutoPending( event_name );
    
    if ( screen ) then
      private.auto_pending = { name = tutorial_name, screen = screen };
      private.PendingBegin( tutorial_name, screen );
    end;
    
    mediator.Unsubscribe( event_name, event_handler );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Add ( name, func, depend, func_active )

  if ( func_active ) and ( not func_active() ) then
    return;
  end;
  
  private.BACKUP[ name ] = {};
  table.insert( private.list, { name = name, func = func, depend = depend } );

end;
--------------------------------------------------------------------------------
function private.Set ( name, tag, data )

  private.BACKUP[ name ][ tag ] = private.BACKUP[ name ][ tag ] or {};
  private.BACKUP[ name ][ tag ].data = data;

end;
--------------------------------------------------------------------------------
function private.Get ( name, tag )

  if ( private.BACKUP[ name ][ tag ] ) then
    return private.BACKUP[ name ][ tag ].data;
  end;

end;
--------------------------------------------------------------------------------
function private.Backup ( name, tag, data )

  private.BACKUP[ name ][ tag ] = private.BACKUP[ name ][ tag ] or {};

  if ( not private.BACKUP[ name ][ tag ].enabled ) then
    private.BACKUP[ name ][ tag ].enabled = true;
    private.BACKUP[ name ][ tag ].data    = data;
  end;

end;
--------------------------------------------------------------------------------
function private.Restore ( name, tag )

  if ( private.BACKUP[ name ][ tag ]         ) and
     ( private.BACKUP[ name ][ tag ].enabled ) then

    private.BACKUP[ name ][ tag ].enabled = false;
    return private.BACKUP[ name ][ tag ].data;

  end;

end;
--------------------------------------------------------------------------------
function private.Complete ( tutorial_name )

  m_tutorial.Complete( tutorial_name );
  
  local pending = private.auto_pending;
  
  if ( pending and pending.name == tutorial_name ) then
    private.PendingEnd( tutorial_name, pending.screen );
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Tutorial_SetOnEvent (
    tutorial_name,
    event_name,
    delegates,
    func_validate
  )

  if ( not m_tutorial.IsCompleted( tutorial_name ) ) then
  
    local func_start;

    func_start = function ( arg )

      --DbgTrace(
      --  "[ INFO ][ v_tutorial ] Attempt to start '"..tutorial_name.."'"
      --);

      if ( func_validate            ) and
         ( not func_validate( arg ) ) then
        --DbgTrace(
        --  "[ ERROR ][ v_tutorial ] Validation failed for '"..tutorial_name.."'"
        --);
        return;
      end;

      private.Unsubscribe( tutorial_name, event_name, func_start );

      local path =
        "assets/content/scripts/tutorials/mod_int_tutorial_"..tutorial_name;

      ne.mod.LoadLazily( path );

      delegates = delegates or {};

      delegates.GetMatch3Cell = function ()
        return {
          size  = private.cell_size,
          scale = private.cell_scale
        };
      end;

      delegates.Subscribe           = private.Subscribe;
      delegates.Unsubscribe         = private.Unsubscribe;
      delegates.Backup              = private.Backup;
      delegates.Restore             = private.Restore;
      delegates.Complete            = function () private.Complete( tutorial_name ); end;
      delegates.LogKontagentEvent   = private.Kontagent_LogEvent;
      delegates.GetDialogSettings   = private.Dialog_Settings;
      delegates.ShowDialogInstantly = private.Dialog_ShowInstantly;
      delegates.ShowDialogFromQueue = private.Dialog_ShowFromQueue;
      delegates.HideDialog          = private.Dialog_Hide;
      delegates.StartPending        = private.PendingBegin;
      delegates.FinishPending       = private.PendingEnd;
      delegates.Unload              = function () private.Tutorial_Unset( tutorial_name ); end;

      common.delegate.Set( "int_tutorial_"..tutorial_name, delegates, true );

      _G[ "int_tutorial_"..tutorial_name ].Start( arg );

    end;

    --DbgTrace(
    --  "[ INFO ][ v_tutorial ] Set '"..tutorial_name.."' on '"..event_name.."'"
    --);
    private.Subscribe( tutorial_name, event_name, func_start );

  else

    DbgTrace(
      "[ ERROR ][ v_tutorial ] Attempt to set already completed '"..
      tutorial_name.."' on '"..event_name.."'"
    );
  
  end;

end;
--------------------------------------------------------------------------------
function private.Tutorial_IsStageMatched (
    received_stage_name,
    required_stage_name
  )

  if ( received_stage_name ~= required_stage_name ) then return false; end;

  local is_completed =
      ( v_stage.Stage_StorageData( received_stage_name ) or {} ).completed;

  if ( is_completed ) then return false; end;

  return true;

end;
--------------------------------------------------------------------------------
function private.Tutorial_IsStageCompleted ( stage_name )

  local is_completed =
    ( v_stage.Stage_StorageData( stage_name ) or {} ).completed;

  return is_completed;

end;
--------------------------------------------------------------------------------
function private.Tutorial_Unset ( tutorial_name )

  _G[ "int_tutorial_"..tutorial_name ] = nil;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Dialog_ShowInstantly (
    text_id,
    positions,
    hole,
    func_close,
    text_params,
    map_params,
    interface_lock
  )

  ObjSet( "zmv_map", { candrag = false, canzoom = false } );

  local func_start = function ()

    int_dialog_tutorial.Show(
      text_id,
      positions,
      hole,
      func_close,
      text_params,
      interface_lock
    );

  end;

  map_params = map_params or {
    notify = false,
    input  = false,
    active = false };

  v_interface.window.Show(
    private.this.dialog,
    func_start,
    {
      notify     = map_params.notify,
      map_input  = map_params.input,
      map_active = map_params.active
    }
  );

end;
--------------------------------------------------------------------------------
function private.Dialog_ShowFromQueue (
    text_id,
    positions,
    hole,
    func_close,
    text_params,
    map_params,
    interface_lock,
    func_scale
  )

  ObjSet( "zmv_map", { candrag = false, canzoom = false } );

  local func_start = function ()

    local func_show = function()

      int_dialog_tutorial.Show(
        text_id,
        positions,
        hole,
        func_close,
        text_params,
        interface_lock
      );

    end;

    if ( func_scale ) then

      func_scale( func_show );

    else

      func_show();

    end;

  end;

  map_params = map_params or
    {
      notify = false,
      input  = false,
      active = false };

  v_interface.window.AddToQueue(
    private.this.dialog,
    func_start,
    {
      notify     = map_params.notify,
      map_input  = map_params.input,
      map_active = map_params.active
    }
  );

end;
--------------------------------------------------------------------------------
function private.Dialog_Hide ( func_end, fast_hide )

  ObjSet( "zmv_map", { candrag = true, canzoom = true } );

  int_dialog_tutorial.Hide( function()

    v_interface.window.Hide( private.this.dialog );

    if ( func_end ) then

      func_end();
  
    end;

  end, 

  fast_hide );

end;
--------------------------------------------------------------------------------
function private.Dialog_Settings ( focus, pref_vertical, pref_horizontal )
  
  -- JK: you can pass position directly
  local focus_pos = focus;
  
  -- JK: or let it auto-detect
  if ( type( focus ) ~= "table" ) then
  
    focus_pos = ui.obj.Position( focus );
    
  end;
  
  --DbgTrace( "Focus is @ "..focus_pos[ 1 ].." x "..focus_pos[ 2 ] );
  
  local percent_w = private.app_w / 100;
  local percent_h = private.app_h / 100;
  local landscape = private.landscape;

  local center_h = 50 * percent_h;
  local center_w = 50 * percent_w;
  local top_edge = 20 * percent_h;
  local btm_edge = 80 * percent_h;
  local lft_edge = 20 * percent_w;
  local rgt_edge = 80 * percent_w;
  local upper_h  = 35 * percent_h;
  local lower_h  = 70 * percent_h;

  local focus_diff_x = ( focus_pos[ 1 ] - center_w );
  local focus_diff_y = ( focus_pos[ 2 ] - center_h );
  local proximity    = 1 * percent_w;

  local is_focus_on_top   = ( focus_diff_y < proximity );
  local is_focus_on_right = ( focus_diff_x > proximity );
  
  local dialog_settings = {
    start_x = 130 * percent_w,
    pos_x   = center_w,
    pos_y   = upper_h,
  };
  
  -- JK: if focus is near the edge, place the text in center
  if ( focus_pos[ 1 ] <= lft_edge ) or
     ( focus_pos[ 1 ] >= rgt_edge ) or
     ( focus_pos[ 2 ] <= top_edge ) or
     ( focus_pos[ 2 ] >= btm_edge ) then

    if ( focus_pos[ 1 ] >= rgt_edge ) then
     
      dialog_settings.start_x = -30 * percent_w;
     
    end;
    
    if ( focus_pos[ 2 ] <= top_edge ) or
       ( focus_pos[ 2 ] >= btm_edge ) then
     
      dialog_settings.pos_y = center_h;
    
    elseif ( is_focus_on_top ) then
    
      dialog_settings.pos_y = lower_h;
    
    end;
    
    dialog_settings = private.Dialog_Settings_TrySetPrefs(
      dialog_settings,
      pref_vertical,
      pref_horizontal,
      percent_h,
      percent_w
    );
    
    dialog_settings = private.Dialog_Settings_SetCharacter(
      dialog_settings,
      percent_h,
      percent_w,
      focus_pos
    );
    
    return dialog_settings;
     
  end;
  
  if ( landscape ) then

    dialog_settings.pos_x = 30 * percent_w;
    dialog_settings.pos_y = 30 * percent_h;
    
    if ( focus_pos[ 1 ] < center_w ) then
      
      dialog_settings.pos_x  = 70 * percent_w;
      
    end;
    
    if ( is_focus_on_top ) then
    
      dialog_settings.pos_y = lower_h;
    
    end;
    
    -- JK: in stage left top quarter must be clean as hero is there
    if ( v_stage.IsActive()                     ) and
       ( not v_interface.window.IsAnyOnScreen() ) and
       ( focus_pos[ 1 ] >= center_w             ) then
      
      dialog_settings.pos_y = btm_edge;
      
    end;
    
  elseif ( is_focus_on_top ) then
    
    dialog_settings.pos_y = lower_h;
    
  end;
  
  dialog_settings = private.Dialog_Settings_TrySetPrefs(
    dialog_settings,
    pref_vertical,
    pref_horizontal,
    percent_h,
    percent_w
  );
  
  dialog_settings = private.Dialog_Settings_SetCharacter(
    dialog_settings,
    percent_h,
    percent_w,
    focus_pos
  );
  
  -- JK: if character looks to the right, slide from the left
  if ( dialog_settings.ch_scale == -1 ) then
    
    dialog_settings.start_x = -30 * percent_w;
    
  end;
  
  return dialog_settings;
  
end;
--------------------------------------------------------------------------------
function private.Dialog_Settings_TrySetPrefs (
    dialog_settings,
    pref_vertical,
    pref_horizontal,
    percent_h,
    percent_w
  )
  
  if ( pref_vertical ) then
  
    dialog_settings.pos_y = pref_vertical * percent_h;
      
  end;

  -- JK: horizontal pref works in landscape only
  if ( private.landscape ) and ( pref_horizontal ) then
    
    dialog_settings.pos_x = pref_horizontal * percent_w;
    
    if ( dialog_settings.pos_x > 50 * percent_w ) then
    
      dialog_settings.ch_scale = 1;
      
    end;
      
  end;
  
  return dialog_settings;
  
end;
--------------------------------------------------------------------------------
function private.Dialog_Settings_SetCharacter (
    dialog_settings,
    percent_h,
    percent_w,
    focus_pos
  )

  local center_w  = 50 * percent_w;
  local proximity = 1 * percent_w;

  local dialog_diff = ( dialog_settings.pos_x - center_w );
  local focus_diff  = ( focus_pos[ 1 ]        - center_w );

  local is_dialog_on_right = ( dialog_diff > proximity );
  local is_focus_on_right  = ( focus_diff  > proximity );
  local is_space_on_top    = ( dialog_settings.pos_y >= 20 * percent_h );

  dialog_settings.ch_scale = tern( is_dialog_on_right )(   1,  -1 );
  dialog_settings.ch_scale = tern( is_focus_on_right  )(  -1,   1 );
  dialog_settings.ch_pos_y = tern( is_space_on_top    )( -96, 160 );
  
  return dialog_settings;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Entity_Add ( tutorial_name, entity_name, entity_type )

  if ( "booster" == entity_type ) then
    if not ( m_tutorial.Entity_IsAdded( tutorial_name, entity_name ) ) then
      v_boosters.TryChangeCount( entity_name, 1, "tutorial" );
      m_tutorial.Complete( tutorial_name.."_"..entity_name );
    end;
  elseif ( "stickerpack" == entity_type ) then
    if not ( m_tutorial.Entity_IsAdded( tutorial_name, entity_name ) ) then
      v_fireflies.Stickerpack_TryObtain( 1 );
      m_tutorial.Complete( tutorial_name.."_"..entity_name );
    end;
  end;

end;
--------------------------------------------------------------------------------
function private.Entity_Free_Show ( tutorial_name, entity_name, object )

  if ( m_tutorial.Entity_IsAdded( tutorial_name, entity_name ) ) then

    local obj_name = ui.obj.GetNameWithoutPrefix( object );
    local params   = ObjGet( object );

    ObjCreate( "txt_"..obj_name.."_free", "text" );
    ObjAttach( "txt_"..obj_name.."_free", object );
    ObjSet( "txt_"..obj_name.."_free", 
    {
      res = "assets/fonts/sml_title",
      text = "str_tutorial_free",
      pos_x = 0.5 * math.abs( params.inputrect_x ) or 25,
      pos_y = math.abs( params.inputrect_y ) - 10 or 40, 
      pos_z = 10,
      scale_x = 0,
      scale_y = 0,
      ang = 0.3,
      align = 1,
      fontsize = 25,
      fontcolor_r = 1,
      fontcolor_g = 0,
      fontcolor_b = 0,
      display_outline = true,
      outline_size = 3,
      outline_color_r = 1, 
      outline_color_g = 1, 
      outline_color_b = 1 
    } );

    private.backup_entity_count = ObjGet( "anm_"..obj_name.."_icon" ).count;
    ObjSet( "anm_"..obj_name.."_icon", { count = "" } );

    ObjAnimate( "txt_"..obj_name.."_free", "scale_xy", 0, 0, function ()

      private.Entity_Free_Animate( "txt_"..obj_name.."_free", true );

    end, 
    { 
      0.0, 0, 0, 0, 
      0.3, 0, 1, 1 
    } );


  end;

end;
--------------------------------------------------------------------------------
function private.Entity_Free_Animate ( object, animate )

  if ( animate ) then
    ObjAnimate( object, "ang", 1, 0, "", 
    { 
      0.0, 3, 0.3, 
      0.6, 3, 0.0,
      1.2, 3, 0.3 
    } );
    return;
  end;

  ObjStopAnimate( object, "ang" );

end;
--------------------------------------------------------------------------------
function private.Entity_Free_Hide ( object )

  local obj_name = ui.obj.GetNameWithoutPrefix( object );

  local func_end_anim = function ()

    ObjDelete( "txt_"..obj_name.."_free" );
    ObjSet( "anm_"..obj_name.."_icon",
            { count = private.backup_entity_count } );

  end;

  private.Entity_Free_Animate( "txt_"..obj_name.."_free", false );

  ObjAnimate( "txt_"..obj_name.."_free", "scale_xy", 0, 0, func_end_anim, 
  { 
    0.0, 0, 1, 1, 
    0.3, 0, 0, 0 
  } );

end;
--------------------------------------------------------------------------------
function private.Entity_Index ( entities_list, pattern, point )

  for i = #entities_list, 1, - 1 do
    local entity_type = m_tutorial.Entity_Type( entities_list[ i ] );
    if not ( entity_type == point ) then
      table.remove( entities_list, i );
    end;
  end;  

  local index = 1;

  for i = 1, #entities_list do
    if ( entities_list[ i ] == pattern ) then
      index = i; break;
    end;
  end;

  return index;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Boosters_Backup ( tutorial_name, exclude_index, other, custom_name )

  local func_check = function ( index )

    if ( exclude_index ) then

      return not ( exclude_index == index );

    end;

    return true;  

  end;
  
  local parent_name = "obj_int_panel_stage_booster";
  
  if ( custom_name ) then
    
    parent_name = custom_name;
    
  end;

  for i = 1, 3 do

    if func_check( i ) then

      local object_name = parent_name.."_"..i;

      local backup_name = "otherbooster_tap_"..i;

      if ( other ) then

        object_name = parent_name.."_"..i.."_count";

        backup_name = "otherboosterbuy_tap_"..i;

      end;

      local tap_params = ui.control.TapGet( object_name );

      if ( tap_params ) and ( tap_params.func_true ) then

        private.Backup( tutorial_name, backup_name, tap_params.func_true );

        ui.control.TapSet( object_name, { func_true = function () end } );

      end;

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Boosters_Restore ( tutorial_name, exclude_index, other, custom_name )

  local func_check = function ( index )

    if ( exclude_index ) then

      return not ( exclude_index == index );

    end;

    return true;  

  end;
  
  local parent_name = "obj_int_panel_stage_booster";
  
  if ( custom_name ) then
    
    parent_name = custom_name;
    
  end;  

  for i = 1, 3 do

    local object_name = parent_name.."_"..i;

    local backup_name = "otherbooster_tap_"..i;

    if ( other ) then

      object_name = parent_name.."_"..i.."_count";

      backup_name = "otherboosterbuy_tap_"..i;

    end;

    if func_check( i ) then

      local func_restored = private.Restore( tutorial_name, backup_name );

      if ( func_restored ) then

        ui.control.TapSet( object_name, { func_true = func_restored } );

      end;

    end;

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stage_animation ()
  
  private.Subscribe(
    "stage_animation",
    "View_Tutorial_StageOpened",
    private.stage_animation_first
  );
  
end;
--------------------------------------------------------------------------------
function private.stage_animation_first ( stage_name )
  
  local data = v_stage.Stage_StaticData( stage_name );
  
  if ( data.tutorial ) and ( not m_tutorial.IsCompleted( stage_name ) ) then
    
    private.stage_animation_data = data;
    
    v_stage.SetPresetEffect( private.stage_animation_next );
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.stage_animation_next ( func_after )
  
  func_after = func_after or function () end;
  
  if ( not private.stage_animation_data ) then
    func_after();
    return;
  end;
  
  local tutorial_name = private.stage_animation_data.name;
  
  local delegates = {
    Close = function ()
      
      int_dialog_tutorial_animation.Hide(
        function ()
          v_interface.window.Hide( private.this.dialog_animation );
          func_after();
        end
      );
      
    end
  };
  
  v_interface.window.Show(
    private.this.dialog_animation,
    function ()
      int_dialog_tutorial_animation.Show( private.stage_animation_data );
    end,
    nil,
    delegates
  );
  
  private.stage_animation_last( tutorial_name );
  
end;
--------------------------------------------------------------------------------
function private.stage_animation_last ( tutorial_name )
  
  private.stage_animation_data = nil;
  
  private.Complete( tutorial_name );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stage_1_tap ()

  local delegates = {
    GetCurrentSectionName = function ()
      local  section_name = v_sections.Section_NameCurrent();
      return section_name;
    end
  };

  private.Tutorial_SetOnEvent(
    "stage_1_tap",
    "View_Map_Update_Finished",
    delegates
  );

end;
--------------------------------------------------------------------------------
function private.stage_1_start_dialog ()
  
  private.Subscribe(
    "stage_1_start_dialog",
    "View_Stage_DialogStartActivated",
    private.stage_1_start_dialog_next
  );
  
end;
--------------------------------------------------------------------------------
function private.stage_1_start_dialog_next ( params )
  
  private.Unsubscribe(
    "stage_1_start_dialog",
    "View_Stage_DialogStartActivated",
    private.stage_1_start_dialog_next
  );
  
  if ( ( params.name ~= "stage_1" ) or ( v_stage.Stage_StorageData( "stage_1" ) or {} ).completed ) then
    private.stage_1_start_dialog_last();
    return;
  end;
  
  private.Kontagent_LogEvent( "tutorial",
  {
    name  = "tut_level_1_play"
  } );  
  
  local focus_name = int_dialog_stage_start.GetObjectPlayName();
  local dialog_settings = private.Dialog_Settings( focus_name );
  
  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type   = "flashlight",

    res_id  = 3,
    focus   = focus_name,
    scale_x = 2.0,
    scale_y = 1.0,
    
    lock   = true,
    black  = true,
    hand   = {
      angle = 0.5 * math.pi,
      type  = "click"
    }
  };
  
  v_interface.window.ClearOuterTap();
  
  local func_accept = int_dialog_stage_start.__GetDelegateScheme().Accept;
  
  int_dialog_stage_start.__GetDelegateScheme().Accept = function ()
    private.Dialog_Hide();
    func_accept( {} );
    int_dialog_stage_start.__GetDelegateScheme().Accept = func_accept;
  end;
  
  private.Dialog_ShowInstantly( "stage_1_start_dialog_1", positions, hole );
  
  private.stage_1_start_dialog_last();
  
end;
--------------------------------------------------------------------------------
function private.stage_1_start_dialog_last ()
  
  private.Complete( "stage_1_start_dialog" );
  
end;
--------------------------------------------------------------------------------
function private.stage_1_play ()

  private.Subscribe( "stage_1_play", "View_Stage_DialogStartDeactivated", private.stage_1_play_last );

end;

function private.stage_1_play_last ( stage_data )

  if ( stage_data.name == "stage_1" and stage_data.start and not m_tutorial.IsCompleted( "stage_1_play" ) ) then

    private.Unsubscribe( "stage_1_play", "View_Stage_DialogStartDeactivated", private.stage_1_play_last ); 

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_level_1_playdone"
    } );  

    private.Complete( "stage_1_play" );

  end;    

end;
--------------------------------------------------------------------------------
function private.stage_1 ()

  local func_validate = function ( stage_name )
    local  is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_1"
    );
    return is_valid;
  end;

  local delegates = {
    GetTargetCount = function ()
      local  target_count = v_stage.Stage_GetCurrentData().targets[ 2 ].count;
      return target_count;
    end
  };

  private.Tutorial_SetOnEvent(
    "stage_1",
    "View_Tutorial_StageOpened",
    delegates,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.stage_1_win ()

  private.Subscribe( "stage_1_win", "Model_Stage_StageWon", private.stage_1_win_last );

end;

function private.stage_1_win_last ( stage_data )

  if ( stage_data.name == "stage_1" and not m_tutorial.IsCompleted( "stage_1_win" ) ) then

    private.Unsubscribe( "stage_1_win", "Model_Stage_StageWon", private.stage_1_win_last ); 

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_level_1_win"
    } );  

    private.Complete( "stage_1_win" );

  end;    

end;
--------------------------------------------------------------------------------
function private.stage_1_reward ()

  private.Subscribe( "stage_1_reward", "View_Map_Update_StageRewardCollected", private.stage_1_reward_last );

end;

function private.stage_1_reward_last ( stage_name )

  if ( stage_name == "stage_1" and not m_tutorial.IsCompleted( "stage_1_reward" ) ) then

    private.Unsubscribe( "stage_1_reward", "View_Map_Update_StageRewardCollected", private.stage_1_reward_last );

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_level_1_reward"
    } );  

    private.Complete( "stage_1_reward" );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stage_1_modification_1 ()

  local func_validate = function ( stage_name )
    local  is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_1"
    );
    return is_valid;
  end;

  local delegates = {
    GetTargetCount = function ()
      local  target_count = v_stage.Stage_GetCurrentData().targets[ 2 ].count;
      return target_count;
    end
  };

  private.Tutorial_SetOnEvent(
    "stage_1_modification_1",
    "View_Tutorial_StageOpened",
    delegates,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.swordstone ()

  local func_validate = function ( stage_name )
    local is_valid =
      ( v_stage.Stage_StorageData( "stage_1" ) or {} ).completed;
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "swordstone",
    "View_Tutorial_MapOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stage_2_tap ()

  private.Subscribe( "stage_2_tap", "View_Stage_DialogStartActivated", private.stage_2_tap_last );

end;

function private.stage_2_tap_last ( stage_data )

  if ( stage_data.name == "stage_2" and not m_tutorial.IsCompleted( "stage_2_tap" ) ) then

    private.Unsubscribe( "stage_2_tap", "View_Stage_DialogStartActivated", private.stage_2_tap_last ); 

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_level_2"
    } );  

    private.Complete( "stage_2_tap" );

  end;    

end;
--------------------------------------------------------------------------------
function private.stage_2_play ()

  private.Subscribe( "stage_2_play", "View_Stage_DialogStartDeactivated", private.stage_2_play_last );

end;

function private.stage_2_play_last ( stage_data )

  if ( stage_data.name == "stage_2" and stage_data.start and not m_tutorial.IsCompleted( "stage_2_play" ) ) then

    private.Unsubscribe( "stage_2_play", "View_Stage_DialogStartDeactivated", private.stage_2_play_last ); 

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_level_2_play"
    } );  

    private.Complete( "stage_2_play" );

  end;    

end;
--------------------------------------------------------------------------------
function private.stage_2 ()

  local func_validate = function ( stage_name )
    local is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_2"
    );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "stage_2",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.stage_2_win ()

  private.Subscribe( "stage_2_win", "Model_Stage_StageWon", private.stage_2_win_last );

end;

function private.stage_2_win_last ( stage_data )

  if ( stage_data.name == "stage_2" and not m_tutorial.IsCompleted( "stage_2_win" ) ) then

    private.Unsubscribe( "stage_2_win", "Model_Stage_StageWon", private.stage_2_win_last ); 

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_level_2_win"
    } );  

    private.Complete( "stage_2_win" );

  end;    

end;
--------------------------------------------------------------------------------
function private.stage_2_reward ()

  private.Subscribe( "stage_2_reward", "View_Map_Update_StageRewardCollected", private.stage_2_reward_last );

end;

function private.stage_2_reward_last ( stage_name )

  if ( stage_name == "stage_2" and not m_tutorial.IsCompleted( "stage_2_reward" ) ) then

    private.Unsubscribe( "stage_2_reward", "View_Map_Update_StageRewardCollected", private.stage_2_reward_last );

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_level_2_reward"
    } );  

    private.Complete( "stage_2_reward" );

  end;    

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stage_2_modification_1 ()

  local func_validate = function ( stage_name )
    local  is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_2"
    );
    return is_valid;
  end;

  local delegates = {
    GetTargetCount = function ()
      local  target_count = v_stage.Stage_GetCurrentData().targets[ 2 ].count;
      return target_count;
    end
  };

  private.Tutorial_SetOnEvent(
    "stage_2_modification_1",
    "View_Tutorial_StageOpened",
    delegates,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.stage_2_modification_2 ()

  local func_validate = function ( stage_name )
    local  is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_2"
    );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "stage_2_modification_2",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.quest ()

  private.Subscribe( "quest", "View_Tutorial_MapOpened",
    private.quest_first
  );
  
end;

function private.quest_first ()

  if ( not m_tutorial.IsCompleted( "quest" ) ) then

    if ( ( v_stage.Stage_StorageData( "stage_2" ) or {} ).completed ) then

      private.Unsubscribe( "quest", "View_Tutorial_MapOpened",
                           private.quest_first );

      private.Subscribe( "quest", "View_Map_Update_Character_PopupDisplayed",
                         private.quest_next1 );

    end;

  end;

end; 

function private.quest_next1 ( character_name )

  if not ( character_name == "wizard" ) then
    return;
  end;
  
  private.Unsubscribe( 
    "quest",
    "View_Map_Update_Character_PopupDisplayed",
    private.quest_next1
  );
  
  private.PendingBegin( "quest", "map" );

  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_cage"
  } );

  local func_close = function ()

    local func_OnTutorialHide = function ()

      private.quest_last1();

      mediator.Notify( "View_Tutorial_TutorialQuestCompleted" );

    end;

    private.Dialog_Hide( func_OnTutorialHide );

  end;
  
  local pref_vertical = 25;
  
  if ( private.landscape ) then
    
    pref_vertical = nil;
    
  end;
  
  local dialog_settings =
    private.Dialog_Settings( "obj_map_characters_wizard", pref_vertical );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type        = "arrow",

    focus       = "obj_map_characters_wizard",
    offset_x    = 0,
    offset_y    = - 90,
    angle       = 0,
    scale       = 1.0,

    lock        = true
  };

  private.Dialog_ShowInstantly( "quest", positions, hole, func_close, 
                                nil, { active = true } );

end;

function private.quest_last1 ()

  private.Complete( "quest" );
  private.PendingEnd( "quest", "map" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stage_3_tap ()

  private.Subscribe( "stage_3_tap", "View_Stage_DialogStartActivated", private.stage_3_tap_last );

end;

function private.stage_3_tap_last ( stage_data )

  if ( stage_data.name == "stage_3" and not m_tutorial.IsCompleted( "stage_3_tap" ) ) then

    private.Unsubscribe( "stage_3_tap", "View_Stage_DialogStartActivated", private.stage_3_tap_last ); 

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_level_3"
    } );  

    private.Complete( "stage_3_tap" );

  end;    

end;
--------------------------------------------------------------------------------
function private.stage_3_play ()

  private.Subscribe( "stage_3_play", "View_Stage_DialogStartDeactivated", private.stage_3_play_last );

end;

function private.stage_3_play_last ( stage_data )

  if ( stage_data.name == "stage_3" and stage_data.start and not m_tutorial.IsCompleted( "stage_3_play" ) ) then

    private.Unsubscribe( "stage_3_play", "View_Stage_DialogStartDeactivated", private.stage_3_play_last ); 

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_level_3_play"
    } );  

    private.Complete( "stage_3_play" );

  end;    

end;
--------------------------------------------------------------------------------
function private.stage_3 ()

  private.Subscribe(
    "stage_3",
    "View_Tutorial_StageOpened",
    private.stage_3_first
  );

end;

function private.stage_3_first ( stage_name )

  if ( stage_name == "stage_3"                 ) and
     ( not ( v_stage.Stage_StorageData( "stage_3" ) or {} ).completed ) and
     ( not m_tutorial.IsCompleted( "stage_3" ) ) then

    private.Unsubscribe(
      "stage_3",
      "View_Tutorial_StageOpened",
      private.stage_3_first
    ); 
  
    private.Subscribe(
      "stage_3",
      "View_Stage_DialogTargetsHided",
      private.stage_3_next1
    );

  end;

end;

function private.stage_3_next1 ()

  private.Unsubscribe(
    "stage_3",
    "View_Stage_DialogTargetsHided",
    private.stage_3_next1
  );

  private.Subscribe(
    "stage_3",
    "Model_Match3_FieldActivated",
    private.stage_3_hide1
  );

  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_level_3_matchl"
  } );  

  local match3_pos = GetObjPosByObj( "match" );

  local pos_x = match3_pos[ 1 ] + 2 * private.cell_size * 5;
  local pos_y = match3_pos[ 2 ] + 2 * private.cell_size * 5;

  local dialog_settings =
    private.Dialog_Settings( { pos_x, pos_y } );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type           = "flashlight",

    res_id         = 13,
    pos_x          = pos_x - private.cell_size,
    pos_y          = pos_y,
    angle          = 0,
    scale          = private.cell_scale,

    lock           = true,
    hand           = {
         angle     = 0.5 * math.pi,
         offset_x  = -2 * private.cell_size,
         offset_y  = -2 * private.cell_size,
         type      = "move"
    }
  };

  local func_close = function ()

    --

  end;

  private.Dialog_ShowInstantly( "stage_3_1", positions, hole, func_close );

  private.stage_3_last1();

end;

function private.stage_3_hide1 ()
  
  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_level_3_matchldone"
  } );   

  private.Unsubscribe(
    "stage_3",
    "Model_Match3_FieldActivated",
    private.stage_3_hide1
  );

  private.Dialog_Hide();
  
end;

function private.stage_3_last1 ()

  private.Complete( "stage_3" );

end;
--------------------------------------------------------------------------------
function private.stage_3_win ()

  private.Subscribe( "stage_3_win", "Model_Stage_StageWon", private.stage_3_win_last );

end;

function private.stage_3_win_last ( stage_data )

  if ( stage_data.name == "stage_3" and 
       not m_tutorial.IsCompleted( "stage_3_win" ) ) then

    private.Unsubscribe( "stage_3_win",
                     "Model_Stage_StageWon",
                     private.stage_3_win_last ); 

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_level_3_win"
    } );  

    private.Complete( "stage_3_win" );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stage_3_modification_1 ()

  local func_validate = function ( stage_name )
    local  is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_3"
    );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "stage_3_modification_1",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.stage_3_modification_2 ()

  local func_validate = function ( stage_name )
    local  is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_3"
    );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "stage_3_modification_2",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stage_4_tap ()

  private.Subscribe( "stage_4_tap", "View_Stage_DialogStartActivated", private.stage_4_tap_last );

end;

function private.stage_4_tap_last ( stage_data )

  if ( stage_data.name == "stage_4" and not m_tutorial.IsCompleted( "stage_4_tap" ) ) then

    private.Unsubscribe( "stage_4_tap", "View_Stage_DialogStartActivated", private.stage_4_tap_last ); 

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_level_4"
    } );  

    private.Complete( "stage_4_tap" );

  end;    

end;
--------------------------------------------------------------------------------
function private.stage_4_play ()

  private.Subscribe( "stage_4_play", "View_Stage_DialogStartDeactivated", private.stage_4_play_last );

end;

function private.stage_4_play_last ( stage_data )

  if ( stage_data.name == "stage_4" and stage_data.start and not m_tutorial.IsCompleted( "stage_4_play" ) ) then

    private.Unsubscribe( "stage_4_play", "View_Stage_DialogStartDeactivated", private.stage_4_play_last ); 

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_level_4_play"
    } );  

    private.Complete( "stage_4_play" );

  end;    

end;
--------------------------------------------------------------------------------
function private.stage_4 ()

  private.Subscribe(
    "stage_4",
    "View_Tutorial_StageOpened",
    private.stage_4_first
  );

end;

function private.stage_4_first ( stage_name )

  if ( stage_name == "stage_4"                 ) and
     ( not ( v_stage.Stage_StorageData( "stage_4" ) or {} ).completed ) and
     ( not m_tutorial.IsCompleted( "stage_4" ) ) then

    private.Unsubscribe(
      "stage_4",
      "View_Tutorial_StageOpened",
      private.stage_4_first
    ); 

    private.Subscribe(
      "stage_4",
      "View_Stage_DialogTargetsHided",
      private.stage_4_next1
    );

  end;

end;

function private.stage_4_next1 ()

  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_level_4_fireflies"
  } );

  local index = private.Entity_Index( v_stage.Stage_GetCurrentData().boosters, "hammer", "stage" ); 

  private.Unsubscribe( "stage_4", "View_Stage_DialogTargetsHided", private.stage_4_next1 );

  private.Backup( "stage_4", "booster_tap", ui.control.EventsGet( "obj_int_panel_stage_booster_"..index ).func_on_click );

  ui.control.EventsClear( "obj_int_panel_stage_booster_"..index );

  local focus_obj = "anm_int_panel_stage_top_goals_goal2";
  
  local dialog_settings =
    private.Dialog_Settings( focus_obj );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type        = "arrow",

    focus       = focus_obj,
    offset_x    = 0,
    offset_y    = 0,
    scale       = 1,
    angle       = math.pi,

    lock        = true
  };

  local func_close = function ()

    private.Dialog_Hide( function () private.stage_4_next2( index ); end );
    
  end;

  private.Dialog_ShowInstantly( "stage_4_1", positions, hole, func_close );

end;

function private.stage_4_next2 ( index )

  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_level_4_hammer"
  } );

  local dialog_settings =
    private.Dialog_Settings( "obj_int_panel_stage_booster_"..index );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = ui.obj.Position( "obj_int_panel_stage_booster_"..index )[ 2 ] - 200,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type        = "flashlight",  

    res_id      = 3,
    focus       = "obj_int_panel_stage_booster_"..index,
    scale       = 1,
  
    lock        = true,
    black       = true,
    hand        = {
      angle     = 0.5 * math.pi,
      offset_x  = 0,
      offset_y  = 0,
      type      = "click"
    }
  };

  local func_close = function ()

    private.Dialog_Hide( function () 

      private.stage_4_next3( index ); 

    end );

  end;

  ui.control.EventsSet( "obj_int_panel_stage_booster_"..index, { func_on_click = func_close } );

  private.Dialog_ShowInstantly( "stage_4_2", positions, hole );
  
  private.Entity_Add( "stage_4", "hammer", "booster" );

  private.Entity_Free_Show( "stage_4", "hammer", "obj_int_panel_stage_booster_"..index );

end;

function private.stage_4_next3 ( index )

  private.Subscribe( "stage_4", "View_Boosters_BoosterUsed", private.stage_4_last1 );
  private.Subscribe(
    "stage_4",
    "View_Boosters_BoosterUsageFailed",
    private.stage_4_booster_reset
  );

  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_level_4_hammerdone"
  } );

  local func_booster_tap = private.Restore( "stage_4", "booster_tap" );

  ui.control.EventsSet( "obj_int_panel_stage_booster_"..index, { func_on_click = func_booster_tap } );

  if ( func_booster_tap ) then

    func_booster_tap();

    private.Entity_Free_Hide( "obj_int_panel_stage_booster_"..index );

  end;
  
  local match3_pos = GetObjPosByObj( "match" );

  local pos_x = match3_pos[ 1 ] + 2 * private.cell_size * 4 - private.cell_size;
  local pos_y = match3_pos[ 2 ] + 2 * private.cell_size * 5 + private.cell_size;
  
  local dialog_settings =
    private.Dialog_Settings( { pos_x, pos_y } );
  
  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type           = "flashlight",

    res_id         = 3,
    pos_x          = pos_x,
    pos_y          = pos_y,
    angle          = 0,
    scale          = 0.6,

    lock           = true,
    hand           = {
         angle     = 0,
         offset_x  = 0,
         offset_y  = 0,
         type      = "click"
    }
  };

  private.Dialog_ShowInstantly( "stage_4_3", positions, hole );  

end;

function private.stage_4_booster_reset ()

  local boosters = v_stage.Stage_GetCurrentData().boosters;
  local index = private.Entity_Index( boosters, "hammer", "stage" );
  ui.control.EventsGet( "obj_int_panel_stage_booster_"..index ).func_on_click();

end;

function private.stage_4_last1 ()

  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_level_4_hammeruse"
  } );

  private.Unsubscribe( "stage_4", "View_Boosters_BoosterUsed", private.stage_4_last1 );
  
  private.Unsubscribe(
    "stage_4",
    "View_Boosters_BoosterUsageFailed",
    private.stage_4_booster_reset
  );
  
  private.Dialog_Hide();
  private.Complete( "stage_4" );

end;
--------------------------------------------------------------------------------
function private.stage_4_win ()

  private.Subscribe( "stage_4_win", "Model_Stage_StageWon", private.stage_4_win_last );

end;

function private.stage_4_win_last ( stage_data )

  if ( stage_data.name == "stage_4" and 
    not m_tutorial.IsCompleted( "stage_4_win" ) ) then

    private.Unsubscribe( "stage_4_win",
                     "Model_Stage_StageWon",
                     private.stage_4_win_last ); 

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_level_4_win"
    } );  

    private.Complete( "stage_4_win" );

  end;    

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stage_4_modification_1 ()

  local func_validate = function ( stage_name )
    local  is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_4"
    );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "stage_4_modification_1",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stage_5_tap ()

  private.Subscribe( "stage_5_tap", "View_Stage_DialogStartActivated", private.stage_5_tap_last );

end;

function private.stage_5_tap_last ( stage_data )

  if ( stage_data.name == "stage_5" and not m_tutorial.IsCompleted( "stage_5_tap" ) ) then

    private.Unsubscribe( "stage_5_tap", "View_Stage_DialogStartActivated", private.stage_5_tap_last ); 

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_level_5"
    } );  

    private.Complete( "stage_5_tap" );

  end;    

end;
--------------------------------------------------------------------------------
function private.stage_5_play ()

  private.Subscribe( "stage_5_play", "View_Stage_DialogStartDeactivated", private.stage_5_play_last );

end;

function private.stage_5_play_last ( stage_data )

  if ( stage_data.name == "stage_5" and stage_data.start and not m_tutorial.IsCompleted( "stage_5_play" ) ) then

    private.Unsubscribe( "stage_5_play", "View_Stage_DialogStartDeactivated", private.stage_5_play_last ); 

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_level_5_play"
    } );  

    private.Complete( "stage_5_play" );

  end;    

end;
--------------------------------------------------------------------------------
function private.stage_5 ()

  private.Subscribe(
    "stage_5",
    "View_Tutorial_StageOpened",
    private.stage_5_first
  );

end;

function private.stage_5_first ( stage_name )

  if ( stage_name == "stage_5" ) and 
     ( not ( v_stage.Stage_StorageData( "stage_5" ) or {} ).completed ) and
     ( not m_tutorial.IsCompleted( "stage_5" ) ) then

    private.Unsubscribe(
      "stage_5",
      "View_Tutorial_StageOpened",
      private.stage_5_first
    ); 
  
    private.Subscribe(
      "stage_5",
      "View_Stage_DialogTargetsHided",
      private.stage_5_next1
    );

  end;

end;

function private.stage_5_next1 ()

  private.Unsubscribe(
    "stage_5",
    "View_Stage_DialogTargetsHided",
    private.stage_5_next1
  );

  private.Subscribe(
    "stage_5",
    "Model_Match3_FieldActivated",
    private.stage_5_hide1
  );

  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_level_5_matchbigcross"
  } );  

  local match3_pos = GetObjPosByObj( "match" );

  local pos_x = match3_pos[ 1 ] + 2 * private.cell_size * 5;
  local pos_y = match3_pos[ 2 ] + 2 * private.cell_size * 4;

  local dialog_settings =
    private.Dialog_Settings( { pos_x, pos_y } );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type           = "flashlight",

    res_id         = 4,
    pos_x          = pos_x - private.cell_size,
    pos_y          = pos_y,
    angle          = 0.5 * math.pi,
    scale          = private.cell_scale,

    lock           = true,
    hand           = {
         angle     = 0.5 * math.pi,
         offset_x  = 0,
         offset_y  = 0,
         type      = "move"
    }
  };

  local func_close = function ()

    --

  end;

  private.Dialog_ShowInstantly( "stage_5_1", positions, hole, func_close );

  private.stage_5_last1();

end;

function private.stage_5_hide1 ()
  
  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_level_5_matchbigcrossdone"
  } );   

  private.Unsubscribe(
    "stage_5",
    "Model_Match3_FieldActivated",
    private.stage_5_hide1
  );

  private.Dialog_Hide();
  
end;

function private.stage_5_last1 ()

  private.Complete( "stage_5" );

end;
--------------------------------------------------------------------------------
function private.stage_5_win ()

  private.Subscribe( "stage_5_win", "Model_Stage_StageWon", private.stage_5_win_last );

end;

function private.stage_5_win_last ( stage_data )

  if ( stage_data.name == "stage_5" and 
    not m_tutorial.IsCompleted( "stage_5_win" ) ) then

    private.Unsubscribe( "stage_5_win",
                     "Model_Stage_StageWon",
                     private.stage_5_win_last ); 

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_level_5_win"
    } );  

    private.Complete( "stage_5_win" );

  end;    

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stage_5_modification_1 ()

  local func_validate = function ( stage_name )
    local  is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_5"
    );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "stage_5_modification_1",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stickerpack ()

  private.Subscribe(
    "stickerpack",
    "Interface_DialogFireflies_Opened",
    private.stickerpack_next
  );

end;

function private.stickerpack_next ()
    
  private.Unsubscribe(
    "stickerpack",
    "Interface_DialogFireflies_Opened",
    private.stickerpack_next
  );

  if ( m_tutorial.IsCompleted( "stickerpack" ) ) then return; end;
  if ( v_fireflies.Stickerpack_Count().total == 0 ) then return; end;

  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_stickerpack"
  } );

  local focus_obj = "btn_int_dialog_fireflies_stickerpack_get"; 
  
  local dialog_settings = private.Dialog_Settings( focus_obj );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type     = "flashlight",
    focus    = focus_obj,
    
    res_id   = 3,
    offset_x = 0,
    offset_y = 0,
    angle    = 0,
    scale_x  = 2.1,
    scale_y  = 1.0,
    
    lock     = true,
    black    = true,
    wait     = true,
    hand = {  
         offset_x  = 0,
         offset_y  = 0,
         scale_x   = - 1,
         angle     = - 1.57, 
         type      = "click"
    }
  };    
  
  local func_close = function ()

    private.Dialog_Hide( function()
      
      -- JK: lock here, unlock in next tutorial ( private.stickerpool )
      ui.screen.Lock( true );
    
      private.stickerpack_last();
     
      local func_stickerpack_tap =
        private.Restore( "stickerpack", "stickerpack_tap" );

      ui.button.EventSet(
        "btn_int_dialog_fireflies_stickerpack_get",
        func_stickerpack_tap
      );

      if ( func_stickerpack_tap ) then

        func_stickerpack_tap( true );

      end;
     
    end );
    
  end;

  private.Backup(
    "stickerpack",
    "stickerpack_tap",
    ui.button.EventGet( "btn_int_dialog_fireflies_stickerpack_get" )
  );

  ui.button.EventSet(
    "btn_int_dialog_fireflies_stickerpack_get",
    func_close
  );
  
  private.Dialog_ShowInstantly( "stickerpack_1", positions, hole, func_close );
  
end;

function private.stickerpack_last ()
  
  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_stickerpackdone"
  } );  

  private.Complete( "stickerpack" );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stickerpool ()

  private.Subscribe(
    "stickerpool",
    "Interface_DialogFireflies_HasStickable",
    private.stickerpool_first
  );

end;

function private.stickerpool_first ( stickable_stickers )

  local sticker = stickable_stickers[ 1 ];
  
  private.Unsubscribe(
    "stickerpool",
    "Interface_DialogFireflies_HasStickable",
    private.stickerpool_first
  );
    
  if ( not m_tutorial.IsCompleted( "stickerpool" ) ) then
    
    v_fireflies.Dialog_GoToSticker( sticker );
  
    common.Defer(
      "tutorial_stickerpool_delay", 0.5,
      function () private.stickerpool_next1( sticker ); end
    );

  end;

end;

function private.stickerpool_next1 ( sticker )

  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_sticker_move"
  } );
  
  local focus_obj = "obj_int_dialog_fireflies_stickerpool";
  
  local dialog_settings = private.Dialog_Settings( focus_obj, 75 );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local start_pos  =
    v_fireflies.Dialog_GetStickerPoolPos( sticker.name );
  local finish_pos =
    v_fireflies.Dialog_StickerPosition( sticker.index );
  
  v_fireflies.Dialog_SetButtonsInput( false );
  v_interface.window.DisableOuterTap();
    
  local hole = 
  {
    type   = "sticker",
    start  = start_pos,
    finish = finish_pos,
    lock   = false,
    black  = false,
  };
  
  private.Subscribe(
    "stickerpool",
    "Interface_DialogFireflies_StickerSticked",
    private.stickerpool_last1
  );

  private.Dialog_ShowInstantly( "stickerpool_1", positions, hole );

  -- JK: unlock here, locked in previous tutorial ( private.stickerpack )
  ui.screen.Lock( false );
  
end;

function private.stickerpool_last1 ()
  
  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_sticker_movedone"
  } );    
  
  v_fireflies.Dialog_SetButtonsInput( true );
  v_interface.window.EnableOuterTap();
  
  private.Complete( "stickerpool" );

  private.Unsubscribe(
    "stickerpool",
    "Interface_DialogFireflies_StickerSticked",
    private.stickerpool_last1
  );

  private.Dialog_Hide();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stickerpack_buy ()

  private.Subscribe(
    "stickerpack_buy",
    "Interface_DialogFireflies_BuyMode",
    private.stickerpack_buy_first
  );

end;

function private.stickerpack_buy_first ()

  if ( not m_tutorial.IsCompleted( "stickerpool" ) ) then

    return;

  end;

  private.Unsubscribe(
    "stickerpack_buy",
    "Interface_DialogFireflies_BuyMode",
    private.stickerpack_buy_first
  );
  
  local dialog_settings =
    private.Dialog_Settings( "btn_int_dialog_fireflies_stickerpack_buy" );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
 
  local hole = 
  {
    type  = "arrow",
    focus = "btn_int_dialog_fireflies_stickerpack_buy",
    
    offset_x = 0,
    offset_y = 0,
    angle    = 0,
    scale    = 1.0, 

    lock     = true
  };
  
  private.Dialog_ShowInstantly( "stickerpack_buy_1", positions, hole, function()
 
    private.Dialog_Hide( private.stickerpack_buy_last );
 
  end );
  
end;

function private.stickerpack_buy_last ()
  
  private.Complete( "stickerpack_buy" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stage_6_tap ()

  private.Subscribe( "stage_6_tap", "View_Stage_DialogStartActivated", private.stage_6_tap_last );

end;

function private.stage_6_tap_last ( stage_data )

  if ( stage_data.name == "stage_6" and not m_tutorial.IsCompleted( "stage_6_tap" ) ) then

    private.Unsubscribe( "stage_6_tap", "View_Stage_DialogStartActivated", private.stage_6_tap_last ); 

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_level_6"
    } );  

    private.Complete( "stage_6_tap" );

  end;    

end;
--------------------------------------------------------------------------------
function private.stage_6_play ()

  private.Subscribe( "stage_6_play", "View_Stage_DialogStartDeactivated", private.stage_6_play_last );

end;

function private.stage_6_play_last ( stage_data )

  if ( stage_data.name == "stage_6" and stage_data.start and not m_tutorial.IsCompleted( "stage_6_play" ) ) then

    private.Unsubscribe( "stage_6_play", "View_Stage_DialogStartDeactivated", private.stage_6_play_last ); 

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_level_6_play"
    } );  

    private.Complete( "stage_6_play" );

  end;    

end;
--------------------------------------------------------------------------------
function private.stage_6 ()

  private.Subscribe(
    "stage_6",
    "View_Tutorial_StageOpened",
    private.stage_6_first
  );

end;

function private.stage_6_first ( stage_name )

  if ( stage_name == "stage_6"                 ) and
     ( not ( v_stage.Stage_StorageData( "stage_6" ) or {} ).completed ) and
     ( not m_tutorial.IsCompleted( "stage_6" ) ) then

    private.Unsubscribe(
      "stage_6",
      "View_Tutorial_StageOpened",
      private.stage_6_first
    );

    private.Subscribe(
      "stage_6",
      "View_Stage_DialogTargetsHided",
      private.stage_6_next1
    );

  end;

end;

function private.stage_6_next1 ()

  private.Unsubscribe(
    "stage_6",
    "View_Stage_DialogTargetsHided",
    private.stage_6_next1
  );

  private.Subscribe(
    "stage_6",
    "Model_Match3_FieldActivated",
    private.stage_6_hide1
  );

  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_level_6_match5"
  } );

  local match3_pos = GetObjPosByObj( "match" );

  local pos_x = match3_pos[ 1 ] + 2 * private.cell_size * 3 + private.cell_size;
  local pos_y = match3_pos[ 2 ] + 2 * private.cell_size * 3;
  
  local dialog_settings =
    private.Dialog_Settings( { pos_x, pos_y } );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type           = "flashlight",

    res_id         = 5,
    pos_x          = pos_x,
    pos_y          = pos_y,
    angle          = 0,
    scale          = private.cell_scale,

    lock           = true,
    hand           = {
         angle     = 0.5 * math.pi,
         offset_x  = 0,
         offset_y  = 0,
         type      = "move"
    }
  };

  local func_close = function ()

    --
    
  end;

  private.Dialog_ShowInstantly( "stage_6_1", positions, hole, func_close );

  private.stage_6_last();

end;

function private.stage_6_hide1 ()

  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_level_6_match5done"
  } );

  private.Unsubscribe(
    "stage_6",
    "Model_Match3_FieldActivated",
    private.stage_6_hide1
  );

  private.Dialog_Hide();

end;

function private.stage_6_last ()

  private.Complete( "stage_6" );

end;
--------------------------------------------------------------------------------
function private.stage_6_win ()

  private.Subscribe( "stage_6_win", "Model_Stage_StageWon", private.stage_6_win_last );

end;

function private.stage_6_win_last ( stage_data )

  if ( stage_data.name == "stage_6" and 
    not m_tutorial.IsCompleted( "stage_6_win" ) ) then

    private.Unsubscribe( "stage_6_win",
                     "Model_Stage_StageWon",
                     private.stage_6_win_last ); 

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_level_6_win"
    } );  

    private.Complete( "stage_6_win" );

  end;    

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stage_6_modification_1 ()

  local func_validate = function ( stage_name )
    local  is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_6"
    );
    return is_valid;
  end;

  local delegates = {
    SetBoosterPanelInput = function ( value, point )
      v_stage.PanelBoosters_SetInput( value, point );
    end
  };

  private.Tutorial_SetOnEvent(
    "stage_6_modification_1",
    "View_Tutorial_StageOpened",
    delegates,
    func_validate
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.dailyquest_countdown ()

  local delegates = {
    IsDailyquestSuspended = function ()
      local  is_suspended = v_dailyquest.IsSuspended();
      return is_suspended;
    end
  };

  private.Tutorial_SetOnEvent(
    "dailyquest_countdown",
    "View_DailyQuest_CountdownStarted",
    delegates
  );

end;
--------------------------------------------------------------------------------
function private.achievements ()

  local func_validate = function ( achievement_data )
    return ( achievement_data.name == "only_forward" );
  end;

  local delegates = {
    DisableScrolling = function ()
      v_player.DialogMain_DisableAchievementsScrolling();
    end,
    EnableScrolling = function ()
      v_player.DialogMain_EnableAchievementsScrolling();
    end
  };

  private.Tutorial_SetOnEvent(
    "achievements",
    "Model_Achievements_Completed",
    delegates,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.questlist ()

  local func_validate = function ( stage_name )
    local  is_valid = private.Tutorial_IsStageCompleted( "stage_8" );
    return is_valid;
  end;

  local delegates = {
    GetButtonName = function ()
      local  button_name = v_interface.panel.right.QuestsButtonName();
      return button_name;
    end
  };

  private.Tutorial_SetOnEvent(
    "questlist",
    "View_Map_Update_Finished",
    delegates,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.questlist_quest ()

  local func_validate = function ( quests_list )
    return ( #quests_list > 0 );
  end;

  local delegates = {
    GetScrollerName = function ()
      local  scc_name = v_quests.Questlist_GetScrollerName();
      return scc_name;
    end
  };

  private.Tutorial_SetOnEvent(
    "questlist_quest",
    "View_Questlist_Opened",
    delegates,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.stage_9 ()

  local func_validate = function ( stage_name )
    local is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_9"
    );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "stage_9",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.magicgift ()

  private.Subscribe( "magicgift", "View_Stage_DialogStartActivated", private.magicgift_first );

end;

function private.magicgift_first ( stage_data )

  local stage_name = stage_data.name;

  if ( not m_tutorial.IsCompleted( "magicgift" ) ) then

    if ( stage_name == "stage_10" ) then

      local index = private.Entity_Index( stage_data.boosters, "magicgift", "start" );

      private.Unsubscribe( "magicgift", "View_Stage_DialogStartActivated", private.magicgift_first );
      
      private.Entity_Add( "magicgift", "magicgift", "booster" );
      private.Entity_Free_Show( "magicgift", "magicgift", "btn_int_dialog_stage_start_boosters_"..index );

      private.magicgift_next1( stage_data.boosters );

    end;

  end;

end;

function private.magicgift_next1 ( boosters )
  
  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_magicgift"
  } );

  local index = private.Entity_Index( boosters, "magicgift", "start" );

  local dialog_settings =
    private.Dialog_Settings( "btn_int_dialog_stage_start_boosters_"..index );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type        = "arrow",

    focus       = "btn_int_dialog_stage_start_boosters_"..index,
    offset_x    = 0,
    offset_y    = - 50,
    angle       = 0,
    scale       = 1,

    lock        = false
  };

  local func_close = function ()

    private.Dialog_Hide();

    private.Entity_Free_Hide( "btn_int_dialog_stage_start_boosters_"..index );

  end;

  private.Dialog_ShowInstantly( "magicgift", positions, hole, func_close );

  private.magicgift_last1();

end;

function private.magicgift_last1 ()

  private.Complete( "magicgift" );

end;
--------------------------------------------------------------------------------
function private.skin ()

  local func_validate = function ( skin_name )
    local  skin_data = v_player.Skin_Data( skin_name );
    local  is_valid  = skin_data.wearable;
    return is_valid;
  end;

  local delegates = {
    IsSkinActive = function ( skin_name )
      local  skin_data = v_player.Skin_Data( skin_name );
      return skin_data.weared;
    end,
    IsSkinWithoutParts = function ( skin_name )
      local  skin_data = v_player.Skin_Data( skin_name );
      return ( ( not skin_data.parts ) or ( skin_data.parts < 1 ) );
    end,
    FocusOnSkin = function ( skin_name, func_end )
      v_player.DialogMain_FocusOnSkin( skin_name, func_end );
    end,
    IsSelectButtonAvailable = function ()
      local  is_available = v_player.DialogMain_HasSkinSelectButton();
      return is_available;
    end
  };

  private.Tutorial_SetOnEvent(
    "skin",
    "Model_Skins_SkinCollected",
    delegates
  );
  
end;
--------------------------------------------------------------------------------
function private.build ()

  private.Subscribe(
    "build",
    "View_Tutorial_MapOpened",
    private.build_first
  );

end;

function private.build_first ()

  if ( not m_tutorial.IsCompleted( "build" ) ) then

    if ( ( v_stage.Stage_StorageData( "stage_11" ) or {} ).completed ) then

      private.Unsubscribe(
        "build",
        "View_Tutorial_MapOpened",
        private.build_first
      );
      
      private.Subscribe(
        "build",
        "View_Map_Update_Finished",
        private.build_next1
      );

    end;

  end;

end;

function private.build_next1 ()

  private.Unsubscribe(
    "build",
    "View_Map_Update_Finished",
    private.build_next1
  );
  
  private.PendingBegin( "build", "map" );

  local func_close = function ()

    private.Dialog_Hide( private.build_last1 );

  end;

  local dialog_settings =
    private.Dialog_Settings( "spr_map_build_indicator_build_1" );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type        = "flashlight",  

    res_id      = 3,
    focus       = "spr_map_build_indicator_build_1",
    scale       = 1.25,
  
    lock        = true,
    black       = true,
    hand        = {
      angle     = 0.5 * math.pi,
      offset_x  = 0,
      offset_y  = 0,
      type      = "click"
    }
  };
  
  local events        = ui.control.EventsGet( "obj_map_builds_build_1" );
  local func_on_click = events.func_on_click;
  
  events.func_on_click = function ()
  
    func_close();
    func_on_click();
    events.func_on_click = func_on_click;
  
  end;
  ui.control.EventsSet( "obj_map_builds_indicator_build_1", events );

  local map_params =
  {
    notify = false,
    input  = true,
    active = true
  }; 

  v_map.Focus( "obj_map_builds_build_1",
    { 
      scale      = 1.0,
      -- time       = 0.6,
      func_after = function ()
      
        private.Dialog_ShowInstantly(
          "build",
          positions,
          hole,
          func_close,
          nil,
          map_params
        );
      
      end;
    }
  );

end;

function private.build_last1 ()

  private.Complete( "build" );
  private.PendingEnd( "build", "map" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.mine ()

  private.Subscribe( "mine", "View_Tutorial_MapOpened", private.mine_first );

end;

function private.mine_first ()

  if ( not m_tutorial.IsCompleted( "mine" ) ) then

    if ( ( v_stage.Stage_StorageData( "stage_12" ) or {} ).completed ) then

      private.Unsubscribe( "mine", "View_Tutorial_MapOpened", private.mine_first ); 

      private.Subscribe( "mine", "View_Map_Update_Mine_Animated", private.mine_next1 );

    end;

  end;

end;

function private.mine_next1 ()

  private.Unsubscribe( "mine", "View_Map_Update_Mine_Animated", private.mine_next1 );
  
  private.PendingBegin( "mine", "map" );
  
  local events = ui.control.EventsGet( "btn_map_mines_mine_1" );
  local func_on_click = events.func_on_click; -- backup

  local func_close = function ( click )
  
    events.func_on_click = func_on_click; -- restore
  
    private.Dialog_Hide( private.mine_last1 );
    if ( click ) then func_on_click(); end;
    
  end;
  
  events.func_on_click = function () func_close( true ); end;
  
  local dialog_settings = private.Dialog_Settings( "obj_map_mines_mine_1" );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type        = "arrow",

    focus       = "obj_map_mines_mine_1",
    offset_x    = 0,
    offset_y    = - 40,
    angle       = 0,
    scale       = 1.0,

    lock        = true
  };

  private.Dialog_ShowInstantly( 
    "mine", positions, hole, func_close, nil, { active = true }
  );

end;

function private.mine_last1 ()

  private.Complete( "mine" );
  private.PendingEnd( "mine", "map" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stage_14 ()

  private.Subscribe(
    "stage_14",
    "View_Tutorial_StageOpened",
    private.stage_14_first
  );

end;

function private.stage_14_first ( stage_name )

  if ( stage_name == "stage_14"                 ) and
     ( not m_tutorial.IsCompleted( "stage_14" ) ) then
    
    private.Unsubscribe(
      "stage_14",
      "View_Tutorial_StageOpened",
      private.stage_14_first
    ); 
    
    private.Subscribe(
      "stage_14",
      "View_Stage_DialogTargetsHided",
      private.stage_14_next1
    );

  end;

end;

function private.stage_14_next1 ()
  
  private.Unsubscribe( "stage_14", "View_Stage_DialogTargetsHided", private.stage_14_next1 );

  local items_data = v_stage.Stage_GetCurrentData().items; 

  local match3_pos = GetObjPosByObj( "match" );

  local pos_x = match3_pos[ 1 ] + 2 * private.cell_size * items_data[ 1 ].col - private.cell_size;
  local pos_y = match3_pos[ 2 ] + 2 * private.cell_size * items_data[ 1 ].row + private.cell_size;

  local pref_vertical = 12.5;
  
  if ( private.landscape ) then 
    
    pref_vertical = nil;
    
  end;
  
  local dialog_settings =
    private.Dialog_Settings( { pos_x, pos_y }, pref_vertical );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale,
    character_pos_y = dialog_settings.ch_pos_y
  };
  
  local hole = 
  {
    type        = "arrow",

    pos_x       = pos_x,
    pos_y       = pos_y,
    scale       = 1,
    angle       = 0,
    length      = 460,

    lock        = true
  };

  local func_close = function ()

    private.Dialog_Hide( private.stage_14_last1 );
    
  end;

  private.Dialog_ShowInstantly( "stage_14_1", positions, hole, func_close );

end;

function private.stage_14_last1 ()

  private.Complete( "stage_14" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.booster_shop ()
  
  private.Subscribe(
    "booster_shop",
    "View_Tutorial_MapOpened",
    private.booster_shop_first
  );
  
end;

function private.booster_shop_first ()
  if ( ( v_stage.Stage_StorageData( "stage_14" ) or {} ).completed ) and
     ( not m_tutorial.IsCompleted( "booster_shop" ) ) then
    private.Unsubscribe( "booster_shop", "View_Tutorial_MapOpened",
                         private.booster_shop_first );

    private.Subscribe( "booster_shop", "View_Map_Update_Finished",
                        private.booster_shop_next1 );
   end;
end;
                     
function private.booster_shop_next1 ()

  private.Unsubscribe( "booster_shop", "View_Map_Update_Finished",
                         private.booster_shop_next1 );
                         
  private.PendingBegin( "booster_shop", "map" );
  
  local focus_obj = "btn_int_panel_interface_bottom_shop";

  local dialog_settings = private.Dialog_Settings( focus_obj );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type  = "flashlight",
    focus = focus_obj,
    
    res_id   = 3,
    offset_x = 0,
    offset_y = -70,
    angle    = 0,
    scale    = 1.25,

    lock     = true,
    black    = true,
    wait     = true,
    hand = {  
         offset_x  = 0,
         offset_y  = 0,
         scale_x   = - 1,
         angle     = - 0.5 * math.pi, 
         type      = "click"
    }    
  };
  
  local func_close = function ()
    
    local func_shop_tap = private.Restore( "booster_shop", "booster_shop_tap" );
    ui.control.EventsSet( focus_obj, { func_on_click = func_shop_tap } );
    
    private.Dialog_Hide( function()
      
      private.Complete( "booster_shop" );
      private.PendingEnd( "booster_shop", "map" );
      
      if ( func_shop_tap ) then
        int_dialog_shop.SetForcedTab( "boosters" );
        func_shop_tap();
      end;
    
    end );
      
  end;
  
  private.Backup(
    "booster_shop",
    "booster_shop_tap",
    ui.control.EventsGet( focus_obj ).func_on_click
  );

  ui.control.EventsSet( focus_obj, { func_on_click = func_close } );

  private.Dialog_ShowInstantly( "booster_shop_1", positions, hole, nil, nil,
                                { active = true, notify = false } );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.booster_buy () 
  
  v_shop.Notify( true );
  
  private.Subscribe(
    "booster_buy",
    "View_Shop_TabOpened",
    private.booster_buy_first
  ); 
  
end;


function private.booster_buy_first ( name )
    
  if ( name ~= "boosters"                           ) or
     ( m_tutorial.IsCompleted( "booster_buy" )      ) or 
     ( not m_tutorial.IsCompleted( "booster_shop" ) ) then
    
    return;
    
  end;

  private.Unsubscribe(
    "booster_buy",
    "View_Shop_TabOpened",
    private.booster_buy_first
  );

  private.booster_buy_next();

end;

function private.booster_buy_next ()

  -- JK: Workaround, as normal focusing doesn't work
  local focus_obj = "btn_int_dialog_shop_close";
  
  local v_pref = 50;
  
  local shift_x, shift_y = -285, 89;
  local scale_x, scale_y = 0.98, 0.65;
  
  if ( private.landscape ) then
    
    shift_x, shift_y = -220, 70;
    scale_x, scale_y = 0.76, 0.5;
    
    v_pref = 70;
 
  end;
  
  local dialog_settings = private.Dialog_Settings( focus_obj, v_pref );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type     = "flashlight",
    focus    = focus_obj,
    res_id   = 8,
    offset_x = shift_x,
    offset_y = shift_y,
    angle    = 0,
    scale_x  = scale_x,
    scale_y  = scale_y,
    lock     = true,
    black    = true,
    input    = true
  };

  local func_close = function ()

    private.Dialog_Hide( private.booster_buy_last );

  end;
  
  private.Dialog_ShowInstantly( "booster_buy_1", positions, hole, func_close );
  
end;

function private.booster_buy_last ()
  
  if ( not m_tutorial.IsCompleted( "booster_shop" ) ) then
    
    private.Complete( "booster_shop" );
    
  end;
  
  private.Complete( "booster_buy" );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stage_16_tap ()

  private.Subscribe( "stage_16_tap", "View_Stage_DialogStartActivated", private.stage_16_tap_last );

end;

function private.stage_16_tap_last ( stage_data )

  if ( stage_data.name == "stage_16" and not m_tutorial.IsCompleted( "stage_16_tap" ) ) then

    private.Unsubscribe( "stage_16_tap", "View_Stage_DialogStartActivated", private.stage_16_tap_last ); 

    private.Boosters_Backup( "stage_16_tap", "", false, "btn_int_dialog_stage_start_boosters" );
    
  end;    

end;
--------------------------------------------------------------------------------
function private.stage_16 ()

  local func_validate = function ( stage_name )
    local is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_16"
    );
    return is_valid;
  end;

  local delegates = {
    AnimateSorcery = function ()
      v_stage.PanelBoosters_StartAnim( "sorcery" );
    end,
    GetSorceryIndex = function ()
      local index = private.Entity_Index( 
        v_stage.Stage_GetCurrentData().boosters,
        "sorcery",
        "stage"
      );
      return index;
    end,
    AddEntity      = private.Entity_Add,
    GetEntityIndex = private.Entity_Index,
    ShowEntity     = private.Entity_Free_Show,
    HideEntity     = private.Entity_Free_Hide,
    BackupBooster  = private.Boosters_Backup,
    RestoreBooster = private.Boosters_Restore,
    CompleteOther  = function ( tutorial_name )
      private.Complete( tutorial_name );
    end,
    GetStageItems = function ()
      local  items = v_stage.Stage_GetCurrentData().items;
      return items;
    end
  };

  private.Tutorial_SetOnEvent(
    "stage_16",
    "View_Tutorial_StageOpened",
    delegates,
    func_validate
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.spellbook ()

  local func_validate = function ()
    local  is_valid = private.Tutorial_IsStageCompleted( "stage_18" );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "spellbook",
    "View_Spells_SpellbookAppear",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.heavenhelp ()

  local func_validate = function ( stage_data )
    return ( stage_data.name == "stage_20" );
  end;

  local delegates = {
    GetHeavenhelpIndex = function ( stage_data )
      local index = private.Entity_Index(
        stage_data.boosters, "heavenhelp", "start"
      );
      return index;
    end,
    AddEntity  = private.Entity_Add,
    ShowEntity = private.Entity_Free_Show,
    HideEntity = private.Entity_Free_Hide
  };

  private.Tutorial_SetOnEvent(
    "heavenhelp",
    "View_Stage_DialogStartActivated",
    delegates,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.mine_construct ()
  
  local mine_data   = v_mines.Mine_Data( "mine_1" );
  local event_name  = "Model_Mines_MineWaiting";
  local start_point = "first";

  local func_validate = function ( mine_name )
    return ( mine_name == "mine_1" );
  end;

  if ( mine_data.state == "ready" ) then

    func_validate = function ( section_number )
      return ( section == 1 );
    end;

    start_point = "next1";
    event_name  = "View_Map_Section_Switched";

  end;

  local delegates = {
    GetStartPoint = function () return start_point; end
  };

  private.Tutorial_SetOnEvent(
    "mine_construct",
    event_name,
    delegates,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.stage_26 ()

  local func_validate = function ( stage_name )
    return ( stage_name == "stage_26" );
  end;

  local delegates = {
    GetCurrentStageData = function ()
      local  stage_data = v_stage.Stage_GetCurrentData();
      return stage_data;
    end,
    AddEntity           = private.Entity_Add,
    GetEntityIndex      = private.Entity_Index,
    ShowEntity          = private.Entity_Free_Show,
    HideEntity          = private.Entity_Free_Hide,
    BackupBooster       = private.Boosters_Backup,
    RestoreBooster      = private.Boosters_Restore,
    AnimateBoosterPanel = function ( arg )
      v_stage.PanelBoosters_StartAnim( arg );
    end,
    OnStageAbandoned = function ()

      private.PendingEnd( "stage_26", "stage" );
      private.stage_26();

    end,
    SetPauseButtonInput = function ( value )
      v_stage.PanelSettings_SetInput ( value );
    end
  };

  private.Tutorial_SetOnEvent(
    "stage_26",
    "View_Tutorial_StageOpened",
    delegates,
    func_validate
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.nolimits ()

  private.Subscribe(
    "nolimits",
    "View_Stage_DialogStartActivated",
    private.nolimits_first
  );

end;

function private.nolimits_first ( stage_data )

  local stage_name = stage_data.name;
  local completed  = m_tutorial.IsCompleted( "nolimits" );

  if ( stage_name == "stage_37" and not completed ) then

    private.Unsubscribe(
      "nolimits",
      "View_Stage_DialogStartActivated",
      private.nolimits_first
    );

    private.nolimits_point_to( stage_data.boosters );

  end;

end;

function private.nolimits_point_to( boosters )

  local index = private.Entity_Index( boosters, "nolimits", "start" );
  
  private.Entity_Add( "nolimits", "nolimits", "booster" );
  private.Entity_Free_Show(
    "nolimits",
    "nolimits",
    "btn_int_dialog_stage_start_boosters_"..index
  );

  local dialog_settings =
    private.Dialog_Settings( "btn_int_dialog_stage_start_boosters_"..index );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type        = "arrow",

    focus       = "btn_int_dialog_stage_start_boosters_"..index,
    offset_x    = 0,
    offset_y    = - 50,
    angle       = 0,
    scale       = 1,

    lock        = false
  };

  local func_close = function ()

    private.Dialog_Hide();
    private.Entity_Free_Hide( "btn_int_dialog_stage_start_boosters_"..index );

  end;

  private.Dialog_ShowInstantly( "nolimits", positions, hole, func_close );
  private.nolimits_complete();

end;

function private.nolimits_complete ()

  private.Complete( "nolimits" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stage_44 ()

  local func_validate = function ( stage_name )
    local is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_44"
    );
    return is_valid;
  end;

  local delegates = {
    AddEntity  = private.Entity_Add,
    ShowEntity = private.Entity_Free_Show,
    HideEntity = private.Entity_Free_Hide,
    GetCascadeIndex     = function ()
      local index = private.Entity_Index(
        v_stage.Stage_GetCurrentData().boosters, "cascade", "stage"
      );
      return index;
    end
  };

  private.Tutorial_SetOnEvent(
    "stage_44",
    "View_Tutorial_StageOpened",
    delegates,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.stage_45 ()

  local func_validate = function ( stage_name )
    local is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_45"
    );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "stage_45",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.market_construct ()

  if ( v_sections.Section_NameCurrent() ~= "section_1" ) or 
     ( v_shop.Market_IsBuilt()                         ) then
    private.Complete( "market_construct" );
    return;    
  end;

  local delegates = {
    IsDialogShown = function ()
      local  is_shown =
        v_interface.window.IsOnScreen( "int_dialog_map_build_market" );
      return is_shown;
    end
  };

  if ( v_shop.Market_IsUnlocked() ) then
    private.Tutorial_SetOnEvent(
      "market_construct",
      "View_Map_MapOpened",
      delegates
    );
  else
    private.Tutorial_SetOnEvent(
      "market_construct",
      m_shop.msg.MarketUnlocked,
      delegates
    );
  end;

end;
--------------------------------------------------------------------------------
function private.market ()
  
  local delegates = {
    IsSellMarketDisabled = function ()
      local is_disabled = v_shop.Market_IsSellDisabled();
      return is_disabled;
    end
  };
  
  v_shop.Notify( true );
  
  private.Tutorial_SetOnEvent(
    "market",
    "View_Shop_TabOpened",
    delegates
  );
  
end;
--------------------------------------------------------------------------------
function private.stage_69 ()

  local func_validate = function ( stage_name )
    local is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_69"
    );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "stage_69",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.downright ()

  local func_validate = function ( stage_data )
    return ( stage_data.name == "stage_76" );
  end;

  local delegates = {
    AddEntity  = private.Entity_Add,
    GetDownrightIndex = function ( stage_data )
      local index = private.Entity_Index(
        stage_data.boosters, "downright", "start"
      );
      return index;
    end,
    ShowEntity = private.Entity_Free_Show,
    HideEntity = private.Entity_Free_Hide
  };

  private.Tutorial_SetOnEvent(
    "downright",
    "View_Stage_DialogStartActivated",
    delegates,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.stage_79 ()

  local func_validate = function ( stage_name )
    local is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_79"
    );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "stage_79",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.stage_80 ()

  local func_validate = function ( stage_name )
    local is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_80"
    );
    return is_valid;
  end;

  local delegates = {
    AddEntity  = private.Entity_Add,
    GetFireburstIndex = function ()
      local index = private.Entity_Index(
        v_stage.Stage_GetCurrentData().boosters, "fireburst", "stage"
      );
      return index;
    end,
    ShowEntity = private.Entity_Free_Show,
    HideEntity = private.Entity_Free_Hide
  };

  private.Tutorial_SetOnEvent(
    "stage_80",
    "View_Tutorial_StageOpened",
    delegates,
    func_validate
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.dragonspeech ()

  private.Subscribe( "dragonspeech", "View_Map_ZoneAppear", private.dragonspeech_first );

end;

function private.dragonspeech_first ( params )

  if ( params.build_name == "puzzle_1" ) then

    private.Unsubscribe( "dragonspeech", "View_Map_ZoneAppear", private.dragonspeech_first );

    if ( not m_tutorial.IsCompleted( "dragonspeech" ) ) then
    
      private.Subscribe(
        "dragonspeech",
        "View_Map_Update_CompletedStageAnimated",
        private.dragonspeech_next1
      );

    end;

  end;

end;

function private.dragonspeech_next1 ()

  private.Unsubscribe(
    "dragonspeech",
    "View_Map_Update_CompletedStageAnimated",
    private.dragonspeech_next1
  );
  
  private.PendingBegin( "dragonspeech", "map" );
  
  v_map.Focus(
    "obj_map_enemy_reddragon_stage_110",
    {
      scale = 1.0,
      time = 0.6,
      func_after = function ()
      
        ui.popup.Cloud_Show( 
          "monolog", 
          "obj_map_enemy_reddragon_stage_110", 
          "str_tutorial_dragonspeech",
          nil,
          { -50, -130 },
          nil,
          { 
            func_preclose = function ()
              private.PendingEnd( "dragonspeech", "map" );
            end
          }
        );
      
      end
    }
  );

  private.dragonspeech_last1();

end;

function private.dragonspeech_last1 ()

  private.Complete( "dragonspeech" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.worldmap ()
  
  private.Tutorial_SetOnEvent(
    "worldmap",
    "View_Map_Update_PortalUnlocked"
  );

end;
--------------------------------------------------------------------------------
function private.stage_111 ()

  local func_validate = function ( stage_name )
    local is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_111"
    );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "stage_111",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.carrier ()

  private.Subscribe(
    "carrier",
    "Model_Stage_StageWon",
    private.carrier_first
  );

end;

function private.carrier_first ()

  if ( not m_tutorial.IsCompleted( "carrier" ) ) then

    if ( ( v_stage.Stage_StorageData( "stage_121" ) or {} ).completed ) then

      private.Unsubscribe(
        "carrier",
        "Model_Stage_StageWon",
        private.carrier_first
      );

      if ( v_carriers.IsAvailable() ) and
         ( not v_carriers.IsBuilt() ) then

        private.Subscribe(
          "carrier",
          "View_Map_Update_ChangesDone",
          private.carrier_begin
        );

      else

        private.carrier_last( true );

      end;

    end;

  end;

end;

function private.carrier_begin ()

  private.PendingBegin( "carrier", "map" );

  ui.screen.Lock( true );

  private.Unsubscribe(
    "carrier",
    "View_Map_Update_ChangesDone",
    private.carrier_begin
  );

  local carrier_obj = v_carriers.GetCurrentMapObjectName();

  v_map.Focus(
    carrier_obj,
    {
      func_after = function () private.carrier_next1( carrier_obj ); end
    }
  );

end;

function private.carrier_next1 ( carrier_obj )  

  local events = ui.control.EventsGet( carrier_obj );
  local func_on_click = events.func_on_click; -- backup

  local func_close = function ( click )
  
    events.func_on_click = func_on_click; -- restore
  
    private.Dialog_Hide( private.carrier_next2 );
    if ( click ) then func_on_click(); end;
    
  end;
  
  events.func_on_click = function () func_close( true ); end;
  
  local dialog_settings = private.Dialog_Settings( carrier_obj );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type   = "flashlight",

    res_id = 3,
    focus  = carrier_obj,
    scale  = 1.5,
  
    lock   = true,
    black  = true,
    hand   = {
      angle = 0.5 * math.pi,
      type  = "click"
    }
  };

  private.Dialog_ShowInstantly( 
    "carrier_1", positions, hole, func_close, nil, { active = true, input = true }
  );

  ui.screen.Lock( false );

end;

function private.carrier_next2 ()

  local btn_build       = v_carriers.Dialog_GetBuildButtonName();
  local v_pref          = tern( private.landscape )( 25, 63 );
  local h_pref          = tern( private.landscape )( 75, nil );
  local dialog_settings = private.Dialog_Settings( btn_build, v_pref, h_pref );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type     = "arrow",

    focus    = btn_build,
    offset_x = -40,
    offset_y = -40,
    angle    = 1,
    scale    = 1.0,

    lock     = true
  };

  local func_close = function ()
  
    private.Dialog_Hide( private.carrier_last );
    
  end;

  private.Dialog_ShowInstantly( 
    "carrier_2", positions, hole, func_close
  );

end;

function private.carrier_last ( ignore_pending )

  private.Complete( "carrier" );

  if ( not ignore_pending ) then

    private.PendingEnd( "carrier", "map" );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stage_125_tap ()

  private.Subscribe( "stage_125_tap", "View_Stage_DialogStartActivated", private.stage_125_tap_last );

end;

function private.stage_125_tap_last ( stage_data )

  if ( stage_data.name == "stage_125" and not m_tutorial.IsCompleted( "stage_125_tap" ) ) then

    private.Unsubscribe( "stage_125_tap", "View_Stage_DialogStartActivated", private.stage_125_tap_last ); 

    private.Boosters_Backup( "stage_125_tap", "", false, "btn_int_dialog_stage_start_boosters" );
    
  end;    

end;
--------------------------------------------------------------------------------
function private.stage_125 ()

  local func_validate = function ( stage_name )

    local is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_125"
    );

    return is_valid;

  end;

  local delegates = {
    RestoreBooster = private.Boosters_Restore,
    CompleteOther  = function ( tutorial_name )

      private.Complete( tutorial_name );

    end
  };

  private.Tutorial_SetOnEvent(
    "stage_125",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stage_221 ()
  
  local func_validate = function ( stage_name )
    local is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_221"
    );
    return is_valid;
  end;

  local delegates = {
    AddEntity  = private.Entity_Add,
    GetFireburstIndex = function ()
      local index = private.Entity_Index(
        v_stage.Stage_GetCurrentData().boosters, "fireburst", "stage"
      );
      return index;
    end,
    ShowEntity = private.Entity_Free_Show,
    HideEntity = private.Entity_Free_Hide
  };

  private.Tutorial_SetOnEvent(
    "stage_221",
    "View_Tutorial_StageOpened",
    delegates,
    func_validate
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.count_bomb ()
  
  private.Subscribe(
    "count_bomb",
    "View_Tutorial_StageOpened",
    private.count_bomb_start
  );

end;

function private.count_bomb_start ( stage_name )

  if ( stage_name == "stage_230"                  ) and
     ( not m_tutorial.IsCompleted( "count_bomb" ) ) then

    private.Unsubscribe(
      "count_bomb",
      "View_Tutorial_StageOpened",
      private.count_bomb_start
    ); 
  
    private.Subscribe(
      "count_bomb",
      "View_Stage_DialogTargetsHided",
      private.count_bomb_point_to
    );

  end;

end;

function private.count_bomb_point_to ()

  private.Unsubscribe(
    "count_bomb",
    "View_Stage_DialogTargetsHided",
    private.count_bomb_point_to
  );

  local match3_pos = GetObjPosByObj( "match" );

  local pos_x = match3_pos[ 1 ] + 2 * private.cell_size * 4 - private.cell_size;
  local pos_y = match3_pos[ 2 ] + 2 * private.cell_size * 4;

  local dialog_settings = private.Dialog_Settings( { pos_x, pos_y } );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type           = "arrow",

    pos_x          = pos_x,
    pos_y          = pos_y,
    angle          = 0,
    scale          = private.cell_scale,

    lock           = true
  };

  local func_close = function ()

    private.Dialog_Hide( private.count_bomb_show_match );

  end;

  private.Dialog_ShowInstantly( "count_bomb_1", positions, hole, func_close );

end;

function private.count_bomb_show_match ()

  local match3_pos = GetObjPosByObj( "match" );

  local pos_x = match3_pos[ 1 ] + 2 * private.cell_size * 3 - private.cell_size;
  local pos_y = match3_pos[ 2 ] + 2 * private.cell_size * 5;

  local prefs = {
    portrait  = { nil, nil },
    landscape = { 30, 70 }
  };
  
  local orientation = ui.screen.Orientation();
  
  local dialog_settings =
    private.Dialog_Settings(
      { pos_x, pos_y },
      prefs[ orientation ][ 1 ],
      prefs[ orientation ][ 2 ]
    );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type           = "flashlight",

    res_id         = 1,
    pos_x          = pos_x,
    pos_y          = pos_y,
    angle          = - 0.5 * math.pi,
    scale          = private.cell_scale,

    lock           = true,
    hand           = {
         angle     = 0.5 * math.pi,
         scale_y   = - 1, 
         offset_x  = 0,
         offset_y  = 0,
         type      = "move"
    }
  };

  local function func_close ()

    private.Unsubscribe( "count_bomb",
                         "Model_Match3_FieldActivated", func_close );

    private.Dialog_Hide( private.count_bomb_complete );

  end;

  private.Subscribe( "count_bomb", "Model_Match3_FieldActivated", func_close );
  private.Dialog_ShowInstantly( "count_bomb_2", positions, hole );

end;

function private.count_bomb_complete ()

  private.Complete( "count_bomb" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.dailyquest ()

  local dialog_name = "int_dialog_dailyquest";

  local func_validate = function ()
    local  is_valid = private.Tutorial_IsStageCompleted( "stage_6" );
    return is_valid;
  end;

  local delegates = {
    IsDailyquestCountdown = function ()
      local  is_countdown = ( v_dailyquest.Countdown_IsNeeded() ) or
                            ( v_dailyquest.Countdown_IsActive() );
      return is_countdown;
    end,
    IsDailyquestReady = function ()
      local  is_ready = v_dailyquest.IsReady();
      return is_ready;
    end,
    IsDialogOnScreen = function ()
      local  is_on_screen = v_interface.window.IsOnScreen( dialog_name );
      return is_on_screen;
    end,
    ClearOuterTap = function ()
      v_interface.window.ClearOuterTap();
    end,
    SetOuterTap = function ()
      v_interface.window.SetOuterTap( dialog_name );
    end
  };

  private.Tutorial_SetOnEvent(
    "dailyquest",
    "View_Map_Update_Finished",
    delegates,
    func_validate
  );
  
end;
--------------------------------------------------------------------------------
function private.dailyquest_win ()

  private.Subscribe( "dailyquest_win", "View_DailyQuest_RewardShown", private.dailyquest_win_first );

end;

function private.dailyquest_win_first ( challenge_num )

  private.Unsubscribe( "dailyquest_win", "View_DailyQuest_RewardShown", private.dailyquest_win_first );

  if not ( m_tutorial.IsCompleted( "dailyquest_win" ) ) then

    private.dailyquest_win_next1( challenge_num );

  end;

end;

function private.dailyquest_win_next1 ( challenge_num )

  local positions = 
  {
    dlg_start   = 1200,
  };

  local challenge_pos = ui.obj.Position(
    "anm_int_dialog_dailyquest_challenge_"..challenge_num,
    "anm_int_dialog_dailyquest_window"
  );
  
  local arrow_pos  = { challenge_pos[ 1 ], challenge_pos[ 2 ] - 50 };
  local hole = 
  {
    type     = "arrow",
    focus    = "obj_int_dialog_dailyquest_challenge_"..challenge_num.."_reward",
    offset_x = arrow_pos[ 1 ],
    offset_y = arrow_pos[ 2 ],
    scale    = 0.8,
    angle    = 0,
    lock     = false
  };

  if ( challenge_num == 1 ) or ( challenge_num == 2 ) then

    positions.dlg_pos_y = private.app_h / 2 + 230;

  else

    positions.dlg_pos_y = private.app_h / 2 + 50;

  end;
    
  if ( challenge_num == 2 ) then

    positions.character_start = -135;
    positions.character_end   = -135;

  else

    positions.character_start = 135;
    positions.character_end   = 135;

  end;

  local func_close = function ()

    private.Dialog_Hide( private.dailyquest_win_last1 );

  end;

  private.Dialog_ShowInstantly( "dailyquest_win_1", positions, hole, func_close );

  private.Complete( "dailyquest_win" );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stickers_duplicates ()

  if not m_tutorial.IsCompleted( "stickers_duplicates" ) then
    
    private.Subscribe(
      "stickers_duplicates",
      "Interface_DialogFireflies_ButtonSellShown",
      private.stickers_duplicates_first
    );
    
  end;

end;

function private.stickers_duplicates_first ( btn_name )
  
  private.Unsubscribe(
    "stickers_duplicates",
    "Interface_DialogFireflies_ButtonSellShown",
    private.stickers_duplicates_first
  );
  
  local pos_btn = ui.obj.Position( btn_name );

  local shift_x, shift_y = 0, -215;
  
  if ( private.landscape ) then
    
    shift_x = 46;
    shift_y = -165;
 
  end;
  
  local coords   = pos_btn;

  local coords   = {
    pos_btn[1] + shift_x,
    pos_btn[2] + shift_y
  };

  local pref_v = nil;
  
  if ( not private.landscape ) then
    
    pref_v = 65;
 
  end;
  
  local dialog_settings = private.Dialog_Settings( coords, pref_v );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type  = "arrow",
    
    pos_x = coords[ 1 ],
    pos_y = coords[ 2 ],
    angle    = 0.0,
    scale    = 1.0, 

    lock     = false
  };
  
  private.Dialog_ShowInstantly( "stickers_duplicates_1", positions, hole, function()
 
    private.Dialog_Hide( private.stickers_duplicates_last1 );
 
  end );
  
end;

function private.stickers_duplicates_last1 ()
  
  private.Complete( "stickers_duplicates" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.bonuspopup ()
  
  private.Subscribe( "bonuspopup", "View_Stage_PopupBonus_Activated", private.bonuspopup_first );

end;

function private.bonuspopup_first ()

  private.Unsubscribe( "bonuspopup", "View_Stage_PopupBonus_Activated", private.bonuspopup_first );
  
  local stage_name = v_stage.Stage_GetCurrentData().name;
  
  if ( stage_name == "stage_6" ) then
    
    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_level_6_damage"
    } );   
    
  end;

  local dialog_settings =
    private.Dialog_Settings( "obj_int_popup_stage_bonus" );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type        = "arrow",

    focus       = "obj_int_popup_stage_bonus",
    offset_y    = 50,
    scale       = 1,
    angle       = 3.1415,

    lock = true
  };

  local func_close = function ()

    private.Dialog_Hide( private.bonuspopup_last );
    
  end;

  private.Dialog_ShowInstantly( "bonuspopup", positions, hole, func_close );
  
  private.Complete( "bonuspopup" );

end;

function private.bonuspopup_last ()

  --
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.chest ()

  private.Subscribe( "chest", "View_Chests_ChestCollected", private.chest_last );

end;

function private.chest_last ( chest_name )

  if ( chest_name == "chest_1" and not m_tutorial.IsCompleted( "chest" ) ) then

    private.Unsubscribe( "chest", "View_Chests_ChestCollected", private.chest_last ); 

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_present"
    } );  

    private.Complete( "chest" );

  end;    

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.fortunewheel ()

  private.Subscribe(
    "fortunewheel",
    "View_Stage_FortuneWheel_Opened",
    private.fortunewheel_first
  );

  if ( ui.obj.Exists( "btn_int_dialog_stage_continue_video" ) ) then
    ne.obj.Detach( "btn_int_dialog_stage_continue_video" );
  end;
  
end;

function private.fortunewheel_first ()

  private.Unsubscribe(
    "fortunewheel",
    "View_Stage_FortuneWheel_Opened",
    private.fortunewheel_first
  );
  
  if ( not m_tutorial.IsCompleted( "fortunewheel" ) ) then

    private.fortunewheel_next1();

  end;

end;

function private.fortunewheel_next1 ()
  
  if ( v_stage.FortuneWheel_IsOnCooldown() ) then

    private.Complete( "fortunewheel" );
    return;

  end;

  v_stage.FortuneWheel_SetPopupEnabled( false );

  local focus_object    = "anm_int_dialog_stage_fortunewheel_body";
  local dialog_settings = private.Dialog_Settings( focus_object );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type        = "flashlight",  

    res_id      = 6,
    angle       = 1.575707,
    scale       = 1.0,
  
    lock        = true,
    black       = false,
    focus       = focus_object,
    hand        = {
      angle     = 0.5 * math.pi,
      offset_x  = 0,
      offset_y  = 0,
      type      = "click"
    }
  };

  local backup_control = "btn_int_dialog_stage_fortunewheel_open";

  local func_close = function ()
    
    private.Subscribe(
      "fortunewheel",
      "View_Stage_FortuneWheel_ReadyToSpin",
      private.fortunewheel_spin
    );
                     
    private.Dialog_Hide();

    local func_fortunewheel_tap = private.Restore( "fortunewheel", "fortunewheel_tap" );

    ui.control.EventsSet(
      backup_control,
      { func_on_click = func_fortunewheel_tap }
    );

    if ( func_fortunewheel_tap ) then

      func_fortunewheel_tap();

    end;

  end;

  private.Backup(
    "fortunewheel",
    "fortunewheel_tap",
    ui.control.EventsGet( backup_control ).func_on_click 
  );

  ui.control.EventsSet( backup_control, { func_on_click = func_close } );

  private.Dialog_ShowInstantly( "fortunewheel", positions, hole );
  
  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_wheel"
  } );

end;

function private.fortunewheel_spin ()

  private.Unsubscribe(
    "fortunewheel",
    "View_Stage_FortuneWheel_ReadyToSpin",
    private.fortunewheel_spin
  );  
  
  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_wheel_tap"
  } );

  if ( ui.obj.Exists( "btn_int_dialog_stage_continue_video" ) ) then
    ne.obj.Attach( "btn_int_dialog_stage_continue_video", "anm_int_dialog_stage_continue_window" );
  end;
  
  local focus_object    = "btn_int_dialog_stage_fortunewheel_run";
  local dialog_settings = private.Dialog_Settings( focus_object );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type        = "flashlight",

    res_id      = 3,
    angle       = 0,
    scale_x     = 1.5,
    scale_y     = 0.8,
  
    lock        = true,
    black       = false,
    focus       = focus_object,
    hand        = {
      angle     = 0.5 * math.pi,
      offset_x  = 0,
      offset_y  = 0,
      type      = "click"
    }
  };

  local backup_control = "btn_int_dialog_stage_fortunewheel_run";
  local control_tap = "btn_int_dialog_stage_fortunewheel_tap";
  
  ui.input.Off( control_tap );
  
  local func_close = function ()

    local func_fortunewheelspin_tap = private.Restore( "fortunewheel", "fortunewheelspin_tap" );

    ui.input.On( control_tap );
    
    ui.button.EventSet(
      backup_control,
      func_fortunewheelspin_tap
    );

    if ( func_fortunewheelspin_tap ) then

      func_fortunewheelspin_tap();

    end;

    private.Dialog_Hide();
    
    v_stage.FortuneWheel_SetPopupEnabled( true );

  end;

  private.Backup(
    "fortunewheel",
    "fortunewheelspin_tap",
    ui.button.EventGet( backup_control ) );

  ui.button.EventSet(
    backup_control,
    func_close
  );
  
  private.Dialog_ShowInstantly( "fortunewheel_spin", positions, hole );
  
  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_wheel_spin"
  } );  

  private.Complete( "fortunewheel" );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.fortunewheel_activate ()

  private.Subscribe( "fortunewheel_activate", "View_Stage_FortuneWheel_Turn", private.fortunewheel_activate_last );

end;

function private.fortunewheel_activate_last ()

  if ( not m_tutorial.IsCompleted( "fortunewheel_activate" ) ) then

    private.Unsubscribe( "fortunewheel_activate", "View_Stage_FortuneWheel_Turn", private.fortunewheel_activate_last ); 

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_wheel_activate"
    } );  

    private.Complete( "fortunewheel_activate" );

  end;    

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.fortunewheel_win ()

  private.Subscribe( "fortunewheel_win", "View_Stage_FortuneWheel_Reward", private.fortunewheel_win_last );

end;

function private.fortunewheel_win_last ()

  if ( not m_tutorial.IsCompleted( "fortunewheel_win" ) ) then

    private.Unsubscribe( "fortunewheel_win", "View_Stage_FortuneWheel_Reward", private.fortunewheel_win_last ); 

    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_wheel_win"
    } );  

    private.Complete( "fortunewheel_win" );

  end;    

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.spiders ()
  
  private.Subscribe(
    "spiders",
    "View_Tutorial_StageOpened", 
    private.spiders_first
  );
  
end;

function private.spiders_first ( stage_name )

  if ( stage_name == "stage_331"               ) and 
     ( not m_tutorial.IsCompleted( "spiders" ) ) then
    
    private.Unsubscribe(
      "spiders",
      "View_Tutorial_StageOpened",
      private.spiders_first
    );
    
    private.Subscribe(
      "spiders",
      "View_Stage_DialogTargetsHided",
      private.spiders_next1
    );

  end;
  
end;

function private.spiders_next1 ()
  
  private.Unsubscribe( "spiders", "View_Stage_DialogTargetsHided", private.spiders_next1 );
  
  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_spiders"
  } );  
  
  local match3_pos = GetObjPosByObj( "match" );

  local pos_x = match3_pos[ 1 ] + 2 * private.cell_size * 4 - private.cell_size;
  local pos_y = match3_pos[ 2 ] + 2 * private.cell_size * 4 - private.cell_size;

  local dialog_settings =
    private.Dialog_Settings( { pos_x, pos_y } );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type  = "arrow",

    pos_x = pos_x,
    pos_y = pos_y,
    scale = 1,
    angle = math.pi,
    lock  = true
  };

  local func_close = function ()
     
    private.Dialog_Hide( private.spiders_last );

  end;

  private.Dialog_ShowInstantly( "stage_331_1", positions, hole, func_close );

end;

function private.spiders_last ()

  private.Complete( "spiders" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.fans ()
  
  private.Subscribe(
    "fans",
    "View_Tutorial_StageOpened",
    private.fans_first
  );
  
end;

function private.fans_first ( stage_name )

  if ( stage_name == "stage_339"            ) and 
     ( not m_tutorial.IsCompleted( "fans" ) ) then
    
    private.Unsubscribe(
      "fans",
      "View_Tutorial_StageOpened",
      private.fans_first
    );
    
    private.Subscribe(
      "fans",
      "View_Stage_DialogTargetsHided",
      private.fans_next1
    );

  end;
  
end;

function private.fans_next1 ()
  
  private.Unsubscribe( "fans", "View_Stage_DialogTargetsHided", private.fans_next1 );
  
  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_fans"
  } );
  
  local match3_pos = GetObjPosByObj( "match" );

  local pos_x = match3_pos[ 1 ] + 2 * private.cell_size * 4 - private.cell_size;
  local pos_y = match3_pos[ 2 ] + 2 * private.cell_size * 9 - private.cell_size;

  local dialog_settings =
    private.Dialog_Settings( { pos_x, pos_y } );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type  = "arrow",

    pos_x = pos_x,
    pos_y = pos_y,
    scale = 1,
    angle = 0,
  
    lock  = false
  };

  local func_close = function ()
     
    private.Dialog_Hide( private.fans_last );

  end;

  private.Dialog_ShowInstantly( "stage_339_1", positions, hole, func_close );

end;

function private.fans_last ()

  private.Complete( "fans" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.anticrystal ()

  private.Subscribe(
    "anticrystal",
    "View_Tutorial_StageOpened",
    private.anticrystal_first
  );

end;

function private.anticrystal_first ( stage_name )

  if ( stage_name == "stage_340"                   ) and
     ( not m_tutorial.IsCompleted( "anticrystal" ) ) then

    private.Unsubscribe(
      "anticrystal",
      "View_Tutorial_StageOpened",
      private.anticrystal_first
    ); 

    private.Subscribe(
      "anticrystal",
      "View_Stage_DialogTargetsHided",
      private.anticrystal_next1
    );

  end;

end;

function private.anticrystal_next1 ()
  
  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_anticrystal"
  } );  

  local index = private.Entity_Index( v_stage.Stage_GetCurrentData().boosters, "anticrystal", "stage" ); 

  private.Unsubscribe( "anticrystal", "View_Stage_DialogTargetsHided", private.anticrystal_next1 );

  private.Backup( "anticrystal", "booster_tap", ui.control.EventsGet( "obj_int_panel_stage_booster_"..index ).func_on_click );

  local dialog_settings =
    private.Dialog_Settings( "obj_int_panel_stage_booster_"..index );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type        = "flashlight",  

    res_id      = 3,
    focus       = "obj_int_panel_stage_booster_"..index,
    scale       = 1,
  
    lock        = true,
    black       = false,
    hand        = {
      angle     = 0.5 * math.pi,
      offset_x  = 0,
      offset_y  = 0,
      type      = "click"
    }
  };

  local func_close = function ()
    
    private.Kontagent_LogEvent( "tutorial", 
    {
      name  = "tut_anticrystaldone"
    } );

    private.Dialog_Hide( function () private.anticrystal_next2( index ); end );

  end;

  ui.control.EventsSet( "obj_int_panel_stage_booster_"..index, { func_on_click = func_close } );

  private.Dialog_ShowInstantly( "stage_343_1", positions, hole );
  
  private.Entity_Add( "anticrystal", "anticrystal", "booster" ); --?? stage_343

  private.Entity_Free_Show( "anticrystal", "anticrystal", "obj_int_panel_stage_booster_"..index );--?? stage_343

end;

function private.anticrystal_next2 ( index )
  
  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_anticrystal_activate"
  } );     

  private.Subscribe( "anticrystal", "View_Boosters_BoosterUsed", private.anticrystal_last1 );

  local func_booster_tap = private.Restore( "anticrystal", "booster_tap" ); -- ??stage_343

  ui.control.EventsSet( "obj_int_panel_stage_booster_"..index, { func_on_click = func_booster_tap } );

  if ( func_booster_tap ) then

    func_booster_tap();

    private.Entity_Free_Hide( "obj_int_panel_stage_booster_"..index );

  end;

  local match3_pos = GetObjPosByObj( "match" );

  local pos_x = match3_pos[ 1 ] + 2 * private.cell_size * 5 - private.cell_size;
  local pos_y = match3_pos[ 2 ] + 2 * private.cell_size * 9 - private.cell_size;

  local dialog_settings =
    private.Dialog_Settings( { pos_x, pos_y } );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type        = "flashlight",

    res_id      = 3,
    pos_x       = pos_x,
    pos_y       = pos_y,
    scale       = 0.5,
  
    lock        = true,
    black       = false,
    hand       = {
      offset_x  = 0,
      offset_y  = 0,
      scale     = 1,
      angle     = 0,
      type      = "click"
    }
  };

  private.Dialog_ShowInstantly( "stage_343_2", positions, hole );

  private.Complete( "anticrystal" );

end;

function private.anticrystal_last1 ()
  
  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_anticrystal_activatedone"
  } );    

  private.Unsubscribe( "anticrystal", "View_Boosters_BoosterUsed", private.anticrystal_last1 );

  private.Dialog_Hide();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.candy ()

  private.Subscribe( "candy",
    "View_Tutorial_StageOpened",
    private.candy_first
  );

end;

function private.candy_first ( stage_name )

  if ( stage_name == "stage_441"             ) and
     ( not m_tutorial.IsCompleted( "candy" ) ) then

    private.Unsubscribe( "candy",
      "View_Tutorial_StageOpened",
      private.candy_first
    ); 

    private.Subscribe( "candy",
      "View_Stage_DialogTargetsHided",
      private.candy_show_match
    );

  end;
  
end;
  
function private.candy_show_match ()

  private.Unsubscribe( "candy",
    "View_Stage_DialogTargetsHided",
    private.candy_show_match
  );

  local match3_pos = GetObjPosByObj( "match" );

  local pos_x = match3_pos[ 1 ] + 2 * private.cell_size * 7;
  local pos_y = match3_pos[ 2 ] + 2 * private.cell_size * 6.5;
  
  local orientation = ui.screen.Orientation();
  
  local dialog_settings = private.Dialog_Settings( { pos_x, pos_y } );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type           = "flashlight",

    res_id         = 1,
    pos_x          = pos_x,
    pos_y          = pos_y,
    angle          = 0,
    scale          = private.cell_scale,

    lock           = true,
    hand           = {
         angle     = 0,
         offset_x  = 0,
         offset_y  = 0,
         type      = "move"
    }
  };

  local function func_close ()

    private.Unsubscribe( "candy",
      "Model_Match3_FieldActivated",
      func_close
    );

    private.Dialog_Hide( private.candy_point_counter );

  end;

  private.Subscribe( "candy",
    "Model_Match3_FieldActivated",
    func_close
  );
  
  private.Dialog_ShowInstantly( "candy_1", positions, hole );

end;

function private.candy_point_counter ()

  local dialog_settings =
    private.Dialog_Settings( "txt_int_panel_stage_targets_target_2_count" );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type        = "arrow",

    focus       = "txt_int_panel_stage_targets_target_2_count",
    offset_x    = 0,
    offset_y    = 0,
    scale       = 1,
    angle       = 1.25 * math.pi,
  
    lock        = true
  };

  local func_close = function ()

    private.Dialog_Hide( private.candy_complete );
    
  end;

  private.Dialog_ShowInstantly( 
    "candy_2",
    positions, 
    hole, 
    func_close
  );

end;

function private.candy_complete ()

  private.Complete( "candy" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.bottles ()

  private.Subscribe(
    "bottles",
    "View_Tutorial_StageOpened",
    private.bottles_first
  );

end;

function private.bottles_first ( stage_name )

  if ( stage_name == "stage_441"               ) and
     ( not m_tutorial.IsCompleted( "bottles" ) ) then

    private.Unsubscribe(
      "bottles",
      "View_Tutorial_StageOpened",
      private.bottles_first
    ); 

    private.Subscribe(
      "bottles",
      "View_Stage_DialogTargetsHided",
      private.bottles_point_to
    );

  end;

end;

function private.bottles_point_to ()
  
  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_bottles"
  } );

  private.Unsubscribe(
    "bottles",
    "View_Stage_DialogTargetsHided",
    private.bottles_point_to
  );

  local match3_pos = GetObjPosByObj( "match" );

  local pos_x = match3_pos[ 1 ] + 2 * private.cell_size * 4 - private.cell_size;
  local pos_y = match3_pos[ 2 ] + 2 * private.cell_size * 5;

  local dialog_settings = private.Dialog_Settings( { pos_x, pos_y } );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type           = "arrow",

    pos_x          = pos_x,
    pos_y          = pos_y,
    angle          = 0,
    scale          = private.cell_scale,

    lock           = true
  };

  local func_close = function ()

    private.Dialog_Hide( private.bottles_show_match );

  end;

  private.Dialog_ShowInstantly( "bottles_1", positions, hole, func_close );

end;

function private.bottles_show_match ()
  
  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_bottlesdone"
  } );

  local match3_pos = GetObjPosByObj( "match" );

  local pos_x = match3_pos[ 1 ] + 2 * private.cell_size * 5;
  local pos_y = match3_pos[ 2 ] + 2 * private.cell_size * 2.5;
  
  local orientation = ui.screen.Orientation();
  
  local dialog_settings = private.Dialog_Settings( { pos_x, pos_y } );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type           = "flashlight",

    res_id         = 1,
    pos_x          = pos_x,
    pos_y          = pos_y,
    angle          = 0,
    scale          = private.cell_scale,

    lock           = true,
    hand           = {
         angle     = 0,
         offset_x  = 0,
         offset_y  = 0,
         type      = "move"
    }
  };

  local function func_close ()

    private.Unsubscribe( "bottles",
      "Model_Match3_FieldActivated",
      func_close
    );

    private.Dialog_Hide( private.bottles_complete );

  end;

  private.Subscribe( "bottles",
    "Model_Match3_FieldActivated",
    func_close
  );
  
  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_bottles_match"
  } );
  
  private.Dialog_ShowInstantly( "bottles_2", positions, hole );

end;

function private.bottles_complete ()
  
  private.Kontagent_LogEvent( "tutorial", 
  {
    name  = "tut_bottles_matchdone"
  } );

  private.Complete( "bottles" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.paint ()

  local func_validate = function ( stage_name )
    local  is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_455"
    );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "paint",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.wind ()

  local func_validate = function ( stage_name )
    local  is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_551"
    );
    return is_valid;
  end;

  local delegates = {
    GetMatch3FieldCenter = function ()
      local  pos_x, pos_y = v_match3.GetFieldCenter();
      return pos_x, pos_y;
    end,
    ShowMatch3Effect = function ()
      v_stage.Match3_ShowEffect( "windflaw", true );
    end,
    HideMatch3Effect = function ()
      v_stage.Match3_HideEffect();
    end
  };

  private.Tutorial_SetOnEvent(
    "wind",
    "View_Tutorial_StageOpened",
    delegates,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.snowflakes ()

  local func_validate = function ( stage_name )
    local is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_559"
    );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "snowflakes",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.stage_661 ()

  local func_validate = function ( stage_name )
    local is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_661"
    );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "stage_661",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.stage_683 ()

  local func_validate = function ( stage_name )
    local is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_683"
    );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "stage_683",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.stage_tutorial_1_minievent ()

  local func_validate = function ( stage_name )
    local is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_tutorial_1_minievent"
    );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "stage_tutorial_1_minievent",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.stage_771 ()

  local func_validate = function ( stage_name )
    local is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_771"
    );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "stage_771",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.stage_793 ()

  local func_validate = function ( stage_name )
    local is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_793"
    );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "stage_793",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.stage_903 ()

  local func_validate = function ( stage_name )
    local is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_903"
    );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "stage_903",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
function private.stage_1211 ()

  local func_validate = function ( stage_name )
    local is_valid = private.Tutorial_IsStageMatched(
      stage_name,
      "stage_1211"
    );
    return is_valid;
  end;

  private.Tutorial_SetOnEvent(
    "stage_1211",
    "View_Tutorial_StageOpened",
    nil,
    func_validate
  );

end;
--------------------------------------------------------------------------------
-- function ...unused... () end;
--------------------------------------------------------------------------------
function private.resources ()

  private.Subscribe( "resources", "View_Tutorial_MapOpened", private.resources_first );

end;

function private.resources_first ()

  if ( not m_tutorial.IsCompleted( "resources" ) ) then

    if ( ( v_stage.Stage_StorageData( "stage_4" ) or {} ).completed ) then

      private.Unsubscribe( "resources", "View_Tutorial_MapOpened", private.resources_first ); 

      common.Defer( "tutorial_resources_delay", 0.5, private.resources_next1 );

    end;

  end;

end;

function private.resources_next1 ()

  local positions = 
  {
    dlg_start   = 1200,
    dlg_pos_y   = private.app_h / 2 - 200
  };

  local hole = 
  {
    type        = "arrow",

    focus       = "obj_int_panel_interface_top_resource1",
    offset_x    = - 65,
    offset_y    = 0,
    angle       = 3.1415,
    scale       = 1.0,

    lock        = true
  };

  local func_close = function ( execute_tap )

    private.Dialog_Hide( 

      private.resources_last1

    );

  end;

  private.Dialog_ShowFromQueue( "resources_1", positions, hole, func_close );

  private.Complete( "resources" );

end;

function private.resources_last1 ()

  v_map.Focus( "obj_map_marker" );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.replay ()

  local func_validate = function ( stage_name )
    local  is_valid = private.Tutorial_IsStageCompleted( "stage_7" );
    return is_valid;
  end;

  local delegates = {
    GetStageState = function ()
      local  stage_state = ( v_stage.Stage_StorageData( "stage_5" ) or {} ).state;
      return stage_state;
    end
  };

  private.Tutorial_SetOnEvent(
    "replay",
    "View_Tutorial_MapOpened",
    delegates,
    func_validate
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.robber_cloud ()

  private.Subscribe( "robber_cloud", "View_Map_HeroPopupActivated", private.robber_cloud_last );

end;

function private.robber_cloud_last ( stage_name )

  if ( stage_name == "stage_6" and not m_tutorial.IsCompleted( "robber_cloud" ) ) then

    private.Unsubscribe( "robber_cloud", "View_Map_HeroPopupActivated", private.robber_cloud_last ); 

    private.Complete( "robber_cloud" );

  end;    

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.swordstone_cloud ()

  private.Subscribe( "swordstone_cloud", "View_Map_HeroPopupActivated", private.swordstone_cloud_last );

end;

function private.swordstone_cloud_last ( stage_name )

  if ( stage_name == "stage_2" and not m_tutorial.IsCompleted( "swordstone_cloud" ) ) then

    private.Unsubscribe( "swordstone_cloud", "View_Map_HeroPopupActivated", private.swordstone_cloud_last ); 

    private.Complete( "swordstone_cloud" );

  end;    

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.prisoner_cloud ()

  private.Subscribe( "prisoner_cloud", "View_Map_HeroPopupActivated", private.prisoner_cloud_last );

end;

function private.prisoner_cloud_last ( stage_name )

  if ( stage_name == "stage_3" and not m_tutorial.IsCompleted( "prisoner_cloud" ) ) then

    private.Unsubscribe( "prisoner_cloud", "View_Map_HeroPopupActivated", private.prisoner_cloud_last );

    private.Complete( "prisoner_cloud" );

  end;    

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stage_7 ()

  private.Subscribe(
    "stage_7",
    "View_Tutorial_StageOpened",
    private.stage_7_first
  );

end;

function private.stage_7_first ( stage_name )

  if ( stage_name == "stage_7"                 ) and
     ( not m_tutorial.IsCompleted( "stage_7" ) ) then
    
    private.Unsubscribe(
      "stage_7",
      "View_Tutorial_StageOpened",
      private.stage_7_first
    );
    
    private.Subscribe(
      "stage_7",
      "View_Stage_DialogTargetsHided",
      private.stage_7_next1
    );

  end;

end;

function private.stage_7_next1 ()

  private.Unsubscribe(
    "stage_7",
    "View_Stage_DialogTargetsHided",
    private.stage_7_next1
  );
  
  local match3_pos = GetObjPosByObj( "match" );

  local pos_x = match3_pos[ 1 ] + 2 * private.cell_size * 3 + private.cell_size;
  local pos_y = match3_pos[ 2 ] + 2 * private.cell_size * 4;
  
  local v_pref = 25;
  
  if ( ui.screen.IsLandscape() ) then
    
    v_pref = nil;
    
  end;
  
  local dialog_settings = private.Dialog_Settings(
    { pos_x, pos_y },
    v_pref
  );
    
  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type           = "flashlight",

    res_id         = 10,
    pos_x          = pos_x + private.cell_size,
    pos_y          = pos_y,
    angle          = 0,
    scale          = private.cell_scale,

    lock           = false,
    black          = true,
    input          = true
  };

  local func_close = function ()

    private.Dialog_Hide( private.stage_7_last1 );

  end;

  private.Dialog_ShowInstantly( "stage_7_1", positions, hole, func_close );

end;

function private.stage_7_last1 ()

  private.Complete( "stage_7" );

end;
--------------------------------------------------------------------------------