-- name=int_tutorial_achievements
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Subscribe           = function () end,
  Unsubscribe         = function () end,
  Complete            = function () end,
  Backup              = function () end,
  Restore             = function () end,
  LogKontagentEvent   = function () end,
  GetDialogSettings   = function () end,
  ShowDialogInstantly = function () end,
  HideDialog          = function () end,
  StartPending        = function () end,
  FinishPending       = function () end,
  Unload              = function () end,
  DisableScrolling    = function () end,
  EnableScrolling     = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  --

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Start ( achievement_data )

  private.achievements_first( achievement_data );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.achievements_first ( achievement_data )
    
  _f.Subscribe(
    "achievements",
    "View_Map_Update_Finished",
    private.achievements_next1
  );

end;
--------------------------------------------------------------------------------
function private.achievements_next1 ()
  
  _f.LogKontagentEvent( "tutorial", { name = "tut_acheivements_tap" } );  

  _f.Unsubscribe(
    "achievements",
    "View_Map_Update_Finished",
    private.achievements_next1
  );
  
  _f.StartPending( "achievements", "map" );
  
  local focus_obj       = "btn_int_panel_interface_bottom_player";
  local dialog_settings = _f.GetDialogSettings( focus_obj );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type     = "flashlight",
    focus    = focus_obj,    
    res_id   = 3,
    offset_x = 0,
    offset_y = -70,
    angle    = 0,
    scale    = 1.0,
    lock     = true,
    black    = false,
    wait     = true,
    hand = {  
      offset_x = 0,
      offset_y = 0,
      scale_x  = - 1,
      angle    = - 0.785, 
      type     = "click"
    }
  };
  
  local func_on_click = ui.control.EventsGet( focus_obj ).func_on_click;

  local func_close = function ()
    
    ui.control.EventsSet( focus_obj, { func_on_click = func_on_click } );
    
    _f.Subscribe(
      "achievements",
      "Interface_Hero_AchievementUpdated",
      private.achievements_next2
    );
    
    _f.HideDialog( function ()
    
      if ( func_on_click ) then 
        int_dialog_player_main.SetForcedTab( "achievements" );
        func_on_click(); 
      end;
    
    end );

  end;

  ui.control.EventsSet( focus_obj, { func_on_click = func_close } );

  _f.ShowDialogInstantly(
    "achievements_1",
    positions,
    hole,
    func_close,
    nil,
    { active = true, notify = false }
  );
  
  private.achievements_last1();
 
end;
--------------------------------------------------------------------------------
function private.achievements_next2 ( achievement_name )
  
  if not ( achievement_name == "only_forward" ) then
    
    return;
    
  end;
  
  _f.LogKontagentEvent( "tutorial", { name = "tut_acheivements_tapdone" } );
  
  _f.Unsubscribe(
    "achievements",
    "Interface_Hero_AchievementUpdated",
    private.achievements_next2
  );
  
  _f.DisableScrolling();

  local button_obj =
    "btn_int_dialog_player_main_achievements_slot_button_only_forward";
  
  local focus_obj = "anm_int_dialog_player_main_achievements_slot_only_forward";
  
  local dialog_settings = _f.GetDialogSettings( focus_obj );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local scale_x,  scale_y  = 1, 1;
  local offset_x, offset_y = 0, 16;
  
  if ( ui.screen.IsLandscape() ) then
    
    scale_x,  scale_y  = 0.75, 0.75;
    offset_x, offset_y = 0, 15;
    
  end;
  
  local hole = 
  {
    type     = "flashlight",
    focus    = focus_obj,    
    res_id   = 8,
    angle    = 0,
    scale_x  = scale_x,
    scale_y  = scale_y,
    offset_x = offset_x,
    offset_y = offset_y,
    lock     = true,
    black    = true,
    wait     = true,
    hand     = {  
      offset_x = -10,
      offset_y = 40,
      angle    = 0, 
      type     = "click"
    }
  };

  local func_close = function ()

    local func_achievements_reward =
      _f.Restore(
        "achievements",
        "achievements_reward"
      );

    ui.button.EventSet(
      button_obj,
      func_achievements_reward
    );
    
    _f.LogKontagentEvent( "tutorial", { name = "tut_acheivements_rewarddone" } );
  
    _f.HideDialog( function ()
    
      if ( func_achievements_reward ) then func_achievements_reward(); end;
      
      _f.EnableScrolling();
   
    end );
    
  end;

  _f.Backup(
    "achievements",
    "achievements_reward",
    ui.button.EventGet( button_obj )
  );

  ui.button.EventSet( button_obj, func_close );
  
  _f.LogKontagentEvent( "tutorial", { name = "tut_acheivements_reward" } );
  _f.ShowDialogInstantly( "achievements_2", positions, hole );

end;
--------------------------------------------------------------------------------
function private.achievements_last1 ()

  _f.Complete();
  _f.FinishPending( "achievements", "map" );
  _f.Unload();

end;
--------------------------------------------------------------------------------