-- name=int_tutorial_dailyquest
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Subscribe             = function () end,
  Unsubscribe           = function () end,
  Complete              = function () end,
  LogKontagentEvent     = function () end,
  GetDialogSettings     = function () end,
  ShowDialogInstantly   = function () end,
  HideDialog            = function () end,
  StartPending          = function () end,
  FinishPending         = function () end,
  Unload                = function () end,
  IsDailyquestCountdown = function () end,
  IsDailyquestReady     = function () end,
  IsDialogOnScreen      = function () end,
  ClearOuterTap         = function () end,
  SetOuterTap           = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  --

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Start ()

  private.dailyquest_prepare();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.dailyquest_prepare ()
  
  if ( _f.IsDailyquestCountdown() ) then

    _f.Subscribe(
      "dailyquest",
      "Model_DailyQuest_CountdownEnded",
      private.dailyquest_first
    );

  else

    if ( not _f.IsDailyquestReady() ) then

      _f.Subscribe(
        "dailyquest",
        "View_DailyQuest_Ready",
        private.dailyquest_first
      );

    else

      private.dailyquest_first();

    end;

  end;
  
end;
--------------------------------------------------------------------------------
function private.dailyquest_first ()

  _f.Unsubscribe(
    "dailyquest",
    "Model_DailyQuest_CountdownEnded",
    private.dailyquest_first
  );

  _f.Unsubscribe(
    "dailyquest",
    "View_DailyQuest_Ready",
    private.dailyquest_first
  );

  if ( _f.IsDialogOnScreen() ) then

    private.dailyquest_next1();

  else

    _f.Subscribe(
      "dailyquest",
      "Interface_Window_WindowOpened",
      private.dailyquest_first1
    );

  end;

end;
--------------------------------------------------------------------------------
function private.dailyquest_first1 ( window_name )

  if ( window_name ~= "int_dialog_dailyquest" ) then return; end;

  _f.Unsubscribe(
    "dailyquest",
    "Interface_Window_WindowOpened",
    private.dailyquest_first1
  );

  private.dailyquest_next1();

end;
--------------------------------------------------------------------------------
function private.dailyquest_next1 ()
  
  _f.LogKontagentEvent( "tutorial", { name = "tut_arena_chest" } );

  local focus_obj = "obj_int_dialog_dailyquest_cloud";
  local dialog_settings = _f.GetDialogSettings( focus_obj, 85 );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type     = "arrow",
    focus    = focus_obj,    
    offset_x = 0,
    offset_y = 0,
    angle    = math.rad( 180 ),
    scale    = 1.0,
    lock     = true
  };
  
  ui.input.Off( "btn_int_dialog_dailyquest_close" );
  
  _f.ClearOuterTap();

  local func_close = function ()

    _f.HideDialog( private.dailyquest_next2 );

  end;

  _f.ShowDialogInstantly( "dailyquest_1", positions, hole, func_close );

end;
--------------------------------------------------------------------------------
function private.dailyquest_next2 ()
  
  _f.LogKontagentEvent( "tutorial", { name = "tut_arena_battle" } );

  local positions = 
  {
    dlg_start   = 1200,
    dlg_pos_y   = ui.screen.Height() / 2 + 240
  };

  local hole = 
  {
    type     = "arrow",
    focus    = "anm_int_dialog_dailyquest_challenge_1",    
    offset_x = 0,
    offset_y = 0,
    angle    = math.pi,
    scale    = 1.0,
    lock     = true
  };

  local func_close = function ()

    _f.HideDialog( private.dailyquest_last1 );

  end;

  _f.ShowDialogInstantly( "dailyquest_2", positions, hole, func_close );

end;
--------------------------------------------------------------------------------
function private.dailyquest_last1 ()

  ui.input.On( "btn_int_dialog_dailyquest_close" );

  _f.SetOuterTap();
  _f.Complete();
  _f.Unload();

end;
--------------------------------------------------------------------------------