-- name=int_tutorial_market_construct
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Subscribe           = function () end,
  Unsubscribe         = function () end,
  Backup              = function () end,
  Restore             = function () end,
  Complete            = function () end,
  LogKontagentEvent   = function () end,
  GetDialogSettings   = function () end,
  ShowDialogInstantly = function () end,
  HideDialog          = function () end,
  StartPending        = function () end,
  FinishPending       = function () end,
  Unload              = function () end,
  IsDialogShown       = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  --

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Start ()

  private.market_construct_first();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.market_construct_first ()
  
  _f.Subscribe(
    "market_construct",
    "View_Map_Update_Finished",
    private.market_construct_next1
  );

end;
--------------------------------------------------------------------------------
function private.market_construct_next1 ()

  _f.Unsubscribe(
    "market_construct",
    "View_Map_Update_Finished",
    private.market_construct_next1
  );

  local focus_obj = "obj_map_builds_market";
  
  _f.StartPending( "market_construct", "map" );

  local func_close = function ( execute_tap )

    local func_market_construct_tap = _f.Restore(
      "market_construct",
      "market_construct_tap"
    );

    ui.control.TapSet(
      focus_obj,
      { func_true = func_market_construct_tap }
    );

    _f.HideDialog( function ()
    
      if ( execute_tap ) then func_market_construct_tap(); end;
      
      private.market_construct_next2();
     
    end );

  end;

  _f.Backup(
    "market_construct",
    "market_construct_tap",
    ui.control.TapGet( focus_obj ).func_true
  );

  ui.control.TapSet(
    focus_obj,
    { func_true = function () func_close( true ); end }
  );

  local dialog_settings = _f.GetDialogSettings( focus_obj );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type     = "flashlight",
    res_id   = 3,
    focus    = focus_obj,
    scale    = 1,
    offset_x = 0,
    offset_y = -20,
    lock     = true,
    black    = false,
    hand     = {
      angle    = 0.5 * math.pi,
      offset_x = 0,
      offset_y = 0,
      type     = "click"
    }
  };

  _f.ShowDialogInstantly(
    "market_construct_1",
    positions,
    hole, 
    func_close,
    nil,
    { input = true, active = true }
  );

end;
--------------------------------------------------------------------------------
function private.market_construct_next2 ()

  common.Defer(
    "tutorial_market_construct_delay", 0.5,
    function ()
  
      if ( _f.IsDialogShown() ) then
    
        private.market_construct_next3();
        
      else
        
        private.market_construct_last1();
        
      end;
 
    end
  );

end;
--------------------------------------------------------------------------------
function private.market_construct_next3 ()

  local focus_obj = "btn_int_dialog_map_build_market_accept";

  local func_close = function ( execute_tap )

    local func_market_construct_tap = _f.Restore( 
      "market_construct",
      "market_construct_tap"
    );

    ui.button.EventSet( focus_obj, func_market_construct_tap );

    _f.HideDialog( private.market_construct_last1 );

    if ( execute_tap ) then func_market_construct_tap(); end;

  end;

  _f.Backup(
    "market_construct",
    "market_construct_tap",
    ui.button.EventGet( focus_obj )
  );

  ui.button.EventSet( focus_obj, function () func_close( true ); end );
  
  local dialog_settings = _f.GetDialogSettings( focus_obj );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type     = "flashlight",
    res_id   = 3,
    focus    = "btn_int_dialog_map_build_market_accept",
    scale_x  = 1.6,
    scale_y  = 0.8,
    offset_x = 0,
    offset_y = 0,  
    lock     = true,
    black    = false,
    hand     = {
      angle    = 0.5 * math.pi,
      offset_x = 0,
      offset_y = 0,
      type     = "click"
    }
  };

  _f.ShowDialogInstantly(
    "market_construct_2",
    positions,
    hole,
    func_close
  );

end;
--------------------------------------------------------------------------------
function private.market_construct_last1 ()

  _f.Complete();
  _f.FinishPending( "market_construct", "map" );
  _f.Unload();

end;
--------------------------------------------------------------------------------