-- name=int_tutorial_mine_construct
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Subscribe           = function () end,
  Unsubscribe         = function () end,
  Backup              = function () end,
  Restore             = function () end,
  Complete            = function () end,
  LogKontagentEvent   = function () end,
  GetDialogSettings   = function () end,
  ShowDialogInstantly = function () end,
  HideDialog          = function () end,
  StartPending        = function () end,
  FinishPending       = function () end,
  Unload              = function () end,
  GetStartPoint       = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  --

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Start ( stage_name )

  local start_point = _f.GetStartPoint();
  local start_func  = private[ "mine_construct_"..start_point ];
  
  if ( start_func ) then

    start_func();

  else

    DbgTrace(
      "[ ERROR ][ int_tutorial_mine_construct ] Start function not found."
    );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.mine_construct_first ( mine_name )

  _f.Subscribe(
    "mine_construct",
    "View_Map_Update_Mine_Animated",
    private.mine_construct_next1
  );

end;
--------------------------------------------------------------------------------
function private.mine_construct_next1 ()
  
  _f.Unsubscribe(
    "mine_construct",
    "View_Map_Update_Mine_Animated",
    private.mine_construct_next1
  );

  _f.StartPending( "mine_construct", "map" );

  local focus_obj     = "btn_map_mines_mine_1";   
  local events        = ui.control.EventsGet( focus_obj );
  local func_on_click = events.func_on_click;

  local func_close = function ()

    events.func_on_click = func_on_click;

    _f.HideDialog( function ()
    
      func_on_click();
      private.mine_construct_next2();
     
    end );

  end;
  
  events.func_on_click = func_close;

  local dialog_settings = _f.GetDialogSettings( focus_obj );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type     = "flashlight",
    res_id   = 3,
    focus    = focus_obj,
    scale    = 1.5,
    offset_x = 0,
    offset_y = -40,
    lock     = true,
    black    = false,
    hand     = {
      angle    = 0.5 * math.pi,
      offset_x = 0,
      offset_y = 0,
      type     = "click"
    }
  };

  local map_params = {
    notify = false,
    input  = true,
    active = true
  };

  _f.ShowDialogInstantly(
    "mine_construct_1",
    positions,
    hole,
    func_close,
    nil,
    map_params
  );

end;
--------------------------------------------------------------------------------
function private.mine_construct_next2 ()

  common.Defer(
    "tutorial_mine_construct_delay", 0.3,
    private.mine_construct_next3
  );

end;
--------------------------------------------------------------------------------
function private.mine_construct_next3 ()

  local focus_obj = "anm_int_dialog_map_mine_resource1";

  local func_close = function ( execute_tap )

    local func_mine_construct_tap = _f.Restore(
      "mine_construct",
      "mine_construct_tap"
    );

    ui.control.TapSet( focus_obj, { func_true = func_mine_construct_tap } );

    _f.HideDialog( private.mine_construct_next4 );

    if ( execute_tap ) then func_mine_construct_tap(); end;

  end;

  _f.Backup(
    "mine_construct",
    "mine_construct_tap",
    ui.control.TapGet( focus_obj ).func_true
  );

  ui.control.TapSet(
    focus_obj,
    { func_true = function () func_close( true ); end }
  );

  local dialog_settings = _f.GetDialogSettings( focus_obj, 80, 70 );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type     = "flashlight",
    res_id   = 3,
    focus    = focus_obj,
    scale    = 1,
    offset_x = 0,
    offset_y = -15,
    lock     = true,
    black    = false,
    hand     = {
      angle    = 0.3 * math.pi,
      offset_x = 0,
      offset_y = 0,
      type     = "click"
    }
  };

  _f.ShowDialogInstantly( "mine_construct_2", positions, hole, func_close );

end;
--------------------------------------------------------------------------------
function private.mine_construct_next4 ()

  local focus_obj = "btn_int_dialog_map_mine_select";

  local func_close = function ()

    local func_mine_construct_tap = _f.Restore(
      "mine_construct",
      "mine_construct_tap"
    );

    ui.button.EventSet( focus_obj, func_mine_construct_tap );

    _f.HideDialog( private.mine_construct_last1 );

    func_mine_construct_tap();

  end;

  _f.Backup(
    "mine_construct",
    "mine_construct_tap",
    ui.button.EventGet( focus_obj )
  );

  ui.button.EventSet( focus_obj, func_close );

  local dialog_settings = _f.GetDialogSettings( focus_obj );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type     = "flashlight",
    res_id   = 3,
    focus    = "btn_int_dialog_map_mine_select",
    scale_x  = 2.5,
    scale_y  = 1.0,
    offset_x = 0,
    offset_y = 0,  
    lock     = true,
    black    = false,
    hand     = {
      angle    = 0.5 * math.pi,
      offset_x = 0,
      offset_y = 0,
      type     = "click"
    }
  };

  _f.ShowDialogInstantly( "mine_construct_3", positions, hole, func_close );
  _f.Complete();

end;
--------------------------------------------------------------------------------
function private.mine_construct_last1 ()

  _f.FinishPending( "mine_construct", "map" );
  _f.Unload();

end;
--------------------------------------------------------------------------------