-- name=int_tutorial_skin
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  Subscribe               = function () end,
  Unsubscribe             = function () end,
  Complete                = function () end,
  Backup                  = function () end,
  Restore                 = function () end,
  LogKontagentEvent       = function () end,
  GetDialogSettings       = function () end,
  ShowDialogInstantly     = function () end,
  HideDialog              = function () end,
  StartPending            = function () end,
  FinishPending           = function () end,
  Unload                  = function () end,
  IsSkinActive            = function () end,
  IsSkinWithoutParts      = function () end,
  FocusOnSkin             = function () end,
  IsSelectButtonAvailable = function () end
};

local _this = {
  is_started = false
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  --

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Start ( skin_name )

  _this.skin_name  = skin_name;
  _this.is_started = true;

  private.skin_first();

end;
--------------------------------------------------------------------------------
function public.IsStarted ()

  return _this.is_started;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.skin_first ()
    
  _f.Subscribe(
    "skin",
    "View_Map_Update_Quest_RewardDisplayed",
    private.skin_next1
  );

  _f.Subscribe(
    "skin",
    "Interface_DialogHero_Opened",
    private.skin_next2
  );

end;
--------------------------------------------------------------------------------
function private.skin_next1 ()
 
  _f.Unsubscribe(
    "skin",
    "View_Map_Update_Quest_RewardDisplayed",
    private.skin_next1
  );
  
  _f.StartPending( "skin", "map" );
  _f.LogKontagentEvent( "tutorial", { name = "tut_skin_tap" } );

  local focus_obj = "btn_int_panel_interface_bottom_player";
  local dialog_settings = _f.GetDialogSettings( focus_obj );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type     = "flashlight",
    res_id   = 3,
    focus    = focus_obj,
    scale    = 1.25,
    offset_x = 0,
    offset_y = -70,  
    lock     = true,
    black    = true,
    hand     = {
      angle    = 0.5 * math.pi,
      offset_x = 0,
      offset_y = 0,
      type     = "click"
    }
  };

  local func_hero_tap = ui.control.EventsGet( focus_obj ).func_on_click;

  local func_close = function ()

    _f.Subscribe(
      "skin",
      "Interface_DialogHero_Opened",
      private.skin_next2
    );

    _f.HideDialog();
    ui.control.EventsSet( focus_obj, { func_on_click = func_hero_tap } );
    func_hero_tap();

  end;

  ui.control.EventsSet( focus_obj, { func_on_click = func_close } );

  _f.ShowDialogInstantly(
    "skin_1",
    positions,
    hole,
    nil,
    nil,
    {
      active = true,
      notify = false
    }
  );

end;
--------------------------------------------------------------------------------
function private.skin_next2 ()

  _f.Unsubscribe(
    "skin",
    "View_Map_Update_Quest_RewardDisplayed",
    private.skin_next1
  );

  _f.Unsubscribe(
    "skin",
    "Interface_DialogHero_Opened",
    private.skin_next2
  );

  _f.StartPending( "skin", "map" );
  _f.LogKontagentEvent( "tutorial", { name = "tut_skin_tapdone" } );

  -- JK: if somehow the skin is already active, force complete the tutorial
  if ( _f.IsSkinActive( _this.skin_name ) ) then

    private.skin_last1();
    return;

  end;
  
  _f.LogKontagentEvent( "tutorial", { name = "tut_skin_emblem" } );

  if ( _f.IsSkinWithoutParts( _this.skin_name ) ) then
     
    ui.screen.Lock( true );
    
    _f.FocusOnSkin(
      _this.skin_name,
      function ()      
        ui.screen.Lock( false );
        private.skin_next3();    
      end
    );
 
    return;

  end;

  local focus_obj = "anm_int_dialog_player_main_skin_info_progress";
  local dialog_settings = _f.GetDialogSettings( focus_obj );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type     = "arrow",
    focus    = focus_obj,
    angle    = 0,
    scale    = 1.0,
    offset_x = 0,
    offset_y = 0,
    lock     = true
  };

  local func_close = function ()
    
    _this.is_started = false;
    
    mediator.Notify( "View_Tutorial_SkinProgressComplete" );

    _f.HideDialog( private.skin_next3 );

  end;

  ui.screen.Lock( true );
  
  _f.FocusOnSkin(
    _this.skin_name,
    function ()
  
      _f.ShowDialogInstantly( "skin_2", positions, hole, func_close );    
      ui.screen.Lock( false );
  
    end
  );

end;
--------------------------------------------------------------------------------
function private.skin_next3 ()
  
  _f.LogKontagentEvent( "tutorial", { name = "tut_skin_outfit" } );
  
  local focus_obj = "obj_int_dialog_player_main_skin_hero";
  local v_pref    = tern( ui.screen.IsLandscape() )( nil, 65 );
  
  local dialog_settings = _f.GetDialogSettings( focus_obj, v_pref );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type     = "arrow",
    focus    = focus_obj,
    angle    = math.rad( 315 ),
    scale    = 1.0,
    offset_x =  50,
    offset_y = -150,
    lock     = true
  };

  local func_close = function ()

    _f.HideDialog( private.skin_next4 );

  end;
  
  _f.ShowDialogInstantly( "skin_3", positions, hole, func_close );

end;
--------------------------------------------------------------------------------
function private.skin_next4 ()

  -- JK: if somehow the skin tab has no select button,
  -- force complete the tutorial
  if ( not _f.IsSelectButtonAvailable() ) then

    private.skin_last1();
    return;

  end;

  _f.LogKontagentEvent( "tutorial", { name  = "tut_skin_wear" } );

  local focus_obj = "btn_int_dialog_player_main_skin_select";
  local v_pref    = 90;
  
  local dialog_settings = _f.GetDialogSettings( focus_obj, v_pref );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type    = "flashlight",
    res_id  = 3,
    focus   = focus_obj,
    scale_x = 1.4,
    scale_y = 0.7,    
    lock    = true,
    black   = false,
    hand    = {
      angle    = 0.5 * math.pi,
      offset_x = 0,
      offset_y = 0,
      type     = "click"
    }
  };

  local func_close = function ()

    local func_btn_wear_tap = _f.Restore( "skin", "btn_wear_tap" );

    ui.button.EventSet( focus_obj, func_btn_wear_tap );

    func_btn_wear_tap();

    _f.HideDialog( private.skin_last1 );

  end;

  _f.Backup(
    "skin",
    "btn_wear_tap",
    ui.button.EventGet( focus_obj )
  );

  ui.button.EventSet( focus_obj, func_close );

  _f.ShowDialogInstantly( "skin_4", positions, hole );

end;
--------------------------------------------------------------------------------
function private.skin_last1 ()
  
  _f.LogKontagentEvent( "tutorial", { name = "tut_skin_weardone" } );
  _f.Complete();
  _f.FinishPending( "skin", "map" );
  _f.Unload();

end;
--------------------------------------------------------------------------------