-- name=int_tutorial_snowflakes
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  GetMatch3Cell        = function () end,
  Subscribe            = function () end,
  Unsubscribe          = function () end,
  Complete             = function () end,
  LogKontagentEvent    = function () end,
  GetDialogSettings    = function () end,
  ShowDialogInstantly  = function () end,
  HideDialog           = function () end,
  Unload               = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  --

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Start ( stage_name )

  private.snowflakes_start();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.snowflakes_start ()

  _f.Subscribe(
    "snowflakes",
    "View_Stage_DialogTargetsHided",
    private.snowflakes_step1
  );

end;
--------------------------------------------------------------------------------
function private.snowflakes_step1 ()

  _f.Unsubscribe(
    "snowflakes",
    "View_Stage_DialogTargetsHided",
    private.snowflakes_step1
  );
  
  _f.LogKontagentEvent( "tutorial", { name = "tut_snowflakes" } );
  
  local match3_pos = GetObjPosByObj( "match" );

  local cell = _f.GetMatch3Cell();

  local pos_x = match3_pos[ 1 ] + 2 * cell.size * 1;
  local pos_y = match3_pos[ 2 ] + 2 * cell.size * 6;
  
  local dialog_settings = _f.GetDialogSettings( { pos_x, pos_y } );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type   = "flashlight",
    res_id = 15,
    pos_x  = pos_x,
    pos_y  = pos_y,
    angle  = 0,
    scale  = 1,
    lock   = false,
    black  = true,
    input  = true
  };

  local func_close = function ()

    _f.HideDialog( private.snowflakes_step2 );

  end;

  _f.ShowDialogInstantly( "snowflakes_1", positions, hole, func_close );

end;
--------------------------------------------------------------------------------
function private.snowflakes_step2 ()
  
  _f.LogKontagentEvent( "tutorial", { name = "tut_snowflakesdone" } );

  local match3_pos = GetObjPosByObj( "match" );

  local cell = _f.GetMatch3Cell();

  local pos_x = match3_pos[ 1 ] + 2 * cell.size * 1;
  local pos_y = match3_pos[ 2 ] + 2 * cell.size * 5.5;
  
  local dialog_settings = _f.GetDialogSettings( { pos_x, pos_y } );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type   = "flashlight",
    res_id = 1,
    pos_x  = pos_x,
    pos_y  = pos_y,
    angle  = 0,
    scale  = cell.scale,
    lock   = true,
    hand   = {
      angle    = 0,
      offset_x = 0,
      offset_y = 0,
      type     = "move"
    }
  };

  local function func_close ()

    ui.screen.Lock( true );

    _f.Unsubscribe(
      "snowflakes",
      "Model_Match3_FieldActivated",
      func_close
    );

    _f.HideDialog();

    local func_OnGoalReceived;

    func_OnGoalReceived = function ()

      _f.Unsubscribe(
        "snowflakes",
        "Interface_PanelStageTop_GoalReceived",
        func_OnGoalReceived
      );      

      private.snowflakes_step3();

      ui.screen.Lock( false );

    end;

    _f.Subscribe(
      "snowflakes",
      "Interface_PanelStageTop_GoalReceived",
      func_OnGoalReceived
    );

  end;

  _f.LogKontagentEvent( "tutorial", { name = "tut_snowflakes_match" } );
  
  _f.Subscribe(
    "snowflakes",
    "Model_Match3_FieldActivated",
    func_close
  );

  _f.ShowDialogInstantly( "snowflakes_2", positions, hole );

end;
--------------------------------------------------------------------------------
function private.snowflakes_step3 ()
  
  _f.LogKontagentEvent( "tutorial", { name = "tut_snowflakes_matchdone" } );

  local match3_pos = GetObjPosByObj( "match" );

  local cell = _f.GetMatch3Cell();

  local pos_x = match3_pos[ 1 ] + 2 * cell.size * 8;
  local pos_y = match3_pos[ 2 ] + 2 * cell.size * 4;
  
  local dialog_settings = _f.GetDialogSettings( { pos_x, pos_y } );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type   = "flashlight",
    res_id = 15,
    pos_x  = pos_x,
    pos_y  = pos_y,
    angle  = 0,
    lock   = false,
    black  = true,
    input  = true
  };

  local function func_close ()

    _f.HideDialog( private.snowflakes_complete );

  end;

  _f.LogKontagentEvent( "tutorial", { name = "tut_snowflakes_find" } );
  _f.ShowDialogInstantly( "snowflakes_3", positions, hole, func_close );

end;
--------------------------------------------------------------------------------
function private.snowflakes_complete ()
  
  _f.LogKontagentEvent( "tutorial", { name = "tut_snowflakes_finddone" } );
  _f.Complete();
  _f.Unload();

end;
--------------------------------------------------------------------------------