-- name=int_tutorial_stage_16
--------------------------------------------------------------------------------
-- function DelegateScheme ()
--------------------------------------------------------------------------------
local _f = {
  GetMatch3Cell       = function () end,
  Subscribe           = function () end,
  Unsubscribe         = function () end,
  Backup              = function () end,
  Restore             = function () end,
  Complete            = function () end,
  LogKontagentEvent   = function () end,
  GetDialogSettings   = function () end,
  ShowDialogInstantly = function () end,
  HideDialog          = function () end,
  Unload              = function () end,
  AnimateSorcery      = function () end,
  GetSorceryIndex     = function () end,
  AddEntity           = function () end,
  GetEntityIndex      = function () end,
  ShowEntity          = function () end,
  HideEntity          = function () end,
  BackupBooster       = function () end,
  RestoreBooster      = function () end,
  CompleteOther       = function () end,
  GetStageItems       = function () end
};
--------------------------------------------------------------------------------
function public.__GetDelegateScheme ()

  return _f;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  --

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Start ( stage_name )

  private.stage_16_first();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.stage_16_first ()

  _f.Subscribe(
    "stage_16",
    "View_Stage_DialogTargetsHided",
    private.stage_16_next1
  );

  local sorcery_index = _f.GetSorceryIndex();

  _f.Backup(
    "stage_16",
    "sorcery_tap",
    ui.control.EventsGet( "obj_int_panel_stage_booster_"..sorcery_index ).func_on_click
  );

  ui.control.EventsSet(
    "obj_int_panel_stage_booster_"..sorcery_index,
    { func_on_click = function () end }
  );

  _f.BackupBooster( "stage_16", sorcery_index );

  _f.Backup(
    "stage_16",
    "settings_tap", 
    ui.control.EventsGet( "btn_int_panel_stage_settings_panel" ).func_on_click
  );

  ui.control.EventsSet(
    "btn_int_panel_stage_settings_panel",
    { func_on_click = function () end }
  );

end;
--------------------------------------------------------------------------------
function private.stage_16_next1 ()
  
  _f.Unsubscribe(
    "stage_16",
    "View_Stage_DialogTargetsHided",
    private.stage_16_next1
  );

  local index           = _f.GetSorceryIndex();
  local focus_obj       = "obj_int_panel_stage_booster_"..index;  
  local dialog_settings = _f.GetDialogSettings( focus_obj );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };

  local hole = 
  {
    type   = "flashlight",
    res_id = 3,
    focus  = focus_obj,
    scale  = 1,  
    lock   = true,
    black  = false,
    hand   = {
      angle     = 0.5 * math.pi,
      offset_x  = 0,
      offset_y  = 0,
      type      = "click"
    }
  };

  local func_close = function ()

    _f.HideDialog( function () private.stage_16_next2( index ); end );
    _f.HideEntity( "obj_int_panel_stage_booster_"..index );

  end;

  ui.control.EventsSet(
    "obj_int_panel_stage_booster_"..index,
    { func_on_click = func_close }
  );

  _f.BackupBooster( "stage_16", false, true );
  _f.ShowDialogInstantly( "stage_16_1", positions, hole );  
  _f.AddEntity( "stage_16", "sorcery", "booster" );
  _f.ShowEntity( "stage_16", "sorcery", "obj_int_panel_stage_booster_"..index );

end;
--------------------------------------------------------------------------------
function private.stage_16_next2 ( index )
  
  _f.Subscribe(
    "stage_16",
    "View_Boosters_BoosterUsed",
    private.stage_16_hide2
  );

  _f.Subscribe(
    "stage_16",
    "View_Boosters_BoosterUsageFailed",
    private.stage_16_reset
  );

  local items_data = _f.GetStageItems();

  -- Fake tap.
  common.MessageSend(
    Event_Match3_BoosterActivated,
    {
      name     = "sorcery",
      row      = items_data[ 1 ].row * 5,
      col      = items_data[ 1 ].col,
      vertical = true 
    }
  );
  
  _f.AnimateSorcery();
  
  local match3_pos = GetObjPosByObj( "match" );

  local cell = _f.GetMatch3Cell();

  local pos_x = match3_pos[ 1 ] + 2 * cell.size * items_data[ 1 ].col - cell.size;
  local pos_y = match3_pos[ 2 ] + 2 * cell.size * items_data[ 1 ].row * 4 + cell.size;

  local dialog_settings = _f.GetDialogSettings( { pos_x, pos_y } );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type   = "flashlight",
    res_id = 3,
    pos_x  = pos_x,
    pos_y  = pos_y,
    scale  = 0.8,
    angle  = 1.575707,
    lock   = true,
    wait   = true,
    black  = true,
    hand = {
      offset_x = 0,
      offset_y = 0,
      angle    = 0, 
      type     = "click"
    }
  };

  local func_close = function ()

    _f.HideDialog();

  end;

  _f.ShowDialogInstantly( "stage_16_2", positions, hole, func_close );

end;
--------------------------------------------------------------------------------
function private.stage_16_hide2 ( params )

  local booster_name = params.name;
  local index        = _f.GetSorceryIndex();
  
  if ( booster_name == "sorcery" ) then

    local func_hide = function ()
      
      local focus_obj       = "obj_int_panel_booster_sorcery_vertical";
      local dialog_settings = _f.GetDialogSettings( focus_obj );

      local positions = 
      {
        dlg_start       = dialog_settings.start_x,
        dlg_end         = dialog_settings.pos_x,
        dlg_pos_y       = dialog_settings.pos_y,
        character_scale = dialog_settings.ch_scale
      };

      local hole = 
      {
        type   = "flashlight",
        res_id = 3,
        focus  = focus_obj,
        scale  = 0.8,
        angle  = 1.575707,
        lock   = true,
        wait   = true,
        black  = true,
        hand = {
          offset_x = 0,
          offset_y = 0,
          angle    = 0, 
          type     = "click"
        }
      };

      local func_close = function ()

        --

      end;

      _f.ShowDialogInstantly( "stage_16_3", positions, hole, func_close );

      _f.Backup(
        "stage_16",
        "lock_tap",
        ui.control.TapGet( "obj_int_panel_booster_sorcery_lock" ).func_true
      );

      ui.control.TapSet(
        "obj_int_panel_booster_sorcery_lock", 
        { 
          func_true = ui.control.TapGet( focus_obj ).func_true 
        }
      );

    end;

    _f.HideDialog( func_hide );
    
    return;

  end;

  if ( booster_name == "sorcery_last" ) then

    _f.Unsubscribe(
      "stage_16",
      "View_Boosters_BoosterUsed",
      private.stage_16_hide2
    );

    _f.Unsubscribe(
      "stage_16",
      "View_Boosters_BoosterUsageFailed",
      private.stage_16_reset
    );

    local func_lock_tap = _f.Restore( "stage_16", "lock_tap" );

    ui.control.TapSet(
      "obj_int_panel_booster_sorcery_lock",
      { func_true = func_lock_tap }
    );

    local func_sorcery_tap = _f.Restore( "stage_16", "sorcery_tap" );

    ui.control.EventsSet(
      "obj_int_panel_stage_booster_"..index,
      { func_on_click = func_sorcery_tap }
    );

    _f.RestoreBooster( "stage_16", index );
    _f.RestoreBooster( "stage_16", false, true );

    local func_settings_tap = _f.Restore( "stage_16", "settings_tap" );

    ui.control.EventsSet(
      "btn_int_panel_stage_settings_panel",
      { func_on_click = func_settings_tap }
    );
    
    _f.HideDialog( private.stage_16_next3 );

    return;

  end;

end;
--------------------------------------------------------------------------------
function private.stage_16_next3 ()

  _f.Subscribe(
    "stage_16",
    "Model_Match3_FieldActivated",
    private.stage_16_hide3
  );

  local match3_pos = GetObjPosByObj( "match" );

  local cell = _f.GetMatch3Cell();

  local pos_x = match3_pos[ 1 ] + 2 * cell.size - cell.size;
  local pos_y = match3_pos[ 2 ] + 2 * cell.size * 3;

  local v_pref = 75;
  
  if ( ui.screen.IsLandscape() ) then
    
    v_pref = nil;
    
  end;
  
  local dialog_settings = _f.GetDialogSettings( { pos_x, pos_y }, v_pref );

  local positions = 
  {
    dlg_start       = dialog_settings.start_x,
    dlg_end         = dialog_settings.pos_x,
    dlg_pos_y       = dialog_settings.pos_y,
    character_scale = dialog_settings.ch_scale
  };
  
  local hole = 
  {
    type   = "flashlight",
    res_id = 12,
    pos_x  = pos_x,
    pos_y  = pos_y,
    scale  = cell.scale,
    angle  = 1.575707,
    hand   = {
      offset_x  = 0,
      offset_y  = 2 * cell.size,
      angle     = 1.575707,
      type      = "move"
    },
    black  = true,
    lock   = true
  };

  local func_close = function ()

    _f.HideDialog();

  end;

  _f.ShowDialogInstantly( "stage_16_4", positions, hole, func_close );
  
  _f.RestoreBooster(
    "stage_16_tap",
    "",
    false,
    "btn_int_dialog_stage_start_boosters"
  );

  _f.Complete();
  _f.CompleteOther( "stage_16_tap" );

end;
--------------------------------------------------------------------------------
function private.stage_16_hide3 ()

  _f.Unsubscribe(
    "stage_16",
    "Model_Match3_FieldActivated",
    private.stage_16_hide3
  );

  _f.HideDialog();

end;
--------------------------------------------------------------------------------
function private.stage_16_reset ()

  local items_data = _f.GetStageItems();

  -- Fake tap.
  common.MessageSend(
    Event_Match3_BoosterActivated,
    {
      name     = "sorcery",
      row      = items_data[ 1 ].row * 5,
      col      = items_data[ 1 ].col,
      vertical = true 
    }
  );
  
  _f.AnimateSorcery();

end;
--------------------------------------------------------------------------------