-- name=bot
--------------------------------------------------------------------------------
function public.__InitModule ()
  private.CSV_DELIMETER = ";";
  private.by_step = false;
  private.Info("Initialized.");
end;
--------------------------------------------------------------------------------
function public.Start ( attempt_count, section, stage, state, outfile_stat_all, outfile_stat_summary )
  
  private.attempt_count   = attempt_count;
  private.current_section = section;
  private.current_stage   = stage;
  private.current_state   = state;
  
  private.outfile_stat_all     = outfile_stat_all or "stat_"..private.current_section.."_"..private.current_stage.."_"..private.current_state;
  private.outfile_stat_summary = outfile_stat_summary or "stat_all";
  
  private.stat_all = {};
  
  private.shovel_delta = ne.config.GetData("match3bot.shovel_delta");
  
  private.StartMatch3();

  private.goals = {};
  local goals_state = private.GetAllGoals();  
  -- find main goal 1
  local found = false;
  for name, count in pairs( goals_state ) do
    if (string.find(name, "GainScore") == 1) then
    else 
      found = true;
    private.goals.main_1 = name;
      goals_state[name] = nil;
      break;
    end;  
  end
  if ( not found ) then
  private.goals.main_1 = "";
  end;
  
  -- find main goal 2
  found = false;
  for name, count in pairs( goals_state ) do
    if (string.find(name, "GainScore") == 1) then
  else 
    found = true;
    private.goals.main_2 = name;
    goals_state[name] = nil;
    break;
  end;  
  end
  if ( not found ) then
  private.goals.main_2 = "";
  end;
  
  -- find simple goal
  found = false;
  for name, count in pairs( goals_state ) do
    if (string.find(name, "GainScore") == 1) then
    found = true;
    private.goals.simple = name;
    goals_state[name] = nil;
    break;
  end;  
  end
  if ( not found ) then
    private.goals.simple = "";
  end;
  
end;
--------------------------------------------------------------------------------
function public.SetStageType ( stage_type )
  
  private.current_stage_type = stage_type;
  
end;
--------------------------------------------------------------------------------
function private.GetActiveGoals ()

  local target_list  = v_stage.GetCurrentTargets();
  local active_goals = {};
  
  for target_name, target_data in pairs( target_list ) do
    for index, target_entry in pairs( target_data ) do
      if ( not target_entry.done ) then
        local subname = target_entry.name or "";
        if ( subname ~= "" ) then subname = "_"..subname; end;
        table.insert( active_goals, target_name..subname );
      end;
    end;
  end;
  
  return active_goals;
  
end;
--------------------------------------------------------------------------------
function private.GetAllGoals()

  local goals_state = {};  
  local target_list  = v_stage.GetCurrentTargets();        
  for target_type, value in pairs( target_list ) do
    for target_index, target in pairs( value ) do  
      local subname = "";
    if (target.name) then subname = "_"..target.name; end;
    local name = target_type..subname;
    goals_state[name] = target.count - target.progress;  
    end;
  end;  
  return goals_state;

end;
--------------------------------------------------------------------------------
function private.OnEndMatch3()

  v_stage.Unsubscribe();
  
  local goals_state = private.GetAllGoals();  
  local active_goals = private.GetActiveGoals();    
  local win = #active_goals == 0;   
  
  --------------------------------------------------------------------------------
  local helper = {};
  function helper.CreateStr(name, value)
    helper.str_stat = value;
    helper.str_stat_header = name;
    helper.str_array = {};
  end;
  function helper.GetHeader()
    return helper.str_stat_header;
  end;
  function helper.GetStr()
    return helper.str_stat;
  end;
  function helper.GetArray()
    return helper.str_array;
  end;
  function helper.AddStr(name, value)
    helper.str_stat = helper.str_stat..private.CSV_DELIMETER..value;
    helper.str_stat_header = helper.str_stat_header..private.CSV_DELIMETER..name;
    helper.str_array[ name ] = value;
  end;
  function helper.FormatStr(name, value)
    if (win) then
      helper.str_stat = helper.str_stat..private.CSV_DELIMETER..value..private.CSV_DELIMETER;
      helper.str_array[ name.."_WIN" ] = value;
      helper.str_array[ name.."_FAIL" ] = "";
    else
      helper.str_stat = helper.str_stat..private.CSV_DELIMETER..private.CSV_DELIMETER..value;
      helper.str_array[ name.."_WIN" ] = "";
      helper.str_array[ name.."_FAIL" ] = value;
    end;
    helper.str_stat_header = helper.str_stat_header..private.CSV_DELIMETER..name.."_WIN"..private.CSV_DELIMETER..name.."_FAIL";
  end;
  --------------------------------------------------------------------------------
    
  -- 1) attempt
  helper.CreateStr("attempt_number", private.attempt_count);    
  -- 2) win
  if (win) then
    helper.AddStr("win", 1);
  else 
    helper.AddStr("win", 0);
  end;    
  -- 3,4) moves 
  helper.FormatStr("moves_at_end", m_stage.GetCurrentMoves() or 0);  
  -- 5-6) main goal 1
  helper.FormatStr("goal_main_1", goals_state[ private.goals.main_1 ] or "");  
  -- 7-8) main goal 2
  helper.FormatStr("goal_main_2", goals_state[ private.goals.main_2 ] or "");  
  -- 9-10) simple goal
  helper.FormatStr("goal_simple", goals_state[ private.goals.simple ] or "");
  
  -- 11) shovel main 1 
  -- 12) shovel main 2 
  -- 13) shovel simple 
  if (win) then
    helper.AddStr("shovel_goal_main_1", private.shovels[ private.goals.main_1 ] or ""); 
    helper.AddStr("shovel_goal_main_2", private.shovels[ private.goals.main_2 ] or ""); 
    helper.AddStr("shovel_goal_simple", private.shovels[ private.goals.simple ] or ""); 
  else 
    helper.AddStr("shovel_goal_main_1", "");
    helper.AddStr("shovel_goal_main_2", "");
    helper.AddStr("shovel_goal_simple", "");  
  end;
    
  if (not private.FileExists(private.outfile_stat_all)) then
    private.AppendToFile(private.outfile_stat_all, helper.GetHeader());
  end;
  private.AppendToFile(private.outfile_stat_all, helper.GetStr());
  
  table.insert( private.stat_all, helper.GetArray() );  
  private.attempt_count = private.attempt_count - 1;

end;
--------------------------------------------------------------------------------
function private.CalcDispersia(a)

  local dispersia = 0;

  local numbers = {};
  local numbers_sum = 0;
  for i = 1, #a do
    if (type(a[i]) == "number") then
      table.insert(numbers, a[i]);
      numbers_sum = numbers_sum + a[i];
    end;
  end;

  if ( #numbers > 1 ) then
    local m = numbers_sum/#numbers;

    local sum_of_pow_of_sub_with_m = 0;
    for i = 1, #numbers do
      sum_of_pow_of_sub_with_m = sum_of_pow_of_sub_with_m + (numbers[i] - m)^2;
    end;

    dispersia = sum_of_pow_of_sub_with_m / (#numbers - 1);
  end;
  return tonumber(string.format("%.5f", dispersia));
  
end;
--------------------------------------------------------------------------------
function private.CalcMediana(a)

  local mediana = 0;

  local numbers = {};
  for i = 1, #a do
    if (type(a[i]) == "number") then
      table.insert(numbers, a[i]);
    end;
  end;
  
  if ( #numbers > 1 ) then
    table.sort(numbers);

    local middle = #numbers / 2;
    if ( #numbers % 2 == 1 ) then
      mediana = numbers[ math.ceil(middle) ];
    else 
      mediana = (numbers[middle] + numbers[middle + 1]) / 2;
    end;
  end;

  return tonumber(string.format("%.5f", mediana));
  
end;
--------------------------------------------------------------------------------
function private.EndBotSession()

  local attempts = #private.stat_all;
  local wins = 0;
  local params = {
    "moves_at_end_WIN",
    "moves_at_end_FAIL",
    "goal_main_1_WIN",
    "goal_main_1_FAIL",
    "goal_main_2_WIN",
    "goal_main_2_FAIL",
    "goal_simple_WIN",
    "goal_simple_FAIL",
    "shovel_goal_main_1",
    "shovel_goal_main_2",
    "shovel_goal_simple"
  };
  local stat_arrays = {};
  for i = 1, #params do
    stat_arrays[params[i]] = {};
  end;

  for i = 1, #private.stat_all do
    wins = wins + private.stat_all[i].win;
    for param, value in pairs (stat_arrays) do
      table.insert( stat_arrays[ param ], private.stat_all[i][param] );
    end;
  end;
    
  local headers = "description;win_rate";
  local str_stat1 = private.current_section.."_"..tostring(private.current_stage).."_"..tostring(private.current_state).."_"..tostring(private.current_state).."(#"..tostring(#private.stat_all)..")";
  local str_stat2 = str_stat1.."_mediana"..private.CSV_DELIMETER;
  str_stat1 = str_stat1.."_dispersia"..private.CSV_DELIMETER..tostring(wins/attempts);
  
  for i = 1, #params do
    headers = headers..private.CSV_DELIMETER..params[i];
    str_stat1 = str_stat1..private.CSV_DELIMETER..tostring(private.CalcMediana(stat_arrays[params[i]]));
    str_stat2 = str_stat2..private.CSV_DELIMETER..tostring(private.CalcDispersia(stat_arrays[params[i]]));
  end;

  if (not private.FileExists(private.outfile_stat_summary)) then
    private.AppendToFile(private.outfile_stat_summary, headers);
  end;
  private.AppendToFile(private.outfile_stat_summary, str_stat1);
  private.AppendToFile(private.outfile_stat_summary, str_stat2);
    
  ne.dbg.Quit();
  private.Info("That's all!");
end;
--------------------------------------------------------------------------------
function private.FileExists(name)
   local f=io.open(name..".csv","r")
   if f~=nil then io.close(f) return true else return false end
end
--------------------------------------------------------------------------------
function private.AppendToFile (filename, text)

  local file = io.open(filename..".csv", "a");
  file:write(text.."\n");
  file:close();
  
end;
--------------------------------------------------------------------------------
function private.StartMatch3()

  if ( private.attempt_count > 0 ) then
  
    private.Info("StartMatch3: attempt "..private.attempt_count);
  
    private.shovels = {};
    dbg_match3.Set();
    
    ObjSet( "match", { ap = false, ap_loglevel = ne.config.GetData("match3bot.loglevel") } );
  
    ne.msg.Subscribe( "Command_Match3_ChainActivated", private.OnTargetUpdated );
    ne.msg.Subscribe( "Event_Match3_FieldShuffle", private.OnTargetUpdated );
    ne.msg.Subscribe( "Event_Match3_Explosion", private.OnMatch3Explosion );
        
    if ( not private.step ) then
      public.Step();
    end;

    ne.msg.Send( Command_Match3_DidBecomeActive );
  
  else 
    private.EndBotSession();
  end;
  
end;
--------------------------------------------------------------------------------
function private.FinishMatch()

  private.Info('FinishMatch');
  ne.msg.Unsubscribe( "Command_Match3_ChainActivated", private.OnTargetUpdated );
  ne.msg.Unsubscribe( "Event_Match3_FieldShuffle", private.OnTargetUpdated );
  ne.msg.Unsubscribe( "Event_Match3_Explosion", private.OnMatch3Explosion );
  ne.msg.Subscribe( "Event_Match3_FieldEstablished", private.OnFieldEstablished ); 
  ObjSet( "match", { ap = false } ); 
  
end;
--------------------------------------------------------------------------------
function private.OnMatch3Explosion()

  private.OnEndMatch3();
  private.StartMatch3();
  
end;
--------------------------------------------------------------------------------
function private.OnFieldEstablished( params )

  ne.msg.Unsubscribe( "Event_Match3_FieldEstablished", private.OnFieldEstablished );  
  
  private.OnEndMatch3();
  private.StartMatch3();
  
end;
--------------------------------------------------------------------------------
function private.CheckShovel() 
  
  local goals_state = private.GetAllGoals(); 
  for target, current_value in pairs( goals_state ) do
    if ( current_value > 0 and current_value <= private.shovel_delta[target] ) then
      private.shovels[target] = m_stage.GetCurrentMoves();
    end;  
  end;

end;
--------------------------------------------------------------------------------
function private.OnTargetUpdated( params )

  if (private.by_step) then 
    private.Info("OnTargetUpdated: next step");
    ObjSet( "match", { ap = false, ap_loglevel = ne.config.GetData("match3bot.loglevel") } );
    return; 
  else 
    public.Step(); 
  end;
  
end;
--------------------------------------------------------------------------------
function public.SwitchStep()
  private.by_step = not private.by_step;
end;
--------------------------------------------------------------------------------
function public.Step()

  local can_continue = m_stage.TryContinue();
  local can_finalize = m_stage.TryFinalize();
  
  local can_continue = ( can_continue ) and ( not can_finalize );
  
  if ( private.current_stage_type == "fight" ) then
    can_continue = ( not can_finalize );
  end;
  
  private.Info( "Can continue: "..tostring( can_continue ) );
  
  if ( can_continue ) then
  
    local active_goals = private.GetActiveGoals();  
    if ( #active_goals > 0 ) then      
      private.CheckShovel();
      ObjSet( "match", { ap = true, ap_goals = active_goals } );

    elseif ( private.current_stage_type == "fight" ) then
      
      local hero  = m_stage.GetCurrentHero();
      local enemy = m_stage.GetCurrentEnemy();
      
      private.Info( "Hero health: "..hero.health_current );
      private.Info( "Enemy health: "..enemy.health_current );
      
      ObjSet( "match", { ap = true } );
    else
      private.Info('no goals, enemy died');
      private.FinishMatch();
    end;
  
  else
    private.Info('Finalize');
    private.FinishMatch();
  end;
  
end;
--------------------------------------------------------------------------------
function private.Info( text )
  DbgTrace("[BOT] "..tostring(text));
end;
--------------------------------------------------------------------------------