-- name=dbg_match3_user
--------------------------------------------------------------------------------
-- function *** Init *** ()
--------------------------------------------------------------------------------
function public.__InitModule()
  
  private.match3 = {};
  
  private.current_section = 1;
  private.current_stage   = 1;
  private.current_state   = 1;

  ne.mod.LoadLazily( "assets/content/scripts/mod_common" );
  
  common.mod.Load({
    "mediator",
    "lib_rpc",
    "lib_db",
    "ui",
    "v_features",
    "v_log",
    "v_game",
    "v_interface",
    "v_tutorial",
    "v_screen",
    "v_hearts",
    "v_boosters",
    "v_player",
    "v_potions",
    "v_match3",
    "v_stage"
  });

  mediator.CheatDisableQueue();
  
  private.Info("Initialized.");
  
end;
--------------------------------------------------------------------------------
function public.__InitObjects()  
  ObjAttach("obj_dbg_match3_user", "ng_application");
  ne.obj.Set("btn_dbg_match3_user_reset",
  {
    event_mup = private.ResetMatch
  });
  ne.obj.Set("btn_dbg_match3_user_add_moves",
  {
    event_mup = private.AddMoves
  });
  ne.obj.Set("btn_dbg_match3_user_use_hammer",
  {
    event_mup = private.UseHammer
  });
end;
--------------------------------------------------------------------------------
function public.Start( settings )
  
  private.current_section = settings.section or private.current_section;
  private.current_stage   = settings.stage   or private.current_stage;
  private.current_state   = settings.state   or private.current_state;

  lib_db.AddLevels("stages_" .. private.current_section);
  lib_db.Database_Detach();
  
  private.LaunchMatch();
  
end;
--------------------------------------------------------------------------------
function private.LaunchMatch()
  private.PrepareStartMatch();
  private.SetMatch();
  private.StartMatch();
end;
--------------------------------------------------------------------------------
function private.PrepareStartMatch()
  
  ne.msg.Subscribe( "Event_Match3_Explosion", private.OnFieldExplosion );
  ne.msg.Subscribe( "Event_Match3_FieldEstablished", private.OnFieldEstablished ); 
  ne.msg.Subscribe( "Event_Match3_FieldActivated", private.OnFieldActivated ); 
  mediator.Unsubscribe( "Model_Stage_TargetUpdated", private.SetGoals );
  mediator.Unsubscribe( "Model_Stage_CellDestroyed", private.SetGoals );
  mediator.Unsubscribe( "Model_Stage_ItemDropped", private.SetGoals );
  
end;
--------------------------------------------------------------------------------
function private.SetMatch()
  
  local state   = private.current_state;
  local stage   = private.current_stage; 
  local section = private.current_section;
  
  local stage_data = m_stage.GetStaticData( "stage_"..stage, state );  
  
  private.match3 = {};
  
  private.match3.state   = state;
  private.match3.stage   = stage;
  private.match3.section = section;
      
  ObjDelete( "match" );
  ObjCreate( "match", "matchbox" );
  
  v_stage.Init();
  
  stage_data = m_stage.Start(
    v_player.Stats_GetPlayerData(),
    "stage_"..private.match3.stage,
    private.match3.state,
    stage_data.state[ state ].enemy,
    true
    );
  
  ObjSet( "match", 
  { 
    pos_x     = -30,
    pos_y     = 40,
    stage     = "stage_"..private.match3.stage, 
    state     = private.match3.state,
    boosters  = {},
    section   = private.match3.section
  } );
  
  ObjAttach( "match", "obj_matchbox" );
  
  private.SetMoves(0);
  private.SetScore();
  
  local goals   = m_stage.GetCurrentTargets();
  local package = ( m_stage.GetCurrentStageData().section or "section_1" );
  
  private.SetupGoals(goals);
  int_panel_stage_top.Goals_Setup( goals, package );
  
  int_panel_stage_top.Show();
  
  int_panel_stage_score.SetPos( 110, 120 );
  int_panel_stage_score.Show({r = 1.0, g = 1.0, b = 1.0});
  
end;
--------------------------------------------------------------------------------
function private.StartMatch()
  private.SetStatusMatch("Run");
  ne.msg.Send( Command_Match3_DidBecomeActive );
end;
--------------------------------------------------------------------------------
function private.EndMatch()
  
  ne.msg.Subscribe( "Event_Match3_Explosion", private.OnFieldExplosion );
  ne.msg.Subscribe( "Event_Match3_FieldEstablished", private.OnFieldEstablished ); 
  ne.msg.Subscribe( "Event_Match3_FieldActivated", private.OnFieldActivated ); 
  mediator.Unsubscribe( "Model_Stage_TargetUpdated", private.SetGoals );
  mediator.Unsubscribe( "Model_Stage_CellDestroyed", private.SetGoals );
  mediator.Unsubscribe( "Model_Stage_ItemDropped", private.SetGoals );
  
end;
--------------------------------------------------------------------------------
function private.ResetMatch()
  private.EndMatch();
  private.LaunchMatch();
end;
--------------------------------------------------------------------------------
function private.SetStatusMatch(status)
  if (status == "Lose" or status == "Win") then
    ne.obj.Set("obj_matchbox", {input = false});
    ne.obj.Set("btn_dbg_match3_user_use_hammer", {input = false});
  else
    ne.obj.Set("obj_matchbox", {input = true});
    ne.obj.Set("btn_dbg_match3_user_use_hammer", {input = true});
  end;
  ne.obj.Set("txt_dbg_match3_user_action", { text = status });
end;
--------------------------------------------------------------------------------
--function *** ()
--------------------------------------------------------------------------------
function private.OnFieldActivated()
  private.SetMoves(1);
end;
--------------------------------------------------------------------------------
function private.OnFieldEstablished()
  private.Info("Match3_FieldEstablished");
  private.SetGoals();
  if (private.IsGoalsCompleted()) then
    private.SetStatusMatch("Win");
    private.EndMatch();
  end;
  if (m_stage.GetCurrentMoves() <= 0) then
    private.SetStatusMatch("Lose");
    private.EndMatch();
  end;
end;
--------------------------------------------------------------------------------
function private.OnFieldExplosion()
  private.Info("Match3_FieldExplosion");
  private.SetStatusMatch("Lose");
end;
--------------------------------------------------------------------------------
function private.IsGoalsCompleted()
  local result = true;
  local goals = m_stage.GetCurrentTargets();
  for goal_type, goal_data in pairs( goals ) do
    for goal_index, goal_value in pairs( goal_data ) do
      result = (goal_value.done or goal_value.progress >= goal_value.count) and result;
    end;
  end;
  return result;
end;
--------------------------------------------------------------------------------
--function *** ()
--------------------------------------------------------------------------------
function private.SetupGoals(goals)
  private.goals = {
    number   = 0,
    data     = {},
    complete = 0
  };
  for goal_type, goal_data in pairs( goals ) do
    for goal_index, goal_value in pairs( goal_data ) do
      private.goals.number = math.max( private.goals.number, goal_index );
      private.goals.data[ goal_index ] = private.goals.data[ goal_index ] or {};
      private.goals.data[ goal_index ].done = goal_value.done or false;
    end;
  end;
end;
--------------------------------------------------------------------------------
function private.Goals_Progress ( goal_data )
  if ( goal_data.type == "GainScore" ) then
    return ui.text.FormatNumber( goal_data.count, true );
  end;
  return ui.text.FormatNumber( goal_data.progress ).."/"..ui.text.FormatNumber( goal_data.count );
end;
--------------------------------------------------------------------------------
function private.SetGoals()
  local goals = m_stage.GetCurrentTargets();
  for goal_type, goal_data in pairs( goals ) do
    for goal_index, goal_value in pairs( goal_data ) do
      private.goals.number = math.max( private.goals.number, goal_index );
      private.goals.data[ goal_index ] = private.goals.data[ goal_index ] or {};
      ui.obj.Set(
        "anm_int_panel_stage_top_goals_goal"..goal_index,
        {
          tt_count = private.Goals_Progress( goal_value )
        }
      );
      local done = goal_value.done or goal_value.progress >= goal_value.count;
      if (done and not private.goals.data[ goal_index ].done) then
        private.goals.complete = private.goals.complete + 1;
        private.goals.data[ goal_index ].done = true;
        ui.anim.Play(
          "anm_int_panel_stage_top_goals_goal"..goal_index,
          "complete_goal",
          function () end
        );
      end;
    end;
  end;
  private.SetScore();
end;
--------------------------------------------------------------------------------
function private.SetMoves(delta)
  int_panel_stage_top.Moves_Update(m_stage.GetCurrentMoves(), delta);
end;
--------------------------------------------------------------------------------
function private.AddMoves(delta)
  if (private.IsGoalsCompleted()) then
    return;
  end;
  m_stage.TryUseContinue("extramoves");
  private.SetMoves(5);
  private.SetStatusMatch("Run");
end;
--------------------------------------------------------------------------------
function private.SetScore()
  int_panel_stage_score.Update(m_stage.GetCurrentScore());
end;
--------------------------------------------------------------------------------
function private.UseHammer()
  ne.msg.Send(
  Event_Match3_BoosterActivated,
  {
    name     = "hammer",
    row      = nil,
    col      = nil,
    vertical = nil }
  );
end;
--------------------------------------------------------------------------------
--function *** ()
--------------------------------------------------------------------------------
function private.Info( text )
  DbgTrace("[dbg_match3_user] "..tostring(text));
end;
--------------------------------------------------------------------------------