-- name=loader
--------------------------------------------------------------------------------
function public.Start ()

  math.randomseed( os.time() );
  
  private.app_url =
  {
    ios    = "https://itunes.apple.com/app/id905121703",
    google = "https://play.google.com/store/apps/details?id=com.bigfishgames.sirmatchgoogle",
    amazon = "http://www.amazon.com/gp/mas/dl/android?p=com.bigfishgames.sirmatchamazon"
  };

  private.obj =
  {
    application = "ng_application",
    screen      = "scr_loader",
    logo =
    {
      container = "obj_loader_logo",
      pubdev    = "spr_loader_logo_pubdev",
      pubdev_1  = "spr_loader_logo_pubdev_1"
    },
    splash =
    {
      container  = "obj_loader_splash",
      background = "spr_loader_splash_background",
      scene      = "anm_loader_splash_scene",
      planet     = "anm_loader_splash_scene_ground",
      gamelogo   = "anm_loader_splash_scene_logo",
      progress   = "anm_loader_splash_scene_progress",
      message    = 
      { 
        conatiner = "obj_loader_splash_scene_message",
        animation = "anm_loader_splash_scene_message",
      }
    },
    message =
    {
      container = "obj_loader_message"
    }
  };
  
  local is_landscape = ( GetAppWidth() > GetAppHeight() );
  private.orientation = ( is_landscape and "landscape" ) or "portrait";

  local platform = GetPlatform();
  
  if ( platform == "pc" and IsEditor() and debug ) then
  
    private.is_debug = true;
    DbgTrace( "[ LOADER ] Debug mode" );
  
    ne.mod.LoadLazily( "assets/content/scripts/mod_dbg_strict"   );
    ne.mod.LoadLazily( "assets/content/scripts/mod_dbg_safe_api" );
    ne.mod.LoadLazily( "assets/content/scripts/mod_dbg_simulate" );
    
    local simulate_platform = ConfigGetInfo( "simulate.platform" ) or "";
    local simulate_vendor = ConfigGetInfo( "simulate.vendor" ) or "";
  
    if( simulate_platform ~= "" and simulate_vendor ~= "" ) then
      dbg_simulate.Setup( simulate_platform, simulate_vendor );
      platform = GetPlatform();
    end;

  end;
  
  private.is_pc      = ( platform == "pc" );
  private.is_android = ( platform == "android" );
  private.is_ios     = ( platform == "ios" );

  private.free_space_needed = 200;
  if ( private.is_pc ) then private.free_space_needed = 1000; end;

  private.is_online    = true;
  private.request_list = {};

  if ( private.is_android ) then
    ne.msg.Subscribe( "Event_Purchase_ProductInformationReady", function ()
      private.product_info_ready = true;
    end );
  end;

  private.LoadStandardCursors();
  
  private.SystemMessages_Init( platform );
  
  private.PartsLoad();

  private.SetObjects();

  if ( not wiz.HasFreeDiskspace() ) then

    local func_continue = function ()
      ne.msg.Send( Event_RestartEngine );
    end;
    private.SystemDialog_Show( "nofreespace", func_continue );

  end;
  
  local maintenance = parts.application.GetmaintenanceState();
  
  if ( maintenance ~= 0 ) then
    
    local msg_type = "maintenance_"..maintenance;
    
    private.LoadEvent_SetState( msg_type, true, { maintenance = maintenance } );
    
  end;
  
  if ( private.is_ios ) then wiz.bfg.ccs.HideButton(); end;
  
  -- Loader_TryContinue execute condition
  private.can_continue = false;
  
  -- NOTE: parts_app_wasrestarted created by PARTS in _G
  if ( not parts_app_wasrestarted ) then

    private.ShowLogoPubDev();

  else

    private.Splash_Show();

  end;

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function private.GetUserId ()
  
  local user_id = AuthGetUserId();
  
  if ( user_id ~= "" ) then
    
    return tonumber( user_id );
    
  end;

  return 0;
  
end;
--------------------------------------------------------------------------------
function public.IsProductInfoReady ()

  return private.product_info_ready;

end;
--------------------------------------------------------------------------------
function public.Content_GetPacketsState ()

  return private.content.packets.applied;

end;
--------------------------------------------------------------------------------
function public.Content_SetPacketState ( packet_name, packet_state )

  private.content.packets.applied = private.content.packets.applied or {};
  if ( not private.content.packets.applied[ packet_name ] or
       private.content.packets.applied[ packet_name ] ~= packet_state ) then
    private.content.packets.applied[ packet_name ] = packet_state;
  end;

end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function private.LoadStandardCursors ()

  SetSysCursor( GetSysCursor() );
 
  local count = 0;

  local CursorLoad = function ( cur_name, cur_path, cur_idx )

    _G[ cur_name ] = count;
    
    CursorLoad( cur_path, cur_idx );
    
    count = count + 1;
  
  end;

  CursorLoad( "CURSOR_NULL",    "", 0 );
  CursorLoad( "CURSOR_DEFAULT", "assets/cursors/cursor_default_a", 2 );
  CursorLoad( "CURSOR_HAND",    "assets/cursors/cursor_hand_a",    4 );

  SetCursor( CURSOR_DEFAULT );

  --[[CursorLoad( "CURSOR_NULL",    "",                              0 );
  CursorLoad( "CURSOR_DEFAULT", "assets/cursors/cursor",         2 );
  CursorLoad( "CURSOR_LOUPE",   "assets/cursors/cursor_loupe",  10 );
  CursorLoad( "CURSOR_GET",     "assets/cursors/cursor_get",     4 );
  CursorLoad( "CURSOR_HAND",    "assets/cursors/cursor_hand",    4 );
  CursorLoad( "CURSOR_USE",     "assets/cursors/cursor_use",     4 );
  CursorLoad( "CURSOR_DOWN",    "assets/cursors/cursor_down",   10 );
  CursorLoad( "CURSOR_UP",      "assets/cursors/cursor_up",     10 );
  CursorLoad( "CURSOR_LEFT",    "assets/cursors/cursor_left",   10 );
  CursorLoad( "CURSOR_RIGHT",   "assets/cursors/cursor_right",  10 );
  CursorLoad( "CURSOR_FINGER",  "assets/cursors/cursor_finger", 10 );]]

end;
--------------------------------------------------------------------------------
function private.PartsLoad ()

  private.load_events =
  {
    {
      name     = "oldclientversion",
      active   = false,
      need_tap = true,
      auto_msg = false,
      handler  = private.Handler_OldClientVersion
    },
    {
      name     = "statecorrupted", -- broken save
      active   = false,
      need_tap = true,
      auto_msg = true,
      handler  = private.Handler_StateCorrupted
    },
    {
      name     = "nointernetconnection", -- content obsolete + offline
      active   = false,
      need_tap = true,
      auto_msg = true,
      handler  = private.Handler_NoInternetConnection
    },
    {
      name     = "nofreespace", -- content obsolete + no free space
      active   = false,
      need_tap = true,
      auto_msg = true,
      handler  = private.Handler_NoFreeSpace
    },
    {
      name     = "contentobsolete",
      active   = false,
      need_tap = true,
      auto_msg = true,
      handler  = private.Handler_ContentObsolete
    },
    {
      name     = "prioritysession",
      active   = false,
      need_tap = true,
      auto_msg = true,
      handler  = private.Handler_PrioritySession
    },
    {
      name     = "conflictstorage",
      active   = false,
      need_tap = true,
      auto_msg = false,
      handler  = private.Handler_ConflictStorage
    },
    {
      name     = "newversion",
      active   = false,
      need_tap = true,
      auto_msg = false,
      handler  = private.Handler_NewVersion
    },
    {
      name     = "loadingcomplete",
      active   = false,
      need_tap = false,
      auto_msg = false,
      handler  = private.Handler_LoadModules
    },
    {
      name     = "maintenance_3",
      active   = false,
      need_tap = true,
      auto_msg = false,
      handler  = private.Handler_Maintenance
    },
    {
      name     = "maintenance_1",
      active   = false,
      need_tap = true,
      auto_msg = false,
      handler  = private.Handler_Maintenance
    },
    {
      name     = "maintenance_2",
      active   = false,
      need_tap = true,
      auto_msg = false,
      handler  = private.Handler_Maintenance
    }
  };

  private.message_handlers = {};

  private.PartsWaitForResult(
    "Event_Application_GameVersionOutdated",
    function ()
      private.LoadEvent_SetState( "oldclientversion", true );
    end
  );
  
  private.PartsWaitForResult(
    "Event_Application_ClientVersionOutdated",
    function ()
      private.LoadEvent_SetState( "oldclientversion", true );
    end
  );
  
  private.PartsWaitForResult(
    "Event_Application_StateCorrupted",
    function ( params )
      private.LoadEvent_SetState( "statecorrupted", true );
      
      private.Trace( "On save corrupted." );
      local onSetSaveCorrupted;
      onSetSaveCorrupted = function ()
        private.Trace( "On save corrupted. Stop heartbeats." );
        parts.net.heartbeat.Stop();
        ne.msg.Unsubscribe( "Server_save_onSetSaveCorrupted", onSetSaveCorrupted );
      end;
      ne.msg.Subscribe( "Server_save_onSetSaveCorrupted", onSetSaveCorrupted );
      NetInvokeServerMethod(
        "save",
        "setSaveCorrupted",
        { info = params },
        1
      );
    end
  );
  
  private.PartsWaitForResult(
    "Event_Application_LoadingComplete",
    function ()
      private.Parts_TryCompleteLoading( false );
    end
  );

  private.PartsWaitForResult(
    "Event_Application_InternetError",
    function ( params )
      private.is_online = false;
      if ( params and params.error == "KickedByOthers" ) then
        parts.net.heartbeat.Pause(true);
        private.LoadEvent_SetState( "prioritysession", true );
      else
        private.Parts_TryReconnect();
      end;
    end
  );
  
  private.PartsWaitForResult(
    "Event_Storage_Conflict",
    function ( params )
      private.LoadEvent_SetState( "conflictstorage", true, params );
    end
  );
  
  private.PartsWaitForResult(
    "Event_Application_RunMigrations",
    function ( params )
      parts.storage.data.Write( "sml/need_run_migration", params.version );
    end
  );
  
  private.PartsWaitForResult(
    "Event_Auth_Maintenance",
    function ( params )
      
      if ( params.maintenance == 0 ) then return; end;
     
      private.LoadEvent_SetState( "maintenance_"..params.maintenance, true, params );
      
    end
  );
  
  ne.msg.Subscribe( Event_Auth_SignIn,           private.AuthSignIn );
  ne.msg.Send( Event_Menu_LoadingComplete );
  
  local platform = GetPlatform();
  if ( platform == "android" ) then
    platform = wiz.Android.GetPlatform();
  end;

  private.WaitingForAnswer(
    "version", 
    "IsNewClientVersionAvailable",
    { platform = platform },
    function ( params )
      if ( params.available ) then
        private.LoadEvent_SetState( "newversion", true );
      end;
    end
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.LoadEvent_SetState ( event_name, state, params )

  for _, event in pairs( private.load_events ) do
  
    if ( event.name == event_name ) then
      
      if ( state and params ) then
        event.params = params;
      end;
      
      event.active = state;
      break;
      
    end;
  
  end;
  
  if ( state ) then
    local msg = "LoadEvent_SetState: event '%s' activated";
    private.Loader_TryContinue( msg:format( event_name ) );
  end;

end;
--------------------------------------------------------------------------------
function private.LoadEvent_GetData ( event_name )

  for _, event in pairs( private.load_events ) do
  
    if ( event.name == event_name ) then return event; end;
  
  end;
  
  return nil;

end;
--------------------------------------------------------------------------------
function private.LoadEvent_HandleActive ()

  local handling_event = nil;
  for _, event in pairs( private.load_events ) do
  
    if ( event.active ) then
      handling_event = event;
      break;
    end;
  
  end;
  
  if ( handling_event ) then
  
    private.Trace( "LoadEvent_HandleActive: "..handling_event.name );
  
    handling_event.active = false;
    
    local func_handler;
    if ( handling_event.auto_msg ) then
    
      func_handler = function ()
      
        private.SystemDialog_Show(
          handling_event.name,
          function ()
            handling_event.handler( handling_event.params );
            local msg = "LoadEvent_HandleActive: handled '%s'";
            private.Loader_TryContinue( msg:format( handling_event.name ) );
          end
        );
      
      end;
      
    else
    
      func_handler = function ()
        handling_event.handler( handling_event.params );
      end;
    
    end;
    
    if ( handling_event.need_tap ) then
      private.TapToContinue_Show( func_handler );
    else
      func_handler();
    end;
  
  else
  
    private.Trace( "LoadEvent_HandleActive: none" );
  
  end;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Handler_Maintenance ( params )
  
  local func_continue = nil;
  local msg_type = "maintenance_"..params.maintenance;
  
  if ( params.maintenance == 3 ) then
    func_continue = function ()
      ne.msg.Send( Event_RestartEngine );
    end;
  end;
  
  private.SystemDialog_Show( msg_type, func_continue );
  
end;
--------------------------------------------------------------------------------
function private.Handler_OldClientVersion ()

  local func_continue = function ()
  
    local platform = GetPlatform();

    if ( platform == "pc" ) then
      
      wiz.Quit();
      return;
      
    else
      
      if ( platform == "android" ) then
        platform = wiz.Android.GetPlatform();
      end;
      
      wiz.bfg.OpenUrl( private.app_url[ platform ] );
      
      private.LoadEvent_SetState( "oldclientversion", true );
      
    end;
    
  end;
  
  local user_id = private.GetUserId();
  DbgTrace( "User ID: "..user_id );

  private.SystemDialog_Show( "oldclientversion", func_continue );

end;
--------------------------------------------------------------------------------
function private.Handler_StateCorrupted ()
  
  wiz.RemoveLocalSave();

end;
--------------------------------------------------------------------------------
function private.Handler_NoInternetConnection ()

  if ( not private.is_online ) then
  
    private.Parts_TryReconnect();
    private.LoadEvent_SetState( "nointernetconnection", true );
  
  end;

end;
--------------------------------------------------------------------------------
function private.Handler_NoFreeSpace ()

  RestartApplication();

end;
--------------------------------------------------------------------------------
function private.Handler_ContentObsolete ()

  private.PartsUnsubscribeAll();
  private.Content_StartDownload();

end;
--------------------------------------------------------------------------------
function private.Handler_PrioritySession ()

  RestartApplication();

end;
--------------------------------------------------------------------------------
function private.Handler_ConflictStorage ( params )

  local clientTime = params.clientSaveTime;
  local serverTime = params.serverSaveTime;
  
  local func_message_main, func_message_local, func_message_server;

  func_message_main = function ()
    
    local clientInfoData = private.PartsStorageConflict_GetClientInfo();
    local serverInfoData = private.PartsStorageConflict_GetServerInfo();
    
    if ( not serverInfoData ) then
    
      private.SystemDialog_Show(
        "conflictstorage_force_local",
        function ()

          parts.storage.ResolveConflict( "local" );
          private.Loader_TryContinue( "Handler_ConflictStorage: force local" );

        end
      );
      return;
    
    end;
    
    local clientTimeFormatted = os.date( "%Y-%m-%d", clientTime );
    local serverTimeFormatted = os.date( "%Y-%m-%d", serverTime );
    
    if clientTimeFormatted == serverTimeFormatted then
      clientTimeFormatted = os.date( "%H:%M:%S", clientTime );
      serverTimeFormatted = os.date( "%H:%M:%S", serverTime );
    end;

    private.SystemDialog_Show( "conflictstorage", 
      func_message_local, 
      func_message_server,
      { clientTimeFormatted, clientInfoData },
      { serverTimeFormatted, serverInfoData }
    );

  end;

  func_message_local = function ()

    private.SystemDialog_Show(
      "conflictstorage_local",
      function ()

        parts.storage.ResolveConflict( "local" );
        parts.application.Save(true);
        private.Loader_TryContinue( "Handler_ConflictStorage: local" );

      end,
      func_message_main
    );

  end;

  func_message_server = function ()

    private.SystemDialog_Show(
      "conflictstorage_server",
      function ()

        parts.storage.ResolveConflict( "server" );
        parts.application.Save(true);
        private.Loader_TryContinue( "Handler_ConflictStorage: server" );

      end,
      func_message_main
    );

  end;

  func_message_main();

end;
--------------------------------------------------------------------------------
function private.Handler_NewVersion ()

  local func_download = function ()

    local platform = GetPlatform();

    if ( platform == "pc" ) then
      
      wiz.Quit();
      return;
      
    else
      
      if ( platform == "android" ) then
        platform = wiz.Android.GetPlatform();
      end;
      
      private.LoadEvent_SetState( "newversion", true );
      
      wiz.bfg.OpenUrl( private.app_url[ platform ] );
      
    end;

  end;

  local func_continue = function ()
    private.Loader_TryContinue( "Handler_NewVersion: continue" );
  end;

  private.SystemDialog_Show( "newversion", func_download, func_continue );

end;
--------------------------------------------------------------------------------
function private.Handler_LoadModules ()
  
  private.Content_ApplyPackets();
  private.Module_LoadChain();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Parts_TryCompleteLoading ( reconnect )

  private.Trace( "Parts_TryCompleteLoading" );
  
  local version = VersionGetComponentVersion( "game" );
  DbgTrace( "[ VERSION ] "..tostring( version ) );
  DbgTrace( "[ USER ]    "..tostring( private.GetUserId() ) );
  DbgTrace( "[ SERVER ]  "..ConfigGetInfo( "server.host" ) );
  
  local obsolete = private.Content_IsObsolete();

  if ( obsolete.result ) then
    
    private.Trace( "Parts_TryCompleteLoading: content" );

    if ( not private.is_online ) then
    
      private.Trace( "Parts_TryCompleteLoading: content: offline event" );
      private.LoadEvent_SetState( "nointernetconnection", true );

    else

      if ( parts_app_wasrestarted or reconnect ) then
      
        private.Trace( "Parts_TryCompleteLoading: content: reconnect" );
        
        -- DG: skip tap and message
        local data = private.LoadEvent_GetData( "contentobsolete" );
        if ( data ) then
          data.need_tap = false;
          data.auto_msg = false;
        end;

      end;
      
      private.Trace( "Parts_TryCompleteLoading: content: event" );
      private.LoadEvent_SetState( "contentobsolete", true );

    end;
    
    return;

  elseif ( obsolete.reason ) and ( obsolete.reason == "nofreespace" ) then
  
    private.Trace( "Parts_TryCompleteLoading: nofreespace" );
    private.LoadEvent_SetState( "nofreespace", true );
    return;
    
  end;
  
  --if not ( reconnect ) then
  
    private.Trace( "Parts_TryCompleteLoading: loadingcomplete" );
    private.LoadEvent_SetState( "loadingcomplete", true );
  
  --end;
  
end;
--------------------------------------------------------------------------------
function private.Parts_TryReconnect ()

  private.PartsWaitForResult( "Event_HeartBeat", private.Parts_OnReconnect );

end;
--------------------------------------------------------------------------------
function private.Parts_OnReconnect ()

  private.is_online = true;
  private.Info_Update();
  private.Parts_TryCompleteLoading( true );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.AuthSignIn ()
  
  local events_table = { 
    name = "create_raveid",
    default1 = "social",
    default2 = "account_profile",
    default3 = "",
    params = { 
      v = 0,
      l = 1, 
      login_channel = ( wiz.social.GetChannel() or {} ).login_channel
    }
  };

  events_table.params = private.Kontagent_AdditionalDetails( events_table or {} );
  
  wiz.bfg.kontagent.LogEvent( events_table );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.SystemMessages_Init ( platform )

  local system_message_type = {
    ios     = "native",
    pc      = "ingame",
    android = "ingame" 
  };

  private.sysmsg_type    = system_message_type[ platform ];
  private.messages_queue = {};

end;
--------------------------------------------------------------------------------
function private.SystemMessages_Add ( message_type, message_handler )

  for i = #private.messages_queue, 1, - 1 do

    if ( private.messages_queue[ i ].type == message_type ) then

      return;

    end;

  end;   

  table.insert( private.messages_queue, { type = message_type, handler = message_handler } );

end;
--------------------------------------------------------------------------------
function private.SystemMessages_ShowNext ()

  if ( #private.messages_queue > 0 ) then

    local message = table.remove( private.messages_queue, #private.messages_queue );

    if ( message.handler ) then

      message.handler();

    end;

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.PartsWaitForResult ( message_type, message_handler )

  if ( private.message_handlers[ message_type ] ) then

    private.Trace( "Error: handler for < "..message_type.." > already exist!" );

    return;

  end;

  local func_handler = function ( msg_type, msg_params )

    ne.msg.Unsubscribe( message_type, private.message_handlers[ message_type ] );

    if ( message_handler ) then

      message_handler( msg_params );

    end;

  end;

  private.message_handlers[ message_type ] = func_handler;

  ne.msg.Subscribe( message_type, func_handler );

end;
--------------------------------------------------------------------------------
function private.PartsUnsubscribeAll ()

  for message_name, message_handler in pairs( private.message_handlers ) do

    ne.msg.Unsubscribe( message_name, message_handler );

  end;
  
  ne.msg.Unsubscribe( Event_Auth_SignIn, private.AuthSignIn );

  private.message_handlers = {};

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.PartsStorageConflict_GetClientInfo ()
  
  local _, player_info_string = parts.storage.data.Read( "parts/cloudstorage/player/info" );
  local player_info = wiz.json.Decode( player_info_string );
  
  local query;
  
  if not ( player_info ) then
    query = {
      query = "find_all",
      collection = "player/info"
    };
    player_info = DbExecuteQuery( query ).result or {};
  end;
  
  local client_data = {};
  
  client_data.stage_last = private.PartsStorageConflict_GetStage( player_info[ 1 ].stage_last );
  
  local _, sections_string = parts.storage.data.Read( "parts/cloudstorage/player/sections" );
  local sections = wiz.json.Decode( sections_string );

  if not ( sections ) then
    query = {
      query = "find_all",
      collection = "player/sections"
    };
    sections = DbExecuteQuery( query ).result or {};
  end;
 
  client_data.section = private.PartsStorageConflict_GetSection( sections );
  client_data.gold = player_info[ 1 ].gold;
  
  return client_data;
  
end;
--------------------------------------------------------------------------------
function private.PartsStorageConflict_GetServerInfo ()
  
  local conflictData = parts.storage.GetConflictSave();
  if not ( conflictData and 
           conflictData[ "player/info" ] and 
           conflictData[ "player/sections" ] ) then
    
    return nil;
    
  end;
  
  local player_info = wiz.json.Decode( conflictData[ "player/info" ] );
  if not ( player_info ) then return nil; end;
  
  local server_data = {};
  
  server_data.stage_last = private.PartsStorageConflict_GetStage( player_info[ 1 ].stage_last );

  local sections = wiz.json.Decode( conflictData[ "player/sections" ] );
  if not ( sections ) then return nil; end;
  
  server_data.section = private.PartsStorageConflict_GetSection( sections );
  server_data.gold = player_info[ 1 ].gold;
  
  return server_data;
  
end;
--------------------------------------------------------------------------------
function private.PartsStorageConflict_GetStage ( stage_name )
  
  local stage_number = tonumber( stage_name:match( "_(%d+)" ) );
  
  if ( stage_name:match( "_bonus" ) ) then
    return "bonus";
  end;
  
  if ( stage_number > 110 ) then
    local section_number, residue = math.modf( stage_number / 110 );
    if ( residue == 0 ) then
      section_number = section_number - 1;
    end;
    
    stage_number = stage_number - 110 * section_number;
  end;
  
  return stage_number;
  
end;
--------------------------------------------------------------------------------
function private.PartsStorageConflict_GetSection ( sections )
  
  local section = "section_1";
  
  for s = #sections, 1, -1  do
    if ( sections[ s ].enabled ) then
      section = sections[ s ].name;
      break;
    end;
  end;
  
  --section = tonumber( section:match( "_(%d+)" ) );
  
  return section;
  
end;
--------------------------------------------------------------------------------
function private.Loader_TryContinue ( source )

  private.Trace( "Loader_TryContinue: from "..source );

  if not ( private.can_continue ) then

    private.Trace( "Loader_TryContinue: splash not finished" );
    return;

  end;
  
  if ( not private.is_on_screen_system_msg ) then
  
    private.LoadEvent_HandleActive();
  
  end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function public.Module_LoadChain ( only_depend, force, func_finish )
  
  private.Module_LoadChain( only_depend, force, func_finish );
  
end;
--------------------------------------------------------------------------------
function private.Module_Load ( module_name, is_instant )
  
  --DbgTrace( "Load: "..module_name )

  local path = "assets/content/scripts/mod_"..module_name;
  
  if ( is_instant ) then

    ne.mod.Load( path );
    
  else
  
    ne.mod.LoadLazily( path );
    
  end;

end;
--------------------------------------------------------------------------------
function private.Module_LoadChain ( only_depend, force, func_finish )

  --- Profiling ----------------------------------------------------------------
  local profiling = {};

  if ( private.is_debug ) then

    -- JK: profiling start
    profiling = {
      start   = os.clock(),
      stop    = 0,
      load    = {},
      init    = {}
    };

  end;

  private.Trace( "Modules_Load: started." );

  private.modulesloaded = false;

  --- List modules to load -----------------------------------------------------
  local modules = {
    core = {
      "common",
      "mediator",
      "lib_db",
      "lib_rpc",
      "lib_content",
      "ui",
      "v_features"
    },
    game = {
      "v_game",
      "v_interface",
      "v_screen",
      "v_log"
    },
    depend = {
      { name = "v_achievements", condition = "default" },
      { name = "v_player",       condition = "default" },
      { name = "v_map",          condition = "stage_1" },
      { name = "v_mines",        condition = "stage_1" }, -- WHY? Change to stage_11
      { name = "v_builds",       condition = "stage_1" }, -- WHY? Change to stage_11
      { name = "v_characters",   condition = "stage_1" },
      { name = "v_enemies",      condition = "stage_1" },
      { name = "v_chests",       condition = "stage_1" },
      { name = "v_fights",       condition = "stage_1" },
      { name = "v_spells",       condition = "stage_1" },
      { name = "v_friends_new",  condition = "stage_1" },
      { name = "v_stage",        condition = "default" },
      { name = "v_match3",       condition = "default" },
      { name = "v_sections",     condition = "default" },
      { name = "v_settings",     condition = "default" },
      { name = "v_boosters",     condition = "default" },
      { name = "v_tutorial",     condition = "default" },
      { name = "v_shop",         condition = "default" },
      { name = "v_items",        condition = "default" },
      { name = "v_story",        condition = "default" },
      { name = "v_quests",       condition = "stage_1" },
      { name = "v_dailyquest",   condition = "stage_6" },
      { name = "v_dailybonus",   condition = "stage_1" },
      { name = "v_social",       condition = "default" },
      { name = "v_hearts",       condition = "default" },
      { name = "v_fireflies",    condition = "stage_11" },
      { name = "v_minievent",    condition = "stage_16" },
      { name = "v_carriers",     condition = "stage_1" }, -- WHY? Change to stage_3
      { name = "v_potions",      condition = "default" }  -- WHY? Change to stage_1
    }
  };

  --- Define config ------------------------------------------------------------
  local total    = 0;
  local progress = 0;
  local cursor   = 1;

  local at_once = {
    load = 10,
    init = 10
  };

  local step       = 1;
  local load_chain = {};

  local need_init = {
    core   = true,
    game   = true,
    depend = true
  };

  for m = #modules.depend, 1, - 1 do
    
    local module_data = modules.depend[ m ];
    
    if ( _G[ module_data.name ] or
         ( not force and
         not private.Depend_IsConditionMet( module_data.condition, module_data.name ) ) ) then

      table.remove( modules.depend, m );
      
    end;
      
  end;

  for group_name, module_group in pairs( modules ) do

    local mult = 1;

    if ( need_init[ group_name ] ) then

      -- JK: twice progress to init
      mult = 2;

    end;

    total = total + #module_group * mult;

  end;

  --- Progress method ----------------------------------------------------------
  local function func_progress ( func_exec, func_continue, group_total )

    progress = progress + step;

    local function func_step ()

      if ( cursor <= group_total ) then

        func_exec();
        func_continue();

      else

        func_exec();
        cursor = 1;

        local func_next = table.remove( load_chain, 1 );
        if  ( func_next ) then func_next(); end;

      end;

    end;
    
    if ( only_depend ) then
      func_step();
    else
      local percents = math.floor( 100 * progress / total );
      private.Splash_Animate_ModuleLoad( percents, func_step );
    end;

  end;

  --- Load method --------------------------------------------------------------
  local function func_load ( module_list )

    for m, module_name in pairs( module_list ) do
  
      --local start = os.clock();

      private.Module_Load( module_name );

      --local stop = os.clock();
      --profiling.load[ module_name ] = ( stop - start ) * 1000;

    end;

  end;

  --- Init method --------------------------------------------------------------
  local function func_init ( module_list, params )

    for m = 1, #module_list do

      local module_name = module_list[ m ];

      if ( _G[ module_name ].Init ) then

        --local start = os.clock();

        _G[ module_name ].Init( params );

        --local stop = os.clock();
        --profiling.init[ module_name ] = ( stop - start ) * 1000;

      end;

    end;

    mediator.Notify( "Module_Load", module_list );
    
  end;

  --- List method to collect modules being loaded at once ----------------------
  local function func_list ( group_data, operation )

    --DbgTrace( "Cursor: "..cursor );

    local list = {};

    for m = 1, at_once[ operation ] do
      
      local module_data = group_data[ cursor ];
      local module_name = module_data;
      
      if ( type( group_data[ cursor ] ) == "table" ) then
        module_name = group_data[ cursor ].name;
      end;

      if ( module_name ) then

        table.insert( list, module_name );

        cursor = cursor + 1;

      end;

    end;

    step = #list;

    return list;

  end;

  --- Load core modules --------------------------------------------------------
  local function func_load_core ()
    
    local group_data  = modules.core;
    local module_list = func_list( group_data, "load" );

    func_progress(
      function ()

        func_load( module_list );

      end,
      func_load_core,
      #group_data
    );

  end;

  --- Init core modules --------------------------------------------------------
  local function func_init_core ()
    
    local group_data  = modules.core;
    local module_list = func_list( group_data, "init" );

    func_progress(
      function ()

        func_init( module_list );

      end,
      func_init_core,
      #group_data
    );

  end;

  --- Load game modules -------------------------------------------------------
  local function func_load_game ()

    local group_data  = modules.game;
    local module_list = func_list( group_data, "load" );

    func_progress(
      function ()

        func_load( module_list );

      end,
      func_load_game,
      #group_data
    );

  end;

  --- Init game modules --------------------------------------------------------
  local function func_init_game ()

    local group_data  = modules.game;
    local module_list = func_list( group_data, "init" );

    func_progress(
      function ()

        local params = {
          is_debug    = private.is_debug,
          is_online   = private.is_online,
          sysmsg_type = private.sysmsg_type,
          screen      =  {
            name      = "scr_loader",
            callbacks = {
              func_close = private.OnScreenClose
            } 
          }
        };

        func_init( module_list, params );

      end,
      func_init_game,
      #group_data
    );

  end;

  --- Load dependencies --------------------------------------------------------
  local function func_load_depend ()

    local group_data  = modules.depend;
    local module_list = func_list( group_data, "load" );

    func_progress(
      function ()

        func_load( module_list );

      end,
      func_load_depend,
      #group_data
    );

  end;

  --- Init dependencies --------------------------------------------------------
  local function func_init_depend ()

    local group_data  = modules.depend;
    local module_list = func_list( group_data, "init" );

    func_progress(
      function ()

        func_init( module_list );

      end,
      func_init_depend,
      #group_data
    );

  end;
  
  --- Sync features fore fresh player ------------------------------------------
  local function func_sync_features ()
    
    local feature_name = "gold_balance";
    local local_sv     = v_features.GetFeatureData( feature_name );
    local player_info  = lib_db.Player.Info();
    local ingame_time  = parts.time.watch.GetTime( "wtc_ingame_time" ) or 0;
    
    --Log.i( nil, "In-game time: "..tostring( ingame_time ).." sec" );
    
    local function after_sync ()
      local func_next = table.remove( load_chain, 1 );
      func_next();
    end;
    
    -- JK: fresh players or abtest participants
    if ( ingame_time < 1    ) or
       ( player_info.abtest ) then
    
      if ( private.is_online ) then
        
        local func_OnSynced;
        
        func_OnSynced = function ()
          ne.msg.Unsubscribe( "Event_Feature_Updated", func_OnSynced );
          
          local remote_sv = parts.features.GetFeatureData( feature_name )[ 1 ];
          
          if ( remote_sv ) then
            
            if ( not player_info.abtest ) then
              lib_db.Player.Set( "abtest", true );
            end;
            
            if ( remote_sv ~= local_sv ) then
              v_features.SaveFeatureData( feature_name, remote_sv );
              lib_db.TryReloadSettings( remote_sv );
            end;
            
          end;
          
          after_sync();
          
        end;
        
        if ( v_features.IsSynced() ) then
          func_OnSynced();
        else
          ne.msg.Subscribe(
            "Event_Feature_Updated",
            func_OnSynced
          );
        end;
        
      else
        
        local settings_variant = local_sv or 0;
        lib_db.TryReloadSettings( settings_variant );
        
        after_sync();
        
      end;
      
    -- JK: old players
    else
      
      -- JK: if coming to old save from new, clear the local storage
      if ( local_sv ) then
        v_features.RemoveFeatureData( feature_name );
        lib_db.TryReloadSettings( 0 );
      end;
      
      after_sync();
      
    end;
    
  end;

  --- Sync custom features for fresh player ------------------------------------------
  local function func_sync_features_for_fresh_player ()
    
    local ingame_time = parts.time.watch.GetTime( "wtc_ingame_time" ) or 0;
    
    local function after_sync ()
      local func_next = table.remove( load_chain, 1 );
      func_next();
    end;
    
    if ( ingame_time < 1 ) and ( private.is_online ) then
      
      local func_OnSynced;
      
      func_OnSynced = function ()
        
        ne.msg.Unsubscribe( "Event_Feature_Updated", func_OnSynced );
        
        local func_set_data_for_feature = function ( feature_name )
          if ( not parts.features.IsActive( feature_name ) ) then
            lib_db.Player.Set( feature_name.."_feature", true );
          end;
          v_features.Feature_SaveState( feature_name );
        end;
        --ES: added new features for new player if need
        local feature_list = {
          "gold_mine_fresh_player",
          "fortune_wheel_fresh_player"
        };
        
        for i = 1, #feature_list do
          func_set_data_for_feature( feature_list[ i ] );
        end;
        
        after_sync();
        
      end;
      
      if ( v_features.IsSynced() ) then
        func_OnSynced();
      else
        ne.msg.Subscribe( "Event_Feature_Updated", func_OnSynced );
      end;
      
    else
      after_sync();
    end;
    
  end;
  
  --- Finish loading -----------------------------------------------------------
  func_finish = func_finish or function ()

    private.modulesloaded = true;
    private.Trace( "Modules_Load: finished." );

    if ( private.is_debug ) then

      profiling.stop = os.clock();

      DbgTrace(
        "[ LOADER ] Modules loaded: "..total.." operations in "..
        ( profiling.stop - profiling.start ).." seconds."
      );

      --DbgTrace( profiling.load );
      --DbgTrace( profiling.init );
      
      -- JK: Here comes The Architect, inspecting each method call
      dbg_strict.StartArchitect();

    end;

    private.Splash_Animate_Last();
    private.Splash_Finish();

    -- JK: release mediator queue
    mediator.Queue_Release();

  end;

  --- Load chain ---------------------------------------------------------------
  load_chain = {
    func_load_core,
    func_load_game,
    func_init_core,
    --func_sync_features,
    func_sync_features_for_fresh_player,
    func_load_depend,
    func_init_game,
    func_init_depend,
    func_finish
  };
  
  if ( only_depend ) then
    load_chain = {
      func_load_depend,
      func_init_depend,
      func_finish
    };
  end;

  --- Fire!!! ------------------------------------------------------------------
  local func_start = table.remove( load_chain, 1 );
  
  if ( only_depend ) then
    func_start();
    return;
  end;
  
  private.Splash_Animate_ProgressBarOpen( false, func_start );
  
end;
--------------------------------------------------------------------------------
function private.Depend_IsConditionMet ( condition, module_name )

  if ( not condition or
       condition == "default" ) then
    return true;
  end;
  
  local query =
  {
    query      = "find",
    collection = "player/stages:name",
    criteria   = { like = condition }
  };
  
  local query_stage_last =
  {
    query      = "find",
    collection = "player/info"
  };
  
  local func_execute = DbExecuteQuery;
  
  if ( _G.lib_db ) then
    
    query =
    {
      query      = "find",
      group      = "player",
      collection = "stages",
      indice     = "name",
      criteria   = { like = condition }
    };
    
    query_stage_last =
    {
      query      = "find",
      group      = "player",
      collection = "info"
    };

    func_execute = lib_db.Execute;
    
  end;
  
  local condition_met = ( func_execute( query ).result or {} ).completed;
  
  if ( not condition_met ) then
    
    local player_info = func_execute( query_stage_last ).result or {};
    local stage_last  = player_info.stage_last or
                        ( player_info[ 1 ] or {} ).stage_last or "stage_1";
    
    local condition_number  = tonumber( condition:match( "%d+" ) );
    local stage_last_number = tonumber( stage_last:match( "%d+" ) );

    if ( condition_number < stage_last_number ) then
      
      return true;
      
    end;
    
  end;
  
  return condition_met;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.ShowLogoPubDev ()

  if ( private.is_ios ) then
    ObjSet( "tmr_loader_content",
    {
      time    = 1.0,
      playing = true,
      endtrig = function ()
        wiz.SplashHide();
        private.ShowLogoPubDev_1();
      end
    });
    return;
  end;

  wiz.SplashHide();
  
  if ( private.is_pc ) then
    ObjSet( private.obj.logo.pubdev, { alp = 1.0 } );
    private.KeepLogoPubDev();
    return;
  end;
  
  private.ShowLogoPubDev_1();
  
end;
--------------------------------------------------------------------------------
function private.KeepLogoPubDev ()

  private.Anim_KeepLogo( "pubdev", private.HideLogoPubDev );

end;
--------------------------------------------------------------------------------
function private.HideLogoPubDev ()

  private.Anim_HideLogo( "pubdev", private.ShowLogoPubDev_1 );

end;
--------------------------------------------------------------------------------
function private.ShowLogoPubDev_1 ()

  private.Anim_ShowLogo( "pubdev_1", private.KeepLogoPubDev_1 );
  
end;
--------------------------------------------------------------------------------
function private.KeepLogoPubDev_1 ()

  private.Anim_KeepLogo( "pubdev_1", private.HideLogoPubDev_1 );

end;
--------------------------------------------------------------------------------
function private.HideLogoPubDev_1 ()

  private.Anim_HideLogo( "pubdev_1", private.Splash_Show );

end;
--------------------------------------------------------------------------------
function private.Splash_Show ()

  local bg_height = 1136;
  local bg_scale  = math.max( 1, ( GetAppHeight() / bg_height ) );

  ObjSet(
    private.obj.splash.background,
    {
      scale_x = bg_scale,
      scale_y = bg_scale
    }
  );

  ObjSet(
    private.obj.splash.container,
    { active = true, visible = true, input = true }
  );

  AnimPlay( private.obj.splash.message.animation, "continue_static", "" );
  
  ObjAnimate(
    private.obj.splash.container, "alp", 0, 0,
    function ()
    
      private.Splash_FastLoad( function ()
      
        private.can_continue = true;
        private.Loader_TryContinue( "Splash_Show" );
      
      end );

      --[[private.Splash_Animate_Begin( function ()
      
        private.can_continue = true;
        private.Loader_TryContinue( "Splash_Show" );
      
      end );]]

    end,
    { 
      0.0, 0, 0.0, 
      0.3, 0, 1.0 
    }
  );

end;
--------------------------------------------------------------------------------
function private.Splash_FastLoad ( func_after )
  
  local volume = GetSoundVolume( "soundtrack" );
  SetSoundVolume( "soundtrack", math.min( volume, 0.5 ) );
  
  local aud_music = "assets/loader/resources/aud_loader_music";
  ne.snd.Play( aud_music, "soundtrack", 1, "", 0 );
  
  AnimPlay( private.obj.splash.planet, "ground_static", "" );
  AnimPlay( private.obj.splash.gamelogo, "logo_static", "" );
  AnimPlay( private.obj.splash.scene, "appear_static_"..private.orientation, "" );
  
  if ( func_after ) then func_after(); end;
  
end;
--------------------------------------------------------------------------------
function private.Splash_Animate_Begin ( func_after )

  local volume = GetSoundVolume( "soundtrack" );
  SetSoundVolume( "soundtrack", math.min( volume, 0.5 ) );

  local aud_music = "assets/loader/resources/aud_loader_music";
  ne.snd.Play( aud_music, "soundtrack", 1, "", 0 );

  -- FIXME no sound
  AnimPlay( private.obj.splash.planet, "ground_appear", function ()
    AnimPlay( private.obj.splash.planet, "ground_static", "" );
  end );
  
  AnimPlay( private.obj.splash.gamelogo, "logo_appear", function ()
    AnimPlay( private.obj.splash.gamelogo, "logo_static", "" );
  end );
  
  local scene_appear   = "appear_"..private.orientation;
  local scene_static   = "appear_static_"..private.orientation;

  AnimPlay( private.obj.splash.scene, scene_appear,
    function ()
      AnimPlay( private.obj.splash.scene, scene_static, "" );
      if ( func_after ) then func_after(); end;
    end
  );

end;
--------------------------------------------------------------------------------
function private.Splash_Animate_ProgressBarOpen ( need_walk, func_after )

  local walk_cycle = "walk_cycle_"..private.orientation;
  local walk_start = "walk_start_"..private.orientation;
  local appear_static = "appear_static2_"..private.orientation;
  local appear_progressbar = "appear_progressbar";
  
  local function func_progressbar_appear ( func_on_end )

    AnimPlay( private.obj.splash.progress, appear_progressbar, function ()
      
      ne.obj.Set( private.obj.splash.progress, 
      { 
        playing = false, 
        animfunc = "progressbar",
        frame = 0
      } );
      
      ( func_on_end or function () end )();
    
    end );
    
  end;
  
  local function func_prepare_walk ()
    
    ne.obj.Set(
      private.obj.splash.scene,
      {
        animfunc = walk_cycle,
        playing  = false,
        frame    = 0
      }
    );
    
    ne.obj.Set(
      private.obj.splash.planet,
      {
        animfunc = "ground_cycle",
        playing  = false,
        frame    = 0
      }
    );
    
    ( func_after or function () end )();
      
  end;
  
  if ( need_walk ) then   

    AnimPlay( private.obj.splash.scene, walk_start, func_prepare_walk );
  
  else
  
    private.pbar_frame_prev = 0;
  
    AnimPlay( private.obj.splash.scene,    appear_static, "" );
    ObjSet( private.obj.splash.gamelogo, { playing = false } );
    func_progressbar_appear( func_after );
  
  end;

end;
--------------------------------------------------------------------------------
function private.Splash_Animate_ModuleLoad ( percents, func_after )

  local pbar_frame_next = math.floor( percents * 1.25 );
  
  ne.obj.Animate(
    private.obj.splash.progress,
    "frame", 0, 0,
    func_after or "",
    {
      0.000, 0, private.pbar_frame_prev,
      0.017, 0, pbar_frame_next
    }
  );
  
  private.pbar_frame_prev = pbar_frame_next;

end;
--------------------------------------------------------------------------------
function private.Splash_Animate_Last ( func_after )

  private.walk_frame_prev = nil;
  private.pbar_frame_prev = nil;

  local snd_appear2 = "assets/loader/resources/aud_loader_game";
  ne.snd.Play( snd_appear2, "sfx", 0, "", 0 );

  local scene_appear_2 = "appear2_"..private.orientation;
  local scene_static_2 = "appear2_static_"..private.orientation;
  
  ObjSet( private.obj.splash.gamelogo, { playing = true } );

  AnimPlay( private.obj.splash.scene, scene_appear_2, 
    function ()
      AnimPlay( private.obj.splash.scene, scene_static_2, "" );
      if ( func_after ) then func_after(); end;
    end
  );

end;
--------------------------------------------------------------------------------
function private.Splash_Animate_Downloading ()

  private.Splash_Animate_InfinityCharacterWalk();
  private.Splash_Animate_InfinityGroundSpin();

end;
--------------------------------------------------------------------------------
function private.Splash_Animate_InfinityCharacterWalk ()
  
  ne.obj.Set(
    private.obj.splash.scene,
    {
      animfunc = "walk_cycle_"..private.orientation,
      playing  = true,
      frame    = 0,
      endtrig  = function ()
        private.Splash_Animate_InfinityCharacterWalk();
      end;
    }
  );

end;
--------------------------------------------------------------------------------
function private.Splash_Animate_InfinityGroundSpin ()
  
  ne.obj.Set(
    private.obj.splash.planet,
    {
      animfunc = "ground_cycle",
      playing  = true,
      frame    = 0,
      endtrig  = function ()
        private.Splash_Animate_InfinityGroundSpin();
      end;
    }
  );

end;
--------------------------------------------------------------------------------
function private.Splash_ProgressBar_SetText ( str_desc, str_count )

  ne.obj.Set(
    private.obj.splash.progress,
    { 
      tt_desc  = str_desc,
      tt_count = str_count
    } 
  );

end;
--------------------------------------------------------------------------------
function private.Splash_ProgressBar_SetBar ( percents )

  ne.obj.Set(
    private.obj.splash.progress,
    { 
      frame = math.floor( percents * 1.25 );
    } 
  );

end;
--------------------------------------------------------------------------------
function private.Splash_Finish ()

  private.Info_Show();
  
  private.PartsUnsubscribeAll();
  
  v_log.Playhaven_LogPlacement( "load_game" );
  
  v_game.Splash_Show( "obj_loader_splash" );

end;
--------------------------------------------------------------------------------
function private.Anim_ShowLogo ( target, func_next )

  ObjAnimate( "spr_loader_logo_"..target, "alp", 0, 0, func_next, 
  {
    0.0, 0, 0.0,
    0.4, 0, 1.0
  } );

end;
--------------------------------------------------------------------------------
function private.Anim_KeepLogo ( target, func_next )
  
  ObjSet( "spr_loader_logo_"..target, { input = true } );

  ObjAnimate( "spr_loader_logo_"..target, "alp", 0, 0, func_next,
  {
    0.0, 0, 1.0,
    1.3, 0, 1.0
  } );

end;
--------------------------------------------------------------------------------
function private.Anim_HideLogo ( target, func_next )

  ObjSet( "spr_loader_logo_"..target, { input = false } );

  ObjAnimate( "spr_loader_logo_"..target, "alp", 0, 0, func_next,
  {
    0.0, 0, 1.0,
    0.4, 0, 0.0
  } );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.OnScreenClose ( func_after )

  ne.snd.Stop( "assets/loader/resources/aud_loader_music", 1 );
  ne.snd.Stop( "assets/loader/resources/aud_loader_game",  1 );
  
  if ( func_after ) then func_after() end;

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.TapToContinue_Show ( func_on_click )

  ne.obj.Set(
    private.obj.splash.container,
    {
      event_mup = function ()
      
        private.TapToContinue_Hide();
        if ( func_on_click ) then func_on_click(); end;
      
      end
    }
  );
  
  AnimPlay( private.obj.splash.message.animation, "continue_diappear", function ()
    
    ne.obj.Set( private.obj.splash.message.animation,
    {
      tt_continue = "str_loader_message_taptocontinue"
    } );
  
    AnimPlay( private.obj.splash.message.animation, "continue_appear", function ()
  
      AnimPlay( private.obj.splash.message.animation, "continue_static", "" );
  
    end );
  
  end );

end;
--------------------------------------------------------------------------------
function private.TapToContinue_Hide ()

  ne.obj.Set( private.obj.splash.container, { event_mup = "" } );
  AnimPlay( private.obj.splash.message.animation, "continue_diappear", "" );

end;
--------------------------------------------------------------------------------
--function ... ()
--------------------------------------------------------------------------------
function private.Info_Show ()

  private.Info_Update();

  ObjSet( "obj_loader_info",
  {
    active = true,
    visible = true,
    alp = 0,
    pos_y = GetAppTopSafeArea()
  } );

  ObjAnimate( "obj_loader_info", "alp", 0, 0, "", 
  { 
    0.0, 0, 0.0, 
    0.3, 0, 1.0 
  } );

end;
--------------------------------------------------------------------------------
function private.Info_Update ()
  
  if ( not private.modulesloaded ) then
    return;
  end;
  
  local client_version  = ConfigGetInfo( "parts.version.currentVersion" ) or "UNKNOWN";
  local content_main_version = parts.version.GetCurrent( "content_main" ) or "UNKNOWN";
  local server_version  = parts.version.GetCurrent( "server" );
  local content_levels_version = parts.version.GetCurrent( "content_levels" ) or "UNKNOWN";
  
  if ( "" == server_version ) then

    server_version = "UNKNOWN";

  end;

  local server_host  = private.Info_GetServer( "server.host" ) or "UNKNOWN";
  local content_main_host = private.Info_GetServer( "parts.content.contentMainHost" ) or "UNKNOWN";
  local content_levels_host = private.Info_GetServer( "parts.content.contentLevelsHost" ) or "UNKNOWN";
  server_version = server_version.." ( "..server_host.." ) ";
  content_main_version = content_main_version.." ( "..content_main_host.." ) ";
  content_levels_version = content_levels_version.." ( "..content_levels_host.." ) ";
  
  ui.obj.Set(
    "txt_loader_info",
    {
      text    = "str_loader_info",
      param0  = private.GetUserId(),
      param1  = client_version,
      param2  = server_version,
      param3  = content_main_version,
      param4  = content_levels_version
    }
  );

end;
--------------------------------------------------------------------------------
function private.Info_GetServer ( param_type )

  local url_server = tostring( ConfigGetInfo( param_type ) );

  local server_list = 
  {
    { host = "http://dev1.fairysaga.elephant-games.com/", type = "EG_DEV1" },
    { host = "http://dev2.fairysaga.elephant-games.com/", type = "EG_DEV2" },
    { host = "http://dev3.fairysaga.elephant-games.com/", type = "EG_DEV3" },
    { host = "http://dev4.fairysaga.elephant-games.com/", type = "EG_DEV4" },
    { host = "http://f2p-qa.bigfishgames.com/FairySaga1/", type = "BFG_QA1" },
    { host = "http://f2p-qa.bigfishgames.com/FairySaga2/", type = "BFG_QA2" },
    { host = "http://f2p-st.bigfishgames.com/FairySaga/", type = "BFG_STAGE" },
    { host = "http://f2p.bigfishgames.com/FairySaga/", type = "BFG_PROD" },
    { host = "http://cdnf2pqa-a.akamaihd.net/FairySaga1/", type = "CDN_QA1" },
    { host = "http://cdnf2pqa-a.akamaihd.net/FairySaga2/", type = "CDN_QA2" },
    { host = "http://cdnf2pst-a.akamaihd.net/FairySaga/", type = "CDN_STAGE" },
    { host = "http://cdnf2p-a.akamaihd.net/FairySaga/", type = "CDN_PROD" },
    { host = "http://f2p-qa.bigfishgames.com/SirMatchPC1/", type = "BFG_QA1_PC" },
    { host = "http://f2p-qa.bigfishgames.com/SirMatchPC2/", type = "BFG_QA2_PC" },
    { host = "http://f2p-st.bigfishgames.com/SirMatchPC/", type = "BFG_STAGE_PC" },
    { host = "http://f2p.bigfishgames.com/SirMatchPC/", type = "BFG_PROD_PC" },
    { host = "http://cdnf2pqa-a.akamaihd.net/SirMatchPC1/", type = "CDN_QA1_PC" },
    { host = "http://cdnf2pqa-a.akamaihd.net/SirMatchPC2/", type = "CDN_QA2_PC" },
    { host = "http://cdnf2pst-a.akamaihd.net/SirMatchPC/", type = "CDN_STAGE_PC" },
    { host = "http://cdnf2p-a.akamaihd.net/SirMatchPC/", type = "CDN_PROD_PC" }
  };

  for i = 1, #server_list do

    if ( server_list[ i ].host:find( url_server, 1, true ) ) then

      return server_list[ i ].type;

    end;

  end;

  return "???";

end;
--------------------------------------------------------------------------------
function private.SetObjects ()

  local screen_w = GetAppWidth();
  local screen_h = GetAppHeight();
  
  local multiset = {};
  
  table.insert( multiset, { private.obj.logo.container, 
    {
      pos_x = 0.5 * screen_w,
      pos_y = 0.5 * screen_h
    }
  } );
  
  table.insert( multiset, { private.obj.logo.pubdev,
    {
      input     = false,
      event_mup = private.HideLogoPubDev,
      alp       = 0.0
    }
  } );
  
  table.insert( multiset, { private.obj.logo.pubdev_1,
    {
      input     = false,
      event_mup = private.HideLogoPubDev_1,
      alp       = 0.0
    }
  } );
  
  table.insert( multiset, { private.obj.splash.container,
    {
      pos_x = 0.5 * screen_w,
      pos_y = 0.5 * screen_h,
      alp   = 0.0,
      inputrect_init = true,
      inputrect_w    = screen_w,
      inputrect_h    = screen_h,
      inputrect_x    = -0.5 * screen_w,
      inputrect_y    = -0.5 * screen_h
    }
  } );
  
  table.insert( multiset, { private.obj.splash.message.animation,
    {
      tt_continue = "str_loader_message_connecting"
    }
  } );
  
  table.insert( multiset, { private.obj.splash.progress,
    {
      tt_desc  = "str_loader_message_wait",
      tt_count = ""
    }
  } );
  
  table.insert( multiset, { private.obj.message.container,
    {
      pos_x   = 0.5 * screen_w,
      pos_y   = 0.5 * screen_h,
      active  = false, 
      visible = false,
      input   = false
    } 
  } );

  ObjMultiSet( multiset );
  ObjAttach( private.obj.screen, private.obj.application );

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function private.WaitingForAnswer ( request_target, request_method, request_params, answer_handler )

  table.insert( private.request_list,
  {
    request_target = request_target,
    request_method = request_method,
    request_params = request_params,
    answer_handler = answer_handler
  } );

  local answer_message = private._GetAnswerMessageByRequest( request_target, request_method );

  MsgSubscribe( answer_message, private.OnWaitingFinished );

  private.SendRequest( request_target, request_method, request_params, 1 );

end;
--------------------------------------------------------------------------------
function private.OnWaitingFinished ( answer_message, answer_params )

  MsgUnsubscribe( answer_message, private.OnWaitingFinished );

  for h = #private.request_list, 1, -1 do

    local element = private.request_list[ h ];

    if ( answer_message == private._GetAnswerMessageByRequest( element.request_target, element.request_method ) ) then

      if ( element.answer_handler ) then

        element.answer_handler( answer_params );

      end;

      table.remove( private.request_list, h );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.SendRequest ( request_target, request_method, request_params )

  local function _FormatRequestString ( request_string )

    return string.lower( string.sub( request_string, 1, 1 ) )
      ..string.sub( request_string, 2 );

  end;
  
  NetInvokeServerMethod( _FormatRequestString( request_target ), 
                         _FormatRequestString( request_method ), 
                         request_params, 
                         1 );

end;

function private._GetAnswerMessageByRequest ( request_target, request_method )

  return "Server_"..request_target.."_on"..request_method;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Content_IsObsolete ()

  local result, value = parts.storage.data.Read( "sml/allow_sections" );

  local sections_count = private.Content_SectionCount();
  
  if ( not value        ) or 
     ( value == ""      ) or 
     ( value == "false" ) or
     ( value == "true"  ) or
     ( tonumber( value ) < sections_count ) then

    value = sections_count;
    parts.storage.data.Write( "sml/allow_sections", value );

  end;

  private.content = {
    strings = {
      progress  = ne.strings.Get( "str_loader_content_downloading" ),
      unpacking = ne.strings.Get( "str_loader_content_unpacking" )
    },
    sections = tonumber( value ),
    packets = {
      list     = {},
      total    = 0,
      done     = 0,
      applied  = {}
    },
    progress = {
      done  = 0,
      count = 0
    },
    fail = {
      name  = "",
      retry = 0
    },
    milestone = {
      value = 0,
      step  = 25
    }
  };

  private.content.packets.list  = private.Content_GetPacketDownloadList();
  private.content.packets.total = #private.content.packets.list;

  local game_version = tonumber( string.match( VersionGetComponentVersion( "game" ), "%d+.(%d+).%d+.%d+" ) );
  local content_version = tonumber( string.match( parts.version.GetSaved( "content_main" ), "%d+.(%d+).%d+.%d+" ) );

  if ( not private.is_online and
       content_version and
       content_version > 0 and
       content_version < game_version ) then

    private.Trace( "Content_IsObsolete: content_version < game_version." );

    return {
      result = true
    };

  end;

  if ( private.content.packets.total < 1 ) then

    --private.content = nil;

    return {
      result = false,
      reason = "notobsolete"
    };

  end;
  
  --- Is no free space ---------------------------------------------------------
  if ( not private.Content_HasFreeSpace( private.content.packets.list ) ) then
    
    return {
      result = false,
      reason = "nofreespace"
    };
    
  end;
  
  private.content.milestone.value = private.content.milestone.value + 
                                    private.content.milestone.step;

  --- Content is obsolete ------------------------------------------------------
  private.Trace(
    "Need to get "..private.content.packets.total.." content packets for "..
    private.content.sections.." sections."
  );

  return {
    result = true
  };

end;
--------------------------------------------------------------------------------
function private.Content_GetPacketDownloadList ()

  local packet_list = parts.content.GetPacketList();

  local need_exclude = function ( packet )

    if ( not packet.force_update ) then
      
      local packet_number = tonumber( packet.name:match( "_(%d+)" ) ) or 0;
      local is_minievent  = ( packet.name:match( "^(minievent_).*$" ) ~= nil );
      
      if ( packet_number > private.content.sections            ) or
         ( packet_number == 0 or private.content.sections == 1 ) or
         ( is_minievent                                        ) then
        
        return true;
         
      end;
      
    end;
    
    return false;
    
  end;
  
  for p = #packet_list, 1, - 1 do

    local packet = packet_list[ p ];

    if ( not packet.state                          ) or
       ( packet.state ~= "downloading"             ) or
       ( need_exclude( packet )                    ) then
       
       if ( packet.state and packet.state == "ready"  ) then
         public.Content_SetPacketState( packet.name, "ready" );
       end;
       
      table.remove( packet_list, p );

    end;
    
  end;

  return packet_list;

end;
--------------------------------------------------------------------------------
function private.Content_HasFreeSpace ()
  
  local total   = 0;
  local packets = private.Content_GetPacketDownloadList();
  
  for _, packet in pairs( packets ) do

    if ( packet.size ) then

      total = total + ( math.tointeger( packet.size ) or 0 );
      
    end;
  
  end;
  private.Trace( "[ CONTENT ] HasFreeSpace: total size packed = "..total );
  
  --total = 2 * total;
  --private.Trace( "[ CONTENT ] HasFreeSpace: total size x2 = "..total );
  
  private.free_space_needed = math.ceil( total / 1000000 );
  
  private.Trace( "[ CONTENT ] HasFreeSpace: free space needed (MB) = "..private.free_space_needed );
  
  return ne.file.HasFreeSpace( "save/", total );
  
end;
--------------------------------------------------------------------------------
function private.Content_SectionCount ()
  
  local query = 
  {
    query      = "find_all",
    collection = "player/sections",
  };

  local sections = DbExecuteQuery( query ).result;
  
  local enabled_count = 0;
    
  for s = 1, #sections do
    
    if ( sections[ s ].enabled ) then
      
      enabled_count = enabled_count + 1;
      
    end;
    
  end;
  
  private.Trace(
    "[ CONTENT ] Section_ContentAllowed: < "..enabled_count.." >"
  );
  
  return enabled_count;

end;
--------------------------------------------------------------------------------
function private.Content_GetNextPacket ()

  private.content.packets.list = private.Content_GetPacketDownloadList();

  return private.content.packets.list[ 1 ];

end;
--------------------------------------------------------------------------------
function private.Content_TryDownloadNext ()

  local packet = private.Content_GetNextPacket();
  if  ( packet ) then

    private.content.progress.done  = 0;
    private.content.progress.count = 0;
    
    local message = "private.Content_TryDownloadNext %s";
    private.Trace( message:format( packet.name ) );

    wiz.StartBackground();
    ContentStartDownloading( packet.name, private.Content_DownloadCallback );

  end;

end;

--------------------------------------------------------------------------------
function private.Content_DownloadCallback (
    name,
    status,
    total,
    progress
  )
  
  private.Trace( "private.Content_DownloadCallback: package "..name );
  private.Trace( "private.Content_DownloadCallback: status "..status );

  local message = string.format(
    "%s ( %i / %i )",
    name,
    progress,
    total
  );

  local progress_data = {
    status = status,
    overall = {
      total    = private.content.packets.total,
      progress = private.content.packets.done
    },
    packet = {
      total    = total,
      progress = progress
    }
  };

  --- Downloading --------------------------------------------------------------
  if ( status == "progress" ) then

    local timeout = 10; -- sec

    local func_DisableTimer = function ( result )
      if ( result == "success" ) then
        private.Trace( "[ INFO ][ loader ] Content download resumed." );
      else
        private.Trace( "[ ERROR ][ loader ] Content download failed." );
      end;
      ne.obj.Set(
        "tmr_loader_content_progress",
        {
          active  = false,
          time    = 0,
          playing = false,
          endtrig = ""
        }
      );
      private.content_progress_awaits = nil;
    end;

    local func_EnableTimer = function ()
      private.content_progress_awaits = true;
      local func_OnTimerEnd = function ()
        func_DisableTimer( "fail" );
        ContentStopDownloading();
        wiz.EndBackground();
        private.Content_ApplyPackets();
        private.Content_OnDownloadFailed();
      end;
      ne.obj.Set(
        "tmr_loader_content_progress",
        {
          active  = true,
          time    = timeout,
          playing = true,
          endtrig = func_OnTimerEnd
        }
      );
    end;

    if ( private.content.progress.done == progress ) then

      if ( private.content_progress_awaits ) then
        func_DisableTimer( "success" );
      end;

      private.content.progress.count = private.content.progress.count + 1;

    end;

    if ( 3 == private.content.progress.count ) then

      private.Trace( "Content FAIL on progress: "..message );
      private.Trace( "Waiting "..timeout.."s..." );

      func_EnableTimer();

      return;

    end;

    private.Trace( "Downloading "..message );

    private.content.progress.done = progress;

    private.Content_UpdateProgressbar( progress_data );

  --- Unpacking ----------------------------------------------------------------
  elseif ( status == "unpacking" ) then

    ShowLoadingWheel( true );

    private.Trace( "Unpacking "..name );

    private.size_prev = nil;

    private.Content_UpdateProgressbar( progress_data );

  --- Success ------------------------------------------------------------------
  elseif ( status == "success" ) then

    ShowLoadingWheel( false );

    private.content.packets.done   = private.content.packets.done   + 1;
    progress_data.overall.progress = progress_data.overall.progress + 1;
    
    local overall_ratio    = progress_data.overall.progress / progress_data.overall.total;
    local overall_percents = math.floor( overall_ratio * 100 );
    
    while ( overall_percents > private.content.milestone.value ) do
      private.Kontagent_LogTechnicalEvent( private.content.milestone.value.."_complete" );
      private.content.milestone.value = private.content.milestone.value + 
                                        private.content.milestone.step;
    end;

    private.Trace(
      "Packet ready: "..name.."( "..
      progress_data.overall.progress.." / "..progress_data.overall.total..
      " )"
    );

    if ( private.content.packets.total == private.content.packets.done ) then

      private.Content_UpdateProgressbar( progress_data );
      private.Content_FinishDownload();
      wiz.EndBackground();
      private.Content_ApplyPackets();

    else

      -- JK: next step is downloading
      progress_data.status = "progress";

      private.Content_UpdateProgressbar( progress_data );
      private.Content_TryDownloadNext();

    end;

  --- Fail ---------------------------------------------------------------------
  elseif ( status == "failed" ) then

    if ( private.content.fail.name ~= name ) then

      private.content.fail.name  = name;
      private.content.fail.retry = 0;

    end;

    if ( private.content.fail.retry == 3 ) then

      private.Trace( "Content FAIL on retry: "..message );

      ContentStopDownloading();
      wiz.EndBackground();
      private.Content_ApplyPackets();
      private.Content_OnDownloadFailed();

      return;

    end;

    private.content.fail.retry = private.content.fail.retry + 1;

    ObjSet(
      "tmr_loader_content",
      {
        time    = 10,
        playing = true,
        endtrig = private.Content_TryDownloadNext
      }
    );

  end;

end;
--------------------------------------------------------------------------------
function private.Content_OnDownloadFailed ()

  private.SystemDialog_Show(
    "contentfailed",
    function () RestartApplication( true ); end
  );

end;
--------------------------------------------------------------------------------
function private.Content_StartDownload ()

  private.Kontagent_LogTechnicalEvent( "begin" );
  
  private.Splash_ProgressBar_SetText(
    private.content.strings.progress,
    private.content.packets.done.." / "..private.content.packets.total
  );
  
  local function func_after ()
  
    private.Splash_Animate_Downloading();
    private.Content_TryDownloadNext();
  
  end;
  
  private.Splash_Animate_ProgressBarOpen( true, func_after );

end;
--------------------------------------------------------------------------------
function private.Content_UpdateProgressbar ( progress_data )

  local description = private.content.strings[ progress_data.status ];

  local overall = progress_data.overall;
  local packet  = progress_data.packet;
  
  local overall_ratio    = overall.progress / overall.total;
  local overall_percents = math.floor( overall_ratio * 100 );
  
  local percents = overall_percents;

  if ( progress_data.status == "progress" ) then

    if ( packet.total > 0 ) and ( packet.progress > 0 ) then

      local p_segment  = math.floor( 100 / private.content.packets.total );
      local p_ratio    = packet.progress / packet.total;
      local p_percents = math.floor( p_ratio * 100 );
      local p_addon    = math.floor( p_segment * p_ratio );
      
      percents = overall_percents + p_addon;
      
      local str_downloading = "str_loader_content_downloading";
      description = ne.strings.Get( str_downloading, p_percents.."%" );

    end;

  end;
  
  private.Splash_ProgressBar_SetText(
    description,
    overall.progress.." / "..overall.total
  );
  
  private.Splash_ProgressBar_SetBar( percents );

end;
--------------------------------------------------------------------------------
function private.Content_ApplyPackets ()
  
  local packets = parts.content.GetPacketList();

  for i = 1, #packets do
    
    if ( packets[ i ].name and
         packets[ i ].state and
         packets[ i ].state == "downloaded" and 
         packets[ i ].downloaded ) then
      private.Trace( "Content: apply packet "..packets[ i ].name );
      parts.content.ApplyPacket( packets[ i ].name );
      public.Content_SetPacketState( packets[ i ].name, "ready" ); 
    end;
    
  end;
  parts.application.Save( true );
  
end;
--------------------------------------------------------------------------------
function private.Content_FinishDownload ()

  private.Kontagent_LogTechnicalEvent( "complete" );
  
  private.Trace( "Finished downloading content." );

  ObjAnimate(
    private.obj.splash.container,
    "color_rgb", 0, 0,
    function () RestartApplication( true ); end,
    {
      0.0, 0, 1.0, 1.0, 1.0,
      0.3, 0, 0.0, 0.0, 0.0
    }
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Kontagent_LogTechnicalEvent ( event_name )

  local events_table = { 
    name = event_name,
    default1 = "technical",
    default2 = "download_1",
    default3 = "",
    params = { 
      v = 0, 
      l = 1
  } };

  events_table.params = private.Kontagent_AdditionalDetails( events_table or {} );
  
  wiz.bfg.kontagent.LogEvent( events_table );

end;
--------------------------------------------------------------------------------
function private.Kontagent_AdditionalDetails ( events_table )

  local additional_data = events_table.params;
  
  local timeplayed = function ()
    local token = "wtc_ingame_time";
    if ( TimeIsWatchExist( token ) ) then
      local time = TimeGetWatchTime( token );
      return tostring( time );
    end;
    return tostring( 0 );
  end;

  local ispaying = function ()
    if ( private.Kontagent_GetPlayerInfo( "ispaying" ) ) then
      return 1;
    end;
    return 0;
  end;

  local isfb = function ()
    if ( ( wiz.social.GetChannel() or {} ).login_channel == "Facebook" ) then
      return 1;
    end;
    return 0;
  end;

  local is_googleplay_connected = function ()
    if ( wiz.Android.GooglePlayIsConnected() ) then
      return 1;
    end;
    return 0;
  end;
  
  local isbfg = function ()
    if ( ( wiz.social.GetChannel() or {} ).login_channel == "BFG" ) then
      return 1;
    end;
    return 0;
  end;

  additional_data.appVersion = string.match( 
                        VersionGetComponentVersion( "game" ), "(%d+.%d+.%d+)" );
  additional_data.cntVersion = string.match( 
                        parts.version.GetCurrent( "content_main" ), "%d+.%d+.%d+.(%d+)" );
  additional_data.timeplayed = timeplayed();  
  additional_data.host       = private.Info_GetServer( "server.host" ) or "UNKNOWN";
  additional_data.isgc       = is_googleplay_connected();
  additional_data.isfb       = isfb();
  additional_data.ispaying   = ispaying();
  additional_data.isbfg      = isbfg();
  
  if ( private.is_PC ) then
    
    additional_data.friends   = "0";
    
    local bfginfo = wiz.bfg.lot.GetBigFishInfo();
    
    additional_data.wrappingID = bfginfo.wrappingID;
    additional_data.gmguid     = bfginfo.gmguid;
    additional_data.bfgudID    = bfginfo.bfgudID;
    additional_data.userID     = bfginfo.userID;
    
  else
  
    additional_data.friends   = nil;
      
  end;
  
  return additional_data;

end;
--------------------------------------------------------------------------------
function private.Kontagent_GetPlayerInfo ( target )

  local query =
  {
    query      = "find",
    collection = "player/info",
    result     = target
  };
  
  return DbExecuteQuery( query ).result;

end;  
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function private.SystemDialog_Show ( dialog_type, func_btn1, func_btn2, text_params_btn1, text_params_btn2 )

  local is_ingame = ( private.sysmsg_type == "ingame" ) or 
                    ( dialog_type == "conflictstorage" ) or
                    ( string.find( dialog_type, "maintenance" ) );
  
  local func_hide = function()

    private.is_on_screen_system_msg = false;

    if ( is_ingame ) then

      private.InGameMessage_Hide();

    end;

    if ( private.after_close_handler ) then

      private.after_close_handler();

    end;

  end;

  local buttons = {};
  local postfix = "";
  if ( is_ingame ) then
    postfix = "_ingame";
  end;
  if ( private.is_pc ) then
    postfix = "_pc";
  end;

  if ( func_btn1 ) then

    text_params_btn1 = text_params_btn1 or {};
    
    table.insert(
      buttons,
      {
        title  = ne.strings.Get(
          "str_systemmessage_"..dialog_type.."_button1"..postfix,
          text_params_btn1[ 1 ],
          text_params_btn1[ 2 ]
        ),
        action = function()

          func_hide();
          func_btn1();

        end,
        params = text_params_btn1
      }
    );

  end;

  if ( func_btn2 ) then
    
    text_params_btn2 = text_params_btn2 or {};

    table.insert(
      buttons,
      {
        title  = ne.strings.Get(
          "str_systemmessage_"..dialog_type.."_button2"..postfix,
          text_params_btn2[ 1 ],
          text_params_btn2[ 2 ]
        ),
        action = function()

          func_hide();
          func_btn2();

        end,
        params = text_params_btn2
      }
    );

  end;
  
  private.is_on_screen_system_msg = true;
  
  local message_param = "";
  local user_id = private.GetUserId();
  if ( user_id ~= 0 ) then
    message_param = ne.strings.Get( "str_loader_userid", user_id );
  end;
  
  if ( dialog_type == "nofreespace" ) then
    message_param = private.free_space_needed;
  end;
  
  if ( is_ingame ) then
    
    local body = "str_systemmessage_"..dialog_type.."_body";
    
    if ( private.is_pc and 
       ( dialog_type == "oldclientversion" or
         dialog_type == "newversion" ) ) then
      
       body = body..postfix;
       
    end;
    
    private.InGameMessage_Show( {
      title   = ne.strings.Get( "str_systemmessage_"..dialog_type.."_title" ),
      body    = ne.strings.Get( body, message_param ),
      buttons = buttons
    } );
    
    return;
    
  end;

  wiz.ShowSystemMessage( {
    title   = ne.strings.Get( "str_systemmessage_"..dialog_type.."_title" ),
    body    = ne.strings.Get( "str_systemmessage_"..dialog_type.."_body", message_param ),
    buttons = buttons
  } );

end;
--------------------------------------------------------------------------------
function private.InGameMessage_Show ( message )
  
  local buttons_max = 2;
  local multiset    = {};

  table.insert( multiset, {
    "obj_loader_message", { active = true, input = true, visible = true }
  } );
  
  table.insert( multiset, {
    "txt_loader_message_title", { text = message.title }
  } );

  table.insert( multiset, {
    "txt_loader_message_text", { text = message.body }
  } );
  
  for b = 1, #message.buttons do
    
    local button = message.buttons[ b ];
 
    table.insert( multiset, {
      "btn_loader_message_"..b,
      { 
        active    = true, 
        input     = true, 
        visible   = true,
        pos_x     = 320 * ( b - 1 ),
        event_mup = button.action }
    } );
    
    table.insert( multiset, {
      "anm_loader_message_"..b, { title = button.title }
    } );
    
    table.insert( multiset, {
      "anm_loader_message_"..b.."_timer", {
        active  = false,
        visible = false }
    } );

    table.insert( multiset, {
      "anm_loader_message_"..b.."_saveinfo", {
        active  = false,
        visible = false }
    } );
    
    if ( button.params[ 1 ] ) then

      table.insert( multiset, {
        "anm_loader_message_"..b.."_timer", {
          active  = true,
          visible = true,
          tt_time = button.params[ 1 ] or "" }
      } );

    end;
    
    if ( button.params[ 2 ] ) then
      
      local section = "str_systemmessage_"..( button.params[ 2 ].section 
                      or "section_1" );
      local gold  = button.params[ 2 ].gold or "0";
      
      local stage_number = button.params[ 2 ].stage_last or "stage_1";
      local stage_animfunc = "stage";
      if ( stage_number == "bonus" ) then
        stage_animfunc = "stage_bonus";
      end;
      
      ObjSet( "anm_loader_message_"..b.."_saveinfo_stage", { tt_stage = stage_number} );
      AnimPlay( "anm_loader_message_"..b.."_saveinfo_stage", stage_animfunc, "" );
      
      table.insert( multiset, {
        "anm_loader_message_"..b.."_saveinfo", {
          active  = true,
          visible = true,
          tt_section = section,
          tt_gold = gold }
      } );
      
    end;
    
  end;
  
  table.insert( multiset, {
    "obj_loader_message_buttons", { pos_x = - 160 * ( #message.buttons - 1 ) }
  } ); 
  
  if ( #message.buttons < buttons_max ) then
    
    for b = #message.buttons + 1, buttons_max do
    
      local button = message.buttons[ b ];
   
      table.insert( multiset, {
        "btn_loader_message_"..b, 
        { 
          active  = false,
          input   = false,
          visible = false }
      } );

    end;
    
  end;
  
  ObjMultiSet( multiset );
  
  --local bg_params = ObjGet( "spr_loader_logo_game" );
  --
  --ObjAnimate( "spr_loader_logo_game", "color_rgb", 0, 0, "",
  --{
  --  0.0, 0, bg_params.color_r, bg_params.color_g, bg_params.color_b,
  --  0.3, 0, 0.2, 0.2, 0.2
  --} );
    
  local bg_params = ObjGet( private.obj.splash.container );

  ObjAnimate( private.obj.splash.container, "color_rgb", 0, 0, "",
  {
    0.0, 0, bg_params.color_r, bg_params.color_g, bg_params.color_b,
    0.3, 0, 0.2, 0.2, 0.2
  } );
  
end;
--------------------------------------------------------------------------------
function private.InGameMessage_Hide ( func_end )
  
  local multiset_table = {};
  
  table.insert( multiset_table, {
    "obj_loader_message", { active = false, input = false, visible = false }
  } );

  ObjMultiSet( multiset_table );
  
  local bg_params = ObjGet( private.obj.splash.container );
  
  --ObjAnimate( "spr_loader_logo_game", "color_rgb", 0, 0, "", 
  --{
  --  0.0, 0, bg_params.color_r, bg_params.color_g, bg_params.color_b,
  --  0.3, 0, 1.0, 1.0, 1.0  
  --} );
  --
  ObjAnimate( private.obj.splash.container, "color_rgb", 0, 0, func_end, 
  {
    0.0, 0, bg_params.color_r, bg_params.color_g, bg_params.color_b,
    0.3, 0, 1.0, 1.0, 1.0  
  } );
  
end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function private.Trace ( text )

  local allow_trace = false;

  if ( allow_trace ) then

    DbgTrace( "[ LOADER ] "..text );

  end;

end;
--------------------------------------------------------------------------------
